#ifndef ARRAY_H
#define ARRAY_H "$Header:$"
//-----------------------------------------------------------------------
//    $Id: table.h,v 1.21 2003/05/06 23:15:36 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//-----------------------------------------------------------------------
//
//modified from deal.II/base/include/base/table.h 
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <algorithm>
#include <cstddef>
#include <iostream>

#include <MemoryDebugger.H>
#include <NumPtr.H>

//class T must have a constructor with no arguments
//compare to DEAL TableBase
template<int N,typename T> class ArrayBase {
  protected:
    NumPtr<T> val;
    int table_size[N];
    template <int, typename> friend class Array;

    int position(const int indices[N]) const {
      CHECK_TEST(indices[0]<table_size[0]);
      int s=indices[0];
      for (int n=1;n<N;n++) {
        CHECK_TEST(indices[n]<table_size[n]);
        s=s*table_size[n]+indices[n];
      }
      return s;
    }
    T& entry(const int indices[N]) { 
      return val[position(indices)]; 
    }
    const T& entry(const int indices[N]) const {
      return val[position(indices)]; 
    }
    const T* data() const { return val.getData(); }
  public:
    ArrayBase() {
      for (int i=0;i<N;i++) table_size[i]=0;
    }
    void clear() {
      if (val.getNumber()>0) {
        T default_entry;
        for (int i=0;i<val.getNumber();i++) {
          val[i]=default_entry;
        }
      }
    }
    void reinit(const int new_sizes[N]) {
      int new_size=1;
      for (int n=0;n<N;n++) {
        table_size[n]=new_sizes[n];
        new_size*=new_sizes[n];
      }
      if (new_size==0) {
        val.cleanup();
        for (int n=0;n<N;n++) table_size[n]=0;
        return;
      }
      if (val.getNumber()<new_size) {
        val.cleanup();
        val.allocate(new_size);
      }
      clear();
    }

    ArrayBase(const int sizes[N]) {
      reinit(sizes);
    }
    ArrayBase(const ArrayBase<N,T> &src) {
      reinit(src.table_size);
      if (src.val.getNumber()>0) val.copyFrom(src.val);
    }
    template <typename T2> ArrayBase(const ArrayBase<N,T2> &src) {
      reinit(src.table_size);
      for (int i=0;i<val.getNumber();i++) val[i]=src.val[i];
    }
    virtual ~ArrayBase() {}
    ArrayBase<N,T>& operator=(const ArrayBase<N,T> &m) {
      reinit(m.table_size);
      if (val.getNumber()>0) val.copyFrom(m.val);
      return *this;
    }
    template<typename T2> ArrayBase<N,T>& operator=(
    const ArrayBase<N,T2> &m) {
      reinit (m.table_size);
      for (int i=0;i<val.getNumber();i++) val[i]=m.val[i];
      return *this;
    }
    int size(int i) const { return table_size[i]; }
    bool empty() const { return val.getNumber()<=0; }
    template<typename T2> void fill(const T2 *entries) {
      for (int i=0;i<val.getNumber();i++) val[i]=entries[i];
    }
    template<typename T2> void fillConstant(const T2 &t2) {
      for (int i=0;i<val.getNumber();i++) val[i]=t2;
    }
    T& operator()(const int indices[N]) {
      return entry(indices);
    }
    const T& operator()(const int indices[N]) const {
      return entry(indices);
    }
    virtual void printOn(ostream &os) const {
      os << "ArrayBase<N,T>: table_size =";
      for (int n=0;n<N;n++) os << " " << table_size[n];
      os << endl;
      os << "\tval:" << endl;
      val.printOn(os);
    }
};

template <int N,typename T> class Array : public ArrayBase<N,T> {};

//compare to DEAL Table<1,T>
template <typename T> class Array<1,T> : public ArrayBase<1,T> {
  public:
    Array() {}
    Array(int size) {
     int sizes[1]; 
     sizes[0]=size; 
     this->reinit(sizes);
    }
    const T& operator[](int i) const {
      int indices[1]; 
      indices[0]=i; 
      return this->entry(indices);
    }
    T& operator[](int i) {
      int indices[1]; 
      indices[0]=i; 
      return this->entry(indices);
    }
    const T& operator()(int i) const {
      int indices[1]; 
      indices[0]=i; 
      return this->entry(indices);
    }
    T& operator()(int i) {
      int indices[1]; 
      indices[0]=i; 
      return this->entry(indices);
    }
    virtual void printOn(ostream &os) const {
      os << "Array<1,T>: " << endl;
      ArrayBase<1,T>::printOn(os);
    }
};

//compare to DEAL Table<2,T>
template <typename T> class Array<2,T> : public ArrayBase<2,T> {
  public:
    Array() {}
    Array(int size0,int size1) {
      int sizes[2]; 
      sizes[0]=size0; 
      sizes[1]=size1; 
      ArrayBase<2,T>::reinit(sizes);
    }
    void reinit(int size0,int size1) {
      int sizes[2]; 
      sizes[0]=size0; 
      sizes[1]=size1; 
      ArrayBase<2,T>::reinit(sizes);
    }
    const T& operator()(int index0,int index1) const {
      int indices[2]; 
      indices[0]=index0; 
      indices[1]=index1; 
      return this->entry(indices);
    }
    T& operator()(int index0,int index1) {
      int indices[2]; 
      indices[0]=index0; 
      indices[1]=index1; 
      return this->entry(indices);
    }
    virtual void printOn(ostream &os) const {
      os << "Array<2,T>:" << endl;
      ArrayBase<2,T>::printOn(os);
    }
};

#endif
