// "$Header:$"
//---------------------------  tria_boundary.cc  --------------------------
//   $Id: tria_boundary.cc,v 1.28 2003/04/30 03:40:26 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//---------------------------  tria_boundary.cc  --------------------------
//---------------------------  tria_boundary_lib.cc  ----------------------
//   $Id: tria_boundary_lib.cc,v 1.46 2003/01/08 17:58:03 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//---------------------------  tria_boundary_lib.cc  ----------------------
//
//modified from deal.II/deal.II/source/grid/tria_boundary.cc
//  and tria_boundary_lib.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Boundary.H>
#include <Tracer.H>
#include <cmath>
//In general, we have
// int_{G_x} f(x) d s_x 
// = int_{G_xi} f(mu(xi)) || (d mu/d xi)^{-T} n(xi) det d mu/d xi || d s_xi

//On lines (ie Shape1), if y(xi_0) parameterizes the boundary with
//  y(0) = x_2 and y(1) = x_3
//then
//  mu(xi) = y(xi_0) xi_1 + [ x_0 ( 1 - xi_0 ) + x_1 xi_0 ] ( 1 - xi_1 )
//so at xi_1=1 we have
//  || (d mu/d xi)^{-T} n(xi) det d mu/d xi || = || d y/d xi_0 ||

//On quads or triangles (ie Shape2 in 3D), we have
//  || (d mu/d xi)^{-T} n(xi) det d mu/d xi || 
//= || d y/d xi_0  x  d y/d xi_1 ||

//In the various Boundary classes, we have 
//  y(xi) = projectToSurface(x_2 (1-xi_0) + x_3 xi_0)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
Point<SPACEDIM> Boundary::getNewPointOn(const Shape &element,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices) const {
  Point<SPACEDIM> x=element.center(tessellation_vertices);
  projectToSurface(x);
  return x;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void Boundary::getIntermediatePointsOn(const Shape1 &element,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  element.getIntermediatePoints(tessellation_vertices,points);
  for (int i=0;i<points.getNumber();i++) projectToSurface(points[i]);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Boundary::computeInteriorLatticePoints(const Shape2 &element,
int degree,int bf,NumPtr<Point<SPACEDIM> > &points) const {
//points are computed in interior of element only:
  element.computeInteriorLatticePoints(degree,bf,points);
//we can either call element::interiorLatticePointIndices and
//  projectToSurface(points[indices[i]]), or
//  just project everything. boundary points should already be on this
//  boundary (maybe on the intersection with another boundary), so
//  projectToSurface will make no change in those.
  for (int i=0;i<points.getNumber();i++) projectToSurface(points[i]);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void Boundary::getMappedPointsOn(const Shape1 &element,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
const NumPtr<Point<1> > &canonical_points,
NumPtr<Point<SPACEDIM> > &points,NumPtr<Point<SPACEDIM> > &dpointsdxi)
const {
  dynamic_cast<const Interval*>(&element)->getMappedPoints(
    tessellation_vertices,canonical_points,points,dpointsdxi);
  int n=canonical_points.getNumber();
  for (int i=0;i<n;i++) projectToSurface(points[i],dpointsdxi[i]);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
const Boundary *boundary;
const Boundary *other_boundary;
Point<3> given_point;
bool debug_now;
extern "C" {
  void F77NAME(hybrj)(
    void (*fcn)(const int&,const double*,double*,double*,const int&,
      const int&),
    const int &n,double *x,double *fvec,double *fjac,const int &ldfjac,
    const double &xtol,const int &maxfev,double *diag,const int &mode,
    const double &factor,const int &nprint,int &info,int &nfev,
    int &njev,double *r,const int &lr,double *qtf,double *wa1,double *wa2,
    double *wa3,double *wa4);
  void hybrjFcn(const int &n,const double *x,double *fvec,
  double *fjac,const int &ldfjac,const int &iflag) {
    CHECK_SAME(n,5);
    Point<3> px(x[0],x[1],x[2]); 
    REAL lagrange(x[3]),other_lagrange(x[4]);
    if (iflag==0) { // print
      return;
    }
    Tensor<1,3> cd=boundary->constraintDerivative(px);
    Tensor<1,3> ocd=other_boundary->constraintDerivative(px);
    for (int i=0;i<3;i++) {
      fvec[i]=x[i]-given_point[i]-cd[i]*lagrange-ocd[i]*other_lagrange;
    }
    fvec[3]=boundary->constraint(px);
    fvec[4]=other_boundary->constraint(px);
    if (iflag==2) {
      Tensor<2,3> c2d=boundary->constraintSecondDerivative(px);
      Tensor<2,3> oc2d=other_boundary->constraintSecondDerivative(px);
      for (int j=0;j<3;j++) {
        int j5=j*5;
        for (int i=0;i<3;i++) {
          fjac[i+j5]=-c2d[i][j]*lagrange-oc2d[i][j]*other_lagrange;
        }
        fjac[j+j5]+=1.;
        fjac[3+j5]=cd[j];
        fjac[4+j5]=ocd[j];
      }
      int j=3;
      int j5=j*5;
      for (int i=0;i<3;i++) fjac[i+j5]=-cd[i];
      fjac[3+j5]=0.;
      fjac[4+j5]=0.;
      j=4;
      j5=j*5;
      for (int i=0;i<3;i++) fjac[i+j5]=-ocd[i];
      fjac[3+j5]=0.;
      fjac[4+j5]=0.;
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void printHybrjInfo(int info,REAL xtol,int maxfev) {
  switch (info) {
    case 0:
      cout << "\timproper input parameters" << endl;
      break;
    case 1:
      cout << "\trelative error between two consecutive iterates is "
           << "at most " << xtol << endl;
      break;
    case 2:
      cout << "\tnumber of calls to fcn with iflag = 1 has reached "
           << maxfev << endl;
      break;
    case 3:
      cout << "\txtol = " << xtol << " is too small. no further "
           << "improvement in the approximate solution x is possible" 
           << endl;
      break;
    case 4:
      cout << "\titeration is not making good progress, as measured by "
           << "the improvement from the last five jacobian evaluations" 
           << endl;
      break;
    case 5:
      cout << "\titeration is not making good progress, as measured by "
           << "the improvement from the last ten iterations" 
           << endl;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
Point<3> Boundary::getNewPointOnIntersectionWith(const Boundary *ob,
const Shape1 &line,const NumPtr<Point<3> > &vertices) const {
  boundary=this;
  other_boundary=ob;
  given_point=(vertices[line.vertexIndex(0)]
              +vertices[line.vertexIndex(1)])*0.5;

  int n=5;
  double x[5]={given_point[0],given_point[1],given_point[2],0.,0.};
  double fvec[5];
  double fjac[25];
  int ldfjac=5;
  double xtol=DBL_EPSILON*1.e4;
  int maxfev=100;
  double diag[5];
  int mode=1;
  double factor=100.;
  int nprint=-1;
  int info=0;
  int nfev=0;
  int njev=0;
  double r[15];
  int lr=15;
  double qtf[5];
  double wa1[5];
  double wa2[5];
  double wa3[5];
  double wa4[5];
  F77NAME(hybrj)(&hybrjFcn,n,x,fvec,fjac,ldfjac,xtol,maxfev,diag,
    mode,factor,nprint,info,nfev,njev,r,lr,qtf,wa1,wa2,wa3,wa4);
# ifdef DEBUG
  if (info!=1) printHybrjInfo(info,xtol,maxfev);
# endif
  Point<3> ret(x[0],x[1],x[2]); 
  debug_now=false;
  return ret;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Boundary::getIntermediatePointsOnIntersectionWith(
const Boundary *ob,const Shape1 &line,
const NumPtr<Point<3> > &vertices,
NumPtr<Point<3> > &points) const {
  boundary=this;
  other_boundary=ob;
  int npts=points.getNumber();
  given_point=vertices[line.vertexIndex(0)];
  Point<3> dx=(vertices[line.vertexIndex(1)]-given_point)
                    /static_cast<REAL>(npts+1);
  given_point+=dx;
  for (int i=0;i<npts;i++,given_point+=dx) {
    int n=5;
    double x[5]={given_point[0],given_point[1],given_point[2],0.,0.};
    double fvec[5];
    double fjac[25];
    int ldfjac=5;
    double xtol=DBL_EPSILON*1.e4;
    int maxfev=100;
    double diag[5];
    int mode=1;
    double factor=100.;
    int nprint=-1;
    int info=0;
    int nfev=0;
    int njev=0;
    double r[15];
    int lr=15;
    double qtf[5];
    double wa1[5];
    double wa2[5];
    double wa3[5];
    double wa4[5];
    F77NAME(hybrj)(&hybrjFcn,n,x,fvec,fjac,ldfjac,xtol,maxfev,diag,
      mode,factor,nprint,info,nfev,njev,r,lr,qtf,wa1,wa2,wa3,wa4);
# ifdef DEBUG
    if (info!=1) printHybrjInfo(info,xtol,maxfev);
# endif
    for (int j=0;j<3;j++) points[i][j]=x[j];
  }
}
#endif

Tensor<2,SPACEDIM> PlaneBoundary::zero_matrix(true);
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
PlaneBoundary::PlaneBoundary(const Point<1> &pid,const Point<1> &v) : 
point_inside_domain(pid) {
  if (v[0]>pid[0]) {
    normal[0]=1.;
    point_dot_normal=v[0];
  } else {
    normal[0]=-1.;
    point_dot_normal=-v[0];
  }
}
#elif (SPACEDIM==2)
PlaneBoundary::PlaneBoundary(const Point<2> &pid,const Point<2> &v0,
const Point<2> &v1) : point_inside_domain(pid) {
  Tensor<1,2> n;
  Tensor<1,2> dv(v1-v0);
  crossProduct(n,dv);
  normal=n.getValues();
  normal/=sqrt(normal.square());
  if (normal*(point_inside_domain-v0)>0.) normal=-normal;
  point_dot_normal=normal*v0;
}
#else
PlaneBoundary::PlaneBoundary(const Point<3> &pid,const Point<3> &v0,
const Point<3> &v1,const Point<3> &v2) : point_inside_domain(pid) {
  Tensor<1,3> n;
  Tensor<1,3> dv1(v1-v0),dv2(v2-v0);
  crossProduct(n,dv1,dv2);
  normal=n.getValues();
  normal/=sqrt(normal.square());
  if (normal*(point_inside_domain-v0)>0.) normal=-normal;
  point_dot_normal=normal*v0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
Point<SPACEDIM> PlaneBoundary::getNewPointOn(const Shape &element,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices) const {
  return element.center(tessellation_vertices);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void PlaneBoundary::getIntermediatePointsOn(const Shape1 &element,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  element.getIntermediatePoints(tessellation_vertices,points);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void PlaneBoundary::computeInteriorLatticePoints(const Shape2 &element,
int degree,int bf,NumPtr<Point<SPACEDIM> > &points) const {
  element.computeInteriorLatticePoints(degree,bf,points);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void PlaneBoundary::getMappedPointsOn(const Shape1 &element,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
const NumPtr<Point<1> > &canonical_points,
NumPtr<Point<SPACEDIM> > &points,NumPtr<Point<SPACEDIM> > &dpointsdxi)
const {
  dynamic_cast<const Interval*>(&element)->getMappedPoints(
    tessellation_vertices,canonical_points,points,dpointsdxi);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void PlaneBoundary::getMappedMeasureOn(const Shape1 &element,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
const NumPtr<Point<1> > &canonical_points,
NumPtr<Point<SPACEDIM> > &points,NumPtr<Point<SPACEDIM> > &normals,
NumPtr<REAL> &surface_measure) const {
  NumPtr<Point<SPACEDIM> > dpointsdxi;
  dynamic_cast<const Interval*>(&element)->getMappedPoints(
    tessellation_vertices,canonical_points,points,dpointsdxi);

  int n=canonical_points.getNumber();
  normals.allocate(n);
  normals.initialize(normal); // constraintDerivative is constant

  surface_measure.allocate(n);
  Point<SPACEDIM> dydxi=dpointsdxi[0]-normal*(normal*dpointsdxi[0]);
  surface_measure.initialize(sqrt(dydxi.square()));
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//this routine assumes that none of the faces of element lie on a
//  different Boundary
//it may be more accurate to use 
//  DoFHandler::integrateDifferenceOnMappedBoundary
//  than DoFHandler::integrateDifferenceOnBoundary when the domain
//  uses more than one Boundary
void PlaneBoundary::getMappedMeasureOn(const Shape2 &element,
const NumPtr<Point<3> > &tessellation_vertices,
const NumPtr<Point<2> > &canonical_points,
NumPtr<Point<3> > &points,NumPtr<Point<3> > &normals,
NumPtr<REAL> &surface_measure) const {
  NumPtr<Point<3> > dpointsdxi0;
  NumPtr<Point<3> > dpointsdxi1;
  element.getMappedPoints(tessellation_vertices,canonical_points,points,
    dpointsdxi0,dpointsdxi1);
  int n=canonical_points.getNumber();
  normals.allocate(n);
  normals.initialize(normal); // constraintDerivative is constant

  surface_measure.allocate(n);
  for (int i=0;i<n;i++) {
    projectToSurface(points[i],dpointsdxi0[i],dpointsdxi1[i]);
    Tensor<1,SPACEDIM> tens;
    Tensor<1,SPACEDIM> dydxi0(dpointsdxi0[i]);
    Tensor<1,SPACEDIM> dydxi1(dpointsdxi1[i]);
    crossProduct(tens,dydxi0,dydxi1);
    REAL sm;
    product(sm,tens,tens);
    surface_measure[i]=sqrt(sm);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
Point<3> PlaneBoundary::getNewPointOnIntersectionWith(
const Boundary *ob,const Shape1 &line,const NumPtr<Point<3> > &vertices)
const {
  if (ob->isPlanar()) {
    given_point=line.center(vertices);
    return given_point;
  }
  return Boundary::getNewPointOnIntersectionWith(ob,line,vertices);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void PlaneBoundary::getIntermediatePointsOnIntersectionWith(
const Boundary *ob,const Shape1 &line,
const NumPtr<Point<3> > &vertices,
NumPtr<Point<3> > &points) const {
  if (ob->isPlanar()) {
    int npts=points.getNumber();
    given_point=vertices[line.vertexIndex(0)];
    Point<3> dx=(vertices[line.vertexIndex(1)]-given_point)
               /static_cast<REAL>(npts+1);
    given_point+=dx;
    for (int i=0;i<npts;i++,given_point+=dx) {
      points[i]=given_point;
    }
  } else {
    Boundary::getIntermediatePointsOnIntersectionWith(ob,line,vertices,
      points);
  }
}
#endif

#if (SPACEDIM>1)
bool BallBoundary::identity_set=false;
Tensor<2,SPACEDIM> BallBoundary::identity_matrix(true);
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
BallBoundary::BallBoundary(const Point<SPACEDIM> &c,REAL r) : center(c),
radius(r) {
  if (!identity_set) {
    for (int i=0;i<SPACEDIM;i++) identity_matrix[i][i]=1.;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void BallBoundary::getMappedMeasureOn(const Shape1 &element,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
const NumPtr<Point<1> > &canonical_points,
NumPtr<Point<SPACEDIM> > &points,NumPtr<Point<SPACEDIM> > &normals,
NumPtr<REAL> &surface_measure) const {
  NumPtr<Point<SPACEDIM> > dpointsdxi;
  dynamic_cast<const Interval*>(&element)->getMappedPoints(
    tessellation_vertices,canonical_points,points,dpointsdxi);

  int n=canonical_points.getNumber();
  normals.allocate(n);
  surface_measure.allocate(n);

  for (int i=0;i<n;i++) {
    projectToSurface(points[i],dpointsdxi[i]);
    surface_measure[i]=sqrt(dpointsdxi[i].square());
    Point<SPACEDIM> xc=constraintDerivative(points[i]).getValues();
    normals[i]=xc/sqrt(xc.square());
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//this routine assumes that none of the faces of element lie on a
//  different Boundary
//it may be more accurate to use 
//  DoFHandler::integrateDifferenceOnMappedBoundary
//  than DoFHandler::integrateDifferenceOnBoundary when the domain
//  uses more than one Boundary
void BallBoundary::getMappedMeasureOn(const Shape2 &element,
const NumPtr<Point<3> > &tessellation_vertices,
const NumPtr<Point<2> > &canonical_points,
NumPtr<Point<3> > &points,NumPtr<Point<3> > &normals,
NumPtr<REAL> &surface_measure) const {
  NumPtr<Point<SPACEDIM> > dpointsdxi0;
  NumPtr<Point<SPACEDIM> > dpointsdxi1;
  element.getMappedPoints(tessellation_vertices,canonical_points,
    points,dpointsdxi0,dpointsdxi1);

  int n=canonical_points.getNumber();
  normals.allocate(n);
  surface_measure.allocate(n);
  for (int i=0;i<n;i++) {
    projectToSurface(points[i],dpointsdxi0[i],dpointsdxi1[i]);
    Tensor<1,3> tens;
    Tensor<1,SPACEDIM> dydxi0(dpointsdxi0[i]);
    Tensor<1,SPACEDIM> dydxi1(dpointsdxi1[i]);
    crossProduct(tens,dydxi0,dydxi1);
    REAL sm;
    product(sm,tens,tens);
    surface_measure[i]=sqrt(sm);
    Point<3> xc=constraintDerivative(points[i]).getValues();
    normals[i]=xc/sqrt(xc.square());
  }
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void CylinderBoundary::projectToSurface(Point<3> &x) const {
  Point<3> xc(x-center);
  REAL dot=axis*xc;
  Point<3> proj=xc-axis*dot;
  REAL rho=sqrt(proj.square());
  x=center+(xc*radius-axis*(dot*(radius-rho)))/rho;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void CylinderBoundary::projectToSurface(Point<3> &x,Point<3> &dxdxi)
const {
  Point<3> xc(x-center);
  REAL dot=axis*xc;
  Point<3> proj=xc-axis*dot;
  REAL rho=sqrt(proj.square());
  proj/=rho;
  dxdxi=(dxdxi-axis*(axis*dxdxi)*(1.-rho/radius)-proj*(proj*dxdxi))
       *radius/rho;
  x=center+(xc*radius-axis*(dot*(radius-rho)))/rho;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void CylinderBoundary::projectToSurface(Point<3> &x,Point<3> &dxdxi0,
Point<3> &dxdxi1) const {
  Point<3> xc(x-center);
  REAL dot=axis*xc;
  Point<3> proj=xc-axis*dot;
  REAL rho=sqrt(proj.square());
  proj/=rho;
  REAL term=radius/rho;
  REAL term1=1.-rho/radius;
  dxdxi0=(dxdxi0-axis*((axis*dxdxi0)*term1)-proj*(proj*dxdxi0))*term;
  dxdxi1=(dxdxi1-axis*((axis*dxdxi1)*term1)-proj*(proj*dxdxi1))*term;
  x=center+(xc*radius-axis*(dot*(radius-rho)))/rho;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//this routine assumes that none of the faces of element lie on a
//  different Boundary
//it may be more accurate to use 
//  DoFHandler::integrateDifferenceOnMappedBoundary
//  than DoFHandler::integrateDifferenceOnBoundary when the domain
//  uses more than one Boundary
void CylinderBoundary::getMappedMeasureOn(const Shape2 &element,
const NumPtr<Point<3> > &tessellation_vertices,
const NumPtr<Point<2> > &canonical_points,
NumPtr<Point<3> > &points,NumPtr<Point<3> > &normals,
NumPtr<REAL> &surface_measure) const {
  NumPtr<Point<3> > dpointsdxi0;
  NumPtr<Point<3> > dpointsdxi1;
  element.getMappedPoints(tessellation_vertices,canonical_points,
    points,dpointsdxi0,dpointsdxi1);

  int n=canonical_points.getNumber();
  normals.allocate(n);
  surface_measure.allocate(n);
  for (int i=0;i<n;i++) {
    projectToSurface(points[i],dpointsdxi0[i],dpointsdxi1[i]);
    Tensor<1,3> tens;
    Tensor<1,SPACEDIM> dydxi0(dpointsdxi0[i]);
    Tensor<1,SPACEDIM> dydxi1(dpointsdxi1[i]);
    crossProduct(tens,dydxi0,dydxi1);
    REAL sm;
    product(sm,tens,tens);
    surface_measure[i]=sqrt(sm);

    Point<3> xc=constraintDerivative(points[i]).getValues();
    normals[i]=xc/sqrt(xc.square());
  }
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void ConeBoundary::projectToSurface(Point<3> &x) const {
  Point<3> xc(x-center);
  REAL rho=sqrt(xc.square());
  if (rho>0.) {
    REAL delta=axis*xc;
    REAL disc=cosine_angle*sqrt((rho-delta)*(rho+delta));
    if (delta<0.) disc=-disc;
    REAL lagrange=(delta-rho*cosine_angle)*(delta+rho*cosine_angle)
                 /(sine_angle*(delta*sine_angle+disc));
    REAL norm_ymc=(delta-lagrange*sine_angle*sine_angle)/cosine_angle;
    x=center+(xc-axis*lagrange)
            *(norm_ymc/(norm_ymc-lagrange*cosine_angle));
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void ConeBoundary::projectToSurface(Point<3> &x,Point<3> &dxdxi) const {
  Point<3> xc(x-center);
  REAL delta=axis*xc;
  REAL rho=sqrt(xc.square());
  REAL disc=cosine_angle*sqrt((rho-delta)*(rho+delta));
  if (delta<0.) disc=-disc;
  REAL lagrange=(rho*cosine_angle-delta)*(rho*cosine_angle+delta)
               /(sine_angle*(delta*sine_angle+disc));
  REAL norm_ymc=(delta+lagrange*sine_angle*sine_angle)/cosine_angle;
  x=center+(xc+axis*lagrange)
          *(norm_ymc/(norm_ymc+lagrange*cosine_angle));

  Tensor<1,3> grad=constraintDerivative(x);
  Tensor<2,3> hess=constraintSecondDerivative(x);
  hess*=(-lagrange);
  for (int i=0;i<3;i++) hess[i][i]+=1.;
  Tensor<2,3> hess_inverse=invert(hess);
  Tensor<1,3> hess_inverse_times_grad;
  product(hess_inverse_times_grad,hess_inverse,grad);
  REAL g_dot_hinv_times_g;
  product(g_dot_hinv_times_g,grad,hess_inverse_times_grad);
  Tensor<1,3> tdxdxi(dxdxi);
  Tensor<1,3> ret;
  product(ret,hess_inverse,tdxdxi);
  REAL dot;
  product(dot,hess_inverse_times_grad,tdxdxi);
  hess_inverse_times_grad*=dot/g_dot_hinv_times_g;
  ret-=hess_inverse_times_grad;
  dxdxi=ret.getValues();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void ConeBoundary::projectToSurface(Point<3> &x,Point<3> &dxdxi0,
Point<3> &dxdxi1) const {
  Point<3> xc(x-center);
  REAL delta=axis*xc;
  REAL rho=sqrt(xc.square());
  REAL disc=cosine_angle*sqrt((rho-delta)*(rho+delta));
  if (delta<0.) disc=-disc;
  REAL lagrange=(rho*cosine_angle-delta)*(rho*cosine_angle+delta)
               /(sine_angle*(delta*sine_angle+disc));
  REAL norm_ymc=(delta+lagrange*sine_angle*sine_angle)/cosine_angle;
  x=center+(xc+axis*lagrange)
          *(norm_ymc/(norm_ymc+lagrange*cosine_angle));

  Tensor<1,3> grad=constraintDerivative(x);
  Tensor<2,3> hess=constraintSecondDerivative(x);
  hess*=(-lagrange);
  for (int i=0;i<3;i++) hess[i][i]+=1.;
  Tensor<2,3> hess_inverse=invert(hess);
  Tensor<1,3> hess_inverse_times_grad;
  product(hess_inverse_times_grad,hess_inverse,grad);
  REAL g_dot_hinv_times_g;
  product(g_dot_hinv_times_g,grad,hess_inverse_times_grad);

  Tensor<1,3> tdxdxi0(dxdxi0);
  Tensor<1,3> ret;
  product(ret,hess_inverse,tdxdxi0);
  REAL dot;
  product(dot,hess_inverse_times_grad,tdxdxi0);
  hess_inverse_times_grad*=dot/g_dot_hinv_times_g;
  ret-=hess_inverse_times_grad;
  dxdxi0=ret.getValues();

  Tensor<1,3> tdxdxi1(dxdxi1);
  product(ret,hess_inverse,tdxdxi1);
  product(dot,hess_inverse_times_grad,tdxdxi1);
  hess_inverse_times_grad*=dot/g_dot_hinv_times_g;
  ret-=hess_inverse_times_grad;
  dxdxi1=ret.getValues();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//this routine assumes that none of the faces of element lie on a
//  different Boundary
//it may be more accurate to use 
//  DoFHandler::integrateDifferenceOnMappedBoundary
//  than DoFHandler::integrateDifferenceOnBoundary when the domain
//  uses more than one Boundary
void ConeBoundary::getMappedMeasureOn(const Shape2 &element,
const NumPtr<Point<3> > &tessellation_vertices,
const NumPtr<Point<2> > &canonical_points,
NumPtr<Point<3> > &points,NumPtr<Point<3> > &normals,
NumPtr<REAL> &surface_measure) const {
  NumPtr<Point<SPACEDIM> > dpointsdxi0;
  NumPtr<Point<SPACEDIM> > dpointsdxi1;
  element.getMappedPoints(tessellation_vertices,canonical_points,
    points,dpointsdxi0,dpointsdxi1);

  int n=canonical_points.getNumber();
  normals.allocate(n);
  surface_measure.allocate(n);
  for (int i=0;i<n;i++) {
    projectToSurface(points[i],dpointsdxi0[i],dpointsdxi1[i]);
    Tensor<1,3> tens;
    Tensor<1,SPACEDIM> dydxi0(dpointsdxi0[i]);
    Tensor<1,SPACEDIM> dydxi1(dpointsdxi1[i]);
    crossProduct(tens,dydxi0,dydxi1);
    REAL sm;
    product(sm,tens,tens);
    surface_measure[i]=sqrt(sm);

    Point<3> xc=constraintDerivative(points[i]).getValues();
    normals[i]=xc*sqrt(xc.square());
  }
}
#endif
