#ifndef BOUNDARY_H
#define BOUNDARY_H "$Header:$"
//----------------------------  tria_boundary.h  ---------------------------
//    $Id: tria_boundary.h,v 1.45 2003/04/30 03:40:39 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  tria_boundary.h  ---------------------------
//
//modified from deal.II/deal.II/include/grid/tria_boundary.h by
//   John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Shape.H>
#include <Tensor.H>

//compare to DEAL Boundary
class Boundary {
  protected:
    virtual void projectToSurface(Point<SPACEDIM> &x) const {
      OBSOLETE("should have called derived class function");
    }
#if (SPACEDIM>1)
    virtual void projectToSurface(Point<SPACEDIM> &x,
    Point<SPACEDIM> &dpdxi) const {
      OBSOLETE("should have called derived class function");
    }
#endif
#if (SPACEDIM==3)
    virtual void projectToSurface(Point<3> &x,Point<3> &dpdxi0,
    Point<3> &dpdxi1) const {
      OBSOLETE("should have called derived class function");
    }
#endif
  public:
    virtual ~Boundary() {}
#if (SPACEDIM>1)
//  called in Tessellation::executeRefinement
    virtual Point<SPACEDIM> getNewPointOn(const Shape&,
      const NumPtr<Point<SPACEDIM> >&) const;
//  canonically equally-spaced:
//  called in mappingSupportPointsOnBoundaryOf in Mapping.C
    virtual void getIntermediatePointsOn(const Shape1&,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
# if (SPACEDIM==3)
    virtual void computeInteriorLatticePoints(const Shape2&,int,int,
      NumPtr<Point<SPACEDIM> >&) const;
# endif
//  canonical locations given:
//  called in lenoirSupportPointsOnTetrahedron in Mapping.C
    virtual void getMappedPointsOn(const Shape1&,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<1> >&,
      NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
# if (SPACEDIM==3)
    virtual Point<3> getNewPointOnIntersectionWith(const Boundary*,
      const Shape1&,const NumPtr<Point<3> >&) const;
    virtual void getIntermediatePointsOnIntersectionWith(const Boundary*,
      const Shape1&,const NumPtr<Point<3> >&,NumPtr<Point<3> >&) const;
# endif

    virtual bool isPlanar() const { return false; }
    virtual REAL constraint(const Point<SPACEDIM>&) const=0;
    virtual Tensor<1,SPACEDIM> constraintDerivative(
      const Point<SPACEDIM>&) const=0;
    virtual Tensor<2,SPACEDIM> constraintSecondDerivative(
      const Point<SPACEDIM>&) const=0;
//  called in DoFHandler::integrateDifferenceOnBoundary
    virtual void getMappedMeasureOn(const CODIMENSION1_SHAPE_CLASS&,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<SPACEDIM-1> >&,
      NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&,
      NumPtr<REAL>&) const=0;
#endif
    virtual void printOn(ostream &os) const {
      os << "Boundary:" << endl;
    }
};

//compare to DEAL StraightBoundary
class PlaneBoundary : public Boundary {
  private:
    const Point<SPACEDIM> point_inside_domain;
    static Tensor<2,SPACEDIM> zero_matrix;
    Point<SPACEDIM> normal;
    REAL point_dot_normal;

//  return point in plane nearest x
    void projectToSurface(Point<SPACEDIM> &x) const {
      x-=normal*constraint(x);
    }
#if (SPACEDIM>1)
    void projectToSurface(Point<SPACEDIM> &x,Point<SPACEDIM> &dpdxi) 
    const {
      dpdxi-=normal*(normal*dpdxi);
      x-=normal*constraint(x);
    }
#endif
#if (SPACEDIM==3)
    void projectToSurface(Point<3> &x,Point<3> &dpdxi0,Point<3> &dpdxi1)
    const {
      dpdxi0-=normal*(normal*dpdxi0);
      dpdxi1-=normal*(normal*dpdxi1);
      x-=normal*constraint(x);
    }
    void projectToIntersectionWith(const PlaneBoundary *pb,Point<3> &x)
    const {
      REAL con=constraint(x);
      REAL other_con=pb->constraint(x);
      const Point<SPACEDIM> &other_normal=pb->normal;
      Tensor<1,3> tn(normal);
      Tensor<1,3> other_tn(other_normal);
      Tensor<1,3> cp;
      crossProduct(cp,tn,other_tn);
      REAL cpn;
      product(cpn,cp,cp);
      REAL dot=normal*other_normal;
      x-=(normal*(con-dot*other_con)+other_normal*(other_con-dot*con))/cpn;
    }
#endif
  public:
#if (SPACEDIM==1)
    PlaneBoundary(const Point<1>&,const Point<1>&);
#elif (SPACEDIM==2)
    PlaneBoundary(const Point<2>&,const Point<2>&,const Point<2>&);
#else
    PlaneBoundary(const Point<3>&,const Point<3>&,const Point<3>&,
      const Point<3>&);
#endif
    virtual bool isPlanar() const { return true; }
    REAL constraint(const Point<SPACEDIM> &v) const {
      return normal*v-point_dot_normal;
    }
    Tensor<1,SPACEDIM> constraintDerivative(const Point<SPACEDIM> &v) 
    const {
      Tensor<1,SPACEDIM> n(normal);
      return n;
    }
    Tensor<2,SPACEDIM> constraintSecondDerivative(const Point<SPACEDIM>&)
    const {
      return zero_matrix;
    }
#if (SPACEDIM>1)
    virtual Point<SPACEDIM> getNewPointOn(const Shape&,
      const NumPtr<Point<SPACEDIM> >&) const;
    virtual void getIntermediatePointsOn(const Shape1&,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
# if (SPACEDIM==3)
    virtual void computeInteriorLatticePoints(const Shape2&,int,int,
      NumPtr<Point<SPACEDIM> >&) const;
# endif
    virtual void getMappedPointsOn(const Shape1&,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<1> >&,
      NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
    virtual void getMappedMeasureOn(const CODIMENSION1_SHAPE_CLASS&,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<SPACEDIM-1> >&,
      NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&,
      NumPtr<REAL>&) const;
# if (SPACEDIM==3)
    virtual Point<3> getNewPointOnIntersectionWith(const Boundary*,
      const Shape1&,const NumPtr<Point<3> >&) const;
    virtual void getIntermediatePointsOnIntersectionWith(const Boundary*,
      const Shape1&,const NumPtr<Point<3> >&,NumPtr<Point<3> >&) const;
# endif
#endif
    virtual void printOn(ostream &os) const {
      os << "PlaneBoundary: normal = " << normal 
         << "\n\tpoint_inside_domain = " << point_inside_domain 
         << "\n\tpoint_dot_normal = " << point_dot_normal << endl;
      Boundary::printOn(os);
    }
};

#if (SPACEDIM>1)
//compare to DEAL HyperBallBoundary
class BallBoundary : public Boundary {
  protected:
    static Tensor<2,SPACEDIM> identity_matrix;
    static bool identity_set;
    const Point<SPACEDIM> center;
    const REAL radius;
//  return point on ball nearest x
    void projectToSurface(Point<SPACEDIM> &x) const {
      Point<SPACEDIM> xc=x-center;
      xc*=radius/sqrt(xc.square());
      x=xc+center;
    }
    void projectToSurface(Point<SPACEDIM> &x,Point<SPACEDIM> &dpdxi) 
    const {
      Point<SPACEDIM> xc=x-center;
      REAL nxc=sqrt(xc.square());
      xc/=nxc;
      dpdxi=(dpdxi-xc*(xc*dpdxi))*(radius/nxc);
      x=center+xc*radius;
    }
#if (SPACEDIM==3)
    void projectToSurface(Point<SPACEDIM> &x,Point<SPACEDIM> &dpdxi0,
    Point<SPACEDIM> &dpdxi1) const {
      Point<SPACEDIM> xc=x-center;
      REAL nxc=sqrt(xc.square());
      xc/=nxc;
      REAL term=radius/nxc;
      dpdxi0=(dpdxi0-xc*(xc*dpdxi0))*term;
      dpdxi1=(dpdxi1-xc*(xc*dpdxi1))*term;
      x=center+xc*radius;
    }
#endif
  public:
    BallBoundary(const Point<SPACEDIM>&,REAL);
    REAL constraint(const Point<SPACEDIM> &v) const {
      return ((v-center).square()-radius*radius)*0.5;
    }
    Tensor<1,SPACEDIM> constraintDerivative(const Point<SPACEDIM> &v) 
    const {
      Tensor<1,SPACEDIM> t(v-center);
      return t;
    }
    Tensor<2,SPACEDIM> constraintSecondDerivative(const Point<SPACEDIM>&)
    const {
      return identity_matrix;
    }
    inline Point<SPACEDIM> getCenter() const { return center; }
    inline REAL getRadius() const { return radius; }

    virtual void getMappedMeasureOn(const CODIMENSION1_SHAPE_CLASS&,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<SPACEDIM-1> >&,
      NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&,
      NumPtr<REAL>&) const;
    virtual void printOn(ostream &os) const {
      os << "BallBoundary: center = " << center << endl;
      os << "\tradius = " << radius << endl;
      Boundary::printOn(os);
    }
};
#endif

#if (SPACEDIM==3)
//compare to DEAL CylinderBoundary
class CylinderBoundary : public Boundary {
  private:
    Point<3> axis;
    Point<3> center;
    REAL radius;
    Tensor<2,3> second_derivative_matrix;

    void projectToSurface(Point<3>&) const;
    void projectToSurface(Point<SPACEDIM>&,Point<SPACEDIM>&) const;
    void projectToSurface(Point<SPACEDIM>&,Point<SPACEDIM>&,
      Point<SPACEDIM>&) const;
  public:
    CylinderBoundary(const Point<3> &a,const Point<3> &c,REAL r) : 
    axis(a),center(c),radius(r) {
      REAL anorm=sqrt(axis.square());
      CHECK_POSITIVE(anorm);
      axis/=anorm;
      for (int i=0;i<3;i++) {
        second_derivative_matrix[i]=-axis*axis[i];
        second_derivative_matrix[i][i]+=1.;
      }
    }
    REAL constraint(const Point<3> &v) const {
      Point<3> vc=v-center;
      REAL dot=axis*vc;
      return ((vc-axis*dot).square()-radius*radius)*0.5;
    }
    Tensor<1,3> constraintDerivative(const Point<3> &v) const {
      Point<3> vc=v-center;
      Tensor<1,3> t(vc-axis*(axis*vc));
      return t;
    }
    Tensor<2,3> constraintSecondDerivative(const Point<3>&) const {
      return second_derivative_matrix;
    }
    inline Point<3> getAxis() const { return axis; }
    inline Point<3> getCenter() const { return center; }
    inline REAL getRadius() const { return radius; }

    virtual void getMappedMeasureOn(const Shape2&,
      const NumPtr<Point<3> >&,const NumPtr<Point<2> >&,
      NumPtr<Point<3> >&,NumPtr<Point<3> >&,NumPtr<REAL>&) const;
    virtual void printOn(ostream &os) const {
      os << "CylinderBoundary: axis = " << axis << endl;
      os << "\taxis = " << axis << endl;
      os << "\tcenter = " << center << endl;
      os << "\tradius = " << radius << endl;
      Boundary::printOn(os);
    }
};
#endif

#if (SPACEDIM==3)
//in 2D, use two PlaneBoundary's: computations are simpler
//not in DEAL
class ConeBoundary : public Boundary {
  protected:
    Point<SPACEDIM> axis;
    const Point<SPACEDIM> center;
    REAL angle;
    REAL cosine_angle;
    REAL sine_angle;
    REAL tangent_angle;
    void projectToSurface(Point<SPACEDIM>&) const;
    void projectToSurface(Point<SPACEDIM> &,Point<SPACEDIM>&) const;
    void projectToSurface(Point<SPACEDIM>&,Point<SPACEDIM>&,
      Point<SPACEDIM>&) const;
  public:
    ConeBoundary(const Point<SPACEDIM> &a,const Point<SPACEDIM> &c,
    REAL ang) : axis(a),center(c),angle(ang) {
      CHECK_TEST(angle>0. && angle<M_PI && angle!=M_PI_2);
      REAL anorm=sqrt(axis.square());
      CHECK_POSITIVE(anorm);
      axis/=anorm;
      tangent_angle=tan(angle);
      cosine_angle=cos(angle);
      sine_angle=sin(angle);
    }
    REAL constraint(const Point<SPACEDIM> &v) const {
      Point<SPACEDIM> vc=v-center;
      return sqrt(vc.square())*cosine_angle-axis*vc;
    }
    Tensor<1,SPACEDIM> constraintDerivative(const Point<SPACEDIM> &v) 
    const {
      Point<SPACEDIM> vc=v-center;
      Tensor<1,SPACEDIM> t(vc*(cosine_angle/sqrt(vc.square()))-axis);
      return t;
    }
    Tensor<2,SPACEDIM> constraintSecondDerivative(const Point<SPACEDIM> &v)
    const {
      Point<SPACEDIM> vc=v-center;
      REAL vcn=sqrt(vc.square());
      vc/=vcn;
      Tensor<2,SPACEDIM> t(false);
      for (int i=0;i<SPACEDIM;i++) {
        t[i]=vc*(-vc[i]);
        t[i][i]+=1.;
      }
      t*=cosine_angle/vcn;
      return t;
    }
    inline Point<SPACEDIM> getAxis() const { return axis; }
    inline Point<SPACEDIM> getCenter() const { return center; }
    inline REAL getAngle() const { return angle; }

    virtual void getMappedMeasureOn(const Shape2&,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<2> >&,
      NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&,
      NumPtr<REAL>&) const;
    virtual void printOn(ostream &os) const {
      os << "ConeBoundary: axis = " << axis << endl;
      os << "\tcenter = " << center << endl;
      os << "\tangle = " << angle << endl;
      Boundary::printOn(os);
    }
};
#endif

#endif
