// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <BoundaryCondition.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//face_q_points=fe_face_values.getQuadraturePoints(element,f)
//face_jxw_values=fe_face_values.getJxWValues(element,f)
//fsv=*fe_face_values.getShapeValues(element)

void BoundaryCondition::faceNeumannValues(
const NumPtr<Point<SPACEDIM> > &face_q_points,
const NumPtr<REAL> &face_jxw_values,
int offset,const Array<2,REAL> &fsv,Vector &cell_rhs) const {
  int shapes_per_cell=fsv.size(0);
  int nc=boundary_function->nComponents();
  for (int c=0;c<nc;c++) {
    if (isNatural(c)) {
#if (SPACEDIM==1)
      const Point<SPACEDIM> &qp(face_q_points[0]);
      REAL neumann_value=boundary_function->value(qp,c);
      for (int i=0;i<shapes_per_cell;i++) {
        cell_rhs[i*nc+c]+=neumann_value*fsv(i,0);
      }
#else
      for (int q=0;q<face_q_points.getNumber();q++) {
        const Point<SPACEDIM> &qp(face_q_points[q]);
        REAL neumann_value=
          boundary_function->value(qp,c)*face_jxw_values[q];
        for (int i=0;i<shapes_per_cell;i++) {
          cell_rhs[i*nc+c]+=neumann_value*fsv(i,q+offset);
        }
      }
#endif
    }
  }
}
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BoundaryCondition::faceNeumannValues(
const NumPtr<Point<SPACEDIM> > &face_q_points,
const NumPtr<REAL> &face_jxw_values,
const NumPtr<Point<SPACEDIM> > &normal_vectors,
int offset,const Array<2,Tensor<1,SPACEDIM> > &fsv,Vector &cell_rhs)
const {
  int shapes_per_cell=fsv.size(0);
  int nc=boundary_function->nComponents();
  for (int c=0;c<nc;c++) {
    if (isNatural(c)) {
      for (int q=0;q<face_q_points.getNumber();q++) {
        const Point<SPACEDIM> &qp(face_q_points[q]);
        REAL neumann_value=
          boundary_function->value(qp,c)*face_jxw_values[q];
        const Point<SPACEDIM> &nq=normal_vectors[q];
        for (int s=0;s<shapes_per_cell;s++) {
          const Tensor<1,SPACEDIM> &fsvsq=fsv(s,q+offset);
          REAL dot=0.;
          for (int d=0;d<SPACEDIM;d++) dot+=nq[d]*fsvsq[d];
          cell_rhs[s*nc+c]+=neumann_value*dot;
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BoundaryCondition::printOn(ostream &os) const {
  os << "BoundaryCondition: boundary = " << boundary
     << "\n\tboundary_function = " << boundary_function << endl;
  for (int c=0;c<is_essential.getNumber();c++) {
    cout << "\tis_essential[" << c << "] = "
         << PRINT_BOOLEAN(is_essential[c]) << endl;
  }
  if (boundary) boundary->printOn(os);
  if (boundary_function) boundary_function->printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimeDependentBoundaryCondition::faceNeumannValues(
const NumPtr<Point<SPACEDIM> > &face_q_points,REAL time,
const NumPtr<REAL> &face_jxw_values,
int offset,const Array<2,REAL> &fsv,Vector &cell_rhs) const {
  int shapes_per_cell=fsv.size(0);
  TimeDependentFunctionSystem *tdfs=getFunctionSystem();
  int nc=tdfs->nComponents();
  for (int c=0;c<nc;c++) {
    if (isNatural(c)) {
#if (SPACEDIM==1)
      const Point<SPACEDIM> &qp(face_q_points[0]);
      REAL neumann_value=tdfs->value(qp,time,c);
      for (int i=0;i<shapes_per_cell;i++) {
        cell_rhs[i*nc+c]+=neumann_value*fsv(i,0);
      }
#else
      for (int q=0;q<face_q_points.getNumber();q++) {
        const Point<SPACEDIM> &qp(face_q_points[q]);
        REAL neumann_value=
          tdfs->value(qp,time,c)*face_jxw_values[q];
        for (int i=0;i<shapes_per_cell;i++) {
          cell_rhs[i*nc+c]+=neumann_value*fsv(i,q+offset);
        }
      }
#endif
    }
  }
}
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimeDependentBoundaryCondition::faceNeumannValues(
const NumPtr<Point<SPACEDIM> > &face_q_points,REAL time,
const NumPtr<REAL> &face_jxw_values,
const NumPtr<Point<SPACEDIM> > &normal_vectors,
int offset,const Array<2,Tensor<1,SPACEDIM> > &fsv,Vector &cell_rhs)
const {
  int shapes_per_cell=fsv.size(0);
  TimeDependentFunctionSystem *tdfs=getFunctionSystem();
  int nc=tdfs->nComponents();
  for (int c=0;c<nc;c++) {
    if (isNatural(c)) {
      for (int q=0;q<face_q_points.getNumber();q++) {
        const Point<SPACEDIM> &qp(face_q_points[q]);
        REAL neumann_value=tdfs->value(qp,time,c)*face_jxw_values[q];
        const Point<SPACEDIM> &nq=normal_vectors[q];
        for (int s=0;s<shapes_per_cell;s++) {
          const Tensor<1,SPACEDIM> &fsvsq=fsv(s,q+offset);
          REAL dot=0.;
          for (int d=0;d<SPACEDIM;d++) dot+=nq[d]*fsvsq[d];
          cell_rhs[s*nc+c]+=neumann_value*dot;
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimeDependentBoundaryCondition::printOn(ostream &os) const {
  os << "TimeDependentBoundaryCondition:" << endl;
  BoundaryCondition::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void HcurlBoundaryCondition::faceNeumannValues(
const NumPtr<Point<1> > &face_q_points,
const NumPtr<REAL> &face_jxw_values,int f,int offset,
const Array<2,Tensor<1,1> > &fsv,Vector &cell_rhs) const {
  int shapes_per_cell=fsv.size(0);
  int nc=boundary_function->nComponents();
  for (int c=0;c<nc;c++) {
    if (isNatural(c)) {
      const Point<1> &qp(face_q_points[0]);
      REAL neumann_value=boundary_function->value(qp,c);
      if (f==0) neumann_value=-neumann_value;
      for (int i=0;i<shapes_per_cell;i++) {
        cell_rhs[i*nc+c]+=neumann_value*fsv(i,offset)[0];
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void HcurlBoundaryCondition::faceNeumannValues(
const NumPtr<Point<2> > &face_q_points,
const NumPtr<REAL> &face_jxw_values,
const NumPtr<Point<2> > &normal_vectors,
int offset,const Array<2,Tensor<1,2> > &fsv,Vector &cell_rhs)
const {
  int shapes_per_cell=fsv.size(0);
  int nc=boundary_function->nComponents();
  for (int c=0;c<nc;c++) {
    if (isNatural(c)) {
      for (int q=0;q<face_q_points.getNumber();q++) {
        Tensor<1,2> nq(normal_vectors[q]);
        const Point<2> &qp(face_q_points[q]);
        REAL neumann_value=boundary_function->value(qp,c);
        for (int i=0;i<shapes_per_cell;i++) {
          cell_rhs[i*nc+c]+=(fsv(i,q+offset)*nq)*neumann_value;
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlBoundaryCondition::faceNeumannValues(
const NumPtr<Point<3> > &face_q_points,
const NumPtr<REAL> &face_jxw_values,
const NumPtr<Point<3> > &normal_vectors,
int offset,const Array<2,Tensor<1,3> > &fsv,Vector &cell_rhs)
const {
  int shapes_per_cell=fsv.size(0);
  int nc=tensor_boundary_function->nComponents();
  for (int c=0;c<nc;c++) {
    if (isNatural(c)) {
      for (int q=0;q<face_q_points.getNumber();q++) {
        const Point<3> &qp(face_q_points[q]);
        Tensor<1,3> nq(normal_vectors[q]);
        Tensor<1,3> bv_x_normal;
        crossProduct(bv_x_normal,tensor_boundary_function->value(qp,c),nq);
        bv_x_normal*=face_jxw_values[q];
        for (int i=0;i<shapes_per_cell;i++) {
          cell_rhs[c+i*nc]+=bv_x_normal*fsv(i,q+offset);
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlBoundaryCondition::printOn(ostream &os) const {
  os << "HcurlBoundaryCondition: tensor_boundary_function = "
     << tensor_boundary_function << endl;
  if (tensor_boundary_function) tensor_boundary_function->printOn(cout);
  BoundaryCondition::printOn(os);
}
