#ifndef BOUNDARY_CONDITION_H
#define BOUNDARY_CONDITION_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Boundary.H>
#include <FunctionSystem.H>
#include <MaterialModel.H>

//not in DEAL
class BoundaryCondition {
  protected:
    const Boundary *boundary;
    FunctionSystem *boundary_function;
    NumPtr<bool> is_essential;
  public:
    BoundaryCondition() : boundary(0),
      boundary_function(0) {}
    BoundaryCondition(const Boundary *b) : boundary(b),
      boundary_function(0) {}
    virtual ~BoundaryCondition() {
    }

//  must call setFunctionSystem before calling the next six procedures:
    virtual bool isEssential(int c) const { return is_essential[c]; }
    virtual bool isEssential() const { 
      for (int c=0;c<is_essential.getNumber();c++) {
        if (is_essential[c]) return true;
      }
      return false;
    }
    virtual bool isNatural(int c) const { return !is_essential[c]; }
    virtual bool isNatural() const { 
      for (int c=0;c<is_essential.getNumber();c++) {
        if (!is_essential[c]) return true;
      }
      return false;
    }
//  if is_essential.getNumber()==0,
//    probably forgot to call setFunctionSystem beforehand
    virtual void makeEssential(int c=0) { is_essential[c]=true; }
    virtual void makeNatural(int c=0) { is_essential[c]=false; }

    const Boundary* getBoundary() const { return boundary; }
    virtual void setBoundary(const Boundary *b) { boundary=b; }
    virtual FunctionSystem* getFunctionSystem() const { 
      return boundary_function; 
    }
    virtual void setFunctionSystem(FunctionSystem *fs) { 
      CHECK_TEST(!boundary_function);
      boundary_function=fs; 
      is_essential.allocate(fs->nComponents());
      is_essential.initialize(true);
    }
//  for pde - div ( A grad u ) + b . grad u + c u = f
//    return int_{face bdry cap domain bdry} v_i normal . ( A grad u ) ds 
//  for pde - div ( A grad u + b grad u ) + c u = f
//    return int_{face bdry cap domain bdry} 
//               v_i normal . ( A grad u + b u ) ds 
//  normal . ( A grad u + b u ) is provided by function value
    virtual void faceNeumannValues(const NumPtr<Point<SPACEDIM> >&,
      const NumPtr<REAL>&,int,const Array<2,REAL>&,Vector&) const;
# if (SPACEDIM>1)
    virtual void faceNeumannValues(const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,const NumPtr<Point<SPACEDIM> >&,int,
    const Array<2,Tensor<1,SPACEDIM> >&,Vector&) const;
# endif
    virtual void printOn(ostream&) const;
};

//not in DEAL
class TimeDependentBoundaryCondition : public BoundaryCondition {
  private:
    virtual void faceNeumannValues(const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,int,const Array<2,REAL>&,Vector&) const {
      OBSOLETE("prohibit BoundaryCondition version of this function");
    }
# if (SPACEDIM>1)
    virtual void faceNeumannValues(const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,const NumPtr<Point<SPACEDIM> >&,int,
    const Array<2,Tensor<1,SPACEDIM> >&,Vector&) const {
      OBSOLETE("prohibit BoundaryCondition version of this function");
    }
# endif
  public:
    TimeDependentBoundaryCondition();
    TimeDependentBoundaryCondition(const Boundary *b) : 
      BoundaryCondition(b) {}
    virtual ~TimeDependentBoundaryCondition() { }

    virtual TimeDependentFunctionSystem* getFunctionSystem() const { 
      return dynamic_cast<TimeDependentFunctionSystem*>(boundary_function); 
    }
    virtual void setFunctionSystem(FunctionSystem *fs) { 
      CHECK_TEST(!boundary_function);
      boundary_function=dynamic_cast<TimeDependentFunctionSystem*>(fs); 
      CHECK_POINTER(boundary_function);
      is_essential.allocate(fs->nComponents());
      is_essential.initialize(true);
    }
//  for pde - div ( A grad u ) + b . grad u + c u = f
//    return int_{face bdry cap domain bdry} v_i normal . ( A grad u ) ds 
//  for pde - div ( A grad u + b grad u ) + c u = f
//    return int_{face bdry cap domain bdry} 
//               v_i normal . ( A grad u + b u ) ds 
//  normal . ( A grad u + b u ) is provided by function value
    virtual void faceNeumannValues(const NumPtr<Point<SPACEDIM> >&,REAL,
      const NumPtr<REAL>&,int,const Array<2,REAL>&,Vector&) const;
# if (SPACEDIM>1)
    virtual void faceNeumannValues(const NumPtr<Point<SPACEDIM> >&,REAL,
    const NumPtr<REAL>&,const NumPtr<Point<SPACEDIM> >&,int,
    const Array<2,Tensor<1,SPACEDIM> >&,Vector&) const;
# endif
    virtual void printOn(ostream&) const;
};

//1D,2D: natural bc ==> scalar boundary value, essential bc ==> tensor 
//3D: either bc ==> tensor boundary value
//not in DEAL
class HcurlBoundaryCondition : public BoundaryCondition {
  private:
//  in 1D and 2D, natural boundary values are scalar-valued
    TensorFunctionSystem *tensor_boundary_function;

#if (SPACEDIM==3)
    virtual FunctionSystem* getFunctionSystem() const { 
      OBSOLETE("inappropriate for this class");
      return BoundaryCondition::getFunctionSystem();
    }
    virtual void setFunctionSystem(FunctionSystem *fs) { 
      OBSOLETE("must provide a TensorFunctionSystem");
      BoundaryCondition::setFunctionSystem(fs);
    }
#endif
    virtual void faceNeumannValues(const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,int,const Array<2,REAL>&,Vector&) const {
      OBSOLETE("inappropriate for this class");
    }
  public:
    HcurlBoundaryCondition() : tensor_boundary_function(0) {}
    HcurlBoundaryCondition(const Boundary *b) : BoundaryCondition(b),
      tensor_boundary_function(0) {}
    virtual ~HcurlBoundaryCondition() { }

#if (SPACEDIM<=2)
    virtual void setFunctionSystem(FunctionSystem *fs) { 
      CHECK_TEST(!boundary_function);
      boundary_function=fs; 
      is_essential.allocate(fs->nComponents());
      is_essential.initialize(false);
    }
#endif
    TensorFunctionSystem* getTensorFunctionSystem() const { 
      return tensor_boundary_function; 
    }
    virtual void setFunctionSystem(TensorFunctionSystem *tfs) { 
      CHECK_TEST(!tensor_boundary_function);
      tensor_boundary_function=tfs; 
      is_essential.allocate(tfs->nComponents());
      is_essential.initialize(true);
    }
#if (SPACEDIM==1)
    virtual void faceNeumannValues(const NumPtr<Point<1> >&,
      const NumPtr<REAL>&,int,int,const Array<2,Tensor<1,1> >&,Vector&)
      const;
#else
    virtual void faceNeumannValues(const NumPtr<Point<SPACEDIM> >&,
      const NumPtr<REAL>&,const NumPtr<Point<SPACEDIM> >&,int,
      const Array<2,Tensor<1,SPACEDIM> >&,Vector&) const;
# endif
    virtual void printOn(ostream&) const;
};
#endif
