#ifndef MATERIAL_MODEL_H
#define MATERIAL_MODEL_H
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Array.H"
#include "Matrix.H"
#include "NumPtr.H"
#include "Point.H"
#include "Tensor.H"

//In DEAL, this was represented by the material_id of the TriaAccessor
class MaterialModel {
  public:
    virtual ~MaterialModel() {}
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &shape_values,
    const Array<2,Tensor<1,SPACEDIM> > &shape_gradients,
    Matrix &stiffness_matrix,Vector &rhs) const {
//    should not be called for NonlinearMaterialModel
      OBSOLETE("should have been defined in derived class");
    }
    virtual void printOn(ostream &os) const {
      cout << "\tMaterialModel:" << endl;
    }
};

//a common example
class LaplaceMaterialModel : public virtual MaterialModel {
  private:
    REAL (*rightHandSideValue)(const Point<SPACEDIM>&);
  public:
    LaplaceMaterialModel(REAL (*rhsv)(const Point<SPACEDIM>&)) :
      rightHandSideValue(rhsv) {}
    virtual ~LaplaceMaterialModel() {}
    virtual void elementStiffnessMatrixAndRHS(
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
      const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
      Matrix&,Vector&) const;
    virtual void printOn(ostream &os) const {
      cout << "\tLaplaceMaterialModel: rightHandSideValue = " 
           << rightHandSideValue << endl;
      MaterialModel::printOn(os);
    }

};

class NonlinearMaterialModel : public virtual MaterialModel {
  public:
//  this should compute the negative of the restoring force:
//    boundary forces are added in BoundaryCondition::faceNeumannValues
    virtual void elementResidual(const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,const Array<2,REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >&,
    const NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void printOn(ostream &os) const {
      cout << "\tNonlinearMaterialModel" << endl; 
      MaterialModel::printOn(os);
    }
};

//for pde in space and time, with pde coefficients varying in time
class TransientMaterialModel : public virtual MaterialModel {
  private:
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &shape_values,
    const Array<2,Tensor<1,SPACEDIM> > &shape_gradients,
    Matrix &stiffness_matrix,Vector &rhs) const {
      OBSOLETE("prohibit MaterialModel version");
    }
  public:
    virtual ~TransientMaterialModel() {}
//  for L2 projection of initial data:
    virtual void elementL2ProjectionMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,REAL>&,Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
//  for elliptic projection of initial data:
    virtual void elementEllipticProjectionMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >&,Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
//  for general timestep:
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,REAL,const NumPtr<REAL>&,
    const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
    Matrix&,Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void printOn(ostream &os) const {
      cout << "\tTransientMaterialModel:" << endl;
      MaterialModel::printOn(cout);
    }
};

//for pde in space and time, with pde coefficients independent of time:
//  avoid re-computation of mass and stiffness matrix during timestepping
class TimeIndependentTransientMaterialModel : 
public virtual TransientMaterialModel {
  private:
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &shape_values,
    const Array<2,Tensor<1,SPACEDIM> > &shape_gradients,
    Matrix &stiffness_matrix,Vector &rhs) const {
      OBSOLETE("prohibit MaterialModel version");
    }
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,REAL,const NumPtr<REAL>&,
    const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
    Matrix&,Matrix&,Vector&) const {
      OBSOLETE("prohibit TransientMaterialModel function");
    }
  public:
    virtual ~TimeIndependentTransientMaterialModel() {}
//  for general timestep:
    virtual void elementRHS(const NumPtr<Point<SPACEDIM> >&,REAL,
    const NumPtr<REAL>&,const Array<2,REAL>&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void printOn(ostream &os) const {
      cout << "\tTimeIndependentTransientMaterialModel:" << endl;
      TransientMaterialModel::printOn(cout);
    }
};

class HdivMaterialModel : public virtual MaterialModel {
  private:
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&, const NumPtr<REAL>&,
    const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
    Matrix&,Vector&) const {
//    should not be called for HdivMaterialModel
      OBSOLETE("prohibit MaterialModel version of this function");
    }
  public:
    virtual ~HdivMaterialModel() {}
    virtual bool requiresDerivative() const { return false; }
//  Matrix and Vector are indexed by dofs on element
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&, const NumPtr<REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >*,const Array<2,REAL>*,
    const Array<2,Tensor<2,SPACEDIM> >*,const Array<2,REAL>*,
    Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void printOn(ostream &os) const {
      cout << "\tHdivMaterialModel:" << endl;
      MaterialModel::printOn(os);
    }
};

class HcurlMaterialModel : public virtual MaterialModel {
  private:
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&, const NumPtr<REAL>&,
    const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
    Matrix&,Vector&) const {
      OBSOLETE("prohibit MaterialModel version of this function");
    }
  public:
    virtual ~HcurlMaterialModel() {}
//  test if material model requires more than just the curl
    virtual bool requiresDerivative() const { return false; }
#if (SPACEDIM<=2)
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,const Array<2,Tensor<1,SPACEDIM> >*,
    const Array<2,REAL>*,const Array<2,REAL>*,Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
#else
    virtual void elementStiffnessMatrixAndRHS(const NumPtr<Point<3> >&,
    const NumPtr<REAL>&,const Array<2,Tensor<1,3> >*,
    const Array<2,Tensor<1,3> >*,const Array<2,Tensor<1,3> >*,
    Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
#endif
    virtual void printOn(ostream &os) const {
      cout << "\tHcurlMaterialModel:" << endl;
      MaterialModel::printOn(os);
    }
};

class HcurlTransientMaterialModel : public TransientMaterialModel {
//public HcurlMaterialModel 
  private:
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
    Matrix&,Vector&) const {
      OBSOLETE("prohibit MaterialModel version of this function");
    }
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,REAL,const NumPtr<REAL>&,
    const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
    Matrix&,Matrix&,Vector&) const {
      OBSOLETE("prohibit TransientMaterialModel version of this function");
    }
    virtual void elementL2ProjectionMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,REAL>&,Matrix&,Vector&) const {
      OBSOLETE("prohibit TransientMaterialModel version of this function");
    }
/*
#if (SPACEDIM<=2)
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >*,const Array<2,REAL>*,
    Matrix&,Vector&) const {
      OBSOLETE("prohibit HcurlMaterialModel version of this function");
    }
#else
    virtual void elementStiffnessMatrixAndRHS(const NumPtr<Point<3> >&,
    const NumPtr<REAL>&,const Array<2,Tensor<1,3> >*,
    const Array<2,Tensor<1,3> >*,Matrix&,Vector&) const {
      OBSOLETE("prohibit HcurlMaterialModel version of this function");
    }
#endif
*/
/*
#if (SPACEDIM==2)
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >*,
    const Array<2,Tensor<2,SPACEDIM> >*,const Array<2,REAL>*,
    Matrix&,Vector&) const {
      OBSOLETE("does not apply to this class");
    }
#endif
*/
  public:
    virtual ~HcurlTransientMaterialModel() {}
    virtual bool requiresDerivative() const { return false; }
    virtual void elementL2ProjectionMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,
    const Array<2,Tensor<1,SPACEDIM> > *shape_values,
#if (SPACEDIM<=2)
    const Array<2,REAL> *multiplier_shape_values,
#else
    const Array<2,Tensor<1,3> > *multiplier_shape_values,
#endif
    Matrix &mass_matrix,Vector &rhs) const {
      OBSOLETE("should have been defined in derived class");
    }
#if (SPACEDIM!=3)
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >*,
    const Array<2,Tensor<2,SPACEDIM> >*,const Array<2,REAL>*,
    Matrix&,Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
#else
    virtual void elementStiffnessMatrixAndRHS(const NumPtr<Point<3> >&,
    const NumPtr<REAL>&,const Array<2,Tensor<1,3> >*,
    const Array<2,Tensor<1,3> >*,const Array<2,Tensor<1,3> >*,
    Matrix&,Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
#endif
    virtual void printOn(ostream &os) const {
      cout << "\tHcurlTransientMaterialModel:" << endl;
      TransientMaterialModel::printOn(cout);
//    HcurlMaterialModel::printOn(cout);
    }
};

#endif
