#ifndef _VECTOR_H_
#define _VECTOR_H_ "$Header:$"
//      LAPACK++ (V. 1.1)
//      (C) 1992-1996 All Rights Reserved.
//
//      Lapack++ "Shared" Vector Double Class
//
//----------------------------  vector.h  ---------------------------
//    $Id: vector.h,v 1.42 2003/05/31 20:43:24 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  vector.h  ---------------------------
//
//modified from LAPACK++ file vd.h and deal.II/lac/include/lac/vector.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>
#include <assert.h>
#include "Errors.H"
#include "MemoryDebugger.H"
#include "Tracer.H"

//this class used to be templated; we only used double for the template
//not in DEAL
class LAVector {
  protected:
#ifdef DEBUG
    static REAL undefined_;
#endif
    static REAL zero_,one_;
    REAL *data;
    int sz;
    bool owns_data;
  public:                                                            
    virtual void fillWith(REAL);
    void copy(const LAVector&); // calls memcpy

    LAVector() : data(0),sz(0),owns_data(true) {;}
    LAVector(int n) : data(0),sz(n),owns_data(true) { 
      if (sz>0) {
        data=OPERATOR_NEW_BRACKET(REAL,sz);
#ifdef DEBUG
        fillWith(undefined_);
#endif
      }
    }
    LAVector(int n,REAL scalar) : data(0),sz(n),owns_data(true) { 
      if (sz>0) {
        data=OPERATOR_NEW_BRACKET(REAL,n); 
        fillWith(scalar); 
      }
    }
    LAVector(const LAVector &m) : data(0),sz(0),owns_data(true) { 
      copy(m); 
    }
    LAVector(REAL *d,int s) : data(d),sz(s),owns_data(false) {}
    virtual ~LAVector() { 
      if (owns_data && data!=0) delete data; data=0; 
    }

    int resize(int n) {
      if (n==sz) return sz;
      CHECK_TEST(owns_data);
      if (data!=0) delete data; 
      sz=n; data=0;
      if (sz>0) {
        data=OPERATOR_NEW_BRACKET(REAL,sz);
        owns_data=true;
      }
      return sz;
    }

    REAL* addr() const { return data; }
    REAL* addr(int i) { 
      CHECK_BOUNDS(i,0,sz)
      return data+i; 
    }
    const REAL* addr(int i) const { 
      CHECK_BOUNDS(i,0,sz)
      return data+i; 
    }
    REAL& operator[](int i) { 
      CHECK_BOUNDS(i,0,sz)
      return data[i]; 
    }
    REAL operator[](int i) const { 
      CHECK_BOUNDS(i,0,sz)
      return data[i]; 
    }
    int size() const { return sz; }
    bool ownsData() const { return owns_data; }

    virtual LAVector& operator=(const LAVector &m) { 
      copy(m); return *this; 
    }
    LAVector& operator=(REAL s) { fillWith(s); return *this; } 
    LAVector& operator+=(const LAVector&); // calls lapack daxpy
    LAVector& operator-=(const LAVector&); // call lapack daxpy
    LAVector& operator*=(REAL); // calls lapack dscal
    LAVector& operator/=(REAL); // calls lapack dscal
    LAVector* operator+(const LAVector &B) const {
      LAVector *v=OPERATOR_NEW LAVector(*this); *v += B; return v;
    }
    LAVector* operator-(const LAVector &B) const {
      LAVector *v=OPERATOR_NEW LAVector(*this); *v -= B; return v;
    }
    LAVector* operator*(REAL d) const {
      LAVector *v=OPERATOR_NEW LAVector(*this); *v *= d; return v;
    }
    LAVector* operator/(REAL d) const {
      LAVector *v=OPERATOR_NEW LAVector(*this); *v /= d; return v;
    }
    REAL l2Norm() const; // calls lapack dnrm2

//  this function allows operator<< to access class members 
//  without being a friend
    virtual void printOn(ostream&) const;
};                                                                     
inline ostream& operator<<(ostream &s,const LAVector &m) {
  m.printOn(s); return s;
}

//compare to DEAL Vector, which does not use LaPack
class Vector {
  protected:
    LAVector d;
  public:
    Vector() { }
    explicit Vector(int n) : d(n) { }
    Vector(int n,REAL scalar) : d(n,scalar) { } 
    Vector(const Vector &m) : d(m.d) { }
    Vector(REAL *r,int s) : d(r,s) {}
    virtual ~Vector() { d.LAVector::~LAVector(); }
    Vector* clone() const { return OPERATOR_NEW Vector(d.size()); }
    Vector* cloneEmpty() const { return OPERATOR_NEW Vector; }
    Vector* subVector(int offset,int sz) {
      CHECK_TEST(offset+sz<=d.size());
      return OPERATOR_NEW Vector(d.addr()+offset,sz);
    }
    int size() const { return d.size(); }
    bool ownsData() const { return d.ownsData(); }
    void reinit(int n,bool fast=false) {
      d.resize(n);
      if (!fast) d.fillWith(0.);
    }
    REAL* addr() { return d.addr(); }
    const REAL* addr() const { return d.addr(); }
    REAL& operator[](int i) { return d[i]; }
    REAL operator[](int i) const { return d[i]; }
    Vector& operator=(const Vector &v) {
      CHECK_SAME(d.size(),v.d.size());
      d.copy(v.d);
      return *this;
    }
    Vector& operator=(REAL s) { d=s; return *this; }
    Vector& operator+=(const Vector &v) { 
      d+=v.d;
      return *this;
    }
    Vector& operator-=(const Vector &v) { 
      d-=v.d;
      return *this;
    }
    void scale(REAL a) { d*=a; }
    void equ(REAL,const Vector&); // this=r*vector; calls lapack
    REAL operator*(const Vector&) const; // dot product; call lapack
    REAL linftyNorm() const;
    void add(REAL,const Vector&); //this+=r*vector; call lapack
    void swapWith(Vector&);
    REAL l2Norm() const { return d.l2Norm(); }
    bool allZero() const { return linftyNorm()<=0.; }
    void printOn(ostream &os) const {
      os << "Vector:" << endl;
      d.printOn(os);
    }
};
#endif 
