// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <cmath>
#include <iostream>
#include <math.h> // for HUGE_VAL,M_PI
#include <stdlib.h>

#include "Debug.H"
#include "GUIEnumInputParameter.H"
#include "GUIInputs.H"
#include "MemoryDebugger.H"
#include "Polynomial.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "TimedObject.H"
#include "Types.H"
#include "XColormap.H"
#include "XYGraphTool.H"

#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

enum POLYNOMIAL{MONOMIAL,
  C0LAGRANGE,C1LAGRANGE,C2LAGRANGE,LEGENDRE,
  HIERARCHICAL,HIERARCHICAL_SIDE};
POLYNOMIAL polynomial=HIERARCHICAL;
int ipolynomial=polynomial;
const char *polynomial_name[7]=
  {"Monomial","C0Lagrange","C1Lagrange","C2Lagrange","Legendre",
   "Hierarchical","HierarchicalSide"};

GUI_INPUT_PARAMETER_LIST_TYPE *main_list=0;
BOOLEAN skip_gui=FALSE;
char *display_name=0;
REAL winsize=0.5;

INTEGER order=1;
//INTEGER continuity=0;

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void processCommandLine(int argc,char *argv[],char *display_name,
BOOLEAN &skip_gui,ifstream &in_file) {
//TRACER_CALL(t,"processCommandLine");
  if (argc<2) skip_gui=FALSE;
  else {
    in_file.open(argv[1],ios::in);
    CHECK_TEST(!in_file.fail());
    if (in_file) {
      INTEGER i=2;
      while (i<argc) {
        if (strcmp(argv[i],"-d")==0) {
          display_name=OPERATOR_NEW_BRACKET(char,strlen(argv[++i])+1);
          strcpy(display_name,argv[i]);
        } else {
          cout << " >>>> error...invalid command line option:"
               << argv[i] << endl;
          abort();
        }
        i++;
      }
    } else {
      cerr << "\ncannot open " << argv[1] << " for input" << endl;
      exit(1);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void readMainInput(ifstream &in_file,InputParameterList *main_list,
BOOLEAN &skip_gui) {
//TRACER_CALL(t,"readMainInput");
  BOOLEAN found_main=FALSE;
  char name[LENGTH_NAME], comment[LENGTH_COMMENT];

  in_file.clear(ios::goodbit);
  in_file.seekg(0,ios::beg);
  while ( in_file >> setw(LENGTH_NAME) >> name ) {
    if ( strcmp(name,"Main") == 0 ) {
      in_file.getline( comment, LENGTH_COMMENT);
      found_main=TRUE;
      while ( in_file >> setw(LENGTH_NAME) >> name ) {
#ifdef DEBUG
//      cout << "\tname = " << name << endl;
#endif
        if ( strcmp(name,"end") == 0 ) break;
        else if (strcmp(name,"skip_gui") == 0) {
          int iskip_gui; // type bool not read correctly, so use int
          in_file >> iskip_gui;
          skip_gui= iskip_gui!=0;
        }
        else main_list->formattedRead(in_file,name);
        in_file.getline( comment, LENGTH_COMMENT);
#ifdef DEBUG
//      cout << "\tcomment = " << comment << endl;
#endif
      }
    } else in_file.getline(comment,LENGTH_COMMENT);
  }
#ifdef DEBUG
//main_list->printOn(cout);
#endif
  if ( !found_main ) skip_gui=FALSE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template double __cmath_power<double>(double,unsigned);
template class InputParameter<BOOLEAN>;
template class InputParameter<INTEGER>;
template class InputParameter<REAL>;
#include "NumPtr.C"
INSTANTIATE_NUMPTR(BOOLEAN)
INSTANTIATE_NUMPTR(NumPtr<REAL>)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void makeMainList(GUI_INPUT_PARAMETER_LIST_TYPE *&main_list) {
//TRACER_CALL(t,"makeMainList");
  main_list=OPERATOR_NEW GUI_INPUT_PARAMETER_LIST_TYPE("Main");

  { const char *group="Polynomial Parameters";
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      ipolynomial,"polynomial",polynomial_name[0],polynomial_name[1],
      polynomial_name[2],polynomial_name[3],polynomial_name[4],
      polynomial_name[5],polynomial_name[6]));
    main_list->append(OPERATOR_NEW GUIInputParameter<INTEGER>(
      order,"order",1,32,group));
//  main_list->append(OPERATOR_NEW GUIInputParameter<INTEGER>(
//    continuity,"continuity",0,2,group));
  }

  { const char *group="Graphics";
    main_list->append(OPERATOR_NEW GUIInputParameter<REAL>(winsize,
      "winsize",0.,1.,group));
  }
#ifdef DEBUG
//main_list->printOn(cout);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkMainInput() {
//TRACER_CALL(t,"checkMainInput");
//order=min(order,2);
  polynomial=static_cast<POLYNOMIAL>(ipolynomial);
//if (polynomial==LAGRANGE) {
//  order=min(order,6);
//  continuity=min(continuity,(order-1)/2);
//}
  if (polynomial==C1LAGRANGE) {
    order=max(order,3);
    int n=(order-1)/2;
    order=2*n+1;
  }
  if (polynomial==C2LAGRANGE) {
    order=max(order,5);
    int n=(order-2)/3;
    order=3*n+2;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void richardsonSlopes(int nmax,Polynomial *poly,REAL x,
const NumPtr<REAL> &f,NumPtr<REAL> &fp) {
  NumPtr<NumPtr<REAL> > table(nmax);
  for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
  REAL h=0.1;
  for (int n=1;n<=nmax;n++,h*=0.5) {
    NumPtr<REAL> fnew(f.getNumber());
    poly->values(x+h,fnew);
    for (int i=0;i<f.getNumber();i++) {
      table[nmax-n][i]=(fnew[i]-f[i])/h;
    }
    REAL p=2.;
    for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
      for (int i=0;i<f.getNumber();i++) {
        table[m][i]=table[m-1][i]+(table[m-1][i]-table[m][i])/(p-1.);
      }
    }
  }
  fp.allocate(f.getNumber());
  for (int i=0;i<f.getNumber();i++) {
    fp[i]=table[nmax-1][i];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runPolynomial() {
//TRACER_CALL(t,"runPolynomial");
  Polynomial *lp=0;
  switch (polynomial) {
    case MONOMIAL:
      lp=OPERATOR_NEW Monomial(order);
      break;
    case C0LAGRANGE:
      lp=OPERATOR_NEW C0LagrangePolynomial(order);
      break;
    case C1LAGRANGE:
      lp=OPERATOR_NEW C1LagrangePolynomial(order);
      break;
    case C2LAGRANGE:
      lp=OPERATOR_NEW C2LagrangePolynomial(order);
      break;
    case LEGENDRE:
      lp=OPERATOR_NEW LegendrePolynomial();
      break;
    case HIERARCHICAL:
      lp=OPERATOR_NEW HierarchicalPolynomial();
      break;
    case HIERARCHICAL_SIDE:
      lp=OPERATOR_NEW HierarchicalSidePolynomial();
      break;
    default:
      OBSOLETE("not programmed");
  }
#ifdef DEBUG
//cout << "\torder = " << order << endl;
#endif
  REAL xmin=0.;
  REAL xmax=1.;
  REAL ymin=0.;
  REAL ymax=1.;
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool gt("Polynomial","x","basis polynomial",
    xmin,xmax,ymin,ymax,&cmap,0,winsize);
  gt.setbgColor("white");
  REAL dx=gt.getPixelWidth();

#ifdef DEBUG
//cout << "\tvalues" << endl;
#endif
  NumPtr<REAL> v(order+1);
  for (REAL x=0.;x<=1.;x+=dx) {
    lp->values(x,v);
    for (int n=0;n<=order;n++) {
      ymin=min(ymin,v[n]);
      ymax=max(ymax,v[n]);
    }
  }
  gt.rescale(xmin,xmax,ymin,ymax);
  gt.newPage();

  gt.setfgColor("black");
  gt.drawAxes();
  for (int n=0;n<=order;n++) {
#ifdef DEBUG
//  cout << "\tn = " << n << endl;
#endif
    gt.setfgColor("black");
    lp->values(0.,v);
    gt.movePen(0.,v[n]);
    for (REAL x=dx;x<=1.;x+=dx) {
      lp->values(x,v);
      gt.drawLine(x,v[n]);
    }
    gt.setfgColor(n+1,order+1);
    NumPtr<REAL> vd(1);
    lp->derivatives(0.,n,vd);
    gt.movePen(0.,vd[0]);
    for (REAL x=dx;x<=1.;x+=dx) {
      lp->derivatives(x,n,vd);
      gt.drawLine(x,vd[0]);
    }
  }
  gt.flush();
  {
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }

#ifdef DEBUG
//cout << "\tslope" << endl;
#endif
  NumPtr<REAL> s(order+1);
  REAL x=drand48();
  x=drand48();
  lp->values(x,v);
  NumPtr<REAL> vv(order+1);
  REAL eps=sqrt(DBL_EPSILON);
  lp->values(x+eps,vv);
  lp->slopes(x,s);
  NumPtr<REAL> deriv;
  richardsonSlopes(order,lp,x,v,deriv);
  for (int n=0;n<=order;n++) {
    REAL exact=s[n];
//  REAL deriv=(vv[n]-v[n])/eps;
    REAL error=abs(deriv[n]-exact);
    if (abs(exact)>0.) error/=abs(exact);
    if (error>65536.*DBL_EPSILON) {
      cout << "\tn = " << n << " x = " << x << " deriv = " << deriv[n]
           << " " << exact << " error = " << error << " " 
           << error/DBL_EPSILON << endl;
    }
  }

  for (REAL x=0.;x<=1.;x+=dx) {
    lp->slopes(x,s);
    for (int n=0;n<=order;n++) {
      ymin=min(ymin,s[n]);
      ymax=max(ymax,s[n]);
    }
  }
  gt.rescale(xmin,xmax,ymin,ymax);
  gt.newPage();

  gt.setfgColor("black");
  gt.drawAxes();
  for (int n=0;n<=order;n++) {
#ifdef DEBUG
//  cout << "\tn = " << n << endl;
//  if (continuity==0) {
//    cout << lp->slope(static_cast<REAL>(n)/static_cast<REAL>(order),n)
//         << endl;
//  }
#endif
    gt.setfgColor("black");
    lp->slopes(0.,s);
    gt.movePen(0.,s[n]);
    for (REAL x=dx;x<=1.;x+=dx) {
      lp->slopes(x,s);
      gt.drawLine(x,s[n]);
    }
    gt.setfgColor(n+1,order+1);
    NumPtr<REAL> vd(2);
    lp->derivatives(0.,n,vd);
    gt.movePen(0.,vd[1]);
    for (REAL x=dx;x<=1.;x+=dx) {
      lp->derivatives(x,n,vd);
      gt.drawLine(x,vd[1]);
    }
  }
  gt.flush();
  {
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }

  if (order>=2) {
#ifdef DEBUG
//  cout << "\tslope2" << endl;
#endif
    NumPtr<REAL> s2(order+1);
    REAL x=drand48();
    lp->slopes(x,s);
    NumPtr<REAL> ss(order+1);
    REAL eps=sqrt(DBL_EPSILON);
    lp->slopes(x+eps,ss);
    lp->slope2s(x,s2);
    for (int n=0;n<=order;n++) {
      REAL exact=s2[n];
      REAL deriv=(ss[n]-s[n])/eps;
      REAL error=abs(deriv-exact);
      if (abs(exact)>0.) error/=abs(exact);
      if (error>1024.*eps) {
        cout << "\tn = " << n << " x = " << x << " deriv = " << deriv
             << " " << exact << " error = " << error << " " 
             << error/eps << endl;
      }
    }

    for (REAL x=0.;x<=1.;x+=dx) {
      lp->slope2s(x,s2);
      for (int n=0;n<=order;n++) {
        ymin=min(ymin,s2[n]);
        ymax=max(ymax,s2[n]);
      }
    }
    gt.rescale(xmin,xmax,ymin,ymax);
    gt.newPage();

    gt.setfgColor("black");
    gt.drawAxes();
    for (int n=0;n<=order;n++) {
#ifdef DEBUG
//    cout << "\tn = " << n << endl;
//    if (continuity==0) {
//      cout<< lp->slope2(static_cast<REAL>(n)/static_cast<REAL>(order),n)
//           << endl;
//    }
#endif
      gt.setfgColor("black");
      lp->slope2s(0.,s2);
      gt.movePen(0.,s2[n]);
      for (REAL x=dx;x<=1.;x+=dx) {
        lp->slope2s(x,s2);
        gt.drawLine(x,s2[n]);
      }
      gt.setfgColor(n+1,order+1);
      NumPtr<REAL> vd(3);
      lp->derivatives(0.,n,vd);
      gt.movePen(0.,vd[2]);
      for (REAL x=dx;x<=1.;x+=dx) {
        lp->derivatives(x,n,vd);
        gt.drawLine(x,vd[2]);
      }
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }

  delete lp;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runMain(BOOLEAN /*called_before*/) {
//TRACER_CALL(t,"runMain");
  runPolynomial();
  if (polynomial==C0LAGRANGE) {
    C0LagrangePolynomial c0lp(order);
    REAL dx=1./static_cast<REAL>(order);
    NumPtr<REAL> v(order+1);
    for (int n=0;n<=order;n++) {
      c0lp.values(static_cast<REAL>(n)*dx,v);
      for (int i=0;i<=order;i++) {
        REAL exact=(i==n ? 1. : 0.);
        REAL error=abs(exact-v[i]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tn,i = " << n << " " << i << " v = " << v[i] << " "
               << exact << " error = " << error << " " 
               << error/DBL_EPSILON << endl;
        }
      }
    }
  }
  if (polynomial==LEGENDRE) {
    LegendrePolynomial lp;
    NumPtr<REAL> v(order+1);
    lp.values(0.,v);
    for (int n=0;n<=order;n++) {
      REAL error=(n%2==0 ?  abs(v[n]-1.) : abs(v[n]+1.));
      if (error>16.*DBL_EPSILON) {
        cout << "\tlp[" << n << "](0.) = " << v[n] << " error = "
             << error << " " << error/DBL_EPSILON << endl;
      }
    }
    lp.values(1.,v);
    for (int n=0;n<=order;n++) {
      REAL error=abs(v[n]-1.);
      if (error>16.*DBL_EPSILON) {
        cout << "\tlp[" << n << "](1.) = " << v[n] << " error = "
             << error << " " << error/DBL_EPSILON << endl;
      }
    }

    REAL x=drand48();
    NumPtr<REAL> s(order+1);
    lp.values(x,v);
    lp.slopes(x,s);
    for (int n=2;n<=order;n++) {
      REAL error=abs(0.5*(s[n]-s[n-2])-v[n-1]*static_cast<REAL>(2*n-1));
      if (error>32.*DBL_EPSILON) {
        cout << "\tderivative diference[" << n << "] = " << error << " " 
             << error/DBL_EPSILON << endl;
      }
    }

    GaussianQuadrature<1> q(order+1);
    int np=q.numberPoints();
    NumPtr<REAL> quad(order+1);
    for (int n=0;n<=order;n++) {
      quad.initialize(0.);
      for (int p=0;p<np;p++) {
        const Point<1> &x=q.point(p);
        lp.values(x[0],v);
        for (int i=0;i<=order;i++) {
          quad[i]+=v[i]*v[n]*q.weight(p);
        }
      }
      for (int i=0;i<=order;i++) {
        REAL error=(i==n ? abs(quad[i]*static_cast<REAL>(2*n+1)-1.)
                         : abs(quad[i]*static_cast<REAL>(2*n+1)));
        if (error>32.*DBL_EPSILON) {
          cout << "\tintegral(lp[" << n << "],lp[" << i << "]) = "
               << quad[i] << " error = " << error << " " 
               << error/DBL_EPSILON << endl;
        }
      }
    }
  }
  if (polynomial==HIERARCHICAL) {
//  check orthogonality 
    HierarchicalPolynomial hp;
    NumPtr<REAL> v(order+1);
    hp.values(0.,v);
    for (int n=0;n<=order;n++) {
      REAL error=(n==0 ?  abs(v[n]-1.) : abs(v[n]));
      if (error>16.*DBL_EPSILON) {
        cout << "\thp[" << n << "](0.) = " << v[n] << " error = "
             << error << " " << error/DBL_EPSILON << endl;
      }
    }
    hp.values(1.,v);
    for (int n=0;n<=order;n++) {
      REAL error=(n==1 ?  abs(v[n]-1.) : abs(v[n]));
      if (error>16.*DBL_EPSILON) {
        cout << "\thp[" << n << "](1.) = " << v[n] << " error = "
             << error << " " << error/DBL_EPSILON << endl;
      }
    }

    GaussianQuadrature<1> q(order+1);
    int np=q.numberPoints();
    NumPtr<REAL> quad(order+1);
    quad.initialize(0.);
/*
    LegendrePolymomial lp;
    NumPtr<REAL> lv(order);
    REAL x=drand48();
    for (int p=0;p<np;p++) {
      REAL qp=q.point(p);
      REAL z=0.5*(qp*(1.+x)
    }
*/

    NumPtr<REAL> s(order+1);
    for (int n=2;n<=order;n++) {
      quad.initialize(0.);
      for (int p=0;p<np;p++) {
        const Point<1> &x=q.point(p);
        hp.slopes(x[0],s);
        for (int i=0;i<=order;i++) {
          quad[i]+=s[i]*s[n]*q.weight(p);
        }
      }
      for (int i=0;i<=order;i++) {
        if (i!=n) {
          REAL error=abs(quad[i]/quad[n]);
          if (error>16.*DBL_EPSILON) {
            cout << "\tintegral(hp[" << n << "],hp[" << i << "]) = "
                 << quad[i] << " error = " << error << " " 
                 << error/DBL_EPSILON << endl;
          }
        }
      }
    }
  }
  if (polynomial==HIERARCHICAL_SIDE) {
    REAL x=drand48();
    HierarchicalSidePolynomial hsp;
    HierarchicalPolynomial hp;
    NumPtr<REAL> hsp_vals(order+1);
    hsp.values(x,hsp_vals);
    NumPtr<REAL> hp_vals(order+3);
    hp.values(x,hp_vals);
    REAL q=x*(1.-x);
    for (int n=0;n<=order;n++) {
      REAL error=abs(hsp_vals[n]*q-hp_vals[n+2]);
      if (error>32.*DBL_EPSILON*abs(hp_vals[n+2])) {
        cout << "\thsp_vals[" << n << "] = " << hsp_vals[n]*q << " "
             << hp_vals[n+2] << " " << error;
        if (abs(hp_vals[n+2])>0.) {
          cout << " " << error/(DBL_EPSILON*abs(hp_vals[n+2])) << endl;
        } else cout << endl;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void cleanup() {
//TRACER_CALL(t,"cleanup");
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//use this for things that need to be done at the end of GUI operations
void shutdown() {
//TRACER_CALL(t,"shutdown");
//things allocated in processCommandLine:
  if (display_name) delete display_name; display_name=0;

//things allocated in makeMainList:
  if (main_list) {
    while (main_list->notEmpty()) {
      delete main_list->delAfter(0);
    }
    delete main_list; main_list=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
//cout << "\tin main" << endl;
#ifdef DEBUG
//setTraps();
#endif
#if MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  ifstream in_file;
  processCommandLine(argc,argv,display_name,skip_gui,in_file);
  makeMainList(main_list);
  if (in_file) {
    readMainInput(in_file,main_list,skip_gui);
    if (skip_gui) checkMainInput();
  }

  if (skip_gui) {
    runMain(0);
    cleanup();
    shutdown();
  } else {
#ifdef USE_GTK
    GTKGUI gui(argv[0],display_name,main_list,&runMain,
      &checkMainInput,&cleanup,&shutdown,TRUE);
#else
    GUI gui(argv[0],display_name,main_list,&runMain,&checkMainInput,
      &cleanup,&shutdown,TRUE);
#endif
    gui.createFileMenu();
    gui.createViewMenu();
    gui.createHelpMenu();
    gui.createMainWindow(argc,argv);
    gui.eventLoop();
  }
  return EXIT_SUCCESS;
}
