// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <fstream>
#include <iomanip>
#include <iostream>
#include <stdlib.h>
#include "Shape.H"
#include "Matrix.H"
#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "ShapeFunction.H"
#include "Surface.H"
#include "SurfaceDataSet.H"
#include "Tracer.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif
REAL winsize=0.5;

extern "C" {
void F77NAME(dpotrf)(char *uplo,int *n,double *a,int *lda,int *info);
void F77NAME(dpotri)(char *uplo,int *n,double *a,int *lda,int *info);
}

void debugPoly(const ShapeFunction<SPACEDIM> *pols,int degree) {
//TRACER_CALL(t,"debugPoly");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int n=pols->getNumber();
  const TensorProductPolynomials<SPACEDIM> *tpp=
    dynamic_cast<const TensorProductPolynomials<SPACEDIM>*>(pols);
  if (tpp) {
//  TRACER_CALL(t,"debugPoly TensorProductPolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
#if (SPACEDIM==1)
    const C0LagrangeIntervalPolynomials *c0lip=
      dynamic_cast<const C0LagrangeIntervalPolynomials*>(pols);
    const HierarchicalIntervalPolynomials *hip=
      dynamic_cast<const HierarchicalIntervalPolynomials*>(pols);
    if (c0lip==0 && hip==0)
#elif (SPACEDIM==2)
    const C0LagrangeQuadrilateralPolynomials *c0lqp=
      dynamic_cast<const C0LagrangeQuadrilateralPolynomials*>(pols);
    const HierarchicalQuadrilateralPolynomials *hqp=
      dynamic_cast<const HierarchicalQuadrilateralPolynomials*>(pols);
    if (c0lqp==0 && hqp==0)
#else
    const C0LagrangeHexahedronPolynomials *c0lhp=
      dynamic_cast<const C0LagrangeHexahedronPolynomials*>(pols);
    const HierarchicalHexahedronPolynomials *hhp=
      dynamic_cast<const HierarchicalHexahedronPolynomials*>(pols);
    if (c0lhp==0 && hhp==0)
#endif
    {
      NumPtr<Point<SPACEDIM> > points;
      D_CHOOSE(Interval,Quadrilateral,Hexahedron)::
        staticEquallySpacedLatticePoints(degree,points);
      for (int pi=0;pi<n;pi++) {
        NumPtr<REAL> v;
        pols->computeValues(points[pi],v);
        for (int s=0;s<n;s++) {
          REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
          if (error>DBL_EPSILON*512.) {
            cout << "\tcomputeValue(" << s << "," << points[pi] 
                 << ") = " << v[s] << ", error = " << error << " "
                 << error/DBL_EPSILON << endl; 
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
    }
  }
#if (SPACEDIM==1)
  const C0LagrangeIntervalPolynomials *c0lip=
    dynamic_cast<const C0LagrangeIntervalPolynomials*>(pols);
  if (c0lip) {
//  TRACER_CALL(t,"debugPoly C0LagrangeIntervalPolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    int vertices[2]={0,1};
    Interval e(0,vertices);
//  NumPtr<int> renumber;
//  e.reorderEquallySpacedLatticePoints(degree,renumber);
    Interval::staticEquallySpacedLatticePoints(degree,points);
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error = " << error 
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }
#endif
#if (SPACEDIM==1)
  const HierarchicalIntervalPolynomials *hip=
    dynamic_cast<const HierarchicalIntervalPolynomials*>(pols);
  if (hip) {
//  TRACER_CALL(t,"debugPoly HierarchicalIntervalPolynomials");
//  Ciarlet's nodal variables are either
//    point evaluation at the vertices, or
//    integration over the canonical interval of products of the derivatives
    for (int pi=0;pi<2;pi++) {
      Point<1> p=Interval::canonicalVertex(pi);
      NumPtr<REAL> v;
      pols->computeValues(p,v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*64.) {
          cout << "\tcomputeValue(" << s << "," << p << ") = " << v[s] 
               << ", error = " << error  << endl; 
          CHECK_TEST(error<=DBL_EPSILON*64.);
        }
      }
    }
    GaussianQuadrature<1> gq(degree);
    const NumPtr<Point<1> > &points=gq.points();
    const NumPtr<REAL> &weights=gq.weights();
    for (int pi=2;pi<n;pi++) {
      NumPtr<REAL> v(n);
      v.initialize(0.);
      for (int q=0;q<points.getNumber();q++) {
        NumPtr<Tensor<1,1> > g;
        pols->computeGrads(points[q],g);
        for (int s=0;s<n;s++) {
          v[s]+=g[s][0]*g[pi][0]*weights[q];
        }
      }
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*64.) {
          cout << "\tquadrature(" << pi << "," << s << "," << pi << ") = " 
               << v[s] << ", error = " << error  << endl; 
          CHECK_TEST(error<=DBL_EPSILON*64.);
        }
      }
    }
  }
/*
  {
    REAL xmin=0.;
    REAL xmax=1.;
    REAL ymin=0.;
    REAL ymax=1.;
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
    XYGraphTool gt("HierarchicalIntervalPolynomials","x",
      "basis polynomial",xmin,xmax,ymin,ymax,&cmap,0,winsize);
    gt.setbgColor("white");
    REAL dx=gt.getPixelWidth();
    Point<1> p;
    for (REAL x=xmin;x<=xmax;x+=dx) {
      p[0]=x;
      pols->computeValues(x,v);
      for (int s=0;s<n;s++) {
        ymin=min(ymin,v[s]);
        ymax=max(ymax,v[s]);
      }
    }
    gt.rescale(xmin,xmax,ymin,ymax);
    gt.newPage();
    gt.setfgColor("black");
    gt.drawAxes();
    for (int s=0;s<n;s++) {
      gt.setfgColor(s+1,n+1);
      for (REAL x=xmin;x<=xmax;x+=dx) {
        p[0]=x;
        pols->computeValues(x,v);
        if (x<=xmin) gt.movePen(x,v[s]);
        else gt.drawLine(x,v[s]);
      }
    }
    gt.flush();
    {
      XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
  }
*/
#endif
#if (SPACEDIM==2)
  const C0LagrangeTrianglePolynomials *ltp=
    dynamic_cast<const C0LagrangeTrianglePolynomials*>(pols);
  if (ltp) {
//  TRACER_CALL(t,"debugPoly LagrangeTrianglePolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    Triangle::staticEquallySpacedLatticePoints(degree,points);
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }
  const C0LagrangeTrianglePolynomials *c0ltp=
    dynamic_cast<const C0LagrangeTrianglePolynomials*>(pols);
  if (c0ltp) {
//  TRACER_CALL(t,"debugPoly C0LagrangeTrianglePolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    int vertices[3]={0,1,2};
    Triangle e(0,vertices,vertices);
//  NumPtr<int> renumber;
//  e.reorderEquallySpacedLatticePoints(degree,renumber);
    Triangle::staticEquallySpacedLatticePoints(degree,points);
#ifdef DEBUG
//  for (int pi=0;pi<points.getNumber();pi++) {
//    cout << "\tpoints[" << pi << "] = " << points[pi] << endl;
//  }
//  for (int pi=0;pi<renumber.getNumber();pi++) {
//    cout << "\trenumber[" << pi << "] = " << renumber[pi] << endl;
//  }
#endif
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
#ifdef DEBUG
//    cout << "\n\tpoints[" << pi << "] = " << points[pi] << endl;
//    for (int s=0;s<v.getNumber();s++) {
//      cout << "\tv[" << s << "] = " << v[s] << endl;
//    }
#endif
      for (int s=0;s<n;s++) {
        REAL vs=v[s];
        REAL error=(s==pi ? abs(vs-1.) : abs(vs));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << vs << ", error = " << error 
               << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }
  const HierarchicalTrianglePolynomials *htp=
    dynamic_cast<const HierarchicalTrianglePolynomials*>(pols);
  if (htp) {
//  TRACER_CALL(t,"debugPoly HierarchicalTrianglePolynomials");
//  Ciarlet's nodal variables are either
//    point evaluation at the vertices, or WHAT?
    for (int pi=0;pi<3;pi++) {
      Point<2> p=Triangle::canonicalVertex(pi);
      NumPtr<REAL> v;
      pols->computeValues(p,v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*64.) {
          cout << "\tcomputeValue(" << s << "," << p << ") = " << v[s] 
               << ", error = " << error  << endl; 
          CHECK_TEST(error<=DBL_EPSILON*64.);
        }
      }
    }
    TriangleProductGaussianQuadrature gq(degree+1);
    const NumPtr<Point<2> > &points=gq.points();
    const NumPtr<REAL> &weights=gq.weights();

    NumPtr<REAL> m(n*n);
    m.initialize(0.);
    for (int q=0;q<points.getNumber();q++) {
      NumPtr<REAL> v;
      pols->computeValues(points[q],v);
      const REAL &w=weights[q];
      for (int j=0;j<n;j++) {
        REAL vj=v[j]*w;
        for (int i=0;i<=j;i++) m[i+j*n]+=v[i]*vj;
      }
    }
    for (int j=0;j<n;j++) {
      for (int i=j+1;i<n;i++) m[i+j*n]=m[j+i*n];
    }
#ifdef DEBUG
//  for (int i=0;i<n;i++) {
//    cout << "\tm[" << i << ",*] =";
//    for (int j=0;j<n;j++) {
//      cout << " " << m[i+j*n];
//    }
//    cout << endl;
//  }
#endif
    int info;
    char uplo='L';
    F77NAME(dpotrf)(&uplo,&n,m.getData(),&n,&info);
    CHECK_SAME(info,0);
    F77NAME(dpotri)(&uplo,&n,m.getData(),&n,&info);
    CHECK_SAME(info,0);
#ifdef DEBUG
//  for (int i=0;i<n;i++) {
//    cout << "\tm inverse[" << i << ",*] =";
//    for (int j=0;j<=i;j++) {
//      cout << " " << m[i+j*n];
//    }
//    for (int j=i+1;j<n;j++) {
//      cout << " " << m[j+i*n];
//    }
//    cout << endl;
//  }
#endif
  }
  const C0LagrangeQuadrilateralPolynomials *c0qtp=
    dynamic_cast<const C0LagrangeQuadrilateralPolynomials*>(pols);
  if (c0qtp) {
//  TRACER_CALL(t,"debugPoly C0LagrangeQuadrilateralPolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    int vertices[4]={0,1,2,3};
    Quadrilateral e(0,vertices,vertices);
//  NumPtr<int> renumber;
//  e.reorderEquallySpacedLatticePoints(degree,renumber);
    Quadrilateral::staticEquallySpacedLatticePoints(degree,points);
#ifdef DEBUG
//  for (int pi=0;pi<points.getNumber();pi++) {
//    cout << "\tpoints[" << pi << "] = " << points[pi] << endl;
//  }
//  for (int pi=0;pi<renumber.getNumber();pi++) {
//    cout << "\trenumber[" << pi << "] = " << renumber[pi] << endl;
//  }
#endif
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error = " << error  
               << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }
  const HierarchicalQuadrilateralPolynomials *hqp=
    dynamic_cast<const HierarchicalQuadrilateralPolynomials*>(pols);
  if (hqp) {
//  TRACER_CALL(t,"debugPoly HierarchicalQuadrilateralPolynomials");
//  Ciarlet's nodal variables are either
//    point evaluation at the vertices, or
//    integration over the canonical square of products of the derivatives
    for (int pi=0;pi<4;pi++) {
      Point<2> p=Quadrilateral::canonicalVertex(pi);
      NumPtr<REAL> v;
      pols->computeValues(p,v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*64.) {
          cout << "\tcomputeValue(" << s << "," << p << ") = " << v[s] 
               << ", error = " << error  << endl; 
          CHECK_TEST(error<=DBL_EPSILON*64.);
        }
      }
    }
    GaussianQuadrature<2> gq(degree+1);
    const NumPtr<Point<2> > &points=gq.points();
    const NumPtr<REAL> &weights=gq.weights();
    NumPtr<REAL> m(n*n);
    m.initialize(0.);
    for (int q=0;q<points.getNumber();q++) {
      NumPtr<REAL> v;
      pols->computeValues(points[q],v);
      const REAL &w=weights[q];
      for (int j=0;j<n;j++) {
        REAL vj=v[j]*w;
        for (int i=0;i<=j;i++) m[i+j*n]+=v[i]*vj;
      }
    }
    for (int j=0;j<n;j++) {
      for (int i=j+1;i<n;i++) m[i+j*n]=m[j+i*n];
    }
#ifdef DEBUG
//  for (int i=0;i<n;i++) {
//    cout << "\tm[" << i << ",*] =";
//    for (int j=0;j<n;j++) {
//      cout << " " << m[i+j*n];
//    }
//    cout << endl;
//  }
#endif
    int info;
    char uplo='L';
    F77NAME(dpotrf)(&uplo,&n,m.getData(),&n,&info);
    CHECK_SAME(info,0);
    F77NAME(dpotri)(&uplo,&n,m.getData(),&n,&info);
    CHECK_SAME(info,0);
#ifdef DEBUG
//  for (int i=0;i<n;i++) {
//    cout << "\tm inverse[" << i << ",*] =";
//    for (int j=0;j<=i;j++) {
//      cout << " " << m[i+j*n];
//    }
//    for (int j=i+1;j<n;j++) {
//      cout << " " << m[j+i*n];
//    }
//    cout << endl;
//  }
#endif
  }
#endif
#if (SPACEDIM==3)
  const C0LagrangeTetrahedronPolynomials *ltep=
    dynamic_cast<const C0LagrangeTetrahedronPolynomials*>(pols);
  if (ltep) {
//  TRACER_CALL(t,"debugPoly LagrangeTetrahedronPolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    Tetrahedron::staticEquallySpacedLatticePoints(degree,points);
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*256.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*256.);
        }
      }
    }
  }
  const C0LagrangeTetrahedronPolynomials *c0ltep=
    dynamic_cast<const C0LagrangeTetrahedronPolynomials*>(pols);
  if (c0ltep) {
//  TRACER_CALL(t,"debugPoly C0LagrangeTetrahedronPolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    int vertices[4]={0,1,2,3};
    int lines[6]={0,1,2,3,4,5};
    Tetrahedron e(0,lines,vertices,vertices);
//  NumPtr<int> renumber;
//  e.reorderEquallySpacedLatticePoints(degree,renumber);
    Tetrahedron::staticEquallySpacedLatticePoints(degree,points);
#ifdef DEBUG
//  for (int pi=0;pi<points.getNumber();pi++) {
//    cout << "\tpoints[" << pi << "] = " << points[pi] << endl;
//  }
//  for (int pi=0;pi<renumber.getNumber();pi++) {
//    cout << "\trenumber[" << pi << "] = " << renumber[pi] << endl;
//  }
#endif
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
#ifdef DEBUG
//    cout << "\n\tpoints[" << pi << "] = " << points[pi] << endl;
//    for (int s=0;s<v.getNumber();s++) {
//      cout << "\tv[" << s << "] = " << v[s] << endl;
//    }
#endif
      for (int s=0;s<n;s++) {
        REAL vs=v[s];
        REAL error=(s==pi ? abs(vs-1.) : abs(vs));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << vs << ", error = " << error 
               << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
      }
    }
  }
  const C0LagrangePrismPolynomials *lwp=
    dynamic_cast<const C0LagrangePrismPolynomials*>(pols);
  if (lwp) {
//  TRACER_CALL(t,"debugPoly LagrangePrismPolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    Prism::staticEquallySpacedLatticePoints(degree,points);
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*256.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*256.);
        }
      }
    }
  }
  const C0LagrangePrismPolynomials *c0lwp=
    dynamic_cast<const C0LagrangePrismPolynomials*>(pols);
  if (c0lwp) {
//  TRACER_CALL(t,"debugPoly C0LagrangePrismPolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    int vertices[6]={0,1,2,3,4,5};
    int lines[9]={0,1,2,3,4,5,6,7,8};
    int faces[9]={0,1,2,3,4};
    Prism e(0,lines,faces,vertices);
//  NumPtr<int> renumber;
//  e.reorderEquallySpacedLatticePoints(degree,renumber);
    Prism::staticEquallySpacedLatticePoints(degree,points);
#ifdef DEBUG
//  for (int pi=0;pi<points.getNumber();pi++) {
//    cout << "\tpoints[" << pi << "] = " << points[pi] << endl;
//  }
//  for (int pi=0;pi<renumber.getNumber();pi++) {
//    cout << "\trenumber[" << pi << "] = " << renumber[pi] << endl;
//  }
#endif
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
#ifdef DEBUG
//    cout << "\n\tpoints[" << pi << "] = " << points[pi] << endl;
//    for (int s=0;s<v.getNumber();s++) {
//      cout << "\tv[" << s << "] = " << v[s] << endl;
//    }
#endif
      for (int s=0;s<n;s++) {
        REAL vs=v[s];
        REAL error=(s==pi ? abs(vs-1.) : abs(vs));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << vs << ", error = " << error 
               << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
      }
    }
  }
  const C0LagrangeHexahedronPolynomials *c0lhp=
    dynamic_cast<const C0LagrangeHexahedronPolynomials*>(pols);
  if (c0lhp) {
//  TRACER_CALL(t,"debugPoly C0LagrangeHexahedronPolynomials");
//  Ciarlet's nodal variables are point evaluation at canonical support pts
    NumPtr<Point<SPACEDIM> > points;
    int vertices[8]={0,1,2,3,4,5,6,7};
    int faces[6]={0,1,2,3,4,5};
    int lines[12]={0,1,2,3,4,5,6,7,8,9,10,11};
    Hexahedron e(0,lines,faces,vertices);
//  NumPtr<int> renumber;
//  e.reorderEquallySpacedLatticePoints(degree,renumber);
    Hexahedron::staticEquallySpacedLatticePoints(degree,points);
#ifdef DEBUG
//  for (int pi=0;pi<points.getNumber();pi++) {
//    cout << "\tpoints[" << pi << "] = " << points[pi] << endl;
//  }
//  for (int pi=0;pi<renumber.getNumber();pi++) {
//    cout << "\trenumber[" << pi << "] = " << renumber[pi] << endl;
//  }
#endif
    for (int pi=0;pi<n;pi++) {
      NumPtr<REAL> v;
      pols->computeValues(points[pi],v);
      for (int s=0;s<n;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error = " << error  
               << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }
  const HierarchicalTetrahedronPolynomials *htep=
    dynamic_cast<const HierarchicalTetrahedronPolynomials*>(pols);
  if (htep) {
//  TRACER_CALL(t,"debugPoly HierarchicalTetrahedronPolynomials");
//  Ciarlet's nodal variables are ???
    int vertices[4]={0,1,2,3};
    int lines[6]={0,1,2,3,4,5};
    Tetrahedron e(0,lines,vertices,vertices);
    for (int pi=0;pi<4;pi++) {
      Point<3> p=Tetrahedron::canonicalVertex(pi);
      NumPtr<REAL> v;
      htep->computeValuesFor(&e,p,v);
      for (int s=0;s<n;s++) {
        REAL vs=v[s];
        REAL error=(s==pi ? abs(vs-1.) : abs(vs));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s << "," << p << ") = " << vs 
               << ", error = " << error << " " << error/DBL_EPSILON 
               << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
      }
    }
  }
  const HierarchicalPrismPolynomials *hwp=
    dynamic_cast<const HierarchicalPrismPolynomials*>(pols);
  if (hwp) {
//  TRACER_CALL(t,"debugPoly HierarchicalPrismPolynomials");
//  Ciarlet's nodal variables are ???
    int vertices[6]={0,1,2,3,4,5};
    int lines[9]={0,1,2,3,4,5,6,7,8};
    int faces[5]={0,1,2,3,4};
    Prism e(0,lines,faces,vertices);
    for (int pi=0;pi<6;pi++) {
      Point<3> p=Prism::canonicalVertex(pi);
      NumPtr<REAL> v;
      hwp->computeValuesFor(&e,p,v);
      for (int s=0;s<n;s++) {
        REAL vs=v[s];
        REAL error=(s==pi ? abs(vs-1.) : abs(vs));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s << "," << p << ") = " << vs 
               << ", error = " << error << " " << error/DBL_EPSILON 
               << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
      }
    }
  }
  const HierarchicalHexahedronPolynomials *hhp=
    dynamic_cast<const HierarchicalHexahedronPolynomials*>(pols);
  if (hhp) {
//  TRACER_CALL(t,"debugPoly HierarchicalHexahedronPolynomials");
//  Ciarlet's nodal variables are ???
    int vertices[8]={0,1,2,3,4,5,6,7};
    int lines[12]={0,1,2,3,4,5,6,7,8,9,10,11};
    int faces[6]={0,1,2,3,4,5};
    Hexahedron e(0,lines,faces,vertices);
    for (int pi=0;pi<8;pi++) {
      Point<3> p=Hexahedron::canonicalVertex(pi);
      NumPtr<REAL> v;
      pols->computeValuesFor(&e,p,v);
      for (int s=0;s<n;s++) {
        REAL vs=v[s];
        REAL error=(s==pi ? abs(vs-1.) : abs(vs));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s << "," << p << ") = " << vs 
               << ", error = " << error << " " << error/DBL_EPSILON 
               << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
      }
    }
    GaussianQuadrature<3> gq(degree+1);
    const NumPtr<Point<3> > &points=gq.points();
    const NumPtr<REAL> &weights=gq.weights();

    NumPtr<REAL> m(n*n);
    m.initialize(0.);
    for (int q=0;q<points.getNumber();q++) {
      NumPtr<REAL> v;
      pols->computeValuesFor(&e,points[q],v);
      const REAL &w=weights[q];
      for (int j=0;j<n;j++) {
        REAL vj=v[j]*w;
        for (int i=0;i<=j;i++) m[i+j*n]+=v[i]*vj;
      }
    }
    for (int j=0;j<n;j++) {
      for (int i=j+1;i<n;i++) m[i+j*n]=m[j+i*n];
    }
#ifdef DEBUG
//  for (int i=0;i<n;i++) {
//    cout << "\tm[" << i << ",*] =";
//    for (int j=0;j<n;j++) {
//      cout << " " << m[i+j*n];
//    }
//    cout << endl;
//  }
#endif
    int info;
    char uplo='L';
    F77NAME(dpotrf)(&uplo,&n,m.getData(),&n,&info);
    CHECK_SAME(info,0);
    F77NAME(dpotri)(&uplo,&n,m.getData(),&n,&info);
    CHECK_SAME(info,0);
#ifdef DEBUG
//  for (int i=0;i<n;i++) {
//    cout << "\tm inverse[" << i << ",*] =";
//    for (int j=0;j<=i;j++) {
//      cout << " " << m[i+j*n];
//    }
//    for (int j=i+1;j<n;j++) {
//      cout << " " << m[j+i*n];
//    }
//    cout << endl;
//  }
#endif
  }
#endif
}

#if (SPACEDIM==2)
void debugPoly(const ShapeFunction<SPACEDIM> *pols,int degree0,
int degree1) {
//TRACER_CALL(t,"debugPoly");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int n=pols->getNumber();
  const C0LagrangeQuadrilateralPolynomials *c0lqp=
    dynamic_cast<const C0LagrangeQuadrilateralPolynomials*>(pols);
  CHECK_POINTER(c0lqp);
//Ciarlet's nodal variables are point evaluation at canonical support pts
  NumPtr<Point<SPACEDIM> > points;
  int vertices[4]={0,1,2,3};
  Quadrilateral e(0,vertices,vertices);
  Quadrilateral::staticEquallySpacedLatticePoints(degree0,degree1,points);
#ifdef DEBUG
//for (int pi=0;pi<points.getNumber();pi++) {
//  cout << "\tpoints[" << pi << "] = " << points[pi] << endl;
//}
//for (int pi=0;pi<renumber.getNumber();pi++) {
//  cout << "\trenumber[" << pi << "] = " << renumber[pi] << endl;
//}
#endif
  for (int pi=0;pi<n;pi++) {
    NumPtr<REAL> v;
    pols->computeValues(points[pi],v);
    for (int s=0;s<n;s++) {
      REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
      if (error>DBL_EPSILON*512.) {
        cout << "\tcomputeValue(" << s << "," << points[pi] 
             << ") = " << v[s] << ", error = " << error  
             << " " << error/DBL_EPSILON << endl; 
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
    }
  }
}
#endif
#if (SPACEDIM==3)
void debugPoly(const ShapeFunction<SPACEDIM> *pols,int degree0,
int degree1,int degree2) {
//TRACER_CALL(t,"debugPoly");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int n=pols->getNumber();
  const C0LagrangeHexahedronPolynomials *c0lhp=
    dynamic_cast<const C0LagrangeHexahedronPolynomials*>(pols);
  CHECK_POINTER(c0lhp);
//Ciarlet's nodal variables are point evaluation at canonical support pts
  NumPtr<Point<SPACEDIM> > points;
  int lines[12]={0,1,2,3,4,5,6,7,8,9,10,11};
  int faces[6]={0,1,2,3,4,5};
  int vertices[8]={0,1,2,3,4,5,6,7};
  Hexahedron e(0,lines,faces,vertices);
  Hexahedron::staticEquallySpacedLatticePoints(degree0,degree1,degree2,
    points);
#ifdef DEBUG
//for (int pi=0;pi<points.getNumber();pi++) {
//  cout << "\tpoints[" << pi << "] = " << points[pi] << endl;
//}
//for (int pi=0;pi<renumber.getNumber();pi++) {
//  cout << "\trenumber[" << pi << "] = " << renumber[pi] << endl;
//}
#endif
  for (int pi=0;pi<n;pi++) {
    NumPtr<REAL> v;
    pols->computeValues(points[pi],v);
    for (int s=0;s<n;s++) {
      REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
      if (error>DBL_EPSILON*512.) {
        cout << "\tcomputeValue(" << s << "," << points[pi] 
             << ") = " << v[s] << ", error = " << error  
             << " " << error/DBL_EPSILON << endl; 
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void debugDerivsCase(const Shape1 &e,const ShapeFunction<SPACEDIM> *pols,
int degree) {
//TRACER_CALL(t,"debugDerivsCase");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int n=pols->getNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  Point<SPACEDIM> p(D_DECL(drand48(),drand48(),drand48()));
  if (e.numberVertices()<=6) p[1]*=1.-p[0];
  if (e.numberVertices()==4) p[2]*=1.-p[0]-p[1];
  NumPtr<REAL> v;
  NumPtr<Tensor<1,SPACEDIM> > g;
  NumPtr<Tensor<2,SPACEDIM> > j;
  const HierarchicalTetrahedronPolynomials *htep=
    dynamic_cast<const HierarchicalTetrahedronPolynomials*>(pols);
  const HierarchicalPrismPolynomials *hwp=
    dynamic_cast<const HierarchicalPrismPolynomials*>(pols);
  const HierarchicalHexahedronPolynomials *hhp=
    dynamic_cast<const HierarchicalHexahedronPolynomials*>(pols);
  pols->computeValuesFor(&e,p,v);
  pols->computeGradsFor(&e,p,g);
  pols->computeGradGradsFor(&e,p,j);
  for (int d=0;d<SPACEDIM;d++) {
    Point<SPACEDIM> pp(p);
    pp[d]+=eps;
    NumPtr<REAL> vv;
    NumPtr<Tensor<1,SPACEDIM> > gg;
    pols->computeValuesFor(&e,pp,vv);
    pols->computeGradsFor(&e,pp,gg);
    for (int pi=0;pi<n;pi++) {
      REAL exact=g[pi][d];
      REAL deriv=(vv[pi]-v[pi])/eps;
      REAL error=abs(deriv-exact);
      REAL den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*262144.) {
        cout << "\tcomputeGrad(" << pi << "," << p << ")[" << d 
             << "] = " << exact << " " << deriv << ", error = "
             << error << " " << error/eps << " degree = " << degree
             << endl;
        CHECK_TEST(error<=eps*262144.);
      }
//
      for (int dd=0;dd<SPACEDIM;dd++) {
        REAL exact=j[pi][d][dd];
        REAL deriv=(gg[pi][dd]-g[pi][dd])/eps;
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>eps*262144.) {
          cout << "\tcomputeGradGrad(" << pi << "," << p << ")[" 
               << d << "][" << dd << "] = " << exact << " " 
               << deriv << ", error = " << error << " " << error/eps
               << endl;
          CHECK_TEST(error<=eps*262144.);
        }
      }
//
    }
  }
//
}
#endif

void richardsonDerivs(int nmax,const ShapeFunction<SPACEDIM> *pols,
const Point<SPACEDIM> &x,const NumPtr<REAL> &f,
NumPtr<Tensor<1,SPACEDIM> > &fp) {
//TRACER_CALL(t,"richardsonDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  fp.allocate(f.getNumber());
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<NumPtr<REAL> > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<REAL> fnew;
      Point<SPACEDIM> xnew; xnew=x; xnew[d]+=h;
      pols->computeValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        table[nmax-n][s]=(fnew[s]-f[s])/h;
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
        }
      }
    }
    for (int s=0;s<f.getNumber();s++) {
      fp[s][d]=table[nmax-1][s];
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardson2ndDerivs(int nmax,const ShapeFunction<SPACEDIM> *pols,
const Point<SPACEDIM> &x,const NumPtr<Tensor<1,SPACEDIM> > &fp,
NumPtr<Tensor<2,SPACEDIM> > &fpp) {
  fpp.allocate(fp.getNumber());
  for (int i=0;i<SPACEDIM;i++) {
    for (int j=0;j<SPACEDIM;j++) {
      NumPtr<NumPtr<REAL> > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<1,SPACEDIM> > fpnew;
        Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
        pols->computeGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          table[nmax-n][s]=(fpnew[s][i]-fp[s][i])/h;
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        fpp[s][i][j]=table[nmax-1][s];
      }
    }
  }
}

void debugDerivs(const ShapeFunction<SPACEDIM> *pols,int degree) {
//TRACER_CALL(t,"debugDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int n=pols->getNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  Point<SPACEDIM> p(D_DECL(drand48(),drand48(),drand48()));
  NumPtr<REAL> v;
  NumPtr<Tensor<1,SPACEDIM> > g;
  NumPtr<Tensor<2,SPACEDIM> > j;
#if (SPACEDIM==3)
  const HierarchicalTetrahedronPolynomials *htep=
    dynamic_cast<const HierarchicalTetrahedronPolynomials*>(pols);
  const HierarchicalPrismPolynomials *hwp=
    dynamic_cast<const HierarchicalPrismPolynomials*>(pols);
  const HierarchicalHexahedronPolynomials *hhp=
    dynamic_cast<const HierarchicalHexahedronPolynomials*>(pols);
  if (htep) {
    TRACER_CALL(t,"debugDerivs");
    PermutationGenerator pg(4);
    int l[6];
    for (int i=0;i<6;i++) l[i]=i;
    int v[4];
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//    cout << "\tp =";
//    for (int i=0;i<p.getNumber();i++) cout << " " << p[i];
//    cout << endl;
#endif
      for (int i=0;i<4;i++) v[i]=p[i];
      Tetrahedron e(0,l,v,v);
      debugDerivsCase(e,pols,degree);
    }
    return;
  } else if (hwp) {
    PermutationGenerator pg(6);
    int l[9];
    int f[5];
    int v[6];
    for (int i=0;i<5;i++) f[i]=i;
    for (int i=0;i<9;i++) l[i]=i;
    while (pg.hasMore()) 
    {
      const NumPtr<int> &p=pg.getNext();
//    NumPtr<int> p(6);
//    p[0]=0;
//    p[1]=1;
//    p[2]=2;
//    p[3]=3;
//    p[4]=4;
//    p[5]=5;
#ifdef DEBUG
//    cout << "\tp =";
//    for (int i=0;i<p.getNumber();i++) cout << " " << p[i];
//    cout << endl;
#endif
      for (int i=0;i<6;i++) v[i]=p[i];
      Prism e(0,l,f,v);
      debugDerivsCase(e,pols,degree);
    }
    return;
  } else if (hhp) {
//  TRACER_CALL(t,"debugDerivs");
//  PermutationGenerator pg(8);
    int l[12];
    int f[6];
    int v[8];
    for (int i=0;i<6;i++) f[i]=i;
    for (int i=0;i<12;i++) l[i]=i;
//  while (pg.hasMore()) 
//  for (int count=0;count<1000;count++)
    {
//    const NumPtr<int> &p=pg.getNext();
      NumPtr<int> p(8);
//    RandomPermutation(p);
      p[0]=0;
      p[1]=1;
      p[2]=2;
      p[3]=3;
      p[4]=7;
      p[5]=6;
      p[6]=5;
      p[7]=4;
#ifdef DEBUG
      cout << "\tp =";
      for (int i=0;i<p.getNumber();i++) cout << " " << p[i];
      cout << endl;
#endif
      for (int i=0;i<8;i++) v[i]=p[i];
      Hexahedron e(0,l,f,v);
      debugDerivsCase(e,pols,degree);
    }
    return;
  }
#endif
  pols->computeValues(p,v);
  pols->computeGrads(p,g);
  pols->computeGradGrads(p,j);
  NumPtr<Tensor<1,SPACEDIM> > gg;
  NumPtr<Tensor<2,SPACEDIM> > jj;
  richardsonDerivs(degree,pols,p,v,gg);
  richardson2ndDerivs(degree,pols,p,g,jj);
  for (int d=0;d<SPACEDIM;d++) {
//  Point<SPACEDIM> pp(p);
//  pp[d]+=eps;
//  NumPtr<REAL> vv;
//  pols->computeValues(pp,vv);
//  pols->computeGrads(pp,gg);
    for (int pi=0;pi<n;pi++) {
      REAL exact=g[pi][d];
//    REAL deriv=(vv[pi]-v[pi])/eps;
      REAL deriv=gg[pi][d];
      REAL error=abs(deriv-exact);
      REAL den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>DBL_EPSILON*1048576.) {
        cout << "\tcomputeGrad(" << pi << "," << p << ")[" << d 
             << "] = " << exact << " " << deriv << ", error = "
             << error << " " << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*1048576.);
      }
      for (int dd=0;dd<SPACEDIM;dd++) {
        REAL exact=j[pi][d][dd];
//      REAL deriv=(gg[pi][dd]-g[pi][dd])/eps;
        REAL deriv=jj[pi][d][dd];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tcomputeGradGrad(" << pi << "," << p << ")[" 
               << d << "][" << dd << "] = " << exact << " " 
               << deriv << ", error = " << error << " "
               << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
  }
//
}

#if (SPACEDIM>=2)
void debugDerivs(int degree,const ShapeFunction<SPACEDIM> *pols) {
//TRACER_CALL(t,"debugDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int n=pols->getNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  Point<SPACEDIM> p(D_DECL(drand48(),drand48(),drand48()));
  NumPtr<REAL> v;
  NumPtr<Tensor<1,SPACEDIM> > g;
  NumPtr<Tensor<2,SPACEDIM> > j;
  pols->computeValues(p,v);
  pols->computeGrads(p,g);
  pols->computeGradGrads(p,j);
  NumPtr<Tensor<1,SPACEDIM> > gg;
  NumPtr<Tensor<2,SPACEDIM> > jj;
  richardsonDerivs(degree,pols,p,v,gg);
  richardson2ndDerivs(degree,pols,p,g,jj);
  for (int d=0;d<SPACEDIM;d++) {
//  Point<SPACEDIM> pp(p);
//  pp[d]+=eps;
//  NumPtr<REAL> vv;
//  NumPtr<Tensor<1,SPACEDIM> > gg;
//  pols->computeValues(pp,vv);
//  pols->computeGrads(pp,gg);
    for (int pi=0;pi<n;pi++) {
      REAL exact=g[pi][d];
//    REAL deriv=(vv[pi]-v[pi])/eps;
      REAL deriv=gg[pi][d];
      REAL error=abs(deriv-exact);
      REAL den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>DBL_EPSILON*1048576.) {
        cout << "\tcomputeGrad(" << pi << "," << p << ")[" << d 
             << "] = " << exact << " " << deriv << ", error = "
             << error << " " << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*1048576.);
      }
      for (int dd=0;dd<SPACEDIM;dd++) {
        REAL exact=j[pi][d][dd];
//      REAL deriv=(gg[pi][dd]-g[pi][dd])/eps;
        REAL deriv=jj[pi][d][dd];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tcomputeGradGrad(" << pi << "," << p << ")[" 
               << d << "][" << dd << "] = " << exact << " " 
               << deriv << ", error = " << error << " "
               << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
  }
}
#endif

void debugOrderCase(const Shape1 &e,const ShapeFunction<SPACEDIM> *pols,
int degree) {
//TRACER_CALL(z,"debugOrderCase");
#if (SPACEDIM==1)
  bool is_hierarchical=
    (dynamic_cast<const HierarchicalIntervalPolynomials*>(pols)!=0);
#elif (SPACEDIM==2)
  bool is_hierarchical=
    (dynamic_cast<const HierarchicalTrianglePolynomials*>(pols)!=0) ||
    (dynamic_cast<const HierarchicalQuadrilateralPolynomials*>(pols)!=0);
#else
  bool is_hierarchical=
    (dynamic_cast<const HierarchicalTetrahedronPolynomials*>(pols)!=0) ||
    (dynamic_cast<const HierarchicalPrismPolynomials*>(pols)!=0) ||
    (dynamic_cast<const HierarchicalHexahedronPolynomials*>(pols)!=0);
  const HierarchicalTetrahedronPolynomials *htp=
    dynamic_cast<const HierarchicalTetrahedronPolynomials*>(pols);
  const HierarchicalPrismPolynomials *hwp=
    dynamic_cast<const HierarchicalPrismPolynomials*>(pols);
  const HierarchicalHexahedronPolynomials *hhp=
    dynamic_cast<const HierarchicalHexahedronPolynomials*>(pols);
  const Shape3 *e3=dynamic_cast<const Shape3*>(&e);
#endif
  bool requires_shape_to_compute_values=
    pols->requiresShapeToComputeValues();
#if (SPACEDIM>1)
  const Shape2 *e2=dynamic_cast<const Shape2*>(&e);
#endif

  if (!requires_shape_to_compute_values) {
    NumPtr<int> shape_renumber;
    pols->shapeFunctionsOrder(&e,shape_renumber);
#ifdef DEBUG
//  for (int i=0;i<shape_renumber.getNumber();i++) {
//    cout << "\tshape_renumber[" << i << "] = " << shape_renumber[i]
//         << endl;
//  }
#endif
    NumPtr<int> renumber_inverse(shape_renumber.getNumber());
    renumber_inverse.initialize(INT_MAX);
    for (int i=0;i<shape_renumber.getNumber();i++) {
      renumber_inverse[abs(shape_renumber[i])]=i;
    }
    for (int i=0;i<renumber_inverse.getNumber();i++) { 
      CHECK_TEST(renumber_inverse[i]!=INT_MAX);
    }
  }

  NumPtr<Point<SPACEDIM> > lattice_points;
  e.equallySpacedLatticePoints(degree,lattice_points);

  for (int v=0;v<e.numberVertices();v++) {
//  TRACER_CALL(z,"debugOrderCase");
    int qv=e.equallySpacedLatticePointIndexOfVertex(v,degree);
    int value_index=(is_hierarchical ? v : qv);
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[qv],values);
    } else pols->computeValues(lattice_points[qv],values);
    REAL error=abs(values[value_index]-1.);
    if (error>512.*DBL_EPSILON) {
      cout << "\tvertex shape value at vertex " << v << " = "
           << values[value_index] << " " << error << " "
           << error/DBL_EPSILON << endl;
      CHECK_TEST(error<=DBL_EPSILON*512.);
    }
    for (int v2=0;v2<e.numberVertices();v2++) {
      if (v2==v) continue;
      value_index=(is_hierarchical ? v2 : 
        e.equallySpacedLatticePointIndexOfVertex(v2,degree));
      REAL error=abs(values[value_index]);
      if (error>512.*DBL_EPSILON) {
        cout << "\tvertex shape value at vertex " << v << " = "
             << values[value_index] << " " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
    }
  }
#if (SPACEDIM==3)
  for (int l=0;l<e3->numberCodimension2Shapes();l++) {
    NumPtr<int> indices; // fe order
    if (is_hierarchical) {
      int nl=degree-1;
      int m=e.numberVertices()+l*nl;
      indices.allocate(nl);
      for (int i=0;i<nl;i++) indices[i]=m+i;
    } else {
      e3->equallySpacedLatticePointIndicesOfLineInterior(l,degree,indices);
    }
    for (int v=0;v<e.numberVertices();v++) {
      int qv=e.equallySpacedLatticePointIndexOfVertex(v,degree);
      NumPtr<REAL> values;
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[qv],values);
      } else pols->computeValues(lattice_points[qv],values);
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tedge shape value at vertex " << v << " = "
               << values[indices[i]] << " " << error << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
    for (int l2=0;l2<e3->numberCodimension2Shapes();l2++) {
      if (l2==l) continue;
      NumPtr<int> indices2;
      e3->equallySpacedLatticePointIndicesOfLineInterior(l2,degree,
        indices2);
      for (int i=0;i<indices2.getNumber();i++) {
        NumPtr<REAL> values;
        if (requires_shape_to_compute_values) {
          pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
        } else {
          pols->computeValues(lattice_points[indices2[i]],values);
        }
        for (int i=0;i<indices.getNumber();i++) {
          REAL error=abs(values[indices[i]]);
          if (error>512.*DBL_EPSILON) {
            cout << "\tedge shape value at line " << l2 << " #" << i 
                 << " = " << values[indices[i]] << " " 
                 << error << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
    }
  }
#endif
#if (SPACEDIM>1)
  for (int f=0;f<e.numberCodimension1Shapes();f++) {
    NumPtr<int> indices; // fe order
    if (is_hierarchical) {
      int nv=e.numberVertices();
      int m=nv;
#if (SPACEDIM==3)
      m+=e3->numberCodimension2Shapes()*(degree-1);
#endif
      for (int f2=0;f2<f;f2++) {
        m+=numberInteriorEquallySpacedLatticePointsOnFace(nv,f2,degree);
      }
      int nof=numberInteriorEquallySpacedLatticePointsOnFace(nv,f,degree);
      indices.allocate(nof);
      for (int i=0;i<nof;i++) indices[i]=m+i;
    } else {
      e2->equallySpacedLatticePointIndicesOfFaceInterior(f,degree,indices);
    }
#ifdef DEBUG
//  cout << "\n\tface = " << f << endl;
//  for (int i=0;i<indices.getNumber();i++) {
//    cout << "\tface indices[" << i << "] = " << indices[i] << endl;
//  }
#endif
    for (int v=0;v<e.numberVertices();v++) {
      int qv=e.equallySpacedLatticePointIndexOfVertex(v,degree); //fe order
      NumPtr<REAL> values; // fe order
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[qv],values);
      } else pols->computeValues(lattice_points[qv],values);
#ifdef DEBUG
//    cout << "\tlattice_points[" << qv << "] = " << lattice_points[qv]
//         << endl;
//    for (int i=0;i<indices.getNumber();i++) {
//      cout << "\tvalues[" << indices[i] << "] = "
//           << values[indices[i]] << endl;
//    }
#endif
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tface shape value at vertex " << v << " = "
               << values[indices[i]] << " " << error << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
# if (SPACEDIM==3)
    for (int l=0;l<e3->numberCodimension2Shapes();l++) {
      NumPtr<int> indices2;
      e3->equallySpacedLatticePointIndicesOfLineInterior(l,degree,
        indices2);
      for (int i=0;i<indices2.getNumber();i++) {
        NumPtr<REAL> values;
        if (requires_shape_to_compute_values) {
          pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
        } else pols->computeValues(lattice_points[indices2[i]],values);
        for (int i=0;i<indices.getNumber();i++) {
          REAL error=abs(values[indices[i]]);
          if (error>512.*DBL_EPSILON) {
            cout << "\tface shape value at line " << l << " #" << i 
                 << " = " << values[indices[i]] << " " 
                 << error << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
    }
# endif
    for (int f2=0;f2<e.numberCodimension1Shapes();f2++) {
      if (f2==f) continue;
      NumPtr<int> indices2; // fe_order
      e2->equallySpacedLatticePointIndicesOfFaceInterior(f2,degree,
        indices2);
#ifdef DEBUG
//    cout << "\n\tf2 = " << f2 << endl;
//    for (int i=0;i<indices2.getNumber();i++) {
//      cout << "\tface indices2[" << i << "] = " << indices2[i] << endl;
//    }
#endif
      for (int i=0;i<indices2.getNumber();i++) {
        NumPtr<REAL> values;
        if (requires_shape_to_compute_values) {
          pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
        } else pols->computeValues(lattice_points[indices2[i]],values);
#ifdef DEBUG
//      cout << "\tlattice_points[" << indices2[i] << "] = " 
//           << lattice_points[indices2[i]] << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tvalues[" << indices[i] << "] = "
//             << values[indices[i]] << endl;
//      }
#endif
        for (int i=0;i<indices.getNumber();i++) {
          REAL error=abs(values[indices[i]]);
          if (error>512.*DBL_EPSILON) {
            cout << "\tface shape value at face " << f2 << " #" << i 
                 << " = " << values[indices[i]] << " " << error 
                 << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
    }
  }
#endif
  NumPtr<int> indices; // fe order
  if (is_hierarchical) {
    int n=pols->getNumber();
    int m=e.numberInteriorEquallySpacedLatticePoints(degree);
    indices.allocate(m);
    for (int i=0;i<m;i++) {
      indices[i]=n-m+i;
    }
  } else {
    e.interiorLatticePointIndices(degree,indices);
  }
#ifdef DEBUG
//for (int i=0;i<indices.getNumber();i++) {
//  cout << "\tinterior indices[" << i << "] = " << indices[i] << endl;
//}
#endif
  for (int v=0;v<e.numberVertices();v++) { // v is in dof order
//  TRACER_CALL(z,"debugOrderCase");
    int qv=e.equallySpacedLatticePointIndexOfVertex(v,degree); // fe order
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[qv],values);
    } else pols->computeValues(lattice_points[qv],values);
#ifdef DEBUG
//  cout << "\tv = " << v << " qv = " << qv << endl;
//  cout << "\tvertex = " << lattice_points[qv] << endl;
//  for (int i=0;i<indices.getNumber();i++) {
//    cout << "\tvalues[" << indices[i] << "] = " << values[indices[i]] 
//         << endl;
//  }
#endif
    for (int i=0;i<indices.getNumber();i++) {
      REAL error=abs(values[indices[i]]);
      if (error>512.*DBL_EPSILON) {
        cout << "\tinterior shape value at index " << indices[i] << " = "
             << values[indices[i]] << " " << error 
             << " " << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
    }
  }
#if (SPACEDIM==3)
  for (int l=0;l<e3->numberCodimension2Shapes();l++) {
    NumPtr<int> indices2;
    e3->equallySpacedLatticePointIndicesOfLineInterior(l,degree,indices2);
#ifdef DEBUG
//  cout << "\tline = " << l << endl;
//  for (int i=0;i<indices.getNumber();i++) {
//    cout << "\tline indices[" << i << "] = " << indices2[i] << endl;
//  }
#endif
    for (int i=0;i<indices2.getNumber();i++) {
      NumPtr<REAL> values;
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
      } else pols->computeValues(lattice_points[indices2[i]],values);
#ifdef DEBUG
//    cout << "\tlattice_points[" << indices2[i] << "] = " 
//         << lattice_points[indices2[i]] << endl;
//    for (int j=0;j<indices.getNumber();j++) {
//      cout << "\tvalues[" << indices[j] << "] = " << values[indices[j]] 
//           << endl;
//    }
#endif
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>DBL_EPSILON) {
          cout << "\tinterior shape value at line " << l << " #" << i 
               << " = " << values[indices[i]] << " " 
               << error << " " << error/DBL_EPSILON << endl;
        }
      }
    }
  }
#endif
#if (SPACEDIM>1)
  for (int f=0;f<e.numberCodimension1Shapes();f++) {
    NumPtr<int> indices2;
    e2->equallySpacedLatticePointIndicesOfFaceInterior(f,degree,indices2);
#ifdef DEBUG
//  cout << "\tface = " << f << endl;
//  for (int i=0;i<indices2.getNumber();i++) {
//    cout << "\tface indices[" << i << "] = " << indices2[i] << endl;
//  }
#endif
    for (int i=0;i<indices2.getNumber();i++) {
      NumPtr<REAL> values;
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
      } else pols->computeValues(lattice_points[indices2[i]],values);
#ifdef DEBUG
//    cout << "\tlattice_points[" << indices2[i] << "] = " 
//         << lattice_points[indices2[i]] << endl;
//    for (int j=0;j<indices.getNumber();j++) {
//      cout << "\tvalues[" << indices[j] << "] = " << values[indices[j]] 
//           << endl;
//    }
#endif
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tinterior shape value at face " << f << " #" << i 
               << " = " << values[indices[i]] << " " 
               << error << " " << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }
#endif

  LAMatrix mat(pols->getNumber(),lattice_points.getNumber());
  LAMatrix rhs(pols->getNumber(),1);
  for (int j=0;j<lattice_points.getNumber();j++) {
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[j],values);
    } else pols->computeValues(lattice_points[j],values);
    for (int i=0;i<values.getNumber();i++) mat(i,j)=values[i];
    rhs(j,0)=0.;
  }
//make sure that matrix is nonsingular
  LAMatrix *soln=factorNSolve(mat,rhs); 
  delete soln;
}

#if (SPACEDIM==2)
void debugOrderCase(const Shape1 &e,const ShapeFunction<SPACEDIM> *pols,
int degree0,int degree1) {
//TRACER_CALL(z,"debugOrderCase");
  CHECK_POINTER(dynamic_cast<const C0LagrangeQuadrilateralPolynomials*>(pols));
  bool requires_shape_to_compute_values=
    pols->requiresShapeToComputeValues();
  const Quadrilateral *e2=dynamic_cast<const Quadrilateral*>(&e);
  CHECK_POINTER(e2);

  if (!requires_shape_to_compute_values) {
    NumPtr<int> shape_renumber;
    pols->shapeFunctionsOrder(&e,shape_renumber);
#ifdef DEBUG
//  for (int i=0;i<shape_renumber.getNumber();i++) {
//    cout << "\tshape_renumber[" << i << "] = " << shape_renumber[i]
//         << endl;
//  }
#endif
    NumPtr<int> renumber_inverse(shape_renumber.getNumber());
    renumber_inverse.initialize(INT_MAX);
    for (int i=0;i<shape_renumber.getNumber();i++) {
      renumber_inverse[abs(shape_renumber[i])]=i;
    }
    for (int i=0;i<renumber_inverse.getNumber();i++) { 
      CHECK_TEST(renumber_inverse[i]!=INT_MAX);
    }
  }

  NumPtr<Point<SPACEDIM> > lattice_points;
  e2->staticEquallySpacedLatticePoints(degree0,degree1,lattice_points);

  for (int v=0;v<e.numberVertices();v++) {
//  TRACER_CALL(z,"debugOrderCase");
    int qv=e2->staticEquallySpacedLatticePointIndexOfVertex(v,degree0,
      degree1);
    int value_index=qv;
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[qv],values);
    } else pols->computeValues(lattice_points[qv],values);
    REAL error=abs(values[value_index]-1.);
    if (error>512.*DBL_EPSILON) {
      cout << "\tvertex shape value at vertex " << v << " = "
           << values[value_index] << " " << error << " "
           << error/DBL_EPSILON << endl;
      CHECK_TEST(error<=DBL_EPSILON*512.);
    }
    for (int v2=0;v2<e.numberVertices();v2++) {
      if (v2==v) continue;
      value_index=e2->staticEquallySpacedLatticePointIndexOfVertex(v2,
        degree0,degree1);
      REAL error=abs(values[value_index]);
      if (error>512.*DBL_EPSILON) {
        cout << "\tvertex shape value at vertex " << v << " = "
             << values[value_index] << " " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
    }
  }
  for (int f=0;f<e.numberCodimension1Shapes();f++) {
//  TRACER_CALL(z,"debugOrderCase");
    NumPtr<int> indices; // fe order
    e2->staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
      degree0,degree1,indices);
#ifdef DEBUG
//  e2->printOn(cout);
//  cout << "\n\tface = " << f << endl;
//  cout << "\n\tdegree = " << degree0 << " " << degree1 << endl;
//  for (int i=0;i<indices.getNumber();i++) {
//    cout << "\tface indices[" << i << "] = " << indices[i] << endl;
//  }
#endif
    for (int v=0;v<e.numberVertices();v++) {
      int qv=e2->staticEquallySpacedLatticePointIndexOfVertex(v,degree0,
        degree1); //fe order
      NumPtr<REAL> values; // fe order
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[qv],values);
      } else pols->computeValues(lattice_points[qv],values);
#ifdef DEBUG
//    cout << "\tlattice_points[" << qv << "] = " << lattice_points[qv]
//         << endl;
//    for (int i=0;i<indices.getNumber();i++) {
//      cout << "\tvalues[" << indices[i] << "] = "
//           << values[indices[i]] << endl;
//    }
#endif
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tface shape value at vertex " << v << " = "
               << values[indices[i]] << " " << error << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
    for (int f2=0;f2<e.numberCodimension1Shapes();f2++) {
      if (f2==f) continue;
      NumPtr<int> indices2; // fe_order
      e2->equallySpacedLatticePointIndicesOfFaceInterior(f2,degree0,
        degree1,indices2);
#ifdef DEBUG
//    cout << "\n\tf2 = " << f2 << endl;
//    for (int i=0;i<indices2.getNumber();i++) {
//      cout << "\tface indices2[" << i << "] = " << indices2[i] << endl;
//    }
#endif
      for (int i=0;i<indices2.getNumber();i++) {
        NumPtr<REAL> values;
        if (requires_shape_to_compute_values) {
          pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
        } else pols->computeValues(lattice_points[indices2[i]],values);
#ifdef DEBUG
//      cout << "\tlattice_points[" << indices2[i] << "] = " 
//           << lattice_points[indices2[i]] << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tvalues[" << indices[i] << "] = "
//             << values[indices[i]] << endl;
//      }
#endif
        for (int i=0;i<indices.getNumber();i++) {
          REAL error=abs(values[indices[i]]);
          if (error>512.*DBL_EPSILON) {
            cout << "\tface shape value at face " << f2 << " #" << i 
                 << " = " << values[indices[i]] << " " << error 
                 << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
    }
  }
  NumPtr<int> indices; // fe order
  e2->staticInteriorLatticePointIndices(degree0,degree1,indices);
#ifdef DEBUG
//for (int i=0;i<indices.getNumber();i++) {
//  cout << "\tinterior indices[" << i << "] = " << indices[i] << endl;
//}
#endif
  for (int v=0;v<e.numberVertices();v++) { // v is in dof order
//  TRACER_CALL(z,"debugOrderCase");
    int qv=e2->staticEquallySpacedLatticePointIndexOfVertex(v,degree0,
      degree1); // fe order
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[qv],values);
    } else pols->computeValues(lattice_points[qv],values);
#ifdef DEBUG
//  cout << "\tv = " << v << " qv = " << qv << endl;
//  cout << "\tvertex = " << lattice_points[qv] << endl;
//  for (int i=0;i<indices.getNumber();i++) {
//    cout << "\tvalues[" << indices[i] << "] = " << values[indices[i]] 
//         << endl;
//  }
#endif
    for (int i=0;i<indices.getNumber();i++) {
      REAL error=abs(values[indices[i]]);
      if (error>512.*DBL_EPSILON) {
        cout << "\tinterior shape value at index " << indices[i] << " = "
             << values[indices[i]] << " " << error 
             << " " << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
    }
  }
  for (int f=0;f<e.numberCodimension1Shapes();f++) {
    NumPtr<int> indices2;
    e2->equallySpacedLatticePointIndicesOfFaceInterior(f,degree0,degree1,
      indices2);
#ifdef DEBUG
//  cout << "\tface = " << f << endl;
//  for (int i=0;i<indices2.getNumber();i++) {
//    cout << "\tface indices[" << i << "] = " << indices2[i] << endl;
//  }
#endif
    for (int i=0;i<indices2.getNumber();i++) {
      NumPtr<REAL> values;
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
      } else pols->computeValues(lattice_points[indices2[i]],values);
#ifdef DEBUG
//    cout << "\tlattice_points[" << indices2[i] << "] = " 
//         << lattice_points[indices2[i]] << endl;
//    for (int j=0;j<indices.getNumber();j++) {
//      cout << "\tvalues[" << indices[j] << "] = " << values[indices[j]] 
//           << endl;
//    }
#endif
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tinterior shape value at face " << f << " #" << i 
               << " = " << values[indices[i]] << " " 
               << error << " " << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }

  LAMatrix mat(pols->getNumber(),lattice_points.getNumber());
  LAMatrix rhs(pols->getNumber(),1);
  for (int j=0;j<lattice_points.getNumber();j++) {
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[j],values);
    } else pols->computeValues(lattice_points[j],values);
    for (int i=0;i<values.getNumber();i++) mat(i,j)=values[i];
    rhs(j,0)=0.;
  }
//make sure that matrix is nonsingular
  LAMatrix *soln=factorNSolve(mat,rhs); 
  delete soln;
}
#endif

#if (SPACEDIM==3)
void debugOrderCase(const Shape1 &e,const ShapeFunction<SPACEDIM> *pols,
int degree0,int degree1,int degree2) {
//TRACER_CALL(z,"debugOrderCase");
  CHECK_POINTER(dynamic_cast<const C0LagrangeHexahedronPolynomials*>(pols));
  const Hexahedron *e3=dynamic_cast<const Hexahedron*>(&e);
  CHECK_POINTER(e3);
  bool requires_shape_to_compute_values=
    pols->requiresShapeToComputeValues();
  const Shape2 *e2=dynamic_cast<const Shape2*>(&e);

  if (!requires_shape_to_compute_values) {
    NumPtr<int> shape_renumber;
    pols->shapeFunctionsOrder(&e,shape_renumber);
#ifdef DEBUG
//  for (int i=0;i<shape_renumber.getNumber();i++) {
//    cout << "\tshape_renumber[" << i << "] = " << shape_renumber[i]
//         << endl;
//  }
#endif
    NumPtr<int> renumber_inverse(shape_renumber.getNumber());
    renumber_inverse.initialize(INT_MAX);
    for (int i=0;i<shape_renumber.getNumber();i++) {
      renumber_inverse[abs(shape_renumber[i])]=i;
    }
    for (int i=0;i<renumber_inverse.getNumber();i++) { 
      CHECK_TEST(renumber_inverse[i]!=INT_MAX);
    }
  }

  NumPtr<Point<SPACEDIM> > lattice_points;
  e3->staticEquallySpacedLatticePoints(degree0,degree1,degree2,
    lattice_points);

  for (int v=0;v<e.numberVertices();v++) {
//  TRACER_CALL(z,"debugOrderCase");
    int qv=e3->staticEquallySpacedLatticePointIndexOfVertex(v,degree0,
      degree1,degree2);
    int value_index=qv;
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[qv],values);
    } else pols->computeValues(lattice_points[qv],values);
    REAL error=abs(values[value_index]-1.);
    if (error>512.*DBL_EPSILON) {
      cout << "\tvertex shape value at vertex " << v << " = "
           << values[value_index] << " " << error << " "
           << error/DBL_EPSILON << endl;
      CHECK_TEST(error<=DBL_EPSILON*512.);
    }
    for (int v2=0;v2<e.numberVertices();v2++) {
      if (v2==v) continue;
      value_index=e3->staticEquallySpacedLatticePointIndexOfVertex(v2,
        degree0,degree1,degree2);
      REAL error=abs(values[value_index]);
      if (error>512.*DBL_EPSILON) {
        cout << "\tvertex shape value at vertex " << v << " = "
             << values[value_index] << " " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
    }
  }
  for (int l=0;l<e3->numberCodimension2Shapes();l++) {
    NumPtr<int> indices; // fe order
    e3->equallySpacedLatticePointIndicesOfLineInterior(l,degree0,degree1,
      degree2,indices);
    for (int v=0;v<e.numberVertices();v++) {
      int qv=e3->staticEquallySpacedLatticePointIndexOfVertex(v,degree0,
        degree1,degree2);
      NumPtr<REAL> values;
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[qv],values);
      } else pols->computeValues(lattice_points[qv],values);
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tedge shape value at vertex " << v << " = "
               << values[indices[i]] << " " << error << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
    for (int l2=0;l2<e3->numberCodimension2Shapes();l2++) {
      if (l2==l) continue;
      NumPtr<int> indices2;
      e3->equallySpacedLatticePointIndicesOfLineInterior(l2,degree0,
        degree1,degree2,indices2);
      for (int i=0;i<indices2.getNumber();i++) {
        NumPtr<REAL> values;
        if (requires_shape_to_compute_values) {
          pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
        } else {
          pols->computeValues(lattice_points[indices2[i]],values);
        }
        for (int i=0;i<indices.getNumber();i++) {
          REAL error=abs(values[indices[i]]);
          if (error>512.*DBL_EPSILON) {
            cout << "\tedge shape value at line " << l2 << " #" << i 
                 << " = " << values[indices[i]] << " " 
                 << error << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
    }
  }
  for (int f=0;f<e.numberCodimension1Shapes();f++) {
    NumPtr<int> indices; // fe order
    e3->equallySpacedLatticePointIndicesOfFaceInterior(f,degree0,degree1,
      degree2,indices);
#ifdef DEBUG
//  cout << "\n\tface = " << f << endl;
//  for (int i=0;i<indices.getNumber();i++) {
//    cout << "\tface indices[" << i << "] = " << indices[i] << endl;
//  }
#endif
    for (int v=0;v<e.numberVertices();v++) {
      int qv=e3->staticEquallySpacedLatticePointIndexOfVertex(v,degree0,
        degree1,degree2); //fe order
      NumPtr<REAL> values; // fe order
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[qv],values);
      } else pols->computeValues(lattice_points[qv],values);
#ifdef DEBUG
//    cout << "\tlattice_points[" << qv << "] = " << lattice_points[qv]
//         << endl;
//    for (int i=0;i<indices.getNumber();i++) {
//      cout << "\tvalues[" << indices[i] << "] = "
//           << values[indices[i]] << endl;
//    }
#endif
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tface shape value at vertex " << v << " = "
               << values[indices[i]] << " " << error << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
    for (int l=0;l<e3->numberCodimension2Shapes();l++) {
      NumPtr<int> indices2;
      e3->equallySpacedLatticePointIndicesOfLineInterior(l,degree0,
        degree1,degree2,indices2);
      for (int i=0;i<indices2.getNumber();i++) {
        NumPtr<REAL> values;
        if (requires_shape_to_compute_values) {
          pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
        } else pols->computeValues(lattice_points[indices2[i]],values);
        for (int i=0;i<indices.getNumber();i++) {
          REAL error=abs(values[indices[i]]);
          if (error>512.*DBL_EPSILON) {
            cout << "\tface shape value at line " << l << " #" << i 
                 << " = " << values[indices[i]] << " " 
                 << error << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
    }
    for (int f2=0;f2<e.numberCodimension1Shapes();f2++) {
      if (f2==f) continue;
      NumPtr<int> indices2; // fe_order
      e3->equallySpacedLatticePointIndicesOfFaceInterior(f2,degree0,
        degree1,degree2,indices2);
#ifdef DEBUG
//    cout << "\n\tf2 = " << f2 << endl;
//    for (int i=0;i<indices2.getNumber();i++) {
//      cout << "\tface indices2[" << i << "] = " << indices2[i] << endl;
//    }
#endif
      for (int i=0;i<indices2.getNumber();i++) {
        NumPtr<REAL> values;
        if (requires_shape_to_compute_values) {
          pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
        } else pols->computeValues(lattice_points[indices2[i]],values);
#ifdef DEBUG
//      cout << "\tlattice_points[" << indices2[i] << "] = " 
//           << lattice_points[indices2[i]] << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tvalues[" << indices[i] << "] = "
//             << values[indices[i]] << endl;
//      }
#endif
        for (int i=0;i<indices.getNumber();i++) {
          REAL error=abs(values[indices[i]]);
          if (error>512.*DBL_EPSILON) {
            cout << "\tface shape value at face " << f2 << " #" << i 
                 << " = " << values[indices[i]] << " " << error 
                 << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
    }
  }
  NumPtr<int> indices; // fe order
  e3->staticInteriorLatticePointIndices(degree0,degree1,degree2,indices);
#ifdef DEBUG
//for (int i=0;i<indices.getNumber();i++) {
//  cout << "\tinterior indices[" << i << "] = " << indices[i] << endl;
//}
#endif
  for (int v=0;v<e.numberVertices();v++) { // v is in dof order
//  TRACER_CALL(z,"debugOrderCase");
    int qv=e3->staticEquallySpacedLatticePointIndexOfVertex(v,degree0,
      degree1,degree2); // fe order
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[qv],values);
    } else pols->computeValues(lattice_points[qv],values);
#ifdef DEBUG
//  cout << "\tv = " << v << " qv = " << qv << endl;
//  cout << "\tvertex = " << lattice_points[qv] << endl;
//  for (int i=0;i<indices.getNumber();i++) {
//    cout << "\tvalues[" << indices[i] << "] = " << values[indices[i]] 
//         << endl;
//  }
#endif
    for (int i=0;i<indices.getNumber();i++) {
      REAL error=abs(values[indices[i]]);
      if (error>512.*DBL_EPSILON) {
        cout << "\tinterior shape value at index " << indices[i] << " = "
             << values[indices[i]] << " " << error 
             << " " << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
    }
  }
  for (int l=0;l<e3->numberCodimension2Shapes();l++) {
    NumPtr<int> indices2;
    e3->equallySpacedLatticePointIndicesOfLineInterior(l,degree0,degree1,
      degree2,indices2);
#ifdef DEBUG
//  cout << "\tline = " << l << endl;
//  for (int i=0;i<indices.getNumber();i++) {
//    cout << "\tline indices[" << i << "] = " << indices2[i] << endl;
//  }
#endif
    for (int i=0;i<indices2.getNumber();i++) {
      NumPtr<REAL> values;
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
      } else pols->computeValues(lattice_points[indices2[i]],values);
#ifdef DEBUG
//    cout << "\tlattice_points[" << indices2[i] << "] = " 
//         << lattice_points[indices2[i]] << endl;
//    for (int j=0;j<indices.getNumber();j++) {
//      cout << "\tvalues[" << indices[j] << "] = " << values[indices[j]] 
//           << endl;
//    }
#endif
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>DBL_EPSILON) {
          cout << "\tinterior shape value at line " << l << " #" << i 
               << " = " << values[indices[i]] << " " 
               << error << " " << error/DBL_EPSILON << endl;
        }
      }
    }
  }
  for (int f=0;f<e.numberCodimension1Shapes();f++) {
    NumPtr<int> indices2;
    e3->equallySpacedLatticePointIndicesOfFaceInterior(f,degree0,degree1,
      degree2,indices2);
#ifdef DEBUG
//  cout << "\tface = " << f << endl;
//  for (int i=0;i<indices2.getNumber();i++) {
//    cout << "\tface indices[" << i << "] = " << indices2[i] << endl;
//  }
#endif
    for (int i=0;i<indices2.getNumber();i++) {
      NumPtr<REAL> values;
      if (requires_shape_to_compute_values) {
        pols->computeValuesFor(&e,lattice_points[indices2[i]],values);
      } else pols->computeValues(lattice_points[indices2[i]],values);
#ifdef DEBUG
//    cout << "\tlattice_points[" << indices2[i] << "] = " 
//         << lattice_points[indices2[i]] << endl;
//    for (int j=0;j<indices.getNumber();j++) {
//      cout << "\tvalues[" << indices[j] << "] = " << values[indices[j]] 
//           << endl;
//    }
#endif
      for (int i=0;i<indices.getNumber();i++) {
        REAL error=abs(values[indices[i]]);
        if (error>512.*DBL_EPSILON) {
          cout << "\tinterior shape value at face " << f << " #" << i 
               << " = " << values[indices[i]] << " " 
               << error << " " << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }

  LAMatrix mat(pols->getNumber(),lattice_points.getNumber());
  LAMatrix rhs(pols->getNumber(),1);
  for (int j=0;j<lattice_points.getNumber();j++) {
    NumPtr<REAL> values;
    if (requires_shape_to_compute_values) {
      pols->computeValuesFor(&e,lattice_points[j],values);
    } else pols->computeValues(lattice_points[j],values);
    for (int i=0;i<values.getNumber();i++) mat(i,j)=values[i];
    rhs(j,0)=0.;
  }
//make sure that matrix is nonsingular
  LAMatrix *soln=factorNSolve(mat,rhs); 
  delete soln;
}
#endif

void debugOrder(const ShapeFunction<SPACEDIM> *pols,int degree) {
//TRACER_CALL(z,"debugOrder");
#if (SPACEDIM==1)
  PermutationGenerator pg(2);
  while (pg.hasMore()) {
    const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//  cout << "\tp =";
//  for (int i=0;i<p.getNumber();i++) cout << " " << p[i];
//  cout << endl;
#endif
    int v[2];
    for (int i=0;i<2;i++) v[i]=p[i];
    Interval e(0,v);
    debugOrderCase(e,pols,degree);
  }
#elif (SPACEDIM==2)
  if (dynamic_cast<const TrianglePolynomials*>(pols)) {
    PermutationGenerator pg(3);
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
      int v[3];
      for (int i=0;i<3;i++) v[i]=p[i];
      Triangle e(0,v,v);
      debugOrderCase(e,pols,degree);
    }
  } else {
    PermutationGenerator pg(4);
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
      int v[4];
      for (int i=0;i<4;i++) v[i]=p[i];
      Quadrilateral e(0,v,v);
      debugOrderCase(e,pols,degree);
    }
  }
#else
  if (dynamic_cast<const TetrahedronPolynomials*>(pols)) {
    PermutationGenerator pg(4);
    int l[6];
    for (int i=0;i<6;i++) l[i]=i;
    int v[4];
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
      for (int i=0;i<4;i++) v[i]=p[i];
      Tetrahedron e(0,l,v,v);
      debugOrderCase(e,pols,degree);
    }
  } else if (dynamic_cast<const PrismPolynomials*>(pols)) {
//  TRACER_CALL(z,"debugOrder");
    PermutationGenerator pg(6);
    int l[9];
    int f[5];
    int v[6];
    for (int i=0;i<5;i++) f[i]=i;
    for (int i=0;i<9;i++) l[i]=i;
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//    cout << "\tp =";
//    for (int i=0;i<p.getNumber();i++) cout << " " << p[i];
//    cout << endl;
#endif
      for (int i=0;i<6;i++) v[i]=p[i];
      Prism e(0,l,f,v);
      debugOrderCase(e,pols,degree);
    }
  } else {
//  TRACER_CALL(z,"debugOrder");
//  PermutationGenerator pg(8);
//  while (pg.hasMore()) 
    for (int count=0;count<1000;count++)
    {
//    const NumPtr<int> &p=pg.getNext();
      NumPtr<int> p(8);
      RandomPermutation(p);
//    p[0]=0;
//    p[1]=1;
//    p[2]=2;
//    p[3]=3;
//    p[4]=5;
//    p[5]=6;
//    p[6]=4;
//    p[7]=7;
#ifdef DEBUG
//    cout << "\tp =";
//    for (int i=0;i<p.getNumber();i++) cout << " " << p[i];
//    cout << endl;
#endif
      int l[12];
      int f[6];
      int v[8];
      for (int i=0;i<8;i++) v[i]=p[i];
      for (int i=0;i<6;i++) f[i]=i;
      for (int i=0;i<12;i++) l[i]=i;
      Hexahedron e(0,l,f,v);
      debugOrderCase(e,pols,degree);
    }
  }
#endif
}

#if (SPACEDIM==2)
void debugOrder(const ShapeFunction<SPACEDIM> *pols,int degree0,
int degree1) {
//TRACER_CALL(z,"debugOrder");
  CHECK_POINTER(dynamic_cast<const C0LagrangeQuadrilateralPolynomials*>(pols));
  PermutationGenerator pg(4);
  while (pg.hasMore()) {
    const NumPtr<int> &p=pg.getNext();
    int v[4];
    for (int i=0;i<4;i++) v[i]=p[i];
    Quadrilateral e(0,v,v);
    debugOrderCase(e,pols,degree0,degree1);
  }
}
#endif

#if (SPACEDIM==3)
const ShapeFunction<SPACEDIM> *plot_polys=0;
int plot_shape=-1;
int plot_nv=-1;
REAL surface_function(const Vector3 &v) {
  CHECK_POINTER(plot_polys);
  int n=plot_polys->getNumber();
  NumPtr<REAL> vals(n);
  Point<SPACEDIM> p(v[0],v[1],v[2]);
  plot_polys->computeValues(p,vals);
  if (plot_nv==8) return vals[plot_shape];
  if (plot_nv==6) return (v[0]+v[1]<=0.5 ? vals[plot_shape] : 0.);
  else return (v[0]+v[1]+v[2]<=0.5 ? vals[plot_shape] : 0.);
}
#endif

void plotPoly(const ShapeFunction<SPACEDIM> *pols,int nv,
VirtualGraphTool *g) {
//TRACER_CALL(t,"plotPoly");
  int n=pols->getNumber();
#if (SPACEDIM==1)
  XYGraphTool *gt=dynamic_cast<XYGraphTool*>(g);
  CHECK_POINTER(gt);
  gt->setbgColor("white");
  REAL dx=gt->getPixelWidth();
  Point<1> p;
  REAL xmin=0.;
  REAL xmax=1.;
  REAL ymin=DBL_MAX;
  REAL ymax=-DBL_MAX;
  for (REAL x=xmin;x<=xmax;x+=dx) {
    p[0]=x;
    NumPtr<REAL> v;
    pols->computeValues(p,v);
    for (int s=0;s<n;s++) {
      ymin=min(ymin,v[s]);
      ymax=max(ymax,v[s]);
    }
  }
  gt->rescale(xmin,xmax,ymin,ymax);
  gt->newPage();
  gt->setfgColor("black");
  gt->drawAxes();
  for (int s=0;s<n;s++) {
    gt->setfgColor(s+1,n+1);
    for (REAL x=xmin;x<=xmax;x+=dx) {
      p[0]=x;
      NumPtr<REAL> v;
      pols->computeValues(p,v);
      if (x<=xmin) gt->movePen(x,v[s]);
      else gt->drawLine(x,v[s]);
    }
  }
  gt->flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
#elif (SPACEDIM==2)
  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(g);
  CHECK_POINTER(vgt);
  Vector3 low(0.,0.,DBL_MAX);
  Vector3 high(1.,1.,-DBL_MAX);
  int nptsf=100;
  int nptsc=10;
  for (int s=0;s<n;s++) {
    cout << "\tplotting shape " << s << endl;
    vgt->newPage();
    REAL dxf=(high[0]-low[0])/static_cast<REAL>(nptsf);
    REAL dyf=(high[1]-low[1])/static_cast<REAL>(nptsf);
    REAL dxc=(high[0]-low[0])/static_cast<REAL>(nptsc);
    REAL dyc=(high[1]-low[1])/static_cast<REAL>(nptsc);
    low[2]=DBL_MAX;
    high[2]=-DBL_MAX;
    Point<2> p;
    if (nv==4) {
      REAL *valuesf=OPERATOR_NEW_BRACKET(REAL,(nptsf+1)*(nptsf+1));
      int ij=0;
      for (int j=0;j<=nptsf;j++) {
        p[1]=dyf*static_cast<REAL>(j); 
        for (int i=0;i<=nptsf;i++) {
          p[0]=dxf*static_cast<REAL>(i); 
          NumPtr<REAL> v;
          pols->computeValues(p,v);
          valuesf[ij++]=v[s];
          low[2]=min(low[2],v[s]);
          high[2]=max(high[2],v[s]);
        }
      }
      REAL dif=(high[2]-low[2])*.005;
      low[2]-=dif;
      high[2]+=dif;
      vgt->rescale(low,high);
      REAL *valuesc=OPERATOR_NEW_BRACKET(REAL,(nptsc+1)*(nptsc+1));
      ij=0;
      for (int j=0;j<=nptsc;j++) {
        p[1]=dyc*static_cast<REAL>(j); 
        for (int i=0;i<=nptsc;i++) {
          p[0]=dxc*static_cast<REAL>(i); 
          NumPtr<REAL> v;
          pols->computeValues(p,v);
          valuesc[ij++]=v[s];
        }
      }
      REAL valdif=1./(high[2]-low[2]);
      { VolGraphTool::DrawDelimiter dd(TRUE,vgt);
        vgt->drawBox(low,high);
      }
      vgt->newPage();
      vgt->expose();
      { BOOLEAN do_best=FALSE;
        VolGraphTool::DrawDelimiter dd(do_best,vgt);
        int ij=0;
        REAL y=0.;
        for (int j=0;j<nptsc;j++) {
          REAL x=0.;
          for (int i=0;i<=nptsc;i++) {
            Vector3 v0(x,y,valuesc[ij]);
            Vector3 v1(x+dxc,y,valuesc[ij+1]);
            Vector3 v2(x+dxc,y+dyc,valuesc[ij+nptsc+2]);
            Vector3 v3(x,y+dyc,valuesc[ij+nptsc+1]);
            Vector3 vave=(v0+v1+v2+v3)*0.25;

            REAL frac=((v0[2]+v1[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v0,v1,vave);

            frac=((v1[2]+v2[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v1,v2,vave);

            frac=((v2[2]+v3[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v2,v3,vave);

            frac=((v3[2]+v0[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v3,v0,vave);
            ij++;
            x+=dxc;
          }
          y+=dyc;
        }
      }
      { BOOLEAN do_best=TRUE;
        VolGraphTool::DrawDelimiter dd(do_best,vgt);
        int ij=0;
        REAL y=0.;
        for (int j=0;j<nptsf;j++) {
          REAL x=0.;
          for (int i=0;i<=nptsf;i++) {
            Vector3 v0(x,y,valuesf[ij]);
            Vector3 v1(x+dxf,y,valuesf[ij+1]);
            Vector3 v2(x+dxf,y+dyf,valuesf[ij+nptsf+2]);
            Vector3 v3(x,y+dyf,valuesf[ij+nptsf+1]);
            Vector3 vave=(v0+v1+v2+v3)*0.25;

            REAL frac=((v0[2]+v1[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v0,v1,vave);

            frac=((v1[2]+v2[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v1,v2,vave);

            frac=((v2[2]+v3[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v2,v3,vave);

            frac=((v3[2]+v0[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v3,v0,vave);
            ij++;
            x+=dxf;
          }
          y+=dyf;
        }
      }
      vgt->eventLoop(FALSE);
      delete [] valuesf; valuesf=0;
      delete [] valuesc; valuesc=0;
    } else {
//    TRACER_CALL(t,"plotPoly");
      REAL *valuesf=OPERATOR_NEW_BRACKET(REAL,((nptsf+1)*(nptsf+2))/2);
      int ij=0;
      for (int j=0;j<=nptsf;j++) {
        p[1]=dyf*static_cast<REAL>(j); 
        for (int i=0;i+j<=nptsf;i++) {
          p[0]=dxf*static_cast<REAL>(i); 
          NumPtr<REAL> v;
          pols->computeValues(p,v);
          valuesf[ij++]=v[s];
          low[2]=min(low[2],v[s]);
          high[2]=max(high[2],v[s]);
        }
      }
      REAL dif=(high[2]-low[2])*.005;
      low[2]-=dif;
      high[2]+=dif;
      vgt->rescale(low,high);
      REAL *valuesc=OPERATOR_NEW_BRACKET(REAL,((nptsc+1)*(nptsc+2))/2);
      ij=0;
      for (int j=0;j<=nptsc;j++) {
        p[1]=dyc*static_cast<REAL>(j); 
        for (int i=0;i+j<=nptsc;i++) {
          p[0]=dxc*static_cast<REAL>(i); 
          NumPtr<REAL> v;
          pols->computeValues(p,v);
          valuesc[ij++]=v[s];
        }
      }
      REAL valdif=1./(high[2]-low[2]);
      { VolGraphTool::DrawDelimiter dd(TRUE,vgt);
        vgt->drawBox(low,high);
      }
      vgt->newPage();
      vgt->expose();
      { BOOLEAN do_best=FALSE;
        VolGraphTool::DrawDelimiter dd(do_best,vgt);
        int ij=0;
        REAL y=0.;
        for (int j=0;j<nptsc;j++) {
          REAL x=0.;
          for (int i=0;i+j<nptsc;i++) {
            Vector3 v0(x,y,valuesc[ij]);
            Vector3 v1(x+dxc,y,valuesc[ij+1]);
            Vector3 v2(x+dxc,y+dyc,valuesc[ij+nptsc+2-j]);
            Vector3 v3(x,y+dyc,valuesc[ij+nptsc+1-j]);
            Vector3 vave=(v0+v1+v2+v3)*0.25;
            if (i+j+1>=nptsc) vave=(v1+v3)*0.5;

            REAL frac=((v0[2]+v1[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v0,v1,vave);

            if (i+j+1<nptsc) {
              frac=((v1[2]+v2[2]+vave[2])/3.-low[2])*valdif;
              frac=max(0.,min(1.,frac));
              vgt->colorTriangle(frac,v1,v2,vave);

              frac=((v2[2]+v3[2]+vave[2])/3.-low[2])*valdif;
              frac=max(0.,min(1.,frac));
              vgt->colorTriangle(frac,v2,v3,vave);
            }

            frac=((v3[2]+v0[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v3,v0,vave);
            ij++;
            x+=dxc;
          }
          ij++;
          y+=dyc;
        }
      }
      { BOOLEAN do_best=TRUE;
        VolGraphTool::DrawDelimiter dd(do_best,vgt);
        int ij=0;
        REAL y=0.;
        for (int j=0;j<nptsf;j++) {
          REAL x=0.;
          for (int i=0;i+j<nptsf;i++) {
            Vector3 v0(x,y,valuesf[ij]);
            Vector3 v1(x+dxf,y,valuesf[ij+1]);
            Vector3 v2(x+dxf,y+dyf,valuesf[ij+nptsf+2-j]);
            Vector3 v3(x,y+dyf,valuesf[ij+nptsf+1-j]);
            Vector3 vave=(v0+v1+v2+v3)*0.25;
            if (i+j+1>=nptsf) vave=(v1+v3)*0.5;

            REAL frac=((v0[2]+v1[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v0,v1,vave);

            if (i+j+1<nptsf) {
              frac=((v1[2]+v2[2]+vave[2])/3.-low[2])*valdif;
              frac=max(0.,min(1.,frac));
              vgt->colorTriangle(frac,v1,v2,vave);

              frac=((v2[2]+v3[2]+vave[2])/3.-low[2])*valdif;
              frac=max(0.,min(1.,frac));
              vgt->colorTriangle(frac,v2,v3,vave);
            }

            frac=((v3[2]+v0[2]+vave[2])/3.-low[2])*valdif;
            frac=max(0.,min(1.,frac));
            vgt->colorTriangle(frac,v3,v0,vave);
            ij++;
            x+=dxf;
          }
          ij++;
          y+=dyf;
        }
      }
      vgt->eventLoop(FALSE);
      delete [] valuesf; valuesf=0;
    }
  }
#else
  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(g);
  CHECK_POINTER(vgt);
  int nptsf=20;
  int nptsc=5;
  plot_polys=pols;
  plot_nv=nv;
  Vector3 low(0.,0.,0.);
  Vector3 high(1.,1.,1.);
  vgt->rescale(low,high);
  REAL dx=1./static_cast<REAL>(nptsf);
  REAL vmax=-DBL_MAX;
  REAL vmin=DBL_MAX;
  Point<SPACEDIM> p;
  for (int s=0;s<n;s++) {
    int ijk=0;
    for (int k=0;k<=nptsf;k++) {
      p[2]=dx*static_cast<REAL>(k); 
      for (int j=0;j<=nptsf;j++) {
        p[1]=dx*static_cast<REAL>(j); 
        for (int i=0;i<=nptsf;i++) {
          p[0]=dx*static_cast<REAL>(i); 
          NumPtr<REAL> v;
          pols->computeValues(p,v);
          vmin=min(vmin,v[s]);
          vmax=max(vmax,v[s]);
        }
      }
    }

    plot_shape=s;
    vgt->newPage();
    {
      SurfaceDataSet dataset(vgt,low,high,nptsf,nptsc,surface_function,
        vmin,vmax);
      Surface<SurfaceDataSet> *object=
        OPERATOR_NEW Surface<SurfaceDataSet>(&dataset,vgt,16);
      vgt->initialize(object);
      vgt->eventLoop();
      vgt->terminatePlotObj();
    }
  }
#endif
}

int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif
    Palette pal;
#if (SPACEDIM==1)
    REAL xmin=0.;
    REAL xmax=1.;
    REAL ymin=DBL_MAX;
    REAL ymax=-DBL_MAX;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
    XYGraphTool *g=OPERATOR_NEW XYGraphTool("ShapeFunctions","x","y",
      xmin,xmax,ymin,ymax,&cmap,0,winsize);
#else
    Vector3 low(0.,0.,0.);
    Vector3 high(1.,1.,1.);
    BOOLEAN use_lighting=TRUE;
    BOOLEAN rotate_best=FALSE;
    BOOLEAN show_bounding_box=TRUE;
    VolGraphTool *g=OPERATOR_NEW VolGraphTool("ShapeFunction",low,high,
      pal,use_lighting,winsize,rotate_best,show_bounding_box);
#endif
/*
    {
      TRACER_CALL(t,"TensorProductPolynomials");
      for (int degree=1;degree<=D_CHOOSE(10,6,3);degree++) 
//    int degree=2;
      {
        cout << "\ndegree = " << degree << endl;
        C0LagrangePolynomial *lp=
          OPERATOR_NEW C0LagrangePolynomial(degree);
        TensorProductPolynomials<SPACEDIM> pols(lp,degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,1<<SPACEDIM,g);
      }
    }
*/
#if (SPACEDIM==1)
//
    {
      TRACER_CALL(t,"C0LagrangeIntervalPolynomials");
      for (int degree=1;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        C0LagrangeIntervalPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,2,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"HierarchicalIntervalPolynomials");
      for (int degree=1;degree<=10;degree++) 
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        HierarchicalIntervalPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,2,g);
      }
    }
//
#endif
#if (SPACEDIM==2)
//
    {
      TRACER_CALL(t,"C0LagrangeTrianglePolynomials");
      for (int degree=1;degree<=4;degree++) 
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        C0LagrangeTrianglePolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,3,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"C0LagrangeTrianglePolynomials");
      for (int degree=1;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        C0LagrangeTrianglePolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,3,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"C0LagrangeQuadrilateralPolynomials");
      for (int degree=1;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        C0LagrangeQuadrilateralPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,4,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"C0LagrangeQuadrilateralPolynomials");
      for (int degree1=1;degree1<=4;degree1++) {
        for (int degree0=1;degree0<=4;degree0++) {
          cout << "\tdegree = " << degree0 << " " << degree1
               << endl;
          C0LagrangeQuadrilateralPolynomials pols(degree0,degree1);
          debugPoly(&pols,degree0,degree1);
          debugDerivs(max(degree0,degree1),&pols);
          debugOrder(&pols,degree0,degree1);
//        plotPoly(&pols,4,g);
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"HierarchicalTrianglePolynomials");
      for (int degree=1;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        HierarchicalTrianglePolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,3,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"HierarchicalQuadrilateralPolynomials");
      for (int degree=1;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        HierarchicalQuadrilateralPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,4,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"HomogeneousPolynomials");
      for (int degree=1;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        HomogeneousPolynomials<2> pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
//      debugOrder(&pols,degree);
//      plotPoly(&pols,4,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"HomogeneousLagrangePolynomials2");
      for (int degree=1;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        HomogeneousLagrangePolynomials2 pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
//      debugOrder(&pols,degree);
//      plotPoly(&pols,4,g);
      }
    }
//
#endif
#if (SPACEDIM==3)
//
    {
      TRACER_CALL(t,"C0LagrangeTetrahedronPolynomials");
      for (int degree=1;degree<=6;degree++) 
//    int degree=5;
      {
        cout << "\tdegree = " << degree << endl;
        C0LagrangeTetrahedronPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,4,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"C0LagrangePrismPolynomials");
      for (int degree=1;degree<=5;degree++) 
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        C0LagrangePrismPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,6,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"C0LagrangeHexahedronPolynomials");
      for (int degree=1;degree<=3;degree++)
//    int degree=4;
      {
        cout << "\tdegree = " << degree << endl;
        C0LagrangeHexahedronPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,8,g);
      }
    }
//
//
//  {
//    TRACER_CALL(t,"HierarchicalTetrahedronPolynomials");
//    for (int degree=1;degree<=5;degree++) 
//    {
//      cout << "\tdegree = " << degree << endl;
//      HierarchicalTetrahedronPolynomials pols(degree);
//      debugOrder(&pols,degree);
//      debugDerivs(&pols,degree);
//      debugPoly(&pols,degree);
//    }
//  }
//
//
    {
      TRACER_CALL(t,"HierarchicalPrismPolynomials");
      for (int degree=1;degree<=4;degree++) 
//    int degree=4;
      {
        cout << "\tdegree = " << degree << endl;
        HierarchicalPrismPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,6,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"HierarchicalHexahedronPolynomials");
      for (int degree=1;degree<=4;degree++) 
//    int degree=4;
      {
        cout << "\tdegree = " << degree << endl;
        HierarchicalHexahedronPolynomials pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
        debugOrder(&pols,degree);
//      plotPoly(&pols,8,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"HomogeneousPolynomials");
      for (int degree=1;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        HomogeneousPolynomials<3> pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
//      debugOrder(&pols,degree);
//      plotPoly(&pols,4,g);
      }
    }
//
//
    {
      TRACER_CALL(t,"HomogeneousLagrangePolynomials3");
      for (int degree=1;degree<=9;degree++) {
        cout << "\tdegree = " << degree << endl;
        HomogeneousLagrangePolynomials3 pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(&pols,degree);
//      debugOrder(&pols,degree);
//      plotPoly(&pols,4,g);
      }
    }
//
#endif
    delete g;
  }
  return EXIT_SUCCESS;
}

#include "ShapeFunction.C"
#if (SPACEDIM==2)
//template class HomogeneousPolynomials<2>;
#endif
#if (SPACEDIM==3)
//template class HomogeneousPolynomials<3>;
template class Surface<SurfaceDataSet>;
template class ToggleButtonArrayCallbackStructure< Surface<SurfaceDataSet> >;
#endif
