// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <set>
#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Shape.H"
#include "Tracer.H"

void debugChild(const Shape &parent,const Shape &child,int c,
const NumPtr<int> &line_interior_child_vertices,
const Point<SPACEDIM> &cp) {
//TRACER_CALL(t,"debugChild");
  const D_CHOOSE(Shape1,Shape2,Shape3) *ps=
    dynamic_cast<const D_CHOOSE(Shape1,Shape2,Shape3)*>(&parent);
  const D_CHOOSE(Shape1,Shape2,Shape3) *cs=
    dynamic_cast<const D_CHOOSE(Shape1,Shape2,Shape3)*>(&child);
  Point<SPACEDIM> pp=ps->parentCoordinateForCanonical(cp,&child,c,
    line_interior_child_vertices);
#ifdef DEBUG
//cout << "\tc = " << c << endl;
//cout << "\tcp = " << cp << endl;
//cout << "\tpp = " << pp << endl;
#endif
  CHECK_TEST(ps->containsCanonical(pp));
  Point<SPACEDIM> cppp=ps->childCoordinateOfCanonical(pp,&child,c,
    line_interior_child_vertices);
#ifdef DEBUG
//cout << "\tcppp = " << cppp << endl;
#endif
  CHECK_TEST(cs->containsCanonical(cppp));
  REAL error=cp.distance(cppp)/sqrt(cp.square());
  if (error>16.*DBL_EPSILON) {
    parent.printOn(cout);
    child.printOn(cout);
    cout << "\tc = " << c << endl;
    cout << "\tcp = " << cp << endl;
    cout << "\tpp = " << pp << endl;
    cout << "\tcppp = " << cppp << endl;
    cout << "\terror = " << error/DBL_EPSILON << endl;
    OBSOLETE("fix bug");
  }
}
#if (SPACEDIM==2)
void debugTriangleChild(int c,int pv[3],int cv[3],
const NumPtr<int> &line_interior_child_vertices) {
  int l[3]={-1,-2,-3};
  Triangle parent(0,l,pv);
  Triangle child(1,l,cv);
  Point<2> cp(drand48(),0.);
  cp[1]=drand48()*(1.-cp[0]);
  debugChild(parent,child,c,line_interior_child_vertices,cp);
}
void debugQuadrilateralChild(int c,int pv[4],int cv[4],
const NumPtr<int> &line_interior_child_vertices) {
  int l[4]={-1,-2,-3,-4};
  Quadrilateral parent(0,l,pv);
  Quadrilateral child(1,l,cv);
  Point<2> cp(drand48(),drand48());
  debugChild(parent,child,c,line_interior_child_vertices,cp);
}
#endif
#if (SPACEDIM==3)
void debugTetrahedronChild(int c,int pv[4],int cv[4],
const NumPtr<int> &line_interior_child_vertices) {
  int l[6]={-1,-2,-3,-4,-5,-6};
  int f[4]={-1,-2,-3};
  Tetrahedron parent(1,l,f,pv);
  Tetrahedron child(1,l,f,cv);
  Point<3> cp(drand48(),0.,0.);
  cp[1]=drand48()*(1.-cp[0]);
  cp[2]=drand48()*(1.-cp[0]-cp[1]);
  debugChild(parent,child,c,line_interior_child_vertices,cp);
}
void debugPrismChild(int c,int pv[6],int cv[6],
const NumPtr<int> &line_interior_child_vertices) {
  int f[5]={-1,-2,-3,-4,-5};
  int l[9]={-1,-2,-3,-4,-5,-6,-7,-8,-9};
  Prism parent(1,l,f,pv);
  Prism child(1,l,f,cv);
  Point<3> cp(drand48(),0.,drand48());
  cp[1]=drand48()*(1.-cp[0]);
  debugChild(parent,child,c,line_interior_child_vertices,cp);
}
void debugHexahedronChild(int c,int pv[8],int cv[8],
const NumPtr<int> &line_interior_child_vertices) {
  int f[6]={-1,-2,-3,-4,-5,-6};
  int l[12]={-1,-2,-3,-4,-5,-6,-7,-8,-9,-10,-11,-12};
  Hexahedron parent(1,l,f,pv);
  Hexahedron child(1,l,f,cv);
  Point<3> cp(drand48(),drand48(),drand48());
  debugChild(parent,child,c,line_interior_child_vertices,cp);
}
#endif

int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  for (int nv=2;nv<=8;nv++) {
//    PermutationGenerator pg(nv);
//    while (pg.hasMore()) {
//      const NumPtr<int> &p=pg.getNext();
//      for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//      cout << endl;
//    }
//    cout << endl;
//  }

#if (SPACEDIM==1)
    { 
      TRACER_CALL(t,"Interval");
      int pv[2]={0,1};
      Interval parent(0,pv);
      int cv[2][2]={ {0,2}, {2,1} };
      NumPtr<Point<SPACEDIM> > vertices(3);
        vertices[0]=Point<SPACEDIM>(0.);
        vertices[1]=Point<SPACEDIM>(1.);
        vertices[2]=Point<SPACEDIM>(.5);
      NumPtr<int> line_interior_child_vertices(0);
      for (int c=0;c<childrenPerShape();c++) {
        Interval child(0,cv[c]);
        const Shape *chile=
          child.reorientFor(&parent,c,line_interior_child_vertices);
        CHECK_TEST(chile->volume(vertices)>0.);
        for (int i=0;i<chile->numberVertices();i++) {
          Point<SPACEDIM> cp=Interval::canonicalVertex(i);
          CHECK_SAME(vertices[chile->vertexIndex(i)],
                     Interval::parentCoordinateOfCanonical(cp,c));
        }
        if (chile!=&child) delete chile;
      }
    }
    { 
      TRACER_CALL(t,"Interval");
      PermutationGenerator pg(3);
      NumPtr<int> licv;
      while (pg.hasMore()) {
        const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//      for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//      cout << endl;
#endif
        int pv[2]={p[0],p[1]};
        Interval parent(0,pv);
        for (int c=0;c<childrenPerShape();c++) {
          int cv[2]={pv[c],p[2]};
          Interval child(1,cv);
          Point<1> cp(drand48());
          debugChild(parent,child,c,licv,cp);
        }
      }
      cout << endl;
    }
#endif

#if (SPACEDIM==2)
//
    {
      TRACER_CALL(t,"Triangle");
      int pv[3]={0,1,2};
      int l[3]={-1,-2,-3};
      Triangle parent(-1,l,pv);
      NumPtr<Point<SPACEDIM> > vertices(6);
        vertices[0]=Point<SPACEDIM>(0.,0.);
        vertices[1]=Point<SPACEDIM>(1.,0.);
        vertices[2]=Point<SPACEDIM>(0.,1.);
      int li[3]={3,4,5};
      PermutationGenerator pg(3);
      while (pg.hasMore()) {
        const NumPtr<int> &p=pg.getNext();
        NumPtr<int> line_interior_child_vertices(3);
        for (int i=0;i<3;i++) {
          line_interior_child_vertices[i]=li[p[i]];
        }
        vertices[line_interior_child_vertices[0]]=Point<SPACEDIM>(0.,.5);
        vertices[line_interior_child_vertices[1]]=Point<SPACEDIM>(.5,0.);
        vertices[line_interior_child_vertices[2]]=Point<SPACEDIM>(.5,.5);
        int cv[4][3]={ 
          {0,line_interior_child_vertices[1],
             line_interior_child_vertices[0]},
          {1,line_interior_child_vertices[2],
             line_interior_child_vertices[1]},
          {2,line_interior_child_vertices[0],
             line_interior_child_vertices[2]},
          {line_interior_child_vertices[0],
             line_interior_child_vertices[1],
             line_interior_child_vertices[2]}
        };
        for (int c=0;c<childrenPerShape();c++) {
          Triangle child(0,l,cv[c]);
          const Shape *chile=
            child.reorientFor(&parent,c,line_interior_child_vertices);
          CHECK_TEST(chile->volume(vertices)>0.);
          for (int i=0;i<chile->numberVertices();i++) {
            Point<SPACEDIM> cp=Triangle::canonicalVertex(i);
            CHECK_SAME(vertices[chile->vertexIndex(i)],
                       Triangle::parentCoordinateOfCanonical(cp,c));
            if (!(vertices[chile->vertexIndex(i)]==
            Triangle::parentCoordinateOfCanonical(cp,c))) {
              cout << "\tparent:" << endl;
              parent.printOn(cout);
              cout << "\tchild[" << c << "] = " << endl;
              child.printOn(cout);
              cout << "\tchile = " << endl;
              chile->printOn(cout);
              cout << "\ti = " << i << endl;
              cout << "\tvertices[" << chile->vertexIndex(i) << "] = " 
                   << vertices[chile->vertexIndex(i)] << endl;
              cout << "\tparentCoordinateOfCanonical(" << cp << ") = " 
                   << Triangle::parentCoordinateOfCanonical(cp,c) << endl;
              OBSOLETE("vertices bad");
            }
          }
          if (chile!=&child) delete chile;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"Triangle");
      NumPtr<int> line_interior_child_vertices(3);
      int c;
      { c=0;
        PermutationGenerator pg(2);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[0]+1;
          line_interior_child_vertices[1]=p[1]+1;
          line_interior_child_vertices[2]=3;
          int pv[3]={0,4,5};
          int cv[3]={pv[0],line_interior_child_vertices[1],
                     line_interior_child_vertices[0]};
          debugTriangleChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=1;
        PermutationGenerator pg(3);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int pv[3]={0,p[0]+1,4};
          line_interior_child_vertices[0]=5;
          line_interior_child_vertices[1]=p[1]+1;
          line_interior_child_vertices[2]=p[2]+1;
          int cv[3]={pv[1],line_interior_child_vertices[2],
                     line_interior_child_vertices[1]};
          debugTriangleChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=2;
        PermutationGenerator pg(3);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int pv[3]={0,4,p[0]+1};
          line_interior_child_vertices[0]=p[1]+1;
          line_interior_child_vertices[2]=p[2]+1;
          line_interior_child_vertices[1]=5;
          int cv[3]={pv[2],line_interior_child_vertices[0],
                     line_interior_child_vertices[2]};
          debugTriangleChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=3;
        PermutationGenerator pg(3);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int pv[3]={0,4,5};
          line_interior_child_vertices[0]=p[0]+1;
          line_interior_child_vertices[1]=p[1]+1;
          line_interior_child_vertices[2]=p[2]+1;
          int cv[3]={line_interior_child_vertices[0],
                     line_interior_child_vertices[1],
                     line_interior_child_vertices[2]};
          debugTriangleChild(c,pv,cv,line_interior_child_vertices);
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"Quadrilateral");
      int pv[4]={0,1,2,3};
      int l[4]={-1,-2,-3,-4};
      Quadrilateral parent(-1,l,pv);
      NumPtr<Point<SPACEDIM> > vertices(9);
        vertices[0]=Point<SPACEDIM>(0.,0.);
        vertices[1]=Point<SPACEDIM>(1.,0.);
        vertices[2]=Point<SPACEDIM>(0.,1.);
        vertices[3]=Point<SPACEDIM>(1.,1.);
        vertices[8]=Point<SPACEDIM>(.5,.5);
      int li[4]={4,5,6,7};
      PermutationGenerator pg(4);
      while (pg.hasMore()) {
        const NumPtr<int> &p=pg.getNext();
        NumPtr<int> line_interior_child_vertices(4);
        for (int i=0;i<4;i++) {
          line_interior_child_vertices[i]=li[p[i]];
        }
        vertices[line_interior_child_vertices[0]]=Point<SPACEDIM>(0.,.5);
        vertices[line_interior_child_vertices[1]]=Point<SPACEDIM>(1.,.5);
        vertices[line_interior_child_vertices[2]]=Point<SPACEDIM>(.5,0.);
        vertices[line_interior_child_vertices[3]]=Point<SPACEDIM>(.5,1.);
        int cv[4][4]={ 
          {0,line_interior_child_vertices[2],
             line_interior_child_vertices[0],8},
          {1,line_interior_child_vertices[1],
             line_interior_child_vertices[2],8},
          {2,line_interior_child_vertices[0],
             line_interior_child_vertices[3],8},
          {3,line_interior_child_vertices[3],
             line_interior_child_vertices[1],8}
        };
        for (int c=0;c<childrenPerShape();c++) {
          Quadrilateral child(0,l,cv[c]);
//        Shape *chile=
//          child.reorientFor(&parent,c,line_interior_child_vertices);
//        CHECK_TEST(chile->volume(vertices)>0.);
          for (int i=0;i<child.numberVertices();i++) {
            Point<SPACEDIM> cp=Quadrilateral::canonicalVertex(i);
            CHECK_SAME(vertices[child.vertexIndex(i)],
                       Quadrilateral::parentCoordinateOfCanonical(cp,c));
          }
//        if (chile!=child) delete chile;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"Quadrilateral");
      NumPtr<int> line_interior_child_vertices(4);
      int c;
      { c=0;
        PermutationGenerator pg(3);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[1]+1;
          line_interior_child_vertices[1]=7;
          line_interior_child_vertices[2]=p[0]+1;
          line_interior_child_vertices[3]=8;
          int pv[4]={0,4,5,6};
          int cv[4]={pv[0],line_interior_child_vertices[2],
                     line_interior_child_vertices[0],p[2]+1};
          debugQuadrilateralChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=1;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=7;
          line_interior_child_vertices[1]=p[1]+1;
          line_interior_child_vertices[2]=p[2]+1;
          line_interior_child_vertices[3]=8;
          int pv[4]={0,p[0]+1,5,6};
          int cv[4]={pv[1],line_interior_child_vertices[1],
                     line_interior_child_vertices[2],p[3]+1};
          debugQuadrilateralChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=2;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[1]+1;
          line_interior_child_vertices[1]=7;
          line_interior_child_vertices[2]=8;
          line_interior_child_vertices[3]=p[2]+1;
          int pv[4]={0,5,p[0]+1,6};
          int cv[4]={pv[2],line_interior_child_vertices[0],
                     line_interior_child_vertices[3],p[3]+1};
          debugQuadrilateralChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=3;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=7;
          line_interior_child_vertices[1]=p[2]+1;
          line_interior_child_vertices[2]=8;
          line_interior_child_vertices[3]=p[1]+1;
          int pv[4]={0,5,6,p[0]+1};
          int cv[4]={pv[3],line_interior_child_vertices[3],
                     line_interior_child_vertices[1],p[3]+1};
          debugQuadrilateralChild(c,pv,cv,line_interior_child_vertices);
        }
      }
    }
//
#endif

#if (SPACEDIM==3)
//
    {
      TRACER_CALL(t,"Tetrahedron");
      int pv[4]={0,1,2,3};
      int f[4]={-1,-2,-3};
      int l[6]={-1,-2,-3,-4,-5,-6};
      Tetrahedron parent(-1,l,f,pv);
      NumPtr<Point<SPACEDIM> > vertices(10);
        vertices[0]=Point<SPACEDIM>(0.,0.,0.);
        vertices[1]=Point<SPACEDIM>(1.,0.,0.);
        vertices[2]=Point<SPACEDIM>(0.,1.,0.);
        vertices[3]=Point<SPACEDIM>(0.,0.,1.);
      int li[6]={4,5,6,7,8,9};
      PermutationGenerator pg(6);
      int count=0;
      while (pg.hasMore())
      {
        count++;
        const NumPtr<int> &p=pg.getNext();
//      NumPtr<int> p(6); for (int i=0;i<6;i++) p[i]=i;
#ifdef DEBUG
        if (count%1000==0) cout << "\tcount = " << count << endl;
//      cout << "\tp[" << count << "] = ";
//      for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//      cout << endl;
#endif
        NumPtr<int> line_interior_child_vertices(6);
        for (int i=0;i<6;i++) {
          line_interior_child_vertices[i]=li[p[i]];
        }
        vertices[line_interior_child_vertices[0]]=
          Point<SPACEDIM>(.5,0.,0.);
        vertices[line_interior_child_vertices[1]]=
          Point<SPACEDIM>(.5,.5,0.);
        vertices[line_interior_child_vertices[2]]=
          Point<SPACEDIM>(0.,.5,0.);
        vertices[line_interior_child_vertices[3]]=
          Point<SPACEDIM>(0.,0.,.5);
        vertices[line_interior_child_vertices[4]]=
          Point<SPACEDIM>(.5,0.,.5);
        vertices[line_interior_child_vertices[5]]=
          Point<SPACEDIM>(0.,.5,.5);
        int cv[8][4]={ 
          {0,line_interior_child_vertices[0],
             line_interior_child_vertices[2],
             line_interior_child_vertices[3]},
          {1,line_interior_child_vertices[1],
             line_interior_child_vertices[0],
             line_interior_child_vertices[4]},
          {2,line_interior_child_vertices[2],
             line_interior_child_vertices[1],
             line_interior_child_vertices[5]},
          {3,line_interior_child_vertices[4],
             line_interior_child_vertices[3],
             line_interior_child_vertices[5]},
          {line_interior_child_vertices[4],
            line_interior_child_vertices[3],
            line_interior_child_vertices[5],
            line_interior_child_vertices[0]},
          {line_interior_child_vertices[4],
            line_interior_child_vertices[5],
            line_interior_child_vertices[1],
            line_interior_child_vertices[0]},
          {line_interior_child_vertices[2],
            line_interior_child_vertices[5],
            line_interior_child_vertices[0],
            line_interior_child_vertices[1]},
          {line_interior_child_vertices[5],
            line_interior_child_vertices[2],
            line_interior_child_vertices[0],
            line_interior_child_vertices[3]}
        };
        for (int c=0;c<childrenPerShape();c++) {
          Tetrahedron child(0,l,f,cv[c]);
          const Shape *chile=
            child.reorientFor(&parent,c,line_interior_child_vertices);

          NumPtr<set<int> > child_face_vertex_indices(4);
          for (int f=0;f<4;f++) {
            set<int> &cfvi=child_face_vertex_indices[f];
            for (int v=0;v<3;v++) {
              cfvi.insert(child.vertexIndex(
                child.elementVertexFromFace(f,v)));
            }
          }
          for (int f=0;f<4;f++) {
            set<int> cfvi;
            for (int v=0;v<3;v++) {
              cfvi.insert(chile->vertexIndex(
                chile->elementVertexFromFace(f,v)));
            }
            bool found=false;
            for (int f2=0;f2<4;f2++) {
              if (cfvi==child_face_vertex_indices[f2]) {
                found=true;
                break;
              }
            }
            CHECK_TEST(found);
          }
          NumPtr<set<int> > child_line_vertex_indices(6);
          for (int l=0;l<6;l++) {
            set<int> &clvi=child_line_vertex_indices[l];
            for (int v=0;v<2;v++) {
              clvi.insert(child.vertexIndex(
                child.elementVertexFromLine(l,v)));
            }
          }
          for (int l=0;l<6;l++) {
            set<int> clvi;
            for (int v=0;v<2;v++) {
              clvi.insert(chile->vertexIndex(
                chile->elementVertexFromLine(l,v)));
            }
            bool found=false;
            for (int l2=0;l2<6;l2++) {
              if (clvi==child_line_vertex_indices[l2]) {
                found=true;
                break;
              }
            }
            CHECK_TEST(found);
          }

          for (int i=0;i<child.numberVertices();i++) {
            Point<SPACEDIM> cp=Tetrahedron::canonicalVertex(i);
            Point<SPACEDIM> pp=
              Tetrahedron::parentCoordinateOfCanonical(cp,c);
            if (!(vertices[chile->vertexIndex(i)]==pp)) {
              cout << "\tc = " << c << endl;
              cout << "\tparent:" << endl;
              parent.printOn(cout);
              cout << "\tchild:" << endl;
              child.printOn(cout);
              for (int j=0;j<line_interior_child_vertices.getNumber();j++){
                cout << "\tline_interior_child_vertices[" << j << "] = "
                     << line_interior_child_vertices[j] << endl;
              }
              cout << "\tchile:" << endl;
              chile->printOn(cout);
              cout << "\tvertices[" << chile->vertexIndex(i) << "] = "
                   << vertices[chile->vertexIndex(i)] << endl;
              cout << "\tcp = " << cp << endl;
              cout << "\tpp = " << pp << endl;
              CHECK_SAME(vertices[chile->vertexIndex(i)],pp);
            }
            if (!(vertices[child.vertexIndex(i)]==
            parent.parentCoordinateForCanonical(cp,&child,c,
            line_interior_child_vertices))) {
              cout << "\tparent:" << endl;
              parent.printOn(cout);
              cout << "\tchild[" << c << "] = " << endl;
              child.printOn(cout);
              cout << "\tcv = " << cv[c][0] << " " << cv[c][1] << " "
                   << cv[c][2] << " " << cv[c][3] << endl;
              cout << "\tchile = " << endl;
              chile->printOn(cout);
              for (int l=0;
              l<line_interior_child_vertices.getNumber();l++) {
                cout << "\tline_interior_child_vertices[" << l << "] = "
                     << line_interior_child_vertices[l] << endl;
              }
              for (int v=0;v<vertices.getNumber();v++) {
                cout << "\tvertices[" << v << "] = " << vertices[v]
                     << endl;
              }
              cout << "\tvertices[child->vertexIndex(" << i << ")] = "
                   << vertices[child.vertexIndex(i)] << endl;
              cout << "\tTetrahedron::canonicalVertex(" << i << ") = "
                   << cp << endl;
              cout << "Tetrahedron::parentCoordinateOfCanonical(cp,"
                   << c << ") = "
                   << Tetrahedron::parentCoordinateOfCanonical(cp,c)
                   << endl;
              OBSOLETE("vertices bad");
            }
          }
          chile->volume(vertices);
          if (chile!=&child) delete chile;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"Tetrahedron");
      NumPtr<int> line_interior_child_vertices(6);
      int c;
      { c=0;
        PermutationGenerator pg(3);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[0]+2;
          line_interior_child_vertices[1]=7;
          line_interior_child_vertices[2]=p[1]+2;
          line_interior_child_vertices[3]=p[2]+2;
          line_interior_child_vertices[4]=8;
          line_interior_child_vertices[5]=9;
          int pv[4]={0,1,5,6};
          int cv[4]={pv[0],line_interior_child_vertices[0],
                     line_interior_child_vertices[2],
                     line_interior_child_vertices[3]};
          debugTetrahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=1;
        PermutationGenerator pg(3);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[1]+2;
          line_interior_child_vertices[1]=p[0]+2;
          line_interior_child_vertices[2]=7;
          line_interior_child_vertices[3]=8;
          line_interior_child_vertices[4]=p[2]+2;
          line_interior_child_vertices[5]=9;
          int pv[4]={0,1,5,6};
          int cv[4]={pv[1],line_interior_child_vertices[1],
                     line_interior_child_vertices[0],
                     line_interior_child_vertices[4]};
          debugTetrahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=2;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=7;
          line_interior_child_vertices[1]=p[2]+2;
          line_interior_child_vertices[2]=p[1]+2;
          line_interior_child_vertices[3]=8;
          line_interior_child_vertices[4]=9;
          line_interior_child_vertices[5]=p[3]+2;
          int pv[4]={0,1,p[0]+2,6};
          int cv[4]={pv[2],line_interior_child_vertices[2],
                     line_interior_child_vertices[1],
                     line_interior_child_vertices[5]};
          debugTetrahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=3;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=7;
          line_interior_child_vertices[1]=8;
          line_interior_child_vertices[2]=9;
          line_interior_child_vertices[3]=p[2]+2;
          line_interior_child_vertices[4]=p[1]+2;
          line_interior_child_vertices[5]=p[3]+2;
          int pv[4]={0,1,6,p[0]+2};
          int cv[4]={pv[3],line_interior_child_vertices[4],
                     line_interior_child_vertices[3],
                     line_interior_child_vertices[5]};
          debugTetrahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=4;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[0]+2;
          line_interior_child_vertices[1]=8;
          line_interior_child_vertices[2]=9;
          line_interior_child_vertices[3]=p[1]+2;
          line_interior_child_vertices[4]=p[2]+2;
          line_interior_child_vertices[5]=p[3]+2;
          int pv[4]={0,1,6,7};
          int cv[4]={line_interior_child_vertices[0],
                     line_interior_child_vertices[3],
                     line_interior_child_vertices[4],
                     line_interior_child_vertices[5]};
          debugTetrahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=5;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[0]+2;
          line_interior_child_vertices[1]=p[1]+2;
          line_interior_child_vertices[2]=8;
          line_interior_child_vertices[3]=9;
          line_interior_child_vertices[4]=p[3]+2;
          line_interior_child_vertices[5]=p[2]+2;
          int pv[4]={0,1,6,7};
          int cv[4]={line_interior_child_vertices[0],
                     line_interior_child_vertices[1],
                     line_interior_child_vertices[5],
                     line_interior_child_vertices[4]};
          debugTetrahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=6;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[0]+2;
          line_interior_child_vertices[1]=p[1]+2;
          line_interior_child_vertices[2]=p[2]+2;
          line_interior_child_vertices[3]=8;
          line_interior_child_vertices[4]=9;
          line_interior_child_vertices[5]=p[3]+2;
          int pv[4]={0,1,6,7};
          int cv[4]={line_interior_child_vertices[0],
                     line_interior_child_vertices[1],
                     line_interior_child_vertices[2],
                     line_interior_child_vertices[5]};
          debugTetrahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=7;
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[0]+2;
          line_interior_child_vertices[1]=8;
          line_interior_child_vertices[2]=p[2]+2;
          line_interior_child_vertices[3]=p[1]+2;
          line_interior_child_vertices[4]=9;
          line_interior_child_vertices[5]=p[3]+2;
          int pv[4]={0,1,6,7};
          int cv[4]={line_interior_child_vertices[0],
                     line_interior_child_vertices[2],
                     line_interior_child_vertices[3],
                     line_interior_child_vertices[5]};
          debugTetrahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"Prism");
      int pv[6]={0,1,2,3,4,5};
      int f[5]={-1,-2,-3,-4,-5};
      int l[9]={-1,-2,-3,-4,-5,-6,-7,-8,-9};
      Prism parent(-1,l,f,pv);
      NumPtr<Point<SPACEDIM> > vertices(18);
        vertices[0]=Point<SPACEDIM>(0.,0.,0.);
        vertices[1]=Point<SPACEDIM>(1.,0.,0.);
        vertices[2]=Point<SPACEDIM>(0.,1.,0.);
        vertices[3]=Point<SPACEDIM>(0.,0.,1.);
        vertices[4]=Point<SPACEDIM>(1.,0.,1.);
        vertices[5]=Point<SPACEDIM>(0.,1.,1.);
      int li[12]={6,7,8,9,10,11,12,13,14,15,16,17};
//    PermutationGenerator pg(12);
//    int count=0;
//    while (pg.hasMore()) 
      for (int count=0;count<20000;count++)
      {
//      count++;
//      if (count%100000==0) cout << count << endl;
//      const NumPtr<int> &p=pg.getNext();
        NumPtr<int> p(12);
        RandomPermutation(p);
#ifdef DEBUG
        if (count%1000==0) cout << "\tcount = " << count << endl;
//      cout << "\tp[" << count << "] = ";
//      for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//      cout << endl;
#endif
        NumPtr<int> line_interior_child_vertices(12);
        for (int i=0;i<12;i++) {
          line_interior_child_vertices[i]=li[p[i]];
        }
        vertices[line_interior_child_vertices[0]]=
          Point<SPACEDIM>(.5,0.,0.);
        vertices[line_interior_child_vertices[1]]=
          Point<SPACEDIM>(.5,.5,0.);
        vertices[line_interior_child_vertices[2]]=
          Point<SPACEDIM>(0.,.5,0.);
        vertices[line_interior_child_vertices[3]]=
          Point<SPACEDIM>(.5,0.,1.);
        vertices[line_interior_child_vertices[4]]=
          Point<SPACEDIM>(.5,.5,1.);
        vertices[line_interior_child_vertices[5]]=
          Point<SPACEDIM>(0.,.5,1.);
        vertices[line_interior_child_vertices[6]]=
          Point<SPACEDIM>(0.,0.,.5);
        vertices[line_interior_child_vertices[7]]=
          Point<SPACEDIM>(1.,0.,.5);
        vertices[line_interior_child_vertices[8]]=
          Point<SPACEDIM>(0.,1.,.5);
        vertices[line_interior_child_vertices[9]]=
          Point<SPACEDIM>(0.,.5,.5);
        vertices[line_interior_child_vertices[10]]=
          Point<SPACEDIM>(.5,0.,.5);
        vertices[line_interior_child_vertices[11]]=
          Point<SPACEDIM>(.5,.5,.5);
        int cv[8][6]={ 
         {0,line_interior_child_vertices[0],
           line_interior_child_vertices[2],
           line_interior_child_vertices[6],
           line_interior_child_vertices[10],
           line_interior_child_vertices[9]},
         {1,line_interior_child_vertices[1],
           line_interior_child_vertices[0],
           line_interior_child_vertices[7],
           line_interior_child_vertices[11],
           line_interior_child_vertices[10]},
         {2,line_interior_child_vertices[2],
           line_interior_child_vertices[1],
           line_interior_child_vertices[8],
           line_interior_child_vertices[9],
           line_interior_child_vertices[11]},
         {line_interior_child_vertices[11],
           line_interior_child_vertices[10],
           line_interior_child_vertices[9],
           line_interior_child_vertices[1],
           line_interior_child_vertices[0],
           line_interior_child_vertices[2]},
         {3,line_interior_child_vertices[5],
           line_interior_child_vertices[3],
           line_interior_child_vertices[6],
           line_interior_child_vertices[9],
           line_interior_child_vertices[10]},
         {4,line_interior_child_vertices[3],
           line_interior_child_vertices[4],
           line_interior_child_vertices[7],
           line_interior_child_vertices[10],
           line_interior_child_vertices[11]},
         {5,line_interior_child_vertices[4],
           line_interior_child_vertices[5],
           line_interior_child_vertices[8],
           line_interior_child_vertices[11],
           line_interior_child_vertices[9]},
         {line_interior_child_vertices[4],
           line_interior_child_vertices[3],
           line_interior_child_vertices[5],
           line_interior_child_vertices[11],
           line_interior_child_vertices[10],
           line_interior_child_vertices[9]} 
        };
        for (int c=0;c<childrenPerShape();c++) {
          Prism child(0,l,f,cv[c]);
          const Shape *chile=
            child.reorientFor(&parent,c,line_interior_child_vertices);

          NumPtr<set<int> > child_face_vertex_indices(5);
          for (int f=0;f<2;f++) {
            set<int> &cfvi=child_face_vertex_indices[f];
            for (int v=0;v<3;v++) {
              cfvi.insert(child.vertexIndex(
                child.elementVertexFromFace(f,v)));
            }
          }
          for (int f=0;f<2;f++) {
            set<int> cfvi;
            for (int v=0;v<3;v++) {
              cfvi.insert(chile->vertexIndex(
                chile->elementVertexFromFace(f,v)));
            }
            bool found=false;
            for (int f2=0;f2<2;f2++) {
              if (cfvi==child_face_vertex_indices[f2]) {
                found=true;
                break;
              }
            }
            CHECK_TEST(found);
          }
          for (int f=2;f<5;f++) {
            set<int> &cfvi=child_face_vertex_indices[f];
            for (int v=0;v<4;v++) {
              cfvi.insert(child.vertexIndex(
                child.elementVertexFromFace(f,v)));
            }
          }
          for (int f=2;f<5;f++) {
            set<int> cfvi;
            for (int v=0;v<4;v++) {
              cfvi.insert(chile->vertexIndex(
                chile->elementVertexFromFace(f,v)));
            }
            bool found=false;
            for (int f2=2;f2<5;f2++) {
              if (cfvi==child_face_vertex_indices[f2]) {
                found=true;
                break;
              }
            }
            CHECK_TEST(found);
          }
          NumPtr<set<int> > child_line_vertex_indices(9);
          for (int l=0;l<9;l++) {
            set<int> &clvi=child_line_vertex_indices[l];
            for (int v=0;v<2;v++) {
              clvi.insert(child.vertexIndex(
                child.elementVertexFromLine(l,v)));
            }
          }
          for (int l=0;l<9;l++) {
            set<int> clvi;
            for (int v=0;v<2;v++) {
              clvi.insert(chile->vertexIndex(
                chile->elementVertexFromLine(l,v)));
            }
            bool found=false;
            for (int l2=0;l2<9;l2++) {
              if (clvi==child_line_vertex_indices[l2]) {
                found=true;
                break;
              }
            }
            CHECK_TEST(found);
          }

          for (int i=0;i<child.numberVertices();i++) {
            Point<SPACEDIM> cp=Prism::canonicalVertex(i);
            Point<SPACEDIM> pp=
              Prism::parentCoordinateOfCanonical(cp,c);
            if (!(vertices[chile->vertexIndex(i)]==pp)) {
              cout << "\tc = " << c << endl;
              cout << "\tparent:" << endl;
              parent.printOn(cout);
              cout << "\tchild:" << endl;
              child.printOn(cout);
              for (int j=0;j<line_interior_child_vertices.getNumber();j++){
                cout << "\tline_interior_child_vertices[" << j << "] = "
                     << line_interior_child_vertices[j] << endl;
              }
              cout << "\tchile:" << endl;
              chile->printOn(cout);
              cout << "\tvertices[" << chile->vertexIndex(i) << "] = "
                   << vertices[chile->vertexIndex(i)] << endl;
              cout << "\tcp = " << cp << endl;
              cout << "\tpp = " << pp << endl;
              CHECK_SAME(vertices[chile->vertexIndex(i)],pp);
            }
            if (!(vertices[child.vertexIndex(i)]==
            parent.parentCoordinateForCanonical(cp,&child,c,
            line_interior_child_vertices))) {
              cout << "\tparent:" << endl;
              parent.printOn(cout);
              cout << "\tchild[" << c << "] = " << endl;
              child.printOn(cout);
              cout << "\tcv = " << cv[c][0] << " " << cv[c][1] << " "
                   << cv[c][2] << " " << cv[c][3] << " " << cv[c][4]
                   << " " << cv[c][5] << endl;
              cout << "\tchile = " << endl;
              chile->printOn(cout);
              for (int l=0;
              l<line_interior_child_vertices.getNumber();l++) {
                cout << "\tline_interior_child_vertices[" << l << "] = "
                     << line_interior_child_vertices[l] << endl;
              }
              for (int v=0;v<vertices.getNumber();v++) {
                cout << "\tvertices[" << v << "] = " << vertices[v]
                     << endl;
              }
              cout << "\tvertices[child->vertexIndex(" << i << ")] = "
                   << vertices[child.vertexIndex(i)] << endl;
              cout << "\tPrism::canonicalVertex(" << i << ") = "
                   << cp << endl;
              cout << "Prism::parentCoordinateOfCanonical(cp,"
                   << c << ") = "
                   << Prism::parentCoordinateOfCanonical(cp,c)
                   << endl;
              OBSOLETE("vertices bad");
            }
          }
          chile->volume(vertices);
          if (chile!=&child) delete chile;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"Prism");
      NumPtr<int> line_interior_child_vertices(9);
      int c;
      { c=0;
        PermutationGenerator pg(5);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[0]+1;
          line_interior_child_vertices[1]=11;
          line_interior_child_vertices[2]=p[1]+1;
          line_interior_child_vertices[3]=12;
          line_interior_child_vertices[4]=13;
          line_interior_child_vertices[5]=14;
          line_interior_child_vertices[6]=p[2]+1;
          line_interior_child_vertices[7]=15;
          line_interior_child_vertices[8]=16;
          int pv[6]={0,6,7,8,9,10};
          int cv[6]={pv[0],line_interior_child_vertices[0],
                     line_interior_child_vertices[2],
                     line_interior_child_vertices[6],p[3]+1,p[4]+1};
          debugPrismChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=1;
        PermutationGenerator pg(6);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[2]+1;
          line_interior_child_vertices[1]=p[1]+1;
          line_interior_child_vertices[2]=11;
          line_interior_child_vertices[3]=12;
          line_interior_child_vertices[4]=13;
          line_interior_child_vertices[5]=14;
          line_interior_child_vertices[6]=15;
          line_interior_child_vertices[7]=p[3]+1;
          line_interior_child_vertices[8]=16;
          int pv[6]={0,p[0]+1,7,8,9,10};
          int cv[6]={pv[1],line_interior_child_vertices[1],
                     line_interior_child_vertices[0],
                     line_interior_child_vertices[7],p[4]+1,p[5]+1};
          debugPrismChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=2;
        PermutationGenerator pg(6);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=11;
          line_interior_child_vertices[1]=p[2]+1;
          line_interior_child_vertices[2]=p[1]+1;
          line_interior_child_vertices[3]=12;
          line_interior_child_vertices[4]=13;
          line_interior_child_vertices[5]=14;
          line_interior_child_vertices[6]=15;
          line_interior_child_vertices[7]=16;
          line_interior_child_vertices[8]=p[3]+1;
          int pv[6]={0,7,p[0]+1,8,9,10};
          int cv[6]={pv[2],line_interior_child_vertices[2],
                     line_interior_child_vertices[1],
                     line_interior_child_vertices[8],p[4]+1,p[5]+1};
          debugPrismChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=3;
        PermutationGenerator pg(6);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[1]+1;
          line_interior_child_vertices[1]=p[0]+1;
          line_interior_child_vertices[2]=p[2]+1;
          line_interior_child_vertices[3]=12;
          line_interior_child_vertices[4]=13;
          line_interior_child_vertices[5]=14;
          line_interior_child_vertices[6]=15;
          line_interior_child_vertices[7]=16;
          line_interior_child_vertices[8]=17;
          int pv[6]={0,7,8,9,10,11};
          int cv[6]={p[5]+1,p[4]+1,p[3]+1,line_interior_child_vertices[1],
                     line_interior_child_vertices[0],
                     line_interior_child_vertices[2]};
          debugPrismChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=4;
        PermutationGenerator pg(6);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=11;
          line_interior_child_vertices[1]=12;
          line_interior_child_vertices[2]=13;
          line_interior_child_vertices[3]=p[4]+1;
          line_interior_child_vertices[4]=14;
          line_interior_child_vertices[5]=p[5]+1;
          line_interior_child_vertices[6]=p[0]+1;
          line_interior_child_vertices[7]=15;
          line_interior_child_vertices[8]=16;
          int pv[6]={0,7,8,p[3]+1,9,10};
          int cv[6]={line_interior_child_vertices[6],p[1]+1,p[2]+1,pv[3],
                     line_interior_child_vertices[3],
                     line_interior_child_vertices[5]};
          debugPrismChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=5;
        PermutationGenerator pg(6);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=11;
          line_interior_child_vertices[1]=12;
          line_interior_child_vertices[2]=13;
          line_interior_child_vertices[3]=p[5]+1;
          line_interior_child_vertices[4]=p[4]+1;
          line_interior_child_vertices[5]=14;
          line_interior_child_vertices[6]=15;
          line_interior_child_vertices[7]=p[0]+1;
          line_interior_child_vertices[8]=16;
          int pv[6]={0,7,8,9,p[3]+1,10};
          int cv[6]={line_interior_child_vertices[7],p[1]+1,p[2]+1,pv[4],
                     line_interior_child_vertices[4],
                     line_interior_child_vertices[3]};
          debugPrismChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=6;
        PermutationGenerator pg(6);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=11;
          line_interior_child_vertices[1]=12;
          line_interior_child_vertices[2]=13;
          line_interior_child_vertices[3]=14;
          line_interior_child_vertices[4]=p[5]+1;
          line_interior_child_vertices[5]=p[4]+1;
          line_interior_child_vertices[6]=15;
          line_interior_child_vertices[7]=16;
          line_interior_child_vertices[8]=p[0]+1;
          int pv[6]={0,7,8,9,10,p[3]+1};
          int cv[6]={line_interior_child_vertices[8],p[1]+1,p[2]+1,pv[5],
                     line_interior_child_vertices[5],
                     line_interior_child_vertices[4]};
          debugPrismChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=7;
        PermutationGenerator pg(6);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=12;
          line_interior_child_vertices[1]=13;
          line_interior_child_vertices[2]=14;
          line_interior_child_vertices[3]=p[1]+1;
          line_interior_child_vertices[4]=p[0]+1;
          line_interior_child_vertices[5]=p[2]+1;
          line_interior_child_vertices[6]=15;
          line_interior_child_vertices[7]=16;
          line_interior_child_vertices[8]=17;
          int pv[6]={0,7,8,9,10,11};
          int cv[6]={line_interior_child_vertices[4],
                     line_interior_child_vertices[3],
                     line_interior_child_vertices[5],p[3]+1,p[4]+1,p[5]+1};
          debugPrismChild(c,pv,cv,line_interior_child_vertices);
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"Hexahedron");
      int pv[8]={0,1,2,3,4,5,6,7};
      int f[6]={-1,-2,-3,-4,-5,-6};
      int l[12]={-1,-2,-3,-4,-5,-6,-7,-8,-9,-10,-11,-12};
      Hexahedron parent(-1,l,f,pv);
      NumPtr<Point<SPACEDIM> > vertices(27);
        vertices[0]=Point<SPACEDIM>(0.,0.,0.);
        vertices[1]=Point<SPACEDIM>(1.,0.,0.);
        vertices[2]=Point<SPACEDIM>(0.,1.,0.);
        vertices[3]=Point<SPACEDIM>(1.,1.,0.);
        vertices[4]=Point<SPACEDIM>(0.,0.,1.);
        vertices[5]=Point<SPACEDIM>(1.,0.,1.);
        vertices[6]=Point<SPACEDIM>(0.,1.,1.);
        vertices[7]=Point<SPACEDIM>(1.,1.,1.);
        vertices[20]=Point<SPACEDIM>(0.,.5,.5);
        vertices[21]=Point<SPACEDIM>(1.,.5,.5);
        vertices[22]=Point<SPACEDIM>(.5,0.,.5);
        vertices[23]=Point<SPACEDIM>(.5,1.,.5);
        vertices[24]=Point<SPACEDIM>(.5,.5,0.);
        vertices[25]=Point<SPACEDIM>(.5,.5,1.);
        vertices[26]=Point<SPACEDIM>(.5,.5,.5);
      int li[12]={8,9,10,11,12,13,14,15,16,17,18,19};
//    PermutationGenerator pg(12);
//    int count=0;
//    while (pg.hasMore()) 
      for (int count=0;count<1000;count++)
      {
//      count++;
//      if (count%10000==0) cout << count << endl;
//      const NumPtr<int> &p=pg.getNext();
        NumPtr<int> p(12);
        RandomPermutation(p);
#ifdef DEBUG
        if (count%1000==0) cout << "\tcount = " << count << endl;
//      cout << "\tp[" << count << "] = ";
//      for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//      cout << endl;
#endif
        NumPtr<int> line_interior_child_vertices(12);
        for (int i=0;i<12;i++) {
          line_interior_child_vertices[i]=li[p[i]];
        }
        vertices[line_interior_child_vertices[0]]=
          Point<SPACEDIM>(0.,.5,0.);
        vertices[line_interior_child_vertices[1]]=
          Point<SPACEDIM>(1.,.5,0.);
        vertices[line_interior_child_vertices[2]]=
          Point<SPACEDIM>(.5,0.,0.);
        vertices[line_interior_child_vertices[3]]=
          Point<SPACEDIM>(.5,1.,0.);
        vertices[line_interior_child_vertices[4]]=
          Point<SPACEDIM>(0.,.5,1.);
        vertices[line_interior_child_vertices[5]]=
          Point<SPACEDIM>(1.,.5,1.);
        vertices[line_interior_child_vertices[6]]=
          Point<SPACEDIM>(.5,0.,1.);
        vertices[line_interior_child_vertices[7]]=
          Point<SPACEDIM>(.5,1.,1.);
        vertices[line_interior_child_vertices[8]]=
          Point<SPACEDIM>(0.,0.,.5);
        vertices[line_interior_child_vertices[9]]=
          Point<SPACEDIM>(1.,0.,.5);
        vertices[line_interior_child_vertices[10]]=
          Point<SPACEDIM>(0.,1.,.5);
        vertices[line_interior_child_vertices[11]]=
          Point<SPACEDIM>(1.,1.,.5);
//      vertices[line_interior_child_vertices[12]]=
//        Point<SPACEDIM>(0.,.5,.5);
//      vertices[line_interior_child_vertices[13]]=
//        Point<SPACEDIM>(1.,.5,.5);
//      vertices[line_interior_child_vertices[14]]=
//        Point<SPACEDIM>(.5,0.,.5);
//      vertices[line_interior_child_vertices[15]]=
//        Point<SPACEDIM>(.5,1.,.5);
//      vertices[line_interior_child_vertices[16]]=
//        Point<SPACEDIM>(.5,.5,0.);
//      vertices[line_interior_child_vertices[17]]=
//        Point<SPACEDIM>(.5,.5,1.);
        int cv[8][8]={ 
         {0,line_interior_child_vertices[2],
           line_interior_child_vertices[0],
           24,line_interior_child_vertices[8],22,20,26},
         {1,line_interior_child_vertices[1],
           line_interior_child_vertices[2],
           24,line_interior_child_vertices[9],21,22,26},
         {2,line_interior_child_vertices[0],
           line_interior_child_vertices[3],
           24,line_interior_child_vertices[10],20,23,26},
         {3,line_interior_child_vertices[3],
           line_interior_child_vertices[1],
           24,line_interior_child_vertices[11],23,21,26},
         {4,line_interior_child_vertices[4],
           line_interior_child_vertices[6],
           25,line_interior_child_vertices[8],20,22,26},
         {5,line_interior_child_vertices[6],
           line_interior_child_vertices[5],
           25,line_interior_child_vertices[9],22,21,26},
         {6,line_interior_child_vertices[7],
           line_interior_child_vertices[4],
           25,line_interior_child_vertices[10],23,20,26},
         {7,line_interior_child_vertices[5],
           line_interior_child_vertices[7],
           25,line_interior_child_vertices[11],21,23,26},
        };
        for (int c=0;c<childrenPerShape();c++) {
          Hexahedron child(0,l,f,cv[c]);
          const Shape *chile=
            child.reorientFor(&parent,c,line_interior_child_vertices);
          for (int i=0;i<chile->numberVertices();i++) {
            Point<SPACEDIM> cp=Hexahedron::canonicalVertex(i);
            Point<SPACEDIM> pp=
              Hexahedron::parentCoordinateOfCanonical(cp,c);
            if (!(vertices[chile->vertexIndex(i)]==pp)) {
              cout << "\tc = " << c << endl;
              cout << "\tparent:" << endl;
              parent.printOn(cout);
              cout << "\tchild:" << endl;
              child.printOn(cout);
              for (int j=0;j<line_interior_child_vertices.getNumber();j++){
                cout << "\tline_interior_child_vertices[" << j << "] = "
                     << line_interior_child_vertices[j] << endl;
              }
              cout << "\tchile:" << endl;
              chile->printOn(cout);
              cout << "\tvertices[" << chile->vertexIndex(i) << "] = "
                   << vertices[chile->vertexIndex(i)] << endl;
              cout << "\tcp = " << cp << endl;
              cout << "\tpp = " << pp << endl;
              CHECK_SAME(vertices[chile->vertexIndex(i)],pp);
            }
            if (!(vertices[child.vertexIndex(i)]==
            parent.parentCoordinateForCanonical(cp,&child,c,
            line_interior_child_vertices))) {
              cout << "\tparent:" << endl;
              parent.printOn(cout);
              cout << "\tchild[" << c << "] = " << endl;
              child.printOn(cout);
              cout << "\tcv = " << cv[c][0] << " " << cv[c][1] << " "
                   << cv[c][2] << " " << cv[c][3] << " " << cv[c][4]
                   << " " << cv[c][5] << " " << cv[c][6] << " "
                   << cv[c][7] << endl;
//            cout << "\tchile = " << endl;
//            chile->printOn(cout);
              for (int li=0;
              li<line_interior_child_vertices.getNumber();li++) {
                cout << "\tline_interior_child_vertices[" << li << "] = "
                     << line_interior_child_vertices[li] << endl;
              }
              for (int v=0;v<vertices.getNumber();v++) {
                cout << "\tvertices[" << v << "] = " << vertices[v]
                     << endl;
              }
              cout << "\tvertices[child->vertexIndex(" << i << ")] = "
                   << vertices[chile->vertexIndex(i)] << endl;
              cout << "\tHexahedron::canonicalVertex(" << i << ") = "
                   << cp << endl;
              cout << "\tHexahedron::parentCoordinateOfCanonical(cp,"
                   << c << ") = "
                   << Hexahedron::parentCoordinateOfCanonical(cp,c)
                   << endl;
              OBSOLETE("vertices bad");
            }
          }
          CHECK_TEST(chile->volume(vertices)>0.);
          if (chile!=&child) delete chile;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"Hexahedron");
      NumPtr<int> line_interior_child_vertices(12);
      int c;
      { c=0;
        TRACER_CALL(t,"Hexahedron");
        PermutationGenerator pg(7);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=p[1]+2;
          line_interior_child_vertices[1]=15;
          line_interior_child_vertices[2]=p[0]+2;
          line_interior_child_vertices[3]=16;
          line_interior_child_vertices[4]=17;
          line_interior_child_vertices[5]=18;
          line_interior_child_vertices[6]=19;
          line_interior_child_vertices[7]=20;
          line_interior_child_vertices[8]=p[2]+2;
          line_interior_child_vertices[9]=21;
          line_interior_child_vertices[10]=22;
          line_interior_child_vertices[11]=23;
          int pv[8]={0,1,9,10,11,12,13,14};
          int cv[8]={pv[0],line_interior_child_vertices[2],
                     line_interior_child_vertices[0],p[3]+2,
                     line_interior_child_vertices[8],p[4]+2,p[5]+2,p[6]+2};
          debugHexahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=1;
        TRACER_CALL(t,"Hexahedron");
        PermutationGenerator pg(7);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          line_interior_child_vertices[0]=15;
          line_interior_child_vertices[1]=p[0]+2;
          line_interior_child_vertices[2]=p[1]+2;
          line_interior_child_vertices[3]=16;
          line_interior_child_vertices[4]=17;
          line_interior_child_vertices[5]=18;
          line_interior_child_vertices[6]=19;
          line_interior_child_vertices[7]=20;
          line_interior_child_vertices[8]=21;
          line_interior_child_vertices[9]=p[2]+2;
          line_interior_child_vertices[10]=22;
          line_interior_child_vertices[11]=23;
          int pv[8]={0,1,9,10,11,12,13,14};
          int cv[8]={pv[1],line_interior_child_vertices[1],
                     line_interior_child_vertices[2],p[3]+2,
                     line_interior_child_vertices[9],p[4]+2,p[5]+2,p[6]+2};
          debugHexahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=2;
        TRACER_CALL(t,"Hexahedron");
        PermutationGenerator pg(8);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//        cout << endl;
//        for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//        cout << endl;
#endif
          line_interior_child_vertices[0]=p[1]+2;
          line_interior_child_vertices[1]=15;
          line_interior_child_vertices[2]=16;
          line_interior_child_vertices[3]=p[2]+2;
          line_interior_child_vertices[4]=17;
          line_interior_child_vertices[5]=18;
          line_interior_child_vertices[6]=19;
          line_interior_child_vertices[7]=20;
          line_interior_child_vertices[8]=21;
          line_interior_child_vertices[9]=22;
          line_interior_child_vertices[10]=p[3]+2;
          line_interior_child_vertices[11]=23;
          int pv[8]={0,1,p[0]+2,10,11,12,13,14};
          int cv[8]={pv[2],line_interior_child_vertices[0],
                     line_interior_child_vertices[3],p[4]+2,
                     line_interior_child_vertices[10],p[5]+2,p[6]+2,
                     p[7]+2};
          debugHexahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=3;
        TRACER_CALL(t,"Hexahedron");
        PermutationGenerator pg(8);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//        if (p[0]<2) continue;
//        cout << endl;
//        for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//        cout << endl;
#endif
          line_interior_child_vertices[0]=15;
          line_interior_child_vertices[1]=p[2]+2;
          line_interior_child_vertices[2]=16;
          line_interior_child_vertices[3]=p[1]+2;
          line_interior_child_vertices[4]=17;
          line_interior_child_vertices[5]=18;
          line_interior_child_vertices[6]=19;
          line_interior_child_vertices[7]=20;
          line_interior_child_vertices[8]=21;
          line_interior_child_vertices[9]=22;
          line_interior_child_vertices[10]=23;
          line_interior_child_vertices[11]=p[3]+2;
          int pv[8]={0,1,10,p[0]+2,11,12,13,14};
          int cv[8]={pv[3],line_interior_child_vertices[3],
                     line_interior_child_vertices[1],p[4]+2,
                     line_interior_child_vertices[11],p[5]+2,p[6]+2,
                     p[7]+2};
          debugHexahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=4;
        TRACER_CALL(t,"Hexahedron");
        PermutationGenerator pg(8);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//        if (p[0]<2) continue;
//        cout << endl;
//        for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//        cout << endl;
#endif
          line_interior_child_vertices[0]=15;
          line_interior_child_vertices[1]=16;
          line_interior_child_vertices[2]=17;
          line_interior_child_vertices[3]=18;
          line_interior_child_vertices[4]=p[1]+2;
          line_interior_child_vertices[5]=19;
          line_interior_child_vertices[6]=p[2]+2;
          line_interior_child_vertices[7]=20;
          line_interior_child_vertices[8]=p[3]+2;
          line_interior_child_vertices[9]=21;
          line_interior_child_vertices[10]=22;
          line_interior_child_vertices[11]=23;
          int pv[8]={0,1,10,11,p[0]+2,12,13,14};
          int cv[8]={pv[4],line_interior_child_vertices[4],
                     line_interior_child_vertices[6],p[4]+2,
                     line_interior_child_vertices[8],p[5]+2,p[6]+2,
                     p[7]+2};
          debugHexahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=5;
        TRACER_CALL(t,"Hexahedron");
        PermutationGenerator pg(8);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//        if (p[0]<2) continue;
//        cout << endl;
//        for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//        cout << endl;
#endif
          line_interior_child_vertices[0]=15;
          line_interior_child_vertices[1]=16;
          line_interior_child_vertices[2]=17;
          line_interior_child_vertices[3]=18;
          line_interior_child_vertices[4]=19;
          line_interior_child_vertices[5]=p[2]+2;
          line_interior_child_vertices[6]=p[1]+2;
          line_interior_child_vertices[7]=20;
          line_interior_child_vertices[8]=21;
          line_interior_child_vertices[9]=p[3]+2;
          line_interior_child_vertices[10]=22;
          line_interior_child_vertices[11]=23;
          int pv[8]={0,1,10,11,12,p[0]+2,13,14};
          int cv[8]={pv[5],line_interior_child_vertices[6],
                     line_interior_child_vertices[5],p[4]+2,
                     line_interior_child_vertices[9],p[5]+2,p[6]+2,
                     p[7]+2};
          debugHexahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=6;
        TRACER_CALL(t,"Hexahedron");
        PermutationGenerator pg(8);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//        if (p[0]<2) continue;
//        cout << endl;
//        for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//        cout << endl;
#endif
          line_interior_child_vertices[0]=15;
          line_interior_child_vertices[1]=16;
          line_interior_child_vertices[2]=17;
          line_interior_child_vertices[3]=18;
          line_interior_child_vertices[4]=p[2]+2;
          line_interior_child_vertices[5]=19;
          line_interior_child_vertices[6]=20;
          line_interior_child_vertices[7]=p[1]+2;
          line_interior_child_vertices[8]=21;
          line_interior_child_vertices[9]=22;
          line_interior_child_vertices[10]=p[3]+2;
          line_interior_child_vertices[11]=23;
          int pv[8]={0,1,10,11,12,13,p[0]+2,14};
          int cv[8]={pv[6],line_interior_child_vertices[7],
                     line_interior_child_vertices[4],p[4]+2,
                     line_interior_child_vertices[10],p[5]+2,p[6]+2,
                     p[7]+2};
          debugHexahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
      { c=7;
        TRACER_CALL(t,"Hexahedron");
        PermutationGenerator pg(8);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
#ifdef DEBUG
//        if (p[0]<2) continue;
//        cout << endl;
//        for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//        cout << endl;
#endif
          line_interior_child_vertices[0]=15;
          line_interior_child_vertices[1]=16;
          line_interior_child_vertices[2]=17;
          line_interior_child_vertices[3]=18;
          line_interior_child_vertices[4]=19;
          line_interior_child_vertices[5]=p[1]+2;
          line_interior_child_vertices[6]=20;
          line_interior_child_vertices[7]=p[2]+2;
          line_interior_child_vertices[8]=21;
          line_interior_child_vertices[9]=22;
          line_interior_child_vertices[10]=23;
          line_interior_child_vertices[11]=p[3]+2;
          int pv[8]={0,1,10,11,12,13,14,p[0]+2};
          int cv[8]={pv[7],line_interior_child_vertices[5],
                     line_interior_child_vertices[7],p[4]+2,
                     line_interior_child_vertices[11],p[5]+2,p[6]+2,
                     p[7]+2};
          debugHexahedronChild(c,pv,cv,line_interior_child_vertices);
        }
      }
    }
//
#endif

  }
  return EXIT_SUCCESS;
}
#include "NumPtr.C"
INSTANTIATE_NUMPTR(set<int>)
