// "$Header: /home/faculty/johnt/cvs/deal_new/examples/LinearElements1D.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "ConditionNumber.H"
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Multigrid.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM!=2)
barf;
#endif

//- u'' - pi^2 u = sin(pi x)
//  u(0) = 0 = u(1)

int N_CYCLES=6;
REAL small=M_PI;
bool rot=false;
REAL con=1.;

REAL exactSolutionValue(const Point<SPACEDIM> &p) {
  if (rot) return exp(-small*(p[1]-p[0]))*sin(M_PI*(p[1]+p[0]));
  else return exp(-small*p[1])*sin(M_PI*p[0]);
}
Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &p) {
  Tensor<1,SPACEDIM> return_val;
  if (rot) {
    REAL e=exp(-small*(p[1]-p[0]));
    REAL s=M_PI*(p[0]+p[1]);
    REAL si=small*sin(s);
    REAL co=M_PI*cos(s);
    return_val[0]=e*(co+si);
    return_val[1]=e*(co-si);
  } else {
    REAL e=exp(-small*p[1]);
    REAL s=M_PI*p[0];
    return_val[0]=M_PI*e*cos(s);
    return_val[1]=-small*e*sin(s);
  }
  return return_val;
}
class ExactFunctionSystem : public FunctionSystem {
  protected:
    explicit ExactFunctionSystem(const FunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ExactFunctionSystem() {}
    virtual ~ExactFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> > &gradients,
    const Point<SPACEDIM> &p) const {
      gradients[0]=exactSolutionGradient(p);
    }
};
class DirichletFunctionSystem : public FunctionSystem {
  private:
    int dir;
    bool right_hand;
  protected:
    explicit DirichletFunctionSystem(const DirichletFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    DirichletFunctionSystem(int d,bool rh) : dir(d),right_hand(rh) {}
    virtual ~DirichletFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
};
class NeumannFunctionSystem : public FunctionSystem {
  private:
    int dir;
    bool right_hand;
  protected:
    explicit NeumannFunctionSystem(const NeumannFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    NeumannFunctionSystem(int d,bool rh) : dir(d),right_hand(rh) { }
    virtual ~NeumannFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> grad=exactSolutionGradient(p);
      if (rot) {
        if (dir==0) {
          REAL g0=grad[0]+grad[1]+con*(grad[0]-grad[1]);
          return (right_hand ? g0 : -g0);
        } else {
          REAL g1=grad[0]+grad[1]-con*(grad[0]-grad[1]);
          return (right_hand ? g1 : -g1);
        }
      } else {
        if (dir==0) return (right_hand ? grad[0] : -grad[0]);
        else return (right_hand ? con*grad[1] : -con*grad[1]);
      }
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      Tensor<1,SPACEDIM> grad=exactSolutionGradient(p);
      if (rot) {
        if (dir==0) {
          REAL g0=grad[0]+grad[1]+con*(grad[0]-grad[1]);
          function_values[0]=(right_hand ? g0 : -g0);
        } else {
          REAL g1=grad[0]+grad[1]-con*(grad[0]-grad[1]);
          function_values[0]=(right_hand ? g1 : -g1);
        }
      } else {
        if (dir==0) function_values[0]=(right_hand ? grad[0] : -grad[0]);
        else function_values[0]=(right_hand ? con*grad[1] : -con*grad[1]);
      }
    }
};

class BabuskaSuriLockingMaterialModel : public MaterialModel {
  public:
    BabuskaSuriLockingMaterialModel() { }
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      if (rot) {
        for (int q=0;q<q_points.getNumber();q++) {
          REAL jxw=jxw_values[q];
          for (int i=0;i<dofs_per_cell;i++) {
            Tensor<1,SPACEDIM> sgi=sg(i,q);
            sgi*=jxw;
            for (int j=0;j<dofs_per_cell;j++) {
              Tensor<1,SPACEDIM> sgj=sg(j,q);
              cell_matrix(i,j)+=(sgi[0]+sgi[1])    *(sgj[0]+sgj[1])
                               +(sgi[0]-sgi[1])*con*(sgj[0]-sgj[1]);
            }
          }
        }
      } else {
        for (int q=0;q<q_points.getNumber();q++) {
          REAL jxw=jxw_values[q];
          for (int i=0;i<dofs_per_cell;i++) {
            Tensor<1,SPACEDIM> sgi=sg(i,q);
            sgi*=jxw;
            for (int j=0;j<dofs_per_cell;j++) {
              Tensor<1,SPACEDIM> sgj=sg(j,q);
              cell_matrix(i,j)+=sgi[0]*sgj[0]+sgi[1]*con*sgj[1];
            }
          }
        }
      }
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(128);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  int degree=(argc>1 ? atoi(argv[1]) : 1);
  small=(argc>2 ? atof(argv[2]) : M_PI);
  rot=(argc>3 ? strcmp(argv[3],"true")==0 : false);
  con=M_PI*M_PI/(small*small);
  cout << "\tdegree = " << degree << endl;
  cout << "\tsmall = " << small << endl;
  cout << "\trot = " << rot << endl;

  Tessellation tess;

  bool set_face_ids=false;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
//rectangular domain consisting of nelements intervals
  Point<2> low(-1.,-1.);
  Point<2> edge[2]; edge[0]=Point<2>(2.,0.);
                    edge[1]=Point<2>(0.,2.);
  int number_elements[2]={2,2};
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);

  BabuskaSuriLockingMaterialModel bslmm;
  tess.setMaterialModel(&bslmm);

  Point<2> origin(0.,0.);
  boundary.allocate(4);
  for (int d=0;d<2;d++) {
    boundary[2*d]=OPERATOR_NEW PlaneBoundary(origin,low,low+edge[1-d]);
    boundary[2*d+1]=OPERATOR_NEW PlaneBoundary(origin,low+edge[d],
      low+edge[0]+edge[1]);
  }

  boundary_condition.allocate(5);

  boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
  boundary_condition[0]->setFunctionSystem(OPERATOR_NEW
    NeumannFunctionSystem(0,false));
  boundary_condition[0]->makeNatural();

  boundary_condition[1]=
    OPERATOR_NEW BoundaryCondition(boundary[1]);
  boundary_condition[1]->setFunctionSystem(OPERATOR_NEW
    NeumannFunctionSystem(0,true));
  boundary_condition[1]->makeNatural();

  boundary_condition[2]=OPERATOR_NEW BoundaryCondition(boundary[2]);
  boundary_condition[2]->setFunctionSystem(OPERATOR_NEW
    NeumannFunctionSystem(1,false));
  boundary_condition[2]->makeNatural();

  boundary_condition[3]=
    OPERATOR_NEW BoundaryCondition(boundary[3]);
  boundary_condition[3]->setFunctionSystem(OPERATOR_NEW
    NeumannFunctionSystem(1,true));
  boundary_condition[3]->makeNatural();

  boundary_condition[4]=OPERATOR_NEW BoundaryCondition(boundary[0]);
  boundary_condition[4]->setFunctionSystem(
    OPERATOR_NEW DirichletFunctionSystem(0,false));
  boundary_condition[4]->makeEssential();

  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<SPACEDIM> center=face_element->center(vertices);
      for (int i=0;i<boundary.getNumber();i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          if (i==0) {
            if (center[1]<0.) { // Dirichlet
              codimension1_level->
                setBoundaryCondition(boundary_condition[4]);
            } else { // Neumann
              codimension1_level->
                setBoundaryCondition(boundary_condition[0]);
            }
          } else { // Neumann
            codimension1_level->
              setBoundaryCondition(boundary_condition[i]);
          }
          break;
        }
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
  tess.checkFaceBoundaryCondition(boundary_condition);
#endif

//linear coordinate mapping
//  mapping shape functions are computed on canonical elements now
  MappingL mapping;

//Lagrange polynomials to represent solution of problem
//  finite element shape functions are computed on canonical elements now
  FE_L fe(degree);

//Gaussian quadrature
  map<int,Quadrature<SPACEDIM> > qmap;
  qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
    GaussianQuadrature<2>::argGivenOrder(2*degree+1));

  map<int,Quadrature<SPACEDIM-1> > nqmap;
  nqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree+1));

  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={-1.,-1.},rhi[2]={1.,1.};
//AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
//AreaGraphTool contour_gt("solution contours","x0","x1",rlo,rhi,&cmap,
//  0,0.5);
  AreaGraphTool plot_gt("solution plot","x0","y1",rlo,rhi,&cmap,0,0.5);
//BOOLEAN use_lighting=FALSE;
//Vector3 vlow(0.,0.,0.);
//Vector3 vhigh(1.,1.,1.);
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}

  for (int cycle=0;cycle<N_CYCLES;cycle++) {

//  data structure to keep track of degrees of freedom in solver
    MGDofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();
    n_dofs[cycle]=dof_handler.numberDofs();
    cout << "\n\tNumber of degrees of freedom = " << n_dofs[cycle] << endl;

//  data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.DofHandler::makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//  storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

    dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix,system_rhs);

//  compute essential boundary values
    map<int,REAL> boundary_values;
    dof_handler.interpolateBoundaryValues(mapping,boundary_values);

//  modify linear system and rhs for essential boundary conditions
    solution.reinit(n_dofs[cycle]);
    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);
    REAL rcond=symmetricConditionNumber(system_matrix);
    cout << "\tcondition number = " << rcond << endl;

//  bi-conjugate gradient solver
    SolverControl cg_solver_control(2*solution.size(),
      DBL_EPSILON*1.e2,false,false);
    CGSolver cg(cg_solver_control);
    PreconditionIdentity pi;
    cg.solve(system_matrix,solution,system_rhs,pi);

    {
      DataOut data_out(dof_handler,mapping);
      data_out.addDataVector(solution,"solution");
      int n_subdivisions=degree;
      data_out.buildPatches(n_subdivisions);
//    data_out.drawGrid(&mesh_gt);
      bool quit_button=D_CHOOSE(false,false,true);
      data_out.plot("solution",&plot_gt,quit_button);
//    data_out.contourPlot("solution",&contour_gt);
//    data_out.plotSurface("solution",&surface_gt);
#if (SPACEDIM==1)
      if (!quit_button) {
      XYGraphTool::WINDOW_TYPE::QuitButton qb;
      }
#elif (SPACEDIM==2)
      if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
      }
#endif
    }

    {
      ExactFunctionSystem efs;
      FunctionSystem *weight=0;
      L2_error[cycle]=dof_handler.integrateDifference(mapping,solution,
        efs,qmap,L2_NORM,weight,1.);
      H1_error[cycle]=dof_handler.integrateDifference(mapping,solution,
        efs,qmap,H1_NORM,weight,1.);
      Linfty_error[cycle]=dof_handler.integrateDifference(mapping,
        solution,efs,qmap,LINFTY_NORM,weight,1.);
    }

    if (cycle<N_CYCLES-1) {
      tess.refineGlobal(1);
    }
    delete sparsity_pattern;
  }

  if (N_CYCLES>1) {
    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }


  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]->getFunctionSystem();
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
#include "NumPtr.C"
INSTANTIATE_NUMPTR(Tessellation*);
