// "$Header: /home/faculty/johnt/cvs/deal_new/examples/LinearElements1D.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "ConditionNumber.H"
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "RefinementStudy.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM!=1)
barf;
#endif

// delta u'' + 2 u' + u = 1 , 0 < x < 1
// u(0) = 0 = u(1)

int N_CYCLES=4;
REAL diffusion=.1; // diffusion=1. leads to a floating point exception

class LinSegelMaterialModel : public MaterialModel {
//Lin & Segel p. 285
  public:
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=diffusion*jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
            cell_matrix(i,j)+=(sgi[0]-2.*si)*sg(j,q)[0]-si*sv(j,q);
          }
        }
      }
    }
};
class ExactFunctionSystem : public FunctionSystem {
//Lin & Segel p. 286
  private:
    REAL root1,root2;
  public:
    ExactFunctionSystem() {
      root1=-1./(1.+sqrt(1.-diffusion));
      root2=1./(diffusion*root1);
    }
    ~ExactFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return (exp(root1*p[0])-exp(root2*p[0]))/(exp(root1)-exp(root2));
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=
        (exp(root1*p[0])-exp(root2*p[0]))/(exp(root1)-exp(root2));
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int) 
    const {
      Tensor<1,SPACEDIM> grad;
      grad[0]=(root1*exp(root1*p[0])-root2*exp(root2*p[0]))
             /(exp(root1)-exp(root2));
      return grad;
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> > &grads,
    const Point<SPACEDIM> &p) const {
      Tensor<1,SPACEDIM> &grad=grads[0];
      grad[0]=(root1*exp(root1*p[0])-root2*exp(root2*p[0]))
             /(exp(root1)-exp(root2));
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(128);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  int degree=(argc>1 ? atoi(argv[1]) : 1);
  diffusion=(argc>2 ? atof(argv[2]) : .1);
  cout << "\tdegree = " << degree << endl;
  cout << "\tdiffusion = " << diffusion << endl;

  Tessellation tess;

  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  Point<1> low(0.);
  Point<1> edge[1]; edge[0]=Point<1>(1.);
  int number_elements[1]={2};
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);

  LinSegelMaterialModel lsmm;
  tess.setMaterialModel(&lsmm);

  ConstantFunctionSystem cfs0(0.);
  ConstantFunctionSystem cfs1(1.);
  boundary_condition[0]->setFunctionSystem(&cfs0);
  boundary_condition[0]->makeEssential(); // Dirichlet boundary
  boundary_condition[1]->setFunctionSystem(&cfs1);
  boundary_condition[1]->makeEssential(); // Dirichlet boundary

//linear coordinate mapping
//  mapping shape functions are computed on canonical elements now
  MappingL mapping;

//Lagrange polynomials to represent solution of problem
//  finite element shape functions are computed on canonical elements now
  FE_L fe(degree);

//Gaussian quadrature
  map<int,Quadrature<SPACEDIM> > qmap;
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree-1));

  map<int,Quadrature<SPACEDIM-1> > nqmap;

  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool *plot_gt=OPERATOR_NEW XYGraphTool("solution","x","solution",
    0.,1.,0.,1.,&cmap,0,0.5);
  plot_gt->setbgColor("white");
  plot_gt->setfgColor("black");
  for (int cycle=0;cycle<N_CYCLES;cycle++) {

//  data structure to keep track of degrees of freedom in solver
    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();
    n_dofs[cycle]=dof_handler.numberDofs();
    cout << "\n\tNumber of degrees of freedom = " << n_dofs[cycle] << endl;

//data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle]);
    solution.reinit(n_dofs[cycle]);

    dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix,system_rhs);

//  compute essential boundary values
    map<int,REAL> boundary_values;
    dof_handler.interpolateBoundaryValues(mapping,boundary_values);

//  modify linear system and rhs for essential boundary conditions
    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);
    REAL rcond=unSymmetricConditionNumber(system_matrix);
    cout << "\tcondition number = " << rcond << endl;

//  initial guess
//  this works only for FE_L because nodal values are point values
    ExactFunctionSystem efs;
    NumPtr<Point<1> > eslp;
    Interval::staticEquallySpacedLatticePoints(degree,eslp);
    Quadrature<1> eslpq(eslp);
    map<int,Quadrature<1> > initial_guess_qmap;
    initial_guess_qmap[2]=eslpq;
    UPDATE_FLAGS flags=UPDATE_VALUES | UPDATE_Q_POINTS;
    FEValues initial_guess_fe_values(mapping,fe,initial_guess_qmap,flags);
    for (bool doit=tess.firstActiveShape();doit;
    doit=tess.nextActiveShape()) {
      LEVEL_CLASS *present_level=tess.presentLevel();
      const SHAPE_CLASS *element=present_level->getShape();
      initial_guess_fe_values.mapShapes(tess,element,present_level);
      const NumPtr<Point<SPACEDIM> > &q_points=
        initial_guess_fe_values.getQuadraturePoints(element);
      int n_q_points=q_points.getNumber();
      NumPtr<REAL> asv(n_q_points);
      efs.FunctionSystem::value(asv,q_points,0);
      NumPtr<int> local_dof_indices;
      dof_handler.getDofIndices(element,present_level,local_dof_indices);
      int dofs_per_cell=fe.dofsPerCell(element);
      for (int i=0;i<dofs_per_cell;i++) {
        solution[local_dof_indices[i]]=asv[i];
      }
    }

    RelativeErrorControl solver_control(2*solution.size(),
      1.e-8,false,false);
    PreconditionIdentity pi;
//  GMRESSolver gmres(solver_control,5);
//  gmres.solve(system_matrix,solution,system_rhs,pi);
//  MinResSolver minres(solver_control);
//  minres.solve(system_matrix,solution,system_rhs,pi);
    BicgstabSolver bicg(solver_control);
    bicg.solve(system_matrix,solution,system_rhs,pi);
//  REAL spectral_radius=
//    1.+4.*diffusion*static_cast<REAL>(solution.size()-1);
//  RichardsonSolver rich(solver_control,1./spectral_radius,false);
//  rich.solve(system_matrix,solution,system_rhs,pi);

    FunctionSystem *weight=0;
    L2_error[cycle]=dof_handler.integrateDifference(mapping,solution,
      efs,qmap,L2_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tL2_error[" << cycle << "] = " << L2_error[cycle] << endl;
#endif
    
    H1_error[cycle]=dof_handler.integrateDifference(mapping,solution,
      efs,qmap,H1_NORM,weight,1.);
   
    Linfty_error[cycle]=dof_handler.integrateDifference(mapping,
      solution,efs,qmap,LINFTY_NORM,weight,1.);

//  plot solution
    DataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
    data_out.buildPatches(degree);
    bool quit_button=D_CHOOSE(false,false,true);
    data_out.plot("solution",plot_gt,quit_button);
#if (SPACEDIM==1)
    if (!quit_button) {
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    if (!quit_button) {
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif

    if (cycle<N_CYCLES-1) tess.refineGlobal(1);
    delete sparsity_pattern;
  }
#if (SPACEDIM<=2)
  delete plot_gt; plot_gt=0;
#endif

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
#include "NumPtr.C"
INSTANTIATE_NUMPTR(Tessellation*);
