// "$Header: /home/faculty/johnt/cvs/deal_new/examples/LinearElements1D.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM!=1)
barf // hard-wired for 1D
#endif

//-u''+u=0, u(-1)=-sinh(1), u'(1)=cosh(1) ==> u(x)=sinh(x)
int degree=1;
int nelements=10;
int N_CYCLES=4;

//description of pde right-hand side and exact solution (to compute errors)
REAL rightHandSideValue(const Point<1> &p) { return 0.; }
REAL exactSolutionValue(const Point<1> &p) { return sinh(p[0]); }
Tensor<1,1> exactSolutionGradient(const Point<1> &p) {
  Tensor<1,1> return_val;
  return_val[0]=cosh(p[0]);
  return return_val;
}

//class to describe essential boundary condition
class DirichletFunctionSystem : public FunctionSystem {
  protected:
    explicit DirichletFunctionSystem(const FunctionSystem &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    DirichletFunctionSystem() {;}
    virtual ~DirichletFunctionSystem() {;}
    virtual REAL value(const Point<1> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<1> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int) const {
      return exactSolutionGradient(p);
    }
    void gradient(NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
    const Point<SPACEDIM> &p) const {
      return_gradients[0]=exactSolutionGradient(p);
    }
};
class NeumannFunctionSystem : public FunctionSystem {
  protected:
    explicit NeumannFunctionSystem(const NeumannFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    NeumannFunctionSystem() {}
    virtual ~NeumannFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      return g[0];
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      for (int i=0;i<function_values.size();i++) function_values[i]=g[i];
    }
};
class SimpleMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<1> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,1> sgi=sg(i,q);
          sgi*=jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
//          element-wise contribution to stiffness matrix is
//            integral grad v_i . grad v_j + v_i * v_j
//          ie, pde is -div ( 1 grad u ) + 1 u = ...
            cell_matrix(i,j)+=sgi*sg(j,q);
            cell_matrix(i,j)+=si*sv(j,q);
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << "\tsinh solution" << endl;
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

//rectangular domain consisting of nelements intervals
  Point<1> low(-1.);
  Point<1> edge[1]; edge[0]=Point<1>(2.);
  int number_elements[1]={nelements};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);

  SimpleMaterialModel sm;
  tess.setMaterialModel(&sm);

//boundary conditions
  DirichletFunctionSystem dfs;
  boundary_condition[0]->setFunctionSystem(&dfs);
  boundary_condition[0]->makeEssential(); // Dirichlet boundary

  NeumannFunctionSystem nfs;
  boundary_condition[1]->setFunctionSystem(&nfs);
  boundary_condition[1]->makeNatural(); // Neumann boundary

//linear coordinate mapping
//  mapping shape functions are computed on canonical elements now
  MappingL mapping;

//Lagrange polynomials to represent solution of problem
//  finite element shape functions are computed on canonical elements now
  FE_L fe(degree);

//Gaussian quadrature
  map<int,Quadrature<1> > qmap;
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree-1));

//Dirichlet bc evaluation at Neumann boundary (right)
  map<int,Quadrature<1> > dbc_qmap;
  Point<1> q0(1.);
  dbc_qmap[Interval::verticesPerShape()]=Quadrature<1>(q0);
//Neumann bc evaluation at Dirichlet boundary (left)
  map<int,Quadrature<1> > nbc_qmap;
  Point<1> q1(-1.);
  nbc_qmap[Interval::verticesPerShape()]=Quadrature<1>(q1);

//data structure to hold mapped shape function information
  UPDATE_FLAGS update_flags=
    UPDATE_VALUES | UPDATE_GRADIENTS | UPDATE_JxW_VALUES | 
    UPDATE_Q_POINTS;
  FEValues fe_values(mapping,fe,qmap,update_flags); 
  FEValues dbc_fe_values(mapping,fe,dbc_qmap,UPDATE_VALUES); 
  FEValues nbc_fe_values(mapping,fe,nbc_qmap,UPDATE_GRADIENTS); 

//quadrature rule on essential boundaries (not important in 1d)
  map<int,Quadrature<0> > dqmap; // essential bc's

  double log_n_dofs[N_CYCLES];
  double L2_error[N_CYCLES];
  double H1_error[N_CYCLES];
  double Linfty_error[N_CYCLES];
  double DBC_error[N_CYCLES];
  double NBC_error[N_CYCLES];
  double log10=log(10.);

  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool plot_gt("solution plot","x","y",0.,1.,0.,1.,&cmap,0,0.5);
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
#ifdef DEBUG
    cout << "cycle = " << cycle << endl;
#endif

//  data structure to keep track of degrees of freedom in solver
    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();
    int n_dofs=dof_handler.numberDofs();
    log_n_dofs[cycle]=log(static_cast<double>(n_dofs))/log10;
    cout << "\n\tNumber of degrees of freedom = " << n_dofs << endl;

//  data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs,n_dofs,
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//  storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs,0.);
    Vector solution(n_dofs,0.);

//assemble linear system
//  this will call simpleMaterialModel::elementStiffnessMatrixAndRHS
    map<int,Quadrature<0> > fqmap;
    dof_handler.assembleLinearSystem(mapping,qmap,fqmap,
      system_matrix,system_rhs);

//  compute essential boundary values
    map<int,REAL> boundary_values;
    dof_handler.interpolateBoundaryValues(mapping,boundary_values,dqmap);

//  modify linear system and right-hand side for Dirichlet bc's
    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);

//  conjugate gradient solver
    SolverControl 
      cg_solver_control(solution.size(),DBL_EPSILON*1.e2,false,false);
    CGSolver cg(cg_solver_control);
    PreconditionIdentity pi;
    cg.solve(system_matrix,solution,system_rhs,pi);

//  compute various error measures
    int n_active_cells=tess.numberActiveShapes();
    FunctionSystem *weight=0;
    REAL error=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,L2_NORM,weight,1.);
    L2_error[cycle]=log(error)/log10;
#ifdef DEBUG
    cout << "\tL2_error(" << cycle << ") = " << error << endl;
#endif
    
    error=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,H1_NORM,weight,1.);
    H1_error[cycle]=log(error)/log10;
#ifdef DEBUG
    cout << "\tH1_error(" << cycle << ") = " << error << endl;
#endif
 
    error=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,LINFTY_NORM,weight,1.);
    Linfty_error[cycle]=log(error)/log10;
#ifdef DEBUG
    cout << "\tLinfinity_error(" << cycle << ") = " << error << endl;
#endif

    map<int,Quadrature<SPACEDIM-1> > bc_qmap;
    DBC_error[cycle]=dof_handler.integrateDifferenceOnBoundary(mapping,
      solution,dfs,qmap,bc_qmap,L2_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tL2_NORM on boundary(" << cycle << ") = " << error << endl;
#endif
    NBC_error[cycle]=dof_handler.integrateDifferenceOnBoundary(mapping,
      solution,dfs,qmap,bc_qmap,H1_SEMINORM,weight,1.);
#ifdef DEBUG
    cout << "\tH1_SEMINORM on boundary(" << cycle << ") = " << error 
         << endl;
#endif

//  plot solution
    DataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
    int n_subdivisions=640/nelements;
    data_out.buildPatches(n_subdivisions);
    bool quit_button=D_CHOOSE(false,false,true);
    data_out.plot("solution",&plot_gt,quit_button);
#if (SPACEDIM==1)
    if (!quit_button) {
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    if (!quit_button) {
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif

    if (cycle<N_CYCLES-1) tess.refineGlobal(1);
    delete sparsity_pattern;
  }

  double ymin=HUGE_VAL;
  double ymax=-HUGE_VAL;
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
    ymin=min(ymin,L2_error[cycle]);
    ymax=max(ymax,L2_error[cycle]);
  }
  XYGraphTool L2_gt("L2 Error","log_10(degrees of freedom)",
    "log_10(L2 error)",log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,
    &cmap,0,0.5);
  L2_gt.setbgColor("white");
  L2_gt.setfgColor("black");
  L2_gt.drawAxes();
  L2_gt.setfgColor("blue");
  L2_gt.movePen(log_n_dofs[0],L2_error[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    L2_gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
  }
  L2_gt.flush();

  ymin=HUGE_VAL;
  ymax=-HUGE_VAL;
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
    ymin=min(ymin,H1_error[cycle]);
    ymax=max(ymax,H1_error[cycle]);
  }
  XYGraphTool H1_gt("H1 Error","log_10(degrees of freedom)",
    "log_10(H1 error)",log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,
    &cmap,0,0.5);
  H1_gt.setbgColor("white");
  H1_gt.setfgColor("black");
  H1_gt.drawAxes();
  H1_gt.setfgColor("blue");
  H1_gt.movePen(log_n_dofs[0],H1_error[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    H1_gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
  }
  H1_gt.flush();

  ymin=HUGE_VAL;
  ymax=-HUGE_VAL;
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
    ymin=min(ymin,Linfty_error[cycle]);
    ymax=max(ymax,Linfty_error[cycle]);
  }
  XYGraphTool Linfty_gt("Linfty Error","log_10(degrees of freedom)",
    "log_10(Linfty error)",log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,
    &cmap,0,0.5);
  Linfty_gt.setbgColor("white");
  Linfty_gt.setfgColor("black");
  Linfty_gt.drawAxes();
  Linfty_gt.setfgColor("blue");
  Linfty_gt.movePen(log_n_dofs[0],Linfty_error[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    Linfty_gt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
  }
  Linfty_gt.flush();

  ymin=HUGE_VAL;
  ymax=-HUGE_VAL;
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
    ymin=min(ymin,DBC_error[cycle]);
    ymax=max(ymax,DBC_error[cycle]);
  }
  XYGraphTool DBC_gt("Dirichlet BC Error","log_10(degrees of freedom)",
    "log_10(Linfty error)",log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,
    &cmap,0,0.5);
  DBC_gt.setbgColor("white");
  DBC_gt.setfgColor("black");
  DBC_gt.drawAxes();
  DBC_gt.setfgColor("blue");
  DBC_gt.movePen(log_n_dofs[0],DBC_error[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    DBC_gt.drawLine(log_n_dofs[cycle],DBC_error[cycle]);
  }
  DBC_gt.flush();

  ymin=HUGE_VAL;
  ymax=-HUGE_VAL;
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
    ymin=min(ymin,NBC_error[cycle]);
    ymax=max(ymax,NBC_error[cycle]);
  }
  XYGraphTool NBC_gt("Dirichlet BC Error","log_10(degrees of freedom)",
    "log_10(Linfty error)",log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,
    &cmap,0,0.5);
  NBC_gt.setbgColor("white");
  NBC_gt.setfgColor("black");
  NBC_gt.drawAxes();
  NBC_gt.setfgColor("blue");
  NBC_gt.movePen(log_n_dofs[0],NBC_error[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    NBC_gt.drawLine(log_n_dofs[cycle],NBC_error[cycle]);
  }
  NBC_gt.flush();

  XYGraphTool::WINDOW_TYPE::QuitButton qb;

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) { 
    delete boundary_condition[i]; boundary_condition[i]=0; 
  }
  for (int i=0;i<boundary.getNumber();i++) { 
    delete boundary[i]; boundary[i]=0; 
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
