// "$Header: /home/faculty/johnt/cvs/deal_new/examples/ConvergenceStudy.C,v 1.1 2009/08/20 17:27:33 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <iostream>

#include <AreaGraphTool.H>
#include "Boundary.H"
#include "BoundaryCondition.H"
#include "ConstraintMatrix.H"
#include "DataOut.H"
#include "DofHandler.H"
#include "DofRenumbering.H"
#include "FiniteElement.H"
#include "FEValues.H"
#include "GenerateTess.H"
#include "KellyErrorEstimator.H"
#include "Mapping.H"
#include "Point.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "Tensor.H"
#include "Types.H"

// - div grad u + u = sum of exponentials
// some Dirichlet and some Neumann bc's

enum REFINEMENT_MODE { GLOBAL_REFINEMENT, ADAPTIVE_REFINEMENT };
#define N_CYCLES 7

static const int n_source_centers=3;    
#if (SPACEDIM==1)
const Point<1> source_centers[3] =
  { Point<1>(-1./3.), Point<1>(0.), Point<1>(1./3.) };
#endif
#if (SPACEDIM==2)
const Point<2> source_centers[3] = 
  { Point<2>(-0.5,  0.5), Point<2>(-0.5, -0.5), Point<2>( 0.5, -0.5) };
#endif
const REAL width=1./3.;

REAL rightHandSideValue(const Point<SPACEDIM> &p) {
  REAL return_value=0.;
  for (int i=0;i<n_source_centers;i++) {
    Point<SPACEDIM> shifted_point=p-source_centers[i];
    REAL sps=shifted_point.square();
    REAL w2=width*width;
    return_value+=exp(-sps/w2)
      *((2.*static_cast<REAL>(SPACEDIM)-4.*sps/w2)/w2+1.);
  }
  return return_value;
}

REAL exactSolutionValue(const Point<SPACEDIM> &p) {
  REAL return_value=0.;
  REAL w2=width*width;
  for (int i=0;i<n_source_centers;i++) {
    const Point<SPACEDIM> shifted_point=p-source_centers[i];
    return_value+=exp(-shifted_point.square()/w2); 
  }
  return return_value;
}

Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &p) {
  Tensor<1,SPACEDIM> return_value;
  REAL w2=width*width;
  for (int i=0;i<n_source_centers;i++) {
    Tensor<1,SPACEDIM> shifted_point(p-source_centers[i]);
    shifted_point*= -2.*exp(-shifted_point*shifted_point/w2)/w2;
    return_value+=shifted_point;
  }
  return return_value;
}

class DirichletFunctionSystem : public FunctionSystem {
  protected:
    explicit DirichletFunctionSystem(const FunctionSystem &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    DirichletFunctionSystem() {;}
    virtual ~DirichletFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector&) const {
      OBSOLETE("not programmed");
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
};
class NeumannFunctionSystem : public FunctionSystem {
  private:
    int dir;
    bool left_hand;
  protected:
    explicit NeumannFunctionSystem(const NeumannFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    NeumannFunctionSystem(int d,bool lh) : dir(d),left_hand(lh) {}
    virtual ~NeumannFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      return (left_hand ? -g[dir] : g[dir]);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      function_values[0]=(left_hand ? -g[dir] : g[dir]);
    }
};

class SimpleMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
            cell_matrix(i,j)+=sgi*sg(j,q)+si*sv(j,q);
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};

void solveLaplaceProblem(int degree,
REFINEMENT_MODE refinement_mode,
D_CHOOSE(XYGraphTool,AreaGraphTool,VolGraphTool) *mgt,
D_CHOOSE(XYGraphTool,AreaGraphTool,VolGraphTool) *sgt) {

  Tessellation tess;

  Point<SPACEDIM> low(D_DECL(-1.,-1.,-1.));
  Point<SPACEDIM> edge[SPACEDIM];
  edge[0]=Point<SPACEDIM>(D_DECL(2.,0.,0.));
#if (SPACEDIM>1)
  edge[1]=Point<SPACEDIM>(D_DECL(0.,2.,0.));
#endif
#if (SPACEDIM==3)
  edge[2]=Point<SPACEDIM>(D_DECL(0.,0.,2.));
#endif
  int number_elements[SPACEDIM]={D_DECL(1,1,1)};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);
  SimpleMaterialModel smm;
  tess.setMaterialModel(&smm);

  DirichletFunctionSystem dfs;
  for (int d=0;d<SPACEDIM;d++) {
    boundary_condition[2*d]->setFunctionSystem(&dfs);
    boundary_condition[2*d]->makeEssential();//Dirichlet back, left & bottom
    boundary_condition[2*d+1]->setFunctionSystem(
      OPERATOR_NEW NeumannFunctionSystem(d,false));
    boundary_condition[2*d+1]->makeNatural();//Neumann front, right & top
  }

  tess.refineGlobal(1);

  MappingL mapping;

  FE_L fe(degree);
  GaussianQuadrature<SPACEDIM-1> face_quadrature_formula(
    GaussianQuadrature<SPACEDIM-1>::argGivenOrder(
      D_CHOOSE(degree,degree,degree+1))); 

  map<int,Quadrature<SPACEDIM> > qmap;
  int nv=D_CHOOSE(Interval::verticesPerShape(),
                  Quadrilateral::verticesPerShape(),
                  Hexahedron::verticesPerShape());
  qmap[nv]=GaussianQuadrature<SPACEDIM>(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(
      D_CHOOSE(2*degree-1,2*degree+1,2*degree+1)));

//natural bc's
//should be exact for polynomials of shape function degree
//  plus exact solution degree-1
//  plus Jacobian degree
  map<int,Quadrature<SPACEDIM-1> > nqmap;
  nv=D_CHOOSE(0,Interval::verticesPerShape(),
              Quadrilateral::verticesPerShape());
  nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
    GaussianQuadrature<SPACEDIM-1>::argGivenOrder(
      D_CHOOSE(2*degree-1,2*degree+1,2*degree+1)));

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];
  cout << "               active           L2              H1              Linfinity" 
       << endl;
  cout << "      cycle    cells   dofs     error           error           error" << endl;
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();

    cuthillMcKeeRenumbering(dof_handler);

    ConstraintMatrix hanging_node_constraints;
    hanging_node_constraints.makeHangingNodeConstraints(dof_handler);
    hanging_node_constraints.close();
    n_dofs[cycle]=dof_handler.numberDofs();

    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);

    hanging_node_constraints.condense(*sparsity_pattern);
    sparsity_pattern->compress();

    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

    dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix,system_rhs);
    hanging_node_constraints.condense(system_matrix);
    hanging_node_constraints.condense(system_rhs);

    map<int,REAL> boundary_values;
    dof_handler.interpolateBoundaryValues(mapping,boundary_values);

    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);

    RelativeErrorControl solver_control(solution.size(),DBL_EPSILON*1.e2,
      false,false);
    CGSolver cg(solver_control);
    PreconditionSSOR preconditioner(system_matrix,1.2);
    cg.solve(system_matrix,solution,system_rhs,preconditioner);

    hanging_node_constraints.distribute(solution);

    n_active_cells[cycle]=tess.numberActiveShapes();
    FunctionSystem *weight=0;
    L2_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,L2_NORM,weight,1.);

    H1_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,H1_SEMINORM,weight,1.);

    Linfty_error[cycle]=dof_handler.integrateDifference(mapping,solution,
      dfs,qmap,LINFTY_NORM,weight,1.);

//  cout << "Cycle " << cycle << ':' << endl
//            << "   Number of active cells:       "
//            << n_active_cells[cycle] << endl
//            << "   Number of degrees of freedom: "
//            << n_dofs[cycle] << endl;
    cout << "\t" << cycle << "\t" << n_active_cells[cycle]
      << "\t" << n_dofs[cycle] << "\t" << L2_error[cycle] << " \t" 
      << H1_error[cycle] << " \t" << Linfty_error[cycle] << endl;

    DataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
    int n_subdivisions=2;
    data_out.buildPatches(n_subdivisions);
    bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>=2)
    if (mgt) data_out.drawGrid(mgt,quit_button);
#endif
//  data_out.contourPlot(gt);
    if (sgt) data_out.plot("solution",sgt,quit_button);
#if (SPACEDIM<3)
    if (mgt!=0 || sgt!=0) {
#if (SPACEDIM==1)
      if (!quit_button) {
        XYGraphTool::WINDOW_TYPE::QuitButton qb;
      }
#endif
#if (SPACEDIM==2)
      if (!quit_button) {
        AreaGraphTool::WINDOW_TYPE::QuitButton qb;
      }
#endif
    }
#endif
    if (cycle<N_CYCLES-1) {
      switch (refinement_mode) {
        case GLOBAL_REFINEMENT: {
          tess.refineGlobal(1);
          break;
        }
        case ADAPTIVE_REFINEMENT: {
          Vector estimated_error_per_cell(n_active_cells[cycle]);
          KellyErrorEstimator error_estimator(mapping,dof_handler,
            nqmap,solution,estimated_error_per_cell);
          tess.refineAndCoarsenFixedNumber(estimated_error_per_cell,0.3,
            0.03);
          tess.executeCoarseningAndRefinement();
          break;
        }
      }
    }
    delete sparsity_pattern;
  }
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

  XYGraphTool nacgt("number active cells vs cycle","cycle",
    "number active cells",0,N_CYCLES-1,n_active_cells[0],
    n_active_cells[N_CYCLES-1],&cmap,0,0.5);
  nacgt.newPage();
  nacgt.setbgColor("white");
  nacgt.setfgColor("black");
  nacgt.drawAxes();
  nacgt.setfgColor("blue");
  nacgt.movePen(0,n_active_cells[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    nacgt.drawLine(cycle,n_active_cells[cycle]);
  }
  nacgt.flush();

  XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
    "number degress of freedom",0,N_CYCLES-1,n_dofs[0],
    n_dofs[N_CYCLES-1],&cmap,0,0.5);
  ndgt.newPage();
  ndgt.setbgColor("white");
  ndgt.setfgColor("black");
  ndgt.drawAxes();
  ndgt.setfgColor("blue");
  ndgt.movePen(0,n_dofs[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    ndgt.drawLine(cycle,n_dofs[cycle]);
  }
  ndgt.flush();

  REAL ymin=HUGE_VAL;
  REAL ymax=-HUGE_VAL;
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    ymin=min(ymin,L2_error[cycle]);
    ymax=max(ymax,L2_error[cycle]);
  }
  XYGraphTool l2gt("L2 error vs cycle","cycle","L2 error",0,N_CYCLES-1,
    ymin,ymax,&cmap,0,0.5);
  l2gt.newPage();
  l2gt.setbgColor("white");
  l2gt.setfgColor("black");
  l2gt.drawAxes();
  l2gt.setfgColor("blue");
  l2gt.movePen(0,L2_error[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    l2gt.drawLine(cycle,L2_error[cycle]);
  }
  l2gt.flush();

  ymin=HUGE_VAL;
  ymax=-HUGE_VAL;
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    ymin=min(ymin,H1_error[cycle]);
    ymax=max(ymax,H1_error[cycle]);
  }
  XYGraphTool h1gt("H1 error vs cycle","cycle","H1 error",0,N_CYCLES-1,
    ymin,ymax,&cmap,0,0.5);
  h1gt.newPage();
  h1gt.setbgColor("white");
  h1gt.setfgColor("black");
  h1gt.drawAxes();
  h1gt.setfgColor("blue");
  h1gt.movePen(0,H1_error[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    h1gt.drawLine(cycle,H1_error[cycle]);
  }
  h1gt.flush();

  ymin=HUGE_VAL;
  ymax=-HUGE_VAL;
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    ymin=min(ymin,Linfty_error[cycle]);
    ymax=max(ymax,Linfty_error[cycle]);
  }
  XYGraphTool linftygt("Linfty error vs cycle","cycle","Linfty error",
    0,N_CYCLES-1,ymin,ymax,&cmap,0,0.5);
  linftygt.newPage();
  linftygt.setbgColor("white");
  linftygt.setfgColor("black");
  linftygt.drawAxes();
  linftygt.setfgColor("blue");
  linftygt.movePen(0,Linfty_error[0]);
  for (int cycle=1;cycle<N_CYCLES;cycle++) {
    linftygt.drawLine(cycle,Linfty_error[cycle]);
  }
  linftygt.flush();

  XYGraphTool::WINDOW_TYPE::QuitButton qb;

  delete boundary_data; boundary_data=0;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    if (boundary_condition[i]->isNatural()) {
      delete boundary_condition[i]->getFunctionSystem();
    }
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
}

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  Palette pal("hot_cold");
  pal.insertEntry(0,"white");
  pal.insertEntry(1,"black");
  pal.insertEntry(2,"blue");
  pal.insertEntry(19,"cyan");
  pal.insertEntry(39,"green");
  pal.insertEntry(49,"yellow");
  pal.insertEntry(59,"magenta");
  pal.insertEntry(79,"red");
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool sgt=("solution","x","solution",0.,1.,0.,1.,&cmap,0,0.5);
#endif
#if (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2],rhi[2];
  rlo[0]=rlo[1]=0.;
  rhi[0]=rhi[1]=1.;
  AreaGraphTool mgt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
  AreaGraphTool sgt("solution","x0","x1",rlo,rhi,&cmap,0,0.5);
#endif
#if (SPACEDIM==3)
  Vector3 low(0.,0.,0.),high(1.,1.,1.);
  bool use_lighting=false;
  VolGraphTool mgt("grid",low,high,pal,use_lighting,0.5);
  VolGraphTool sgt("solution",low,high,pal,use_lighting,0.5);
#endif
  {
    cout << "Solving with linear elements, adaptive refinement\n"
         << "=============================================" << endl;
    solveLaplaceProblem(1,ADAPTIVE_REFINEMENT,&mgt,&sgt);
    cout << endl;
  }
  {
    cout << "Solving with linear elements, global refinement\n" 
         << "=============================================" << endl;
    solveLaplaceProblem(1,GLOBAL_REFINEMENT,&mgt,&sgt);
    cout << endl;
  }
  {
    cout << "Solving with quadratic elements, global refinement\n" 
         << "=============================================" << endl;
    solveLaplaceProblem(2,GLOBAL_REFINEMENT,&mgt,&sgt);
    cout << endl;
  }
  return EXIT_SUCCESS;
}
