// "$Header: /home/faculty/johnt/cvs/deal_new/examples/LinearElements1D.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "RefinementStudy.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

//- div grad u = delta(a)

int degree=1;
int N_CYCLES=D_CHOOSE(7,5,3);
Point<SPACEDIM> pole;
#if (SPACEDIM>1)
REAL radius=1.;
Point<SPACEDIM> center(D_DECL(0.,0.,0.));
Point<SPACEDIM> reflected_pole;
REAL pole_norm;
#endif

REAL exactSolutionValue(const Point<SPACEDIM> &x) { 
#if (SPACEDIM==1)
  return (x[0]<pole[0] ? (1.-pole[0])*x[0] : pole[0]*(1.-x[0]));
#elif (SPACEDIM==2)
  if (pole_norm>0.) {
    REAL xpn=sqrt((x-pole).square());
    if (xpn<=0.) return HUGE_VAL;
    REAL xrpn=sqrt((x-reflected_pole).square());
    if (xrpn<=0.) return -HUGE_VAL;
    return -0.5*log(xpn*radius/(xrpn*pole_norm))/M_PI;
  } else {
    REAL xn=sqrt(x.square());
    if (xn<=0.) return HUGE_VAL;
    return -0.5*log(xn/radius)/M_PI;
  }
#else
  if (pole_norm>0.) {
    REAL xpn=sqrt((x-pole).square());
    if (xpn<=0.) return HUGE_VAL;
    REAL xrpn=sqrt((x-reflected_pole).square());
    if (xrpn<=0.) return -HUGE_VAL;
    return -0.25*(radius/(pole_norm*xrpn)-1./xpn)/M_PI;
  } else {
    REAL xn=sqrt(x.square());
    if (xn<=0.) return HUGE_VAL;
    return -0.25*(1./radius-1./xn)/M_PI;
  }
#endif
}
Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &x) {
  Point<SPACEDIM> rv;
#if (SPACEDIM==1)
  rv[0]=(x[0]<pole[0] ? 1.-pole[0] : - pole[0]);
#elif (SPACEDIM==2)
  if (pole_norm>0.) {
    Point<SPACEDIM> xp(x-pole);
    REAL xpn2=xp.square();
    if (xpn2<=0.) rv=-HUGE_VAL;
    else {
      Point<SPACEDIM> xrp(x-reflected_pole);
      REAL xrpn2=xrp.square();
      if (xrpn2<=0.) rv=HUGE_VAL;
      else rv=(xp/xpn2-xrp/xrpn2)*(-0.5/M_PI);
    }
  } else {
    REAL xn2=x.square();
    if (xn2<=0.) rv=-HUGE_VAL;
    else rv=x*(-0.5/(M_PI*xn2));
  }
#else
  if (pole_norm>0.) {
    Point<SPACEDIM> xp(x-pole);
    REAL xpn2=xp.square();
    if (xpn2<=0.) rv=HUGE_VAL;
    else {
      Point<SPACEDIM> xrp(x-reflected_pole);
      REAL xrpn2=xrp.square();
      if (xrpn2<=0.) rv=-HUGE_VAL;
      else rv=(xrp*(radius/(pole_norm*xrpn2))-xp/xpn2)*(-0.25/M_PI);
    }
  } else {
    REAL xn2=x.square();
    if (xn2<=0.) rv=HUGE_VAL;
    else rv=x*(-0.25/(M_PI*xn2));
  }
#endif
  return Tensor<1,SPACEDIM>(rv);
}

class ExactFunctionSystem : public FunctionSystem {
  protected:
    explicit ExactFunctionSystem(const FunctionSystem &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ExactFunctionSystem() {;}
    virtual ~ExactFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
};
class BogusMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,const Array<2,REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >&,Matrix&,Vector&) const {
      OBSOLETE("should not be called");
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(128);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  drand48();
#if (SPACEDIM==1)
//rectangular domain consisting of nelements intervals
  pole[0]=M_PI/6.;
  Point<1> low(0.);
  Point<1> edge[1]; edge[0]=Point<1>(1.);
  int number_elements[1]={10};
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);
#elif (SPACEDIM==2)
  pole[0]=sqrt(2.);
  pole[1]=sqrt(3.);
  pole_norm=sqrt(pole.square());
  pole*=(M_PI/(6.*pole_norm));
  pole_norm=M_PI/6.;
  reflected_pole=pole*pow(radius/pole_norm,2);
  BoundaryData *boundary_data=generateBallTriangles(tess,center,radius,
    set_face_ids,boundary,boundary_condition);
#else
  pole[0]=drand48();
  pole[1]=drand48();
  pole[2]=drand48();
  pole_norm=sqrt(pole.square());
  pole*=(M_PI/(6.*pole_norm));
  pole_norm=M_PI/6.;
  reflected_pole=pole*pow(radius/pole_norm,2);
  BoundaryData *boundary_data=generateBallTetrahedra(tess,center,radius,
    set_face_ids,boundary,boundary_condition);
#endif

  BogusMaterialModel bmm;
  tess.setMaterialModel(&bmm);

  ConstantFunctionSystem dfs(0.);
  for (int i=0;i<boundary_condition.getNumber();i++) {
    boundary_condition[i]->setFunctionSystem(&dfs);
    boundary_condition[i]->makeEssential(); // Dirichlet boundary
  }

//linear coordinate mapping
//  mapping shape functions are computed on canonical elements now
  MappingL mapping;

//Lagrange polynomials to represent solution of problem
//  finite element shape functions are computed on canonical elements now
  FE_L fe(degree);

//Gaussian quadrature
  map<int,Quadrature<SPACEDIM> > qmap;
#if (SPACEDIM==1)
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree-1));
#elif (SPACEDIM==2)
  qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
    TriangleGaussianQuadrature::argGivenOrder(2*degree-1));
#else
  qmap[Tetrahedron::verticesPerShape()]=TetrahedronGaussianQuadrature(
    TetrahedronGaussianQuadrature::argGivenOrder(2*degree-1));
#endif

  map<int,Quadrature<SPACEDIM-1> > fqmap;
#if (SPACEDIM==2)
  fqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree-1));
#elif (SPACEDIM==3)
  fqmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
    TriangleGaussianQuadrature::argGivenOrder(2*degree-1));
#endif

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

  NumPtr<Vector> solutions(N_CYCLES);
  NumPtr<DofHandler*> dof_handlers(N_CYCLES);
  NumPtr<Tessellation*> tessellations(N_CYCLES);
  NumPtr<LEVEL_CLASS*> coarsest_levels(N_CYCLES);

#if (SPACEDIM==1)
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool *plot_gt=OPERATOR_NEW XYGraphTool("solution error","x",
    "error",0.,1.,0.,1.,&cmap,0,0.5);
  plot_gt->setbgColor("white");
  plot_gt->setfgColor("black");
#elif (SPACEDIM==2)
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.};
  REAL rhi[2]={1.,1.};
  AreaGraphTool *plot_gt=OPERATOR_NEW AreaGraphTool("solution error","x",
    "y",rlo,rhi,&cmap,0,0.5);
  plot_gt->setbgColor("white");
  plot_gt->setfgColor("black");
#endif
  for (int cycle=0;cycle<N_CYCLES;cycle++) {

    LEVEL_CLASS *copy_coarser_level=0;
    for (LEVEL_CLASS *level=tess.coarsestLevel();level;
    level=level->finerLevel()) {
      if (copy_coarser_level==0) {
        copy_coarser_level=coarsest_levels[cycle]=
          OPERATOR_NEW LEVEL_CLASS();
        coarsest_levels[cycle]->copyLevel(*level);
      } else {
        LEVEL_CLASS *cl=OPERATOR_NEW LEVEL_CLASS(copy_coarser_level);
        cl->copyLevel(*level);
        copy_coarser_level=cl;
      }
    }
    tessellations[cycle]=OPERATOR_NEW Tessellation();
    tessellations[cycle]->copyTessellation(tess,coarsest_levels[cycle]);

//  data structure to keep track of degrees of freedom in solver
    dof_handlers[cycle]=OPERATOR_NEW DofHandler(*tessellations[cycle],fe);
    dof_handlers[cycle]->distributeDofs();
    n_dofs[cycle]=dof_handlers[cycle]->numberDofs();
    cout << "\n\tNumber of degrees of freedom = " << n_dofs[cycle] << endl;

//data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handlers[cycle]->maxCouplingsBetweenDofs());
    dof_handlers[cycle]->makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

//  from DofHandler::assembleLinearSystem
//    cannot call MaterialModel::elementStiffnessMatrixAndRHS
//    because the rightHandSide is a GENERALIZED function
    {
      UPDATE_FLAGS update_flags= UPDATE_VALUES | UPDATE_GRADIENTS | 
        UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
      FEValues fe_values(mapping,fe,qmap,update_flags);
#if (SPACEDIM>1)
      UPDATE_FLAGS face_update_flags=UPDATE_VALUES | UPDATE_Q_POINTS |
        UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES;
      FEFaceValues fe_face_values(mapping,fe,fqmap,face_update_flags);
#endif
      const NumPtr<Point<SPACEDIM> > &vertices=tessellations[cycle]->getVertices();
      for (bool doit=tessellations[cycle]->firstActiveShape();doit;
      doit=tessellations[cycle]->nextActiveShape()) {
        LEVEL_CLASS *present_level=tessellations[cycle]->presentLevel();
        const SHAPE_CLASS *element=present_level->getShape();
        int dofs_per_cell=fe.dofsPerCell(element);

//      local stiffness matrix and right-hand side
        Matrix cell_matrix(dofs_per_cell,dofs_per_cell);
        Vector cell_rhs(dofs_per_cell);
        cell_matrix=0.;
        cell_rhs=0.;

//      shape functions and quadrature points are mapped 
//        from canonical elements to present element
        fe_values.mapShapes(*tessellations[cycle],element,present_level);
        const NumPtr<Point<SPACEDIM> > &q_points=
          fe_values.getQuadraturePoints(element);
        const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
        const Array<2,REAL> &sv=*fe_values.getShapeValues(element);
        const Array<2,Tensor<1,SPACEDIM> > &sg=
          *fe_values.getShapeGradients(element);

        for (int q=0;q<q_points.getNumber();q++) {
          REAL jxw=jxw_values[q];
          for (int i=0;i<dofs_per_cell;i++) {
            Tensor<1,SPACEDIM> sgi=sg(i,q);
            sgi*=jxw;
            for (int j=0;j<dofs_per_cell;j++) {
              cell_matrix(i,j)+=sgi*sg(j,q);
            }
          }
        }
//      delta function rhs:
        if (element->contains(pole,vertices)) {
          int nv=element->numberVertices();
          Point<SPACEDIM> cp=element->canonicalCenter();
          mapping.canonicalPoint(element,*tessellations[cycle],fe_values.mappingData(nv),
            pole,cp);
          Quadrature<SPACEDIM> q(cp);
          UPDATE_FLAGS flags=UPDATE_VALUES;
          FiniteElementDataBase *fedb=fe.getFEData(nv,flags,&q);
          FiniteElementDataBase *fedb2=
            fe.fixShapeFunctionsFor(element,fedb);
          for (int s=0;s<dofs_per_cell;s++) {
            cell_rhs[s]+=fedb2->shapeValue(s,0);
          }
          if (fedb2!=fedb) delete fedb2; fedb2=0;
          delete fedb; fedb=0;
        }

//      neumann boundary conditions
#if (SPACEDIM==1)
        for (int f=0;f<element->numberCodimension1Shapes();f++) {
          if (present_level->neighborLevel(f)==0) {
            const BoundaryCondition *bc=
              present_level->Level0::getBoundaryCondition(f);
            CHECK_POINTER(bc);
            if (bc->isNatural()) {
              NumPtr<Point<1> > p(1); 
              p[0]=Interval::canonicalVertex(f);
              map<int,Quadrature<1> > fqmap_1d;
              fqmap_1d[element->numberVertices()]=Quadrature<1>(p);
              FEValues face_fe_values(mapping,fe,fqmap_1d,
                UPDATE_VALUES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES); 
              face_fe_values.mapShapes(*tessellations[cycle],element,
                present_level);
              const Array<2,REAL> &fsv=
                *face_fe_values.getShapeValues(element);
              const NumPtr<Point<1> > &face_q_points=
                face_fe_values.getQuadraturePoints(element);
              NumPtr<Point<1> > normal_vectors(1);
                normal_vectors[0][0]=(f==0 ? -1. : 1.);
              NumPtr<REAL> face_jxw_values(1); face_jxw_values[0]=1.;
              bc->faceNeumannValues(face_q_points,face_jxw_values,
                0,fsv,cell_rhs);
            }
          }
        }
#else
        int offset=0;
        for (int f=0;f<element->numberCodimension1Shapes();f++) {
          int face_index=element->codimension1ShapeIndex(f);
          const BoundaryCondition *bc=present_level->
            CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
          if (bc!=0) {
            if (bc->isNatural()) {
              const CODIMENSION1_SHAPE_CLASS *face_element=present_level->
                CODIMENSION1_LEVEL_CLASS::getShape(face_index);
              fe_face_values.mapShapes(*tessellations[cycle],element,
                present_level,f);
              const Array<2,REAL> &fsv=
                *fe_face_values.getShapeValues(element);
              NumPtr<Point<SPACEDIM> > face_q_points=
                fe_face_values.getQuadraturePoints(element,f);
              NumPtr<REAL> face_jxw_values=
                fe_face_values.getJxWValues(element,f);
              const NumPtr<Point<SPACEDIM> > &normal_vectors=
                fe_face_values.getNormalVectors(element,f);
              bc->faceNeumannValues(face_q_points,face_jxw_values,
                offset,fsv,cell_rhs);
            }
          }
          offset+=
            fe_face_values.getFaceQuadrature(element,f).numberPoints();
        }
#endif
        NumPtr<int> local_dof_indices;
        dof_handlers[cycle]->getDofIndices(element,present_level,
          local_dof_indices);
        for (int i=0;i<dofs_per_cell;i++) {
          int ldfi=local_dof_indices[i];
          for (int j=0;j<dofs_per_cell;j++) {
            system_matrix.add(ldfi,local_dof_indices[j],cell_matrix(i,j));
          }
          system_rhs[ldfi]+=cell_rhs[i];
        }
      }
    }

//  compute essential boundary values
    map<int,REAL> boundary_values;
    dof_handlers[cycle]->interpolateBoundaryValues(mapping,boundary_values);

//  modify linear system and rhs for essential boundary conditions
    solutions[cycle].reinit(n_dofs[cycle]);
    bool preserve_symmetry=true;
    dof_handlers[cycle]->applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handlers[cycle]->applyBoundaryValues(boundary_values,
      solutions[cycle]);

//  conjugate gradient solver
    SolverControl cg_solver_control(2*solutions[cycle].size(),
      DBL_EPSILON*1.e2,false,false);
    CGSolver cg(cg_solver_control);
    PreconditionIdentity pi;
    cg.solve(system_matrix,solutions[cycle],system_rhs,pi);

    ExactFunctionSystem efs;
//  compute various error measures
    int n_active_cells=tessellations[cycle]->numberActiveShapes();
    FunctionSystem *weight=0;
    L2_error[cycle]=dof_handlers[cycle]->integrateDifference(mapping,
      solutions[cycle],efs,qmap,L2_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tL2_error = " << L2_error[cycle] << endl;
#endif
    
    H1_error[cycle]=dof_handlers[cycle]->integrateDifference(mapping,
      solutions[cycle],efs,qmap,H1_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tH1_error = " << H1_error[cycle] << endl;
#endif
 
    Linfty_error[cycle]=dof_handlers[cycle]->integrateDifference(mapping,
      solutions[cycle],efs,qmap,LINFTY_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tLinfinity_error = " << Linfty_error[cycle] << endl;
#endif

#if (SPACEDIM==1)
//  plot solution
    plot_gt->setfgColor("black");
    plot_gt->newPage();
    plot_gt->rescale(0.,1.,-Linfty_error[cycle],Linfty_error[cycle]);
//  plot_gt->rescale(0.,1.,0.,pole[0]*(1.-pole[0]));
    plot_gt->drawAxes();
    int nelements=tessellations[cycle]->numberActiveShapes();
    int n_pts=640/nelements+1;
    NumPtr<Point<1> > plot_points(n_pts);
    REAL dx=1./static_cast<REAL>(n_pts-1);
    for (int i=0;i<n_pts;i++) {
      plot_points[i][0]=static_cast<REAL>(i)*dx;
    }
    Quadrature<1> plot_quadrature(plot_points);
    map<int,Quadrature<SPACEDIM> > plot_qmap;
    plot_qmap[Interval::verticesPerShape()]=plot_quadrature;
    FEValues plot_fe_values(mapping,fe,plot_qmap,
      UPDATE_VALUES | UPDATE_Q_POINTS);
    plot_gt->setfgColor("blue");
    for (bool doit=tessellations[cycle]->firstActiveShape();doit;
    doit=tessellations[cycle]->nextActiveShape()) {
      LEVEL_CLASS *present_level=tessellations[cycle]->presentLevel();
      const SHAPE_CLASS *element=present_level->getShape();
      plot_fe_values.mapShapes(*tessellations[cycle],element,present_level);
      const NumPtr<Point<1> > &q_points=
        plot_fe_values.getQuadraturePoints(element);

      NumPtr<REAL> solution_values;
      dof_handlers[cycle]->getFunctionValues(plot_fe_values,element,
        present_level,solutions[cycle],solution_values);
      const Point<1> &x=q_points[0];
//    plot_gt->movePen(x,exactSolutionValue(x));
      plot_gt->movePen(x[0],solution_values[0]-exactSolutionValue(x));
      for (int i=1;i<q_points.getNumber();i++) {
        const Point<1> &x=q_points[i];
//      plot_gt->drawLine(x,exactSolutionValue(x));
        plot_gt->drawLine(x[0],solution_values[i]-exactSolutionValue(x));
      }
    }
    plot_gt->flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
#elif (SPACEDIM==2)
    DataOut data_out(*dof_handlers[cycle],mapping);
    data_out.addDataVector(solutions[cycle],"solution");
    int n_subdivisions=128/(1<<cycle);
    data_out.buildPatches(n_subdivisions);
    bool quit_button=D_CHOOSE(false,false,true);
    data_out.plot("solution",plot_gt,quit_button);
#if (SPACEDIM==1)
    if (!quit_button) {
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    if (!quit_button) {
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif
#endif

    if (cycle<N_CYCLES-1) tess.refineGlobal(1);
    delete sparsity_pattern;
  }
#if (SPACEDIM<=2)
  delete plot_gt; plot_gt=0;
#endif

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }

  for (int i=0;i<tessellations.getNumber();i++) {
    delete tessellations[i]; tessellations[i]=0;
  }
  for (int i=0;i<dof_handlers.getNumber();i++) {
    delete dof_handlers[i]; dof_handlers[i]=0;
  }
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
#include "NumPtr.C"
INSTANTIATE_NUMPTR(Tessellation*);
