// "$Header: /home/faculty/johnt/cvs/deal_new/examples/LinearElements1D.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM!=1)
barf // hard-wired for 1D
#endif

// c(x) = 1, x<a; c(x) = 2, x>a
//-[ c u']'=0, u(0)=0, u(1)=1 ==> 
//  u(x)= 2x/(1+a), x<a
//  u(x) = (x+a)/(1+a), x>a
int degree=1;
int nelements=10;
REAL jump=M_PI/6.; // irrational (near 1/2) ==> not a mesh point
int N_CYCLES=7;

//description of pde right-hand side and exact solution (to compute errors)
REAL pdeCoefficient(const Point<1> &p) { return (p[0]<jump ? 1. : 2.); }
REAL rightHandSideValue(const Point<1> &p) { return 0.; }
REAL exactSolutionValue(const Point<1> &p) { 
  return (p[0]<jump ? 2.*p[0]/(1.+jump) : (p[0]+jump)/(1.+jump));
}
Tensor<1,1> exactSolutionGradient(const Point<1> &p) {
  Tensor<1,1> return_val;
  return_val[0]=(p[0]<jump ? 2./(1.+jump) : 1./(1.+jump));
  return return_val;
}

//class to describe essential boundary condition
class DirichletFunctionSystem : public FunctionSystem {
  protected:
    explicit DirichletFunctionSystem(const FunctionSystem &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    DirichletFunctionSystem() {;}
    virtual ~DirichletFunctionSystem() {;}
    virtual REAL value(const Point<1> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<1> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,1> gradient(const Point<1> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,1> >&,
    const Point<1>&) const {
      OBSOLETE("not programmed");
    }
};
class SimpleMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<1> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL coef_value=pdeCoefficient(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,1> sgi=sg(i,q);
          sgi*=coef_value*jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
//          element-wise contribution to stiffness matrix is
//            integral grad v_i . pdeCoefficient * grad v_j
//          ie, pde is -div ( pdeCoefficient grad u ) = rhs
            cell_matrix(i,j)+=sgi*sg(j,q);
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

//rectangular domain consisting of nelements intervals
  Point<1> low(0.);
  Point<1> edge[1]; edge[0]=Point<1>(1.);
  int number_elements[1]={nelements};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);

  SimpleMaterialModel smm;
  tess.setMaterialModel(&smm);

  DirichletFunctionSystem dfs;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    boundary_condition[i]->setFunctionSystem(&dfs);
    boundary_condition[i]->makeEssential(); // Dirichlet boundary
  }

//linear coordinate mapping
//  mapping shape functions are computed on canonical elements now
  MappingL mapping;

//Lagrange polynomials to represent solution of problem
//  finite element shape functions are computed on canonical elements now
  FE_L fe(degree);

//Gaussian quadrature
  map<int,Quadrature<1> > qmap;
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree-1));

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool plot_gt("solution error","x","error",0.,1.,0.,1.,&cmap,0,0.5);
  plot_gt.setbgColor("white");
  plot_gt.setfgColor("black");
  for (int cycle=0;cycle<N_CYCLES;cycle++) {

//  data structure to keep track of degrees of freedom in solver
    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();
    n_dofs[cycle]=dof_handler.numberDofs();
    cout << "\n\tNumber of degrees of freedom = " << n_dofs[cycle] << endl;

//data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

    map<int,Quadrature<0> > fqmap;
    dof_handler.assembleLinearSystem(mapping,qmap,fqmap,
      system_matrix,system_rhs);

//  compute essential boundary values
    map<int,REAL> boundary_values;
    dof_handler.interpolateBoundaryValues(mapping,boundary_values);

//  modify linear system and rhs for essential boundary conditions
    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);

//  conjugate gradient solver
    SolverControl 
      cg_solver_control(2*solution.size(),DBL_EPSILON*1.e2,false,false);
    CGSolver cg(cg_solver_control);
    PreconditionIdentity pi;
    cg.solve(system_matrix,solution,system_rhs,pi);

//  compute various error measures
    int n_active_cells=tess.numberActiveShapes();
    FunctionSystem *weight=0;
    L2_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,L2_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tL2_error = " << L2_error[cycle] << endl;
#endif
    
    H1_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,H1_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tH1_error = " << H1_error[cycle] << endl;
#endif
 
    Linfty_error[cycle]=dof_handler.integrateDifference(mapping,solution,
      dfs,qmap,LINFTY_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tLinfinity_error = " << Linfty_error[cycle] << endl;
#endif

//  plot solution
    plot_gt.newPage();
    plot_gt.rescale(0.,1.,-Linfty_error[cycle],Linfty_error[cycle]);
    plot_gt.drawAxes();
    int nelements=tess.numberActiveShapes();
    int n_pts=640/nelements+1;
    NumPtr<Point<1> > plot_points(n_pts);
    REAL dx=1./static_cast<REAL>(n_pts-1);
    for (int i=0;i<n_pts;i++) {
      plot_points[i][0]=static_cast<REAL>(i)*dx;
    }
    Quadrature<1> plot_quadrature(plot_points);
    map<int,Quadrature<1> > plot_qmap;
    plot_qmap[Interval::verticesPerShape()]=plot_quadrature;
    FEValues plot_fe_values(mapping,fe,plot_qmap,
      UPDATE_VALUES | UPDATE_Q_POINTS);
    for (bool doit=tess.firstActiveShape();doit;doit=tess.nextActiveShape())
    {
      LEVEL_CLASS *present_level=tess.presentLevel();
      const SHAPE_CLASS *element=present_level->getShape();
      plot_fe_values.mapShapes(tess,element,present_level);
      const NumPtr<Point<1> > &q_points=
        plot_fe_values.getQuadraturePoints(element);

      NumPtr<REAL> solution_values;
      dof_handler.getFunctionValues(plot_fe_values,element,present_level,
        solution,solution_values);
      const Point<1> &x=q_points[0];
      plot_gt.movePen(x[0],solution_values[0]-exactSolutionValue(x));
      for (int i=1;i<q_points.getNumber();i++) {
        const Point<1> &x=q_points[i];
        plot_gt.drawLine(x[0],solution_values[i]-exactSolutionValue(x));
      }
    }
    plot_gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;

    if (cycle<N_CYCLES-1) tess.refineGlobal(1);
    delete sparsity_pattern;
  }

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    XYGraphTool nacgt("number active cells vs cycle","cycle",
      "number active cells",0,N_CYCLES-1,n_active_cells[0],
      n_active_cells[N_CYCLES-1],&cmap,0,0.5);
    nacgt.newPage();
    nacgt.setbgColor("white");
    nacgt.setfgColor("black");
    nacgt.drawAxes();
    nacgt.setfgColor("blue");
    nacgt.movePen(0,n_active_cells[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      nacgt.drawLine(cycle,n_active_cells[cycle]);
    }
    nacgt.flush();

    XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
      "number degress of freedom",0,N_CYCLES-1,n_dofs[0],
      n_dofs[N_CYCLES-1],&cmap,0,0.5);
    ndgt.newPage();
    ndgt.setbgColor("white");
    ndgt.setfgColor("black");
    ndgt.drawAxes();
    ndgt.setfgColor("blue");
    ndgt.movePen(0,n_dofs[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ndgt.drawLine(cycle,n_dofs[cycle]);
    }
    ndgt.flush();

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
#ifdef DEBUG
      cout << "\tlog_n_dofs[" << cycle << "] = " << log_n_dofs[cycle]
           << endl;
      cout << "\tL2_error[" << cycle << "] = " << L2_error[cycle]
           << endl;
#endif
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
