// "$Header: /home/faculty/johnt/cvs/deal_new/examples/EllipticQuadratic.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"

/* $Id: step-6.cc,v 1.33 2003/01/09 20:32:45 wolf Exp $ */
/* Author: Wolfgang Bangerth, University of Heidelberg, 2000 */

/*    $Id: step-6.cc,v 1.33 2003/01/09 20:32:45 wolf Exp $       */
/*    Version: $Name: Version-4-0-0 $                                               */
/*                                                                */
/*    Copyright (C) 2000, 2001, 2002, 2003 by the deal.II authors */
/*                                                                */
/*    This file is subject to QPL and may not be  distributed     */
/*    without copyright and license information. Please refer     */
/*    to the file deal.II/doc/license.html for the  text  and     */
/*    further information on this license.                        */
//modified from deal.II/examples/step-6/step-6.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>
#include <map>

#include "Boundary.H"
#include <BoundaryCondition.H>
#include <ConstraintMatrix.H>
#include <DataOut.H>
#include <FiniteElement.H>
#include <GenerateTess.H>
#include <KellyErrorEstimator.H>
#include <MemoryDebugger.H>
#include <Multigrid.H>
#include <Precondition.H>
#include "SetTraps.H"
#include <Tessellation.H>
#include <Tracer.H>
#include <VolGraphTool.H>

//problem is
//  - div [ k(x) grad u ] = 1 in domain
//  u = 0                     on boundary

class VariableCoefficientMaterialModel : public MaterialModel {
  private:
    inline REAL coefficientValue(const Point<SPACEDIM> &p) const {
      return (p.square()<0.25 ? 20. : 1.);
    }
  public:
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      int n_q_points=q_points.getNumber();
      for (int q=0;q<n_q_points;q++) {
        REAL coefficient_value=coefficientValue(q_points[q]);
        REAL jxw=jxw_values[q];
        REAL rhsv_jxw=jxw; // i.e., rhs = 1.
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=coefficient_value*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
            cell_matrix(i,j)+=sgi*sg(j,q);
          }
          cell_rhs[i]+=sv(i,q)*rhsv_jxw;
        }
      }
    }
    REAL processBoundaryGradient(const Point<SPACEDIM> &p,
    const Tensor<1,SPACEDIM> &gradient,const Point<SPACEDIM> &normal,int)
    const {
      return coefficientValue(p)*(gradient.getValues()*normal);
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif

  Tessellation tess;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  bool set_face_ids=true;
#if (SPACEDIM==1)
  Point<1> low(-1.);
  Point<1> edge[1]; edge[0]=Point<1>(2.);
  int number_elements[1]={1};
  SubCellData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);
#else
  Point<SPACEDIM> center(D_DECL(0.,0.,0.));
  REAL radius=1.;
  BoundaryData *boundary_data=generateBall(tess,center,radius,set_face_ids,
    boundary,boundary_condition);
#endif

  VariableCoefficientMaterialModel vcmm;
  tess.setMaterialModel(&vcmm);

  ConstantFunctionSystem cfs(0.);
  for (int i=0;i<boundary_condition.getNumber();i++) {
    boundary_condition[i]->setFunctionSystem(&cfs);
    boundary_condition[i]->makeEssential();
//  boundary_condition[i]->setMaterialModel(&vcmm);
  }

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;

//quadratic or tensor product quadratic polynomials to represent solution
  int degree=2;
  FE_L fe(degree);

//quadrature rule for stiffness matrix and rhs
  GaussianQuadrature<SPACEDIM> quadrature_formula(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(
    D_CHOOSE(2*degree,2*degree+1,2*degree+1)));;
//NewtonCotesQuadrature<SPACEDIM> quadrature_formula(3);
  map<int,Quadrature<SPACEDIM> > qmap;
  int nv=D_CHOOSE(Interval::verticesPerShape(),
                  Quadrilateral::verticesPerShape(),
                  Hexahedron::verticesPerShape());
  qmap[nv]=quadrature_formula;

//graphics setup
  Palette pal;
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool plot_gt("solution plot","x","y",0.,1.,0.,1.,&cmap,0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
//AreaGraphTool contour_gt("solution contours","x0","x1",rlo,rhi,&cmap,
//  0,0.5);
  AreaGraphTool plot_gt("solution plot","x0","y1",rlo,rhi,&cmap,0,0.5);
  BOOLEAN use_lighting=FALSE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
    0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
    surface_gt.drawBox(vlow,vhigh);
  }
#else
  BOOLEAN use_lighting=FALSE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool plot_gt("solution plot",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&plot_gt);
    plot_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
    0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
    surface_gt.drawBox(vlow,vhigh);
  }
#endif

//In the KellyErrorEstimator the square of the jump in the solution 
//  gradient is integrated over each face.
//The square of the jump has order 2*degree, so we want the quadrature rule
//  to have order 2*degree+1.
  map<int,Quadrature<SPACEDIM-1> > error_estimator_quadrature_map;
#if (SPACEDIM==2)
  error_estimator_quadrature_map[Interval::verticesPerShape()]=
    GaussianQuadrature<1>(GaussianQuadrature<1>::argGivenOrder(2*degree+1));
#endif
#if (SPACEDIM==3)
  error_estimator_quadrature_map[Quadrilateral::verticesPerShape()]=
    GaussianQuadrature<2>(GaussianQuadrature<2>::argGivenOrder(2*degree+1));
#endif

//loop over local mesh refinements
  for (int cycle=1;cycle<8;cycle++) {
    cout << "Cycle " << cycle << ':' << endl;
    cout << "   Number of active cells:       "
         << tess.numberActiveShapes() << endl;

//  data structure to keep track of degrees of freedom in solver
    MGDofHandler mg_dof_handler(tess,fe);
    mg_dof_handler.distributeDofs();

//  data structure to keep track of conforming grid constraints for
//    hanging nodes
    ConstraintMatrix hanging_node_constraints;
    hanging_node_constraints.makeHangingNodeConstraints(mg_dof_handler);
    hanging_node_constraints.close();
#ifdef DEBUG
//  hanging_node_constraints.printOn(cout);
#endif

//  data structure to keep track of linear system storage
    int n_dofs=mg_dof_handler.numberDofs();
    cout << "   Number of degrees of freedom: " << n_dofs << endl;
    SparsityPattern *sparsity_pattern=OPERATOR_NEW SparsityPattern(n_dofs,
      n_dofs,mg_dof_handler.maxCouplingsBetweenDofs());
    mg_dof_handler.DofHandler::makeSparsityPattern(*sparsity_pattern);

//  insert hanging node constraint equations into sparsity pattern
    hanging_node_constraints.condense(*sparsity_pattern);
    sparsity_pattern->compress();

//  storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs,0.);
    Vector solution(n_dofs,0.);

//  for each shape
//    compute mapping shape functions at canonical quadrature points
//    compute finite element shape fcns at canonical quadrature points
//    allocate space for mapped shape functions
//  location of shape function canonical support points is determined in
//    FE_Q::initializeCanonicalSupportPoints; in first element have
//          6 7 8
//        3   4   5
//      0     1     2
//  location of quadrature points determined in
//    NewtonCotesQuadrature<SPACEDIM>; in first element have
//          2 5 8
//        1   4   7
//      0     3     6
//  ordering of degrees of freedom in each element determined in
//    DofHandler::getDofValues; in first element have
//          2 7 3
//        4   8   5
//      0     6     1

//  quadrature rule for Neumann bc; 
//    quadrature rule not used, but number of quadrature points is used
    map<int,Quadrature<SPACEDIM-1> > nqmap;
#if (SPACEDIM==2)
    nqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
      GaussianQuadrature<1>::argGivenOrder(2*degree+1));
#elif (SPACEDIM==3)
    nqmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
      GaussianQuadrature<2>::argGivenOrder(2*degree+1));
#endif

//  assemble linear system
    mg_dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix,system_rhs);

//  replace stiffness equations with hanging node constraint equations
//  constrained nodes are still in the linear system:
//    for each hanging node, matrix row = row of identity matrix
//    rhs = 0
//  equations for non-hanging nodes are modified to eliminate hanging nodes
//  hanging node values are computed after system solve in distribute fcn
    hanging_node_constraints.condense(system_matrix);
    hanging_node_constraints.condense(system_rhs);

//  compute essential boundary values at mapped points on bdry
//  calls FunctionSystem in BoundaryCondition's to set boundary values
    map<int,REAL> boundary_values;
    mg_dof_handler.interpolateBoundaryValues(mapping,boundary_values);

//  modify linear system and rhs for essential boundary conditions
    bool preserve_symmetry=true;
    mg_dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    mg_dof_handler.applyBoundaryValues(boundary_values,solution);

//  solve linear system
    SolverControl solver_control(solution.size(),DBL_EPSILON*1.e4,
      false,false);
    CGSolver cg(solver_control);
//  PreconditionSSOR preconditioner(system_matrix,1.2);
    PreconditionIdentity preconditioner;
    cg.solve(system_matrix,solution,system_rhs,preconditioner);
    

//  evaluate hanging nodes
    hanging_node_constraints.distribute(solution);

//  compute solution values for graphical display
    DataOut data_out(mg_dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
    int n_subdivisions=2;
    data_out.buildPatches(n_subdivisions);
    bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>=2)
    data_out.drawGrid(&mesh_gt,quit_button);
#endif
#if (SPACEDIM==2)
//  data_out.contourPlot("solution",&contour_gt);
#endif
    data_out.plot("solution",&plot_gt,quit_button);
#if (SPACEDIM>=2)
    data_out.plotSurface("solution",&surface_gt);
#endif
#if (SPACEDIM==1)
    if (!quit_button) {
      XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif

    if (cycle<7) {
//    refine those 30 per cent of cells with the highest error values,
//    and coarsen the 3 per cent of cells with the lowest values.
      Vector estimated_error_per_cell(tess.numberActiveShapes());
      KellyErrorEstimator error_estimator(mapping,mg_dof_handler,
        error_estimator_quadrature_map,solution,estimated_error_per_cell);

      tess.refineAndCoarsenFixedNumber(estimated_error_per_cell,0.3,
        0.03);
      tess.executeCoarseningAndRefinement();
    }
    delete sparsity_pattern;
  }
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }

  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
