// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <iostream>

#include "Boundary.H"
#include "DataOut.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#ifdef USE_GTK
#include "GTKGUI.H"
#else
#include "GUI.H"
#endif
#include "GUIEnumInputParameter.H"
#include "GUIInputs.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Multigrid.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "TimedObject.H"
#include "Tracer.H"
#include "Vector.H"

GUI_INPUT_PARAMETER_LIST_TYPE *main_list=0;
BOOLEAN skip_gui=FALSE;
char *display_name=0;
double winsize=0.5;

#if (SPACEDIM==1)
enum DOMAIN_SHAPE { RECTANGULAR_DOMAIN };
const char *domain_shape_name[1]={ "rectangular" };
#else
enum DOMAIN_SHAPE { RECTANGULAR_DOMAIN, SPHERICAL_DOMAIN };
const char *domain_shape_name[2]={ "rectangular","spherical" };
#endif
DOMAIN_SHAPE domain_shape=RECTANGULAR_DOMAIN;
int int_domain_shape=domain_shape;
Point<SPACEDIM> center(D_DECL(0.,0.,0.));
REAL radius=1.;
Point<SPACEDIM> low;
Point<SPACEDIM> edge[SPACEDIM];

#if (SPACEDIM==1)
  int n_cycles=7;
#elif (SPACEDIM==2)
  int n_cycles=4;
  enum MESH_ELEMENTS { 
    QUADRILATERAL_MESH_ELEMENTS,
    TRIANGLE_MESH_ELEMENTS 
  };
  const char *mesh_elements_name[2]={"quadrilateral","triangle"};
  MESH_ELEMENTS mesh_elements=QUADRILATERAL_MESH_ELEMENTS;
  int int_mesh_elements=mesh_elements;
#else
  int n_cycles=3;
  enum MESH_ELEMENTS { 
    HEXAHEDRAL_MESH_ELEMENTS,
    PRISM_MESH_ELEMENTS,
    TETRAHEDRAL_MESH_ELEMENTS 
  };
  const char *mesh_elements_name[3]={"hexahedron","prism","tetrahedron"};
  MESH_ELEMENTS mesh_elements=HEXAHEDRAL_MESH_ELEMENTS;
  int int_mesh_elements=mesh_elements;
  int dir=0;
#endif
enum TEST_PROBLEM { 
  LINEAR_SOLUTION,    // -u''= 0, u (-1)=-1, u (1)=1 ==> u(x)=x
  QUADRATIC_SOLUTION, // -u''= 2, u (-1)= 1, u'(1)=2 ==> u(x)=x^2
  CUBIC_SOLUTION,     // -u''=6x, u'(-1)= 3, u (1)=1 ==> u(x)=x^3
  SINH_SOLUTION//-u''+u=0, u(-1)=-sinh(1), u(1)=sinh(1) ==> u(x)=sinh(x)
};
const char *test_problem_name[4]={
  "linear solution","quadratric solution","cubic solution","sinh solution"
};
//TEST_PROBLEM test_problem=LINEAR_SOLUTION;
TEST_PROBLEM test_problem=SINH_SOLUTION;
int int_test_problem=test_problem;

enum FINITE_ELEMENT_TYPE {
  LAGRANGE_FINITE_ELEMENT,
  HIERARCHICAL_FINITE_ELEMENT
};
const char *finite_element_type_name[2]={ 
  "lagrange finite element","hierarchical finite element"
};
FINITE_ELEMENT_TYPE finite_element_type=LAGRANGE_FINITE_ELEMENT;
int int_finite_element_type=finite_element_type;
int degree=1;

int mapping_degree=1;

enum QUADRATURE_TYPE { 
  GAUSSIAN_QUADRATURE,LOBATTO_QUADRATURE,NEWTON_COATES_QUADRATURE,
  FULL_NEWTON_COATES_QUADRATURE,PRODUCT_GAUSSIAN_QUADRATURE
};
const char *quadrature_type_name[5]={ 
  "gaussian quadrature","lobatto quadrature","newton-coates quadrature",
  "full newton-coates quadrature","product gaussian quadrature"
};
QUADRATURE_TYPE quadrature_type=GAUSSIAN_QUADRATURE;
int int_quadrature_type=quadrature_type;
int quadrature_order=1;

enum SOLVER_TYPE { CONJUGATE_GRADIENTS_SOLVER,MULTIGRID_SOLVER };
const char *solver_type_name[3]={ "conjugate gradients","multigrid" };
//SOLVER_TYPE solver_type=CONJUGATE_GRADIENTS_SOLVER;
SOLVER_TYPE solver_type=MULTIGRID_SOLVER;
int int_solver_type=solver_type;
REAL tolerance=DBL_EPSILON*1.e4;
int max_iterations=100;
bool log_history=false;
bool log_result=false;

enum PRECONDITIONER_TYPE { 
  MULTIGRID_PRECONDITIONER,IDENTITY_PRECONDITIONER,
  JACOBI_PRECONDITIONER,SSOR_PRECONDITIONER
};
const char *preconditioner_type_name[4]={ 
  "multigrid","identity","jacobi","ssor"
};
PRECONDITIONER_TYPE preconditioner_type=IDENTITY_PRECONDITIONER;
int int_preconditioner_type=preconditioner_type;
REAL relaxation=1.;

enum MULTIGRID_RELAXATION_TYPE { 
  SOR_RELAXATION,SSOR_RELAXATION
};
const char *relaxation_type_name[2]={ /* "jacobi", */ "sor","ssor" };
MULTIGRID_RELAXATION_TYPE relaxation_type=SOR_RELAXATION;
int int_relaxation_type=relaxation_type;

#if (SPACEDIM==1)
XYGraphTool *plot_gt=0;
#endif
#if (SPACEDIM==2)
enum PLOT_TYPE { COLOR_PLOT,CONTOUR_PLOT,SURFACE_PLOT };
const char *plot_type_name[3]={ "color","contour","surface" };
PLOT_TYPE plot_type=COLOR_PLOT;
int int_plot_type=plot_type;
bool plot_mesh=false;
AreaGraphTool *mesh_gt=0;
AreaGraphTool *plot_gt=0;
AreaGraphTool *contour_gt=0;
VolGraphTool *surface_gt=0;
#elif (SPACEDIM==3)
enum PLOT_TYPE { CLIP_PLANE_PLOT,SURFACE_PLOT };
const char *plot_type_name[2]={ "clip plane","surface" };
PLOT_TYPE plot_type=CLIP_PLANE_PLOT;
int int_plot_type=plot_type;
int number_surfaces=1;
bool plot_mesh=false;
VolGraphTool *mesh_gt=0;
VolGraphTool *plot_gt=0;
VolGraphTool *surface_gt=0;
#endif

REAL rightHandSideValue(const Point<SPACEDIM> &p) {
#if (SPACEDIM==1)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return -2.;
//    return -4.;
    case CUBIC_SOLUTION:
      return -6.*p[0];
    case SINH_SOLUTION:
    case LINEAR_SOLUTION:
    default:
      return 0.;
  }
#elif (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return -12.; // - 2 * trace of A=[[1,2],[2,5]]
//    return -4.;
    case CUBIC_SOLUTION:
      return -6.*(p[0]+p[1]);
    case SINH_SOLUTION: {
      switch (domain_shape) {
        case RECTANGULAR_DOMAIN:
          return 0.;
        case SPHERICAL_DOMAIN: {
          return -9.*sqrt(p[0]*p[0]+p[1]*p[1]); 
        }
      }
    }
    case LINEAR_SOLUTION:
    default:
      return 0.;
  }
#else
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return -40.; // - 2 * trace of A=[[1,2,3],[2,5,8],[3,8,14]]
//    return -4.;
    case CUBIC_SOLUTION:
      return -6.*(p[0]+p[1]+p[2]);
    case SINH_SOLUTION:
    case LINEAR_SOLUTION:
    default:
      return 0.;
  }
#endif
}

REAL exactSolutionValue(const Point<SPACEDIM> &p) {
#if (SPACEDIM==1)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return pow(p[0],2);
//    return (1.-2.*p[0])*(1.-p[0]);
    case CUBIC_SOLUTION:
      return pow(p[0],3);
    case SINH_SOLUTION:
      return sinh(p[0]);
    case LINEAR_SOLUTION:
    default:
      return 1.-p[0];
  }
#elif (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return p[0]*(   p[0]+2.*p[1])
            +p[1]*(2.*p[0]+5.*p[1]);
//    return (1.-2.*p[0])*(1.-p[0]);
    case CUBIC_SOLUTION:
      return pow(p[0],3)+pow(p[1],3);
    case SINH_SOLUTION: {
      switch (domain_shape) {
        case RECTANGULAR_DOMAIN:
          return sin(M_PI*p[0])*sinh(M_PI*p[1]);
        case SPHERICAL_DOMAIN: {
          REAL r=sqrt(p[0]*p[0]+p[1]*p[1]);
          return pow(r,3);
        }
      }
    }
    case LINEAR_SOLUTION:
    default:
      return p[0]+2.*p[1];
//    return 1.-p[0];
  }
#else
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return p[0]*(   p[0]+2.*p[1]+ 3.*p[2])
            +p[1]*(2.*p[0]+5.*p[1]+ 8.*p[2])
            +p[2]*(3.*p[0]+8.*p[1]+14.*p[2]);
//    return (1.-2.*p[0])*(1.-p[0]);
    case CUBIC_SOLUTION:
      return pow(p[0],3)+pow(p[1],3)+pow(p[2],3);
    case SINH_SOLUTION:
      switch (domain_shape) {
        case RECTANGULAR_DOMAIN:
          return sin(M_PI*p[0])*cos(M_PI*p[1])*sinh(M_SQRT2*M_PI*p[2]);
        case SPHERICAL_DOMAIN: {
          return sin(M_PI*p[0])*cos(M_PI*p[1])*sinh(M_SQRT2*M_PI*p[2]);
        }
      }
    case LINEAR_SOLUTION:
    default:
      return p[0]+2.*p[1]+3.*p[2];
//    return 1.-p[0];
  }
#endif
}

Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &p) {
#if (SPACEDIM==1)
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val(p);
      return_val*=2.;
//    Tensor<1,SPACEDIM> return_val;
//    return_val[0]=-3.+4.*p[0];
      return return_val;
    }
    case CUBIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val(p);
      return_val*=3.*p[0];
      return return_val;
    }
    case SINH_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=cosh(p[0]);
      return return_val;
    }
    case LINEAR_SOLUTION:
    default: {
      Tensor<1,SPACEDIM> return_val;
      return_val.initialize(-1.);
      return return_val;
    }
  }
#elif (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=2.*p[0]+ 4.*p[1];
      return_val[1]=4.*p[0]+10.*p[1]; // 2 * A * x
//    return_val[0]=-3.+4.*p[0];
//    return_val[1]=0.;
      return return_val;
    }
    case CUBIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=3.*p[0]*p[0];
      return_val[1]=3.*p[1]*p[1];
      return return_val;
    }
    case SINH_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      switch (domain_shape) {
        case RECTANGULAR_DOMAIN:
          return_val[0]=M_PI*cos(M_PI*p[0])*sinh(M_PI*p[1]);
          return_val[1]=M_PI*sin(M_PI*p[0])*cosh(M_PI*p[1]);
          break;
        case SPHERICAL_DOMAIN: {
          REAL r=sqrt(p[0]*p[0]+p[1]*p[1]);
          return_val[0]=3.*r*p[0];
          return_val[1]=3.*r*p[1];
          break;
        }
      }
      return return_val;
    }
    case LINEAR_SOLUTION:
    default: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=1.;
      return_val[1]=2.;
//    return_val[0]=-1.;
//    return_val[1]=0.;
      return return_val;
    }
  }
#else
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=2.*p[0]+ 4.*p[1]+ 6.*p[2];
      return_val[1]=4.*p[0]+10.*p[1]+16.*p[2];
      return_val[2]=6.*p[0]+16.*p[1]+28.*p[2]; // 2 * A * x
//    return_val[0]=-3.+4.*p[0];
//    return_val[1]=0.;
      return return_val;
    }
    case CUBIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=3.*p[0]*p[0];
      return_val[1]=3.*p[1]*p[1];
      return_val[2]=3.*p[2]*p[2];
      return return_val;
    }
    case SINH_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      switch (domain_shape) {
        case RECTANGULAR_DOMAIN: {
          return_val[0]=M_PI*cos(M_PI*p[0])*cos(M_PI*p[1])
                            *sinh(M_SQRT2*M_PI*p[2]);
          return_val[1]=-M_PI*sin(M_PI*p[0])*sin(M_PI*p[1])
                             *sinh(M_SQRT2*M_PI*p[2]);
          return_val[2]=M_SQRT2*M_PI*sin(M_PI*p[0])*cos(M_PI*p[1])
                               *cosh(M_SQRT2*M_PI*p[2]);
          break;
        }
        case SPHERICAL_DOMAIN: {
          return_val[0]=M_PI*cos(M_PI*p[0])*cos(M_PI*p[1])
                            *sinh(M_SQRT2*M_PI*p[2]);
          return_val[1]=-M_PI*sin(M_PI*p[0])*sin(M_PI*p[1])
                             *sinh(M_SQRT2*M_PI*p[2]);
          return_val[2]=M_SQRT2*M_PI*sin(M_PI*p[0])*cos(M_PI*p[1])
                               *cosh(M_SQRT2*M_PI*p[2]);
          break;
        }
      }
      return return_val;
    }
    case LINEAR_SOLUTION:
    default: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=1.;
      return_val[1]=2.;
      return_val[2]=3.;
//    return_val[0]=-1.;
//    return_val[1]=0.;
//    return_val[2]=0.;
      return return_val;
    }
  }
#endif
}

class ExactFunctionSystem : public FunctionSystem {
  protected:
    explicit ExactFunctionSystem(const BoundaryCondition &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ExactFunctionSystem() {;}
    virtual ~ExactFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
};
class RectangleNeumannFunctionSystem : public FunctionSystem {
  private:
    int dir;
    bool left_hand;
  protected:
    explicit RectangleNeumannFunctionSystem(
    const RectangleNeumannFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    RectangleNeumannFunctionSystem(int d,bool lh) : dir(d),left_hand(lh) {}
    virtual ~RectangleNeumannFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      return (left_hand ? -g[dir] : g[dir]);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      function_values[0]=(left_hand ? -g[dir] : g[dir]);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
};
class BallNeumannFunctionSystem : public FunctionSystem {
  protected:
  explicit BallNeumannFunctionSystem(const BallNeumannFunctionSystem&) {
    OBSOLETE("copy constructor disallowed");
  }
  public:
    BallNeumannFunctionSystem() {}
    virtual ~BallNeumannFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      return g.getValues()*p/sqrt(p.square());
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      function_values[0]=g.getValues()*p/sqrt(p.square());
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
};
#ifdef TIMING
//the most expensive routines are the linear system solve and
//  interpolateBoundaryValues
//the latter requires computation of the shape functions on an element
//  at iteratively computed canonical points corresponding to boundary:
  TimedObjectList tl;
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void processCommandLine(int argc,char *argv[],char *display_name,
BOOLEAN &skip_gui,ifstream &in_file) {
  if (argc<2) skip_gui=FALSE;
  else {
    in_file.open(argv[1],ios::in);
    CHECK_TEST(!in_file.fail());
    if (in_file) {
      INTEGER i=2; 
      while (i<argc) {
        if (strcmp(argv[i],"-d")==0) {
          display_name=OPERATOR_NEW_BRACKET(char,strlen(argv[++i])+1);
          strcpy(display_name,argv[i]);
        } else {
          cout << " >>>> error...invalid command line option:"
               << argv[i] << endl;
          abort();
        }    
        i++; 
      }    
    } else {
      cerr << "\ncannot open " << argv[1] << " for input" << endl;
      exit(1);
    }    
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void readMainInput(ifstream &in_file,InputParameterList *main_list,
BOOLEAN &skip_gui) {
  BOOLEAN found_main=FALSE;
  char name[LENGTH_NAME], comment[LENGTH_COMMENT];

  in_file.clear(ios::goodbit);
  in_file.seekg(0,ios::beg);
  while ( in_file >> setw(LENGTH_NAME) >> name ) {
    if ( strcmp(name,"Main") == 0 ) {
      in_file.getline( comment, LENGTH_COMMENT);
      found_main=TRUE;
      while ( in_file >> setw(LENGTH_NAME) >> name ) {
        if ( strcmp(name,"end") == 0 ) break;
        else if (strcmp(name,"skip_gui") == 0) { 
          int iskip_gui; // type bool not read correctly, so use int
          in_file >> iskip_gui;
          skip_gui= iskip_gui!=0;
        }    
        else main_list->formattedRead(in_file,name);
        in_file.getline( comment, LENGTH_COMMENT);
      }    
    } else in_file.getline(comment,LENGTH_COMMENT);
  }
  if ( !found_main ) skip_gui=FALSE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template class InputParameter<BOOLEAN>;
template class InputParameter<INTEGER>;
template class InputParameter<REAL>;

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//make list of input parameters to be read by GUI
void makeMainList(GUI_INPUT_PARAMETER_LIST_TYPE *&main_list) {
  main_list=OPERATOR_NEW GUI_INPUT_PARAMETER_LIST_TYPE("Main");
  { const char *group="Problem Parameters";
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_test_problem,"test problem",test_problem_name[0],
      test_problem_name[1],test_problem_name[2],test_problem_name[3]));
#if (SPACEDIM>1)
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_domain_shape,"domain shape",domain_shape_name[0],
      domain_shape_name[1]));
#endif
    main_list->append(OPERATOR_NEW GUIInputParameter<REAL>(radius,
      "radius",0.,DBL_MAX,group) );
  }
  { const char *group="Shape Parameters";
#if (SPACEDIM==2)
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_mesh_elements,"mesh elements",mesh_elements_name[0],
      mesh_elements_name[1]));
#elif (SPACEDIM==3)
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_mesh_elements,"mesh elements",mesh_elements_name[0],
      mesh_elements_name[1],mesh_elements_name[2]));
    main_list->append(OPERATOR_NEW GUIInputParameter<int>(
      dir,"Prism direction",0,3,group) );
#endif
  }
  { const char *group="Finite Element Parameters";
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_finite_element_type,"finite element",finite_element_type_name[0],
      finite_element_type_name[1]));
    main_list->append(OPERATOR_NEW GUIInputParameter<int>(
      degree,"shape function degree",0,INT_MAX,group) );
  }
  { const char *group="Mapping Parameters";
    main_list->append(OPERATOR_NEW GUIInputParameter<int>(
      mapping_degree,"mapping degree",1,INT_MAX,group) );
  }
  { const char *group="Quadrature Parameters";
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_quadrature_type,"quadrature type",quadrature_type_name[0],
      quadrature_type_name[1],quadrature_type_name[2],
      quadrature_type_name[3],quadrature_type_name[4]));
    main_list->append(OPERATOR_NEW GUIInputParameter<int>(
      quadrature_order,"quadrature_order",1,INT_MAX,group) );
  }
  { const char *group="Solver Parameters";
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_solver_type,"solver type",solver_type_name[0],
      solver_type_name[1]));
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_preconditioner_type,"preconditioner type",
      preconditioner_type_name[0],preconditioner_type_name[1],
      preconditioner_type_name[2],preconditioner_type_name[3]));
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_relaxation_type,"relaxation type",relaxation_type_name[0],
      relaxation_type_name[1] /* ,relaxation_type_name[2] */ ));
    main_list->append(OPERATOR_NEW GUIInputParameter<int>(
      max_iterations,"max iterations",0,INT_MAX,group) );
    main_list->append(OPERATOR_NEW GUIInputParameter<REAL>(
      tolerance,"tolerance",0.,DBL_MAX,group) );
    main_list->append(OPERATOR_NEW GUIInputParameter<REAL>(
      relaxation,"preconditioner relaxation",0.,DBL_MAX,group) );
    main_list->append(OPERATOR_NEW GUIInputParameter<bool>(
      log_history,"log history",false,true,group) );
    main_list->append(OPERATOR_NEW GUIInputParameter<bool>(
      log_result,"log result",false,true,group) );
  }
  { const char *group="Mesh Refinement Parameters";
    main_list->append(OPERATOR_NEW GUIInputParameter<int>(
      n_cycles,"number cycles",0,INT_MAX,group) );
  }
#if (SPACEDIM>1)
  { const char *group="Plot Parameters";
    main_list->append(OPERATOR_NEW GUIInputParameter<bool>(
      plot_mesh,"plot mesh",false,true,group) );
#if (SPACEDIM==2)
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_plot_type,"plot type",plot_type_name[0],plot_type_name[1],
      plot_type_name[2]));
#else
    main_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
      int_plot_type,"plot type",plot_type_name[0],plot_type_name[1]));
    main_list->append(OPERATOR_NEW GUIInputParameter<int>(
      number_surfaces,"number surfaces",0,16,group) );
#endif
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkMainInput() {
  domain_shape=static_cast<DOMAIN_SHAPE>(int_domain_shape);
  test_problem=static_cast<TEST_PROBLEM>(int_test_problem);
  quadrature_type=static_cast<QUADRATURE_TYPE>(int_quadrature_type);
  solver_type=static_cast<SOLVER_TYPE>(int_solver_type);
  preconditioner_type=
    static_cast<PRECONDITIONER_TYPE>(int_preconditioner_type);
  relaxation_type=
    static_cast<MULTIGRID_RELAXATION_TYPE>(int_relaxation_type);
  finite_element_type=
    static_cast<FINITE_ELEMENT_TYPE>(int_finite_element_type);
#if (SPACEDIM>1)
  mesh_elements=static_cast<MESH_ELEMENTS>(int_mesh_elements);
  plot_type=static_cast<PLOT_TYPE>(int_plot_type);
#endif

  switch (domain_shape) {
    case RECTANGULAR_DOMAIN:
      if (mapping_degree>1) {
        cerr << "it is unnecessary to use a coordinate mapping of degree "
             << mapping_degree << " on a rectangular domain"
             << "\na mapping of degree 1 in each coordinate direction "
             << " is sufficient" << endl;
      }
      break;
#if (SPACEDIM>1)
    case SPHERICAL_DOMAIN:
      if (mapping_degree==1) {
        cerr << "it is necessary to use a coordinate mapping of degree "
             << degree << " on a spherical domain to preserve the order"
             << " of the finite element approximation"
             << endl;
      }
      break;
#endif
  }
#if (SPACEDIM==1)
  int goal=Interval::staticNumberEquallySpacedLatticePoints(degree-1);
  switch (quadrature_type) {
    case GAUSSIAN_QUADRATURE:
    case PRODUCT_GAUSSIAN_QUADRATURE:
      if (GaussianQuadrature<1>::nPointsGivenOrder(quadrature_order)<goal)
      {
        cout << "\tnot enough quadrature points to produce"
             << " nonsingular stiffness matrix" << endl;
        while(GaussianQuadrature<1>::nPointsGivenOrder(quadrature_order)
        <goal) {
          quadrature_order++;
        }
        cout << "\tquadrature_order reset to " << quadrature_order 
             << endl;
      }
      break;
    case LOBATTO_QUADRATURE:
      if (LobattoQuadrature<1>::nPointsGivenOrder(quadrature_order)<goal)
      {
        cout << "\tnot enough quadrature points to produce"
             << " nonsingular stiffness matrix" << endl;
        while(LobattoQuadrature<1>::nPointsGivenOrder(quadrature_order)
        <goal) {
          quadrature_order++;
        }
        cout << "\tquadrature_order reset to " << quadrature_order 
             << endl;
      }
      break;
    case NEWTON_COATES_QUADRATURE:
    case FULL_NEWTON_COATES_QUADRATURE:
      if (NewtonCotesQuadrature<1>::nPointsGivenOrder(quadrature_order)<
      goal) {
        cout << "\tnot enough quadrature points to produce"
             << " nonsingular stiffness matrix" << endl;
        while(NewtonCotesQuadrature<1>::nPointsGivenOrder(quadrature_order)
        <goal) {
          quadrature_order++;
        }
        cout << "\tquadrature_order reset to " << quadrature_order 
             << endl;
      }
      break;
  }
  if (quadrature_order<2*degree-1) {
    cout << "\tquadrature_order = " << quadrature_order << " too small"
         << "\n\tshould be >= " << 2*degree-1 << " to preserve order "
         << "of finite element shape functions" << endl;
  }
#endif
#if (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS: {
      int goal=
        Quadrilateral::staticNumberEquallySpacedLatticePoints(degree);
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE:
        case PRODUCT_GAUSSIAN_QUADRATURE:
          if (GaussianQuadrature<2>::nPointsGivenOrder(quadrature_order)<
          goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(GaussianQuadrature<2>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        case LOBATTO_QUADRATURE:
          if (LobattoQuadrature<2>::nPointsGivenOrder(quadrature_order)<
          goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(LobattoQuadrature<2>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        case NEWTON_COATES_QUADRATURE:
        case FULL_NEWTON_COATES_QUADRATURE:
          if (NewtonCotesQuadrature<2>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(NewtonCotesQuadrature<2>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
      }
      if (quadrature_order<2*degree+1) {
        cout << "\tquadrature_order = " << quadrature_order 
             << " too small"
             << "\n\tshould be >= " << 2*degree+1 
             << " to preserve order of finite element shape functions" 
             << endl;
      }
      break;
    }
    case TRIANGLE_MESH_ELEMENTS: {
      int goal=
        Triangle::staticNumberEquallySpacedLatticePoints(degree-1);
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE: {
          if (TriangleGaussianQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleGaussianQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case LOBATTO_QUADRATURE: {
          if (TriangleLobattoQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleLobattoQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case FULL_NEWTON_COATES_QUADRATURE:
          if (TriangleFullNewtonCotesQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleFullNewtonCotesQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        case NEWTON_COATES_QUADRATURE: {
          if (TriangleNewtonCotesQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleNewtonCotesQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case PRODUCT_GAUSSIAN_QUADRATURE:
          if (TriangleProductGaussianQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleProductGaussianQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
      }
      if (quadrature_order<2*degree-1) {
        cout << "\tquadrature_order = " << quadrature_order 
             << " too small"
             << "\n\tshould be >= " << 2*degree-1 
             << " to preserve order of finite element shape functions" 
             << endl;
      }
      break;
    }
  }
#endif
#if (SPACEDIM==3)
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      int goal=
        Hexahedron::staticNumberEquallySpacedLatticePoints(degree);
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE:
        case PRODUCT_GAUSSIAN_QUADRATURE:
          if (GaussianQuadrature<3>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(GaussianQuadrature<3>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        case LOBATTO_QUADRATURE:
          if (LobattoQuadrature<3>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(LobattoQuadrature<3>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        case NEWTON_COATES_QUADRATURE:
        case FULL_NEWTON_COATES_QUADRATURE:
          if (NewtonCotesQuadrature<3>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(NewtonCotesQuadrature<3>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
      }
      if (quadrature_order<2*degree+1) {
        cout << "\tquadrature_order = " << quadrature_order 
             << " too small"
             << "\n\tshould be >= " << 2*degree+1 
             << " to preserve order of finite element shape functions" 
             << endl;
      }
      break;
    }
    case PRISM_MESH_ELEMENTS: {
      int goal=
        Prism::staticNumberEquallySpacedLatticePoints(degree);
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE: {
//        TriangleGaussianQuadrature x GaussianQuadrature
          if (TriangleGaussianQuadrature::nPointsGivenOrder(
          quadrature_order)*GaussianQuadrature<1>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleGaussianQuadrature::nPointsGivenOrder(
            quadrature_order)*GaussianQuadrature<1>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case LOBATTO_QUADRATURE: {
//        TriangleLobattoQuadrature x LobattoQuadrature
          if (TriangleLobattoQuadrature::nPointsGivenOrder(
          quadrature_order)*LobattoQuadrature<1>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleLobattoQuadrature::nPointsGivenOrder(
            quadrature_order)*LobattoQuadrature<1>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case FULL_NEWTON_COATES_QUADRATURE:
          if (TriangleFullNewtonCotesQuadrature::nPointsGivenOrder(
          quadrature_order)*NewtonCotesQuadrature<1>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleFullNewtonCotesQuadrature::nPointsGivenOrder(
            quadrature_order)*NewtonCotesQuadrature<1>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        case NEWTON_COATES_QUADRATURE: {
          if (TriangleNewtonCotesQuadrature::nPointsGivenOrder(
          quadrature_order)*NewtonCotesQuadrature<1>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleNewtonCotesQuadrature::nPointsGivenOrder(
            quadrature_order)*NewtonCotesQuadrature<1>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case PRODUCT_GAUSSIAN_QUADRATURE:
          if (TriangleProductGaussianQuadrature::nPointsGivenOrder(
          quadrature_order)*GaussianQuadrature<1>::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TriangleProductGaussianQuadrature::nPointsGivenOrder(
            quadrature_order)*GaussianQuadrature<1>::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
      }
      if (quadrature_order<2*degree+1) {
        cout << "\tquadrature_order = " << quadrature_order 
             << " too small"
             << "\n\tshould be >= " << 2*degree+1 
             << " to preserve order of finite element shape functions" 
             << endl;
      }
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
      int goal=
        Tetrahedron::staticNumberEquallySpacedLatticePoints(degree);
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE: {
          if (TetrahedronGaussianQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TetrahedronGaussianQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case LOBATTO_QUADRATURE: {
//        TriangleLobattoQuadrature x LobattoQuadrature
          if (TetrahedronLobattoQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TetrahedronLobattoQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case FULL_NEWTON_COATES_QUADRATURE:
          if (TetrahedronFullNewtonCotesQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TetrahedronFullNewtonCotesQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        case NEWTON_COATES_QUADRATURE: {
          if (TetrahedronNewtonCotesQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TetrahedronNewtonCotesQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
        }
        case PRODUCT_GAUSSIAN_QUADRATURE:
          if (TetrahedronProductGaussianQuadrature::nPointsGivenOrder(
          quadrature_order)<goal) {
            cout << "\tnot enough quadrature points to produce"
                 << " nonsingular stiffness matrix" << endl;
            while(TetrahedronProductGaussianQuadrature::nPointsGivenOrder(
            quadrature_order)<goal) {
              quadrature_order++;
            }
            cout << "\tquadrature_order reset to " << quadrature_order 
                 << endl;
          }
          break;
      }
      if (quadrature_order<2*degree-1) {
        cout << "\tquadrature_order = " << quadrature_order 
             << " too small"
             << "\n\tshould be >= " << 2*degree-1 
             << " to preserve order of finite element shape functions" 
             << endl;
      }
      break;
    }
  }
#endif
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      cout << "\tquadratic solution" << endl;
      break;
    case CUBIC_SOLUTION:
      cout << "\tcubic solution" << endl;
      break;
    case SINH_SOLUTION:
      cout << "\tsinh solution" << endl;
      break;
    case LINEAR_SOLUTION:
    default:
      cout << "\tlinear solution" << endl;
  }
  cout << "\tdegree = " << degree << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* initializeMesh(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  BoundaryData *boundary_data=0;
  for (int i=0;i<SPACEDIM;i++) {
    low[i]=-radius;
    Point<SPACEDIM> &ei=edge[i];
    for (int j=0;j<SPACEDIM;j++) {
      ei[j]=(i==j ? 2.*radius : 0.);
    }
  }
  switch (domain_shape) {
    case RECTANGULAR_DOMAIN: {
      bool set_face_ids=true;
      int number_elements[SPACEDIM]={D_DECL(2,2,2)};
#if (SPACEDIM==1)
      boundary_data=generateParallelpiped(tess,low,edge,
        number_elements,set_face_ids,boundary,boundary_condition);
#elif (SPACEDIM==2)
      if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
        boundary_data=generateParallelpiped(tess,low,edge,number_elements,
          set_face_ids,boundary,boundary_condition);
      } else {
        boundary_data=generateParallelpipedTriangles(tess,low,edge,
          number_elements,set_face_ids,boundary,boundary_condition);
      }
#else
      switch (mesh_elements) {
        case HEXAHEDRAL_MESH_ELEMENTS:
          boundary_data=generateParallelpiped(tess,low,edge,
            number_elements,set_face_ids,boundary,boundary_condition);
          break;
        case PRISM_MESH_ELEMENTS:
          boundary_data=generateParallelpipedPrisms(tess,low,edge,
            number_elements,dir,set_face_ids,boundary,boundary_condition);
          break;
        case TETRAHEDRAL_MESH_ELEMENTS:
          boundary_data=generateParallelpipedTetrahedra(tess,low,edge,
            number_elements,set_face_ids,boundary,boundary_condition);
          break;
        default:
          OBSOLETE("bad mesh_elements");
      }
#endif
      break;
    }
#if (SPACEDIM>1)
    case SPHERICAL_DOMAIN: {
      bool set_face_ids=false;
# if (SPACEDIM==2)
      if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
        boundary_data=generateBall(tess,center,radius,set_face_ids,
          boundary,boundary_condition);
      } else {
        boundary_data=generateBallTriangles(tess,center,radius,
          set_face_ids,boundary,boundary_condition);
      }
# elif (SPACEDIM==3)
      switch (mesh_elements) {
        case HEXAHEDRAL_MESH_ELEMENTS:
          boundary_data=generateBall(tess,center,radius,set_face_ids,
            boundary,boundary_condition);
          break;
        case PRISM_MESH_ELEMENTS:
          boundary_data=generateBallPrismsWithTetrahedra(tess,center,
            radius,set_face_ids,boundary,boundary_condition);
//        boundary_data=generateBallPrismsWithHexahedra(tess,center,
//          radius);
          break;
        case TETRAHEDRAL_MESH_ELEMENTS:
          boundary_data=generateBallTetrahedra(tess,center,radius,
            set_face_ids,boundary,boundary_condition);
          break;
        default:
          OBSOLETE("bad mesh_elements");
      }
# endif
      break;
    }
#endif
    default:
      OBSOLETE("unknown domain shape");
  }
  return boundary_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void setBoundaryConditionsInRectangle(Tessellation &tess,
FunctionSystem &fs,NumPtr<BoundaryCondition*> &boundary_condition) {
  REAL fudge=sqrt(DBL_EPSILON);
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
//    DIRICHLET on left, back and bottom, NEUMANN on right, front and top
      for (int d=0;d<SPACEDIM;d++) {
        boundary_condition[2*d]->setFunctionSystem(&fs);
        boundary_condition[2*d]->makeEssential();
        boundary_condition[2*d+1]->setFunctionSystem(
          OPERATOR_NEW RectangleNeumannFunctionSystem(d,false));
        boundary_condition[2*d+1]->makeNatural();
      }
      break;
    }
    case CUBIC_SOLUTION: {
//    NEUMANN on left, back and bottom, DIRICHLET on right, front and top
      for (int d=0;d<SPACEDIM;d++) {
        boundary_condition[2*d]->setFunctionSystem(
          OPERATOR_NEW RectangleNeumannFunctionSystem(d,true));
        boundary_condition[2*d]->makeNatural();
        boundary_condition[2*d+1]->setFunctionSystem(&fs);
        boundary_condition[2*d+1]->makeEssential();
      }
      break;
    }
    case LINEAR_SOLUTION:
    case SINH_SOLUTION:
    default: {
//    DIRICHLET on all
      for (int d=0;d<boundary_condition.getNumber();d++) {
        boundary_condition[d]->setFunctionSystem(&fs);
        boundary_condition[d]->makeEssential();
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void setBoundaryConditionsInSphere(Tessellation &tess,
FunctionSystem &fs,NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  boundary.allocate(1);
  boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
  boundary_condition.allocate(2);
  for (int i=0;i<2;i++) {
    boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[0]);
  }
  boundary_condition[0]->setFunctionSystem(&fs);
  boundary_condition[0]->makeEssential(); // Dirichlet
  boundary_condition[1]->
    setFunctionSystem(OPERATOR_NEW BallNeumannFunctionSystem());
  boundary_condition[1]->makeNatural();   // Neumann

  REAL fudge=sqrt(DBL_EPSILON);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<SPACEDIM> fc=face_element->center(tess.getVertices());
      switch (test_problem) {
        case QUADRATIC_SOLUTION: {
#if (SPACEDIM==2)
//        DIRICHLET on SW, NEUMANN on NE
          codimension1_level->setBoundaryCondition(fc[0]+fc[1]<fudge
            ? boundary_condition[0] : boundary_condition[1]);
#else
          codimension1_level->setBoundaryCondition(
            fc[1]<radius/sqrt(3.)+fudge
            ? boundary_condition[0] : boundary_condition[1]);
#endif
          break;
        }
        case CUBIC_SOLUTION: {
#if (SPACEDIM==2)
//        NEUMANN on SW, DIRICHLET on NE
          codimension1_level->setBoundaryCondition(fc[0]+fc[1]<fudge
            ? boundary_condition[1] : boundary_condition[0]);
#else
          codimension1_level->setBoundaryCondition(
            fc[1]<radius/sqrt(3.)+fudge
            ? boundary_condition[1] : boundary_condition[0]);
#endif
          break;
        }
        case LINEAR_SOLUTION:
        case SINH_SOLUTION: // actually radially symmetric solution
        default:
//        DIRICHLET all around
          codimension1_level->setBoundaryCondition(boundary_condition[0]);
          break;
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void initializeQuadratures(
map<int,Quadrature<SPACEDIM> > &qmap) {
#if (SPACEDIM==1)
  switch (quadrature_type) {
    case GAUSSIAN_QUADRATURE:
    case PRODUCT_GAUSSIAN_QUADRATURE: {
      int n=GaussianQuadrature<1>::argGivenOrder(quadrature_order);
      qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(n);
      break;
    }
    case LOBATTO_QUADRATURE: {
      int n=LobattoQuadrature<1>::argGivenOrder(quadrature_order);
      qmap[Interval::verticesPerShape()]=LobattoQuadrature<1>(n);
      break;
    }
    case NEWTON_COATES_QUADRATURE:
    case FULL_NEWTON_COATES_QUADRATURE: {
      int n=NewtonCotesQuadrature<1>::argGivenOrder(quadrature_order);
      qmap[Interval::verticesPerShape()]=NewtonCotesQuadrature<1>(n);
      break;
    }
  }
#elif (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS: {
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE:
        case PRODUCT_GAUSSIAN_QUADRATURE: {
          int n=GaussianQuadrature<2>::argGivenOrder(quadrature_order);
          qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(n);
          break;
        }
        case LOBATTO_QUADRATURE: {
          int n=LobattoQuadrature<2>::argGivenOrder(quadrature_order);
          qmap[Quadrilateral::verticesPerShape()]=LobattoQuadrature<2>(n);
          break;
        }
        case NEWTON_COATES_QUADRATURE:
        case FULL_NEWTON_COATES_QUADRATURE: {
          int n=NewtonCotesQuadrature<2>::argGivenOrder(quadrature_order);
          qmap[Quadrilateral::verticesPerShape()]=
            NewtonCotesQuadrature<2>(n);
          break;
        }
      }
      break;
    }
    case TRIANGLE_MESH_ELEMENTS: { 
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE: {
          int n=
            TriangleGaussianQuadrature::argGivenOrder(quadrature_order);
          qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(n);
          break;
        }
        case LOBATTO_QUADRATURE: {
          int n=
            TriangleLobattoQuadrature::argGivenOrder(quadrature_order);
          qmap[Triangle::verticesPerShape()]=TriangleLobattoQuadrature(n);
          break;
        }
        case NEWTON_COATES_QUADRATURE: {
          int n=
            TriangleNewtonCotesQuadrature::argGivenOrder(quadrature_order);
          qmap[Triangle::verticesPerShape()]=
            TriangleNewtonCotesQuadrature(n);
          break;
        }
        case FULL_NEWTON_COATES_QUADRATURE: {
          int n=TriangleFullNewtonCotesQuadrature::argGivenOrder(
            quadrature_order);
          qmap[Triangle::verticesPerShape()]=
            TriangleFullNewtonCotesQuadrature(n);
          break;
        }
        case PRODUCT_GAUSSIAN_QUADRATURE: {
          int n=TriangleProductGaussianQuadrature::argGivenOrder(
            quadrature_order);
          qmap[Quadrilateral::verticesPerShape()]=
            TriangleProductGaussianQuadrature(n);
          break;
        }
      }
      break;
    }
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE:
        case PRODUCT_GAUSSIAN_QUADRATURE: {
          int n=GaussianQuadrature<3>::argGivenOrder(quadrature_order);
          qmap[Hexahedron::verticesPerShape()]=
            GaussianQuadrature<3>(n);
          break;
        }
        case LOBATTO_QUADRATURE: {
          int n=LobattoQuadrature<3>::argGivenOrder(quadrature_order);
          qmap[Hexahedron::verticesPerShape()]=
            LobattoQuadrature<3>(n);
          break;
        }
        case NEWTON_COATES_QUADRATURE:
        case FULL_NEWTON_COATES_QUADRATURE: {
          int n=NewtonCotesQuadrature<3>::argGivenOrder(quadrature_order);
          qmap[Hexahedron::verticesPerShape()]=
            NewtonCotesQuadrature<3>(n);
          break;
        }
      }
      break;
    }
    case PRISM_MESH_ELEMENTS: {
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE: {
          int nt=TriangleGaussianQuadrature::argGivenOrder(
            quadrature_order);
          int ni=GaussianQuadrature<1>::argGivenOrder(quadrature_order);
          qmap[Prism::verticesPerShape()]=ProductQuadrature(
            TriangleGaussianQuadrature(nt),GaussianQuadrature<1>(ni));
          break;
        }
        case LOBATTO_QUADRATURE: {
          int nt=TriangleLobattoQuadrature::argGivenOrder(
            quadrature_order);
          int ni=LobattoQuadrature<1>::argGivenOrder(quadrature_order);
          qmap[Prism::verticesPerShape()]=ProductQuadrature(
            TriangleLobattoQuadrature(nt),LobattoQuadrature<1>(ni));
          break;
        }
        case NEWTON_COATES_QUADRATURE: {
          int nt=TriangleNewtonCotesQuadrature::argGivenOrder(
            quadrature_order);
          int ni=NewtonCotesQuadrature<1>::argGivenOrder(quadrature_order);
          qmap[Prism::verticesPerShape()]=ProductQuadrature(
            TriangleNewtonCotesQuadrature(nt),
            NewtonCotesQuadrature<1>(ni));
          break;
        }
        case FULL_NEWTON_COATES_QUADRATURE: {
          int nt=TriangleFullNewtonCotesQuadrature::argGivenOrder(
            quadrature_order);
          int ni=NewtonCotesQuadrature<1>::argGivenOrder(quadrature_order);
          qmap[Prism::verticesPerShape()]=ProductQuadrature(
            TriangleFullNewtonCotesQuadrature(nt),
            NewtonCotesQuadrature<1>(ni));
          break;
        }
        case PRODUCT_GAUSSIAN_QUADRATURE: {
          int nt=TriangleProductGaussianQuadrature::argGivenOrder(
            quadrature_order);
          int ni=GaussianQuadrature<1>::argGivenOrder(quadrature_order);
          qmap[Prism::verticesPerShape()]=ProductQuadrature(
            TriangleProductGaussianQuadrature(nt),
            GaussianQuadrature<1>(ni));
          break;
        }
      }
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
      switch (quadrature_type) {
        case GAUSSIAN_QUADRATURE: {
          int n=TetrahedronGaussianQuadrature::argGivenOrder(
            quadrature_order);
          qmap[Tetrahedron::verticesPerShape()]=
            TetrahedronGaussianQuadrature(n);
          break;
        }
        case LOBATTO_QUADRATURE: {
          int n=TetrahedronLobattoQuadrature::argGivenOrder(
            quadrature_order);
          qmap[Tetrahedron::verticesPerShape()]=
            TetrahedronLobattoQuadrature(n);
          break;
        }
        case NEWTON_COATES_QUADRATURE: {
        case FULL_NEWTON_COATES_QUADRATURE:
          int n=TetrahedronNewtonCotesQuadrature::argGivenOrder(
            quadrature_order);
          qmap[Tetrahedron::verticesPerShape()]=
            TetrahedronNewtonCotesQuadrature(n);
          break;
        }
        case PRODUCT_GAUSSIAN_QUADRATURE: {
          int n=TetrahedronProductGaussianQuadrature::argGivenOrder(
            quadrature_order);
          qmap[Tetrahedron::verticesPerShape()]=
            TetrahedronProductGaussianQuadrature(n);
          break;
        }
      }
      break;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void initializeFaceQuadratures(FiniteElement &fe,
map<int,Quadrature<SPACEDIM-1> > &nqmap) {
#if (SPACEDIM==2)
  {
    int nv=Interval::verticesPerShape();
    switch (quadrature_type) {
      case GAUSSIAN_QUADRATURE:
      case PRODUCT_GAUSSIAN_QUADRATURE: {
        int n=GaussianQuadrature<1>::argGivenOrder(quadrature_order);
        nqmap[nv]=GaussianQuadrature<1>(n);
        break;
      }
      case LOBATTO_QUADRATURE: {
        int n=LobattoQuadrature<1>::argGivenOrder(quadrature_order);
        nqmap[nv]=LobattoQuadrature<1>(n);
        break;
      }
      case NEWTON_COATES_QUADRATURE:
      case FULL_NEWTON_COATES_QUADRATURE: {
        int n=NewtonCotesQuadrature<1>::argGivenOrder(quadrature_order);
        nqmap[nv]=NewtonCotesQuadrature<1>(n);
        break;
      }
    }
  }
#elif (SPACEDIM==3)
  {
    int nvq=Quadrilateral::verticesPerShape();
    int nvt=Triangle::verticesPerShape();
    switch (quadrature_type) {
      case GAUSSIAN_QUADRATURE: {
        int nq=GaussianQuadrature<2>::argGivenOrder(quadrature_order);
        nqmap[nvq]=GaussianQuadrature<2>(nq);
        int nt=TriangleGaussianQuadrature::argGivenOrder(quadrature_order);
        nqmap[nvt]=TriangleGaussianQuadrature(nt);
        break;
      }
      case LOBATTO_QUADRATURE: {
        int nq=LobattoQuadrature<2>::argGivenOrder(quadrature_order);
        nqmap[nvq]=LobattoQuadrature<2>(nq);
        int nt=TriangleLobattoQuadrature::argGivenOrder(quadrature_order);
        nqmap[nvt]=TriangleLobattoQuadrature(nt);
        break;
      }
      case NEWTON_COATES_QUADRATURE: {
        int nq=NewtonCotesQuadrature<2>::argGivenOrder(quadrature_order);
        nqmap[nvq]=NewtonCotesQuadrature<2>(nq);
        int nt=
          TriangleNewtonCotesQuadrature::argGivenOrder(quadrature_order);
        nqmap[nvt]=TriangleNewtonCotesQuadrature(nt);
        break;
      }
      case FULL_NEWTON_COATES_QUADRATURE: {
        int nq=NewtonCotesQuadrature<2>::argGivenOrder(quadrature_order);
        nqmap[nvq]=NewtonCotesQuadrature<2>(nq);
        int nt=TriangleFullNewtonCotesQuadrature::argGivenOrder(
          quadrature_order);
        nqmap[nvt]=TriangleFullNewtonCotesQuadrature(nt);
        break;
      }
      case PRODUCT_GAUSSIAN_QUADRATURE: {
        int nq=GaussianQuadrature<2>::argGivenOrder(quadrature_order);
        nqmap[nvq]=GaussianQuadrature<2>(nq);
        int nt=TriangleProductGaussianQuadrature::argGivenOrder(
          quadrature_order);
        nqmap[nvt]=TriangleProductGaussianQuadrature(nt);
        break;
      }
    }
  }
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
class SimpleMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
            cell_matrix(i,j)+=sgi*sg(j,q);
#if (SPACEDIM==1)
            if (test_problem==SINH_SOLUTION) {
              cell_matrix(i,j)+=si*sv(j,q);
            }
#endif
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void solveLinearSystem(Tessellation &tess,FiniteElement &fe,
Mapping &mapping,const SolverMatrix &system_matrix,
const Vector &system_rhs,Vector &solution) {
  PreconditionIdentity pi;
  switch (solver_type) {
    case CONJUGATE_GRADIENTS_SOLVER: {
      switch (preconditioner_type) {
        case MULTIGRID_PRECONDITIONER: {
          SolverControl solver_control(max_iterations,tolerance,false,
            false);
          CGSolver cg(solver_control);
          MGDofHandler mgdh(tess,fe);
          mgdh.distributeDofs();
          NumPtr<set<int> > dirichlet_boundary_dofs;
          mgdh.findDirichletBoundaryDofs(mapping,dirichlet_boundary_dofs);
          MGTransfer transfer(mgdh);
          switch (relaxation_type) {
            case SOR_RELAXATION: {
              PreconditionSOR smoother(relaxation); 
                // relaxation=1 ==> Gauss-Seidel
              Multigrid mg(mgdh,system_matrix,solution,transfer,
                smoother,cg,pi,dirichlet_boundary_dofs,0);
              SolverControl mg_solver_control(max_iterations,tolerance,
                log_history,log_result);
              CGSolver mgcg(mg_solver_control);
              PreconditionMG pcmg(mg);
              mgcg.solve(system_matrix,solution,system_rhs,pcmg);
              break;
            }
            case SSOR_RELAXATION: {
              PreconditionSSOR smoother(relaxation); 
              Multigrid mg(mgdh,system_matrix,solution,transfer,
                smoother,cg,pi,dirichlet_boundary_dofs,0);
              SolverControl mg_solver_control(max_iterations,tolerance,
                log_history,log_result);
              CGSolver mgcg(mg_solver_control);
              PreconditionMG pcmg(mg);
              mgcg.solve(system_matrix,solution,system_rhs,pcmg);
              break;
            }
          }
          break;
        }
        case IDENTITY_PRECONDITIONER: {
          SolverControl solver_control(max_iterations,tolerance,
            log_history,log_result);
          CGSolver cg(solver_control);
          cg.solve(system_matrix,solution,system_rhs,pi);
          break;
        }
        case JACOBI_PRECONDITIONER: {
          SolverControl solver_control(max_iterations,tolerance,
            log_history,log_result);
          CGSolver cg(solver_control);
          PreconditionJacobi pj(system_matrix,relaxation);
          cg.solve(system_matrix,solution,system_rhs,pj);
          break;
        }
        case SSOR_PRECONDITIONER: {
          SolverControl solver_control(max_iterations,tolerance,
            log_history,log_result);
          CGSolver cg(solver_control);
          PreconditionSSOR ps(system_matrix,relaxation);
          cg.solve(system_matrix,solution,system_rhs,ps);
          break;
        }
      }
      break;
    }
    case MULTIGRID_SOLVER: {
      SolverControl solver_control(max_iterations,tolerance,false,
        false);
      CGSolver cg(solver_control);
      MGDofHandler mgdh(tess,fe);
      mgdh.distributeDofs();
      MGTransfer transfer(mgdh);
      NumPtr<set<int> > dirichlet_boundary_dofs;
      mgdh.findDirichletBoundaryDofs(mapping,dirichlet_boundary_dofs);
      switch (relaxation_type) {
        case SOR_RELAXATION: {
          PreconditionSOR smoother(relaxation); 
            // relaxation=1 ==> Gauss-Seidel
          Multigrid mg(mgdh,system_matrix,solution,transfer,smoother,
            cg,pi,dirichlet_boundary_dofs,0);
          SolverControl mg_solver_control(max_iterations,tolerance,
            log_history,log_result);
          mg.solve(mg_solver_control,system_matrix,system_rhs,solution);
          break;
        }
        case SSOR_RELAXATION: {
          PreconditionSSOR smoother(relaxation); 
          Multigrid mg(mgdh,system_matrix,solution,transfer,smoother,
            cg,pi,dirichlet_boundary_dofs,0);
          SolverControl mg_solver_control(max_iterations,tolerance,
            log_history,log_result);
          mg.solve(mg_solver_control,system_matrix,system_rhs,solution);
          break;
        }
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void plotSolution(DofHandler &dof_handler,Mapping &mapping,
int n_active_cells,Vector &solution) {
  DataOut data_out(dof_handler,mapping);
  data_out.addDataVector(solution,"solution");
#if (SPACEDIM==1)
  int n_subdivisions=640/n_active_cells;
#else
  int n_subdivisions=degree;
#endif
  data_out.buildPatches(n_subdivisions);

  bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM==1)
  if (plot_gt!=0) data_out.plot("solution",plot_gt,quit_button);
  if (!quit_button) {
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#elif (SPACEDIM==2)
  if (mesh_gt!=0) data_out.drawGrid(mesh_gt,quit_button);
  switch (plot_type) {
    case COLOR_PLOT:
      data_out.plot("solution",plot_gt,quit_button);
      if (!quit_button) {
        AreaGraphTool::WINDOW_TYPE::QuitButton qb;
      }
      break;
    case CONTOUR_PLOT:
      data_out.contourPlot("solution",contour_gt,quit_button);
      break;
    case SURFACE_PLOT:
      data_out.plotSurface("solution",surface_gt);
      break;
  }
#else
  if (mesh_gt!=0) data_out.drawGrid(mesh_gt,quit_button);
  switch (plot_type) {
    case CLIP_PLANE_PLOT:
      data_out.plot("solution",plot_gt,quit_button);
      break;
    case SURFACE_PLOT:
      data_out.plotSurface("solution",surface_gt);
      break;
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void plotErrors(const int *n_active_cells,
const int *n_dofs,REAL *log_n_dofs,REAL *L2_error,
REAL *H1_error,REAL *Linfty_error) {
  if (n_cycles<=1) return;
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

  XYGraphTool nacgt("number active cells vs cycle","cycle",
    "number active cells",0,n_cycles-1,n_active_cells[0],
    n_active_cells[n_cycles-1],&cmap,0,0.5);
  nacgt.newPage();
  nacgt.setbgColor("white");
  nacgt.setfgColor("black");
  nacgt.drawAxes();
  nacgt.setfgColor("blue");
  nacgt.movePen(0,n_active_cells[0]);
  for (int cycle=1;cycle<n_cycles;cycle++) {
    nacgt.drawLine(cycle,n_active_cells[cycle]);
  }
  nacgt.flush();

  XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
    "number degress of freedom",0,n_cycles-1,n_dofs[0],
    n_dofs[n_cycles-1],&cmap,0,0.5);
  ndgt.newPage();
  ndgt.setbgColor("white");
  ndgt.setfgColor("black");
  ndgt.drawAxes();
  ndgt.setfgColor("blue");
  ndgt.movePen(0,n_dofs[0]);
  for (int cycle=1;cycle<n_cycles;cycle++) {
    ndgt.drawLine(cycle,n_dofs[cycle]);
  }
  ndgt.flush();

  double log10=log(10.);
  for (int cycle=0;cycle<n_cycles;cycle++) {
    log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
      /(static_cast<REAL>(SPACEDIM)*log10);
    L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                        : log(DBL_EPSILON))/log10;
    H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                        : log(DBL_EPSILON))/log10;
    Linfty_error[cycle]=(Linfty_error[cycle]>0. 
      ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
  }

  REAL ymin=L2_error[0];
  REAL ymax=L2_error[0];
  for (int cycle=1;cycle<n_cycles;cycle++) {
    ymin=min(ymin,L2_error[cycle]);
    ymax=max(ymax,L2_error[cycle]);
  }
  XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
    "log_10(degrees of freedom)","log_10(L2 error)",
    log_n_dofs[0],log_n_dofs[n_cycles-1],ymin,ymax,&cmap,0,0.5);
  l2gt.newPage();
  l2gt.setbgColor("white");
  l2gt.setfgColor("black");
  l2gt.drawAxes();
  l2gt.setfgColor("blue");
  l2gt.movePen(log_n_dofs[0],L2_error[0]);
  for (int cycle=1;cycle<n_cycles;cycle++) {
    l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
  }
  l2gt.flush();

  ymin=H1_error[0];
  ymax=H1_error[0];
  for (int cycle=1;cycle<n_cycles;cycle++) {
    ymin=min(ymin,H1_error[cycle]);
    ymax=max(ymax,H1_error[cycle]);
  }
  XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
    "log_10(degrees of freedom)","log_10(H1 error)",
    log_n_dofs[0],log_n_dofs[n_cycles-1],ymin,ymax,&cmap,0,0.5);
  h1gt.newPage();
  h1gt.setbgColor("white");
  h1gt.setfgColor("black");
  h1gt.drawAxes();
  h1gt.setfgColor("blue");
  h1gt.movePen(log_n_dofs[0],H1_error[0]);
  for (int cycle=1;cycle<n_cycles;cycle++) {
    h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
  }
  h1gt.flush();

  ymin=Linfty_error[0];
  ymax=Linfty_error[0];
  for (int cycle=1;cycle<n_cycles;cycle++) {
    ymin=min(ymin,Linfty_error[cycle]);
    ymax=max(ymax,Linfty_error[cycle]);
  }
  XYGraphTool linftygt(
    "log_10(Linfinity error) vs log_10(degrees of freedom)",
    "log_10(degrees of freedom)","log_10(Linfinity error)",
    log_n_dofs[0],log_n_dofs[n_cycles-1],ymin,ymax,&cmap,0,0.5);
  linftygt.newPage();
  linftygt.setbgColor("white");
  linftygt.setfgColor("black");
  linftygt.drawAxes();
  linftygt.setfgColor("blue");
  linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
  for (int cycle=1;cycle<n_cycles;cycle++) {
    linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
  }
  linftygt.flush();

  XYGraphTool::WINDOW_TYPE::QuitButton qb;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runMain(BOOLEAN /*called_before*/) {
#ifdef TIMING
//TimedObject *z3=
//  OPERATOR_NEW TimedObject("main interpolateBoundaryValues",tl);
//TimedObject *z5=OPERATOR_NEW TimedObject("main solve",tl);
//TimedObject *z10=OPERATOR_NEW TimedObject("main refine",tl);
#endif
  Tessellation tess;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=
    initializeMesh(tess,boundary,boundary_condition);

  ExactFunctionSystem efs;
  switch (domain_shape) {
    case RECTANGULAR_DOMAIN:
      setBoundaryConditionsInRectangle(tess,efs,boundary_condition);
      break;
#if (SPACEDIM>1)
    case SPHERICAL_DOMAIN:
      setBoundaryConditionsInSphere(tess,efs,boundary,boundary_condition);
      tess.synchronize(boundary_data);
      break;
#endif
  }

  SimpleMaterialModel smm;
  tess.setMaterialModel(&smm);

  Mapping *mapping=OPERATOR_NEW MappingL(mapping_degree);

  FiniteElement *fe=0;
  switch (finite_element_type) {
    case LAGRANGE_FINITE_ELEMENT:
      fe=OPERATOR_NEW FE_L(degree);
      break;
    case HIERARCHICAL_FINITE_ELEMENT:
      fe=OPERATOR_NEW FE_H(degree);
      break;
  }

  map<int,Quadrature<SPACEDIM> > qmap;
  initializeQuadratures(qmap);

  map<int,Quadrature<SPACEDIM-1> > nqmap; // natural bc's
#if (SPACEDIM>1)
  initializeFaceQuadratures(*fe,nqmap);
#endif

  int *n_active_cells=OPERATOR_NEW_BRACKET(int,n_cycles);
  int *n_dofs=OPERATOR_NEW_BRACKET(int,n_cycles);
  REAL *log_n_dofs=OPERATOR_NEW_BRACKET(REAL,n_cycles);
  REAL *L2_error=OPERATOR_NEW_BRACKET(REAL,n_cycles);
  REAL *H1_error=OPERATOR_NEW_BRACKET(REAL,n_cycles);
  REAL *Linfty_error=OPERATOR_NEW_BRACKET(REAL,n_cycles);

//graphics setup
  Palette pal;
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  plot_gt=OPERATOR_NEW XYGraphTool("solution plot","x","y",0.,1.,0.,1.,
    &cmap,0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  if (plot_mesh) {
    mesh_gt=OPERATOR_NEW AreaGraphTool("grid","x0","x1",rlo,rhi,&cmap,
      0,0.5);
  }
  switch (plot_type) {
    case COLOR_PLOT:
      plot_gt=OPERATOR_NEW AreaGraphTool("solution plot","x0","y1",
        rlo,rhi,&cmap,0,0.5);
      break;
    case CONTOUR_PLOT:
      contour_gt=OPERATOR_NEW AreaGraphTool("solution contours",
        "x0","x1",rlo,rhi,&cmap,0,0.5);
      break;
    case SURFACE_PLOT: {
      BOOLEAN use_lighting=FALSE;
      Vector3 vlow(0.,0.,0.);
      Vector3 vhigh(1.,1.,1.);
      surface_gt=OPERATOR_NEW VolGraphTool("solution surface",
        vlow,vhigh,pal,use_lighting,0.5);
      { VolGraphTool::DrawDelimiter dd(TRUE,surface_gt);
        surface_gt->drawBox(vlow,vhigh);
      }
      break;
    }
  }
#else
  BOOLEAN use_lighting=FALSE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  if (plot_mesh) {
    mesh_gt=OPERATOR_NEW VolGraphTool("grid",vlow,vhigh,pal,
      use_lighting,0.5);
    { VolGraphTool::DrawDelimiter dd(TRUE,mesh_gt);
      mesh_gt->drawBox(vlow,vhigh);
    }
  }
  switch (plot_type) {
    case CLIP_PLANE_PLOT: {
      plot_gt=OPERATOR_NEW VolGraphTool("solution plot",vlow,vhigh,
        pal,use_lighting,0.5);
      { VolGraphTool::DrawDelimiter dd(TRUE,plot_gt);
        plot_gt->drawBox(vlow,vhigh);
      }
      break;
    }
    case SURFACE_PLOT: {
      surface_gt=OPERATOR_NEW VolGraphTool("solution surface",
        vlow,vhigh,pal,use_lighting,0.5);
      { VolGraphTool::DrawDelimiter dd(TRUE,surface_gt);
        surface_gt->drawBox(vlow,vhigh);
      }
      break;
    }
  }
#endif

//loop over global mesh refinements
  for (int cycle=0;cycle<n_cycles;cycle++) {

//  data structure to keep track of degrees of freedom in solve
    DofHandler dof_handler(tess,*fe);
    dof_handler.distributeDofs();
    n_dofs[cycle]=dof_handler.numberDofs();
    cout << "\n\tNumber of degrees of freedom = " << n_dofs[cycle]
         << endl;

//  data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//  storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

//  assemble linear system
    dof_handler.assembleLinearSystem(*mapping,qmap,nqmap,system_matrix,
      system_rhs);

//  compute essential boundary values at mapped points on bdry
//  calls FunctionSystem in BoundaryCondition's to set boundary values
    map<int,REAL> boundary_values;
    {
#ifdef TIMING
//  Timer y(z3);
#endif
    dof_handler.interpolateBoundaryValues(*mapping,boundary_values);
    }

//  modify linear system and rhs for essential boundary condition
    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);
    solveLinearSystem(tess,*fe,*mapping,system_matrix,system_rhs,solution);

    n_active_cells[cycle]=tess.numberActiveShapes();
    FunctionSystem *weight=0;
    L2_error[cycle]=dof_handler.integrateDifference(*mapping,solution,efs,
      qmap,L2_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tL2_error[" << cycle << "] = " << L2_error[cycle] << endl;
#endif
    H1_error[cycle]=dof_handler.integrateDifference(*mapping,solution,efs,
      qmap,H1_NORM,weight,1.);
    Linfty_error[cycle]=dof_handler.integrateDifference(*mapping,solution,
      efs,qmap,LINFTY_NORM,weight,1.);

    plotSolution(dof_handler,*mapping,n_active_cells[cycle],solution);
    if (cycle<n_cycles-1) {
#ifdef TIMING
//    Timer y(z10);
#endif
      tess.refineGlobal(1);
    }
    delete sparsity_pattern;
  }
  plotErrors(n_active_cells,n_dofs,log_n_dofs,L2_error,H1_error,
    Linfty_error);
#ifdef TIMING
//tl.printOn(cout);
//tl.~TimedObjectList();
#endif

  delete n_active_cells;
  delete n_dofs;
  delete log_n_dofs;
  delete L2_error;
  delete H1_error;
  delete Linfty_error;
  if (boundary_data) delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    if (boundary_condition[i]->isNatural()) {
      delete boundary_condition[i]->getFunctionSystem();
    }
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  delete mapping;
  delete fe;
#if (SPACEDIM==1)
  if (plot_gt!=0) delete plot_gt; plot_gt=0;
#endif
#if (SPACEDIM==2)
  if (mesh_gt!=0) delete mesh_gt; mesh_gt=0;
  if (plot_gt!=0) delete plot_gt; plot_gt=0;
  if (contour_gt!=0) delete contour_gt; contour_gt=0;
  if (surface_gt!=0) delete surface_gt; surface_gt=0;
#endif
#if (SPACEDIM==3)
  if (mesh_gt!=0) delete mesh_gt; mesh_gt=0;
  if (plot_gt!=0) delete plot_gt; plot_gt=0;
  if (surface_gt!=0) delete surface_gt; surface_gt=0;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//use this for things that need to be done at the end of each call to
//  runMain by the GUI
void cleanup() {
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//use this for things that need to be done at the end of GUI operations
void shutdown() {
//things allocated in processCommandLine:
  if (display_name) delete display_name; display_name=0;

//things allocated in makeMainList:
  while (main_list->notEmpty()) {
    delete main_list->delAfter(0);
  }
  delete main_list; main_list=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << setprecision(16);
  ifstream in_file;
  processCommandLine(argc,argv,display_name,skip_gui,in_file);
  makeMainList(main_list);
  if (in_file) {
    readMainInput(in_file,main_list,skip_gui);
    if (skip_gui) checkMainInput();
  }

//  open graphical user interface; turn control of events to it
  if (skip_gui) {
    runMain(0);
    cleanup();
    shutdown();
  } else {
#ifdef USE_GTK
    GTKGUI gui(argv[0],display_name,main_list,&runMain,
      &checkMainInput,&cleanup,&shutdown,TRUE);
#else
    GUI gui(argv[0],display_name,main_list,&runMain,&checkMainInput,
      &cleanup,&shutdown,TRUE);
#endif
    gui.createFileMenu();
    gui.createViewMenu();
    gui.createHelpMenu();
    gui.createMainWindow(argc,argv);
    gui.eventLoop();
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int,
  pair<int const, double>,
  _Select1st<pair<int const, double> >,
  less<int>,
  allocator<pair<int const, double> >
>;
