// "$Header: /home/faculty/johnt/cvs/deal_new/examples/RectangularDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2010 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "DataOut.H"
#include "DofHandler.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "TimedObject.H"
#include "TimeIntegrator.H"
#include "Tracer.H"
#include "Vector.H"

bool tessellation_fixed=true;
bool coefs_transient=false;
bool bcs_transient=false;
bool use_elliptic_projection=true;
bool use_deferred_correction=true;
enum BASE_TIME_INTEGRATOR_TYPE {
  FORWARD_EULER_TIME_INTEGRATOR,
  BACKWARD_EULER_TIME_INTEGRATOR,
  CRANK_NICOLSON_TIME_INTEGRATOR,
  TRBDF_TIME_INTEGRATOR
};
BASE_TIME_INTEGRATOR_TYPE base_time_integrator_type=
  CRANK_NICOLSON_TIME_INTEGRATOR;

#if (SPACEDIM==1)
#elif (SPACEDIM==2)
  enum MESH_ELEMENTS { 
    QUADRILATERAL_MESH_ELEMENTS,
    TRIANGULAR_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=QUADRILATERAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TRIANGULAR_MESH_ELEMENTS;
#else
  enum MESH_ELEMENTS { 
    HEXAHEDRAL_MESH_ELEMENTS,
    PRISM_MESH_ELEMENTS,
    TETRAHEDRAL_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=HEXAHEDRAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=PRISM_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TETRAHEDRAL_MESH_ELEMENTS;
//int dir=0;
#endif
int degree=1;
int nelements=20;
//int nelements=20;


//constant functions for testing transient boundary conditions
class FakeConstantFunctionSystem : public TimeDependentFunctionSystem {
  private:
    REAL val;

    explicit FakeConstantFunctionSystem(
    const FakeConstantFunctionSystem &bc) : 
    TimeDependentFunctionSystem(bc) {
      OBSOLETE("copy constructor disallowed");
    }
    virtual REAL value(const Point<SPACEDIM> &point,int component) const {
      OBSOLETE("prohibit FunctionSystem version");
      return TimeDependentFunctionSystem::value(point,component);
    }
    virtual void value(const Point<SPACEDIM> &point,Vector &v) const {
      OBSOLETE("prohibit FunctionSystem version");
      TimeDependentFunctionSystem::value(point,v);
    }
    void value(NumPtr<REAL> &return_values,
    const NumPtr<Point<SPACEDIM> > &points,int component) const {
      OBSOLETE("prohibit FunctionSystem version");
      return TimeDependentFunctionSystem::value(return_values,points,
        component);
    }
    void value(NumPtr<Vector> &return_values,
    const NumPtr<Point<SPACEDIM> > &points) const {
      OBSOLETE("prohibit FunctionSystem version");
      return TimeDependentFunctionSystem::value(return_values,points);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &point,
    int component) const {
      OBSOLETE("prohibit FunctionSystem version");
      return TimeDependentFunctionSystem::gradient(point,component);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> > & return_gradients,
    const Point<SPACEDIM> &points) const {
      OBSOLETE("prohibit FunctionSystem version");
      TimeDependentFunctionSystem::gradient(return_gradients,points);
    }
    void gradient(NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
    const NumPtr<Point<SPACEDIM> > &points,int component) const {
      OBSOLETE("prohibit FunctionSystem version");
      TimeDependentFunctionSystem::gradient(return_gradients,points,
        component);
    }
    void gradient(NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &return_gradients,
    const NumPtr<Point<SPACEDIM> > &points) const {
      OBSOLETE("prohibit FunctionSystem version");
      TimeDependentFunctionSystem::gradient(return_gradients,points);
    }
  public:
    FakeConstantFunctionSystem(REAL v,int nc=1) :
      TimeDependentFunctionSystem(nc),val(v) {}
    virtual REAL value(const Point<SPACEDIM>&,REAL,int) const {
      return val;
    }
    virtual void value(const Point<SPACEDIM>&,REAL,Vector &v) const {
      v=val;
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM>&,REAL,int)
    const {
      Tensor<1,SPACEDIM> return_gradient;
      return_gradient.initialize(0.);
      return return_gradient;
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
    const Point<SPACEDIM>&,REAL) const {
      if (return_gradients.getNumber()!=n_components) {
        return_gradients.cleanup();
        return_gradients.allocate(n_components);
      }
      for (int c=0;c<n_components;c++) {
        return_gradients[c].initialize(0.);
      }
    }
};

// 1D:
//   partial_t u - partial_x^2 u = 0
//   u(0,t) = u_0
//   partial_x u(1,t) = u_1
//   u(x,0) = u_0 + u_1 x + sin(pi x/2)
//   ==> u(x,t) = u_0 + u_1 x + sin(pi x/2) exp(- pi^2 t/4)
// 2D:
//   partial_t u - partial_{x_0}^2 u - partial_{x_1}^2 u
//     = pi^2 sin(pi x_0/2) sin(pi x_1) exp(-pi^2 t/4)
//   u(0,x_1,t) = 0
//   u_x(1,x_1,t) = 0
//   u(x_0,0,t) = 0
//   u(x_0,1,t) = 0
//   u(x_0,x_1,0) = sin(pi x_0/2) sin(pi x_1)
//   ==> u(x,t) = sin(pi x_0/2) sin(pi x_1) e^{- pi^2 t/4)
// 3D:
//  partial_t u - partial_{x_0}^2 u - partial_{x_1}^2 u - partial_{x_2}^2 u
//     = 5 pi^2 sin(pi x_0/2) sin(pi x_1) exp(-pi^2 t/4)
//   u(0,x_1,x_2,t) = 0
//   u_x(1,x_1,x_2,t) = 0
//   u(x_0,x_1,x_2,t) = 0
//   u(x_0,0,x_2,t) = 0
//   u(x_0,1,x_2,t) = 0
//   u(x_0,x_1,0,t) = 0
//   u(x_0,x_1,1,t) = 0
//   u(x_0,x_1,x_2,0) = sin(pi x_0/2) sin(pi x_1) sin(2 pi x_2)
//   ==> u(x,t) = sin(pi x_0/2) sin(pi x_1) sin(2 pi x_2) e^{- pi^2 t/4)

//the following class is used to test transient parts of the generic code,
//  even though the HeatModel has constant coefficients
class TransientHeatModel : public TransientMaterialModel {
  private:
    REAL diffusion;

    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
    Matrix&,Vector&) const {
      OBSOLETE("does not apply to this class");
    }
  public:
    TransientHeatModel(REAL d=1.) : diffusion(d) { }
    REAL diffusionValue() const { return diffusion; }
    REAL exactSolution(const Point<SPACEDIM> &x,REAL t) const {
#if (SPACEDIM==1)
      return 1.+x[0]+sin(M_PI_2*x[0])*exp(-M_PI_2*M_PI_2*t);
#elif (SPACEDIM==2)
      return sin(M_PI_2*x[0])*sin(M_PI*x[1])*exp(-M_PI_2*M_PI_2*t);
#else
      return sin(M_PI_2*x[0])*sin(M_PI*x[1])*sin(2.*M_PI*x[2])
            *exp(-M_PI_2*M_PI_2*t);
#endif
    }
    Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &x,
    REAL t) const {
      Tensor<1,SPACEDIM> grad;
      REAL et=exp(-M_PI_2*M_PI_2*t);
#if (SPACEDIM==1)
      grad[0]=1.+M_PI_2*cos(M_PI_2*x[0])*et;
#elif (SPACEDIM==2)
      REAL arg0=M_PI_2*x[0];
      REAL arg1=M_PI*x[1];
      grad[0]=M_PI_2*cos(arg0)*sin(arg1)*et;
      grad[1]=M_PI*sin(arg0)*cos(arg1)*et;
#else
      REAL arg0=M_PI_2*x[0];
      REAL arg1=M_PI*x[1];
      REAL arg2=2.*M_PI*x[2];
      REAL sin0=sin(arg0);
      REAL sin1=sin(arg1);
      REAL sin2=sin(arg2);
      grad[0]=M_PI_2*cos(arg0)*sin1*sin2*et;
      grad[1]=M_PI*sin0*cos(arg1)*sin2*et;
      grad[2]=2.*M_PI*sin0*sin1*cos(arg2)*et;
#endif
      return grad;
    }
    REAL pdeInhomogeneity(const Point<SPACEDIM> &x,REAL t) const {
#if (SPACEDIM==1)
      return 0.;
#elif (SPACEDIM==2)
      return M_PI*M_PI*exactSolution(x,t);
#else
      return 5.*M_PI*M_PI*exactSolution(x,t);
#endif
    }
//  used for L2 projection of initial data
    void elementL2ProjectionMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    Matrix &cell_mass_matrix,Vector &cell_rhs) const {
      int dpc=sv.size(0);
      CHECK_SAME(cell_mass_matrix.size(0),dpc);
      CHECK_SAME(cell_mass_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        REAL u0=exactSolution(q_points[q],0.);
        for (int i=0;i<dpc;i++) {
          REAL svi=sv(i,q);
          svi*=jxw;
          for (int j=0;j<dpc;j++) {
            cell_mass_matrix(i,j)+=svi*sv(j,q);
          }
          cell_rhs[i]+=svi*u0;
        }
      }
    }
//  used for L2 projection of initial data
    void elementEllipticProjectionMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,Tensor<1,SPACEDIM> > &sg,
    Matrix &cell_stiffness_matrix,Vector &cell_rhs) const {
      int dpc=sg.size(0);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        Tensor<1,SPACEDIM> g0=exactSolutionGradient(q_points[q],0.);
        for (int i=0;i<dpc;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=jxw*diffusion;
          for (int j=0;j<dpc;j++) {
            cell_stiffness_matrix(i,j)+=sgi*sg(j,q);
          }
          cell_rhs[i]+=(sgi*g0)*diffusion;
        }
      }
    }
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,REAL time,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,
    Matrix &cell_mass_matrix,Matrix &cell_stiffness_matrix,
    Vector &cell_rhs) const {
      int dpc=sv.size(0);
      CHECK_SAME(cell_mass_matrix.size(0),dpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        REAL rhs=pdeInhomogeneity(q_points[q],time);
        for (int i=0;i<dpc;i++) {
          REAL svi=sv(i,q)*jxw;
          Tensor<1,SPACEDIM> sgi=sg(i,q)*jxw*diffusion;
          for (int j=0;j<dpc;j++) {
            cell_mass_matrix(i,j)+=svi*sv(j,q);
            cell_stiffness_matrix(i,j)+=(sgi*sg(j,q));
          }
          cell_rhs[i]+=svi*rhs;
        }
      }
    }
    virtual void printOn(ostream &os) const {
      os << "\tTransientHeatModel" << endl;
      TransientMaterialModel::printOn(os);
    }
};

//constant coefficient heat equation with transient forcing
class TimeIndependentTransientHeatModel : 
public TimeIndependentTransientMaterialModel {
  private:
    REAL diffusion;

    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &shape_values,
    const Array<2,Tensor<1,SPACEDIM> > &shape_gradients,
    Matrix &stiffness_matrix,Vector &rhs) const {
      OBSOLETE("prohibit MaterialModel version");
    }

  public:
    TimeIndependentTransientHeatModel(REAL d=1.) : diffusion(d) { }
    REAL diffusionValue() const { return diffusion; }
    REAL exactSolution(const Point<SPACEDIM> &x,REAL t) const {
#if (SPACEDIM==1)
      return 1.+x[0]+sin(M_PI_2*x[0])*exp(-M_PI_2*M_PI_2*t);
#elif (SPACEDIM==2)
      return sin(M_PI_2*x[0])*sin(M_PI*x[1])*exp(-M_PI_2*M_PI_2*t);
#else
      return sin(M_PI_2*x[0])*sin(M_PI*x[1])*sin(2.*M_PI*x[2])
            *exp(-M_PI_2*M_PI_2*t);
#endif
    }
    Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &x,
    REAL t) const {
      Tensor<1,SPACEDIM> grad;
      REAL et=exp(-M_PI_2*M_PI_2*t);
#if (SPACEDIM==1)
      grad[0]=1.+M_PI_2*cos(M_PI_2*x[0])*et;
#elif (SPACEDIM==2)
      REAL arg0=M_PI_2*x[0];
      REAL arg1=M_PI*x[1];
      grad[0]=M_PI_2*cos(arg0)*sin(arg1)*et;
      grad[1]=M_PI*sin(arg0)*cos(arg1)*et;
#else
      REAL arg0=M_PI_2*x[0];
      REAL arg1=M_PI*x[1];
      REAL arg2=2.*M_PI*x[2];
      REAL sin0=sin(arg0);
      REAL sin1=sin(arg1);
      REAL sin2=sin(arg2);
      grad[0]=M_PI_2*cos(arg0)*sin1*sin2*et;
      grad[1]=M_PI*sin0*cos(arg1)*sin2*et;
      grad[2]=2.*M_PI*sin0*sin1*cos(arg2)*et;
#endif
      return grad;
    }
    REAL pdeInhomogeneity(const Point<SPACEDIM> &x,REAL t) const {
#if (SPACEDIM==1)
      return 0.;
#elif (SPACEDIM==2)
      return M_PI*M_PI*exactSolution(x,t);
#else
      return 5.*M_PI*M_PI*exactSolution(x,t);
#endif
    }
//  used for L2 projection of initial data
    void elementL2ProjectionMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    Matrix &cell_mass_matrix,Vector &cell_rhs) const {
      int dpc=sv.size(0);
      CHECK_SAME(cell_mass_matrix.size(0),dpc);
      CHECK_SAME(cell_mass_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        REAL u0=exactSolution(q_points[q],0.);
        for (int i=0;i<dpc;i++) {
          REAL svi=sv(i,q)*jxw;
          for (int j=0;j<dpc;j++) {
            cell_mass_matrix(i,j)+=svi*sv(j,q);
          }
          cell_rhs[i]+=svi*u0;
        }
      }
    }
//  used for elliptic projection of initial data
    void elementEllipticProjectionMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_stiffness_matrix,
    Vector &cell_rhs) const {
      int dpc=sg.size(0);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        Tensor<1,SPACEDIM> g0=exactSolutionGradient(q_points[q],0.);
        for (int i=0;i<dpc;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q)*jxw*diffusion;
          for (int j=0;j<dpc;j++) {
            cell_stiffness_matrix(i,j)+=(sgi*sg(j,q));
          }
          cell_rhs[i]+=(sgi*g0)*diffusion;
        }
      }
    }
//  used for time-dependent tessellations
    virtual void elementStiffnessMatrixAndRHS(
  const NumPtr<Point<SPACEDIM> > &q_points,REAL time,
  const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_mass_matrix,
    Matrix &cell_stiffness_matrix,Vector &cell_rhs) const {
      int dpc=sv.size(0);
      CHECK_SAME(cell_mass_matrix.size(0),dpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        REAL rhs=pdeInhomogeneity(q_points[q],time);
        for (int i=0;i<dpc;i++) {
          REAL svi=sv(i,q)*jxw;
          Tensor<1,SPACEDIM> sgi=sg(i,q)*jxw;
          for (int j=0;j<dpc;j++) {
            cell_mass_matrix(i,j)+=svi*sv(j,q);
            cell_stiffness_matrix(i,j)+=(sgi*sg(j,q))*diffusion;
          }
          cell_rhs[i]+=svi*rhs;
        }
      }
    }
//  used for fixed tessellations
    void elementRHS(const NumPtr<Point<SPACEDIM> > &q_points,
    REAL time,const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    Vector &cell_rhs) const {
      int dpc=sv.size(0);
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        REAL rhs=pdeInhomogeneity(q_points[q],time)*jxw;
        for (int i=0;i<dpc;i++) {
          cell_rhs[i]+=sv(i,q)*rhs;
        }
      }
    }
    virtual void printOn(ostream &os) const {
      os << "\tTimeIndependentTransientHeatModel" << endl;
      TimeIndependentTransientMaterialModel::printOn(os);
    }
};

#ifdef TIMING
//the most expensive routines are the linear system solve and
//  interpolateBoundaryValues
//the latter requires computation of the shape functions on an element
//  at iteratively computed canonical points corresponding to boundary:
//TimedObjectList tl;
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
#ifdef TIMING
//TimedObject *z3=
//  OPERATOR_NEW TimedObject("main interpolateBoundaryValues",tl);
//TimedObject *z5=OPERATOR_NEW TimedObject("main solve",tl);
//TimedObject *z10=OPERATOR_NEW TimedObject("main refine",tl);
#endif
#ifdef DEBUG
  cout << "\ttessellation_fixed = " << PRINT_BOOLEAN(tessellation_fixed)
       << endl;
  cout << "\tcoefs_transient = " << PRINT_BOOLEAN(coefs_transient)
       << endl;
  cout << "\tbcs_transient = " << PRINT_BOOLEAN(bcs_transient) << endl;
  cout << "\tuse_elliptic_projection = "
       << PRINT_BOOLEAN(use_elliptic_projection) << endl;
#endif
  Tessellation tess;
  const LEVEL_CLASS *level=tess.coarsestLevel();

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> low(D_DECL(0.,0.,0.));
//Point<SPACEDIM> low(0.,0.,0.);
  Point<SPACEDIM> edge[SPACEDIM];
    edge[0]=Point<SPACEDIM>(D_DECL(1.,0.,0.));
#if (SPACEDIM>=2)
    edge[1]=Point<SPACEDIM>(D_DECL(0.,1.,0.));
#endif
#if (SPACEDIM==3)
    edge[2]=Point<SPACEDIM>(0.,0.,1.);
#endif
  int number_elements[SPACEDIM]={D_DECL(nelements,nelements,nelements)};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=0;
#if (SPACEDIM==1)
  boundary_data=generateParallelpiped(tess,low,edge,number_elements,
    set_face_ids,boundary,boundary_condition);
  for (int i=0;i<2;i++) {
    const Boundary *b=boundary_condition[i]->getBoundary();
    delete boundary_condition[i];
    boundary_condition[i]=(bcs_transient ?
      OPERATOR_NEW TimeDependentBoundaryCondition(boundary[i]) :
      OPERATOR_NEW BoundaryCondition(boundary[i]) );
    boundary_condition[i]->setBoundary(b);
    level->CODIMENSION1_LEVEL_CLASS::setBoundaryCondition(i,
      boundary_condition[i]);
  }
#elif (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS: {
      boundary_data=generateParallelpiped(tess,low,edge,number_elements,
        set_face_ids,boundary,boundary_condition);
      for (int i=0;i<4;i++) {
        const Boundary *b=boundary_condition[i]->getBoundary();
        delete boundary_condition[i];
        boundary_condition[i]=(bcs_transient ?
          OPERATOR_NEW TimeDependentBoundaryCondition(boundary[i]) :
          OPERATOR_NEW BoundaryCondition(boundary[i]) );
        boundary_condition[i]->setBoundary(b);
      }
      const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          bool found=false;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            const BoundaryCondition *bc=boundary_condition[i];
            const Boundary *b=bc->getBoundary();
            bool all_on_boundary=true;
            for (int v=0;v<face_element->numberVertices();v++) {
              int vi=face_element->vertexIndex(v);
              const Point<SPACEDIM> &ver=vertices[vi];
              if (abs(b->constraint(ver))>1.e-5) {
                all_on_boundary=false;
                break;
              }
            }
            if (all_on_boundary) {
              codimension1_level->setBoundaryCondition(bc);
              found=true;
              break;
            }
          }
        }
      }
      break;
    }
    case TRIANGULAR_MESH_ELEMENTS: {
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<2> > vertices(5);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[3]+=edge[1];
      vertices[4]+=edge[0]; vertices[4]+=edge[1];
      vertices[2]=(vertices[0]+vertices[1]+vertices[3]+vertices[4])*0.25;
      NumPtr<CellData> cells(4);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(3);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;

      cells[1].vertex(0)=1;
      cells[1].vertex(1)=4;
      cells[1].vertex(2)=2;

      cells[2].vertex(0)=2;
      cells[2].vertex(1)=4;
      cells[2].vertex(2)=3;

      cells[3].vertex(0)=0;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=3;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(4);
      boundary_condition.allocate(6);
      Point<2> center(vertices[2]);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,vertices[0],
        vertices[3]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,vertices[1],
        vertices[4]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,vertices[0],
        vertices[1]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,vertices[3],
        vertices[4]);
      for (int i=0;i<4;i++) {
        boundary_condition[i]=(bcs_transient ?
          OPERATOR_NEW TimeDependentBoundaryCondition(boundary[i]) :
          OPERATOR_NEW BoundaryCondition(boundary[i]) );
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      boundary_data=generateParallelpiped(tess,low,edge,number_elements,
        set_face_ids,boundary,boundary_condition);
      for (int i=0;i<6;i++) {
        const Boundary *b=boundary_condition[i]->getBoundary();
        delete boundary_condition[i];
        boundary_condition[i]=(bcs_transient ?
          OPERATOR_NEW TimeDependentBoundaryCondition(boundary[i]) :
          OPERATOR_NEW BoundaryCondition(boundary[i]) );
        boundary_condition[i]->setBoundary(b);
      }
      const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          bool found=false;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            const BoundaryCondition *bc=boundary_condition[i];
            const Boundary *b=bc->getBoundary();
            bool all_on_boundary=true;
            for (int v=0;v<face_element->numberVertices();v++) {
              int vi=face_element->vertexIndex(v);
              const Point<SPACEDIM> &ver=vertices[vi];
              if (abs(b->constraint(ver))>1.e-5) {
                all_on_boundary=false;
                break;
              }
            }
            if (all_on_boundary) {
              codimension1_level->setBoundaryCondition(bc);
              found=true;
              break;
            }
          }
        }
      }
      for (bool doit=tess.firstCodimension2ActiveShape();doit;
      doit=tess.nextCodimension2ActiveShape()) {
        Level1 *codimension2_level=
          tess.presentCodimension2Level();
        pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
          codimension2_level->getBoundaryConditions();
        const BoundaryCondition *bc=bcs.first;
        if (bc==0) continue;
        Shape1 *face_element=codimension2_level->getShape();
        codimension2_level->setBoundaryCondition(0);
        int count=0;
        for (int i=0;i<boundary.getNumber();i++) {
          const BoundaryCondition *bc=boundary_condition[i];
          const Boundary *b=bc->getBoundary();
          bool all_on_boundary=true;
          for (int v=0;v<face_element->numberVertices();v++) {
            int vi=face_element->vertexIndex(v);
            const Point<SPACEDIM> &ver=vertices[vi];
            if (abs(b->constraint(ver))>1.e-5) {
              all_on_boundary=false;
              break;
            }
          }
          if (all_on_boundary) {
            codimension2_level->setBoundaryCondition(bc);
            count++;
          }
        }
      }
      break;
    }
    case PRISM_MESH_ELEMENTS: {
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<3> > vertices(10);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[0]*0.5; vertices[2]+=edge[1]*0.5; 
      vertices[3]+=edge[1];
      vertices[4]+=edge[0]; vertices[4]+=edge[1];
      vertices[5]+=edge[2];
      vertices[6]+=edge[0]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]*0.5; vertices[7]+=edge[1]*0.5; 
                                vertices[7]+=edge[2];
      vertices[8]+=edge[1]; vertices[8]+=edge[2];
      vertices[9]+=edge[0]; vertices[9]+=edge[1]; vertices[9]+=edge[2];
      NumPtr<CellData> cells(4);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(6);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=5;
      cells[0].vertex(4)=6;
      cells[0].vertex(5)=7;

      cells[1].vertex(0)=0;
      cells[1].vertex(1)=2;
      cells[1].vertex(2)=3;
      cells[1].vertex(3)=5;
      cells[1].vertex(4)=7;
      cells[1].vertex(5)=8;

      cells[2].vertex(0)=2;
      cells[2].vertex(1)=4;
      cells[2].vertex(2)=3;
      cells[2].vertex(3)=7;
      cells[2].vertex(4)=9;
      cells[2].vertex(5)=8;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=4;
      cells[3].vertex(2)=2;
      cells[3].vertex(3)=6;
      cells[3].vertex(4)=9;
      cells[3].vertex(5)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<6;i++) {
        boundary_condition[i]=(bcs_transient ?
          OPERATOR_NEW TimeDependentBoundaryCondition(boundary[i]) :
          OPERATOR_NEW BoundaryCondition(boundary[i]) );
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              found=true;
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<3> > vertices(8);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[1];
      vertices[3]+=edge[0]; vertices[3]+=edge[1];
      vertices[4]+=edge[2];
      vertices[5]+=edge[0]; vertices[5]+=edge[2];
      vertices[6]+=edge[1]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]; vertices[7]+=edge[1]; vertices[7]+=edge[2];
      NumPtr<CellData> cells(6);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(4);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=4;

      cells[1].vertex(0)=1;
      cells[1].vertex(1)=4;
      cells[1].vertex(2)=5;
      cells[1].vertex(3)=6;

      cells[2].vertex(0)=1;
      cells[2].vertex(1)=2;
      cells[2].vertex(2)=4;
      cells[2].vertex(3)=6;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=7;
      cells[3].vertex(3)=3;

      cells[4].vertex(0)=1;
      cells[4].vertex(1)=5;
      cells[4].vertex(2)=7;
      cells[4].vertex(3)=6;

      cells[5].vertex(0)=1;
      cells[5].vertex(1)=2;
      cells[5].vertex(2)=6;
      cells[5].vertex(3)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<6;i++) {
        boundary_condition[i]=(bcs_transient ?
          OPERATOR_NEW TimeDependentBoundaryCondition(boundary[i]) :
          OPERATOR_NEW BoundaryCondition(boundary[i]) );
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif
#if (SPACEDIM>=2)
  {
    tess.checkFaceBoundaryCondition(boundary_condition);
  }
#endif
#if (SPACEDIM==3)
  tess.checkLineBoundaryCondition(boundary_condition);
#endif

  TransientMaterialModel *hm;
  if (coefs_transient) hm=OPERATOR_NEW TransientHeatModel;
  else hm=OPERATOR_NEW TimeIndependentTransientHeatModel;
  tess.setMaterialModel(hm);

//specified solution at left is essential boundary condition
  FunctionSystem *efs;
  if (bcs_transient) efs=OPERATOR_NEW FakeConstantFunctionSystem(1.);
  else efs=OPERATOR_NEW ConstantFunctionSystem(1.);
  boundary_condition[0]->setFunctionSystem(efs);
  boundary_condition[0]->makeEssential();

//specified solution on right is natural boundary condition
  FunctionSystem *nfs;
  if (bcs_transient) nfs=OPERATOR_NEW FakeConstantFunctionSystem(1.);
  else nfs=OPERATOR_NEW ConstantFunctionSystem(1.);
  boundary_condition[1]->setFunctionSystem(nfs);
  boundary_condition[1]->makeNatural();

//other boundaries (if they exist) are homogeneous and essential
  FunctionSystem *zefs;
  if (bcs_transient) zefs=OPERATOR_NEW FakeConstantFunctionSystem(0.);
  else zefs=OPERATOR_NEW ConstantFunctionSystem(0.);
  for (int d=2;d<2*SPACEDIM;d++) {
    boundary_condition[d]->setFunctionSystem(zefs);
    boundary_condition[d]->makeEssential();
  }

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;

//set polynomials for each shape, face quadratures and dofs per object
  FE_L fe(degree);

//quadrature rules
  map<int,Quadrature<SPACEDIM> > qmap;
#if (SPACEDIM==1)
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree+1));
#elif (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS:
      qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
        GaussianQuadrature<2>::argGivenOrder(2*degree+1));
      break;
    case TRIANGULAR_MESH_ELEMENTS:
      qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree+1));
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
      qmap[Hexahedron::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
        GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+1));
      break;
    case PRISM_MESH_ELEMENTS:
      qmap[Prism::verticesPerShape()]=
        ProductQuadrature(
          TriangleGaussianQuadrature(
            TriangleGaussianQuadrature::argGivenOrder(2*degree+1)),
          GaussianQuadrature<1>(
            GaussianQuadrature<1>::argGivenOrder(2*degree+1)));
      break;
    case TETRAHEDRAL_MESH_ELEMENTS:
      qmap[Tetrahedron::verticesPerShape()]=TetrahedronGaussianQuadrature(
        TetrahedronGaussianQuadrature::argGivenOrder(2*degree+1));
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif

//boundary quadrature rules
  map<int,Quadrature<SPACEDIM-1> > nqmap;
#if (SPACEDIM>=2)
//natural bc's
//should be exact for polynomials of shape function degree
//  plus exact solution degree-1
//  plus Jacobian degree
# if (SPACEDIM==2)
  nqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
        GaussianQuadrature<1>::argGivenOrder(2*degree+1));
# else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<2>(
        GaussianQuadrature<2>::argGivenOrder(2*degree+1));
      break;
    }
    case PRISM_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<2>(
        GaussianQuadrature<2>::argGivenOrder(2*degree+1));
      nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree-1));
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
      int nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree-1));
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
# endif
#endif

//graphics setup
  Palette pal;
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool *plot_gt=OPERATOR_NEW XYGraphTool(
    "solution vs position","x","u",0.,1.,0.,1.,&cmap,0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
  AreaGraphTool *plot_gt=OPERATOR_NEW AreaGraphTool(
    "solution","x0","x1",rlo,rhi,&cmap,0,0.5);
#else
  BOOLEAN use_lighting=TRUE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool *plot_gt=OPERATOR_NEW VolGraphTool(
    "solution",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,plot_gt);
    plot_gt->drawBox(vlow,vhigh);
  }
//VolGraphTool surface_gt("solution",vlow,vhigh,pal,use_lighting,0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
#endif

//  data structure to keep track of degrees of freedom in solver
    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();
    int n_dofs=dof_handler.numberDofs();
#ifdef DEBUG
    cout << "\tnumber of active elements = " 
         << tess.numberActiveShapes() << endl;
    cout << "\tNumber of degrees of freedom = " << n_dofs << endl;
    cout << "\tassemble initial linear system..." << endl;
#endif
//  data structure to keep track of linear system storage
    int band_width=dof_handler.maxCouplingsBetweenDofs();
#ifdef DEBUG
//  cout << "\tband_width = " << band_width << endl;
#endif
    SparsityPattern *sparsity_pattern=OPERATOR_NEW SparsityPattern(
      n_dofs,n_dofs,band_width,true);
    dof_handler.makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();
    PreconditionIdentity pi;
    RelativeErrorControl cg_solver_control(n_dofs,DBL_EPSILON*1.e6,
      false,false);
    CGSolver cg(cg_solver_control);
    TimeControl tc(1000,10.,DBL_EPSILON);
    bool preserve_symmetry=true;

    int ns=0;
    if (use_deferred_correction) {
      ns=DeferredCorrectionTimeIntegrator::numberSolutions(degree+1);
    } else {
      switch (base_time_integrator_type) {
        case FORWARD_EULER_TIME_INTEGRATOR:
          ns=ForwardEulerTimeIntegrator::numberSolutions();
          break;
        case BACKWARD_EULER_TIME_INTEGRATOR:
          ns=BackwardEulerTimeIntegrator::numberSolutions();
          break;
        case CRANK_NICOLSON_TIME_INTEGRATOR:
          ns=CrankNicolsonTimeIntegrator::numberSolutions();
          break;
        case TRBDF_TIME_INTEGRATOR:
          ns=TRBDFTimeIntegrator::numberSolutions();
          break;
        default:
          OBSOLETE("unknown time integrator type");
      }
    }
    int nm=(dynamic_cast<TimeIndependentTransientMaterialModel*>(hm)
      ? 1 : ns);
    NumPtr<Vector*> solution(ns); 
      solution.initialize(0);
    NumPtr<Vector*> rhs(ns); 
      rhs.initialize(0);
    NumPtr<SparseMatrix*> mass_matrix(nm); 
      mass_matrix.initialize(0);
    NumPtr<SparseMatrix*> stiffness_matrix(nm); 
      stiffness_matrix.initialize(0);

    TimeIntegrator *ti=0;
    switch (base_time_integrator_type) {
      case FORWARD_EULER_TIME_INTEGRATOR:
        ti=OPERATOR_NEW ForwardEulerTimeIntegrator(tc,dof_handler,
          mapping,qmap,nqmap,pi,cg,solution,rhs,tessellation_fixed,
          preserve_symmetry);
        break;
      case BACKWARD_EULER_TIME_INTEGRATOR:
        ti=OPERATOR_NEW BackwardEulerTimeIntegrator(tc,dof_handler,
          mapping,qmap,nqmap,pi,cg,solution,rhs,tessellation_fixed,
          preserve_symmetry);
        break;
      case CRANK_NICOLSON_TIME_INTEGRATOR:
        ti=OPERATOR_NEW CrankNicolsonTimeIntegrator(tc,dof_handler,
          mapping,qmap,nqmap,pi,cg,solution,rhs,tessellation_fixed,
          preserve_symmetry);
        break;
      case TRBDF_TIME_INTEGRATOR:
        ti=OPERATOR_NEW TRBDFTimeIntegrator(tc,dof_handler,mapping,qmap,
          nqmap,pi,cg,solution,rhs,tessellation_fixed,preserve_symmetry);
        break;
      default:
        OBSOLETE("unknown time integrator type");
    }
//
    DeferredCorrectionTimeIntegrator *dcti=0;
    if (use_deferred_correction) {
      dcti=OPERATOR_NEW 
        DeferredCorrectionTimeIntegrator(*ti,degree+1,solution,rhs);
    }

//  compute initial data
    REAL dx=1./static_cast<REAL>(nelements);
    if (use_elliptic_projection) {
      ti->ellipticProjectionInitialData(sparsity_pattern,solution,rhs,
        mass_matrix,stiffness_matrix);
    } else {
      ti->l2ProjectionInitialData(sparsity_pattern,solution,rhs,
        mass_matrix,stiffness_matrix);
    }
//  compute solution values for graphical display
    DataOut data_out(dof_handler,mapping);
    data_out.addDataVector(*solution[0],"solution");
//  int n_subdivisions=degree+1;
    int n_subdivisions=5;
    data_out.buildPatches(n_subdivisions);
    bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>1)
    data_out.drawGrid(&mesh_gt,quit_button);
#endif
    data_out.plot("solution",plot_gt,quit_button);
//  data_out.plotSurface("solution",&surface_gt);
#if (SPACEDIM==1)
    if (!quit_button) {
      XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif
    data_out.clearInputDataReferences();

//  Forward Euler, degree=1:
//  REAL dt=0.9*dx*dx/(6.*(coefs_transient 
//    ?  dynamic_cast<TransientHeatModel*>(hm)->diffusionValue()
//    :  dynamic_cast<TimeIndependentTransientHeatModel*>(hm)
//       ->diffusionValue()));
//  Backward Euler, Crank Nicolsons or TRBDF: same temporal & spatial order
    REAL dt=dx;
    while (true) {
#ifdef TIMING
//    Timer y(z5);
#endif
      TIME_INTEGRATOR_STATE state=TIME_INTEGRATOR_STEPPING;
      if (use_deferred_correction) {
        state=dcti->takeStep(dt,sparsity_pattern,solution,rhs,
          mass_matrix,stiffness_matrix);
      } else {
        state=ti->takeStep(dt,sparsity_pattern,solution,rhs,mass_matrix,
          stiffness_matrix);
      }
      data_out.replaceDataVector(*solution[0],"solution");
      data_out.reBuildPatches();
      data_out.plot("solution",plot_gt,quit_button);
/*
//    data_out.plotSurface("solution",&surface_gt);
*/
      data_out.clearInputDataReferences();
      if (state==TIME_INTEGRATOR_DONE) break;
    }
#if (SPACEDIM==1)
    {
      XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif

  for (int i=0;i<solution.getNumber();i++) {
    if (solution[i]) delete solution[i];
    solution[i]=0;
  }
  for (int i=0;i<rhs.getNumber();i++) {
    if (rhs[i]) delete rhs[i];
    rhs[i]=0;
  }
  for (int i=0;i<mass_matrix.getNumber();i++) {
    if (mass_matrix[i]) delete mass_matrix[i];
    mass_matrix[i]=0;
  }
  for (int i=0;i<stiffness_matrix.getNumber();i++) {
    if (stiffness_matrix[i]) delete stiffness_matrix[i];
    stiffness_matrix[i]=0;
  }
  delete sparsity_pattern;
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  if (ti) delete ti; ti=0;
  if (dcti) delete dcti; dcti=0;
  delete plot_gt;
  delete hm;
  delete efs;
  delete nfs;
  delete zefs;
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
