// "$Header: /home/faculty/johnt/cvs/deal_new/examples/HyperElasticity.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "BoundaryCondition.H"
#include "ConstraintMatrix.H"
#include "DataOut.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "MooneyRivlin.H"
#include "NonlinearDofHandler.H"
#include "NonlinearSolver.H"
#include "Precondition.H"
#include "Tessellation.H"
#include "Tracer.H"

//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
class ZeroDisplacementFunctionSystem : public FunctionSystem {
  public:
    ZeroDisplacementFunctionSystem() : FunctionSystem(SPACEDIM) {;}
    REAL value(const Point<SPACEDIM> &p,int c) const { return 0.; }
    void value(const Point<SPACEDIM> &p,Vector &return_value) const {
      return_value=0.;
    }
    virtual void printOn(ostream &os) const {
      os << "\tZeroDisplacementFunctionSystem" << endl;
      FunctionSystem::printOn(os);
    }
};

class RollerFunctionSystem : public FunctionSystem {
  private:
    NumPtr<bool> constrained_component;
  public:
    RollerFunctionSystem() : FunctionSystem(SPACEDIM),
    constrained_component(SPACEDIM) {
      constrained_component.initialize(false);
      constrained_component[SPACEDIM-1]=true;
    }
    REAL value(const Point<SPACEDIM> &p,int c) const { 
      if (constrained_component[c]) return 0.; // zero displacement
      return 0.; // zero force
    }
    void value(const Point<SPACEDIM> &p,Vector &return_value) const {
      return_value=HUGE_VAL;
      for (int c=0;c<constrained_component.getNumber();c++) {
        if (constrained_component[c]) return_value[c]=0.;//zero displacement
        else return_value[c]=0.; // zero force
      }
    }
    const NumPtr<bool>& constrainedComponent() const {
      return constrained_component;
    }
    virtual void printOn(ostream &os) const {
      os << "\tRollerFunctionSystem" << endl;
      for (int c=0;c<constrained_component.getNumber();c++) {
        os << "\t\tconstrained_component[" << c << "] = "
           << constrained_component[c] << endl;
      }
      FunctionSystem::printOn(os);
    }
};

class SpecifiedTractionFunctionSystem : public FunctionSystem {
  private:
    Point<SPACEDIM> (*specified_function)(const Point<SPACEDIM>&); 
  public:
    SpecifiedTractionFunctionSystem(
    Point<SPACEDIM> (*sf)(const Point<SPACEDIM>&)) : 
      FunctionSystem(SPACEDIM),specified_function(sf) {;}
    REAL value(const Point<SPACEDIM> &p,int c) const { 
      Point<SPACEDIM> f(specified_function(p));
      return f[c];
    }
    void value(const Point<SPACEDIM> &p,Vector &return_value) const {
      Point<SPACEDIM> f(specified_function(p));
      for (int c=0;c<return_value.size();c++) return_value[c]=f[c];
    }
    virtual void printOn(ostream &os) const {
      os << "\tSpecifiedTractionFunctionSystem" << endl;
      FunctionSystem::printOn(os);
    }
};
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL f_size=1.e-3;
Point<SPACEDIM> boundaryForce(const Point<SPACEDIM> &p) {
#if (SPACEDIM==1)
  Point<SPACEDIM> boundary_force(-f_size);
#elif (SPACEDIM==2)
  Point<SPACEDIM> boundary_force(0.,-f_size);
#else
  Point<SPACEDIM> boundary_force(0.,f_size,-f_size);
#endif
  return boundary_force;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> zeroForce(const Point<SPACEDIM>&) {
  return Point<SPACEDIM>(D_DECL(0.,0.,0.));
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
double lame=1.;
double shear=0.3;
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main (int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  Tessellation tess;
  cout << "Solving problem in " << SPACEDIM << " space dimensions." 
       << endl;

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> low(D_DECL(-1.0,0.,0.));
  Point<SPACEDIM> edge[SPACEDIM];
  edge[0]=Point<SPACEDIM>(D_DECL(1.2,0.,0.));
#if (SPACEDIM>1)
  edge[1]=Point<SPACEDIM>(D_DECL(0.,0.1,0.));
#endif
#if (SPACEDIM==3)
  edge[2]=Point<SPACEDIM>(D_DECL(0.,0.,0.1));
#endif
  int number_elements[SPACEDIM]={D_DECL(12,1,1)};
  bool set_face_ids=false;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);

//nonlinear material model
  MooneyRivlinMaterialModel mooney_rivlin(lame,shear,zeroForce); 
//LinearElasticityMaterialModel mooney_rivlin(lame,shear,bodyForce);
  tess.setMaterialModel(&mooney_rivlin);

//PlaneBoundary's all around
  boundary.allocate(2*SPACEDIM);
#if (SPACEDIM==1)
  Point<1> center(low+edge[0]*0.5);
#elif (SPACEDIM==2)
  Point<2> center(low+(edge[0]+edge[1])*0.5);
#else
  Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
#endif
  for (int d=0;d<SPACEDIM;d++) {
    boundary[2*d]=OPERATOR_NEW
    PlaneBoundary(center,
      D_DECL(low,low+edge[(d+1)%SPACEDIM],low+edge[(d+2)%SPACEDIM]));
    boundary[2*d+1]=OPERATOR_NEW
      PlaneBoundary(center,
        D_DECL(low+edge[d],low+edge[d]+edge[(d+1)%SPACEDIM],
               low+edge[d]+edge[(d+2)%SPACEDIM]));
  }

//BoundaryCondition's (2 on bottom in 2D and 3D)
  boundary_condition.allocate(D_CHOOSE(2,5,7));
  for (int i=0;i<2*SPACEDIM;i++) {
    boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
  }
#if (SPACEDIM>1)
  boundary_condition[2*SPACEDIM]=
    OPERATOR_NEW BoundaryCondition(boundary[2*(SPACEDIM-1)]);
#endif

//clamped boundary at left (ie, at back in 3d)
  ZeroDisplacementFunctionSystem zdfs;
  boundary_condition[0]->setFunctionSystem(&zdfs);
  for (int d=0;d<SPACEDIM;d++) boundary_condition[0]->makeEssential(d);

//specified traction at right (ie at front in 3d)
  SpecifiedTractionFunctionSystem stfs(boundaryForce);
  boundary_condition[1]->setFunctionSystem(&stfs);
  for (int d=0;d<SPACEDIM;d++) boundary_condition[1]->makeNatural(d);

//free surface on (almost) all other faces
  SpecifiedTractionFunctionSystem ztfs(zeroForce);
  for (int i=2;i<2*SPACEDIM;i++) {
    boundary_condition[i]->setFunctionSystem(&ztfs);
    for (int d=0;d<SPACEDIM;d++) boundary_condition[i]->makeNatural(d);
  }

#if (SPACEDIM>1)
//roller surface on small portion of bottom
  RollerFunctionSystem rfs;
  boundary_condition[2*SPACEDIM]->setFunctionSystem(&rfs);
  const NumPtr<bool> &cc=rfs.constrainedComponent();
  for (int d=0;d<SPACEDIM;d++) {
    if (cc[d]) {
      boundary_condition[2*SPACEDIM]->makeEssential(d);
    } else {
      boundary_condition[2*SPACEDIM]->makeNatural(d);
    }
  }
#endif

//store BoundaryCondition's in Tessellation
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
      Point<SPACEDIM> center=face_element->center(vertices);
      bool found=false;
      for (int i=0;i<2*SPACEDIM;i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[i]);
          found=true;
        }
      }
      CHECK_TEST(found);
#if (SPACEDIM>1)
//    roller boundary condition
      if (codimension1_level->getBoundaryCondition()==
      boundary_condition[2*(SPACEDIM-1)]) {
        const Point<SPACEDIM> &v0=vertices[face_element->vertexIndex(0)];
        const Point<SPACEDIM> &v1=vertices[face_element->vertexIndex(1)];
# if (SPACEDIM==2)
        if (v0[0]<-0.05 && v1[0]>-0.05) {
          codimension1_level->setBoundaryCondition(boundary_condition[4]);
        }
# else
        const Point<SPACEDIM> &v2=vertices[face_element->vertexIndex(2)];
        const Point<SPACEDIM> &v3=vertices[face_element->vertexIndex(3)];
        REAL v1min=min(min(v0[1],v1[1]),min(v2[1],v3[1]));
        REAL v1max=max(max(v0[1],v1[1]),max(v2[1],v3[1]));
        if (v1min<-0.05 && v1max>-0.05) {
          codimension1_level->setBoundaryCondition(boundary_condition[6]);
        }
# endif
      }
#endif
    }
  }
  tess.synchronize(boundary_data);
#if (SPACEDIM>1)
  tess.checkFaceBoundaryCondition(boundary_condition);
#endif
#if (SPACEDIM==3)
  tess.checkLineBoundaryCondition(boundary_condition);
#endif

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;
  
//polynomials to represent solution with SPACEDIM components
  int degree=1;
  FE_L fe(degree,SPACEDIM);

//quadrature rules
  GaussianQuadrature<SPACEDIM> quadrature_formula(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(
    D_CHOOSE(2*degree-1,2*degree+1,2*degree+1)));
  map<int,Quadrature<SPACEDIM> > qmap;
  int nv=D_CHOOSE(Interval::verticesPerShape(),
                  Quadrilateral::verticesPerShape(),
                  Hexahedron::verticesPerShape());
  qmap[nv]=quadrature_formula;

  map<int,Quadrature<SPACEDIM-1> > fqmap; // natural bc's
#if (SPACEDIM==2)
  fqmap[Interval::verticesPerShape()]=GaussianQuadrature<SPACEDIM-1>(
    GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree-1));
#elif (SPACEDIM==3)
  fqmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<SPACEDIM-1>(
    GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+1));
#endif
//data structure to keep track of conforming grid constraints for
//  hanging nodes
  ConstraintMatrix hanging_node_constraints;

//data structure to keep track of degrees of freedom in solve
  NonlinearDofHandler nonlinear_dof_handler(tess,fe,mapping,
    hanging_node_constraints);
  int n_dofs=nonlinear_dof_handler.numberDofs();
  cout << "   Number of degrees of freedom: " << n_dofs << endl;

//class to solve nonlinear system
  Vector *deformation=OPERATOR_NEW Vector(n_dofs,0.);
  NonlinearSolver<NonlinearDofHandler> nonlinear_solver(
    nonlinear_dof_handler,&NonlinearDofHandler::nonlinearSystemEvaluation,
    *deformation);

//linear iteration control
  ReductionControl linear_system_control(n_dofs,sqrt(sqrt(DBL_EPSILON)),
    false,false);
//since we use finite differences to approximate the jacobian, 
//  we shouldnt set the relative error tolerance smaller than sqrt(epsilon)
  RelativeErrorControl nonlinear_system_control(max(30,30*n_dofs),
    sqrt(DBL_EPSILON),true,true);
  nonlinear_system_control.setRhsNorm(f_size*sqrt(n_dofs));

//solve nonlinear system
  nonlinear_solver.krylovSubspaceIteration(linear_system_control,
    nonlinear_system_control);
//nonlinear_solver.minpackHybridIteration(nonlinear_system_control);

//evaluate hanging nodes
  hanging_node_constraints.distribute(*deformation);

//plot deformation
  int n_subdivisions=2;
#if (SPACEDIM==1)
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool plot_gt("deformation plot","x","y",0.,1.,0.,1.,&cmap,0,0.5);
  DataOut data_out(nonlinear_dof_handler,mapping);
  data_out.addDataVector (*deformation,"deformation");
  data_out.buildPatches(n_subdivisions);
  data_out.plot("deformation",&plot_gt);
#elif (SPACEDIM==2)
  NumPtr<Point<SPACEDIM> > p;
  Quadrilateral::staticEquallySpacedLatticePoints(n_subdivisions,p);
  int fps=Quadrilateral::facesPerShape();
  int vps=Quadrilateral::verticesPerShape();
  NumPtr<NumPtr<int> > face_indices(fps); 
  int np=0;
  for (int f=0;f<fps;f++) {
    Quadrilateral::staticEquallySpacedLatticePointIndicesOfFace(f,
      n_subdivisions,face_indices[f]);
    np+=face_indices[f].getNumber();
  }
  NumPtr<Point<SPACEDIM> > qp(np);
  int q=0;
  for (int f=0;f<fps;f++) {
    for (int qf=0;qf<face_indices[f].getNumber();qf++,q++) {
      qp[q]=p[face_indices[f][qf]];
    }
  }
  Quadrature<SPACEDIM> face_quad(qp);
  map<int,Quadrature<SPACEDIM> > face_quad_map;
  face_quad_map[vps]=face_quad;
  NumPtr<Point<SPACEDIM> > lagrangian_points(qp.getNumber());
  REAL rlo[SPACEDIM]={D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL)};
  REAL rhi[SPACEDIM]={D_DECL(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL)};
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *present_level=tess.presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    mapping.elementPoints(element,tess,present_level,qp,lagrangian_points);
    FEValues fev(mapping,fe,face_quad_map,UPDATE_VALUES | UPDATE_Q_POINTS);
    fev.mapShapes(tess,element,present_level);
    NumPtr<Vector> displacement;
    nonlinear_dof_handler.getFunctionValues(fev,element,present_level,
      *deformation,displacement);
    q=0;
    for (int f=0;f<fps;f++) {
      for (int qf=0;qf<face_indices[f].getNumber();qf++,q++) {
        const Point<SPACEDIM> &lpq=lagrangian_points[q];
        const Vector &dq=displacement[q];
        for (int d=0;d<SPACEDIM;d++) {
          REAL eulerian_point=lpq[d]+dq[d];
          rlo[d]=min(rlo[d],eulerian_point);
          rhi[d]=max(rhi[d],eulerian_point);
        }
      }
    }
  }
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  AreaGraphTool gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *present_level=tess.presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    mapping.elementPoints(element,tess,present_level,qp,lagrangian_points);
    FEValues fev(mapping,fe,face_quad_map,UPDATE_VALUES | UPDATE_Q_POINTS);
    fev.mapShapes(tess,element,present_level);
    NumPtr<Vector> displacement;
    nonlinear_dof_handler.getFunctionValues(fev,element,present_level,
      *deformation,displacement);
    q=0;
    for (int f=0;f<Quadrilateral::facesPerShape();f++) {
      Point<SPACEDIM> eulerian_point=lagrangian_points[q];
      const Vector &dq=displacement[q];
      for (int d=0;d<SPACEDIM;d++) eulerian_point[d]+=dq[d];
      gt.movePen(eulerian_point[0],eulerian_point[1]);
      q++;
      for (int qf=1;qf<face_indices[f].getNumber();qf++,q++) {
        Point<SPACEDIM> eulerian_point=lagrangian_points[q];
        const Vector &dq=displacement[q];
        for (int d=0;d<SPACEDIM;d++) eulerian_point[d]+=dq[d];
        gt.drawLine(eulerian_point[0],eulerian_point[1]);
      }
    }
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  for (int i=0;i<face_indices.getNumber();i++) face_indices[i].cleanup();
#else
  NumPtr<Point<SPACEDIM> > p;
  Hexahedron::staticEquallySpacedLatticePoints(n_subdivisions,p);
  int lps=Hexahedron::linesPerShape();
  int vps=Hexahedron::verticesPerShape();
  NumPtr<NumPtr<int> > line_indices(lps); 
  int np=0;
  for (int l=0;l<lps;l++) {
    Hexahedron::staticEquallySpacedLatticePointIndicesOfLine(l,
      n_subdivisions,line_indices[l]);
    np+=line_indices[l].getNumber();
  }
  NumPtr<Point<SPACEDIM> > qp(np);
  int q=0;
  for (int l=0;l<lps;l++) {
    for (int ql=0;ql<line_indices[l].getNumber();ql++,q++) {
      qp[q]=p[line_indices[l][ql]];
    }
  }
  Quadrature<SPACEDIM> line_quad(qp);
  map<int,Quadrature<SPACEDIM> > line_quad_map;
  line_quad_map[vps]=line_quad;
  NumPtr<Point<SPACEDIM> > lagrangian_points(qp.getNumber());
  REAL rlo[SPACEDIM]={HUGE_VAL,HUGE_VAL,HUGE_VAL};
  REAL rhi[SPACEDIM]={-HUGE_VAL,-HUGE_VAL,-HUGE_VAL};
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *present_level=tess.presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    mapping.elementPoints(element,tess,present_level,qp,
      lagrangian_points);
    FEValues fev(mapping,fe,line_quad_map,
      UPDATE_VALUES | UPDATE_Q_POINTS);
    fev.mapShapes(tess,element,present_level);
    NumPtr<Vector> displacement;
    nonlinear_dof_handler.getFunctionValues(fev,element,present_level,*deformation,
      displacement);
    q=0;
    for (int l=0;l<lps;l++) {
      for (int ql=0;ql<line_indices[l].getNumber();ql++,q++) {
        const Point<SPACEDIM> &lpq=lagrangian_points[q];
        const Vector &dq=displacement[q];
        for (int d=0;d<SPACEDIM;d++) {
          REAL eulerian_point=lpq[d]+dq[d];
          rlo[d]=min(rlo[d],eulerian_point);
          rhi[d]=max(rhi[d],eulerian_point);
        }
      }
    }
  }
  Palette pal;
  BOOLEAN use_lighting=FALSE;
  Vector3 vlow(rlo[0],rlo[1],rlo[2]);
  Vector3 vhigh(rhi[0],rhi[1],rhi[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;
  VolGraphTool gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&gt);
    gt.drawBox(vlow,vhigh);
  }
  gt.setfgColor("blue");
  for (BOOLEAN do_best=FALSE;;do_best=TRUE) {
    VolGraphTool::DrawDelimiter dd(do_best,&gt);
    for (bool doit=tess.firstActiveShape();doit;
    doit=tess.nextActiveShape()) {
      LEVEL_CLASS *present_level=tess.presentLevel();
      const SHAPE_CLASS *element=present_level->getShape();
      mapping.elementPoints(element,tess,present_level,qp,
        lagrangian_points);
      FEValues fev(mapping,fe,line_quad_map,
        UPDATE_VALUES | UPDATE_Q_POINTS);
      fev.mapShapes(tess,element,present_level);
      NumPtr<Vector> displacement;
      nonlinear_dof_handler.getFunctionValues(fev,element,present_level,
        *deformation,displacement);
      q=0;
      for (int l=0;l<lps;l++) {
        int ql=0;
        const Point<SPACEDIM> &lpq=lagrangian_points[q];
        const Vector &dq=displacement[q];
        Vector3 v0;
        for (int d=0;d<SPACEDIM;d++) v0[d]=lpq[d]+dq[d];
        for (ql++,q++;ql<line_indices[l].getNumber();ql++,q++) {
          const Point<SPACEDIM> &lpq=lagrangian_points[q];
          const Vector &dq=displacement[q];
          Vector3 v1;
          for (int d=0;d<SPACEDIM;d++) v1[d]=lpq[d]+dq[d];
          gt.drawLine(v0,v1);
          v0=v1;
        }
      }
    }
    if (do_best==TRUE) break;
  }
  gt.expose();
  gt.eventLoop();
#endif

  delete boundary_data; boundary_data=0;
  delete deformation; deformation=0;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int,
  pair<int const, double>,
  _Select1st<pair<int const, double> >,
  less<int>,
  allocator<pair<int const, double> >
>;
#include "NonlinearSolver.C"
template class NonlinearSolver<NonlinearDofHandler>;
#include "Solver.C"
template SOLVER_STATE GMRESSolver::daskrSolve<
  NonlinearSolver<NonlinearDofHandler> 
>(
  NonlinearSolver<NonlinearDofHandler> const&, 
  void (NonlinearSolver<NonlinearDofHandler>::*)(
   Vector&, 
   Vector const&
  ) const, 
  Vector&, 
  Vector const&, 
  SolverPreconditioner const&
);
template SOLVER_STATE GMRESSolver::templateSolve<
  NonlinearSolver<NonlinearDofHandler> 
>(
  NonlinearSolver<NonlinearDofHandler> const&, 
  void (NonlinearSolver<NonlinearDofHandler>::*)(
   Vector&, 
   Vector const&
  ) const, 
  Vector&, 
  Vector const&, 
  SolverPreconditioner const&
);
