// "$Header: /home/faculty/johnt/cvs/deal_new/examples/InhomogeneousLaplaceProblem.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"

/* $Id: step-4.cc,v 1.22 2003/01/09 20:32:45 wolf Exp $ */
/* Author: Wolfgang Bangerth, University of Heidelberg, 1999 */

/*    $Id: step-4.cc,v 1.22 2003/01/09 20:32:45 wolf Exp $       */
/*    Version: $Name: Version-4-0-0 $                                              */
/*                                                                */
/*    Copyright (C) 1999, 2000, 2001, 2002, 2003 by the deal.II authors */
/*                                                                */
/*    This file is subject to QPL and may not be  distributed     */
/*    without copyright and license information. Please refer     */
/*    to the file deal.II/doc/license.html for the  text  and     */
/*    further information on this license.                        */

//modified from deal.II/examples/step-4/step-4.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <iostream>
#include <map>

#include "Boundary.H"
#include <BoundaryCondition.H>
#include <DataOut.H>
#include <DofHandler.H>
#include <FiniteElement.H>
#include <GenerateTess.H>
#include <Mapping.H>
#include <MemoryDebugger.H>
#include <Multigrid.H>
#include <Precondition.H>
#include "SetTraps.H"
#include <Tracer.H>
#include <Tessellation.H>

//problem is
//  - div ( 1 grad u ) = - 4 in domain
//  u(x) = x . x             on boundary of domain
//solution is u(x) = || x ||^2

//class to describe essential boundary condition
class DirichletFunctionSystem : public FunctionSystem {
  protected:
    explicit DirichletFunctionSystem(const DirichletFunctionSystem &fs) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    DirichletFunctionSystem() : FunctionSystem() {};
    virtual ~DirichletFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return p.square();
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &return_value) const
    {
      return_value=p.square();
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      Tensor<1,SPACEDIM> return_gradient(p*2.);
      return return_gradient;
    }
    void gradient(NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
    const Point<SPACEDIM> &p) const {
      if (return_gradients.getNumber()!=n_components) {
        return_gradients.cleanup();
        return_gradients.allocate(n_components);
      }
      for (int c=0;c<n_components;c++) return_gradients[c]=gradient(p,c);
    }
};

REAL rightHandSideValue(const Point<SPACEDIM>&) { 
  return D_CHOOSE(2.,4.,6.); 
}

int main (int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  Tessellation tess;
  cout << "Solving problem in " << SPACEDIM 
            << " space dimensions." << endl;

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> low(D_DECL(-1.,-1.,-1.));
  Point<SPACEDIM> edge[SPACEDIM];
  edge[0]=Point<SPACEDIM>(D_DECL(2.,0.,0.));
#if (SPACEDIM>1)
  edge[1]=Point<SPACEDIM>(D_DECL(0.,2.,0.));
#endif
#if (SPACEDIM==3)
  edge[2]=Point<SPACEDIM>(D_DECL(0.,0.,2.));
#endif
  int number_elements[SPACEDIM]={D_DECL(1,1,1)};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);

  LaplaceMaterialModel lmm(rightHandSideValue);
  tess.setMaterialModel(&lmm); // same model in all elements

  DirichletFunctionSystem dfs;
  for (int i=0;i<2*SPACEDIM;i++) {
    boundary_condition[i]->setFunctionSystem(&dfs);
    boundary_condition[i]->makeEssential();
  }

//refine four times everywhere
  tess.refineGlobal(4);
#ifdef DEBUG
  cout << "   Number of active cells: "
       << tess.numberActiveShapes() << endl
       << "   Total number of cells: " << tess.numberShapes()
       << endl;
#endif

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;

//Lagrange polynomials to represent solution of Laplace equation
  int degree=1;
  FE_L fe(degree);

//Gaussian quadrature
  GaussianQuadrature<SPACEDIM> quadrature_formula(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(
    D_CHOOSE(2*degree-1,2*degree+1,2*degree+1)));
  map<int,Quadrature<SPACEDIM> > qmap;
  int nv=D_CHOOSE(Interval::verticesPerShape(),
                  Quadrilateral::verticesPerShape(),
                  Hexahedron::verticesPerShape());
  qmap[nv]=quadrature_formula;

//data structure to keep track of degrees of freedom in solver
  MGDofHandler mg_dof_handler(tess,fe);
  mg_dof_handler.distributeDofs();
  int n_dofs=mg_dof_handler.numberDofs();
  cout << "   Number of degrees of freedom: " << n_dofs << endl;

//data structure to keep track of linear system storage
  SparsityPattern *sparsity_pattern=OPERATOR_NEW SparsityPattern(n_dofs,
    n_dofs,mg_dof_handler.maxCouplingsBetweenDofs());
  mg_dof_handler.DofHandler::makeSparsityPattern(*sparsity_pattern);
  sparsity_pattern->compress();

//storage for stiffness matrix, right-hand side and solution
  SparseMatrix system_matrix(sparsity_pattern);
  Vector system_rhs(n_dofs,0.);
  Vector solution(n_dofs,0.);

//assemble linear system
//  this will call simpleMaterialModel::elementStiffnessMatrixAndRHS
  map<int,Quadrature<SPACEDIM-1> > nqmap; // quadrature rule for Neumann bc
#if (SPACEDIM==2)
  nqmap[Interval::verticesPerShape()]=
    GaussianQuadrature<1>(GaussianQuadrature<1>::argGivenOrder(2*degree+1));
#elif (SPACEDIM==3)
  nqmap[Quadrilateral::verticesPerShape()]=
    GaussianQuadrature<2>(GaussianQuadrature<2>::argGivenOrder(2*degree+1));
#endif
  mg_dof_handler.assembleLinearSystem(mapping,qmap,nqmap,system_matrix,
    system_rhs);

//compute essential boundary values at mapped equally spaced points on bdry 
//calls FunctionSystem in BoundaryCondition's to set boundary values
  map<int,REAL> boundary_values; // output boundary values
  mg_dof_handler.interpolateBoundaryValues(mapping,boundary_values);

//modify linear system and right-hand side for essential boundary conditions
  bool preserve_symmetry=true;
  mg_dof_handler.applyBoundaryValues(boundary_values,system_matrix,
    system_rhs,preserve_symmetry);
  mg_dof_handler.applyBoundaryValues(boundary_values,solution);

//conjugate gradient solver for coarsest level in multigrid
//  cg_solver_control max interations and tolerance will be reset in
//    the Multigrid constructo
  SolverControl solver_control(solution.size(),DBL_EPSILON*1.e4,
    false,false);
  CGSolver cg(solver_control);
  PreconditionIdentity pi;

//multigrid solver
  SolverControl
    mg_solver_control(solution.size(),DBL_EPSILON*1.e4,false,true);
  MGTransfer transfer(mg_dof_handler);
  PreconditionSOR smoother; // omega=1 ==> Gauss-Seidel
  NumPtr<set<int> > dirichlet_boundary_dofs;
  mg_dof_handler.findDirichletBoundaryDofs(mapping,dirichlet_boundary_dofs);
  Multigrid mg(mg_dof_handler,system_matrix,solution,transfer,smoother,
    cg,pi,dirichlet_boundary_dofs,0);

//solve linear system
  mg.solve(mg_solver_control,system_matrix,system_rhs,solution);

  Palette pal;
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool plot_gt("solution plot","x","y",0.,1.,0.,1.,&cmap,0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
  AreaGraphTool contour_gt("solution contours","x0","x1",rlo,rhi,&cmap,
    0,0.5);
  AreaGraphTool plot_gt("solution plot","x0","y1",rlo,rhi,&cmap,0,0.5);
  BOOLEAN use_lighting=FALSE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
    0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
    surface_gt.drawBox(vlow,vhigh);
  }
#else
  BOOLEAN use_lighting=FALSE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool plot_gt("solution plot",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&plot_gt);
    plot_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
    0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
    surface_gt.drawBox(vlow,vhigh);
  }
#endif

  DataOut data_out(mg_dof_handler,mapping);
  data_out.addDataVector (solution,"solution");
  int n_subdivisions=2;
  data_out.buildPatches(n_subdivisions);
  bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>=2)
  data_out.drawGrid(&mesh_gt,quit_button);
#endif
#if (SPACEDIM==2)
  data_out.contourPlot("solution",&contour_gt,quit_button);
#endif
  data_out.plot("solution",&plot_gt,quit_button);
#if (SPACEDIM>=2)
  data_out.plotSurface("solution",&surface_gt);
#endif
#if (SPACEDIM==1)
  if (!quit_button) {
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#elif (SPACEDIM==2)
  if (!quit_button) {
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
  
  delete sparsity_pattern;
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
