// "$Header: /home/faculty/johnt/cvs/deal_new/examples/LinearElements1D.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "DataOut.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM!=1)
barf // hard-wired for 1D
#endif

//-u''+u=0, u(-1)=-sinh(1), u(1)=sinh(1) ==> u(x)=sinh(x)
int degree=1;
int nelements=10;

//description of pde right-hand side and exact solution (to compute errors)
REAL rightHandSideValue(const Point<1> &p) { return 0.; }
REAL exactSolutionValue(const Point<1> &p) { return sinh(p[0]); }
Tensor<1,1> exactSolutionGradient(const Point<1> &p) {
  Tensor<1,1> return_val;
  return_val[0]=cosh(p[0]);
  return return_val;
}

//class to describe essential boundary condition
class GivenFunctionSystem : public FunctionSystem {
  protected:
    explicit GivenFunctionSystem(const BoundaryCondition &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    GivenFunctionSystem() {;}
    virtual ~GivenFunctionSystem() {;}
    virtual REAL value(const Point<1> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<1> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,1> gradient(const Point<1> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,1> >&,
    const Point<1>&) const {
      OBSOLETE("not programmed");
    }
};
class SimpleMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<1> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,1> sgi=sg(i,q);
          sgi*=jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
//          element-wise contribution to stiffness matrix is
//            integral grad v_i . grad v_j + v_i * v_j
//          ie, pde is -div ( 1 grad u ) + 1 u = ...
            cell_matrix(i,j)+=sgi*sg(j,q);
            cell_matrix(i,j)+=si*sv(j,q);
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << "\tsinh solution" << endl;
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

//rectangular domain consisting of nelements intervals
  Point<1> low(-1.);
  Point<1> edge[1]; edge[0]=Point<1>(2.);
  int number_elements[1]={nelements};
  bool set_face_ids=true;
  NumPtr<Boundary*> bs;
  NumPtr<BoundaryCondition*> bcs;
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,bs,bcs);

  SimpleMaterialModel sm;
  tess.setMaterialModel(&sm);

//boundary conditions
  GivenFunctionSystem gfs;
  for (int i=0;i<bcs.getNumber();i++) { // Dirichlet boundary
    BoundaryCondition *bc=bcs[i];
    bc->setFunctionSystem(&gfs);
    bc->makeEssential();
  }

//linear coordinate mapping
//  mapping shape functions are computed on canonical elements now
  MappingL mapping;

//Lagrange polynomials to represent solution of problem
//  finite element shape functions are computed on canonical elements now
  FE_L fe(degree);

//Gaussian quadrature
  map<int,Quadrature<1> > qmap;
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree-1));

//quadrature rule on essential boundaries (not important in 1d)
  map<int,Quadrature<0> > dqmap; // essential bc's

//data structure to keep track of degrees of freedom in solver
  DofHandler dof_handler(tess,fe);
  dof_handler.distributeDofs();
  int n_dofs=dof_handler.numberDofs();
  cout << "\n\n\n\tNumber of degrees of freedom = " << n_dofs << endl;

//data structure to keep track of linear system storage
  SparsityPattern *sparsity_pattern=
    OPERATOR_NEW SparsityPattern(n_dofs,n_dofs,
    dof_handler.maxCouplingsBetweenDofs());
  dof_handler.makeSparsityPattern(*sparsity_pattern);
  sparsity_pattern->compress();

//storage for stiffness matrix, right-hand side and solution
  SparseMatrix system_matrix(sparsity_pattern);
  Vector system_rhs(n_dofs,0.);
  Vector solution(n_dofs,0.);

//assemble linear system
//  this will call simpleMaterialModel::elementStiffnessMatrixAndRHS
//  and would call simpleMaterialModel::processBoundaryGradient
//    if there were any Neumann boundary conditions
  map<int,Quadrature<0> > fqmap;
  dof_handler.assembleLinearSystem(mapping,qmap,fqmap,
    system_matrix,system_rhs);

//compute essential boundary values
  map<int,REAL> boundary_values;
  dof_handler.interpolateBoundaryValues(mapping,boundary_values,dqmap);

//modify linear system and right-hand side for essential boundary conditions
  bool preserve_symmetry=true;
  dof_handler.applyBoundaryValues(boundary_values,system_matrix,
    system_rhs,preserve_symmetry);
  dof_handler.applyBoundaryValues(boundary_values,solution);

//conjugate gradient solver
  SolverControl 
    cg_solver_control(solution.size(),DBL_EPSILON*1.e2,false,false);
  CGSolver cg(cg_solver_control);
  PreconditionIdentity pi;
  cg.solve(system_matrix,solution,system_rhs,pi);

//compute various error measures
  int n_active_cells=tess.numberActiveShapes();
  FunctionSystem *weight=0;
  REAL error=dof_handler.integrateDifference(mapping,solution,gfs,
    qmap,L2_NORM,weight,1.);
#ifdef DEBUG
  cout << "\tL2_error = " << error << endl;
#endif
    
  error=dof_handler.integrateDifference(mapping,solution,gfs,
    qmap,H1_NORM,weight,1.);
#ifdef DEBUG
  cout << "\tH1_error = " << error << endl;
#endif
 
  error=dof_handler.integrateDifference(mapping,solution,gfs,
    qmap,LINFTY_NORM,weight,1.);
#ifdef DEBUG
  cout << "\tLinfinity_error = " << error << endl;
#endif

  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool plot_gt("solution plot","x","y",0.,1.,0.,1.,&cmap,0,0.5);
//plot solution
  DataOut data_out(dof_handler,mapping);
  data_out.addDataVector(solution,"solution");
  int n_subdivisions=640/nelements;
  data_out.buildPatches(n_subdivisions);
  bool quit_button=false;
  data_out.plot("solution",&plot_gt,quit_button);
  if (!quit_button) {
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }

  delete sparsity_pattern;
  delete boundary_data;
  for (int i=0;i<bcs.getNumber();i++) { delete bcs[i]; bcs[i]=0; }
  for (int i=0;i<bs.getNumber();i++) { delete bs[i]; bs[i]=0; }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
