// "$Header:$"
#include "MooneyRivlin.H"

//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//  must have SPACEDIM <= n_components <= 3
Tensor<2,3> MooneyRivlinMaterialModel::greenDeformationTensor(
const NumPtr<Tensor<1,SPACEDIM> > &displacement_gradient) const {
//displacement_gradient is nc x SPACEDIM
//TRACER_CALL(t,"MooneyRivlinMaterialModel::greenDeformationTensor");
#ifdef DEBUG
//for (int c=0;c<displacement_gradient.getNumber();c++) {
//  cout << "\tdisplacement_gradient[" << c << "][*] = ";
//  for (int d=0;d<SPACEDIM;d++) {
//    cout << displacement_gradient[c][d] << " ";
//  }
//  cout << endl;
//}
#endif
  Tensor<2,3> green_deformation_tensor; // 3 x 3
  int n_components=displacement_gradient.getNumber();
  green_deformation_tensor.initialize(0.);
  for (int j=0;j<n_components;j++) {
    const Tensor<1,SPACEDIM> &dgj=displacement_gradient[j];
    for (int i=j;i<n_components;i++) {
      const Tensor<1,SPACEDIM> &dgi=displacement_gradient[j];
      green_deformation_tensor[i][j]=dgi*dgj+dgj[i]+dgi[j];
    }
    green_deformation_tensor[j][j]+=1.;
    for (int i=0;i<j;i++) {
      green_deformation_tensor[i][j]=green_deformation_tensor[j][i];
    }
  }
  for (int j=n_components;j<3;j++) green_deformation_tensor[j][j]=1.;
#ifdef DEBUG
//for (int i=0;i<3;i++) {
//  cout << "\tgreen_deformation_tensor[" << i << "][*] = ";
//  for (int j=0;j<3;j++) {
//    cout << green_deformation_tensor[i][j] << " ";
//  }
//  cout << endl;
//}
#endif
  return green_deformation_tensor;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MooneyRivlinMaterialModel::freeEnergy(
const Tensor<2,3> &green_deformation_tensor,NumPtr<REAL> &values) const {
//TRACER_CALL(t,"MooneyRivlinMaterialModel::freeEnergy");
//Mooney-Rivlin model
  REAL det=determinant(green_deformation_tensor);
  REAL log_det=log(det);
  values.allocate(4);
//free energy
  values[0]=0.25*lame*pow(log_det,2)
           +shear*(trace(green_deformation_tensor)-log_det);
//free energy derivatives wrt invariants
  values[1]=shear; // derivative wrt trace
  values[2]=0.; // derivative wrt second invariant
  values[3]=(0.5*lame*log_det-shear)/det; // derivative wrt det
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Tensor<2,3> MooneyRivlinMaterialModel::stressTensor(
const NumPtr<Tensor<1,SPACEDIM> > &displacement_gradient) const {
//TRACER_CALL(t,"MooneyRivlinMaterialModel::stressTensor");
  Tensor<2,3> green_deformation_tensor=
    greenDeformationTensor(displacement_gradient);
  NumPtr<REAL> values;
  freeEnergy(green_deformation_tensor,values);
  Tensor<2,3> trace_derivative=traceDerivative(green_deformation_tensor);
  trace_derivative*=values[1];
  Tensor<2,3> determinant_derivative=
    determinantDerivative(green_deformation_tensor);
  determinant_derivative*=values[3];
  Tensor<2,3> second_piola_kirchhoff_stress_tensor=
    trace_derivative+determinant_derivative;
  Tensor<2,3> first_piola_kirchhoff_stress_tensor=
    second_piola_kirchhoff_stress_tensor;
  int n_components=displacement_gradient.getNumber();
  for (int i=0;i<n_components;i++) {
    for (int j=0;j<SPACEDIM;j++) {
      for (int k=0;k<SPACEDIM;k++) {
        first_piola_kirchhoff_stress_tensor[i][j]+=
          displacement_gradient[i][k]*
          second_piola_kirchhoff_stress_tensor[k][j];
      }
    }
  }
  return first_piola_kirchhoff_stress_tensor;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MooneyRivlinMaterialModel::elementResidual(
const NumPtr<Point<SPACEDIM> > &q_points,
const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
const Array<2,Tensor<1,SPACEDIM> > &sg,
const NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &input_gradients,
Vector &cell_output) const {
//TRACER_CALL(t,"NonlinearMechanicsSystem::forceBalance");
//if (!checkElementVolumes(displacement)) return false;
#ifdef DEBUG
//for (int i=0;i<displacement.size();i++) {
//  cout << "\tdisplacement[" << i << "] = " << displacement[i] << endl;
//}
#endif
  int dofs_per_cell=cell_output.size();
  int shapes_per_cell=sv.size(0);
//input_gradients[q][c][d], 
//  0<=q<=nqp, 0<=c<component,s 0<=d<SPACEDIM
  for (int q=0;q<q_points.getNumber();q++) {
    REAL jxw=jxw_values[q];
    Point<SPACEDIM> rhsv(bodyForce(q_points[q]));
    rhsv*=jxw;
    const NumPtr<Tensor<1,SPACEDIM> > &displacement_gradient=
      input_gradients[q];
    Tensor<2,3> first_piola_kirchhoff_stress=
      stressTensor(displacement_gradient);
    for (int i=0;i<shapes_per_cell;i++) {
      Tensor<1,SPACEDIM> sgi=sg(i,q);
      sgi*=jxw;
      int ic=i*SPACEDIM;
      for (int ci=0;ci<SPACEDIM;ci++) {
        for (int cj=0;cj<SPACEDIM;cj++) {
//        restoring force
          cell_output[ic+ci]-=
            sgi[cj]*first_piola_kirchhoff_stress[cj][ci];
        }
//      body force
        cell_output[ic+ci]+=sv(i,q)*rhsv[ci];
      }
    }
  }
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
bool MooneyRivlinMaterialModel::checkElementVolumes(
DofHandler &dof_handler,const Vector &global_displacement) const {
//TRACER_CALL(t,"NonlinearMechanicsSystem::checkElementVolumes");
  int nv=1<<SPACEDIM;
  Tessellation &tess=dof_handler.getTessellation();
  NumPtr<Point<SPACEDIM> > qp(nv);
  for (int v=0;v<nv;v++) qp[v]=canonicalVertex(nv,v);
  Quadrature<SPACEDIM> q(qp);
  map<int,Quadrature<SPACEDIM> > qmap;
  qmap[nv]=q;
  MappingL mapping;
  FEValues fev(mapping,dof_handler.getFe(),qmap,UPDATE_VALUES);
  const NumPtr<Point<SPACEDIM> > &lagrangian_vertices=
    tess.getVertices();
  NumPtr<Point<SPACEDIM> > 
    eulerian_vertices(lagrangian_vertices.getNumber());
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *present_level=tess.presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    fev.mapShapes(tess,element,present_level);
    NumPtr<Vector> local_displacement;
    dof_handler.getFunctionValues(fev,element,present_level,
      global_displacement,local_displacement);
    for (int v=0;v<nv;v++) {
      int vi=element->vertexIndex(v);
      const Point<SPACEDIM> &lv=lagrangian_vertices[vi];
      Point<SPACEDIM> &ev=eulerian_vertices[vi];
      const Vector &dv=local_displacement[v];
      for (int d=0;d<SPACEDIM;d++) ev[d]=lv[d]+dv[d];
    }
    if (element->volume(eulerian_vertices)<=0.) return false;
  }
  return true;
}
#endif
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MooneyRivlinMaterialModel::printOn(ostream &os) const {
  os << "\tMooneyRivlinMaterialModel: lame = " << lame 
     << "\n\tshear = " << shear
     << "\n\tbodyForce  " << bodyForce << endl; 
  NonlinearMaterialModel::printOn(os);
}

//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//must have SPACEDIM <= n_components <= 3
Tensor<2,3> LinearElasticityMaterialModel::infinitesimalStrainTensor(
const NumPtr<Tensor<1,SPACEDIM> > &displacement_gradient) const {
//TRACER_CALL(t,"infinitesimalStrainTensor");
//displacement_gradient is nc x SPACEDIM
#ifdef DEBUG
//for (int c=0;c<displacement_gradient.getNumber();c++) {
//  cout << "\tdisplacement_gradient[" << c << "][*] = ";
//  for (int d=0;d<SPACEDIM;d++) {
//    cout << displacement_gradient[c][d] << " ";
//  }
//  cout << endl;
//}
#endif
  Tensor<2,3> infinitesimal_strain_tensor; // 3 x 3
  int n_components=displacement_gradient.getNumber();
  infinitesimal_strain_tensor.initialize(0.);
  for (int j=0;j<n_components;j++) {
    for (int i=j;i<n_components;i++) {
      infinitesimal_strain_tensor[i][j]=
        0.5*(displacement_gradient[i][j]+displacement_gradient[j][i]);
    }
    for (int i=0;i<j;i++) {
      infinitesimal_strain_tensor[i][j]=
        infinitesimal_strain_tensor[j][i];
    }
  }
#ifdef DEBUG
//for (int i=0;i<3;i++) {
//  cout << "\tinfinitesimal_strain_tensor[" << i << "][*] = ";
//  for (int j=0;j<3;j++) {
//    cout << infinitesimal_strain_tensor[i][j] << " ";
//  }
//  cout << endl;
//}
#endif
  return infinitesimal_strain_tensor;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Tensor<2,3> LinearElasticityMaterialModel::stressTensor(
const NumPtr<Tensor<1,SPACEDIM> > &displacement_gradient) const {
//TRACER_CALL(t,"LinearMechanicsSystem::stressTensor");
#ifdef DEBUG
//for (int c=0;c<displacement_gradient.getNumber();c++) {
//  const Tensor<1,SPACEDIM> &dgc=displacement_gradient[c];
//  cout << "\tdisplacement_gradient[" << c << ",*] = ";
//  for (int d=0;d<SPACEDIM;d++) {
//    cout << dgc[d] << " ";
//  }
//  cout << endl;
//}
#endif
  Tensor<2,3> infinitesimal_strain_tensor=
    infinitesimalStrainTensor(displacement_gradient);
  REAL trace_strain=lame*trace(infinitesimal_strain_tensor);
  infinitesimal_strain_tensor*=2.*shear;
#ifdef DEBUG
//for (int i=0;i<3;i++) {
//  cout << "\tinfinitesimal_strain_tensor[" << i << ",*] = ";
//  for (int j=0;j<3;j++) {
//    cout << infinitesimal_strain_tensor[i][j] << " ";
//  }
//  cout << endl;
//}
#endif
  Tensor<2,3> stress_tensor=infinitesimal_strain_tensor;
  for (int i=0;i<3;i++) stress_tensor[i][i]+=trace_strain;
#ifdef DEBUG
//cout << "\tstress_tensor" << endl;
//for (int i=0;i<3;i++) {
//  cout << "\t\t";
//  for (int j=0;j<3;j++) {
//    cout << stress_tensor[i][j] << " ";
//  }
//  cout << endl;
//}
#endif
  return stress_tensor;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void LinearElasticityMaterialModel::elementResidual(
const NumPtr<Point<SPACEDIM> > &q_points,
const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
const Array<2,Tensor<1,SPACEDIM> > &sg,
const NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &input_gradients,
Vector &cell_output) const {
  int dofs_per_cell=cell_output.size();
  int shapes_per_cell=sv.size(0);
//input_gradients[q][c][d], 
//  0<=q<=nqp, 0<=c<component,s 0<=d<SPACEDIM
  for (int q=0;q<q_points.getNumber();q++) {
    REAL jxw=jxw_values[q];
    Point<SPACEDIM> rhsv(bodyForce(q_points[q]));
    rhsv*=jxw;
#ifdef DEBUG
//  cout << "\tq = " << q << endl;
//  cout << "\trhsv[" << q << "] = ";
//  for (int i=0;i<rhsv.size();i++) cout << rhsv[i] << " ";
//  cout << endl;
#endif
    const NumPtr<Tensor<1,SPACEDIM> > &displacement_gradient=
      input_gradients[q];
#ifdef DEBUG
//  for (int c=0;c<displacement_gradient.getNumber();c++) {
//    const Tensor<1,SPACEDIM> &dgqc=displacement_gradient[c];
//    cout << "\tdisplacement_gradient[" << q << "," << c << ",*] = ";
//    for (int d=0;d<SPACEDIM;d++) {
//      cout << dgqc[d] << " ";
//    }
//    cout << endl;
//  }
#endif
    Tensor<2,3> stress_tensor=stressTensor(displacement_gradient);
#ifdef DEBUG
//  cout << "\tstress_tensor[" << q << "] = " << endl;
//  for (int i=0;i<3;i++) {
//    cout << "\t\t";
//    for (int j=0;j<3;j++) {
//      cout << stress_tensor[i][j] << " ";
//    }
//    cout << endl;
//  }
#endif
    for (int i=0;i<shapes_per_cell;i++) {
      Tensor<1,SPACEDIM> sgi=sg(i,q);
      sgi*=jxw;
      int ic=i*SPACEDIM;
#ifdef DEBUG
//    cout << "\tsv = " << sv(i,q) << endl;
//    cout << "\tsgi = ";
//    for (int ci=0;ci<SPACEDIM;ci++) cout << sgi[ci] << " ";
//    cout << endl;
#endif
      for (int ci=0;ci<SPACEDIM;ci++) {
        for (int cj=0;cj<SPACEDIM;cj++) {
//        restoring force
          cell_output[ic+ci]-=sgi[cj]*stress_tensor[cj][ci];
        }
//      body force
        cell_output[ic+ci]+=sv(i,q)*rhsv[ci];
      }
    }
  }
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool LinearElasticityMaterialModel::checkElementVolumes(
DofHandler &dof_handler,const Vector &global_displacement) const {
//TRACER_CALL(t,"LinearElasticityMaterialModel::checkElementVolumes");
  int nv=1<<SPACEDIM;
  NumPtr<Point<SPACEDIM> > qp(nv);
  for (int v=0;v<nv;v++) qp[v]=canonicalVertex(nv,v);
  Quadrature<SPACEDIM> q(qp);
  map<int,Quadrature<SPACEDIM> > qmap;
  qmap[nv]=q;
  MappingL mapping;
  FEValues fev(mapping,dof_handler.getFe(),qmap,UPDATE_VALUES);
  Tessellation &tess=dof_handler.getTessellation();
  const NumPtr<Point<SPACEDIM> > &lagrangian_vertices=
    tess.getVertices();
  NumPtr<Point<SPACEDIM> > 
    eulerian_vertices(lagrangian_vertices.getNumber());
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *present_level=tess.presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    fev.mapShapes(tess,element,present_level);
    NumPtr<Vector> local_displacement;
    dof_handler.getFunctionValues(fev,element,present_level,
      global_displacement,local_displacement);
    for (int v=0;v<nv;v++) {
      int vi=element->vertexIndex(v);
      const Point<SPACEDIM> &lv=lagrangian_vertices[vi];
      Point<SPACEDIM> &ev=eulerian_vertices[vi];
      const Vector &dv=local_displacement[v];
      for (int d=0;d<SPACEDIM;d++) ev[d]=lv[d]+dv[d];
    }
    if (element->volume(eulerian_vertices)<=0.) return false;
  }
  return true;
}
#ifdef DEBUG
void LinearElasticityMaterialModel::checkLinearity(
NonlinearDofHandler &dof_handler) const {
  TRACER_CALL(t,"LinearElasticityMaterialModel::checkLinearity");
  int n_dofs=dof_handler.numberDofs();
  Vector x(n_dofs,0.);
  Vector random_x(n_dofs);
  drand48();
  for (int i=0;i<n_dofs;i++) random_x[i]=drand48();
  Vector f(n_dofs);
  dof_handler.nonlinearSystemEvaluation(x,f); // f(0)

  Vector random_f(n_dofs);
  dof_handler.nonlinearSystemEvaluation(random_x,random_f);
  random_f-=f;
  random_f.scale(2.); // 2 * [ f(x) - f(0) ]

  Vector random_f2(n_dofs);
  random_x.scale(2.);
  dof_handler.nonlinearSystemEvaluation(random_x,random_f2);
  random_f2-=f; // f( 2 * x ) - f(0)

  for (int i=0;i<n_dofs;i++) {
    REAL error=random_f[i]-random_f2[i];
    REAL sz=max(abs(random_f[i]),abs(random_f2[i]));
    if (sz>0.) error/=sz;
    if (error>DBL_EPSILON*1028.) {
      cout << "\trandom_f[" << i << "] = " << random_f[i] << " "
           << random_f2[i] << " " << error << endl;
    }
  }
}
void LinearElasticityMaterialModel::checkJacobian(
NonlinearDofHandler &dof_handler) const {
  TRACER_CALL(t,"LinearElasticityMaterialModel::checkJacobian");
  int n_dofs=dof_handler.numberDofs();
  Matrix jacobian(n_dofs,n_dofs);
  Vector x(n_dofs,0.);
  Vector f(n_dofs);
  dof_handler.nonlinearSystemEvaluation(x,f); // f(0)

  Vector axis_v(n_dofs);
  Vector jv(n_dofs);
  REAL epsilon=sqrt(DBL_EPSILON);
  for (int j=0;j<n_dofs;j++) {
    axis_v=0.;
    axis_v[j]=epsilon;
    dof_handler.nonlinearSystemEvaluation(axis_v,jv);
    jv-=f;
    jv.scale(1./epsilon);
    for (int i=0;i<n_dofs;i++) jacobian(i,j)=jv[i];
  }

  Vector random_x(n_dofs);
  for (int i=0;i<n_dofs;i++) random_x[i]=drand48();
  jacobian.vmult(jv,random_x); // J * x

  Vector random_f(n_dofs);
  dof_handler.nonlinearSystemEvaluation(random_x,random_f);
  random_f-=f; // f(x) - f(0)

  for (int i=0;i<n_dofs;i++) {
    REAL error=random_f[i]-jv[i];
    REAL sz=max(abs(random_f[i]),abs(jv[i]));
    if (sz>0.) error/=sz;
    if (error>DBL_EPSILON*1028.) {
      cout << "\trandom_f[" << i << "] = " << random_f[i] << " "
           << jv[i] << " " << error << endl;
    }
  }
}
#endif
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void LinearElasticityMaterialModel::printOn(ostream &os) const {
  os << "\tLinearElasticityMaterialModel: lame = " << lame 
     << "\n\tshear = " << shear
     << "\n\tbodyForce  " << bodyForce << endl; 
  NonlinearMaterialModel::printOn(os);
}
