// "$Header: /home/faculty/johnt/cvs/deal_new/examples/LinearElements1D.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "ConditionNumber.H"
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "RefinementStudy.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM!=1)
barf;
#endif

//- u'' - pi^2 u = sin(pi x)
//  u(0) = 0 = u(1)

int degree=1;
int N_CYCLES=7;

REAL rightHandSideValue(const Point<1> &p) { return sin(M_PI*p[0]); }

class NonCoerciveMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<1> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      REAL pi2=M_PI*M_PI;
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,1> sgi=sg(i,q);
          sgi*=jxw;
          REAL si=sv(i,q)*pi2*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
            cell_matrix(i,j)+=sgi*sg(j,q)-si*sv(j,q);
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(128);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
//rectangular domain consisting of nelements intervals
  Point<1> low(0.);
  Point<1> edge[1]; edge[0]=Point<1>(1.);
  int number_elements[1]={10};
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);

  NonCoerciveMaterialModel ncmm;
  tess.setMaterialModel(&ncmm);

  ConstantFunctionSystem cfs(0.);
  for (int i=0;i<boundary_condition.getNumber();i++) {
    boundary_condition[i]->setFunctionSystem(&cfs);
    boundary_condition[i]->makeEssential(); // Dirichlet boundary
  }

//linear coordinate mapping
//  mapping shape functions are computed on canonical elements now
  MappingL mapping;

//Lagrange polynomials to represent solution of problem
//  finite element shape functions are computed on canonical elements now
  FE_L fe(degree);

//Gaussian quadrature
  map<int,Quadrature<SPACEDIM> > qmap;
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree-1));

  map<int,Quadrature<SPACEDIM-1> > fqmap;

  int n_dofs[N_CYCLES];
  NumPtr<Vector> solutions(N_CYCLES);
  NumPtr<DofHandler*> dof_handlers(N_CYCLES);
  NumPtr<Tessellation*> tessellations(N_CYCLES);
  NumPtr<LEVEL_CLASS*> coarsest_levels(N_CYCLES);

  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool *plot_gt=OPERATOR_NEW XYGraphTool("solution","x","error",
    0.,1.,0.,1.,&cmap,0,0.5);
  plot_gt->setbgColor("white");
  plot_gt->setfgColor("black");
  for (int cycle=0;cycle<N_CYCLES;cycle++) {

    LEVEL_CLASS *copy_coarser_level=0;
    for (LEVEL_CLASS *level=tess.coarsestLevel();level;
    level=level->finerLevel()) {
      if (copy_coarser_level==0) {
        copy_coarser_level=coarsest_levels[cycle]=
          OPERATOR_NEW LEVEL_CLASS();
        coarsest_levels[cycle]->copyLevel(*level);
      } else {
        LEVEL_CLASS *cl=OPERATOR_NEW LEVEL_CLASS(copy_coarser_level);
        cl->copyLevel(*level);
        copy_coarser_level=cl;
      }
    }
    tessellations[cycle]=OPERATOR_NEW Tessellation();
    tessellations[cycle]->copyTessellation(tess,coarsest_levels[cycle]);

//  data structure to keep track of degrees of freedom in solver
    dof_handlers[cycle]=OPERATOR_NEW DofHandler(*tessellations[cycle],fe);
    dof_handlers[cycle]->distributeDofs();
    n_dofs[cycle]=dof_handlers[cycle]->numberDofs();
    cout << "\n\tNumber of degrees of freedom = " << n_dofs[cycle] << endl;

//  data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handlers[cycle]->maxCouplingsBetweenDofs());
    dof_handlers[cycle]->makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//  storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

    dof_handlers[cycle]->assembleLinearSystem(mapping,qmap,fqmap,
      system_matrix,system_rhs);

//  compute essential boundary values
    map<int,REAL> boundary_values;
    dof_handlers[cycle]->interpolateBoundaryValues(mapping,boundary_values);

//  modify linear system and rhs for essential boundary conditions
    solutions[cycle].reinit(n_dofs[cycle]);
    bool preserve_symmetry=true;
    dof_handlers[cycle]->applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handlers[cycle]->applyBoundaryValues(boundary_values,
      solutions[cycle]);
    REAL rcond=symmetricConditionNumber(system_matrix);
    cout << "\tcondition number = " << rcond << endl;

//  bi-conjugate gradient solver
    SolverControl bicg_solver_control(2*solutions[cycle].size(),
      DBL_EPSILON*rcond*1.e2,true,true);
    BicgstabSolver bicg(bicg_solver_control);
    PreconditionIdentity pi;
    bicg.solve(system_matrix,solutions[cycle],system_rhs,pi);

//  plot solution
    REAL ymin=HUGE_VAL;
    REAL ymax=-HUGE_VAL;
    for (int i=0;i<solutions[cycle].size();i++) {
      REAL si=solutions[cycle][i];
      ymin=min(ymin,si);
      ymax=max(ymax,si);
    }
    cout << "\tsolution max,min = " << ymax << " " << ymin << endl;
    plot_gt->setfgColor("black");
    plot_gt->newPage();
    plot_gt->rescale(0.,1.,ymin,ymax);
    plot_gt->drawAxes();
    int nelements=tessellations[cycle]->numberActiveShapes();
    int n_pts=640/nelements+1;
    NumPtr<Point<1> > plot_points(n_pts);
    REAL dx=1./static_cast<REAL>(n_pts-1);
    for (int i=0;i<n_pts;i++) {
      plot_points[i][0]=static_cast<REAL>(i)*dx;
    }
    Quadrature<1> plot_quadrature(plot_points);
    map<int,Quadrature<SPACEDIM> > plot_qmap;
    plot_qmap[Interval::verticesPerShape()]=plot_quadrature;
    FEValues plot_fe_values(mapping,fe,plot_qmap,
      UPDATE_VALUES | UPDATE_Q_POINTS);
    plot_gt->setfgColor("blue");
    for (bool doit=tessellations[cycle]->firstActiveShape();doit;
    doit=tessellations[cycle]->nextActiveShape()) {
      LEVEL_CLASS *present_level=tessellations[cycle]->presentLevel();
      const SHAPE_CLASS *element=present_level->getShape();
      plot_fe_values.mapShapes(*tessellations[cycle],element,present_level);
      const NumPtr<Point<1> > &q_points=
        plot_fe_values.getQuadraturePoints(element);

      NumPtr<REAL> solution_values;
      dof_handlers[cycle]->getFunctionValues(plot_fe_values,element,
        present_level,solutions[cycle],solution_values);
      REAL x=q_points[0][0];
      plot_gt->movePen(x,solution_values[0]);
      for (int i=1;i<q_points.getNumber();i++) {
        x=q_points[i][0];
        plot_gt->drawLine(x,solution_values[i]);
      }
    }
    plot_gt->flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;

    if (cycle<N_CYCLES-1) tess.refineGlobal(1);
    delete sparsity_pattern;
  }
#if (SPACEDIM<=2)
  delete plot_gt; plot_gt=0;
#endif

//
  {
    cout << "\tRefinementStudy:" << endl;

    NORM_TYPE nt=L2_NORM;
    REAL norm_exponent=2.; // needed for lp norms
    RefinementStudy rs(mapping,qmap,dof_handlers,solutions,nt,
      norm_exponent);
  }
//

  for (int i=0;i<tessellations.getNumber();i++) {
    delete tessellations[i]; tessellations[i]=0;
  }
  for (int i=0;i<dof_handlers.getNumber();i++) {
    delete dof_handlers[i]; dof_handlers[i]=0;
  }
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
#include "NumPtr.C"
INSTANTIATE_NUMPTR(Tessellation*);
