// "$Header: /home/faculty/johnt/cvs/deal_new/examples/RectangularDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "ConstraintMatrix.H"
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "KellyErrorEstimator.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM==1)
# define N_CYCLES 7
#elif (SPACEDIM==2)
# define N_CYCLES 4
#else
# define N_CYCLES 3
#endif
int degree=1;
int nelements=2;

// -div grad u = 0 in (-1,1) x (-1,1)
//  u(-1,x_1) = u(1,x_1) = 0 for x_1 in (-1,1)
//  u(x_0,-1) = sin( pi x_0 ) sinh( - pi ) for x_0 in (-1,1)
//  d u / d x_1(x_0,1) = d u / d x_1(x_0,1) = pi sin( pi x_0 ) cosh( pi )
//    for x_0 in (-1,1)
//  ==> u(x_0,x_1) = sin( pi x_0 ) sinh( pi x_1 )

REAL rightHandSideValue(const Point<SPACEDIM> &p) { return 0.; }

REAL exactSolutionValue(const Point<SPACEDIM> &p) {
#if (SPACEDIM==1)
  return sinh(p[0]);
#elif (SPACEDIM==2)
  return sin(M_PI*p[0])*sinh(M_PI*p[1]);
#else
  return sin(M_PI*p[0])*cos(M_PI*p[1])*sinh(M_SQRT2*M_PI*p[2]);
#endif
}
Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &p) {
  Tensor<1,SPACEDIM> return_val;
#if (SPACEDIM==1)
  return_val[0]=cosh(p[0]);
#elif (SPACEDIM==2)
  Point<SPACEDIM> q=p*M_PI;
  return_val[0]=M_PI*cos(q[0])*sinh(q[1]);
  return_val[1]=M_PI*sin(q[0])*cosh(q[1]);
#else
  Point<SPACEDIM> q=p*M_PI;
  REAL s0=sin(q[0]);
  REAL c1=cos(q[1]);
  REAL s2=sinh(M_SQRT2*q[2]);
  return_val[0]=M_PI*cos(q[0])*c1*s2;
  return_val[1]=-M_PI*s0*sin(q[1])*s2;
  return_val[2]=M_SQRT2*M_PI*s0*c1*cosh(M_SQRT2*q[2]);
#endif
  return return_val;
}

class DirichletFunctionSystem : public FunctionSystem {
  protected:
    explicit DirichletFunctionSystem(const FunctionSystem &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    DirichletFunctionSystem() {;}
    virtual ~DirichletFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
};
class NeumannFunctionSystem : public FunctionSystem {
  private:
    int dir;
    bool left_hand;
  protected:
    explicit NeumannFunctionSystem(const NeumannFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    NeumannFunctionSystem(int d,bool lh) : dir(d),left_hand(lh) {}
    virtual ~NeumannFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      return (left_hand ? -g[dir] : g[dir]);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      function_values[0]=(left_hand ? -g[dir] : g[dir]);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
};

class SimpleMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
            cell_matrix(i,j)+=sgi*sg(j,q);
#if (SPACEDIM==1)
            cell_matrix(i,j)+=si*sv(j,q);
#endif
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> low(D_DECL(-1.,-1.,-1.));
  Point<SPACEDIM> edge[SPACEDIM];
    edge[0]=Point<SPACEDIM>(D_DECL(2.,0.,0.));
#if (SPACEDIM>1)
    edge[1]=Point<SPACEDIM>(D_DECL(0.,2.,0.));
#endif
#if (SPACEDIM==3)
    edge[2]=Point<SPACEDIM>(0.,0.,2.);
#endif
  int number_elements[SPACEDIM]={D_DECL(nelements,nelements,nelements)};
  bool set_face_ids=true;
  BoundaryData *boundary_data=0;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  boundary_data=generateParallelpiped(tess,low,edge,number_elements,
    set_face_ids,boundary,boundary_condition);

  SimpleMaterialModel sm;
  tess.setMaterialModel(&sm);

  DirichletFunctionSystem dfs;
  for (int d=0;d<SPACEDIM;d++) {
    boundary_condition[2*d]->setFunctionSystem(&dfs);
    boundary_condition[2*d]->makeEssential(); // Dirichlet boundary
    boundary_condition[2*d+1]->setFunctionSystem(
      OPERATOR_NEW NeumannFunctionSystem(d,false));
    boundary_condition[2*d+1]->makeNatural();
  }

  MappingL mapping;
  FE_L fe(degree);

  map<int,Quadrature<SPACEDIM> > qmap;
#if (SPACEDIM==1)
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree-1));
#elif (SPACEDIM==2)
  qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+1));
#else
  qmap[Hexahedron::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+1));
#endif

//essential bc's
//must have same number of points as face shapes
  map<int,Quadrature<SPACEDIM-1> > dbc_qmap; 
#if (SPACEDIM==2)
  {
    int nv=Interval::verticesPerShape();
    int dpf=fe.dofsPerFace(nv);
    int m=LobattoQuadrature<1>::argGivenNPoints(dpf);
    dbc_qmap[nv]=LobattoQuadrature<SPACEDIM-1>(m);
  }
#elif (SPACEDIM==3)
  {
    int nv=Quadrilateral::verticesPerShape();
    int dpf=fe.dofsPerFace(nv);
    int m=LobattoQuadrature<2>::argGivenNPoints(dpf);
    dbc_qmap[nv]=LobattoQuadrature<2>(m);
  }
#endif

//measure error in boundary conditions
  map<int,Quadrature<SPACEDIM-1> > bc_qmap;
#if (SPACEDIM==2)
  bc_qmap[Interval::verticesPerShape()]=GaussianQuadrature<SPACEDIM-1>(
    GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree));
#endif
#if (SPACEDIM==3)
  bc_qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<SPACEDIM-1>(
    GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
#endif

//natural bc's
//should be exact for polynomials of shape function degree
//  plus exact solution degree-1
//  plus Jacobian degree
  map<int,Quadrature<SPACEDIM-1> > nbc_qmap;
#if (SPACEDIM==2)
  {
    int nv=Interval::verticesPerShape();
    nbc_qmap[nv]=GaussianQuadrature<SPACEDIM-1>(
      GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
  }
#elif (SPACEDIM==3)
  {
    int nv=Quadrilateral::verticesPerShape();
    nbc_qmap[nv]=GaussianQuadrature<SPACEDIM-1>(
      GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
  }
#endif

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];
  double DBC_error[N_CYCLES];
  double NBC_error[N_CYCLES];

  Palette pal;
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
#else
  BOOLEAN use_lighting=TRUE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
#endif

  for (int cycle=0;cycle<N_CYCLES;cycle++) {
#ifdef DEBUG
    cout << "cycle = " << cycle << endl;
#endif

    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();
    n_dofs[cycle]=dof_handler.numberDofs();

#ifdef DEBUG
    cout << "\tnumber of active elements = " 
         << tess.numberActiveShapes() << endl;
    cout << "\tNumber of degrees of freedom = " << n_dofs[cycle]
         << endl;
    cout << "\tassemble linear system..." << endl;
#endif
    ConstraintMatrix hanging_node_constraints;
    hanging_node_constraints.makeHangingNodeConstraints(dof_handler);
    hanging_node_constraints.close();

    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);
//  insert hanging node constraint equations into sparsity pattern
    hanging_node_constraints.condense(*sparsity_pattern);
    sparsity_pattern->compress();

    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

    dof_handler.assembleLinearSystem(mapping,qmap,nbc_qmap,
      system_matrix,system_rhs);

//  replace stiffness equations with hanging node constraint equations
    hanging_node_constraints.condense(system_matrix);
    hanging_node_constraints.condense(system_rhs);

    map<int,REAL> boundary_values;
    {
#ifdef TIMING
//    Timer y(z3);
#endif
      dof_handler.interpolateBoundaryValues(mapping,boundary_values,
        dbc_qmap);
    }

    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);

    {
#ifdef DEBUG
      cout << "\tsolve linear system..." << endl;
#endif
      RelativeErrorControl cg_solver_control(solution.size(),
        DBL_EPSILON*1.e2,false,false);
      CGSolver cg(cg_solver_control);
//    PreconditionIdentity pi;
      PreconditionJacobi pjac(system_matrix);
      cg.solve(system_matrix,solution,system_rhs,pjac);
    }
//  apply hanging node constraints to solution of linear system
    hanging_node_constraints.distribute(solution);
    {
#ifdef DEBUG
      cout << "\testimate error..." << endl;
#endif
      n_active_cells[cycle]=tess.numberActiveShapes();
      FunctionSystem *weight=0;
      L2_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
        qmap,L2_NORM,weight,1.);
#ifdef DEBUG
      cout << "\tL2_error[" << cycle << "] = " << L2_error[cycle] << endl;
#endif
    
      H1_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
        qmap,H1_NORM,weight,1.);
   
      Linfty_error[cycle]=dof_handler.integrateDifference(mapping,solution,
        dfs,qmap,LINFTY_NORM,weight,1.);

      DBC_error[cycle]=dof_handler.integrateDifferenceOnMappedBoundary(
        mapping,solution,dfs,qmap,bc_qmap,L2_NORM,weight,1.);
      NBC_error[cycle]=dof_handler.integrateDifferenceOnMappedBoundary(
        mapping,solution,dfs,qmap,bc_qmap,H1_SEMINORM,weight,1.);
#ifdef DEBUG
      cout << "\tDBC_error[" << cycle << "] = " << DBC_error[cycle] << endl;
      cout << "\tNBC_error[" << cycle << "] = " << NBC_error[cycle] << endl;
#endif
    }

//
    DataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
#if (SPACEDIM==1)
    int n_subdivisions=640/n_active_cells[cycle];
#else
    int n_subdivisions=degree;
#endif
    data_out.buildPatches(n_subdivisions);
    bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>=2)
    data_out.drawGrid(&mesh_gt,quit_button);
#endif
#if (SPACEDIM==2)
//  data_out.contourPlot("solution",&contour_gt,quit_button);
#endif
//  data_out.plot("solution",&plot_gt,quit_button);
#if (SPACEDIM>=2)
//  data_out.plotSurface("solution",&surface_gt);
#endif
#if (SPACEDIM==1)
//  if (!quit_button) {
//    XYGraphTool::WINDOW_TYPE::QuitButton qb;
//  }
#elif (SPACEDIM==2)
    if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif
//
    if (cycle<N_CYCLES-1) {
#ifdef DEBUG
      cout << "\trefine mesh..." << endl;
#endif
      tess.refineGlobal(1);
    }
    delete sparsity_pattern;
  }

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    XYGraphTool nacgt("number active cells vs cycle","cycle",
      "number active cells",0,N_CYCLES-1,n_active_cells[0],
      n_active_cells[N_CYCLES-1],&cmap,0,0.5);
    nacgt.newPage();
    nacgt.setbgColor("white");
    nacgt.setfgColor("black");
    nacgt.drawAxes();
    nacgt.setfgColor("blue");
    nacgt.movePen(0,n_active_cells[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      nacgt.drawLine(cycle,n_active_cells[cycle]);
    }
    nacgt.flush();

    XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
      "number degress of freedom",0,N_CYCLES-1,n_dofs[0],
      n_dofs[N_CYCLES-1],&cmap,0,0.5);
    ndgt.newPage();
    ndgt.setbgColor("white");
    ndgt.setfgColor("black");
    ndgt.drawAxes();
    ndgt.setfgColor("blue");
    ndgt.movePen(0,n_dofs[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ndgt.drawLine(cycle,n_dofs[cycle]);
    }
    ndgt.flush();

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
      DBC_error[cycle]=(DBC_error[cycle]>0. ?  log(DBC_error[cycle]) : 
        log(DBL_EPSILON))/log10;
      NBC_error[cycle]=(NBC_error[cycle]>0. ?  log(NBC_error[cycle]) : 
        log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    ymin=HUGE_VAL;
    ymax=-HUGE_VAL;
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,DBC_error[cycle]);
      ymax=max(ymax,DBC_error[cycle]);
    }
    XYGraphTool DBC_gt("Dirichlet BC Error","log_10(degrees of freedom)",
      "log_10(L2 boundary error)",log_n_dofs[0],log_n_dofs[N_CYCLES-1],
      ymin,ymax,&cmap,0,0.5);
    DBC_gt.setbgColor("white");
    DBC_gt.setfgColor("black");
    DBC_gt.drawAxes();
    DBC_gt.setfgColor("blue");
    DBC_gt.movePen(log_n_dofs[0],DBC_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      DBC_gt.drawLine(log_n_dofs[cycle],DBC_error[cycle]);
    }
    DBC_gt.flush();

    ymin=HUGE_VAL;
    ymax=-HUGE_VAL;
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,NBC_error[cycle]);
      ymax=max(ymax,NBC_error[cycle]);
    }
    XYGraphTool NBC_gt("Neumann BC Error","log_10(degrees of freedom)",
      "log_10(L2 normal deriv error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    NBC_gt.setbgColor("white");
    NBC_gt.setfgColor("black");
    NBC_gt.drawAxes();
    NBC_gt.setfgColor("blue");
    NBC_gt.movePen(log_n_dofs[0],NBC_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      NBC_gt.drawLine(log_n_dofs[cycle],NBC_error[cycle]);
    }
    NBC_gt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    if (boundary_condition[i]->isNatural()) {
      delete boundary_condition[i]->getFunctionSystem();
    }
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
