// "$Header: /home/faculty/johnt/cvs/deal_new/examples/SphericalDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <iostream>

#include "Boundary.H"
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM!=2)
barf // use RectangularDomain
#endif
#define N_CYCLES 5

// - div [ c(x) grad u ] = 1

int degree=1;
REAL jump=M_PI/6.;

REAL pdeCoefficient(const Point<SPACEDIM> &p) {
  REAL r=sqrt(p.square());
  return (r<jump ? 1. : 2.);
}

REAL exactSolutionValue(const Point<SPACEDIM> &p) {
  REAL r=sqrt(p.square());
  if (r<jump) return (jump-r)*(jump+r)*0.25+(1.-jump)*(1.+jump)*0.125;
  return (1.-r)*(1.+r)*0.125;
}

REAL rightHandSideValue(const Point<SPACEDIM> &p) { return 1.; }

Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &p) {
  REAL r=sqrt(p.square());
  Tensor<1,SPACEDIM> return_val(p);
  return_val*=(r<jump ? -0.5 : -0.25);
  return return_val;
}

class DirichletFunctionSystem : public FunctionSystem {
  protected:
    explicit DirichletFunctionSystem(const FunctionSystem &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    DirichletFunctionSystem() {;}
    virtual ~DirichletFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
};

class SimpleMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL coef_value=pdeCoefficient(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=coef_value*jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
//          element-wise contribution to stiffness matrix is
//            integral grad v_i . pdeCoefficient * grad v_j
//          ie, pde is -div ( pdeCoefficient grad u ) = rhs
            cell_matrix(i,j)+=sgi*sg(j,q);
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

  Point<SPACEDIM> center(D_DECL(0.,0.,0.));
  REAL radius=1.;
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=generateBallTriangles(tess,center,radius,
    set_face_ids,boundary,boundary_condition);

  SimpleMaterialModel smm;
  tess.setMaterialModel(&smm);

  DirichletFunctionSystem dfs;
  boundary_condition[0]->setFunctionSystem(&dfs);
  boundary_condition[0]->makeEssential(); // Dirichlet boundary

  MappingL mapping(degree);
  FE_L fe(degree);

  map<int,Quadrature<SPACEDIM> > qmap;
  qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
    TriangleGaussianQuadrature::argGivenOrder(2*degree-1)); 

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

  for (int cycle=0;cycle<N_CYCLES;cycle++) {
    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();
    n_dofs[cycle]=dof_handler.numberDofs();
    cout << "\n\tNumber of degrees of freedom: " << n_dofs[cycle]
         << endl;

    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

//  quadrature rule for Neumann bc; only no. quad pts is used
    map<int,Quadrature<1> > nqmap;
    nqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
      GaussianQuadrature<1>::argGivenOrder(2*degree-1));

    dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix,system_rhs);

    map<int,REAL> boundary_values;
    dof_handler.interpolateBoundaryValues(mapping,boundary_values);

    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);

    SolverControl solver_control(solution.size(),DBL_EPSILON*1.e4,false,
      false);
    CGSolver cg(solver_control);
    PreconditionIdentity pi;
    cg.solve(system_matrix,solution,system_rhs,pi);

    n_active_cells[cycle]=tess.numberActiveShapes();
    FunctionSystem *weight=0;
    L2_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,L2_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tL2_error[" << cycle << "] = " << L2_error[cycle] << endl;
#endif
    
    H1_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,H1_NORM,weight,1.);
   
    Linfty_error[cycle]=dof_handler.integrateDifference(mapping,solution,
      dfs,qmap,LINFTY_NORM,weight,1.);

    if (cycle<N_CYCLES-1) tess.refineGlobal(1);
    delete sparsity_pattern;
  }

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    XYGraphTool nacgt("number active cells vs cycle","cycle",
      "number active cells",0,N_CYCLES-1,n_active_cells[0],
      n_active_cells[N_CYCLES-1],&cmap,0,0.5);
    nacgt.newPage();
    nacgt.setbgColor("white");
    nacgt.setfgColor("black");
    nacgt.drawAxes();
    nacgt.setfgColor("blue");
    nacgt.movePen(0,n_active_cells[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      nacgt.drawLine(cycle,n_active_cells[cycle]);
    }
    nacgt.flush();

    XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
      "number degress of freedom",0,N_CYCLES-1,n_dofs[0],
      n_dofs[N_CYCLES-1],&cmap,0,0.5);
    ndgt.newPage();
    ndgt.setbgColor("white");
    ndgt.setfgColor("black");
    ndgt.drawAxes();
    ndgt.setfgColor("blue");
    ndgt.movePen(0,n_dofs[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ndgt.drawLine(cycle,n_dofs[cycle]);
    }
    ndgt.flush();

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#ifdef TIMING
//tl.printOn(cout);
//tl.~TimedObjectList();
#endif

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
