// "$Header: /home/faculty/johnt/cvs/deal_new/examples/SphericalDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <iostream>

#include "Boundary.H"
#include "DataOut.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM==1)
barf // use RectangularDomain
#elif (SPACEDIM==2)
# define N_CYCLES 4
  enum MESH_ELEMENTS { 
    QUADRILATERAL_MESH_ELEMENTS,
    TRIANGLE_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=QUADRILATERAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TRIANGLE_MESH_ELEMENTS;
#else
# define N_CYCLES 3
  enum MESH_ELEMENTS { 
    HEXAHEDRAL_MESH_ELEMENTS,
    PRISM_MESH_ELEMENTS,
    TETRAHEDRAL_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=HEXAHEDRAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=PRISM_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TETRAHEDRAL_MESH_ELEMENTS;
  int dir=0;
#endif
enum TEST_PROBLEM { 
  LINEAR_SOLUTION,
  QUADRATIC_SOLUTION,
  CUBIC_SOLUTION,
  RADIALLY_SYMMETRIC_SOLUTION
};
TEST_PROBLEM test_problem=LINEAR_SOLUTION;
int degree=1;

REAL exactSolutionValue(const Point<SPACEDIM> &p) {
#if (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return p[0]*(   p[0]+2.*p[1])
            +p[1]*(2.*p[0]+5.*p[1]);
//    return (1.-2.*p[0])*(1.-p[0]);
    case CUBIC_SOLUTION:
      return pow(p[0],3)+pow(p[1],3);
    case RADIALLY_SYMMETRIC_SOLUTION: {
      REAL r=sqrt(p[0]*p[0]+p[1]*p[1]);
      return pow(r,3);
    }
    case LINEAR_SOLUTION:
    default:
      return p[0]+2.*p[1];
//    return 1.-p[0];
  }
#elif (SPACEDIM==3)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return p[0]*(   p[0]+2.*p[1]+ 3.*p[2])
            +p[1]*(2.*p[0]+5.*p[1]+ 8.*p[2])
            +p[2]*(3.*p[0]+8.*p[1]+14.*p[2]);
//    return (1.-2.*p[0])*(1.-p[0]);
    case CUBIC_SOLUTION:
      return pow(p[0],3)+pow(p[1],3)+pow(p[2],3);
    case RADIALLY_SYMMETRIC_SOLUTION:
      return sin(M_PI*p[0])*cos(M_PI*p[1])*sinh(M_SQRT2*M_PI*p[2]);
    case LINEAR_SOLUTION:
    default:
      return p[0]+2.*p[1]+3.*p[2];
//    return 1.-p[0];
  }
#endif
}

REAL rightHandSideValue(const Point<SPACEDIM> &p) {
#if (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return -12.; // - 2 * trace of A=[[1,2],[2,5]]
//    return -4.;
    case CUBIC_SOLUTION:
      return -6.*(p[0]+p[1]);
    case RADIALLY_SYMMETRIC_SOLUTION:
      return -9.*sqrt(p[0]*p[0]+p[1]*p[1]);
    case LINEAR_SOLUTION:
    default:
      return 0.;
  }
#elif (SPACEDIM==3)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return -40.; // - 2 * trace of A=[[1,2,3],[2,5,8],[3,8,14]]
//    return -4.;
    case CUBIC_SOLUTION:
      return -6.*(p[0]+p[1]+p[2]);
    case LINEAR_SOLUTION:
    case RADIALLY_SYMMETRIC_SOLUTION:
    default:
      return 0.;
  }
#endif
}

Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &p) {
#if (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=2.*p[0]+ 4.*p[1];
      return_val[1]=4.*p[0]+10.*p[1]; // 2 * A * x
//    return_val[0]=-3.+4.*p[0];
//    return_val[1]=0.;
      return return_val;
    }
    case CUBIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=3.*p[0]*p[0];
      return_val[1]=3.*p[1]*p[1];
      return return_val;
    }
    case RADIALLY_SYMMETRIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      REAL r=sqrt(p[0]*p[0]+p[1]*p[1]);
      return_val[0]=3.*r*p[0];
      return_val[1]=3.*r*p[1];
      return return_val;
    }
    case LINEAR_SOLUTION:
    default: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=1.;
      return_val[1]=2.;
//    return_val[0]=-1.;
//    return_val[1]=0.;
      return return_val;
    }
  }
#elif (SPACEDIM==3)
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=2.*p[0]+ 4.*p[1]+ 6.*p[2];
      return_val[1]=4.*p[0]+10.*p[1]+16.*p[2];
      return_val[2]=6.*p[0]+16.*p[1]+28.*p[2]; // 2 * A * x
//    return_val[0]=-3.+4.*p[0];
//    return_val[1]=0.;
      return return_val;
    }
    case CUBIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=3.*p[0]*p[0];
      return_val[1]=3.*p[1]*p[1];
      return_val[2]=3.*p[2]*p[2];
      return return_val;
    }
    case RADIALLY_SYMMETRIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=M_PI*cos(M_PI*p[0])*cos(M_PI*p[1])
                        *sinh(M_SQRT2*M_PI*p[2]);
      return_val[1]=-M_PI*sin(M_PI*p[0])*sin(M_PI*p[1])
                         *sinh(M_SQRT2*M_PI*p[2]);
      return_val[2]=M_SQRT2*M_PI*sin(M_PI*p[0])*cos(M_PI*p[1])
                           *cosh(M_SQRT2*M_PI*p[2]);
      return return_val;
    }
    case LINEAR_SOLUTION:
    default: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=1.;
      return_val[1]=2.;
      return_val[2]=3.;
//    return_val[0]=-1.;
//    return_val[1]=0.;
//    return_val[2]=0.;
      return return_val;
    }
  }
#endif
}

class ExactFunctionSystem : public FunctionSystem {
  protected:
    explicit ExactFunctionSystem(const FunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ExactFunctionSystem() {;}
    virtual ~ExactFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
};
class NeumannFunctionSystem : public FunctionSystem {
  protected:
  explicit NeumannFunctionSystem(const NeumannFunctionSystem&) {
    OBSOLETE("copy constructor disallowed");
  }
  public:
    NeumannFunctionSystem() {}
    virtual ~NeumannFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      return (g.getValues()*p)/sqrt(p.square());
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      function_values[0]=(g.getValues()*p)/sqrt(p.square());
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
};
#ifdef TIMING
//TimedObjectList tl;
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
#ifdef TIMING
//TimedObject *z3=
//  OPERATOR_NEW TimedObject("main interpolateBoundaryValues",tl);
//TimedObject *z5=OPERATOR_NEW TimedObject("main solve",tl);
//TimedObject *z10=OPERATOR_NEW TimedObject("main refine",tl);
#endif
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      cout << "quadratic solution" << endl;
      break;
    case CUBIC_SOLUTION:
      cout << "cubic solution" << endl;
      break;
    case RADIALLY_SYMMETRIC_SOLUTION:
      cout << "radially symmetric solution" << endl;
      break;
    case LINEAR_SOLUTION:
    default:
      cout << "linear solution" << endl;
  }
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> center(D_DECL(0.,0.,0.));
  REAL radius=1.;
  BoundaryData *boundary_data=0;
  NumPtr<Boundary*> boundary(1);
  NumPtr<BoundaryCondition*> boundary_condition(2);

  bool set_face_ids=false;
#if (SPACEDIM==2)
  if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
    boundary_data=generateBall(tess,center,radius,set_face_ids,boundary,
      boundary_condition);
  } else {
    boundary_data=generateBallTriangles(tess,center,radius,set_face_ids,
      boundary,boundary_condition);
  }
#elif (SPACEDIM==3)
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
      boundary_data=generateBall(tess,center,radius,set_face_ids,boundary,
        boundary_condition);
      break;
    case PRISM_MESH_ELEMENTS:
      boundary_data=generateBallPrismsWithTetrahedra(tess,center,radius,
        set_face_ids,boundary,boundary_condition);
//    boundary_data=generateBallPrismsWithHexahedra(tess,center,radius);
      break;
    case TETRAHEDRAL_MESH_ELEMENTS:
      boundary_data=generateBallTetrahedra(tess,center,radius,set_face_ids,
        boundary,boundary_condition);
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif
  LaplaceMaterialModel lmm(rightHandSideValue);
  tess.setMaterialModel(&lmm);

  ExactFunctionSystem efs;
  NeumannFunctionSystem nfs;
  boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
  for (int i=0;i<2;i++) {
    boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[0]);
  }
  boundary_condition[0]->setFunctionSystem(&efs);
  boundary_condition[0]->makeEssential(); // Dirichlet
  boundary_condition[1]->setFunctionSystem(&nfs);
  boundary_condition[1]->makeNatural();   // Neumann
  REAL fudge=sqrt(DBL_EPSILON);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<SPACEDIM> fc=face_element->center(tess.getVertices());
      switch (test_problem) {
        case QUADRATIC_SOLUTION: {
#if (SPACEDIM==2)
//        DIRICHLET on SW, NEUMANN on NE
          codimension1_level->setBoundaryCondition(fc[0]+fc[1]<fudge 
            ? boundary_condition[0] : boundary_condition[1]);
#else
          codimension1_level->setBoundaryCondition(
            fc[1]<radius/sqrt(3.)+fudge
            ? boundary_condition[0] : boundary_condition[1]);
#endif
          break;
        }
        case CUBIC_SOLUTION: {
#if (SPACEDIM==2)
//        NEUMANN on SW, DIRICHLET on NE
          codimension1_level->setBoundaryCondition(fc[0]+fc[1]<fudge 
            ? boundary_condition[1] : boundary_condition[0]);
#else
          codimension1_level->setBoundaryCondition(
            fc[1]<radius/sqrt(3.)+fudge
            ? boundary_condition[1] : boundary_condition[0]);
#endif
          break;
        }
        case LINEAR_SOLUTION:
        case RADIALLY_SYMMETRIC_SOLUTION:
        default: {
//        DIRICHLET all around
//        NEUMANN on SW, DIRICHLET on NE
          codimension1_level->setBoundaryCondition(boundary_condition[0]);
          break;
        }
      }
    }
  }
  tess.synchronize(boundary_data);

//coordinate mapping
  MappingL mapping(degree);

//Lagrange polynomials to represent solution of problem
  FE_L fe(degree);
//FE_H fe(degree);

//quadrature rules
  map<int,Quadrature<SPACEDIM> > qmap;
#if (SPACEDIM==2)
  if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
    qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
      GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+1));
  } else {
    qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
      TriangleGaussianQuadrature::argGivenOrder(2*degree-1)); 
  }
#elif (SPACEDIM==3)
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
      qmap[Hexahedron::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
        GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+1));
      break;
    case PRISM_MESH_ELEMENTS:
      qmap[Prism::verticesPerShape()]=ProductQuadrature(
        TriangleGaussianQuadrature(
          TriangleGaussianQuadrature::argGivenOrder(2*degree-1)),
        GaussianQuadrature<1>(
          GaussianQuadrature<1>::argGivenOrder(2*degree-1)));
      break;
    case TETRAHEDRAL_MESH_ELEMENTS:
      qmap[Tetrahedron::verticesPerShape()]=TetrahedronGaussianQuadrature(
        TetrahedronGaussianQuadrature::argGivenOrder(2*degree-1)); 
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif

//natural bc's
  map<int,Quadrature<SPACEDIM-1> > nqmap;
#if (SPACEDIM==2)
  {
    int nv=Interval::verticesPerShape();
    int m=GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree-1);
    nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(m);
  }
#elif (SPACEDIM==3)
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
    case PRISM_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      int m=GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+1);
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(m);

      nv=Triangle::verticesPerShape();
      m=TriangleGaussianQuadrature::argGivenOrder(2*degree-1);
      nqmap[nv]=TriangleGaussianQuadrature(m);
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree-1));
      nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree-1));
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
# endif

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

//graphics setup
  Palette pal;
#if (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
//AreaGraphTool contour_gt("solution contours","x0","x1",rlo,rhi,&cmap,
//  0,0.5);
//AreaGraphTool plot_gt("solution plot","x0","y1",rlo,rhi,&cmap,0,0.5);
//BOOLEAN use_lighting=FALSE;
//Vector3 vlow(0.,0.,0.);
//Vector3 vhigh(1.,1.,1.);
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
#elif (SPACEDIM==3)
//
  BOOLEAN use_lighting=TRUE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
//VolGraphTool plot_gt("solution plot",vlow,vhigh,pal,use_lighting,0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&plot_gt);
//  plot_gt.drawBox(vlow,vhigh);
//}
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
//
#endif

//loop over global mesh refinements
  for (int cycle=0;cycle<N_CYCLES;cycle++) {

//  data structure to keep track of degrees of freedom in solve
    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();
    n_dofs[cycle]=dof_handler.numberDofs();
    cout << "\n\tNumber of degrees of freedom: " << n_dofs[cycle]
         << endl;

//  data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//  storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

//  assemble linear system
    dof_handler.assembleLinearSystem(mapping,qmap,nqmap,system_matrix,
      system_rhs);

//  compute essential boundary values at mapped points on bdry
//  calls FunctionSystem in BoundaryCondition's to set boundary values
    map<int,REAL> boundary_values;
    {
#ifdef TIMING
//    Timer y(z3);
#endif
      dof_handler.interpolateBoundaryValues(mapping,boundary_values);
    }

//  modify linear system and rhs for essential boundary condition
    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);

//  solve linear system
    SolverControl solver_control(solution.size(),DBL_EPSILON*1.e4,false,
      false);
    CGSolver cg(solver_control);
    PreconditionIdentity pi;
    cg.solve(system_matrix,solution,system_rhs,pi);

//  estimate error
    n_active_cells[cycle]=tess.numberActiveShapes();
    FunctionSystem *weight=0;
    L2_error[cycle]=dof_handler.integrateDifference(mapping,solution,efs,
      qmap,L2_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tL2_error[" << cycle << "] = " << L2_error[cycle] << endl;
#endif
    
    H1_error[cycle]=dof_handler.integrateDifference(mapping,solution,efs,
      qmap,H1_NORM,weight,1.);
   
    Linfty_error[cycle]=dof_handler.integrateDifference(mapping,solution,
      efs,qmap,LINFTY_NORM,weight,1.);

//  compute solution values for graphical display
    DataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
#if (SPACEDIM==1)
    int n_subdivisions=640/n_active_cells[cycle];
#else
    int n_subdivisions=degree;
#endif
    data_out.buildPatches(n_subdivisions);
    bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>=2)
    data_out.drawGrid(&mesh_gt,quit_button);
#endif
#if (SPACEDIM==2)
//  data_out.contourPlot("solution",&contour_gt););
#endif
//  data_out.plot("solution",&plot_gt);
#if (SPACEDIM==3)
//  data_out.plotSurface("solution",&surface_gt);
#endif
#if (SPACEDIM==1)
//  if (!quit_button) {
//    XYGraphTool::WINDOW_TYPE::QuitButton qb;
//  }
#elif (SPACEDIM==2)
    if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif
    if (cycle<N_CYCLES-1) {
#ifdef TIMING
//    Timer y(z10);
#endif

//    refine tessellation
      tess.refineGlobal(1);
    }
    delete sparsity_pattern;
  }

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    XYGraphTool nacgt("number active cells vs cycle","cycle",
      "number active cells",0,N_CYCLES-1,n_active_cells[0],
      n_active_cells[N_CYCLES-1],&cmap,0,0.5);
    nacgt.newPage();
    nacgt.setbgColor("white");
    nacgt.setfgColor("black");
    nacgt.drawAxes();
    nacgt.setfgColor("blue");
    nacgt.movePen(0,n_active_cells[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      nacgt.drawLine(cycle,n_active_cells[cycle]);
    }
    nacgt.flush();

    XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
      "number degress of freedom",0,N_CYCLES-1,n_dofs[0],
      n_dofs[N_CYCLES-1],&cmap,0,0.5);
    ndgt.newPage();
    ndgt.setbgColor("white");
    ndgt.setfgColor("black");
    ndgt.drawAxes();
    ndgt.setfgColor("blue");
    ndgt.movePen(0,n_dofs[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ndgt.drawLine(cycle,n_dofs[cycle]);
    }
    ndgt.flush();

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#ifdef TIMING
//tl.printOn(cout);
//tl.~TimedObjectList();
#endif

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
