// "$Header: /home/faculty/johnt/cvs/deal_new/examples/SphericalDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <iostream>

#include "Boundary.H"
#include <ConstraintMatrix.H>
#include "DataOut.H"
#include "FEValues.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include <KellyErrorEstimator.H>
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM!=2)
barf // use RectangularDomain
#endif
#define N_CYCLES 8

int degree=3;
REAL radius=1.;
REAL angle=M_PI*1.5;
REAL pi_over_angle=M_PI/angle;
REAL poam1=pi_over_angle-1.;
Point<SPACEDIM> center(D_DECL(0.,0.,0.));

REAL exactSolutionValue(const Point<SPACEDIM> &p) {
  REAL r=sqrt(p.square());
  if (r<=0.) return 0.;
  REAL theta=acos(p[0]/r);
  if (p[1]<0.) theta=2.*M_PI-theta;
  return pow(r,pi_over_angle)*sin(theta*pi_over_angle);
}

REAL rightHandSideValue(const Point<SPACEDIM> &p) { return 0.; }

Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &p) {
  REAL r=sqrt(p.square());
  Tensor<1,SPACEDIM> return_val;
  if (radius<=0.) {
    if (poam1>0.) {
      return_val[0]=0.;
      return_val[1]=0.;
    } else if (poam1<0.) {
      return_val[0]=HUGE_VAL;
      return_val[1]=HUGE_VAL;
    } else {
      return_val[0]=0.;
      return_val[1]=1.;
    }
  } else {
    REAL theta=acos(p[0]/r);
    if (p[1]<0.) theta=2.*M_PI-theta;
    return_val[0]=sin(theta*poam1);
    return_val[1]=cos(theta*poam1);
    return_val*=pi_over_angle*pow(radius,poam1);
  } 
  return return_val;
}

class DirichletFunctionSystem : public FunctionSystem {
  protected:
    explicit DirichletFunctionSystem(const FunctionSystem &bc) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    DirichletFunctionSystem() {;}
    virtual ~DirichletFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
};

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  cout << "\tdegree = " << degree << endl;

  Tessellation tess;

  Point<SPACEDIM> center(D_DECL(0.,0.,0.));
  REAL radius=1.;
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=generateBallSectorTriangles(tess,center,
    radius,angle,set_face_ids,boundary,boundary_condition);

  LaplaceMaterialModel lmm(rightHandSideValue);
  tess.setMaterialModel(&lmm);

  DirichletFunctionSystem dfs;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    boundary_condition[i]->setFunctionSystem(&dfs);
    boundary_condition[i]->makeEssential(); // Dirichlet boundary
  }

  MappingL mapping(degree);
  FE_L fe(degree);

  map<int,Quadrature<SPACEDIM> > qmap;
  qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
    TriangleGaussianQuadrature::argGivenOrder(2*degree-1)); 

  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
//AreaGraphTool contour_gt("solution contours","x0","x1",rlo,rhi,&cmap,
//  0,0.5);
  AreaGraphTool plot_gt("solution plot","x0","y1",rlo,rhi,&cmap,0,0.5);

  map<int,Quadrature<SPACEDIM-1> > error_estimator_quadrature_map;
  error_estimator_quadrature_map[Interval::verticesPerShape()]=
    GaussianQuadrature<1>(GaussianQuadrature<1>::argGivenOrder(2*degree+1));

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

//tess.refineGlobal(1);
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
    DofHandler dof_handler(tess,fe);
    dof_handler.distributeDofs();

    ConstraintMatrix hanging_node_constraints;
    hanging_node_constraints.makeHangingNodeConstraints(dof_handler);
    hanging_node_constraints.close();

    n_dofs[cycle]=dof_handler.numberDofs();
    cout << "\n\tNumber of degrees of freedom: " << n_dofs[cycle]
         << endl;

    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      dof_handler.maxCouplingsBetweenDofs());
    dof_handler.makeSparsityPattern(*sparsity_pattern);

    hanging_node_constraints.condense(*sparsity_pattern);
    sparsity_pattern->compress();

    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

//  quadrature rule for Neumann bc; only no. quad pts is used
    map<int,Quadrature<1> > nqmap;
    nqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
      GaussianQuadrature<1>::argGivenOrder(2*degree-1));

    dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix,system_rhs);

    hanging_node_constraints.condense(system_matrix);
    hanging_node_constraints.condense(system_rhs);

    map<int,REAL> boundary_values;
    dof_handler.interpolateBoundaryValues(mapping,boundary_values);

    bool preserve_symmetry=true;
    dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    dof_handler.applyBoundaryValues(boundary_values,solution);

    SolverControl solver_control(solution.size(),DBL_EPSILON*1.e4,false,
      false);
    CGSolver cg(solver_control);
    PreconditionIdentity pi;
    cg.solve(system_matrix,solution,system_rhs,pi);
    hanging_node_constraints.distribute(solution);

    n_active_cells[cycle]=tess.numberActiveShapes();
    FunctionSystem *weight=0;
    L2_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,L2_NORM,weight,1.);
#ifdef DEBUG
    cout << "\tL2_error[" << cycle << "] = " << L2_error[cycle] << endl;
#endif
    
    H1_error[cycle]=dof_handler.integrateDifference(mapping,solution,dfs,
      qmap,H1_NORM,weight,1.);
   
    Linfty_error[cycle]=dof_handler.integrateDifference(mapping,solution,
      dfs,qmap,LINFTY_NORM,weight,1.);

    DataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution",DataOut::TYPE_DOF_DATA);
    int n_subdivisions=degree;
//  int n_subdivisions=100;
    data_out.buildPatches(n_subdivisions);
    bool quit_button=false;
    data_out.drawGrid(&mesh_gt,quit_button);
//  data_out.contourPlot("solution",&contour_gt,quit_button);
    data_out.plot("solution",&plot_gt,quit_button);
    if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }

    if (cycle<N_CYCLES-1) {
      Vector estimated_error_per_cell(tess.numberActiveShapes());
      KellyErrorEstimator error_estimator(mapping,dof_handler,
        error_estimator_quadrature_map,solution,estimated_error_per_cell);
//    tess.refineAndCoarsenFixedNumber(estimated_error_per_cell,0.3,
//      0.03);
      tess.refineAndCoarsenFixedFraction(estimated_error_per_cell,0.3,
        0.03);
      tess.executeCoarseningAndRefinement();
//    tess.refineGlobal(1);
    }
    delete sparsity_pattern;
  }

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    XYGraphTool nacgt("number active cells vs cycle","cycle",
      "number active cells",0,N_CYCLES-1,n_active_cells[0],
      n_active_cells[N_CYCLES-1],&cmap,0,0.5);
    nacgt.newPage();
    nacgt.setbgColor("white");
    nacgt.setfgColor("black");
    nacgt.drawAxes();
    nacgt.setfgColor("blue");
    nacgt.movePen(0,n_active_cells[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      nacgt.drawLine(cycle,n_active_cells[cycle]);
    }
    nacgt.flush();

    XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
      "number degress of freedom",0,N_CYCLES-1,n_dofs[0],
      n_dofs[N_CYCLES-1],&cmap,0,0.5);
    ndgt.newPage();
    ndgt.setbgColor("white");
    ndgt.setfgColor("black");
    ndgt.drawAxes();
    ndgt.setfgColor("blue");
    ndgt.movePen(0,n_dofs[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ndgt.drawLine(cycle,n_dofs[cycle]);
    }
    ndgt.flush();

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#ifdef TIMING
//tl.printOn(cout);
//tl.~TimedObjectList();
#endif

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  } 
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
