// "$Header:$"
//----------------------------  data_out.cc  ---------------------------
//  $Id: data_out.cc,v 1.68 2003/04/07 15:39:02 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  data_out.cc  ---------------------------
//
//modified from deal.II/deal.II/source/numerics/data_out.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <DataOut.H>
#include <FEValues.H>
#include <Surface.H>
#include <Tracer.H>
#include <TypedPlotObj.H>
#include <Vector.H>
#include <VolGraphTool.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::replaceDataVector(const Vector &vec,const string &name) {
  bool found=false;
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) {
        dof_data[i]->replaceData(&vec);
        found=true;
        break;
      }
    }
    if (found) break;
  }
  if (found) return;
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) {
        cell_data[i]->replaceData(&vec);
        found=true;
        break;
      }
    }
    if (found) break;
  }
  CHECK_TEST(found);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::clearDataVectors() {
  for (int i=0;i<dof_data.getNumber();i++) {
    delete dof_data[i]; dof_data[i]=0;
  }
  dof_data.cleanup();
  for (int i=0;i<cell_data.getNumber();i++) {
    delete cell_data[i]; cell_data[i]=0;
  }
  cell_data.cleanup();
  for (int i=0;i<patches.getNumber();i++) {
    delete patches[i]; patches[i]=0;
  }
  patches.cleanup();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::addDataVector(const Vector &vec,
const NumPtr<string> &new_names,DATA_VECTOR_TYPE type) {
  for (int i=0;i<new_names.getNumber();i++) {
    CHECK_SAME(new_names[i].
      find_first_not_of("abcdefghijklmnopqrstuvwxyz"
      "ABCDEFGHIJKLMNOPQRSTUVWXYZ" "0123456789_<>()"),string::npos);
  }
  DataEntry *new_entry=OPERATOR_NEW DataEntry(&vec,new_names);
  switch (type) {
    case TYPE_CELL_DATA:
      CHECK_SAME(vec.size(),
                 dofs->getTessellation().numberActiveShapes());
      CHECK_SAME(new_names.getNumber(),1);
      break;
    case TYPE_DOF_DATA:
      CHECK_SAME(vec.size(),dofs->numberDofs());
      CHECK_SAME(new_names.getNumber(),dofs->getFe().nComponents());
      break;
    default:
      if (vec.size()==dofs->getTessellation().numberActiveShapes()) {
        type=TYPE_CELL_DATA;
        CHECK_SAME(new_names.getNumber(),1);
      } else {
        type=TYPE_DOF_DATA;
        CHECK_SAME(new_names.getNumber(),dofs->getFe().nComponents());
      }
  }
  if (type==TYPE_DOF_DATA) {
    NumPtr<DataEntry*> new_dof_data(dof_data.getNumber()+1);
    new_dof_data.copyFrom(dof_data);
    new_dof_data[dof_data.getNumber()]=new_entry;
    dof_data.copy(new_dof_data);
  } else {
    NumPtr<DataEntry*> new_cell_data(cell_data.getNumber()+1);
    new_cell_data.copyFrom(cell_data);
    new_cell_data[cell_data.getNumber()]=new_entry;
    cell_data.copy(new_cell_data);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::addDataVector(const Vector &vec,const string &name,
DATA_VECTOR_TYPE type) {
  int n_components=dofs->getFe().nComponents();
  NumPtr<string> new_names;
  if ((n_components==1) ||
  (vec.size()==dofs->getTessellation().numberActiveShapes())) {
    new_names.allocate(1);
    new_names[0]=name;
  } else {
    new_names.allocate(n_components);
    for (int i=0; i<n_components; ++i) {
      ostringstream namebuf;
      namebuf << '_' << i;
      namebuf << ends;
      new_names[i]=name+namebuf.str();
    }
  }
  addDataVector(vec,new_names,type);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::clearInputDataReferences() {
  for (int i=0;i<dof_data.getNumber();i++) {
    dof_data[i]->clearPointers();
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    cell_data[i]->clearPointers();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::getDatasetNames(NumPtr<string> &names_out) const {
  int n=0;
  for (int i=0;i<dof_data.getNumber();i++) {
    n+=dof_data[i]->getNames().getNumber();
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    n+=cell_data[i]->getNames().getNumber();
  }
  names_out.allocate(n);
  n=0;
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_names=dof_data[i]->getNames();
    for (int j=0;j<dof_names.getNumber();j++,n++) {
      names_out[n]=dof_names[j];
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_names=cell_data[i]->getNames();
    for (int j=0;j<cell_names.getNumber();j++,n++) {
      names_out[n]=cell_names[j];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::buildSomePatches(int n_subdivisions, 
NumPtr<NumPtr<int> > &cell_to_patch_index_map) {
  static const MappingL mapping;
  UPDATE_FLAGS flags=
    UPDATE_VALUES | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  FEValues fe_patch_values(mapping,dofs->getFe(),
    Patch::getQuadratureMap(),flags);

  int n_components=dofs->getFe().nComponents();
  int n_datasets=
    dof_data.getNumber()*n_components+cell_data.getNumber();
  int cell_number=0;
  int p=0;
  Tessellation &tess=dofs->getTessellation();
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),p++) {
    const SHAPE_CLASS *shape=tess.presentShape();
    LEVEL_CLASS *present_level=tess.presentLevel();
    Patch *patch=patches[p];
    if (n_datasets>0) {
      fe_patch_values.mapShapes(tess,shape,present_level);
      int n_q_points=patch->numberQuadraturePoints();
      for (int dataset=0;dataset<dof_data.getNumber(); 
      dataset++) {
        if (n_components==1) {
          NumPtr<REAL> patch_values;
          dofs->getFunctionValues(fe_patch_values,shape,present_level,
            *dof_data[dataset]->singleData(),patch_values);
          patch->setData(dataset,patch_values);
        } else {
          NumPtr<Vector> patch_values_system;
          dofs->getFunctionValues(fe_patch_values,shape,present_level,
            *dof_data[dataset]->singleData(),patch_values_system);
          patch->setData(dataset,n_components,patch_values_system);
        }
      }
      int dnc=dof_data.getNumber()*n_components;
      for (int dataset=0;dataset<this->cell_data.getNumber(); 
      dataset++) {
        REAL value=(*cell_data[dataset]->singleData())[cell_number];
        patch->setData(dataset+dnc,value);
      }
    }
    for (int f=0;f<shape->numberCodimension1Shapes();f++) {
      int neighbor_index=shape->codimension1ShapeIndex(f);
      LEVEL_CLASS *neighbor_level=present_level->neighborLevel(f);
      if (neighbor_index<0) continue;
      if (present_level!=neighbor_level) continue;

      int neighbor_level_number=
        neighbor_level->numberCoarserLevels();
      NumPtr<int> &cell_to_patch_index_map_this_level=
        cell_to_patch_index_map[neighbor_level_number];
      if (neighbor_index>=
      cell_to_patch_index_map_this_level.getNumber()) {
        continue;
      }
      if (cell_to_patch_index_map_this_level[neighbor_index]==
      Patch::no_neighbor) {
        continue;
      }

      patch->setNeighbor(f,neighbor_index);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::buildPatches(int n_subdivisions) {
  Patch::setQuadratures(n_subdivisions);
  CHECK_POINTER(dofs);
  patches.cleanup();
  Tessellation &tess=dofs->getTessellation();
  int n_levels=tess.numberLevels();
  NumPtr<NumPtr<int> > cell_to_patch_index_map(n_levels);
  int l=0;
  for (LEVEL_CLASS *level=tess.coarsestLevel();level;
  level=level->finerLevel(),l++) {
    int max_index=0;
    for (bool doit=tess.firstActiveShape(level);doit;
    doit=tess.nextActiveShapeInLevel()) {
      max_index=max(max_index,level->presentIndex());
    }
    cell_to_patch_index_map[l].allocate(max_index+1);
    cell_to_patch_index_map[l].initialize(Patch::no_neighbor);
  }
  int n_patches=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),n_patches++) {
    LEVEL_CLASS *level=tess.presentLevel();
    cell_to_patch_index_map[level->numberCoarserLevels()]
      [level->presentIndex()]=n_patches;
  }
  int n_components=dofs->getFe().nComponents();
  int n_datasets=
    dof_data.getNumber()*n_components+cell_data.getNumber();
  patches.allocate(n_patches);
  int p=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),p++) {
    const SHAPE_CLASS *e=tess.presentShape();
    patches[p]=OPERATOR_NEW Patch(e,tess.presentLevel(),tess,
      *mapping,n_datasets,p);
  }

  buildSomePatches(n_subdivisions,cell_to_patch_index_map);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::reBuildPatches() {
  CHECK_POINTER(dofs);
  Tessellation &tess=dofs->getTessellation();
  int n_components=dofs->getFe().nComponents();
  int n_datasets=
    dof_data.getNumber()*n_components+cell_data.getNumber();
  static const MappingL mapping_l;
  UPDATE_FLAGS flags= UPDATE_VALUES | UPDATE_JxW_VALUES |
    UPDATE_Q_POINTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
  FEValues fe_patch_values(mapping_l,dofs->getFe(),
    Patch::getQuadratureMap(),flags);

  int p=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),p++) {
    const SHAPE_CLASS *shape=tess.presentShape();
    LEVEL_CLASS *present_level=tess.presentLevel();
    Patch *patch=patches[p];
    if (n_datasets>0) {
      fe_patch_values.mapShapes(tess,shape,present_level);
      int n_q_points=patch->numberQuadraturePoints();
      for (int dataset=0;dataset<dof_data.getNumber();dataset++) {
        if (n_components==1) {
          NumPtr<REAL> patch_values;
          dofs->getFunctionValues(fe_patch_values,shape,present_level,
            *dof_data[dataset]->singleData(),patch_values);
          patch->setData(dataset,patch_values);
        } else {
          NumPtr<Vector> patch_values_system;
          dofs->getFunctionValues(fe_patch_values,shape,present_level,
            *dof_data[dataset]->singleData(),patch_values_system);
          patch->setData(dataset,n_components,patch_values_system);
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

#include <AreaGraphTool.H>
#include <Palette.H>
#include <Vector3.H>
#include <VolGraphTool.H>
#include <XColormap.H>
#include <XYGraphTool.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::plot(const char *name,VirtualGraphTool *gt,
bool quit_button) const {
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) plot(name,i,gt,quit_button);
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) plot(name,i,gt,quit_button);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void DataOut::plot(const char *name,int data_set,VirtualGraphTool *gt,
bool quit_button) const {
  XYGraphTool *xygt=dynamic_cast<XYGraphTool*>(gt);
  CHECK_POINTER(xygt);
  Point<1> low(HUGE_VAL);
  Point<1> high(-HUGE_VAL);
  int n_datasets=dof_data.getNumber()
    *dofs->getFe().nComponents()+cell_data.getNumber();
  REAL dlow(HUGE_VAL);
  REAL dhigh(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,dlow,dhigh);
  }
  xygt->rescale(low[0],high[0],dlow,dhigh);
  xygt->newPage();
  xygt->setbgColor("white");
  xygt->setfgColor("black");
  xygt->drawAxes();
  xygt->setfgColor("blue");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plot(*xygt,data_set);
  }
  xygt->flush();
  if (quit_button) XYGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif

#if (SPACEDIM==2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::plot(const char *name,int data_set,VirtualGraphTool *gt,
bool quit_button) const {
  AreaGraphTool *agt=dynamic_cast<AreaGraphTool*>(gt);
  CHECK_POINTER(agt);
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  REAL ymin=HUGE_VAL;
  REAL ymax=-HUGE_VAL;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
  }
  agt->rescale(low[0],high[0],low[1],high[1]);
  agt->newPage();
  agt->setbgColor("white");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plot(*agt,data_set,ymin,ymax);
  }
  agt->setfgColor("black");
  agt->drawAxes();
  agt->flush();
  if (quit_button) AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DataOut::plot(const char *name,int data_set,VirtualGraphTool *gt,
bool) const {
  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(gt);
  CHECK_POINTER(vgt);
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  REAL ymin=DBL_MAX;
  REAL ymax=-DBL_MAX;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
  }
  Vector3 vlow(low[0],low[1],low[2]);
  Vector3 vhigh(high[0],high[1],high[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;
  vgt->rescale(vlow,vhigh);

  vgt->newPage();
  vgt->expose();
  PatchDataSet *pds=OPERATOR_NEW PatchDataSet(vgt,vlow,vhigh,patches,
    dofs,mapping,dof_data[data_set]->singleData());
  TypedPlotObj<PatchDataSet> *tpo=
    OPERATOR_NEW TypedPlotObj<PatchDataSet>(pds);
  vgt->initialize(tpo);
  vgt->plotClipPlanes();
  vgt->eventLoop();
  vgt->terminatePlotObj();
  delete pds;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void DataOut::plotSurface(const char *name,VolGraphTool *gt) const {
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) plotSurface(name,j,gt);
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) plotSurface(name,j,gt);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void DataOut::plotSurface(const char *name,int data_set,
VolGraphTool *gt) const{
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  REAL ymin=DBL_MAX;
  REAL ymax=-DBL_MAX;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
  }
  Vector3 vlow(low[0],low[1],ymin),vhigh(high[0],high[1],ymax);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;
  gt->rescale(vlow,vhigh);
  gt->newPage();
  gt->expose();

  { BOOLEAN do_best=FALSE;
    VolGraphTool::DrawDelimiter dd(do_best,gt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->plotSurface(*gt,data_set,vlow[2],vhigh[2],do_best);
    }
  }
  { BOOLEAN do_best=TRUE;
    VolGraphTool::DrawDelimiter dd(do_best,gt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->plotSurface(*gt,data_set,vlow[2],vhigh[2],do_best);
    }
  }
  gt->eventLoop();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void DataOut::plotSurface(const char *name,int data_set,
VolGraphTool *gt) const{
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  REAL ymin=DBL_MAX;
  REAL ymax=-DBL_MAX;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
  }
  Vector3 vlow(low[0],low[1],low[2]);
  Vector3 vhigh(high[0],high[1],high[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;
  gt->rescale(vlow,vhigh);
  gt->newPage();
  PatchSurfaceDataSet psds(gt,vlow,vhigh,dofs,mapping,patches,
    ymin,ymax,data_set);
  Surface<PatchSurfaceDataSet> *s=
    OPERATOR_NEW Surface<PatchSurfaceDataSet>(&psds,gt);
  gt->initialize(s);
  gt->eventLoop();
  gt->terminatePlotObj();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void DataOut::drawGrid(VirtualGraphTool *gt,bool quit_button)
const {
  AreaGraphTool *agt=dynamic_cast<AreaGraphTool*>(gt);
  CHECK_POINTER(agt);
  CHECK_TEST(patches.getNumber()>0);
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
  }
  agt->rescale(low[0],high[0],low[1],high[1]);
  agt->setbgColor("white");
  agt->newPage();
  agt->flush();
  agt->setfgColor("black");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->draw(*agt);
  }
  agt->drawAxes();
  agt->flush();
  if (quit_button) AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void DataOut::drawGrid(VirtualGraphTool *gt,bool) const {
  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(gt);
  CHECK_POINTER(vgt);
  CHECK_TEST(patches.getNumber()>0);
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
  }
  Vector3 vlow(low[0],low[1],low[2]);
  Vector3 vhigh(high[0],high[1],high[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;
  vgt->rescale(vlow,vhigh);
  vgt->newPage();
  vgt->expose();

  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=vgt->getColorArray();
  VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=color_array->getColor(0.);
  vgt->setfgColor("black");
  { BOOLEAN do_best=FALSE;
    VolGraphTool::DrawDelimiter dd(do_best,vgt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->draw(*vgt);
    }
  }
  { BOOLEAN do_best=TRUE;
    VolGraphTool::DrawDelimiter dd(do_best,vgt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->draw(*vgt);
    }
  }
  vgt->expose();
  vgt->eventLoop();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void DataOut::contourPlot(const char *name,AreaGraphTool *gt,
bool quit_button) const {
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) contourPlot(name,j,gt,quit_button);
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) contourPlot(name,j,gt,quit_button);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void DataOut::contourPlot(const char *name,int data_set,
AreaGraphTool *gt,bool quit_button) const {
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
  }
  gt->rescale(low[0],high[0],low[1],high[1]);
  gt->newPage();
  int ncontours=30;
  REAL ymin=DBL_MAX,ymax=-DBL_MAX;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getDataBounds(data_set,ymin,ymax);
  }
  REAL dy=(ymax-ymin)/static_cast<REAL>(ncontours+1);
  gt->setfgColor("black");
  gt->drawBoxGivenCorners(low[0],low[1],high[0],high[1]);
  REAL value=ymin+dy*0.5;
  for (int ic=0;ic<ncontours;ic++,value+=dy) {
    gt->setfgColor(ic,ncontours);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->contour(*gt,value,data_set);
    }
  }
  gt->flush();
  if (quit_button) AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void PatchDataSet::plot(BOOLEAN,const AxisClipPlane *cp,REAL ymin,
REAL ymax) {
  REAL location=graph_tool->getLocation(cp);
  int dir=cp->getDirection();
  HAND h=cp->getHand();
  REAL fudge=graph_tool->getLength()[dir]*0.015625;
  if (h==LEFT_HAND) location += fudge;
  else location -= fudge;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plotClipPlane(*graph_tool,ymin,ymax,location,dir,h,dofs,
      mapping,single_data);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void PatchDataSet::plot(BOOLEAN do_best,const AxisClipPlane *cp) {
  int data_set=0;
  REAL ymin=HUGE_VAL;
  REAL ymax=-HUGE_VAL;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getDataBounds(data_set,ymin,ymax);
  }
  plot(do_best,cp,ymin,ymax);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void PatchDataSet::plot(BOOLEAN,REAL) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void PatchDataSet::plot(BOOLEAN) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
REAL PatchDataSet::getVarMin() const {
  OBSOLETE("not programmed"); 
  return HUGE_VAL;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
REAL PatchDataSet::getVarMax() const {
  OBSOLETE("not programmed"); 
  return HUGE_VAL;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void PatchSurfaceDataSet::plot(BOOLEAN,const AxisClipPlane *cp) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void PatchSurfaceDataSet::plot(BOOLEAN do_best,REAL val) {
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->
      plotSurface(*graph_tool,data_set,val,var_min,var_max,do_best);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "NumPtr.C"
INSTANTIATE_NUMPTR(string);
INSTANTIATE_NUMPTR(DataOut::DataEntry*);
INSTANTIATE_NUMPTR(Patch*);
INSTANTIATE_NUMPTR(NumPtr<int>);

#if (SPACEDIM==3)
template class TypedPlotObj<PatchDataSet>;
template class Surface<PatchSurfaceDataSet>;
#endif
