#ifndef DATA_OUT_H
#define DATA_OUT_H "$Header:$"
//----------------------------  data_out.h  ---------------------------
//   $Id: data_out.h,v 1.46 2003/05/30 19:05:54 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  data_out.h  ---------------------------
//
//modified from deal.II/deal.II/include/numerics/data_out.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <DofHandler.H>
#include <Patch.H>
#include <Types.H>

#if (deal_II_dimension==1)
class XYGraphTool;
#endif
#if (deal_II_dimension==2)
class AreaGraphTool;
class Vector3;
#endif
#if (deal_II_dimension>=2)
class VolGraphTool;
#endif
#if (deal_II_dimension==3)
class AxisClipPlane;
#endif

//compare to DEAL DataOut
class DataOut {
  protected:
    class DataEntry {
      private:
        int size;
        const Vector *single_data;
        NumPtr<string> names;
      public:
        DataEntry(const Vector *data=0,
        const NumPtr<string> &n=NumPtr<string>()) : size(0),
        single_data(data),names(n.getNumber()) {
          if (single_data) size=single_data->size();
          names.copyFrom(n); 
        }
        void replaceData(const Vector *new_data) {
          CHECK_SAME(size,new_data->size());
          single_data=new_data;
        }
        void clearPointers() { single_data=0; }
        const NumPtr<string>& getNames() const { return names; }
        const Vector* singleData() const { return single_data; }
    };
    const DofHandler *dofs;
    const Mapping *mapping;
    NumPtr<DataEntry*> dof_data;
    NumPtr<DataEntry*> cell_data;
    NumPtr<Patch*> patches;

    virtual void getDatasetNames(NumPtr<string>&) const;
    void clearDataVectors();
  private:
    void buildSomePatches(int,NumPtr<NumPtr<int> >&);
    void plot(const char*,int,VirtualGraphTool*,bool) const;
#if (SPACEDIM==2)
    void contourPlot(const char*,int,AreaGraphTool*,bool) const;
#endif
#if (SPACEDIM>=2)
    void plotSurface(const char*,int,VolGraphTool*) const;
#endif
  public:
//  TYPE_DOF_DATA : size = dof_handler.numberDofs, 1 name
//  TYPE_CELL_DATA : size = tess.numberActiveShapes, nComponents names
//  TYPE_AUTOMATIC : if size==tess.numberActiveShapes then 1 name
//                   else size arbitrary, nComponents names
    enum DATA_VECTOR_TYPE {TYPE_DOF_DATA,TYPE_CELL_DATA,TYPE_AUTOMATIC};

    DataOut(const DofHandler &d,const Mapping &m) : dofs(&d),mapping(&m)
      {;}
    virtual ~DataOut() { clearDataVectors(); }
    virtual void replaceDataVector(const Vector&,const string&);
    void addDataVector(const Vector&,const NumPtr<string>&,
      DATA_VECTOR_TYPE=TYPE_AUTOMATIC);
    void addDataVector(const Vector&,const string &name,
      DATA_VECTOR_TYPE=TYPE_AUTOMATIC);
    void clearInputDataReferences();
    virtual void buildPatches(int n_subdivisions=1);
    virtual void reBuildPatches();
    void plot(const char*,VirtualGraphTool*,bool) const;
#if (SPACEDIM==2)
    void contourPlot(const char*,AreaGraphTool*,bool) const;
#endif
#if (SPACEDIM>=2)
    void drawGrid(VirtualGraphTool*,bool) const;
    void plotSurface(const char*,VolGraphTool*) const;
#endif
};

#if (SPACEDIM==3)
class PatchDataSet {
  private:
    VolGraphTool *graph_tool;
    const Vector3 &low;
    const Vector3 &high;
    const NumPtr<Patch*> &patches;
    const DofHandler *dofs;
    const Mapping *mapping;
    const Vector *single_data;
  public:
    PatchDataSet(VolGraphTool *gt,const Vector3 &l,const Vector3 &h,
    const NumPtr<Patch*> &p,const DofHandler *dh,const Mapping *m,
    const Vector *sd) : graph_tool(gt),low(l),high(h),patches(p),
      dofs(dh),mapping(m),single_data(sd) { }
    virtual ~PatchDataSet() { }
    VolGraphTool* getGraphTool() const { return graph_tool; }
    BOOLEAN drawSurface() const { return FALSE; }
    void plot(BOOLEAN,const AxisClipPlane*);
    void plot(BOOLEAN,const AxisClipPlane*,REAL,REAL);
    void plot(BOOLEAN,REAL);
    void plot(BOOLEAN);
    REAL getVarMin() const;
    REAL getVarMax() const;
    virtual void printOn(ostream&) const {}
};
#endif

#if (SPACEDIM==3)
class PatchSurfaceDataSet {
  private:
    VolGraphTool *graph_tool;
    const Vector3 &low;
    const Vector3 &high;
    const DofHandler *dofs;
    const Mapping *mapping;
    const NumPtr<Patch*> &patches;
    REAL var_min,var_max;
    int data_set;
  public:
    PatchSurfaceDataSet(VolGraphTool *gt,const Vector3 &l,
    const Vector3 &h,const DofHandler *dh,const Mapping *m,
    const NumPtr<Patch*> &p,REAL vn,REAL vx,int ds) : 
      graph_tool(gt),low(l),high(h),dofs(dh),mapping(m),patches(p),
      var_min(vn),var_max(vx),data_set(ds) { }
    virtual ~PatchSurfaceDataSet() { }
    VolGraphTool* getGraphTool() const { return graph_tool; }
    REAL getVarMin() const { return var_min; }
    REAL getVarMax() const { return var_max; }
    BOOLEAN drawSurface() const { return TRUE; }
    void plot(BOOLEAN,const AxisClipPlane*);
    void plot(BOOLEAN,REAL);
    virtual void printOn(ostream&) const {}
};
#endif

#endif
