#ifndef DOF_HANDLER_H
#define DOF_HANDLER_H "$Header:$"
//----------------------------  dof_handler.h  ---------------------------
//    $Id: dof_handler.h,v 1.90 2003/02/04 21:56:48 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  dof_handler.h  ---------------------------
//----------------------------  dof_tools.h  ---------------------------
//    $Id: dof_tools.h,v 1.68 2003/02/17 19:46:15 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  dof_tools.h  ---------------------------
//
//modified from deal.II/deal.II/include/dofs/dof_handler.h and dof_tools.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <map>
#include <set>
#include <BoundaryCondition.H>
#include <CompressedSparsityPattern.H>
#include <DDecl.H>
#include <DofLevel.H>
#include <FEValues.H>
#include <FiniteElement.H>
#include <NumPtr.H>
#include <SparseMatrix.H>
#include <Tessellation.H>

enum NORM_TYPE {
  MEAN_NORM, L1_NORM, L2_NORM, LINFTY_NORM, H1_SEMINORM, H1_NORM,
  LP_NORM, W1P_SEMINORM, W1P_NORM
};
ostream& operator<<(ostream&,NORM_TYPE);

//compare to DEAL DoFHandler
class DofHandler {
  private:
//  can't use map<const LEVEL_CLASS*,DOF_LEVEL_CLASS*> because
//    getLevel takes pointer into LEVEL_CLASS* data structure
    NumPtr<DOF_LEVEL_CLASS*> levels;
    int used_dofs;
    NumPtr<int> vertex_dofs;

    DofHandler(const DofHandler&) {
      OBSOLETE("copy constructor disallowed");
    }
    DofHandler& operator=(const DofHandler&) {
      OBSOLETE("copy constructor disallowed");
    }
    void reserveSpace();
    int distributeDofsOnCell(const SHAPE_CLASS*,LEVEL_CLASS*,int);
  protected:
    Tessellation *tess;
    FiniteElement *selected_fe;

//  deletes levels and vertex_dofs
//    called from DofHandler::reserveSpace and MGDofHandler::clear
    void clearSpace();
    const DofLevel1* getLevel(const Level1 *level) const {
      return levels[level->numberCoarserLevels()];
    }
#if (SPACEDIM>1)
    const DofLevel2* getLevel(const Level2 *level) const {
      return levels[level->numberCoarserLevels()];
    }
//  called from dofsOnChildren
    int numberDofsOnChildren(const CODIMENSION1_SHAPE_CLASS*) const;
#endif
#if (SPACEDIM==3)
    const DofLevel3* getLevel(const Level3 *level) const {
      return levels[level->numberCoarserLevels()];
    }
#endif
//  called from getDofIndices, numberDofsOnChildren and 
//    interpolateBoundaryValues
    void vertexDofIndices(const Shape*,NumPtr<int>&) const;
//  if shape is refined, gets dofs for interpolation
//  called from getFunctionValues, getFunctionGrads and
//    getFunction2ndDerivatives    
    void getInterpolatedDofValues(const SHAPE_CLASS*,const LEVEL_CLASS*,
      const Vector&,Vector&) const;
  public:
    DofHandler(Tessellation &t,FiniteElement &fe) : 
      levels(t.numberLevels()),tess(&t),selected_fe(&fe),used_dofs(0)
      { levels.initialize(0); }
    virtual ~DofHandler() { clearSpace(); }

    int numberDofs() const { return used_dofs; }
    FiniteElement& getFe() const {
      CHECK_POINTER(selected_fe);
      return *selected_fe;
    }
    Tessellation& getTessellation() const { return *tess; }

//  pulls local dof's for element out of global dof's 
//  called from getInterpolatedDofValues and 
//    RefinementStudy::recursivelyIntegrateError
//  public for RefinementStudy
    void getDofValues(const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
      Vector&) const;
//  stores first global dof in each active element of tessellation
//  called from user programs
    virtual void distributeDofs();
//  called from computeCuthillMcKeeRenumbering, Multigrid::Multigrid and
//    user programs
    int maxCouplingsBetweenDofs() const;
//  called from computeCuthillMcKeeRenumbering and user programs
    void makeSparsityPattern(SparsityPatternBase&) const;
//  called from DofHandler::getDofValues, DofHandler::makeSparsityPattern
//    DofHandler::interpolateBoundaryValues,
//    ConstraintMatrix::makeHangingNodeConstraints
//    MGTransferPrebuilt::copyToMG,MGTransferPrebuilt::copyFromMG,
//      MGTransferPrebuilt::copyFromMGAdd and user programs
//  useful in forming global stiffness matrix from cell matrices
    void getDofIndices(const Shape1*,const Level1*,NumPtr<int>&) const;
#if (SPACEDIM>1)
    void getDofIndices(const Shape2*,const Level2*,NumPtr<int>&) const;
#endif
#if (SPACEDIM==3)
    void getDofIndices(const Shape3*,const Level3*,NumPtr<int>&) const;
#endif
//  Assembles linear system except for Dirichlet boundary values.
//    For TransientMaterialModel, linear system is for elliptic projection.
//    Called from user programs
    void assembleLinearSystem(const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,SparseMatrix&,Vector&) const;
//  Assembles L2 projection linear system for TransientMaterialModel
//    Called from user programs
    void assembleL2ProjectionLinearSystem(const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,SparseMatrix&,Vector&) const;
//  Assembles L2 projection linear system for TransientMaterialModel
//    Called from user programs
    void assembleEllipticProjectionLinearSystem(const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,SparseMatrix&,Vector&) const;
//  Assembles linear system for TransientMaterialModel
//    Called from user programs
    void assembleTransientLinearSystem(REAL,const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,SparseMatrix&,SparseMatrix&,
      Vector&) const;
    void assembleTransientLinearSystem(REAL,const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,Vector&) const;
//  sets Dirichlet boundary values
//    called from user programs
    void interpolateBoundaryValues(const Mapping&,map<int,REAL>&,REAL=0.)
      const;
    void interpolateBoundaryValues(const Mapping&,map<int,REAL>&,
      map<int,Quadrature<SPACEDIM-1> >&,REAL=0.) const;
//  puts Dirichlet boundary values into linear system
//    called from user programs
    void applyBoundaryValues(const map<int,double>&,SparseMatrix&,
      Vector&,bool) const;
    void applyBoundaryValues(const map<int,double>&,Vector&) const;
//  computes norm of difference of numerical and exact solutions
//    called from user programs
    REAL integrateDifference(const Mapping&,const Vector&,
      const FunctionSystem&,const map<int,Quadrature<SPACEDIM> >&,
      NORM_TYPE,const FunctionSystem*,REAL);
    REAL integrateDifferenceOnBoundary(const Mapping&,const Vector&,
      const FunctionSystem&,const map<int,Quadrature<SPACEDIM> >&,
      const map<int,Quadrature<SPACEDIM-1> >&,NORM_TYPE,
      const FunctionSystem*,REAL);
    REAL integrateDifferenceOnMappedBoundary(const Mapping&,const Vector&,
      const FunctionSystem&,const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,NORM_TYPE,
      const FunctionSystem*,REAL);

#if (SPACEDIM>1)
//  called from ConstraintMatrix::makeHangingNodeConstraints
    void dofsOnChildren(const CODIMENSION1_SHAPE_CLASS*,
      const CODIMENSION1_LEVEL_CLASS*,NumPtr<int>&) const;
#endif
//  called from cuthillMcKeeRenumbering:
    void renumberDofs(const NumPtr<int>&);

//  for eventual call to FiniteElement::sumShapeValues
//    T = double or Vector
//  called from DataOut::buildSomePatches, Patch::plotClipPlane,
//    RefinementStudy::RefinementStudy,
//    RefinementStudy::recursivelyIntegrateError
    template<class T> void getFunctionValues(const FEValues &fe_values,
    const SHAPE_CLASS *e,const LEVEL_CLASS *level,
    const Vector &fe_function,NumPtr<T> &values) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_values.getFunctionValues(dof_values,e,values);
    }
#if (SPACEDIM>1)
    template<class T> void getFunctionValues(
    const FEFaceValues &fe_face_values,const SHAPE_CLASS *e,int f,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &values) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_face_values.getFunctionValues(dof_values,e,f,values);
    }
#endif
//  for eventual call to FiniteElement::sumShapeDerivatives
//    T = Tensor<1,SPACEDIM> or NumPtr<Tensor<1,SPACEDIM>
//  called from KellyErrorEstimator::integrateOverRegularFace and
//    KellyErrorEstimator::integrateOverIrregularFace,
//    KellyErrorEstimator::work, RefinementStudy::RefinementStudy and
//    RefinementStudy::recursivelyIntegrateError
    template<class T> void getFunctionDerivatives(
    const FEValues &fe_values,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &gradients) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_values.getFunctionDerivatives(dof_values,e,gradients);
    }
#if (SPACEDIM>1)
    template<class T> void getFunctionDerivatives(
    const FEFaceValues &fe_face_values,const SHAPE_CLASS *e,int f,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &gradients) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_face_values.getFunctionDerivatives(dof_values,e,f,gradients);
    }
#endif
//  for eventual call to FiniteElement::sumShape2ndDerivatives
//    T = Tensor<2,SPACEDIM> or NumPtr<Tensor<2,SPACEDIM> >
    template<class T> void getFunction2ndDerivatives(
    const FEValues &fe_values,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &second_derivatives) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_values.getFunction2ndDerivatives(dof_values,e,second_derivatives);
    }
#if (SPACEDIM>1)
    template<class T>void getFunction2ndDerivatives(
    const FEFaceValues &fe_face_values,const SHAPE_CLASS *e,int f,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &second_derivatives) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_face_values.getFunction2ndDerivatives(dof_values,e,f,
        second_derivatives);
    }
#endif
    void setValues(const Mapping&,const FunctionSystem*,Vector&) const;
    virtual void printOn(ostream&) const;
    virtual void printDofs() const;
};

#endif
