// "$Header:$"
//----------------------------  dof_renumbering.cc  -----------------------
//  $Id: dof_renumbering.cc,v 1.54 2003/04/21 20:14:03 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  dof_renumbering.cc  -----------------------
//
//modified from deal.II/deal.II/source/dofs/dof_renumbering.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include "DofRenumbering.H"

#include <algorithm>
#include <map>
#include "ConstraintMatrix.H"
#include "Errors.H"
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void computeCuthillMcKeeRenumbering(vector<int> &new_indices,
const DofHandler &dof_handler,bool reversed_numbering,
bool use_constraints,const vector<int> &starting_indices) {
  int n_dofs=dof_handler.numberDofs();
  SparsityPattern sparsity(n_dofs,n_dofs,
    dof_handler.maxCouplingsBetweenDofs());
  dof_handler.makeSparsityPattern(sparsity);

  if (use_constraints) {
    ConstraintMatrix constraints;
    constraints.makeHangingNodeConstraints(dof_handler);
    constraints.close();
    constraints.condense(sparsity);
  }
  CHECK_SAME(new_indices.size(),n_dofs);
  vector<int> last_round_dofs (starting_indices);
  for (int i=0; i<last_round_dofs.size(); ++i) {
    if (last_round_dofs[i]>=n_dofs) {
      last_round_dofs[i]=INT_MAX;
    }
  }
  remove_if(last_round_dofs.begin(),last_round_dofs.end(),
    bind2nd(equal_to<int>(),INT_MAX));
  if (last_round_dofs.size()==0) {
    int starting_point=INT_MAX;
    int min_coordination=n_dofs;
    for (int row=0;row<n_dofs;row++) {
      int j;
      for (j=sparsity.getRowstartIndex(row);
      j<sparsity.getRowstartIndex(row+1); ++j) {
        if (sparsity.getColumnNumber(j)==
        SparsityPattern::invalid_entry) {
          break;
        }
      }
      if (j-sparsity.getRowstartIndex(row)<min_coordination) {
        min_coordination=j-sparsity.getRowstartIndex(row);
        starting_point=row;
      }
    }
    if (starting_point==INT_MAX) starting_point=0;
    last_round_dofs.push_back (starting_point);
  }
  int next_free_number=0;
  for (int i=0;i!=last_round_dofs.size();i++) {
    new_indices[last_round_dofs[i]]=next_free_number++;
  }
  bool all_dofs_renumbered=false;
  while (!all_dofs_renumbered) {
    vector<int> next_round_dofs;
    for (int i=0; i<last_round_dofs.size(); ++i) {
      for (int j=sparsity.getRowstartIndex(last_round_dofs[i]);
      j<sparsity.getRowstartIndex(last_round_dofs[i]+1); ++j) {
        if (sparsity.getColumnNumber(j)==
        SparsityPattern::invalid_entry) {
          break;
        } else {
          next_round_dofs.push_back(sparsity.getColumnNumber(j));
        }
      }
    }
    sort(next_round_dofs.begin(),next_round_dofs.end());
    vector<int>::iterator end_sorted;
    end_sorted=unique(next_round_dofs.begin(),next_round_dofs.end());
    next_round_dofs.erase(end_sorted,next_round_dofs.end());
    for (int s=next_round_dofs.size()-1;s>=0;s--) {
      if (new_indices[next_round_dofs[s]]!=INT_MAX) {
        next_round_dofs.erase (next_round_dofs.begin() + s);
      }
    }
    all_dofs_renumbered=(next_round_dofs.size()==0);
    if (all_dofs_renumbered) continue;
    multimap<int, int> dofs_by_coordination;
    for (vector<int>::iterator s=next_round_dofs.begin();
    s!=next_round_dofs.end();s++) {
      int coordination=0;
      for (int j=sparsity.getRowstartIndex(*s);
      j<sparsity.getRowstartIndex(*s+1);j++) {
        if (sparsity.getColumnNumber(j)==
        SparsityPattern::invalid_entry) {
          break;
        } else ++coordination;
      }
      const pair<int, int> new_entry(coordination,*s);
      dofs_by_coordination.insert(new_entry);
    }
    multimap<int, int>::iterator i;
    for (i = dofs_by_coordination.begin();
    i!=dofs_by_coordination.end();i++) { 
      new_indices[i->second] = next_free_number++;
    }
    last_round_dofs = next_round_dofs;
  }
#ifdef DEBUG
  CHECK_SAME(find(new_indices.begin(),new_indices.end(),INT_MAX),
             new_indices.end());
  CHECK_SAME(next_free_number,n_dofs);
#endif
  if (reversed_numbering) {
    for (vector<int>::iterator i=new_indices.begin();
    i!=new_indices.end();i++) {
      *i=n_dofs-*i-1;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void cuthillMcKeeRenumbering(DofHandler &dof_handler,
bool reversed_numbering,bool use_constraints,
const vector<int> &starting_indices) {
  int n_dofs=dof_handler.numberDofs();
  vector<int> renumbering(n_dofs,INT_MAX);
  computeCuthillMcKeeRenumbering(renumbering,dof_handler,
    reversed_numbering,use_constraints,starting_indices);
  NumPtr<int> my_renumbering(renumbering.size());
  for (int i=0;i<my_renumbering.getNumber();i++) {
    my_renumbering[i]=renumbering[i];
  }
  dof_handler.renumberDofs(my_renumbering);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> remove_if<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  binder2nd<equal_to<int> > 
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  binder2nd<equal_to<int> >
); 
template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> unique<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > > 
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >
);
template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> adjacent_find<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > > 
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >
);
template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> __find<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  unsigned int
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  unsigned int const&,
  random_access_iterator_tag
);
template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> __find_if<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  binder2nd<equal_to<int> > 
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  binder2nd<equal_to<int> >, 
  random_access_iterator_tag
);
template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> remove_copy_if<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  binder2nd<equal_to<int> > 
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  binder2nd<equal_to<int> >
);
template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> fill_n<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  unsigned int, 
  int
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  unsigned int, 
  int const&
);
template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> __find<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  int
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  int const&, 
  random_access_iterator_tag
);

template void fill<
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  int
>(
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >, 
  int const&
);

template class _Rb_tree<
  int, 
  pair<int const, int>, 
  _Select1st<pair<int const, int> >, 
  less<int>, 
  allocator<pair<int const, int> > 
>;

template class vector<int, allocator<int> >;
