// "$Header:$"
//----------------------------  fe_values.cc  ---------------------------
//  $Id: fe_values.cc,v 1.113 2003/01/17 19:24:49 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  fe_values.cc  ---------------------------
//
//modified from deal.II/deal.II/source/fe/fe_values.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <FEValues.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//PROJECT_TO_FACE calls this and the next routine with different flags
FEValuesData::FEValuesData(int nv,int nqp,const FiniteElement &fe,
UPDATE_FLAGS uf) : update_flags(uf),number_quadrature_points(nqp),
number_vertices(nv),shape_values_ok(false),shape_gradients_ok(false),
shape_2nd_derivatives_ok(false),JxW_values_ok(false),
quadrature_points_ok(false),finite_element(&fe) {
  int n_nonzero_shape_components=
    finite_element->nNonzeroShapeComponents(number_vertices);
  if (update_flags & UPDATE_VALUES) {
    shape_values.reinit(n_nonzero_shape_components,
                        number_quadrature_points);
#ifdef INDEF
    shape_values.fillConstant(HUGE_VAL);
#endif
  }
  if (update_flags & UPDATE_GRADIENTS) {
    shape_gradients.reinit(n_nonzero_shape_components,
                           number_quadrature_points);
#ifdef INDEF
    Tensor<1,SPACEDIM> bogus_tensor;
      bogus_tensor.initialize(HUGE_VAL);
    shape_gradients.fillConstant(bogus_tensor);
#endif
  }

  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
    shape_2nd_derivatives.reinit(n_nonzero_shape_components,
                                 number_quadrature_points);
#ifdef INDEF
    Tensor<2,SPACEDIM> bogus_2tensor;
      bogus_2tensor.initialize(HUGE_VAL);
    shape_2nd_derivatives.fillConstant(bogus_2tensor);
#endif
  }
  if (update_flags & UPDATE_Q_POINTS) {
    quadrature_points.cleanup();
    quadrature_points.allocate(number_quadrature_points);
#ifdef INDEF
    Point<SPACEDIM> bogus_point(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
    quadrature_points.initialize(bogus_point);
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FEValuesData::initializeJxWValues(int nqp,UPDATE_FLAGS update_flags)
{
  if (update_flags & UPDATE_JxW_VALUES) {
    JxW_values.cleanup();
    JxW_values.allocate(nqp);
#ifdef INDEF
    JxW_values.initialize(HUGE_VAL);
#endif
    JxW_values_ok=false;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FEValuesData::mapShapes(MappingDataBase *mapping_data,
FiniteElementDataBase *fe_data,const SHAPE_CLASS *e,
const Tessellation &tess,const LEVEL_CLASS *level) {
  shape_values_ok=false;
  shape_gradients_ok=false;
  shape_2nd_derivatives_ok=false;
  JxW_values_ok=false;
  quadrature_points_ok=false;
#ifdef INDEF
  JxW_values.initialize(HUGE_VAL);
  shape_values.fillConstant(HUGE_VAL);
  {
    Tensor<1,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    shape_gradients.fillConstant(bogus_tensor);
  }
  {
    Tensor<2,SPACEDIM> bogus_2tensor;
    bogus_2tensor.initialize(HUGE_VAL);
    shape_2nd_derivatives.fillConstant(bogus_2tensor);
  }
#endif
  mapping_data->computeFEValues(e,tess,level,quadrature_points,
    JxW_values,quadrature_points_ok,JxW_values_ok);
  FiniteElementDataBase *fedb=
    finite_element->fixShapeFunctionsFor(e,fe_data);
  if (update_flags & UPDATE_VALUES) {
    for (int s=0;s<shape_values.size(0);s++) {
      for (int q=0;q<shape_values.size(1);q++) {
        shape_values(s,q)=fedb->shapeValue(s,q);
      }
    }
    shape_values_ok=true;
  }
  if (update_flags & UPDATE_GRADIENTS) {
    for (int s=0;s<shape_gradients.size(0);s++) {
      for (int q=0;q<shape_gradients.size(1);q++) {
        mapping_data->postMultiplyByJacobianInverse(
          shape_gradients(s,q),q,fedb->shapeDerivative(s,q));
      }
    }
    shape_gradients_ok=true;
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
//  requires second derivatives of FE shapes and mapping transformation
    OBSOLETE("not programmed");
    shape_2nd_derivatives_ok=true;
  }
  if (fedb!=fe_data) delete fedb;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEValuesData::mapFaceShapes(MappingDataBase *mapping_data,
FiniteElementDataBase *fe_data,const SHAPE_CLASS *e,
int face_no,const Tessellation &tess,const LEVEL_CLASS *level,
const Quadrature<SPACEDIM-1> &fq,
NumPtr<Tensor<1,SPACEDIM> > &boundary_forms,
NumPtr<Point<SPACEDIM> > &normal_vectors,
int offset,bool &boundary_forms_ok,bool &normal_vectors_ok) {
  CHECK_TEST(face_no<e->numberCodimension1Shapes());
  int nq=fq.numberPoints();
  bool no_jxw=JxW_values.getNumber()==0;
  NumPtr<REAL> face_jxw_values(JxW_values,(no_jxw ? 0 : offset),
    (no_jxw ? 0 : nq));

#ifdef INDEF
  face_jxw_values.initialize(HUGE_VAL);
  {
    Tensor<1,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    boundary_forms.initialize(bogus_tensor);
  }
  {
    Point<SPACEDIM> bogus_point(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
    normal_vectors.initialize(bogus_point);
  }
#endif
  bool no_points=quadrature_points.getNumber()==0;
  NumPtr<Point<SPACEDIM> > qp(quadrature_points,(no_points ? 0 : offset),
    (no_points ? 0 : nq));
  bool is_subface=false;
  mapping_data->computeFEFaceValues(e,tess,level,face_no,is_subface,
    nq,offset,fq.weights(),qp,face_jxw_values,boundary_forms,
    normal_vectors,quadrature_points_ok,JxW_values_ok,boundary_forms_ok,
    normal_vectors_ok);
  FiniteElementDataBase *fedb=
    finite_element->fixShapeFunctionsFor(e,fe_data);
  if (update_flags & UPDATE_VALUES) {
    for (int s=0;s<shape_values.size(0);s++) {
      for (int q=0;q<nq;q++) {
        shape_values(s,q+offset)=fedb->shapeValue(s,q+offset);
      }
    }
    shape_values_ok=true;
  }
  if (update_flags & UPDATE_GRADIENTS) {
    for (int s=0;s<shape_gradients.size(0);s++) {
      for (int q=0;q<nq;q++) {
        mapping_data->postMultiplyByJacobianInverse(
          shape_gradients(s,q+offset),q,fedb->shapeDerivative(s,q+offset));
      }
    }
    shape_gradients_ok=true;
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
//  requires second derivatives of FE shapes and mapping transformation
    OBSOLETE("not programmed");
    shape_2nd_derivatives_ok=true;
  }
  if (fedb!=fe_data) delete fedb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEValuesData::mapSubFaceShapes(MappingDataBase *mapping_data,
FiniteElementDataBase *fe_data,const SHAPE_CLASS *e,
int face_no,const Tessellation &tess,const LEVEL_CLASS *level,
const Quadrature<SPACEDIM-1> &fq,
NumPtr<Tensor<1,SPACEDIM> > &boundary_forms,
NumPtr<Point<SPACEDIM> > &normal_vectors,
int fq_offset,int offset,bool &boundary_forms_ok,
bool &normal_vectors_ok) {
  CHECK_TEST(face_no<e->numberCodimension1Shapes());
  int nq=fq.numberPoints();
  bool no_jxw=JxW_values.getNumber()==0;
  NumPtr<REAL> face_jxw_values(JxW_values,(no_jxw ? 0 : fq_offset),
    (no_jxw ? 0 : nq));

#ifdef INDEF
  face_jxw_values.initialize(HUGE_VAL);
  {
    Tensor<1,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    boundary_forms.initialize(bogus_tensor);
  }
  {
    Point<SPACEDIM> bogus_point(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
    normal_vectors.initialize(bogus_point);
  }
#endif
  bool no_points=quadrature_points.getNumber()==0;
  NumPtr<Point<SPACEDIM> > qp(quadrature_points,
    (no_points ? 0 : fq_offset),(no_points ? 0 : nq));
  bool is_subface=true;
  mapping_data->computeFEFaceValues(e,tess,level,face_no,is_subface,
    fq.numberPoints(),offset,fq.weights(),qp,face_jxw_values,
    boundary_forms,normal_vectors,quadrature_points_ok,JxW_values_ok,
    boundary_forms_ok,normal_vectors_ok);
  FiniteElementDataBase *fedb=
    finite_element->fixShapeFunctionsFor(e,fe_data);
  if (update_flags & UPDATE_VALUES) {
    for (int s=0;s<shape_values.size(0);s++) {
      for (int q=0;q<nq;q++) {
        shape_values(s,q+offset)=fedb->shapeValue(s,q+offset);
      }
    }
    shape_values_ok=true;
  }
  if (update_flags & UPDATE_GRADIENTS) {
    for (int s=0;s<shape_gradients.size(0);s++) {
      for (int q=0;q<nq;q++) {
        mapping_data->postMultiplyByJacobianInverse(
          shape_gradients(s,q+offset),q,fedb->shapeDerivative(s,q+offset));
      }
    }
    shape_gradients_ok=true;
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
//  requires second derivatives of FE shapes and mapping transformation
    OBSOLETE("not programmed");
    shape_2nd_derivatives_ok=true;
  }
  if (fedb!=fe_data) delete fedb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FEValuesData::printOn(ostream &os) const {
  os << "\tFEValuesData: this = " << this << endl;
//os << "\t\tshape_values is " << shape_values.size(0) << "x" 
//   << shape_values.size(1) << endl;
//for (int i=0;i<shape_values.size(0);i++) {
//  for (int j=0;j<shape_values.size(1);j++) {
//    os << "\t\t\tshape_value[" << i << "," << j << "] = "
//       << shape_values.operator()(i,j) << endl;
//  }
//}
//os << "\t\tshape_gradients is " << shape_gradients.size(0) << "x" 
//   << shape_gradients.size(1) << endl;
//for (int i=0;i<shape_gradients.size(0);i++) {
//  for (int j=0;j<shape_gradients.size(1);j++) {
//    os << "\t\t\tshape_gradients[" << i << "," << j << "] = "
//       << shape_gradients.operator()(i,j) << endl;
//  }
//}
//os << "\t\tshape_2nd_derivatives is "
//   << shape_2nd_derivatives.size(0) << "x" 
//   << shape_2nd_derivatives.size(1) << endl;
//os << "\t\tJxW_values has "
//   << JxW_values.getNumber() << " entries" << endl;
//for (int i=0;i<JxW_values.getNumber();i++) {
//  os << "\t\t\tJxW_values[" << i << "] = " << JxW_values[i] << endl;
//}
  os << "\t\tJxW_values has "
     << JxW_values.getNumber() << " entries" << endl;
  for (int i=0;i<JxW_values.getNumber();i++) {
    os << "\t\t\tJxW_values[" << i << "] = " << JxW_values[i] << endl;
  }
//os << "\t\tquadrature_points has "
//   << quadrature_points.getNumber() << " entries" << endl;
//for (int i=0;i<quadrature_points.getNumber();i++) {
//  os << "\t\t\tquadrature_point[" << i << "] = " 
//     << quadrature_points[i] << endl;
//}
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
FEValuesBase::~FEValuesBase() {
  for (map<int,FEValuesData*>::iterator
  fevd_iterator=fe_values_data_map.begin();
  fevd_iterator!=fe_values_data_map.end();fevd_iterator++) {
    delete fevd_iterator->second; fevd_iterator->second=0;
  }
  for (map<int,FiniteElementDataBase*>::iterator
  fedm_iterator=fe_data_map.begin();
  fedm_iterator!=fe_data_map.end();fedm_iterator++) {
    delete fedm_iterator->second; fedm_iterator->second=0;
  }
  for (map<int,MappingDataBase*>::iterator
  mdm_iterator=mapping_data_map.begin();
  mdm_iterator!=mapping_data_map.end();mdm_iterator++) {
    delete mdm_iterator->second; mdm_iterator->second=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FEValuesBase::printOn(ostream &os) const {
  os << "\tFEValuesBase :" << endl;
//cout << "\tmapping_data_map args = ";
//for (map<int,MappingDataBase*>::const_iterator md_iterator=
//mapping_data_map.begin();md_iterator!=mapping_data_map.end();
//md_iterator++) {
//  cout << md_iterator->first << " ";
//}
//cout << endl;
  for (map<int,MappingDataBase*>::const_iterator md_iterator=
  mapping_data_map.begin();md_iterator!=mapping_data_map.end();
  md_iterator++) {
    os << "\t\tmapping_data_map[" << md_iterator->first << "] = "
       << md_iterator->second << endl;
    md_iterator->second->printOn(os);
  }
//for (map<int,FiniteElementDataBase*>::const_iterator fed_iterator=
//fe_data_map.begin();fed_iterator!=fe_data_map.end();fed_iterator++) {
//  os << "\t\tfe_data_map[" << fed_iterator->first << "] = "
//     << fed_iterator->second << endl;
//  fed_iterator->second->printOn(os);
//}
//for (map<int,FEValuesData*>::const_iterator fevd_iterator=
//fe_values_data_map.begin();fevd_iterator!=fe_values_data_map.end();
//fevd_iterator++) {
//  os << "\t\tfe_values_data_map[" << fevd_iterator->first << "] = "
//     << fevd_iterator->second << endl;
//  fevd_iterator->second->printOn(os);
//}
//for (map<int,int>::const_iterator nqp_iterator=
//n_quadrature_points.begin();nqp_iterator!=n_quadrature_points.end();
//nqp_iterator++) {
//  os << "\t\tn_quadrature_points[" << nqp_iterator->first << "] = "
//     << nqp_iterator->second << endl;
//}
//for (map<int,int>::const_iterator dpc_iterator=
//dofs_per_cell.begin();dpc_iterator!=dofs_per_cell.end();
//dpc_iterator++) {
//  os << "\t\tdofs_per_cell[" << dpc_iterator->first << "] = "
//     << dpc_iterator->second << endl;
//}
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this is called outside a loop over grid elements
//  computes mapping shape functions at canonical quadrature points
//    and stores in mapping_data_map[nv]
//  computes finite element shape functions at canonical quadrature pts
//    and stores in fe_data_map[nv]
//  sets shape_function_to_row_table, allocates shape_values, 
//  shape_gradients, quadrature_points, JxW_values
//    these are stored in FEValuesData
FEValues::FEValues(const Mapping &m,FiniteElement &fe_in,
const map<int,Quadrature<SPACEDIM> > &q,UPDATE_FLAGS flags) :
FEValuesBase(flags,m,fe_in) {
  if (update_flags & UPDATE_GRADIENTS) {
    update_flags |= UPDATE_MAPPING_JACOBIAN_INVERSE;
  }
  for (map<int,Quadrature<SPACEDIM> >::const_iterator
  q_iterator=q.begin();q_iterator!=q.end();q_iterator++) {
    int nv=q_iterator->first;
    const Quadrature<SPACEDIM> &q_nv=q_iterator->second;
    insertNQuadraturePoint(nv,q_nv.numberPoints());
    insertDofsPerCell(nv,fe_in.dofsPerCell(nv));
    quadrature[nv]=q_nv;

    mapping_data_map[nv]=mapping->getMappingData(nv,update_flags,&q_nv);
    fe_data_map[nv]=fe->getFEData(nv,update_flags,&q_nv);
    int np=q_nv.numberPoints();
    FEValuesData *fevd=fe_values_data_map[nv]=
      OPERATOR_NEW FEValuesData(nv,np,*fe,update_flags);
    fevd->initializeJxWValues(np,update_flags);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FEValues::printOn(ostream &os) const {
  os << "\tFEValues:" << endl;
//for (map<int,Quadrature<SPACEDIM> >::const_iterator 
//q_iterator=quadrature.begin();q_iterator!=quadrature.end();
//q_iterator++) {
//  os << "\t\tquadrature[" << q_iterator->first << "] has "
//     << q_iterator->second.numberPoints() << " points" << endl;
//}
  FEValuesBase::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
int FEFaceValues::faceOffset(const SHAPE_CLASS *e,int face_no) const {
  int offset=0;
  for (int f=0;f<face_no;f++) {
    int vpf=e->numberVerticesOnFace(f);
    offset+=getFaceQuadrature(vpf).numberPoints();
  }
  return offset;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues::initializeBoundaryFormsAndNormals(int nv,
const Quadrature<SPACEDIM-1> &qf) {
  int nqp=qf.numberPoints();
  if (update_flags & UPDATE_BOUNDARY_FORMS) {
    CHECK_TEST(boundary_forms.find(nv)==boundary_forms.end());
    NumPtr<Tensor<1,SPACEDIM> > &bf=boundary_forms[nv];
    bf.cleanup();
    bf.allocate(nqp);
  }
  if (update_flags & UPDATE_NORMAL_VECTORS) {
    CHECK_TEST(normal_vectors.find(nv)==normal_vectors.end());
    NumPtr<Point<SPACEDIM> > &normals=normal_vectors[nv];
    normals.cleanup();
    normals.allocate(nqp);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
//for each possible shape in this number of dimensions,
//  generates a Quadrature on all faces
//  compute unit tangents, shapes, etc at quadrature points
FEFaceValues::FEFaceValues(const Mapping &m,FiniteElement &fe_in,
map<int,Quadrature<SPACEDIM-1> > &q,UPDATE_FLAGS flags_in) : 
FEValuesBase(flags_in,m,fe_in),boundary_forms_ok(false),
normal_vectors_ok(false) {
  for (map<int,Quadrature<SPACEDIM-1> >::const_iterator
  q_iterator=q.begin();q_iterator!=q.end();q_iterator++) {
    int nv=q_iterator->first;
    const Quadrature<SPACEDIM-1> &q_nv=q_iterator->second;
    face_quadrature[nv]=q_nv;
  }
  update_flags |= UPDATE_BOUNDARY_FORMS;
  if (update_flags & UPDATE_GRADIENTS) {
    update_flags |= UPDATE_MAPPING_JACOBIAN_INVERSE;
  }
  const map<int,int> &dpc_map=fe->dofsPerCellMap();
  for (map<int,int>::const_iterator dpc_iterator=
  dpc_map.begin();dpc_iterator!=dpc_map.end();dpc_iterator++) {
    int nv=dpc_iterator->first;
//original DEAL code defined mapping_data and fe_data for projected
//  points around all faces, then called FEValuesData::initialize
//  for a single face
//this worked because all elements were Quadrilaterals/Hexahegons
//  and all faces were Intervals/Quadrilaterals with the same number
//  of shape functions on all faces
//in this code, we have the same number of quadrature points on all
//  faces of the same shape, but not necessarily the same number of
//  element shape functions to be evaluated on a face of a given shape
//instead, we call FEValuesData::initialize for the collection of
//  projected quadrature points around all faces
//this necessitates changes in FiniteElement::fillFEFaceValues and
//  Mapping::fillFEFaceValues
    int nfp=0;
    for (int f=0;f<facesPerShape(nv);f++) {
      int vpf=numberVerticesOnFace(nv,f);
      map<int,Quadrature<SPACEDIM-1> >::const_iterator
        q_iterator=q.find(vpf);
      if (q_iterator==q.end()) {
//      main program did not expect a face with vpf vertices, but
//        FiniteElement constructor had to allow for all Shapes
//      so, we construct a bogus Quadrature for vpf: point=0, weight=inf
        NumPtr<Point<SPACEDIM-1> > points(1);
        for (int d=0;d<SPACEDIM-1;d++) points[0][d]=0.;
        NumPtr<REAL> weights(1);
        weights[0]=HUGE_VAL;
        q[vpf]=Quadrature<SPACEDIM-1>(points,weights);
        q_iterator=q.find(vpf);
      }
      nfp+=q_iterator->second.numberPoints();
    }
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<facesPerShape(nv);f++) {
      int vpf=numberVerticesOnFace(nv,f);
      map<int,Quadrature<SPACEDIM-1> >::const_iterator
        q_iterator=q.find(vpf);
      int nqp=q_iterator->second.numberPoints();
      projectToFace(nv,f,q_iterator->second.points(),q_points_face);
      weights_face.copyFrom(q_iterator->second.weights());
      q_points_face+=nqp;
      weights_face+=nqp;
    }
    Quadrature<SPACEDIM> *qf=all_face_quadrature[nv]=
      OPERATOR_NEW Quadrature<SPACEDIM>(q_points,weights);
    MappingDataBase *mdb=mapping_data_map[nv]=
      mapping->getMappingData(nv,flags_in,qf);
    FiniteElementDataBase *fed=fe_data_map[nv]=
      fe->getFEData(nv,flags_in,qf);
    FEValuesData *fe_values_data=fe_values_data_map[nv]=
      OPERATOR_NEW FEValuesData(nv,qf->numberPoints(),*fe,
                                flags_in^UPDATE_JxW_VALUES);
    insertNQuadraturePoint(nv,nfp);
    insertDofsPerCell(nv,fe_in.dofsPerCell(nv));
    fe_values_data->initializeJxWValues(nfp,flags_in);
  }
  for (map<int,Quadrature<SPACEDIM-1> >::const_iterator
  q_iterator=q.begin();q_iterator!=q.end();q_iterator++) {
    int nv=q_iterator->first;
    initializeBoundaryFormsAndNormals(nv,q_iterator->second);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
FEFaceValues::~FEFaceValues() {
  for (map<int,Quadrature<SPACEDIM>*>::iterator afq_iterator=
  all_face_quadrature.begin();afq_iterator!=all_face_quadrature.end();
  afq_iterator++) {
    delete afq_iterator->second; afq_iterator->second=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues::mapShapes(const Tessellation &tess,
const SHAPE_CLASS *e,const LEVEL_CLASS *level,int f) {
  int nv=e->numberVerticesOnFace(f);
  NumPtr<Point<SPACEDIM> > empty_nvec;
  NumPtr<Tensor<1,SPACEDIM> > empty_bf;

  map<int,NumPtr<Tensor<1,SPACEDIM> > >::iterator
    bf_iterator=boundary_forms.find(nv);
  map<int,NumPtr<Point<SPACEDIM> > >::iterator
    nv_iterator=normal_vectors.find(nv);

  const Quadrature<SPACEDIM-1> &fq=getFaceQuadrature(e,f);
  getFEValuesData(e)->mapFaceShapes(mappingData(e),
    getFEData(e),e,f,tess,level,fq,
    (update_flags & UPDATE_BOUNDARY_FORMS ? bf_iterator->second : 
                    empty_bf),
    (update_flags & UPDATE_NORMAL_VECTORS ? nv_iterator->second : 
                    empty_nvec),
    faceOffset(e,f),boundary_forms_ok,normal_vectors_ok);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues::getFunctionValues(const Vector &dof_values,
const SHAPE_CLASS *e,int face_no,NumPtr<REAL> &values) const {
  CHECK_TEST(update_flags & UPDATE_VALUES); 
  int offset=faceOffset(e,face_no);
  int nqp=getFaceQuadrature(e,face_no).numberPoints();
  fe->sumShapeValues(e,*getShapeValues(e),nqp,faceOffset(e,face_no),
    dof_values,values);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues::getFunctionValues(const Vector &dof_values,
const SHAPE_CLASS *e,int face_no,NumPtr<Vector> &values) const {
  CHECK_TEST(update_flags & UPDATE_VALUES); 
  int nqp=getFaceQuadrature(e,face_no).numberPoints();
  int offset=faceOffset(e,face_no);
  const Array<2,REAL> &sv=*getShapeValues(e);
  fe->sumShapeValues(e,sv,nqp,offset,dof_values,values);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues::getFunctionDerivatives(const Vector &dof_values,
const SHAPE_CLASS *e,int face_no,NumPtr<Tensor<1,SPACEDIM> > &gradients) 
const {
  CHECK_TEST(update_flags & UPDATE_GRADIENTS); 
  int offset=faceOffset(e,face_no);
  int nqp=getFaceQuadrature(e,face_no).numberPoints();
  const Array<2,Tensor<1,SPACEDIM> > &sg=*getShapeGradients(e);
  fe->sumShapeDerivatives(e,sg,nqp,offset,dof_values,gradients);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues::getFunctionDerivatives(const Vector &dof_values,
const SHAPE_CLASS *e,int face_no,
NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &gradients) const {
  CHECK_TEST(update_flags & UPDATE_GRADIENTS); 
  int nqp=getFaceQuadrature(e,face_no).numberPoints();
  int offset=faceOffset(e,face_no);
  const Array<2,Tensor<1,SPACEDIM> > &sg=*getShapeGradients(e);
  fe->sumShapeDerivatives(e,sg,nqp,offset,dof_values,gradients);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues::getFunction2ndDerivatives(const Vector &dof_values,
const SHAPE_CLASS *e,int face_no,
NumPtr<Tensor<2,SPACEDIM> > &second_derivatives) const {
  CHECK_TEST(update_flags & UPDATE_SECOND_DERIVATIVES); 
  int offset=faceOffset(e,face_no);
  int nqp=getFaceQuadrature(e,face_no).numberPoints();
  const Array<2,Tensor<2,SPACEDIM> > &s2d=*getShape2ndDerivatives(e);
  fe->sumShape2ndDerivatives(e,s2d,nqp,offset,dof_values,
    second_derivatives);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues:: getFunction2ndDerivatives(const Vector &dof_values,
const SHAPE_CLASS *e,int face_no,
NumPtr<NumPtr<Tensor<2,SPACEDIM> > > &second_derivs) const {
  CHECK_TEST(update_flags & UPDATE_SECOND_DERIVATIVES); 
  int nqp=getFaceQuadrature(e,face_no).numberPoints();
  int offset=faceOffset(e,face_no);
  const Array<2,Tensor<2,SPACEDIM> > &s2d=*getShape2ndDerivatives(e);
  fe->sumShape2ndDerivatives(e,s2d,nqp,offset,dof_values,second_derivs);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void FEFaceValues::printOn(ostream &os) const {
  os << "\tFEFaceValues:" << endl;
  for (map<int,Quadrature<SPACEDIM-1> >::const_iterator
  fq_iterator=face_quadrature.begin();
  fq_iterator!=face_quadrature.end();fq_iterator++) {
    os << "\t\tface_quadrature[" << fq_iterator->first << "] has "
       << fq_iterator->second.numberPoints() << " points" << endl;
    fq_iterator->second.printOn(os);
  }
//if (normal_vectors_ok) {
//  for (map<int,NumPtr<Point<SPACEDIM> > >::const_iterator
//  nv_iterator=normal_vectors.begin();
//  nv_iterator!=normal_vectors.end();nv_iterator++) {
//    const NumPtr<Point<SPACEDIM> > &nv=nv_iterator->second;
//    os << "\t\tnormal_vectors[" << nv_iterator->first << "] has "
//       << nv.getNumber() << " entries" << endl;
//    for (int i=0;i<nv.getNumber();i++) {
//      os << "\t\t\tnormal_vector[" << i << "] = " << nv[i] << endl;
//    }
//  }
//}
//if (boundary_forms_ok) {
//  for (map<int,NumPtr<Tensor<1,SPACEDIM> > >::const_iterator
//  bf_iterator=boundary_forms.begin();
//  bf_iterator!=boundary_forms.end();bf_iterator++) {
//    const NumPtr<Tensor<1,SPACEDIM> > &bf=bf_iterator->second;
//    os << "\t\tboundary_forms[" << bf_iterator->first << "] has "
//       << bf.getNumber() << " entries" << endl;
//    for (int i=0;i<bf.getNumber();i++) {
//      os << "\t\t\tboundary_form[" << i << "] = " << bf[i] << endl;
//    }
//  }
//}
  FEValuesBase::printOn(os);
}
#endif

template class _Rb_tree<
  int, 
  pair<int const, FEValuesData*>, 
  _Select1st<pair<int const, FEValuesData*> >, 
  less<int>, 
  allocator<pair<int const, FEValuesData*> > 
>;
template class _Rb_tree<
  int, 
  pair<int const, Quadrature<SPACEDIM> >, 
  _Select1st<pair<int const, Quadrature<SPACEDIM> > >, 
  less<int>, 
  allocator<pair<int const, Quadrature<SPACEDIM> > > 
>;
template class _Rb_tree<
  int, 
  pair<int const, Quadrature<SPACEDIM-1> >, 
  _Select1st<pair<int const, Quadrature<SPACEDIM-1> > >, 
  less<int>, 
  allocator<pair<int const, Quadrature<SPACEDIM-1> > > 
>;
template class _Rb_tree<int, int, _Identity<int>, 
         less<int>, allocator<int> >;
template class _Rb_tree<
  int, 
  pair<int const, MappingDataBase*>, 
  _Select1st<pair<int const, MappingDataBase*> >, 
  less<int>, 
  allocator<pair<int const, MappingDataBase*> > 
>;
template class _Rb_tree<
  int, 
  pair<int const, FiniteElementDataBase*>, 
  _Select1st<pair<int const, FiniteElementDataBase*> >, 
  less<int>, 
  allocator<pair<int const, FiniteElementDataBase*> > 
>;
#if (SPACEDIM>=2)
template class _Rb_tree<
  int, 
  pair<int const, NumPtr<Tensor<1, SPACEDIM> > >, 
  _Select1st<pair<int const, NumPtr<Tensor<1, SPACEDIM> > > >, 
  less<int>, 
  allocator<pair<int const, NumPtr<Tensor<1, SPACEDIM> > > > 
>;
#endif
