#ifndef FE_VALUES_H
#define FE_VALUES_H "$Header:$"
//---------------------------  fe_values.h  ---------------------------
//   $Id: fe_values.h,v 1.94 2003/03/11 16:50:52 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  fe_values.h  ---------------------------
//
//modified from deal.II/deal.II/include/fe/fe_values.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <algorithm>
#include <Array.H>
#include <FiniteElement.H>
#include <Quadrature.H>
#include <Tensor.H>
#include <Types.H>

//data for finite element shape function(x);
//  ie, mapped from FiniteElementDataBase values
//  the mapping affects derivatives, not values
//the functions in this class should be used only by FEValuesBase and 
//  derived classes
//compare to DEAL FEValuesData
class FEValuesData {
  protected:
//  these are set in
//    FEFaceValues::mapShapes
//    FESubfaceValues::mapShapes
//    FEValues::mapShapes
    UPDATE_FLAGS update_flags;
    int number_quadrature_points;
    int number_vertices;
    bool shape_values_ok;
    bool shape_gradients_ok;
    bool shape_2nd_derivatives_ok;
    bool JxW_values_ok;
    bool quadrature_points_ok;
    const FiniteElement *finite_element;
    Array<2,REAL> shape_values;
    Array<2,Tensor<1,SPACEDIM> > shape_gradients;
    Array<2,Tensor<2,SPACEDIM> > shape_2nd_derivatives;
    NumPtr<REAL> JxW_values;
    NumPtr<Point<SPACEDIM> > quadrature_points;
  public:
    FEValuesData(int,int,const FiniteElement&,UPDATE_FLAGS);
    virtual ~FEValuesData() {}
//  called from FEFaceValues::FEFaceValues and 
//    FESubfaceValues::FESubfaceValues
    void initializeJxWValues(int,UPDATE_FLAGS);
//  called from FEValues::mapShapes
    void mapShapes(MappingDataBase*,FiniteElementDataBase*,
      const SHAPE_CLASS*,const Tessellation&,const LEVEL_CLASS*);
//  called from FEFaceValues::mapShapes
    void mapFaceShapes(MappingDataBase*,FiniteElementDataBase*,
      const SHAPE_CLASS*,int,const Tessellation&,const LEVEL_CLASS*,
      const Quadrature<SPACEDIM-1>&,NumPtr<Tensor<1,SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&,
//    const FiniteElement&,UPDATE_FLAGS,
      int,bool&,bool&);
//  called from FESubfaceValues::mapShapes
    void mapSubFaceShapes(MappingDataBase*,FiniteElementDataBase*,
      const SHAPE_CLASS*,int,const Tessellation&,const LEVEL_CLASS*,
      const Quadrature<SPACEDIM-1>&,NumPtr<Tensor<1,SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&,
//    const FiniteElement&,UPDATE_FLAGS,
      int,int,bool&,bool&);
//  in dof order; called from FEValuesBase::getShapeValues
    const Array<2,REAL>& getShapeValues() const { 
      CHECK_TEST(shape_values_ok);
      return shape_values; 
    }
//  in dof order; called from FEValuesBase::getShapeGradients
    const Array<2,Tensor<1,SPACEDIM> >& getShapeGradients() const { 
      CHECK_TEST(shape_gradients_ok);
      return shape_gradients;
    }
//  in dof order; called from FEValuesBase::getShape2ndDerivatives
    const Array<2,Tensor<2,SPACEDIM> >& getShape2ndDerivatives() const { 
      CHECK_TEST(shape_2nd_derivatives_ok);
      return shape_2nd_derivatives;
    }
//  called from FEValuesBase::getJxWValues
    const NumPtr<REAL>& getJxWValues() const { 
      CHECK_TEST(JxW_values_ok);
      return JxW_values; 
    }
//  called from FEValuesBase::getQuadraturePoints
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints() const {
      CHECK_TEST(quadrature_points_ok);
      return quadrature_points;
    }
    virtual void printOn(ostream&) const;
};

//compare to DEAL FEValuesBase
class FEValuesBase {
  protected:
    map<int,int> n_quadrature_points;
    map<int,int> dofs_per_cell;
    map<int,MappingDataBase*> mapping_data_map;
    map<int,FiniteElementDataBase*> fe_data_map;
    map<int,FEValuesData*> fe_values_data_map;
    const Mapping *mapping;
    FiniteElement *fe;
    UPDATE_FLAGS update_flags;

    void insertNQuadraturePoint(int nv,int nqp) {
      n_quadrature_points[nv]=nqp;
    }
    void insertDofsPerCell(int nv,int dpc) {
      dofs_per_cell[nv]=dpc;
    }
    const FEValuesData* getFEValuesData(int nv) const {
      map<int,FEValuesData*>::const_iterator fevdm_iterator=
        fe_values_data_map.find(nv);
      if (fevdm_iterator!=fe_values_data_map.end()) {
        return fevdm_iterator->second;
      }
      return 0;
    }
    const FEValuesData* getFEValuesData(const Shape *e) const {
      return getFEValuesData(e->numberVertices());
    }
    FEValuesData* getFEValuesData(int nv) {
      map<int,FEValuesData*>::iterator fevdm_iterator=
        fe_values_data_map.find(nv);
      if (fevdm_iterator!=fe_values_data_map.end()) {
        return fevdm_iterator->second;
      }
      return 0;
    }
    FEValuesData* getFEValuesData(const Shape *e) {
      return getFEValuesData(e->numberVertices());
    }
    const Array<2,REAL>* getShapeValues(int nv) const { 
      const FEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeValues()) : 0);
    }
    const Array<2,Tensor<1,SPACEDIM> >* getShapeGradients(int nv) const { 
      const FEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeGradients()) : 0);
    }
    const Array<2,Tensor<2,SPACEDIM> >* getShape2ndDerivatives(int nv) 
    const { 
      const FEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShape2ndDerivatives()) 
                             : 0);
    }

    const NumPtr<REAL>& getJxWValues(int nv) const {
      const FEValuesData *fe_values_data=getFEValuesData(nv);
      CHECK_POINTER(fe_values_data);
      return fe_values_data->getJxWValues() ;
    }
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints(int nv) const {
      const FEValuesData *fe_values_data=getFEValuesData(nv);
      CHECK_POINTER(fe_values_data);
      return fe_values_data->getQuadraturePoints() ;
    }
    int dofsPerCell(int nv) const { 
      map<int,int>::const_iterator dpc_iterator=
        dofs_per_cell.find(nv);
      CHECK_TEST(dpc_iterator!=dofs_per_cell.end());
      return dpc_iterator->second;
    }
    FiniteElement& getFe() const { return *fe; }
//  if shape_function->requiresShapeToComputeValues,
//  can't compute values here because we can't get the shape_function
    FiniteElementDataBase* getFEData(const Shape *e) {
      map<int,FiniteElementDataBase*>::iterator
        fedm_iterator=fe_data_map.find(e->numberVertices());
      CHECK_TEST(fedm_iterator!=fe_data_map.end());
      return fedm_iterator->second;
    }
  public:
    FEValuesBase(UPDATE_FLAGS flags,const Mapping &m,
    FiniteElement &fe_in) :
      update_flags(flags),mapping(&m),fe(&fe_in) { }
    virtual ~FEValuesBase();
    const Mapping& getMapping() const { return *mapping; }
    const NumPtr<REAL>& getJxWValues(const Shape *e) const {
      return getJxWValues(e->numberVertices());
    }
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints(const Shape *e) 
    const {
      return getQuadraturePoints(e->numberVertices());
    }
    int dofsPerCell(const Shape *e) const { 
      return dofsPerCell(e->numberVertices());
    }
    int numberQuadraturePoints(const Shape *e) const {
      map<int,int>::const_iterator nqp_iterator=
        n_quadrature_points.find(e->numberVertices());
      CHECK_TEST(nqp_iterator!=n_quadrature_points.end());
      return nqp_iterator->second;
    }
    MappingDataBase* mappingData(int nv) {
      map<int,MappingDataBase*>::iterator
        mdm_iterator=mapping_data_map.find(nv);
      CHECK_TEST(mdm_iterator!=mapping_data_map.end());
      return mdm_iterator->second;
    }
    const MappingDataBase* mappingData(int nv) const {
      map<int,MappingDataBase*>::const_iterator
        mdm_iterator=mapping_data_map.find(nv);
      CHECK_TEST(mdm_iterator!=mapping_data_map.end());
      return mdm_iterator->second;
    }
    MappingDataBase* mappingData(const Shape *e) {
      return mappingData(e->numberVertices());
    }
    const MappingDataBase* mappingData(const Shape *e) const {
      return mappingData(e->numberVertices());
    }
//  if shape_function->requiresShapeToComputeValues,
//  can't compute values here because we can't get the shape_function
    const FiniteElementDataBase* getFEData(const Shape *e) const {
      map<int,FiniteElementDataBase*>::const_iterator
        fedm_iterator=fe_data_map.find(e->numberVertices());
      CHECK_TEST(fedm_iterator!=fe_data_map.end());
      return fedm_iterator->second;
    }
//  called from user programs:
    const Array<2,REAL>* getShapeValues(const Shape *e) const { 
      return getShapeValues(e->numberVertices());
    }
    const Array<2,Tensor<1,SPACEDIM> >* getShapeGradients(
    const Shape *e) const { 
      return getShapeGradients(e->numberVertices());
    }
    const Array<2,Tensor<2,SPACEDIM> >* getShape2ndDerivatives(
    const Shape *e) const { 
      return getShape2ndDerivatives(e->numberVertices());
    }
    virtual void printOn(ostream&) const;
};

//compare to DEAL FEValues
class FEValues : public FEValuesBase {
  private:
    map<int,Quadrature<SPACEDIM> > quadrature;

    Quadrature<SPACEDIM>& getQuadrature(int nv) {
      map<int,Quadrature<SPACEDIM> >::iterator q_iterator
        =quadrature.find(nv);
      CHECK_TEST(q_iterator!=quadrature.end());
      return q_iterator->second;
    }
    const Quadrature<SPACEDIM>& getQuadrature(int nv) const {
      map<int,Quadrature<SPACEDIM> >::const_iterator q_iterator
        =quadrature.find(nv);
      CHECK_TEST(q_iterator!=quadrature.end());
      return q_iterator->second;
    }
  public:
    FEValues(const Mapping&,FiniteElement&,
      const map<int,Quadrature<SPACEDIM> >&,UPDATE_FLAGS);
    virtual ~FEValues() {;}
//  compute mapped values of shape functions/derivatives in dof order
    void mapShapes(const Tessellation &tess,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level) {
      getFEValuesData(e)->
        mapShapes(mappingData(e),getFEData(e),e,tess,level);
    }
    const Quadrature<SPACEDIM>& getQuadrature(const Shape *e) const {
      return getQuadrature(e->numberVertices());
    }
//  for eventual call to FiniteElement::sumShapeValues
//    T = double or Vector
    template<class T> void getFunctionValues(const Vector &dof_values,
    const SHAPE_CLASS *e,NumPtr<T> &values) const {
      CHECK_TEST(update_flags & UPDATE_VALUES);
      fe->sumShapeValues(e,*getShapeValues(e),numberQuadraturePoints(e),
        0,dof_values,values);
    }
//  for eventual call to FiniteElement::sumShapeDerivatives
//    T = Tensor<1,SPACEDIM> or NumPtr<Tensor<1,SPACEDIM>
    template<class T> void getFunctionDerivatives(
    const Vector &dof_values,const SHAPE_CLASS *e,NumPtr<T> &gradients) 
    const {
      CHECK_TEST(update_flags & UPDATE_GRADIENTS); 
      fe->sumShapeDerivatives(e,*getShapeGradients(e),
        numberQuadraturePoints(e),0,dof_values,gradients);
    }
//  for eventual call to FiniteElement::sumShape2ndDerivatives
//    T = Tensor<2,SPACEDIM> or NumPtr<Tensor<2,SPACEDIM> >
    template<class T> void getFunction2ndDerivatives(
    const Vector &dof_values,const SHAPE_CLASS *e,
    NumPtr<T> &second_derivatives) const {
      CHECK_TEST(update_flags & UPDATE_SECOND_DERIVATIVES); 
      fe->sumShape2ndDerivatives(e,*getShape2ndDerivatives(e),
        numberQuadraturePoints(e),0,dof_values,second_derivatives);
    }
    virtual void printOn(ostream&) const;
};

# if (SPACEDIM>=2)
//merged with DEAL FEFaceValuesBase
//compare to DEAL FEFaceValues
class FEFaceValues : public FEValuesBase {
  private:
    map<int,Quadrature<SPACEDIM>*> all_face_quadrature;
  protected:
    map<int,Quadrature<SPACEDIM-1> > face_quadrature;
    map<int,NumPtr<Point<SPACEDIM> > > normal_vectors;
//  (d x / d xi)^{-T} n_{xi} det(d x / d xi):
    map<int,NumPtr<Tensor<1,SPACEDIM> > > boundary_forms;
    bool boundary_forms_ok;
    bool normal_vectors_ok;

    void initializeBoundaryFormsAndNormals(int,
      const Quadrature<SPACEDIM-1>&);

    const Quadrature<SPACEDIM-1>& getFaceQuadrature(int nv) const {
      map<int,Quadrature<SPACEDIM-1> >::const_iterator
        q_iterator=face_quadrature.find(nv);
      CHECK_TEST(q_iterator!=face_quadrature.end());
      return q_iterator->second;
    }
    const NumPtr<Point<SPACEDIM> >& getNormalVectors(int nv) const {
      CHECK_TEST(normal_vectors_ok);
      map<int,NumPtr<Point<SPACEDIM> > >::const_iterator
        nv_iterator=normal_vectors.find(nv);
      CHECK_TEST(nv_iterator!=normal_vectors.end());
      return nv_iterator->second;
    }
    int faceOffset(const SHAPE_CLASS*,int) const;
  public:
    FEFaceValues(const Mapping&,FiniteElement&, 
      map<int,Quadrature<SPACEDIM-1> >&,UPDATE_FLAGS);
    virtual ~FEFaceValues();
    const Quadrature<SPACEDIM-1>& getFaceQuadrature(const Shape2 *e,int f) 
    const {
      CHECK_TEST(f<e->numberCodimension1Shapes());
      return getFaceQuadrature(e->numberVerticesOnFace(f));
    }
    void getCanonicalQuadraturePoints(const SHAPE_CLASS *e,int f,
    NumPtr<Point<SPACEDIM> > &cp) const {
      int s=faceOffset(e,f);
      map<int,Quadrature<SPACEDIM>*>::const_iterator afq_iterator=
        all_face_quadrature.find(e->numberVertices());
      CHECK_TEST(afq_iterator!=all_face_quadrature.end());
      cp=NumPtr<Point<SPACEDIM> >(afq_iterator->second->points(),
        s,getFaceQuadrature(e,f).numberPoints());
    }
    NumPtr<Point<SPACEDIM> > getQuadraturePoints(const SHAPE_CLASS *e,
    int f) const {
      int s=faceOffset(e,f);
      return NumPtr<Point<SPACEDIM> >(
        getFEValuesData(e)->getQuadraturePoints(),
        s,getFaceQuadrature(e,f).numberPoints());
    }
    NumPtr<REAL> getJxWValues(const SHAPE_CLASS *e,int f) const {
      int s=faceOffset(e,f);
      return NumPtr<REAL>(getFEValuesData(e)->getJxWValues(),s,
        getFaceQuadrature(e,f).numberPoints());
    }
    const NumPtr<Point<SPACEDIM> >& getNormalVectors(const Shape2 *e,
    int f) const {
      return getNormalVectors(e->numberVerticesOnFace(f));
    }

    void mapShapes(const Tessellation&,const SHAPE_CLASS*,
      const LEVEL_CLASS*,int);
    void getFunctionValues(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<REAL>&) const;
    void getFunctionValues(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<Vector>&) const;
    void getFunctionDerivatives(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<Tensor<1,SPACEDIM> >&) const;
    void getFunctionDerivatives(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
    void getFunction2ndDerivatives(const Vector&,
      const SHAPE_CLASS*,int,NumPtr<Tensor<2,SPACEDIM> >&) const;
    void getFunction2ndDerivatives(const Vector&,
      const SHAPE_CLASS*,int,NumPtr<NumPtr<Tensor<2,SPACEDIM> > >&) const;
    virtual void printOn(ostream&) const;
};
# endif

#endif
