// "$Header:$"
//----------------------------  fe.cc  ---------------------------
//  $Id: fe.cc,v 1.116 2003/05/07 20:15:22 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  fe.cc  ---------------------------
//
//modified from deal.II/deal.II/source/fe/fe.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "FiniteElement.H"
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//quad must be destructed after this is destructed
FiniteElementDataBase::FiniteElementDataBase(UPDATE_FLAGS flags,
const Quadrature<SPACEDIM> *quad,const ShapeFunction<SPACEDIM> *sff) : 
update_flags(flags),shape_values_ok(false),shape_derivatives_ok(false),
shape_second_derivatives_ok(false),shape_values(0,0),
shape_derivatives(0,0),shape_second_derivatives(0,0),quadrature(quad),
shape_function_family(sff) {
  int n_q_points=quadrature->numberPoints();
  int n_shapes=sff->getNumber();
  if (flags & UPDATE_VALUES) shape_values.reinit(n_shapes,n_q_points);
  if (flags & UPDATE_GRADIENTS) {
    shape_derivatives.reinit(n_shapes,n_q_points);
  }
  if (flags & UPDATE_SECOND_DERIVATIVES) {
    shape_second_derivatives.reinit(n_shapes,n_q_points);
  }
  if (!sff->requiresShapeToComputeValues()) {
    if (flags & (UPDATE_VALUES | UPDATE_GRADIENTS | 
    UPDATE_SECOND_DERIVATIVES)) {
      for (int q=0;q<n_q_points;q++) {
        if (update_flags & UPDATE_VALUES) {
          NumPtr<REAL> values;
          sff->computeValues(quadrature->point(q),values);
          for (int s=0;s<n_shapes;s++) {
            shape_values(s,q)=values[s];
          }
        }
        if (update_flags & UPDATE_GRADIENTS) {
          NumPtr<Tensor<1,SPACEDIM> > grads;
          sff->computeGrads(quadrature->point(q),grads);
          for (int s=0;s<n_shapes;s++) {
            shape_derivatives(s,q)=grads[s];
          }
        }
        if (update_flags & UPDATE_SECOND_DERIVATIVES) {
          NumPtr<Tensor<2,SPACEDIM> > grad_grads;
          sff->computeGradGrads(quadrature->point(q),grad_grads);
          for (int s=0;s<n_shapes;s++) {
            shape_second_derivatives(s,q)=grad_grads[s];
          }
        }
      }
      shape_values_ok=(update_flags & UPDATE_VALUES);
      shape_derivatives_ok=(update_flags & UPDATE_GRADIENTS);
      shape_second_derivatives_ok=
        (update_flags & UPDATE_SECOND_DERIVATIVES);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
FiniteElementDataBase::FiniteElementDataBase(
const FiniteElementDataBase *fedb,const Shape *e) :
update_flags(fedb->update_flags),
shape_values_ok(false),shape_derivatives_ok(false),
shape_second_derivatives_ok(false),
shape_values(fedb->shape_values.size(0),fedb->shape_values.size(1)),
shape_derivatives(fedb->shape_derivatives.size(0),
                  fedb->shape_derivatives.size(1)),
shape_second_derivatives(fedb->shape_second_derivatives.size(0),
                         fedb->shape_second_derivatives.size(1)),
quadrature(fedb->quadrature),
shape_function_family(fedb->shape_function_family) {
  NumPtr<int> renumber;
  shape_function_family->shapeFunctionsOrder(e,renumber);
  if (update_flags & UPDATE_VALUES) {
    const Array<2,REAL> &sv=fedb->getShapeValues();
    for (int fe_s=0;fe_s<sv.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sv.size(1);q++) {
        REAL svsq=sv(fe_s,q);
        shape_values(abs(dof_s),q)=(dof_s>=0 ? svsq : -svsq);
      }
    }
    shape_values_ok=true;
  }
  if (update_flags & UPDATE_GRADIENTS) {
    const Array<2,Tensor<1,SPACEDIM> > &sg=fedb->getShapeDerivatives();
    for (int fe_s=0;fe_s<sg.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sg.size(1);q++) {
        const Tensor<1,SPACEDIM> &sgsq=sg(fe_s,q);
        shape_derivatives(abs(dof_s),q)=(dof_s>=0 ? sgsq : - sgsq);
      }
    }
    shape_derivatives_ok=true;
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
    const Array<2,Tensor<2,SPACEDIM> > &sgg=
      fedb->getShapeSecondDerivatives();
    for (int fe_s=0;fe_s<sgg.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sgg.size(1);q++) {
        const Tensor<2,SPACEDIM> &sggsq=sgg(fe_s,q);
        shape_second_derivatives(abs(dof_s),q)=
          (dof_s>=0 ? sggsq : -sggsq);
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElementDataBase::computeShapesFor(const Shape *e) {
  if (!shape_function_family->requiresShapeToComputeValues()) return;
  int n_points=quadrature->numberPoints();
  if (update_flags & UPDATE_VALUES) {
    for (int q=0;q<n_points;q++) {
      NumPtr<REAL> values;
      shape_function_family->
        computeValuesFor(e,quadrature->point(q),values);
      for (int s=0;s<values.getNumber();s++) {
        shape_values(s,q)=values[s];
      }
    }
    shape_values_ok=true;
  }
  if (update_flags & UPDATE_GRADIENTS) {
    for (int q=0;q<n_points;q++) {
      NumPtr<Tensor<1,SPACEDIM> > grads;
      shape_function_family->
        computeGradsFor(e,quadrature->point(q),grads);
      for (int s=0;s<grads.getNumber();s++) {
        shape_derivatives(s,q)=grads[s];
      }
    }
    shape_derivatives_ok=true;
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
    for (int q=0;q<n_points;q++) {
      NumPtr<Tensor<2,SPACEDIM> > grad_grads;
      shape_function_family->
        computeGradGradsFor(e,quadrature->point(q),grad_grads);
      for (int s=0;s<grad_grads.getNumber();s++) {
        shape_second_derivatives(s,q)=grad_grads[s];
      }
    }
    shape_second_derivatives_ok=true;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElementDataBase::printOn(ostream &os) const {
  os << "FiniteElementDataBase: update_flags = " << update_flags 
     << "\n\tshape_values_ok = " << PRINT_BOOLEAN(shape_values_ok)
     << "\n\tshape_derivatives_ok = " 
     << PRINT_BOOLEAN(shape_derivatives_ok)
     << "\n\tshape_second_derivatives_ok = " 
     << PRINT_BOOLEAN(shape_second_derivatives_ok)
     << endl;
//if (shape_values_ok) {
//  os << "\tshape_values:" << endl;
//  for (int i=0;i<shape_values.size(0);i++) {
//    for (int j=0;j<shape_values.size(1);j++) {
//      os << "\t\tshape_values[" << i << "," << j << "] = "
//         << shape_values(i,j) << endl;
//    }
//  }
//}
  if (shape_derivatives_ok) {
    os << "\tshape_derivatives:" << endl;
    for (int i=0;i<shape_derivatives.size(0);i++) {
      for (int j=0;j<shape_derivatives.size(1);j++) {
        os << "\t\tshape_derivatives[" << i << "," << j << "] = "
           << shape_derivatives(i,j) << endl;
      }
    }
  }
//if (shape_second_derivatives_ok) {
//  os << "\tshape_second_derivatives:" << endl;
//  for (int i=0;i<shape_second_derivatives.size(0);i++) {
//    for (int j=0;j<shape_second_derivatives.size(1);j++) {
//      os << "\t\tshape_second_derivatives[" << i << "," << j << "] = "
//         << shape_second_derivatives(i,j) << endl;
//    }
//  }
//}
  cout << "\tquadrature = " << quadrature << endl;
  quadrature->printOn(os);
//shape_function_family->printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
FiniteElement::~FiniteElement() {
  while (!polynomial_space.empty()) {
    map<int,ShapeFunction<SPACEDIM>*>::iterator 
      ps_iterator=polynomial_space.begin();
    delete ps_iterator->second;
    polynomial_space.erase(ps_iterator);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::setDofsPerObject() {
#if (SPACEDIM==1)
  dofs_per_cell[Interval::verticesPerShape()]=
    Interval::verticesPerShape()*dofs_per_vertex
    +Interval::linesPerShape()*dofs_per_line;
#endif
#if (SPACEDIM==2)
  for (map<int,int>::const_iterator 
  dpo_iterator=dofs_per_quad.begin();
  dpo_iterator!=dofs_per_quad.end();dpo_iterator++) {
    int nv=dpo_iterator->first;
    dofs_per_cell[nv]=nv*dofs_per_vertex
      +linesPerShape(nv)*dofs_per_line
      +dofs_per_quad[nv];
    for (int f=0;f<facesPerShape(nv);f++) {
      int nvof=numberVerticesOnFace(nv,f);
      int lpf=linesPerFace(nv,f);
      int dpf=nvof*dofs_per_vertex+lpf*dofs_per_line;
      dofs_per_face[nvof]=dpf;
    }
  }
#endif
#if (SPACEDIM==3)
  int ffli=0;
  int ffqi=0;
  for (map<int,int>::const_iterator 
  dpo_iterator=dofs_per_quad.begin();
  dpo_iterator!=dofs_per_quad.end();dpo_iterator++) {
    int nv=dpo_iterator->first;
    ffli+=nv;
    if (nv==Quadrilateral::verticesPerShape()) {
      ffqi+=Quadrilateral::linesPerShape();
    } else {
      CHECK_TEST(nv==Triangle::verticesPerShape());
      ffqi+=Triangle::linesPerShape();
    }
  }
  ffli*=dofs_per_vertex;
  ffqi=ffli+ffqi*dofs_per_line;
  for (map<int,int>::const_iterator 
  dpo_iterator=dofs_per_hex.begin();
  dpo_iterator!=dofs_per_hex.end();dpo_iterator++) {
    int nv=dpo_iterator->first;
    dofs_per_hex[nv]=dpo_iterator->second;
    int fli=nv*dofs_per_vertex;
    int fqi=fli+linesPerShape(nv)*dofs_per_line;
    int dpc=nv*dofs_per_vertex
      +linesPerShape(nv)*dofs_per_line
      +dpo_iterator->second;
    int fhi=fqi;
    for (int f=0;f<facesPerShape(nv);f++) {
      int vpf=numberVerticesOnFace(nv,f);
      int lpf=linesPerFace(nv,f);
      map<int,int>::const_iterator dpq_iterator=
        dofs_per_quad.find(vpf);
      CHECK_TEST(dpq_iterator!=dofs_per_quad.end());
      dpc+=dpq_iterator->second;
      int dpf=vpf*dofs_per_vertex
        +lpf*dofs_per_line+dpq_iterator->second;
      dofs_per_face[vpf]=dpf;
      fhi+=dpq_iterator->second;
    }
    dofs_per_cell[nv]=dpc;
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int FiniteElement::maxDofsPerCell() const {
  int max_dofs_per_cell=0;
  for (map<int,int>::const_iterator dpc_iterator=
  dofs_per_cell.begin();dpc_iterator!=dofs_per_cell.end();
  dpc_iterator++) {
    max_dofs_per_cell=max(max_dofs_per_cell,dpc_iterator->second);
  }
  return max_dofs_per_cell;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//dof_values come from element vertices in element's ordering, then
//  from line elements' interior, in line element's ordering, then
//  from face elements' interior, in face element's ordering, then
//  from element's interior, in element's ordering
//see DoFHandler::getDofIndices for details
void FiniteElement::sumShapeValues(const SHAPE_CLASS *e,
const Array<2,REAL> &sv,int nqp,int offset,const Vector &dof_values,
NumPtr<REAL> &values) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    REAL val=0.;
    for (int s=0;s<dpc;s++) val+=sv(s,q+offset)*dof_values[s]; 
    values[q]=val;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::sumShapeValues(const SHAPE_CLASS *e,
const Array<2,REAL> &sv,int nqp,int offset,const Vector &dof_values,
NumPtr<Vector> &values) const{
  int shapes_per_cell=dofsPerCell(e)/components;
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    Vector &v=values[q];
    v.reinit(components);
    for (int s=0;s<shapes_per_cell;s++) {
      const REAL &svq=sv(s,q+offset);
      int sc=s*components;
      for (int c=0;c<components;c++) {
        v[c]+=svq*dof_values[c+sc]; 
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::sumShapeDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<1,SPACEDIM> > &sg,int nqp,int offset,
const Vector &dof_values,NumPtr<Tensor<1,SPACEDIM> > &gradients) 
const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  gradients.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    for (int s=0;s<dpc;s++) {
      Tensor<1,SPACEDIM> tmp(sg(s,q+offset));
      tmp*=dof_values[s];
      gradients[q]+=tmp;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::sumShapeDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<1,SPACEDIM> > &sg,int nqp,int offset,
const Vector &dof_values,
NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &gradients) const {
  int shapes_per_cell=dofsPerCell(e)/components;
  if (gradients.getNumber()==0) gradients.allocate(nqp);
#ifdef DEBUG
  else CHECK_SAME(gradients.getNumber(),nqp);
#endif
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<1,SPACEDIM> > &g=gradients[q];
    if (g.getNumber()==0) g.allocate(components);
#ifdef DEBUG
    else CHECK_SAME(g.getNumber(),components);
#endif
    Tensor<1,SPACEDIM> empty_tensor;
    g.initialize(empty_tensor);
    for (int s=0;s<shapes_per_cell;s++) {
      int sc=s*components;
      const Tensor<1,SPACEDIM> &sgq=sg(s,q+offset);
      for (int c=0;c<components;c++) {
        Tensor<1,SPACEDIM> tmp(sgq);
        tmp*=dof_values[c+sc];
        g[c]+=tmp;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::sumShape2ndDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<2,SPACEDIM> > &sgg,int nqp,int offset,
const Vector &dof_values,
NumPtr<Tensor<2,SPACEDIM> > &second_derivatives) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  second_derivatives.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    for (int s=0;s<dpc;s++) {
      Tensor<2,SPACEDIM> tmp(sgg(s,q+offset));
      tmp*=dof_values[s];
      second_derivatives[q]+=tmp;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::sumShape2ndDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<2,SPACEDIM> > &sgg,int nqp,int offset,
const Vector &dof_values,
NumPtr<NumPtr<Tensor<2,SPACEDIM> > > &second_derivatives) const {
  int shapes_per_cell=dofsPerCell(e)/components;
  second_derivatives.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<2,SPACEDIM> > &gg=second_derivatives[q];
    gg.allocate(components);
    Tensor<2,SPACEDIM> empty_tensor;
    gg.initialize(empty_tensor);
    for (int s=0;s<shapes_per_cell;s++) {
      int sc=s*components;
      const Tensor<2,SPACEDIM> &sggq=sgg(s,q+offset);
      for (int c=0;c<components;c++) {
        Tensor<2,SPACEDIM> tmp(sggq);
        tmp*=dof_values[c+sc];
        gg[c]+=tmp;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
//for a boundary_function with few masked components,
//  the factorNSolve call on all components is wasteful
//guessing that this case is infrequent in any one problem,
//  we didn't bother to solve only for the masked components:
//  we would have to fix the LAMatrix class to have separate
//    factor and solve procedures
//face_element argument for compatibility with NonPrimitiveFiniteElement
void FiniteElement::setBoundaryValues(const SHAPE_CLASS *element,
const CODIMENSION1_SHAPE_CLASS *face_element,int face_no,
const NumPtr<int> &element_dofs,const NumPtr<int> &face_dofs,
const NumPtr<Point<SPACEDIM> > &canonical_locations,
const NumPtr<Point<SPACEDIM> > &mapped_locations,
const BoundaryCondition *bc,
map<int,REAL> &boundary_values,REAL time) const {
  int shapes_per_face=mapped_locations.getNumber();
  LAMatrix bv(shapes_per_face,components);
  if (components==1) {
    NumPtr<REAL> dof_values_scalar(shapes_per_face);
    const TimeDependentBoundaryCondition *tdbc=
      dynamic_cast<const TimeDependentBoundaryCondition*>(bc);
    if (tdbc) {
      TimeDependentFunctionSystem *tdfs=tdbc->getFunctionSystem();
      tdfs->value(dof_values_scalar,mapped_locations,time,0);
    } else {
      bc->getFunctionSystem()->value(dof_values_scalar,mapped_locations,0);
    }
    for (int s=0;s<shapes_per_face;s++) {
      bv(s,0)=dof_values_scalar[s];
    }
  } else {
    NumPtr<Vector> dof_values_system;
    const TimeDependentBoundaryCondition *tdbc=
      dynamic_cast<const TimeDependentBoundaryCondition*>(bc);
    if (tdbc) {
      TimeDependentFunctionSystem *tdfs=tdbc->getFunctionSystem();
      tdfs->value(dof_values_system,mapped_locations,time);
    } else {
      bc->getFunctionSystem()->value(dof_values_system,mapped_locations);
    }
    for (int c=0;c<components;c++) {
      if (bc->isEssential(c)) {
        for (int s=0;s<shapes_per_face;s++) {
          bv(s,c)=dof_values_system[s][c];
        }
      } else {
        for (int s=0;s<shapes_per_face;s++) bv(s,c)=0.;
      }
    }
  }
  LAMatrix face_shape_values(shapes_per_face,shapes_per_face);
  NumPtr<int> nsof;
  ShapeFunction<SPACEDIM> *sf=shapeFunctionFamily(element);
//nsof[face_fe_index]=element_fe_index
  sf->nonzeroShapesOnFace(element,face_no,nsof);
  CHECK_SAME(shapes_per_face,nsof.getNumber());

//element_renumber[element_fe_index]=element_dof_index
  NumPtr<int> element_renumber; 
  sf->shapeFunctionsOrder(element,element_renumber);

//face_dofs[face_dof_index]=global_dof_index 
  map<int,int> face_dofs_map;
  for (int i=0;i<face_dofs.getNumber();i++) {
    face_dofs_map[face_dofs[i]]=i;
  }
//global_dof_index=element_dofs[element_dof_index]
//                =element_dofs[element_renumber[element_fe_index]]
//                =element_dofs[element_renumber[nsof[face_fe_index]]]
  map<int,int> nsof_map;
  for (int i=0;i<nsof.getNumber();i++) {
    int element_dof_index=element_renumber[nsof[i]];
    int global_dof_index=element_dofs[abs(element_dof_index)*components];
    nsof_map[element_dof_index>=0 ? global_dof_index : -global_dof_index]
      =i;
  }
  NumPtr<int> face_renumber(nsof.getNumber());
  for (map<int,int>::const_iterator nm_iterator=nsof_map.begin();
  nm_iterator!=nsof_map.end();nm_iterator++) {
    int global_dof_index=nm_iterator->first;
    map<int,int>::const_iterator fdm_iterator=
      face_dofs_map.find(abs(global_dof_index));
    CHECK_TEST(fdm_iterator!=face_dofs_map.end());
//  face_renumber[face_fe_index]=face_dof_index
    int face_dof_index=(fdm_iterator->second)/components;
    face_renumber[nm_iterator->second]=
      (global_dof_index>=0 ? face_dof_index : - face_dof_index);
  }

  Quadrature<SPACEDIM> quad(canonical_locations);
  FiniteElementDataBase *fedb=
    getFEData(element->numberVertices(),UPDATE_VALUES,&quad);
  fedb->computeShapesFor(element);
  for (int face_fe_s=0;face_fe_s<shapes_per_face;face_fe_s++) {
    int face_dof_s=face_renumber[face_fe_s];
    for (int q=0;q<shapes_per_face;q++) {
      face_shape_values(q,abs(face_dof_s))=
        fedb->shapeValue(nsof[face_fe_s],q);
    }
  }
  LAMatrix *soln=factorNSolve(face_shape_values,bv);
  for (int face_fe_s=0;face_fe_s<shapes_per_face;face_fe_s++) {
    int sc=face_fe_s*components;
    for (int c=0;c<components;c++) {
      if (bc->isEssential(c)) {
        int global_dof=face_dofs[sc+c];
        boundary_values[global_dof]=(*soln)(face_fe_s,c);
        if (face_renumber[face_fe_s]<0) {
          boundary_values[global_dof]=-boundary_values[global_dof];
        }
      }
    }
  }
  delete soln;
  delete fedb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::equallySpacedLatticePointDataBase(
map<int,FiniteElementDataBase*> &fedb_map,
map<int,Quadrature<SPACEDIM>*> &quad_map) const {
  for (map<int,ShapeFunction<SPACEDIM>*>::const_iterator
  sf_iterator=polynomial_space.begin();
  sf_iterator!=polynomial_space.end();sf_iterator++) {
    int nv=sf_iterator->first;
    NumPtr<Point<SPACEDIM> > element_lattice_points;
    equallySpacedLatticePoints(nv,degree,element_lattice_points);
    Quadrature<SPACEDIM> *quad=quad_map[nv]=
      OPERATOR_NEW Quadrature<SPACEDIM>(element_lattice_points);
    FiniteElementDataBase *fedb=getFEData(nv,UPDATE_VALUES,quad);
    fedb_map[nv]=fedb;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::setValues(const SHAPE_CLASS *element,
const NumPtr<int> &element_dofs,FiniteElementDataBase *fedb,
const NumPtr<Point<SPACEDIM> > &mapped_locations,
const FunctionSystem *given_function,Vector &values) const {
  int shapes_per_element=mapped_locations.getNumber();
//given_function(mapped_location[q]); order of q irrelevant
  LAMatrix gf(shapes_per_element,components);
  if (components==1) {
    NumPtr<REAL> dof_values(shapes_per_element);
    given_function->value(dof_values,mapped_locations,0);
    for (int s=0;s<shapes_per_element;s++) {
      gf(s,0)=dof_values[s];
    }
  } else {
    NumPtr<Vector> dof_values;
    given_function->value(dof_values,mapped_locations);
    for (int c=0;c<components;c++) {
      for (int s=0;s<shapes_per_element;s++) {
        gf(s,c)=dof_values[s][c];
      }
    }
  }
  fedb->computeShapesFor(element);
//shapeValue(q,element_fe_index)
  LAMatrix shape_values(shapes_per_element,shapes_per_element);
  for (int s=0;s<shapes_per_element;s++) {
    for (int q=0;q<shapes_per_element;q++) {
      shape_values(q,s)=fedb->shapeValue(s,q);
    }
  }
//soln(element_fe_index,c)
  LAMatrix *soln=factorNSolve(shape_values,gf);
//element_renumber[element_fe_index]=element_dof_index
  ShapeFunction<SPACEDIM> *sf=shapeFunctionFamily(element);
  NumPtr<int> element_renumber;
  sf->shapeFunctionsOrder(element,element_renumber);
  NumPtr<int> element_renumber_inverse(element_renumber.getNumber());
  for (int i=0;i<element_renumber.getNumber();i++) {
    element_renumber_inverse[abs(element_renumber[i])]=i;
  }
//element_dofs[element_dof_index]=global_dof_index
  for (int element_dof_index=0;element_dof_index<shapes_per_element;
  element_dof_index++) {
    int sc=element_dof_index*components;
    int fe_index=element_renumber_inverse[element_dof_index];
    for (int c=0;c<components;c++) {
      values[element_dofs[sc+c]]=(*soln)(fe_index,c);
      if (element_renumber[fe_index]<0) {
        values[element_dofs[sc+c]]=-values[element_dofs[sc+c]];
      }
    }
  }
  delete soln;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
FiniteElementDataBase* FiniteElement::fixShapeFunctionsFor(
const Shape *e,FiniteElementDataBase *fedb) const {
  ShapeFunction<SPACEDIM> *sf=shapeFunctionFamily(e);
  if (sf->requiresShapeToComputeValues()) {
    fedb->computeShapesFor(e);
    return fedb;
  } else {
    return OPERATOR_NEW FiniteElementDataBase(fedb,e);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FiniteElement::printOn(ostream &os) const {
  os << "\tFiniteElement: degree = " << degree << endl;
  for (map<int,ShapeFunction<SPACEDIM>*>::const_iterator
  ps_iterator=polynomial_space.begin();
  ps_iterator!=polynomial_space.end();ps_iterator++) {
    os << "\tpolynomial_space[" << ps_iterator->first << "]: " << endl;
    ps_iterator->second->printOn(os);
  }
  os << "\tdofs_per_vertex = " << dofs_per_vertex
     << "\n\tdofs_per_line = " << dofs_per_line
     << endl;
#if (SPACEDIM>1)
  for (map<int,int>::const_iterator dpq_iterator=
  dofs_per_quad.begin();dpq_iterator!=dofs_per_quad.end();
  dpq_iterator++) {
    os << "\tdofs_per_quad[" << dpq_iterator->first << "] = "
       << dpq_iterator->second << endl;
  }
  for (map<int,int>::const_iterator
  dpf_iterator=dofs_per_face.begin();dpf_iterator!=
  dofs_per_face.end();dpf_iterator++) {
    os << "\tdofs_per_face[" << dpf_iterator->first << "] = "
       << dpf_iterator->second << endl;
  }
#endif
#if (SPACEDIM==3)
  for (map<int,int>::const_iterator dph_iterator=
  dofs_per_hex.begin();dph_iterator!=dofs_per_hex.end();
  dph_iterator++) {
    os << "\tdofs_per_hex[" << dph_iterator->first << "] = "
       << dph_iterator->second << endl;
  }
#endif
  for (map<int,int>::const_iterator dpc_iterator=
  dofs_per_cell.begin();dpc_iterator!=
  dofs_per_cell.end();dpc_iterator++) {
    os << "\tdofs_per_cell[" << dpc_iterator->first << "] = "
       << dpc_iterator->second << endl;
  }
  os << "\tcomponents = " << components << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
FE_H::FE_H(int d,int nc) : FiniteElement(d,nc) {
  CHECK_POSITIVE(d);
#if (SPACEDIM==1)
  polynomial_space[Interval::verticesPerShape()]=
    OPERATOR_NEW HierarchicalIntervalPolynomials(degree);
#endif
#if (SPACEDIM==2)
  polynomial_space[Quadrilateral::verticesPerShape()]=
    OPERATOR_NEW HierarchicalQuadrilateralPolynomials(degree);
  polynomial_space[Triangle::verticesPerShape()]=
    OPERATOR_NEW HierarchicalTrianglePolynomials(degree);
#endif
#if (SPACEDIM==3)
  polynomial_space[Hexahedron::verticesPerShape()]=
    OPERATOR_NEW HierarchicalHexahedronPolynomials(degree);
  polynomial_space[Prism::verticesPerShape()]=
    OPERATOR_NEW HierarchicalPrismPolynomials(degree);
  polynomial_space[Tetrahedron::verticesPerShape()]=
    OPERATOR_NEW HierarchicalTetrahedronPolynomials(degree);
#endif

  dofs_per_vertex=components;
  dofs_per_line=components*
    Interval::staticNumberInteriorEquallySpacedLatticePoints(degree);
#if (SPACEDIM>1)
  dofs_per_quad[Quadrilateral::verticesPerShape()]=components*
    Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(degree);
  dofs_per_quad[Triangle::verticesPerShape()]=components*
    Triangle::staticNumberInteriorEquallySpacedLatticePoints(degree);
#endif
#if (SPACEDIM==3)
  dofs_per_hex[Hexahedron::verticesPerShape()]=components*
    Hexahedron::staticNumberInteriorEquallySpacedLatticePoints(degree);
  dofs_per_hex[Prism::verticesPerShape()]=components*
    Interval::staticNumberInteriorEquallySpacedLatticePoints(degree)*
    Triangle::staticNumberInteriorEquallySpacedLatticePoints(degree);
  dofs_per_hex[Tetrahedron::verticesPerShape()]=components*
    Tetrahedron::staticNumberInteriorEquallySpacedLatticePoints(degree);
#endif
  setDofsPerObject();

//We want the fine solution function to equal the coarse solution fcn:
//  b_s(xi_q) c_s = B_S(parent coord of child xi_q) C_S
//(Einstein summation convention: repeated subscript ==> sum)
//Thus if [ p_{d,s} ] is permutation that reorders 
//  from static fe order (index s)
//  to shape-dependent dof order on the child (index d), and
//if [ Q_{D,S} ] is the permutation that reorders 
//  from fe (index S) 
//  to dof  (index D) on parent,
//then
//  c_d = p_{d,s} [ b_s(xi_q) ]^{-1} 
//        [ B_S(parent coord of child xi_q) ] Q_{D,S}^T C_D
//The permutations p_{d,s} and Q_{D,S} are not used in this routine
//fe order is same as Hierarchical*Polynomials
  int cps=childrenPerShape();
  int cpcd1s=childrenPerCodimension1Shape();
  for (map<int,ShapeFunction<SPACEDIM>*>::const_iterator
  sf_iterator=polynomial_space.begin();
  sf_iterator!=polynomial_space.end();sf_iterator++) {
    int nv=sf_iterator->first;
    const ShapeFunction<SPACEDIM> *sf=sf_iterator->second;
    NumPtr<Point<SPACEDIM> > csp;
    equallySpacedLatticePoints(nv,degree,csp);
    int ncsp=csp.getNumber();
    NumPtr<Matrix> &prolongation_matrices=prolongation[nv];
    prolongation_matrices.allocate(cps);
    if (sf->requiresShapeToComputeValues()) continue;

    LAMatrix mat(ncsp,ncsp);
    LAMatrix rhs(ncsp,ncsp);
    for (int c=0;c<cps;c++) {
      for (int q=0;q<ncsp;q++) {
        const Point<SPACEDIM> &child_point=csp[q];
        Point<SPACEDIM> parent_point=
          parentCoordinateOfCanonical(nv,child_point,c);
        NumPtr<REAL> cv;
        sf->computeValues(child_point,cv);
        NumPtr<REAL> pv;
        sf->computeValues(parent_point,pv);
        for (int s=0;s<sf->getNumber();s++) {
          mat(q,s)=cv[s];
          rhs(q,s)=pv[s];
        }
      }
      LAMatrix *soln=factorNSolve(mat,rhs);
      prolongation_matrices[c].copy(*soln);
      delete soln;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
const Matrix& FE_H::prolongationMatrix(const SHAPE_CLASS *e,int c,
const LEVEL_CLASS *level) {
#if (SPACEDIM>1)
  if (shapeFunctionFamily(e)->requiresShapeToComputeValues()) {
#if (SPACEDIM==2)
    NumPtr<int> 
      line_interior_child_vertices(e->numberCodimension1Shapes());
    for (int l=0;l<line_interior_child_vertices.getNumber();l++) {
      int face_index=e->codimension1ShapeIndex(l);
      int fc,fv;
      line_interior_child_vertices[l]=level->CODIMENSION1_LEVEL_CLASS::
        interiorChildVertex(face_index,fc,fv);
    }
#else
    NumPtr<int> 
      line_interior_child_vertices(e->numberCodimension2Shapes());
    for (int l=0;l<line_interior_child_vertices.getNumber();l++) {
      int line_index=e->codimension2ShapeIndex(l);
      int lc,lv;
      line_interior_child_vertices[l]=level->Level1::
        interiorChildVertex(line_index,lc,lv);
    }
#endif
    int child_index=level->childIndex(e->elementDataIndex(),c);
    const Shape *child_shape=level->finerLevel()->getShape(child_index);
    const ShapeFunction<SPACEDIM> *sf=shapeFunctionFamily(e);
    NumPtr<Point<SPACEDIM> > csp;
    e->equallySpacedLatticePoints(degree,csp);
    int ncsp=csp.getNumber();
    int nv=e->numberVertices();
    NumPtr<Matrix> &prolongation_matrices=prolongation[nv];
    LAMatrix mat(ncsp,ncsp);
    LAMatrix rhs(ncsp,ncsp);
    for (int q=0;q<ncsp;q++) {
      Point<SPACEDIM> child_point=csp[q];
      Point<SPACEDIM> parent_point=
        e->parentCoordinateForCanonical(child_point,child_shape,c,
          line_interior_child_vertices);
      NumPtr<REAL> cv;
      sf->computeValuesFor(child_shape,child_point,cv);
      NumPtr<REAL> pv;
      sf->computeValuesFor(e,parent_point,pv);
      for (int s=0;s<sf->getNumber();s++) {
        mat(q,s)=cv[s];
        rhs(q,s)=pv[s];
      }
    }
    LAMatrix *soln=factorNSolve(mat,rhs);
    prolongation_matrices[c].copy(*soln);
    delete soln;
//*********************************
#ifdef DEBUG
    Point<SPACEDIM> cp(D_DECL(drand48(),drand48(),drand48()));
#if (SPACEDIM==2)
    if (dynamic_cast<const Triangle*>(e)!=0) {
      cp[1]*=(1.-cp[0]);
    }
#endif
#if (SPACEDIM==3)
    if (dynamic_cast<const Tetrahedron*>(e)!=0) {
      cp[1]*=(1.-cp[0]);
      cp[2]*=(1.-cp[1]-cp[1]);
    }
    if (dynamic_cast<const Prism*>(e)!=0) {
      cp[1]*=(1.-cp[0]);
    }
#endif
    Point<SPACEDIM> pp=e->parentCoordinateForCanonical(cp,child_shape,c,
      line_interior_child_vertices);
    for (int ps=0;ps<sf->getNumber();ps++) {
      NumPtr<REAL> pv;
      NumPtr<REAL> cv;
      if (sf->requiresShapeToComputeValues()) {
        sf->computeValuesFor(e,pp,pv);
        sf->computeValuesFor(child_shape,cp,cv);
      } else {
        sf->computeValues(pp,pv);
        sf->computeValues(cp,cv);
      }
      REAL sum=0.;
      for (int cs=0;cs<cv.getNumber();cs++) {
        sum+=cv[cs]*prolongation_matrices[c](cs,ps);
      }
      REAL error=abs(pv[ps]-sum);
      if (abs(pv[ps])>0.) error/=abs(pv[ps]);
      if (error>4194304.*DBL_EPSILON) {
        cout << "\tpv[" << ps << "] = " << pv[ps] << " " << sum
             << " error = " << error << " " << error/DBL_EPSILON << endl;
        OBSOLETE("fix bugs");
      }
    }
#endif
//*********************************
    return prolongation_matrices[c];
  } else
#endif
  return FiniteElement::prolongationMatrix(e,c,level);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void processConstraints(int components,int prolongation_first_row,
int interior_vertex_child,const NumPtr<int> &neighbor_child_on_face,
const NumPtr<NumPtr<int> > &prolongation_rows,
const NumPtr<Matrix> &prolongation_matrices,
int prolongation_column,int col,Matrix &m) {
  int row=0;
  int apc=abs(prolongation_column);
  {
    const Matrix &pm=prolongation_matrices[interior_vertex_child];
    REAL mrc=pm(abs(prolongation_first_row),apc);
    if (prolongation_first_row*prolongation_column<0) mrc=-mrc;
    for (int c=0;c<components;c++) {
      m(row++,col+c)=mrc;
    }
  }
  int cpcd1s=childrenPerCodimension1Shape();
  for (int s=0;s<cpcd1s;s++) {
    int c=neighbor_child_on_face[s];
    const Matrix &pm=prolongation_matrices[c];
    const NumPtr<int> &pr=prolongation_rows[s];
    for (int i=0;i<pr.getNumber();i++) {
      REAL mrc=pm(abs(pr[i]),apc);
      if (pr[i]*prolongation_column<0) mrc=-mrc;
      for (int c=0;c<components;c++) {
        m(row++,col+c)=mrc;
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
//see ConstraintMatrix::makeHangingNodeConstraints
void FE_H::constraints(const LEVEL_CLASS *level,const SHAPE_CLASS *shape,
const CODIMENSION1_SHAPE_CLASS *face_shape,int f,Matrix &m) {
  CHECK_SAME(m.size(0),0);
  CHECK_SAME(m.size(1),0);
// (n_dofs_on_children,n_dofs_on_mother)
  m.resize(face_shape->numberInteriorChildVertices()*dofs_per_vertex
    +dofs_per_line*childrenPerCodimension1Shape(),
    face_shape->numberVertices()*dofs_per_vertex+dofs_per_line);

  int face_index=face_shape->elementDataIndex();
  int neighbor_index=level->neighborIndex(shape,f);
  CHECK_TEST(level->hasChildren(neighbor_index));
  const SHAPE_CLASS *neighbor_shape=level->getShape(neighbor_index);
  int nf=neighbor_shape->faceWithIndexEqualTo(face_index);

  int nnf=neighbor_shape->numberCodimension1Shapes();
  NumPtr<int> line_interior_child_vertices(nnf);
  for (int l=0;l<nnf;l++) {
    int face_index=neighbor_shape->codimension1ShapeIndex(l);
    int fc,fv;
    line_interior_child_vertices[l]=level->CODIMENSION1_LEVEL_CLASS::
      interiorChildVertex(face_index,fc,fv);
  }
  const LEVEL_CLASS *finer_level=level->finerLevel();

//shapeFunctionsOrder is normally called in 
//  FiniteElementDataBase::FiniteElementDataBase
//  right before setting the shape_values
//FiniteElementDataBase::FiniteElementDataBase is called from
//  FiniteElement::fixShapeFunctionsFor
//which is called from FEValuesData::mapShapes
//
//for hierarchical shape functions, this flips the sign for certain dofs
  ShapeFunction<SPACEDIM> *psf=shapeFunctionFamily(neighbor_shape);
  NumPtr<int> parent_renumber;
  psf->shapeFunctionsOrder(neighbor_shape,parent_renumber);

  map<int,NumPtr<int> > child_renumber_map;
  int prolongation_first_row=-1;
  int interior_vertex_child=-1;
  {
    CHECK_TEST(face_shape->numberInteriorChildVertices()==1);
//  first row corresponds to interior child vertex (if it exists)
    int fc,fcv;
    int interior_child_vertex_index=level->CODIMENSION1_LEVEL_CLASS::
      interiorChildVertex(face_index,fc,fcv);
    int icv;
    level->interiorChildVertexOnFaceOf(neighbor_shape,nf,
      interior_vertex_child,icv);
    int neighbor_child_index=
      level->childIndex(neighbor_index,interior_vertex_child);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);
    int ncv=neighbor_child_shape->
      vertexEqualToVertexIndex(interior_child_vertex_index);
    ShapeFunction<SPACEDIM> *csf=shapeFunctionFamily(neighbor_child_shape);
    NumPtr<int> child_renumber;
    csf->shapeFunctionsOrder(neighbor_child_shape,child_renumber);
    prolongation_first_row=child_renumber[ncv];
  }

  int cpcd1s=childrenPerCodimension1Shape();
  NumPtr<NumPtr<int> > prolongation_rows(cpcd1s);
  NumPtr<int> neighbor_child_on_face(cpcd1s);

//other rows correspond to vertices interior to child faces
  for (int s=0;s<cpcd1s;s++) {
    int face_child_index=
      level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,s);
    int nc=-1,ncf=-1;
    level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);
    neighbor_child_on_face[s]=nc;
    int neighbor_child_index=
      level->childIndex(neighbor_index,nc);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);
    map<int,NumPtr<int> >::iterator child_renumber_map_iterator=
      child_renumber_map.find(nc);
    if (child_renumber_map_iterator==child_renumber_map.end()) {
      NumPtr<int> &child_renumber=child_renumber_map[nc];
      ShapeFunction<SPACEDIM> *csf=
        shapeFunctionFamily(neighbor_child_shape);
      csf->shapeFunctionsOrder(neighbor_child_shape,child_renumber);
      child_renumber_map_iterator=child_renumber_map.find(nc);
    }
    NumPtr<int> &child_renumber=child_renumber_map_iterator->second;

    NumPtr<int> &prc=prolongation_rows[s];
    prc.allocate(dofs_per_line);
    int offset=neighbor_shape->numberVertices()+ncf*dofs_per_line;
    for (int j=0;j<dofs_per_line;j++) {
      prc[j]=child_renumber[j+offset];
    }
  }

//hierarchical shape functions require shape only in 3D
  CHECK_TEST(!shapeFunctionFamily(neighbor_shape)->
             requiresShapeToComputeValues());

  const NumPtr<Matrix> &prolongation_matrices=
    prolongationMatrices(neighbor_shape);

  int col=0;
  for (int pv=0;pv<face_shape->numberVertices();pv++) {
//  first columns correspond to vertices of face_shape
    int vi=face_shape->vertexIndex(pv);
    int prolongation_column=neighbor_shape->vertexEqualToVertexIndex(vi);
    processConstraints(components,prolongation_first_row,
      interior_vertex_child,neighbor_child_on_face,prolongation_rows,
      prolongation_matrices,prolongation_column,col,m);
    col++;
  }
  int offset=neighbor_shape->numberVertices()+nf*dofs_per_line;
  for (int j=0;j<dofs_per_line;j++) {
//  other columns correspond to points interior to parent face
    processConstraints(components,prolongation_first_row,
      interior_vertex_child,neighbor_child_on_face,prolongation_rows,
      prolongation_matrices,parent_renumber[offset+j],col,m);
    col++;
  }
  for (map<int,NumPtr<int> >::iterator child_renumber_map_iterator=
  child_renumber_map.begin();
  child_renumber_map_iterator!=child_renumber_map.end();
  child_renumber_map_iterator++) {
    child_renumber_map_iterator->second.cleanup();
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void processConstraints(int components,int prolongation_first_row,
int interior_vertex_child,
const NumPtr<int> &prolongation_second_rows,
const NumPtr<int> &child_for_prolongation_second_rows,
const NumPtr<NumPtr<int> > &prolongation_third_rows,
const NumPtr<int> &child_for_prolongation_third_rows,
const NumPtr<NumPtr<NumPtr<int> > > &prolongation_fourth_rows,
const NumPtr<NumPtr<int> > &child_for_prolongation_fourth_rows,
const NumPtr<NumPtr<int> > &prolongation_fifth_rows,
const NumPtr<int> &child_for_prolongation_fifth_rows,
const NumPtr<Matrix> &prolongation_matrices,
int prolongation_column,int col,Matrix &m) {
  int row=0;
  if (prolongation_first_row>=0) {
    const Matrix &pm=prolongation_matrices[interior_vertex_child];
    m(row++,col)=pm(prolongation_first_row,prolongation_column);
  }
  for (int l=0;l<prolongation_second_rows.getNumber();l++) {
    const Matrix &pm=
      prolongation_matrices[child_for_prolongation_second_rows[l]];
    for (int c=0;c<components;c++) {
      m(row++,col+c)=pm(prolongation_second_rows[l],prolongation_column);
    }
  }
  for (int i=0;i<prolongation_third_rows.getNumber();i++) {
    const NumPtr<int> &pr=prolongation_third_rows[i];
    int dpl=pr.getNumber();
    if (dpl>0) {
      const Matrix &pm=
        prolongation_matrices[child_for_prolongation_third_rows[i]];
      for (int j=0;j<dpl;j++) {
        for (int c=0;c<components;c++) {
          m(row++,col+c)=pm(pr[j],prolongation_column);
        }
      }
    }
  }
  for (int l=0;l<prolongation_fourth_rows.getNumber();l++) {
    const NumPtr<NumPtr<int> > &p4r=prolongation_fourth_rows[l];
    int dpl=p4r.getNumber();
    if (dpl>0) {
      const NumPtr<int> &cfp4r=
        child_for_prolongation_fourth_rows[l];
      for (int s=0;s<dpl;s++) {
        const Matrix &pm=prolongation_matrices[cfp4r[s]];
        const NumPtr<int> &pr=p4r[s];
        for (int j=0;j<pr.getNumber();j++) {
          for (int c=0;c<components;c++) {
            m(row++,col+c)=pm(pr[j],prolongation_column);
          }
        }
      }
    }
  }
  for (int i=0;i<prolongation_fifth_rows.getNumber();i++) {
    const NumPtr<int> &pr=prolongation_fifth_rows[i];
    int dpq=pr.getNumber();
    if (dpq>0) {
      int cfp5r=child_for_prolongation_fifth_rows[i];
      const Matrix &pm=prolongation_matrices[cfp5r];
      for (int j=0;j<dpq;j++) {
        for (int c=0;c<components;c++) {
          m(row++,col+c)=pm(pr[j],prolongation_column);
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//see ConstraintMatrix::makeHangingNodeConstraints
void FE_H::constraints(const LEVEL_CLASS *level,const SHAPE_CLASS *shape,
const CODIMENSION1_SHAPE_CLASS *face_shape,int f,Matrix &m) {
  CHECK_SAME(m.size(0),0);
  CHECK_SAME(m.size(1),0);
//(n_dofs_on_children,n_dofs_on_mother)
  m.resize((face_shape->numberInteriorChildVertices()
     +face_shape->numberCodimension1Shapes())*dofs_per_vertex
    +(face_shape->numberCodimension1Shapes()*childrenPerCodimension2Shape()
     +face_shape->numberInteriorChildCodimension1Shapes())*dofs_per_line
    +childrenPerCodimension1Shape()*dofsPerQuad(face_shape),
     face_shape->numberVertices()*dofs_per_vertex
    +face_shape->numberCodimension1Shapes()*dofs_per_line
    +dofsPerQuad(face_shape));

  int face_index=face_shape->elementDataIndex();
  int neighbor_index=level->neighborIndex(shape,f);
  CHECK_TEST(level->hasChildren(neighbor_index));
  const SHAPE_CLASS *neighbor_shape=level->getShape(neighbor_index);
  int nf=neighbor_shape->faceWithIndexEqualTo(face_index);
  NumPtr<int> line_interior_child_vertices(
    neighbor_shape->numberCodimension2Shapes());
  for (int l=0;l<line_interior_child_vertices.getNumber();l++) {
    int line_index=neighbor_shape->codimension2ShapeIndex(l);
    int lc,lv;
    line_interior_child_vertices[l]=level->Level1::
    interiorChildVertex(line_index,lc,lv);
  }   
  const LEVEL_CLASS *finer_level=level->finerLevel();

  int cpcd1s=childrenPerCodimension1Shape();
  int prolongation_first_row=-1;
  int interior_vertex_child=-1;
//child vertices interior to parent face
  if (face_shape->numberInteriorChildVertices()>0) {
    int fc,fcv;
    int interior_child_vertex_index=level->CODIMENSION1_LEVEL_CLASS::
      interiorChildVertex(face_index,fc,fcv);
    int icv;
    level->interiorChildVertexOnFaceOf(neighbor_shape,nf,
      interior_vertex_child,icv);
    int neighbor_child_index=
      level->childIndex(neighbor_index,interior_vertex_child);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);
    int ncv=neighbor_child_shape->
      vertexEqualToVertexIndex(interior_child_vertex_index);
    prolongation_first_row=ncv;
  }

  int fsnl=face_shape->numberCodimension1Shapes();
  NumPtr<int> prolongation_second_rows(fsnl);
  NumPtr<int> child_for_prolongation_second_rows(fsnl);
//child vertices interior to parent face lines
  for (int l=0;l<fsnl;l++) {
    int face_line_index=face_shape->codimension1ShapeIndex(l);
    int lc,lcv;
    int face_line_interior_vertex_index=level->
      Level1::interiorChildVertex(face_line_index,lc,lcv);
    int face_line_child_index=
      level->Level1::childIndex(face_line_index,lc);
    int fc,fcf;
    level->CODIMENSION1_LEVEL_CLASS::childFaceOn(face_line_child_index,
      face_shape,l,fc,fcf);
    int face_child_index=
      level->CODIMENSION1_LEVEL_CLASS::childIndex(face_shape,fc);
    int nc,ncf;
    level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);
    int neighbor_child_index=level->childIndex(neighbor_shape,nc);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);
    prolongation_second_rows[l]=neighbor_child_shape->
      vertexEqualToVertexIndex(face_line_interior_vertex_index);
    child_for_prolongation_second_rows[l]=nc;
  }

  int niccs=face_shape->numberInteriorChildCodimension1Shapes();
  NumPtr<NumPtr<int> > prolongation_third_rows(niccs);
  NumPtr<int> child_for_prolongation_third_rows(niccs);
//child lines interior to parent face
  if (dofs_per_line>0) {
    for (int i=0;i<niccs;i++) {
      NumPtr<int> &ptr=prolongation_third_rows[i];
      ptr.allocate(dofs_per_line);
      int c0,c1;
      face_shape->interiorChildCodimension1ShapeNeighbors(i,c0,c1);
      int face_child_index0=
        level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,c0);
      int face_child_index1=
        level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,c1);
      const CODIMENSION1_SHAPE_CLASS *face_child_element0=
        finer_level->CODIMENSION1_LEVEL_CLASS::getShape(face_child_index0);
      const CODIMENSION1_SHAPE_CLASS *face_child_element1=
        finer_level->CODIMENSION1_LEVEL_CLASS::getShape(face_child_index1);
      int face_child_line_index=face_child_element0->
        sharedFaceElementIndex(face_child_element1);
      int nc,ncf;
      level->childFaceOn(face_child_index0,neighbor_shape,nf,nc,ncf);
      int neighbor_child_index=level->childIndex(neighbor_shape,nc);
      const SHAPE_CLASS *neighbor_child_shape=
        finer_level->getShape(neighbor_child_index);
      child_for_prolongation_third_rows[i]=nc;
      int ncl=neighbor_child_shape->
        lineWithIndexEqualTo(face_child_line_index);
      int offset=
        neighbor_child_shape->numberVertices()+ncl*dofs_per_line;
      for (int j=0;j<dofs_per_line;j++) {
        ptr[j]=offset+j; 
      }
    }
  }

  NumPtr<NumPtr<NumPtr<int> > > prolongation_fourth_rows(fsnl);
  NumPtr<NumPtr<int> > child_for_prolongation_fourth_rows(fsnl);
  int cpcd2s=childrenPerCodimension2Shape();
//child lines interior to parent face lines
  if (dofs_per_line>0) {
    for (int l=0;l<fsnl;l++) {
      NumPtr<NumPtr<int> > &pfr=prolongation_fourth_rows[l];
      pfr.allocate(cpcd2s);
      NumPtr<int> &cfpfr=child_for_prolongation_fourth_rows[l];
      cfpfr.allocate(cpcd2s);
      int face_line_index=face_shape->codimension1ShapeIndex(l);
      for (int s=0;s<cpcd2s;s++) {
        NumPtr<int> &pfrs=pfr[s];
        pfrs.allocate(dofs_per_line);
        int face_line_child_index=
          level->Level1::childIndex(face_line_index,s);
        int fc,fcl;
        level->CODIMENSION1_LEVEL_CLASS::childFaceOn(face_line_child_index,
          face_shape,l,fc,fcl);
        int face_child_index=
          level->CODIMENSION1_LEVEL_CLASS::childIndex(face_shape,fc);
        int nc,ncf;
        level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);
        int neighbor_child_index=level->childIndex(neighbor_shape,nc);
        const SHAPE_CLASS *neighbor_child_shape=
          finer_level->getShape(neighbor_child_index);
        cfpfr[s]=nc;
        int ncl=neighbor_child_shape->
          lineWithIndexEqualTo(face_line_child_index);
        int offset=neighbor_child_shape->numberVertices()
                  +ncl*dofs_per_line;
        for (int j=0;j<dofs_per_line;j++) {
          pfrs[j]=j+offset;
        }
      }
    }
  }

//face child elements
  NumPtr<NumPtr<int> > prolongation_fifth_rows;
  NumPtr<int> child_for_prolongation_fifth_rows;
  int dpq=dofs_per_quad[face_shape->numberVertices()];
  if (dpq>0) {
    prolongation_fifth_rows.allocate(cpcd1s);
    child_for_prolongation_fifth_rows.allocate(cpcd1s);
    for (int fc=0;fc<cpcd1s;fc++) {
      NumPtr<int> &pfr=prolongation_fifth_rows[fc];
      pfr.allocate(dpq);

      int face_child_index=
        level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,fc);
      const CODIMENSION1_SHAPE_CLASS *face_child_element=
        finer_level->CODIMENSION1_LEVEL_CLASS::getShape(face_child_index);

      int nc,ncf;
      level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);
      int neighbor_child_index=level->childIndex(neighbor_shape,nc);
      const SHAPE_CLASS *neighbor_child_shape=
        finer_level->getShape(neighbor_child_index);
      child_for_prolongation_fifth_rows[fc]=nc;
      int offset=neighbor_child_shape->numberVertices()
        +neighbor_child_shape->numberCodimension2Shapes()*dofs_per_line;
      for (int f=0;f<ncf;f++) {
        offset+=
          dofs_per_quad[neighbor_child_shape->numberVerticesOnFace(f)];
      }
      for (int j=0;j<pfr.getNumber();j++) {
        pfr[j]=j+offset;
      }
    }
  }

  CHECK_TEST(shapeFunctionFamily(neighbor_shape)->requiresShapeToComputeValues());
  {
    for (int fc=0;fc<cpcd1s;fc++) {
      int face_child_index=
        level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,fc);
      int nc,ncf;
      level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);
      prolongationMatrix(neighbor_shape,nc,level);
    }
  }
  const NumPtr<Matrix> &prolongation_matrices=
    prolongationMatrices(neighbor_shape);
  int col=0;
  for (int pv=0;pv<face_shape->numberVertices();pv++) {
//  first columns correspond to vertices of face_shape
    int vi=face_shape->vertexIndex(pv);
    int prolongation_column=neighbor_shape->vertexEqualToVertexIndex(vi);
    processConstraints(components,prolongation_first_row,
      interior_vertex_child,
      prolongation_second_rows,child_for_prolongation_second_rows,
      prolongation_third_rows,child_for_prolongation_third_rows,
      prolongation_fourth_rows,child_for_prolongation_fourth_rows,
      prolongation_fifth_rows,child_for_prolongation_fifth_rows,
      prolongation_matrices,prolongation_column,col,m);
    col++;
  }
  for (int pl=0;pl<fsnl;pl++) {
//  next columns correspond to lines of face_shape
    int nl=neighbor_shape->lineFromFace(face_shape,nf,pl);
    int offset=neighbor_shape->numberVertices()+nl*dofs_per_line;
    for (int i=0;i<dofs_per_line;i++) {
      processConstraints(components,prolongation_first_row,
        interior_vertex_child,prolongation_second_rows,
        child_for_prolongation_second_rows,prolongation_third_rows,
        child_for_prolongation_third_rows,prolongation_fourth_rows,
        child_for_prolongation_fourth_rows,prolongation_fifth_rows,
        child_for_prolongation_fifth_rows,prolongation_matrices,
        i+offset,col,m);
      col++;
    }
    offset+=dofs_per_line;
  }
//final columns correspond to interior of face_shape
  int nieslp=face_shape->numberInteriorEquallySpacedLatticePoints(degree);
  int offset=neighbor_shape->numberVertices()
            +neighbor_shape->numberCodimension2Shapes()*dofs_per_line;
  for (int f=0;f<nf;f++) {
    offset+=dofs_per_quad[neighbor_shape->numberVerticesOnFace(f)];
  }
  for (int i=0;i<nieslp;i++) {
    processConstraints(components,prolongation_first_row,
      interior_vertex_child,
      prolongation_second_rows,child_for_prolongation_second_rows,
      prolongation_third_rows,child_for_prolongation_third_rows,
      prolongation_fourth_rows,child_for_prolongation_fourth_rows,
      prolongation_fifth_rows,child_for_prolongation_fifth_rows,
      prolongation_matrices,i+offset,col,m);
    col++;
  }

  for (int i=0;i<prolongation_third_rows.getNumber();i++) {
    prolongation_third_rows[i].cleanup();
  }
  for (int i=0;i<prolongation_fourth_rows.getNumber();i++) {
    NumPtr<NumPtr<int> > &p4r=prolongation_fourth_rows[i];
    for (int j=0;j<p4r.getNumber();j++) p4r[j].cleanup();
      p4r.cleanup();
    }
  for (int i=0;i<child_for_prolongation_fourth_rows.getNumber();
  i++) {
    child_for_prolongation_fourth_rows[i].cleanup();
  }
  for (int i=0;i<prolongation_fifth_rows.getNumber();i++) {
    prolongation_fifth_rows[i].cleanup();
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FE_H::printOn(ostream &os) const {
  os << "\tFE_H: degree = " << degree << endl;
  for (map<int,ShapeFunction<SPACEDIM>*>::const_iterator
  ps_iterator=polynomial_space.begin();
  ps_iterator!=polynomial_space.end();ps_iterator++) {
    os << "\tpolynomial_space[" << ps_iterator->first << "]: " << endl;
    ps_iterator->second->printOn(os);
  }
  FiniteElement::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
FE_L::FE_L(int d,int nc) : FiniteElement(d,nc) {
  CHECK_POSITIVE(d);
#if (SPACEDIM==1)
  polynomial_space[Interval::verticesPerShape()]=
    OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
#endif
#if (SPACEDIM==2)
  polynomial_space[Quadrilateral::verticesPerShape()]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree);
  polynomial_space[Triangle::verticesPerShape()]=
    OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
#endif
#if (SPACEDIM==3)
  polynomial_space[Hexahedron::verticesPerShape()]=
    OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree);
  polynomial_space[Prism::verticesPerShape()]=
    OPERATOR_NEW C0LagrangePrismPolynomials(degree);
  polynomial_space[Tetrahedron::verticesPerShape()]=
    OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
#endif

  dofs_per_vertex=components;
  dofs_per_line=components*
    Interval::staticNumberInteriorEquallySpacedLatticePoints(degree);
#if (SPACEDIM>1)
  dofs_per_quad[Quadrilateral::verticesPerShape()]=components*
    Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(degree);
  dofs_per_quad[Triangle::verticesPerShape()]=components*
    Triangle::staticNumberInteriorEquallySpacedLatticePoints(degree);
#endif
#if (SPACEDIM==3)
  dofs_per_hex[Hexahedron::verticesPerShape()]=components*
    Hexahedron::staticNumberInteriorEquallySpacedLatticePoints(degree);
  dofs_per_hex[Prism::verticesPerShape()]=components*
    Prism::staticNumberInteriorEquallySpacedLatticePoints(degree);
  dofs_per_hex[Tetrahedron::verticesPerShape()]=components*
    Tetrahedron::staticNumberInteriorEquallySpacedLatticePoints(degree);
#endif
  setDofsPerObject();

  int cps=childrenPerShape();
  int cpcd1s=childrenPerCodimension1Shape();
  for (map<int,ShapeFunction<SPACEDIM>*>::const_iterator 
  sf_iterator=polynomial_space.begin();
  sf_iterator!=polynomial_space.end();sf_iterator++) {
    int nv=sf_iterator->first;
    const ShapeFunction<SPACEDIM> *sf=sf_iterator->second;
    NumPtr<Point<SPACEDIM> > csp;
    equallySpacedLatticePoints(nv,degree,csp);

    NumPtr<Matrix> &prolongation_matrices=prolongation[nv];
    prolongation_matrices.allocate(cps);

    int ncsp=csp.getNumber();
    for (int c=0;c<cps;c++) {
      Matrix &pm=prolongation_matrices[c];
      pm.resize(ncsp,ncsp);
      for (int q=0;q<ncsp;q++) {
//      prolongation is injection of coarse functions into fine space
        NumPtr<REAL> v;
        sf->computeValues(
          parentCoordinateOfCanonical(nv,csp[q],c),v);
        for (int s=0;s<ncsp;s++) pm(q,s)=v[s];
      }
#ifdef DEBUG
      for (int q=0;q<ncsp;q++) {
        REAL sum=0.;
        for (int s=0;s<ncsp;s++) { 
          sum+=pm(q,s);
        }
        if (abs(sum-1.)>=512.*DBL_EPSILON) {
          cout << "\tprolongation matrix sum[" << q << "] = " << sum
               << " " << abs(sum-1.)/DBL_EPSILON << endl;
          CHECK_TEST(abs(sum-1.)<512.*DBL_EPSILON);
        }
      }
#endif
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from ConstraintMatrix::makeHangingNodeConstraints
#if (SPACEDIM==2)
void FE_L::constraints(const LEVEL_CLASS *level,const SHAPE_CLASS *shape,
const CODIMENSION1_SHAPE_CLASS *face_shape,int f,Matrix &m) {
  CHECK_SAME(m.size(0),0);
  CHECK_SAME(m.size(1),0);
// (n_dofs_on_children,n_dofs_on_mother)
  m.resize(face_shape->numberInteriorChildVertices()*dofs_per_vertex
    +dofs_per_line*childrenPerCodimension1Shape(),
    face_shape->numberVertices()*dofs_per_vertex+dofs_per_line);

  int face_index=face_shape->elementDataIndex();
  int neighbor_index=level->neighborIndex(shape,f);
  CHECK_TEST(level->hasChildren(neighbor_index));
  const SHAPE_CLASS *neighbor_shape=level->getShape(neighbor_index);
  int nf=neighbor_shape->faceWithIndexEqualTo(face_index);

  NumPtr<int> line_interior_child_vertices(
    neighbor_shape->numberCodimension1Shapes());
  for (int l=0;l<line_interior_child_vertices.getNumber();l++) {
    int face_index=neighbor_shape->codimension1ShapeIndex(l);
    int fc,fv;
    line_interior_child_vertices[l]=level->CODIMENSION1_LEVEL_CLASS::
      interiorChildVertex(face_index,fc,fv);
  }
  const LEVEL_CLASS *finer_level=level->finerLevel();

  map<int,NumPtr<int> > renumberfc_map;
  int prolongation_first_row=-1;
  int interior_vertex_child=-1;
  if (face_shape->numberInteriorChildVertices()>0) {
//  first row corresponds to interior child vertex (if it exists)
    int fc,fcv;
    int interior_child_vertex_index=level->CODIMENSION1_LEVEL_CLASS::
      interiorChildVertex(face_index,fc,fcv);
    int icv;
    level->interiorChildVertexOnFaceOf(neighbor_shape,nf,
      interior_vertex_child,icv);
    int neighbor_child_index=
      level->childIndex(neighbor_index,interior_vertex_child);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);
    int ncv=neighbor_child_shape->
      vertexEqualToVertexIndex(interior_child_vertex_index);
    int eslpi=neighbor_child_shape->
      equallySpacedLatticePointIndexOfVertex(ncv,degree);

// eslpi is relative to neighbor_child_shape vertex order
// renumberfc[eslpi] must refer to point used by 
//   Shape::parentCoordinateOfCanonical in the computation of the
//   prolongation matrices
    map<int,NumPtr<int> >::iterator renumberfc_map_iterator=
      renumberfc_map.find(interior_vertex_child);
    if (renumberfc_map_iterator==renumberfc_map.end()) {
      NumPtr<int> &renumberfc=renumberfc_map[interior_vertex_child];
      neighbor_shape->reorderChildEquallySpacedLatticePoints(degree,
        neighbor_child_shape,interior_vertex_child,
        line_interior_child_vertices,renumberfc);
      renumberfc_map_iterator=renumberfc_map.find(interior_vertex_child);
    }
    NumPtr<int> &renumberfc=renumberfc_map_iterator->second;
    prolongation_first_row=renumberfc[eslpi];
  }

  int cpcd1s=childrenPerCodimension1Shape();
  NumPtr<NumPtr<int> > prolongation_rows(cpcd1s);
  NumPtr<int> neighbor_child_on_face(cpcd1s);
  map<int,NumPtr<int> > renumberf_map;
  map<int,NumPtr<int> > renumberf_inverse_map;

//other rows correspond to vertices interior to child faces
  for (int s=0;s<cpcd1s;s++) {
    int face_child_index=
      level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,s);
    int nc=-1,ncf=-1;
    level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);

//  renumberfc[parent fe_s]=child fe_s
    neighbor_child_on_face[s]=nc;
    int neighbor_child_index=
      level->childIndex(neighbor_index,nc);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);

//  renumberf[fe_s]=dof_s
    map<int,NumPtr<int> >::iterator renumberf_map_iterator=
      renumberf_map.find(nc);
    if (renumberf_map_iterator==renumberf_map.end()) {
      NumPtr<int> &renumberf=renumberf_map[nc];
      neighbor_child_shape->
        reorderEquallySpacedLatticePoints(degree,renumberf);
      renumberf_map_iterator=renumberf_map.find(nc);
      NumPtr<int> &renumberf_inverse=renumberf_inverse_map[nc];
      renumberf_inverse.allocate(renumberf.getNumber());
      for (int i=0;i<renumberf.getNumber();i++) {
        renumberf_inverse[renumberf[i]]=i;
      }
    }
    NumPtr<int> &renumberf=renumberf_map_iterator->second;
    NumPtr<int> &renumberf_inverse=renumberf_inverse_map.find(nc)->second;

//  prolongation rows correspond to childPoint(parent lattice point)
//    we reorder these to child lattice point order
    NumPtr<int> rfi(renumberf_inverse,
      neighbor_child_shape->numberVertices()+ncf*dofs_per_line,
      dofs_per_line);
    map<int,NumPtr<int> >::iterator renumberfc_map_iterator=
      renumberfc_map.find(nc);
    if (renumberfc_map_iterator==renumberfc_map.end()) {
      NumPtr<int> &renumberfc=renumberfc_map[nc];
      neighbor_shape->reorderChildEquallySpacedLatticePoints(degree,
        neighbor_child_shape,nc,
        line_interior_child_vertices,renumberfc);
      renumberfc_map_iterator=renumberfc_map.find(nc);
    }
    NumPtr<int> &renumberfc=renumberfc_map_iterator->second;

    NumPtr<int> &prc=prolongation_rows[s];
    prc.allocate(rfi.getNumber());
    for (int j=0;j<prc.getNumber();j++) {
      prc[j]=renumberfc[rfi[j]];
    }
  }

  const NumPtr<Matrix> &prolongation_matrices=
    prolongationMatrices(neighbor_shape);

  NumPtr<int> renumberc;
  neighbor_shape->reorderEquallySpacedLatticePoints(degree,renumberc);
  NumPtr<int> renumberc_inverse(renumberc.getNumber());
  for (int i=0;i<renumberc.getNumber();i++) {
    renumberc_inverse[renumberc[i]]=i;
  }

  int col=0;
  for (int pv=0;pv<face_shape->numberVertices();pv++) {
//  first columns correspond to vertices of face_shape
    int vi=face_shape->vertexIndex(pv);
    int v=neighbor_shape->vertexEqualToVertexIndex(vi);
    int prolongation_column=
      neighbor_shape->equallySpacedLatticePointIndexOfVertex(v,degree); 
    processConstraints(components,prolongation_first_row,
      interior_vertex_child,neighbor_child_on_face,prolongation_rows,
      prolongation_matrices,prolongation_column,col,m);
    col++;
  }
  NumPtr<int> interior_vertices;
  neighbor_shape->equallySpacedLatticePointIndicesOfFaceInterior(nf,
    degree,interior_vertices);
  NumPtr<int> rnci(renumberc_inverse,neighbor_shape->numberVertices()
    +nf*dofs_per_line,dofs_per_line);
  for (int j=0;j<rnci.getNumber();j++) {
//  other columns correspond to vertices interior to parent face
    int prolongation_column=rnci[j];
    processConstraints(components,prolongation_first_row,
      interior_vertex_child,neighbor_child_on_face,prolongation_rows,
      prolongation_matrices,prolongation_column,col,m);
    col++;
  }
  for (int c=0;c<cpcd1s;c++) prolongation_rows[c].cleanup();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from ConstraintMatrix::makeHangingNodeConstraints
#if (SPACEDIM==3)
void FE_L::constraints(const LEVEL_CLASS *level,const SHAPE_CLASS *shape,
const CODIMENSION1_SHAPE_CLASS *face_shape,int f,Matrix &m) {
  CHECK_SAME(m.size(0),0);
  CHECK_SAME(m.size(1),0);
// (n_dofs_on_children,n_dofs_on_mother)
  int fsl=face_shape->numberCodimension1Shapes();
  int dpq=dofsPerQuad(face_shape);
  m.resize((face_shape->numberInteriorChildVertices()+fsl)*dofs_per_vertex
    +(fsl*childrenPerCodimension2Shape()
     +face_shape->numberInteriorChildCodimension1Shapes())*dofs_per_line
    +childrenPerCodimension1Shape()*dpq,
     face_shape->numberVertices()*dofs_per_vertex+fsl*dofs_per_line+dpq);

  int face_index=face_shape->elementDataIndex();
  int neighbor_index=level->neighborIndex(shape,f);
  CHECK_TEST(level->hasChildren(neighbor_index));
  const SHAPE_CLASS *neighbor_shape=level->getShape(neighbor_index);
  int nf=neighbor_shape->faceWithIndexEqualTo(face_index);
  NumPtr<int> line_interior_child_vertices(
    neighbor_shape->numberCodimension2Shapes());
  for (int l=0;l<line_interior_child_vertices.getNumber();l++) {
    int line_index=neighbor_shape->codimension2ShapeIndex(l);
    int lc,lv;
    line_interior_child_vertices[l]=level->Level1::
      interiorChildVertex(line_index,lc,lv);
  }

  map<int,NumPtr<int> > renumberfc_map;
  const LEVEL_CLASS *finer_level=level->finerLevel();
  int prolongation_first_row=-1;
  int interior_vertex_child=-1;
  if (face_shape->numberInteriorChildVertices()>0) {
    int fc,fcv;
    int interior_child_vertex_index=level->CODIMENSION1_LEVEL_CLASS::
      interiorChildVertex(face_index,fc,fcv);
    int icv;
    level->interiorChildVertexOnFaceOf(neighbor_shape,nf,
      interior_vertex_child,icv);
    int neighbor_child_index=
      level->childIndex(neighbor_index,interior_vertex_child);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);
    int ncv=neighbor_child_shape->
      vertexEqualToVertexIndex(interior_child_vertex_index);
    int eslpi=neighbor_child_shape->
      equallySpacedLatticePointIndexOfVertex(ncv,degree);

// eslpi is relative to neighbor_child_shape vertex order
// renumberfc[eslpi] must refer to point used by 
//   Shape::parentCoordinateOfCanonical in the computation of the
//   prolongation matrices
    map<int,NumPtr<int> >::iterator renumberfc_map_iterator=
      renumberfc_map.find(interior_vertex_child);
    if (renumberfc_map_iterator==renumberfc_map.end()) {
      NumPtr<int> &renumberfc=renumberfc_map[interior_vertex_child];
      neighbor_shape->reorderChildEquallySpacedLatticePoints(degree,
        neighbor_child_shape,interior_vertex_child,
        line_interior_child_vertices,renumberfc);
      renumberfc_map_iterator=renumberfc_map.find(interior_vertex_child);
    }
    NumPtr<int> &renumberfc=renumberfc_map_iterator->second;
    prolongation_first_row=renumberfc[eslpi];
  }

  int fsnl=face_shape->numberCodimension1Shapes();
  NumPtr<int> prolongation_second_rows(fsnl);
  NumPtr<int> child_for_prolongation_second_rows(fsnl);
  map<int,NumPtr<int> > renumberf_map;
  map<int,NumPtr<int> > renumberf_inverse_map;
//child vertices interior to parent face lines
  for (int l=0;l<fsnl;l++) {
    int face_line_index=face_shape->codimension1ShapeIndex(l);
    int lc,lcv;
    int face_line_interior_vertex_index=level->
      Level1::interiorChildVertex(face_line_index,lc,lcv);
    int face_line_child_index=
      level->Level1::childIndex(face_line_index,lc);
    int fc,fcf;
    level->CODIMENSION1_LEVEL_CLASS::childFaceOn(face_line_child_index,
      face_shape,l,fc,fcf);
    int face_child_index=
      level->CODIMENSION1_LEVEL_CLASS::childIndex(face_shape,fc);
    int nc,ncf;
    level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);
    int neighbor_child_index=level->childIndex(neighbor_shape,nc);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);
    int ncv=neighbor_child_shape->
      vertexEqualToVertexIndex(face_line_interior_vertex_index);
//  renumberf[fe_s]=dof_s
    map<int,NumPtr<int> >::iterator renumberf_map_iterator=
      renumberf_map.find(nc);
    if (renumberf_map_iterator==renumberf_map.end()) {
      NumPtr<int> &renumberf=renumberf_map[nc];
      neighbor_child_shape->
        reorderEquallySpacedLatticePoints(degree,renumberf);
      renumberf_map_iterator=renumberf_map.find(nc);
      NumPtr<int> &renumberf_inverse=renumberf_inverse_map[nc];
      renumberf_inverse.allocate(renumberf.getNumber());
      for (int i=0;i<renumberf.getNumber();i++) {
        renumberf_inverse[renumberf[i]]=i;
      }
    }
    NumPtr<int> &renumberf=renumberf_map_iterator->second;
    NumPtr<int> &renumberf_inverse=renumberf_inverse_map.find(nc)->second;
    map<int,NumPtr<int> >::iterator renumberfc_map_iterator=
      renumberfc_map.find(nc);
    if (renumberfc_map_iterator==renumberfc_map.end()) {
      NumPtr<int> &renumberfc=renumberfc_map[nc];
      neighbor_shape->reorderChildEquallySpacedLatticePoints(degree,
        neighbor_child_shape,nc,
        line_interior_child_vertices,renumberfc);
      renumberfc_map_iterator=renumberfc_map.find(nc);
    }
    NumPtr<int> &renumberfc=renumberfc_map_iterator->second;
    prolongation_second_rows[l]=renumberfc[renumberf_inverse[ncv]];
    child_for_prolongation_second_rows[l]=nc;
  }

  NumPtr<NumPtr<int> > prolongation_third_rows(
    face_shape->numberInteriorChildCodimension1Shapes());
  NumPtr<int> child_for_prolongation_third_rows(
    face_shape->numberInteriorChildCodimension1Shapes());
//child lines interior to parent face
  for (int i=0;i<face_shape->numberInteriorChildCodimension1Shapes();
  i++) {
    NumPtr<int> &ptr=prolongation_third_rows[i];
    ptr.allocate(dofs_per_line);

    int c0,c1;
    face_shape->interiorChildCodimension1ShapeNeighbors(i,c0,c1);
    int face_child_index0=
      level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,c0);
    int face_child_index1=
      level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,c1);
    const Shape2 *face_child_element0=
      finer_level->CODIMENSION1_LEVEL_CLASS::getShape(face_child_index0);
    const Shape2 *face_child_element1=
      finer_level->CODIMENSION1_LEVEL_CLASS::getShape(face_child_index1);
    int face_child_line_index=face_child_element0->
      sharedFaceElementIndex(face_child_element1);
    int fcl=face_child_element0->
      faceWithIndexEqualTo(face_child_line_index);
    int nc,ncf;
    level->childFaceOn(face_child_index0,neighbor_shape,nf,nc,ncf);
    child_for_prolongation_third_rows[i]=nc;
    int neighbor_child_index=level->childIndex(neighbor_shape,nc);
    const SHAPE_CLASS *neighbor_child_shape=
      finer_level->getShape(neighbor_child_index);
    int ncl=
      neighbor_child_shape->lineFromFace(face_child_element0,ncf,fcl);
//  renumberf[fe_s]=dof_s
    map<int,NumPtr<int> >::iterator renumberf_map_iterator=
      renumberf_map.find(nc);
    if (renumberf_map_iterator==renumberf_map.end()) {
      NumPtr<int> &renumberf=renumberf_map[nc];
      neighbor_child_shape->
        reorderEquallySpacedLatticePoints(degree,renumberf);
      renumberf_map_iterator=renumberf_map.find(nc);
      NumPtr<int> &renumberf_inverse=renumberf_inverse_map[nc];
      renumberf_inverse.allocate(renumberf.getNumber());
      for (int i=0;i<renumberf.getNumber();i++) {
        renumberf_inverse[renumberf[i]]=i;
      }
    }
    NumPtr<int> &renumberf=renumberf_map_iterator->second;
    NumPtr<int> &renumberf_inverse=renumberf_inverse_map.find(nc)->second;
//  prolongation rows correspond to childPoint(parent lattice point)
//    we reorder these to child lattice point order
    NumPtr<int> rfi(renumberf_inverse,
      neighbor_child_shape->numberVertices()+ncl*dofs_per_line,
      dofs_per_line);
    map<int,NumPtr<int> >::iterator renumberfc_map_iterator=
      renumberfc_map.find(nc);
    if (renumberfc_map_iterator==renumberfc_map.end()) {
      NumPtr<int> &renumberfc=renumberfc_map[nc];
      neighbor_shape->reorderChildEquallySpacedLatticePoints(degree,
        neighbor_child_shape,nc,
        line_interior_child_vertices,renumberfc);
      renumberfc_map_iterator=renumberfc_map.find(nc);
    }
    NumPtr<int> &renumberfc=renumberfc_map_iterator->second;
    for (int j=0;j<ptr.getNumber();j++) {
      ptr[j]=renumberfc[rfi[j]];
    }
  }

  NumPtr<NumPtr<NumPtr<int> > > prolongation_fourth_rows(fsnl);
  NumPtr<NumPtr<int> > child_for_prolongation_fourth_rows(fsnl);
  int cpcd2s=childrenPerCodimension2Shape();
//child lines interior to parent face lines
  for (int l=0;l<fsnl;l++) {
    NumPtr<NumPtr<int> > &pfr=prolongation_fourth_rows[l];
    pfr.allocate(cpcd2s);
    NumPtr<int> &cfpfr=child_for_prolongation_fourth_rows[l];
    cfpfr.allocate(cpcd2s);

    int face_line_index=face_shape->codimension1ShapeIndex(l);
    for (int s=0;s<cpcd2s;s++) {
      NumPtr<int> &pfrs=pfr[s];
      pfrs.allocate(dofs_per_line);

      int face_line_child_index=
        level->Level1::childIndex(face_line_index,s);
      int fc,fcl;
      level->CODIMENSION1_LEVEL_CLASS::childFaceOn(face_line_child_index,
        face_shape,l,fc,fcl);
      int face_child_index=
        level->CODIMENSION1_LEVEL_CLASS::childIndex(face_shape,fc);
      const CODIMENSION1_SHAPE_CLASS *face_child_shape=finer_level->
        CODIMENSION1_LEVEL_CLASS::getShape(face_child_index);
      int nc,ncf;
      level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);
      cfpfr[s]=nc;
      int neighbor_child_index=level->childIndex(neighbor_shape,nc);
      const SHAPE_CLASS *neighbor_child_shape=
        finer_level->getShape(neighbor_child_index);
      int ncl=
        neighbor_child_shape->lineFromFace(face_child_shape,ncf,fcl);
//    renumberf[fe_s]=dof_s
      map<int,NumPtr<int> >::iterator renumberf_map_iterator=
        renumberf_map.find(nc);
      if (renumberf_map_iterator==renumberf_map.end()) {
        NumPtr<int> &renumberf=renumberf_map[nc];
        neighbor_child_shape->
          reorderEquallySpacedLatticePoints(degree,renumberf);
        renumberf_map_iterator=renumberf_map.find(nc);
        NumPtr<int> &renumberf_inverse=renumberf_inverse_map[nc];
        renumberf_inverse.allocate(renumberf.getNumber());
        for (int i=0;i<renumberf.getNumber();i++) {
          renumberf_inverse[renumberf[i]]=i;
        }
      }
      NumPtr<int> &renumberf=renumberf_map_iterator->second;
      NumPtr<int> &renumberf_inverse=
        renumberf_inverse_map.find(nc)->second;
//    prolongation rows correspond to childPoint(parent lattice point)
//      we reorder these to child lattice point order
      NumPtr<int> rfi(renumberf_inverse,
        neighbor_child_shape->numberVertices()+ncl*dofs_per_line,
        dofs_per_line);
      map<int,NumPtr<int> >::iterator renumberfc_map_iterator=
        renumberfc_map.find(nc);
      if (renumberfc_map_iterator==renumberfc_map.end()) {
        NumPtr<int> &renumberfc=renumberfc_map[nc];
        neighbor_shape->reorderChildEquallySpacedLatticePoints(degree,
          neighbor_child_shape,nc,
          line_interior_child_vertices,renumberfc);
        renumberfc_map_iterator=renumberfc_map.find(nc);
      }
      NumPtr<int> &renumberfc=renumberfc_map_iterator->second;
      for (int j=0;j<pfrs.getNumber();j++) {
        pfrs[j]=renumberfc[rfi[j]];
      }
    }
  }

//face child elements
  NumPtr<NumPtr<int> > prolongation_fifth_rows;
  NumPtr<int> child_for_prolongation_fifth_rows;
  if (dpq>0) {
    int cpcd1s=childrenPerCodimension1Shape();
    prolongation_fifth_rows.allocate(cpcd1s);
    child_for_prolongation_fifth_rows.allocate(cpcd1s);
    for (int fc=0;fc<cpcd1s;fc++) {
      int face_child_index=
        level->CODIMENSION1_LEVEL_CLASS::childIndex(face_index,fc);
      const CODIMENSION1_SHAPE_CLASS *face_child_element=
        finer_level->CODIMENSION1_LEVEL_CLASS::getShape(face_child_index);

      int nc,ncf;
      level->childFaceOn(face_child_index,neighbor_shape,nf,nc,ncf);
      int neighbor_child_index=level->childIndex(neighbor_shape,nc);
      const SHAPE_CLASS *neighbor_child_shape=
        finer_level->getShape(neighbor_child_index);
      map<int,NumPtr<int> >::iterator renumberf_map_iterator=
        renumberf_map.find(nc);
      if (renumberf_map_iterator==renumberf_map.end()) {
        NumPtr<int> &renumberf=renumberf_map[nc];
        neighbor_child_shape->
          reorderEquallySpacedLatticePoints(degree,renumberf);
        renumberf_map_iterator=renumberf_map.find(nc);
        NumPtr<int> &renumberf_inverse=renumberf_inverse_map[nc];
        renumberf_inverse.allocate(renumberf.getNumber());
        for (int i=0;i<renumberf.getNumber();i++) {
          renumberf_inverse[renumberf[i]]=i;
        }
      }
      NumPtr<int> &renumberf=renumberf_map_iterator->second;
      NumPtr<int> &renumberf_inverse=
        renumberf_inverse_map.find(nc)->second;
      int offset=neighbor_child_shape->numberVertices()
        +neighbor_child_shape->numberCodimension2Shapes()*dofs_per_line;
      for (int f2=0;f2<ncf;f2++) {
        offset+=
          dofs_per_quad[neighbor_child_shape->numberVerticesOnFace(f2)];
      }
      NumPtr<int> rfi(renumberf_inverse,offset,dpq);
      child_for_prolongation_fifth_rows[fc]=nc;
      NumPtr<int> &pfr=prolongation_fifth_rows[fc];
      pfr.allocate(dpq);

      map<int,NumPtr<int> >::iterator renumberfc_map_iterator=
        renumberfc_map.find(nc);
      if (renumberfc_map_iterator==renumberfc_map.end()) {
        NumPtr<int> &renumberfc=renumberfc_map[nc];
        neighbor_shape->reorderChildEquallySpacedLatticePoints(degree,
          neighbor_child_shape,nc,
          line_interior_child_vertices,renumberfc);
        renumberfc_map_iterator=renumberfc_map.find(nc);
      }
      NumPtr<int> &renumberfc=renumberfc_map_iterator->second;
      for (int j=0;j<pfr.getNumber();j++) {
        pfr[j]=renumberfc[rfi[j]];
      }
    }
  }

  const NumPtr<Matrix> &prolongation_matrices=prolongationMatrices(shape);
  int col=0;
  for (int pv=0;pv<face_shape->numberVertices();pv++) {
    int vi=face_shape->vertexIndex(pv);
    int v=neighbor_shape->vertexEqualToVertexIndex(vi);
    int prolongation_column=
      neighbor_shape->equallySpacedLatticePointIndexOfVertex(v,degree);
    processConstraints(components,prolongation_first_row,
      interior_vertex_child,
      prolongation_second_rows,child_for_prolongation_second_rows,
      prolongation_third_rows,child_for_prolongation_third_rows,
      prolongation_fourth_rows,child_for_prolongation_fourth_rows,
      prolongation_fifth_rows,child_for_prolongation_fifth_rows,
      prolongation_matrices,prolongation_column,col,m);
    col++;
  }
  for (int pl=0;pl<face_shape->numberCodimension1Shapes();pl++) {
    int nl=neighbor_shape->lineFromFace(face_shape,nf,pl);
    NumPtr<int> indices;
    neighbor_shape->
      equallySpacedLatticePointIndicesOfLineInterior(nl,degree,indices);
    for (int i=0;i<indices.getNumber();i++) {
      processConstraints(components,prolongation_first_row,
        interior_vertex_child,
        prolongation_second_rows,child_for_prolongation_second_rows,
        prolongation_third_rows,child_for_prolongation_third_rows,
        prolongation_fourth_rows,child_for_prolongation_fourth_rows,
        prolongation_fifth_rows,child_for_prolongation_fifth_rows,
        prolongation_matrices,indices[i],col,m);
      col++;
    }
  }
  NumPtr<int> indices;
  neighbor_shape->equallySpacedLatticePointIndicesOfFaceInterior(nf,
    degree,indices);
  for (int i=0;i<indices.getNumber();i++) {
    processConstraints(components,prolongation_first_row,
      interior_vertex_child,
      prolongation_second_rows,child_for_prolongation_second_rows,
      prolongation_third_rows,child_for_prolongation_third_rows,
      prolongation_fourth_rows,child_for_prolongation_fourth_rows,
      prolongation_fifth_rows,child_for_prolongation_fifth_rows,
      prolongation_matrices,indices[i],col,m);
    col++;
  }

  for (int i=0;i<prolongation_third_rows.getNumber();i++) {
    prolongation_third_rows[i].cleanup();
  }
  for (int i=0;i<prolongation_fourth_rows.getNumber();i++) {
    NumPtr<NumPtr<int> > &p4r=prolongation_fourth_rows[i];
    for (int j=0;j<p4r.getNumber();j++) p4r[j].cleanup();
    p4r.cleanup();
  }
  for (int i=0;i<child_for_prolongation_fourth_rows.getNumber();
  i++) {
    child_for_prolongation_fourth_rows[i].cleanup();
  }
  for (int i=0;i<prolongation_fifth_rows.getNumber();i++) {
    prolongation_fifth_rows[i].cleanup();
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FE_L::printOn(ostream &os) const {
  os << "\tFE_L:" << endl;
  FiniteElement::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

#include "NumPtr.C"
template class NumPtrBase<NumPtr<NumPtr<int> >,int>;
template class NumPtr<NumPtr<NumPtr<int> > >;

INSTANTIATE_NUMPTR(Matrix);
#if (SPACEDIM>1)
template class NumPtrBase<NumPtr<Matrix>,int>;
template class NumPtr<NumPtr<Matrix> >;
#endif
template class NumPtrBase<NumPtr<bool>,int>;
template class NumPtr<NumPtr<bool> >;
template class NumPtrBase<pair<int,int>,int>;
template class NumPtr<pair<int,int> >;
template class NumPtrBase<pair<pair<int,int>,int>,int>;
template class NumPtr<pair<pair<int,int>,int> >;

template class _Rb_tree<
  int,
  pair<int const,int>,
  _Select1st<pair<int const,int> >,
  less<int>,
  allocator<pair<int const, int> > 
>;

template class _Rb_tree<
  int, 
  pair<int const,bool>, 
  _Select1st<pair<int const,bool> >, 
  less<int>, 
  allocator<pair<int const, bool> > 
>;

template class _Rb_tree<
  int, 
  pair<int const,NumPtr<int> >, 
  _Select1st<pair<int const, NumPtr<int> > >, 
  less<int>, 
  allocator<pair<int const, NumPtr<int> > > 
>;

template class _Rb_tree<
  int, 
  pair<int const,NumPtr<bool> >, 
  _Select1st<pair<int const,NumPtr<bool> > >, 
  less<int>, 
  allocator<pair<int const, NumPtr<bool> > > 
>;

template class _Rb_tree<
  int, 
  pair<int const, NumPtr<NumPtr<bool> > >, 
  _Select1st<pair<int const, NumPtr<NumPtr<bool> > > >, 
  less<int>, 
  allocator<pair<int const, NumPtr<NumPtr<bool> > > > 
>;

template class _Rb_tree<
  int, 
  pair<int const,NumPtr<pair<int,int> > >, 
  _Select1st<
    pair<int const,NumPtr<pair<int,int> > > 
  >, 
  less<int>, 
  allocator<
    pair<int const,NumPtr<pair<int, int> > > 
  >
>;

template class _Rb_tree<
  int, 
  pair<
    int const, 
    NumPtr<pair<pair<int, int>, int> > 
  >, 
  _Select1st<
    pair<
      int const, 
      NumPtr<pair<pair<int, int>, int> > 
    > 
  >, 
  less<int>, 
  allocator<
    pair<
      int const, 
      NumPtr<pair<pair<int, int>, int> > 
    > 
  > 
>;

template class _Rb_tree<
  int, 
  pair<int const,NumPtr<Point<0> > >, 
  _Select1st<pair<int const,NumPtr<Point<0> > > >, 
  less<int>, 
  allocator<pair<int const, NumPtr<Point<0> > > > 
>;

template class _Rb_tree<
  int, 
  pair<int const,NumPtr<Matrix> >, 
  _Select1st<pair<int const,NumPtr<Matrix> > >,
  less<int>, 
  allocator<pair<int const,NumPtr<Matrix> > >
>;

template class _Rb_tree<
  int, 
  pair<const int, ShapeFunction<SPACEDIM>*>, 
  _Select1st<pair<const int,
             ShapeFunction<SPACEDIM>*> 
            >, 
  less<int>, 
  allocator<pair<const int,ShapeFunction<SPACEDIM>*> >
>;
template class _Rb_tree<
  int, 
  pair<const int, NumPtr<Point<SPACEDIM> > >, 
  _Select1st<pair<const int, NumPtr<Point<SPACEDIM> > > >, 
  less<int>, 
  allocator<pair<const int, NumPtr<Point<SPACEDIM> > > > 
>;
#if (SPACEDIM>1)
template class _Rb_tree<
  int, 
  pair<const int, NumPtr<Point<SPACEDIM-1> > >, 
  _Select1st<pair<const int, NumPtr<Point<SPACEDIM-1> > > >, 
  less<int>, 
  allocator<pair<const int, NumPtr<Point<SPACEDIM-1> > > > 
>;
template class _Rb_tree<
  int, 
  pair<int const, NumPtr<NumPtr<Matrix> > >, 
  _Select1st<pair<int const, NumPtr<NumPtr<Matrix> > > >, 
  less<int>, 
  allocator<pair<int const, NumPtr<NumPtr<Matrix> > > > 
>;
#endif
#if (SPACEDIM==3)
template class _Rb_tree<
  int, 
  pair<int const, Shape*>, 
  _Select1st<pair<int const, Shape*> >, 
  less<int>, 
  allocator<pair<int const, Shape*> > 
>;
#endif
template class _Rb_tree<
  int, 
  pair<int const, Quadrature<SPACEDIM>*>, 
  _Select1st<pair<int const, Quadrature<SPACEDIM>*> >, 
  less<int>, 
  allocator<pair<int const, Quadrature<SPACEDIM>*> > 
>;
