#ifndef FINITE_ELEMENT_H
#define FINITE_ELEMENT_H "$Header:$"
//----------------------------  fe.h  ---------------------------
//   $Id: fe.h,v 1.112 2003/01/08 17:57:51 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  fe.h  ---------------------------
//
//modified from deal.II/deal.II/include/fe/fe.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <BoundaryCondition.H>
#include <Mapping.H> // for UPDATE_FLAGS, not Mapping
#include <Matrix.H>

// data for finite element shape function(xi)
//compare to DEAL FiniteElementBase::InternalDataBase
class FiniteElementDataBase {
  private:
    UPDATE_FLAGS update_flags;
    bool shape_values_ok;
    bool shape_derivatives_ok;
    bool shape_second_derivatives_ok;
// shape_values(s,q) = shape_s(xi_q); ie, defined on canonical elements
    Array<2,REAL> shape_values;
    Array<2,Tensor<1,SPACEDIM> > shape_derivatives;
    Array<2,Tensor<2,SPACEDIM> > shape_second_derivatives;
    const Quadrature<SPACEDIM> *quadrature;
    const ShapeFunction<SPACEDIM> *shape_function_family;

    FiniteElementDataBase(const FiniteElementDataBase&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    FiniteElementDataBase(UPDATE_FLAGS,const Quadrature<SPACEDIM>*,
      const ShapeFunction<SPACEDIM>*);
    FiniteElementDataBase(const FiniteElementDataBase*,const Shape*);
    virtual ~FiniteElementDataBase() {;}
    UPDATE_FLAGS updateFlags() const { return update_flags; }
    const Quadrature<SPACEDIM> *getQuadrature() const {
      return quadrature;
    }
    bool shapeValuesOK() const { return shape_values_ok; }
    bool shapeDerivativesOK() const { return shape_derivatives_ok; }
    bool shapeSecondDerivativesOK() const { 
      return shape_second_derivatives_ok; 
    }
    const REAL& shapeValue(int s,int q) const { 
      CHECK_TEST(shape_values_ok);
      return shape_values(s,q); 
    }
    const Tensor<1,SPACEDIM>& shapeDerivative(int s,int q) const {
      CHECK_TEST(shape_derivatives_ok);
      return shape_derivatives(s,q);
    }
    const Tensor<2,SPACEDIM>& shapeSecondDerivative(int s,int q) const {
      CHECK_TEST(shape_second_derivatives_ok);
      return shape_second_derivatives(s,q);
    }
    const Array<2,REAL>& getShapeValues() const { 
      CHECK_TEST(shape_values_ok);
      return shape_values; 
    }
    const Array<2,Tensor<1,SPACEDIM> >& getShapeDerivatives() const { 
      CHECK_TEST(shape_derivatives_ok);
      return shape_derivatives;
    }
    const Array<2,Tensor<2,SPACEDIM> >& getShapeSecondDerivatives() const { 
      CHECK_TEST(shape_second_derivatives_ok);
      return shape_second_derivatives;
    }
    virtual void printOn(ostream&) const;
    void computeShapesFor(const Shape*);
};

class FiniteElement {
  protected:
    int degree;
    map<int,ShapeFunction<SPACEDIM>*> polynomial_space;

//  the following = components * shape_functions per whatever
    int dofs_per_vertex;
    int dofs_per_line;
#if (SPACEDIM>1)
    map<int,int> dofs_per_quad; // 2D Shapes only
    map<int,int> dofs_per_face; // codimension 1 Shapes only
#endif
#if (SPACEDIM==3)
    map<int,int> dofs_per_hex; // 3D Shapes only
#endif
    map<int,int> dofs_per_cell; // sum over dimension and codimensions
    int components;

//  the following used to be in FiniteElementBase:
//  We want the fine solution function to equal the coarse solution fcn:
//    b_s(xi_q) c_s = B_S(parent coord of child xi_q) C_S
//  (Einstein summation convention: repeated subscript ==> sum)
//  Thus if [ p_{d,s} ] is permutation that reorders 
//    from static fe order (index s)
//    to shape-dependent dof order on the child (index d), and
//  if [ Q_{D,S} ] is the permuation that reorders 
//    from fe (index S) 
//    to dof  (index D) on parent,
//  then
//    c_d = p_{d,s} [ b_s(xi_q) ]^{-1} 
//          [ B_S(parent coord of child xi_q) ] Q_{D,S}^T C_D
    map<int,NumPtr<Matrix> > prolongation;

    const NumPtr<Matrix>& prolongationMatrices(int nv) const {
      map<int,NumPtr<Matrix> >::const_iterator i=
        prolongation.find(nv);
      CHECK_TEST(i!=prolongation.end());
      return i->second;
    }
    virtual const NumPtr<Matrix>& prolongationMatrices(const Shape *e) 
    const {
      return prolongationMatrices(e->numberVertices());
    }
#if (SPACEDIM>1)
    int interfaceConstraintsSize(const Shape*,int) const;
#endif
//  call from derived class constructor after setting 
//    dofs_per_vertex,dofs_per_line,dofs_per_quad,dofs_per_hex
    virtual void setDofsPerObject();
  public:
    FiniteElement(int d,int nc) : components(nc),degree(d) { }
    virtual ~FiniteElement();

    int dofsPerVertex() const { return dofs_per_vertex; }
    virtual int dofsPerLine() const { return dofs_per_line; }
#if (SPACEDIM>1)
    virtual int dofsPerQuad(int nv) const {
      map<int,int>::const_iterator i=
        dofs_per_quad.find(nv);
      CHECK_TEST(i!=dofs_per_quad.end());
      return i->second;
    }
    virtual int dofsPerQuad(const Shape *e) const {
      CHECK_SAME(e->numberDimensions(),2);
      return dofsPerQuad(e->numberVertices());
    }
    int dofsPerFace(int nv) const { 
      map<int,int>::const_iterator i=
        dofs_per_face.find(nv);
      CHECK_TEST(i!=dofs_per_face.end());
      return i->second;
    }
    int dofsPerFace(const Shape *e) const { 
      CHECK_SAME(e->numberDimensions(),SPACEDIM-1);
      return dofsPerFace(e->numberVertices());
    }
    int dofsPerFace(const Shape2 *e,int f) const { 
      CHECK_SAME(e->numberDimensions(),SPACEDIM);
      return dofsPerFace(e->numberVerticesOnFace(f));
    }
#endif
#if (SPACEDIM==3)
    virtual int dofsPerHex(int nv) const {
      map<int,int>::const_iterator i=
        dofs_per_hex.find(nv);
      CHECK_TEST(i!=dofs_per_hex.end());
      return i->second;
    }
    virtual int dofsPerHex(const Shape *e) const {
      CHECK_SAME(e->numberDimensions(),3);
      return dofsPerHex(e->numberVertices());
    }
#endif
    virtual int dofsPerCell(int nv) const {
      map<int,int>::const_iterator i=
        dofs_per_cell.find(nv);
      CHECK_TEST(i!=dofs_per_cell.end());
      return i->second;
    }
    virtual int dofsPerCell(const Shape *e) const {
      CHECK_SAME(e->numberDimensions(),SPACEDIM);
      return dofsPerCell(e->numberVertices());
    }
    virtual int maxDofsPerCell() const;
    const map<int,int>& dofsPerCellMap() const {
      return dofs_per_cell;
    }
    int nComponents() const { return components; }
//  called from FEValuesData::FEValuesData
    virtual int nNonzeroShapeComponents(int nv) const {
      return dofsPerCell(nv)/components;
    }
    inline int nNonzeroShapeComponents(const Shape *e) const {
      return dofsPerCell(e)/components;
    }
    int getDegree() const { return degree; }
//  public for testFE only
    ShapeFunction<SPACEDIM>* shapeFunctionFamily(int nv) const {
      map<int,ShapeFunction<SPACEDIM>*>::
        const_iterator ps_iterator=
        polynomial_space.find(nv);
      CHECK_TEST(ps_iterator!=polynomial_space.end());
      return ps_iterator->second;
    }
    ShapeFunction<SPACEDIM>* shapeFunctionFamily(
    const Shape *e) const {
      CHECK_SAME(e->numberDimensions(),SPACEDIM);
      return shapeFunctionFamily(e->numberVertices());
    }
//  compute shape(xi_q), d shape/d xi( xi_q ), d^2 shape/d xi^2(xi_q)
//    shape order is same as Shape::equallySpacedLatticePoints
//  called from FEValues::FEValues, FEFaceValues::FEFaceValues and
//    FESubfaceValues::FESubfaceValues
    virtual FiniteElementDataBase* getFEData(int nv,UPDATE_FLAGS flags,
    const Quadrature<SPACEDIM> *quadrature) const {
      return OPERATOR_NEW FiniteElementDataBase(flags,quadrature,
        shapeFunctionFamily(nv));
    }

    virtual FiniteElementDataBase* fixShapeFunctionsFor(const Shape*,
      FiniteElementDataBase*) const;
    virtual void sumShapeValues(const SHAPE_CLASS*,const Array<2,REAL>&,
      int,int,const Vector&,NumPtr<REAL>&) const;
    virtual void sumShapeValues(const SHAPE_CLASS*,const Array<2,REAL>&,
      int,int,const Vector&,NumPtr<Vector>&) const;
//  gradients[q][d], 0<=q<nqp, 0<=d<SPACEDIM
    virtual void sumShapeDerivatives(const SHAPE_CLASS*,
      const Array<2,Tensor<1,SPACEDIM> >&,int,int,const Vector&,
      NumPtr<Tensor<1,SPACEDIM> >&) const;
//  gradients[q][c][d], 0<=q<nqp, 0<=c<components, 0<=d<SPACEDIM
    virtual void sumShapeDerivatives(const SHAPE_CLASS*,
      const Array<2,Tensor<1,SPACEDIM> >&,int,int,const Vector&,
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
    virtual void sumShape2ndDerivatives(const SHAPE_CLASS*,
      const Array<2,Tensor<2,SPACEDIM> >&,int,int,const Vector&,
      NumPtr<Tensor<2,SPACEDIM> >&) const;
    virtual void sumShape2ndDerivatives(const SHAPE_CLASS*,
      const Array<2,Tensor<2,SPACEDIM> >&,int,int,const Vector&,
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > >&) const;
#if (SPACEDIM>1)
    virtual void setBoundaryValues(const SHAPE_CLASS*,
      const CODIMENSION1_SHAPE_CLASS *face_element,int,
      const NumPtr<int>&,const NumPtr<int>&,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<SPACEDIM> >&,
      const BoundaryCondition*,map<int,REAL>&,REAL=0.) const;
#endif
//  the following is public for testFE only
    const Matrix& prolongationMatrix(int nv,int c) {
      map<int,NumPtr<Matrix> >::iterator i=prolongation.find(nv);
      CHECK_TEST(i!=prolongation.end());
      Matrix &m=(i->second)[c];
      CHECK_TEST(m.size(1)!=0);
      return m;
    }
    virtual const Matrix& prolongationMatrix(const SHAPE_CLASS *e,int c,
    const LEVEL_CLASS*) {
      return prolongationMatrix(e->numberVertices(),c);
    }
#if (SPACEDIM>1)
    virtual void constraints(const LEVEL_CLASS*,const SHAPE_CLASS*,
      const CODIMENSION1_SHAPE_CLASS*,int,Matrix&)=0;
#endif
    void equallySpacedLatticePointDataBase(
      map<int,FiniteElementDataBase*>&,map<int,Quadrature<SPACEDIM>*>&)
      const;
    void setValues(const SHAPE_CLASS*,const NumPtr<int>&,
      FiniteElementDataBase*,const NumPtr<Point<SPACEDIM> >&,
      const FunctionSystem*,Vector&) const;
    virtual void printOn(ostream&) const;
};

//compare to DEAL FE_Q_Hierarchical
class FE_H : public FiniteElement {
  protected:    
  public:
    FE_H(int,int=1);
    ~FE_H() {;}
    virtual const Matrix& prolongationMatrix(const SHAPE_CLASS*,int,
      const LEVEL_CLASS*);
#if (SPACEDIM>1)
    virtual void constraints(const LEVEL_CLASS*,const SHAPE_CLASS*,
      const CODIMENSION1_SHAPE_CLASS*,int,Matrix&);
#endif
    void printOn(ostream&) const;
};

//compare to DEAL FE_Q
class FE_L : public FiniteElement {
  protected:    
  public:
    FE_L(int,int=1);
    ~FE_L() {}
#if (SPACEDIM>1)
    virtual void constraints(const LEVEL_CLASS*,const SHAPE_CLASS*,
      const CODIMENSION1_SHAPE_CLASS*,int,Matrix&);
#endif
    void printOn(ostream&) const;
};

#endif
