// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <HcurlDataOut.H>
#include <HcurlFEValues.H>
#include <Surface.H>
#include <Tracer.H>
#include <TypedPlotObj.H>
#include <Vector.H>
#include <VolGraphTool.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::DataEntry::printOn(ostream &os) const {
  os << "HcurlDataOut::DataEntry : names = ";
  for (int i=0;i<names.getNumber();i++) os << names[i] << " ";
  os << endl;
  os << "\n\tsingle_data = " << single_data << endl;
  if (single_data!=0) {
    for (int i=0;i<single_data->size();i++) {
      os << "\tsingle_data[" << i << "] = " << (*single_data)[i]
         << endl;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::clearDataVectors() {
  for (int i=0;i<dof_data.getNumber();i++) {
    delete dof_data[i]; dof_data[i]=0;
  }
  dof_data.cleanup();
  for (int i=0;i<multiplier_data.getNumber();i++) {
    delete multiplier_data[i]; multiplier_data[i]=0;
  }
  multiplier_data.cleanup();
  for (int i=0;i<cell_data.getNumber();i++) {
    delete cell_data[i]; cell_data[i]=0;
  }
  cell_data.cleanup();
  for (int i=0;i<patches.getNumber();i++) {
    delete patches[i]; patches[i]=0;
  }
  patches.cleanup();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::addDataVector(const Vector &vec,
const NumPtr<string> &new_names,DATA_VECTOR_TYPE type) {
  DataEntry *new_entry=OPERATOR_NEW DataEntry(&vec,new_names);
  switch (type) {
    case TYPE_DOF_DATA:
      CHECK_SAME(vec.size(),dofs->totalSolutionDofs());
      CHECK_SAME(new_names.getNumber(),dofs->getFe().nComponents());
      break;
    case TYPE_MULTIPLIER_DATA:
      CHECK_SAME(vec.size(),dofs->numberDofs()-dofs->totalSolutionDofs());
      CHECK_SAME(new_names.getNumber(),dofs->getFe().nComponents());
      break;
    case TYPE_CELL_DATA:
      CHECK_SAME(vec.size(),
                 dofs->getTessellation().numberActiveShapes());
      CHECK_SAME(new_names.getNumber(),1);
      break;
    default:
      if (vec.size()==dofs->getTessellation().numberActiveShapes()) {
        type=TYPE_CELL_DATA;
        CHECK_SAME(new_names.getNumber(),1);
      } else if (vec.size()==dofs->totalSolutionDofs()) {
        type=TYPE_DOF_DATA;
        CHECK_SAME(new_names.getNumber(),dofs->getFe().nComponents());
      } else {
        type=TYPE_MULTIPLIER_DATA;
        CHECK_SAME(vec.size(),
                   dofs->numberDofs()-dofs->totalSolutionDofs());
        CHECK_SAME(new_names.getNumber(),dofs->getFe().nComponents());
      }
  }
  if (type==TYPE_DOF_DATA) {
    NumPtr<DataEntry*> new_dof_data(dof_data.getNumber()+1);
    new_dof_data.copyFrom(dof_data);
    new_dof_data[dof_data.getNumber()]=new_entry;
    dof_data.copy(new_dof_data);
  } else if (type==TYPE_MULTIPLIER_DATA) {
    NumPtr<DataEntry*> new_multiplier_data(multiplier_data.getNumber()+1);
    new_multiplier_data.copyFrom(multiplier_data);
    new_multiplier_data[multiplier_data.getNumber()]=new_entry;
    multiplier_data.copy(new_multiplier_data);
  } else {
    NumPtr<DataEntry*> new_cell_data(cell_data.getNumber()+1);
    new_cell_data.copyFrom(cell_data);
    new_cell_data[cell_data.getNumber()]=new_entry;
    cell_data.copy(new_cell_data);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::addDataVector(const Vector &vec,const string &name,
DATA_VECTOR_TYPE type) {
  int n_components=dofs->getFe().nComponents();
  NumPtr<string> new_names;
  if ((n_components==1) ||
  (vec.size()==dofs->getTessellation().numberActiveShapes())) {
    new_names.allocate(1);
    new_names[0]=name;
  } else {
    new_names.allocate(n_components);
    for (int i=0; i<n_components; ++i) {
      ostringstream namebuf;
      namebuf << '_' << i;
      namebuf << ends;
      new_names[i]=name+namebuf.str();
    }
  }
  addDataVector(vec,new_names,type);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::replaceDataVector(const Vector &vec,
const string &name) {
  bool found=false;
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) {
        dof_data[i]->replaceData(&vec);
        found=true;
        break;
      }
    }
    if (found) break;
  }
  if (found) return;
  for (int i=0;i<multiplier_data.getNumber();i++) {
    const NumPtr<string> &multiplier_data_names=
      multiplier_data[i]->getNames();
    for (int j=0;j<multiplier_data_names.getNumber();j++) {
      if (name==multiplier_data_names[j]) {
        multiplier_data[i]->replaceData(&vec);
        found=true;
        break;
      }
    }
    if (found) break;
  }
  if (found) return;
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) {
        cell_data[i]->replaceData(&vec);
        found=true;
        break;
      }
    }
    if (found) break;
  }
  CHECK_TEST(found);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::clearInputDataReferences() {
  for (int i=0;i<dof_data.getNumber();i++) {
    dof_data[i]->clearPointers();
  }
  for (int i=0;i<multiplier_data.getNumber();i++) {
    multiplier_data[i]->clearPointers();
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    cell_data[i]->clearPointers();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::getDatasetNames(NumPtr<string> &names_out) const {
  int n=0;
  for (int i=0;i<dof_data.getNumber();i++) {
    n+=dof_data[i]->getNames().getNumber();
  }
  for (int i=0;i<multiplier_data.getNumber();i++) {
    n+=multiplier_data[i]->getNames().getNumber();
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    n+=cell_data[i]->getNames().getNumber();
  }
  names_out.allocate(n);
  n=0;
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_names=dof_data[i]->getNames();
    for (int j=0;j<dof_names.getNumber();j++,n++) {
      names_out[n]=dof_names[j];
    }
  }
  for (int i=0;i<multiplier_data.getNumber();i++) {
    const NumPtr<string> &dof_names=multiplier_data[i]->getNames();
    for (int j=0;j<dof_names.getNumber();j++,n++) {
      names_out[n]=dof_names[j];
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_names=cell_data[i]->getNames();
    for (int j=0;j<cell_names.getNumber();j++,n++) {
      names_out[n]=cell_names[j];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::buildPatches(int n_subdivisions) {
  HcurlPatch::setQuadratures(n_subdivisions);
  CHECK_POINTER(dofs);
  patches.cleanup();
  Tessellation &tess=dofs->getTessellation();
  int n_levels=tess.numberLevels();
  NumPtr<NumPtr<int> > cell_to_patch_index_map(n_levels);
  int l=0;
  for (LEVEL_CLASS *level=tess.coarsestLevel();level;
  level=level->finerLevel(),l++) {
    int max_index=0;
    for (bool doit=tess.firstActiveShape(level);doit;
    doit=tess.nextActiveShapeInLevel()) {
      max_index=max(max_index,level->presentIndex());
    }
    cell_to_patch_index_map[l].allocate(max_index+1);
    cell_to_patch_index_map[l].initialize(HcurlPatch::no_neighbor);
  }
  int n_patches=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),n_patches++) {
    LEVEL_CLASS *level=tess.presentLevel();
    cell_to_patch_index_map[level->numberCoarserLevels()]
      [level->presentIndex()]=n_patches;
  }
  int n_components=dofs->getFe().nComponents();
  int n_datasets=cell_data.getNumber()
    +(dof_data.getNumber()+multiplier_data.getNumber())*n_components;
  patches.allocate(n_patches);
  int p=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),p++) {
    const SHAPE_CLASS *e=tess.presentShape();
    patches[p]=OPERATOR_NEW HcurlPatch(e,tess.presentLevel(),tess,
      *mapping,n_datasets,p);
  }

  static const MappingL mapping_l;
  UPDATE_FLAGS flags= UPDATE_VALUES | UPDATE_JxW_VALUES |
    UPDATE_Q_POINTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
  HcurlFEValues fe_patch_values(mapping_l,dofs->getFe(),
    HcurlPatch::getQuadratureMap(),flags);

  int cell_number=0;
  p=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),p++) {
    const SHAPE_CLASS *shape=tess.presentShape();
    LEVEL_CLASS *present_level=tess.presentLevel();
    HcurlPatch *patch=patches[p];
    if (n_datasets>0) {
      fe_patch_values.mapShapes(tess,shape,present_level);
      int n_q_points=patch->numberQuadraturePoints();
      for (int dataset=0;dataset<dof_data.getNumber();dataset++) {
        if (n_components==1) {
          NumPtr<Tensor<1,SPACEDIM> > patch_values;
          dofs->getSolutionFunctionValues(fe_patch_values,shape,
            present_level,*dof_data[dataset]->singleData(),patch_values);
          patch->setData(dataset,patch_values);
        } else {
          NumPtr<NumPtr<Tensor<1,SPACEDIM> > > patch_values_system;
          dofs->getSolutionFunctionValues(fe_patch_values,shape,
            present_level,*dof_data[dataset]->singleData(),
            patch_values_system);
          patch->setData(dataset,n_components,patch_values_system);
        }
      }
      int multiplier_offset=dofs->totalSolutionDofs();
      for (int dataset=0;dataset<multiplier_data.getNumber();dataset++) {
        if (n_components==1) {
          NumPtr<HCURL_MULTIPLIER_DATA_TYPE> patch_multiplier_values;
          dofs->getMultiplierFunctionValues(fe_patch_values,shape,
            present_level,*multiplier_data[dataset]->singleData(),
            multiplier_offset,patch_multiplier_values);
          patch->setMultiplierData(dataset,patch_multiplier_values);
        } else {
          NumPtr<NumPtr<Tensor<1,SPACEDIM> > > patch_values_system;
          NumPtr<NumPtr<HCURL_MULTIPLIER_DATA_TYPE> >
            patch_multiplier_values_system;
          dofs->getMultiplierFunctionValues(fe_patch_values,shape,
            present_level,*multiplier_data[dataset]->singleData(),
            multiplier_offset,patch_multiplier_values_system);
          patch->setMultiplierData(dataset,n_components,
            patch_multiplier_values_system);
        }
      }
    }
    for (int f=0;f<shape->numberCodimension1Shapes();f++) {
      int neighbor_index=shape->codimension1ShapeIndex(f);
      LEVEL_CLASS *neighbor_level=present_level->neighborLevel(f);
      if (neighbor_index<0) continue;
      if (present_level!=neighbor_level) continue;

      int neighbor_level_number=
        neighbor_level->numberCoarserLevels();
      NumPtr<int> &cell_to_patch_index_map_this_level=
        cell_to_patch_index_map[neighbor_level_number];
      if (neighbor_index>=
      cell_to_patch_index_map_this_level.getNumber()) {
        continue;
      }
      if (cell_to_patch_index_map_this_level[neighbor_index]==
      HcurlPatch::no_neighbor) {
        continue;
      }

      patch->setNeighbor(f,neighbor_index);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::reBuildPatches() {
  CHECK_POINTER(dofs);
  Tessellation &tess=dofs->getTessellation();
  int n_components=dofs->getFe().nComponents();
  int n_datasets=cell_data.getNumber()
    +(dof_data.getNumber()+multiplier_data.getNumber())*n_components;

  static const MappingL mapping_l;
  UPDATE_FLAGS flags= UPDATE_VALUES | UPDATE_JxW_VALUES |
    UPDATE_Q_POINTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
  HcurlFEValues fe_patch_values(mapping_l,dofs->getFe(),
    HcurlPatch::getQuadratureMap(),flags);

  int p=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),p++) {
    const SHAPE_CLASS *shape=tess.presentShape();
    LEVEL_CLASS *present_level=tess.presentLevel();
    HcurlPatch *patch=patches[p];
    if (n_datasets>0) {
      fe_patch_values.mapShapes(tess,shape,present_level);
      int n_q_points=patch->numberQuadraturePoints();
      for (int dataset=0;dataset<dof_data.getNumber();dataset++) {
        if (n_components==1) {
          NumPtr<Tensor<1,SPACEDIM> > patch_values;
          dofs->getSolutionFunctionValues(fe_patch_values,shape,
            present_level,*dof_data[dataset]->singleData(),patch_values);
          patch->setData(dataset,patch_values);
        } else {
          NumPtr<NumPtr<Tensor<1,SPACEDIM> > > patch_values_system;
          dofs->getSolutionFunctionValues(fe_patch_values,shape,
            present_level,*dof_data[dataset]->singleData(),
            patch_values_system);
          patch->setData(dataset,n_components,patch_values_system);
        }
      }
      int multiplier_offset=dofs->totalSolutionDofs();
      for (int dataset=0;dataset<multiplier_data.getNumber();dataset++) {
        if (n_components==1) {
          NumPtr<HCURL_MULTIPLIER_DATA_TYPE> patch_values;
          dofs->getMultiplierFunctionValues(fe_patch_values,shape,
            present_level,*multiplier_data[dataset]->singleData(),
            multiplier_offset,patch_values);
          patch->setMultiplierData(dataset,patch_values);
        } else {
          NumPtr<NumPtr<HCURL_MULTIPLIER_DATA_TYPE> > patch_values_system;
          dofs->getMultiplierFunctionValues(fe_patch_values,shape,
            present_level,*multiplier_data[dataset]->singleData(),
            multiplier_offset,patch_values_system);
          patch->setMultiplierData(dataset,n_components,
            patch_values_system);
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

#include <AreaGraphTool.H>
#include <Palette.H>
#include <Vector3.H>
#include <VolGraphTool.H>
#include <XColormap.H>
#include <XYGraphTool.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::plot(const char *name,VirtualGraphTool *gt,
bool quit_button) const {
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) plot(name,i,gt,quit_button);
    }
  }
  for (int i=0;i<multiplier_data.getNumber();i++) {
    const NumPtr<string> &multiplier_data_names=
      multiplier_data[i]->getNames();
    for (int j=0;j<multiplier_data_names.getNumber();j++) {
      if (name==multiplier_data_names[j]) {
        plotMultipliers(name,i,gt,quit_button);
      }
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) plot(name,i,gt,quit_button);
    }
  }
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void HcurlDataOut::plot(const char *name,int data_set,
VirtualGraphTool *gt,bool quit_button) const {
  Point<1> low(HUGE_VAL);
  Point<1> high(-HUGE_VAL);
  Tensor<1,1> dlow(HUGE_VAL);
  Tensor<1,1> dhigh(-HUGE_VAL);
  REAL multiplier_dlow(HUGE_VAL);
  REAL multiplier_dhigh(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,dlow,dhigh);
  }
  XYGraphTool *xygt;
  xygt=dynamic_cast<XYGraphTool*>(gt);
  CHECK_POINTER(xygt);
  xygt->rescale(low[0],high[0],dlow[0],dhigh[0]);
  xygt->newPage();
  xygt->setbgColor("white");
  xygt->setfgColor("black");
  xygt->drawAxes();
  xygt->setfgColor("blue");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plot(xygt,data_set);
  }
  xygt->flush();
  if (quit_button) XYGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void HcurlDataOut::plotMultipliers(const char *name,int data_set,
VirtualGraphTool *gt,bool quit_button) const {
  Point<1> low(HUGE_VAL);
  Point<1> high(-HUGE_VAL);
  REAL multiplier_dlow(HUGE_VAL);
  REAL multiplier_dhigh(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getMultiplierDataBounds(data_set,multiplier_dlow,
      multiplier_dhigh);
  }
  XYGraphTool *xygt;
  xygt=dynamic_cast<XYGraphTool*>(gt);
  CHECK_POINTER(xygt);
  xygt->rescale(low[0],high[0],multiplier_dlow,multiplier_dhigh);
  xygt->newPage();
  xygt->setbgColor("white");
  xygt->setfgColor("black");
  xygt->drawAxes();
  xygt->setfgColor("blue");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plotMultipliers(xygt,data_set);
  }
  xygt->flush();
  if (quit_button) XYGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void HcurlDataOut::plot(const char *name,int data_set,
VirtualGraphTool *gt,bool quit_button) const {
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  Tensor<1,2> ymin; ymin.initialize(HUGE_VAL);
  Tensor<1,2> ymax; ymax.initialize(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
  }
  AreaGraphTool *agt;
  agt=dynamic_cast<AreaGraphTool*>(gt);
  CHECK_POINTER(agt);
  agt->rescale(low[0],high[0],low[1],high[1]);
  agt->newPage();
  agt->setbgColor("white");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plot(agt,data_set,ymin,ymax);
  }
  agt->setfgColor("black");
  agt->drawAxes();
  agt->flush();
  if (quit_button) AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void HcurlDataOut::plotMultipliers(const char *name,int data_set,
VirtualGraphTool *gt,bool quit_button) const {
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  HCURL_MULTIPLIER_DATA_TYPE multiplier_ymin(HUGE_VAL);
  HCURL_MULTIPLIER_DATA_TYPE multiplier_ymax(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getMultiplierDataBounds(data_set,multiplier_ymin,
      multiplier_ymax);
  }
  AreaGraphTool *agt;
  agt=dynamic_cast<AreaGraphTool*>(gt);
  CHECK_POINTER(agt);
  agt->rescale(low[0],high[0],low[1],high[1]);
  agt->newPage();
  agt->setbgColor("white");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plotMultipliers(agt,data_set,multiplier_ymin,
      multiplier_ymax);
  }
  agt->setfgColor("black");
  agt->drawAxes();
  agt->flush();
  if (quit_button) AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlDataOut::plot(const char *name,int data_set,
VirtualGraphTool *gt,bool quit_button) const {
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  Tensor<1,3> ymin; ymin.initialize(HUGE_VAL);
  Tensor<1,3> ymax; ymax.initialize(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
  }
  Vector3 vlow(low[0],low[1],low[2]);
  Vector3 vhigh(high[0],high[1],high[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;

  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(gt);
  CHECK_POINTER(vgt);
  vgt->rescale(vlow,vhigh);
  vgt->newPage();
  vgt->expose();
  for (bool do_best=FALSE;;do_best=!do_best) {
    VolGraphTool::DrawDelimiter dd(do_best,vgt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->drawVectors(vgt,data_set,vlow,vhigh,ymin,ymax);
    }
    if (do_best) break;
  }
  if (quit_button) vgt->eventLoop();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlDataOut::plotMultipliers(const char *name,int data_set,
VirtualGraphTool *gt,bool quit_button) const {
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  Tensor<1,3> multiplier_ymin; multiplier_ymin.initialize(HUGE_VAL);
  Tensor<1,3> multiplier_ymax; multiplier_ymax.initialize(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getMultiplierDataBounds(data_set,multiplier_ymin,
      multiplier_ymax);
  }
  Vector3 vlow(low[0],low[1],low[2]);
  Vector3 vhigh(high[0],high[1],high[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;

  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(gt);
  CHECK_POINTER(vgt);
  vgt->rescale(vlow,vhigh);
  vgt->newPage();
  vgt->expose();
  for (bool do_best=FALSE;;do_best=!do_best) {
    VolGraphTool::DrawDelimiter dd(do_best,vgt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->drawVectors(vgt,data_set,vlow,vhigh,
        multiplier_ymin,multiplier_ymax);
    }
    if (do_best) break;
  }
  if (quit_button) vgt->eventLoop();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void HcurlDataOut::drawGrid(VirtualGraphTool *gt,bool quit_button) const
{
  AreaGraphTool *agt=dynamic_cast<AreaGraphTool*>(gt);
  CHECK_POINTER(agt);
  CHECK_TEST(patches.getNumber()>0);
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
  }
  agt->rescale(low[0],high[0],low[1],high[1]);
  agt->setbgColor("white");
  agt->newPage();
  agt->flush();
  agt->setfgColor("black");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->draw(*agt);
  }
  agt->drawAxes();
  agt->flush();
  if (quit_button) {
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlDataOut::drawGrid(VirtualGraphTool *gt,bool quit_button) const
{
  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(gt);
  CHECK_POINTER(vgt);
  CHECK_TEST(patches.getNumber()>0);
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
  }
  Vector3 vlow(low[0],low[1],low[2]);
  Vector3 vhigh(high[0],high[1],high[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;
  vgt->rescale(vlow,vhigh);
  vgt->newPage();
  vgt->expose();

  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=vgt->getColorArray();
  VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=color_array->getColor(0.);
  vgt->setfgColor("black");
  { BOOLEAN do_best=FALSE;
    VolGraphTool::DrawDelimiter dd(do_best,vgt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->draw(*vgt);
    }
  }
  { BOOLEAN do_best=TRUE;
    VolGraphTool::DrawDelimiter dd(do_best,vgt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->draw(*vgt);
    }
  }
  vgt->expose();
  if (quit_button) vgt->eventLoop();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDataOut::printOn(ostream &os) const {
  os << "HcurlDataOut : dofs = " << dofs
     << "\n\tmapping = " << mapping;
  for (int i=0;i<dof_data.getNumber();i++) {
    os << "\n\tdof_data[" << i << "] :" << endl;
    dof_data[i]->printOn(os);
  }
  for (int i=0;i<multiplier_data.getNumber();i++) {
    os << "\n\tmultiplier_data[" << i << "] :" << endl;
    multiplier_data[i]->printOn(os);
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    os << "\n\tcell_data[" << i << "] :" << endl;
    cell_data[i]->printOn(os);
  }
  for (int i=0;i<patches.getNumber();i++) {
    os << "\n\tpatches[" << i << "] :" << endl;
    patches[i]->printOn(os);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlPatchDataSet::plot(BOOLEAN,const AxisClipPlane *cp,
const HCURL_MULTIPLIER_DATA_TYPE &ymin,
const HCURL_MULTIPLIER_DATA_TYPE &ymax) {
  REAL location=graph_tool->getLocation(cp);
  int dir=cp->getDirection();
  HAND h=cp->getHand();
  REAL fudge=graph_tool->getLength()[dir]*0.015625;
  if (h==LEFT_HAND) location += fudge;
  else location -= fudge;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plotClipPlane(d,*graph_tool,ymin,ymax,location,dir,h,
      dofs,mapping,single_data);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//
#if (SPACEDIM==3)
void HcurlPatchDataSet::plot(BOOLEAN do_best,const AxisClipPlane *cp) {
  int data_set=0;
  HCURL_MULTIPLIER_DATA_TYPE ymin; ymin.initialize(HUGE_VAL);
  HCURL_MULTIPLIER_DATA_TYPE ymax; ymax.initialize(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMultiplierDataBounds(data_set,ymin,ymax);
  }
  plot(do_best,cp,ymin,ymax);
}
#endif
//
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlPatchDataSet::plot(BOOLEAN,REAL) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlPatchDataSet::plot(BOOLEAN) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
REAL HcurlPatchDataSet::getVarMin() const {
  OBSOLETE("not programmed"); 
  return HUGE_VAL;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
REAL HcurlPatchDataSet::getVarMax() const {
  OBSOLETE("not programmed"); 
  return HUGE_VAL;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlPatchSurfaceDataSet::plot(BOOLEAN,const AxisClipPlane *cp) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlPatchSurfaceDataSet::plot(BOOLEAN do_best,REAL val) {
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->
      plotSurface(d,*graph_tool,data_set,val,var_min,var_max,do_best);
  }
}
#endif

#include "NumPtr.C"
INSTANTIATE_NUMPTR(HcurlDataOut::DataEntry*);
INSTANTIATE_NUMPTR(HcurlPatch*);

#if (SPACEDIM==3)
template class TypedPlotObj<HcurlPatchDataSet>;
template class Surface<HcurlPatchSurfaceDataSet>;
#endif
