#ifndef HCURL_DATA_OUT_H
#define HCURL_DATA_OUT_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <HcurlDofHandler.H>
#include <HcurlPatch.H>
#include <Types.H>

# if (SPACEDIM==1)
class XYGraphTool;
# else
class VolGraphTool;
#   if (SPACEDIM==2)
class AreaGraphTool;
class Vector3;
#   else
class AxisClipPlane;
#   endif
# endif

class HcurlDataOut {
  public :
    enum DATA_VECTOR_TYPE {
      TYPE_DOF_DATA,TYPE_MULTIPLIER_DATA,TYPE_CELL_DATA,TYPE_AUTOMATIC
    };
  protected :
    class DataEntry {
      private:
        int size;
        const Vector *single_data;
        NumPtr<string> names;
      public:
        DataEntry(const Vector *data=0,
        const NumPtr<string> &n=NumPtr<string>()) : size(0),
        single_data(data),names(n.getNumber()) {
          if (single_data) size=single_data->size();
          names.copyFrom(n); 
        }
        void replaceData(const Vector *new_data) {
          CHECK_SAME(size,new_data->size());
          single_data=new_data;
        }
        void clearPointers() { single_data=0; }
        const NumPtr<string>& getNames() const { return names; }
        const Vector* singleData() const { return single_data; }
        void printOn(ostream&) const;
    };
    const HcurlDofHandler *dofs;
    const Mapping *mapping;
    NumPtr<DataEntry*> dof_data;
    NumPtr<DataEntry*> multiplier_data;
    NumPtr<DataEntry*> cell_data;
    NumPtr<HcurlPatch*> patches;

    virtual void getDatasetNames(NumPtr<string>&) const;
    void clearDataVectors();
  private:
    void plot(const char*,int,VirtualGraphTool*,bool) const;
    void plotMultipliers(const char*,int,VirtualGraphTool*,bool)
      const;
# if (SPACEDIM==2)
    void multiplierContourPlot(const char*,int,AreaGraphTool*,bool) const;
# endif
# if (SPACEDIM>=2)
    void plotMultipliersSurface(const char*,int,VolGraphTool*,bool) const;
# endif
  public:
    HcurlDataOut(const HcurlDofHandler &d,const Mapping &m) : dofs(&d),
      mapping(&m) {;}
    virtual ~HcurlDataOut() { clearDataVectors(); }
    void addDataVector(const Vector&,const NumPtr<string>&,
      DATA_VECTOR_TYPE=TYPE_AUTOMATIC);
    void addDataVector(const Vector&,const string &name,
      DATA_VECTOR_TYPE=TYPE_AUTOMATIC);
    void replaceDataVector(const Vector &vec,const string &name);
    void clearInputDataReferences();
    virtual void buildPatches(int n_subdivisions=1);
    virtual void reBuildPatches();
    void plot(const char*,VirtualGraphTool*,bool=true) const;
# if (SPACEDIM>=2)
    void drawGrid(VirtualGraphTool*,bool=true) const;
# endif
    virtual void printOn(ostream&) const;
};

# if (SPACEDIM==3)
class HcurlPatchDataSet {
  private:
    int d;
    VolGraphTool *graph_tool;
    const Vector3 &low;
    const Vector3 &high;
    const NumPtr<HcurlPatch*> &patches;
    const HcurlDofHandler *dofs;
    const Mapping *mapping;
    const Vector *single_data;

    void plot(BOOLEAN,const AxisClipPlane*,
      const HCURL_MULTIPLIER_DATA_TYPE&,const HCURL_MULTIPLIER_DATA_TYPE&);
  public:
    HcurlPatchDataSet(int dd,VolGraphTool *gt,const Vector3 &l,
    const Vector3 &h,const NumPtr<HcurlPatch*> &p,const HcurlDofHandler *dh,
    const Mapping *m,const Vector *sd) : d(dd),graph_tool(gt),low(l),
      high(h),patches(p),dofs(dh),mapping(m),single_data(sd) { }
    virtual ~HcurlPatchDataSet() { }
    VolGraphTool* getGraphTool() const { return graph_tool; }
    BOOLEAN drawSurface() const { return FALSE; }
    void plot(BOOLEAN,const AxisClipPlane*);
    void plot(BOOLEAN,REAL);
    void plot(BOOLEAN);
    REAL getVarMin() const;
    REAL getVarMax() const;
    virtual void printOn(ostream&) const {}
};
# endif

# if (SPACEDIM==3)
class HcurlPatchSurfaceDataSet {
  private:
    int d;
    VolGraphTool *graph_tool;
    const Vector3 &low;
    const Vector3 &high;
    const HcurlDofHandler *dofs;
    const Mapping *mapping;
    const NumPtr<HcurlPatch*> &patches;
    REAL var_min,var_max;
    int data_set;
  public:
    HcurlPatchSurfaceDataSet(int dd,VolGraphTool *gt,const Vector3 &l,
    const Vector3 &h,const HcurlDofHandler *dh,const Mapping *m,
    const NumPtr<HcurlPatch*> &p,REAL vn,REAL vx,int ds) : d(dd), 
      graph_tool(gt),low(l),high(h),dofs(dh),mapping(m),patches(p),
      var_min(vn),var_max(vx),data_set(ds) { }
    virtual ~HcurlPatchSurfaceDataSet() { }
    VolGraphTool* getGraphTool() const { return graph_tool; }
    REAL getVarMin() const { return var_min; }
    REAL getVarMax() const { return var_max; }
    BOOLEAN drawSurface() const { return TRUE; }
    void plot(BOOLEAN,const AxisClipPlane*);
    void plot(BOOLEAN,REAL);
    virtual void printOn(ostream&) const {}
};
# endif

#endif
