// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <HcurlDofHandler.H>

#include <algorithm>
#include <Boundary.H>
#include <FEValues.H>

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::clearSpace() {  
  for (int i=0;i<levels.getNumber();i++) {
    if (levels[i]) delete levels[i]; levels[i]=0;
  }
  levels.cleanup();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::reserveSpace() {
  int number_levels=tess->numberLevels();
  CHECK_TEST(number_levels>0);
  clearSpace();
  levels.allocate(number_levels);
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
#if (SPACEDIM==3)
    int cd2n=0;
    Level1 *cd2l=static_cast<Level1*>(level);
    int dpl=selected_fe->dofsPerLine();
    for (bool doit=cd2l->first();doit;doit=cd2l->next()) {
      Shape1 *e=cd2l->getShape();
      e->setDofIndex(cd2n);
      cd2n+=dpl;
    }
#endif
    int cd1n=0;
#if (SPACEDIM==1)
    cd1n=tess->getVertices().getNumber()*selected_fe->dofsPerFace(1);
#else
    CODIMENSION1_LEVEL_CLASS *cd1l=
      static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
    for (bool doit=cd1l->first();doit;doit=cd1l->next()) {
      CODIMENSION1_SHAPE_CLASS *e=cd1l->getShape();
      e->setDofIndex(cd1n);
      cd1n+=selected_fe->dofsPerFace(e);
    }
#endif
    int n=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      SHAPE_CLASS *e=level->getShape();
      e->setDofIndex(n);
      n+=selected_fe->dofsPerShape(e);
    }
    int mn=0;
    map<int,int> element_data_index_to_multiplier_index_map;
    for (bool doit=level->first();doit;doit=level->next()) {
      SHAPE_CLASS *e=level->getShape();
      element_data_index_to_multiplier_index_map[e->elementDataIndex()]=
        mn;
      mn+=selected_fe->multiplierDofsPerShape(e);
    }
#if (SPACEDIM<=2)
    levels[i]=OPERATOR_NEW HcurlDofLevel(cd1n,n,mn,
      element_data_index_to_multiplier_index_map);
#else
    levels[i]=OPERATOR_NEW HcurlDofLevel(cd2n,cd1n,n,mn,
      element_data_index_to_multiplier_index_map);
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::distributeDofs() {
  CHECK_TEST(tess->numberLevels()>0);
  reserveSpace();
  NumPtr<bool> user_flags;
  tess->saveUserFlags(user_flags);
  const_cast<Tessellation*>(tess)->clearUserFlags();
  int next_free_dof=0;   
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    HcurlDofLevel *dof_level=getLevel(level);
#if (SPACEDIM==3)
    int dpl=selected_fe->dofsPerLine();
    if (dpl>0) {
      for (int l=0;l<element->numberCodimension2Shapes();l++) {
        const Shape1 *line_element=
          level->Level1::getShape(element->codimension2ShapeIndex(l));
        dof_level->setCodimension2Dofs(line_element->dofIndex(),dpl,
          next_free_dof);
      }
    }
#endif
#if (SPACEDIM==1)
    int dpf=selected_fe->dofsPerFace(1);
    if (dpf>0) {
      for (int v=0;v<element->numberVertices();v++) {
        dof_level->setCodimension1Dofs(element->vertexIndex(v),dpf,
          next_free_dof);
      }
    }
#else
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      const CODIMENSION1_SHAPE_CLASS *face_element=
        level->CODIMENSION1_LEVEL_CLASS::getShape(
          element->codimension1ShapeIndex(f));
      int dpf=selected_fe->dofsPerFace(face_element);
      if (dpf>0) {
        dof_level->setCodimension1Dofs(face_element->dofIndex(),dpf,
          next_free_dof);
      }
    }
#endif
    int dps=selected_fe->dofsPerShape(element);
    if (dps>0) dof_level->setDofs(element->dofIndex(),dps,next_free_dof);
    level->setUserFlag(element);
  }
  total_solution_dofs=next_free_dof;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    HcurlDofLevel *dof_level=getLevel(level);
    int mdps=selected_fe->multiplierDofsPerShape(element);
    if (mdps>0) {
      dof_level->setMultiplierDofs(element->elementDataIndex(),mdps,
        next_free_dof);
    }
  }
  used_dofs=next_free_dof;
  const_cast<Tessellation*>(tess)->loadUserFlags(user_flags);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getDofIndices(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,NumPtr<int> &dof_indices) const {
  const CODIMENSION1_LEVEL_CLASS *codimension1_level=
    static_cast<const CODIMENSION1_LEVEL_CLASS*>(level);
  int dps=selected_fe->dofsPerShape(element);
  int mdps=selected_fe->multiplierDofsPerShape(element);
  int sz=dps+mdps;
#if (SPACEDIM==3)
  int nl=element->numberCodimension2Shapes();
  for (int l=0;l<nl;l++) {
    const Shape1 *line_element=
      level->Level1::getShape(element->codimension2ShapeIndex(l));
    sz+=selected_fe->dofsPerLine();
  }
#endif
#if (SPACEDIM==1)
  int dpf=selected_fe->dofsPerFace(1);
  sz+=dpf*element->numberVertices();
#else
  int nf=element->numberCodimension1Shapes();
  for (int f=0;f<nf;f++) {
    const CODIMENSION1_SHAPE_CLASS *face_element=
      level->CODIMENSION1_LEVEL_CLASS::getShape(
        element->codimension1ShapeIndex(f));
    sz+=selected_fe->dofsPerFace(face_element);
  }
#endif
  dof_indices.allocate(sz);
#ifdef INDEF
  dof_indices.initialize(invalid_dof_index);
#endif
  HcurlDofLevel *dof_level=getLevel(level);
  int offset=0;
#if (SPACEDIM==3)
  for (int l=0;l<nl;l++) {
    const Shape1 *line_element=
      level->Level1::getShape(element->codimension2ShapeIndex(l));
    int dpl=selected_fe->dofsPerLine();
    NumPtr<int> line_dofs(dof_indices,offset,dpl);
    dof_level->getCodimension2Dofs(line_element->dofIndex(),line_dofs);
    offset+=dpl;
  }
#endif
#if (SPACEDIM==1)
  for (int v=0;v<element->numberVertices();v++) {
    NumPtr<int> face_dofs(dof_indices,offset,dpf);
    dof_level->getCodimension1Dofs(element->vertexIndex(v),face_dofs);
    offset+=dpf;
  }
#else
  for (int f=0;f<nf;f++) {
    const CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape(element->codimension1ShapeIndex(f));
    int dpf=selected_fe->dofsPerFace(face_element);
    NumPtr<int> face_dofs(dof_indices,offset,dpf);
    dof_level->getCodimension1Dofs(face_element->dofIndex(),face_dofs);
    offset+=dpf;
  }
#endif
  NumPtr<int> element_dofs(dof_indices,offset,dps);
  dof_level->getDofs(element->dofIndex(),element_dofs);

  offset+=dps;
  NumPtr<int> multiplier_dofs(dof_indices,offset,mdps);
  dof_level->getMultiplierDofs(element->elementDataIndex(),
    multiplier_dofs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getSolutionDofIndices(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,NumPtr<int> &dof_indices) const {
  const CODIMENSION1_LEVEL_CLASS *codimension1_level=
    static_cast<const CODIMENSION1_LEVEL_CLASS*>(level);
  int dps=selected_fe->dofsPerShape(element);
  int sz=dps;
#if (SPACEDIM==3)
  int nl=element->numberCodimension2Shapes();
  for (int l=0;l<nl;l++) {
    const Shape1 *line_element=
      level->Level1::getShape(element->codimension2ShapeIndex(l));
    sz+=selected_fe->dofsPerLine();
  }
#endif
#if (SPACEDIM==1)
  int dpf=selected_fe->dofsPerFace(1);
  sz+=dpf*element->numberVertices();
#else
  int nf=element->numberCodimension1Shapes();
  for (int f=0;f<nf;f++) {
    const CODIMENSION1_SHAPE_CLASS *face_element=
      level->CODIMENSION1_LEVEL_CLASS::getShape(
        element->codimension1ShapeIndex(f));
    sz+=selected_fe->dofsPerFace(face_element);
  }
#endif
  dof_indices.allocate(sz);
#ifdef INDEF
  dof_indices.initialize(invalid_dof_index);
#endif
  HcurlDofLevel *dof_level=getLevel(level);
  int offset=0;
#if (SPACEDIM==3)
  for (int l=0;l<nl;l++) {
    const Shape1 *line_element=
      level->Level1::getShape(element->codimension2ShapeIndex(l));
    int dpl=selected_fe->dofsPerLine();
    NumPtr<int> line_dofs(dof_indices,offset,dpl);
    dof_level->getCodimension2Dofs(line_element->dofIndex(),line_dofs);
    offset+=dpl;
  }
#endif
#if (SPACEDIM==1)
  for (int v=0;v<element->numberVertices();v++) {
    NumPtr<int> face_dofs(dof_indices,offset,dpf);
    dof_level->getCodimension1Dofs(element->vertexIndex(v),face_dofs);
    offset+=dpf;
  }
#else
  for (int f=0;f<nf;f++) {
    const CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape(element->codimension1ShapeIndex(f));
    int dpf=selected_fe->dofsPerFace(face_element);
    NumPtr<int> face_dofs(dof_indices,offset,dpf);
    dof_level->getCodimension1Dofs(face_element->dofIndex(),face_dofs);
    offset+=dpf;
  }
#endif
  NumPtr<int> element_dofs(dof_indices,offset,dps);
  dof_level->getDofs(element->dofIndex(),element_dofs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getMultiplierDofIndices(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,NumPtr<int> &dof_indices) const {
  dof_indices.allocate(selected_fe->multiplierDofsPerShape(element));
#ifdef INDEF
  dof_indices.initialize(invalid_dof_index);
#endif
  getLevel(level)->getMultiplierDofs(element->elementDataIndex(),
    dof_indices);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlDofHandler::getFaceDofIndices(
const CODIMENSION1_SHAPE_CLASS *face_element,
const CODIMENSION1_LEVEL_CLASS *level,NumPtr<int> &face_dof_indices)
const {
  int dpf=selected_fe->dofsPerFace(face_element);
  int sz=dpf;
  face_dof_indices.allocate(sz);
# ifdef INDEF
  face_dof_indices.initialize(invalid_dof_index);
# endif
  HcurlDofLevel *dof_level=getLevel(level);
  NumPtr<int> fdi(face_dof_indices,0,dpf);
  dof_level->getCodimension1Dofs(face_element->dofIndex(),fdi);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getFaceDofIndices(const SHAPE_CLASS *element,
int f,const LEVEL_CLASS *level,NumPtr<int> &face_dof_indices) const {
#if (SPACEDIM==1)
  face_dof_indices.allocate(selected_fe->dofsPerFace(1));
# ifdef INDEF
  face_dof_indices.initialize(invalid_dof_index);
# endif
  HcurlDofLevel *dof_level=getLevel(level);
  dof_level->
    getCodimension1Dofs(element->vertexIndex(f),face_dof_indices);
#else
  int face_index=element->codimension1ShapeIndex(f);
  const CODIMENSION1_SHAPE_CLASS *face_element=level->
    CODIMENSION1_LEVEL_CLASS::getShape(face_index);
  getFaceDofIndices(face_element,
    static_cast<const CODIMENSION1_LEVEL_CLASS*>(level),
    face_dof_indices);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlDofHandler::getLineDofIndices(const Shape1 *line_element,
const Level1 *level,NumPtr<int> &line_dof_indices) const {
  int sz=selected_fe->dofsPerLine();
  line_dof_indices.allocate(sz);
# ifdef INDEF
  line_dof_indices.initialize(invalid_dof_index);
# endif
  HcurlDofLevel *dof_level=getLevel(level);
  dof_level->
    getCodimension2Dofs(line_element->dofIndex(),line_dof_indices);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlDofHandler::dofsOnChildren(
const CODIMENSION1_SHAPE_CLASS *face_element,
const CODIMENSION1_LEVEL_CLASS *face_level,
NumPtr<int> &dofs_on_children) const {
  CHECK_TEST(face_level->hasChildren(face_element));
# if (SPACEDIM==2)
  int ndoc=childrenPerCodimension1Shape()
          *selected_fe->dofsPerFace(face_element);
# else
  int dpl=selected_fe->dofsPerLine();
  int ndoc=(face_element->numberCodimension1Shapes()
           *childrenPerCodimension2Shape()
           +face_element->numberInteriorChildCodimension1Shapes())*dpl
           +childrenPerCodimension1Shape()
           *selected_fe->dofsPerFace(face_element);
# endif
  dofs_on_children.allocate(ndoc);
  NumPtr<int> doc(dofs_on_children);

  const CODIMENSION1_LEVEL_CLASS *face_child_level=
    face_level->finerLevel();
  HcurlDofLevel *dof_level=getLevel(face_child_level);
  int face_index=face_element->elementDataIndex();
# if (SPACEDIM==3)
  const Level1 *line_level=static_cast<const Level1*>(face_level);
  const Level1 *face_line_child_level=line_level->finerLevel();
  for (int i=0;i<face_element->numberInteriorChildCodimension1Shapes();
  i++) {
    int c0,c1;
    face_element->
      interiorChildCodimension1ShapeNeighbors(i,c0,c1);
    int face_child_index0=face_level->childIndex(face_index,c0);
    int face_child_index1=face_level->childIndex(face_index,c1);
    const Shape2 *face_child_element0=
      face_child_level->getShape(face_child_index0);
    const Shape2 *face_child_element1=
      face_child_level->getShape(face_child_index1);
    int face_child_line_index=face_child_element0->
      sharedFaceElementIndex(face_child_element1);
    const Shape1 *face_child_line_element=
      face_line_child_level->getShape(face_child_line_index);
    NumPtr<int> face_line_child_dof_indices(dpl);
    dof_level->getCodimension2Dofs(face_child_line_element->dofIndex(),
      face_line_child_dof_indices);
    doc.copyFrom(face_line_child_dof_indices);
    doc+=face_line_child_dof_indices.getNumber();
  }
# endif
  int dpf=selected_fe->dofsPerFace(face_element);
  for (int child=0;child<childrenPerCodimension1Shape();child++) {
    int face_child_index=face_level->childIndex(face_index,child);
    const CODIMENSION1_SHAPE_CLASS *face_child=
      face_child_level->getShape(face_child_index);
    NumPtr<int> face_child_dof_indices(dpf);
    dof_level->getCodimension1Dofs(face_child->dofIndex(),
      face_child_dof_indices);
    doc.copyFrom(face_child_dof_indices);
    doc+=face_child_dof_indices.getNumber();
  }
  CHECK_SAME(0,doc.getNumber());
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getDofValues(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,const Vector &values,Vector &local_values) 
const {
  NumPtr<int> dof_indices;
  getDofIndices(element,level,dof_indices);
  local_values.reinit(dof_indices.getNumber(),true);
  for (int i=0;i<dof_indices.getNumber();i++) {
    local_values[i]=values[dof_indices[i]];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getSolutionDofValues(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,const Vector &values,Vector &local_values) 
const {
  NumPtr<int> dof_indices;
  getSolutionDofIndices(element,level,dof_indices);
  local_values.reinit(dof_indices.getNumber(),true);
  for (int i=0;i<dof_indices.getNumber();i++) {
    local_values[i]=values[dof_indices[i]];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getMultiplierDofValues(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,const Vector &values,int offset,
Vector &local_values) const {
  NumPtr<int> dof_indices;
  getMultiplierDofIndices(element,level,dof_indices);
  local_values.reinit(dof_indices.getNumber(),true);
  for (int i=0;i<dof_indices.getNumber();i++) {
    local_values[i]=values[dof_indices[i]-offset];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getInterpolatedDofValues(
const SHAPE_CLASS *element,const LEVEL_CLASS *level,
const Vector &values,Vector &interpolated_values) const {
  CHECK_SAME(values.size(),numberDofs());
  int dofs_per_cell=selected_fe->dofsPerCell(element);
  if (!level->hasChildren(element)) {
    getDofValues(element,level,values,interpolated_values);
  } else {
    OBSOLETE("not programmed");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getInterpolatedSolutionDofValues(
const SHAPE_CLASS *element,const LEVEL_CLASS *level,
const Vector &values,Vector &interpolated_values) const {
  CHECK_SAME(values.size(),totalSolutionDofs());
  int dofs_per_cell=selected_fe->dofsPerCell(element);
  if (!level->hasChildren(element)) {
    getSolutionDofValues(element,level,values,interpolated_values);
  } else {
    OBSOLETE("not programmed");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::getInterpolatedMultiplierDofValues(
const SHAPE_CLASS *element,const LEVEL_CLASS *level,
const Vector &values,int offset,Vector &interpolated_values) const {
  CHECK_SAME(values.size(),numberDofs()-offset);
  int dofs_per_cell=selected_fe->dofsPerCell(element);
  if (!level->hasChildren(element)) {
    getMultiplierDofValues(element,level,values,offset,
      interpolated_values);
  } else {
    OBSOLETE("not programmed");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::makeSparsityPattern(SparsityPatternBase &sp) const {
  CHECK_SAME(sp.nRows(),used_dofs);
  CHECK_SAME(sp.nCols(),sp.nRows());
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    NumPtr<int> dofs_on_this_cell;
    getDofIndices(element,present_level,dofs_on_this_cell);
    int dofs_per_cell=dofs_on_this_cell.getNumber();
    for (int i=0;i<dofs_per_cell;i++) {
      for (int j=0;j<dofs_per_cell;j++) {
        sp.add(dofs_on_this_cell[i],dofs_on_this_cell[j]);
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::makeTransientSparsityPatterns(
SparsityPatternBase &mass_sp,SparsityPatternBase &stiffness_sp) const {
  CHECK_SAME(mass_sp.nRows(),
    D_CHOOSE(used_dofs,used_dofs+total_solution_dofs,3*used_dofs));
  CHECK_SAME(mass_sp.nCols(),
    D_CHOOSE(used_dofs,used_dofs+total_solution_dofs,3*used_dofs));
  CHECK_SAME(stiffness_sp.nRows(),
    D_CHOOSE(used_dofs,used_dofs+total_solution_dofs,3*used_dofs));
  CHECK_SAME(stiffness_sp.nCols(),
    D_CHOOSE(used_dofs,used_dofs+total_solution_dofs,3*used_dofs));
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);
    int multiplier_dofs_per_cell=
      selected_fe->multiplierDofsPerShape(element);
    NumPtr<int> dofs_on_this_cell;
    getDofIndices(element,present_level,dofs_on_this_cell);
#if (SPACEDIM==1)
    for (int i=0;i<dofs_per_cell;i++) {
      int soln_dofi=dofs_on_this_cell[i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=dofs_on_this_cell[j];
        mass_sp.add(soln_dofi,soln_dofj);
        stiffness_sp.add(soln_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int multiplier_dofi=dofs_on_this_cell[dofs_per_cell+i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=dofs_on_this_cell[dofs_per_cell+j];
        mass_sp.add(multiplier_dofi,multiplier_dofj);
        stiffness_sp.add(multiplier_dofi,multiplier_dofj);
      }
    }
#elif (SPACEDIM==2)
    for (int i=0;i<dofs_per_cell;i++) {
      int soln_dofi=dofs_on_this_cell[i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=dofs_on_this_cell[j];
        mass_sp.add(soln_dofi,soln_dofj);
        stiffness_sp.add(soln_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
    }
    for (int i=0;i<dofs_per_cell;i++) {
      int soln_dofi=total_solution_dofs+dofs_on_this_cell[i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=total_solution_dofs+dofs_on_this_cell[j];
        mass_sp.add(soln_dofi,soln_dofj);
        stiffness_sp.add(soln_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int multiplier_dofi=total_solution_dofs
                         +dofs_on_this_cell[dofs_per_cell+i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=total_solution_dofs+dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        mass_sp.add(multiplier_dofi,multiplier_dofj);
        stiffness_sp.add(multiplier_dofi,multiplier_dofj);
      }
    }
#else
    for (int i=0;i<dofs_per_cell;i++) {
      int soln_dofi=dofs_on_this_cell[i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=dofs_on_this_cell[j];
        mass_sp.add(soln_dofi,soln_dofj);
        stiffness_sp.add(soln_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=used_dofs+total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=2*used_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
    }
    for (int i=0;i<dofs_per_cell;i++) {
      int soln_dofi=total_solution_dofs+dofs_on_this_cell[i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=total_solution_dofs+dofs_on_this_cell[j];
        mass_sp.add(soln_dofi,soln_dofj);
        stiffness_sp.add(soln_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=2*total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=2*used_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
    }
    for (int i=0;i<dofs_per_cell;i++) {
      int soln_dofi=2*total_solution_dofs+dofs_on_this_cell[i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=2*total_solution_dofs+dofs_on_this_cell[j];
        mass_sp.add(soln_dofi,soln_dofj);
        stiffness_sp.add(soln_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=2*total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=used_dofs+total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        stiffness_sp.add(soln_dofi,multiplier_dofj);
      }
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int multiplier_dofi=2*total_solution_dofs
                         +dofs_on_this_cell[dofs_per_cell+i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=total_solution_dofs+dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=2*total_solution_dofs+dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=2*total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        mass_sp.add(multiplier_dofi,multiplier_dofj);
        stiffness_sp.add(multiplier_dofi,multiplier_dofj);
      }
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int multiplier_dofi=used_dofs+total_solution_dofs
                         +dofs_on_this_cell[dofs_per_cell+i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=2*total_solution_dofs+dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=used_dofs+total_solution_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        mass_sp.add(multiplier_dofi,multiplier_dofj);
        stiffness_sp.add(multiplier_dofi,multiplier_dofj);
      }
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int multiplier_dofi=2*used_dofs
                         +dofs_on_this_cell[dofs_per_cell+i];
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<dofs_per_cell;j++) {
        int soln_dofj=total_solution_dofs+dofs_on_this_cell[j];
        stiffness_sp.add(multiplier_dofi,soln_dofj);
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int multiplier_dofj=2*used_dofs
                           +dofs_on_this_cell[dofs_per_cell+j];
        mass_sp.add(multiplier_dofi,multiplier_dofj);
        stiffness_sp.add(multiplier_dofi,multiplier_dofj);
      }
    }
    OBSOLETE("OK?");
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
int HcurlDofHandler::maxCouplingsBetweenDofs() const {
  return min(3*selected_fe->dofsPerFace(1)
            +2*selected_fe->dofsPerShape(2)
            +2*selected_fe->multiplierDofsPerShape(2),used_dofs);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
int HcurlDofHandler::maxCouplingsBetweenDofs() const {
//for each line, find set of elements containing it
  map<const Shape1*,set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> > >
    line_active_element_set_map;
  map<const Shape1*,set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> > >
    line_element_set_map;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
# if (SPACEDIM==2)
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int line_element_index=element->codimension1ShapeIndex(f);
      const Shape1 *line_element=
        level->Level1::getShape(line_element_index);
      line_active_element_set_map[line_element].
        insert(make_pair(element,level));
    }
# else
    for (int l=0;l<element->numberCodimension2Shapes();l++) {
      int line_element_index=element->codimension2ShapeIndex(l);
      const Shape1 *line_element=
        level->Level1::getShape(line_element_index);
      line_active_element_set_map[line_element].
        insert(make_pair(element,level));
    }
# endif
  }
  for (bool doit=tess->firstShape();doit;doit=tess->nextShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
# if (SPACEDIM==2)
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int line_element_index=element->codimension1ShapeIndex(f);
      const Shape1 *line_element=
        level->Level1::getShape(line_element_index);
      line_element_set_map[line_element].
        insert(make_pair(element,level));
    }
# else
    for (int l=0;l<element->numberCodimension2Shapes();l++) {
      int line_element_index=element->codimension2ShapeIndex(l);
      const Shape1 *line_element=
        level->Level1::getShape(line_element_index);
      line_element_set_map[line_element].
        insert(make_pair(element,level));
    }
# endif
  }
  int max_couplings=0;
  LEVEL_CLASS *tess_coarsest_level=tess->coarsestLevel();
  LEVEL_CLASS *tess_finest_level=tess_coarsest_level->finestLevel();
  for (map<const Shape1*,
  set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> > >::
  const_iterator laesm_iterator=line_active_element_set_map.begin();
  laesm_iterator!=line_active_element_set_map.end();laesm_iterator++) {
    const Shape1 *line_element=laesm_iterator->first;
    const set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >
      &line_element_set=laesm_iterator->second;
//  find finest and coarsest levels of elements with this line
    const LEVEL_CLASS *finest_level=tess_coarsest_level;
    const LEVEL_CLASS *coarsest_level=tess_finest_level;
    for (set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >
    ::const_iterator les_iterator=line_element_set.begin();les_iterator!=
    line_element_set.end();les_iterator++) {
      const LEVEL_CLASS *level=les_iterator->second;
      if (level->isFinerThan(finest_level)) finest_level=level;
      if (level->isCoarserThan(coarsest_level)) coarsest_level=level;
    }
    int couplings=0;
    int line_element_index=line_element->elementDataIndex();
    if (finest_level==coarsest_level) {
//    all indices of faces and lines in elements containing this line
      set<int> codimension1_shape_indices;
# if (SPACEDIM==3)
      set<int> codimension2_shape_indices;
# endif
      for (set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >
      ::const_iterator les_iterator=line_element_set.begin();
      les_iterator!=line_element_set.end();les_iterator++) {
        const SHAPE_CLASS *element=les_iterator->first;
# if (SPACEDIM==2)
        couplings+=2*selected_fe->dofsPerShape(element)
                  +selected_fe->multiplierDofsPerShape(element);
# else
        couplings+=3*selected_fe->dofsPerShape(element)
                  +3*selected_fe->multiplierDofsPerShape(element);
# endif
        for (int f=0;f<element->numberCodimension1Shapes();f++) {
          codimension1_shape_indices.
            insert(element->codimension1ShapeIndex(f));
        }
# if (SPACEDIM==3)
        for (int l=0;l<element->numberCodimension2Shapes();l++) {
          codimension2_shape_indices.
            insert(element->codimension2ShapeIndex(l));
        }
# endif
      }
# if (SPACEDIM==2)
      couplings+=codimension1_shape_indices.size()
                *selected_fe->dofsPerFace(2);
# else
      couplings+=codimension2_shape_indices.size()
                *selected_fe->dofsPerLine();
      for (set<int>::const_iterator face_iterator=
      codimension1_shape_indices.begin();face_iterator!=
      codimension1_shape_indices.end();face_iterator++) {
        const CODIMENSION1_SHAPE_CLASS *face=finest_level->
          CODIMENSION1_LEVEL_CLASS::getShape(*face_iterator);
        couplings+=selected_fe->dofsPerFace(face);
      }
# endif
    } else {
//    faces and lines in coarse active elements containing this line
      set<int> coarse_codimension1_shape_indices;
# if (SPACEDIM==3)
      set<int> coarse_codimension2_shape_indices;
# endif
      for (set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >::
      const_iterator les_iterator=line_element_set.begin();
      les_iterator!=line_element_set.end();les_iterator++) {
        const LEVEL_CLASS *level=les_iterator->second;
        if (level==coarsest_level) {
          const SHAPE_CLASS *element=les_iterator->first;
          couplings+=selected_fe->dofsPerShape(element);
#if (SPACEDIM==2)
          couplings+=selected_fe->multiplierDofsPerShape(element);
#endif
          for (int f=0;f<element->numberCodimension1Shapes();f++) {
            coarse_codimension1_shape_indices.
              insert(element->codimension1ShapeIndex(f));
          }
# if (SPACEDIM==3)
          for (int l=0;l<element->numberCodimension2Shapes();l++) {
            coarse_codimension2_shape_indices.
              insert(element->codimension2ShapeIndex(l));
          }
# endif
        }
      }
//    coarse elements having children and containing this line
      set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> > les_difference;
      for (set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >::
      const_iterator les_iterator=line_element_set.begin();
      les_iterator!=line_element_set.end();les_iterator++) {
        const pair<const SHAPE_CLASS*,const LEVEL_CLASS*> &p=
          *les_iterator;
        if (p.second!=coarsest_level) continue;
        if (p.second->hasChildren(p.first)) {
          les_difference.insert(*les_iterator);
        }
      }
//    indices of children of this line
      set<int> fine_line_indices;
# if (SPACEDIM==2)
      for (int c=0;c<childrenPerCodimension1Shape();c++)
# else
      for (int c=0;c<childrenPerCodimension2Shape();c++)
# endif
      {
        fine_line_indices.insert(finest_level->
          Level1::childIndex(line_element_index,c));
      }
//    coarse element children containing children of this line
      set<int> fine_element_indices;
      const set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >
        &line_inactive_element_set=line_element_set_map[line_element];
      for (set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >::
      const_iterator lesdi=les_difference.begin();lesdi!=
      les_difference.end();lesdi++) {
        const LEVEL_CLASS *level=lesdi->second;
        const SHAPE_CLASS *element=lesdi->first;
        CHECK_TEST(level->hasChildren(element));
        for (int c=0;c<childrenPerShape();c++) {
          int child_index=
            level->childIndex(element->elementDataIndex(),c);
          const SHAPE_CLASS *child_element=
            finest_level->getShape(child_index);
# if (SPACEDIM==2)
          for (int f=0;f<child_element->numberCodimension1Shapes();f++) {
            set<int>::const_iterator fli_iterator=fine_line_indices.
              find(child_element->codimension1ShapeIndex(f));
            if (fli_iterator!=fine_line_indices.end()) {
              fine_element_indices.insert(child_index);
              break;
            }
          }
# else
          for (int l=0;l<child_element->numberCodimension2Shapes();l++) {
            set<int>::const_iterator fli_iterator=fine_line_indices.
              find(child_element->codimension2ShapeIndex(l));
            if (fli_iterator!=fine_line_indices.end()) {
              fine_element_indices.insert(child_index);
              break;
            }
          }
# endif
        }
      }
//    faces and lines of coarse element children
      set<int> fine_codimension1_shape_indices;
# if (SPACEDIM==3)
      set<int> fine_codimension2_shape_indices;
# endif
      for (set<int>::const_iterator fei_iterator=
      fine_element_indices.begin();fei_iterator!=
      fine_element_indices.end();fei_iterator++) {
        const SHAPE_CLASS *child_element=
          finest_level->getShape(*fei_iterator);
        for (int f=0;f<child_element->numberCodimension1Shapes();f++) {
          fine_codimension1_shape_indices.insert(
            child_element->codimension1ShapeIndex(f));
        }
# if (SPACEDIM==3)
        for (int l=0;l<child_element->numberCodimension2Shapes();l++) {
          fine_codimension2_shape_indices.insert(
            child_element->codimension2ShapeIndex(l));
        }
# endif
        couplings+=selected_fe->dofsPerShape(child_element);
#if (SPACEDIM==2)
        couplings+=selected_fe->multiplierDofsPerShape(child_element);
#endif
      }
# if (SPACEDIM==2)
      couplings+=(fine_codimension1_shape_indices.size()
                 +coarse_codimension1_shape_indices.size())
                *selected_fe->dofsPerFace(2);
# else
      couplings+=(fine_codimension2_shape_indices.size()
                 +coarse_codimension2_shape_indices.size())
                *selected_fe->dofsPerLine();
      for (set<int>::const_iterator fcd1si_iterator=
      fine_codimension1_shape_indices.begin();fcd1si_iterator!=
      fine_codimension1_shape_indices.end();fcd1si_iterator++) {
        const CODIMENSION1_SHAPE_CLASS *child_face_element=
          finest_level->CODIMENSION1_LEVEL_CLASS::
            getShape(*fcd1si_iterator);
        couplings+=selected_fe->dofsPerFace(child_face_element);
      }
      for (set<int>::const_iterator ccd1si_iterator=
      coarse_codimension1_shape_indices.begin();ccd1si_iterator!=
      coarse_codimension1_shape_indices.end();ccd1si_iterator++) {
        const CODIMENSION1_SHAPE_CLASS *face_element=
          coarsest_level->CODIMENSION1_LEVEL_CLASS::
            getShape(*ccd1si_iterator);
          couplings+=selected_fe->dofsPerFace(face_element);
      }
# endif
    }
    max_couplings=max(max_couplings,couplings);
  }
  return min(max_couplings,numberDofs());
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::assembleLinearSystem(const Mapping &mapping,
const map<int,Quadrature<SPACEDIM> > &qmap,
map<int,Quadrature<SPACEDIM-1> > &fqmap,SparseMatrix &system_matrix,
Vector &system_rhs) const {
  UPDATE_FLAGS update_flags= UPDATE_VALUES | UPDATE_CURL | 
    UPDATE_DETERMINANT | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  for (bool doit=tess->firstShape();doit;doit=tess->nextShapeInLevel()) {
    HcurlMaterialModel *mm=dynamic_cast<HcurlMaterialModel*>(
      tess->presentLevel()->getMaterialModel());
    CHECK_POINTER(mm);
    if (mm->requiresDerivative()) {
      update_flags |= UPDATE_GRADIENTS;
      break;
    }
  }
  HcurlFEValues fe_values(mapping,*selected_fe,qmap,update_flags);
#if (SPACEDIM>=2)
  UPDATE_FLAGS face_update_flags=UPDATE_VALUES | UPDATE_Q_POINTS |
    UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES | 
    UPDATE_MAPPING_JACOBIAN_INVERSE;
  HcurlFEFaceValues fe_face_values(mapping,*selected_fe,fqmap,
    face_update_flags);
#endif
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);

//  local stiffness matrix and right-hand side
    int dofs_per_cell=local_dof_indices.getNumber();
    Matrix cell_matrix(dofs_per_cell,dofs_per_cell);
    Vector cell_rhs(dofs_per_cell);
    cell_matrix=0.;
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,Tensor<1,SPACEDIM> > *sv=
      fe_values.getShapeValues(element);
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE > *msv=
      fe_values.getMultiplierShapeValues(element);
    int multiplier_dofs_per_shape=
      selected_fe->multiplierDofsPerShape(element);

    HcurlMaterialModel *mm=dynamic_cast<HcurlMaterialModel*>(
      present_level->getMaterialModel());
    CHECK_POINTER(mm);

//  computation of local stiffness matrix and right-hand side
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE> *sc=
      fe_values.getShapeCurl(element);
    mm->elementStiffnessMatrixAndRHS(q_points,jxw_values,sv,sc,msv,
      cell_matrix,cell_rhs);

//  neumann boundary conditions
//    - int_Omega delta E . grad times H dx
//    = - int_Omega H . grad times delta E dx
//    + int_{partial Omega} H . normal times delta E da
//  in 2D: normal times delta E = tangent . delta E
//    - int_Omega delta E . grad times H_2 dx
//    = - int_Omega H_2 ( grad times delta E )_2 dx
//    + int_{partial Omega} H_2 tangent . delta E da
//  in 1D:
//    - int_a^b delta E_1 (- partial_0 H_2) dx
//    = - int_a^b H_2 partial_0 delta E_1 dx
//    + H_2 delta E_1 |_a^b
#if (SPACEDIM==1)
//  copied from HdivDofHandler:
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (present_level->neighborLevel(f)==0) {
        const BoundaryCondition *bc=
          present_level->Level0::getBoundaryCondition(f);
        if (bc) {
          if (bc->isNatural()) {
            NumPtr<Point<1> > p(1);
            p[0]=Interval::canonicalVertex(f);
            map<int,Quadrature<1> > fqmap_1d;
            fqmap_1d[element->numberVertices()]=Quadrature<1>(p);
            HcurlFEValues face_fe_values(mapping,*selected_fe,fqmap_1d,
              UPDATE_VALUES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES |
              UPDATE_MAPPING_JACOBIAN_INVERSE);
            face_fe_values.mapShapes(*tess,element,present_level);
            const Array<2,Tensor<1,1> > &fsv=
              *face_fe_values.getShapeValues(element);
            const NumPtr<Point<1> > &face_q_points=
              face_fe_values.getQuadraturePoints(element);
            NumPtr<REAL> face_jxw_values(1); face_jxw_values[0]=1.;
            const HcurlBoundaryCondition *hbc=
              dynamic_cast<const HcurlBoundaryCondition*>(bc);
            CHECK_POINTER(hbc);
            hbc->faceNeumannValues(face_q_points,face_jxw_values,f,0,fsv,
              cell_rhs);
          }
        }
      }
    }
#else
// int_mapping(Gamma) m(mapping^{-1}(x)) n_{mapping(Gamma)}(x) . u(x) da
// = int_Gamma m(xi) n_{mapping(Gamma)}(mapping(xi)) . u(mapping(xi))
//   || D mapping^{-1}(xi) n_Gamma det D mapping(xi) || da_Gamma
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const BoundaryCondition *bc=present_level->
        CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
      if (bc) {
        if (bc->isNatural()) {
//        HcurlBoundaryCondition can be natural for some components, if
//          n_components > 1; we haven't programmed that case
//        CHECK_TEST(dynamic_cast<const HcurlBoundaryCondition*>(bc)==0);
          const CODIMENSION1_SHAPE_CLASS *face_element=present_level->
            CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          fe_face_values.mapShapes(*tess,element,present_level,f);
          const Array<2,Tensor<1,SPACEDIM> > &fsv=
            *fe_face_values.getShapeValues(element);
          NumPtr<Point<SPACEDIM> > face_q_points=
            fe_face_values.getQuadraturePoints(element,f);
          NumPtr<REAL> face_jxw_values=
            fe_face_values.getJxWValues(element,f);
          const NumPtr<Point<SPACEDIM> > &normal_vectors=
            fe_face_values.getNormalVectors(element,f);
          const HcurlBoundaryCondition *hbc=
            dynamic_cast<const HcurlBoundaryCondition*>(bc);
          CHECK_POINTER(hbc);
          hbc->faceNeumannValues(face_q_points,face_jxw_values,
            normal_vectors,offset,fsv,cell_rhs);
        }
      }
      offset+=
        fe_face_values.getFaceQuadrature(element,f).numberPoints();
    }
#endif

//  add element-wise contributions to global arrays
    int ndofs=local_dof_indices.getNumber();
    CHECK_SAME(ndofs,cell_matrix.size(0));
    for (int i=0;i<ndofs;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<ndofs;j++) {
        system_matrix.add(ldfi,local_dof_indices[j],cell_matrix(i,j));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::assembleLinearSystemForInitialData(
const Mapping &mapping,const map<int,Quadrature<SPACEDIM> > &qmap,
SparseMatrix &system_mass_matrix,Vector &system_rhs) const {
  int system_ns=
    D_CHOOSE(used_dofs,used_dofs+total_solution_dofs,3*used_dofs);
  CHECK_SAME(system_mass_matrix.size(0),system_ns);
  CHECK_SAME(system_mass_matrix.size(0),system_mass_matrix.size(1));
  CHECK_SAME(system_mass_matrix.size(0),system_rhs.size());
  UPDATE_FLAGS update_flags= UPDATE_VALUES | UPDATE_DETERMINANT | 
    UPDATE_JxW_VALUES | UPDATE_Q_POINTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
  HcurlFEValues fe_values(mapping,*selected_fe,qmap,update_flags);
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);
    int multiplier_dofs_per_cell=
      selected_fe->multiplierDofsPerShape(element);
    int ns=D_CHOOSE(dofs_per_cell+multiplier_dofs_per_cell,
      2*dofs_per_cell+multiplier_dofs_per_cell,
      3*(dofs_per_cell+multiplier_dofs_per_cell));

//  local matrices and right-hand side
    Matrix cell_mass_matrix(ns,ns);
    Vector cell_rhs(ns);
    cell_mass_matrix=0.;
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,Tensor<1,SPACEDIM> > *sv=
      fe_values.getShapeValues(element);
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE> *msv=
      fe_values.getMultiplierShapeValues(element);

    HcurlTransientMaterialModel *mm=
      dynamic_cast<HcurlTransientMaterialModel*>(
        present_level->getMaterialModel());
    CHECK_POINTER(mm);

//  computation of local mass matrix and right-hand side
    mm->elementL2ProjectionMatrixAndRHS(q_points,jxw_values,sv,msv,
      cell_mass_matrix,cell_rhs);
//  add element-wise contributions to global arrays
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);
#if (SPACEDIM==1)
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_mass_matrix.add(ldfi,local_dof_indices[j],
          cell_mass_matrix(i,j));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int dpcpi=dofs_per_cell+i;
      int ldfi=local_dof_indices[dpcpi];
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int dpcpj=dofs_per_cell+j;
        system_mass_matrix.add(ldfi,local_dof_indices[dpcpj],
          cell_mass_matrix(dpcpi,dpcpj));
      }
      system_rhs[ldfi]+=cell_rhs[dpcpi];
    }
#elif (SPACEDIM==2)
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_mass_matrix.add(ldfi,local_dof_indices[j],
          cell_mass_matrix(i,j));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
    for (int i=0;i<dofs_per_cell;i++) {
      int dpcpi=dofs_per_cell+i;
      int ldfi=total_solution_dofs+local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_mass_matrix.add(ldfi,
          total_solution_dofs+local_dof_indices[j],
          cell_mass_matrix(dpcpi,dofs_per_cell+j));
      }
      system_rhs[ldfi]+=cell_rhs[dpcpi];
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int dpcpi=2*dofs_per_cell+i;
      int ldfi=total_solution_dofs+local_dof_indices[dofs_per_cell+i];
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        system_mass_matrix.add(ldfi,
          total_solution_dofs+local_dof_indices[dofs_per_cell+j],
          cell_mass_matrix(dpcpi,2*dofs_per_cell+j));
      }
      system_rhs[ldfi]+=cell_rhs[dpcpi];
    }
#else
    OBSOLETE("fix");
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::assembleTransientLinearSystem(
const Mapping &mapping,const map<int,Quadrature<SPACEDIM> > &qmap,
map<int,Quadrature<SPACEDIM-1> > &fqmap,
SparseMatrix &system_mass_matrix,SparseMatrix &system_stiffness_matrix,
Vector &system_rhs) const {
  UPDATE_FLAGS update_flags= UPDATE_VALUES | UPDATE_CURL | 
    UPDATE_DETERMINANT | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  for (bool doit=tess->firstShape();doit;doit=tess->nextShapeInLevel()) {
    HcurlTransientMaterialModel *mm=
      dynamic_cast<HcurlTransientMaterialModel*>(
        tess->presentLevel()->getMaterialModel());
    CHECK_POINTER(mm);
    if (mm->requiresDerivative()) {
      update_flags |= UPDATE_GRADIENTS;
      break;
    }
  }
  HcurlFEValues fe_values(mapping,*selected_fe,qmap,update_flags);
#if (SPACEDIM>=2)
  UPDATE_FLAGS face_update_flags=UPDATE_VALUES | UPDATE_Q_POINTS |
    UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES | 
    UPDATE_MAPPING_JACOBIAN_INVERSE;
  HcurlFEFaceValues fe_face_values(mapping,*selected_fe,fqmap,
    face_update_flags);
#endif
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);
    int multiplier_dofs_per_cell=
      selected_fe->multiplierDofsPerShape(element);
    int ns=D_CHOOSE(dofs_per_cell+multiplier_dofs_per_cell,
      2*dofs_per_cell+multiplier_dofs_per_cell,
      3*(dofs_per_cell+multiplier_dofs_per_cell));

//  local matrices and right-hand side
    Matrix cell_mass_matrix(ns,ns);
    Matrix cell_stiffness_matrix(ns,ns);
    Vector cell_rhs(ns);
    cell_mass_matrix=0.;
    cell_stiffness_matrix=0.;
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,Tensor<1,SPACEDIM> > *sv=
      fe_values.getShapeValues(element);
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE> *msv=
      fe_values.getMultiplierShapeValues(element);
#if (SPACEDIM<=2)
    const Array<2,Tensor<2,SPACEDIM> > *sg=
      fe_values.getShapeGradients(element);
#else
    const Array<2,Tensor<1,3> > *sc=fe_values.getShapeCurl(element);
#endif

    HcurlTransientMaterialModel *mm=
      dynamic_cast<HcurlTransientMaterialModel*>(
        present_level->getMaterialModel());
    CHECK_POINTER(mm);

//  computation of local stiffness matrix and right-hand side
#if (SPACEDIM<=2)
    mm->elementStiffnessMatrixAndRHS(q_points,jxw_values,sv,sg,msv,
      cell_mass_matrix,cell_stiffness_matrix,cell_rhs);
#else
    mm->elementStiffnessMatrixAndRHS(q_points,jxw_values,sv,sc,msv,
      cell_mass_matrix,cell_stiffness_matrix,cell_rhs);
#endif

//  neumann boundary conditions
#if (SPACEDIM==1)
//  copied from HdivDofHandler:
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (present_level->neighborLevel(f)==0) {
        const BoundaryCondition *bc=
          present_level->Level0::getBoundaryCondition(f);
        if (bc) {
          if (bc->isNatural()) {
            NumPtr<Point<1> > p(1);
            p[0]=Interval::canonicalVertex(f);
            map<int,Quadrature<1> > fqmap_1d;
            fqmap_1d[element->numberVertices()]=Quadrature<1>(p);
            HcurlFEValues face_fe_values(mapping,*selected_fe,fqmap_1d,
              UPDATE_VALUES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES |
              UPDATE_MAPPING_JACOBIAN_INVERSE);
            face_fe_values.mapShapes(*tess,element,present_level);
            const Array<2,Tensor<1,1> > &fsv=
              *face_fe_values.getShapeValues(element);
            const NumPtr<Point<1> > &face_q_points=
              face_fe_values.getQuadraturePoints(element);
            NumPtr<REAL> face_jxw_values(1); face_jxw_values[0]=1.;
            const HcurlBoundaryCondition *hbc=
              dynamic_cast<const HcurlBoundaryCondition*>(bc);
            CHECK_POINTER(hbc);
            hbc->faceNeumannValues(face_q_points,face_jxw_values,f,0,fsv,
              cell_rhs);
          }
        }
      }
    }
#else
// int_mapping(Gamma) m(mapping^{-1}(x)) n_{mapping(Gamma)}(x) . u(x) da
// = int_Gamma m(xi) n_{mapping(Gamma)}(mapping(xi)) . u(mapping(xi))
//   || D mapping^{-1}(xi) n_Gamma det D mapping(xi) || da_Gamma
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const BoundaryCondition *bc=present_level->
        CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
      if (bc) {
        if (bc->isNatural()) {
//        HcurlBoundaryCondition can be natural for some components, if
//          n_components > 1; we haven't programmed that case
//        CHECK_TEST(dynamic_cast<const HcurlBoundaryCondition*>(bc)==0);
          const CODIMENSION1_SHAPE_CLASS *face_element=present_level->
            CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          fe_face_values.mapShapes(*tess,element,present_level,f);
          const Array<2,Tensor<1,SPACEDIM> > &fsv=
            *fe_face_values.getShapeValues(element);
          NumPtr<Point<SPACEDIM> > face_q_points=
            fe_face_values.getQuadraturePoints(element,f);
          NumPtr<REAL> face_jxw_values=
            fe_face_values.getJxWValues(element,f);
          const NumPtr<Point<SPACEDIM> > &normal_vectors=
            fe_face_values.getNormalVectors(element,f);
          const HcurlBoundaryCondition *hbc=
            dynamic_cast<const HcurlBoundaryCondition*>(bc);
          CHECK_POINTER(hbc);
          hbc->faceNeumannValues(face_q_points,face_jxw_values,
            normal_vectors,offset,fsv,cell_rhs);
        }
      }
      offset+=
        fe_face_values.getFaceQuadrature(element,f).numberPoints();
    }
#endif

//  add element-wise contributions to global arrays
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);
#if (SPACEDIM==1)
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        int ldfj=local_dof_indices[j];
        system_mass_matrix.add(ldfi,ldfj,cell_mass_matrix(i,j));
        system_stiffness_matrix.add(ldfi,ldfj,cell_stiffness_matrix(i,j));
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int dpcpj=dofs_per_cell+j;
        int ldfj=local_dof_indices[dpcpj];
        system_stiffness_matrix.add(ldfi,ldfj,
          cell_stiffness_matrix(i,dpcpj));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int dpcpi=dofs_per_cell+i;
      int ldfi=local_dof_indices[dpcpi];
      for (int j=0;j<dofs_per_cell;j++) {
        int ldfj=local_dof_indices[j];
        system_stiffness_matrix.add(ldfi,ldfj,
          cell_stiffness_matrix(dpcpi,j));
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int dpcpj=dofs_per_cell+j;
        int ldfj=local_dof_indices[dpcpj];
        system_mass_matrix.add(ldfi,ldfj,cell_mass_matrix(dpcpi,dpcpj));
        system_stiffness_matrix.add(ldfi,ldfj,
          cell_stiffness_matrix(dpcpi,dpcpj));
      }
      system_rhs[ldfi]+=cell_rhs[dpcpi];
    }
#elif (SPACEDIM==2)
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_mass_matrix.add(ldfi,local_dof_indices[j],
          cell_mass_matrix(i,j));
        system_stiffness_matrix.add(ldfi,local_dof_indices[j],
          cell_stiffness_matrix(i,j));
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int dpcpj=dofs_per_cell+j;
        int ldfj=total_solution_dofs+local_dof_indices[dpcpj];
        system_stiffness_matrix.add(ldfi,ldfj,
          cell_stiffness_matrix(i,dofs_per_cell+dpcpj));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
    for (int i=0;i<dofs_per_cell;i++) {
      int dpcpi=dofs_per_cell+i;
      int ldfi=total_solution_dofs+local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_mass_matrix.add(ldfi,
          total_solution_dofs+local_dof_indices[j],
          cell_mass_matrix(dpcpi,dofs_per_cell+j));
        system_stiffness_matrix.add(ldfi,
          total_solution_dofs+local_dof_indices[j],
          cell_stiffness_matrix(dpcpi,dofs_per_cell+j));
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        int dpcpj=dofs_per_cell+j;
        int ldfj=total_solution_dofs+local_dof_indices[dpcpj];
        system_stiffness_matrix.add(ldfi,ldfj,
          cell_stiffness_matrix(dpcpi,dofs_per_cell+dpcpj));
      }
      system_rhs[ldfi]+=cell_rhs[dpcpi];
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      int dpcpi=2*dofs_per_cell+i;
      int ldfi=total_solution_dofs+local_dof_indices[dofs_per_cell+i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_stiffness_matrix.add(ldfi,local_dof_indices[j],
          cell_stiffness_matrix(dpcpi,j));
      }
      for (int j=0;j<dofs_per_cell;j++) {
        system_stiffness_matrix.add(ldfi,
          total_solution_dofs+local_dof_indices[j],
          cell_stiffness_matrix(dpcpi,dofs_per_cell+j));
      }
      for (int j=0;j<multiplier_dofs_per_cell;j++) {
        system_mass_matrix.add(ldfi,
          total_solution_dofs+local_dof_indices[dofs_per_cell+j],
          cell_mass_matrix(dpcpi,2*dofs_per_cell+j));
        system_stiffness_matrix.add(ldfi,
          total_solution_dofs+local_dof_indices[dofs_per_cell+j],
          cell_stiffness_matrix(dpcpi,2*dofs_per_cell+j));
      }
      system_rhs[ldfi]+=cell_rhs[dpcpi];
    }
#else
    OBSOLETE("fix");
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::interpolateBoundaryValues(const Mapping &mapping,
map<int,REAL> &boundary_values) const {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
#if (SPACEDIM==1)
  map<int,Quadrature<SPACEDIM> > qmap;
  qmap[Interval::verticesPerShape()]=NewtonCotesQuadrature<1>(2);
  HcurlFEValues 
    fe_values(mapping,*selected_fe,qmap,UPDATE_MAPPING_JACOBIAN);
#else
  UPDATE_FLAGS face_update_flags=
    UPDATE_Q_POINTS | UPDATE_MAPPING_JACOBIAN_INVERSE; 
  map<int,Quadrature<SPACEDIM-1> > &fqmap=
    selected_fe->faceQuadratureMap();
  HcurlFEFaceValues fe_face_values(mapping,*selected_fe,fqmap,
    face_update_flags);
# if (SPACEDIM==3)
  const Quadrature<1> &eq=selected_fe->edgeQuadrature();
  HcurlFEEdgeValues fe_edge_values(mapping,*selected_fe);
# endif
#endif
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    for (int face_no=0;face_no<element->numberCodimension1Shapes();
    face_no++) {
      if (level->neighborLevel(face_no)==0) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
#if (SPACEDIM==1)
        const HcurlBoundaryCondition *bc=
          dynamic_cast<const HcurlBoundaryCondition*>(
          codimension1_level->getBoundaryCondition(face_no));
        if (bc) {
          if (bc->isEssential()) {
            fe_values.mapShapes(*tess,element,level);
            const MappingDataBase *md=fe_values.mappingData(element);
            NumPtr<Tensor<1,1> > function_values(n_components);
            const Point<SPACEDIM> &boundary_point=tess->
              getVertices()[element->codimension1ShapeIndex(face_no)];
            TensorFunctionSystem *fs=bc->getTensorFunctionSystem();
            CHECK_POINTER(fs);
            fs->value(boundary_point,function_values);
            NumPtr<int> face_dof_indices;
            getFaceDofIndices(element,face_no,level,face_dof_indices);
            for (int c=0;c<n_components;c++) {
              if (bc->isEssential(c)) {
                Tensor<1,1> bv;
                md->preMultiplyByJacobianTranspose(bv,face_no,
                  function_values[c]);
                boundary_values[face_dof_indices[c]]=bv[0];
              }
            }
          }
        }
#else
        int face_index=element->codimension1ShapeIndex(face_no);
        const HcurlBoundaryCondition *bc=
          dynamic_cast<const HcurlBoundaryCondition*>(
          codimension1_level->getBoundaryCondition(face_index));
        if (bc) {
          if (bc->isEssential()) {
            const CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape(face_index);
            fe_face_values.mapShapes(*tess,element,level,face_no);
            const MappingDataBase *md=fe_face_values.mappingData(element);
            NumPtr<Point<SPACEDIM> > dof_locations=
              fe_face_values.getQuadraturePoints(element,face_no);
            const Quadrature<SPACEDIM-1> &fq=
              fe_face_values.getFaceQuadrature(element,face_no);
            const NumPtr<Point<SPACEDIM-1> > &canonical_locations=
              fq.points();
            NumPtr<int> face_dofs;
            getFaceDofIndices(element,face_no,level,face_dofs);
            NumPtr<int> element_dofs;
            getDofIndices(element,level,element_dofs);
            NumPtr<NumPtr<Tensor<1,SPACEDIM> > > 
              bc_face_values(dof_locations.getNumber());
            TensorFunctionSystem *fs=bc->getTensorFunctionSystem();
            if (n_components==1) {
              NumPtr<Tensor<1,SPACEDIM> > bc_face_values_scalar(
                dof_locations.getNumber());
              fs->value(bc_face_values_scalar,dof_locations,0);
              for (int q=0;q<bc_face_values.getNumber();q++) {
                bc_face_values[q].allocate(1);
                bc_face_values[q][0]=bc_face_values_scalar[q];
              }
            } else {
              fs->value(bc_face_values,dof_locations);
            }
            NumPtr<NumPtr<Tensor<1,SPACEDIM> > > 
              transformed_bc_face_values(bc_face_values.getNumber());
            for (int q=0;q<bc_face_values.getNumber();q++) {
              transformed_bc_face_values[q].allocate(n_components);
              for (int c=0;c<n_components;c++) {
                md->preMultiplyByJacobianTranspose(
                  transformed_bc_face_values[q][c],q,bc_face_values[q][c]);
              }
            }
            selected_fe->setFaceBoundaryValues(element,face_no,
              face_dofs,canonical_locations,fq.weights(),
              transformed_bc_face_values,bc,boundary_values);
# if (SPACEDIM==3)
//          edge values may be set twice in neighboring faces
//          this would take some effort to prevent
            fe_edge_values.mapShapes(*tess,element,level,face_no);
            const MappingDataBase *emd=fe_edge_values.mappingData(element);
            const Quadrature<1> &lq=fe_edge_values.getEdgeQuadrature();
            const NumPtr<Point<1> > &edge_canonical_locations=lq.points();
            int nl=face_element->numberCodimension1Shapes();
            for (int l=0;l<nl;l++) {
              Level1 *level1=static_cast<Level1*>(level);
              int line_index=face_element->codimension1ShapeIndex(l);
              const Shape1 *line_element=level1->getShape(line_index);
              NumPtr<Point<SPACEDIM> > dof_locations=
                fe_edge_values.getQuadraturePoints(element,l);
              NumPtr<int> line_dofs;
              getLineDofIndices(line_element,level,line_dofs);
              NumPtr<NumPtr<Tensor<1,3> > > 
                bc_edge_values(dof_locations.getNumber());
              if (n_components==1) {
                NumPtr<Tensor<1,SPACEDIM> > bc_edge_values_scalar(
                  dof_locations.getNumber());
                fs->value(bc_edge_values_scalar,dof_locations,0);
                for (int q=0;q<bc_edge_values.getNumber();q++) {
                  bc_edge_values[q].allocate(1);
                  bc_edge_values[q][0]=bc_edge_values_scalar[q];
                }
              } else {
                fs->value(bc_edge_values,dof_locations);
              }
              NumPtr<NumPtr<Tensor<1,3> > > 
                transformed_bc_edge_values(bc_edge_values.getNumber());
              for (int q=0;q<bc_edge_values.getNumber();q++) {
                transformed_bc_edge_values[q].allocate(n_components);
                for (int c=0;c<n_components;c++) {
                  md->preMultiplyByJacobianTranspose(
                    transformed_bc_edge_values[q][c],q,
                    bc_edge_values[q][c]);
                }
              }
              selected_fe->setLineBoundaryValues(element,face_element,
                face_no,l,line_dofs,edge_canonical_locations,lq.weights(),
                transformed_bc_edge_values,bc,boundary_values);
            }
# endif
          }
        }
#endif
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::interpolateBoundaryValues(const Mapping &mapping,
map<int,REAL> &boundary_values,map<int,Quadrature<SPACEDIM-1> > &dqmap)
const {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    for (int face_no=0;
    face_no<element->numberCodimension1Shapes();face_no++) {
      int face_index=element->codimension1ShapeIndex(face_no);
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
#if (SPACEDIM==1)
      OBSOLETE("fix like other interpolateBoundaryValues");
#else
      OBSOLETE("fix like other interpolateBoundaryValues");
#endif
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//eliminates boundary dof from matrix and right-hand-side
//  for curved boundaries with normal != axis vector, this is incorrect
void HcurlDofHandler::applyBoundaryValues(
const map<int,REAL> &boundary_values,SparseMatrix &matrix,
Vector &right_hand_side,bool preserve_symmetry)
const {
  int n_rows=matrix.size(0);
  CHECK_SAME(matrix.size(1),n_rows);
  CHECK_SAME(n_rows,right_hand_side.size());
  if (boundary_values.size()==0) return;
  const SparsityPattern &sparsity=matrix.getSparsityPattern();
  const NumPtr<int> &sparsity_colnums=sparsity.getColumnNumbers();

  REAL first_nonzero_diagonal_entry=1.;
  for (int i=0;i<n_rows;i++) {
    REAL diag=matrix.diagElement(i);
    if (abs(diag)>0.) {
      first_nonzero_diagonal_entry=diag;
      break;
    }
  }
//sum_j A_ij x_j = b_i, x_k given ==>
//  A_kk x_k = (A_kk x_k)
//  for i!=k, sum_{j!=k} A_ij x_j = b_i - A_ik x_k
  for (map<int,REAL>::const_iterator bv_iterator=
  boundary_values.begin();bv_iterator!=boundary_values.end();
  bv_iterator++) {
    int dof_number=bv_iterator->first;
    CHECK_TEST(dof_number<n_rows);
//  set matrix row to zero, except for diagonal
    int last=sparsity.getRowstartIndex(dof_number+1);
    for (int j=sparsity.getRowstartIndex(dof_number);j<last;j++) {
      if (sparsity.getColumnNumber(j)!=dof_number) {
        matrix.globalEntry(j)=0.;
      }
    }
//  set diagonal entry to nonzero, set new right-hand-side for this row
    REAL diag=matrix.diagElement(dof_number);
    REAL boundary_value=bv_iterator->second;
    if (abs(diag)<=0.0) {
      diag=first_nonzero_diagonal_entry;
      matrix.set(dof_number,dof_number,first_nonzero_diagonal_entry);
    }
    right_hand_side[dof_number]=boundary_value*diag;
//  zero matrix column off diagonal, modify right-hand-side in other rows
    if (preserve_symmetry) {
      for (int j=sparsity.getRowstartIndex(dof_number);j<last;j++) {
        if (sparsity.getColumnNumber(j)!=dof_number) {
          int row=sparsity_colnums[j];
          int p=sparsity.getRowstartIndex(row+1);
          for (int c=sparsity.getRowstartIndex(row)+1;
          c<sparsity.getRowstartIndex(row+1);c++) {
            if (sparsity_colnums[c]>=dof_number) {
              p=c; break;
            }
          }
          CHECK_TEST((sparsity_colnums[p]==dof_number) &&
           (p!=sparsity.getRowstartIndex(row+1)));
          int global_entry=p-sparsity.getRowstartIndex(0);
          REAL &mij=matrix.globalEntry(global_entry);
          right_hand_side[row]-=mij*boundary_value;
          mij=0.;
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//sets solution at boundary dofs
void HcurlDofHandler::applyBoundaryValues(
const map<int,REAL> &boundary_values,Vector &solution) const {
  if (boundary_values.size()==0) return;
  for (map<int,REAL>::const_iterator dof=
  boundary_values.begin();dof!=boundary_values.end();dof++) {
    solution[dof->first]=dof->second;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::printOn(ostream &os) const {
  os << "HcurlDofHandler:" << endl;
  os << "\ttess = " << tess
     << "\n\tselected_fe = " << selected_fe
     << "\n\tused_dofs = " << used_dofs
     << "\n\tlevels.getNumber = " << levels.getNumber()
     << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::printElementDofs() const {
  cout << "\n\telement dofs" << endl;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel()) {
    for (bool doit=tess->firstShape(level);
    doit && tess->presentLevel()==level;doit=tess->nextShapeInLevel()) {
      const SHAPE_CLASS *element=level->getShape();
      cout << "\tdof_index[" << element->elementDataIndex() << "] = "
           << element->dofIndex() << endl;
    }
  }
#if (SPACEDIM>=2)
  cout << "\n\tface dofs" << endl;
  for (CODIMENSION1_LEVEL_CLASS *level=
  static_cast<CODIMENSION1_LEVEL_CLASS*>(tess->coarsestLevel());
  level;level=level->finerLevel()) {
    for (bool doit=tess->firstCodimension1Shape(level);
    doit && tess->presentCodimension1Level()==level;
    doit=tess->nextCodimension1ShapeInLevel()) {
      const CODIMENSION1_SHAPE_CLASS *element=level->getShape();
      cout << "\tdof_index[" << element->elementDataIndex() << "] = "
           << element->dofIndex() << endl;
    }
  }
#endif
#if (SPACEDIM==3)
  cout << "\n\tline dofs" << endl;
  for (Level1 *level=static_cast<Level1*>(tess->coarsestLevel());
  level;level=level->finerLevel()) {
    for (bool doit=tess->firstCodimension2Shape(level);
    doit && tess->presentCodimension2Level()==level;
    doit=tess->nextCodimension2ShapeInLevel()) {
      const Shape1 *element=level->getShape();
      cout << "\tdof_index[" << element->elementDataIndex() << "] = "
           << element->dofIndex() << endl;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofHandler::printDofs() const {
  cout << "\n\telement dofs:" << endl;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel()) {
    for (bool doit=tess->firstShape(level);
    doit && tess->presentLevel()==level;doit=tess->nextShapeInLevel()) {
      const SHAPE_CLASS *element=level->getShape();
      NumPtr<int> dof_indices;
      getDofIndices(element,level,dof_indices);
      cout << "\tdof_indices[" << element->elementDataIndex() << "] = ";
      for (int d=0;d<dof_indices.getNumber();d++) {
        cout << dof_indices[d] << " ";
      }
      cout << endl;
    }
  }
#if (SPACEDIM>=2)
  cout << "\n\tface dofs:" << endl;
  for (CODIMENSION1_LEVEL_CLASS *level=
  static_cast<CODIMENSION1_LEVEL_CLASS*>(tess->coarsestLevel());
  level;level=level->finerLevel()) {
    for (bool doit=tess->firstCodimension1Shape(level);
    doit && tess->presentCodimension1Level()==level;
    doit=tess->nextCodimension1ShapeInLevel()) {
      const CODIMENSION1_SHAPE_CLASS *element=level->getShape();
      NumPtr<int> dof_indices;
      getFaceDofIndices(element,level,dof_indices);
      cout << "\tdof_indices[" << element->elementDataIndex() << "] = ";
      for (int d=D_CHOOSE(0,0,selected_fe->dofsPerLine()*
      element->numberCodimension1Shapes());d<dof_indices.getNumber();d++) {
        cout << dof_indices[d] << " ";
      }
      cout << endl;
    }
  }
#else
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel()) {
    HcurlDofLevel *dof_level=getLevel(level);
    dof_level->printOn(cout);
  }
#endif
#if (SPACEDIM==3)
  cout << "\n\tedge dofs:" << endl;
  for (Level1 *level=static_cast<Level1*>(tess->coarsestLevel());
  level;level=level->finerLevel()) {
    for (bool doit=tess->firstCodimension2Shape(level);
    doit && tess->presentCodimension2Level()==level;
    doit=tess->nextCodimension2ShapeInLevel()) {
      const Shape1 *element=level->getShape();
      NumPtr<int> dof_indices;
      getLineDofIndices(element,level,dof_indices);
      cout << "\tdof_indices[" << element->elementDataIndex() << "] = ";
      for (int d=0;d<dof_indices.getNumber();d++) {
        cout << dof_indices[d] << " ";
      }
      cout << endl;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template void HcurlDofHandler::getFunctionValues<Tensor<1,SPACEDIM> >(
  const HcurlFEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,
  const Vector&,NumPtr<Tensor<1,SPACEDIM> >&) const;
template void HcurlDofHandler::getFunctionValues<NumPtr<Tensor<1,SPACEDIM> > >(
  const HcurlFEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,
  const Vector&,NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
#if (SPACEDIM>=2)
template void HcurlDofHandler::getFunctionValues<Tensor<1,SPACEDIM> >(
  const HcurlFEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<Tensor<1,SPACEDIM> >&) const;
template void HcurlDofHandler::getFunctionValues<NumPtr<Tensor<1,SPACEDIM> > >(
  const HcurlFEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
#endif

typedef Tensor<2,SPACEDIM> SCALAR_GRAD_TYPE;
typedef NumPtr<Tensor<2,SPACEDIM> > VECTOR_GRAD_TYPE;

#if (SPACEDIM>=2)
template class _Rb_tree<
  int,
  pair<int const, HcurlFEValuesData*>,
  _Select1st<pair<int const, HcurlFEValuesData*> >,
  less<int>,
  allocator<pair<int const, HcurlFEValuesData*> >
>;
#endif
#if (SPACEDIM==3)
template class _Rb_tree<
  Shape1 const*,
  pair<Shape1 const* const,
       set<pair<Shape3 const*, Level3 const*>,
           less<pair<Shape3 const*, Level3 const*> >,
           allocator<pair<Shape3 const*, Level3 const*> >
          >
      >,
  _Select1st<pair<Shape1 const* const,
                  set<pair<Shape3 const*, Level3 const*>,
                      less<pair<Shape3 const*, Level3 const*> >,
                      allocator<pair<Shape3 const*, Level3 const*> >
                     >
                 >
            >,
  less<Shape1 const*>,
  allocator<pair<Shape1 const* const,
                 set<pair<Shape3 const*, Level3 const*>,
                     less<pair<Shape3 const*, Level3 const*> >,
                     allocator<pair<Shape3 const*, Level3 const*> >
                    >
                >
           >
>;
#endif

#include "NumPtr.C"
INSTANTIATE_NUMPTR(HcurlDofLevel*);
