#ifndef HCURL_DOF_HANDLER_H
#define HCURL_DOF_HANDLER_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <map>
#include <set>
#include <BoundaryCondition.H>
#include <CompressedSparsityPattern.H>
#include <DDecl.H>
#include <DofHandler.H>
#include <HcurlDofLevel.H>
#include <HcurlFEValues.H>
#include <HcurlFiniteElement.H>
#include <NumPtr.H>
#include <SparseMatrix.H>
#include <Tessellation.H>

class HcurlDofHandler {
  private:
//  can't use map<const LEVEL_CLASS*,DOF_LEVEL_CLASS*> because
//    getLevel takes pointer into LEVEL_CLASS* data structure
    NumPtr<HcurlDofLevel*> levels;
    int used_dofs;
    int total_solution_dofs;

    HcurlDofHandler(const HcurlDofHandler&) {
      OBSOLETE("copy constructor disallowed");
    }
    HcurlDofHandler& operator=(const HcurlDofHandler&) {
      OBSOLETE("copy constructor disallowed");
    }
    void reserveSpace();
  protected:
    Tessellation *tess;
    HcurlFiniteElement *selected_fe;

//  deletes levels and vertex_dofs
//    called from HcurlDofHandler::reserveSpace and MGHcurlDofHandler::clear
    void clearSpace();
    HcurlDofLevel* getLevel(const LEVEL_CLASS *level) const {
      return levels[level->numberCoarserLevels()];
    }
#if (SPACEDIM>=2)
    HcurlDofLevel* getLevel(const CODIMENSION1_LEVEL_CLASS *level) const {
      return levels[level->numberCoarserLevels()];
    }
#endif
#if (SPACEDIM==3)
    HcurlDofLevel* getLevel(const Level1 *level) const {
      return levels[level->numberCoarserLevels()];
    }
#endif
//  if shape is refined, gets dofs for interpolation
//  called from getFunctionValues, getFunctionGrads and
//    getFunction2ndDerivatives    
    void getInterpolatedDofValues(const SHAPE_CLASS*,const LEVEL_CLASS*,
      const Vector&,Vector&) const;
    void getInterpolatedSolutionDofValues(const SHAPE_CLASS*,
      const LEVEL_CLASS*,const Vector&,Vector&) const;
    void getInterpolatedMultiplierDofValues(const SHAPE_CLASS*,
      const LEVEL_CLASS*,const Vector&,int,Vector&) const;
  public:
    HcurlDofHandler(Tessellation &t,HcurlFiniteElement &fe) : 
    levels(t.numberLevels()),tess(&t),selected_fe(&fe),used_dofs(0) {
      levels.initialize(0); 
    }
    virtual ~HcurlDofHandler() { clearSpace(); }

    int numberDofs() const { return used_dofs; } // both soln & multiplier
    int totalSolutionDofs() const { return total_solution_dofs; }
    HcurlFiniteElement& getFe() const {
      CHECK_POINTER(selected_fe);
      return *selected_fe;
    }
    Tessellation& getTessellation() const { return *tess; }

//  pulls local solution and multiplier dof's out of global dof's 
//  called from getInterpolatedDofValues and 
//    RefinementStudy::recursivelyIntegrateError
//  public for RefinementStudy
    void getDofValues(const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
      Vector&) const;
    void getSolutionDofValues(const SHAPE_CLASS*,const LEVEL_CLASS*,
      const Vector&,Vector&) const;
    void getMultiplierDofValues(const SHAPE_CLASS*,const LEVEL_CLASS*,
      const Vector&,int offset,Vector&) const;
//  stores first global dof in each active element of tessellation
//  called from user programs
    virtual void distributeDofs();
//  called from computeCuthillMcKeeRenumbering, Multigrid::Multigrid and
//    user programs
    int maxCouplingsBetweenDofs() const;
//  called from computeCuthillMcKeeRenumbering and user programs
    void makeSparsityPattern(SparsityPatternBase&) const;
    void makeTransientSparsityPatterns(SparsityPatternBase&,
      SparsityPatternBase&) const;
//  gets solution and multiplier dof indices
//  called from HcurlDofHandler::getDofValues,
//    HcurlDofHandler::makeSparsityPattern
//    HcurlDofHandler::interpolateBoundaryValues,
//    ConstraintMatrix::makeHangingNodeConstraints
//    MGTransferPrebuilt::copyToMG,MGTransferPrebuilt::copyFromMG,
//      MGTransferPrebuilt::copyFromMGAdd and user programs
//  useful in forming global stiffness matrix from cell matrices
    void getDofIndices(const SHAPE_CLASS*,const LEVEL_CLASS*,
      NumPtr<int>&) const;
    void getSolutionDofIndices(const SHAPE_CLASS*,const LEVEL_CLASS*,
      NumPtr<int>&) const;
    void getMultiplierDofIndices(const SHAPE_CLASS*,const LEVEL_CLASS*,
      NumPtr<int>&) const;
    void getFaceDofIndices(const SHAPE_CLASS*,int,const LEVEL_CLASS*,
      NumPtr<int>&) const;
#if (SPACEDIM>=2)
    void getFaceDofIndices(const CODIMENSION1_SHAPE_CLASS*,
      const CODIMENSION1_LEVEL_CLASS*,NumPtr<int>&) const;
#endif
#if (SPACEDIM==3)
    void getLineDofIndices(const Shape1*,const Level1*,NumPtr<int>&) const;
#endif
//  assembles linear system except for Dirichlet boundary values
//    called from user programs
    void assembleLinearSystem(const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,SparseMatrix&,Vector&) const;
    void assembleLinearSystemForInitialData(const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,SparseMatrix&,Vector&) const;
    void assembleTransientLinearSystem(const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,SparseMatrix&,SparseMatrix&,
      Vector&) const;
//  sets Dirichlet boundary values
//    called from user programs
    void interpolateBoundaryValues(const Mapping&,map<int,REAL>&) const;
    void interpolateBoundaryValues(const Mapping&,map<int,REAL>&,
      map<int,Quadrature<SPACEDIM-1> >&) const;
//  puts Dirichlet boundary values into linear system
//    called from user programs
    void applyBoundaryValues(const map<int,double>&,SparseMatrix&,
      Vector&,bool) const;
    void applyBoundaryValues(const map<int,double>&,Vector&) const;

#if (SPACEDIM>=2)
//  called from ConstraintMatrix::makeHangingNodeConstraints
    void dofsOnChildren(const CODIMENSION1_SHAPE_CLASS*,
      const CODIMENSION1_LEVEL_CLASS*,NumPtr<int>&) const;
#endif
//  T = Tensor<1,SPACEDIM> or NumPtr<Tensor<1,SPACEDIM> >
//  called from HcurlDofHandler::integrateDifference,
//    HcurlDofHandler::integrateDifferenceOnBoundary,
//    HcurlDofHandler::integrateDifferenceOnMappedBoundary,
//    DataOut::buildSomePatches, 
//    Patch::plotClipPlane,
//    RefinementStudy::RefinementStudy,
//    RefinementStudy::recursivelyIntegrateError
    template<class T> void getFunctionValues(const HcurlFEValues &fe_values,
    const SHAPE_CLASS *e,const LEVEL_CLASS *level,
    const Vector &fe_function,NumPtr<T> &values) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_values.getFunctionValues(dof_values,e,values);
    }
#if (SPACEDIM>=2)
    template<class T> void getFunctionValues(
    const HcurlFEFaceValues &fe_face_values,const SHAPE_CLASS *e,int f,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &values) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_face_values.getFunctionValues(dof_values,e,f,values);
    }
#endif
//  T = Tensor<1,SPACEDIM> or NumPtr<Tensor<1,SPACEDIM> >
    template<class T> void getSolutionFunctionValues(
    const HcurlFEValues &fe_values,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level,const Vector &fe_function,NumPtr<T> &values)
    const {
      CHECK_SAME(fe_function.size(),totalSolutionDofs());
      Vector dof_values;
      getInterpolatedSolutionDofValues(e,level,fe_function,dof_values);
      fe_values.getFunctionValues(dof_values,e,values);
    }
//  T = REAL or Vector
    template<class T> void getMultiplierFunctionValues(
    const HcurlFEValues &fe_values,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level,const Vector &fe_function,int offset,
    NumPtr<T> &values) const {
      CHECK_SAME(fe_function.size(),numberDofs()-offset);
      Vector dof_values;
      getInterpolatedMultiplierDofValues(e,level,fe_function,offset,
        dof_values);
      fe_values.getMultiplierFunctionValues(dof_values,e,values);
    }
    virtual void printOn(ostream&) const;
    virtual void printDofs() const;
    virtual void printElementDofs() const;
};

#endif
