// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <HcurlFEValues.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
HcurlFEValuesData::HcurlFEValuesData(int nv,int nqp,
const HcurlFiniteElement &fe,UPDATE_FLAGS uf) : update_flags(uf),
number_quadrature_points(nqp),number_vertices(nv),
shape_values_ok(false),shape_gradients_ok(false),
shape_2nd_derivatives_ok(false),shape_curl_ok(false),
multiplier_shape_values_ok(false),multiplier_shape_gradients_ok(false),
multiplier_shape_2nd_derivatives_ok(false),JxW_values_ok(false),
quadrature_points_ok(false),finite_element(&fe) {
  int n_nonzero_shape_components=
    finite_element->nNonzeroShapeComponents(number_vertices);
  int n_nonzero_multiplier_shape_components=
    finite_element->nNonzeroMultiplierShapeComponents(number_vertices);
  if (update_flags & UPDATE_VALUES) {
    shape_values.reinit(n_nonzero_shape_components,
                        number_quadrature_points);
    multiplier_shape_values.reinit(n_nonzero_multiplier_shape_components,
                                   number_quadrature_points);
#ifdef INDEF
    Tensor<1,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    shape_values.fillConstant(bogus_tensor);
# if (SPACEDIM<=2)
    multiplier_shape_values.fillConstant(HUGE_VAL);
# else
    multiplier_shape_values.fillConstant(bogus_tensor);
# endif
#endif
  }
  if (update_flags & UPDATE_GRADIENTS) {
    shape_gradients.reinit(n_nonzero_shape_components,
                           number_quadrature_points);
#ifdef INDEF
    Tensor<2,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    shape_gradients.fillConstant(bogus_tensor);
#endif
  }
  if (update_flags & UPDATE_CURL) {
    shape_curl.reinit(n_nonzero_shape_components,
                      number_quadrature_points);
#ifdef INDEF
# if (SPACEDIM<=2)
    shape_curl.fillConstant(HUGE_VAL);
# else
    Tensor<1,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    shape_curl.fillConstant(bogus_tensor);
# endif
#endif
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
    shape_2nd_derivatives.reinit(n_nonzero_shape_components,
                                 number_quadrature_points);
#ifdef INDEF
    Tensor<3,SPACEDIM> bogus_2tensor;
    bogus_2tensor.initialize(HUGE_VAL);
    shape_2nd_derivatives.fillConstant(bogus_2tensor);
#endif
  }
  if (update_flags & UPDATE_Q_POINTS) {
    quadrature_points.cleanup();
    quadrature_points.allocate(number_quadrature_points);
#ifdef INDEF
    Point<SPACEDIM> bogus_point(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
    quadrature_points.initialize(bogus_point);
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFEValuesData::initializeJxWValues(int nqp,
UPDATE_FLAGS update_flags) {
  if (update_flags & UPDATE_JxW_VALUES) {
#ifdef DEBUG
//  cout << "\tnqp = " << nqp << endl;
#endif
    JxW_values.cleanup();
    JxW_values.allocate(nqp);
#ifdef INDEF
    JxW_values.initialize(HUGE_VAL);
#endif
    JxW_values_ok=false;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFEValuesData::mapShapes(MappingDataBase *mapping_data,
HcurlFiniteElementDataBase *fe_data,const SHAPE_CLASS *e,
const Tessellation &tess,const LEVEL_CLASS *level) {
  if (fe_data==0) return;
  shape_values_ok=false;
  shape_gradients_ok=false;
  shape_curl_ok=false;
  shape_2nd_derivatives_ok=false;
  multiplier_shape_values_ok=false;
  JxW_values_ok=false;
  quadrature_points_ok=false;
#ifdef INDEF
  JxW_values.initialize(HUGE_VAL);
  {
    Tensor<1,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    shape_values.fillConstant(bogus_tensor);
# if (SPACEDIM<=2)
    shape_curl.fillConstant(HUGE_VAL);
    multiplier_shape_values.fillConstant(HUGE_VAL);
# else
    shape_curl.fillConstant(bogus_tensor);
    multiplier_shape_values.fillConstant(bogus_tensor);
# endif
  }
  {
    Tensor<2,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    shape_gradients.fillConstant(bogus_tensor);
  }
  {
    Tensor<3,SPACEDIM> bogus_2tensor;
    bogus_2tensor.initialize(HUGE_VAL);
    shape_2nd_derivatives.fillConstant(bogus_2tensor);
  }
#endif
  mapping_data->computeFEValues(e,tess,level,quadrature_points,
    JxW_values,quadrature_points_ok,JxW_values_ok);
  HcurlFiniteElementDataBase *fedb=
    finite_element->fixShapeFunctionsFor(e,fe_data);
  if (update_flags & UPDATE_VALUES) {
    int offset=0;
#if (SPACEDIM==2)
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransform(shape_values(s+offset,q),q,
            fedb->shapeValue(s+offset,q));
        }
      }
      offset+=dpf;
    }
#elif (SPACEDIM==3)
    for (int l=0;l<e->numberCodimension2Shapes();l++) {
      int dpl=finite_element->dofsPerLine();
      for (int s=0;s<dpl;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransform(shape_values(s+offset,q),q,
            fedb->shapeValue(s+offset,q));
        }
      }
      offset+=dpl;
    }
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      int dpf2=dpf/2;
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransform(shape_values(s+offset,q),q,
            fedb->shapeValue(s+offset,q));
        }
      }
      offset+=dpf;
    }
#endif
    for (int s=offset;s<shape_values.size(0);s++) {
      for (int q=0;q<shape_values.size(1);q++) {
        mapping_data->hcurlTransform(shape_values(s,q),q,
          fedb->shapeValue(s,q));
      }
    }
    shape_values_ok=true;
    for (int s=0;s<multiplier_shape_values.size(0);s++) {
      for (int q=0;q<multiplier_shape_values.size(1);q++) {
        multiplier_shape_values(s,q)=fedb->multiplierShapeValue(s,q);
      }
    }
    multiplier_shape_values_ok=true;
  }
  if (update_flags & UPDATE_CURL) {
    int offset=0;
#if (SPACEDIM==2)
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          const Tensor<2,SPACEDIM> &sg=
            fedb->shapeDerivative(s+offset,q);
          REAL curl=sg[1][0]-sg[0][1];
          mapping_data->hcurlTransformCurl(shape_curl(s+offset,q),q,curl);
        }
      }
      offset+=dpf;
    }
#elif (SPACEDIM==3)
    for (int l=0;l<e->numberCodimension2Shapes();l++) {
      int dpl=finite_element->dofsPerLine();
      for (int s=0;s<dpl;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          const Tensor<2,SPACEDIM> &sg=
            fedb->shapeDerivative(s+offset,q);
          Tensor<1,3> curl;
          curl[0]=sg[2][1]-sg[1][2];
          curl[1]=sg[0][2]-sg[2][0];
          curl[2]=sg[1][0]-sg[0][1];
          mapping_data->hcurlTransformCurl(shape_curl(s+offset,q),q,curl);
        }
      }
      offset+=dpl;
    }
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          const Tensor<2,SPACEDIM> &sg=
            fedb->shapeDerivative(s+offset,q);
          Tensor<1,3> curl;
          curl[0]=sg[2][1]-sg[1][2];
          curl[1]=sg[0][2]-sg[2][0];
          curl[2]=sg[1][0]-sg[0][1];
          mapping_data->hcurlTransformCurl(shape_curl(s+offset,q),q,curl);
        }
      }
      offset+=dpf;
    }
#endif
    for (int s=offset;s<shape_curl.size(0);s++) {
      for (int q=0;q<shape_curl.size(1);q++) {
        const Tensor<2,SPACEDIM> &sg=fedb->shapeDerivative(s,q);
#if (SPACEDIM==1)
//      similar to hcurlTransformDerivative:
        REAL value=fedb->shapeValue(s,q)[0];
        REAL curl=sg[0][0];
        mapping_data->hcurlTransformCurl(shape_curl(s,q),q,curl,value);
#elif (SPACEDIM==2)
        REAL curl=sg[1][0]-sg[0][1];
        mapping_data->hcurlTransformCurl(shape_curl(s,q),q,curl);
#else
        Tensor<1,3> curl;
        curl[0]=sg[2][1]-sg[1][2];
        curl[1]=sg[0][2]-sg[2][0];
        curl[2]=sg[1][0]-sg[0][1];
        mapping_data->hcurlTransformCurl(shape_curl(s,q),q,curl);
#endif
      }
    }
    shape_curl_ok=true;
  }
  if (update_flags & UPDATE_GRADIENTS) {
    int offset=0;
#if (SPACEDIM==2)
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransformDerivative(
            shape_gradients(s+offset,q),q,fedb->shapeValue(s+offset,q),
            fedb->shapeDerivative(s+offset,q));
        }
      }
      offset+=dpf;
    }
#elif (SPACEDIM==3)
    for (int l=0;l<e->numberCodimension2Shapes();l++) {
      int dpl=finite_element->dofsPerLine();
      for (int s=0;s<dpl;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransformDerivative(
            shape_gradients(s+offset,q),q,fedb->shapeValue(s+offset,q),
            fedb->shapeDerivative(s+offset,q));
        }
      }
      offset+=dpl;
    }
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransformDerivative(
            shape_gradients(s+offset,q),q,fedb->shapeValue(s+offset,q),
            fedb->shapeDerivative(s+offset,q));
        }
      }
      offset+=dpf;
    }
#endif
    for (int s=offset;s<shape_gradients.size(0);s++) {
      for (int q=0;q<shape_gradients.size(1);q++) {
        mapping_data->hcurlTransformDerivative(shape_gradients(s,q),q,
          fedb->shapeValue(s,q),fedb->shapeDerivative(s,q));
      }
    }
    shape_gradients_ok=true;
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
//  requires second derivatives of FE shapes and mapping transformation
    OBSOLETE("not programmed");
    shape_2nd_derivatives_ok=true;
  }
  if (fedb!=fe_data) delete fedb;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlFEValuesData::mapFaceShapes(MappingDataBase *mapping_data,
HcurlFiniteElementDataBase *fe_data,const SHAPE_CLASS *e,
int face_no,const Tessellation &tess,const LEVEL_CLASS *level,
const Quadrature<SPACEDIM-1> &fq,
NumPtr<Tensor<1,SPACEDIM> > &boundary_forms,
NumPtr<Point<SPACEDIM> > &normal_vectors,
int q_offset,bool &boundary_forms_ok,bool &normal_vectors_ok) {
  CHECK_TEST(face_no<e->numberCodimension1Shapes());
  int nq=fq.numberPoints();
  bool no_jxw=JxW_values.getNumber()==0;
  NumPtr<REAL> face_jxw_values(JxW_values,(no_jxw ? 0 : q_offset),
    (no_jxw ? 0 : nq));

# ifdef INDEF
  face_jxw_values.initialize(HUGE_VAL);
  {
    Tensor<1,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    boundary_forms.initialize(bogus_tensor);
  }
  {
    Point<SPACEDIM> bogus_point(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
    normal_vectors.initialize(bogus_point);
  }
# endif
  bool no_points=quadrature_points.getNumber()==0;
  NumPtr<Point<SPACEDIM> > qp(quadrature_points,(no_points ? 0 : q_offset),
    (no_points ? 0 : nq));
  bool is_subface=false;
  mapping_data->computeFEFaceValues(e,tess,level,face_no,is_subface,
    nq,q_offset,fq.weights(),qp,face_jxw_values,boundary_forms,
    normal_vectors,quadrature_points_ok,JxW_values_ok,boundary_forms_ok,
    normal_vectors_ok);
  HcurlFiniteElementDataBase *fedb=
    finite_element->fixShapeFunctionsFor(e,fe_data);
  if (update_flags & UPDATE_VALUES) {
    int s_offset=0;
#if (SPACEDIM==2)
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      bool tangent_points_backward=level->lineTangentPointsBackward(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<nq;q++) {
          mapping_data->hcurlTransform(
            shape_values(s+s_offset,q+q_offset),q,
            fedb->shapeValue(s+s_offset,q+q_offset));
          if (tangent_points_backward) {
            shape_values(s+s_offset,q+q_offset)*=-1.;
          }
        }
      }
      s_offset+=dpf;
    }
#else
    for (int l=0;l<e->numberCodimension2Shapes();l++) {
      int dpl=finite_element->dofsPerLine();
      bool tangent_points_backward=level->lineTangentPointsBackward(e,l);
      for (int s=0;s<dpl;s++) {
        for (int q=0;q<nq;q++) {
          mapping_data->hcurlTransform(
            shape_values(s+s_offset,q+q_offset),q,
            fedb->shapeValue(s+s_offset,q+q_offset));
          if (tangent_points_backward) {
            shape_values(s+s_offset,q+q_offset)*=-1.;
          }
        }
      }
      s_offset+=dpl;
    }
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      bool normal_points_inward=level->faceNormalPointsInward(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<nq;q++) {
          mapping_data->hcurlTransform(
            shape_values(s+s_offset,q+q_offset),q,
            fedb->shapeValue(s+s_offset,q+q_offset));
          if (normal_points_inward) {
            shape_values(s+s_offset,q+q_offset)*=-1.;
          }
        }
      }
      s_offset+=dpf;
    }
#endif
    for (int s=s_offset;s<shape_values.size(0);s++) {
      for (int q=0;q<nq;q++) {
        mapping_data->hcurlTransform(shape_values(s,q+q_offset),
          q,fedb->shapeValue(s,q+q_offset));
      }
    }
    shape_values_ok=true;
  }
  if (update_flags & UPDATE_GRADIENTS) {
    OBSOLETE("not programmed");
    shape_gradients_ok=true;
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
//  requires second derivatives of FE shapes and mapping transformation
    OBSOLETE("not programmed");
    shape_2nd_derivatives_ok=true;
  }
  if (fedb!=fe_data) delete fedb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlFEValuesData::mapSubFaceShapes(MappingDataBase *mapping_data,
HcurlFiniteElementDataBase *fe_data,const SHAPE_CLASS *e,
int face_no,const Tessellation &tess,const LEVEL_CLASS *level,
const Quadrature<SPACEDIM-1> &fq,
NumPtr<Tensor<1,SPACEDIM> > &boundary_forms,
NumPtr<Point<SPACEDIM> > &normal_vectors,
int fq_offset,int offset,bool &boundary_forms_ok,
bool &normal_vectors_ok) {
  CHECK_TEST(face_no<e->numberCodimension1Shapes());
  int nq=fq.numberPoints();
  bool no_jxw=JxW_values.getNumber()==0;
  NumPtr<REAL> face_jxw_values(JxW_values,(no_jxw ? 0 : fq_offset),
    (no_jxw ? 0 : nq));

# ifdef INDEF
  face_jxw_values.initialize(HUGE_VAL);
  {
    Tensor<1,SPACEDIM> bogus_tensor;
    bogus_tensor.initialize(HUGE_VAL);
    boundary_forms.initialize(bogus_tensor);
  }
  {
    Point<SPACEDIM> bogus_point(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
    normal_vectors.initialize(bogus_point);
  }
# endif
  bool no_points=quadrature_points.getNumber()==0;
  NumPtr<Point<SPACEDIM> > qp(quadrature_points,
    (no_points ? 0 : fq_offset),(no_points ? 0 : nq));
//mapping_data->fillFESubfaceValues(offset,e,tess,level,face_no,
//  fq,qp,face_jxw_values,boundary_forms,normal_vectors);
  bool is_subface=true;
  mapping_data->computeFEFaceValues(e,tess,level,face_no,is_subface,
    fq.numberPoints(),offset,fq.weights(),qp,face_jxw_values,
    boundary_forms,normal_vectors,quadrature_points_ok,JxW_values_ok,
    boundary_forms_ok,normal_vectors_ok);
  HcurlFiniteElementDataBase *fedb=fe_data;
  if (update_flags & UPDATE_VALUES) {
    int offset=0;
#if (SPACEDIM==2)
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      bool tangent_points_backward=level->lineTangentPointsBackward(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransform(shape_values(s+offset,q),q,
            fedb->shapeValue(s+offset,q));
          if (tangent_points_backward) shape_values(s+offset,q)*=-1.;
        }
      }
      offset+=dpf;
    }
#else
    for (int l=0;l<e->numberCodimension2Shapes();l++) {
      int dpl=finite_element->dofsPerLine();
      bool tangent_points_backward=level->lineTangentPointsBackward(e,l);
      for (int s=0;s<dpl;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransform(shape_values(s+offset,q),q,
            fedb->shapeValue(s+offset,q));
          if (tangent_points_backward) shape_values(s+offset,q)*=-1.;
        }
      }
      offset+=dpl;
    }
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int dpf=finite_element->dofsPerFace(e,f);
      bool normal_points_inward=level->faceNormalPointsInward(e,f);
      for (int s=0;s<dpf;s++) {
        for (int q=0;q<shape_values.size(1);q++) {
          mapping_data->hcurlTransform(shape_values(s+offset,q),q,
            fedb->shapeValue(s+offset,q));
          if (normal_points_inward) shape_values(s+offset,q)*=-1.;
        }
      }
      offset+=dpf;
    }
#endif
    for (int s=offset;s<shape_values.size(0);s++) {
      for (int q=0;q<shape_values.size(1);q++) {
        mapping_data->hcurlTransform(shape_values(s,q),q,
          fedb->shapeValue(s,q));
      }
    }
    shape_values_ok=true;
  }
  if (update_flags & UPDATE_GRADIENTS) {
    shape_gradients_ok=true;
  }
  if (update_flags & UPDATE_SECOND_DERIVATIVES) {
//  requires second derivatives of FE shapes and mapping transformation
    OBSOLETE("not programmed");
    shape_2nd_derivatives_ok=true;
  }
  if (fedb!=fe_data) delete fedb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFEValuesData::printOn(ostream &os) const {
  os << "\tHcurlFEValuesData: this = " << this << endl;
  os << "\t\tshape_values is " << shape_values.size(0) << "x" 
     << shape_values.size(1) << endl;
  if (shape_values_ok) {
    for (int i=0;i<shape_values.size(0);i++) {
      for (int j=0;j<shape_values.size(1);j++) {
        os << "\t\t\tshape_value[" << i << "," << j << "] = "
           << shape_values.operator()(i,j) << endl;
      }
    }
  }
//os << "\t\tshape_gradients is " << shape_gradients.size(0) << "x" 
//   << shape_gradients.size(1) << endl;
//for (int i=0;i<shape_gradients.size(0);i++) {
//  for (int j=0;j<shape_gradients.size(1);j++) {
//    os << "\t\t\tshape_gradients[" << i << "," << j << "] = "
//       << shape_gradients.operator()(i,j) << endl;
//  }
//}
//os << "\t\tshape_curl is " << shape_curl.size(0) << "x" 
//   << shape_curl.size(1) << endl;
//for (int i=0;i<shape_curl.size(0);i++) {
//  for (int j=0;j<shape_curl.size(1);j++) {
//    os << "\t\t\tshape_curl[" << i << "," << j << "] = "
//       << shape_curl.operator()(i,j) << endl;
//  }
//}
//os << "\t\tshape_2nd_derivatives is "
//   << shape_2nd_derivatives.size(0) << "x" 
//   << shape_2nd_derivatives.size(1) << endl;
//os << "\t\tJxW_values has "
//   << JxW_values.getNumber() << " entries" << endl;
//for (int i=0;i<JxW_values.getNumber();i++) {
//  os << "\t\t\tJxW_values[" << i << "] = " << JxW_values[i] << endl;
//}
//os << "\t\tmultiplier_shape_values is "
//   << multiplier_shape_values.size(0) << "x" 
//   << multiplier_shape_values.size(1) << endl;
//for (int i=0;i<multiplier_shape_values.size(0);i++) {
//  for (int j=0;j<multiplier_shape_values.size(1);j++) {
//    os << "\t\t\tmultiplier_shape_value[" << i << "," << j << "] = "
//       << multiplier_shape_values.operator()(i,j) << endl;
//  }
//}
//os << "\t\tJxW_values has "
//   << JxW_values.getNumber() << " entries" << endl;
//for (int i=0;i<JxW_values.getNumber();i++) {
//  os << "\t\t\tJxW_values[" << i << "] = " << JxW_values[i] << endl;
//}
//os << "\t\tquadrature_points has "
//   << quadrature_points.getNumber() << " entries" << endl;
//if (quadrature_points_ok) {
//  for (int i=0;i<quadrature_points.getNumber();i++) {
//    os << "\t\t\tquadrature_point[" << i << "] = " 
//       << quadrature_points[i] << endl;
//  }
//}
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
HcurlFEValuesBase::~HcurlFEValuesBase() {
  for (map<int,HcurlFEValuesData*>::iterator
  fevd_iterator=fe_values_data_map.begin();
  fevd_iterator!=fe_values_data_map.end();fevd_iterator++) {
    delete fevd_iterator->second; fevd_iterator->second=0;
  }
  for (map<int,HcurlFiniteElementDataBase*>::iterator
  fedm_iterator=fe_data_map.begin();
  fedm_iterator!=fe_data_map.end();fedm_iterator++) {
    delete fedm_iterator->second; fedm_iterator->second=0;
  }
  for (map<int,MappingDataBase*>::iterator
  mdm_iterator=mapping_data_map.begin();
  mdm_iterator!=mapping_data_map.end();mdm_iterator++) {
    delete mdm_iterator->second; mdm_iterator->second=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFEValuesBase::printFEData(ostream &os) const {
  os << "\tfe_data_map = " << &fe_data_map << endl;
  for (map<int,HcurlFiniteElementDataBase*>::const_iterator
  fedm_iterator=fe_data_map.begin();fedm_iterator!=fe_data_map.end();
  fedm_iterator++) {
    os << "\t\tfe_data_map[" << fedm_iterator->first << "] = "
       << fedm_iterator->second << endl;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFEValuesBase::printOn(ostream &os) const {
  os << "\tHcurlFEValuesBase :" << endl;
//cout << "\tmapping_data_map args = ";
//for (map<int,MappingDataBase*>::const_iterator md_iterator=
//mapping_data_map.begin();md_iterator!=mapping_data_map.end();
//md_iterator++) {
//  cout << md_iterator->first << " ";
//}
//cout << endl;
  for (map<int,MappingDataBase*>::const_iterator md_iterator=
  mapping_data_map.begin();md_iterator!=mapping_data_map.end();
  md_iterator++) {
    os << "\t\tmapping_data_map[" << md_iterator->first << "] = "
       << md_iterator->second << endl;
    md_iterator->second->printOn(os);
  }
//for (map<int,HcurlFiniteElementDataBase*>::const_iterator fed_iterator=
//fe_data_map.begin();fed_iterator!=fe_data_map.end();fed_iterator++) {
//  os << "\t\tfe_data_map[" << fed_iterator->first << "] = "
//     << fed_iterator->second << endl;
//  fed_iterator->second->printOn(os);
//}
  for (map<int,HcurlFEValuesData*>::const_iterator fevd_iterator=
  fe_values_data_map.begin();fevd_iterator!=fe_values_data_map.end();
  fevd_iterator++) {
    os << "\t\tfe_values_data_map[" << fevd_iterator->first << "] = "
       << fevd_iterator->second << endl;
    fevd_iterator->second->printOn(os);
  }
//for (map<int,int>::const_iterator nqp_iterator=
//n_quadrature_points.begin();nqp_iterator!=n_quadrature_points.end();
//nqp_iterator++) {
//  os << "\t\tn_quadrature_points[" << nqp_iterator->first << "] = "
//     << nqp_iterator->second << endl;
//}
//for (map<int,int>::const_iterator dpc_iterator=
//dofs_per_cell.begin();dpc_iterator!=dofs_per_cell.end();
//dpc_iterator++) {
//  os << "\t\tdofs_per_cell[" << dpc_iterator->first << "] = "
//     << dpc_iterator->second << endl;
//}
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this is called outside a loop over grid elements
//  computes mapping shape functions at canonical quadrature points
//    and stores in mapping_data_map[nv]
//  computes finite element shape functions at canonical quadrature pts
//    and stores in fe_data_map[nv]
//  sets shape_function_to_row_table, allocates shape_values, 
//  shape_gradients, quadrature_points, JxW_values
//    these are stored in FEValuesData
HcurlFEValues::HcurlFEValues(const Mapping &m,HcurlFiniteElement &fe_in,
const map<int,Quadrature<SPACEDIM> > &q,UPDATE_FLAGS flags) :
HcurlFEValuesBase(flags,m,fe_in) {
  if (update_flags & UPDATE_CURL) {
    update_flags |= UPDATE_GRADIENTS;
    update_flags |= UPDATE_MAPPING_JACOBIAN_INVERSE;
    update_flags |= UPDATE_MAPPING_SECOND_DERIVATIVE;
  }
  for (map<int,Quadrature<SPACEDIM> >::const_iterator
  q_iterator=q.begin();q_iterator!=q.end();q_iterator++) {
    int nv=q_iterator->first;
    const Quadrature<SPACEDIM> &q_nv=q_iterator->second;
    insertNQuadraturePoint(nv,q_nv.numberPoints());
    insertDofsPerCell(nv,fe_in.dofsPerCell(nv));
    quadrature[nv]=q_nv;

    mapping_data_map[nv]=mapping->getMappingData(nv,update_flags,&q_nv);
    fe_data_map[nv]=fe->getFEData(nv,update_flags,&q_nv);
    int np=q_nv.numberPoints();
    HcurlFEValuesData *fevd=fe_values_data_map[nv]=
      OPERATOR_NEW HcurlFEValuesData(nv,np,*fe,update_flags);
    fevd->initializeJxWValues(np,update_flags);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFEValues::printOn(ostream &os) const {
  os << "\tHcurlFEValues:" << endl;
//for (map<int,Quadrature<SPACEDIM> >::const_iterator 
//q_iterator=quadrature.begin();q_iterator!=quadrature.end();
//q_iterator++) {
//  os << "\t\tquadrature[" << q_iterator->first << "] has "
//     << q_iterator->second.numberPoints() << " points" << endl;
//}
  HcurlFEValuesBase::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
int HcurlFEFaceValues::faceOffset(const SHAPE_CLASS *e,int face_no) const
{
  int offset=0;
  for (int f=0;f<face_no;f++) {
    int vpf=e->numberVerticesOnFace(f);
    offset+=getFaceQuadrature(vpf).numberPoints();
  }
  return offset;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlFEFaceValues::initializeBoundaryFormsAndNormals(int nv,
const Quadrature<SPACEDIM-1> &qf) {
  int nqp=qf.numberPoints();
  if (update_flags & UPDATE_BOUNDARY_FORMS) {
    CHECK_TEST(boundary_forms.find(nv)==boundary_forms.end());
    NumPtr<Tensor<1,SPACEDIM> > &bf=boundary_forms[nv];
    bf.cleanup();
    bf.allocate(nqp);
  }
  if (update_flags & UPDATE_NORMAL_VECTORS) {
    CHECK_TEST(normal_vectors.find(nv)==normal_vectors.end());
    NumPtr<Point<SPACEDIM> > &normals=normal_vectors[nv];
    normals.cleanup();
    normals.allocate(nqp);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
//for each possible shape in this number of dimensions,
//  generates a Quadrature on all faces
//  compute unit tangents, shapes, etc at quadrature points
HcurlFEFaceValues::HcurlFEFaceValues(const Mapping &m,
HcurlFiniteElement &fe_in,map<int,Quadrature<SPACEDIM-1> > &q,
UPDATE_FLAGS flags_in) : HcurlFEValuesBase(flags_in,m,fe_in),
boundary_forms_ok(false),normal_vectors_ok(false) {
//q contains quadrature points for each possible face shape
  for (map<int,Quadrature<SPACEDIM-1> >::const_iterator
  q_iterator=q.begin();q_iterator!=q.end();q_iterator++) {
    int nv=q_iterator->first;
    const Quadrature<SPACEDIM-1> &q_nv=q_iterator->second;
    face_quadrature[nv]=q_nv;
  }
  update_flags |= UPDATE_BOUNDARY_FORMS;
  if (update_flags & UPDATE_CURL) {
    update_flags |= UPDATE_MAPPING_JACOBIAN_INVERSE;
  }
  const map<int,int> &dpc_map=fe->dofsPerCellMap();
  for (map<int,int>::const_iterator dpc_iterator=
  dpc_map.begin();dpc_iterator!=dpc_map.end();dpc_iterator++) {
    int nv=dpc_iterator->first;
//original DEAL code defined mapping_data and fe_data for projected
//  points around all faces, then called FEValuesData::initialize
//  for a single face
//this worked because all elements were Quadrilaterals/Hexahegons
//  and all faces were Intervals/Quadrilaterals with the same number
//  of shape functions on all faces
//in this code, we have the same number of quadrature points on all
//  faces of the same shape, but not necessarily the same number of
//  element shape functions to be evaluated on a face of a given shape
//instead, we call FEValuesData::initialize for the collection of
//  projected quadrature points around all faces
//this necessitates changes in FiniteElement::fillFEFaceValues and
//  Mapping::fillFEFaceValues
    int nfp=0;
    for (int f=0;f<facesPerShape(nv);f++) {
      int vpf=numberVerticesOnFace(nv,f);
      map<int,Quadrature<SPACEDIM-1> >::const_iterator
        q_iterator=q.find(vpf);
      if (q_iterator==q.end()) {
//      main program did not expect a face with vpf vertices, but
//        FiniteElement constructor had to allow for all Shapes
//      so, we construct a bogus Quadrature for vpf: point=0, weight=inf
        NumPtr<Point<SPACEDIM-1> > points(1);
        for (int d=0;d<SPACEDIM-1;d++) points[0][d]=0.;
        NumPtr<REAL> weights(1);
        weights[0]=HUGE_VAL;
        q[vpf]=Quadrature<SPACEDIM-1>(points,weights);
        q_iterator=q.find(vpf);
      }
      nfp+=q_iterator->second.numberPoints();
    }
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<facesPerShape(nv);f++) {
      int vpf=numberVerticesOnFace(nv,f);
      map<int,Quadrature<SPACEDIM-1> >::const_iterator
        q_iterator=q.find(vpf);
      int nqp=q_iterator->second.numberPoints();
      projectToFace(nv,f,q_iterator->second.points(),q_points_face);
      weights_face.copyFrom(q_iterator->second.weights());
      q_points_face+=nqp;
      weights_face+=nqp;
    }
    Quadrature<SPACEDIM> *qf=all_face_quadrature[nv]=
      OPERATOR_NEW Quadrature<SPACEDIM>(q_points,weights);
    MappingDataBase *mdb=mapping_data_map[nv]=
      mapping->getMappingData(nv,flags_in,qf);
    HcurlFiniteElementDataBase *fed=fe_data_map[nv]=
      fe->getFEData(nv,flags_in,qf);
    HcurlFEValuesData *fe_values_data=fe_values_data_map[nv]=
      OPERATOR_NEW HcurlFEValuesData(nv,qf->numberPoints(),*fe,
                                    flags_in^UPDATE_JxW_VALUES);
    insertNQuadraturePoint(nv,nfp);
    insertDofsPerCell(nv,fe_in.dofsPerCell(nv));
    fe_values_data->initializeJxWValues(nfp,flags_in);
  }
  for (map<int,Quadrature<SPACEDIM-1> >::const_iterator
  q_iterator=q.begin();q_iterator!=q.end();q_iterator++) {
    int nv=q_iterator->first;
    initializeBoundaryFormsAndNormals(nv,q_iterator->second);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
HcurlFEFaceValues::~HcurlFEFaceValues() {
  for (map<int,Quadrature<SPACEDIM>*>::iterator afq_iterator=
  all_face_quadrature.begin();afq_iterator!=all_face_quadrature.end();
  afq_iterator++) {
    delete afq_iterator->second; afq_iterator->second=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlFEFaceValues::mapShapes(const Tessellation &tess,
const SHAPE_CLASS *e,const LEVEL_CLASS *level,int f) {
  int nv=e->numberVerticesOnFace(f);
  NumPtr<Point<SPACEDIM> > empty_nvec;
  NumPtr<Tensor<1,SPACEDIM> > empty_bf;

  map<int,NumPtr<Tensor<1,SPACEDIM> > >::iterator
    bf_iterator=boundary_forms.find(nv);
  map<int,NumPtr<Point<SPACEDIM> > >::iterator
    nv_iterator=normal_vectors.find(nv);

  const Quadrature<SPACEDIM-1> &fq=getFaceQuadrature(e,f);
  getFEValuesData(e)->mapFaceShapes(mappingData(e),
    getFEData(e),e,f,tess,level,fq,
    (update_flags & UPDATE_BOUNDARY_FORMS ? bf_iterator->second : 
                    empty_bf),
    (update_flags & UPDATE_NORMAL_VECTORS ? nv_iterator->second : 
                    empty_nvec),
    faceOffset(e,f),boundary_forms_ok,normal_vectors_ok);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlFEFaceValues::getFunctionValues(const Vector &dof_values,
const SHAPE_CLASS *e,int face_no,NumPtr<Tensor<1,SPACEDIM> > &values)
const {
  CHECK_TEST(update_flags & UPDATE_VALUES); 
  int offset=faceOffset(e,face_no);
  int nqp=getFaceQuadrature(e,face_no).numberPoints();
  fe->sumShapeValues(e,*getShapeValues(e),nqp,faceOffset(e,face_no),
    dof_values,values);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlFEFaceValues::getFunctionValues(const Vector &dof_values,
const SHAPE_CLASS *e,int face_no,
NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &values) const {
  CHECK_TEST(update_flags & UPDATE_VALUES); 
  int nqp=getFaceQuadrature(e,face_no).numberPoints();
  int offset=faceOffset(e,face_no);
  const Array<2,Tensor<1,SPACEDIM> > &sv=*getShapeValues(e);
  fe->sumShapeValues(e,sv,nqp,offset,dof_values,values);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HcurlFEFaceValues::printOn(ostream &os) const {
  os << "\tHcurlFEFaceValues:" << endl;
  for (map<int,Quadrature<SPACEDIM-1> >::const_iterator
  fq_iterator=face_quadrature.begin();
  fq_iterator!=face_quadrature.end();fq_iterator++) {
    os << "\t\tface_quadrature[" << fq_iterator->first << "] has "
       << fq_iterator->second.numberPoints() << " points" << endl;
    fq_iterator->second.printOn(os);
  }
//if (normal_vectors_ok) {
//  for (map<int,NumPtr<Point<SPACEDIM> > >::const_iterator
//  nv_iterator=normal_vectors.begin();
//  nv_iterator!=normal_vectors.end();nv_iterator++) {
//    const NumPtr<Point<SPACEDIM> > &nv=nv_iterator->second;
//    os << "\t\tnormal_vectors[" << nv_iterator->first << "] has "
//       << nv.getNumber() << " entries" << endl;
//    for (int i=0;i<nv.getNumber();i++) {
//      os << "\t\t\tnormal_vector[" << i << "] = " << nv[i] << endl;
//    }
//  }
//}
//if (boundary_forms_ok) {
//  for (map<int,NumPtr<Tensor<1,SPACEDIM> > >::const_iterator
//  bf_iterator=boundary_forms.begin();
//  bf_iterator!=boundary_forms.end();bf_iterator++) {
//    const NumPtr<Tensor<1,SPACEDIM> > &bf=bf_iterator->second;
//    os << "\t\tboundary_forms[" << bf_iterator->first << "] has "
//       << bf.getNumber() << " entries" << endl;
//    for (int i=0;i<bf.getNumber();i++) {
//      os << "\t\t\tboundary_form[" << i << "] = " << bf[i] << endl;
//    }
//  }
//}
  HcurlFEValuesBase::printOn(os);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
HcurlFEEdgeValues::HcurlFEEdgeValues(const Mapping &m,
HcurlFiniteElement &fe_in) : HcurlFEValuesBase(UPDATE_Q_POINTS,m,fe_in),
edge_quadrature(fe_in.edgeQuadrature()) {
  const map<int,int> &dpc_map=fe->dofsPerCellMap();
  int nqp=edge_quadrature.numberPoints();
  for (map<int,int>::const_iterator dpc_iterator=dpc_map.begin();
  dpc_iterator!=dpc_map.end();dpc_iterator++) {
    int nv=dpc_iterator->first;
    int nep=linesPerShape(nv)*nqp;
    NumPtr<Point<3> > q_points(nep);
    NumPtr<REAL> weights(nep);
    NumPtr<Point<3> > q_points_edge=q_points;
    NumPtr<REAL> weights_edge=weights;
    for (int l=0;l<linesPerShape(nv);l++) {
      projectToLine(nv,l,edge_quadrature.points(),q_points_edge);
      q_points_edge+=nqp;
      weights_edge+=nqp;
    }
    Quadrature<3> *qe=all_edge_quadrature[nv]=
      OPERATOR_NEW Quadrature<3>(q_points,weights);
    mapping_data_map[nv]=
      mapping->getMappingData(nv,UPDATE_Q_POINTS,qe);
    fe_data_map[nv]=0;
    HcurlFEValuesData *fe_values_data=fe_values_data_map[nv]=
      OPERATOR_NEW HcurlFEValuesData(nv,qe->numberPoints(),*fe,
        UPDATE_Q_POINTS);
    insertNQuadraturePoint(nv,nep);
    insertDofsPerCell(nv,fe_in.dofsPerCell(nv));
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
HcurlFEEdgeValues::~HcurlFEEdgeValues() {
  for (map<int,Quadrature<3>*>::iterator aeq_iterator=
  all_edge_quadrature.begin();aeq_iterator!=all_edge_quadrature.end();
  aeq_iterator++) {
    delete aeq_iterator->second; aeq_iterator->second=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlFEEdgeValues::printOn(ostream &os) const {
  os << "HcurlFEEdgeValues:" << endl;
  HcurlFEValuesBase::printOn(os);
}
#endif

template class _Rb_tree<
  int, 
  pair<int const, HcurlFEValuesData*>, 
  _Select1st<pair<int const, HcurlFEValuesData*> >, 
  less<int>, 
  allocator<pair<int const, HcurlFEValuesData*> > 
>;
