#ifndef HCURL_FE_VALUES_H
#define HCURL_FE_VALUES_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <algorithm>
#include <Array.H>
#include <HcurlFiniteElement.H>
#include <Quadrature.H>
#include <Tensor.H>
#include <Types.H>

//data for finite element shape function(x);
//  ie, mapped from FiniteElementDataBase values
//  the mapping affects derivatives, not values
//the functions in this class should be used only by HcurlFEValuesBase and 
//  derived classes
class HcurlFEValuesData {
  protected:
//  these are set in
//    HcurlFEFaceValues::mapShapes
//    HcurlFESubfaceValues::mapShapes
//    HcurlFEValues::mapShapes
    UPDATE_FLAGS update_flags;
    int number_quadrature_points;
    int number_vertices;
    bool shape_values_ok;
    bool shape_gradients_ok;
    bool shape_2nd_derivatives_ok;
    bool shape_curl_ok;
    bool multiplier_shape_values_ok;
    bool multiplier_shape_gradients_ok;
    bool multiplier_shape_2nd_derivatives_ok;
    bool JxW_values_ok;
    bool quadrature_points_ok;
    const HcurlFiniteElement *finite_element;
    Array<2,Tensor<1,SPACEDIM> > shape_values;
    Array<2,Tensor<2,SPACEDIM> > shape_gradients;
    Array<2,Tensor<3,SPACEDIM> > shape_2nd_derivatives;
    Array<2,HCURL_MULTIPLIER_DATA_TYPE> shape_curl;
    Array<2,HCURL_MULTIPLIER_DATA_TYPE> multiplier_shape_values;
    NumPtr<REAL> JxW_values;
    NumPtr<Point<SPACEDIM> > quadrature_points;
  public:
    HcurlFEValuesData(int,int,const HcurlFiniteElement&,UPDATE_FLAGS);
    virtual ~HcurlFEValuesData() {}
//  called from HcurlFEFaceValues::HcurlFEFaceValues and 
//    HcurlFESubfaceValues::HcurlFESubfaceValues
    void initializeJxWValues(int,UPDATE_FLAGS);
//  called from HcurlFEValues::mapShapes
    void mapShapes(MappingDataBase*,HcurlFiniteElementDataBase*,
      const SHAPE_CLASS*,const Tessellation&,const LEVEL_CLASS*);
//  called from HcurlFEFaceValues::mapShapes
    void mapFaceShapes(MappingDataBase*,HcurlFiniteElementDataBase*,
      const SHAPE_CLASS*,int,const Tessellation&,const LEVEL_CLASS*,
      const Quadrature<SPACEDIM-1>&,NumPtr<Tensor<1,SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&,
//    const HcurlFiniteElement&,UPDATE_FLAGS,
      int,bool&,bool&);
#if (SPACEDIM==3)
    void mapEdgeShapes(MappingDataBase *mapping_data,const SHAPE_CLASS *e,
    int l,const Tessellation &tess,const LEVEL_CLASS *level,
    const Quadrature<1> &eq,int q_offset) {
      int nq=eq.numberPoints();
      NumPtr<Point<3> > qp(quadrature_points,q_offset,nq);
      mapping_data->computeFEEdgeValues(e,tess,level,nq,q_offset,
        qp,quadrature_points_ok);
    }
#endif
//  called from HcurlFESubfaceValues::mapShapes
    void mapSubFaceShapes(MappingDataBase*,HcurlFiniteElementDataBase*,
      const SHAPE_CLASS*,int,const Tessellation&,const LEVEL_CLASS*,
      const Quadrature<SPACEDIM-1>&,NumPtr<Tensor<1,SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&,
//    const HcurlFiniteElement&,UPDATE_FLAGS,
      int,int,bool&,bool&);
//  in dof order; called from HcurlFEValuesBase::getShapeValues
    const Array<2,Tensor<1,SPACEDIM> >& getShapeValues() const { 
      CHECK_TEST(shape_values_ok);
      return shape_values; 
    }
//  in dof order; called from HcurlFEValuesBase::getShapeGradients
    const Array<2,Tensor<2,SPACEDIM> >& getShapeGradients() const { 
      CHECK_TEST(shape_gradients_ok);
      return shape_gradients;
    }
//  in dof order; called from HcurlFEValuesBase::getShapeCurl
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE>& getShapeCurl() const { 
      CHECK_TEST(shape_curl_ok);
      return shape_curl;
    }
//  in dof order; called from HcurlFEValuesBase::getShape2ndDerivatives
    const Array<2,Tensor<3,SPACEDIM> >& getShape2ndDerivatives() const { 
      CHECK_TEST(shape_2nd_derivatives_ok);
      return shape_2nd_derivatives;
    }
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE>& getMultiplierShapeValues()
    const { 
      CHECK_TEST(multiplier_shape_values_ok);
      return multiplier_shape_values; 
    }
//  called from HcurlFEValuesBase::getJxWValues
    const NumPtr<REAL>& getJxWValues() const { 
      CHECK_TEST(JxW_values_ok);
      return JxW_values; 
    }
//  called from HcurlFEValuesBase::getQuadraturePoints
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints() const {
      CHECK_TEST(quadrature_points_ok);
      return quadrature_points;
    }
    virtual void printOn(ostream&) const;
};

class HcurlFEValuesBase {
  protected:
    map<int,int> n_quadrature_points;
    map<int,int> dofs_per_cell;
    map<int,MappingDataBase*> mapping_data_map;
    map<int,HcurlFiniteElementDataBase*> fe_data_map;
    map<int,HcurlFEValuesData*> fe_values_data_map;
    const Mapping *mapping;
    HcurlFiniteElement *fe;
    UPDATE_FLAGS update_flags;

    void insertNQuadraturePoint(int nv,int nqp) {
      n_quadrature_points[nv]=nqp;
    }
    void insertDofsPerCell(int nv,int dpc) {
      dofs_per_cell[nv]=dpc;
    }
    const HcurlFEValuesData* getFEValuesData(int nv) const {
      map<int,HcurlFEValuesData*>::const_iterator fevdm_iterator=
        fe_values_data_map.find(nv);
      if (fevdm_iterator!=fe_values_data_map.end()) {
        return fevdm_iterator->second;
      }
      return 0;
    }
    const HcurlFEValuesData* getFEValuesData(const Shape *e) const {
      return getFEValuesData(e->numberVertices());
    }
    HcurlFEValuesData* getFEValuesData(int nv) {
      map<int,HcurlFEValuesData*>::iterator fevdm_iterator=
        fe_values_data_map.find(nv);
      if (fevdm_iterator!=fe_values_data_map.end()) {
        return fevdm_iterator->second;
      }
      return 0;
    }
    HcurlFEValuesData* getFEValuesData(const Shape *e) {
      return getFEValuesData(e->numberVertices());
    }
    const Array<2,Tensor<1,SPACEDIM> >* getShapeValues(int nv) const { 
      const HcurlFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeValues()) : 0);
    }
    const Array<2,Tensor<2,SPACEDIM> >* getShapeGradients(int nv) const { 
      const HcurlFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeGradients()) : 0);
    }
#if (SPACEDIM<=2)
    const Array<2,REAL>* getShapeCurl(int nv) const { 
      const HcurlFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeCurl()) : 0);
    }
#else
    const Array<2,Tensor<1,3> >* getShapeCurl(int nv) const { 
      const HcurlFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeCurl()) : 0);
    }
#endif
    const Array<2,Tensor<3,SPACEDIM> >* getShape2ndDerivatives(int nv) 
    const { 
      const HcurlFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShape2ndDerivatives()) 
                             : 0);
    }
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE>* getMultiplierShapeValues(
    int nv) const { 
      const HcurlFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ?
              &(fe_values_data->getMultiplierShapeValues()) : 0);
    }

    const NumPtr<REAL>& getJxWValues(int nv) const {
      const HcurlFEValuesData *fe_values_data=getFEValuesData(nv);
      CHECK_POINTER(fe_values_data);
      return fe_values_data->getJxWValues() ;
    }
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints(int nv) const {
      const HcurlFEValuesData *fe_values_data=getFEValuesData(nv);
      CHECK_POINTER(fe_values_data);
      return fe_values_data->getQuadraturePoints() ;
    }
    int dofsPerCell(int nv) const { 
      map<int,int>::const_iterator dpc_iterator=
        dofs_per_cell.find(nv);
      CHECK_TEST(dpc_iterator!=dofs_per_cell.end());
      return dpc_iterator->second;
    }
    HcurlFiniteElement& getFe() const { return *fe; }
//  if shape_function->requiresShapeToComputeValues,
//  can't compute values here because we can't get the shape_function
    HcurlFiniteElementDataBase* getFEData(const Shape *e) {
      map<int,HcurlFiniteElementDataBase*>::iterator
        fedm_iterator=fe_data_map.find(e->numberVertices());
      CHECK_TEST(fedm_iterator!=fe_data_map.end());
      return fedm_iterator->second;
    }
  public:
    HcurlFEValuesBase(UPDATE_FLAGS flags,const Mapping &m,
    HcurlFiniteElement &fe_in) :
      update_flags(flags),mapping(&m),fe(&fe_in) { }
    virtual ~HcurlFEValuesBase();
    const Mapping& getMapping() const { return *mapping; }
    const NumPtr<REAL>& getJxWValues(const Shape *e) const {
      return getJxWValues(e->numberVertices());
    }
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints(const Shape *e) 
    const {
      return getQuadraturePoints(e->numberVertices());
    }
    int dofsPerCell(const Shape *e) const { 
      return dofsPerCell(e->numberVertices());
    }
    int numberQuadraturePoints(const Shape *e) const {
      map<int,int>::const_iterator nqp_iterator=
        n_quadrature_points.find(e->numberVertices());
      CHECK_TEST(nqp_iterator!=n_quadrature_points.end());
      return nqp_iterator->second;
    }
    MappingDataBase* mappingData(int nv) {
      map<int,MappingDataBase*>::iterator
        mdm_iterator=mapping_data_map.find(nv);
      CHECK_TEST(mdm_iterator!=mapping_data_map.end());
      return mdm_iterator->second;
    }
    const MappingDataBase* mappingData(int nv) const {
      map<int,MappingDataBase*>::const_iterator
        mdm_iterator=mapping_data_map.find(nv);
      CHECK_TEST(mdm_iterator!=mapping_data_map.end());
      return mdm_iterator->second;
    }
    MappingDataBase* mappingData(const Shape *e) {
      return mappingData(e->numberVertices());
    }
    const MappingDataBase* mappingData(const Shape *e) const {
      return mappingData(e->numberVertices());
    }
//  if shape_function->requiresShapeToComputeValues,
//  can't compute values here because we can't get the shape_function
    const HcurlFiniteElementDataBase* getFEData(const Shape *e) const {
      map<int,HcurlFiniteElementDataBase*>::const_iterator
        fedm_iterator=fe_data_map.find(e->numberVertices());
      CHECK_TEST(fedm_iterator!=fe_data_map.end());
      return fedm_iterator->second;
    }
//  called from HcurlDofHandler::assembleLinearSystem
    const Array<2,Tensor<1,SPACEDIM> >* getShapeValues(const Shape *e)
    const { 
      return getShapeValues(e->numberVertices());
    }
    const Array<2,Tensor<2,SPACEDIM> >* getShapeGradients(
    const Shape *e) const { 
      return getShapeGradients(e->numberVertices());
    }
#if (SPACEDIM<=2)
    const Array<2,REAL>* getShapeCurl(const Shape *e) const { 
      return getShapeCurl(e->numberVertices());
    }
#else
    const Array<2,Tensor<1,3> >* getShapeCurl(const Shape *e) const { 
      return getShapeCurl(e->numberVertices());
    }
#endif
    const Array<2,Tensor<3,SPACEDIM> >* getShape2ndDerivatives(
    const Shape *e) const { 
      return getShape2ndDerivatives(e->numberVertices());
    }
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE>* getMultiplierShapeValues(
    const Shape *e) const { 
      return getMultiplierShapeValues(e->numberVertices());
    }
    virtual void printFEData(ostream&) const;
    virtual void printOn(ostream&) const;
};

class HcurlFEValues : public HcurlFEValuesBase {
  private:
    map<int,Quadrature<SPACEDIM> > quadrature;

    Quadrature<SPACEDIM>& getQuadrature(int nv) {
      map<int,Quadrature<SPACEDIM> >::iterator q_iterator
        =quadrature.find(nv);
      CHECK_TEST(q_iterator!=quadrature.end());
      return q_iterator->second;
    }
    const Quadrature<SPACEDIM>& getQuadrature(int nv) const {
      map<int,Quadrature<SPACEDIM> >::const_iterator q_iterator
        =quadrature.find(nv);
      CHECK_TEST(q_iterator!=quadrature.end());
      return q_iterator->second;
    }
  public:
    HcurlFEValues(const Mapping&,HcurlFiniteElement&,
      const map<int,Quadrature<SPACEDIM> >&,UPDATE_FLAGS);
    virtual ~HcurlFEValues() {;}
//  compute mapped values of shape functions/derivatives in dof order
    void mapShapes(const Tessellation &tess,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level) {
      getFEValuesData(e)->
        mapShapes(mappingData(e),getFEData(e),e,tess,level);
    }
    const Quadrature<SPACEDIM>& getQuadrature(const Shape *e) const {
      return getQuadrature(e->numberVertices());
    }
//  called from HcurlDofHandler::getFunctionValues
//    T = Tensor<1,SPACEDIM> or NumPtr<Tensor<1,SPACEDIM> >
    template<class T> void getFunctionValues(const Vector &dof_values,
    const SHAPE_CLASS *e,NumPtr<T> &values) const {
      CHECK_TEST(update_flags & UPDATE_VALUES);
      fe->sumShapeValues(e,*getShapeValues(e),numberQuadraturePoints(e),
        0,dof_values,values);
    }
//  for eventual call to FiniteElement::sumShapeDerivatives
//    T = Tensor<2,SPACEDIM> or NumPtr<Tensor<2,SPACEDIM> >
    template<class T> void getFunctionDerivatives(
    const Vector &dof_values,const SHAPE_CLASS *e,NumPtr<T> &gradients) 
    const {
      CHECK_TEST(update_flags & UPDATE_GRADIENTS); 
      fe->sumShapeDerivatives(e,*getShapeGradients(e),
        numberQuadraturePoints(e),0,dof_values,gradients);
    }
//  for eventual call to FiniteElement::sumShapeCurl
//    T = REAL or Tensor<1,3>
    template<class T> void getFunctionCurl(
    const Vector &dof_values,const SHAPE_CLASS *e,NumPtr<T> &curl) 
    const {
      CHECK_TEST(update_flags & UPDATE_CURL); 
      fe->sumShapeCurl(e,*getShapeCurl(e),
        numberQuadraturePoints(e),0,dof_values,curl);
    }
//  called from HcurlDofHandler::getMultiplierFunctionValues
//    T = REAL or Vector
    template<class T> void getMultiplierFunctionValues(
    const Vector &dof_values,const SHAPE_CLASS *e,NumPtr<T> &values)
    const {
      CHECK_TEST(update_flags & UPDATE_VALUES);
      fe->sumMultiplierShapeValues(e,*getMultiplierShapeValues(e),
        numberQuadraturePoints(e),0,dof_values,values);
    }
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM>=2)
class HcurlFEFaceValues : public HcurlFEValuesBase {
  private:
    map<int,Quadrature<SPACEDIM>*> all_face_quadrature;

//  multipliers have no boundary trace values
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE>* getMultiplierShapeValues(
    const Shape *e) const { 
      OBSOLETE("should not be called for this class");
    }
  protected:
    map<int,Quadrature<SPACEDIM-1> > face_quadrature;
    map<int,NumPtr<Point<SPACEDIM> > > normal_vectors;
//  (d x / d xi)^{-T} n_{xi} det(d x / d xi):
    map<int,NumPtr<Tensor<1,SPACEDIM> > > boundary_forms;
    bool boundary_forms_ok;
    bool normal_vectors_ok;

    void initializeBoundaryFormsAndNormals(int,
      const Quadrature<SPACEDIM-1>&);

    const Quadrature<SPACEDIM-1>& getFaceQuadrature(int nv) const {
      map<int,Quadrature<SPACEDIM-1> >::const_iterator
        q_iterator=face_quadrature.find(nv);
      CHECK_TEST(q_iterator!=face_quadrature.end());
      return q_iterator->second;
    }
    const NumPtr<Point<SPACEDIM> >& getNormalVectors(int nv) const {
      CHECK_TEST(normal_vectors_ok);
      map<int,NumPtr<Point<SPACEDIM> > >::const_iterator
        nv_iterator=normal_vectors.find(nv);
      CHECK_TEST(nv_iterator!=normal_vectors.end());
      return nv_iterator->second;
    }
    int faceOffset(const SHAPE_CLASS*,int) const;
  public:
    HcurlFEFaceValues(const Mapping&,HcurlFiniteElement&, 
      map<int,Quadrature<SPACEDIM-1> >&,UPDATE_FLAGS);
    virtual ~HcurlFEFaceValues();
    const Quadrature<SPACEDIM-1>& getFaceQuadrature(const Shape2 *e,int f) 
    const {
      CHECK_TEST(f<e->numberCodimension1Shapes());
      return getFaceQuadrature(e->numberVerticesOnFace(f));
    }
    void getCanonicalQuadraturePoints(const SHAPE_CLASS *e,int f,
    NumPtr<Point<SPACEDIM> > &cp) const {
      int s=faceOffset(e,f);
      map<int,Quadrature<SPACEDIM>*>::const_iterator afq_iterator=
        all_face_quadrature.find(e->numberVertices());
      CHECK_TEST(afq_iterator!=all_face_quadrature.end());
      cp=NumPtr<Point<SPACEDIM> >(afq_iterator->second->points(),
        s,getFaceQuadrature(e,f).numberPoints());
    }
    NumPtr<Point<SPACEDIM> > getQuadraturePoints(const SHAPE_CLASS *e,
    int f) const {
      int s=faceOffset(e,f);
      return NumPtr<Point<SPACEDIM> >(
        getFEValuesData(e)->getQuadraturePoints(),
        s,getFaceQuadrature(e,f).numberPoints());
    }
    NumPtr<REAL> getJxWValues(const SHAPE_CLASS *e,int f) const {
      int s=faceOffset(e,f);
      return NumPtr<REAL>(getFEValuesData(e)->getJxWValues(),s,
        getFaceQuadrature(e,f).numberPoints());
    }
    const NumPtr<Point<SPACEDIM> >& getNormalVectors(const Shape2 *e,
    int f) const {
      return getNormalVectors(e->numberVerticesOnFace(f));
    }

    void mapShapes(const Tessellation&,const SHAPE_CLASS*,
      const LEVEL_CLASS*,int);
//  called from HcurlDofHandler::getFunctionValues
    void getFunctionValues(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<Tensor<1,SPACEDIM> >&) const;
    void getFunctionValues(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
    virtual void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class HcurlFEEdgeValues : public HcurlFEValuesBase {
  protected:
    map<int,Quadrature<3>*> all_edge_quadrature;
    const Quadrature<1> &edge_quadrature;

    int edgeOffset(const SHAPE_CLASS *e,int l) const {
      return fe->edgeQuadrature().numberPoints()*l;
    }
  public:
    HcurlFEEdgeValues(const Mapping&,HcurlFiniteElement&);
    virtual ~HcurlFEEdgeValues();
    const Quadrature<1>& getEdgeQuadrature() const {
      return edge_quadrature;
    }
    void getCanonicalQuadraturePoints(const SHAPE_CLASS *e,int l,
    NumPtr<Point<3> > &cp) const {
      map<int,Quadrature<3>*>::const_iterator aeq_iterator=
        all_edge_quadrature.find(e->numberVertices());
      CHECK_TEST(aeq_iterator!=all_edge_quadrature.end());
      cp=NumPtr<Point<3> >(aeq_iterator->second->points(),
        edgeOffset(e,l),edge_quadrature.numberPoints());
    }
    NumPtr<Point<3> > getQuadraturePoints(const SHAPE_CLASS *e,
    int l) const {
      return NumPtr<Point<3> >(
        getFEValuesData(e)->getQuadraturePoints(),
        edgeOffset(e,l),edge_quadrature.numberPoints());
    }

    void mapShapes(const Tessellation &tess,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level,int f) {
      int lpf=linesPerFace(e->numberVertices(),f);
      for (int l=0;l<lpf;l++) {
        getFEValuesData(e)->mapEdgeShapes(mappingData(e),e,l,tess,level,
          edge_quadrature,edgeOffset(e,e->elementLineFromFace(f,l)));
      }
    }
    virtual void printOn(ostream&) const;
};
#endif

#endif
