// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "HcurlFiniteElement.H"
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//quad must be destructed after this is destructed
HcurlFiniteElementDataBase::HcurlFiniteElementDataBase(
UPDATE_FLAGS flags,const Quadrature<SPACEDIM> *quad,
const TensorShapeFunction<SPACEDIM> *sff,
const HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE *msff) :
update_flags(flags),shape_values_ok(false),shape_derivatives_ok(false),
shape_second_derivatives_ok(false),shape_values(0,0),
shape_derivatives(0,0),shape_second_derivatives(0,0),
shape_function_family(sff),
multiplier_shape_values_ok(false),multiplier_shape_derivatives_ok(false),
multiplier_shape_second_derivatives_ok(false),
multiplier_shape_values(0,0),
multiplier_shape_function_family(msff),quadrature(quad) {
  int n_q_points=quadrature->numberPoints();
  int n_shapes=sff->getNumber();
  int n_multiplier_shapes=msff->getNumber();
  if (flags & UPDATE_VALUES) {
    shape_values.reinit(n_shapes,n_q_points);
    multiplier_shape_values.reinit(n_multiplier_shapes,n_q_points);
  }
  if (flags & (UPDATE_GRADIENTS | UPDATE_CURL)) {
    shape_derivatives.reinit(n_shapes,n_q_points);
  }
  if (flags & UPDATE_SECOND_DERIVATIVES) {
    shape_second_derivatives.reinit(n_shapes,n_q_points);
  }
  if (!sff->requiresShapeToComputeValues()) {
    if (flags & (UPDATE_VALUES | UPDATE_GRADIENTS | 
    UPDATE_SECOND_DERIVATIVES)) {
      for (int q=0;q<n_q_points;q++) {
        if (update_flags & UPDATE_VALUES) {
          NumPtr<Tensor<1,SPACEDIM> > values;
          sff->computeValues(quadrature->point(q),values);
          for (int s=0;s<n_shapes;s++) {
            shape_values(s,q)=values[s];
          }
          shape_values_ok=true;
        }
        if (update_flags & (UPDATE_GRADIENTS | UPDATE_CURL)) {
          NumPtr<Tensor<2,SPACEDIM> > grads;
          sff->computeGrads(quadrature->point(q),grads);
          for (int s=0;s<n_shapes;s++) {
            shape_derivatives(s,q)=grads[s];
          }
          shape_derivatives_ok=true;
        }
        if (update_flags & UPDATE_SECOND_DERIVATIVES) {
          NumPtr<Tensor<3,SPACEDIM> > grad_grads;
          sff->computeGradGrads(quadrature->point(q),grad_grads);
          for (int s=0;s<n_shapes;s++) {
            shape_second_derivatives(s,q)=grad_grads[s];
          }
          shape_second_derivatives_ok=true;
        }
      }
    }
  }
  if (!msff->requiresShapeToComputeValues()) {
    if (flags & (UPDATE_VALUES | UPDATE_GRADIENTS | 
    UPDATE_SECOND_DERIVATIVES)) {
      for (int q=0;q<n_q_points;q++) {
        if (update_flags & UPDATE_VALUES) {
          NumPtr<HCURL_MULTIPLIER_DATA_TYPE> values;
          msff->computeValues(quadrature->point(q),values);
          for (int s=0;s<n_multiplier_shapes;s++) {
            multiplier_shape_values(s,q)=values[s];
          }
          multiplier_shape_values_ok=true;
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
HcurlFiniteElementDataBase::HcurlFiniteElementDataBase(
const HcurlFiniteElementDataBase *fedb,const Shape *e) :
update_flags(fedb->update_flags),shape_values_ok(false),
shape_derivatives_ok(false),shape_second_derivatives_ok(false),
shape_values(fedb->shape_values.size(0),fedb->shape_values.size(1)),
shape_derivatives(fedb->shape_derivatives.size(0),
                  fedb->shape_derivatives.size(1)),
shape_second_derivatives(fedb->shape_second_derivatives.size(0),
                         fedb->shape_second_derivatives.size(1)),
multiplier_shape_values_ok(false),multiplier_shape_derivatives_ok(false),
multiplier_shape_second_derivatives_ok(false),
multiplier_shape_values(fedb->multiplier_shape_values.size(0),
  fedb->multiplier_shape_values.size(1)),
quadrature(fedb->quadrature),
shape_function_family(fedb->shape_function_family),
multiplier_shape_function_family(fedb->multiplier_shape_function_family)
{
  NumPtr<int> renumber;
  shape_function_family->shapeFunctionsOrder(e,renumber);
  if (fedb->shape_values_ok) {
    const Array<2,Tensor<1,SPACEDIM> > &sv=fedb->shape_values;
    for (int fe_s=0;fe_s<sv.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sv.size(1);q++) {
        const Tensor<1,SPACEDIM> &svsq=sv(fe_s,q);
        shape_values(abs(dof_s),q)=(dof_s>=0 ? svsq : -svsq);
      }
    }
    shape_values_ok=true;
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE> &msv=
      fedb->multiplier_shape_values;
    for (int s=0;s<msv.size(0);s++) {
      for (int q=0;q<msv.size(1);q++) {
        multiplier_shape_values(s,q)=msv(s,q);
      }
    }
    multiplier_shape_values_ok=true;
  }
  if (fedb->shape_derivatives_ok) {
    const Array<2,Tensor<2,SPACEDIM> > &sg=fedb->shape_derivatives;
    for (int fe_s=0;fe_s<sg.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sg.size(1);q++) {
        const Tensor<2,SPACEDIM> &sgsq=sg(fe_s,q);
        shape_derivatives(abs(dof_s),q)=(dof_s>=0 ? sgsq : - sgsq);
      }
    }
    shape_derivatives_ok=true;
  }
  if (fedb->shape_second_derivatives_ok) {
    const Array<2,Tensor<3,SPACEDIM> > &sgg=
      fedb->shape_second_derivatives;
    for (int fe_s=0;fe_s<sgg.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sgg.size(1);q++) {
        const Tensor<3,SPACEDIM> &sggsq=sgg(fe_s,q);
        shape_second_derivatives(abs(dof_s),q)=
          (dof_s>=0 ? sggsq : -sggsq);
      }
    }
    shape_second_derivatives_ok=true;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElementDataBase::computeShapesFor(const Shape *e) {
  if (shape_function_family->requiresShapeToComputeValues()) {
    int n_points=quadrature->numberPoints();
    if (update_flags & UPDATE_VALUES) {
      for (int q=0;q<n_points;q++) {
        NumPtr<Tensor<1,SPACEDIM> > values;
        shape_function_family->
          computeValues(quadrature->point(q),values);
        for (int s=0;s<values.getNumber();s++) {
          shape_values(s,q)=values[s];
        }
      }
      shape_values_ok=true;
    }
    if (update_flags & UPDATE_GRADIENTS) {
      for (int q=0;q<n_points;q++) {
        NumPtr<Tensor<2,SPACEDIM> > grads;
        shape_function_family->
          computeGrads(quadrature->point(q),grads);
        for (int s=0;s<grads.getNumber();s++) {
          shape_derivatives(s,q)=grads[s];
        }
      }
      shape_derivatives_ok=true;
    }
    if (update_flags & UPDATE_SECOND_DERIVATIVES) {
      for (int q=0;q<n_points;q++) {
        NumPtr<Tensor<3,SPACEDIM> > grad_grads;
        shape_function_family->
          computeGradGrads(quadrature->point(q),grad_grads);
        for (int s=0;s<grad_grads.getNumber();s++) {
          shape_second_derivatives(s,q)=grad_grads[s];
        }
      }
      shape_second_derivatives_ok=true;
    }
  }
  if (multiplier_shape_function_family->requiresShapeToComputeValues()) {
    OBSOLETE("not programmed");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElementDataBase::printOn(ostream &os) const {
  os << "HcurlFiniteElementDataBase: update_flags = " << update_flags 
     << "\n\tshape_values_ok = " << PRINT_BOOLEAN(shape_values_ok)
     << "\n\tshape_derivatives_ok = " 
     << PRINT_BOOLEAN(shape_derivatives_ok)
     << "\n\tshape_second_derivatives_ok = " 
     << PRINT_BOOLEAN(shape_second_derivatives_ok)
     << "\n\tmultiplier_shape_values_ok = "
     << PRINT_BOOLEAN(multiplier_shape_values_ok)
     << "\n\tmultiplier_shape_derivatives_ok = " 
     << PRINT_BOOLEAN(multiplier_shape_derivatives_ok)
     << "\n\tmultiplier_tshape_second_derivatives_ok = " 
     << PRINT_BOOLEAN(multiplier_shape_second_derivatives_ok)
     << endl;
  if (shape_values_ok) {
    os << "\tshape_values:" << endl;
    for (int s=0;s<shape_values.size(0);s++) {
      for (int q=0;q<shape_values.size(1);q++) {
        os << "\t\tshape_values[" << s << "," << q << "] = ";
        for (int d=0;d<SPACEDIM;d++) {
          REAL svsqd=shape_values(s,q)[d];
          os << (abs(svsqd)>sqrt(DBL_EPSILON) ?  svsqd : 0.) << " ";
        }
        os << endl;
      }
    }
  }
//if (shape_derivatives_ok) {
//  os << "\tshape_derivatives:" << endl;
//  for (int s=0;s<shape_derivatives.size(0);s++) {
//    for (int q=0;q<shape_derivatives.size(1);q++) {
//      os << "\t\tshape_derivatives[" << s << "," << q << "] = "
//         << shape_derivatives(s,q) << endl;
//    }
//  }
//}
//if (shape_second_derivatives_ok) {
//  os << "\tshape_second_derivatives:" << endl;
//  for (int s=0;s<shape_second_derivatives.size(0);s++) {
//    for (int q=0;q<shape_second_derivatives.size(1);q++) {
//      os << "\t\tshape_second_derivatives[" << s << "," << q << "] = "
//         << shape_second_derivatives(s,q) << endl;
//    }
//  }
//}
  if (multiplier_shape_values_ok) {
    os << "\tmultiplier_shape_values:" << endl;
    for (int i=0;i<multiplier_shape_values.size(0);i++) {
      for (int j=0;j<multiplier_shape_values.size(1);j++) {
        os << "\t\tmultiplier_shape_values[" << i << "," << j << "] = "
           << multiplier_shape_values(i,j) << endl;
      }
    }
  }
  cout << "\tquadrature = " << quadrature << endl;
  quadrature->printOn(os);
  shape_function_family->printOn(os);
  multiplier_shape_function_family->printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
HcurlFiniteElement::~HcurlFiniteElement() {
  while (!polynomial_space.empty()) {
    map<int,HcurlDualTensorShapeFunction*>::iterator 
      ps_iterator=polynomial_space.begin();
    HcurlDualTensorShapeFunction *ps=ps_iterator->second;
    if (ps!=0) delete ps;
    polynomial_space.erase(ps_iterator);
  }
  while (!multiplier_polynomial_space.empty()) {
    map<int,HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE*>::iterator 
      ps_iterator=multiplier_polynomial_space.begin();
    HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE *ps=ps_iterator->second;
    if (ps!=0) delete ps;
    multiplier_polynomial_space.erase(ps_iterator);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::setDofsPerObject() {
  for (map<int,HcurlDualTensorShapeFunction*>::const_iterator
  ps_iterator=polynomial_space.begin();
  ps_iterator!=polynomial_space.end();ps_iterator++) {
    int nv=ps_iterator->first;
    HcurlDualTensorShapeFunction *ps=ps_iterator->second;
    if (ps==0) {
      dofs_per_cell[nv]=0;
      dofs_per_shape[nv]=0;
#if (SPACEDIM==1)
      dofs_per_codimension1_shape[1]=0;
#elif (SPACEDIM==2)
      dofs_per_codimension1_shape[2]=0;
#else
      dofs_per_line=0;
      if (nv==4) {
        dofs_per_codimension1_shape[3]=0;
      } else if (nv==8) {
        dofs_per_codimension1_shape[4]=0;
      }
#endif
    } else {
      dofs_per_cell[nv]=ps->getNumber()*components;
      dofs_per_shape[nv]=ps->getNumberInteriorDofs()*components;
#if (SPACEDIM==1)
      dofs_per_codimension1_shape[1]=ps->getNumberEdgeDofs()*components;
#elif (SPACEDIM==2)
      dofs_per_codimension1_shape[2]=ps->getNumberEdgeDofs()*components;
#else
      dofs_per_line=ps->getNumberEdgeDofs()*components;
      if (nv==4) {
        dofs_per_codimension1_shape[3]=ps->getNumberFaceDofs()*components;
      } else if (nv==8) {
        dofs_per_codimension1_shape[4]=ps->getNumberFaceDofs()*components;
      }
#endif
    }
  }
  for (map<int,HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE*>::const_iterator
  mps_iterator=multiplier_polynomial_space.begin();mps_iterator!=
  multiplier_polynomial_space.end();mps_iterator++) {
    int nv=mps_iterator->first;
    HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE *mps=mps_iterator->second;
    if (mps==0) {
      multiplier_dofs_per_shape[nv]=0;
    } else {
      multiplier_dofs_per_shape[nv]=mps->getNumber()*components;
    }
  }
//the following are for essential boundary conditions
#if (SPACEDIM==2)
  int nv=Interval::verticesPerShape();
  int nrp=dofsPerFace(nv)/components;
  NumPtr<Point<1> > random_points(nrp);
  for (int i=0;i<nrp;i++) random_points[i][0]=drand48();
  Quadrature<1> random_quad(random_points);
  fbvmap[nv]=random_quad;
#elif (SPACEDIM==3)
  int nv=Quadrilateral::verticesPerShape();
  int nrp=(dofsPerFace(nv)+Quadrilateral::facesPerShape()*dofsPerLine())
         /components;
  NumPtr<Point<2> > random_pointsq(nrp);
  for (int i=0;i<nrp;i++) {
    for (int d=0;d<2;d++) random_pointsq[i][d]=drand48();
  }
  Quadrature<2> random_quadq(random_pointsq);
  fbvmap[nv]=random_quadq;

  nv=Triangle::verticesPerShape();
  nrp=(dofsPerFace(nv)+Triangle::facesPerShape()*dofsPerLine())
     /components;
  NumPtr<Point<2> > random_pointst(nrp);
  for (int i=0;i<nrp;i++) {
    random_pointst[i][0]=drand48();
    random_pointst[i][1]=(1.-random_pointst[i][0])*drand48();
  }
  Quadrature<2> random_quadt(random_pointst);
  fbvmap[nv]=random_quadt;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int HcurlFiniteElement::maxDofsPerCell() const {
  int max_dofs_per_cell=0;
  for (map<int,int>::const_iterator dpc_iterator=
  dofs_per_cell.begin();dpc_iterator!=dofs_per_cell.end();
  dpc_iterator++) {
    max_dofs_per_cell=max(max_dofs_per_cell,dpc_iterator->second);
  }
  return max_dofs_per_cell;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int HcurlFiniteElement::maxMultiplierDofsPerShape() const {
  int max_multiplier_dofs_per_shape=0;
  for (map<int,int>::const_iterator dpc_iterator=
  multiplier_dofs_per_shape.begin();dpc_iterator!=
  multiplier_dofs_per_shape.end();dpc_iterator++) {
    max_multiplier_dofs_per_shape=
      max(max_multiplier_dofs_per_shape,dpc_iterator->second);
  }
  max_multiplier_dofs_per_shape+=max_multiplier_dofs_per_shape;
  return max_multiplier_dofs_per_shape;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//dof_values come from face elements, in face element's ordering, then
//  from element's interior, in element's ordering
//see HcurlDoFHandler::getDofIndices for details
void HcurlFiniteElement::sumShapeValues(const SHAPE_CLASS *e,
const Array<2,Tensor<1,SPACEDIM> > &sv,int nqp,int offset,
const Vector &dof_values,NumPtr<Tensor<1,SPACEDIM> > &values) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    Tensor<1,SPACEDIM> val;
    for (int s=0;s<dpc;s++) {
      const Tensor<1,SPACEDIM> &svsq=sv(s,q+offset);
      for (int d=0;d<SPACEDIM;d++) val[d]+=svsq[d]*dof_values[s]; 
    }
    values[q]=val;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::sumShapeValues(const SHAPE_CLASS *e,
const Array<2,Tensor<1,SPACEDIM> > &sv,int nqp,int offset,
const Vector &dof_values,NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &values)
const {
  int shapes_per_cell=dofsPerCell(e)/components;
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<1,SPACEDIM> > &v=values[q];
    v.allocate(components);
    for (int s=0;s<shapes_per_cell;s++) {
      const Tensor<1,SPACEDIM> &svq=sv(s,q+offset);
      int sc=s*components;
      for (int c=0;c<components;c++) {
        for (int d=0;d<SPACEDIM;d++) v[c][d]+=svq[d]*dof_values[c+sc]; 
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::sumMultiplierShapeValues(const SHAPE_CLASS *e,
const Array<2,HCURL_MULTIPLIER_DATA_TYPE> &sv,int nqp,int offset,
const Vector &dof_values,NumPtr<HCURL_MULTIPLIER_DATA_TYPE> &values)
const {
  CHECK_SAME(components,1);
  int dpc=multiplierDofsPerShape(e);
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
#if (SPACEDIM<=2)
    HCURL_MULTIPLIER_DATA_TYPE val=0.;
#else
    HCURL_MULTIPLIER_DATA_TYPE val;
#endif
    for (int s=0;s<dpc;s++) val+=sv(s,q+offset)*dof_values[s]; 
    values[q]=val;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::sumMultiplierShapeValues(const SHAPE_CLASS *e,
const Array<2,HCURL_MULTIPLIER_DATA_TYPE> &sv,int nqp,int offset,
const Vector &dof_values,
NumPtr<NumPtr<HCURL_MULTIPLIER_DATA_TYPE> > &values) const {
  int shapes_per_cell=multiplierDofsPerShape(e)/components;
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    NumPtr<HCURL_MULTIPLIER_DATA_TYPE> &v=values[q];
    v.allocate(components);
    for (int s=0;s<shapes_per_cell;s++) {
      const HCURL_MULTIPLIER_DATA_TYPE &svq=sv(s,q+offset);
      int sc=s*components;
      for (int c=0;c<components;c++) v[c]+=svq*dof_values[c+sc]; 
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::sumShapeDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<2,SPACEDIM> > &sg,int nqp,int offset,
const Vector &dof_values,NumPtr<Tensor<2,SPACEDIM> > &gradients) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  gradients.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    for (int s=0;s<dpc;s++) {
      Tensor<2,SPACEDIM> tmp(sg(s,q+offset));
      tmp*=dof_values[s];
      gradients[q]+=tmp;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::sumShapeDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<2,SPACEDIM> > &sg,int nqp,int offset,
const Vector &dof_values,
NumPtr<NumPtr<Tensor<2,SPACEDIM> > > &gradients) const {
  int shapes_per_cell=dofsPerCell(e)/components;
  if (gradients.getNumber()==0) gradients.allocate(nqp);
#ifdef DEBUG
  else CHECK_SAME(gradients.getNumber(),nqp);
#endif
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<2,SPACEDIM> > &g=gradients[q];
    if (g.getNumber()==0) g.allocate(components);
#ifdef DEBUG
    else CHECK_SAME(g.getNumber(),components);
#endif
    Tensor<2,SPACEDIM> empty_tensor;
    g.initialize(empty_tensor);
    for (int s=0;s<shapes_per_cell;s++) {
      int sc=s*components;
      const Tensor<2,SPACEDIM> &sgq=sg(s,q+offset);
      for (int c=0;c<components;c++) {
        Tensor<2,SPACEDIM> tmp(sgq);
        tmp*=dof_values[c+sc];
        g[c]+=tmp;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM<=2)
void HcurlFiniteElement::sumShapeCurl(const SHAPE_CLASS *e,
const Array<2,REAL> &sc,int nqp,int offset,const Vector &dof_values,
NumPtr<REAL> &curl) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  curl.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    REAL &cq=curl[q];
    cq=0.;
    for (int s=0;s<dpc;s++) cq+=sc(s,q+offset)*dof_values[s];
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM<=2)
void HcurlFiniteElement::sumShapeCurl(const SHAPE_CLASS *e,
const Array<2,REAL> &sc,int nqp,int offset,const Vector &dof_values,
NumPtr<NumPtr<REAL> > &curl) const {
  int shapes_per_cell=dofsPerCell(e)/components;
  if (curl.getNumber()==0) curl.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    NumPtr<REAL> &cq=curl[q];
    if (cq.getNumber()==0) cq.allocate(components);
    for (int c=0;c<cq.getNumber();c++) cq[c]=0.;
    for (int s=0;s<shapes_per_cell;s++) {
      int cs=components*s;
      REAL scq=sc(s,q+offset);
      for (int c=0;c<components;c++) cq[c]+=scq*dof_values[c+cs];
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlFiniteElement::sumShapeCurl(const SHAPE_CLASS *e,
const Array<2,Tensor<1,3> > &sc,int nqp,int offset,
const Vector &dof_values,NumPtr<Tensor<1,3> > &curl) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  curl.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    Tensor<1,3> &cq=curl[q];
    cq.initialize(0.);
    for (int s=0;s<dpc;s++) {
      Tensor<1,3> scsq=sc(s,q+offset);
      scsq*=dof_values[s];
      cq+=scsq;
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HcurlFiniteElement::sumShapeCurl(const SHAPE_CLASS *e,
const Array<2,Tensor<1,3> > &sc,int nqp,int offset,
const Vector &dof_values,NumPtr<NumPtr<Tensor<1,3> > > &curl)
const {
  int shapes_per_cell=dofsPerCell(e)/components;
  if (curl.getNumber()==0) curl.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<1,3> > &cq=curl[q];
    if (cq.getNumber()==0) cq.allocate(components);
    for (int c=0;c<cq.getNumber();c++) cq[c].initialize(0.);
    for (int s=0;s<shapes_per_cell;s++) {
      int cs=s*components;
      const Tensor<1,3> &scq=sc(s,q+offset);
      for (int c=0;c<components;c++) {
        for (int d=0;d<3;d++) cq[c][d]+=scq[d]*dof_values[c+cs];
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//essential boundary conditions for Maxwell's equations might look like
//  n x E = b on the boundary
//this has a solution iff n . b = 0, in which case the soln is not unique
//
//if t_0 and t_1 are tangents to the boundary with t_1 = n x t_0, then
//  t_0 = - n x t_1
//if E = t_0 epsilon_0 + t_1 epsilon_1 and b = t_0 beta_0 + t_1 beta_1 then
//  epsilon_0 = beta_1 and epsilon_1 = - beta_0
//solves the equation n x E = b subject to n . E = 0
//
//the Hcurl basis functions are dual to the dof polynomials via linear
//  functionals on the reference shape
//thus if E = sum_j E_j c_j we should have
//  int_Edge chi_i tangent_Edge . [ t_0 beta_1 - t_1 beta_0 ] ds
//  = int_Edge chi_i tangent_Edge . E ds
//  = sum_j c_j int_Edge chi_i tangent_Edge . E_j d s
//  = c_i
//Here t_0 and t_1 are tangent to the mapped face, and
//  tangent_Edge is tangent to the reference Edge
//Similarly, if n_Face . chi_i = 0 then
//  int_Face chi_i . [ t_0 beta_1 - t_1 beta_0 ] da
//  = int_Face chi_i . E da
//  = sum_j c_j int_Face chi_i . E_j da
//  = c_i
//Thus we need the mapped tangents to the edges and face, as well as the
//  mapped tangential components beta_0 and beta_1 of the essential data
//  in order to compute the linear functionals

#if (SPACEDIM>=2)
void HcurlFiniteElement::setFaceBoundaryValues(const SHAPE_CLASS *element,
int f,const NumPtr<int> &face_dofs,
const NumPtr<Point<SPACEDIM-1> > &canonical_locations,
const NumPtr<REAL> &weights,
const NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &transformed_bc_face_values,
const HcurlBoundaryCondition *bc,
map<int,REAL> &boundary_values) const {
  HcurlDualTensorShapeFunction *sff=shapeFunctionFamily(element);
# if (SPACEDIM==2)
  NumPtr<NumPtr<REAL> > dof_poly_values(canonical_locations.getNumber());
# else
  NumPtr<NumPtr<Tensor<1,2> > >
    dof_poly_values(canonical_locations.getNumber());
# endif
  int nv=element->numberVertices();
#if (SPACEDIM==2)
  Tensor<1,SPACEDIM> face_tangent=lineTangent(nv,f);
  if (element->lineTangentPointsBackward(f)) face_tangent=-face_tangent;
#else
  bool ftwb=element->faceTangentsWindBackward(f);
  Tensor<1,3> ft0=element->faceTangent(f,(ftwb ? 1 : 0));
  Tensor<1,3> ft1=element->faceTangent(f,(ftwb ? 0 : 1));
#endif
  int nfv=element->numberVerticesOnFace(f);
  for (int q=0;q<canonical_locations.getNumber();q++) {
# if (SPACEDIM==2)
    sff->computeEdgeValues(canonical_locations[q],dof_poly_values[q]);
# else
    sff->computeFaceValues(nfv,canonical_locations[q],dof_poly_values[q]);
#endif
  }
  int shape_fcns_per_face=face_dofs.getNumber()/components;
  for (int s=0;s<shape_fcns_per_face;s++) {
    int sc=s*components;
    for (int c=0;c<components;c++) {
      if (bc->isEssential(c)) {
        REAL face_dof_value=0.;
        for (int q=0;q<canonical_locations.getNumber();q++) {
#if (SPACEDIM==2)
          face_dof_value+=dof_poly_values[q][s]*(face_tangent
            *transformed_bc_face_values[q][c])*weights[q];
#else
          Tensor<1,3> tens=ft0*dof_poly_values[q][s][0]
                          +ft1*dof_poly_values[q][s][1];
          face_dof_value+=(tens*transformed_bc_face_values[q][c])
                         *weights[q];
#endif
        }
        boundary_values[face_dofs[sc+c]]=face_dof_value;
      }
    }
  }
}
#endif
#if (SPACEDIM==3)
void HcurlFiniteElement::setLineBoundaryValues(const SHAPE_CLASS *element,
const CODIMENSION1_SHAPE_CLASS *face_element,int f,int fl,
const NumPtr<int> &line_dofs,const NumPtr<Point<1> > &canonical_locations,
const NumPtr<REAL> &weights,
const NumPtr<NumPtr<Tensor<1,3> > > &transformed_bc_face_values,
const HcurlBoundaryCondition *bc,map<int,REAL> &boundary_values) const {
  HcurlDualTensorShapeFunction *sff=shapeFunctionFamily(element);
  NumPtr<NumPtr<REAL> > 
    dof_poly_values(canonical_locations.getNumber());
  int nv=element->numberVertices();
  int l=element->lineFromFace(face_element,f,fl);
  Tensor<1,3> line_tangent=lineTangent(nv,l);
  if (element->lineTangentPointsBackward(l)) line_tangent=-line_tangent;
  for (int q=0;q<canonical_locations.getNumber();q++) {
    sff->computeEdgeValues(canonical_locations[q],dof_poly_values[q]);
  }
  int shape_fcns_per_line=line_dofs.getNumber()/components;
  for (int s=0;s<shape_fcns_per_line;s++) {
    int sc=s*components;
    for (int c=0;c<components;c++) {
      if (bc->isEssential(c)) {
        REAL line_dof_value=0.;
        for (int q=0;q<canonical_locations.getNumber();q++) {
          line_dof_value+=dof_poly_values[q][s]*(line_tangent
            *transformed_bc_face_values[q][c])*weights[q];
        }
        boundary_values[line_dofs[sc+c]]=line_dof_value;
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::equallySpacedLatticePointDataBase(
map<int,HcurlFiniteElementDataBase*> &fedb_map,
map<int,Quadrature<SPACEDIM>*> &quad_map) const {
  for (map<int,HcurlDualTensorShapeFunction*>::const_iterator
  sf_iterator=polynomial_space.begin();
  sf_iterator!=polynomial_space.end();sf_iterator++) {
    int nv=sf_iterator->first;
    NumPtr<Point<SPACEDIM> > element_lattice_points;
    equallySpacedLatticePoints(nv,degree,element_lattice_points);
    Quadrature<SPACEDIM> *quad=quad_map[nv]=
      OPERATOR_NEW Quadrature<SPACEDIM>(element_lattice_points);
    HcurlFiniteElementDataBase *fedb=getFEData(nv,UPDATE_VALUES,quad);
    fedb_map[nv]=fedb;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
HcurlFiniteElementDataBase* HcurlFiniteElement::fixShapeFunctionsFor(
const Shape *e,HcurlFiniteElementDataBase *fedb) const {
  TensorShapeFunction<SPACEDIM> *sf=shapeFunctionFamily(e);
  if (sf->requiresShapeToComputeValues()) {
    computeShapeFunctionCoefficientsFor(e);
    fedb->computeShapesFor(e);
    return fedb;
  } else {
    return OPERATOR_NEW HcurlFiniteElementDataBase(fedb,e);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::printPolynomialSpaces(ostream &os) const {
  for (map<int,HcurlDualTensorShapeFunction*>::const_iterator
  ps_iterator=polynomial_space.begin();
  ps_iterator!=polynomial_space.end();ps_iterator++) {
    os << "\tpolynomial_space[" << ps_iterator->first << "] = "
       << ps_iterator->second << endl;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlFiniteElement::printOn(ostream &os) const {
  os << "\tHcurlFiniteElement: degree = " << degree << endl;
  for (map<int,HcurlDualTensorShapeFunction*>::const_iterator
  ps_iterator=polynomial_space.begin();
  ps_iterator!=polynomial_space.end();ps_iterator++) {
    os << "\tpolynomial_space[" << ps_iterator->first << "]: " << endl;
    if (ps_iterator->second!=0) ps_iterator->second->printOn(os);
  }
  for (map<int,HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE*>::const_iterator
  ps_iterator=multiplier_polynomial_space.begin();
  ps_iterator!=multiplier_polynomial_space.end();ps_iterator++) {
    os << "\tpolynomial_space[" << ps_iterator->first << "]: " << endl;
    if (ps_iterator->second!=0) ps_iterator->second->printOn(os);
  }
  for (map<int,int>::const_iterator dps_iterator=
  dofs_per_shape.begin();dps_iterator!=dofs_per_shape.end();
  dps_iterator++) {
    os << "\tdofs_per_shape[" << dps_iterator->first << "] = "
       << dps_iterator->second << endl;
  }
  for (map<int,int>::const_iterator mdps_iterator=
  multiplier_dofs_per_shape.begin();mdps_iterator!=
  multiplier_dofs_per_shape.end();mdps_iterator++) {
    os << "\tmultiplier_dofs_per_shape[" << mdps_iterator->first
       << "] = " << mdps_iterator->second << endl;
  }
#if (SPACEDIM>1)
  for (map<int,int>::const_iterator
  dpcs_iterator=dofs_per_codimension1_shape.begin();dpcs_iterator!=
  dofs_per_codimension1_shape.end();dpcs_iterator++) {
    os << "\tdofs_per_codimension1_shape[" << dpcs_iterator->first
       << "] = " << dpcs_iterator->second << endl;
  }
#endif
  for (map<int,int>::const_iterator dpc_iterator=
  dofs_per_cell.begin();dpc_iterator!=
  dofs_per_cell.end();dpc_iterator++) {
    os << "\tdofs_per_cell[" << dpc_iterator->first << "] = "
       << dpc_iterator->second << endl;
  }
  os << "\tcomponents = " << components << endl;
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1FiniteElement::Nedelec1FiniteElement(int d,int nc) :
HcurlFiniteElement(d,nc) {
#if (SPACEDIM==1)
  int nv=Interval::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec1SimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
#elif (SPACEDIM==2)
  int nv=Quadrilateral::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec1ParallelpipedDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree);

  nv=Triangle::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec1SimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);

  fqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree+3));
#else
  int nv=Hexahedron::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec1ParallelpipedDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW HcurlMultiplierHexahedronShapeFunction(degree,degree);

//current Nedelec1PrismDualShapeFunction does not necessarily have
//  continuous tangential components on shared element faces for degree>=2
  nv=Prism::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec1PrismDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW HcurlMultiplierPrismShapeFunction(degree,degree+1);

  nv=Tetrahedron::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec1SimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW HcurlMultiplierTetrahedronShapeFunction(degree);

  fqmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
    TriangleGaussianQuadrature::argGivenOrder(2*degree+3));
  fqmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
    GaussianQuadrature<2>::argGivenOrder(2*degree+3));
  edge_quadrature=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree+3));
#endif
  setDofsPerObject();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1FiniteElement::printOn(ostream &os) const {
  os << "\tNedelec1FiniteElement:" << endl;
  HcurlFiniteElement::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2FiniteElement::Nedelec2FiniteElement(int d,int nc) :
HcurlFiniteElement(d,nc) {
  CHECK_POSITIVE(degree);
#if (SPACEDIM==1)
  int nv=Interval::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec2SimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeIntervalPolynomials(degree-1);
#elif (SPACEDIM==2)
  int nv=Quadrilateral::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec2ParallelpipedDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree);

  nv=Triangle::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec2SimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTrianglePolynomials(degree-1);

  fqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree+3));
#else
  int nv=Hexahedron::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec2ParallelpipedDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW HcurlMultiplierHexahedronShapeFunction(degree,degree+1);

//current Nedelec2PrismDualShapeFunction does not necessarily have
//  continuous tangential components on shared element faces for degree>=2
  nv=Prism::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec2PrismDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW HcurlMultiplierPrismShapeFunction(degree,degree);

//current Nedelec2SimplexDualShapeFunction does not necessarily have
//  continuous tangential components on shared element faces for degree>=3
  nv=Tetrahedron::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW Nedelec2SimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW HcurlMultiplierTetrahedronShapeFunction(degree-1);

  fqmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
    TriangleGaussianQuadrature::argGivenOrder(2*degree+3));
  fqmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
    GaussianQuadrature<2>::argGivenOrder(2*degree+3));
#endif
  setDofsPerObject();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2FiniteElement::printOn(ostream &os) const {
  os << "\tNedelec2FiniteElement:" << endl;
  HcurlFiniteElement::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template class _Rb_tree<
  int,
  pair<int const, HcurlDualTensorShapeFunction*>,
  _Select1st<pair<int const, HcurlDualTensorShapeFunction*> >,
  less<int>,
  allocator<pair<int const, HcurlDualTensorShapeFunction*> >
>;
template class _Rb_tree<
  int,
  pair<int const, HcurlFiniteElementDataBase*>,
  _Select1st<pair<int const, HcurlFiniteElementDataBase*> >,
  less<int>,
  allocator<pair<int const, HcurlFiniteElementDataBase*> >
>;
template class _Rb_tree<
  int,
  pair<int const, TensorShapeFunction<3>*>,
  _Select1st<pair<int const, TensorShapeFunction<3>*> >,
  less<int>,
  allocator<pair<int const, TensorShapeFunction<3>*> >
>;
