#ifndef HCURL_FINITE_ELEMENT_H
#define HCURL_FINITE_ELEMENT_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <BoundaryCondition.H>
#include <Mapping.H> // for UPDATE_FLAGS, not Mapping
#include <Matrix.H>
#include <TensorShapeFunction.H>

#if (SPACEDIM<=2)
typedef REAL HCURL_MULTIPLIER_DATA_TYPE;
typedef ShapeFunction<SPACEDIM> HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE;
#else
typedef Tensor<1,3> HCURL_MULTIPLIER_DATA_TYPE;
typedef TensorShapeFunction<3> HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE;
#endif

// data for finite element shape function(xi)
class HcurlFiniteElementDataBase {
  private:
    UPDATE_FLAGS update_flags;
    bool shape_values_ok;
    bool shape_derivatives_ok;
    bool shape_second_derivatives_ok;
//  shape_values(s,q) = shape_s(xi_q); ie, defined on canonical elements
    Array<2,Tensor<1,SPACEDIM> > shape_values;
    Array<2,Tensor<2,SPACEDIM> > shape_derivatives;
    Array<2,Tensor<3,SPACEDIM> > shape_second_derivatives;
    const TensorShapeFunction<SPACEDIM> *shape_function_family;

    bool multiplier_shape_values_ok;
    bool multiplier_shape_derivatives_ok;
    bool multiplier_shape_second_derivatives_ok;
    Array<2,HCURL_MULTIPLIER_DATA_TYPE> multiplier_shape_values;
    const HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE
      *multiplier_shape_function_family;

    const Quadrature<SPACEDIM> *quadrature;

    HcurlFiniteElementDataBase(const HcurlFiniteElementDataBase&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    HcurlFiniteElementDataBase(UPDATE_FLAGS,const Quadrature<SPACEDIM>*,
      const TensorShapeFunction<SPACEDIM>*,
      const HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE*);
    HcurlFiniteElementDataBase(const HcurlFiniteElementDataBase*,
      const Shape*);
    virtual ~HcurlFiniteElementDataBase() {;}
    UPDATE_FLAGS updateFlags() const { return update_flags; }
    const Quadrature<SPACEDIM> *getQuadrature() const {
      return quadrature;
    }
    bool shapeValuesOK() const { return shape_values_ok; }
    bool shapeDerivativesOK() const { return shape_derivatives_ok; }
    bool shapeSecondDerivativesOK() const { 
      return shape_second_derivatives_ok; 
    }
    const Tensor<1,SPACEDIM>& shapeValue(int s,int q) const { 
      CHECK_TEST(shape_values_ok);
      return shape_values(s,q); 
    }
    const Tensor<2,SPACEDIM>& shapeDerivative(int s,int q) const {
      CHECK_TEST(shape_derivatives_ok);
      return shape_derivatives(s,q);
    }
    const Tensor<3,SPACEDIM>& shapeSecondDerivative(int s,int q) const {
      CHECK_TEST(shape_second_derivatives_ok);
      return shape_second_derivatives(s,q);
    }
    bool multiplierShapeValuesOK() const { 
      return multiplier_shape_values_ok; 
    }
//  called from HdivFiniteElement::setMultiplierValues
    const HCURL_MULTIPLIER_DATA_TYPE& multiplierShapeValue(int s,int q)
    const { 
      CHECK_TEST(multiplier_shape_values_ok);
      return multiplier_shape_values(s,q); 
    }
    virtual void printOn(ostream&) const;
    void computeShapesFor(const Shape*);
};

class HcurlFiniteElement {
  protected:
    int degree;
    map<int,HcurlDualTensorShapeFunction*> polynomial_space;
    map<int,HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE*>
      multiplier_polynomial_space;
#if (SPACEDIM>1)
    map<int,Quadrature<SPACEDIM-1> > fqmap;
    map<int,Quadrature<SPACEDIM-1> > fbvmap;
#endif
#if (SPACEDIM==3)
    Quadrature<1> edge_quadrature;
#endif

//  the following = components * shape_functions per whatever
    map<int,int> dofs_per_shape;
    map<int,int> dofs_per_codimension1_shape;
#if (SPACEDIM==3)
    int dofs_per_line;
#endif
    map<int,int> dofs_per_cell; // sum over dimension and codimensions
    map<int,int> multiplier_dofs_per_shape;
    int components;

    int interfaceConstraintsSize(const Shape*,int) const;
//  call from derived class constructor after setting 
//    dofs_per_vertex,dofs_per_line,dofs_per_quad,dofs_per_hex
    virtual void setDofsPerObject();
  public:
    HcurlFiniteElement(int d,int nc) : components(nc),degree(d) { }
    virtual ~HcurlFiniteElement();

    virtual int dofsPerShape(int nv) const {
//    nv=2 in 1D, 3 or 4 in 2D, and 4 or 6 or 8 in 3D
      map<int,int>::const_iterator i=dofs_per_shape.find(nv);
      CHECK_TEST(i!=dofs_per_shape.end());
      return i->second;
    }
    virtual int dofsPerShape(const Shape *e) const {
      CHECK_SAME(e->numberDimensions(),SPACEDIM);
      return dofsPerShape(e->numberVertices());
    }
#if (SPACEDIM==3)
    int dofsPerLine() const { return dofs_per_line; }
#endif
    int dofsPerFace(int nv) const { 
//    nv=1 in 1D, 2 in 2D, and 3 or 4 in 3D
      map<int,int>::const_iterator i=
        dofs_per_codimension1_shape.find(nv);
      CHECK_TEST(i!=dofs_per_codimension1_shape.end());
      return i->second;
    }
    int dofsPerFace(const Shape *e) const { 
      CHECK_SAME(e->numberDimensions(),SPACEDIM-1);
      return dofsPerFace(e->numberVertices());
    }
    int dofsPerFace(const Shape *e,int f) const { 
      CHECK_SAME(e->numberDimensions(),SPACEDIM);
#if (SPACEDIM==1)
      return dofsPerFace(1);
#else
      return dofsPerFace(dynamic_cast<const Shape2*>(e)->
                         numberVerticesOnFace(f));
#endif
    }
    virtual int dofsPerCell(int nv) const { // dofs/shape + dofs/face
      map<int,int>::const_iterator i=dofs_per_cell.find(nv);
      CHECK_TEST(i!=dofs_per_cell.end());
      return i->second;
    }
    virtual int dofsPerCell(const Shape *e) const {
      CHECK_SAME(e->numberDimensions(),SPACEDIM);
      return dofsPerCell(e->numberVertices());
    }
    virtual int maxDofsPerCell() const;
    virtual int maxMultiplierDofsPerShape() const;
    const map<int,int>& dofsPerCellMap() const {
      return dofs_per_cell;
    }
    virtual int multiplierDofsPerShape(int nv) const {
      map<int,int>::const_iterator i=multiplier_dofs_per_shape.find(nv);
      CHECK_TEST(i!=dofs_per_shape.end());
      return i->second;
    }
    virtual int multiplierDofsPerShape(const Shape *e) const {
     CHECK_SAME(e->numberDimensions(),SPACEDIM);
     return multiplierDofsPerShape(e->numberVertices());
    }

    int nComponents() const { return components; }
//  called from FEValuesData::FEValuesData
    virtual int nNonzeroShapeComponents(int nv) const {
      return dofsPerCell(nv)/components;
    }
    inline int nNonzeroShapeComponents(const Shape *e) const {
      return dofsPerCell(e)/components;
    }
    virtual int nNonzeroMultiplierShapeComponents(int nv) const {
      return multiplierDofsPerShape(nv)/components;
    }
    inline int nNonzeroMultiplierShapeComponents(const Shape *e) const {
      return multiplierDofsPerShape(e)/components;
    }
    int getDegree() const { return degree; }
//  public for testFE only
    HcurlDualTensorShapeFunction* shapeFunctionFamily(int nv) const {
      map<int,HcurlDualTensorShapeFunction*>::
        const_iterator ps_iterator=polynomial_space.find(nv);
      CHECK_TEST(ps_iterator!=polynomial_space.end());
      return ps_iterator->second;
    }
    HcurlDualTensorShapeFunction* shapeFunctionFamily(const Shape *e)
    const {
      CHECK_SAME(e->numberDimensions(),SPACEDIM);
      return shapeFunctionFamily(e->numberVertices());
    }
    HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE* multiplierShapeFunctionFamily(
    int nv) const {
      map<int,HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE*>::const_iterator
        ps_iterator=multiplier_polynomial_space.find(nv);
      CHECK_TEST(ps_iterator!=multiplier_polynomial_space.end());
      return ps_iterator->second;
    }
    HCURL_MULTIPLIER_SHAPE_FUNCTION_TYPE* multiplierShapeFunctionFamily(
    const Shape *e) const {
      CHECK_SAME(e->numberDimensions(),SPACEDIM);
      return multiplierShapeFunctionFamily(e->numberVertices());
    }
    void computeShapeFunctionCoefficientsFor(const Shape *e) const {
      shapeFunctionFamily(e)->computeCoefficientsFor(e);
    }
    virtual HcurlFiniteElementDataBase* getFEData(int nv,
    UPDATE_FLAGS flags,const Quadrature<SPACEDIM> *quadrature) const {
      const TensorShapeFunction<SPACEDIM> *sff=shapeFunctionFamily(nv);
      if (sff!=0) {
        return OPERATOR_NEW HcurlFiniteElementDataBase(flags,quadrature,
          shapeFunctionFamily(nv),multiplierShapeFunctionFamily(nv));
      }
      return 0;
    }
#if (SPACEDIM>1)
    const Quadrature<SPACEDIM-1>& faceQuadrature(int nv) const {
      map<int,Quadrature<SPACEDIM-1> >::
        const_iterator fq_iterator=fqmap.find(nv);
      CHECK_TEST(fq_iterator!=fqmap.end());
      return fq_iterator->second;
    }
    const Quadrature<SPACEDIM-1>& faceQuadrature(const SHAPE_CLASS *e)
    const {
      return faceQuadrature(e->numberVertices());
    }
    map<int,Quadrature<SPACEDIM-1> >& faceQuadratureMap() {
      return fqmap;
    }
    const Quadrature<SPACEDIM-1>& boundaryValueQuadrature(int nv) const {
      map<int,Quadrature<SPACEDIM-1> >::
        const_iterator fq_iterator=fbvmap.find(nv);
      CHECK_TEST(fq_iterator!=fbvmap.end());
      return fq_iterator->second;
    }
    const Quadrature<SPACEDIM-1>& boundaryValueQuadrature(
    const SHAPE_CLASS *e) const {
      return boundaryValueQuadrature(e->numberVertices());
    }
    map<int,Quadrature<SPACEDIM-1> >& boundaryValueQuadratureMap() {
      return fbvmap;
    }
#endif
#if (SPACEDIM==3)
    const Quadrature<1>& edgeQuadrature() const {
      return edge_quadrature;
    }
#endif

    virtual HcurlFiniteElementDataBase* fixShapeFunctionsFor(const Shape*,
      HcurlFiniteElementDataBase*) const;

//  called from HcurlFEValues::getFunctionValues,
//    HcurlFEFaceValues::getFunctionValues,
    virtual void sumShapeValues(const SHAPE_CLASS*,
      const Array<2,Tensor<1,SPACEDIM> >&,int,int,const Vector&,
      NumPtr<Tensor<1,SPACEDIM> >&) const;
    virtual void sumShapeValues(const SHAPE_CLASS*,
      const Array<2,Tensor<1,SPACEDIM> >&,int,int,const Vector&,
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
//  gradients[q][d], 0<=q<nqp, 0<=d<SPACEDIM
    virtual void sumShapeDerivatives(const SHAPE_CLASS*,
      const Array<2,Tensor<2,SPACEDIM> >&,int,int,const Vector&,
      NumPtr<Tensor<2,SPACEDIM> >&) const;
//  gradients[q][c][d], 0<=q<nqp, 0<=c<components, 0<=d<SPACEDIM
    virtual void sumShapeDerivatives(const SHAPE_CLASS*,
      const Array<2,Tensor<2,SPACEDIM> >&,int,int,const Vector&,
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > >&) const;
#if (SPACEDIM<=2)
    virtual void sumShapeCurl(const SHAPE_CLASS*,const Array<2,REAL>&,
      int,int,const Vector&,NumPtr<REAL>&) const;
    virtual void sumShapeCurl(const SHAPE_CLASS*,const Array<2,REAL>&,
      int,int,const Vector&,NumPtr<NumPtr<REAL> >&) const;
#else
    virtual void sumShapeCurl(const SHAPE_CLASS*,
      const Array<2,Tensor<1,3> >&,int,int,const Vector&,
      NumPtr<Tensor<1,3> >&) const;
    virtual void sumShapeCurl(const SHAPE_CLASS*,
      const Array<2,Tensor<1,3> >&,int,int,const Vector&,
      NumPtr<NumPtr<Tensor<1,3> > >&) const;
#endif
    virtual void sumMultiplierShapeValues(const SHAPE_CLASS*,
      const Array<2,HCURL_MULTIPLIER_DATA_TYPE>&,int,int,const Vector&,
      NumPtr<HCURL_MULTIPLIER_DATA_TYPE>&) const;
    virtual void sumMultiplierShapeValues(const SHAPE_CLASS*,
      const Array<2,HCURL_MULTIPLIER_DATA_TYPE>&,int,int,const Vector&,
      NumPtr<NumPtr<HCURL_MULTIPLIER_DATA_TYPE> >&) const;
#if (SPACEDIM>=2)
    virtual void setFaceBoundaryValues(const SHAPE_CLASS*,int f,
      const NumPtr<int>&,const NumPtr<Point<SPACEDIM-1> >&,
      const NumPtr<REAL>&,const NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&,
      const HcurlBoundaryCondition*,map<int,REAL>&) const;
#endif
#if (SPACEDIM==3)
    virtual void setLineBoundaryValues(const SHAPE_CLASS*,
      const CODIMENSION1_SHAPE_CLASS*,int,int,const NumPtr<int>&,
      const NumPtr<Point<1> >&,const NumPtr<REAL>&,
      const NumPtr<NumPtr<Tensor<1,3> > >&,const HcurlBoundaryCondition*,
      map<int,REAL>&) const;
#endif
    void equallySpacedLatticePointDataBase(
      map<int,HcurlFiniteElementDataBase*>&,
      map<int,Quadrature<SPACEDIM>*>&)const;
    virtual void printPolynomialSpaces(ostream&) const;
    virtual void printOn(ostream&) const;
};

class Nedelec1FiniteElement : public HcurlFiniteElement {
  protected:    
  public:
    Nedelec1FiniteElement(int,int=1);
    ~Nedelec1FiniteElement() {;}
    void printOn(ostream&) const;
};

class Nedelec2FiniteElement : public HcurlFiniteElement {
  protected:    
  public:
    Nedelec2FiniteElement(int,int=1);
    ~Nedelec2FiniteElement() {;}
    void printOn(ostream&) const;
};

#endif
