#ifndef HCURL_PATCH_H
#define HCURL_PATCH_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Array.H"
#include "HcurlDofHandler.H"
#include "Mapping.H"
#include "NumPtr.H"
#include "Point.H"
#include "Quadrature.H"
#include "Shape.H"
#include "XYGraphTool.H"

class HcurlPatch {
  protected:
    const SHAPE_CLASS *shape;
    const LEVEL_CLASS *level;
    NumPtr<Point<SPACEDIM> > ep;
    NumPtr<int> neighbors;
    int patch_index;
    Array<2,Tensor<1,SPACEDIM> > data;
    Array<2,HCURL_MULTIPLIER_DATA_TYPE> multiplier_data;

    static int n_subdivisions;
    static map<int,Quadrature<SPACEDIM> > q_map;
    const Quadrature<SPACEDIM>& getQuadrature() const;
# if (SPACEDIM==3)
    void plotClipPlane(int d,VolGraphTool&,
      const HCURL_MULTIPLIER_DATA_TYPE&,
      const HCURL_MULTIPLIER_DATA_TYPE&,
      const REAL value[2][2][2],const Vector3 vertex[2][2][2],
      const HcurlDofHandler*,const Mapping*,const Vector*) const;
    void plotTetrahedralClipPlane(int d,VolGraphTool&,
      const HCURL_MULTIPLIER_DATA_TYPE&,
      const HCURL_MULTIPLIER_DATA_TYPE&,REAL,int,HAND,
      const HcurlDofHandler*,const Mapping*,const Vector*,
      const Point<SPACEDIM> x[4]) const;
    void plotPrismClipPlane(int d,VolGraphTool&,
      const HCURL_MULTIPLIER_DATA_TYPE&,
      const HCURL_MULTIPLIER_DATA_TYPE&,REAL,int,HAND,
      const HcurlDofHandler*,const Mapping*,const Vector*,
      const Point<SPACEDIM> x[6]) const;
# endif
  public:
    static const int no_neighbor;
    static void setQuadratures(int);
    static const map<int,Quadrature<SPACEDIM> >& 
      getQuadratureMap() { return q_map; }

    HcurlPatch(const SHAPE_CLASS*,const LEVEL_CLASS*,const Tessellation&,
      const Mapping&,int,int);
    virtual ~HcurlPatch();
    int numberQuadraturePoints() const {
      return getQuadrature().numberPoints();
    }
    Tensor<1,SPACEDIM> getData(int dataset,int q_point) const {
      return data(dataset,q_point);
    }
    void setData(int dataset,const Tensor<1,SPACEDIM> &val) {
      for (int q=0;q<data.size(1);q++) {
        data(dataset,q)=val;
      }
    }
    void setData(int dataset,const NumPtr<Tensor<1,SPACEDIM> > &values) {
      for (int q=0;q<values.getNumber();q++) {
        data(dataset,q)=values[q];
      }
    }
    void setData(int dataset,int n_components,
    const NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &values) {
      int dc=dataset*n_components;
      for (int c=0;c<n_components;c++) {
        for (int q=0;q<values.getNumber();q++) {
          data(dc+c,q)=values[q][c];
        }
      }
    }
    void setMultiplierData(int dataset,HCURL_MULTIPLIER_DATA_TYPE &val) {
      for (int q=0;q<data.size(1);q++) multiplier_data(dataset,q)=val;
    }
    void setMultiplierData(int dataset,
    const NumPtr<HCURL_MULTIPLIER_DATA_TYPE> &values) {
      for (int q=0;q<values.getNumber();q++) {
        multiplier_data(dataset,q)=values[q];
      }
    }
    void setMultiplierData(int dataset,int n_components,
    const NumPtr<NumPtr<HCURL_MULTIPLIER_DATA_TYPE> > &values) {
      int dc=dataset*n_components;
      for (int c=0;c<n_components;c++) {
        for (int q=0;q<values.getNumber();q++) {
          multiplier_data(dc+c,q)=values[q][c];
        }
      }
    }
    int dataSize(int i) const { return data.size(i); }
    int multiplierDataSize(int i) const { return multiplier_data.size(i); }
    void setNeighbor(int f,int n) {
      neighbors[f]=n;
    }
    int nSubdivisions() const { return n_subdivisions; }
    void getMeshBounds(Point<SPACEDIM>&,Point<SPACEDIM>&) const;
    void getDataBounds(int,Tensor<1,SPACEDIM>&,Tensor<1,SPACEDIM>&) const;
    void getMultiplierDataBounds(int,HCURL_MULTIPLIER_DATA_TYPE&,
      HCURL_MULTIPLIER_DATA_TYPE&) const;
#if (SPACEDIM==1)
    void plot(XYGraphTool*,int) const;
    void plotMultipliers(XYGraphTool*,int) const;
#endif
#if (SPACEDIM==2)
    void plot(AreaGraphTool*,int,const Tensor<1,2>&,
      const Tensor<1,2>&) const;
    void plotMultipliers(AreaGraphTool*,int,
      const HCURL_MULTIPLIER_DATA_TYPE&,
      const HCURL_MULTIPLIER_DATA_TYPE&) const;
    void plotSurface(int,VolGraphTool*,int,REAL,REAL,BOOLEAN) const;
    void plotMultipliersSurface(VolGraphTool*,int,
      HCURL_MULTIPLIER_DATA_TYPE,HCURL_MULTIPLIER_DATA_TYPE,BOOLEAN) const;
    void draw(AreaGraphTool&) const;
    void multiplierContour(AreaGraphTool*,
      const HCURL_MULTIPLIER_DATA_TYPE&,int) const;
#endif
#if (SPACEDIM==3)
    void plotClipPlane(int,VolGraphTool&,const HCURL_MULTIPLIER_DATA_TYPE&,
      const HCURL_MULTIPLIER_DATA_TYPE&,const REAL,int,HAND,
      const HcurlDofHandler*,const Mapping*,const Vector*) const;
    void plotSurface(int,VolGraphTool&,int,REAL,REAL,REAL,BOOLEAN) const;
    void draw(VolGraphTool&) const;
    void drawVectors(VolGraphTool*,int,const Vector3&,const Vector3&,
      const Tensor<1,3>&,const Tensor<1,3>&) const;
#endif
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM==3)
void drawSurfaceWithinCube(VolGraphTool&,REAL,REAL value[2][2][2],
  Vector3 vertex[2][2][2]);
#endif

#endif
