// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <HdivDataOut.H>
#include <HdivFEValues.H>
#include <Surface.H>
#include <Tracer.H>
#include <TypedPlotObj.H>
#include <Vector.H>
#include <VolGraphTool.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::clearDataVectors() {
  for (int i=0;i<dof_data.getNumber();i++) {
    delete dof_data[i]; dof_data[i]=0;
  }
  dof_data.cleanup();
  for (int i=0;i<cell_data.getNumber();i++) {
    delete cell_data[i]; cell_data[i]=0;
  }
  cell_data.cleanup();
  for (int i=0;i<patches.getNumber();i++) {
    delete patches[i]; patches[i]=0;
  }
  patches.cleanup();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::addDataVector(const Vector &vec,
const NumPtr<string> &new_names,DATA_VECTOR_TYPE type) {
  for (int i=0;i<new_names.getNumber();i++) {
    CHECK_SAME(new_names[i].
      find_first_not_of("abcdefghijklmnopqrstuvwxyz"
      "ABCDEFGHIJKLMNOPQRSTUVWXYZ" "0123456789_<>()"),string::npos);
  }
  DataEntry *new_entry=OPERATOR_NEW DataEntry(&vec,new_names);
  switch (type) {
    case TYPE_CELL_DATA:
      CHECK_SAME(vec.size(),
                 dofs->getTessellation().numberActiveShapes());
      CHECK_SAME(new_names.getNumber(),1);
      break;
    case TYPE_DOF_DATA:
      CHECK_SAME(vec.size(),dofs->numberDofs());
      CHECK_SAME(new_names.getNumber(),dofs->getFe().nComponents());
      break;
    default:
      if (vec.size()==dofs->getTessellation().numberActiveShapes()) {
        type=TYPE_CELL_DATA;
        CHECK_SAME(new_names.getNumber(),1);
      } else {
        type=TYPE_DOF_DATA;
        CHECK_SAME(new_names.getNumber(),dofs->getFe().nComponents());
      }
  }
  if (type==TYPE_DOF_DATA) {
    NumPtr<DataEntry*> new_dof_data(dof_data.getNumber()+1);
    new_dof_data.copyFrom(dof_data);
    new_dof_data[dof_data.getNumber()]=new_entry;
    dof_data.copy(new_dof_data);
  } else {
    NumPtr<DataEntry*> new_cell_data(cell_data.getNumber()+1);
    new_cell_data.copyFrom(cell_data);
    new_cell_data[cell_data.getNumber()]=new_entry;
    cell_data.copy(new_cell_data);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::addDataVector(const Vector &vec,const string &name,
DATA_VECTOR_TYPE type) {
  int n_components=dofs->getFe().nComponents();
  NumPtr<string> new_names;
  if ((n_components==1) ||
  (vec.size()==dofs->getTessellation().numberActiveShapes())) {
    new_names.allocate(1);
    new_names[0]=name;
  } else {
    new_names.allocate(n_components);
    for (int i=0; i<n_components; ++i) {
      ostringstream namebuf;
      namebuf << '_' << i;
      namebuf << ends;
      new_names[i]=name+namebuf.str();
    }
  }
  addDataVector(vec,new_names,type);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::clearInputDataReferences() {
  for (int i=0;i<dof_data.getNumber();i++) {
    dof_data[i]->clearPointers();
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    cell_data[i]->clearPointers();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::getDatasetNames(NumPtr<string> &names_out) const {
  int n=0;
  for (int i=0;i<dof_data.getNumber();i++) {
    n+=dof_data[i]->getNames().getNumber();
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    n+=cell_data[i]->getNames().getNumber();
  }
  names_out.allocate(n);
  n=0;
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_names=dof_data[i]->getNames();
    for (int j=0;j<dof_names.getNumber();j++,n++) {
      names_out[n]=dof_names[j];
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_names=cell_data[i]->getNames();
    for (int j=0;j<cell_names.getNumber();j++,n++) {
      names_out[n]=cell_names[j];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::buildPatches(int n_subdivisions) {
  HdivPatch::setQuadratures(n_subdivisions);
  CHECK_POINTER(dofs);
  patches.cleanup();
  Tessellation &tess=dofs->getTessellation();
  int n_levels=tess.numberLevels();
  NumPtr<NumPtr<int> > cell_to_patch_index_map(n_levels);
  int l=0;
  for (LEVEL_CLASS *level=tess.coarsestLevel();level;
  level=level->finerLevel(),l++) {
    int max_index=0;
    for (bool doit=tess.firstActiveShape(level);doit;
    doit=tess.nextActiveShapeInLevel()) {
      max_index=max(max_index,level->presentIndex());
    }
    cell_to_patch_index_map[l].allocate(max_index+1);
    cell_to_patch_index_map[l].initialize(HdivPatch::no_neighbor);
  }
  int n_patches=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),n_patches++) {
    LEVEL_CLASS *level=tess.presentLevel();
    cell_to_patch_index_map[level->numberCoarserLevels()]
      [level->presentIndex()]=n_patches;
  }
  int n_components=dofs->getFe().nComponents();
  int n_datasets=
    dof_data.getNumber()*n_components+cell_data.getNumber();
  patches.allocate(n_patches);
  int p=0;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),p++) {
    const SHAPE_CLASS *e=tess.presentShape();
    patches[p]=OPERATOR_NEW HdivPatch(e,tess.presentLevel(),tess,
      *mapping,n_datasets,p);
  }

  static const MappingL mapping_l;
  UPDATE_FLAGS flags=
    UPDATE_VALUES | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  HdivFEValues fe_patch_values(mapping_l,dofs->getFe(),
    HdivPatch::getQuadratureMap(),flags);

  int cell_number=0;
  p=0;
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),p++) {
    const SHAPE_CLASS *shape=tess.presentShape();
    LEVEL_CLASS *present_level=tess.presentLevel();
    HdivPatch *patch=patches[p];
    if (n_datasets>0) {
      fe_patch_values.mapShapes(tess,shape,present_level);
      int n_q_points=patch->numberQuadraturePoints();
      for (int dataset=0;dataset<dof_data.getNumber();dataset++) {
        if (n_components==1) {
          NumPtr<Tensor<1,SPACEDIM> > patch_values;
          dofs->getFunctionValues(fe_patch_values,shape,present_level,
            *dof_data[dataset]->singleData(),patch_values);
          patch->setData(dataset,patch_values);
          NumPtr<REAL> patch_multiplier_values;
          dofs->getMultiplierFunctionValues(fe_patch_values,shape,
            present_level,*dof_data[dataset]->singleData(),
            patch_multiplier_values);
          patch->setMultiplierData(dataset,patch_multiplier_values);
        } else {
          NumPtr<NumPtr<Tensor<1,SPACEDIM> > > patch_values_system;
          dofs->getFunctionValues(fe_patch_values,shape,present_level,
              *dof_data[dataset]->singleData(),patch_values_system);
          patch->setData(dataset,n_components,patch_values_system);
          NumPtr<Vector> patch_multiplier_values_system;
          dofs->getMultiplierFunctionValues(fe_patch_values,shape,
            present_level,*dof_data[dataset]->singleData(),
            patch_multiplier_values_system);
          patch->setMultiplierData(dataset,n_components,
            patch_multiplier_values_system);
        }
      }
      int dnc=dof_data.getNumber()*n_components;
      for (int dataset=0;dataset<this->cell_data.getNumber();dataset++)
      {
        REAL value=(*cell_data[dataset]->singleData())[cell_number];
        patch->setMultiplierData(dataset+dnc,value);
      }
    }
    for (int f=0;f<shape->numberCodimension1Shapes();f++) {
      int neighbor_index=shape->codimension1ShapeIndex(f);
      LEVEL_CLASS *neighbor_level=present_level->neighborLevel(f);
      if (neighbor_index<0) continue;
      if (present_level!=neighbor_level) continue;

      int neighbor_level_number=
        neighbor_level->numberCoarserLevels();
      NumPtr<int> &cell_to_patch_index_map_this_level=
        cell_to_patch_index_map[neighbor_level_number];
      if (neighbor_index>=
      cell_to_patch_index_map_this_level.getNumber()) {
        continue;
      }
      if (cell_to_patch_index_map_this_level[neighbor_index]==
      HdivPatch::no_neighbor) {
        continue;
      }

      patch->setNeighbor(f,neighbor_index);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

#include <AreaGraphTool.H>
#include <Palette.H>
#include <Vector3.H>
#include <VolGraphTool.H>
#include <XColormap.H>
#include <XYGraphTool.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::plot(const char *name,NumPtr<VirtualGraphTool*> &gt,
bool quit_button) const {
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) plot(name,i,gt,quit_button);
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) plot(name,i,gt,quit_button);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void HdivDataOut::plot(const char *name,int data_set,
NumPtr<VirtualGraphTool*> &gt,bool quit_button) const {
  Point<1> low(HUGE_VAL);
  Point<1> high(-HUGE_VAL);
  int n_datasets=dof_data.getNumber()
    *dofs->getFe().nComponents()+cell_data.getNumber();
  Tensor<1,SPACEDIM> dlow; dlow.initialize(HUGE_VAL);
  Tensor<1,SPACEDIM> dhigh; dhigh.initialize(-HUGE_VAL);
  REAL multiplier_dlow(HUGE_VAL);
  REAL multiplier_dhigh(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,dlow,dhigh);
    patches[p]->getMultiplierDataBounds(data_set,multiplier_dlow,
      multiplier_dhigh);
  }
  CHECK_SAME(gt.getNumber(),2);
  NumPtr<XYGraphTool*> xygt(2);
  for (int d=0;d<2;d++) {
    xygt[d]=dynamic_cast<XYGraphTool*>(gt[d]);
    CHECK_POINTER(xygt[d]);
    if (d==0) {
      xygt[d]->rescale(low[0],high[0],dlow[0],dhigh[0]);
    } else {
      xygt[SPACEDIM]->rescale(low[0],high[0],multiplier_dlow,
        multiplier_dhigh);
    }
    xygt[d]->newPage();
    xygt[d]->setbgColor("white");
    xygt[d]->setfgColor("black");
    xygt[d]->drawAxes();
    xygt[d]->setfgColor("blue");
  }
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plot(xygt,data_set);
  }
  for (int d=0;d<2;d++) xygt[d]->flush();
  if (quit_button) XYGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif

#if (SPACEDIM==2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::plot(const char *name,int data_set,
NumPtr<VirtualGraphTool*> &gt,bool quit_button) const {
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  Tensor<1,SPACEDIM> ymin; ymin.initialize(HUGE_VAL);
  Tensor<1,SPACEDIM> ymax; ymax.initialize(-HUGE_VAL);
  REAL multiplier_ymin(HUGE_VAL);
  REAL multiplier_ymax(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
    patches[p]->getMultiplierDataBounds(data_set,multiplier_ymin,
      multiplier_ymax);
  }
  Tensor<1,SPACEDIM+1> dmin,dmax;
  for (int d=0;d<SPACEDIM;d++) {
    dmin[d]=ymin[d];
    dmax[d]=ymax[d];
  }
  dmin[SPACEDIM]=multiplier_ymin;
  dmax[SPACEDIM]=multiplier_ymax;
  CHECK_SAME(gt.getNumber(),2);
  NumPtr<AreaGraphTool*> agt(2);
  for (int d=0;d<2;d++) {
    agt[d]=dynamic_cast<AreaGraphTool*>(gt[d]);
    CHECK_POINTER(agt[d]);
    agt[d]->rescale(low[0],high[0],low[1],high[1]);
    agt[d]->newPage();
    agt[d]->setbgColor("white");
  }
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plot(agt,data_set,dmin,dmax);
  }
  for (int d=0;d<2;d++) {
    agt[d]->setfgColor("black");
    agt[d]->drawAxes();
    agt[d]->flush();
  }
  if (quit_button) AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDataOut::plot(const char *name,int data_set,
NumPtr<VirtualGraphTool*> &gt,bool) const {
  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(gt[0]);
  CHECK_POINTER(vgt);
  {
    Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
    Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
    Tensor<1,SPACEDIM> ymin; ymin.initialize(HUGE_VAL);
    Tensor<1,SPACEDIM> ymax; ymax.initialize(-HUGE_VAL);
    REAL multiplier_ymin(HUGE_VAL);
    REAL multiplier_ymax(-HUGE_VAL);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->getMeshBounds(low,high);
      patches[p]->getDataBounds(data_set,ymin,ymax);
    }
    Vector3 vlow(low[0],low[1],low[2]);
    Vector3 vhigh(high[0],high[1],high[2]);
    Vector3 vdif=(vhigh-vlow)*0.015625;
    vlow-=vdif;
    vhigh+=vdif;

    vgt->rescale(vlow,vhigh);
    vgt->newPage();
    vgt->expose();
    for (bool do_best=FALSE;;do_best=!do_best) {
      VolGraphTool::DrawDelimiter dd(do_best,vgt);
      for (int p=0;p<patches.getNumber();p++) {
        patches[p]->drawVectors(vgt,data_set,vlow,vhigh,ymin,ymax);
      }
      if (do_best) break;
    }
    vgt->eventLoop();
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void HdivDataOut::plotSurface(const char *name,
NumPtr<VolGraphTool*> &gt) const {
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) plotSurface(name,j,gt);
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) plotSurface(name,j,gt);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void HdivDataOut::plotSurface(const char *name,int data_set,
NumPtr<VolGraphTool*> &gt) const{
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  Tensor<1,SPACEDIM> ymin; ymin.initialize(DBL_MAX);
  Tensor<1,SPACEDIM> ymax; ymax.initialize(-DBL_MAX);
  REAL multiplier_ymin(DBL_MAX);
  REAL multiplier_ymax(-DBL_MAX);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
    patches[p]->getMultiplierDataBounds(data_set,multiplier_ymin,
      multiplier_ymax);
  }
  NumPtr<Vector3> vlow(SPACEDIM+1);
  NumPtr<Vector3> vhigh(SPACEDIM+1);
  NumPtr<Vector3> vdif(SPACEDIM+1);
  for (int d=0;d<=SPACEDIM;d++) {
    vlow[d][0]=low[0];
    vlow[d][1]=low[1];
    vlow[d][2]=(d<SPACEDIM ? ymin[d] : multiplier_ymin);
    vhigh[d][0]=high[0];
    vhigh[d][1]=high[1];
    vhigh[d][2]=(d<SPACEDIM ? ymax[d] : multiplier_ymax);
    Vector3 vdif=(vhigh[d]-vlow[d])*0.015625;
    vlow[d]-=vdif;
    vhigh[d]+=vdif;
    gt[d]->rescale(vlow[d],vhigh[d]);
    gt[d]->newPage();
    gt[d]->expose();
  }

  for (int d=0;d<=SPACEDIM;d++) {
    { BOOLEAN do_best=FALSE;
      VolGraphTool::DrawDelimiter dd(do_best,gt[d]);
      for (int p=0;p<patches.getNumber();p++) {
        patches[p]->plotSurface(d,gt[d],data_set,vlow[d][2],
          vhigh[d][2],do_best);
      }
    }
    { BOOLEAN do_best=TRUE;
      VolGraphTool::DrawDelimiter dd(do_best,gt[d]);
      for (int p=0;p<patches.getNumber();p++) {
        patches[p]->plotSurface(d,gt[d],data_set,vlow[d][2],
          vhigh[d][2],do_best);
      }
    }
    gt[d]->eventLoop();
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HdivDataOut::plotSurface(const char *name,int data_set,
NumPtr<VolGraphTool*> &gt) const{
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  Tensor<1,SPACEDIM> ymin; ymin.initialize(DBL_MAX);
  Tensor<1,SPACEDIM> ymax; ymax.initialize(-DBL_MAX);
  REAL multiplier_ymin(DBL_MAX);
  REAL multiplier_ymax(-DBL_MAX);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
    patches[p]->getDataBounds(data_set,ymin,ymax);
    patches[p]->getMultiplierDataBounds(data_set,multiplier_ymin,
      multiplier_ymax);
  }
  Vector3 vlow(low[0],low[1],low[2]);
  Vector3 vhigh(high[0],high[1],high[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;
  for (int d=0;d<=SPACEDIM;d++) {
    gt[d]->rescale(vlow,vhigh);
    gt[d]->newPage();
    HdivPatchSurfaceDataSet psds(d,gt[d],vlow,vhigh,dofs,mapping,patches,
      (d<SPACEDIM ? ymin[d] : multiplier_ymin),
      (d<SPACEDIM ? ymax[d] : multiplier_ymax),data_set);
    Surface<HdivPatchSurfaceDataSet> *s=
      OPERATOR_NEW Surface<HdivPatchSurfaceDataSet>(&psds,gt[d]);
    gt[d]->initialize(s);
    gt[d]->eventLoop();
    gt[d]->terminatePlotObj();
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void HdivDataOut::drawGrid(VirtualGraphTool *gt) const {
  AreaGraphTool *agt=dynamic_cast<AreaGraphTool*>(gt);
  CHECK_POINTER(agt);
  CHECK_TEST(patches.getNumber()>0);
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
  }
  agt->rescale(low[0],high[0],low[1],high[1]);
  agt->setbgColor("white");
  agt->newPage();
  agt->flush();
  agt->setfgColor("black");
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->draw(*agt);
  }
  agt->drawAxes();
  agt->flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HdivDataOut::drawGrid(VirtualGraphTool *gt) const {
  VolGraphTool *vgt=dynamic_cast<VolGraphTool*>(gt);
  CHECK_POINTER(vgt);
  CHECK_TEST(patches.getNumber()>0);
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
  }
  Vector3 vlow(low[0],low[1],low[2]);
  Vector3 vhigh(high[0],high[1],high[2]);
  Vector3 vdif=(vhigh-vlow)*0.015625;
  vlow-=vdif;
  vhigh+=vdif;
  vgt->rescale(vlow,vhigh);
  vgt->newPage();
  vgt->expose();

  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=vgt->getColorArray();
  VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=color_array->getColor(0.);
  vgt->setfgColor("black");
  { BOOLEAN do_best=FALSE;
    VolGraphTool::DrawDelimiter dd(do_best,vgt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->draw(*vgt);
    }
  }
  { BOOLEAN do_best=TRUE;
    VolGraphTool::DrawDelimiter dd(do_best,vgt);
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->draw(*vgt);
    }
  }
  vgt->expose();
  vgt->eventLoop();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void HdivDataOut::contourPlot(const char *name,
NumPtr<AreaGraphTool*> &gt) const {
  for (int i=0;i<dof_data.getNumber();i++) {
    const NumPtr<string> &dof_data_names=dof_data[i]->getNames();
    for (int j=0;j<dof_data_names.getNumber();j++) {
      if (name==dof_data_names[j]) contourPlot(name,j,gt);
    }
  }
  for (int i=0;i<cell_data.getNumber();i++) {
    const NumPtr<string> &cell_data_names=cell_data[i]->getNames();
    for (int j=0;j<cell_data_names.getNumber();j++) {
      if (name==cell_data_names[j]) contourPlot(name,j,gt);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void HdivDataOut::contourPlot(const char *name,int data_set,
NumPtr<AreaGraphTool*> &gt) const {
  Point<2> low(HUGE_VAL,HUGE_VAL);
  Point<2> high(-HUGE_VAL,-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMeshBounds(low,high);
  }
  int ncontours=30;
  Tensor<1,SPACEDIM> ymin; ymin.initialize(DBL_MAX);
  Tensor<1,SPACEDIM> ymax; ymax.initialize(-DBL_MAX);
  REAL multiplier_ymin(DBL_MAX),multiplier_ymax(-DBL_MAX);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getDataBounds(data_set,ymin,ymax);
    patches[p]->getMultiplierDataBounds(data_set,multiplier_ymin,
      multiplier_ymax);
  }
  Tensor<1,SPACEDIM+1> dy;
  for (int d=0;d<SPACEDIM;d++) {
    dy[d]=(ymax[d]-ymin[d])/static_cast<REAL>(ncontours+1);
  }
  dy[SPACEDIM]=(multiplier_ymax-multiplier_ymin)
              /static_cast<REAL>(ncontours+1);
  for (int d=0;d<=SPACEDIM;d++) {
    gt[d]->rescale(low[0],high[0],low[1],high[1]);
    gt[d]->newPage();
    gt[d]->setfgColor("black");
    gt[d]->drawBoxGivenCorners(low[0],low[1],high[0],high[1]);
  }
  Tensor<1,SPACEDIM+1> value;
  for (int d=0;d<=SPACEDIM;d++) value[d]=ymin[d]+dy[d]*0.5;
  for (int ic=0;ic<ncontours;ic++,value+=dy) {
    for (int d=0;d<=SPACEDIM;d++) {
      gt[d]->setfgColor(ic,ncontours);
    }
    for (int p=0;p<patches.getNumber();p++) {
      patches[p]->contour(gt,value,data_set);
    }
  }
  for (int d=0;d<=SPACEDIM;d++) gt[d]->flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HdivPatchDataSet::plot(BOOLEAN,const AxisClipPlane *cp,REAL ymin,
REAL ymax) {
  REAL location=graph_tool->getLocation(cp);
  int dir=cp->getDirection();
  HAND h=cp->getHand();
  REAL fudge=graph_tool->getLength()[dir]*0.015625;
  if (h==LEFT_HAND) location += fudge;
  else location -= fudge;
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->plotClipPlane(*graph_tool,ymin,ymax,location,dir,h,
      dofs,mapping,single_data);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HdivPatchDataSet::plot(BOOLEAN do_best,const AxisClipPlane *cp) {
  int data_set=0;
  REAL ymin(HUGE_VAL);
  REAL ymax(-HUGE_VAL);
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->getMultiplierDataBounds(data_set,ymin,ymax);
  }
  plot(do_best,cp,ymin,ymax);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HdivPatchDataSet::plot(BOOLEAN,REAL) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HdivPatchDataSet::plot(BOOLEAN) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
REAL HdivPatchDataSet::getVarMin() const {
  OBSOLETE("not programmed"); 
  return HUGE_VAL;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
REAL HdivPatchDataSet::getVarMax() const {
  OBSOLETE("not programmed"); 
  return HUGE_VAL;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HdivPatchSurfaceDataSet::plot(BOOLEAN,const AxisClipPlane *cp) {
  OBSOLETE("not programmed yet");
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void HdivPatchSurfaceDataSet::plot(BOOLEAN do_best,REAL val) {
  for (int p=0;p<patches.getNumber();p++) {
    patches[p]->
      plotSurface(d,*graph_tool,data_set,val,var_min,var_max,do_best);
  }
}
#endif

#include "NumPtr.C"
INSTANTIATE_NUMPTR(string);
INSTANTIATE_NUMPTR(HdivDataOut::DataEntry*);
INSTANTIATE_NUMPTR(HdivPatch*);
INSTANTIATE_NUMPTR(NumPtr<int>);
INSTANTIATE_NUMPTR(VirtualGraphTool*);
#if (SPACEDIM==1)
INSTANTIATE_NUMPTR(XYGraphTool*);
#endif
#if (SPACEDIM==2)
INSTANTIATE_NUMPTR(AreaGraphTool*);
INSTANTIATE_NUMPTR(Vector3);
#endif

#if (SPACEDIM==3)
template class TypedPlotObj<HdivPatchDataSet>;
template class Surface<HdivPatchSurfaceDataSet>;
#endif
