#ifndef HDIV_DATA_OUT_H
#define HDIV_DATA_OUT_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <HdivDofHandler.H>
#include <HdivPatch.H>
#include <Types.H>

#if (SPACEDIM==1)
class XYGraphTool;
#endif
#if (SPACEDIM==2)
class AreaGraphTool;
class Vector3;
#endif
#if (SPACEDIM>=2)
class VolGraphTool;
#endif
#if (SPACEDIM==3)
class AxisClipPlane;
#endif

class HdivDataOut {
  protected:
    enum DATA_VECTOR_TYPE {TYPE_DOF_DATA,TYPE_CELL_DATA,TYPE_AUTOMATIC};
    class DataEntry {
      private:
        const Vector *single_data;
        NumPtr<string> names;
      public:
        DataEntry(const Vector *data=0,
        const NumPtr<string> &n=NumPtr<string>()) : single_data(data),
        names(n.getNumber()) {
          names.copyFrom(n); 
        }
        void clearPointers() { single_data=0; }
        const NumPtr<string>& getNames() const { return names; }
        const Vector* singleData() const { return single_data; }
    };
    const HdivDofHandler *dofs;
    const Mapping *mapping;
    NumPtr<DataEntry*> dof_data;
    NumPtr<DataEntry*> cell_data;
    NumPtr<HdivPatch*> patches;

    virtual void getDatasetNames(NumPtr<string>&) const;
    void clearDataVectors();
  private:
    void plot(const char*,int,NumPtr<VirtualGraphTool*>&,bool) const;
#if (SPACEDIM==2)
    void contourPlot(const char*,int,NumPtr<AreaGraphTool*>&) const;
#endif
#if (SPACEDIM>=2)
    void plotSurface(const char*,int,NumPtr<VolGraphTool*>&) const;
#endif
  public:
    HdivDataOut(const HdivDofHandler &d,const Mapping &m) : dofs(&d),
      mapping(&m) {;}
    virtual ~HdivDataOut() { clearDataVectors(); }
    void addDataVector(const Vector&,const NumPtr<string>&,
      DATA_VECTOR_TYPE=TYPE_AUTOMATIC);
    void addDataVector(const Vector&,const string &name,
      DATA_VECTOR_TYPE=TYPE_AUTOMATIC);
    void clearInputDataReferences();
    virtual void buildPatches(int n_subdivisions=1);
    void plot(const char*,NumPtr<VirtualGraphTool*>&,bool) const;
#if (SPACEDIM==2)
    void contourPlot(const char*,NumPtr<AreaGraphTool*>&) const;
#endif
#if (SPACEDIM>=2)
    void drawGrid(VirtualGraphTool*) const;
    void plotSurface(const char*,NumPtr<VolGraphTool*>&) const;
#endif
};

#if (SPACEDIM==3)
class HdivPatchDataSet {
  private:
    int d;
    VolGraphTool *graph_tool;
    const Vector3 &low;
    const Vector3 &high;
    const NumPtr<HdivPatch*> &patches;
    const HdivDofHandler *dofs;
    const Mapping *mapping;
    const Vector *single_data;
  public:
    HdivPatchDataSet(int dd,VolGraphTool *gt,const Vector3 &l,
    const Vector3 &h,const NumPtr<HdivPatch*> &p,const HdivDofHandler *dh,
    const Mapping *m,const Vector *sd) : d(dd),graph_tool(gt),low(l),
      high(h),patches(p),dofs(dh),mapping(m),single_data(sd) { }
    virtual ~HdivPatchDataSet() { }
    VolGraphTool* getGraphTool() const { return graph_tool; }
    BOOLEAN drawSurface() const { return FALSE; }
    void plot(BOOLEAN,const AxisClipPlane*);
    void plot(BOOLEAN,const AxisClipPlane*,REAL,REAL);
    void plot(BOOLEAN,REAL);
    void plot(BOOLEAN);
    REAL getVarMin() const;
    REAL getVarMax() const;
    virtual void printOn(ostream&) const {}
};
#endif

#if (SPACEDIM==3)
class HdivPatchSurfaceDataSet {
  private:
    int d;
    VolGraphTool *graph_tool;
    const Vector3 &low;
    const Vector3 &high;
    const HdivDofHandler *dofs;
    const Mapping *mapping;
    const NumPtr<HdivPatch*> &patches;
    REAL var_min,var_max;
    int data_set;
  public:
    HdivPatchSurfaceDataSet(int dd,VolGraphTool *gt,const Vector3 &l,
    const Vector3 &h,const HdivDofHandler *dh,const Mapping *m,
    const NumPtr<HdivPatch*> &p,REAL vn,REAL vx,int ds) : d(dd), 
      graph_tool(gt),low(l),high(h),dofs(dh),mapping(m),patches(p),
      var_min(vn),var_max(vx),data_set(ds) { }
    virtual ~HdivPatchSurfaceDataSet() { }
    VolGraphTool* getGraphTool() const { return graph_tool; }
    REAL getVarMin() const { return var_min; }
    REAL getVarMax() const { return var_max; }
    BOOLEAN drawSurface() const { return TRUE; }
    void plot(BOOLEAN,const AxisClipPlane*);
    void plot(BOOLEAN,REAL);
    virtual void printOn(ostream&) const {}
};
#endif

#endif
