// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <HdivDofHandler.H>

#include <algorithm>
#include <Boundary.H>
#include <FEValues.H>

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::clearSpace() {  
  for (int i=0;i<levels.getNumber();i++) {
    if (levels[i]) delete levels[i]; levels[i]=0;
  }
  levels.cleanup();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::reserveSpace() {
  int number_levels=tess->numberLevels();
  CHECK_TEST(number_levels>0);
  clearSpace();
  levels.allocate(number_levels);
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
    int cd1n=0;
#if (SPACEDIM==1)
    cd1n=tess->getVertices().getNumber()*selected_fe->dofsPerFace(1);
#else
    CODIMENSION1_LEVEL_CLASS *cd1l=
      static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
    for (bool doit=cd1l->first();doit;doit=cd1l->next()) {
      CODIMENSION1_SHAPE_CLASS *e=cd1l->getShape();
      e->setDofIndex(cd1n);
      cd1n+=selected_fe->dofsPerFace(e);
    }
#endif
    int n=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      SHAPE_CLASS *e=level->getShape();
      e->setDofIndex(n);
      n+=selected_fe->dofsPerShape(e);
    }
    int mn=0;
    map<int,int> element_data_index_to_multiplier_index_map;
    for (bool doit=level->first();doit;doit=level->next()) {
      SHAPE_CLASS *e=level->getShape();
      element_data_index_to_multiplier_index_map[e->elementDataIndex()]=
        mn;
      mn+=selected_fe->multiplierDofsPerShape(e);
    }
    levels[i]=OPERATOR_NEW HdivDofLevel(cd1n,n,mn,
      element_data_index_to_multiplier_index_map);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::distributeDofs() {
  CHECK_TEST(tess->numberLevels()>0);
  reserveSpace();
  NumPtr<bool> user_flags;
  tess->saveUserFlags(user_flags);
  const_cast<Tessellation*>(tess)->clearUserFlags();
  int next_free_dof=0;   
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    HdivDofLevel *dof_level=getLevel(level);
#if (SPACEDIM==1)
    int dpf=selected_fe->dofsPerFace(1);
    if (dpf>0) {
      for (int v=0;v<element->numberVertices();v++) {
        dof_level->setCodimension1Dofs(element->vertexIndex(v),dpf,
          next_free_dof);
      }
    }
#else
    const CODIMENSION1_LEVEL_CLASS *face_level=
      static_cast<const CODIMENSION1_LEVEL_CLASS*>(level);
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      const CODIMENSION1_SHAPE_CLASS *face_element=
        face_level->getShape(element->codimension1ShapeIndex(f));
      int dpf=selected_fe->dofsPerFace(face_element);
      if (dpf>0) {
        dof_level->setCodimension1Dofs(face_element->dofIndex(),dpf,
          next_free_dof);
      }
    }
#endif
    int dps=selected_fe->dofsPerShape(element);
    if (dps>0) dof_level->setDofs(element->dofIndex(),dps,next_free_dof);
  }
  total_solution_dofs=next_free_dof;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    const SHAPE_CLASS *element=tess->presentLevel()->getShape();
    LEVEL_CLASS *level=tess->presentLevel();
    HdivDofLevel *dof_level=getLevel(level);
    int mdps=selected_fe->multiplierDofsPerShape(element);
    if (mdps>0) {
      dof_level->setMultiplierDofs(element->elementDataIndex(),mdps,
        next_free_dof);
    }
    level->setUserFlag(element);
  }
  used_dofs=next_free_dof;
  const_cast<Tessellation*>(tess)->loadUserFlags(user_flags);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::getDofIndices(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,NumPtr<int> &dof_indices) const {
  const CODIMENSION1_LEVEL_CLASS *codimension1_level=
    static_cast<const CODIMENSION1_LEVEL_CLASS*>(level);
  int dps=selected_fe->dofsPerShape(element);
  int mdps=selected_fe->multiplierDofsPerShape(element);
  int sz=dps+mdps;
#if (SPACEDIM==1)
  int dpf=selected_fe->dofsPerFace(1);
  sz+=dpf*element->numberVertices();
#else
  int nf=element->numberCodimension1Shapes();
  for (int f=0;f<nf;f++) {
    const CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape(element->codimension1ShapeIndex(f));
    sz+=selected_fe->dofsPerFace(face_element);
  }
#endif
  dof_indices.allocate(sz);
#ifdef INDEF
  dof_indices.initialize(invalid_dof_index);
#endif
  HdivDofLevel *dof_level=getLevel(level);
  int offset=0;
#if (SPACEDIM==1)
  for (int v=0;v<element->numberVertices();v++) {
    NumPtr<int> face_dofs(dof_indices,offset,dpf);
    dof_level->getCodimension1Dofs(element->vertexIndex(v),face_dofs);
    offset+=dpf;
  }
#else
  for (int f=0;f<nf;f++) {
    const CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape(element->codimension1ShapeIndex(f));
    int dpf=selected_fe->dofsPerFace(face_element);
    NumPtr<int> face_dofs(dof_indices,offset,dpf);
    dof_level->getCodimension1Dofs(face_element->dofIndex(),face_dofs);
    offset+=dpf;
  }
#endif
  NumPtr<int> element_dofs(dof_indices,offset,dps);
  dof_level->getDofs(element->dofIndex(),element_dofs);

  NumPtr<int> multiplier_dofs(dof_indices,offset+dps,mdps);
  dof_level->getMultiplierDofs(element->elementDataIndex(),
    multiplier_dofs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::getFaceDofIndices(const SHAPE_CLASS *element,
int f,const LEVEL_CLASS *level,NumPtr<int> &face_dof_indices) const {
#if (SPACEDIM==1)
  int sz=selected_fe->dofsPerFace(1);
#else
  int sz=selected_fe->dofsPerFace(element,f);
#endif
  face_dof_indices.allocate(sz);
#ifdef INDEF
  face_dof_indices.initialize(invalid_dof_index);
#endif
  HdivDofLevel *dof_level=getLevel(level);
#if (SPACEDIM==1)
  dof_level->
    getCodimension1Dofs(element->vertexIndex(f),face_dof_indices);
#else
  int face_index=element->codimension1ShapeIndex(f);
  const CODIMENSION1_SHAPE_CLASS *face_element=
    static_cast<const CODIMENSION1_LEVEL_CLASS*>(level)->
    getShape(face_index);
  dof_level->
    getCodimension1Dofs(face_element->dofIndex(),face_dof_indices);
#endif
}
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::getFaceDofIndices(
const CODIMENSION1_SHAPE_CLASS *face_element,
const CODIMENSION1_LEVEL_CLASS *level,NumPtr<int> &face_dof_indices)
const {
  int sz=selected_fe->dofsPerFace(face_element);
  face_dof_indices.allocate(sz);
# ifdef INDEF
  face_dof_indices.initialize(invalid_dof_index);
# endif
  HdivDofLevel *dof_level=getLevel(level);
  dof_level->
    getCodimension1Dofs(face_element->dofIndex(),face_dof_indices);
}
#endif
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::dofsOnChildren(
const CODIMENSION1_SHAPE_CLASS *face_element,
const CODIMENSION1_LEVEL_CLASS *face_level,
NumPtr<int> &dofs_on_children) const {
  CHECK_TEST(face_level->hasChildren(face_element));
  int ndoc=childrenPerCodimension1Shape()
          *selected_fe->dofsPerFace(face_element);
  dofs_on_children.allocate(ndoc);

  int dpf=selected_fe->dofsPerFace(face_element);
  NumPtr<int> doc(dofs_on_children);
  const CODIMENSION1_LEVEL_CLASS *face_child_level=
    face_level->finerLevel();
  HdivDofLevel *dof_level=getLevel(face_child_level);
  int face_index=face_element->elementDataIndex();
  for (int child=0;child<childrenPerCodimension1Shape();child++) {
    int face_child_index=face_level->childIndex(face_index,child);
    const CODIMENSION1_SHAPE_CLASS *face_child=
      face_child_level->getShape(face_child_index);
    NumPtr<int> face_child_dof_indices(dpf);
    dof_level->getCodimension1Dofs(face_child->dofIndex(),
      face_child_dof_indices);
    doc.copyFrom(face_child_dof_indices);
    doc+=face_child_dof_indices.getNumber();
  }
  CHECK_SAME(0,doc.getNumber());
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::getDofValues(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,const Vector &values,Vector &local_values) 
const {
  NumPtr<int> dof_indices;
  getDofIndices(element,level,dof_indices);
  local_values.reinit(dof_indices.getNumber(),true);
  for (int i=0;i<dof_indices.getNumber();i++) {
    local_values[i]=values[dof_indices[i]];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::getInterpolatedDofValues(
const SHAPE_CLASS *element,const LEVEL_CLASS *level,
const Vector &values,Vector &interpolated_values) const {
  CHECK_SAME(values.size(),numberDofs());
  int dofs_per_cell=selected_fe->dofsPerCell(element);
  if (!level->hasChildren(element)) {
    getDofValues(element,level,values,interpolated_values);
  } else {
    OBSOLETE("not programmed");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::makeSparsityPattern(SparsityPatternBase &sp_A,
SparsityPatternBase &sp_B) const {
  CHECK_SAME(sp_A.nRows(),total_solution_dofs);
  CHECK_SAME(sp_A.nCols(),total_solution_dofs);
  CHECK_SAME(sp_B.nRows(),used_dofs-total_solution_dofs);
  CHECK_SAME(sp_B.nCols(),total_solution_dofs);
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);
    int multiplier_dofs_per_cell=
      selected_fe->multiplierDofsPerShape(element);
    NumPtr<int> dofs_on_this_cell;
    getDofIndices(element,present_level,dofs_on_this_cell);
    for (int i=0;i<dofs_per_cell;i++) {
      for (int j=0;j<dofs_per_cell;j++) {
        sp_A.add(dofs_on_this_cell[i],dofs_on_this_cell[j]);
      }
    }
    for (int i=0;i<multiplier_dofs_per_cell;i++) {
      for (int j=0;j<dofs_per_cell;j++) {
        sp_B.add(dofs_on_this_cell[i+dofs_per_cell]-total_solution_dofs,
                 dofs_on_this_cell[j]);
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
int HdivDofHandler::maxCouplingsBetweenDofs() const {
  return min(3*selected_fe->dofsPerFace(1)
            +2*selected_fe->dofsPerShape(2)
            +2*selected_fe->multiplierDofsPerShape(2),used_dofs);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
int HdivDofHandler::maxCouplingsBetweenDofs() const {
//for each element face, find set of neighbor elements
  map<const CODIMENSION1_SHAPE_CLASS*,
    set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> > >
    face_element_set_map;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_element_index=element->codimension1ShapeIndex(f);
      const CODIMENSION1_SHAPE_CLASS *face_element=level->
        CODIMENSION1_LEVEL_CLASS::getShape(face_element_index);
      face_element_set_map[face_element].insert(make_pair(element,level));
      LEVEL_CLASS *neighbor_level=level->neighborLevel(f);
      if (neighbor_level) {
        face_element_set_map[face_element].insert(
          make_pair(neighbor_level->getShape(level->neighborIndex(f)),
                    neighbor_level));
      }    
    }    
  }
  int max_couplings=0;
  LEVEL_CLASS *tess_coarsest_level=tess->coarsestLevel();
  LEVEL_CLASS *tess_finest_level=tess_coarsest_level->finestLevel();
  for (map<const CODIMENSION1_SHAPE_CLASS*,
  set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> > >::const_iterator
  fesm_iterator=face_element_set_map.begin();fesm_iterator!=
  face_element_set_map.end();fesm_iterator++) {
    const CODIMENSION1_SHAPE_CLASS* face_element=fesm_iterator->first;
    const set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >
      &neighbor_set=fesm_iterator->second;
    const LEVEL_CLASS *finest_level=tess_coarsest_level;
    const LEVEL_CLASS *coarsest_level=tess_finest_level;
    for (set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >
    ::const_iterator ns_iterator=neighbor_set.begin();ns_iterator!=
    neighbor_set.end();ns_iterator++) {
      const LEVEL_CLASS *level=ns_iterator->second;
      if (level->isFinerThan(finest_level)) finest_level=level;
      if (level->isCoarserThan(coarsest_level)) coarsest_level=level;
    }
    int face_element_index=face_element->elementDataIndex();
    int couplings=0;
    if (finest_level==coarsest_level) {
      couplings=selected_fe->dofsPerFace(face_element);
      for (set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >
      ::const_iterator ns_iterator=neighbor_set.begin();ns_iterator!=
      neighbor_set.end();ns_iterator++) {
        const SHAPE_CLASS *element=ns_iterator->first;
        const LEVEL_CLASS *level=ns_iterator->second;
        couplings+=selected_fe->dofsPerShape(element)
                  +selected_fe->multiplierDofsPerShape(element);
        for (int f=0;f<element->numberCodimension1Shapes();f++) {
          int fi=element->codimension1ShapeIndex(f);
          if (fi!=face_element_index) {
            const CODIMENSION1_SHAPE_CLASS *other_face_element=level->
              CODIMENSION1_LEVEL_CLASS::getShape(fi);
            couplings+=selected_fe->dofsPerFace(other_face_element);
          }
        }
      }
    } else {
      for (set<pair<const SHAPE_CLASS*,const LEVEL_CLASS*> >
      ::const_iterator ns_iterator=neighbor_set.begin();ns_iterator!=
      neighbor_set.end();ns_iterator++) {
        const SHAPE_CLASS *element=ns_iterator->first;
        const LEVEL_CLASS *level=ns_iterator->second;
        if (level==finest_level) {
          couplings+=selected_fe->dofsPerShape(element)
                    +selected_fe->multiplierDofsPerShape(element);
          for (int f=0;f<element->numberCodimension1Shapes();f++) {
            int fi=element->codimension1ShapeIndex(f);
            if (fi!=face_element_index) {
              const CODIMENSION1_SHAPE_CLASS *other_face_element=level->
                CODIMENSION1_LEVEL_CLASS::getShape(fi);
              couplings+=selected_fe->dofsPerFace(other_face_element);
            }
          }
        } else {
          set<int> fine_face_indices;
          for (int c=0;c<childrenPerCodimension1Shape();c++) {
            fine_face_indices.insert(finest_level->
              CODIMENSION1_LEVEL_CLASS::childIndex(face_element_index,c));
          }
          set<int> fine_element_indices;
          for (int c=0;c<childrenPerShape();c++) {
            int child_index=
              level->childIndex(element->elementDataIndex(),c);
            const SHAPE_CLASS *child_element=
              finest_level->getShape(child_index);
            for (int f=0;f<child_element->numberCodimension1Shapes();f++)
            {
              set<int>::const_iterator ffi_iterator=
                fine_face_indices.find(child_element->
                codimension1ShapeIndex(f));
              if (ffi_iterator!=fine_element_indices.end()) {
                fine_element_indices.insert(child_index);
                break;
              }
            }
          }
          for (set<int>::const_iterator fei_iterator=
          fine_element_indices.begin();fei_iterator!=
          fine_element_indices.end();fei_iterator++) {
            const SHAPE_CLASS *child_element=
              finest_level->getShape(*fei_iterator);
            for (int f=0;f<child_element->numberCodimension1Shapes();f++)
            {
              fine_face_indices.insert(
                child_element->codimension1ShapeIndex(f));
            }
            couplings+=selected_fe->dofsPerShape(child_element)
                      +selected_fe->multiplierDofsPerShape(child_element);
          }
          for (set<int>::const_iterator ffi_iterator=
          fine_face_indices.begin();ffi_iterator!=
          fine_face_indices.end();ffi_iterator++) {
            const CODIMENSION1_SHAPE_CLASS *child_face_element=
              finest_level->CODIMENSION1_LEVEL_CLASS::
              getShape(*ffi_iterator);
            couplings+=selected_fe->dofsPerFace(child_face_element);
          }
        }
      }
    }
    max_couplings=max(max_couplings,couplings);
  }
  return min(max_couplings,numberDofs());
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::assembleLinearSystem(const Mapping &mapping,
const map<int,Quadrature<SPACEDIM> > &qmap,
map<int,Quadrature<SPACEDIM-1> > &fqmap,SparseMatrix &system_matrix_A,
SparseMatrix &system_matrix_B,Vector &system_rhs) const {
  UPDATE_FLAGS update_flags= UPDATE_VALUES | UPDATE_DIVERGENCE | 
    UPDATE_DETERMINANT | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  for (bool doit=tess->firstShape();doit;doit=tess->nextShapeInLevel()) {
    HdivMaterialModel *mm=dynamic_cast<HdivMaterialModel*>(
      tess->presentLevel()->getMaterialModel());
    CHECK_POINTER(mm);
    if (mm->requiresDerivative()) {
      update_flags |= UPDATE_GRADIENTS;
      break;
    }
  }
  HdivFEValues fe_values(mapping,*selected_fe,qmap,update_flags);
#if (SPACEDIM>1)
  UPDATE_FLAGS face_update_flags=UPDATE_VALUES | UPDATE_Q_POINTS |
    UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES;
  HdivFEFaceValues fe_face_values(mapping,*selected_fe,fqmap,
    face_update_flags);
#endif
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);
    int multiplier_dofs_per_shape=
      selected_fe->multiplierDofsPerShape(element);

//  local stiffness matrix and right-hand side
    Matrix cell_matrix(dofs_per_cell+multiplier_dofs_per_shape,
                       dofs_per_cell+multiplier_dofs_per_shape);
    Vector cell_rhs(dofs_per_cell+multiplier_dofs_per_shape);
    cell_matrix=0.;
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,Tensor<1,SPACEDIM> > *sv=
      fe_values.getShapeValues(element);
    const Array<2,REAL> *sd=
      fe_values.getShapeDivergence(element);
    const Array<2,REAL> *msv=
      fe_values.getMultiplierShapeValues(element);

    HdivMaterialModel *mm=dynamic_cast<HdivMaterialModel*>(
      present_level->getMaterialModel());
    CHECK_POINTER(mm);
    const Array<2,Tensor<2,SPACEDIM> > *sg=(mm->requiresDerivative()
      ? fe_values.getShapeGradients(element) : 0 );

//  computation of local stiffness matrix and right-hand side
    mm->elementStiffnessMatrixAndRHS(q_points,jxw_values,sv,sd,sg,msv,
      cell_matrix,cell_rhs);

//  neumann boundary conditions
#if (SPACEDIM==1)
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (present_level->neighborLevel(f)==0) {
        const BoundaryCondition *bc=
          present_level->Level0::getBoundaryCondition(f);
        if (bc) {
          if (bc->isNatural()) {
//          the following 9 lines are awkward:
//          the use of FEValues of some sort is necessary in order to
//            get the shapeValue's in the correct order
//          DEAL has trouble with FEFaceValues in 1D because there is
//            no Shape associated with faces of Interval's
//          FEValues wants to compute Jacobians and quadrature points
//            when it computes shapeValue's
            NumPtr<Point<1> > p(1); 
            p[0]=Interval::canonicalVertex(f);
            map<int,Quadrature<1> > fqmap_1d;
            fqmap_1d[element->numberVertices()]=Quadrature<1>(p);
            HdivFEValues face_fe_values(mapping,*selected_fe,fqmap_1d,
              UPDATE_VALUES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES); 
            face_fe_values.mapShapes(*tess,element,present_level);
            const Array<2,Tensor<1,SPACEDIM> > &fsv=
              *face_fe_values.getShapeValues(element);
            const NumPtr<Point<1> > &face_q_points=
              face_fe_values.getQuadraturePoints(element);
            Tensor<1,SPACEDIM> normal_vector;
            normal_vector[0]=(f==0 ? -1. : 1.);
            Array<2,REAL> normal_dot_fsv(fsv.size(0),fsv.size(1));
            for (int s=0;s<fsv.size(0);s++) {
              normal_dot_fsv(s,0)=normal_vector*fsv(s,0);
            }
            NumPtr<REAL> face_jxw_values(1); face_jxw_values[0]=1.;
            bc->faceNeumannValues(face_q_points,face_jxw_values,
              0,normal_dot_fsv,cell_rhs);
          }
        }
      }
    }
#else
// int_mapping(Gamma) m(mapping^{-1}(x)) n_{mapping(Gamma)}(x) . u(x) da
// = int_Gamma m(xi) n_{mapping(Gamma)}(mapping(xi)) . u(mapping(xi))
//   || D mapping^{-1}(xi) n_Gamma det D mapping(xi) || da_Gamma
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const BoundaryCondition *bc=present_level->
        CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
      if (bc) {
        if (bc->isNatural()) {
//        HdivBoundaryCondition can be natural for some components, if
//          n_components > 1; we haven't programmed that case
          const CODIMENSION1_SHAPE_CLASS *face_element=present_level->
            CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          fe_face_values.mapShapes(*tess,element,present_level,f);
          const Array<2,Tensor<1,SPACEDIM> > &fsv=
            *fe_face_values.getShapeValues(element);
          NumPtr<Point<SPACEDIM> > face_q_points=
            fe_face_values.getQuadraturePoints(element,f);
          NumPtr<REAL> face_jxw_values=
            fe_face_values.getJxWValues(element,f);
          const NumPtr<Point<SPACEDIM> > &normal_vectors=
            fe_face_values.getNormalVectors(element,f);
          bc->faceNeumannValues(face_q_points,face_jxw_values,
            normal_vectors,offset,fsv,cell_rhs);
        }
      }
      offset+=
        fe_face_values.getFaceQuadrature(element,f).numberPoints();
    }
#endif

//  add element-wise contributions to global arrays
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_matrix_A.add(ldfi,local_dof_indices[j],cell_matrix(i,j));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
    for (int i=0;i<multiplier_dofs_per_shape;i++) {
      int ldfi=local_dof_indices[i+dofs_per_cell]-total_solution_dofs;
      for (int j=0;j<dofs_per_cell;j++) {
        system_matrix_B.add(ldfi,local_dof_indices[j],
          cell_matrix(i+dofs_per_cell,j));
      }
      system_rhs[ldfi]+=cell_rhs[i+dofs_per_cell];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::interpolateBoundaryValues(const Mapping &mapping,
map<int,REAL> &boundary_values) const {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
#if (SPACEDIM>1)
  UPDATE_FLAGS face_update_flags=UPDATE_Q_POINTS | 
    UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES;
  map<int,Quadrature<SPACEDIM-1> > &fqmap=
    selected_fe->faceQuadratureMap();
  HdivFEFaceValues fe_face_values(mapping,*selected_fe,fqmap,
    face_update_flags);
#endif
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    for (int face_no=0;face_no<element->numberCodimension1Shapes();
    face_no++) {
      if (level->neighborLevel(face_no)==0) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
#if (SPACEDIM==1)
        const BoundaryCondition *bc=
          codimension1_level->getBoundaryCondition(face_no);
        if (bc) {
          if (bc->isEssential()) {
            Vector function_values(n_components);
            const Point<SPACEDIM> &boundary_point=tess->
              getVertices()[element->codimension1ShapeIndex(face_no)];
            FunctionSystem *fs=bc->getFunctionSystem();
            CHECK_POINTER(fs);
            fs->value(boundary_point,function_values);
            NumPtr<int> face_dof_indices;
            getFaceDofIndices(element,face_no,level,face_dof_indices);
            for (int c=0;c<n_components;c++) {
              if (bc->isEssential(c)) {
                boundary_values[face_dof_indices[c]]=function_values[c];
                if (face_no==0) { // outer normal points to left
                  boundary_values[face_dof_indices[c]]=
                    -boundary_values[face_dof_indices[c]];
                }
              }
            }
          }
        }
#else
        int face_index=element->codimension1ShapeIndex(face_no);
        const BoundaryCondition *bc=
          codimension1_level->getBoundaryCondition(face_index);
        if (bc) {
          if (bc->isEssential()) {
            const CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape(face_index);
            fe_face_values.mapShapes(*tess,element,level,face_no);
            NumPtr<Point<SPACEDIM> > dof_locations=
              fe_face_values.getQuadraturePoints(element,face_no);
            const NumPtr<Point<SPACEDIM-1> > &canonical_locations=
              fe_face_values.getFaceQuadrature(element,face_no).points();
            NumPtr<REAL> face_jxw_values=
              fe_face_values.getJxWValues(element,face_no);
            NumPtr<int> face_dofs;
            getFaceDofIndices(element,face_no,level,face_dofs);
            NumPtr<int> element_dofs;
            getDofIndices(element,level,element_dofs);
            selected_fe->setBoundaryValues(element,face_element,
              element_dofs,face_dofs,canonical_locations,dof_locations,
              face_jxw_values,bc,boundary_values);
          }
        }
#endif
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::interpolateBoundaryValues(const Mapping &mapping,
map<int,REAL> &boundary_values,map<int,Quadrature<SPACEDIM-1> > &dqmap)
const {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    for (int face_no=0;
    face_no<element->numberCodimension1Shapes();face_no++) {
      int face_index=element->codimension1ShapeIndex(face_no);
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
#if (SPACEDIM==1)
      OBSOLETE("fix like other interpolateBoundaryValues");
#else
      OBSOLETE("fix like other interpolateBoundaryValues");
#endif
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//eliminates boundary dof from matrix and right-hand-side
//  for curved boundaries with normal != axis vector, this is incorrect
void HdivDofHandler::applyBoundaryValues(
const map<int,REAL> &boundary_values,SparseMatrix &matrix_A,
SparseMatrix &matrix_B,Vector &right_hand_side,bool preserve_symmetry)
const {
  int n_rows_A=matrix_A.size(0);
  CHECK_SAME(matrix_A.size(1),n_rows_A);
  int n_rows_B=matrix_B.size(0);
  CHECK_SAME(n_rows_A+n_rows_B,right_hand_side.size());
  if (boundary_values.size()==0) return;
  const SparsityPattern &sparsity_A=matrix_A.getSparsityPattern();
  const SparsityPattern &sparsity_B=matrix_B.getSparsityPattern();
  const NumPtr<int> &sparsity_colnums_A=sparsity_A.getColumnNumbers();
  const NumPtr<int> &sparsity_colnums_B=sparsity_B.getColumnNumbers();

  REAL first_nonzero_diagonal_entry_A=1.;
  for (int i=0;i<n_rows_A;i++) {
    REAL diag=matrix_A.diagElement(i);
    if (abs(diag)>0.) {
      first_nonzero_diagonal_entry_A=diag;
      break;
    }
  }
//sum_j A_ij x_j = b_i, x_k given ==>
//  A_kk x_k = (A_kk x_k)
//  for i!=k, sum_{j!=k} A_ij x_j = b_i - A_ik x_k
  for (map<int,REAL>::const_iterator bv_iterator=
  boundary_values.begin();bv_iterator!=boundary_values.end();
  bv_iterator++) {
    int dof_number=bv_iterator->first;
    CHECK_TEST(dof_number<n_rows_A);
//  set matrix_A row to zero, except for diagonal
    int last_A=sparsity_A.getRowstartIndex(dof_number+1);
    for (int j=sparsity_A.getRowstartIndex(dof_number);j<last_A;j++) {
      if (sparsity_A.getColumnNumber(j)!=dof_number) {
        matrix_A.globalEntry(j)=0.;
      }
    }
//  set diagonal entry to nonzero, set new right-hand-side for this row
    REAL diag=matrix_A.diagElement(dof_number);
    REAL boundary_value=bv_iterator->second;
    if (abs(diag)<=0.0) {
      diag=first_nonzero_diagonal_entry_A;
      matrix_A.set(dof_number,dof_number,first_nonzero_diagonal_entry_A);
    }
    right_hand_side[dof_number]=boundary_value*diag;
//  zero matrix column off diagonal, modify right-hand-side in other rows
    if (preserve_symmetry) {
      for (int j=sparsity_A.getRowstartIndex(dof_number);j<last_A;j++) {
        if (sparsity_A.getColumnNumber(j)!=dof_number) {
          int row=sparsity_colnums_A[j];
          int p=sparsity_A.getRowstartIndex(row+1);
          for (int c=sparsity_A.getRowstartIndex(row)+1;
          c<sparsity_A.getRowstartIndex(row+1);c++) {
            if (sparsity_colnums_A[c]>=dof_number) {
              p=c; break;
            }
          }
          CHECK_TEST((sparsity_colnums_A[p]==dof_number) &&
           (p!=sparsity_A.getRowstartIndex(row+1)));
          int global_entry=p-sparsity_A.getRowstartIndex(0);
          REAL &mij=matrix_A.globalEntry(global_entry);
          right_hand_side[row]-=mij*boundary_value;
          mij=0.;
        }
      }
    }
    for (int i=0;i<sparsity_B.nRows();i++) {
      for (int k=sparsity_B.getRowstartIndex(i);
      k<sparsity_B.getRowstartIndex(i+1);k++) {
        int j=sparsity_B.getColumnNumber(k);
        if (j==dof_number) {
          REAL &mij=
            matrix_B.globalEntry(k-sparsity_B.getRowstartIndex(0));
          right_hand_side[i+total_solution_dofs]-=mij*boundary_value;
          mij=0.;
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//sets solution at boundary dofs
void HdivDofHandler::applyBoundaryValues(
const map<int,REAL> &boundary_values,Vector &solution) const {
  if (boundary_values.size()==0) return;
  for (map<int,REAL>::const_iterator dof=
  boundary_values.begin();dof!=boundary_values.end();dof++) {
    solution[dof->first]=dof->second;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL HdivDofHandler::integrateDifference(const Mapping &mapping,
const Vector &fe_function,const TensorFunctionSystem &exact_solution,
const map<int,Quadrature<SPACEDIM> > &qmap,NORM_TYPE norm,
const TensorFunctionSystem *weight,REAL exponent_1) {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  if (weight) {
    int wnc=weight->nComponents();
    CHECK_TEST(wnc==1 || wnc==n_components);
  }

  REAL exponent=exponent_1;
  switch (norm) {
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      exponent=2.;
      break;
    case L1_NORM:
      exponent=1.;
      break;
    default:
      break;
  }
  
  UPDATE_FLAGS update_flags=UPDATE_Q_POINTS | UPDATE_JxW_VALUES;
  switch (norm) {
    case H1_SEMINORM:
    case W1P_SEMINORM:
      update_flags|=UPDATE_GRADIENTS;
      break;
    case H1_NORM:
    case W1P_NORM:
      update_flags|=UPDATE_GRADIENTS;
//    no break!
    default:
      update_flags|=UPDATE_VALUES;
      break;
  }  

  HdivFEValues fe_values(mapping,*selected_fe,qmap,update_flags);
  int index=0;
  REAL diff_sum=0.;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape(),index++) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    int n_q_points=q_points.getNumber();

    NumPtr<Tensor<1,SPACEDIM> > tmp_values(n_q_points);
    NumPtr<Tensor<2,SPACEDIM> > tmp_gradients(n_q_points);
    NumPtr<Tensor<1,SPACEDIM> > weight_values(n_q_points);
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > weight_vectors(n_q_points); 
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > psi_values(n_q_points);
    NumPtr<NumPtr<Tensor<2,SPACEDIM> > > psi_grads(n_q_points);
    for (int q=0;q<n_q_points;q++) {
      weight_vectors[q].allocate(n_components);
      psi_values[q].allocate(n_components);
      psi_grads[q].allocate(n_components);
    }
    if (weight) {
      if (n_components>1) {
        weight->value(weight_vectors,q_points);
      } else {
        weight->value(weight_values,q_points,0);
        for (int q=0;q<n_q_points;q++) {
          weight_vectors[q][0]=weight_values[q];
        }
      }
    } else {
      for (int q=0;q<n_q_points;q++) {
        for (int c=0;c<n_components;c++) {
          weight_vectors[q][c].initialize(1.);
        }
      }
    }
     
    if (update_flags & UPDATE_VALUES) {
      if (fe_is_system) {
        exact_solution.value(psi_values,q_points);
      } else {
        exact_solution.value(tmp_values,q_points,0);
        for (int q=0;q<n_q_points;q++) {
          psi_values[q][0]=tmp_values[q];
        }
      }
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > > function_values;
      getFunctionValues(fe_values,element,present_level,fe_function,
        function_values);
      for (int q=0; q<n_q_points; ++q) {
        for (int c=0;c<n_components;c++) {
          psi_values[q][c]-=function_values[q][c];
        }
      }
    }
    if (update_flags & UPDATE_GRADIENTS) {
      if (fe_is_system) {
        exact_solution.gradient(psi_grads,q_points);
      } else {
        exact_solution.gradient(tmp_gradients,q_points,0);
        for (int q=0;q<n_q_points;q++) {
          psi_grads[q][0]=tmp_gradients[q];
        }
      }      
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > > function_grads;
      getFunctionDerivatives(fe_values,element,present_level,fe_function,
        function_grads);
      for (int k=0;k<n_components;k++) {
        for (int q=0;q<n_q_points;++q) {
          psi_grads[q][k]-=function_grads[q][k];
        }
      }
    }
    REAL diff=0;
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    switch (norm) {
      case MEAN_NORM:
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            psi_scalar+=psi_values[q][k]*weight_vectors[q][k];
          }
          diff+=psi_scalar*jxw_values[q];
        }
        break;
      case LP_NORM:
      case L1_NORM:
      case W1P_NORM:
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            for (int d=0;d<SPACEDIM;d++) {
              psi_scalar+=pow(abs(psi_values[q][k][d]),exponent)
                            *weight_vectors[q][k][d];
            }
          }
          diff+=psi_scalar*jxw_values[q];
        }
        break;
      case L2_NORM:
      case H1_NORM: {
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            for (int d=0;d<SPACEDIM;d++) {
              psi_scalar+=psi_values[q][k][d]*psi_values[q][k][d]
                         *weight_vectors[q][k][d];
            }
          }
          diff+=psi_scalar*jxw_values[q];
        }
        break;
      }
      case LINFTY_NORM: {
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            for (int d=0;d<SPACEDIM;d++) {
              REAL newval=
                abs(psi_values[q][k][d])*weight_vectors[q][k][d];
              if (psi_scalar<newval) psi_scalar=newval;
            }
          }
          diff=max(diff,psi_scalar);
        }
        break;
      }
      case H1_SEMINORM:
      case W1P_SEMINORM:
        break;
      default:
        OBSOLETE("not implemented");
        break;
    }
    switch (norm) {
      case MEAN_NORM:
      case LP_NORM:
      case L1_NORM:
      case L2_NORM:
        diff_sum+=diff;
        break;
      case LINFTY_NORM: {
        diff_sum=max(diff_sum,diff);
        break;
      }
      case W1P_SEMINORM:
      case W1P_NORM:
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            for (int d=0;d<SPACEDIM;d++) {
              psi_scalar+=pow(psi_grads[q][k][d]*psi_grads[q][k][d],
                exponent*0.5)*weight_vectors[q][k][d];
            }
          }
          diff+=psi_scalar*jxw_values[q];
        }
        diff_sum+=diff;
        break;
      case H1_SEMINORM:
      case H1_NORM:
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            for (int d=0;d<SPACEDIM;d++) {
              psi_scalar+=(psi_grads[q][k][d]*psi_grads[q][k][d])
                            *weight_vectors[q][k][d];
            }
          }
          diff+=psi_scalar*jxw_values[q];
        }
        diff_sum+=diff;
        break;
      default:
        break;
     }
  }
  switch (norm) {
    case LP_NORM:
    case L1_NORM:
    case W1P_SEMINORM:
    case W1P_NORM:
      diff_sum=pow(diff_sum,1./exponent);
      break;
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      diff_sum=sqrt(diff_sum);
      break;
    default:
      break;
  }
  return diff_sum;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// MEAN_NORM, L1_NORM, L2_NORM, LINFTY_NORM, LP_NORM measure value
// H1_SEMINORM, W1P_SEMINORM measure normal derivative
// H1_NORM, W1P_NORM measure value and normal derivative
REAL HdivDofHandler::integrateDifferenceOnBoundary(const Mapping &mapping,
const Vector &fe_function,const TensorFunctionSystem &exact_solution,
const map<int,Quadrature<SPACEDIM> > &qmap,
const map<int,Quadrature<SPACEDIM-1> > &bc_qmap,NORM_TYPE norm,
const TensorFunctionSystem *weight,REAL exponent_1) {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  if (weight) {
    int wnc=weight->nComponents();
    CHECK_TEST(wnc==1 || wnc==n_components);
  }

  REAL exponent=exponent_1;
  switch (norm) {
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      exponent=2.;
      break;
    case L1_NORM:
      exponent=1.;
      break;
    default:
      break;
  }
  
  UPDATE_FLAGS update_flags=UPDATE_Q_POINTS;
  switch (norm) {
    case H1_SEMINORM:
    case W1P_SEMINORM:
      update_flags|=UPDATE_GRADIENTS;
      break;
    case H1_NORM:
    case W1P_NORM:
      update_flags|=UPDATE_GRADIENTS;
//    no break!
    default:
      update_flags|=UPDATE_VALUES;
      break;
  }  

  HdivFEValues fe_values(mapping,*selected_fe,qmap,
    UPDATE_TRANSFORMATION_VALUES | UPDATE_TRANSFORMATION_GRADIENTS);
  int index=0;
  REAL diff_sum=0.;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape(),index++) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int nv=element->numberVertices();
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (!present_level->neighborLevel(f)) {
#if (SPACEDIM==1)
        NumPtr<Point<1> > p(1); p[0]=Interval::canonicalVertex(f);
        NumPtr<Point<1> > normals(1); normals[0][0]=(f==0 ? -1. : 1.);
        NumPtr<REAL> w(1); w[0]=1.;
        NumPtr<REAL> surface_measure(1); surface_measure[0]=1.;
        Quadrature<1> fq(p,w);
        map<int,Quadrature<1> > fqmap;
        fqmap[element->numberVertices()]=fq;
        HdivFEValues fev(mapping,*selected_fe,fqmap,update_flags);
        fev.mapShapes(*tess,element,present_level);
        int n_q_points=1;
        NumPtr<Point<SPACEDIM> > q_points(1);
          q_points[0]=tess->getVertices()[element->vertexIndex(f)];
#else
        int face_element_index=element->codimension1ShapeIndex(f);
        CODIMENSION1_LEVEL_CLASS *present_codimension1_level=
          static_cast<CODIMENSION1_LEVEL_CLASS*>(present_level);
        CODIMENSION1_SHAPE_CLASS *face_element=
          present_codimension1_level->getShape(face_element_index);
        int fnv=face_element->numberVertices();
        const BoundaryCondition *bc=present_codimension1_level->
          getBoundaryCondition(face_element_index);
        const Boundary *boundary=bc->getBoundary();
        NumPtr<Point<SPACEDIM> > q_points;
        NumPtr<Point<SPACEDIM> > normals;
        NumPtr<REAL> surface_measure;
        map<int,Quadrature<SPACEDIM-1> >::const_iterator
          bcq_iterator=bc_qmap.find(fnv);
        CHECK_TEST(bcq_iterator!=bc_qmap.end());
        tess->getMappedMeasureOn(*face_element,present_level,
          bcq_iterator->second.points(),q_points,normals,
          surface_measure);

        fe_values.mapShapes(*tess,element,present_level);
        int n_q_points=q_points.getNumber();
        NumPtr<Point<SPACEDIM> > c_points(n_q_points);
        const MappingDataBase *md=fe_values.mappingData(element);
        for (int q=0;q<n_q_points;q++) {
          REAL min_dist=HUGE_VAL;
          for (int v=0;v<face_element->numberVertices();v++) {
            const Point<SPACEDIM> &p=
              tess->getVertices()[face_element->vertexIndex(v)];
            if (p.distance(q_points[q])<min_dist) c_points[q]=p;
          }
          mapping.canonicalPoint(element,*tess,md,q_points[q],c_points[q]);
        }

        CHECK_TEST(bcq_iterator!=bc_qmap.end());
        Quadrature<SPACEDIM> fq(c_points,bcq_iterator->second.weights());
        map<int,Quadrature<SPACEDIM> > qm;
        for (map<int,Quadrature<SPACEDIM> >::const_iterator
        q_iterator=qmap.begin();q_iterator!=qmap.end();q_iterator++) {
          Quadrature<SPACEDIM> &qmi=
            const_cast<Quadrature<SPACEDIM>&>(q_iterator->second);
          qm[q_iterator->first]=(q_iterator->first==nv ? fq : qmi);
        }
        HdivFEValues fev(mapping,*selected_fe,qm,update_flags);
        fev.mapShapes(*tess,element,present_level);
#endif
        NumPtr<Tensor<1,SPACEDIM> > tmp_values(n_q_points);
        NumPtr<Tensor<2,SPACEDIM> > tmp_gradients(n_q_points);
        NumPtr<Tensor<1,SPACEDIM> > weight_values(n_q_points);
        NumPtr<NumPtr<Tensor<1,SPACEDIM> > > weight_vectors(n_q_points); 
        NumPtr<NumPtr<Tensor<1,SPACEDIM> > > psi_values(n_q_points);
        NumPtr<NumPtr<Tensor<2,SPACEDIM> > > psi_grads(n_q_points);
        for (int q=0;q<n_q_points;q++) {
          weight_vectors[q].allocate(n_components);
          psi_values[q].allocate(n_components);
          psi_grads[q].allocate(n_components);
        }
        if (weight) {
          if (n_components>1) {
            weight->value(weight_vectors,q_points);
          } else {
            weight->value(weight_values,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              weight_vectors[q][0]=weight_values[q];
            }
          }
        } else {
          for (int q=0;q<n_q_points;q++) {
            for (int c=0;c<n_components;c++) {
              weight_vectors[q][c].initialize(1.);
            }
          }
        }
        if (update_flags & UPDATE_VALUES) {
          if (fe_is_system) {
            exact_solution.value(psi_values,q_points);
          } else {
            exact_solution.value(tmp_values,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              psi_values[q][0]=tmp_values[q];
            }
          }
          NumPtr<NumPtr<Tensor<1,SPACEDIM> > > function_values;
          getFunctionValues(fev,element,present_level,
            fe_function,function_values);
          for (int q=0; q<n_q_points; ++q) {
            for (int c=0;c<n_components;c++) {
              psi_values[q][c]-=function_values[q][c];
            }
          }
        }
        if (update_flags & UPDATE_GRADIENTS) {
          if (fe_is_system) {
            exact_solution.gradient(psi_grads,q_points);
          } else {
            exact_solution.gradient(tmp_gradients,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              psi_grads[q][0]=tmp_gradients[q];
            }
          }      
          NumPtr<NumPtr<Tensor<2,SPACEDIM> > > function_grads;
          getFunctionDerivatives(fev,element,present_level,
            fe_function,function_grads);
          for (int q=0;q<n_q_points;++q) {
            for (int k=0;k<n_components;k++) {
              psi_grads[q][k]-=function_grads[q][k];
            }
          }
        }
        REAL diff=0;
        const NumPtr<REAL> &fq_weights=fq.weights();
        switch (norm) {
          case MEAN_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                psi_scalar+=psi_values[q][k]*weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            break;
          case LP_NORM:
          case L1_NORM:
          case W1P_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  psi_scalar+=pow(abs(psi_values[q][k][d]),exponent)
                                *weight_vectors[q][k][d];
                }
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            break;
          case L2_NORM:
          case H1_NORM: {
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  psi_scalar+=psi_values[q][k][d]*psi_values[q][k][d]
                                *weight_vectors[q][k][d];
                }
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            break;
          }
          case LINFTY_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  REAL newval=
                    abs(psi_values[q][k][d])*weight_vectors[q][k][d];
                  if (psi_scalar<newval) psi_scalar=newval;
                }
              }
              diff=max(diff,psi_scalar);
            }
            break;
          case H1_SEMINORM:
          case W1P_SEMINORM:
            break;
          default:
            OBSOLETE("not implemented");
            break;
        }
        switch (norm) {
          case MEAN_NORM:
          case LP_NORM:
          case L1_NORM:
          case L2_NORM: {
            diff_sum+=diff;
            break;
          }
          case LINFTY_NORM:
            diff_sum=max(diff_sum,diff);
          case W1P_SEMINORM:
          case W1P_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  REAL dot=abs(psi_grads[q][k][d].getValues()*normals[q]);
                  psi_scalar+=pow(dot,exponent)*weight_vectors[q][k][d];
                }
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            diff_sum+=diff;
            break;
          case H1_SEMINORM:
          case H1_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  REAL dot=psi_grads[q][k][d].getValues()*normals[q];
                  psi_scalar+=dot*dot*weight_vectors[q][k][d];
                }
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            diff_sum+=diff;
            break;
          default:
            break;
        }
      }
    }
  }
  switch (norm) {
    case LP_NORM:
    case L1_NORM:
    case W1P_SEMINORM:
    case W1P_NORM:
      diff_sum=pow(diff_sum,1./exponent);
      break;
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      diff_sum=sqrt(diff_sum);
      break;
    default:
      break;
  }
  return diff_sum;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// MEAN_NORM, L1_NORM, L2_NORM, LINFTY_NORM, LP_NORM measure value
// H1_SEMINORM, W1P_SEMINORM measure normal derivative
// H1_NORM, W1P_NORM measure value and normal derivative
REAL HdivDofHandler::integrateDifferenceOnMappedBoundary(
const Mapping &mapping,const Vector &fe_function,
const TensorFunctionSystem &exact_solution,
const map<int,Quadrature<SPACEDIM> > &qmap,
map<int,Quadrature<SPACEDIM-1> > &bc_qmap,NORM_TYPE norm,
const TensorFunctionSystem *weight,REAL exponent_1) {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  if (weight) {
    int wnc=weight->nComponents();
    CHECK_TEST(wnc==1 || wnc==n_components);
  }

  REAL exponent=exponent_1;
  switch (norm) {
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      exponent=2.;
      break;
    case L1_NORM:
      exponent=1.;
      break;
    default:
      break;
  }
  
  UPDATE_FLAGS update_flags=UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES | 
    UPDATE_Q_POINTS;
  switch (norm) {
    case H1_SEMINORM:
    case W1P_SEMINORM:
      update_flags |= UPDATE_GRADIENTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
      break;
    case H1_NORM:
    case W1P_NORM:
      update_flags |= UPDATE_GRADIENTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
//    no break!
    default:
      update_flags|=UPDATE_VALUES;
      break;
  }  

#if (SPACEDIM>=2)
  HdivFEFaceValues bc_fe_values(mapping,*selected_fe,bc_qmap,
    update_flags);
#endif
  int index=0;
  REAL diff_sum=0.;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape(),index++) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int nv=element->numberVertices();
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (!present_level->neighborLevel(f)) {
#if (SPACEDIM==1)
        NumPtr<Point<1> > p(1); p[0]=Interval::canonicalVertex(f);
        NumPtr<Point<1> > normals(1); normals[0][0]=(f==0 ? -1. : 1.);
        NumPtr<REAL> w(1); w[0]=1.;
        NumPtr<REAL> surface_measure(1); surface_measure[0]=1.;
        Quadrature<1> fq(p,w);
        map<int,Quadrature<1> > fqmap;
        fqmap[element->numberVertices()]=fq;
        HdivFEValues face_fe_values(mapping,*selected_fe,fqmap,
          update_flags);
        face_fe_values.mapShapes(*tess,element,present_level);
        int n_q_points=1;
        NumPtr<Point<SPACEDIM> > q_points(1);
          q_points[0]=tess->getVertices()[element->vertexIndex(f)];
#else

        bc_fe_values.mapShapes(*tess,element,present_level,f);
        const NumPtr<Point<SPACEDIM> > &q_points=
          bc_fe_values.getQuadraturePoints(element,f);
        int n_q_points=q_points.getNumber();
        const NumPtr<REAL> &surface_measure=
          bc_fe_values.getJxWValues(element,f);
        const NumPtr<Point<SPACEDIM> > &normals=
          bc_fe_values.getNormalVectors(element,f);
#endif
        NumPtr<Tensor<1,SPACEDIM> > tmp_values(n_q_points);
        NumPtr<Tensor<2,SPACEDIM> > tmp_gradients(n_q_points);
        NumPtr<Tensor<1,SPACEDIM> > weight_values(n_q_points);
        NumPtr<NumPtr<Tensor<1,SPACEDIM> > > weight_vectors(n_q_points); 
        NumPtr<NumPtr<Tensor<1,SPACEDIM> > > psi_values(n_q_points);
        NumPtr<NumPtr<Tensor<2,SPACEDIM> > > psi_grads(n_q_points);
        for (int q=0;q<n_q_points;q++) {
          weight_vectors[q].allocate(n_components);
          psi_values[q].allocate(n_components);
          psi_grads[q].allocate(n_components);
        }
        if (weight) {
          if (n_components>1) {
            weight->value(weight_vectors,q_points);
          } else {
            weight->value(weight_values,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              weight_vectors[q][0]=weight_values[q];
            }
          }
        } else {
          for (int q=0;q<n_q_points;q++) {
            for (int c=0;c<n_components;c++) {
              weight_vectors[q][c].initialize(1.);
            }
          }
        }
        if (update_flags & UPDATE_VALUES) {
          if (fe_is_system) {
            exact_solution.value(psi_values,q_points);
          } else {
            exact_solution.value(tmp_values,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              psi_values[q][0]=tmp_values[q];
            }
          }
          NumPtr<NumPtr<Tensor<1,SPACEDIM> > > function_values;
#if (SPACEDIM==1)
          getFunctionValues(face_fe_values,element,present_level,
            fe_function,function_values);
#else
          getFunctionValues(bc_fe_values,element,f,present_level,
            fe_function,function_values);
#endif
          for (int q=0; q<n_q_points; ++q) {
            for (int c=0;c<n_components;c++) {
              for (int d=0;d<SPACEDIM;d++) {
                psi_values[q][c][d]-=function_values[q][c][d];
              }
            }
          }
        }
        if (update_flags & UPDATE_GRADIENTS) {
          if (fe_is_system) {
            exact_solution.gradient(psi_grads,q_points);
          } else {
            exact_solution.gradient(tmp_gradients,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              psi_grads[q][0]=tmp_gradients[q];
            }
          }      
          NumPtr<NumPtr<Tensor<2,SPACEDIM> > > function_grads;
#if (SPACEDIM==1)
          getFunctionDerivatives(face_fe_values,element,present_level,
            fe_function,function_grads);
#else
          getFunctionDerivatives(bc_fe_values,element,f,present_level,
            fe_function,function_grads);
#endif
          for (int q=0;q<n_q_points;++q) {
            for (int k=0;k<n_components;k++) {
              psi_grads[q][k]-=function_grads[q][k];
            }
          }
        }
        REAL diff=0;
        switch (norm) {
          case MEAN_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                psi_scalar+=psi_values[q][k]*weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q];
            }
            break;
          case LP_NORM:
          case L1_NORM:
          case W1P_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  psi_scalar+=pow(abs(psi_values[q][k][d]),exponent)
                                *weight_vectors[q][k][d];
                }
              }
              diff+=psi_scalar*surface_measure[q];
            }
            break;
          case L2_NORM:
          case H1_NORM: {
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  psi_scalar+=psi_values[q][k][d]*psi_values[q][k][d]
                                *weight_vectors[q][k][d];
                }
              }
              diff+=psi_scalar*surface_measure[q];
            }
            break;
          }
          case LINFTY_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  REAL newval=
                    abs(psi_values[q][k][d])*weight_vectors[q][k][d];
                  if (psi_scalar<newval) psi_scalar=newval;
                }
              }
              diff=max(diff,psi_scalar);
            }
            break;
          case H1_SEMINORM:
          case W1P_SEMINORM:
            break;
          default:
            OBSOLETE("not implemented");
            break;
        }
        switch (norm) {
          case MEAN_NORM:
          case LP_NORM:
          case L1_NORM:
          case L2_NORM: {
            diff_sum+=diff;
            break;
          }
          case LINFTY_NORM:
            diff_sum=max(diff_sum,diff);
          case W1P_SEMINORM:
          case W1P_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  REAL dot=abs(psi_grads[q][k][d].getValues()*normals[q]);
                  psi_scalar+=pow(dot,exponent)*weight_vectors[q][k][d];
                }
              }
              diff+=psi_scalar*surface_measure[q];
            }
            diff_sum+=diff;
            break;
          case H1_SEMINORM:
          case H1_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                for (int d=0;d<SPACEDIM;d++) {
                  REAL dot=psi_grads[q][k][d].getValues()*normals[q];
                  psi_scalar+=dot*dot*weight_vectors[q][k][d];
                }
              }
              diff+=psi_scalar*surface_measure[q];
            }
            diff_sum+=diff;
            break;
          default:
            break;
        }
      }
#if (SPACEDIM>=2)
      offset+=bc_fe_values.getFaceQuadrature(element,f).numberPoints();
#endif
    }
  }
  switch (norm) {
    case LP_NORM:
    case L1_NORM:
    case W1P_SEMINORM:
    case W1P_NORM:
      diff_sum=pow(diff_sum,1./exponent);
      break;
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      diff_sum=sqrt(diff_sum);
      break;
    default:
      break;
  }
  return diff_sum;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::printOn(ostream &os) const {
  os << "HdivDofHandler:" << endl;
  os << "\ttess = " << tess
     << "\n\tselected_fe = " << selected_fe
     << "\n\tused_dofs = " << used_dofs
     << "\n\tlevels.getNumber = " << levels.getNumber()
     << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::printElementDofs() const {
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel()) {
    for (bool doit=tess->firstShape(level);
    doit && tess->presentLevel()==level;doit=tess->nextShapeInLevel()) {
      const SHAPE_CLASS *element=level->getShape();
      cout << "\tdof_index[" << element->elementDataIndex() << "] = "
           << element->dofIndex() << endl;
    }
  }
#if (SPACEDIM>1)
  for (CODIMENSION1_LEVEL_CLASS *level=
  static_cast<CODIMENSION1_LEVEL_CLASS*>(tess->coarsestLevel());
  level;level=level->finerLevel()) {
    for (bool doit=tess->firstCodimension1Shape(level);
    doit && tess->presentCodimension1Level()==level;
    doit=tess->nextCodimension1ShapeInLevel()) {
      const CODIMENSION1_SHAPE_CLASS *element=level->getShape();
      cout << "\tdof_index[" << element->elementDataIndex() << "] = "
           << element->dofIndex() << endl;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofHandler::printDofs() const {
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel()) {
    for (bool doit=tess->firstShape(level);
    doit && tess->presentLevel()==level;doit=tess->nextShapeInLevel()) {
      const SHAPE_CLASS *element=level->getShape();
      NumPtr<int> dof_indices;
      getDofIndices(element,level,dof_indices);
      cout << "\tdof_indices[" << element->elementDataIndex() << "] = ";
      for (int d=0;d<dof_indices.getNumber();d++) {
        cout << dof_indices[d] << " ";
      }
      cout << endl;
    }
  }
#if (SPACEDIM>1)
  for (CODIMENSION1_LEVEL_CLASS *level=
  static_cast<CODIMENSION1_LEVEL_CLASS*>(tess->coarsestLevel());
  level;level=level->finerLevel()) {
    for (bool doit=tess->firstCodimension1Shape(level);
    doit && tess->presentCodimension1Level()==level;
    doit=tess->nextCodimension1ShapeInLevel()) {
      const CODIMENSION1_SHAPE_CLASS *element=level->getShape();
      NumPtr<int> dof_indices;
      getFaceDofIndices(element,level,dof_indices);
      cout << "\tdof_indices[" << element->elementDataIndex() << "] = ";
      for (int d=0;d<dof_indices.getNumber();d++) {
        cout << dof_indices[d] << " ";
      }
      cout << endl;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template void HdivDofHandler::getFunctionValues<Tensor<1,SPACEDIM> >(
  const HdivFEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,
  const Vector&,NumPtr<Tensor<1,SPACEDIM> >&) const;
template void HdivDofHandler::getFunctionValues<NumPtr<Tensor<1,SPACEDIM> > >(
  const HdivFEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,
  const Vector&,NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
#if (SPACEDIM>1)
template void HdivDofHandler::getFunctionValues<Tensor<1,SPACEDIM> >(
  const HdivFEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<Tensor<1,SPACEDIM> >&) const;
template void HdivDofHandler::getFunctionValues<NumPtr<Tensor<1,SPACEDIM> > >(
  const HdivFEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
#endif

typedef Tensor<2,SPACEDIM> SCALAR_GRAD_TYPE;
typedef NumPtr<Tensor<2,SPACEDIM> > VECTOR_GRAD_TYPE;
template void HdivDofHandler::getFunctionDerivatives<SCALAR_GRAD_TYPE>(
  const HdivFEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
  NumPtr<SCALAR_GRAD_TYPE>&) const;
template void HdivDofHandler::getFunctionDerivatives<VECTOR_GRAD_TYPE>(
  const HdivFEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
  NumPtr<VECTOR_GRAD_TYPE>&) const;
#if (SPACEDIM>1)
template void HdivDofHandler::getFunctionDerivatives<SCALAR_GRAD_TYPE>(
  const HdivFEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<SCALAR_GRAD_TYPE>&) const;
template void HdivDofHandler::getFunctionDerivatives<VECTOR_GRAD_TYPE>(
  const HdivFEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<VECTOR_GRAD_TYPE>&) const;
#endif

typedef Tensor<3,SPACEDIM> SCALAR_GRAD_GRAD_TYPE;
typedef NumPtr<Tensor<3,SPACEDIM> > VECTOR_GRAD_GRAD_TYPE;
template void HdivDofHandler::getFunction2ndDerivatives<SCALAR_GRAD_GRAD_TYPE>(
  const HdivFEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
  NumPtr<SCALAR_GRAD_GRAD_TYPE>&) const;
template void HdivDofHandler::getFunction2ndDerivatives<VECTOR_GRAD_GRAD_TYPE>(
  const HdivFEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
  NumPtr<VECTOR_GRAD_GRAD_TYPE>&) const;
#if (SPACEDIM>1)
template void HdivDofHandler::getFunction2ndDerivatives<SCALAR_GRAD_GRAD_TYPE>(
  const HdivFEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<SCALAR_GRAD_GRAD_TYPE>&) const;
template void HdivDofHandler::getFunction2ndDerivatives<VECTOR_GRAD_GRAD_TYPE>(
  const HdivFEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<VECTOR_GRAD_GRAD_TYPE>&) const;
#endif

#if (SPACEDIM>1)
template class _Rb_tree<
  Shape2 const*,
  Shape2 const*,
  _Identity<Shape2 const*>,
  less<Shape2 const*>,
  allocator<Shape2 const*>
>;

template class _Rb_tree<
  int,
  pair<int const, 
       set<Shape2 const*,
           less<Shape2 const*>,
           allocator<Shape2 const*>
          >
      >,
  _Select1st<pair<int const,
                  set<Shape2 const*,
                      less<Shape2 const*>,
                      allocator<Shape2 const*>
                     >
                 >
            >,
  less<int>,
  allocator<pair<int const,
                 set<Shape2 const*,
                     less<Shape2 const*>,
                     allocator<Shape2 const*>
                    >
                >
           >
>;
template class _Rb_tree<
  pair<Shape2 const*, Level2 const*>,
  pair<Shape2 const*, Level2 const*>,
  _Identity<pair<Shape2 const*, Level2 const*> >,
  less<pair<Shape2 const*, Level2 const*> >,
  allocator<pair<Shape2 const*, Level2 const*> >
>;
template class _Rb_tree<
  Shape1 const*,
  pair<Shape1 const* const,
       set<pair<Shape2 const*, Level2 const*>,
           less<pair<Shape2 const*, Level2 const*> >,
           allocator<pair<Shape2 const*, Level2 const*> >
          >
      >,
  _Select1st<pair<Shape1 const* const,
                  set<pair<Shape2 const*, Level2 const*>,
                      less<pair<Shape2 const*, Level2 const*> >,
                      allocator<pair<Shape2 const*, Level2 const*> >
                     >
                 >
            >,
  less<Shape1 const*>,
  allocator<pair<Shape1 const* const,
                 set<pair<Shape2 const*,Level2 const*>,
                     less<pair<Shape2 const*, Level2 const*> >,
                     allocator<pair<Shape2 const*, Level2 const*> >
                    >
                >
           >
>;
#endif
#if (SPACEDIM==3)
template class _Rb_tree<
  Shape2 const*,
  pair<Shape2 const* const,
       set<pair<Shape3 const*, Level3 const*>,
           less<pair<Shape3 const*, Level3 const*> >,
           allocator<pair<Shape3 const*, Level3 const*> >
          >
      >,
  _Select1st<pair<Shape2 const* const,
                  set<pair<Shape3 const*, Level3 const*>,
                      less<pair<Shape3 const*, Level3 const*> >,
                      allocator<pair<Shape3 const*, Level3 const*> >
                     >
                 >
            >,
  less<Shape2 const*>,
  allocator<pair<Shape2 const* const,
                 set<pair<Shape3 const*, Level3 const*>,
                     less<pair<Shape3 const*, Level3 const*> >,
                     allocator<pair<Shape3 const*, Level3 const*> >
                    >
                >
           >
>;
template class _Rb_tree<
  pair<Shape3 const*, Level3 const*>,
  pair<Shape3 const*, Level3 const*>,
  _Identity<pair<Shape3 const*, Level3 const*> >,
  less<pair<Shape3 const*, Level3 const*> >,
  allocator<pair<Shape3 const*, Level3 const*> >
>;
#endif

#include "NumPtr.C"
INSTANTIATE_NUMPTR(HdivDofLevel*);
template class NumPtrBase<NumPtr<Tensor<2,SPACEDIM> >,int>;
template class NumPtr<NumPtr<Tensor<2,SPACEDIM> > >;
