#ifndef HDIV_DOF_HANDLER_H
#define HDIV_DOF_HANDLER_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <map>
#include <set>
#include <BoundaryCondition.H>
#include <CompressedSparsityPattern.H>
#include <DDecl.H>
#include <DofHandler.H>
#include <HdivDofLevel.H>
#include <HdivFEValues.H>
#include <HdivFiniteElement.H>
#include <NumPtr.H>
#include <SparseMatrix.H>
#include <Tessellation.H>

class HdivDofHandler {
  private:
//  can't use map<const LEVEL_CLASS*,DOF_LEVEL_CLASS*> because
//    getLevel takes pointer into LEVEL_CLASS* data structure
    NumPtr<HdivDofLevel*> levels;
    int used_dofs;
    int total_solution_dofs;

    HdivDofHandler(const HdivDofHandler&) {
      OBSOLETE("copy constructor disallowed");
    }
    HdivDofHandler& operator=(const HdivDofHandler&) {
      OBSOLETE("copy constructor disallowed");
    }
    void reserveSpace();
  protected:
    Tessellation *tess;
    HdivFiniteElement *selected_fe;

//  deletes levels and vertex_dofs
//    called from HdivDofHandler::reserveSpace and MGHdivDofHandler::clear
    void clearSpace();
    HdivDofLevel* getLevel(const LEVEL_CLASS *level) const {
      return levels[level->numberCoarserLevels()];
    }
    HdivDofLevel* getLevel(const CODIMENSION1_LEVEL_CLASS *level) const {
      return levels[level->numberCoarserLevels()];
    }
  
//  if shape is refined, gets dofs for interpolation
//  called from getFunctionValues, getFunctionGrads and
//    getFunction2ndDerivatives    
    void getInterpolatedDofValues(const SHAPE_CLASS*,const LEVEL_CLASS*,
      const Vector&,Vector&) const;
  public:
    HdivDofHandler(Tessellation &t,HdivFiniteElement &fe) : 
    levels(t.numberLevels()),tess(&t),selected_fe(&fe),used_dofs(0),
    total_solution_dofs(0) {
      levels.initialize(0); 
    }
    virtual ~HdivDofHandler() { clearSpace(); }

    int numberDofs() const { return used_dofs; } // both soln & multiplier
    int totalSolutionDofs() const { return total_solution_dofs; }
    HdivFiniteElement& getFe() const {
      CHECK_POINTER(selected_fe);
      return *selected_fe;
    }
    Tessellation& getTessellation() const { return *tess; }

//  pulls local solution and multiplier dof's out of global dof's 
//  called from getInterpolatedDofValues and 
//    RefinementStudy::recursivelyIntegrateError
//  public for RefinementStudy
    void getDofValues(const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
      Vector&) const;
//  stores first global dof in each active element of tessellation
//  called from user programs
    virtual void distributeDofs();
//  called from computeCuthillMcKeeRenumbering, Multigrid::Multigrid and
//    user programs
    int maxCouplingsBetweenDofs() const;
//  called from computeCuthillMcKeeRenumbering and user programs
    void makeSparsityPattern(SparsityPatternBase&,
      SparsityPatternBase&) const;
//  gets both solution and multiplier dof indices
//  called from HdivDofHandler::getDofValues,
//    HdivDofHandler::makeSparsityPattern
//    HdivDofHandler::interpolateBoundaryValues,
//    ConstraintMatrix::makeHangingNodeConstraints
//    MGTransferPrebuilt::copyToMG,MGTransferPrebuilt::copyFromMG,
//      MGTransferPrebuilt::copyFromMGAdd and user programs
//  useful in forming global stiffness matrix from cell matrices
    void getDofIndices(const SHAPE_CLASS*,const LEVEL_CLASS*,
      NumPtr<int>&) const;
    void getFaceDofIndices(const SHAPE_CLASS*,int,const LEVEL_CLASS*,
      NumPtr<int>&) const;
#if (SPACEDIM>1)
    void getFaceDofIndices(const CODIMENSION1_SHAPE_CLASS*,
      const CODIMENSION1_LEVEL_CLASS*,NumPtr<int>&) const;
#endif
//  assembles linear system except for Dirichlet boundary values
//    called from user programs
    void assembleLinearSystem(const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,SparseMatrix&,SparseMatrix&,
      Vector&) const;
//  sets Dirichlet boundary values
//    called from user programs
    void interpolateBoundaryValues(const Mapping&,map<int,REAL>&) const;
    void interpolateBoundaryValues(const Mapping&,map<int,REAL>&,
      map<int,Quadrature<SPACEDIM-1> >&) const;
//  puts Dirichlet boundary values into linear system
//    called from user programs
    void applyBoundaryValues(const map<int,double>&,SparseMatrix&,
      SparseMatrix&,Vector&,bool) const;
    void applyBoundaryValues(const map<int,double>&,Vector&) const;
//  computes norm of difference of numerical and exact solutions
//    called from user programs
    REAL integrateDifference(const Mapping&,const Vector&,
      const TensorFunctionSystem&,const map<int,Quadrature<SPACEDIM> >&,
      NORM_TYPE,const TensorFunctionSystem*,REAL);
    REAL integrateDifferenceOnBoundary(const Mapping&,const Vector&,
      const TensorFunctionSystem&,const map<int,Quadrature<SPACEDIM> >&,
      const map<int,Quadrature<SPACEDIM-1> >&,NORM_TYPE,
      const TensorFunctionSystem*,REAL);
    REAL integrateDifferenceOnMappedBoundary(const Mapping&,const Vector&,
      const TensorFunctionSystem&,const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,NORM_TYPE,
      const TensorFunctionSystem*,REAL);

#if (SPACEDIM>1)
//  called from ConstraintMatrix::makeHangingNodeConstraints
    void dofsOnChildren(const CODIMENSION1_SHAPE_CLASS*,
      const CODIMENSION1_LEVEL_CLASS*,NumPtr<int>&) const;
#endif
//  T = Tensor<1,SPACEDIM> or NumPtr<Tensor<1,SPACEDIM> >
//  called from HdivDofHandler::integrateDifference,
//    HdivDofHandler::integrateDifferenceOnBoundary,
//    HdivDofHandler::integrateDifferenceOnMappedBoundary,
//    DataOut::buildSomePatches, 
//    Patch::plotClipPlane,
//    RefinementStudy::RefinementStudy,
//    RefinementStudy::recursivelyIntegrateError
    template<class T> void getFunctionValues(const HdivFEValues &fe_values,
    const SHAPE_CLASS *e,const LEVEL_CLASS *level,
    const Vector &fe_function,NumPtr<T> &values) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_values.getFunctionValues(dof_values,e,values);
    }
#if (SPACEDIM>1)
    template<class T> void getFunctionValues(
    const HdivFEFaceValues &fe_face_values,const SHAPE_CLASS *e,int f,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &values) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_face_values.getFunctionValues(dof_values,e,f,values);
    }
#endif
//  for eventual call to FiniteElement::sumShapeDerivatives
//    T = Tensor<1,SPACEDIM> or NumPtr<Tensor<1,SPACEDIM>
//  called from KellyErrorEstimator::integrateOverRegularFace and
//    KellyErrorEstimator::integrateOverIrregularFace,
//    KellyErrorEstimator::work, RefinementStudy::RefinementStudy and
//    RefinementStudy::recursivelyIntegrateError
    template<class T> void getFunctionDerivatives(
    const HdivFEValues &fe_values,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &gradients) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_values.getFunctionDerivatives(dof_values,e,gradients);
    }
#if (SPACEDIM>1)
    template<class T> void getFunctionDerivatives(
    const HdivFEFaceValues &fe_face_values,const SHAPE_CLASS *e,int f,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &gradients) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_face_values.getFunctionDerivatives(dof_values,e,f,gradients);
    }
#endif
//  for eventual call to FiniteElement::sumShape2ndDerivatives
//    T = Tensor<2,SPACEDIM> or NumPtr<Tensor<2,SPACEDIM> >
    template<class T> void getFunction2ndDerivatives(
    const HdivFEValues &fe_values,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &second_derivatives) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_values.getFunction2ndDerivatives(dof_values,e,second_derivatives);
    }
#if (SPACEDIM>1)
    template<class T>void getFunction2ndDerivatives(
    const HdivFEFaceValues &fe_face_values,const SHAPE_CLASS *e,int f,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &second_derivatives) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_face_values.getFunction2ndDerivatives(dof_values,e,f,
        second_derivatives);
    }
#endif
//  T = REAL or Vector
    template<class T> void getMultiplierFunctionValues(
    const HdivFEValues &fe_values,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level,const Vector &fe_function,
    NumPtr<T> &values) const {
      CHECK_SAME(fe_function.size(),numberDofs());
      Vector dof_values;
      getInterpolatedDofValues(e,level,fe_function,dof_values);
      fe_values.getMultiplierFunctionValues(dof_values,e,values);
    }
    virtual void printOn(ostream&) const;
    virtual void printDofs() const;
    virtual void printElementDofs() const;
};

#endif
