#ifndef HDIV_FE_VALUES_H
#define HDIV_FE_VALUES_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <algorithm>
#include <Array.H>
#include <HdivFiniteElement.H>
#include <Quadrature.H>
#include <Tensor.H>
#include <Types.H>

//data for finite element shape function(x);
//  ie, mapped from FiniteElementDataBase values
//  the mapping affects derivatives, not values
//the functions in this class should be used only by HdivFEValuesBase and 
//  derived classes
class HdivFEValuesData {
  protected:
//  these are set in
//    HdivFEFaceValues::mapShapes
//    HdivFESubfaceValues::mapShapes
//    HdivFEValues::mapShapes
    UPDATE_FLAGS update_flags;
    int number_quadrature_points;
    int number_vertices;
    bool shape_values_ok;
    bool shape_gradients_ok;
    bool shape_2nd_derivatives_ok;
    bool shape_divergence_ok;
    bool multiplier_shape_values_ok;
    bool multiplier_shape_gradients_ok;
    bool multiplier_shape_2nd_derivatives_ok;
    bool JxW_values_ok;
    bool quadrature_points_ok;
    const HdivFiniteElement *finite_element;
    Array<2,Tensor<1,SPACEDIM> > shape_values;
    Array<2,Tensor<2,SPACEDIM> > shape_gradients;
    Array<2,Tensor<3,SPACEDIM> > shape_2nd_derivatives;
    Array<2,REAL> shape_divergence;
    Array<2,REAL> multiplier_shape_values;
    NumPtr<REAL> JxW_values;
    NumPtr<Point<SPACEDIM> > quadrature_points;
  public:
    HdivFEValuesData(int,int,const HdivFiniteElement&,UPDATE_FLAGS);
    virtual ~HdivFEValuesData() {}
//  called from HdivFEFaceValues::HdivFEFaceValues and 
//    HdivFESubfaceValues::HdivFESubfaceValues
    void initializeJxWValues(int,UPDATE_FLAGS);
//  called from HdivFEValues::mapShapes
    void mapShapes(MappingDataBase*,HdivFiniteElementDataBase*,
      const SHAPE_CLASS*,const Tessellation&,const LEVEL_CLASS*);
//  called from HdivFEFaceValues::mapShapes
    void mapFaceShapes(MappingDataBase*,HdivFiniteElementDataBase*,
      const SHAPE_CLASS*,int,const Tessellation&,const LEVEL_CLASS*,
      const Quadrature<SPACEDIM-1>&,NumPtr<Tensor<1,SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&,int,bool&,bool&);
//  called from HdivFESubfaceValues::mapShapes
    void mapSubFaceShapes(MappingDataBase*,HdivFiniteElementDataBase*,
      const SHAPE_CLASS*,int,const Tessellation&,const LEVEL_CLASS*,
      const Quadrature<SPACEDIM-1>&,NumPtr<Tensor<1,SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&,int,int,bool&,bool&);
//  in dof order; called from HdivFEValuesBase::getShapeValues
    const Array<2,Tensor<1,SPACEDIM> >& getShapeValues() const { 
      CHECK_TEST(shape_values_ok);
      return shape_values; 
    }
//  in dof order; called from HdivFEValuesBase::getShapeGradients
    const Array<2,Tensor<2,SPACEDIM> >& getShapeGradients() const { 
      CHECK_TEST(shape_gradients_ok);
      return shape_gradients;
    }
//  in dof order; called from HdivFEValuesBase::getShapeDivergence
    const Array<2,REAL>& getShapeDivergence() const { 
      CHECK_TEST(shape_divergence_ok);
      return shape_divergence;
    }
//  in dof order; called from HdivFEValuesBase::getShape2ndDerivatives
    const Array<2,Tensor<3,SPACEDIM> >& getShape2ndDerivatives() const { 
      CHECK_TEST(shape_2nd_derivatives_ok);
      return shape_2nd_derivatives;
    }
//  in dof order; called from HdivFEValuesBase::getMultiplierShapeValues
    const Array<2,REAL>& getMultiplierShapeValues() const { 
      CHECK_TEST(multiplier_shape_values_ok);
      return multiplier_shape_values; 
    }
//  called from HdivFEValuesBase::getJxWValues
    const NumPtr<REAL>& getJxWValues() const { 
      CHECK_TEST(JxW_values_ok);
      return JxW_values; 
    }
//  called from HdivFEValuesBase::getQuadraturePoints
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints() const {
      CHECK_TEST(quadrature_points_ok);
      return quadrature_points;
    }
    virtual void printOn(ostream&) const;
};

class HdivFEValuesBase {
  protected:
    map<int,int> n_quadrature_points;
    map<int,int> dofs_per_cell;
    map<int,MappingDataBase*> mapping_data_map;
    map<int,HdivFiniteElementDataBase*> fe_data_map;
    map<int,HdivFEValuesData*> fe_values_data_map;
    const Mapping *mapping;
    HdivFiniteElement *fe;
    UPDATE_FLAGS update_flags;

    void insertNQuadraturePoint(int nv,int nqp) {
      n_quadrature_points[nv]=nqp;
    }
    void insertDofsPerCell(int nv,int dpc) {
      dofs_per_cell[nv]=dpc;
    }
    const HdivFEValuesData* getFEValuesData(int nv) const {
      map<int,HdivFEValuesData*>::const_iterator fevdm_iterator=
        fe_values_data_map.find(nv);
      if (fevdm_iterator!=fe_values_data_map.end()) {
        return fevdm_iterator->second;
      }
      return 0;
    }
    const HdivFEValuesData* getFEValuesData(const Shape *e) const {
      return getFEValuesData(e->numberVertices());
    }
    HdivFEValuesData* getFEValuesData(int nv) {
      map<int,HdivFEValuesData*>::iterator fevdm_iterator=
        fe_values_data_map.find(nv);
      if (fevdm_iterator!=fe_values_data_map.end()) {
        return fevdm_iterator->second;
      }
      return 0;
    }
    HdivFEValuesData* getFEValuesData(const Shape *e) {
      return getFEValuesData(e->numberVertices());
    }
    const Array<2,Tensor<1,SPACEDIM> >* getShapeValues(int nv) const { 
      const HdivFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeValues()) : 0);
    }
    const Array<2,Tensor<2,SPACEDIM> >* getShapeGradients(int nv) const { 
      const HdivFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeGradients()) : 0);
    }
    const Array<2,REAL>* getShapeDivergence(int nv) const { 
      const HdivFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShapeDivergence()) : 0);
    }
    const Array<2,Tensor<3,SPACEDIM> >* getShape2ndDerivatives(int nv) 
    const { 
      const HdivFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ? &(fe_values_data->getShape2ndDerivatives()) 
                             : 0);
    }
    const Array<2,REAL>* getMultiplierShapeValues(int nv) const { 
      const HdivFEValuesData *fe_values_data=getFEValuesData(nv);
      return (fe_values_data ?
              &(fe_values_data->getMultiplierShapeValues()) : 0);
    }

    const NumPtr<REAL>& getJxWValues(int nv) const {
      const HdivFEValuesData *fe_values_data=getFEValuesData(nv);
      CHECK_POINTER(fe_values_data);
      return fe_values_data->getJxWValues() ;
    }
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints(int nv) const {
      const HdivFEValuesData *fe_values_data=getFEValuesData(nv);
      CHECK_POINTER(fe_values_data);
      return fe_values_data->getQuadraturePoints() ;
    }
    int dofsPerCell(int nv) const { 
      map<int,int>::const_iterator dpc_iterator=
        dofs_per_cell.find(nv);
      CHECK_TEST(dpc_iterator!=dofs_per_cell.end());
      return dpc_iterator->second;
    }
    HdivFiniteElement& getFe() const { return *fe; }
//  if shape_function->requiresShapeToComputeValues,
//  can't compute values here because we can't get the shape_function
    HdivFiniteElementDataBase* getFEData(const Shape *e) {
      map<int,HdivFiniteElementDataBase*>::iterator
        fedm_iterator=fe_data_map.find(e->numberVertices());
      CHECK_TEST(fedm_iterator!=fe_data_map.end());
      return fedm_iterator->second;
    }
  public:
    HdivFEValuesBase(UPDATE_FLAGS flags,const Mapping &m,
    HdivFiniteElement &fe_in) :
      update_flags(flags),mapping(&m),fe(&fe_in) { }
    virtual ~HdivFEValuesBase();
    const Mapping& getMapping() const { return *mapping; }
    const NumPtr<REAL>& getJxWValues(const Shape *e) const {
      return getJxWValues(e->numberVertices());
    }
    const NumPtr<Point<SPACEDIM> >& getQuadraturePoints(const Shape *e) 
    const {
      return getQuadraturePoints(e->numberVertices());
    }
    int dofsPerCell(const Shape *e) const { 
      return dofsPerCell(e->numberVertices());
    }
    int numberQuadraturePoints(const Shape *e) const {
      map<int,int>::const_iterator nqp_iterator=
        n_quadrature_points.find(e->numberVertices());
      CHECK_TEST(nqp_iterator!=n_quadrature_points.end());
      return nqp_iterator->second;
    }
    MappingDataBase* mappingData(int nv) {
      map<int,MappingDataBase*>::iterator
        mdm_iterator=mapping_data_map.find(nv);
      CHECK_TEST(mdm_iterator!=mapping_data_map.end());
      return mdm_iterator->second;
    }
    const MappingDataBase* mappingData(int nv) const {
      map<int,MappingDataBase*>::const_iterator
        mdm_iterator=mapping_data_map.find(nv);
      CHECK_TEST(mdm_iterator!=mapping_data_map.end());
      return mdm_iterator->second;
    }
    MappingDataBase* mappingData(const Shape *e) {
      return mappingData(e->numberVertices());
    }
    const MappingDataBase* mappingData(const Shape *e) const {
      return mappingData(e->numberVertices());
    }
//  if shape_function->requiresShapeToComputeValues,
//  can't compute values here because we can't get the shape_function
    const HdivFiniteElementDataBase* getFEData(const Shape *e) const {
      map<int,HdivFiniteElementDataBase*>::const_iterator
        fedm_iterator=fe_data_map.find(e->numberVertices());
      CHECK_TEST(fedm_iterator!=fe_data_map.end());
      return fedm_iterator->second;
    }
//  called from HdivDofHandler::assembleLinearSystem
    const Array<2,Tensor<1,SPACEDIM> >* getShapeValues(const Shape *e)
    const { 
      return getShapeValues(e->numberVertices());
    }
    const Array<2,Tensor<2,SPACEDIM> >* getShapeGradients(
    const Shape *e) const { 
      return getShapeGradients(e->numberVertices());
    }
    const Array<2,REAL>* getShapeDivergence(const Shape *e) const { 
      return getShapeDivergence(e->numberVertices());
    }
    const Array<2,Tensor<3,SPACEDIM> >* getShape2ndDerivatives(
    const Shape *e) const { 
      return getShape2ndDerivatives(e->numberVertices());
    }
    const Array<2,REAL>* getMultiplierShapeValues(const Shape *e)
    const { 
      return getMultiplierShapeValues(e->numberVertices());
    }
    virtual void printOn(ostream&) const;
};

class HdivFEValues : public HdivFEValuesBase {
  private:
    map<int,Quadrature<SPACEDIM> > quadrature;

    Quadrature<SPACEDIM>& getQuadrature(int nv) {
      map<int,Quadrature<SPACEDIM> >::iterator q_iterator
        =quadrature.find(nv);
      CHECK_TEST(q_iterator!=quadrature.end());
      return q_iterator->second;
    }
    const Quadrature<SPACEDIM>& getQuadrature(int nv) const {
      map<int,Quadrature<SPACEDIM> >::const_iterator q_iterator
        =quadrature.find(nv);
      CHECK_TEST(q_iterator!=quadrature.end());
      return q_iterator->second;
    }
  public:
    HdivFEValues(const Mapping&,HdivFiniteElement&,
      const map<int,Quadrature<SPACEDIM> >&,UPDATE_FLAGS);
    virtual ~HdivFEValues() {;}
//  compute mapped values of shape functions/derivatives in dof order
    void mapShapes(const Tessellation &tess,const SHAPE_CLASS *e,
    const LEVEL_CLASS *level) {
      getFEValuesData(e)->
        mapShapes(mappingData(e),getFEData(e),e,tess,level);
    }
    const Quadrature<SPACEDIM>& getQuadrature(const Shape *e) const {
      return getQuadrature(e->numberVertices());
    }
//  called from HdivDofHandler::getFunctionValues
//    T = Tensor<1,SPACEDIM> or NumPtr<Tensor<1,SPACEDIM> >
    template<class T> void getFunctionValues(const Vector &dof_values,
    const SHAPE_CLASS *e,NumPtr<T> &values) const {
      CHECK_TEST(update_flags & UPDATE_VALUES);
      fe->sumShapeValues(e,*getShapeValues(e),numberQuadraturePoints(e),
        0,dof_values,values);
    }
//  for eventual call to FiniteElement::sumShapeDerivatives
//    T = Tensor<2,SPACEDIM> or NumPtr<Tensor<2,SPACEDIM> >
    template<class T> void getFunctionDerivatives(
    const Vector &dof_values,const SHAPE_CLASS *e,NumPtr<T> &gradients) 
    const {
      CHECK_TEST(update_flags & UPDATE_GRADIENTS); 
      fe->sumShapeDerivatives(e,*getShapeGradients(e),
        numberQuadraturePoints(e),0,dof_values,gradients);
    }
//  for eventual call to FiniteElement::sumShapeDivergence
//    T = REAL or Vector
    template<class T> void getFunctionDivergence(
    const Vector &dof_values,const SHAPE_CLASS *e,NumPtr<T> &divergence) 
    const {
      CHECK_TEST(update_flags & UPDATE_DIVERGENCE); 
      fe->sumShapeDivergence(e,*getShapeDivergence(e),
        numberQuadraturePoints(e),0,dof_values,divergence);
    }
//  for eventual call to FiniteElement::sumShape2ndDerivatives
//    T = Tensor<3,SPACEDIM> or NumPtr<Tensor<3,SPACEDIM> >
    template<class T> void getFunction2ndDerivatives(
    const Vector &dof_values,const SHAPE_CLASS *e,
    NumPtr<T> &second_derivatives) const {
      CHECK_TEST(update_flags & UPDATE_SECOND_DERIVATIVES); 
      fe->sumShape2ndDerivatives(e,*getShape2ndDerivatives(e),
        numberQuadraturePoints(e),0,dof_values,second_derivatives);
    }
//  called from HdivDofHandler::getMultiplierFunctionValues
//    T = REAL or Vector
    template<class T> void getMultiplierFunctionValues(
    const Vector &dof_values,const SHAPE_CLASS *e,NumPtr<T> &values)
    const {
      CHECK_TEST(update_flags & UPDATE_VALUES);
      fe->sumMultiplierShapeValues(e,*getMultiplierShapeValues(e),
        numberQuadraturePoints(e),0,dof_values,values);
    }
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM>=2)
class HdivFEFaceValues : public HdivFEValuesBase {
  private:
    map<int,Quadrature<SPACEDIM>*> all_face_quadrature;

//  multipliers have no boundary trace values
    const Array<2,REAL>* getMultiplierShapeValues(const Shape *e)
    const { 
      OBSOLETE("should not be called for this class");
    }
  protected:
    map<int,Quadrature<SPACEDIM-1> > face_quadrature;
    map<int,NumPtr<Point<SPACEDIM> > > normal_vectors;
//  (d x / d xi)^{-T} n_{xi} det(d x / d xi):
    map<int,NumPtr<Tensor<1,SPACEDIM> > > boundary_forms;
    bool boundary_forms_ok;
    bool normal_vectors_ok;

    void initializeBoundaryFormsAndNormals(int,
      const Quadrature<SPACEDIM-1>&);

    const Quadrature<SPACEDIM-1>& getFaceQuadrature(int nv) const {
      map<int,Quadrature<SPACEDIM-1> >::const_iterator
        q_iterator=face_quadrature.find(nv);
      CHECK_TEST(q_iterator!=face_quadrature.end());
      return q_iterator->second;
    }
    const NumPtr<Point<SPACEDIM> >& getNormalVectors(int nv) const {
      CHECK_TEST(normal_vectors_ok);
      map<int,NumPtr<Point<SPACEDIM> > >::const_iterator
        nv_iterator=normal_vectors.find(nv);
      CHECK_TEST(nv_iterator!=normal_vectors.end());
      return nv_iterator->second;
    }
    int faceOffset(const SHAPE_CLASS*,int) const;
  public:
    HdivFEFaceValues(const Mapping&,HdivFiniteElement&, 
      map<int,Quadrature<SPACEDIM-1> >&,UPDATE_FLAGS);
    virtual ~HdivFEFaceValues();
    const Quadrature<SPACEDIM-1>& getFaceQuadrature(const Shape2 *e,int f) 
    const {
      CHECK_TEST(f<e->numberCodimension1Shapes());
      return getFaceQuadrature(e->numberVerticesOnFace(f));
    }
    void getCanonicalQuadraturePoints(const SHAPE_CLASS *e,int f,
    NumPtr<Point<SPACEDIM> > &cp) const {
      int s=faceOffset(e,f);
      map<int,Quadrature<SPACEDIM>*>::const_iterator afq_iterator=
        all_face_quadrature.find(e->numberVertices());
      CHECK_TEST(afq_iterator!=all_face_quadrature.end());
      cp=NumPtr<Point<SPACEDIM> >(afq_iterator->second->points(),
        s,getFaceQuadrature(e,f).numberPoints());
    }
    NumPtr<Point<SPACEDIM> > getQuadraturePoints(const SHAPE_CLASS *e,
    int f) const {
      int s=faceOffset(e,f);
      return NumPtr<Point<SPACEDIM> >(
        getFEValuesData(e)->getQuadraturePoints(),
        s,getFaceQuadrature(e,f).numberPoints());
    }
    NumPtr<REAL> getJxWValues(const SHAPE_CLASS *e,int f) const {
      int s=faceOffset(e,f);
      return NumPtr<REAL>(getFEValuesData(e)->getJxWValues(),s,
        getFaceQuadrature(e,f).numberPoints());
    }
    const NumPtr<Point<SPACEDIM> >& getNormalVectors(const Shape2 *e,
    int f) const {
      return getNormalVectors(e->numberVerticesOnFace(f));
    }

    void mapShapes(const Tessellation&,const SHAPE_CLASS*,
      const LEVEL_CLASS*,int);
//  called from HdivDofHandler::getFunctionValues
    void getFunctionValues(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<Tensor<1,SPACEDIM> >&) const;
    void getFunctionValues(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&) const;
    void getFunctionDerivatives(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<Tensor<2,SPACEDIM> >&) const;
    void getFunctionDerivatives(const Vector&,const SHAPE_CLASS*,int,
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > >&) const;
    void getFunction2ndDerivatives(const Vector&,
      const SHAPE_CLASS*,int,NumPtr<Tensor<3,SPACEDIM> >&) const;
    void getFunction2ndDerivatives(const Vector&,
      const SHAPE_CLASS*,int,NumPtr<NumPtr<Tensor<3,SPACEDIM> > >&) const;
    virtual void printOn(ostream&) const;
};
#endif

#endif
