// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "HdivFiniteElement.H"
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//quad must be destructed after this is destructed
HdivFiniteElementDataBase::HdivFiniteElementDataBase(UPDATE_FLAGS flags,
const Quadrature<SPACEDIM> *quad,TensorShapeFunction<SPACEDIM> *sff,
const ShapeFunction<SPACEDIM> *msff) : 
update_flags(flags),shape_values_ok(false),shape_derivatives_ok(false),
shape_second_derivatives_ok(false),shape_values(0,0),
shape_derivatives(0,0),shape_second_derivatives(0,0),
shape_function_family(sff),
multiplier_shape_values_ok(false),multiplier_shape_derivatives_ok(false),
multiplier_shape_second_derivatives_ok(false),
multiplier_shape_values(0,0),
multiplier_shape_function_family(msff),quadrature(quad) {
  int n_q_points=quadrature->numberPoints();
  int n_shapes=sff->getNumber();
  int n_multiplier_shapes=msff->getNumber();
  if (flags & UPDATE_VALUES) {
    shape_values.reinit(n_shapes,n_q_points);
    multiplier_shape_values.reinit(n_multiplier_shapes,n_q_points);
  }
  if (flags & (UPDATE_GRADIENTS | UPDATE_DIVERGENCE)) {
    shape_derivatives.reinit(n_shapes,n_q_points);
  }
  if (flags & UPDATE_SECOND_DERIVATIVES) {
    shape_second_derivatives.reinit(n_shapes,n_q_points);
  }
  if (!sff->requiresShapeToComputeValues()) {
    if (flags & (UPDATE_VALUES | UPDATE_GRADIENTS | 
    UPDATE_SECOND_DERIVATIVES)) {
      for (int q=0;q<n_q_points;q++) {
        if (update_flags & UPDATE_VALUES) {
          NumPtr<Tensor<1,SPACEDIM> > values;
          sff->computeValues(quadrature->point(q),values);
          for (int s=0;s<n_shapes;s++) {
            shape_values(s,q)=values[s];
          }
          shape_values_ok=true;
        }
        if (update_flags & (UPDATE_GRADIENTS | UPDATE_DIVERGENCE)) {
          NumPtr<Tensor<2,SPACEDIM> > grads;
          sff->computeGrads(quadrature->point(q),grads);
          for (int s=0;s<n_shapes;s++) {
            shape_derivatives(s,q)=grads[s];
          }
          shape_derivatives_ok=true;
        }
        if (update_flags & UPDATE_SECOND_DERIVATIVES) {
          NumPtr<Tensor<3,SPACEDIM> > grad_grads;
          sff->computeGradGrads(quadrature->point(q),grad_grads);
          for (int s=0;s<n_shapes;s++) {
            shape_second_derivatives(s,q)=grad_grads[s];
          }
          shape_second_derivatives_ok=true;
        }
      }
    }
  }
  if (!msff->requiresShapeToComputeValues()) {
    if (flags & (UPDATE_VALUES | UPDATE_GRADIENTS | 
    UPDATE_SECOND_DERIVATIVES)) {
      for (int q=0;q<n_q_points;q++) {
        if (update_flags & UPDATE_VALUES) {
          NumPtr<REAL> values;
          msff->computeValues(quadrature->point(q),values);
          for (int s=0;s<n_multiplier_shapes;s++) {
            multiplier_shape_values(s,q)=values[s];
          }
          multiplier_shape_values_ok=true;
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
HdivFiniteElementDataBase::HdivFiniteElementDataBase(
const HdivFiniteElementDataBase *fedb,const Shape *e) :
update_flags(fedb->update_flags),shape_values_ok(false),
shape_derivatives_ok(false),shape_second_derivatives_ok(false),
shape_values(fedb->shape_values.size(0),fedb->shape_values.size(1)),
shape_derivatives(fedb->shape_derivatives.size(0),
                  fedb->shape_derivatives.size(1)),
shape_second_derivatives(fedb->shape_second_derivatives.size(0),
                         fedb->shape_second_derivatives.size(1)),
multiplier_shape_values_ok(false),multiplier_shape_derivatives_ok(false),
multiplier_shape_second_derivatives_ok(false),
multiplier_shape_values(fedb->multiplier_shape_values.size(0),
fedb->multiplier_shape_values.size(1)),
quadrature(fedb->quadrature),
shape_function_family(fedb->shape_function_family),
multiplier_shape_function_family(fedb->multiplier_shape_function_family)
{
  NumPtr<int> renumber;
  shape_function_family->shapeFunctionsOrder(e,renumber);
  if (fedb->shape_values_ok) {
    const Array<2,Tensor<1,SPACEDIM> > &sv=fedb->shape_values;
    for (int fe_s=0;fe_s<sv.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sv.size(1);q++) {
        const Tensor<1,SPACEDIM> &svsq=sv(fe_s,q);
        shape_values(abs(dof_s),q)=(dof_s>=0 ? svsq : -svsq);
      }
    }
    shape_values_ok=true;
    const Array<2,REAL> &msv=fedb->multiplier_shape_values;
    for (int s=0;s<msv.size(0);s++) {
      for (int q=0;q<msv.size(1);q++) {
        multiplier_shape_values(s,q)=msv(s,q);
      }
    }
    multiplier_shape_values_ok=true;
  }
  if (fedb->shape_derivatives_ok) {
    const Array<2,Tensor<2,SPACEDIM> > &sg=fedb->shape_derivatives;
    for (int fe_s=0;fe_s<sg.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sg.size(1);q++) {
        const Tensor<2,SPACEDIM> &sgsq=sg(fe_s,q);
        shape_derivatives(abs(dof_s),q)=(dof_s>=0 ? sgsq : - sgsq);
      }
    }
    shape_derivatives_ok=true;
  }
  if (fedb->shape_second_derivatives_ok) {
    const Array<2,Tensor<3,SPACEDIM> > &sgg=
      fedb->shape_second_derivatives;
    for (int fe_s=0;fe_s<sgg.size(0);fe_s++) {
      int dof_s=renumber[fe_s];
      for (int q=0;q<sgg.size(1);q++) {
        const Tensor<3,SPACEDIM> &sggsq=sgg(fe_s,q);
        shape_second_derivatives(abs(dof_s),q)=
          (dof_s>=0 ? sggsq : -sggsq);
      }
    }
    shape_second_derivatives_ok=true;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElementDataBase::computeShapesFor(const Shape *e) {
  if (shape_function_family->requiresShapeToComputeValues()) {
    HdivDualTensorShapeFunction *hdtsf=
      dynamic_cast<HdivDualTensorShapeFunction*>(shape_function_family);
    hdtsf->computeCoefficientsFor(e);
    int n_points=quadrature->numberPoints();
    if (update_flags & UPDATE_VALUES) {
      for (int q=0;q<n_points;q++) {
        NumPtr<Tensor<1,SPACEDIM> > values;
        shape_function_family->
          computeValues(quadrature->point(q),values);
        for (int s=0;s<values.getNumber();s++) {
          shape_values(s,q)=values[s];
        }
      }
      shape_values_ok=true;
    }
    if (update_flags & UPDATE_GRADIENTS) {
      for (int q=0;q<n_points;q++) {
        NumPtr<Tensor<2,SPACEDIM> > grads;
        shape_function_family->
          computeGrads(quadrature->point(q),grads);
        for (int s=0;s<grads.getNumber();s++) {
          shape_derivatives(s,q)=grads[s];
        }
      }
      shape_derivatives_ok=true;
    }
    if (update_flags & UPDATE_SECOND_DERIVATIVES) {
      for (int q=0;q<n_points;q++) {
        NumPtr<Tensor<3,SPACEDIM> > grad_grads;
        shape_function_family->
          computeGradGrads(quadrature->point(q),grad_grads);
        for (int s=0;s<grad_grads.getNumber();s++) {
          shape_second_derivatives(s,q)=grad_grads[s];
        }
      }
      shape_second_derivatives_ok=true;
    }
  }
  if (multiplier_shape_function_family->requiresShapeToComputeValues()) {
    OBSOLETE("not programmed");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElementDataBase::printOn(ostream &os) const {
  os << "HdivFiniteElementDataBase: update_flags = " << update_flags 
     << "\n\tshape_values_ok = " << PRINT_BOOLEAN(shape_values_ok)
     << "\n\tshape_derivatives_ok = " 
     << PRINT_BOOLEAN(shape_derivatives_ok)
     << "\n\tshape_second_derivatives_ok = " 
     << PRINT_BOOLEAN(shape_second_derivatives_ok)
     << "\n\tmultiplier_shape_values_ok = "
     << PRINT_BOOLEAN(multiplier_shape_values_ok)
     << "\n\tmultiplier_shape_derivatives_ok = " 
     << PRINT_BOOLEAN(shape_derivatives_ok)
     << "\n\tmultiplier_shape_second_derivatives_ok = " 
     << PRINT_BOOLEAN(multiplier_shape_second_derivatives_ok)
     << endl;
  if (shape_values_ok) {
    os << "\tshape_values:" << endl;
    for (int i=0;i<shape_values.size(0);i++) {
      for (int j=0;j<shape_values.size(1);j++) {
        os << "\t\tshape_values[" << i << "," << j << "] = "
           << shape_values(i,j) << endl;
      }
    }
  }
  if (shape_derivatives_ok) {
    os << "\tshape_derivatives:" << endl;
    for (int i=0;i<shape_derivatives.size(0);i++) {
      for (int j=0;j<shape_derivatives.size(1);j++) {
        os << "\t\tshape_derivatives[" << i << "," << j << "] = "
           << shape_derivatives(i,j) << endl;
      }
    }
  }
//if (shape_second_derivatives_ok) {
//  os << "\tshape_second_derivatives:" << endl;
//  for (int i=0;i<shape_second_derivatives.size(0);i++) {
//    for (int j=0;j<shape_second_derivatives.size(1);j++) {
//      os << "\t\tshape_second_derivatives[" << i << "," << j << "] = "
//         << shape_second_derivatives(i,j) << endl;
//    }
//  }
//}
  if (multiplier_shape_values_ok) {
    os << "\tmultiplier_shape_values:" << endl;
    for (int i=0;i<multiplier_shape_values.size(0);i++) {
      for (int j=0;j<multiplier_shape_values.size(1);j++) {
        os << "\t\tmultiplier_shape_values[" << i << "," << j << "] = "
           << multiplier_shape_values(i,j) << endl;
      }
    }
  }
//if (multiplier_shape_derivatives_ok) {
//  os << "\tmultiplier_shape_derivatives:" << endl;
//  for (int i=0;i<multiplier_shape_derivatives.size(0);i++) {
//    for (int j=0;j<multiplier_shape_derivatives.size(1);j++) {
//      os << "\t\tmultiplier_shape_derivatives[" << i << "," << j
//         << "] = " << multiplier_shape_derivatives(i,j) << endl;
//    }
//  }
//}
//if (multiplier_shape_second_derivatives_ok) {
//  os << "\tmultiplier_shape_second_derivatives:" << endl;
//  for (int i=0;i<multiplier_shape_second_derivatives.size(0);i++) {
//    for (int j=0;j<multiplier_shape_second_derivatives.size(1);j++) {
//      os << "\t\tmultiplier_shape_second_derivatives[" << i << ","
//         << j << "] = " << multiplier_shape_second_derivatives(i,j)
//         << endl;
//    }
//  }
//}
  cout << "\tquadrature = " << quadrature << endl;
  quadrature->printOn(os);
  shape_function_family->printOn(os);
  multiplier_shape_function_family->printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
HdivFiniteElement::~HdivFiniteElement() {
  while (!polynomial_space.empty()) {
    map<int,HdivDualTensorShapeFunction*>::iterator 
      ps_iterator=polynomial_space.begin();
    delete ps_iterator->second;
    polynomial_space.erase(ps_iterator);
  }
  while (!multiplier_polynomial_space.empty()) {
    map<int,ShapeFunction<SPACEDIM>*>::iterator 
      ps_iterator=multiplier_polynomial_space.begin();
    delete ps_iterator->second;
    multiplier_polynomial_space.erase(ps_iterator);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::setDofsPerObject() {
  for (map<int,HdivDualTensorShapeFunction*>::const_iterator
  ps_iterator=polynomial_space.begin();
  ps_iterator!=polynomial_space.end();ps_iterator++) {
    int nv=ps_iterator->first;
    HdivDualTensorShapeFunction *ps=ps_iterator->second;
    if (ps!=0) {
      dofs_per_cell[nv]=ps->getNumber()*components;
      dofs_per_shape[nv]=ps->getNumberInteriorDofs();
#if (SPACEDIM==1)
      dofs_per_codimension1_shape[1]=ps->getNumberFaceDofs()*components;
#elif (SPACEDIM==2)
      dofs_per_codimension1_shape[2]=ps->getNumberFaceDofs()*components;
#else
      if (nv==4) {
        dofs_per_codimension1_shape[3]=ps->getNumberFaceDofs()*components;
      } else if (nv==8) {
        dofs_per_codimension1_shape[4]=ps->getNumberFaceDofs()*components;
      }
#endif
    } else {
      dofs_per_cell[nv]=0;
      dofs_per_shape[nv]=0;
#if (SPACEDIM==1)
      dofs_per_codimension1_shape[1]=0;
#elif (SPACEDIM==2)
      dofs_per_codimension1_shape[2]=0;
#else
      if (nv==4) dofs_per_codimension1_shape[3]=0;
      else if (nv==8) dofs_per_codimension1_shape[4]=0;
#endif
    }
  }
  for (map<int,ShapeFunction<SPACEDIM>*>::const_iterator
  mps_iterator=multiplier_polynomial_space.begin();
  mps_iterator!=multiplier_polynomial_space.end();mps_iterator++) {
    int nv=mps_iterator->first;
    ShapeFunction<SPACEDIM> *mps=mps_iterator->second;
    if (mps!=0) {
      multiplier_dofs_per_shape[nv]=mps->getNumber()*components;
    } else multiplier_dofs_per_shape[nv]=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::setProlongation() {
//We want the fine solution function to equal the coarse solution fcn:
//  b_s(xi_q) c_s = B_S(parent coord of child xi_q) C_S
//(Einstein summation convention: repeated subscript ==> sum)
//Thus if [ p_{d,s} ] is permutation that reorders 
//  from static fe order (index s)
//  to shape-dependent dof order on the child (index d), and
//if [ Q_{D,S} ] is the permutation that reorders 
//  from fe (index S) 
//  to dof  (index D) on parent,
//then
//  c_d = p_{d,s} [ b_s(xi_q) ]^{-1} 
//        [ B_S(parent coord of child xi_q) ] Q_{D,S}^T C_D
//The permutations p_{d,s} and Q_{D,S} are not used in this routine
//fe order is same as Hierarchical*Polynomials
  int cps=childrenPerShape();
  int cpcd1s=childrenPerCodimension1Shape();
  for (map<int,HdivDualTensorShapeFunction*>::const_iterator
  sf_iterator=polynomial_space.begin();
  sf_iterator!=polynomial_space.end();sf_iterator++) {
    int nv=sf_iterator->first;
    const HdivDualTensorShapeFunction *sf=sf_iterator->second;
    if (sf==0) continue;
    if (sf->requiresShapeToComputeValues()) continue;
    int nsf=sf->getNumber();
    NumPtr<Point<SPACEDIM> > csp(nsf);
    for (int q=0;q<nsf;q++) {
      for (int d=0;d<SPACEDIM;d++) csp[q][d]=drand48();
    }
    NumPtr<Matrix> &prolongation_matrices=prolongation[nv];
    prolongation_matrices.allocate(cps);

    LAMatrix mat(nsf*SPACEDIM,nsf);
    LAMatrix rhs(nsf*SPACEDIM,nsf);
    for (int c=0;c<cps;c++) {
      for (int q=0;q<nsf;q++) {
        const Point<SPACEDIM> &child_point=csp[q];
        Point<SPACEDIM> parent_point=
          parentCoordinateOfCanonical(nv,child_point,c);
        NumPtr<Tensor<1,SPACEDIM> > cv;
        sf->computeValues(child_point,cv);
        NumPtr<Tensor<1,SPACEDIM> > pv;
        sf->computeValues(parent_point,pv);
        for (int s=0;s<nsf;s++) {
          for (int d=0;d<SPACEDIM;d++) {
            mat(d+q*SPACEDIM,s)=pv[s][d];
            rhs(d+q*SPACEDIM,s)=cv[s][d];
          }
        }
      }
      LAMatrix *soln=factorNSolve(mat,rhs);
      prolongation_matrices[c].copy(*soln);
      delete soln;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int HdivFiniteElement::maxDofsPerCell() const {
  int max_dofs_per_cell=0;
  for (map<int,int>::const_iterator dpc_iterator=
  dofs_per_cell.begin();dpc_iterator!=dofs_per_cell.end();
  dpc_iterator++) {
    max_dofs_per_cell=max(max_dofs_per_cell,dpc_iterator->second);
  }
  return max_dofs_per_cell;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int HdivFiniteElement::maxMultiplierDofsPerShape() const {
  int max_multiplier_dofs_per_shape=0;
  for (map<int,int>::const_iterator mdps_iterator=
  multiplier_dofs_per_shape.begin();
  mdps_iterator!=multiplier_dofs_per_shape.end();mdps_iterator++) {
    max_multiplier_dofs_per_shape=
      max(max_multiplier_dofs_per_shape,mdps_iterator->second);
  }
  return max_multiplier_dofs_per_shape;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//dof_values come from face elements, in face element's ordering, then
//  from element's interior, in element's ordering
//see HdivDoFHandler::getDofIndices for details
void HdivFiniteElement::sumShapeValues(const SHAPE_CLASS *e,
const Array<2,Tensor<1,SPACEDIM> > &sv,int nqp,int offset,
const Vector &dof_values,NumPtr<Tensor<1,SPACEDIM> > &values) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    Tensor<1,SPACEDIM> val;
    for (int s=0;s<dpc;s++) {
      const Tensor<1,SPACEDIM> &svsq=sv(s,q+offset);
      for (int d=0;d<SPACEDIM;d++) val[d]+=svsq[d]*dof_values[s]; 
    }
    values[q]=val;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumMultiplierShapeValues(const SHAPE_CLASS *e,
const Array<2,REAL> &msv,int nqp,int offset,
const Vector &dof_values,NumPtr<REAL> &values) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  int mdps=multiplierDofsPerShape(e);
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    REAL val=0.;
    for (int s=0;s<mdps;s++) val+=msv(s,q+offset)*dof_values[s+dpc]; 
    values[q]=val;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumShapeValues(const SHAPE_CLASS *e,
const Array<2,Tensor<1,SPACEDIM> > &sv,int nqp,int offset,
const Vector &dof_values,NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &values)
const {
  int shapes_per_cell=dofsPerCell(e)/components;
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<1,SPACEDIM> > &v=values[q];
    v.allocate(components);
    for (int s=0;s<shapes_per_cell;s++) {
      const Tensor<1,SPACEDIM> &svq=sv(s,q+offset);
      int sc=s*components;
      for (int c=0;c<components;c++) {
        for (int d=0;d<SPACEDIM;d++) v[c][d]+=svq[d]*dof_values[c+sc]; 
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumMultiplierShapeValues(const SHAPE_CLASS *e,
const Array<2,REAL> &sv,int nqp,int offset,const Vector &dof_values,
NumPtr<Vector> &values) const{
  int shapes_per_cell=dofsPerCell(e)/components;
  values.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    Vector &v=values[q];
    v.reinit(components);
    for (int s=0;s<shapes_per_cell;s++) {
      const REAL &svq=sv(s,q+offset);
      int sc=s*components;
      for (int c=0;c<components;c++) {
        v[c]+=svq*dof_values[c+sc]; 
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumShapeDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<2,SPACEDIM> > &sg,int nqp,int offset,
const Vector &dof_values,NumPtr<Tensor<2,SPACEDIM> > &gradients) 
const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  gradients.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    for (int s=0;s<dpc;s++) {
      Tensor<2,SPACEDIM> tmp(sg(s,q+offset));
      tmp*=dof_values[s];
      gradients[q]+=tmp;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumShapeDivergence(const SHAPE_CLASS *e,
const Array<2,REAL> &sd,int nqp,int offset,const Vector &dof_values,
NumPtr<REAL> &divergence) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  divergence.allocate(nqp);
  divergence.initialize(0.);
  for (int q=0;q<nqp;q++) {
    for (int s=0;s<dpc;s++) divergence[q]+=sd(s,q+offset)*dof_values[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumMultiplierShapeDerivatives(
const SHAPE_CLASS *e,const Array<2,Tensor<1,SPACEDIM> > &sg,int nqp,
int offset,const Vector &dof_values,
NumPtr<Tensor<1,SPACEDIM> > &gradients) const {
  CHECK_SAME(components,1);
  int dpc=INT_MAX; // multiplierDofsPerCell(e);
  gradients.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    for (int s=0;s<dpc;s++) {
      Tensor<1,SPACEDIM> tmp(sg(s,q+offset));
      tmp*=dof_values[s];
      gradients[q]+=tmp;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumShapeDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<2,SPACEDIM> > &sg,int nqp,int offset,
const Vector &dof_values,
NumPtr<NumPtr<Tensor<2,SPACEDIM> > > &gradients) const {
  int shapes_per_cell=dofsPerCell(e)/components;
  if (gradients.getNumber()==0) gradients.allocate(nqp);
#ifdef DEBUG
  else CHECK_SAME(gradients.getNumber(),nqp);
#endif
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<2,SPACEDIM> > &g=gradients[q];
    if (g.getNumber()==0) g.allocate(components);
#ifdef DEBUG
    else CHECK_SAME(g.getNumber(),components);
#endif
    Tensor<2,SPACEDIM> empty_tensor;
    g.initialize(empty_tensor);
    for (int s=0;s<shapes_per_cell;s++) {
      int sc=s*components;
      const Tensor<2,SPACEDIM> &sgq=sg(s,q+offset);
      for (int c=0;c<components;c++) {
        Tensor<2,SPACEDIM> tmp(sgq);
        tmp*=dof_values[c+sc];
        g[c]+=tmp;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumShapeDivergence(const SHAPE_CLASS *e,
const Array<2,REAL> &sd,int nqp,int offset,const Vector &dof_values,
NumPtr<Vector> &divergence) const {
  int shapes_per_cell=dofsPerCell(e)/components;
  if (divergence.getNumber()==0) divergence.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    Vector &dq=divergence[q];
    if (dq.size()==0) dq.reinit(components);
    dq=0.;
    for (int s=0;s<shapes_per_cell;s++) {
      int sc=s*components;
      REAL sdq=sd(s,q+offset);
      for (int c=0;c<components;c++) dq[c]+=sdq*dof_values[c+sc];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumMultiplierShapeDerivatives(
const SHAPE_CLASS *e,const Array<2,Tensor<1,SPACEDIM> > &sg,int nqp,
int offset,const Vector &dof_values,
NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &gradients) const {
  int multiplier_shapes_per_cell=INT_MAX; // multiplierDofsPerCell(e)/components;
  if (gradients.getNumber()==0) gradients.allocate(nqp);
#ifdef DEBUG
  else CHECK_SAME(gradients.getNumber(),nqp);
#endif
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<1,SPACEDIM> > &g=gradients[q];
    if (g.getNumber()==0) g.allocate(components);
#ifdef DEBUG
    else CHECK_SAME(g.getNumber(),components);
#endif
    Tensor<1,SPACEDIM> empty_tensor;
    g.initialize(empty_tensor);
    for (int s=0;s<multiplier_shapes_per_cell;s++) {
      int sc=s*components;
      const Tensor<1,SPACEDIM> &sgq=sg(s,q+offset);
      for (int c=0;c<components;c++) {
        Tensor<1,SPACEDIM> tmp(sgq);
        tmp*=dof_values[c+sc];
        g[c]+=tmp;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumShape2ndDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<3,SPACEDIM> > &sgg,int nqp,int offset,
const Vector &dof_values,
NumPtr<Tensor<3,SPACEDIM> > &second_derivatives) const {
  CHECK_SAME(components,1);
  int dpc=dofsPerCell(e);
  second_derivatives.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    for (int s=0;s<dpc;s++) {
      Tensor<3,SPACEDIM> tmp(sgg(s,q+offset));
      tmp*=dof_values[s];
      second_derivatives[q]+=tmp;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumMultiplierShape2ndDerivatives(
const SHAPE_CLASS *e,const Array<2,Tensor<2,SPACEDIM> > &sgg,int nqp,
int offset,const Vector &dof_values,
NumPtr<Tensor<2,SPACEDIM> > &second_derivatives) const {
  CHECK_SAME(components,1);
  int dpc=INT_MAX; // multiplierDofsPerCell(e);
  second_derivatives.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    for (int s=0;s<dpc;s++) {
      Tensor<2,SPACEDIM> tmp(sgg(s,q+offset));
      tmp*=dof_values[s];
      second_derivatives[q]+=tmp;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumShape2ndDerivatives(const SHAPE_CLASS *e,
const Array<2,Tensor<3,SPACEDIM> > &sgg,int nqp,int offset,
const Vector &dof_values,
NumPtr<NumPtr<Tensor<3,SPACEDIM> > > &second_derivatives) const {
  int shapes_per_cell=dofsPerCell(e)/components;
  second_derivatives.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<3,SPACEDIM> > &gg=second_derivatives[q];
    gg.allocate(components);
    Tensor<3,SPACEDIM> empty_tensor;
    gg.initialize(empty_tensor);
    for (int s=0;s<shapes_per_cell;s++) {
      int sc=s*components;
      const Tensor<3,SPACEDIM> &sggq=sgg(s,q+offset);
      for (int c=0;c<components;c++) {
        Tensor<3,SPACEDIM> tmp(sggq);
        tmp*=dof_values[c+sc];
        gg[c]+=tmp;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::sumMultiplierShape2ndDerivatives(
const SHAPE_CLASS *e,const Array<2,Tensor<2,SPACEDIM> > &sgg,int nqp,
int offset,const Vector &dof_values,
NumPtr<NumPtr<Tensor<2,SPACEDIM> > > &second_derivatives) const {
  int shapes_per_cell=INT_MAX; // multiplierDofsPerCell(e)/components;
  second_derivatives.allocate(nqp);
  for (int q=0;q<nqp;q++) {
    NumPtr<Tensor<2,SPACEDIM> > &gg=second_derivatives[q];
    gg.allocate(components);
    Tensor<2,SPACEDIM> empty_tensor;
    gg.initialize(empty_tensor);
    for (int s=0;s<shapes_per_cell;s++) {
      int sc=s*components;
      const Tensor<2,SPACEDIM> &sggq=sgg(s,q+offset);
      for (int c=0;c<components;c++) {
        Tensor<2,SPACEDIM> tmp(sggq);
        tmp*=dof_values[c+sc];
        gg[c]+=tmp;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void HdivFiniteElement::setBoundaryValues(const SHAPE_CLASS *element,
const CODIMENSION1_SHAPE_CLASS *face_element,
const NumPtr<int> &element_dofs,const NumPtr<int> &face_dofs,
const NumPtr<Point<SPACEDIM-1> > &canonical_locations,
const NumPtr<Point<SPACEDIM> > &mapped_locations,
const NumPtr<REAL> &face_jxw_values,const BoundaryCondition *bc,
map<int,REAL> &boundary_values) const {
  NumPtr<Vector> bc_values(mapped_locations.getNumber());
  if (components==1) {
    NumPtr<REAL> bc_values_scalar(bc_values.getNumber());
    bc->getFunctionSystem()->value(bc_values_scalar,mapped_locations,0);
    for (int s=0;s<bc_values.getNumber();s++) {
      bc_values[s].reinit(1);
      bc_values[s][0]=bc_values_scalar[s];
    }
  } else {
    bc->getFunctionSystem()->value(bc_values,mapped_locations);
  }

  HdivDualTensorShapeFunction *sff=shapeFunctionFamily(element);
  NumPtr<NumPtr<REAL> > 
    dof_poly_values(canonical_locations.getNumber());
  int nv=face_element->numberVertices();
  for (int q=0;q<canonical_locations.getNumber();q++) {
    sff->computeFaceValues(nv,canonical_locations[q],dof_poly_values[q]);
  }
  int shape_fcns_per_face=face_dofs.getNumber()/components;
  for (int s=0;s<shape_fcns_per_face;s++) {
    int sc=s*components;
    for (int c=0;c<components;c++) {
      if (bc->isEssential(c)) {
        REAL face_dof_value=0.;
        for (int q=0;q<canonical_locations.getNumber();q++) {
          face_dof_value+=dof_poly_values[q][s]*bc_values[q][c]
                         *face_jxw_values[q];
        }
        int global_dof=face_dofs[sc+c];
        boundary_values[global_dof]=face_dof_value;
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::equallySpacedLatticePointDataBase(
map<int,HdivFiniteElementDataBase*> &fedb_map,
map<int,Quadrature<SPACEDIM>*> &quad_map) const {
  for (map<int,HdivDualTensorShapeFunction*>::const_iterator
  sf_iterator=polynomial_space.begin();
  sf_iterator!=polynomial_space.end();sf_iterator++) {
    int nv=sf_iterator->first;
    NumPtr<Point<SPACEDIM> > element_lattice_points;
    equallySpacedLatticePoints(nv,degree,element_lattice_points);
    Quadrature<SPACEDIM> *quad=quad_map[nv]=
      OPERATOR_NEW Quadrature<SPACEDIM>(element_lattice_points);
    HdivFiniteElementDataBase *fedb=getFEData(nv,UPDATE_VALUES,quad);
    fedb_map[nv]=fedb;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//if SPACEDIM>1 and given_function is non-zero,
//  the linear system is not square and is solved by least squares
//  so, the boundary conditions may not be satisfied exactly
void HdivFiniteElement::setValues(const SHAPE_CLASS *element,
const NumPtr<int> &element_dofs,HdivFiniteElementDataBase *fedb,
const NumPtr<Point<SPACEDIM> > &mapped_locations,
const TensorFunctionSystem *given_function,Vector &values) const {
  int shapes_per_element=mapped_locations.getNumber();
//given_function(mapped_location[q]); order of q irrelevant
  LAMatrix gf(shapes_per_element*SPACEDIM,components);
  if (components==1) {
    NumPtr<Tensor<1,SPACEDIM> > dof_values(shapes_per_element);
    given_function->value(dof_values,mapped_locations,0);
    for (int s=0;s<shapes_per_element;s++) {
      for (int d=0;d<SPACEDIM;d++) {
        gf(d+s*SPACEDIM,0)=dof_values[s][d];
      }
    }
  } else {
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > dof_values;
    given_function->value(dof_values,mapped_locations);
    for (int c=0;c<components;c++) {
      for (int s=0;s<shapes_per_element;s++) {
        for (int d=0;d<SPACEDIM;d++) {
          gf(d+s*SPACEDIM,c)=dof_values[s][c][d];
        }
      }
    }
  }
//shapeValue(d+q*SPACEDIM,element_fe_index)
  LAMatrix shape_values(shapes_per_element*SPACEDIM,shapes_per_element);
  for (int s=0;s<shapes_per_element;s++) {
    for (int q=0;q<shapes_per_element;q++) {
      const Tensor<1,SPACEDIM> &sv=fedb->shapeValue(s,q);
      for (int d=0;d<SPACEDIM;d++) {
        shape_values(d+q*SPACEDIM,s)=sv[d];
      }
    }
  }
//soln(d+element_fe_index*SPACEDIM,c)
  LAMatrix *soln=factorNSolve(shape_values,gf);
//element_renumber[element_fe_index]=element_dof_index
  TensorShapeFunction<SPACEDIM> *sf=shapeFunctionFamily(element);
  NumPtr<int> element_renumber;
  sf->shapeFunctionsOrder(element,element_renumber);
  NumPtr<int> element_renumber_inverse(element_renumber.getNumber());
  for (int i=0;i<element_renumber.getNumber();i++) {
    element_renumber_inverse[abs(element_renumber[i])]=i;
  }
//element_dofs[element_dof_index]=global_dof_index
  for (int s=0;s<shapes_per_element;s++) {
    int sc=s*components;
    for (int d=0;d<SPACEDIM;d++) {
      int fe_index=element_renumber_inverse[d+sc];
      for (int c=0;c<components;c++) {
        values[element_dofs[d+SPACEDIM*(c+sc)]]=(*soln)(fe_index,c);
        if (element_renumber[fe_index]<0) {
          values[element_dofs[sc+c]]=-values[element_dofs[sc+c]];
        }
      }
    }
  }
  delete soln;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::setMultiplierValues(const SHAPE_CLASS *element,
const NumPtr<int> &element_dofs,HdivFiniteElementDataBase *fedb,
const NumPtr<Point<SPACEDIM> > &mapped_locations,
const FunctionSystem *given_function,Vector &values) const {
  int shapes_per_element=mapped_locations.getNumber();
//given_function(mapped_location[q]); order of q irrelevant
  LAMatrix gf(shapes_per_element,components);
  if (components==1) {
    NumPtr<REAL> dof_values(shapes_per_element);
    given_function->value(dof_values,mapped_locations,0);
    for (int s=0;s<shapes_per_element;s++) {
      gf(s,0)=dof_values[s];
    }
  } else {
    NumPtr<Vector> dof_values;
    given_function->value(dof_values,mapped_locations);
    for (int c=0;c<components;c++) {
      for (int s=0;s<shapes_per_element;s++) {
        gf(s,c)=dof_values[s][c];
      }
    }
  }
//shapeValue(q,element_fe_index)
  LAMatrix shape_values(shapes_per_element,shapes_per_element);
  for (int s=0;s<shapes_per_element;s++) {
    for (int q=0;q<shapes_per_element;q++) {
      shape_values(q,s)=fedb->multiplierShapeValue(s,q);
    }
  }
//soln(element_fe_index,c)
  LAMatrix *soln=factorNSolve(shape_values,gf);
//element_renumber[element_fe_index]=element_dof_index
  ShapeFunction<SPACEDIM> *sf=multiplierShapeFunctionFamily(element);
  NumPtr<int> element_renumber;
  sf->shapeFunctionsOrder(element,element_renumber);
  NumPtr<int> element_renumber_inverse(element_renumber.getNumber());
  for (int i=0;i<element_renumber.getNumber();i++) {
    element_renumber_inverse[abs(element_renumber[i])]=i;
  }
  for (int element_dof_index=0;element_dof_index<shapes_per_element;
  element_dof_index++) {
    int sc=element_dof_index*components;
    int fe_index=element_renumber_inverse[element_dof_index];
    for (int c=0;c<components;c++) {
      values[element_dofs[sc+c]]=(*soln)(fe_index,c);
      if (element_renumber[fe_index]<0) {
        values[element_dofs[sc+c]]=-values[element_dofs[sc+c]];
      }
    }
  }
  delete soln;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
HdivFiniteElementDataBase* HdivFiniteElement::fixShapeFunctionsFor(
const Shape *e,HdivFiniteElementDataBase *fedb) const {
  TensorShapeFunction<SPACEDIM> *sf=shapeFunctionFamily(e);
  if (sf->requiresShapeToComputeValues()) {
    fedb->computeShapesFor(e);
    return fedb;
  } else {
    return OPERATOR_NEW HdivFiniteElementDataBase(fedb,e);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivFiniteElement::printOn(ostream &os) const {
  os << "\tHdivFiniteElement: degree = " << degree << endl;
  for (map<int,HdivDualTensorShapeFunction*>::const_iterator
  ps_iterator=polynomial_space.begin();
  ps_iterator!=polynomial_space.end();ps_iterator++) {
    os << "\tpolynomial_space[" << ps_iterator->first << "]: " << endl;
    ps_iterator->second->printOn(os);
  }
  for (map<int,ShapeFunction<SPACEDIM>*>::const_iterator
  ps_iterator=multiplier_polynomial_space.begin();
  ps_iterator!=multiplier_polynomial_space.end();ps_iterator++) {
    os << "\tmultiplier_polynomial_space[" << ps_iterator->first
       << "]: " << endl;
    ps_iterator->second->printOn(os);
  }
  for (map<int,int>::const_iterator dps_iterator=
  dofs_per_shape.begin();dps_iterator!=dofs_per_shape.end();
  dps_iterator++) {
    os << "\tdofs_per_shape[" << dps_iterator->first << "] = "
       << dps_iterator->second << endl;
  }
  for (map<int,int>::const_iterator
  dpcs_iterator=dofs_per_codimension1_shape.begin();dpcs_iterator!=
  dofs_per_codimension1_shape.end();dpcs_iterator++) {
    os << "\tdofs_per_codimension1_shape[" << dpcs_iterator->first
       << "] = " << dpcs_iterator->second << endl;
  }
  for (map<int,int>::const_iterator mdps_iterator=
  multiplier_dofs_per_shape.begin();
  mdps_iterator!=multiplier_dofs_per_shape.end();mdps_iterator++) {
    os << "\tmultiplier_dofs_per_shape[" << mdps_iterator->first
       << "] = " << mdps_iterator->second << endl;
  }
  for (map<int,int>::const_iterator dpc_iterator=
  dofs_per_cell.begin();dpc_iterator!=
  dofs_per_cell.end();dpc_iterator++) {
    os << "\tdofs_per_cell[" << dpc_iterator->first << "] = "
       << dpc_iterator->second << endl;
  }
  os << "\tcomponents = " << components << endl;
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTFiniteElement::RTFiniteElement(int d,int nc) : HdivFiniteElement(d,nc)
{
#if (SPACEDIM==1)
  int nv=Interval::verticesPerShape();
  polynomial_space[nv]=OPERATOR_NEW RTSimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
#endif
#if (SPACEDIM==2)
  int nv=Quadrilateral::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW RTParallelpipedDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree);

  nv=Triangle::verticesPerShape();
  polynomial_space[nv]=OPERATOR_NEW RTSimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);

  fqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree+2));
#endif
#if (SPACEDIM==3)
  int nv=Hexahedron::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW RTParallelpipedDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree,degree,degree);

  nv=Prism::verticesPerShape();
  polynomial_space[nv]=OPERATOR_NEW RTPrismDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree);

  nv=Tetrahedron::verticesPerShape();
  polynomial_space[nv]=OPERATOR_NEW RTSimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);

  fqmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
    TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  fqmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
    GaussianQuadrature<2>::argGivenOrder(2*degree+2));
#endif
  setDofsPerObject();
  setProlongation();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTFiniteElement::printOn(ostream &os) const {
  os << "\tRTFiniteElement:" << endl;
  HdivFiniteElement::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BDFMFiniteElement::BDFMFiniteElement(int d,int nc) :
HdivFiniteElement(d,nc) {
#if (SPACEDIM==1)
  int nv=Interval::verticesPerShape();
  TensorShapeFunction<SPACEDIM> *ps=polynomial_space[nv]=
    OPERATOR_NEW BDFMSimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
#endif
#if (SPACEDIM==2)
  int nv=Quadrilateral::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW BDFMParallelpipedDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);

  nv=Triangle::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW BDFMSimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);

  fqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
    GaussianQuadrature<1>::argGivenOrder(2*degree+2));
#endif
#if (SPACEDIM==3)
  int nv=Hexahedron::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW BDFMParallelpipedDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);

  nv=Prism::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW BDFMPrismDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);

  nv=Tetrahedron::verticesPerShape();
  polynomial_space[nv]=
    OPERATOR_NEW BDFMSimplexDualShapeFunction(degree);
  multiplier_polynomial_space[nv]=
    OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);

  fqmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
    TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  fqmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
    GaussianQuadrature<2>::argGivenOrder(2*degree+2));
#endif
  setDofsPerObject();
  setProlongation();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
const Matrix& BDFMFiniteElement::prolongationMatrix(
const SHAPE_CLASS *e,int c,const LEVEL_CLASS *level) {
#if (SPACEDIM>1)
  const HdivDualTensorShapeFunction *sf=shapeFunctionFamily(e);
  if (sf->requiresShapeToComputeValues()) {
//  We want the fine solution function to equal the coarse solution fcn:
//    b_s(xi_q) c_s = B_S(parent coord of child xi_q) C_S
//  (Einstein summation convention: repeated subscript ==> sum)
//  Thus if [ p_{d,s} ] is permutation that reorders 
//    from static fe order (index s)
//    to shape-dependent dof order on the child (index d), and
//  if [ Q_{D,S} ] is the permutation that reorders 
//    from fe (index S) 
//    to dof  (index D) on parent,
//  then
//    c_d = p_{d,s} [ b_s(xi_q) ]^{-1} 
//          [ B_S(parent coord of child xi_q) ] Q_{D,S}^T C_D
//  The permutations p_{d,s} and Q_{D,S} are not used in this routine
//  fe order is same as Hierarchical*Polynomials
    int cps=childrenPerShape();
    int cpcd1s=childrenPerCodimension1Shape();
    int nv=e->numberVertices();
    int nsf=sf->getNumber();
    NumPtr<Point<SPACEDIM> > csp(nsf);
    for (int q=0;q<nsf;q++) {
      for (int d=0;d<SPACEDIM;d++) csp[q][d]=drand48();
    }
    NumPtr<Matrix> &prolongation_matrices=prolongation[nv];
    prolongation_matrices.allocate(cps);

    LAMatrix mat(nsf*SPACEDIM,nsf);
    LAMatrix rhs(nsf*SPACEDIM,nsf);
    for (int c=0;c<cps;c++) {
      for (int q=0;q<nsf;q++) {
        const Point<SPACEDIM> &child_point=csp[q];
        Point<SPACEDIM> parent_point=
          parentCoordinateOfCanonical(nv,child_point,c);
        NumPtr<Tensor<1,SPACEDIM> > cv;
        sf->computeValues(child_point,cv);
        NumPtr<Tensor<1,SPACEDIM> > pv;
        sf->computeValues(parent_point,pv);
        for (int s=0;s<nsf;s++) {
          for (int d=0;d<SPACEDIM;d++) {
            mat(d+q*SPACEDIM,s)=pv[s][d];
            rhs(d+q*SPACEDIM,s)=cv[s][d];
          }
        }
      }
      LAMatrix *soln=factorNSolve(mat,rhs);
      prolongation_matrices[c].copy(*soln);
      delete soln;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMFiniteElement::printOn(ostream &os) const {
  os << "\tBDFMFiniteElement:" << endl;
  HdivFiniteElement::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

#include "NumPtr.C"
template class ArrayBase<2,Tensor<3,SPACEDIM> >;
template class Array<2,Tensor<3,SPACEDIM> >;

template class _Rb_tree<
  int,
  pair<int const, Tensor<1,SPACEDIM> >,
  _Select1st<pair<int const, Tensor<1,SPACEDIM> > >,
  less<int>,
  allocator<pair<int const, Tensor<1,SPACEDIM> > > 
>;
template class _Rb_tree<
  int,
  pair<int const, HdivDualTensorShapeFunction*>,
  _Select1st<pair<int const, HdivDualTensorShapeFunction*> >,
  less<int>,
  allocator<pair<int const, HdivDualTensorShapeFunction*> >
>;
