#ifndef HDIV_PATCH_H
#define HDIV_PATCH_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Array.H"
#include "HdivDofHandler.H"
#include "Mapping.H"
#include "NumPtr.H"
#include "Point.H"
#include "Quadrature.H"
#include "Shape.H"
#include "XYGraphTool.H"

class HdivPatch {
  protected:
    const SHAPE_CLASS *shape;
    const LEVEL_CLASS *level;
    NumPtr<Point<SPACEDIM> > ep;
    NumPtr<int> neighbors;
    int patch_index;
    Array<2,Tensor<1,SPACEDIM> > data;
    Array<2,REAL> multiplier_data;

    static int n_subdivisions;
    static map<int,Quadrature<SPACEDIM> > q_map;
    const Quadrature<SPACEDIM>& getQuadrature() const;
#if (SPACEDIM==3)
    void plotClipPlane(VolGraphTool&,REAL,REAL,const REAL value[2][2][2],
      const Vector3 vertex[2][2][2],const HdivDofHandler*,const Mapping*,
      const Vector*) const;
    void plotTetrahedralClipPlane(VolGraphTool&,REAL,REAL,REAL,int,
      HAND,const HdivDofHandler*,const Mapping*,const Vector*,
      const Point<SPACEDIM> x[4]) const;
    void plotPrismClipPlane(VolGraphTool&,REAL,REAL,REAL,int,HAND,
      const HdivDofHandler*,const Mapping*,const Vector*,
      const Point<SPACEDIM> x[6]) const;
#endif
  public:
    static const int no_neighbor;
    static void setQuadratures(int);
    static const map<int,Quadrature<SPACEDIM> >& 
      getQuadratureMap() { return q_map; }

    HdivPatch(const SHAPE_CLASS*,const LEVEL_CLASS*,const Tessellation&,
      const Mapping&,int,int);
    virtual ~HdivPatch();
    int numberQuadraturePoints() const {
      return getQuadrature().numberPoints();
    }
    Tensor<1,SPACEDIM> getData(int dataset,int q_point) const {
      return data(dataset,q_point);
    }
    void setData(int dataset,const Tensor<1,SPACEDIM> &val) {
      for (int q=0;q<data.size(1);q++) {
        data(dataset,q)=val;
      }
    }
    void setData(int dataset,const NumPtr<Tensor<1,SPACEDIM> > &values) {
      for (int q=0;q<values.getNumber();q++) {
        data(dataset,q)=values[q];
      }
    }
    void setData(int dataset,int n_components,
    const NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &values) {
      int dc=dataset*n_components;
      for (int c=0;c<n_components;c++) {
        for (int q=0;q<values.getNumber();q++) {
          data(dc+c,q)=values[q][c];
        }
      }
    }
    void setMultiplierData(int dataset,REAL val) {
      for (int q=0;q<data.size(1);q++) {
        multiplier_data(dataset,q)=val;
      }
    }
    void setMultiplierData(int dataset,const NumPtr<REAL> &values) {
      for (int q=0;q<values.getNumber();q++) {
        multiplier_data(dataset,q)=values[q];
      }
    }
    void setMultiplierData(int dataset,int n_components,
    const NumPtr<Vector> &values) {
      int dc=dataset*n_components;
      for (int c=0;c<n_components;c++) {
        for (int q=0;q<values.getNumber();q++) {
          multiplier_data(dc+c,q)=values[q][c];
        }
      }
    }
    int dataSize(int i) const { return data.size(i); }
    int multiplierDataSize(int i) const { 
      return multiplier_data.size(i);
    }
    void setNeighbor(int f,int n) {
      neighbors[f]=n;
    }
    int nSubdivisions() const { return n_subdivisions; }
    void getMeshBounds(Point<SPACEDIM>&,Point<SPACEDIM>&) const;
    void getDataBounds(int,Tensor<1,SPACEDIM>&,Tensor<1,SPACEDIM>&) const;
    void getMultiplierDataBounds(int,REAL&,REAL&) const;
#if (SPACEDIM==1)
    void plot(NumPtr<XYGraphTool*>&,int) const;
#endif
#if (SPACEDIM==2)
    void plot(NumPtr<AreaGraphTool*>&,int,const Tensor<1,SPACEDIM+1>&,
      const Tensor<1,SPACEDIM+1>&) const;
    void plotSurface(int,VolGraphTool*,int,REAL,REAL,BOOLEAN) const;
    void draw(AreaGraphTool&) const;
    void contour(NumPtr<AreaGraphTool*>&,const Tensor<1,SPACEDIM+1>&,int)
      const;
#endif
#if (SPACEDIM==3)
    void plotClipPlane(VolGraphTool&,REAL,REAL,REAL,int,HAND,
      const HdivDofHandler*,const Mapping*,const Vector*) const;
    void plotSurface(int,VolGraphTool&,int,REAL,REAL,REAL,BOOLEAN) const;
    void draw(VolGraphTool&) const;
    void drawVectors(VolGraphTool*,int,const Vector3&,const Vector3&,
      const Tensor<1,3>&,const Tensor<1,3>&) const;
#endif
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM==3)
void drawSurfaceWithinCube(VolGraphTool&,REAL,REAL value[2][2][2],
  Vector3 vertex[2][2][2]);
#endif

#endif
