// "$Header:$"
#include <KellyErrorEstimator.H>
//----------------------------  error_estimator.cc  -----------------------
//  $Id: error_estimator.cc,v 1.85 2003/03/10 15:20:46 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  error_estimator.cc  -----------------------
//
//modified from deal.II/deal.II/source/numerics/error_estimator.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Mapping.H>

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM!=1)
void KellyErrorEstimator::integrateOverRegularFace(
const DofHandler &dof_handler,
const NumPtr<const Vector*> &solutions,
const NumPtr<bool> &component_mask,
const FunctionSystem *coefficients,
map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> > 
&face_integrals,const SHAPE_CLASS *element,const LEVEL_CLASS *level,
int face_no,FEFaceValues &fe_face_values,
FEFaceValues &fe_face_values_lattice) {
  int nv=element->numberVerticesOnFace(face_no);
  int n_q_points=
    fe_face_values.getFaceQuadrature(element,face_no).numberPoints();
  int n_components=dof_handler.getFe().nComponents();
  int n_solution_vectors=solutions.getNumber();
  Tessellation &tess=dof_handler.getTessellation();
  fe_face_values.mapShapes(tess,element,level,face_no);
  const NumPtr<Point<SPACEDIM> > &face_mapped_points=
    fe_face_values.getQuadraturePoints(element,face_no);
  NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > &gradients_here=
    gradients_here_map[nv];
  for (int n=0;n<n_solution_vectors;n++) {
    dof_handler.getFunctionDerivatives(fe_face_values,element,
      face_no,level,*solutions[n],gradients_here[n]);
  }
  int face_index=element->codimension1ShapeIndex(face_no);
  const BoundaryCondition *bc=
    level->CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
  if (bc==0) {
    const LEVEL_CLASS *neighbor_level=0;
    const SHAPE_CLASS *neighbor=0;
    int neighbor_face=INT_MAX;
    level->neighborOfNeighbor(element,face_no,neighbor_level,neighbor,
      neighbor_face); 
    CHECK_TEST(neighbor_face<neighbor->numberCodimension1Shapes());

//  face lattice points ==> mapped lattice points on neighbor face
    fe_face_values_lattice.mapShapes(tess,neighbor,neighbor_level,
      neighbor_face);
    const NumPtr<Point<SPACEDIM> > &neighbor_mapped_lattice_points=
      fe_face_values_lattice.getQuadraturePoints(neighbor,neighbor_face);
    NumPtr<Point<SPACEDIM> > face_lattice_points;
    fe_face_values_lattice.getCanonicalQuadraturePoints(neighbor,
        neighbor_face,face_lattice_points);
    NumPtr<Point<SPACEDIM> > neighbor_canonical_points(n_q_points);
    const MappingDataBase *fmd=
      fe_face_values_lattice.mappingData(neighbor->numberVertices());
    const Mapping &mapping=fe_face_values_lattice.getMapping();
    for (int q=0;q<n_q_points;q++) {
      const Point<SPACEDIM> &fmp=face_mapped_points[q];
      int minj=-1;
      REAL min_dist=HUGE_VAL;
      for (int j=0;j<face_mapped_points.getNumber();j++) {
        REAL dist=fmp.distance(neighbor_mapped_lattice_points[j]);
        if (dist<min_dist) {
          minj=j;
          min_dist=dist;
        }
      }
      neighbor_canonical_points[q]=face_lattice_points[minj];
      mapping.canonicalPoint(neighbor,tess,fmd,fmp,
        neighbor_canonical_points[q]);
    }
//  mapped lattice points on neighbor face ==> shape derivatives
    Quadrature<SPACEDIM> neighbor_quadrature(neighbor_canonical_points);
    map<int,Quadrature<SPACEDIM> > neighbor_quadrature_map;
    neighbor_quadrature_map[neighbor->numberVertices()]=
      neighbor_quadrature;
    FEValues neighbor_fe_values(mapping,dof_handler.getFe(),
      neighbor_quadrature_map,
      UPDATE_GRADIENTS | UPDATE_MAPPING_JACOBIAN_INVERSE);
    neighbor_fe_values.mapShapes(tess,neighbor,neighbor_level);
    NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > &neighbor_gradients=
      neighbor_gradients_map[nv];
    for (int n=0;n<n_solution_vectors;n++) {
      dof_handler.getFunctionDerivatives(neighbor_fe_values,
        neighbor,neighbor_level,*solutions[n],neighbor_gradients[n]);
      for (int component=0;component<n_components;component++){
        for (int p=0;p<n_q_points;p++) {
          gradients_here[n][p][component]-=
            neighbor_gradients[n][p][component];
        }
      }
    }
  }
  const NumPtr<Point<SPACEDIM> > &normal_vectors=
      fe_face_values.getNormalVectors(element,face_no);
  NumPtr<NumPtr<NumPtr<REAL> > > &normal_derivative_here=
    normal_derivative_here_map[nv];
  for (int n=0;n<n_solution_vectors;n++) {
    for (int component=0;component<n_components;component++) {
      for (int point=0;point<n_q_points;point++) {
        normal_derivative_here[n][point][component]=
          gradients_here[n][point][component].getValues()*
          normal_vectors[point];
      }
    }
  }
  if (coefficients) {
    if (coefficients->nComponents()==1) {
      NumPtr<REAL> &coefficient_values1=coefficient_values1_map[nv];
      coefficients->value(coefficient_values1,face_mapped_points,0);
      for (int n=0;n<n_solution_vectors;n++) {
        for (int component=0;component<n_components;
        component++) {
          for (int point=0;point<n_q_points;point++) {
            normal_derivative_here[n][point][component]*=
              coefficient_values1[point];
          }
        }
      }
    } else {
      NumPtr<Vector> &coefficient_values=coefficient_values_map[nv];
      coefficients->value(coefficient_values,face_mapped_points);
      for (int n=0;n<n_solution_vectors;n++) {
        for (int component=0; component<n_components; 
        component++) {
          for (int point=0;point<n_q_points;point++) {
            normal_derivative_here[n][point][component]*=
              coefficient_values[point][component];
          }
        }
      }
    }
  }
  if (bc!=0) {
    if (bc->isNatural()) {
      FunctionSystem *fs=bc->getFunctionSystem();
      CHECK_POINTER(fs);
      CHECK_SAME(fs->nComponents(),n_components);
      if (n_components==1) {
        NumPtr<REAL> g(n_q_points);
        fs->value(g,face_mapped_points,0);
        for (int n=0;n<n_solution_vectors;n++) {
          for (int point=0;point<n_q_points;point++) {
            normal_derivative_here[n][point][0]-=g[point];
          }
        }
      } else {
        NumPtr<Vector> g(n_q_points);
        for (int i=0;i<g.getNumber();i++) g[i].reinit(n_components);
        fs->value(g,face_mapped_points);
        for (int component=0; component<n_components;++component) {
          if (bc->isNatural(component)) {
            for (int n=0;n<n_solution_vectors;n++) {
              for (int point=0;point<n_q_points;point++) {
                normal_derivative_here[n][point][component]-=
                  g[point][component];
              }
            }
          }
        }
      }
    }
  }
  NumPtr<REAL> JxW_values=
    fe_face_values.getJxWValues(element,face_no);
  CHECK_TEST(face_index>=0);
  const CODIMENSION1_LEVEL_CLASS *face_level=
    static_cast<const CODIMENSION1_LEVEL_CLASS*>(level);
  map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> >::iterator
    fi_iterator=face_integrals.find(make_pair(face_index,face_level));
  CHECK_TEST(fi_iterator!=face_integrals.end());
  NumPtr<REAL> &face_integral=fi_iterator->second;
  face_integral.initialize(0.);
  for (int n=0;n<n_solution_vectors;n++) {
    for (int component=0;component<n_components;component++){
      if (component_mask[component]) {
        for (int p=0;p<n_q_points;p++) {
          face_integral[n]+=
            pow(normal_derivative_here[n][p][component],2)*JxW_values[p];
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM!=1)
void KellyErrorEstimator::integrateOverIrregularFace(
const DofHandler &dof_handler,
const NumPtr<const Vector*> &solutions,
const NumPtr<bool> &component_mask,
const FunctionSystem *coefficients,
map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> > 
&face_integrals,const SHAPE_CLASS *element,const LEVEL_CLASS *level,
int face_no,FEFaceValues &fe_face_values,
FEFaceValues &fe_face_values_lattice) {
  int nv=element->numberVerticesOnFace(face_no);
  int n_q_points=
    fe_face_values.getFaceQuadrature(element,face_no).numberPoints();
  int n_components=dof_handler.getFe().nComponents();
  int n_solution_vectors=solutions.getNumber();
  Tessellation &tess=dof_handler.getTessellation();
  const LEVEL_CLASS *neighbor_level=0;
  const SHAPE_CLASS *neighbor=0;
  int neighbor_face=INT_MAX;
  level->neighborOfNeighbor(element,face_no,neighbor_level,neighbor,
    neighbor_face); 
  CHECK_TEST(neighbor_face<neighbor->numberCodimension1Shapes());
  const LEVEL_CLASS *neighbor_child_level=neighbor_level->finerLevel();
  CHECK_POINTER(neighbor_child_level);
  fe_face_values_lattice.mapShapes(tess,element,level,face_no);
  const NumPtr<Point<SPACEDIM> > &mapped_lattice_points=
    fe_face_values_lattice.getQuadraturePoints(element,face_no);
  NumPtr<Point<SPACEDIM> > canonical_points(n_q_points);
  const MappingDataBase *fmd=fe_face_values_lattice.mappingData(element);
  const Mapping &mapping=fe_face_values_lattice.getMapping();
  NumPtr<Point<SPACEDIM> > face_lattice_points;
  fe_face_values_lattice.getCanonicalQuadraturePoints(element,face_no,
    face_lattice_points);

  for (int subface_no=0; 
  subface_no<childrenPerCodimension1Shape();subface_no++) {
    int nc=0;
    int ncf=0;
    level->childFaceOnFaceOf(neighbor,neighbor_face,subface_no,nc,ncf);
    int neighbor_child_index=
      neighbor_level->childIndex(neighbor->elementDataIndex(),nc);
    CHECK_TEST(neighbor_child_index>=0);
    const SHAPE_CLASS *neighbor_child=
      neighbor_child_level->getShape(neighbor_child_index);
    CHECK_TEST(!neighbor_child_level->hasChildren(neighbor_child_index));

    fe_face_values.
      mapShapes(tess,neighbor_child,neighbor_child_level,ncf);
    const NumPtr<Point<SPACEDIM> > &face_mapped_points=
      fe_face_values.getQuadraturePoints(neighbor_child,ncf);
    NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > &neighbor_gradients=
      neighbor_gradients_map[nv];
    for (int n=0;n<n_solution_vectors;n++) {
      dof_handler.getFunctionDerivatives(fe_face_values,neighbor_child,
        ncf,neighbor_child_level,*solutions[n],neighbor_gradients[n]);
    }
    for (int q=0;q<face_mapped_points.getNumber();q++) {
      const Point<SPACEDIM> &fmp=face_mapped_points[q];
      int minj=-1;
      REAL min_dist=HUGE_VAL;
      for (int j=0;j<face_mapped_points.getNumber();j++) {
        REAL dist=fmp.distance(mapped_lattice_points[j]);
        if (dist<min_dist) {
          minj=j;
          min_dist=dist;
        }
      }
      canonical_points[q]=face_lattice_points[minj];
      mapping.canonicalPoint(element,tess,fmd,fmp,canonical_points[q]);
    }
    Quadrature<SPACEDIM> element_quadrature(canonical_points);
    map<int,Quadrature<SPACEDIM> > element_quadrature_map;
    element_quadrature_map[element->numberVertices()]=element_quadrature;
    FEValues element_fe_values(mapping,dof_handler.getFe(),
      element_quadrature_map,
      UPDATE_GRADIENTS | UPDATE_MAPPING_JACOBIAN_INVERSE);
    element_fe_values.mapShapes(tess,element,level);
    NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > &gradients_here=
      gradients_here_map[nv];
    for (int n=0;n<n_solution_vectors;n++) {
      dof_handler.getFunctionDerivatives(element_fe_values,element,
        level,*solutions[n],gradients_here[n]);
    }
    for (int n=0;n<n_solution_vectors;n++) {
      for (int component=0;component<n_components;component++){
        for (int p=0; p<n_q_points; ++p) {
          gradients_here[n][p][component]-=
            neighbor_gradients[n][p][component];
        }
      }
    }
    const NumPtr<Point<SPACEDIM> > &normal_vectors=
      fe_face_values.getNormalVectors(neighbor_child,ncf);
    NumPtr<NumPtr<NumPtr<REAL> > > &normal_derivative_here=
      normal_derivative_here_map[nv];
    for (int n=0;n<n_solution_vectors;n++) {
      for (int component=0;component<n_components;component++){
        for (int point=0;point<n_q_points;point++) {
          normal_derivative_here[n][point][component]=
            gradients_here[n][point][component].getValues()*
            normal_vectors[point];
        }
      }
    }
    if (coefficients) {
      if (coefficients->nComponents()==1) {
        NumPtr<REAL> &coefficient_values1=coefficient_values1_map[nv];
        coefficients->value(coefficient_values1,face_mapped_points,0);
        for (int n=0;n<n_solution_vectors;n++) {
          for (int component=0;component<n_components; 
          component++) {
            for (int point=0;point<n_q_points;point++) {
              normal_derivative_here[n][point][component]*=
                coefficient_values1[point];
            }
          }
        }
      } else {
        NumPtr<Vector> &coefficient_values=coefficient_values_map[nv];
        coefficients->value(coefficient_values,face_mapped_points);
        for (int n=0;n<n_solution_vectors;n++) {
          for (int component=0;component<n_components; 
          component++) {
            for (int point=0;point<n_q_points;point++) {
              normal_derivative_here[n][point][component]*=
                coefficient_values[point][component];
            }
          }
        }
      }
    }
    NumPtr<REAL> JxW_values=
      fe_face_values.getJxWValues(neighbor_child,ncf);
    int neighbor_child_face_index=
      neighbor_child->codimension1ShapeIndex(ncf);
    CHECK_TEST(neighbor_child_face_index>=0);
    map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> >::
      iterator fi_iterator=face_integrals.
        find(make_pair(neighbor_child_face_index,neighbor_child_level));
    CHECK_TEST(fi_iterator!=face_integrals.end());
    NumPtr<REAL> &face_child_integral=fi_iterator->second;
    face_child_integral.initialize(0.);
    for (int n=0;n<n_solution_vectors;n++) {
      for (int component=0;component<n_components;component++){
        if (component_mask[component]) {
          for (int p=0;p<n_q_points;p++) {
            face_child_integral[n]+=
              pow(normal_derivative_here[n][p][component],2)
              *JxW_values[p];
          }
        }
      }
    }
  }
  int face_index=element->codimension1ShapeIndex(face_no);
  CHECK_TEST(face_index>=0);
  const CODIMENSION1_LEVEL_CLASS *face_level=
    static_cast<const CODIMENSION1_LEVEL_CLASS*>(level);
  map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> >::iterator
    fi_iterator=face_integrals.find(make_pair(face_index,face_level));
  CHECK_TEST(fi_iterator!=face_integrals.end());
  NumPtr<REAL> &face_integral=fi_iterator->second;
  face_integral.initialize(0.);
  const CODIMENSION1_LEVEL_CLASS *face_child_level=
    face_level->finerLevel();
  for (int subface_no=0; 
  subface_no<childrenPerCodimension1Shape();subface_no++) {
    int face_child_index=face_level->childIndex(face_index,subface_no);
    fi_iterator=face_integrals.
      find(make_pair(face_child_index,face_child_level));
    CHECK_TEST(fi_iterator!=face_integrals.end());
    NumPtr<REAL> &face_child_integral=fi_iterator->second;
    for (int n=0;n<n_solution_vectors;n++) {
      face_integral[n]+=face_child_integral[n];
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM!=1)
void KellyErrorEstimator::estimateSome(const Mapping &mapping, 
const DofHandler &dof_handler,
map<int,Quadrature<SPACEDIM-1> > &quadrature_map,
const NumPtr<const Vector*> &solutions,
const NumPtr<bool> &component_mask,const FunctionSystem *coefficients,
map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> > 
&face_integrals) {
  int n_solution_vectors=solutions.getNumber();
  Tessellation &tess=dof_handler.getTessellation();
  UPDATE_FLAGS update_flags=
    UPDATE_GRADIENTS | UPDATE_JxW_VALUES | UPDATE_NORMAL_VECTORS |
    UPDATE_MAPPING_JACOBIAN_INVERSE | UPDATE_Q_POINTS;
  FEFaceValues fe_face_values(mapping,dof_handler.getFe(),
    quadrature_map,update_flags);
//quadratures on triangular faces are not guaranteed to be independent
//  of the ordering of the barycentric coordinates
//FEFaceValues and FESubfaceValues constructors use static shape functions
//thus, we cannot guarantee that two elements sharing a face will map the
//  face quadrature points to the same locations, or 
//  even to the same sets of points
//we compute mapped lattice points and used these to provide initial guess
//  for call to Mapping::canonicalPoints
  int mapping_degree=mapping.getDegree();
  map<int,Quadrature<SPACEDIM-1> > lattice_quadrature_map;
  for (map<int,Quadrature<SPACEDIM-1> >::const_iterator qm_iterator=
  quadrature_map.begin();qm_iterator!=quadrature_map.end();
  qm_iterator++) {
    int nv=qm_iterator->first;
    NumPtr<Point<SPACEDIM-1> > feslp;
    equallySpacedLatticePointsForFace(nv,mapping_degree+1,feslp);
    Quadrature<SPACEDIM-1> q(feslp);
    lattice_quadrature_map[nv]=q;
  }
  FEFaceValues fe_face_values_lattice(mapping,dof_handler.getFe(),
    lattice_quadrature_map,UPDATE_Q_POINTS);
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    const LEVEL_CLASS *level=tess.presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    const CODIMENSION1_LEVEL_CLASS *face_level=
      static_cast<const CODIMENSION1_LEVEL_CLASS*>(level);
    for (int face_no=0; 
    face_no<element->numberCodimension1Shapes();face_no++) {
      int face_index=element->codimension1ShapeIndex(face_no);
      const BoundaryCondition *bc=
        face_level->getBoundaryCondition(face_index);
      bool face_at_boundary=(bc!=0);
      const LEVEL_CLASS *neighbor_level=
        level->neighborLevel(element,face_no);
      if (!face_level->hasChildren(face_index) && !face_at_boundary &&
      neighbor_level==level &&
      level->neighborIndex(element,face_no)<level->presentIndex()) {
        continue;
      }
      if (face_integrals[make_pair(face_index,face_level)][0]>=0.) {
        continue;
      }
      if (!face_at_boundary) {
        if (neighbor_level->isCoarserThan(level)) {
          continue;
        }
      }
      if (face_at_boundary && bc->isEssential()) {
        face_integrals[make_pair(face_index,face_level)].initialize(0.);
        continue;
      }
      if (!face_level->hasChildren(face_index)) {
        integrateOverRegularFace(dof_handler,solutions,
          component_mask,coefficients,face_integrals,element,level,
          face_no,fe_face_values,fe_face_values_lattice);
      } else {
        integrateOverIrregularFace(dof_handler,solutions,component_mask,
          coefficients,face_integrals,element,level,face_no,
          fe_face_values,fe_face_values_lattice);
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM!=1)
void KellyErrorEstimator::work(const Mapping &mapping,
const DofHandler &dof_handler,
map<int,Quadrature<SPACEDIM-1> > &quadrature_map,
const NumPtr<const Vector*> &solutions,NumPtr<Vector*> &errors,
const NumPtr<bool> &component_mask_,
const FunctionSystem *coefficients) {
  int n_components=dof_handler.getFe().nComponents();
  int n_solution_vectors=solutions.getNumber();
#ifdef DEBUG
  CHECK_TEST(n_solution_vectors>0);
  CHECK_SAME(n_solution_vectors,errors.getNumber());
  int cmn=component_mask_.getNumber();
  if (cmn>0) {
    CHECK_SAME(cmn,n_components);
    bool found_true=false;
    for (int i=0;i<cmn;i++) {
      if (component_mask_[i]) {
        found_true=true;
        break;
      }
    }
    CHECK_TEST(found_true);
  }
  if (coefficients) {
    int cnc=coefficients->nComponents();
    CHECK_TEST(cnc==n_components || cnc==1);
  }
  int n_dofs=dof_handler.numberDofs();
  for (int n=0;n<solutions.getNumber();n++) {
    CHECK_SAME(solutions[n]->size(),n_dofs);
  }
#endif
  for (map<int,Quadrature<SPACEDIM-1> >::const_iterator 
  qm_iterator=quadrature_map.begin();qm_iterator!=quadrature_map.end();
  qm_iterator++) {
    int nv=qm_iterator->first;
    int n_q_points=qm_iterator->second.numberPoints();
//  normal_derivative_here[solution no][quadrature point no.][component]
    NumPtr<NumPtr<NumPtr<REAL> > > &normal_derivative_here=
      normal_derivative_here_map[nv];
//  gradients_here[solution no.][face][component][spatial coordinate]
    NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > &gradients_here=
      gradients_here_map[nv];
//  gradients_neighbor[solution no.][face][component][spatial coordinate]
    NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > &neighbor_gradients=
      neighbor_gradients_map[nv];
    NumPtr<REAL> &coefficient_values1=coefficient_values1_map[nv];
    NumPtr<Vector> &coefficient_values=coefficient_values_map[nv];

    normal_derivative_here.allocate(n_solution_vectors);
    gradients_here.allocate(n_solution_vectors);
    neighbor_gradients.allocate(n_solution_vectors);
    coefficient_values1.allocate(n_q_points);
    coefficient_values.allocate(n_q_points);
    for (int i=0;i<n_solution_vectors;i++) {
      normal_derivative_here[i].allocate(n_q_points);
      for (int qp=0;qp<n_q_points;qp++) {
        normal_derivative_here[i][qp].allocate(n_components);
      }
    }
    for (int qp=0;qp<n_q_points;qp++) {
      coefficient_values[qp].reinit(n_components);
    }
  }
  NumPtr<bool> component_mask(n_components);
  if (component_mask_.getNumber()==0) {
    component_mask.initialize(true);
  } else {
    component_mask.copyFrom(component_mask_);
  }
//face_integrals[pair(face,level)][solution no.]
  map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> >
    face_integrals;
  Tessellation &tess=dof_handler.getTessellation();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    const SHAPE_CLASS *element=tess.presentShape();
    const CODIMENSION1_LEVEL_CLASS *face_level=
      static_cast<const CODIMENSION1_LEVEL_CLASS*>(tess.presentLevel());
    for (int face_no=0; 
    face_no<element->numberCodimension1Shapes();face_no++) {
      int face_index=element->codimension1ShapeIndex(face_no);
      NumPtr<REAL> &face_integral=
        face_integrals[make_pair(face_index,face_level)];
      if (!face_integral.ownsData()) {
        face_integral.allocate(n_solution_vectors);
        face_integral.initialize(-HUGE_VAL);
      }
    }
  }
  estimateSome(mapping,dof_handler,quadrature_map,solutions,
    component_mask,coefficients,face_integrals);
  int nas=tess.numberActiveShapes();
  for (int n=0;n<n_solution_vectors;n++) {
    (*errors[n]).reinit(nas,false);
  }
  int present_cell=0;
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),present_cell++) {
    const SHAPE_CLASS *element=tess.presentShape();
    const CODIMENSION1_LEVEL_CLASS *face_level=
      static_cast<const CODIMENSION1_LEVEL_CLASS*>(tess.presentLevel());
    for (int face_no=0; 
    face_no<element->numberCodimension1Shapes();face_no++) {
      int face_index=element->codimension1ShapeIndex(face_no);
      map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> >::
        const_iterator fi_iterator=face_integrals.
        find(make_pair(face_index,face_level));
      CHECK_TEST(fi_iterator!=face_integrals.end());
      const NumPtr<REAL> &face_integral=fi_iterator->second;
      for (int n=0;n<n_solution_vectors;n++) {
        (*errors[n])[present_cell]+=face_integral[n]; 
      }
    }
    REAL factor=element->diameter(vertices)/24.;
    for (int n=0;n<n_solution_vectors;n++) {
      REAL &err=(*errors[n])[present_cell];
      err=sqrt(err*factor);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void KellyErrorEstimator::work(const Mapping &mapping, 
const DofHandler &dof_handler,map<int,Quadrature<0> >&, 
const NumPtr<const Vector*> &solutions,NumPtr<Vector*> &errors,
const NumPtr<bool> &component_mask_,
const FunctionSystem *coefficient) {
  int n_components=dof_handler.getFe().nComponents();
  int n_solution_vectors=solutions.getNumber();
#ifdef DEBUG
  if (component_mask_.getNumber()>0) {
    CHECK_SAME(component_mask_.getNumber(),n_components);
    bool found_true=false;
    for (int i=0;i<component_mask_.getNumber();i++) {
      if (component_mask_[i]) {
        found_true=true;
        break;
      }
    }
    CHECK_TEST(found_true);
  }
  if (coefficient) {
    int cnc=coefficient->nComponents();
    CHECK_TEST(cnc==n_components || cnc==1);
  }
  CHECK_TEST(solutions.getNumber()>0);
  CHECK_SAME(solutions.getNumber(),errors.getNumber());
  int n_dofs=dof_handler.numberDofs();
  for (int n=0;n<solutions.getNumber();n++) {
    CHECK_SAME(solutions[n]->size(),n_dofs);
  }
#endif
  NumPtr<bool> component_mask(n_components);
  if (component_mask_.getNumber()==0) {
    component_mask.initialize(true);
  } else {
    component_mask.copyFrom(component_mask_);
  }
  int n_active_cells=
    dof_handler.getTessellation().numberActiveShapes();
  for (int n=0;n<errors.getNumber();n++) {
    (*errors[n]).reinit(n_active_cells);
  }
//gradients_here[solution no.][face][component][spatial coordinate]
  NumPtr<NumPtr<NumPtr<Tensor<1,1> > > > 
    gradients_here(n_solution_vectors);
//gradients_neighbor[solution no.][face][component][spatial coordinate]
  NumPtr<NumPtr<NumPtr<Tensor<1,1> > > > 
    gradients_neighbor(n_solution_vectors);
  int fps=Interval::facesPerShape();
  for (int i=0;i<n_solution_vectors;i++) {
    gradients_here[i].allocate(fps);
    gradients_neighbor[i].allocate(fps);
    for (int j=0;j<fps;j++) {
      gradients_here[i][j].allocate(n_components);
      gradients_neighbor[i][j].allocate(n_components);
    }
  }
//grad_neighbor[solution no.][component]
  NumPtr<Vector> grad_neighbor(n_solution_vectors);
  for (int i=0;i<n_solution_vectors;i++) {
    grad_neighbor[i].reinit(n_components);
  }
  Vector coefficient_values(n_components);
  if (!coefficient) coefficient_values=1.;

  map<int,Quadrature<SPACEDIM> > quadrature_map;
  LobattoQuadrature<1> quadrature(2);
  quadrature_map[Interval::verticesPerShape()]=quadrature;
  FEValues fe_values(mapping,dof_handler.getFe(),quadrature_map,
    UPDATE_GRADIENTS);
  int cell_index=0;
  Tessellation &tess=dof_handler.getTessellation();
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape(),cell_index++) {
    for (int n=0;n<errors.getNumber();n++) {
      (*errors[n])[cell_index]=0.;
    }
    LEVEL_CLASS *level=tess.presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    fe_values.mapShapes(tess,element,level);
    for (int s=0;s<solutions.getNumber();s++) {
      dof_handler.getFunctionDerivatives(fe_values,element,level,
        *solutions[s],gradients_here[s]);
    }
    REAL diam=element->diameter(vertices);
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int neighbor_index=level->neighborIndex(f);
      int nf=(f==0 ? 1 : 0);
      LEVEL_CLASS *neighbor_level=level->neighborLevel(f);
      bool dirichlet_bc=false;
      if (neighbor_index>=0) {
        while (neighbor_level->hasChildren(neighbor_index)) {
          neighbor_index=neighbor_level->childIndex(neighbor_index,nf);
          neighbor_level=neighbor_level->finerLevel();
        }
        const SHAPE_CLASS *neighbor_cell=
          neighbor_level->getShape(neighbor_index);
        fe_values.mapShapes(tess,neighbor_cell,neighbor_level);
        for (int s=0;s<solutions.getNumber();s++) {
          dof_handler.getFunctionDerivatives(fe_values,neighbor_cell,
            neighbor_level,*solutions[s],gradients_neighbor[s]);
        }
        for (int s=0;s<solutions.getNumber();s++) {
          for (int c=0;c<n_components;c++) {
            grad_neighbor[s][c]=gradients_neighbor[s][nf][c][0];
          }
        }
      } else {
        const BoundaryCondition *bc=
          level->CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(f);
        CHECK_POINTER(bc);
        if (bc->isNatural()) {
          FunctionSystem *fs=bc->getFunctionSystem();
          CHECK_POINTER(fs);
          CHECK_SAME(fs->nComponents(),n_components);
          const Point<SPACEDIM> &vertex=
            vertices[element->vertexIndex(f)];
          if (n_components==1) {
            REAL v=fs->value(vertex,0);
            for (int s=0;s<grad_neighbor.getNumber();s++) {
              grad_neighbor[s][0] = v;
            }
          } else {
            Vector v(n_components);
            fs->value(vertex,v);
            for (int c=0;c<n_components;c++) {
              if (bc->isEssential(c)) v[c]=0.;
            }
            for (int s=0;s<grad_neighbor.getNumber();s++) {
              grad_neighbor[s]=v;
            }
          }
        } else {
          dirichlet_bc=true; // at boundary and no Neumann component
        }
      }
      if (!dirichlet_bc) {
        if (coefficient) {
          const Point<SPACEDIM> &vertex=vertices[element->vertexIndex(f)];
          if (coefficient->nComponents()==1) {
            coefficient_values=coefficient->value(vertex,0);
          } else {
            coefficient->value(vertex,coefficient_values);
          }
        }
        for (int s=0;s<n_solution_vectors;s++) {
          const NumPtr<Tensor<1,1> > &ghsf=gradients_here[s][f];
          const Vector &gns=grad_neighbor[s];
          for (int component=0;component<n_components;component++) {
            if (component_mask[component]) {
              REAL jump=(ghsf[component][0]-gns[component]) 
                        *coefficient_values[component];
              (*errors[s])[cell_index]+=jump*jump*diam;
            }
          }
        }
      }
    }
    for (int s=0;s<errors.getNumber();s++) {
      REAL &err=(*errors[s])[cell_index];
      err=sqrt(err);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
KellyErrorEstimator::KellyErrorEstimator(const Mapping &mapping,
const DofHandler &dof_handler,
map<int,Quadrature<SPACEDIM-1> > &quadrature_map,
const NumPtr<const Vector*> &solutions,NumPtr<Vector*> &errors,
const NumPtr<bool> &component_mask_,
const FunctionSystem *coefficients) {
  work(mapping,dof_handler,quadrature_map,solutions,errors,
    component_mask_,coefficients);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
KellyErrorEstimator::KellyErrorEstimator(const Mapping &mapping,
const DofHandler &dof_handler,
map<int,Quadrature<SPACEDIM-1> > &quadrature_map,
const Vector &solution,Vector &error,const NumPtr<bool> &component_mask,
const FunctionSystem *coefficients) {
  NumPtr<const Vector*> solutions(1);
  solutions[0]=&solution;
  NumPtr<Vector*> errors(1);
  errors[0]=&error;
  work(mapping,dof_handler,quadrature_map,solutions,errors,
    component_mask,coefficients);
}

#include "NumPtr.C"
INSTANTIATE_NUMPTR(Vector*);
INSTANTIATE_NUMPTR(const Vector*);
template class NumPtrBase<NumPtr<Tensor<1,SPACEDIM> >,int>;
template class NumPtr<NumPtr<Tensor<1,SPACEDIM> > >;
template class NumPtrBase<NumPtr<NumPtr<Tensor<1,SPACEDIM> > >, int>;
template class NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > >;
#if (SPACEDIM>1)
template class _Rb_tree<int, 
  pair<int const, NumPtr<NumPtr<NumPtr<double> > > >, 
  _Select1st<pair<int const, NumPtr<NumPtr<NumPtr<double> > > > >,
  less<int>, 
  allocator<pair<int const, NumPtr<NumPtr<NumPtr<double> > > > > 
>;
template class _Rb_tree<int, 
  pair<int const, NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > >,
  _Select1st<pair<int const,
                  NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > > >, 
  less<int>, 
  allocator<pair<int const,
                 NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > > > 
>;
template class _Rb_tree<int, 
  pair<int const, NumPtr<Vector> >, 
  _Select1st<pair<int const, NumPtr<Vector> > >, 
  less<int>, 
  allocator<pair<int const, NumPtr<Vector> > > 
>;
template class _Rb_tree<
  pair<int, CODIMENSION1_LEVEL_CLASS const*>, 
  pair<pair<int, CODIMENSION1_LEVEL_CLASS const*> const, NumPtr<double> >, 
  _Select1st<
    pair<pair<int, CODIMENSION1_LEVEL_CLASS const*> const,NumPtr<double> >
    >, 
  less<pair<int, CODIMENSION1_LEVEL_CLASS const*> >, 
  allocator<
    pair<pair<int, CODIMENSION1_LEVEL_CLASS const*> const,NumPtr<double> >
    > 
>;
#endif
