#ifndef KELLY_ERROR_ESTIMATOR_H
#define KELLY_ERROR_ESTIMATOR_H "$Header:$"
//--------------------------  error_estimator.h  ---------------------------
//   $Id: error_estimator.h,v 1.62 2003/01/08 17:57:57 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  error_estimator.h  -------------------------
//
//modified from deal.II/deal.II/include/numerics/error_estimator.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************


#include <map>

#include "DofHandler.H"
#include "FEValues.H"
#include "Mapping.H"
#include "NumPtr.H"
#include "Point.H"
#include "Tensor.H"
#include "Vector.H"

/**
 *  Implementation of the error estimator by Kelly, Gago, Zienkiewicz
 *  and Babuska.  This error estimator tries to approximate the error
 *  per cell by integration of the jump of the gradient of the
 *  solution along the faces of each cell.  It can be understood as a
 *  gradient recovery estimator; see the survey of Ainsworth for a
 *  complete discussion.
 *
 *  It seem as if this error estimator should only be valid for linear 
 *  trial spaces, and there are indications that for higher order trial
 *  spaces the integrals computed here show superconvergence properties,
 *  i.e. they tend to zero faster than the error itself, thus ruling out
 *  the values as error indicators.
 *
 *  The error estimator really only estimates the error for the 
 *  generalized Poisson equation $-\nabla\cdot a(x) \nabla u = f$ with 
 *  either Dirichlet boundary conditions or generalized Neumann boundary
 *  conditions involving the conormal derivative $a\frac{du}{dn} = g$.
 *
 *  The error estimator returns a vector of estimated errors per cell 
 *  which can be used to feed the @ref{Triangulation}@p{<dim>::refine_*}
 *  functions. This vector contains elements of data type @p{float}, 
 *  rather than @p{REAL}, since accuracy is not so important here, and
 *  since this can save rather a lot of memory, when using many cells.
 *
 *  @sect3{Implementation}
 *
 *  In principle, the implementation of the error estimation is simple:
 *  let $$ \eta_K^2 = \frac h{24} \int_{\partial K} 
 *  \left[a \frac{\partial u_h}{\partial n}\right]^2 do $$
 *  be the error estimator for cell $K$. $[\cdot]$ denotes the jump of 
 *  the argument at the face. In the paper of Ainsworth, $h$ is divided
 *  by $24$, but this factor is a bit esoteric, stemming from 
 *  interpolation estimates and stability constants which may hold for 
 *  the Poisson problem, but may not hold for more general situations. 
 *  In the implementation, this factor is considered, but may lead to 
 *  wrong results. You may scale the vector appropriately afterwards.
 *
 *  To perform the integration, use is made of the @ref{FEFaceValues} 
 *  and @ref{FESubfaceValues} classes. The integration is performed by 
 *  looping over all cells and integrating over faces that are not yet 
 *  treated. This way we avoid integration on faces twice, once for each
 *  time we visit one of the adjacent cells. In a second loop over all 
 *  cells, we sum up the contributions of the faces (which are the 
 *  integrated square of the jumps) of each cell and take the square 
 *  root.
 *
 *  The integration is done using a quadrature formula on the face.
 *  For linear trial functions (@ref{FEQ1}), the @ref{QGauss2} or even 
 *  the @ref{QMidpoint} rule will suffice. For higher order elements, it
 *  is necessary to utilize higher order quadrature formulae as well.
 *
 *  We store the contribution of each face in a @p{map}, as provided by
 *  the C++ standard library, with the iterator pointing to that face 
 *  being the key into the map. In fact, we do not store the indicator 
 *  per face, but only the integral listed above. When looping the 
 *  second time over all cells, we have to sum up the contributions of 
 *  the faces, multiply them with $\frac h{24}$ and take the square 
 *  root. By doing the multiplication with $h$ in the second loop, we 
 *  avoid problems to decide with which $h$ to multiply, that of the 
 *  cell on the one or that of the cell on the other side of the face.
 *
 *  $h$ is taken to be the greatest length of the diagonals of the cell.
 *  For more or less uniform cells without deformed angles, this 
 *  coincides with the diameter of the cell.
 *  
 *
 *  @sect3{Vector-valued functions}
 *
 *  If the finite element field for which the error is to be estimated
 *  is vector-valued, i.e. the finite element has more than one
 *  component, then all the above can be applied to all or only some
 *  components at the same time. The main function of this class takes
 *  a list of flags denoting the components for which components the
 *  error estimator is to be applied; by default, it is a list of only
 *  @p{true}s, meaning that all variables shall be treated.
 *
 *  In case the different components of a field have different
 *  physical meaning (for example velocity and pressure in the Stokes
 *  equations), it would be senseless to use the same coefficient for
 *  all components. In that case, you can pass a function with as many
 *  components as there are components in the finite element field and
 *  each component of the error estimator will then be weighted by the
 *  respective component in this coefficient function. In the other
 *  case, when all components have the same meaning (for example the
 *  displacements in Lame's equations of elasticity), you can specify
 *  a scalar coefficient which will then be used for all components.
 *
 *
 *  @sect3{Boundary values}
 *  
 *  If the face is at the boundary, i.e. there is no neighboring cell to
 *  which the jump in the gradiend could be computed, there are two 
 *  possibilities:
 *  @begin{itemize}
 *  @item The face belongs to a Dirichlet boundary. Then the face is not
 *    considered, which can be justified looking at a dual problem 
 *    technique and should hold exactly if the boundary can be 
 *    approximated exactly by the finite element used (i.e. it is a 
 *    linear boundary for linear finite elements, quadratic for 
 *    isoparametric quadratic elements, etc). For boundaries which can 
 *    not be exactly approximated, one should consider the difference
 *    $z-z_h$ on the face, $z$ being a dual problem's solution which is
 *    zero at the true boundary and $z_h$ being an approximation, which
 *    in most cases will be zero on the numerical boundary. Since on the
 *    numerical boundary $z$ will not be zero in general, we would get 
 *    another term here, but this one is neglected for practical 
 *    reasons, in the hope that the error made here will tend to zero 
 *    faster than the energy error we wish to estimate.
 *
 *    Though no integration is necessary, in the list of face 
 *    contributions we store a zero for this face, which makes summing 
 *    up the contributions of the different faces to the cells easier.
 *
 *  @item The face belongs to a Neumann boundary.  In this case, the
 *    contribution of the face $F\in\partial K$ looks like
 *    $$ \int_F \left|g-a\frac{\partial u_h}{\partial n}\right|^2 ds $$
 *    where $g$ is the Neumann boundary function. If the finite element
 *    is vector-valued, then obviously the function denoting the Neumann
 *    boundary conditions needs to be vector-valued as well.
 *
 *  @item No other boundary conditions are considered.
 *  @end{itemize}
 *  The object describing the boundary conditions is obtained from the
 *  triangulation.
 *
 *  Thanks go to Franz-Theo Suttmeier for clarifications about boundary
 *  conditions.
 *
 *  
 *  @sect3{Handling of hanging nodes}
 *  
 *  The integration along faces with hanging nodes is quite tricky, 
 *  since one of the elements has to be shifted one level up or down. 
 *  See the documentation for the @ref{FESubFaceValues} class for more 
 *  information about technical issues regarding this topic.
 *
 *  In praxi, since we integrate over each face only once, we do this 
 *  when we are on the coarser one of the two cells adjacent to a 
 *  subface (a subface is defined to be the child of a face; seen from 
 *  the coarse cell, it is a subface, while seen from the refined cell 
 *  it is one of its faces). The reason is that finding neighborship 
 *  information is a bit easier then, but that's all practical 
 *  reasoning, nothing fundamental.
 *
 *  Since we integrate from the coarse side of the face, we have the 
 *  mother face readily at hand and store the result of the integration
 *  over that mother face (being the sum of the integrals along the 
 *  subfaces) in the abovementioned map of integrals as well. This 
 *  consumes some memory more than needed, but makes the summing up of 
 *  the face contributions to the cells easier, since then we have the 
 *  information from all faces of all cells at hand and need not think 
 *  about explicitly determining whether a face was refined or not. The
 *  same applies for boundary faces, see above.
 *
 *
 *  @sect3{Multiple solution vectors}
 *
 *  In some cases, for example in time-dependent problems, one would
 *  like to compute the error estimates for several solution vectors
 *  on the same grid at once, with the same coefficients, etc,
 *  e.g. for the solutions on several successive time steps. One could
 *  then call the functions of this class several times for each
 *  solution. However, it is observed that the largest factor in the
 *  computation of the error estimates (in terms of computing time) is
 *  initialization of @ref{FEFaceValues} and @ref{FESubFaceValues}
 *  objects, and iterating through all faces and subfaces. If the
 *  solution vectors live on the same grid, this effort can be reduced
 *  significantly by treating all solution vectors at the same time,
 *  initializing the @ref{FEFaceValues} objects only once per cell and
 *  for all solution vectors at once, and also only looping through
 *  the triangulation only once. For this reason, besides the
 *  @p{estimate} function in this class that takes a single input
 *  vector and returns a single output vector, there is also a
 *  function that accepts several in- and output vectors at the same
 *  time. 
 *
 *  @author Wolfgang Bangerth, 1998, 1999, 2000; parallelization by Thomas Richter, 2000
 */
//compare to DEAL KellyErrorEstimator
class KellyErrorEstimator {
  private:
#if (SPACEDIM!=1)
    map<int,NumPtr<NumPtr<NumPtr<REAL> > > > normal_derivative_here_map;
    map<int,NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > > 
      gradients_here_map;
    map<int,NumPtr<NumPtr<NumPtr<Tensor<1,SPACEDIM> > > > >
      neighbor_gradients_map;
    map<int,NumPtr<REAL> > coefficient_values1_map;
    map<int,NumPtr<Vector> > coefficient_values_map;

    void integrateOverRegularFace(const DofHandler&,
      const NumPtr<const Vector*>&,const NumPtr<bool>&,
      const FunctionSystem*,
      map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> >&,
      const SHAPE_CLASS*,const LEVEL_CLASS*,int,FEFaceValues&,
      FEFaceValues&);
    void integrateOverIrregularFace(const DofHandler&,
      const NumPtr<const Vector*>&,const NumPtr<bool>&,
      const FunctionSystem*,
      map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> >&,
      const SHAPE_CLASS*,const LEVEL_CLASS*,int,FEFaceValues&,
      FEFaceValues&);
    void estimateSome(const Mapping&,const DofHandler&,
      map<int,Quadrature<SPACEDIM-1> >&,const NumPtr<const Vector*>&,
      const NumPtr<bool>&,const FunctionSystem*,
      map<pair<int,const CODIMENSION1_LEVEL_CLASS*>,NumPtr<REAL> >&);
#endif
    void work(const Mapping&,const DofHandler&,
      map<int,Quadrature<SPACEDIM-1> >&,
      const NumPtr<const Vector*>&,NumPtr<Vector*>&,
      const NumPtr<bool>&,const FunctionSystem* = 0);
  public:
    KellyErrorEstimator(const Mapping&,const DofHandler&,
      map<int,Quadrature<SPACEDIM-1> >&,
      const NumPtr<const Vector*>&,NumPtr<Vector*>&,
      const NumPtr<bool>& = NumPtr<bool>(),const FunctionSystem* =0);
    KellyErrorEstimator(const Mapping&,const DofHandler&,
      map<int,Quadrature<SPACEDIM-1> >&,
      const Vector&,Vector&,
      const NumPtr<bool> &component_mask = NumPtr<bool>(),
      const FunctionSystem* =0);
};

#endif
