#ifndef LEVEL_H
#define LEVEL_H "$Header:$"
//----------------------------  tria_accessor.h  ---------------------------
//   $Id: tria_accessor.h,v 1.75 2003/04/11 21:22:31 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  tria_accessor.h  ---------------------------
//
//modified from deal.II/deal.II/include/grid/tria_accessor.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <limits.h> // for UCHAR_MAX
#include <ostream>
#include <set>
#include <vector>
#include <BoundaryCondition.H>
#include <Errors.H>
#include <MaterialModel.H>
#include <NumPtr.H>
#include <Pair.H>
#include <Shape.H>

class ShapeData {
  private:
    Shape *element;
    int parent;
    int child;
    bool used;
    bool user_flag;
    MaterialModel *material_model;
    void *user_pointer;
  public:
    ShapeData() : element(0),parent(-1),child(-1),used(false),
      user_flag(false),material_model(0),user_pointer(0) {}
    ~ShapeData() { }
    ShapeData& operator=(const ShapeData &ed) {
      element=ed.element;
      parent=ed.parent;
      child=ed.child;
      used=ed.used;
      user_flag=ed.user_flag;
      material_model=ed.material_model;
      user_pointer=ed.user_pointer;
      return *this;
    }
    inline bool isUsed() const { return used; }
    inline void setUsedFlag() { used=true; }
    inline void clearUsedFlag() { used=false; }
    inline bool userFlagSet() const { return user_flag; }
    inline void setUserFlag() { user_flag=true; }
    inline void clearUserFlag() { user_flag=false; }
    inline void* userPointer() const { return user_pointer; }
    inline void setUserPointer(void *p) { user_pointer=p; }
    inline void clearUserPointer() { user_pointer=0; }
    inline void setShape(Shape *e) { element=e; }
    inline const Shape* getShape() const { return element; }
    inline Shape* getShape() { return element; }
    inline int parentIndex() const { return parent; }
    inline bool hasParent() const { return parent != -1; }
    inline void setParent(int p) { parent=p; }
    inline int childIndex() const { return child; }
    inline bool hasChildren() const { return child != -1; }
    inline void setChild(int i) { child=i; }
    void printOn(ostream&) const;
#ifdef DEBUG
    void check(int n) const {
      CHECK_TEST(child==-1 || (0 <= child && child < n));
    }
#endif
};

class Level0 {
  protected:
    NumPtr<bool> refine_flags;
    NumPtr<bool> coarsen_flags;
    NumPtr<pair<Level0*,int> > neighbors;
    NumPtr<int> subdomain_ids;
    Level0 *coarser_level;
    Level0 *finer_level;
#if (SPACEDIM==1)
    static const BoundaryCondition* boundary_face_data[2];
#endif
  public:
    explicit Level0(Level0 *cl=0);
    virtual ~Level0();
    void copyLevel(const Level0&);
    void reserveSpace(int);
    virtual Level0* coarserLevel() { return coarser_level; }
    virtual const Level0* coarserLevel() const { return coarser_level; }
    virtual Level0* finerLevel() { return finer_level; }
    virtual const Level0* finerLevel() const { return finer_level; }
    virtual Level0* coarsestLevel() { 
      if (coarser_level==0) return this;
      return coarser_level->coarsestLevel();
    }
    virtual const Level0* coarsestLevel() const { 
      if (coarser_level==0) return this;
      return coarser_level->coarsestLevel();
    }
    virtual Level0* finestLevel() { 
      if (finer_level==0) return this;
      return finer_level->finestLevel();
    }
    virtual const Level0* finestLevel() const { 
      if (finer_level==0) return this;
      return finer_level->finestLevel();
    }
    int numberCoarserLevels() const {
      if (coarser_level==0) return 0;
      return coarser_level->numberCoarserLevels()+1;
    }
    int numberFinerLevels() const {
      if (finer_level==0) return 0;
      return finer_level->numberFinerLevels()+1;
    }
    bool isFinerThan(const Level0 *l) const {
      if (l==coarser_level) return true;
      if (!coarser_level) return false;
      return coarser_level->isFinerThan(l);
    }
    bool isCoarserThan(const Level0 *l) const {
      if (l==finer_level) return true;
      if (!finer_level) return false;
      return finer_level->isCoarserThan(l);
    }
    virtual Level0* neighborLevel(int index,int f) {
      return neighbors[index*2*SPACEDIM+f].first;
    }
    virtual const Level0* neighborLevel(int index,int f) const {
      return neighbors[index*2*SPACEDIM+f].first;
    }
    int neighborIndex(int index,int f) const {
      return neighbors[index*2*SPACEDIM+f].second;
    }
    void setNeighbor(int index,int f,Level0 *new_level,int new_index) {
      int j=index*2*SPACEDIM+f;
      neighbors[j].first=new_level;
      neighbors[j].second=new_index;
    }
    bool refineFlagSet(int index) const {return refine_flags[index];}
    void setRefineFlag(int index) { refine_flags[index]=true; }
    void clearRefineFlag(int index) { refine_flags[index]=false; }
    bool coarsenFlagSet(int index) const {return coarsen_flags[index];}
    void setCoarsenFlag(int index) { coarsen_flags[index]=true; }
    void clearCoarsenFlag(int index) { coarsen_flags[index]=false; }
    int subdomainId(int index) const {
      return subdomain_ids[index];
    }
    void setSubdomainId(int index,int new_subdomain_id) {
      subdomain_ids[index]=new_subdomain_id;
    }
#if (SPACEDIM==1)
    static const BoundaryCondition* getBoundaryCondition(int f) {
      CHECK_BOUNDS(f,0,2);
      return boundary_face_data[f];
    }
    static void setBoundaryCondition(int f,const BoundaryCondition *bc) {
      CHECK_BOUNDS(f,0,2);
      boundary_face_data[f]=bc;
    }
    void printFaceBoundaryConditions() const;
#endif
    virtual void printOn(ostream &os) const;
};

#ifdef LEVEL_FUNCTIONS
#undef LEVEL_FUNCTIONS
#endif
#define LEVEL_FUNCTIONS(n) \
    explicit Level##n(Level##n *cl=0); \
    ~Level##n(); \
    virtual void deleteElementDataShapes(); \
    virtual Level##n * coarserLevel() { \
      return dynamic_cast<Level##n *>(coarser_level); \
    } \
    virtual const Level##n * coarserLevel() const { \
      return dynamic_cast<const Level##n *>(coarser_level); \
    } \
    virtual Level##n * finerLevel() { \
      return dynamic_cast<Level##n *>(finer_level); \
    } \
    virtual const Level##n * finerLevel() const { \
      return dynamic_cast<const Level##n *>(finer_level); \
    } \
    virtual Level##n * coarsestLevel() { \
      return dynamic_cast<Level##n *>(Level0::coarsestLevel()); \
    } \
    virtual const Level##n * coarsestLevel() const { \
      return dynamic_cast<const Level##n *>(Level0::coarsestLevel()); \
    } \
    virtual Level##n * finestLevel() { \
      return dynamic_cast<Level##n *>(Level0::finestLevel()); \
    } \
    virtual const Level##n * finestLevel() const { \
      return dynamic_cast<const Level##n *>(Level0::finestLevel()); \
    } \
    int neighborIndex(int index,int f) const { \
      return Level0::neighborIndex(index,f); \
    } \
    int neighborIndex(const Shape##n *e,int f) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return Level0::neighborIndex(e->elementDataIndex(),f); \
    } \
    int neighborIndex(int f) const { \
      return Level0::neighborIndex(present_index##n,f); \
    } \
    virtual Level##n * neighborLevel(int index,int f) { \
      return dynamic_cast<Level##n *>( Level0::neighborLevel(index,f)); \
    } \
    virtual const Level##n * neighborLevel(int index,int f) const { \
      return dynamic_cast<const Level##n *>( \
        Level0::neighborLevel(index,f)); \
    } \
    virtual Level##n * neighborLevel(const Shape *e,int f) { \
      CHECK_SAME(e->numberDimensions(),n); \
      return dynamic_cast<Level##n *>(Level0::neighborLevel(\
        e->elementDataIndex(),f)); \
    } \
    virtual const Level##n * neighborLevel(const Shape *e,int f) \
    const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return dynamic_cast<const Level##n *>(Level0::neighborLevel(\
        e->elementDataIndex(),f)); \
    } \
    Level##n * neighborLevel(int f) { \
      return dynamic_cast<Level##n *>( \
        Level0::neighborLevel(present_index##n,f)); \
    } \
    const Level##n * neighborLevel(int f) const { \
      return dynamic_cast<const Level##n *>( \
        Level0::neighborLevel(present_index##n,f)); \
    } \
    int presentIndex() const { return present_index##n; } \
    bool iteratorStateValid() const { return present_index##n>=0; } \
    bool iteratorStatePastTheEnd() const { \
      return present_index##n>=element_data##n.getNumber(); \
    } \
    bool first(int index) { \
      int ne=element_data##n.getNumber(); \
      if (ne<=0) return false; \
      present_index##n=index; \
      CHECK_BOUNDS(present_index##n,0,ne); \
      return true; \
    } \
    bool first() { return first(0); } \
    bool last(int index) { \
      int ne=element_data##n.getNumber(); \
      if (ne<=0) return false; \
      if (index==-1) present_index##n=ne-1; \
      else present_index##n=index; \
      CHECK_BOUNDS(present_index##n,0,ne); \
      return true; \
    } \
    bool last() {  return last(-1); } \
    bool next() { \
      ++present_index##n; \
      return (present_index##n<element_data##n.getNumber()); \
    } \
    bool prev() { \
      --present_index##n; \
      return (present_index##n>=0); \
    } \
    const Shape##n * getShape(int index) const { \
      const Shape##n *s=dynamic_cast<const Shape##n *>( \
        element_data##n[index].getShape()); \
      return s; \
    } \
    Shape##n * getShape(int index) { \
      Shape##n *s= \
        dynamic_cast<Shape##n *>(element_data##n[index].getShape()); \
      return s; \
    } \
    const Shape##n * getShape() const { \
      return getShape(present_index##n); \
    } \
    Shape##n * getShape() { \
      return getShape(present_index##n); \
    } \
    int childIndex(int i) const { \
      CHECK_BOUNDS(i,0,Shape##n::childrenPerShape()); \
      return element_data##n[present_index##n].childIndex()+i; \
    } \
    int childIndex(int index,int c) const { \
      CHECK_BOUNDS(c,0,Shape##n::childrenPerShape()); \
      return element_data##n[index].childIndex()+c; \
    } \
    int childIndex(const Shape##n *e,int c) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return childIndex(e->elementDataIndex(),c); \
    } \
    int parentIndex() const { \
      return element_data##n[present_index##n].parentIndex(); \
    } \
    int parentIndex(int index) const { \
      return element_data##n[index].parentIndex(); \
    } \
    int parentIndex(const Shape##n *e) const { \
      return parentIndex(e->elementDataIndex()); \
    } \
    void setShape(Shape##n *element) { \
      if (element!=0) CHECK_SAME(element->numberDimensions(),n); \
      element_data##n[present_index##n].setShape(element); \
    } \
    void setShape(int index,Shape##n *element) { \
      if (element!=0) CHECK_SAME(element->numberDimensions(),n); \
      element_data##n[index].setShape(element); \
    } \
    int numberShapes() const { \
      return element_data##n.getNumber(); \
    } \
    int vertexIndex(int index,int i) const { \
      return element_data##n[index].getShape()->vertexIndex(i); \
    } \
    int vertexIndex(int i) const { \
      return element_data##n[present_index##n].getShape()-> \
        vertexIndex(i); \
    } \
    bool used() const {  \
      return element_data##n[present_index##n].isUsed();  \
    } \
    bool used(int index) const {  \
      return element_data##n[index].isUsed();  \
    } \
    bool used(const Shape##n *e) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return element_data##n[e->elementDataIndex()].isUsed(); \
    } \
    void setUsedFlag() {  \
      element_data##n[present_index##n].setUsedFlag(); \
    } \
    void setUsedFlag(int index) {  \
      element_data##n[index].setUsedFlag(); \
    } \
    void setUsedFlag(const Shape##n *e) {  \
      CHECK_SAME(e->numberDimensions(),n); \
      element_data##n[e->elementDataIndex()].setUsedFlag(); \
    } \
    void clearUsedFlag(int index) {  \
      element_data##n[index].clearUsedFlag(); \
    } \
    void clearUsedFlag(const Shape##n *e) {  \
      CHECK_SAME(e->numberDimensions(),n); \
      element_data##n[e->elementDataIndex()].clearUsedFlag(); \
    } \
    int numberUsedShapes() const { \
      int k=0; \
      for (int i=0;i<element_data##n.getNumber();i++) { \
        k+=element_data##n[i].isUsed(); \
      } \
      return k; \
    } \
    bool userFlagSet(int index) const { \
      return element_data##n[index].userFlagSet(); \
    } \
    bool userFlagSet(const Shape##n *e) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return element_data##n[e->elementDataIndex()].userFlagSet(); \
    } \
    bool userFlagSet() const { \
      return element_data##n[present_index##n].userFlagSet(); \
    } \
    void setUserFlag(int index) { \
      element_data##n[index].setUserFlag(); \
    } \
    void setUserFlag(const Shape##n *e) { \
      element_data##n[e->elementDataIndex()].setUserFlag(); \
    } \
    void setUserFlag() { \
      element_data##n[present_index##n].setUserFlag(); \
    } \
    void clearUserFlag() {  \
      element_data##n[present_index##n].clearUserFlag(); \
    } \
    void clearUserFlag(int index) {  \
      element_data##n[index].clearUserFlag(); \
    } \
    void clearUserFlag(const Shape##n *e) {  \
      CHECK_SAME(e->numberDimensions(),n); \
      element_data##n[e->elementDataIndex()].clearUserFlag(); \
    } \
    void* userPointer(int index) const { \
      return element_data##n[index].userPointer(); \
    } \
    void* userPointer(const Shape##n *e) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return element_data##n[e->elementDataIndex()].userPointer(); \
    } \
    void setUserPointer(int index,void *p) { \
      element_data##n[index].setUserPointer(p); \
    } \
    void setUserPointer(const Shape##n *e,void *p) { \
      CHECK_SAME(e->numberDimensions(),n); \
      element_data##n[e->elementDataIndex()].setUserPointer(p); \
    } \
    void clearUserPointer() {  \
      element_data##n[present_index##n].clearUserPointer(); \
    } \
    void clearUserPointer(int index) {  \
      element_data##n[index].clearUserPointer(); \
    } \
    void clearUserPointer(const Shape##n *e) {  \
      CHECK_SAME(e->numberDimensions(),n); \
      element_data##n[e->elementDataIndex()].clearUserPointer(); \
    } \
    bool hasParent() const { \
      return element_data##n[present_index##n].hasParent(); \
    } \
    bool hasParent(int index) const { \
      return element_data##n[index].hasParent(); \
    } \
    bool hasParent(const Shape##n *e) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return element_data##n[e->elementDataIndex()].hasParent(); \
    } \
    void setParent(int p) { \
      element_data##n[present_index##n].setParent(p); \
    } \
    void setParent(int index,int p) { \
      element_data##n[index].setParent(p); \
    } \
    void setParent(const Shape##n *e,int p) { \
      CHECK_SAME(e->numberDimensions(),n); \
      element_data##n[e->elementDataIndex()].setParent(p); \
    } \
    bool hasChildren() const { \
      return element_data##n[present_index##n].hasChildren(); \
    } \
    bool hasChildren(int index) const { \
      return element_data##n[index].hasChildren(); \
    } \
    bool hasChildren(const Shape##n *e) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return element_data##n[e->elementDataIndex()].hasChildren(); \
    } \
    void setChildren(int i) { \
      element_data##n[present_index##n].setChild(i); \
    } \
    void setChildren(int index,int i) { \
      element_data##n[index].setChild(i); \
    } \
    void setChildren(const Shape##n *e,int i) { \
      CHECK_SAME(e->numberDimensions(),n); \
      element_data##n[e->elementDataIndex()].setChild(i); \
    } \
    int subdomainId(int index) { \
      return Level0::subdomainId(index); \
    } \
    int subdomainId(const Shape##n *e) { \
      CHECK_SAME(e->numberDimensions(),n); \
      return Level0::subdomainId(e->elementDataIndex()); \
    } \
    int subdomainId() { \
      return Level0::subdomainId(present_index##n); \
    } \
    void setSubdomainId(int index,int new_subdomain_id) { \
      Level0::setSubdomainId(index,new_subdomain_id); \
    } \
    void setSubdomainId(const Shape##n *e,int sid) { \
      CHECK_SAME(e->numberDimensions(),n); \
      Level0::setSubdomainId(e->elementDataIndex(),sid); \
    } \
    void setSubdomainId(int new_subdomain_id) { \
      Level0::setSubdomainId(present_index##n,new_subdomain_id); \
    } \
    void setNeighbor(int index,int i,Level##n *new_level, \
    int new_index) { \
      Level0::setNeighbor(index,i,new_level,new_index); \
    } \
    void setNeighbor(const Shape##n *e,int i,Level##n *new_level, \
    int new_index) { \
      CHECK_SAME(e->numberDimensions(),n); \
      Level0::setNeighbor(e->elementDataIndex(),i,new_level,new_index);\
    } \
    void setNeighbor(int i,Level##n *new_level,int new_index) { \
      Level0::setNeighbor(present_index##n,i,new_level,new_index); \
    } \
    bool faceNormalPointsInward(const Shape##n *e,int f) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      const Level##n *nl=neighborLevel(e,f); \
      if (!nl) return false; \
      else if (this==nl) return neighborIndex(e,f)<e->elementDataIndex(); \
      else return isCoarserThan(nl); \
    } \
    bool refineFlagSet() const { \
      CHECK_SAME(SPACEDIM,n); \
      return refine_flags[present_index##n]; \
    } \
    bool refineFlagSet(int index) const { \
      CHECK_SAME(SPACEDIM,n); \
      return refine_flags[index]; \
    } \
    bool refineFlagSet(const Shape##n *e) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      CHECK_SAME(SPACEDIM,n); \
      return refine_flags[e->elementDataIndex()]; \
    } \
    void setRefineFlag(int index) { Level0::setRefineFlag(index); } \
    void setRefineFlag(const Shape##n *e) { \
      CHECK_SAME(e->numberDimensions(),n); \
      Level0::setRefineFlag(e->elementDataIndex()); \
    } \
    void setRefineFlag() { Level0::setRefineFlag(present_index##n); } \
    void clearRefineFlag(int index) { \
      Level0::clearRefineFlag(index); \
    } \
    void clearRefineFlag(const Shape##n *e) { \
      CHECK_SAME(e->numberDimensions(),n); \
      Level0::clearRefineFlag(e->elementDataIndex()); \
    } \
    void clearRefineFlag() { \
      Level0::clearRefineFlag(present_index##n); \
    } \
    bool coarsenFlagSet(int index) const { \
      return Level0::coarsenFlagSet(index); \
    } \
    bool coarsenFlagSet(const Shape##n *e) const { \
      CHECK_SAME(e->numberDimensions(),n); \
      return Level0::coarsenFlagSet(e->elementDataIndex()); \
    } \
    bool coarsenFlagSet() const { \
      return Level0::coarsenFlagSet(present_index##n); \
    } \
    void setCoarsenFlag(int index) { Level0::setCoarsenFlag(index); } \
    void setCoarsenFlag(const Shape##n *e) { \
      CHECK_SAME(e->numberDimensions(),n); \
      Level0::setCoarsenFlag(e->elementDataIndex()); \
    } \
    void setCoarsenFlag() { \
      Level0::setCoarsenFlag(present_index##n); \
    } \
    void clearCoarsenFlag(int index) { \
      Level0::clearCoarsenFlag(index); \
    } \
    void clearCoarsenFlag(const Shape##n *e) { \
      CHECK_SAME(e->numberDimensions(),n); \
      Level0::clearCoarsenFlag(e->elementDataIndex()); \
    } \
    void clearCoarsenFlag() { \
      Level0::clearCoarsenFlag(present_index##n); \
    } \
    void copyLevel(const Level##n &); \
    void reserveSpace(int); \
    void setInteriorChildNeighbors(int,const NumPtr<int>&); \
    void setInteriorChildNeighbors(const NumPtr<int> &child_indices) { \
      setInteriorChildNeighbors(present_index##n,child_indices); \
    } \
    int interiorChildVertex(int,int&,int&) const; \
    virtual void printOn(ostream &os) const;
#define LEVEL_MATERIAL_FUNCTIONS(n) \
    MaterialModel* getMaterialModel() const { \
      return model_data[present_index##n]; \
    } \
    MaterialModel* getMaterialModel(int index) const { \
      return model_data[index]; \
    } \
    MaterialModel* getMaterialModel(const Shape##n *e) const { \
      return model_data[e->elementDataIndex()]; \
    } \
    void setMaterialModel(MaterialModel *mm) { \
      model_data[present_index##n]=mm; \
    } \
    void setMaterialModel(int index,MaterialModel *mm) { \
      model_data[index]=mm; \
    } \
    void setMaterialModel(const Shape##n *e,MaterialModel *mm) { \
      model_data[e->elementDataIndex()]=mm; \
    } \
    void printMaterialModels() const;
#define LEVEL_BOUNDARY_FACE_FUNCTIONS(n) \
    const BoundaryCondition* getBoundaryCondition() const { \
      return boundary_face_data[present_index##n]; \
    } \
    const BoundaryCondition* getBoundaryCondition(int index) const { \
      return boundary_face_data[index]; \
    } \
    const BoundaryCondition* getBoundaryCondition(const Shape##n *e) \
    const { \
      return boundary_face_data[e->elementDataIndex()]; \
    } \
    void setBoundaryCondition(const BoundaryCondition *bc) { \
      boundary_face_data[present_index##n]=bc; \
    } \
    void setBoundaryCondition(int index,const BoundaryCondition *bc) { \
      boundary_face_data[index]=bc; \
    } \
    void setBoundaryCondition(const Shape##n *e, \
    const BoundaryCondition *bc) { \
      CHECK_SAME(e->numberDimensions(),n); \
      boundary_face_data[e->elementDataIndex()]=bc; \
    } \
    void printFaceBoundaryConditions() const;
#ifdef DEBUG
#define LEVEL_DEBUG_FUNCTIONS(n) \
    void checkElementData() const; \
    void checkNeighbors(int) const; \
    void checkNeighbors(); \
    void printChildren() const; \
    void printElementData() const; \
    void printNeighbors(int) const; \
    void printNeighbors() const { printNeighbors(present_index##n); } \
    void printRefineFlags() const; \
    void printCoarsenFlags() const; \
    void printUsedFlags() const; \
    void printUserFlags() const; \
    void plotShapes(const NumPtr<Point<SPACEDIM> >&); \
    void plotShapesIndividually(const NumPtr<Point<SPACEDIM> >&); \
    void plotUsedShapes(const NumPtr<Point<SPACEDIM> >&); \
    void plotUsedShapesIndividually(const NumPtr<Point<SPACEDIM> >&); \
    void plotActiveShapes(const NumPtr<Point<SPACEDIM> >&); \
    void plotActiveShapesIndividually(const NumPtr<Point<SPACEDIM> >&); \
    void plotNeighbors(const NumPtr<Point<SPACEDIM> >&); \
    void plotNeighbors(int,const NumPtr<Point<SPACEDIM> >&);
#endif


class Level1 : public Level0 {
  private:
    NumPtr<ShapeData> element_data1;
  protected:
    NumPtr<MaterialModel*> model_data;
#if (SPACEDIM>=2)
    NumPtr<const BoundaryCondition*> boundary_face_data;
#endif
#if (SPACEDIM==3)
    NumPtr<pair<const BoundaryCondition*,const BoundaryCondition*> > 
      boundary_line_data;
#endif
    int present_index1;
  public:
LEVEL_FUNCTIONS(1)
    void childFaceOnFaceOf(const Shape1*,int,int,int&,int&) const;
#if (SPACEDIM==1)
LEVEL_MATERIAL_FUNCTIONS(1)
#endif
#if (SPACEDIM==2)
LEVEL_BOUNDARY_FACE_FUNCTIONS(1)
#endif
#if (SPACEDIM==3)
    pair<const BoundaryCondition*,const BoundaryCondition*> 
    getBoundaryConditions(int index) const {
      return boundary_line_data[index];
    }
    pair<const BoundaryCondition*,const BoundaryCondition*> 
    getBoundaryConditions() const {
      return boundary_line_data[present_index1];
    }
    pair<const BoundaryCondition*,const BoundaryCondition*> 
    getBoundaryConditions(const Shape1 *e) const {
      return boundary_line_data[e->elementDataIndex()];
    }
    void setBoundaryConditions(int index,
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs) {
      pair<const BoundaryCondition*,const BoundaryCondition*> &p=
        boundary_line_data[index];
      p.first=bcs.first; 
      p.second=bcs.second; 
    }
    void setBoundaryCondition(int index,const BoundaryCondition *bc) {
      pair<const BoundaryCondition*,const BoundaryCondition*> &p=
        boundary_line_data[index];
      if (bc==0) { p.first=0; p.second=0; }
      else if (p.first==0) p.first=bc;
      else if (p.first!=bc) p.second=bc;
    }
    void setBoundaryCondition(const BoundaryCondition *bc) {
      setBoundaryCondition(present_index1,bc);
    }
    void setBoundaryCondition(const Shape1 *e,
    const BoundaryCondition *bc) {
      setBoundaryCondition(e->elementDataIndex(),bc);
    }
    void printLineBoundaryConditions() const;
#endif
#ifdef DEBUG
LEVEL_DEBUG_FUNCTIONS(1)
#endif
};

#if (SPACEDIM>=2)
class Level2 : public Level1 {
  private:
    NumPtr<ShapeData> element_data2;
    int present_index2;
  public:
LEVEL_FUNCTIONS(2)
    void neighborOfNeighbor(const Shape2*,int,const Level2*&,
      const Shape2*&,int&) const;
    void interiorChildVertexOnFaceOf(const Shape2*,int,int&,int&) const;
    void childFaceOnFaceOf(const Shape2*,int,int,int&,int&) const;
    void childFaceOn(int,const Shape2*,int,int&,int&) const;
#if (SPACEDIM==2)
LEVEL_MATERIAL_FUNCTIONS(2)
    bool lineTangentPointsBackward(const Shape2*,int) const;
#endif
#if (SPACEDIM==3)
LEVEL_BOUNDARY_FACE_FUNCTIONS(2)
#endif
#ifdef DEBUG
LEVEL_DEBUG_FUNCTIONS(2)
    void checkNeighborLevelAndIndex(int,int) const;
    void checkNeighborLevelAndIndex(int) const;
    void checkFaceVertices(int) const;
    void checkFaceVertices();
#endif
};
#endif

#if (SPACEDIM==3)
class Level3 : public Level2 {
  private:
    NumPtr<ShapeData> element_data3;
    NumPtr<MaterialModel*> model_data;
    int present_index3;
  public:
LEVEL_FUNCTIONS(3)
    void neighborOfNeighbor(const Shape3*,int,const Level3*&,
      const Shape3*&,int&) const;
    void interiorChildVertexOnFaceOf(const Shape3*,int,int&,int&) const;
    void childFaceOnFaceOf(const Shape3*,int,int,int&,int&) const;
    void childFaceOn(int,const Shape3*,int,int&,int&) const;
LEVEL_MATERIAL_FUNCTIONS(3)
    bool lineTangentPointsBackward(const Shape3*,int) const;
#ifdef DEBUG
LEVEL_DEBUG_FUNCTIONS(3)
    void plotActiveShapes(const NumPtr<Point<SPACEDIM> >&,VolGraphTool&);
    void checkNeighborLevelAndIndex(int,int) const;
    void checkNeighborLevelAndIndex(int) const;
    void checkFaceVertices(int) const;
    void checkFaceVertices();
#endif
};
#endif

#undef LEVEL_FUNCTIONS
#define LEVEL_CLASS D_CHOOSE(Level1,Level2,Level3)
#define CODIMENSION1_LEVEL_CLASS D_CHOOSE(Level0,Level1,Level2)
#endif
