#ifndef MAPPING_H
#define MAPPING_H "$Header:$"
//----------------------------  mapping.h  ---------------------------
//    $Id: mapping.h,v 1.26 2003/04/07 15:42:16 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  mapping.h  ---------------------------
//----------------------------  mapping_q.h  ---------------------------
//    $Id: mapping_q.h,v 1.33 2003/04/07 15:42:16 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  mapping_q.h  ---------------------------
//
//modified from deal.II/deal.II/include/fe/mapping.h and mapping_q.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <cmath>
#include <Array.H>
#include <Point.H>
#include <Quadrature.H>
#include <Matrix.H>
#include <ShapeFunction.H>
#include <Tessellation.H>
#include <Types.H>

enum UPDATE_FLAGS {
  UPDATE_DEFAULT                  = 0,
  UPDATE_VALUES                   = 1<<0, // fe shape fcn values
  UPDATE_GRADIENTS                = 1<<1, // fe shape fcn grads
  UPDATE_SECOND_DERIVATIVES       = 1<<2, // fe shape fcn gradgrads
  UPDATE_DIVERGENCE               = 1<<3, // fe shape fcn divergence
  UPDATE_CURL                     = 1<<4, // fe shape fcn curl
  UPDATE_BOUNDARY_FORMS           = 1<<5, // n ds
  UPDATE_Q_POINTS                 = 1<<6, // x(xi_q)
  UPDATE_DETERMINANT              = 1<<7, // | d x / d xi(xi_q) |
  UPDATE_JxW_VALUES               = 1<<8, // | d x / d xi(xi_q) | w_q
  UPDATE_NORMAL_VECTORS           = 1<<9, // n
  UPDATE_MAPPING_JACOBIAN_INVERSE = 1<<10, // d xi / d x
  UPDATE_MAPPING_JACOBIAN         = 1<<11, // d x / d xi
  UPDATE_MAPPING_SECOND_DERIVATIVE= 1<<12, // d^2 x / d xi^2
  UPDATE_TRANSFORMATION_VALUES    = 1<<13, // mapping b_s( xi )
  UPDATE_TRANSFORMATION_GRADIENTS = 1<<14, // mapping d b_s / d xi
  UPDATE_TRANSFORMATION_SECOND_DERIVATIVE = 1<<15 // mapping d^2 b_s/d xi^2
};
//UPDATE_NORMAL_VECTORS ==> UPDATE_BOUNDARY_FORMS
//UPDATE_BOUNDARY_FORMS ==> UPDATE_MAPPING_JACOBIAN
//UPDATE_JxW_VALUES ==> UPDATE_MAPPING_JACOBIAN
//UPDATE_MAPPING_JACOBIAN_INVERSE ==> UPDATE_MAPPING_JACOBIAN
//UPDATE_MAPPING_JACOBIAN ==> UPDATE_TRANSFORMATION_GRADIENTS
//UPDATE_Q_POINTS ==> UPDATE_TRANSFORMATION_VALUES

inline UPDATE_FLAGS& operator|=(UPDATE_FLAGS &f1,const UPDATE_FLAGS &f2)
{
  f1=static_cast<UPDATE_FLAGS>(f1 | f2);
  return f1;
}

inline UPDATE_FLAGS operator|(const UPDATE_FLAGS &f1,
const UPDATE_FLAGS &f2) {
  UPDATE_FLAGS result=f1;
  result |= f2;
  return result;
}

inline UPDATE_FLAGS& operator^=(UPDATE_FLAGS &f1,const UPDATE_FLAGS &f2)
{
  f1=static_cast<UPDATE_FLAGS>(f1 ^ f2);
  return f1;
}

inline UPDATE_FLAGS operator^(const UPDATE_FLAGS &f1,
const UPDATE_FLAGS &f2) {
  UPDATE_FLAGS result=f1;
  result ^= f2;
  return result;
}

inline UPDATE_FLAGS& operator&=(UPDATE_FLAGS &f1,const UPDATE_FLAGS &f2)
{
  f1=static_cast<UPDATE_FLAGS>(f1 & f2);
  return f1;
}

inline UPDATE_FLAGS operator&(const UPDATE_FLAGS& f1,
const UPDATE_FLAGS& f2) {
  UPDATE_FLAGS result=f1;
  result &= f2;
  return result;
}

ostream& operator<<(ostream&,UPDATE_FLAGS);

//  x = mesh coordinate
//  xi = canonical (unit) coordinate
//  b_s(xi) = basis function for coordinate mapping
//  construct data for mapping x(xi_q) = sum_s X_s b_s(xi_q)
//compare to DEAL Mapping::InternalDataBase
class MappingDataBase {
  protected:
    UPDATE_FLAGS update_once;
    UPDATE_FLAGS update_each;

//  set in derived class::computeMappingSupportPoints
//    depends on element
    bool map_higher_order;

    bool shape_values_ok;
    bool shape_derivatives_ok;
    bool shape_second_derivatives_ok;
    bool higher_order_shape_values_ok;
    bool higher_order_shape_derivatives_ok;
    bool higher_order_shape_second_derivatives_ok;
    bool mapping_support_points_ok;
    bool higher_order_mapping_support_points_ok;
    bool mapping_jacobian_ok;
    bool higher_order_mapping_jacobian_ok;
    bool mapping_jacobian_inverse_ok;
    bool higher_order_mapping_jacobian_inverse_ok;
    bool mapping_second_derivative_ok;
    bool higher_order_mapping_second_derivative_ok;
    bool mapping_determinant_ok;
    bool higher_order_mapping_determinant_ok;

    int degree;
    int n_shape_functions;
    int n_higher_order_shape_functions;
//  set in computeShapes 
//  used in computeShapesFor, computeFEValues
    const Quadrature<SPACEDIM> *quadrature;

//  b_s(xi):
    ShapeFunction<SPACEDIM> *shape_function;
    ShapeFunction<SPACEDIM> *higher_order_shape_function;
//  shape_value(s,q) = b_s(xi_q) ; s in lattice order
//  space allocated in allocateSpace
//  set in computeShapes if shape not required;
//    set in computeShapesFor if shape required
    Array<2,REAL> shape_values; 
    Array<2,REAL> higher_order_shape_values; 
//  shape_value_gradient(s,q) = d b_s / d xi (xi_q) ; s in lattice order
//  space allocated in allocateSpace
//  set in computeShapes if shape not required;
//    set in computeShapesFor if shape required
    Array<2,Tensor<1,SPACEDIM> > shape_derivatives;
    Array<2,Tensor<1,SPACEDIM> > higher_order_shape_derivatives;
//  shape_value_second_derivative(s,q) = d^2 b_s / d xi^2 (xi_q)
//  space allocated in allocateSpace
//  set in computeShapes if shape not required;
//    set in computeShapesFor if shape required
    Array<2,Tensor<2,SPACEDIM> > shape_second_derivatives;
    Array<2,Tensor<2,SPACEDIM> > higher_order_shape_second_derivatives;

//  X_s: s in dof order
//  space allocated in allocateSpace
//  set in derived class::computeMappingSupportPoints(shape)
    NumPtr<Point<SPACEDIM> > mapping_support_points;
    NumPtr<Point<SPACEDIM> > higher_order_mapping_support_points;
//  d x / d xi (xi_q) 
//  space allocated in computeJacobian(shape)
//  set in computeJacobian(shape)
//    requires quadrature_points and mapping_support_points
    NumPtr<Tensor<2,SPACEDIM> > mapping_jacobian; // formerly contravariant
    NumPtr<Tensor<2,SPACEDIM> > higher_order_mapping_jacobian;
//  det d x / d xi (xi_q) 
//  space allocated in computeJacobian(shape)
//  set in computeFEValues(shape)
//    requires quadrature_points and mapping_support_points
    NumPtr<REAL> mapping_determinant;
    NumPtr<REAL> higher_order_mapping_determinant;
//  [ d x / d xi ]^{-1} (xi_q) = covariant(q)
//  space allocated in computeJacobian(shape)
//  set in computeJacobian(shape)
//    requires quadrature_points and mapping_support_points
    NumPtr<Tensor<2,SPACEDIM> > mapping_jacobian_inverse; // was covariant
    NumPtr<Tensor<2,SPACEDIM> > higher_order_mapping_jacobian_inverse;
//  d^2 x / d xi^2 (xi_q) 
//  space allocated in computeJacobian(shape)
//  set in computeJacobian(shape)
//    requires quadrature_points and mapping_support_points
    NumPtr<Tensor<3,SPACEDIM> > mapping_second_derivative;
    NumPtr<Tensor<3,SPACEDIM> > higher_order_mapping_second_derivative;
#if (SPACEDIM>1)
//  n(f) = unit outer normal on face f of reference element
//  space allocated in allocateSpace
//  set in allocateSpace; does not require shape
//    requires number of vertices for shape
    NumPtr<Tensor<1,SPACEDIM> > reference_normal;
#endif
    MappingDataBase(const MappingDataBase&) { OBSOLETE(false); }
//  called from constructors of classes derived from MappingDataBase
    void allocateSpace(UPDATE_FLAGS,const Quadrature<SPACEDIM>*);
//  called from computeFEValues, computeFEFaceValues
//    calls computeMappingSupportPoints
//    computes mapped quadrature points, d x/d xi, (d x/ d xi)^{-1}
    void computeJacobian(const SHAPE_CLASS*,const Tessellation&,
      const LEVEL_CLASS*,int,int,NumPtr<Point<SPACEDIM> >&,bool&);
  public:
    static bool use_full_degree_mapping_on_all_cells;
    MappingDataBase(int);
    virtual ~MappingDataBase() {
      if (higher_order_shape_function) {
        delete higher_order_shape_function; 
        higher_order_shape_function=0; 
      }
      if (shape_function) {
        delete shape_function; shape_function=0; 
      }
    }
//  Only need to evaluate shape functions and/or gradients once
    void setUpdateOnce(UPDATE_FLAGS);
//  Flags for quantities that depend on the Shape
    void setUpdateEach(UPDATE_FLAGS);
    UPDATE_FLAGS getUpdateEach() const { return update_each; }
    bool mapHigherOrder() const { return map_higher_order; }
    bool shapeValuesOK() const {
      return (map_higher_order && higher_order_shape_values_ok) ||
             (!map_higher_order && shape_values_ok);
    }
    bool shapeDerivativesOK() const {
      return (map_higher_order && higher_order_shape_derivatives_ok) ||
             (!map_higher_order && shape_derivatives_ok);
    }
    bool shapeSecondDerivativesOK() const {
      return (map_higher_order && higher_order_shape_second_derivatives_ok)
        || (!map_higher_order && shape_second_derivatives_ok);
    }
    bool requiresShapeToInitialize() const { 
      return (map_higher_order ? 
              higher_order_shape_function->requiresShapeToComputeValues() :
              shape_function->requiresShapeToComputeValues()); 
    }
    int nShapeFunctions() const { 
      return n_shape_functions; 
    }
    int nHigherOrderShapeFunctions() const { 
      return n_higher_order_shape_functions; 
    }
    const Tensor<2,SPACEDIM>& mappingJacobian(int q) const {
      if (map_higher_order) {
        CHECK_TEST(higher_order_mapping_jacobian_ok);
        return higher_order_mapping_jacobian[q];
      } else {
        CHECK_TEST(mapping_jacobian_ok);
        return mapping_jacobian[q];
      }
    }
//  set values and derivatives of mapping shape functions internally 
//    at input canonical points
    inline REAL shapeValue(int s,int q) const {
      CHECK_TEST(shape_values_ok);
      return shape_values(s,q);
    }
    inline REAL& shapeValue(int s,int q) {
      CHECK_TEST(shape_values_ok);
      return shape_values(s,q);
    }
    inline Tensor<1,SPACEDIM> shapeDerivative(int s,int q) const {
      CHECK_TEST(shape_derivatives_ok);
      return shape_derivatives(s,q);
    }
    inline Tensor<1,SPACEDIM>& shapeDerivative(int s,int q) {
      CHECK_TEST(shape_derivatives_ok);
      return shape_derivatives(s,q);
    }
    inline Tensor<2,SPACEDIM> shapeSecondDerivative(int s,int q) const {
      CHECK_TEST(shape_second_derivatives_ok);
      return shape_second_derivatives(s,q);
    }
    inline Tensor<2,SPACEDIM>& shapeSecondDerivative(int s,int q) {
      CHECK_TEST(shape_second_derivatives_ok);
      return shape_second_derivatives(s,q);
    }
    inline REAL higherOrderShapeValue(int s,int q) const {
      CHECK_TEST(higher_order_shape_values_ok);
      return higher_order_shape_values(s,q);
    }
    inline REAL& higherOrderShapeValue(int s,int q) {
      CHECK_TEST(higher_order_shape_values_ok);
      return higher_order_shape_values(s,q);
    }
    inline Tensor<1,SPACEDIM> higherOrderShapeDerivative(int s,int q) 
    const {
      CHECK_TEST(higher_order_shape_derivatives_ok);
      return higher_order_shape_derivatives(s,q);
    }
    inline Tensor<1,SPACEDIM>& higherOrderShapeDerivative(int s,int q) {
      CHECK_TEST(higher_order_shape_derivatives_ok);
      return higher_order_shape_derivatives(s,q);
    }
    inline Tensor<2,SPACEDIM> higherOrderShapeSecondDerivative(int s,int q) 
    const {
      CHECK_TEST(higher_order_shape_second_derivatives_ok);
      return higher_order_shape_second_derivatives(s,q);
    }
    inline Tensor<2,SPACEDIM>& higherOrderShapeSecondDerivative(int s,
    int q) {
      CHECK_TEST(higher_order_shape_second_derivatives_ok);
      return higher_order_shape_second_derivatives(s,q);
    }
    void copyMappingSupportPointsFrom(const MappingDataBase *md) {
      CHECK_SAME(mapping_support_points.getNumber(),
        md->mapping_support_points.getNumber());
      CHECK_SAME(higher_order_mapping_support_points.getNumber(),
        md->higher_order_mapping_support_points.getNumber());
      if (md->mapping_support_points_ok) {
        mapping_support_points.copyFrom(md->mapping_support_points);
        mapping_support_points_ok=true;
      }
      if (md->higher_order_mapping_support_points_ok) {
        higher_order_mapping_support_points.copyFrom(
          md->higher_order_mapping_support_points);
        higher_order_mapping_support_points_ok=true;
      }
      map_higher_order=md->map_higher_order;
    }
//  called from derived class constructor and from canonicalPoint
//    public for testMapping* only
    void computeShapes(const Quadrature<SPACEDIM>*);
//  called from FEValuesData::mapShapes
//    public for testMapping* only
    void computeShapesFor(const Shape*);
//  compute X so that x_d(xi_q)=X_{d,s} b_s(xi_q)
//  called from MappingDataBase::computeJacobian, 
//    Mapping::elementPoint(s), Mapping::canonicalPoint(s) 
    virtual void computeMappingSupportPoints(const SHAPE_CLASS*,
      const Tessellation&,const LEVEL_CLASS*)=0;
//  called from FEValuesData::mapShapes
//  calls computeJacobian to compute mapped quadrature pts and map deriv
//  computes det(mapping derivative) * quadrature weight:
    void computeFEValues(const SHAPE_CLASS*,const Tessellation&,
      const LEVEL_CLASS*,NumPtr<Point<SPACEDIM> >&,NumPtr<REAL>&,bool&,
      bool&);
//  computes x_d=X_{d,s} b_s(xi_0)
    Point<SPACEDIM> elementPoint() const;
//  computes x_{d,q}=X_{d,s} b_s(xi_q)
    void elementPoints(NumPtr<Point<SPACEDIM> >&) const;
//  given y, solve X_{d,s} b_s(xi)=y for xi
//    requires sufficiently good initial guess
    void canonicalPoint(const SHAPE_CLASS*,REAL,const Point<SPACEDIM>&,
      Point<SPACEDIM>&);
//  dest(dest_row,xi_q)_{i} = 
//    ( d x / d xi )(xi_q)_{i,j} * src(src_row,xi_q)_{j} :
    void preMultiplyByJacobian(Tensor<1,SPACEDIM>&,int,
      const Tensor<1,SPACEDIM>&) const;
//  dest(dest_row,xi_q)_{i,j} = 
//    ( d x / d xi )(xi_q)_{i,k} * src(src_row,xi_q)_{k,j} :
    void preMultiplyByJacobian(Tensor<2,SPACEDIM>&,int,
      const Tensor<2,SPACEDIM>&) const;
//  dest(dest_row,xi_q)_{i} = 
//    ( d x / d xi )(xi_q)_{j,i} * src(src_row,xi_q)_{j} :
    void preMultiplyByJacobianTranspose(Tensor<1,SPACEDIM>&,int,
      const Tensor<1,SPACEDIM>&) const;
//  dest_{j} = src_{i} * [ ( d x / d xi )(xi_q) ]^{-1}_{i,j} : 
    void postMultiplyByJacobianInverse(Tensor<1,SPACEDIM>&,int,
      const Tensor<1,SPACEDIM>&) const;
//  dest_{i,j} = src_{i,k} * [ ( d x / d xi )(xi_q) ]^{-1}_{k,j}:
    void postMultiplyByJacobianInverse(Tensor<2,SPACEDIM>&,int,
      const Tensor<2,SPACEDIM>&) const;
//  dest_i = (d x / d xi)(xi_1)_{ij} src_j / det ( d x / d xi )(xi_q)
    void hdivTransform(Tensor<1,SPACEDIM>&,int,const Tensor<1,SPACEDIM>&)
      const;
    void hdivTransformDerivative(Tensor<2,SPACEDIM>&,int,
      const Tensor<1,SPACEDIM>&,const Tensor<2,SPACEDIM>&) const;
    void hdivTransformDivergence(REAL&,int,REAL) const;
    void hcurlTransform(Tensor<1,SPACEDIM>&,int,const Tensor<1,SPACEDIM>&)
      const;
    void hcurlTransformDerivative(Tensor<2,SPACEDIM>&,int,
      const Tensor<1,SPACEDIM>&,const Tensor<2,SPACEDIM>&) const;
#if (SPACEDIM==1)
    void hcurlTransformCurl(REAL&,int,REAL,REAL) const;
#endif
#if (SPACEDIM==2)
    void hcurlTransformCurl(REAL&,int,REAL) const;
#endif
#if (SPACEDIM==3)
    void hcurlTransformCurl(Tensor<1,3>&,int,const Tensor<1,3>&) const;
#endif
#if (SPACEDIM>1)
//  called from FEValuesData::mapFaceShapes, FEValuesData::mapSubFaceShapes
//    calls computeJacobian to compute mapped quad pts, mapping derivs
//    compute mapped tangents, det[ d x/d xi ( xi_q ) ] * w_q, mapped normal
//  note that if requiresShapeToInitialize then 
//    all shape_values for all faces are recomputed for each face
//    such mappings probably should not be used 
    void computeFEFaceValues(const SHAPE_CLASS*,const Tessellation&,
      const LEVEL_CLASS*,int,bool,int,int,const NumPtr<REAL>&,
      NumPtr<Point<SPACEDIM> >&,NumPtr<REAL>&,NumPtr<Tensor<1,SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&,bool&,bool&,bool&,bool&);
#endif
#if (SPACEDIM==3)
    void computeFEEdgeValues(const SHAPE_CLASS*,const Tessellation&,
      const LEVEL_CLASS*,int,int,NumPtr<Point<3> >&,bool&);
#endif
    virtual void printOn(ostream&) const;
};

//compare to DEAL Mapping
class Mapping {
  protected:
    int degree;

  public:
    Mapping(int p=1) : degree(D_CHOOSE(1,p,p)) {}
    virtual ~Mapping () {}
    int getDegree() const { return degree; }
//  given xi in reference shape,
//  compute x_d=X_{d,s} b_s(xi)
//  higher_order_mapping is used if
//    1) higher_order_mapping_support_points.getNumber()>0, and
//    2) use_full_degree_mapping_on_all_cells or element on boundary
    virtual Point<SPACEDIM> elementPoint(const SHAPE_CLASS*,
      const Tessellation&,const LEVEL_CLASS*,const Point<SPACEDIM>&) 
      const;
    virtual void elementPoints(const SHAPE_CLASS*,const Tessellation&,
      const LEVEL_CLASS*,const NumPtr<Point<SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&) const;
//  given x in element (hopefully),
//    solve X_{d,s} b_s(xi) = x_d for xi
    void canonicalPoint(const SHAPE_CLASS*,const Tessellation&,
      const MappingDataBase*,const Point<SPACEDIM>&,Point<SPACEDIM>&) const;
//  void canonicalPoints(const SHAPE_CLASS*,const Tessellation&,
//    const LEVEL_CLASS*,const NumPtr<Point<SPACEDIM> >&,
//    NumPtr<Point<SPACEDIM> >&) const;
    virtual MappingDataBase* getMappingData(int,UPDATE_FLAGS,
      const Quadrature<SPACEDIM>*) const=0;
    virtual void printOn(ostream&) const;
};

//compare to DEAL MappingQ
class MappingL : public Mapping {
  public:
    class InternalData : public MappingDataBase {
      public:
        InternalData(int,int,UPDATE_FLAGS,const Quadrature<SPACEDIM>*);
        virtual ~InternalData() {}
//      compute X so that x_d(xi_q)=X_{d,s} b_s(xi_q)
        virtual void computeMappingSupportPoints(const SHAPE_CLASS*,
          const Tessellation&,const LEVEL_CLASS*);
        virtual void printOn(ostream&) const;
    };
  public:
    MappingL(int p=1) : Mapping(p) {}
    virtual ~MappingL() {}
    InternalData* getMappingData(int nv,UPDATE_FLAGS update_flags,
    const Quadrature<SPACEDIM> *q) const {
      return OPERATOR_NEW InternalData(nv,degree,update_flags,q);
    }
};

#endif
