// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <NonlinearDofHandler.H>

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
NonlinearDofHandler::NonlinearDofHandler(Tessellation &t,
FiniteElement &fe,const Mapping &m,ConstraintMatrix &hnc) : 
DofHandler(t,fe),mapping(m),hanging_node_constraints(hnc) {
  int degree=getFe().getDegree();

  distributeDofs();

  hanging_node_constraints.makeHangingNodeConstraints(*this);
  hanging_node_constraints.close();

  GaussianQuadrature<SPACEDIM> quadrature_formula(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(
    D_CHOOSE(2*degree-1,2*degree+1,2*degree+1)));
  int nv=D_CHOOSE(Interval::verticesPerShape(),
                  Quadrilateral::verticesPerShape(),
                  Hexahedron::verticesPerShape());
  qmap[nv]=quadrature_formula;

  UPDATE_FLAGS update_flags=
    UPDATE_VALUES | UPDATE_GRADIENTS | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  fe_values=OPERATOR_NEW FEValues(mapping,fe,qmap,update_flags);
#if (SPACEDIM>1)
# if (SPACEDIM==2)
  nqmap[Interval::verticesPerShape()]=GaussianQuadrature<SPACEDIM-1>(
  GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree-1));
# else
  nqmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<SPACEDIM-1>(
  GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+1));
# endif
  UPDATE_FLAGS face_update_flags=UPDATE_VALUES | UPDATE_Q_POINTS |
    UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES;
  fe_face_values=
    OPERATOR_NEW FEFaceValues(mapping,fe,nqmap,face_update_flags);
#endif

  interpolateBoundaryValues(mapping,boundary_values);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
NonlinearDofHandler::~NonlinearDofHandler() {
  delete fe_values; fe_values=0;
#if (SPACEDIM>=2)
  delete fe_face_values; fe_face_values=0;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool NonlinearDofHandler::nonlinearSystemEvaluation(const Vector &input,
Vector &output) const {
  output=0.;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);
    Vector cell_output(dofs_per_cell,0.);

    fe_values->mapShapes(*tess,element,level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values->getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values->getJxWValues(element);
    const Array<2,REAL> &sv=*fe_values->getShapeValues(element);
    const Array<2,Tensor<1,SPACEDIM> > &sg=
      *fe_values->getShapeGradients(element);

    NonlinearMaterialModel *nmm=
      dynamic_cast<NonlinearMaterialModel*>(level->getMaterialModel());
    CHECK_POINTER(nmm);

//  input_gradients[q][c][d], 
//    0<=q<=nqp, 0<=c<component,s 0<=d<SPACEDIM
//  first compute the input gradient
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > input_gradients;
    getFunctionDerivatives(*fe_values,element,level,input,input_gradients);
    nmm->elementResidual(q_points,jxw_values,sv,sg,input_gradients,
      cell_output);

    int shapes_per_cell=sv.size(0);
//  neumann boundary condition on face 1
//  homogeneous neumann boundary conditions on
//    faces 2,3 in 2D
//    faces 2,3,4,5 in 3D
#if (SPACEDIM==1)
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (level->neighborLevel(f)==0) {
        const BoundaryCondition *bc=level->Level0::getBoundaryCondition(f);
        CHECK_POINTER(bc);
        if (bc->isNatural()) {
          NumPtr<Point<1> > p(1);
          p[0]=Interval::canonicalVertex(f);
          map<int,Quadrature<1> > fqmap_1d;
          fqmap_1d[element->numberVertices()]=Quadrature<1>(p);
          FEValues face_fe_values(mapping,*selected_fe,fqmap_1d,
            UPDATE_VALUES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES);
          face_fe_values.mapShapes(*tess,element,level);
          const Array<2,REAL> &fsv=
            *face_fe_values.getShapeValues(element);
          const NumPtr<Point<1> > &face_q_points=
            face_fe_values.getQuadraturePoints(element);
          NumPtr<Point<1> > normal_vectors(1);
          normal_vectors[0][0]=(f==0 ? -1. : 1.);
          NumPtr<REAL> face_jxw_values(1); face_jxw_values[0]=1.;
          bc->faceNeumannValues(face_q_points,face_jxw_values,
            // normal_vectors,
            0,fsv,cell_output);
        }
      }
    }
#else
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const BoundaryCondition *bc=level->
        CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
      if (bc!=0) {
        if (bc->isNatural()) {
          const CODIMENSION1_SHAPE_CLASS *face_element=level->
            CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          fe_face_values->mapShapes(*tess,element,level,f);
          const Array<2,REAL> &fsv=
            *fe_face_values->getShapeValues(element);
          NumPtr<Point<SPACEDIM> > face_q_points=
            fe_face_values->getQuadraturePoints(element,f);
          NumPtr<REAL> face_jxw_values=
            fe_face_values->getJxWValues(element,f);
          const NumPtr<Point<SPACEDIM> > &normal_vectors=
            fe_face_values->getNormalVectors(element,f);
          bc->faceNeumannValues(face_q_points,face_jxw_values,
            // normal_vectors,
            offset,fsv,cell_output);
        }
      }
      offset+=fe_face_values->getFaceQuadrature(element,f).numberPoints();
    }
#endif
    NumPtr<int> local_dof_indices;
    getDofIndices(element,level,local_dof_indices);
    for (int i=0;i<dofs_per_cell;i++) {
      output[local_dof_indices[i]]-=cell_output[i];
    }
  }
//insert constraints for dirichlet boundary values
  for (map<int,REAL>::const_iterator bv_iterator=
  boundary_values.begin();bv_iterator!=boundary_values.end();
  bv_iterator++) {
    int dof=bv_iterator->first;
    output[dof]=input[dof]-bv_iterator->second;
  }
//insert constraints for hanging nodes
  hanging_node_constraints.setResidual(input,output);
  return true;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NonlinearDofHandler::printOn(ostream &os) const {
  os << "NonlinearDofHandler:" << endl;
  os << "\tmapping = " << &mapping
     << "\n\thanging_node_constraints = " << &hanging_node_constraints
     << endl;
}
