// "$Header:$"
//----------------------------  data_out_base.cc  -------------------------
//  $Id: data_out_base.cc,v 1.66 2003/02/06 18:14:33 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  data_out_base.cc  -------------------------
//
//modified from deal.II/base/source/data_out_base.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <Patch.H>
#include <FEValues.H>
#include <Mapping.H>
#include <Tracer.H>

int Patch::n_subdivisions=0;
const int Patch::no_neighbor=INT_MAX;
map<int,Quadrature<SPACEDIM> > Patch::q_map;
#define SET_SHAPE_QUADRATURE(shape) \
  { \
    NumPtr<Point<SPACEDIM> > p; \
    shape::staticEquallySpacedLatticePoints(n_subdivisions,p); \
    int n_points=p.getNumber(); \
    NumPtr<REAL> w(n_points); \
    w.initialize(1./static_cast<double>(n_points)); \
    q_map[shape::verticesPerShape()]=Quadrature<SPACEDIM>(p,w); \
  }
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Patch::setQuadratures(int n) {
  while (!q_map.empty()) {
    map<int,Quadrature<SPACEDIM> >::iterator qi=q_map.begin();
    q_map.erase(qi);
  }
  CHECK_TEST(n>0);
  n_subdivisions=n;
#if (SPACEDIM==1)
SET_SHAPE_QUADRATURE(Interval);
#endif
#if (SPACEDIM==2)
SET_SHAPE_QUADRATURE(Triangle);
SET_SHAPE_QUADRATURE(Quadrilateral);
#endif
#if (SPACEDIM==3)
SET_SHAPE_QUADRATURE(Tetrahedron);
SET_SHAPE_QUADRATURE(Prism);
SET_SHAPE_QUADRATURE(Hexahedron);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
const Quadrature<SPACEDIM>& Patch::getQuadrature() const {
  map<int,Quadrature<SPACEDIM> >::const_iterator qm_iterator=
    q_map.find(shape->numberVertices());
  CHECK_TEST(qm_iterator!=q_map.end());
  return qm_iterator->second; 
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Patch::Patch(const SHAPE_CLASS *s,const LEVEL_CLASS *l,
const Tessellation &tess,const Mapping &mapping,int nds,int pi) : 
shape(s),level(l),neighbors(s->numberCodimension1Shapes()),
patch_index(pi) {
  if (nds>0) data.reinit(nds,numberQuadraturePoints());
#ifdef DEBUG
  REAL bogus=HUGE_VAL;
  data.fillConstant(bogus);
#endif
  neighbors.initialize(no_neighbor);

  const NumPtr<Point<SPACEDIM> > &qp=getQuadrature().points();
  ep.allocate(qp.getNumber());
  mapping.elementPoints(shape,tess,level,qp,ep);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Patch::~Patch() {
  while (!q_map.empty()) {
    map<int,Quadrature<SPACEDIM> >::iterator qi=q_map.begin();
    q_map.erase(qi);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Patch::getMeshBounds(Point<SPACEDIM> &low,Point<SPACEDIM> &high) 
const {
  for (int q=0;q<ep.getNumber();q++) {
    const Point<SPACEDIM> &epq=ep[q];
    for (int d=0;d<SPACEDIM;d++) {
      low[d]=min(low[d],epq[d]);
      high[d]=max(high[d],epq[d]);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Patch::getDataBounds(int data_set,REAL &low,REAL &high) 
const {
  for (int j=0;j<data.size(1);j++) { 
    low=min(low,data(data_set,j));
    high=max(high,data(data_set,j));
  } 
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void Patch::plot(XYGraphTool &gt,int data_set) const {
  gt.movePen(ep[0][0],data(data_set,0));
  for (int i=1;i<=n_subdivisions;i++) {
    gt.drawLine(ep[i][0],data(data_set,i)); 
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void Patch::plot(AreaGraphTool &gt,int data_set,REAL ymin,REAL ymax) 
const {
  REAL dy=1./(ymax-ymin);
  if (dynamic_cast<const Triangle*>(shape)) {
    int k=0;
    for (int j=0;j<n_subdivisions;j++) {
      for (int i=0;i+j<n_subdivisions;i++,k++) {
        const Point<2> &x0=ep[k];
        const Point<2> &x1=ep[k+1];
        const Point<2> &x2=ep[k+n_subdivisions+1-j];
        REAL y0=data(data_set,k);
        REAL y1=data(data_set,k+1);
        REAL y2=data(data_set,k+n_subdivisions+1-j);
        REAL y=(y0+y1+y2)/3.;
        REAL frac=(y-ymin)*dy;
        REAL xpos[3]={x0[0],x1[0],x2[0]};
        REAL ypos[3]={x0[1],x1[1],x2[1]};
        gt.colorPolygon(frac,3,xpos,ypos);
        if (i+j+1<n_subdivisions) {
          const Point<2> &x3=ep[k+n_subdivisions+2-j];
          REAL y3=data(data_set,k+n_subdivisions+2-j);
          REAL y=(y3+y1+y2)/3.;
          REAL frac=(y-ymin)*dy;
          xpos[0]=xpos[1];
          ypos[0]=ypos[1];
          xpos[1]=x3[0];
          ypos[1]=x3[1];
          gt.colorPolygon(frac,3,xpos,ypos);
        }
      }
      k++;
    }
  } else {
    int k=0;
    for (int j=0;j<n_subdivisions;j++,k++) {
      for (int i=0;i<n_subdivisions;i++,k++) {
        const Point<2> &x0=ep[k];
        const Point<2> &x1=ep[k+1];
        const Point<2> &x2=ep[k+n_subdivisions+1];
        const Point<2> &x3=ep[k+n_subdivisions+2];
        REAL y0=data(data_set,k);
        REAL y1=data(data_set,k+1);
        REAL y2=data(data_set,k+n_subdivisions+1);
        REAL y3=data(data_set,k+n_subdivisions+2);
        REAL y=(y0+y1+y2+y3)*0.25;
        REAL frac=(y-ymin)*dy;
        REAL xpos[4]={x0[0],x1[0],x3[0],x2[0]};
        REAL ypos[4]={x0[1],x1[1],x3[1],x2[1]};
        gt.colorPolygon(frac,4,xpos,ypos);
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void Patch::plotSurface(VolGraphTool &gt,int data_set,
REAL ymin,REAL ymax,BOOLEAN do_best) const {
  REAL valdif=1./(ymax-ymin);
  if (do_best) {
    if (dynamic_cast<const Triangle*>(shape)) {
      int k=0;
      for (int j=0;j<n_subdivisions;j++) {
        for (int i=0;i+j<n_subdivisions;i++,k++) {
          const Point<2> &x0=ep[k];
          const Point<2> &x1=ep[k+1];
          const Point<2> &x2=ep[k+n_subdivisions+1-j];
          Vector3 v0(x0[0],x0[1],data(data_set,k));
          Vector3 v1(x1[0],x1[1],data(data_set,k+1));
          Vector3 v2(x2[0],x2[1],data(data_set,k+n_subdivisions+1-i));
          REAL frac=((v0[2]+v1[2]+v2[2])/3.-ymin)*valdif;
          frac=max(0.,min(1.,frac));
          gt.colorTriangle(frac,v0,v1,v2);
          if (i+j+1<n_subdivisions) {
            const Point<2> &x3=ep[k+n_subdivisions+2-j];
            Vector3 v3(x3[0],x3[1],data(data_set,k+n_subdivisions+2-j));
            REAL frac=((v1[2]+v3[2]+v2[2])/3.-ymin)*valdif;
            frac=max(0.,min(1.,frac));
            gt.colorTriangle(frac,v1,v3,v2);
          }
        }
        k++;
      }
    } else {
      int k=0;
      for (int j=0;j<n_subdivisions;j++,k++) {
        for (int i=0;i<n_subdivisions;i++,k++) {
          const Point<2> &x0=ep[k];
          const Point<2> &x1=ep[k+1];
          const Point<2> &x2=ep[k+n_subdivisions+1];
          const Point<2> &x3=ep[k+n_subdivisions+2];
          Vector3 v0(x0[0],x0[1],data(data_set,k));
          Vector3 v1(x1[0],x1[1],data(data_set,k+1));
          Vector3 v2(x2[0],x2[1],data(data_set,k+n_subdivisions+1));
          Vector3 v3(x3[0],x3[1],data(data_set,k+n_subdivisions+2));
          Vector3 vave=(v0+v1+v2+v3)*0.25;
          REAL frac=((v0[2]+v1[2]+vave[2])/3.-ymin)*valdif;
          frac=max(0.,min(1.,frac));
          gt.colorTriangle(frac,v0,v1,vave);
      
          frac=((v1[2]+v3[2]+vave[2])/3.-ymin)*valdif;
          frac=max(0.,min(1.,frac));
          gt.colorTriangle(frac,v1,v3,vave);
     
          frac=((v3[2]+v2[2]+vave[2])/3.-ymin)*valdif;
          frac=max(0.,min(1.,frac));
          gt.colorTriangle(frac,v3,v2,vave);
    
          frac=((v2[2]+v0[2]+vave[2])/3.-ymin)*valdif;
          frac=max(0.,min(1.,frac));
          gt.colorTriangle(frac,v2,v0,vave);
        }
      }
    }
  } else {
    if (dynamic_cast<const Triangle*>(shape)) {
      Vector3 v0(ep[0][0],ep[0][1],data(data_set,0));
      Vector3 v1(ep[n_subdivisions][0],ep[n_subdivisions][1],
                 data(data_set,n_subdivisions));
      int k=(n_subdivisions*(n_subdivisions+3))/2;
      Vector3 v2(ep[k][0],ep[k][1],data(data_set,k));
      REAL frac=((v0[2]+v1[2]+v2[2])/3.-ymin)*valdif;
      frac=max(0.,min(1.,frac));
      gt.colorTriangle(frac,v0,v1,v2);
    } else {
      Vector3 v0(ep[0][0],ep[0][1],data(data_set,0));
      Vector3 v1(ep[n_subdivisions][0],ep[n_subdivisions][1],
                 data(data_set,n_subdivisions));
      int k=n_subdivisions*(n_subdivisions+1);
      Vector3 v2(ep[k][0],ep[k][1],data(data_set,k));
      k=n_subdivisions*(n_subdivisions+2);
      Vector3 v3(ep[k][0],ep[k][1],data(data_set,k));
      Vector3 vave=(v0+v1+v2+v3)*0.25;
      REAL frac=((v0[2]+v1[2]+vave[2])/3.-ymin)*valdif;
      frac=max(0.,min(1.,frac));
      gt.colorTriangle(frac,v0,v1,vave);
  
      frac=((v1[2]+v3[2]+vave[2])/3.-ymin)*valdif;
      frac=max(0.,min(1.,frac));
      gt.colorTriangle(frac,v1,v3,vave);
 
      frac=((v3[2]+v2[2]+vave[2])/3.-ymin)*valdif;
      frac=max(0.,min(1.,frac));
      gt.colorTriangle(frac,v3,v2,vave);

      frac=((v2[2]+v0[2]+vave[2])/3.-ymin)*valdif;
      frac=max(0.,min(1.,frac));
      gt.colorTriangle(frac,v2,v0,vave);
    }
  }
}
#endif
#if (SPACEDIM==3)
void drawSurfaceWithinCube(VolGraphTool &gt,REAL frac,
REAL value[2][2][2],Vector3 vertex[2][2][2]) {
  Vector3 intersected[3][2][2];
  for (int dir=0;dir<3;dir++) {
    int id1=(dir+1)%3;
    int id2=(dir+2)%3;
    for (HAND h1=LEFT_HAND;;h1=RIGHT_HAND) {
      for (HAND h2=LEFT_HAND;;h2=RIGHT_HAND) {
        HAND hl[3],hr[3];
        hl[dir]=LEFT_HAND;
        hr[dir]=RIGHT_HAND;
        hl[id1]=hr[id1]=h1;
        hl[id2]=hr[id2]=h2;
        REAL vl=value[hl[0]][hl[1]][hl[2]];
        REAL vr=value[hr[0]][hr[1]][hr[2]];
        if (vl*vr<=ZERO) {
          REAL alpha=(abs(vl-vr)>ZERO ? vl/(vl-vr) : HALF);
          intersected[dir][h1][h2]=
            vertex[hl[0]][hl[1]][hl[2]]*(ONE-alpha)
           +vertex[hr[0]][hr[1]][hr[2]]*alpha;
        }
        if (h2==RIGHT_HAND) break;
      }
      if (h1==RIGHT_HAND) break;
    }
  }
  gt.drawSurfaceWithinCube(frac,intersected,value);
}
void Patch::plotSurface(VolGraphTool &gt,int data_set,REAL val,
REAL var_min,REAL var_max,BOOLEAN do_best) const {
  if (var_max<=var_min) return;
  val=max(var_min,min(var_max,val));
  REAL frac=(val-var_min)/(var_max-var_min);
  REAL value[2][2][2];
  Vector3 vertex[2][2][2];
  int nsd1=n_subdivisions+1;
  if (do_best) {
    if (dynamic_cast<const Hexahedron*>(shape)) {
      for (int k=0;k<n_subdivisions;k++) {
        for (int j=0;j<n_subdivisions;j++) {
          for (int i=0;i<n_subdivisions;i++) {
            for (HAND h2=LEFT_HAND;;h2=RIGHT_HAND) {
              int khn=(k+h2)*nsd1*nsd1;
              for (HAND h1=LEFT_HAND;;h1=RIGHT_HAND) {
                int jhn=(j+h1)*nsd1;
                for (HAND h0=LEFT_HAND;;h0=RIGHT_HAND) {
                  int q=i+h0+jhn+khn;
                  const Point<SPACEDIM> &epq=ep[q];
                  Vector3 &v=vertex[h0][h1][h2];
                  for (int d=0;d<SPACEDIM;d++) v[d]=epq[d];
                  value[h0][h1][h2]=getData(data_set,q)-val;
                  if (h0==RIGHT_HAND) break;
                }
                if (h1==RIGHT_HAND) break;
              }
              if (h2==RIGHT_HAND) break;
            }
            drawSurfaceWithinCube(gt,frac,value,vertex);
          }
        }
      }
    } else if (dynamic_cast<const Prism*>(shape)) {
      for (int k=0;k<n_subdivisions;k++) {
        for (int j=0;j<n_subdivisions;j++) {
          for (int i=0;i+j<n_subdivisions;i++) {
            for (HAND h2=LEFT_HAND;;h2=RIGHT_HAND) {
              int khn=(k+h2)*((n_subdivisions+1)*(n_subdivisions+2))/2;
              for (HAND h1=LEFT_HAND;;h1=RIGHT_HAND) {
                int jhn=((j+h1)*(2*n_subdivisions+3-(j+h1)))/2;
                for (HAND h0=LEFT_HAND;;h0=RIGHT_HAND) {
                  int q=i+jhn+khn;
                  if (i+j+h1<n_subdivisions) q+=h0;
                  const Point<SPACEDIM> &epq=ep[q];
                  Vector3 &v=vertex[h0][h1][h2];
                  for (int d=0;d<SPACEDIM;d++) v[d]=epq[d];
                  value[h0][h1][h2]=getData(data_set,q)-val;
                  if (h0==RIGHT_HAND) break;
                }
                if (h1==RIGHT_HAND) break;
              }
              if (h2==RIGHT_HAND) break;
            }
            drawSurfaceWithinCube(gt,frac,value,vertex);
          }
        }
      }
    } else {
      CHECK_POINTER(dynamic_cast<const Tetrahedron*>(shape));
//    subdivision of tetrahedron involves level surfaces in 7-sided domain
//      which requires modification of marching cubes
//    we avoid subdivision instead
      for (HAND h2=LEFT_HAND;;h2=RIGHT_HAND) {
        int k=h2*n_subdivisions;
        int khn=(k*(n_subdivisions*(n_subdivisions*3+12)+11
                -k*(3*n_subdivisions+6-k)))/6;
        for (HAND h1=LEFT_HAND;;h1=RIGHT_HAND) {
          int j=(h2==LEFT_HAND ? h1*n_subdivisions : 0);
          int jhn=(j*(2*(n_subdivisions-k)+3-j))/2;
          for (HAND h0=LEFT_HAND;;h0=RIGHT_HAND) {
            int q=jhn+khn;
            if (j+k<n_subdivisions) q+=h0*n_subdivisions;
            const Point<SPACEDIM> &epq=ep[q];
            Vector3 &v=vertex[h0][h1][h2];
            for (int d=0;d<SPACEDIM;d++) v[d]=epq[d];
            value[h0][h1][h2]=getData(data_set,q)-val;
            if (h0==RIGHT_HAND) break;
          }
          if (h1==RIGHT_HAND) break;
        }
        if (h2==RIGHT_HAND) break;
      }
      drawSurfaceWithinCube(gt,frac,value,vertex);
    }
  } else {
    if (dynamic_cast<const Hexahedron*>(shape)) {
      for (HAND h2=LEFT_HAND;;h2=RIGHT_HAND) {
        int kn=h2*n_subdivisions*nsd1*nsd1;
        for (HAND h1=LEFT_HAND;;h1=RIGHT_HAND) {
          int jn=h1*n_subdivisions*nsd1;
          for (HAND h0=LEFT_HAND;;h0=RIGHT_HAND) {
            int q=h0*n_subdivisions+jn+kn;
            const Point<SPACEDIM> &epq=ep[q];
            Vector3 &v=vertex[h0][h1][h2];
            for (int d=0;d<SPACEDIM;d++) v[d]=epq[d];
            value[h0][h1][h2]=getData(data_set,q)-val;
            if (h0==RIGHT_HAND) break;
          }
          if (h1==RIGHT_HAND) break;
        }
        if (h2==RIGHT_HAND) break;
      }
      drawSurfaceWithinCube(gt,frac,value,vertex);
    } else if (dynamic_cast<const Prism*>(shape)) {
      for (HAND h2=LEFT_HAND;;h2=RIGHT_HAND) {
        int kn=h2*n_subdivisions*((n_subdivisions+1)*(n_subdivisions+2))/2;
        for (HAND h1=LEFT_HAND;;h1=RIGHT_HAND) {
          int jn=(h1*n_subdivisions*((2-h1)*n_subdivisions+3))/2;
          for (HAND h0=LEFT_HAND;;h0=RIGHT_HAND) {
            int q=jn+kn;
            if (h1==LEFT_HAND) q+=h0*n_subdivisions;
            const Point<SPACEDIM> &epq=ep[q];
            Vector3 &v=vertex[h0][h1][h2];
            for (int d=0;d<SPACEDIM;d++) v[d]=epq[d];
            value[h0][h1][h2]=getData(data_set,q)-val;
            if (h0==RIGHT_HAND) break;
          }
          if (h1==RIGHT_HAND) break;
        }
        if (h2==RIGHT_HAND) break;
      }
      drawSurfaceWithinCube(gt,frac,value,vertex);
    } else {
      CHECK_POINTER(dynamic_cast<const Tetrahedron*>(shape));
      for (HAND h2=LEFT_HAND;;h2=RIGHT_HAND) {
        int k=h2*n_subdivisions;
        int khn=(k*(n_subdivisions*(n_subdivisions*3+12)+11
                -k*(3*n_subdivisions+6-k)))/6;
        for (HAND h1=LEFT_HAND;;h1=RIGHT_HAND) {
          int j=(h2==LEFT_HAND ? h1*n_subdivisions : 0);
          int jhn=(j*(2*(n_subdivisions-k)+3-j))/2;
          for (HAND h0=LEFT_HAND;;h0=RIGHT_HAND) {
            int q=jhn+khn;
            if (j+k<n_subdivisions) q+=h0*n_subdivisions;
            const Point<SPACEDIM> &epq=ep[q];
            Vector3 &v=vertex[h0][h1][h2];
            for (int d=0;d<SPACEDIM;d++) v[d]=epq[d];
            value[h0][h1][h2]=getData(data_set,q)-val;
            if (h0==RIGHT_HAND) break;
          }
          if (h1==RIGHT_HAND) break;
        }
        if (h2==RIGHT_HAND) break;
      }
      drawSurfaceWithinCube(gt,frac,value,vertex);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void Patch::draw(AreaGraphTool &gt) const {
  if (dynamic_cast<const Quadrilateral*>(shape)) {
    const Point<SPACEDIM> &ep0=ep[0];
    gt.movePen(ep0[0],ep0[1]);
    for (int i=1;i<=n_subdivisions;i++) {
      const Point<SPACEDIM> &epi=ep[i];
      gt.drawLine(epi[0],epi[1]);
    }
    for (int i=2*n_subdivisions+1;i<=n_subdivisions*(n_subdivisions+2);
    i+=n_subdivisions+1) {
      const Point<SPACEDIM> &epi=ep[i];
      gt.drawLine(epi[0],epi[1]);
    }
    gt.movePen(ep0[0],ep0[1]);
    for (int i=n_subdivisions+1;i<=n_subdivisions*(n_subdivisions+2);
    i+=n_subdivisions+1) {
      const Point<SPACEDIM> &epi=ep[i];
      gt.drawLine(epi[0],epi[1]);
    }
    for (int i=n_subdivisions*(n_subdivisions+1)+1;
    i<=n_subdivisions*(n_subdivisions+2);i++) {
      const Point<SPACEDIM> &epi=ep[i];
      gt.drawLine(epi[0],epi[1]);
    }
  } else {
    const Point<SPACEDIM> &ep0=ep[0];
    gt.movePen(ep0[0],ep0[1]);
    for (int i=1;i<=n_subdivisions;i++) {
      const Point<SPACEDIM> &epi=ep[i];
      gt.drawLine(epi[0],epi[1]);
    }
    int ij=2*n_subdivisions;
    for (int j=1;j<=n_subdivisions;ij+=n_subdivisions-j,j++) {
      const Point<SPACEDIM> &epi=ep[ij];
      gt.drawLine(epi[0],epi[1]);
    }
    gt.movePen(ep0[0],ep0[1]);
    ij=n_subdivisions+1;
    for (int j=1;j<=n_subdivisions;ij+=n_subdivisions+1-j,j++) {
      const Point<SPACEDIM> &epi=ep[ij];
      gt.drawLine(epi[0],epi[1]);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void Patch::contour(AreaGraphTool &gt,REAL value,int data_set) const {
  if (dynamic_cast<const Triangle*>(shape)) {
    REAL dif[4];
    int k=0;
    for (int j=0;j<n_subdivisions;j++) {
      for (int i=0;i+j<n_subdivisions;i++,k++) {
        const Point<2> &x0=ep[k];
        const Point<2> &x1=ep[k+1];
        const Point<2> &x2=ep[k+n_subdivisions+1-j];
        REAL xpos[4]={x0[0],x1[0],x2[0],x0[0]};
        REAL ypos[4]={x0[1],x1[1],x2[1],x0[1]};
        dif[0]=data(data_set,k)-value;
        dif[1]=data(data_set,k+1)-value;
        dif[2]=data(data_set,k+n_subdivisions+1-j)-value;
        dif[3]=dif[0];
        int count=0;
        for (int n=0;n<3;n++) {
          if (dif[n]*dif[n+1]<=0. && abs(dif[n+1])>0.) {
            REAL pos=dif[n]/(dif[n]-dif[n+1]);
            REAL x=xpos[n]+pos*(xpos[n+1]-xpos[n]);
            REAL y=ypos[n]+pos*(ypos[n+1]-ypos[n]);
            if (count==0) gt.movePen(x,y);
            else {
              gt.drawLine(x,y);
              break;
            }
            count++;
          }
        }
        if (i+j+1<n_subdivisions) {
          xpos[0]=xpos[1];
          ypos[0]=ypos[1];
          const Point<2> &x3=ep[k+n_subdivisions+2-j];
          xpos[1]=x3[0];
          ypos[1]=x3[1];
          xpos[3]=xpos[0];
          ypos[3]=ypos[0];
          dif[0]=dif[1];
          dif[1]=data(data_set,k+n_subdivisions+2-j)-value;
          dif[3]=dif[0];
          int count=0;
          for (int n=0;n<3;n++) {
            if (dif[n]*dif[n+1]<=0. && abs(dif[n+1])>0.) {
              REAL pos=dif[n]/(dif[n]-dif[n+1]);
              REAL x=xpos[n]+pos*(xpos[n+1]-xpos[n]);
              REAL y=ypos[n]+pos*(ypos[n+1]-ypos[n]);
              if (count==0) gt.movePen(x,y);
              else {
                gt.drawLine(x,y);
                break;
              }
              count++;
            }
          }
        }
      }
      k++;
    }
  } else {
    REAL dif[5],xpos[5],ypos[5],xy[4][2];
    int k=0;
    for (int j=0;j<n_subdivisions;j++,k++) {
      for (int i=0;i<n_subdivisions;i++,k++) {
        const Point<2> &x0=ep[k];
        const Point<2> &x1=ep[k+1];
        const Point<2> &x2=ep[k+n_subdivisions+2];
        const Point<2> &x3=ep[k+n_subdivisions+1];
        REAL xpos[5]={x0[0],x1[0],x2[0],x3[0],x0[0]};
        REAL ypos[5]={x0[1],x1[1],x2[1],x3[1],x0[1]};
        dif[0]=data(data_set,k)-value;
        dif[1]=data(data_set,k+1)-value;
        dif[2]=data(data_set,k+n_subdivisions+2)-value;
        dif[3]=data(data_set,k+n_subdivisions+1)-value;
        dif[4]=dif[0];
        int count=0,first,last;
        for (int n=0;n<4;n++) {
          if (dif[n]*dif[n+1]<=0. && abs(dif[n+1])>0.) {
            if (count==0) first=n;
            else last=n;
            count++;
            REAL pos=dif[n]/(dif[n]-dif[n+1]);
            xy[n][0]=xpos[n]+pos*(xpos[n+1]-xpos[n]);
            xy[n][1]=ypos[n]+pos*(ypos[n+1]-ypos[n]);
          }
        }
        if (count==4) {
          gt.movePen(xy[0][0],xy[0][1]);
          gt.drawLine(xy[2][0],xy[2][1]);
          gt.movePen(xy[1][0],xy[1][1]);
          gt.drawLine(xy[3][0],xy[3][1]);
        } else if (count==2) {
          gt.movePen(xy[first][0],xy[first][1]);
          gt.drawLine(xy[last][0],xy[last][1]);
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Patch::plotClipPlane(VolGraphTool &gt,REAL ymax,REAL ymin,
const REAL value[2][2][2],const Vector3 vertex[2][2][2],
const DofHandler *dofs,const Mapping *mapping,const Vector *single_data)
const {
  Vector3 intersected[3][2][2];
  Point<SPACEDIM> center(0.,0.,0.);
  int n=0;
  for (int dir=0;dir<3;dir++) {
    int id1=(dir+1)%3;
    int id2=(dir+2)%3;
    for (HAND h1=LEFT_HAND;;h1=RIGHT_HAND) {
      for (HAND h2=LEFT_HAND;;h2=RIGHT_HAND) {
        HAND hl[3],hr[3];
        hl[dir]=LEFT_HAND;
        hr[dir]=RIGHT_HAND;
        hl[id1]=hr[id1]=h1;
        hl[id2]=hr[id2]=h2;
        REAL vl=value[hl[0]][hl[1]][hl[2]];
        REAL vr=value[hr[0]][hr[1]][hr[2]];
        REAL alpha=ZERO;
        if (vl*vr<=ZERO) {
          alpha=(abs(vl-vr)>ZERO ? vl/(vl-vr) : HALF);
        }
        Vector3 &v=intersected[dir][h1][h2];
        v=vertex[hl[0]][hl[1]][hl[2]]*(ONE-alpha)
         +vertex[hr[0]][hr[1]][hr[2]]*alpha;
        if (vl*vr<=ZERO) {
          for (int i=0;i<SPACEDIM;i++) center[i]+=v[i];
          n++;
        }
        if (h2==RIGHT_HAND) break;
      }
      if (h1==RIGHT_HAND) break;
    }
  }
  center/=static_cast<REAL>(n);
  const Tessellation &tess=dofs->getTessellation();

  REAL min_dist=HUGE_VAL;
  int min_j=-1;
  for (int j=0;j<ep.getNumber();j++) {
    REAL dist=center.distance(ep[j]);
    if (dist<min_dist) {
      min_dist=dist;
      min_j=j;
    }
  }
  const Quadrature<SPACEDIM> &q=getQuadrature();
  const NumPtr<Point<SPACEDIM> > &qp=q.points();
  Point<SPACEDIM> cp=qp[min_j];

  MappingDataBase *md=mapping->getMappingData(shape->numberVertices(),
    UPDATE_TRANSFORMATION_VALUES | UPDATE_TRANSFORMATION_GRADIENTS,&q);
  md->computeMappingSupportPoints(shape,tess,level);
  mapping->canonicalPoint(shape,tess,md,center,cp);

  map<int,Quadrature<SPACEDIM> > qm;
  qm[shape->numberVertices()]=Quadrature<SPACEDIM>(cp);
  FEValues fev(*mapping,dofs->getFe(),qm,
    UPDATE_VALUES | UPDATE_JxW_VALUES | UPDATE_Q_POINTS);
  fev.mapShapes(tess,shape,level);
  NumPtr<REAL> values;
  dofs->getFunctionValues(fev,shape,level,*single_data,values);
  REAL frac=(values[0]-ymin)/(ymax-ymin);
  gt.drawSurfaceWithinCube(frac,intersected,value);
  delete md; md=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Patch::plotTetrahedralClipPlane(VolGraphTool &gt,REAL ymin,
REAL ymax,REAL location,int dir,HAND h,const DofHandler *dofs,
const Mapping *mapping,const Vector *single_data,
const Point<SPACEDIM> x[4]) const {
  int count=0;
  for (int l=0;l<4;l++) count+=(x[l][dir]<location);
  if (count==0 || count==4) return;
  REAL v2=x[2][dir]-location;
  REAL v3=x[3][dir]-location;
  REAL value[2][2][2]={x[0][dir]-location,v3,v2,v3,
                       x[1][dir]-location,v3,v2,v3};
  if (h==LEFT_HAND) {
    REAL *v=&value[0][0][0];
    for (int n=0;n<8;n++) v[n]=-v[n];
  }
  Vector3 vertex[2][2][2]={
    Vector3(x[0][0],x[0][1],x[0][2]),
    Vector3(x[3][0],x[3][1],x[3][2]),
    Vector3(x[2][0],x[2][1],x[2][2]),
    Vector3(x[3][0],x[3][1],x[3][2]),
    Vector3(x[1][0],x[1][1],x[1][2]),
    Vector3(x[3][0],x[3][1],x[3][2]),
    Vector3(x[2][0],x[2][1],x[2][2]),
    Vector3(x[3][0],x[3][1],x[3][2])};
  plotClipPlane(gt,ymax,ymin,value,vertex,dofs,mapping,single_data);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Patch::plotPrismClipPlane(VolGraphTool &gt,REAL ymin,REAL ymax,
REAL location,int dir,HAND h,const DofHandler *dofs,
const Mapping *mapping,const Vector *single_data,
const Point<SPACEDIM> x[6]) const {
  int count=0;
  for (int l=0;l<6;l++) count+=(x[l][dir]<location);
  if (count==0 || count==6) return;
  REAL v2=x[2][dir]-location;
  REAL v5=x[5][dir]-location;
  REAL value[2][2][2]={x[0][dir]-location,x[3][dir]-location,
    v2,v5,x[1][dir]-location,x[4][dir]-location,v2,v5};
  if (h==LEFT_HAND) {
    REAL *v=&value[0][0][0];
    for (int n=0;n<8;n++) v[n]=-v[n];
  }
  Vector3 vertex[2][2][2]={
    Vector3(x[0][0],x[0][1],x[0][2]),
    Vector3(x[3][0],x[3][1],x[3][2]),
    Vector3(x[2][0],x[2][1],x[2][2]),
    Vector3(x[5][0],x[5][1],x[5][2]),
    Vector3(x[1][0],x[1][1],x[1][2]),
    Vector3(x[4][0],x[4][1],x[4][2]),
    Vector3(x[2][0],x[2][1],x[2][2]),
    Vector3(x[5][0],x[5][1],x[5][2])};
  plotClipPlane(gt,ymax,ymin,value,vertex,dofs,mapping,
    single_data);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Patch::plotClipPlane(VolGraphTool &gt,REAL ymin,REAL ymax,
REAL location,int dir,HAND h,const DofHandler *dofs,
const Mapping *mapping,const Vector *single_data) const {
  Point<3> low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Point<3> high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  getMeshBounds(low,high);
  REAL vmin=low[dir];
  REAL vmax=high[dir];
  if (vmin>location || vmax<location) return;
  if (dynamic_cast<const Tetrahedron*>(shape)) {
    Point<SPACEDIM> x[4];
    int m=0;
    int nl=n_subdivisions+1;
    for (int k=0;k<n_subdivisions;k++) {
      for (int j=0;j+k<n_subdivisions;j++) {
        for (int i=0;i+j+k<n_subdivisions;i++,m++) {
          x[0]=ep[m]; // v0
          x[1]=ep[m+1]; // v1
          int nlk=nl-k;
          int nt=(nlk*(nlk+1))/2;
          x[2]=ep[m+nlk-j]; // v2
          x[3]=ep[m+nt]; // v4
          plotTetrahedralClipPlane(gt,ymin,ymax,location,dir,h,dofs,
            mapping,single_data,x);
          if (i+j+k+1<n_subdivisions) {
//          plotted tetrahedron with vertices 0,1,2,4
//          to fill hexadron, plot tetrahedra with vertices
//            1,3,2,4 & 1,3,4,5 & 3,4,5,7 & 2,3,7,4 & 2,4,7,6
            x[0]=ep[m+1]; // v1
            x[1]=ep[m+nlk-j+1]; // v3
            x[2]=ep[m+nlk-j]; // v2
            x[3]=ep[m+nt]; // v4
            plotTetrahedralClipPlane(gt,ymin,ymax,location,dir,h,dofs,
              mapping,single_data,x);
            x[0]=ep[m+1]; // v1
            x[1]=ep[m+nlk-j+1]; // v3
            x[2]=ep[m+nt]; // v4
            x[3]=ep[m+1+nt]; // v5
            plotTetrahedralClipPlane(gt,ymin,ymax,location,dir,h,dofs,
              mapping,single_data,x);
            x[0]=ep[m+nlk-j+1]; // v3
            x[1]=ep[m+nt]; // v4
            x[2]=ep[m+nt+1]; // v5
            x[3]=ep[m+nt+nlk-j]; // v7
            plotTetrahedralClipPlane(gt,ymin,ymax,location,dir,h,dofs,
              mapping,single_data,x);
            x[0]=ep[m+nlk-j]; // v2
            x[1]=ep[m+nlk-j+1]; // v3
            x[2]=ep[m+nt+nlk-j]; // v7
            x[3]=ep[m+nt]; // v4
            plotTetrahedralClipPlane(gt,ymin,ymax,location,dir,h,dofs,
              mapping,single_data,x);
            x[0]=ep[m+nlk-j]; // v2
            x[1]=ep[m+nt]; // v4
            x[2]=ep[m+nt+nlk-j]; // v7
            x[3]=ep[m+nt+nlk-j-1]; // v6
            plotTetrahedralClipPlane(gt,ymin,ymax,location,dir,h,dofs,
              mapping,single_data,x);
          }
        }
        m++;
      }
    }
  } else if (dynamic_cast<const Prism*>(shape)) {
    Point<SPACEDIM> x[6];
    int m=0;
    int nl=n_subdivisions+1;
    int nt=((n_subdivisions+2)*nl)/2;
    for (int k=0;k<n_subdivisions;k++) {
      for (int j=0;j<n_subdivisions;j++) {
        for (int i=0;i+j<n_subdivisions;i++,m++) {
          x[0]=ep[m];
          x[1]=ep[m+1];
          x[2]=ep[m+nl-j];
          x[3]=ep[m+nt];
          x[4]=ep[m+nt+1];
          x[5]=ep[m+nt+nl-j];
          plotPrismClipPlane(gt,ymin,ymax,location,dir,h,dofs,mapping,
            single_data,x);
          if (i+j+1<n_subdivisions) {
//          plotted prism with vertices 0,1,2,4,5,6
//          to fill hexadron, plot prism with vertices 1,2,3,5,6,7
            x[0]=ep[m+nt+nl+1-j];
            x[1]=ep[m+nt+1];
            x[2]=ep[m+nt+nl-j];
            x[3]=ep[m+nl+1-j];
            x[4]=ep[m+1];
            x[5]=ep[m+nl-j];
            plotPrismClipPlane(gt,ymin,ymax,location,dir,h,dofs,mapping,
              single_data,x);
          }
        }
        m++;
      }
    }
  } else { 
    Point<SPACEDIM> x[8];
    int m=0;
    int nl=n_subdivisions+1;
    int nt=nl*nl;
    for (int k=0;k<n_subdivisions;k++) {
      for (int j=0;j<n_subdivisions;j++) {
        for (int i=0;i<n_subdivisions;i++,m++) {
          x[0]=ep[m];
          x[1]=ep[m+1];
          x[2]=ep[m+nl];
          x[3]=ep[m+nl+1];
          x[4]=ep[m+nt];
          x[5]=ep[m+nt+1];
          x[6]=ep[m+nt+nl];
          x[7]=ep[m+nt+nl+1];
          int count=0;
          for (int l=0;l<8;l++) count+=(x[l][dir]<location);
          if (count==0 || count==8) continue;

//        value[h0][h1][h2]:
          REAL value[2][2][2]={x[0][dir]-location,x[4][dir]-location,
            x[2][dir]-location,x[6][dir]-location,x[1][dir]-location,
            x[5][dir]-location,x[3][dir]-location,x[7][dir]-location};
          if (h==LEFT_HAND) {
            REAL *v=&value[0][0][0];
            for (int n=0;n<8;n++) v[n]=-v[n];
          }
//        vertex[h0][h1][h2]:
          Vector3 vertex[2][2][2]={
            Vector3(x[0][0],x[0][1],x[0][2]),
            Vector3(x[4][0],x[4][1],x[4][2]),
            Vector3(x[2][0],x[2][1],x[2][2]),
            Vector3(x[6][0],x[6][1],x[6][2]),
            Vector3(x[1][0],x[1][1],x[1][2]),
            Vector3(x[5][0],x[5][1],x[5][2]),
            Vector3(x[3][0],x[3][1],x[3][2]),
            Vector3(x[7][0],x[7][1],x[7][2])};
          plotClipPlane(gt,ymax,ymin,value,vertex,dofs,mapping,
            single_data);
        }
        m++;
      }
    }
  } 
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Patch::draw(VolGraphTool &gt) const {
  if (dynamic_cast<const Hexahedron*>(shape)) {
    const Point<SPACEDIM> &ep0=ep[0];
    const Point<SPACEDIM> &ep1=ep[n_subdivisions];
    const Point<SPACEDIM> &ep2=ep[n_subdivisions*(n_subdivisions+1)];
    const Point<SPACEDIM> &ep3=ep[n_subdivisions*(n_subdivisions+2)];
    const Point<SPACEDIM> &ep4=
      ep[n_subdivisions*(n_subdivisions+1)*(n_subdivisions+1)];
    const Point<SPACEDIM> &ep5=ep[n_subdivisions
      +n_subdivisions*(n_subdivisions+1)*(n_subdivisions+1)];
    const Point<SPACEDIM> &ep6=
      ep[n_subdivisions*(n_subdivisions+1)*(n_subdivisions+2)];
    const Point<SPACEDIM> &ep7=
      ep[n_subdivisions*(n_subdivisions*(n_subdivisions+3)+3)];
    Vector3 v0; for (int d=0;d<3;d++) v0[d]=ep0[d];
    Vector3 v1; for (int d=0;d<3;d++) v1[d]=ep1[d];
    Vector3 v2; for (int d=0;d<3;d++) v2[d]=ep2[d];
    Vector3 v3; for (int d=0;d<3;d++) v3[d]=ep3[d];
    Vector3 v4; for (int d=0;d<3;d++) v4[d]=ep4[d];
    Vector3 v5; for (int d=0;d<3;d++) v5[d]=ep5[d];
    Vector3 v6; for (int d=0;d<3;d++) v6[d]=ep6[d];
    Vector3 v7; for (int d=0;d<3;d++) v7[d]=ep7[d];
    gt.drawLine(v0,v2);
    gt.drawLine(v1,v3);
    gt.drawLine(v0,v1);
    gt.drawLine(v2,v3);
    gt.drawLine(v4,v6);
    gt.drawLine(v5,v7);
    gt.drawLine(v4,v5);
    gt.drawLine(v6,v7);
    gt.drawLine(v0,v4);
    gt.drawLine(v1,v5);
    gt.drawLine(v2,v6);
    gt.drawLine(v3,v7);
  } else if (dynamic_cast<const Prism*>(shape)) {
    const Point<SPACEDIM> &ep0=ep[0];
    const Point<SPACEDIM> &ep1=ep[n_subdivisions];
    const Point<SPACEDIM> &ep2=ep[(n_subdivisions*(n_subdivisions+3))/2];
    const Point<SPACEDIM> &ep3=
      ep[n_subdivisions*((n_subdivisions+1)*(n_subdivisions+2))/2];
    const Point<SPACEDIM> &ep4=ep[n_subdivisions
      +n_subdivisions*((n_subdivisions+1)*(n_subdivisions+2))/2];
    const Point<SPACEDIM> &ep5=ep[n_subdivisions*(2*n_subdivisions
      +(n_subdivisions*(n_subdivisions+5))/2)];
    Vector3 v0; for (int d=0;d<3;d++) v0[d]=ep0[d];
    Vector3 v1; for (int d=0;d<3;d++) v1[d]=ep1[d];
    Vector3 v2; for (int d=0;d<3;d++) v2[d]=ep2[d];
    Vector3 v3; for (int d=0;d<3;d++) v3[d]=ep3[d];
    Vector3 v4; for (int d=0;d<3;d++) v4[d]=ep4[d];
    Vector3 v5; for (int d=0;d<3;d++) v5[d]=ep5[d];
    gt.drawLine(v0,v1);
    gt.drawLine(v1,v2);
    gt.drawLine(v2,v0);
    gt.drawLine(v3,v4);
    gt.drawLine(v4,v5);
    gt.drawLine(v5,v3);
    gt.drawLine(v0,v3);
    gt.drawLine(v1,v4);
    gt.drawLine(v2,v5);
  } else {
    const Point<SPACEDIM> &ep0=ep[0];
    const Point<SPACEDIM> &ep1=ep[n_subdivisions];
    const Point<SPACEDIM> &ep2=ep[(n_subdivisions*(n_subdivisions+3))/2];
    const Point<SPACEDIM> &ep3=
      ep[(n_subdivisions*(n_subdivisions*(n_subdivisions+6)+11))/6];
    Vector3 v0; for (int d=0;d<3;d++) v0[d]=ep0[d];
    Vector3 v1; for (int d=0;d<3;d++) v1[d]=ep1[d];
    Vector3 v2; for (int d=0;d<3;d++) v2[d]=ep2[d];
    Vector3 v3; for (int d=0;d<3;d++) v3[d]=ep3[d];
    gt.drawLine(v0,v1);
    gt.drawLine(v1,v2);
    gt.drawLine(v2,v0);
    gt.drawLine(v0,v3);
    gt.drawLine(v1,v3);
    gt.drawLine(v2,v3);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Patch::printOn(ostream &os) const {
  os << "Patch: patch_index = " << patch_index
     << "\n\tn_subdivisions = " << n_subdivisions << endl;
//for (int i=0;i<vertices.getNumber();i++) {
//  os << "\tvertices[" << i << "] = " << vertices[i] << endl;
//}
//for (int i=0;i<neighbors.getNumber();i++) {
//  os << "\tneighbors[" << i << "] = " << neighbors[i] << endl;
//}
  os << "\tdata:" << endl;
  for (int i=0;i<data.size(0);i++) {
    for (int j=0;j<data.size(1);j++) {
      os << "\tdata[" << i << "," << j << "] = " << data(i,j) << endl;
    }
  }
}
