//"$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "RefinementStudy.H"
#include "XYGraphTool.H"
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RefinementStudy::recursivelyIntegrateError(
FEValues &finest_fe_values,NumPtr<DofHandler*> &dof_handlers,
const NumPtr<Vector> &solutions,const SHAPE_CLASS *current_element,
const LEVEL_CLASS *current_level) {
  int current_level_number=current_level->numberCoarserLevels();
  if (current_level->hasChildren(current_element)) {
    elements[current_level_number]=current_element;
    int index=current_element->elementDataIndex();
    const LEVEL_CLASS *finer_level=current_level->finerLevel();
    for (int child=0;child<childrenPerShape();child++) {
      child_number[current_level_number]=child;
      int child_index=current_level->childIndex(index,child);
      const SHAPE_CLASS *child_element=
        finer_level->getShape(child_index);
      recursivelyIntegrateError(finest_fe_values,dof_handlers,solutions,
        child_element,finer_level);
    }
  } else {
    const Mapping &mapping=finest_fe_values.getMapping();
//  evaluate shape functions at fine quadrature points
    DofHandler &dof_handler=*dof_handlers[current_level_number];
    Tessellation &tess=dof_handler.getTessellation();
    finest_fe_values.mapShapes(tess,current_element,current_level);

//  these are mapped points:
    const NumPtr<Point<SPACEDIM> > &finest_user_points=
      finest_fe_values.getQuadraturePoints(current_element);
    NumPtr<Point<SPACEDIM> > coarser_points(finest_user_points.getNumber());

    const NumPtr<REAL> &jxw=
      finest_fe_values.getJxWValues(current_element);
//  evaluate sum of fine shape functions times fine solution values
//    at each fine quadrature point
    int n_q_points=jxw.getNumber();
    NumPtr<REAL> finest_values;
    NumPtr<Tensor<1,SPACEDIM> > finest_grads;
    if (update_flags & UPDATE_VALUES) {
      dof_handler.getFunctionValues(finest_fe_values,current_element,
        current_level,solutions[current_level_number],finest_values);
    }
    if (update_flags & UPDATE_GRADIENTS) {
      dof_handler.getFunctionDerivatives(finest_fe_values,current_element,
        current_level,solutions[current_level_number],finest_grads);
    }
    
    int coarser_level_number=current_level_number;
    for (const LEVEL_CLASS *coarser_level=current_level->coarserLevel();
    coarser_level;coarser_level=coarser_level->coarserLevel()) {
      coarser_level_number--;
      const SHAPE_CLASS *coarser_element=elements[coarser_level_number];

      NumPtr<Point<SPACEDIM> > coarse_canonical_lattice_points;
      equallySpacedLatticePoints(coarser_element->numberVertices(),
        mapping.getDegree(),coarse_canonical_lattice_points);
      Quadrature<SPACEDIM> coarse_quadrature(
        coarse_canonical_lattice_points);
      MappingDataBase *md=mapping.getMappingData(
        coarser_element->numberVertices(),
        UPDATE_TRANSFORMATION_VALUES | UPDATE_TRANSFORMATION_GRADIENTS,
        &coarse_quadrature);
      md->computeMappingSupportPoints(coarser_element,tess,coarser_level);
      NumPtr<Point<SPACEDIM> > coarse_mapped_lattice_points(
        coarse_canonical_lattice_points.getNumber());
      md->elementPoints(coarse_mapped_lattice_points);

      for (int i=0;i<finest_user_points.getNumber();i++) {
        REAL min_dist=HUGE_VAL;
        int min_j=-1;
        const Point<SPACEDIM> &point=finest_user_points[i];
        for (int j=0;j<coarse_mapped_lattice_points.getNumber();j++) {
          REAL dist=coarse_mapped_lattice_points[j].distance(point);
          if (dist<min_dist) {
            min_dist=dist;
            min_j=j;
          }
        }
        coarser_points[i]=coarse_mapped_lattice_points[min_j];
        mapping.canonicalPoint(coarser_element,tess,md,
          finest_user_points[i],coarser_points[i]);
      }
      delete md; md=0;
        
      map<int,Quadrature<SPACEDIM> > coarser_qmap;
      Quadrature<SPACEDIM> coarser_quadrature(coarser_points);
      coarser_qmap[coarser_element->numberVertices()]=coarser_quadrature;
      DofHandler &coarser_dof_handler=*dof_handlers[coarser_level_number];
      FEValues coarser_fe_values(mapping,coarser_dof_handler.getFe(),
        coarser_qmap,update_flags);
      coarser_fe_values.mapShapes(tess,coarser_element,coarser_level);

      Vector coarser_dof_values;
      coarser_dof_handler.getDofValues(coarser_element,coarser_level,
        solutions[coarser_level_number],coarser_dof_values);
      NumPtr<REAL> coarser_values;
      NumPtr<Tensor<1,SPACEDIM> > coarser_grads;
      if (update_flags & UPDATE_VALUES) {
        coarser_fe_values.getFunctionValues(coarser_dof_values,
          coarser_element,coarser_values);
      }
      if (update_flags & UPDATE_GRADIENTS) {
        coarser_fe_values.getFunctionDerivatives(coarser_dof_values,
          coarser_element,coarser_grads);
      }
      for (int q=0;q<n_q_points;q++) {
        local_volume[coarser_level_number]+=jxw[q];
      }
      switch (norm_type) {
        case MEAN_NORM:
          for (int q=0;q<n_q_points;q++) {
            local_error[coarser_level_number]+=
              (coarser_values[q]-finest_values[q])*jxw[q];
          }
          break;
        case L1_NORM:
          for (int q=0;q<n_q_points;q++) {
            local_error[coarser_level_number]+=
              abs(coarser_values[q]-finest_values[q])*jxw[q];
          }
          break;
        case L2_NORM:
          for (int q=0;q<n_q_points;q++) {
            REAL dif=coarser_values[q]-finest_values[q];
            local_error[coarser_level_number]+=dif*dif*jxw[q];
          }
          break;
        case H1_SEMINORM:
          for (int q=0;q<n_q_points;q++) {
            Tensor<1,SPACEDIM> dif=coarser_grads[q]-finest_grads[q];
            local_error[coarser_level_number]+=(dif*dif)*jxw[q];
          }
          break;
        case H1_NORM:
          for (int q=0;q<n_q_points;q++) {
            REAL difv=coarser_values[q]-finest_values[q];
            Tensor<1,SPACEDIM> difg=coarser_grads[q]-finest_grads[q];
            local_error[coarser_level_number]+=
              (difv*difv+difg*difg)*jxw[q];
          }
          break;
        case LP_NORM:
          for (int q=0;q<n_q_points;q++) {
            REAL dif=abs(coarser_values[q]-finest_values[q]);
            local_error[coarser_level_number]+=pow(dif,exponent)*jxw[q];
          }
          break;
        case W1P_SEMINORM:
          for (int q=0;q<n_q_points;q++) {
            Tensor<1,SPACEDIM> dif=coarser_grads[q]-finest_grads[q];
            local_error[coarser_level_number]+=
              pow(dif*dif,exponent*0.5)*jxw[q];
          }
          break;
        case W1P_NORM:
          for (int q=0;q<n_q_points;q++) {
            REAL difv=abs(coarser_values[q]-finest_values[q]);
            Tensor<1,SPACEDIM> difg=coarser_grads[q]-finest_grads[q];
            local_error[coarser_level_number]+=
              (pow(difv,exponent)+pow(difg*difg,exponent*0.5))*jxw[q];
          }
          break;
        case LINFTY_NORM:
          for (int q=0;q<n_q_points;q++) {
            local_error[coarser_level_number]=
              max(local_error[coarser_level_number],
                  abs(coarser_values[q]-finest_values[q]));
          }
          break;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RefinementStudy::setUpdateFlags() {
  switch (norm_type) {
    case MEAN_NORM:
    case L1_NORM:
    case L2_NORM:
    case LINFTY_NORM:
    case LP_NORM:
      update_flags=UPDATE_VALUES;
      break;
    case H1_SEMINORM:
    case H1_NORM:
    case W1P_SEMINORM:
    case W1P_NORM:
      update_flags=UPDATE_VALUES | UPDATE_GRADIENTS;
      break;
  }
  if (norm_type!=LINFTY_NORM) update_flags |= UPDATE_JxW_VALUES;
  update_flags |= UPDATE_Q_POINTS;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RefinementStudy::plot(NumPtr<REAL> &total_error,
const NumPtr<Vector> &solutions) const {
  int number_levels=total_error.getNumber();
  REAL log10=log(10.);
  NumPtr<REAL> log_n_dofs(number_levels);
  for (int ln=0;ln<number_levels;ln++) {
    total_error[ln]=log(total_error[ln])/log10;
    log_n_dofs[ln]=
      log(static_cast<REAL>(solutions[ln].size()))/log10;
  }
  REAL ymin=total_error[0];
  REAL ymax=total_error[0];
  for (int ln=1;ln<number_levels;ln++) {
    ymin=min(ymin,total_error[ln]);
    ymax=max(ymax,total_error[ln]);
  }
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool gt("log_10(error) vs log_10(degrees of freedom)",
    "log_10(degrees of freedom)","log_10(error)",
    log_n_dofs[0],log_n_dofs[number_levels-1],ymin,ymax,&cmap,0,0.5);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  gt.movePen(log_n_dofs[0],total_error[0]);
  for (int ln=1;ln<number_levels;ln++) {
    gt.drawLine(log_n_dofs[ln],total_error[ln]);
  }
  REAL dx=(log_n_dofs[number_levels-1]-log_n_dofs[0])/50;
  for (int ln=0;ln<number_levels;ln++) {
    gt.drawDiamondGivenCenter(log_n_dofs[ln],total_error[ln],dx);
  }
  gt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RefinementStudy::RefinementStudy(Mapping &mapping,
const map<int,Quadrature<SPACEDIM> > &qmap,
NumPtr<DofHandler*> &dof_handlers,const NumPtr<Vector> &solutions,
NORM_TYPE nt,REAL ex) : norm_type(nt), exponent(ex) {
  int number_levels=dof_handlers.getNumber();
  Tessellation &tess=dof_handlers[number_levels-1]->getTessellation();
  if (number_levels<=1) return;
  CHECK_TEST(solutions.getNumber()==number_levels);

  setUpdateFlags();
  FEValues fe_values(mapping,dof_handlers[number_levels-1]->getFe(),qmap,
    update_flags);

  local_error.allocate(number_levels-1);
  local_volume.allocate(number_levels-1);
  child_number.allocate(number_levels-1);
  elements.allocate(number_levels);

  NumPtr<REAL> total_error(number_levels-1);
  total_error.initialize(0.);
  NumPtr<REAL> total_volume(number_levels-1);
  total_volume.initialize(0.);
  LEVEL_CLASS *coarsest_level=tess.coarsestLevel();
  for (bool doit=tess.firstShape(coarsest_level);doit;
  doit=tess.nextShapeInLevel()) {
    if (tess.presentLevel()!=coarsest_level) break;
    const SHAPE_CLASS *current_element=coarsest_level->getShape();
    local_error.initialize(0.);
    local_volume.initialize(0.);
    child_number.initialize(-1);
    recursivelyIntegrateError(fe_values,dof_handlers,solutions,
      current_element,coarsest_level);
    if (norm_type==LINFTY_NORM) {
      for (int ln=0;ln<number_levels-1;ln++) {
        total_error[ln]=max(local_error[ln],total_error[ln]);
      }
    } else {
      for (int ln=0;ln<number_levels-1;ln++) {
        total_error[ln]+=local_error[ln];
      }
    }
    for (int ln=0;ln<number_levels-1;ln++) {
      total_volume[ln]+=local_volume[ln];
    }
  }
  switch (norm_type) {
    case MEAN_NORM:
      for (int ln=0;ln<number_levels-1;ln++) {
        total_error[ln]=abs(total_error[ln])/total_volume[ln];
      }
      break;
    case L1_NORM:
      for (int ln=0;ln<number_levels-1;ln++) {
        total_error[ln]=total_error[ln]/total_volume[ln];
      }
      break;
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      for (int ln=0;ln<number_levels-1;ln++) {
        total_error[ln]=sqrt(total_error[ln]/total_volume[ln]);
      }
      break;
    case LP_NORM:
    case W1P_SEMINORM:
    case W1P_NORM:
      for (int ln=0;ln<number_levels-1;ln++) {
        total_error[ln]=pow(total_error[ln]/total_volume[ln],1./exponent);
      }
      break;
    case LINFTY_NORM:
      break;
  }

  plot(total_error,solutions);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RefinementStudy::RefinementStudy(Mapping &mapping,
const map<int,Quadrature<SPACEDIM> > &qmap,
NumPtr<DofHandler*> &dof_handlers,const NumPtr<Vector> &solutions,
const FunctionSystem &exact_solution,NORM_TYPE nt,REAL ex) : 
norm_type(nt),exponent(ex) {
  setUpdateFlags();
  FEValues fe_values(mapping,dof_handlers[0]->getFe(),qmap,
    update_flags);

  Tessellation &tess=dof_handlers[0]->getTessellation();
  int number_levels=tess.numberLevels();

  NumPtr<REAL> total_error(number_levels);
  total_error.initialize(0.);
  NumPtr<REAL> total_volume(number_levels);
  total_volume.initialize(0.);
  int ln=0;
  for (LEVEL_CLASS *level=tess.coarsestLevel();level;
  level=level->finerLevel(),ln++) {
    for (bool doit=tess.firstShape(level);doit;
    doit=tess.nextShapeInLevel()) {
      if (tess.presentLevel()!=level) break;
      const SHAPE_CLASS *element=level->getShape();
      fe_values.mapShapes(tess,element,level);
      const NumPtr<REAL> &jxw=fe_values.getJxWValues(element);
      const NumPtr<Point<SPACEDIM> > &q_points=
        fe_values.getQuadraturePoints(element);
      int n_q_points=q_points.getNumber();
      NumPtr<REAL> values(n_q_points);
      NumPtr<Tensor<1,SPACEDIM> > grads(n_q_points);
      NumPtr<Vector> exact_values(n_q_points);
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > > exact_grads(n_q_points);
      if (update_flags & UPDATE_VALUES) {
        dof_handlers[ln]->getFunctionValues(fe_values,element,level,
          solutions[ln],values);
        exact_solution.value(exact_values,q_points);
      }
      if (update_flags & UPDATE_GRADIENTS) {
        dof_handlers[ln]->getFunctionDerivatives(fe_values,element,level,
          solutions[ln],grads);
        exact_solution.gradient(exact_grads,q_points);
      }
      for (int q=0;q<n_q_points;q++) {
        local_volume[ln]+=jxw[q];
      }
      for (int q=0;q<n_q_points;q++) {
        REAL difv=(update_flags & UPDATE_VALUES
          ? values[q]-exact_values[q][0] : 0.);
        Tensor<1,SPACEDIM> difg;
        if (update_flags & UPDATE_GRADIENTS) {
          difg=grads[q]-exact_grads[q][0];
        }
        switch (norm_type) {
          case MEAN_NORM:
            total_error[ln]+=difv*jxw[q];
            break;
          case L1_NORM:
            total_error[ln]+=abs(difv)*jxw[q];
            break;
          case L2_NORM:
            total_error[ln]+=difv*difv*jxw[q];
            break;
          case H1_SEMINORM:
            total_error[ln]+=(difg*difg)*jxw[q];
            break;
          case H1_NORM:
            total_error[ln]+=(difv*difv+difg*difg)*jxw[q];
            break;
          case LP_NORM:
            total_error[ln]+=pow(abs(difv),exponent)*jxw[q];
            break;
          case W1P_SEMINORM:
            total_error[ln]+=pow(difg*difg,exponent*0.5)*jxw[q];
            break;
          case W1P_NORM:
            total_error[ln]+=(pow(abs(difv),exponent)
                             +pow(difg*difg,exponent*0.5))*jxw[q];
            break;
          case LINFTY_NORM:
            total_error[ln]=max(total_error[ln],abs(difv));
            break;
        }
      }
    }
    switch (norm_type) {
      case MEAN_NORM:
        total_error[ln]=abs(total_error[ln])/total_volume[ln];
        break;
      case L1_NORM:
        total_error[ln]=total_error[ln]/total_volume[ln];
        break;
      case L2_NORM:
      case H1_SEMINORM:
      case H1_NORM:
        total_error[ln]=sqrt(total_error[ln]/total_volume[ln]);
        break;
      case LP_NORM:
      case W1P_SEMINORM:
      case W1P_NORM:
        total_error[ln]=pow(total_error[ln]/total_volume[ln],1./exponent);
        break;
      case LINFTY_NORM:
        break;
    }
  }
  plot(total_error,solutions);
}

#include "NumPtr.C"
INSTANTIATE_NUMPTR(const SHAPE_CLASS*);
INSTANTIATE_NUMPTR(DofHandler*);
