#ifndef TESSELLATION
#define TESSELLATION "$Header:$"
//----------------------------  tria.h  ---------------------------
//   $Id: tria.h,v 1.131 2003/05/28 16:29:36 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  tria.h  ---------------------------
//
//modified from deal.II/deal.II/include/grid/tria.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <BoundaryCondition.H>
#include <Level.H>
#include <NumPtr.H>
#include <Types.H>

//compare to DEAL CellData
class CellData {
  private:
    NumPtr<int> vertices;
    MaterialModel *material_model;
  public:
    CellData() : material_model(0) { }
    CellData(int nv) : vertices(nv),material_model(0) { }
    ~CellData() { }
    void allocate(int nv) { 
      if (vertices.getNumber()==0) vertices.allocate(nv);
    }
    int numberVertices() const { return vertices.getNumber(); }
    const int& vertex(int i) const { return vertices[i]; }
    int& vertex(int i) { return vertices[i]; }
    MaterialModel* getMaterialModel() const { return material_model; }
    void setMaterialModel(MaterialModel *mm) { material_model=mm; }
    void printOn(ostream &os) const {
      os << "CellData: material_model = " << material_model
         << " vertices =";
      for (int i=0;i<vertices.getNumber();i++) {
        os << " " << vertices[i];
      }
      os << endl;
    }
};

//compare to DEAL SubCellData
class FaceData {
  private:
    NumPtr<int> vertices;
    const BoundaryCondition *boundary_condition;
  public:
    FaceData() : boundary_condition(0) { }
    FaceData(int nv) : vertices(nv),boundary_condition(0) { }
    ~FaceData() { }
    void allocate(int nv) { 
      if (vertices.getNumber()==0) vertices.allocate(nv);
    }
    int numberVertices() const { return vertices.getNumber(); }
    const int& vertex(int i) const { return vertices[i]; }
    int& vertex(int i) { return vertices[i]; }
    const BoundaryCondition* getBoundaryCondition() const { 
      return boundary_condition; 
    }
    void setBoundaryCondition(const BoundaryCondition *bc) { 
      boundary_condition=bc;
    }
    void printOn(ostream &os) const {
      os << "FaceData: boundary_condition = " << boundary_condition
         << " vertices =";
      for (int i=0;i<vertices.getNumber();i++) {
        os << " " << vertices[i];
      }
      os << endl;
    }
};

#if (SPACEDIM==3)
//compare to DEAL SubCellData
class LineData {
  private:
    NumPtr<int> vertices;
    pair<const BoundaryCondition*,const BoundaryCondition*> 
      boundary_condition;
  public:
    LineData() : vertices(2) { 
      boundary_condition.first=0; 
      boundary_condition.second=0; 
    }
    ~LineData() { }
    int numberVertices() const { return vertices.getNumber(); }
    const int& vertex(int i) const { return vertices[i]; }
    int& vertex(int i) { return vertices[i]; }
    pair<const BoundaryCondition*,const BoundaryCondition*> 
    getBoundaryConditions() const { 
      return boundary_condition; 
    }
    void setBoundaryConditions(
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs) { 
      boundary_condition.first=bcs.first;
      boundary_condition.second=bcs.second;
    }
    void setBoundaryCondition(const BoundaryCondition *bc) { 
      if (bc==0) { 
        boundary_condition.first=0; 
        boundary_condition.second=0; 
      } else if (boundary_condition.first==0) boundary_condition.first=bc;
      else if (boundary_condition.first!=bc) boundary_condition.second=bc;
    }
    void printOn(ostream &os) const {
      os << "LineData: boundary_condition = <" << boundary_condition.first
         << " , " << boundary_condition.second << "> vertices =";
      for (int i=0;i<vertices.getNumber();i++) {
        os << " " << vertices[i];
      }
      os << endl;
    }
};
#endif

struct BoundaryData {
  NumPtr<FaceData> boundary_faces;
#if (SPACEDIM==3)
  NumPtr<LineData> boundary_lines;
#endif
  void printOn(ostream&) const;
};

//compare to DEAL Triangulation
class Tessellation {
  public:
    enum MeshSmoothing {
      mesh_smoothing_none                               = 0x0,
      mesh_smoothing_limit_level_difference_at_vertices = 0x1,     
      mesh_smoothing_eliminate_unrefined_islands        = 0x2,
      mesh_smoothing_patch_level_1                      = 0x4,
      
      mesh_smoothing_eliminate_refined_inner_islands    = 0x100,
      mesh_smoothing_eliminate_refined_boundary_islands = 0x200,
      mesh_smoothing_do_not_produce_unrefined_islands   = 0x400,
      
      mesh_smoothing_maximum                            = 0xffff
    };
  private:
    LEVEL_CLASS *coarsest_level;
    NumPtr<Point<SPACEDIM> > vertices;
    NumPtr<bool> vertices_used;
    MeshSmoothing mesh_smoothing;
    bool owns_boundary;
    bool levels_own_element_data_shapes;

    LEVEL_CLASS *present_level;
    CODIMENSION1_LEVEL_CLASS *present_codimension1_level;
#if (SPACEDIM==3)
    Level1 *present_codimension2_level;
#endif

    Tessellation(const Tessellation &t) {
      OBSOLETE("copy constructor disallowed");
    }
    void fixCoarsenFlags();
#if (SPACEDIM==3)
    void prepareRefinement();
#endif
    void deleteChildren();
    void executeCoarsening();
    void executeRefinement();

  public:
    Tessellation(MeshSmoothing ms=mesh_smoothing_none) :
    coarsest_level(0),mesh_smoothing(ms),owns_boundary(true),
    levels_own_element_data_shapes(true) {
    }
    virtual ~Tessellation();
    LEVEL_CLASS* coarsestLevel() const { return coarsest_level; }
    LEVEL_CLASS* presentLevel() const { return present_level; }
    CODIMENSION1_LEVEL_CLASS* presentCodimension1Level() { 
      return present_codimension1_level; 
    }
#if (SPACEDIM==3)
    Level1* presentCodimension2Level() { 
      return present_codimension2_level; 
    }
#endif
    void clear() {
      delete coarsest_level;
      vertices.cleanup();
      vertices_used.cleanup();
    }
    virtual void copyTessellation(const Tessellation &old_tess,
    LEVEL_CLASS *cl) {
      CHECK_TEST(vertices.getNumber()==0);
      CHECK_TEST(coarsest_level==0);
      CHECK_POSITIVE(old_tess.vertices.getNumber());
      vertices.allocate(old_tess.vertices.getNumber());
      vertices.copyFrom(old_tess.vertices);
      vertices_used.allocate(old_tess.vertices_used.getNumber());
      vertices_used.copyFrom(old_tess.vertices_used);
      mesh_smoothing= old_tess.mesh_smoothing;
      owns_boundary=false;
      levels_own_element_data_shapes=false;
      coarsest_level=cl;
    }
    virtual BoundaryData* create(const NumPtr<Point<SPACEDIM> >&,
      const NumPtr<CellData>&);
    void synchronize(BoundaryData*);
    void setMaterialModel(MaterialModel*);
    const SHAPE_CLASS*  presentShape() const {
      CHECK_POINTER(present_level);
      return present_level->getShape();
    }
    int presentShapeVertexIndex(int i) const {
      CHECK_POINTER(present_level);
      return present_level->vertexIndex(i);
    }
#if (SPACEDIM>1)
    const CODIMENSION1_SHAPE_CLASS* presentCodimension1Shape() const {
      CHECK_POINTER(present_codimension1_level);
      return present_codimension1_level->getShape();
    }
    const CODIMENSION1_SHAPE_CLASS* presentCodimension1Shape(
    int face_index) const {
      CHECK_POINTER(present_codimension1_level);
      return present_codimension1_level->getShape(face_index);
    }
#endif
#if (SPACEDIM==3)
    const Shape1* presentCodimension2Shape() const {
      CHECK_POINTER(present_codimension2_level);
      return present_codimension2_level->getShape();
    }
    const Shape1* presentCodimension2Shape(int line_index) const {
      CHECK_POINTER(present_codimension2_level);
      return present_codimension2_level->getShape(line_index);
    }
#endif
    bool iteratorStateValid() { 
      if (present_level==0) return false;
      return present_level->iteratorStateValid();
    }
    bool iteratorStatePastTheEnd() { 
      if (iteratorStateValid()) return false;
      if (present_level==0) return false;
      return present_level->iteratorStatePastTheEnd();
    }
    bool iteratorStateInvalid() { return !iteratorStateInvalid(); }
    void setCodimension1Level() {
      present_codimension1_level=
        static_cast<CODIMENSION1_LEVEL_CLASS*>(present_level);
    }
    bool firstShape(const LEVEL_CLASS *level=0,int index=0) {
      present_level=
        (level==0 ? coarsest_level : const_cast<LEVEL_CLASS*>(level));
      return present_level->first(index);
    }
    bool lastShape(const LEVEL_CLASS *level=0,int index=-1) {
      present_level=(level==0 ? coarsest_level->finestLevel() : 
                                const_cast<LEVEL_CLASS*>(level));
      return present_level->last(index);
    }
    bool nextShape() {
      if (present_level==0) return false;
      if (present_level->next()) return true;
      present_level=present_level->finerLevel();
      if (present_level==0) return false;
      present_level->first();
      return (present_level->numberShapes()>0);
    }
    bool nextShapeInLevel() {
      if (present_level==0) return false;
      if (present_level->next()) return true;
      return false;
    }
    bool prevShape() {
      if (present_level==0) return false;
      if (present_level->prev()) return true;
      present_level=present_level->coarserLevel();
      if (present_level==0) return false;
      present_level->last();
      return (present_level->numberShapes()>0);
    }
    bool prevShapeInLevel() {
      if (present_level==0) return false;
      if (present_level->prev()) return true;
      return false;
    }
#if (SPACEDIM>1)
    bool firstCodimension1Shape(
    const CODIMENSION1_LEVEL_CLASS *level=0,int index=0) {
      present_codimension1_level=(level==0 ?
        static_cast<CODIMENSION1_LEVEL_CLASS*>(coarsest_level) : 
        const_cast<CODIMENSION1_LEVEL_CLASS*>(level));
      return present_codimension1_level->first(index);
    }
    bool lastCodimension1Shape(
    const CODIMENSION1_LEVEL_CLASS *level=0,int index=-1) {
      present_codimension1_level=(level==0 ?
        static_cast<CODIMENSION1_LEVEL_CLASS*>(coarsest_level) : 
        const_cast<CODIMENSION1_LEVEL_CLASS*>(level));
      return present_codimension1_level->last(index);
    }
    bool nextCodimension1Shape() {
      if (present_codimension1_level==0) return false;
      if (present_codimension1_level->next()) return true;
      present_codimension1_level=
        present_codimension1_level->finerLevel();
      if (present_codimension1_level==0) return false;
      present_codimension1_level->first();
      return (present_codimension1_level->numberShapes()>0);
    }
    bool nextCodimension1ShapeInLevel() {
      if (present_codimension1_level==0) return false;
      if (present_codimension1_level->next()) return true;
      return false;
    }
    bool prevCodimension1Shape() {
      if (present_codimension1_level==0) return false;
      if (present_codimension1_level->prev()) return true;
      present_codimension1_level=
        present_codimension1_level->coarserLevel();
      if (present_codimension1_level==0) return false;
      present_codimension1_level->last();
      return (present_codimension1_level->numberShapes()>0);
    }
    bool prevCodimension1ShapeInLevel() {
      if (present_codimension1_level==0) return false;
      if (present_codimension1_level->prev()) return true;
      return false;
    }
#endif
#if (SPACEDIM==3)
    bool firstCodimension2Shape(Level1 *level=0,int index=0) {
      present_codimension2_level=(level==0 ?
        static_cast<Level1*>(coarsest_level) : level);
      return present_codimension2_level->first(index);
    }
    bool lastCodimension2Shape(Level1 *level=0,int index=-1) {
      present_codimension2_level=(level==0 ?
        static_cast<Level1*>(coarsest_level) : level);
      return present_codimension2_level->last(index);
    }
    bool nextCodimension2Shape() {
      if (present_codimension2_level==0) return false;
      if (present_codimension2_level->next()) return true;
      present_codimension2_level=
        present_codimension2_level->finerLevel();
      if (present_codimension2_level==0) return false;
      present_codimension2_level->first();
      return (present_codimension2_level->numberShapes()>0);
    }
    bool nextCodimension2ShapeInLevel() {
      if (present_codimension2_level==0) return false;
      if (present_codimension2_level->next()) return true;
      return false;
    }
    bool prevCodimension2Shape() {
      if (present_codimension2_level==0) return false;
      if (present_codimension2_level->prev()) return true;
      present_codimension2_level=
        present_codimension2_level->coarserLevel();
      if (present_codimension2_level==0) return false;
      present_codimension2_level->last();
      return (present_codimension2_level->numberShapes()>0);
    }
    bool prevCodimension2ShapeInLevel() {
      if (present_codimension2_level==0) return false;
      if (present_codimension2_level->prev()) return true;
      return false;
    }
#endif
    bool firstUsedShape(const LEVEL_CLASS *level=0,int index=0) {
      const LEVEL_CLASS *l=(level==0 ? coarsest_level : level);
      if (!firstShape(l,index)) return false;
      CHECK_POINTER(present_level);
      bool test=true;
      while (!present_level->used()) test=nextShape();
      return test;
    }
    bool lastUsedShape(const LEVEL_CLASS *level=0,int index=-1) {
      const LEVEL_CLASS *l=
        (level==0 ? coarsest_level->finestLevel() : level);
      if (!lastShape(l,index)) return false;
      CHECK_POINTER(present_level);
      bool test=true;
      while (!present_level->used()) test=prevShape();
      return test;
    }
    bool nextUsedShape() {
      if (present_level==0) return false;
      while (nextShape()) if (present_level->used()) return true;
      return false;
    }
    bool nextUsedShapeInLevel() {
      if (present_level==0) return false;
      while (nextShapeInLevel()) {
        if (present_level->used()) return true;
      }
      return false;
    }
    bool prevUsedShape() {
      if (present_level==0) return false;
      while(prevShape()) if (present_level->used()) return true;
      return false;
    }
    bool prevUsedShapeInLevel() {
      if (present_level==0) return false;
      while(prevShapeInLevel()) {
        if (present_level->used()) return true;
      }
      return false;
    }
#if (SPACEDIM>1)
    bool firstCodimension1UsedShape(
    const CODIMENSION1_LEVEL_CLASS *level=0,int index=0) {
      if (level==0) level=
        static_cast<CODIMENSION1_LEVEL_CLASS*>(coarsest_level);
      if (!firstCodimension1Shape(level,index)) return false;
      CHECK_POINTER(present_codimension1_level);
      bool test=true;
      while (!present_codimension1_level->used()) {
        test=nextCodimension1Shape();
      }
      return test;
    }
    bool lastCodimension1UsedShape(
    const CODIMENSION1_LEVEL_CLASS *level=0,int index=-1) {
      if (level==0) level=static_cast<CODIMENSION1_LEVEL_CLASS*>(
        coarsest_level->finestLevel());
      if (!lastCodimension1Shape(level,index)) return false;
      CHECK_POINTER(present_codimension1_level);
      bool test=true;
      while (!present_codimension1_level->used()) {
        test=prevCodimension1Shape();
      }
      return test;
    }
    bool nextCodimension1UsedShape() {
      if (present_codimension1_level==0) return false;
      while (nextCodimension1Shape()) {
        if (present_codimension1_level->used()) return true;
      }
      return false;
    }
    bool nextCodimension1UsedShapeInLevel() {
      if (present_codimension1_level==0) return false;
      while (nextCodimension1ShapeInLevel()) {
        if (present_codimension1_level->used()) return true;
      }
      return false;
    }
    bool prevCodimension1UsedShape() {
      if (present_codimension1_level==0) return false;
      while(prevCodimension1Shape()) {
        if (present_codimension1_level->used()) return true;
      }
      return false;
    }
    bool prevCodimension1UsedShapeInLevel() {
      if (present_codimension1_level==0) return false;
      while(prevCodimension1ShapeInLevel()) {
        if (present_codimension1_level->used()) return true;
      }
      return false;
    }
#endif
#if (SPACEDIM==3)
    bool firstCodimension2UsedShape(Level1 *level=0,int index=0) {
      if (level==0) level=
        static_cast<Level1*>(coarsest_level);
      if (!firstCodimension2Shape(level,index)) return false;
      CHECK_POINTER(present_codimension2_level);
      bool test=true;
      while (!present_codimension2_level->used()) {
        test=nextCodimension2Shape();
      }
      return test;
    }
    bool lastCodimension2UsedShape(Level1 *level=0,int index=-1) {
      if (level==0) level=static_cast<Level1*>(
        coarsest_level->finestLevel());
      if (!lastCodimension2Shape(level,index)) return false;
      CHECK_POINTER(present_codimension2_level);
      bool test=true;
      while (!present_codimension2_level->used()) {
        test=prevCodimension2Shape();
      }
      return test;
    }
    bool nextCodimension2UsedShape() {
      if (present_codimension2_level==0) return false;
      while (nextCodimension2Shape()) {
        if (present_codimension2_level->used()) return true;
      }
      return false;
    }
    bool nextCodimension2UsedShapeInLevel() {
      if (present_codimension2_level==0) return false;
      while (nextCodimension2ShapeInLevel()) {
        if (present_codimension2_level->used()) return true;
      }
      return false;
    }
    bool prevCodimension2UsedShape() {
      if (present_codimension2_level==0) return false;
      while(prevCodimension2Shape()) {
        if (present_codimension2_level->used()) return true;
      }
      return false;
    }
    bool prevCodimension2UsedShapeInLevel() {
      if (present_codimension2_level==0) return false;
      while(prevCodimension2ShapeInLevel()) {
        if (present_codimension2_level->used()) return true;
      }
      return false;
    }
#endif
    bool firstActiveShape(const LEVEL_CLASS *level=0,int index=0) {
      const LEVEL_CLASS *l=(level==0 ? coarsest_level : level);
      if (!firstUsedShape(l,index)) return false;
      CHECK_POINTER(present_level);
      bool test=true;
      while (present_level!=0 && test) {
        if (!present_level->hasChildren()) return true;
        test=nextUsedShape();
      }
      return false;
    }
    bool lastActiveShape(const LEVEL_CLASS *level=0,int index=-1) {
      const LEVEL_CLASS *l=
        (level==0 ? coarsest_level->finestLevel() : level);
      if (!lastUsedShape(l,index)) return false;
      CHECK_POINTER(present_level);
      bool test=true;
      while (present_level!=0 && test) {
        if (!present_level->hasChildren()) return true;
        test=prevUsedShape();
      }
      return false;
    }
    bool nextActiveShape() {
      if (present_level==0) return false;
      while (nextUsedShape()) {
        if (present_level==0) return false;
        if (!present_level->hasChildren()) return true;
      }
      return false;
    }
    bool nextActiveShapeInLevel() {
      if (present_level==0) return false;
      while (nextUsedShapeInLevel()) {
        if (present_level==0) return false;
        if (!present_level->hasChildren()) return true;
      }
      return false;
    }
    bool prevActiveShape() {
      if (present_level==0) return false;
      while (prevUsedShape()) {
        if (present_level==0) return false;
        if (!present_level->hasChildren()) return true;
      }
      return false;
    }
    bool prevActiveShapeInLevel() {
      if (present_level==0) return false;
      while (prevUsedShapeInLevel()) {
        if (present_level==0) return false;
        if (!present_level->hasChildren()) return true;
      }
      return false;
    }
#if (SPACEDIM>1)
    bool firstCodimension1ActiveShape(
    const CODIMENSION1_LEVEL_CLASS *level=0,int index=0) {
      if (level==0) level=
        static_cast<CODIMENSION1_LEVEL_CLASS*>(coarsest_level);
      if (!firstCodimension1UsedShape(level,index)) return false;
      CHECK_POINTER(present_codimension1_level);
      bool test=true;
      while (present_codimension1_level!=0 && test) {
        if (!present_codimension1_level->hasChildren()) return true;
        test=nextCodimension1UsedShape();
      }
      return false;
    }
    bool lastCodimension1ActiveShape(
    const CODIMENSION1_LEVEL_CLASS *level=0,int index=-1) {
      if (level==0) level=static_cast<CODIMENSION1_LEVEL_CLASS*>(
        coarsest_level->finestLevel());
      if (!lastCodimension1UsedShape(level,index)) return false;
      CHECK_POINTER(present_codimension1_level);
      bool test=true;
      while (present_codimension1_level!=0 && test) {
        if (!present_codimension1_level->hasChildren()) return true;
        test=prevCodimension1UsedShape();
      }
      return false;
    }
    bool nextCodimension1ActiveShape() {
      if (present_codimension1_level==0) return false;
      while (nextCodimension1UsedShape()) {
        if (present_codimension1_level==0) return false;
        if (!present_codimension1_level->hasChildren()) return true;
      }
      return false;
    }
    bool nextCodimension1ActiveShapeInLevel() {
      if (present_codimension1_level==0) return false;
      while (nextCodimension1UsedShapeInLevel()) {
        if (present_codimension1_level==0) return false;
        if (!present_codimension1_level->hasChildren()) return true;
      }
      return false;
    }
    bool prevCodimension1ActiveShape() {
      if (present_codimension1_level==0) return false;
      while (prevCodimension1UsedShape()) {
        if (present_codimension1_level==0) return false;
        if (!present_codimension1_level->hasChildren()) return true;
      }
      return false;
    }
    bool prevCodimension1ActiveShapeInLevel() {
      if (present_codimension1_level==0) return false;
      while (prevCodimension1UsedShapeInLevel()) {
        if (present_codimension1_level==0) return false;
        if (!present_codimension1_level->hasChildren()) return true;
      }
      return false;
    }
#endif
#if (SPACEDIM==3)
    bool firstCodimension2ActiveShape(Level1 *level=0,int index=0) {
      if (level==0) level=
        static_cast<Level1*>(coarsest_level);
      if (!firstCodimension2UsedShape(level,index)) return false;
      CHECK_POINTER(present_codimension2_level);
      bool test=true;
      while (present_codimension2_level!=0 && test) {
        if (!present_codimension2_level->hasChildren()) return true;
        test=nextCodimension2UsedShape();
      }
      return false;
    }
    bool lastCodimension2ActiveShape(Level1 *level=0,int index=-1) {
      if (level==0) level=static_cast<Level1*>(
        coarsest_level->finestLevel());
      if (!lastCodimension2UsedShape(level,index)) return false;
      CHECK_POINTER(present_codimension2_level);
      bool test=true;
      while (present_codimension2_level!=0 && test) {
        if (!present_codimension2_level->hasChildren()) return true;
        test=prevCodimension2UsedShape();
      }
      return false;
    }
    bool nextCodimension2ActiveShape() {
      if (present_codimension2_level==0) return false;
      while (nextCodimension2UsedShape()) {
        if (present_codimension2_level==0) return false;
        if (!present_codimension2_level->hasChildren()) return true;
      }
      return false;
    }
    bool nextCodimension2ActiveShapeInLevel() {
      if (present_codimension2_level==0) return false;
      while (nextCodimension2UsedShapeInLevel()) {
        if (present_codimension2_level==0) return false;
        if (!present_codimension2_level->hasChildren()) return true;
      }
      return false;
    }
    bool prevCodimension2ActiveShape() {
      if (present_codimension2_level==0) return false;
      while (prevCodimension2UsedShape()) {
        if (present_codimension2_level==0) return false;
        if (!present_codimension2_level->hasChildren()) return true;
      }
      return false;
    }
    bool prevCodimension2ActiveShapeInLevel() {
      if (present_codimension2_level==0) return false;
      while (prevCodimension2UsedShapeInLevel()) {
        if (present_codimension2_level==0) return false;
        if (!present_codimension2_level->hasChildren()) return true;
      }
      return false;
    }
#endif
    void printOn(ostream&) const;
#ifdef DEBUG
    void printVertices() const;
    void printUsedVertices() const;
    void printUnusedVertices() const;
    void printElements(LEVEL_CLASS*);
    void printElements();
    void printActiveElements(LEVEL_CLASS*);
    void printActiveElements();
# if (SPACEDIM>1)
    void printCodimension1Elements(CODIMENSION1_LEVEL_CLASS*);
    void printCodimension1Elements();
    void printActiveCodimension1Elements(CODIMENSION1_LEVEL_CLASS*);
    void printActiveCodimension1Elements();
# endif
# if (SPACEDIM==3)
    void printCodimension2Elements(Level1*);
    void printCodimension2Elements();
    void printActiveCodimension2Elements(Level1*);
    void printActiveCodimension2Elements();
# endif
    void printNeighbors(LEVEL_CLASS*);
    void printNeighbors();
    void checkNeighbors(LEVEL_CLASS*);
    void checkNeighbors();
# if (SPACEDIM>1)
    void checkFaceVertices(LEVEL_CLASS*);
    void checkFaceVertices();
    void checkFaceBoundaryCondition(NumPtr<BoundaryCondition*>&);
#if (SPACEDIM==3)
    void checkLineBoundaryCondition(NumPtr<BoundaryCondition*>&);
#endif
    void checkNeighborLevelAndIndex(LEVEL_CLASS*);
    void checkNeighborLevelAndIndex();
# endif
    void checkElementData() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->checkElementData();
      }
    }
    void printCoarsenFlags(const LEVEL_CLASS *level) const {
      level->printCoarsenFlags();
    }
    void printCoarsenFlags() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->printCoarsenFlags();
      }
    }
    void printRefineFlags(const LEVEL_CLASS *level) const {
      level->printRefineFlags();
    }
    void printRefineFlags() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->printRefineFlags();
      }
    }
    void plotShapes(LEVEL_CLASS *level) const {
      level->plotShapes(vertices);
    }
    void plotShapesIndividually(LEVEL_CLASS *level) const {
      level->plotShapesIndividually(vertices);
    }
    void plotShapes() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->plotShapes(vertices);
      }
    }
    void plotShapesIndividually() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->plotShapesIndividually(vertices);
      }
    }
    void plotUsedShapes(LEVEL_CLASS *level) const {
      level->plotUsedShapes(vertices);
    }
    void plotUsedShapesIndividually(LEVEL_CLASS *level) const {
      level->plotUsedShapesIndividually(vertices);
    }
    void plotUsedShapes() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->plotUsedShapes(vertices);
      }
    }
    void plotUsedShapesIndividually() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->plotUsedShapesIndividually(vertices);
      }
    }
    void plotActiveShapes(LEVEL_CLASS *level) const {
      level->plotActiveShapes(vertices);
    }
    void plotActiveShapesIndividually(LEVEL_CLASS *level) const {
      level->plotActiveShapesIndividually(vertices);
    }
    void plotActiveShapes() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->plotActiveShapes(vertices);
      }
    }
    void plotActiveShapesIndividually() const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->plotActiveShapesIndividually(vertices);
      }
    }
#if (SPACEDIM==3)
    void plotActiveShapes(LEVEL_CLASS *level,VolGraphTool &gt) const {
      level->plotActiveShapes(vertices,gt);
    }
    void plotActiveShapes(VolGraphTool &gt) const {
      for (LEVEL_CLASS *level=coarsest_level;level;
      level=level->finerLevel()) {
        level->plotActiveShapes(vertices,gt);
      }
    }
#endif
#endif
    int numberActiveShapes(const LEVEL_CLASS*);
    int numberActiveShapes();
#if (SPACEDIM>1)
    int numberActiveCodimension1Shapes(
      const CODIMENSION1_LEVEL_CLASS*);
    int numberActiveCodimension1Shapes();
#endif
    void saveCoarsenFlags(NumPtr<bool>&);
    void saveRefineFlags(NumPtr<bool>&);
    void clearUserFlags();
    bool prepareCoarseningAndRefinement ();

#if (SPACEDIM==3)
    void lenoirSupportPointsOnTetrahedron(const Shape3*,int,
      const Level3*,NumPtr<Point<3> >&) const;
    void lenoirSupportPointsOnPrism(const Shape3*,int,const Level3*,
      NumPtr<Point<3> >&) const;
#endif
#if (SPACEDIM>1)
    void mappingSupportPointsOnBoundaryOf(const SHAPE_CLASS*,int,
      const LEVEL_CLASS*,NumPtr<Point<SPACEDIM> >&) const;
    void getMappedMeasureOn(const CODIMENSION1_SHAPE_CLASS&,
      const CODIMENSION1_LEVEL_CLASS*,
      const NumPtr<Point<SPACEDIM-1> >&,NumPtr<Point<SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&,NumPtr<REAL>&) const;
#endif
#if (SPACEDIM==3)
    void getMappedPointsOnLine(int,const Level1*,
      const NumPtr<Point<1> >&,NumPtr<Point<3> >&,
      NumPtr<Point<SPACEDIM> >&) const;
#endif
    void refineGlobal(int);
    void executeCoarseningAndRefinement();
    void refine(const Vector&,REAL);
    void coarsen(const Vector&,REAL);
    void refineAndCoarsenFixedNumber(const Vector&,REAL,REAL);
    void refineAndCoarsenFixedFraction(const Vector&,REAL,REAL);
    void refineOptimize(const Vector&);
    void saveUserFlagsForShapes(NumPtr<bool>&);
#if (SPACEDIM>1)
    void saveUserFlagsForCodimension1Shapes(NumPtr<bool>&);
#endif
#if (SPACEDIM==3)
    void saveUserFlagsForCodimension2Shapes(NumPtr<bool>&);
#endif
    void saveUserFlags(NumPtr<bool>&);
    void loadUserFlagsForShapes(const NumPtr<bool>&);
#if (SPACEDIM>1)
    void loadUserFlagsForCodimension1Shapes(const NumPtr<bool>&);
#endif
#if (SPACEDIM==3)
    void loadUserFlagsForCodimension2Shapes(const NumPtr<bool>&);
#endif
    void loadUserFlags(const NumPtr<bool>&);
    inline int numberShapes(const LEVEL_CLASS *level) const {
      return level->numberShapes();
    }
    int numberShapes() const;
#if (SPACEDIM>1)
    inline int numberCodimension1Shapes(
    const CODIMENSION1_LEVEL_CLASS *level) const {
      return level->numberShapes();
    }
    int numberCodimension1Shapes() const;
#endif
#if (SPACEDIM==3)
    inline int numberCodimension2Shapes(const Level1 *level) const {
      return level->numberShapes();
    }
    int numberCodimension2Shapes() const;
#endif
    int numberLevels();
    const NumPtr<Point<SPACEDIM> >& getVertices() const {
      return vertices;
    }
    int numberUsedVertices() const;
    bool vertexUsed(int v) const { return vertices_used[v]; }
};
ostream& operator<<(ostream&,const Tessellation::MeshSmoothing&);
#endif
