// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <TimeIntegrator.H>
#include <algorithm>
#include "ConditionNumber.H"
#include "MemoryDebugger.H"
#include "Tracer.H"
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
ostream& operator<<(ostream &os,TIME_INTEGRATOR_STATE tis) {
  switch (tis) {
    case TIME_INTEGRATOR_INITIALIZING:
      os << "TIME_INTEGRATOR_INITIALIZING";
      break;
    case TIME_INTEGRATOR_STEPPING:
      os << "TIME_INTEGRATOR_STEPPING";
      break;
    case TIME_INTEGRATOR_DONE:
      os << "TIME_INTEGRATOR_DONE";
      break;
    default:
      os << "unknown";
      break;
  }
  return os;
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TimeIntegrator::TimeIntegrator(TimeControl &cn,const DofHandler &dh,
const Mapping &m,const map<int,Quadrature<SPACEDIM> > &qm,
map<int,Quadrature<SPACEDIM-1> > &fm,const SolverPreconditioner &p,
Solver &s,NumPtr<Vector*> &sol,NumPtr<Vector*> &rhs,bool tf,bool ps,int nsols) : cntrl(cn),
dof_handler(dh),mapping(m),qmap(qm),fqmap(fm),preconditioner(p),
solver(s),tessellation_fixed(tf),preserve_symmetry(ps),
coefs_transient(false),n_dofs(0) {

  if (tessellation_fixed) {
    n_dofs=dof_handler.numberDofs();
    for (int i=0;i<rhs.getNumber();i++) {
      if (rhs[i]==0) rhs[i]=OPERATOR_NEW Vector(n_dofs,0.);
    }
    for (int i=0;i<sol.getNumber();i++) {
      if (sol[i]==0) sol[i]=OPERATOR_NEW Vector(n_dofs,0.);
    }
  }
  Tessellation &tess=dof_handler.getTessellation();
  coefs_transient=false;
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *present_level=tess.presentLevel();
    TransientMaterialModel *tmm=dynamic_cast<TransientMaterialModel*>(
      present_level->getMaterialModel());
    CHECK_POINTER(tmm);
    TimeIndependentTransientMaterialModel *titmm=
      dynamic_cast<TimeIndependentTransientMaterialModel*>(tmm);
    if (titmm==0) { coefs_transient=true; break; }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TimeIntegrator::~TimeIntegrator() {
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimeIntegrator::l2ProjectionInitialData(const SparsityPattern *sp,
NumPtr<Vector*> &solution,NumPtr<Vector*> &rhs,
NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  if (tessellation_fixed) {
    CHECK_SAME(n_dofs,sp->nRows());
    for (int i=0;i<mass_matrix.getNumber();i++) {
      mass_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
    }
    for (int i=0;i<mass_matrix.getNumber();i++) {
      stiffness_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
    }
//  if (!coefs_transient) {
      dof_handler.assembleEllipticProjectionLinearSystem(mapping,qmap,
        *stiffness_matrix[0],*rhs[0]);
//  }
    *rhs[0]=0.;
  } else {
    mass_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
    int n=sp->nRows();
    rhs[0]=OPERATOR_NEW Vector(n,0.);
    solution[0]=OPERATOR_NEW Vector(n,0.);
  }
  dof_handler.assembleL2ProjectionLinearSystem(mapping,qmap,
    *mass_matrix[0],*rhs[0]);
  solver.solve(*mass_matrix[0],*solution[0],*rhs[0],preconditioner);

  if (!tessellation_fixed) {
    delete mass_matrix[0]; mass_matrix[0]=0;
    delete rhs[0]; rhs[0]=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimeIntegrator::ellipticProjectionInitialData(
const SparsityPattern *sp,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  if (tessellation_fixed) {
    CHECK_SAME(n_dofs,sp->nRows());
    for (int i=0;i<mass_matrix.getNumber();i++) {
      stiffness_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
    }
    for (int i=0;i<mass_matrix.getNumber();i++) {
      mass_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
    }
    dof_handler.assembleL2ProjectionLinearSystem(mapping,qmap,
      *mass_matrix[0],*rhs[0]);
    *rhs[0]=0.;
  } else {
    stiffness_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
    int n=sp->nRows();
    rhs[0]=OPERATOR_NEW Vector(n,0.);
    solution[0]=OPERATOR_NEW Vector(n,0.);
  }
  dof_handler.assembleEllipticProjectionLinearSystem(mapping,qmap,
    *stiffness_matrix[0],*rhs[0]);

  SparseMatrix *solver_matrix=0;
  if (coefs_transient) solver_matrix=stiffness_matrix[0];
  else { // don't let applyBoundaryValues destroy mass_matrix[0]
    solver_matrix=stiffness_matrix[0]->clone();
    solver_matrix->copyFrom(*stiffness_matrix[0]);
  }
  map<int,REAL> boundary_values;
  dof_handler.interpolateBoundaryValues(mapping,boundary_values);
  dof_handler.applyBoundaryValues(boundary_values,
    *solver_matrix,*rhs[0],preserve_symmetry);
  dof_handler.applyBoundaryValues(boundary_values,*solution[0]);
  solver.solve(*solver_matrix,*solution[0],*rhs[0],preconditioner);
  if (!tessellation_fixed) {
    delete stiffness_matrix[0]; stiffness_matrix[0]=0;
    delete rhs[0]; rhs[0]=0;
  }
  if (!coefs_transient) delete solver_matrix;
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
ForwardEulerTimeIntegrator::ForwardEulerTimeIntegrator(TimeControl &cn,
const DofHandler &dh,const Mapping &m,
const map<int,Quadrature<SPACEDIM> > &qm,
map<int,Quadrature<SPACEDIM-1> > &fm,const SolverPreconditioner &p,
Solver &s,NumPtr<Vector*> &sol,NumPtr<Vector*> &rhs,bool tf,bool ps) : 
TimeIntegrator(cn,dh,m,qm,fm,p,s,sol,rhs,tf,ps,2) {
  CHECK_TEST(sol.getNumber()>=numberSolutions());
  CHECK_TEST(rhs.getNumber()>=numberSolutions());
  if (tessellation_fixed) {
    rhs_minus_stiffness_matrix_times_old_solution=
      OPERATOR_NEW Vector(n_dofs,0.);
    mass_matrix_times_old_solution=OPERATOR_NEW Vector(n_dofs,0.);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
ForwardEulerTimeIntegrator::~ForwardEulerTimeIntegrator() {
  if (tessellation_fixed) {
    delete rhs_minus_stiffness_matrix_times_old_solution; 
      rhs_minus_stiffness_matrix_times_old_solution=0;
    delete mass_matrix_times_old_solution;
      mass_matrix_times_old_solution=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TIME_INTEGRATOR_STATE ForwardEulerTimeIntegrator::takeStep(REAL dt,
const SparsityPattern *sp,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  REAL old_time=cntrl.getTime();
  REAL new_time=old_time+dt;
  swap(solution[0],solution[1]);
  swap(rhs[0],rhs[1]);
  if (coefs_transient) {
    CHECK_TEST(mass_matrix.getNumber()>=numberSolutions());
    CHECK_TEST(stiffness_matrix.getNumber()>=numberSolutions());
    swap(mass_matrix[0],mass_matrix[1]);
    swap(stiffness_matrix[0],stiffness_matrix[1]);
    if (tessellation_fixed) {
      CHECK_SAME(n_dofs,sp->nRows());
//    avoid re-computing matrices at old time
      mass_matrix[0]->valueArray().initialize(0.);
      stiffness_matrix[0]->valueArray().initialize(0.);
      (*rhs[0])=0.;
    } else {
      for (int i=0;i<2;i++) {
        if (mass_matrix[i]!=0) delete mass_matrix[i]; 
        mass_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
      }
      for (int i=0;i<2;i++) {
        if (stiffness_matrix[i]!=0) delete stiffness_matrix[i]; 
        stiffness_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
      }
      int n=sp->nRows();
      for (int i=0;i<2;i++) {
        if (rhs[i]!=0) delete rhs[i]; 
        rhs[i]=OPERATOR_NEW Vector(n,0.);
      }
      CHECK_SAME(solution[1]->size(),n);
      if (solution[0]!=0) delete solution[0]; 
      solution[0]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*mass_matrix[1],*stiffness_matrix[1],*rhs[1]);
    }
    dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
      fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[0]);
  } else {
    CHECK_TEST(mass_matrix.getNumber()>=1);
    CHECK_TEST(stiffness_matrix.getNumber()>=1);
    if (tessellation_fixed) {
      CHECK_POINTER(rhs[1]);
      CHECK_SAME(n_dofs,sp->nRows());
//    avoid computing any matrices
      (*rhs[1])=0.;
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*rhs[1]);
    } else {
      if (mass_matrix[0]!=0) delete mass_matrix[0]; 
      mass_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      if (stiffness_matrix[0]!=0) delete stiffness_matrix[0]; 
      stiffness_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      int n=sp->nRows();
      if (rhs[1]!=0) delete rhs[1]; 
      rhs[1]=OPERATOR_NEW Vector(n,0.);
      CHECK_SAME(solution[1]->size(),n);
      if (solution[0]!=0) delete solution[0]; 
      solution[0]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[1]);
    }
  }
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);
  SparseMatrix *old_stiffness_matrix=
    (coefs_transient ? stiffness_matrix[1] : stiffness_matrix[0]);
  if (!tessellation_fixed) {
    rhs_minus_stiffness_matrix_times_old_solution=solution[0]->clone();
    mass_matrix_times_old_solution=solution[0]->clone();
  }
  *mass_matrix_times_old_solution=0.;
  old_mass_matrix->vmult(
    *mass_matrix_times_old_solution,*solution[1]);
  *rhs_minus_stiffness_matrix_times_old_solution=0.;
  old_stiffness_matrix->vmult(
    *rhs_minus_stiffness_matrix_times_old_solution,*solution[1]);
  for (int i=0;i<rhs[1]->size();i++) {
    (*rhs_minus_stiffness_matrix_times_old_solution)[i]=
      dt*((*rhs[1])[i]-(*rhs_minus_stiffness_matrix_times_old_solution)[i])
      +(*mass_matrix_times_old_solution)[i];
  }
//avoid destroying matrix entries associated with essential boundaries
  SparseMatrix *new_mass_matrix=mass_matrix[0]->clone();
  new_mass_matrix->copyFrom(*mass_matrix[0]);
  map<int,REAL> new_boundary_values;
  dof_handler.interpolateBoundaryValues(mapping,new_boundary_values,
    fqmap,new_time);
  dof_handler.applyBoundaryValues(new_boundary_values,*new_mass_matrix,
    *rhs_minus_stiffness_matrix_times_old_solution,preserve_symmetry);
  dof_handler.applyBoundaryValues(new_boundary_values,*solution[0]);
  solver.solve(*new_mass_matrix,*solution[0],
    *rhs_minus_stiffness_matrix_times_old_solution,preconditioner);
  delete new_mass_matrix; new_mass_matrix=0;
  if (!tessellation_fixed) {
    delete mass_matrix_times_old_solution;
      mass_matrix_times_old_solution=0;
    delete rhs_minus_stiffness_matrix_times_old_solution;
      rhs_minus_stiffness_matrix_times_old_solution=0;
  }
  return TimeIntegrator::takeStep(dt);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ForwardEulerTimeIntegrator::takeFirstDeferredCorrectionStep(
REAL old_time,REAL dt,const SparsityPattern *sp,
map<int,REAL> &new_boundary_values,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  REAL new_time=old_time+dt;
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);
  SparseMatrix *old_stiffness_matrix=
    (coefs_transient ? stiffness_matrix[1] : stiffness_matrix[0]);
  if (!tessellation_fixed) {
    rhs_minus_stiffness_matrix_times_old_solution=solution[0]->clone();
    mass_matrix_times_old_solution=solution[0]->clone();
  }
  *mass_matrix_times_old_solution=0.;
  old_mass_matrix->vmult(
    *mass_matrix_times_old_solution,*solution[1]);
  *rhs_minus_stiffness_matrix_times_old_solution=0.;
  old_stiffness_matrix->vmult(
    *rhs_minus_stiffness_matrix_times_old_solution,*solution[1]);
  for (int i=0;i<rhs[1]->size();i++) {
    (*rhs_minus_stiffness_matrix_times_old_solution)[i]=
      dt*((*rhs[1])[i]-(*rhs_minus_stiffness_matrix_times_old_solution)[i])
      +(*mass_matrix_times_old_solution)[i];
  }
  SparseMatrix *new_mass_matrix=mass_matrix[0]->clone();
  new_mass_matrix->copyFrom(*mass_matrix[0]);
  dof_handler.interpolateBoundaryValues(mapping,new_boundary_values,
    fqmap,new_time);
  dof_handler.applyBoundaryValues(new_boundary_values,*new_mass_matrix,
    *rhs_minus_stiffness_matrix_times_old_solution,preserve_symmetry);
  dof_handler.applyBoundaryValues(new_boundary_values,*solution[0]);
  solver.solve(*new_mass_matrix,*solution[0],
    *rhs_minus_stiffness_matrix_times_old_solution,preconditioner);
  delete new_mass_matrix; new_mass_matrix=0;
  if (!tessellation_fixed) {
    delete mass_matrix_times_old_solution;
      mass_matrix_times_old_solution=0;
    delete rhs_minus_stiffness_matrix_times_old_solution;
      rhs_minus_stiffness_matrix_times_old_solution=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ForwardEulerTimeIntegrator::takeDeferredCorrectionStep(REAL dt,
const SparsityPattern *sp,const map<int,REAL> &zero_boundary_values,
NumPtr<Vector*> &correction,NumPtr<Vector*> &epsilon,
NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);
  SparseMatrix *old_stiffness_matrix=
    (coefs_transient ? stiffness_matrix[1] : stiffness_matrix[0]);
  if (!tessellation_fixed) {
    rhs_minus_stiffness_matrix_times_old_solution=epsilon[0]->clone();
  }
  old_stiffness_matrix->vmult(
    *rhs_minus_stiffness_matrix_times_old_solution,*correction[1]);
  rhs_minus_stiffness_matrix_times_old_solution->scale(-1.);
  old_mass_matrix->vmultAdd(
    *rhs_minus_stiffness_matrix_times_old_solution,*correction[1]);
  rhs_minus_stiffness_matrix_times_old_solution->add(1.,*epsilon[0]);
  rhs_minus_stiffness_matrix_times_old_solution->add(-1.,*epsilon[1]);
//avoid destroying matrix entries associated with essential boundaries
  SparseMatrix *new_mass_matrix=mass_matrix[0]->clone();
  new_mass_matrix->copyFrom(*mass_matrix[0]);
  dof_handler.applyBoundaryValues(zero_boundary_values,*new_mass_matrix,
    *rhs_minus_stiffness_matrix_times_old_solution,preserve_symmetry);
  dof_handler.applyBoundaryValues(zero_boundary_values,*correction[0]);
  solver.solve(*new_mass_matrix,*correction[0],
    *rhs_minus_stiffness_matrix_times_old_solution,preconditioner);
  delete new_mass_matrix; new_mass_matrix=0;
  if (!tessellation_fixed) {
    delete rhs_minus_stiffness_matrix_times_old_solution;
      rhs_minus_stiffness_matrix_times_old_solution=0;
  }
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BackwardEulerTimeIntegrator::BackwardEulerTimeIntegrator(
TimeControl &cn,const DofHandler &dh,const Mapping &m,
const map<int,Quadrature<SPACEDIM> > &qm,
map<int,Quadrature<SPACEDIM-1> > &fm,const SolverPreconditioner &p,
Solver &s,NumPtr<Vector*> &sol,NumPtr<Vector*> &rhs,bool tf,bool ps) :
TimeIntegrator(cn,dh,m,qm,fm,p,s,sol,rhs,tf,ps,2) {
  CHECK_TEST(sol.getNumber()>=numberSolutions());
  CHECK_TEST(rhs.getNumber()>=numberSolutions());
  if (tessellation_fixed) {
    rhs_times_dt_plus_mass_matrix_times_old_solution=
      OPERATOR_NEW Vector(n_dofs,0.);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BackwardEulerTimeIntegrator::~BackwardEulerTimeIntegrator() {
  if (tessellation_fixed) {
    delete rhs_times_dt_plus_mass_matrix_times_old_solution; 
      rhs_times_dt_plus_mass_matrix_times_old_solution=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TIME_INTEGRATOR_STATE BackwardEulerTimeIntegrator::takeStep(REAL dt,
const SparsityPattern *sp,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  REAL old_time=cntrl.getTime();
  REAL new_time=old_time+dt;
  swap(solution[0],solution[1]);
  swap(rhs[0],rhs[1]);
  if (coefs_transient) {
    CHECK_TEST(mass_matrix.getNumber()>=numberSolutions());
    CHECK_TEST(stiffness_matrix.getNumber()>=numberSolutions());
    swap(mass_matrix[0],mass_matrix[1]);
    swap(stiffness_matrix[0],stiffness_matrix[1]);
    if (tessellation_fixed) {
      CHECK_SAME(n_dofs,sp->nRows());
//    avoid re-computing matrices at old time
      mass_matrix[0]->valueArray().initialize(0.);
      stiffness_matrix[0]->valueArray().initialize(0.);
      (*rhs[0])=0.;
    } else {
      for (int i=0;i<2;i++) {
        if (mass_matrix[i]!=0) delete mass_matrix[i];
        mass_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
      }
      for (int i=0;i<2;i++) {
        if (stiffness_matrix[i]!=0) delete stiffness_matrix[i];
        stiffness_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
      }
      int n=sp->nRows();
      for (int i=0;i<2;i++) {
        if (rhs[i]!=0) delete rhs[i];
        rhs[i]=OPERATOR_NEW Vector(n,0.);
      }
      CHECK_SAME(solution[1]->size(),n);
      if (solution[0]!=0) delete solution[0];
      solution[0]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*mass_matrix[1],*stiffness_matrix[1],*rhs[1]);
    }
    dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
      fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[0]);
  } else {
    CHECK_TEST(mass_matrix.getNumber()>=1);
    CHECK_TEST(stiffness_matrix.getNumber()>=1);
    if (tessellation_fixed) {
      CHECK_POINTER(rhs[0]);
      CHECK_SAME(n_dofs,sp->nRows());
//    avoid computing any matrices
      (*rhs[0])=0.;
      dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
        fqmap,*rhs[0]);
    } else {
      if (mass_matrix[0]!=0) delete mass_matrix[0];
      mass_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      if (stiffness_matrix[0]!=0) delete stiffness_matrix[0];
      stiffness_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      int n=sp->nRows();
      if (rhs[0]!=0) delete rhs[0];
      rhs[0]=OPERATOR_NEW Vector(n,0.);
      CHECK_SAME(solution[1]->size(),n);
      if (solution[0]!=0) delete solution[0];
      solution[0]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
        fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[0]);
    }
  }
  if (!tessellation_fixed) {
    rhs_times_dt_plus_mass_matrix_times_old_solution=solution[0]->clone();
  }
  SparseMatrix mass_matrix_plus_dt_times_stiffness_matrix(sp);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(1.,
    *mass_matrix[0]);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(dt,
    *stiffness_matrix[0]);
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);
  old_mass_matrix->vmult(
    *rhs_times_dt_plus_mass_matrix_times_old_solution,*solution[1]);
  rhs_times_dt_plus_mass_matrix_times_old_solution->add(dt,*rhs[0]);
  map<int,REAL> new_boundary_values;
  dof_handler.interpolateBoundaryValues(mapping,new_boundary_values,
    fqmap,new_time);
  dof_handler.applyBoundaryValues(new_boundary_values,
    mass_matrix_plus_dt_times_stiffness_matrix,
    *rhs_times_dt_plus_mass_matrix_times_old_solution,preserve_symmetry);
  Vector &soln=*solution[0];
  dof_handler.applyBoundaryValues(new_boundary_values,soln);
  solver.solve(mass_matrix_plus_dt_times_stiffness_matrix,soln,
    *rhs_times_dt_plus_mass_matrix_times_old_solution,preconditioner);
  if (!tessellation_fixed) {
    delete rhs_times_dt_plus_mass_matrix_times_old_solution;
    rhs_times_dt_plus_mass_matrix_times_old_solution=0;
  }
  return TimeIntegrator::takeStep(dt);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BackwardEulerTimeIntegrator::takeFirstDeferredCorrectionStep(
REAL old_time,REAL dt,const SparsityPattern *sp,
map<int,REAL> &new_boundary_values,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  REAL new_time=old_time+dt;
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);
  if (!tessellation_fixed) {
    rhs_times_dt_plus_mass_matrix_times_old_solution=solution[0]->clone();
  }
  SparseMatrix mass_matrix_plus_dt_times_stiffness_matrix(sp);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(1.,
    *mass_matrix[0]);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(dt,
    *stiffness_matrix[0]);
  old_mass_matrix->vmult(
    *rhs_times_dt_plus_mass_matrix_times_old_solution,*solution[1]);
  rhs_times_dt_plus_mass_matrix_times_old_solution->add(dt,*rhs[0]);
  dof_handler.interpolateBoundaryValues(mapping,new_boundary_values,
    fqmap,new_time);
  dof_handler.applyBoundaryValues(new_boundary_values,
    mass_matrix_plus_dt_times_stiffness_matrix,
    *rhs_times_dt_plus_mass_matrix_times_old_solution,preserve_symmetry);
  Vector &soln=*solution[0];
  dof_handler.applyBoundaryValues(new_boundary_values,soln);
  solver.solve(mass_matrix_plus_dt_times_stiffness_matrix,soln,
    *rhs_times_dt_plus_mass_matrix_times_old_solution,preconditioner);
  if (!tessellation_fixed) {
    delete rhs_times_dt_plus_mass_matrix_times_old_solution;
    rhs_times_dt_plus_mass_matrix_times_old_solution=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BackwardEulerTimeIntegrator::takeDeferredCorrectionStep(REAL dt,
const SparsityPattern *sp,const map<int,REAL> &zero_boundary_values,
NumPtr<Vector*> &correction,NumPtr<Vector*> &epsilon,
NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);
  if (!tessellation_fixed) {
    rhs_times_dt_plus_mass_matrix_times_old_solution=epsilon[0]->clone();
  }
  old_mass_matrix->vmult(
    *rhs_times_dt_plus_mass_matrix_times_old_solution,*correction[1]);
  rhs_times_dt_plus_mass_matrix_times_old_solution->add(1.,*epsilon[0]);
  rhs_times_dt_plus_mass_matrix_times_old_solution->add(-1.,*epsilon[1]);
  SparseMatrix mass_matrix_plus_dt_times_stiffness_matrix(sp);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(1.,
    *mass_matrix[0]);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(dt,
    *stiffness_matrix[0]);
  dof_handler.applyBoundaryValues(zero_boundary_values,
    mass_matrix_plus_dt_times_stiffness_matrix,
    *rhs_times_dt_plus_mass_matrix_times_old_solution,preserve_symmetry);
  *correction[0]=0.;
  solver.solve(mass_matrix_plus_dt_times_stiffness_matrix,*correction[0],
    *rhs_times_dt_plus_mass_matrix_times_old_solution,preconditioner);
  if (!tessellation_fixed) {
    delete rhs_times_dt_plus_mass_matrix_times_old_solution;
    rhs_times_dt_plus_mass_matrix_times_old_solution=0;
  }
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
CrankNicolsonTimeIntegrator::CrankNicolsonTimeIntegrator(
TimeControl &cn,const DofHandler &dh,const Mapping &m,
const map<int,Quadrature<SPACEDIM> > &qm,
map<int,Quadrature<SPACEDIM-1> > &fm,const SolverPreconditioner &p,
Solver &s,NumPtr<Vector*> &sol,NumPtr<Vector*> &rhs,bool tf,bool ps) :
TimeIntegrator(cn,dh,m,qm,fm,p,s,sol,rhs,tf,ps,2) {
  CHECK_TEST(sol.getNumber()>=numberSolutions());
  CHECK_TEST(rhs.getNumber()>=numberSolutions());
  if (tessellation_fixed) {
    rhs_times_dt_plus_matrix_times_old_solution=rhs[0]->clone();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
CrankNicolsonTimeIntegrator::~CrankNicolsonTimeIntegrator() {
  if (tessellation_fixed) {
    delete rhs_times_dt_plus_matrix_times_old_solution; 
    rhs_times_dt_plus_matrix_times_old_solution=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TIME_INTEGRATOR_STATE CrankNicolsonTimeIntegrator::takeStep(REAL dt,
const SparsityPattern *sp,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  REAL old_time=cntrl.getTime();
  REAL new_time=old_time+dt;
  swap(solution[0],solution[1]);
  swap(rhs[0],rhs[1]);
  if (coefs_transient) {
    CHECK_TEST(mass_matrix.getNumber()>=numberSolutions());
    CHECK_TEST(stiffness_matrix.getNumber()>=numberSolutions());
    swap(mass_matrix[0],mass_matrix[1]);
    swap(stiffness_matrix[0],stiffness_matrix[1]);
    if (tessellation_fixed) {
      CHECK_SAME(n_dofs,sp->nRows());
//    avoid re-computing matrices at old time
      mass_matrix[0]->valueArray().initialize(0.);
      stiffness_matrix[0]->valueArray().initialize(0.);
      (*rhs[0])=0.;
    } else {
      for (int i=0;i<2;i++) {
        if (mass_matrix[i]!=0) delete mass_matrix[i];
        mass_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
      }
      for (int i=0;i<2;i++) {
        if (stiffness_matrix[i]!=0) delete stiffness_matrix[i];
        stiffness_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
      }
      int n=sp->nRows();
      for (int i=0;i<2;i++) {
        if (rhs[i]!=0) delete rhs[i];
        rhs[i]=OPERATOR_NEW Vector(n,0.);
      }
      CHECK_SAME(solution[1]->size(),n);
      if (solution[0]!=0) delete solution[0];
      solution[0]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*mass_matrix[1],*stiffness_matrix[1],*rhs[1]);
    }
    dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
      fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[0]);
  } else {
    CHECK_TEST(mass_matrix.getNumber()>=1);
    CHECK_TEST(stiffness_matrix.getNumber()>=1);
    if (tessellation_fixed) {
      CHECK_POINTER(rhs[0]);
      CHECK_SAME(n_dofs,sp->nRows());
//    avoid computing any matrices
      (*rhs[1])=0.;
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*rhs[1]);
      (*rhs[0])=0.;
      dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
        fqmap,*rhs[0]);
    } else {
      if (mass_matrix[0]!=0) delete mass_matrix[0];
      mass_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      if (stiffness_matrix[0]!=0) delete stiffness_matrix[0];
      stiffness_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      int n=sp->nRows();
      if (rhs[0]!=0) delete rhs[0];
      rhs[0]=OPERATOR_NEW Vector(n,0.);
      CHECK_SAME(solution[1]->size(),n);
      if (solution[0]!=0) delete solution[0];
      solution[0]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
        fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[0]);
      if (rhs[1]!=0) delete rhs[1];
      rhs[1]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*rhs[1]);
    }
  }
  if (!tessellation_fixed) {
    rhs_times_dt_plus_matrix_times_old_solution=solution[0]->clone();
  }

  REAL dt2=0.5*dt;
  SparseMatrix mass_matrix_plus_dt_times_stiffness_matrix(sp);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(1.,*mass_matrix[0]);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(dt2,
    *stiffness_matrix[0]);
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);;
  SparseMatrix *old_stiffness_matrix=
    (coefs_transient ? stiffness_matrix[1] : stiffness_matrix[0]);;
  Vector &old_soln=*solution[1];
  old_stiffness_matrix->vmult(
    *rhs_times_dt_plus_matrix_times_old_solution,old_soln);
  rhs_times_dt_plus_matrix_times_old_solution->scale(-dt2);
  old_mass_matrix->vmultAdd(
    *rhs_times_dt_plus_matrix_times_old_solution,old_soln);
  rhs_times_dt_plus_matrix_times_old_solution->add(dt2,*rhs[0]);
  rhs_times_dt_plus_matrix_times_old_solution->add(dt2,*rhs[1]);
  Vector &new_soln=*solution[0];
  map<int,REAL> new_boundary_values;
  dof_handler.interpolateBoundaryValues(mapping,new_boundary_values,
    fqmap,new_time);
  dof_handler.applyBoundaryValues(new_boundary_values,
    mass_matrix_plus_dt_times_stiffness_matrix,
    *rhs_times_dt_plus_matrix_times_old_solution,preserve_symmetry);
  dof_handler.applyBoundaryValues(new_boundary_values,new_soln);
  solver.solve(mass_matrix_plus_dt_times_stiffness_matrix,new_soln,
    *rhs_times_dt_plus_matrix_times_old_solution,preconditioner);
  if (!tessellation_fixed) {
    delete rhs_times_dt_plus_matrix_times_old_solution;
    rhs_times_dt_plus_matrix_times_old_solution=0;
  }
  return TimeIntegrator::takeStep(dt);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void CrankNicolsonTimeIntegrator::takeFirstDeferredCorrectionStep(
REAL old_time,REAL dt,const SparsityPattern *sp,
map<int,REAL> &new_boundary_values,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  REAL new_time=old_time+dt;
  if (!tessellation_fixed) {
    rhs_times_dt_plus_matrix_times_old_solution=
      solution[0]->clone();
  }
  REAL dt2=0.5*dt;
  SparseMatrix mass_matrix_plus_dt_times_stiffness_matrix(sp);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(1.,
    *mass_matrix[0]);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(dt2,
    *stiffness_matrix[0]);
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);;
  SparseMatrix *old_stiffness_matrix=
    (coefs_transient ? stiffness_matrix[1] : stiffness_matrix[0]);;
  Vector &old_soln=*solution[1];
  old_stiffness_matrix->vmult(
    *rhs_times_dt_plus_matrix_times_old_solution,old_soln);
  rhs_times_dt_plus_matrix_times_old_solution->scale(-dt2);
  old_mass_matrix->vmultAdd(
    *rhs_times_dt_plus_matrix_times_old_solution,old_soln);
  rhs_times_dt_plus_matrix_times_old_solution->add(dt2,*rhs[0]);
  rhs_times_dt_plus_matrix_times_old_solution->add(dt2,*rhs[1]);
  Vector &new_soln=*solution[0];
  dof_handler.interpolateBoundaryValues(mapping,new_boundary_values,
    fqmap,new_time);
  dof_handler.applyBoundaryValues(new_boundary_values,
    mass_matrix_plus_dt_times_stiffness_matrix,
    *rhs_times_dt_plus_matrix_times_old_solution,preserve_symmetry);
  dof_handler.applyBoundaryValues(new_boundary_values,new_soln);
  solver.solve(mass_matrix_plus_dt_times_stiffness_matrix,new_soln,
    *rhs_times_dt_plus_matrix_times_old_solution,preconditioner);
  if (!tessellation_fixed) {
    delete rhs_times_dt_plus_matrix_times_old_solution;
    rhs_times_dt_plus_matrix_times_old_solution=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void CrankNicolsonTimeIntegrator::takeDeferredCorrectionStep(REAL dt,
const SparsityPattern *sp,const map<int,REAL> &zero_boundary_values,
NumPtr<Vector*> &correction,NumPtr<Vector*> &epsilon,
NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  if (!tessellation_fixed) {
    rhs_times_dt_plus_matrix_times_old_solution=
      correction[0]->clone();
  }
  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);;
  SparseMatrix *old_stiffness_matrix=
    (coefs_transient ? stiffness_matrix[1] : stiffness_matrix[0]);;
  old_stiffness_matrix->vmult(
    *rhs_times_dt_plus_matrix_times_old_solution,*correction[1]);
  REAL dt2=0.5*dt;
  rhs_times_dt_plus_matrix_times_old_solution->scale(-dt2);
  old_mass_matrix->vmultAdd(
    *rhs_times_dt_plus_matrix_times_old_solution,*correction[1]);
  rhs_times_dt_plus_matrix_times_old_solution->add(1.,*epsilon[0]);
  rhs_times_dt_plus_matrix_times_old_solution->add(-1.,*epsilon[1]);
  SparseMatrix mass_matrix_plus_dt_times_stiffness_matrix(sp);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(1.,
    *mass_matrix[0]);
  mass_matrix_plus_dt_times_stiffness_matrix.addScaled(dt2,
    *stiffness_matrix[0]);
  dof_handler.applyBoundaryValues(zero_boundary_values,
    mass_matrix_plus_dt_times_stiffness_matrix,
    *rhs_times_dt_plus_matrix_times_old_solution,preserve_symmetry);
  *correction[0]=0.;
  solver.solve(mass_matrix_plus_dt_times_stiffness_matrix,*correction[0],
    *rhs_times_dt_plus_matrix_times_old_solution,preconditioner);
  if (!tessellation_fixed) {
    delete rhs_times_dt_plus_matrix_times_old_solution;
    rhs_times_dt_plus_matrix_times_old_solution=0;
  }
}

const REAL TRBDFTimeIntegrator::gamma=2.-sqrt(2.);
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TRBDFTimeIntegrator::TRBDFTimeIntegrator(TimeControl &cn,
const DofHandler &dh,const Mapping &m,
const map<int,Quadrature<SPACEDIM> > &qm,
map<int,Quadrature<SPACEDIM-1> > &fm,const SolverPreconditioner &p,
Solver &s,NumPtr<Vector*> &sol,NumPtr<Vector*> &rhs,bool tf,bool ps) :
TimeIntegrator(cn,dh,m,qm,fm,p,s,sol,rhs,tf,ps,3) {
  CHECK_TEST(sol.getNumber()>=numberSolutions());
  CHECK_TEST(rhs.getNumber()>=numberSolutions());
  if (tessellation_fixed) {
    rhs_times_dt_plus_matrix_times_old_solution=rhs[0]->clone();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TRBDFTimeIntegrator::~TRBDFTimeIntegrator() {
  if (tessellation_fixed) {
    delete rhs_times_dt_plus_matrix_times_old_solution; 
      rhs_times_dt_plus_matrix_times_old_solution=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TIME_INTEGRATOR_STATE TRBDFTimeIntegrator::takeStep(REAL dt,
const SparsityPattern *sp,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  REAL old_time=cntrl.getTime();
  REAL middle_time=old_time+gamma*dt;
  REAL new_time=old_time+dt;
  swap(solution[0],solution[2]);
  swap(rhs[0],rhs[2]);
  int n=sp->nRows();
  if (coefs_transient) {
    CHECK_TEST(mass_matrix.getNumber()>=numberSolutions());
    CHECK_TEST(stiffness_matrix.getNumber()>=numberSolutions());
    swap(mass_matrix[0],mass_matrix[2]);
    swap(stiffness_matrix[0],stiffness_matrix[2]);
    if (tessellation_fixed) {
      CHECK_SAME(n_dofs,n);
//    avoid re-computing matrices at old time
      for (int i=0;i<2;i++) {
        mass_matrix[i]->valueArray().initialize(0.);
      }
      for (int i=0;i<2;i++) {
        stiffness_matrix[i]->valueArray().initialize(0.);
      }
      for (int i=0;i<2;i++) (*rhs[i])=0.;
    } else {
      for (int i=0;i<3;i++) {
        if (mass_matrix[i]!=0) delete mass_matrix[i];
        mass_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
      }
      for (int i=0;i<3;i++) {
        if (stiffness_matrix[i]!=0) delete stiffness_matrix[i];
        stiffness_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
      }
      for (int i=0;i<3;i++) {
        if (rhs[i]!=0) delete rhs[i];
        rhs[i]=OPERATOR_NEW Vector(n,0.);
      }
      CHECK_SAME(solution[2]->size(),n);
      for (int i=0;i<2;i++) {
        if (solution[i]!=0) delete solution[i];
        solution[i]=OPERATOR_NEW Vector(n,0.);
      }
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*mass_matrix[2],*stiffness_matrix[2],*rhs[2]);
    }
    dof_handler.assembleTransientLinearSystem(middle_time,mapping,qmap,
      fqmap,*mass_matrix[1],*stiffness_matrix[1],*rhs[1]);
    dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
      fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[0]);
  } else {
    CHECK_TEST(mass_matrix.getNumber()>=1);
    CHECK_TEST(stiffness_matrix.getNumber()>=1);
    if (tessellation_fixed) {
      CHECK_POINTER(rhs[0]);
      CHECK_SAME(n_dofs,n);
      (*rhs[0])=0.;
      dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
        fqmap,*rhs[0]);
      CHECK_POINTER(rhs[1]);
      (*rhs[1])=0.;
      dof_handler.assembleTransientLinearSystem(middle_time,mapping,qmap,
        fqmap,*rhs[1]);
      CHECK_POINTER(rhs[2]);
      (*rhs[2])=0.;
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*rhs[2]);
    } else {
      if (mass_matrix[0]!=0) delete mass_matrix[0];
      mass_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      if (stiffness_matrix[0]!=0) delete stiffness_matrix[0];
      stiffness_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      CHECK_SAME(solution[2]->size(),n);
      for (int i=0;i<2;i++) {
        if (solution[i]!=0) delete solution[i];
        solution[i]=OPERATOR_NEW Vector(n,0.);
      }
      if (rhs[0]!=0) delete rhs[0];
      rhs[0]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(new_time,mapping,qmap,
        fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[0]);
      if (rhs[1]!=0) delete rhs[1];
      rhs[1]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(middle_time,mapping,qmap,
        fqmap,*rhs[1]);
      if (rhs[2]!=0) delete rhs[2];
      rhs[2]=OPERATOR_NEW Vector(n,0.);
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*rhs[1]);
    }
  }
  if (!tessellation_fixed) {
    rhs_times_dt_plus_matrix_times_old_solution=rhs[0]->clone();
  }

  SparseMatrix *old_mass_matrix=
    (coefs_transient ? mass_matrix[2] : mass_matrix[0]);
  SparseMatrix *middle_mass_matrix=
    (coefs_transient ? mass_matrix[1] : mass_matrix[0]);
  SparseMatrix *old_stiffness_matrix=
    (coefs_transient ? stiffness_matrix[2] : stiffness_matrix[0]);
  SparseMatrix *middle_stiffness_matrix=
    (coefs_transient ? stiffness_matrix[1] : stiffness_matrix[0]);
  SparseMatrix first_mass_matrix_plus_dt_times_stiffness_matrix(sp);
  first_mass_matrix_plus_dt_times_stiffness_matrix.addScaled(1.,
    *middle_mass_matrix);
  first_mass_matrix_plus_dt_times_stiffness_matrix.addScaled(0.5*gamma*dt,
    *middle_stiffness_matrix);
  SparseMatrix second_mass_matrix_plus_dt_times_stiffness_matrix(sp);
  if (!coefs_transient) {
    second_mass_matrix_plus_dt_times_stiffness_matrix.copyFrom(
      first_mass_matrix_plus_dt_times_stiffness_matrix);
  }

  Vector old_mass_matrix_time_old_solution(n,0.);
  old_stiffness_matrix->vmult(
    *rhs_times_dt_plus_matrix_times_old_solution,*solution[2]);
  rhs_times_dt_plus_matrix_times_old_solution->scale(-0.5*gamma*dt);
  old_mass_matrix->vmult(old_mass_matrix_time_old_solution,*solution[2]);
  (*rhs_times_dt_plus_matrix_times_old_solution)+=
    old_mass_matrix_time_old_solution;
  rhs_times_dt_plus_matrix_times_old_solution->add(0.5*gamma*dt,*rhs[2]);
  rhs_times_dt_plus_matrix_times_old_solution->add(0.5*gamma*dt,*rhs[1]);
  map<int,REAL> boundary_values;
  dof_handler.interpolateBoundaryValues(mapping,boundary_values,
    fqmap,middle_time);
  dof_handler.applyBoundaryValues(boundary_values,
    first_mass_matrix_plus_dt_times_stiffness_matrix,
    *rhs_times_dt_plus_matrix_times_old_solution,preserve_symmetry);
  dof_handler.applyBoundaryValues(boundary_values,*solution[1]);
  solver.solve(first_mass_matrix_plus_dt_times_stiffness_matrix,
    *solution[1],*rhs_times_dt_plus_matrix_times_old_solution,
    preconditioner);

  if (coefs_transient) {
    second_mass_matrix_plus_dt_times_stiffness_matrix.valueArray().
      initialize(0.);
    second_mass_matrix_plus_dt_times_stiffness_matrix.addScaled(1.,
      *mass_matrix[0]);
    second_mass_matrix_plus_dt_times_stiffness_matrix.addScaled(gamma*dt,
      *stiffness_matrix[0]);
  }

  rhs_times_dt_plus_matrix_times_old_solution->
    equ(-(1.-gamma)*(1.-gamma),old_mass_matrix_time_old_solution);
  middle_mass_matrix->
    vmultAdd(*rhs_times_dt_plus_matrix_times_old_solution,*solution[1]);
  rhs_times_dt_plus_matrix_times_old_solution->
    scale(1./(gamma*(2.-gamma)));
  rhs_times_dt_plus_matrix_times_old_solution->add((1.-gamma)*dt,*rhs[0]);
  dof_handler.interpolateBoundaryValues(mapping,boundary_values,
    fqmap,new_time);
  dof_handler.applyBoundaryValues(boundary_values,
    second_mass_matrix_plus_dt_times_stiffness_matrix,
    *rhs_times_dt_plus_matrix_times_old_solution,preserve_symmetry);
  dof_handler.applyBoundaryValues(boundary_values,*solution[0]);
  solver.solve(second_mass_matrix_plus_dt_times_stiffness_matrix,
    *solution[0],*rhs_times_dt_plus_matrix_times_old_solution,
    preconditioner);
  if (!tessellation_fixed) {
    delete rhs_times_dt_plus_matrix_times_old_solution;
    rhs_times_dt_plus_matrix_times_old_solution=0;
  }
  return TimeIntegrator::takeStep(dt);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
DeferredCorrectionTimeIntegrator::DeferredCorrectionTimeIntegrator(
TimeIntegrator &bi,int ordr,NumPtr<Vector*> &sol,NumPtr<Vector*> &rhs) : 
TimeIntegrator(bi.cntrl,bi.dof_handler,bi.mapping,bi.qmap,bi.fqmap,
bi.preconditioner,bi.solver,sol,rhs,bi.tessellation_fixed,
bi.preserve_symmetry,ordr+1),
base_integrator(bi),ord(ordr),integrals(ordr+1),lq(ordr+1),f(ordr+1),
epsilon(ordr+1),correction(ordr+1),My0(0) {
  const NumPtr<Point<1> > &points=lq.points();
  const NumPtr<REAL> &weights=lq.weights();
  f.initialize(0);
  epsilon.initialize(0);
  correction.initialize(0);
  if (tessellation_fixed) {
    My0=OPERATOR_NEW Vector(n_dofs,0.);
    for (int j=0;j<=ord;j++) {
      f[j]=OPERATOR_NEW Vector(n_dofs,0.);
      epsilon[j]=OPERATOR_NEW Vector(n_dofs,0.);
      correction[j]=OPERATOR_NEW Vector(n_dofs,0.);
    }
  }
  for (int k=0;k<=ord;k++) integrals[k].allocate(ord+1);
  REAL div=1./(points[ord][0]-points[0][0]);
  integrals[0].initialize(0.);
  for (int k=1;k<=ord;k++) { // integral from points[0] to points[k]
    REAL dt=(points[k][0]-points[0][0])*div;
    for (int j=0;j<=ord;j++) {
      REAL sum=0.;
      REAL denom=1.;
      for (int i=0;i<=ord;i++) {
        if (i!=j) denom*=points[j][0]-points[i][0];
      }
      for (int q=0;q<=ord;q++) {
        REAL t=points[0][0]+dt*(points[q][0]-points[0][0]);
        REAL num=1.;
        for (int i=0;i<=ord;i++) {
          if (i!=j) num*=t-points[i][0];
        }
        sum+=weights[q]*num/denom;
      }
      integrals[k][j]=sum*dt;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
DeferredCorrectionTimeIntegrator::~DeferredCorrectionTimeIntegrator() {
  if (tessellation_fixed) {
    delete My0; My0=0;
    for (int j=0;j<=ord;j++) {
      delete f[j]; f[j]=0;
      delete epsilon[j]; epsilon[j]=0;
      delete correction[j]; correction[j]=0;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TIME_INTEGRATOR_STATE DeferredCorrectionTimeIntegrator::takeStep(
REAL dt,const SparsityPattern *sp,NumPtr<Vector*> &solution,
NumPtr<Vector*> &rhs,NumPtr<SparseMatrix*> &mass_matrix,
NumPtr<SparseMatrix*> &stiffness_matrix) {
  REAL old_time=cntrl.getTime();
  REAL new_time=old_time+dt;
  swap(solution[0],solution[ord]);
  swap(rhs[0],rhs[ord]);
  int n=sp->nRows();
  const NumPtr<Point<1> > &points=lq.points();
  REAL frac=dt/(points[ord][0]-points[0][0]);
  if (coefs_transient) {
    CHECK_TEST(mass_matrix.getNumber()>=ord+1);
    CHECK_TEST(stiffness_matrix.getNumber()>=ord+1);
    swap(mass_matrix[0],mass_matrix[ord]);
    swap(stiffness_matrix[0],stiffness_matrix[ord]);
    if (tessellation_fixed) {
      CHECK_SAME(n_dofs,n);
//    avoid re-computing matrices at old time
      for (int i=0;i<ord;i++) {
        (*rhs[i])=0.;
        mass_matrix[i]->valueArray().initialize(0.);
        stiffness_matrix[i]->valueArray().initialize(0.);
      }
    } else {
      for (int i=0;i<=ord;i++) {
        if (mass_matrix[i]!=0) delete mass_matrix[i]; 
        mass_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
        if (stiffness_matrix[i]!=0) delete stiffness_matrix[i]; 
        stiffness_matrix[i]=OPERATOR_NEW SparseMatrix(sp);
        if (rhs[i]!=0) delete rhs[i]; 
        rhs[i]=OPERATOR_NEW Vector(n,0.);
      }
      CHECK_SAME(solution[ord]->size(),n);
      for (int i=0;i<ord;i++) {
        if (solution[i]!=0) delete solution[i]; 
        solution[i]=OPERATOR_NEW Vector(n,0.);
      }
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*mass_matrix[ord],*stiffness_matrix[ord],*rhs[ord]);
    }
    for (int s=1;s<=ord;s++) {
      REAL times=old_time+frac*(points[s][0]-points[0][0]);
      (*rhs[ord-s])=0.;
      dof_handler.assembleTransientLinearSystem(times,mapping,qmap,
        fqmap,*mass_matrix[ord-s],*stiffness_matrix[ord-s],
        *rhs[ord-s]);
    }
  } else {
    CHECK_TEST(mass_matrix.getNumber()>=1);
    CHECK_TEST(stiffness_matrix.getNumber()>=1);
    if (tessellation_fixed) {
      CHECK_SAME(n_dofs,n);
//    avoid computing any matrices
      (*rhs[ord])=0.;
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*rhs[ord]);
    } else {
      for (int i=0;i<=ord;i++) {
        if (rhs[i]!=0) delete rhs[i]; 
        rhs[i]=OPERATOR_NEW Vector(n,0.);
      }
      if (mass_matrix[0]!=0) delete mass_matrix[0]; 
      mass_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      if (stiffness_matrix[0]!=0) delete stiffness_matrix[0]; 
      stiffness_matrix[0]=OPERATOR_NEW SparseMatrix(sp);
      CHECK_SAME(solution[ord]->size(),n);
      for (int i=0;i<ord;i++) {
        if (solution[i]!=0) delete solution[i]; 
        solution[i]=OPERATOR_NEW Vector(n,0.);
      }
      (*rhs[ord])=0.;
      dof_handler.assembleTransientLinearSystem(old_time,mapping,qmap,
        fqmap,*mass_matrix[0],*stiffness_matrix[0],*rhs[ord]);
    }
    for (int s=1;s<=ord;s++) {
      REAL times=old_time+frac*(points[s][0]-points[0][0]);
      (*rhs[ord-s])=0.;
      dof_handler.assembleTransientLinearSystem(times,mapping,qmap,
        fqmap,*rhs[ord-s]);
    }
  }

  map<int,REAL> zero_boundary_values;
  REAL auld_time=old_time;
  for (int s=1;s<=ord;s++) {
//  tess_fix  coefs_vary
//  T           F       mass(1), stiffness(1) from t=0
//  T           T       mass(ord), stiffness(ord) 
//  F           F       mass(1), stiffness(1) from step 0
//  F           T       mass(ord), stiffness(ord)
    REAL base_dt=frac*(points[s][0]-points[s-1][0]);
    int offset=ord-s;
    NumPtr<Vector*> solutions(solution,offset);
    NumPtr<Vector*> rhss(rhs,offset);
    map<int,REAL> new_boundary_values;
    if (tessellation_fixed) {
      base_integrator.takeFirstDeferredCorrectionStep(auld_time,base_dt,sp,
        new_boundary_values,solutions,rhss,mass_matrix,stiffness_matrix);
    } else {
      NumPtr<SparseMatrix*> mass_matrixs(mass_matrix,
        coefs_transient ? offset : 0);
      NumPtr<SparseMatrix*> stiffness_matrixs(stiffness_matrix,
        coefs_transient ?  offset : 0);
      base_integrator.takeFirstDeferredCorrectionStep(auld_time,base_dt,sp,
        new_boundary_values,solutions,rhss,mass_matrixs,stiffness_matrixs);
    }
    if (s==ord) {
      for (map<int,REAL>::const_iterator bv_iterator=
      new_boundary_values.begin();bv_iterator!=new_boundary_values.end();
      bv_iterator++) {
        zero_boundary_values[bv_iterator->first]=0.;
      }
    }
    auld_time+=base_dt;
  }
  if (!tessellation_fixed) {
    My0=rhs[0]->clone();
    for (int j=0;j<=ord;j++) {
      f[j]=rhs[0]->clone();
      epsilon[j]=rhs[0]->clone();
      correction[j]=rhs[0]->clone();
    }
  }
  mass_matrix[coefs_transient ? ord : 0]->vmult(*My0,*solution[ord]);
  int biord=base_integrator.order();
  for (int k=1;k*biord<ord;k++) {
    for (int j=0;j<=ord;j++) {
      stiffness_matrix[coefs_transient ? j : 0]->
        vmult(*f[j],*solution[j]);
      f[j]->scale(-1.);
      f[j]->add(1.,*rhs[j]);
    }
    *epsilon[ord]=0.;
    *correction[ord]=0.;
    for (int q=1;q<=ord;q++) {
      Vector &eps=*epsilon[ord-q];
      mass_matrix[coefs_transient ? ord-q : 0]->
        vmult(eps,*solution[ord-q]);
      eps.scale(-1.);
      eps.add(1.,*My0);
      for (int j=0;j<-ord;j++) eps.add(integrals[q][j],*f[ord-j]);
      REAL base_dt=frac*(points[q][0]-points[q-1][0]);
      int offset=ord-q;
      NumPtr<Vector*> correctionq(correction,offset);
      NumPtr<Vector*> epsilonq(epsilon,offset);
      if (tessellation_fixed) {
        base_integrator.takeDeferredCorrectionStep(base_dt,sp,
          zero_boundary_values,correctionq,epsilonq,mass_matrix,
          stiffness_matrix);
      } else {
        NumPtr<SparseMatrix*> mass_matrixq(mass_matrix,
          coefs_transient ? offset : 0);
        NumPtr<SparseMatrix*> stiffness_matrixq(stiffness_matrix,
          coefs_transient ? offset : 0);
        base_integrator.takeDeferredCorrectionStep(base_dt,sp,
          zero_boundary_values,correctionq,epsilonq,mass_matrixq,
          stiffness_matrixq);
      }
    }
    for (int q=0;q<ord;q++) *solution[q]+=*correction[q];
  }
  if (!tessellation_fixed) {
    if (My0) delete My0; My0=0;
    for (int j=0;j<=ord;j++) {
      if (f[j]) delete f[j]; f[j]=0;
      if (epsilon[j]) delete epsilon[j]; epsilon[j]=0;
      if (correction[j]) delete correction[j]; correction[j]=0;
    }
  }
  return TimeIntegrator::takeStep(dt);
}
