#ifndef TIME_INTEGRATOR_H
#define TIME_INTEGRATOR_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <float.h>
#include <map>
#include "DofHandler.H"
#include "NumPtr.H"
#include "Precondition.H"
#include "SparseMatrix.H"
#include "Types.H"

enum TIME_INTEGRATOR_STATE { 
  TIME_INTEGRATOR_INITIALIZING=0,
  TIME_INTEGRATOR_STEPPING,
  TIME_INTEGRATOR_DONE};
ostream& operator<<(ostream&,TIME_INTEGRATOR_STATE);

class TimeControl {
  protected:
    int max_steps;
    int step;
    double max_time;
    double time;
    double tol;
    TIME_INTEGRATOR_STATE state;
  public:
    TimeControl(int ms=1,REAL mt=0.,REAL t=DBL_EPSILON) : 
      max_steps(ms),step(0),max_time(mt),time(0.),tol(t),
      state(TIME_INTEGRATOR_INITIALIZING) {}
    virtual ~TimeControl() {}
    virtual void takeStep(REAL dt) {
      time+=dt;
      step++;
    }
    virtual TIME_INTEGRATOR_STATE check() { 
      if (time<=0.) state=TIME_INTEGRATOR_INITIALIZING;
      else if (time>=max_time || step>=max_steps) {
        state=TIME_INTEGRATOR_DONE;
      } else state=TIME_INTEGRATOR_STEPPING;
      return state; 
    }
    int getStep() const { return step; }
    int maxSteps() const { return max_steps; }
    REAL getTime() const { return time; }
    REAL maxTime() const { return max_time; }
    int setMaxSteps(int newval) {
      int old=max_steps; max_steps=newval; return old;
    }
    double tolerance() const { return tol; }
    double setTolerance(double t) { double old=tol; tol=t; return old; }
    virtual void printOn(ostream &os) const {
      os << "\tTimeControl: max_steps = " << max_steps
         << "\n\tmax_time = " << max_time
         << "\n\ttime = " << time
         << "\n\ttol = " << tol
         << endl;
    }
};

class TimeIntegrator { 
  public:
    TimeControl &cntrl;
    const DofHandler &dof_handler;
    const Mapping &mapping;
    const map<int,Quadrature<SPACEDIM> > &qmap;
    map<int,Quadrature<SPACEDIM-1> > &fqmap;
    bool coefs_transient;
    bool tessellation_fixed;
    bool preserve_symmetry;
    int n_dofs;

    const SolverPreconditioner &preconditioner;
    Solver &solver;

    virtual TIME_INTEGRATOR_STATE takeStep(REAL dt) {
      cntrl.takeStep(dt);
      return cntrl.check();
    }
  public:
    TimeIntegrator(TimeControl&,const DofHandler&,const Mapping&,
      const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,const SolverPreconditioner&,
      Solver&,NumPtr<Vector*>&,NumPtr<Vector*>&,bool,bool,int);
    virtual ~TimeIntegrator();
    Solver& linearSolver() const { return solver; }
    TimeControl& control() const { return cntrl; }
    virtual TIME_INTEGRATOR_STATE takeStep(REAL,const SparsityPattern*,
    NumPtr<Vector*>&,NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,
    NumPtr<SparseMatrix*>&) {
      OBSOLETE("should be defined in derived class");
    }
    void l2ProjectionInitialData(const SparsityPattern*,NumPtr<Vector*>&,
      NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&);
    void ellipticProjectionInitialData(const SparsityPattern*,
      NumPtr<Vector*>&,NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,
      NumPtr<SparseMatrix*>&);
    virtual void takeFirstDeferredCorrectionStep(REAL,REAL,
    const SparsityPattern*,map<int,REAL>&,NumPtr<Vector*>&,
    NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&) {
      OBSOLETE("should be defined in derived class");
    }
    virtual void takeDeferredCorrectionStep(REAL,const SparsityPattern*,
    const map<int,REAL>&,NumPtr<Vector*>&,NumPtr<Vector*>&,
    NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&) {
      OBSOLETE("should be defined in derived class");
    }
    virtual int order() const=0;
    virtual void printOn(ostream &os) const {
      os << "TimeIntegrator:" << endl;
    }
};

class ForwardEulerTimeIntegrator : public TimeIntegrator {
  private:
    Vector *rhs_minus_stiffness_matrix_times_old_solution;
    Vector *mass_matrix_times_old_solution;
    virtual TIME_INTEGRATOR_STATE takeStep(REAL dt) {
      return TimeIntegrator::takeStep(dt);
    }
  public:
    ForwardEulerTimeIntegrator(TimeControl&,const DofHandler&,
      const Mapping&,const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,const SolverPreconditioner&,
      Solver&,NumPtr<Vector*>&,NumPtr<Vector*>&,bool,bool);
    virtual ~ForwardEulerTimeIntegrator();
    virtual TIME_INTEGRATOR_STATE takeStep(REAL,const SparsityPattern*,
      NumPtr<Vector*>&,NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,
      NumPtr<SparseMatrix*>&);
    virtual void takeFirstDeferredCorrectionStep(REAL,REAL,
      const SparsityPattern*,map<int,REAL>&,NumPtr<Vector*>&,
      NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&);
    virtual void takeDeferredCorrectionStep(REAL,const SparsityPattern*,
      const map<int,REAL>&,NumPtr<Vector*>&,NumPtr<Vector*>&,
      NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&);
    static int numberSolutions() { return 2; }
    virtual int order() const { return 1; }
    virtual void printOn(ostream &os) const {
      os << "ForwardEulerTimeIntegrator:" << endl;
      TimeIntegrator::printOn(os);
    }
};

class BackwardEulerTimeIntegrator : public TimeIntegrator {
  private:
    Vector *rhs_times_dt_plus_mass_matrix_times_old_solution;
    virtual TIME_INTEGRATOR_STATE takeStep(REAL dt) {
      return TimeIntegrator::takeStep(dt);
    }
  public:
    BackwardEulerTimeIntegrator(TimeControl&,const DofHandler&,
      const Mapping&,const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,const SolverPreconditioner&,
      Solver&,NumPtr<Vector*>&,NumPtr<Vector*>&,bool,bool);
    virtual ~BackwardEulerTimeIntegrator();
    virtual TIME_INTEGRATOR_STATE takeStep(REAL,const SparsityPattern*,
      NumPtr<Vector*>&,NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,
      NumPtr<SparseMatrix*>&);
    virtual void takeFirstDeferredCorrectionStep(REAL,REAL,
      const SparsityPattern*,map<int,REAL>&,NumPtr<Vector*>&,
      NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&);
    virtual void takeDeferredCorrectionStep(REAL,const SparsityPattern*,
      const map<int,REAL>&,NumPtr<Vector*>&,NumPtr<Vector*>&,
      NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&);
    static int numberSolutions() { return 2; }
    virtual int order() const { return 1; }
    virtual void printOn(ostream &os) const {
      os << "BackwardEulerTimeIntegrator:" << endl;
      TimeIntegrator::printOn(os);
    }
};

class CrankNicolsonTimeIntegrator : public TimeIntegrator {
  private:
    Vector *rhs_times_dt_plus_matrix_times_old_solution;
    virtual TIME_INTEGRATOR_STATE takeStep(REAL dt) {
      return TimeIntegrator::takeStep(dt);
    }
  public:
    CrankNicolsonTimeIntegrator(TimeControl&,const DofHandler&,
      const Mapping&,const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,const SolverPreconditioner&,
      Solver&,NumPtr<Vector*>&,NumPtr<Vector*>&,bool,bool);
    virtual ~CrankNicolsonTimeIntegrator();
    virtual TIME_INTEGRATOR_STATE takeStep(REAL,const SparsityPattern*,
      NumPtr<Vector*>&,NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,
      NumPtr<SparseMatrix*>&);
    virtual void takeFirstDeferredCorrectionStep(REAL,REAL,
      const SparsityPattern*,map<int,REAL>&,NumPtr<Vector*>&,
      NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&);
    virtual void takeDeferredCorrectionStep(REAL,const SparsityPattern*,
      const map<int,REAL>&,NumPtr<Vector*>&,NumPtr<Vector*>&,
      NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&);
    static int numberSolutions() { return 2; }
    virtual int order() const { return 2; }
    virtual void printOn(ostream &os) const {
      os << "CrankNicolsonTimeIntegrator:" << endl;
      TimeIntegrator::printOn(os);
    }
};

class TRBDFTimeIntegrator : public TimeIntegrator {
  private:
    Vector *rhs_times_dt_plus_matrix_times_old_solution;
    static const REAL gamma;
    virtual TIME_INTEGRATOR_STATE takeStep(REAL dt) {
      return TimeIntegrator::takeStep(dt);
    }
    virtual void takeDeferredCorrectionStep(REAL,const SparsityPattern*,
    const map<int,REAL>&,NumPtr<Vector*>&,NumPtr<Vector*>&,
    NumPtr<SparseMatrix*>&,NumPtr<SparseMatrix*>&) {
      OBSOLETE("not programmed");
    }
  public:
    TRBDFTimeIntegrator(TimeControl&,const DofHandler&,
      const Mapping&,const map<int,Quadrature<SPACEDIM> >&,
      map<int,Quadrature<SPACEDIM-1> >&,const SolverPreconditioner&,
      Solver&,NumPtr<Vector*>&,NumPtr<Vector*>&,bool,bool);
    virtual ~TRBDFTimeIntegrator();
    virtual TIME_INTEGRATOR_STATE takeStep(REAL,const SparsityPattern*,
      NumPtr<Vector*>&,NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,
      NumPtr<SparseMatrix*>&);
    static int numberSolutions() { return 3; }
    virtual int order() const { return 2; }
    virtual void printOn(ostream &os) const {
      os << "TRBDFTimeIntegrator:" << endl;
      TimeIntegrator::printOn(os);
    }
};

class DeferredCorrectionTimeIntegrator : public TimeIntegrator {
  private:
    TimeIntegrator &base_integrator;
    LobattoQuadrature<1> lq;
    NumPtr<Vector*> correction;
    NumPtr<Vector*> epsilon;
    NumPtr<Vector*> f;
    Vector *My0;
    int ord;
    NumPtr<NumPtr<REAL> > integrals;
    virtual TIME_INTEGRATOR_STATE takeStep(REAL dt) {
      return TimeIntegrator::takeStep(dt);
    }
  public:
    DeferredCorrectionTimeIntegrator(TimeIntegrator&,int,
      NumPtr<Vector*>&,NumPtr<Vector*>&);
    virtual ~DeferredCorrectionTimeIntegrator();
    virtual TIME_INTEGRATOR_STATE takeStep(REAL,const SparsityPattern*,
      NumPtr<Vector*>&,NumPtr<Vector*>&,NumPtr<SparseMatrix*>&,
      NumPtr<SparseMatrix*>&);
    static int numberSolutions(int ordr) { return ordr+1; }
    virtual int order() const { return ord; }
    virtual void printOn(ostream &os) const {
      os << "DeferredCorrectionTimeIntegrator: ord = " << ord << endl;
      TimeIntegrator::printOn(os);
    }
};

#endif
