// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Boundary.H"
#include "BoundaryCondition.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "HcurlFiniteElement.H"
#include "HdivFiniteElement.H"
//#include "Mapping.H"
#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tessellation.H"
#include "Tracer.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int rbc_count;
class RandomFunctionSystem : public FunctionSystem {
  private:
    int dpf;
    Array<2,REAL> values;
  public: RandomFunctionSystem(int d,int nc=1) : 
  FunctionSystem(nc),values(d,nc),dpf(d) {
    for (int j=0;j<values.size(1);j++) {
      for (int i=0;i<values.size(0);i++) values(i,j)=drand48();
    }
    rbc_count=0;
  }
  ~RandomFunctionSystem() {}
  void restart() { rbc_count=0; }
  REAL value(const Point<SPACEDIM> &p,int c) const { 
    rbc_count=rbc_count%dpf;
    return values(rbc_count++,c); 
  }
  void value(const Point<SPACEDIM> &p,Vector &v) const {
    if (v.size()!=n_components) v.reinit(n_components,true);
    rbc_count=rbc_count%dpf;
    for (int c=0;c<n_components;c++) v[c]=values(rbc_count,c);
    rbc_count++;
  }
  Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int c) const {
    return Tensor<1,SPACEDIM>();
  }
  void gradient(NumPtr<Tensor<1,SPACEDIM> > &g,const Point<SPACEDIM> &p)
  const {
    if (g.getNumber()!=n_components) {
      g.cleanup();
      g.allocate(n_components);
    }
    for (int c=0;c<n_components;c++) g[c].initialize(0.);
  }
  REAL divGrad(const Point<SPACEDIM>&,int) const { return 0.; }
  Vector divGrad(const Point<SPACEDIM>&) const {
    Vector v(n_components,0.);
    return v;
  }
};
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
/*
void plotCanonicalSupportPoints(FiniteElement &fe) {
//TRACER_CALL(t,"plotCanonicalSupportPoints");
#if (SPACEDIM==1)
  int v[2]={0,1};
  Interval e(0,v);
  const NumPtr<Point<SPACEDIM> > &csp=fe.canonicalSupportPoints(&e);

  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL winsize=0.5;
  XYGraphTool gt("plotCanonicalSupportPoints","x","y",0.,1.,-0.5,0.5,
    &cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (int i=0;i<csp.getNumber();i++) {
    char str[5];
    snprintf(str,5,"%i",i);
#ifdef DEBUG
//  cout << "\ti,csp,str = " << i << " " << csp[i] << " " << str << endl;
#endif
    gt.putString(csp[i][0],0.05,str);
  }
  gt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
#else
  OBSOLETE("not programmed");
#endif
}
*/
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void dofs(const FiniteElement &finite_element) {
//TRACER_CALL(t,"dofs");
  int degree=finite_element.getDegree();
  CHECK_SAME(finite_element.dofsPerVertex(),1);
  CHECK_SAME(finite_element.dofsPerLine()+2,degree+1);
  const map<int,int> &dpc_map=
    finite_element.dofsPerCellMap();
  for (map<int,int>::const_iterator dpc_iterator=
  dpc_map.begin();dpc_iterator!=dpc_map.end();dpc_iterator++) {
    int nv=dpc_iterator->first;
    CHECK_SAME(finite_element.dofsPerCell(nv),
               numberEquallySpacedLatticePoints(nv,degree));
#if (SPACEDIM==2)
    CHECK_SAME(finite_element.dofsPerCell(nv),
              finite_element.dofsPerQuad(nv)
              +facesPerShape(nv)*finite_element.dofsPerLine()+nv);
    CHECK_SAME(finite_element.dofsPerFace(2),
               finite_element.dofsPerLine()+2);
#endif
#if (SPACEDIM==3)
    if (nv==4) {
      CHECK_SAME(finite_element.dofsPerCell(nv),
                 finite_element.dofsPerHex(nv)
                +facesPerShape(nv)*finite_element.dofsPerQuad(3)
                +linesPerShape(nv)*finite_element.dofsPerLine()+nv);
      CHECK_SAME(finite_element.dofsPerFace(3),
                 finite_element.dofsPerQuad(3)
                 +finite_element.dofsPerLine()*3+3);
    } else if (nv==6) {
      CHECK_SAME(finite_element.dofsPerCell(nv),
                 finite_element.dofsPerHex(nv)
                +2*finite_element.dofsPerQuad(3)
                +3*finite_element.dofsPerQuad(4)
                +linesPerShape(nv)*finite_element.dofsPerLine()+nv);
      CHECK_SAME(finite_element.dofsPerFace(3),
                 finite_element.dofsPerQuad(3)
                 +finite_element.dofsPerLine()*3+3);
      CHECK_SAME(finite_element.dofsPerFace(4),
                 finite_element.dofsPerQuad(4)
                 +finite_element.dofsPerLine()*4+4);
    } else {
      CHECK_SAME(finite_element.dofsPerCell(nv),
                 finite_element.dofsPerHex(nv)
                +facesPerShape(nv)*finite_element.dofsPerQuad(4)
                +linesPerShape(nv)*finite_element.dofsPerLine()+nv);
      CHECK_SAME(finite_element.dofsPerFace(4),
                 finite_element.dofsPerQuad(4)
                 +finite_element.dofsPerLine()*4+4);
    }
#endif
  }
  CHECK_SAME(finite_element.maxDofsPerCell(),
             finite_element.dofsPerCell(D_CHOOSE(2,4,8)));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
/*
#if (SPACEDIM>1)
void boundaryValues(const SHAPE_CLASS *e,int f,const FiniteElement &fe) {
//TRACER_CALL(t,"boundaryValues");
  int dpf=fe.dofsPerFace(e,f);

  NumPtr<Point<SPACEDIM> > eslp;
  e->equallySpacedLatticePoints(fe.getDegree(),eslp);
  NumPtr<int> face_dofs;
  int nv=e->numberVertices();
  equallySpacedLatticePointIndicesOfFace(nv,f,fe.getDegree(),face_dofs);
  CHECK_SAME(dpf,face_dofs.getNumber());
  NumPtr<Point<SPACEDIM> > face_lattice_points(dpf);
  for (int q=0;q<face_lattice_points.getNumber();q++) {
    face_lattice_points[q]=eslp[face_dofs[q]];
  }
#ifdef DEBUG
//for (int i=0;i<face_lattice_points.getNumber();i++) {
//  cout << "\tface_lattice_points[" << i << "] = " 
//       << face_lattice_points[i] << endl;
//}
//for (int i=0;i<dpf;i++) {
//  cout << "\tface_dofs[" << i << "] = " << face_dofs[i] << endl;
//}
#endif
  CODIMENSION1_SHAPE_CLASS *face_element=0;
  RandomBoundaryCondition rbc(dpf,fe.nComponents());
  NumPtr<bool> component_mask;
  map<int,REAL> boundary_values;
  fe.setBoundaryValues(e,face_element,f,face_dofs,face_lattice_points,
    face_lattice_points,&rbc,component_mask,boundary_values);
#ifdef DEBUG
//for (int i=0;i<dpf;i++) {
//  cout << "\tboundary_values(" << face_dofs[i] << ") = ";
//  for (int c=0;c<fe.nComponents();c++) {
//    cout << boundary_values[face_dofs[i]+c*dpf] << endl;
//  }
//}
#endif

  Quadrature<SPACEDIM> quad(face_lattice_points);
  FiniteElementDataBase *fedb=fe.getFEData(nv,UPDATE_VALUES,quad);
  const Array<2,REAL> &sv=fedb->getShapeValues();
  NumPtr<int> nsof;
  fe.shapeFunctionFamily(e)->nonzeroShapesOnFace(e,f,nsof);
#ifdef DEBUG
//for (int j=0;j<sv.size(1);j++) {
//  for (int i=0;i<nsof.getNumber();i++) {
//    cout << "\tsv(" << nsof[i] << "," << j << ") = " << sv(nsof[i],j) 
//         << endl;
//  }
//}
#endif

  Point<SPACEDIM> pdummy;
  rbc.restart();
  for (int q=0;q<face_lattice_points.getNumber();q++) {
    Vector sum(fe.nComponents(),0.);
    for (int c=0;c<sum.size();c++) {
      for (int s=0;s<nsof.getNumber();s++) {
        sum[c]+=sv(nsof[s],q)*boundary_values[face_dofs[s]+c*dpf];
      }
    }
    Vector v;
    rbc.value(pdummy,v);
    REAL error=0.;
    REAL vmax=0.;
    for (int c=0;c<v.size();c++) {
      error=max(error,abs(sum[c]-v[c]));
      vmax=max(vmax,abs(v[c]));
    }
    if (error>256.*DBL_EPSILON*vmax) {
      cout << "\tface = " << f << " q = " << q << endl;
      cout << "\tboundary_values = [";
      for (int c=0;c<v.size();c++) cout << " " << v[c];
      cout << "], [";
      for (int c=0;c<v.size();c++) cout << " " << sum[c];
      cout << "] error = " << error/(DBL_EPSILON*vmax) << endl;
    }
  }
  delete fedb;
}
#endif
*/
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void prolongationCase(FiniteElement &finite_element,
const SHAPE_CLASS *e,const LEVEL_CLASS *level,int c,
const Point<SPACEDIM> &child_point,const Point<SPACEDIM> &parent_point) {
//TRACER_CALL(t,"prolongationCase");
  ShapeFunction<SPACEDIM> *sf=finite_element.shapeFunctionFamily(e);
  const Matrix &pm=finite_element.prolongationMatrix(e,c,level);
  int ns=sf->getNumber();
  NumPtr<REAL> cv;
  sf->computeValues(child_point,cv);
  NumPtr<REAL> pv;
  sf->computeValues(parent_point,pv);
#ifdef DEBUG
//cout << "\tc = " << c << endl;
//for (int p=0;p<ns;p++) {
//  cout << "\tprolongation(" << p << ",*) =";
//  for (int q=0;q<ns;q++) cout << " " << pm(p,q);
//  cout << endl;
//}
//cout << "\tchild_point = " << child_point << endl;
//cout << "\tparent_point = " << parent_point << endl;
//cout << "\tcv = ";
//for (int i=0;i<cv.getNumber();i++) cout << " " << cv[i];
//cout << endl;
//cout << "\tpv = ";
//for (int i=0;i<pv.getNumber();i++) cout << " " << pv[i];
//cout << endl;
#endif
  for (int p=0;p<ns;p++) {
    REAL sum=0.;
    for (int q=0;q<ns;q++) {
      sum+=pm(q,p)*cv[q];
    }
    REAL error=abs(pv[p]-sum);
    REAL den=max(abs(pv[p]),abs(sum));
//  if (den>0.) error/=den;
    if (error>524288.*DBL_EPSILON) {
      cout << "\tprolongation value(" << p << ") = " << pv[p] << " " 
           << sum << " error = " << error << " " << error/DBL_EPSILON 
           << endl;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void prolongation(FiniteElement &finite_element,const SHAPE_CLASS *e,
const LEVEL_CLASS *level) {
//TRACER_CALL(t,"prolongation");
  drand48();
  Point<SPACEDIM> child_point(D_DECL(drand48(),drand48(),drand48()));
#ifdef DEBUG
//cout << "\tchild_point = " << child_point << endl;
#endif
  for (int c=0;c<childrenPerShape();c++) {
    Point<SPACEDIM> parent_point;
    parent_point=
      parentCoordinateOfCanonical(e->numberVertices(),child_point,c);
    prolongationCase(finite_element,e,level,c,child_point,parent_point);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
/*
//this needs a DoFHandler to work easily: must set dofs on parent and
//  children before summing over shapeValues
#if (SPACEDIM>1)
void constraintMatrix(const Tessellation &tess,const FiniteElement &fe)
{
  LEVEL_CLASS *level=tess.coarsestLevel();
  for (bool doit=tess.firstActiveShape(level);doit;
  doit=tess.nextActiveShapeInLevel()) {
    if (tess.presentLevel()!=level) break;
    const SHAPE_CLASS *e=level->getShape();
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      CODIMENSION1_SHAPE_CLASS *face_shape=level->
        CODIMENSION1_LEVEL_CLASS::getShape((e->codimension1ShapeIndex(f)));
      Matrix m;
      fe.constraints(level,e,face_shape,m);
      Vector vf(m.size(0));
      Vector vc(m.size(1));
      for (int i=0;i<vc.size();i++) vc[i]=drand48();
      m.vmult(vf,vc);
      NumPtr<Point<SPACEDIM> > points(1);
      NumPtr<Point<SPACEDIM-1> > face_points(1);
#if (SPACEDIM==2)
      face_points[0][0]=drand48();
      e->projectToFace(f,face_points,points);
      Quadrature<SPACEDIM-1> face_quad(face_points);
      FiniteElementDataBase *fedb=fe.getFEData(
        face_shape->numberVertices(),UPDATE_VALUES,face_quad);
      Vector sumc(fe.nComponents(),0.);
      Vector sumf(fe.nComponents(),0.);
      for (int c=0;c<sumc.size();c++) {
        int iv=0;
        for (int fv=0;fv<face_shape->numberVertices();fv++) {
          sumc[c]+=fedb->shapeValue(face_shape->
            equallySpacedLatticePointIndexOfVertex(fv,fe.getDegree()),0)
            *vc[iv++];
        }
        NumPtr<int> face_interior_indices;
        equallySpacedLatticePointIndicesOfFaceInterior(e->numberVertices(),
          f,fe.getDegree(),face_interior_indices);
        for (int i=0;i<face_interior_indices.getNumber();i++) {
          sumc[c]+=fedb->shapeValue(indices[i],0)*vc[iv++];
        }
        for (int child=0;child<childrenPerCodimension1Shape();child++) {
          if (face_shape->numberInteriorChildVertices()>0) {
            sumf[c]=e->interiorChildVertexOnFace(f,child,child_vertex); 
          }
        }
      }
#else
      for (int d=0;d<SPACEDIM-1;d++) face_points[0][d]=drand48();
      e->projectToFace(f,face_points,points);
#endif
    }
  }
}
#endif
*/
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void shapeGrad(Tessellation &tess,FiniteElement &finite_element,
UPDATE_FLAGS update_flags) {
//TRACER_CALL(t,"shapeGrad");
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
    const SHAPE_CLASS *e=tess.presentShape();
    const LEVEL_CLASS *level=tess.presentLevel();
#ifdef DEBUG
//  cout << "\nelement = " << endl;
//  e->printOn(cout);
#endif

    ShapeFunction<SPACEDIM> *sf=finite_element.shapeFunctionFamily(e);
#ifdef DEBUG
//  cout << "\nsf = " << endl;
//  sf->printOn(cout);
#endif
    int nsf=sf->getNumber();
    REAL eps=sqrt(DBL_EPSILON);
    {
//    TRACER_CALL(t,"shapeGrad shapeGrad & shapeGradGrad");
      Point<SPACEDIM> cp(D_DECL(drand48(),drand48(),drand48()));
      NumPtr<REAL> v;
      NumPtr<Tensor<1,SPACEDIM> > g;
      NumPtr<Tensor<2,SPACEDIM> > j;
      if (sf->requiresShapeToComputeValues()) {
        sf->computeValuesFor(e,cp,v);
        sf->computeGradsFor(e,cp,g);
        sf->computeGradGradsFor(e,cp,j);
      } else {
        sf->computeValues(cp,v);
        sf->computeGrads(cp,g);
        sf->computeGradGrads(cp,j);
      }
#ifdef DEBUG
//    for (int s=0;s<v.getNumber();s++) {
//      cout << "\tv[" << s << "] = " << v[s] << endl;
//    }
//    for (int s=0;s<v.getNumber();s++) {
//      cout << "\tg[" << s << "] = " << g[s] << endl;
//    }
#endif
      for (int d=0;d<SPACEDIM;d++) {
        Point<SPACEDIM> cpp(cp);
        cpp[d]+=eps;
        NumPtr<REAL> vv;
        NumPtr<Tensor<1,SPACEDIM> > gg;
        if (sf->requiresShapeToComputeValues()) {
          sf->computeValuesFor(e,cpp,vv);
          sf->computeGradsFor(e,cpp,gg);
        } else {
          sf->computeValues(cpp,vv);
          sf->computeGrads(cpp,gg);
        }
#ifdef DEBUG
//      for (int s=0;s<v.getNumber();s++) {
//        cout << "\tvv[" << s << "] = " << v[s] << endl;
//      }
#endif
        for (int s=0;s<nsf;s++) {
          REAL exact=g[s][d];
          REAL deriv=(vv[s]-v[s])/eps;
          REAL error=abs(deriv-exact);
          REAL den=max(abs(exact),abs(deriv));
          if (den>0.) error/=den;
          if (error>eps*65536.) {
            cout << "\tshapeGrad(" << s << "," << cp << ")[" << d
                 << "] = " << exact << " " << deriv << ", error = "
                 << error << " " << error/eps << endl;
          }
/*
          for (int dd=0;dd<SPACEDIM;dd++) {
            REAL exact=j[s][d][dd];
            REAL deriv=(gg[s][dd]-g[s][dd])/eps;
            REAL error=abs(deriv-exact);
            REAL den=max(abs(exact),abs(deriv));
            if (den>0.) error/=den;
            if (error>eps*65536.) {
              cout << "\tshapeGradGrad(" << s << "," << cp << ")["
                   << d << "][" << dd << "] = " << exact << " "
                   << deriv << ", error = " << error 
                   << " " << error/eps << endl;
            }
          }
*/
        }
      }
    }
//
    {
//    TRACER_CALL(t,"shapeGrad getFEData");
      Point<SPACEDIM> cp(D_DECL(drand48(),drand48(),drand48()));
      Quadrature<SPACEDIM> q(cp);
      FiniteElementDataBase *fe_data=finite_element.getFEData(
        e->numberVertices(),update_flags,&q);
      FiniteElementDataBase *fixed=
        finite_element.fixShapeFunctionsFor(e,fe_data);
#ifdef DEBUG
//    for (int s=0;s<sf->getNumber();s++) {
//      cout << "\tv[" << s << "] = " << fe_data->shapeValue(s,0) 
//           << endl;
//    }
//    for (int s=0;s<sf->getNumber();s++) {
//      cout << "\tg[" << s << "] = " << fe_data->shapeDerivative(s,0) 
//           << endl;
//    }
#endif
      for (int d=0;d<SPACEDIM;d++) {
        Point<SPACEDIM> cpp(cp);
        cpp[d]+=eps;
        Quadrature<SPACEDIM> qq(cpp);
        FiniteElementDataBase *fe_data_data=finite_element.getFEData(
          e->numberVertices(),update_flags,&qq);
        FiniteElementDataBase *fixed_fixed=finite_element.
          fixShapeFunctionsFor(e,fe_data_data);
#ifdef DEBUG
//      for (int s=0;s<sf->getNumber();s++) {
//        cout << "\tvv[" << s << "] = " 
//             << fe_data_data->shapeValue(s,0) << endl;
//      }
#endif
        for (int s=0;s<nsf;s++) {
          REAL v=fe_data->shapeValue(s,0);
          REAL vv=fe_data_data->shapeValue(s,0);
          Tensor<1,SPACEDIM> g=fe_data->shapeDerivative(s,0);
          REAL exact=g[d];
          REAL deriv=(vv-v)/eps;
          REAL error=abs(deriv-exact);
          REAL den=max(abs(exact),abs(deriv));
          if (den>0.) error/=den;
          if (error>eps*32768.) {
            cout << "\tshapeDerivative(" << s << "," << cp << ")[" << d
                 << "] = " << exact << " " << deriv << ", error = "
                 << error << " " << error/eps << endl;
          }
/*
          Tensor<1,SPACEDIM> gg=fe_data_data->shapeDerivative(s,0);
          Tensor<2,SPACEDIM> j=fe_data->shapeSecondDerivative(s,0);
          for (int dd=0;dd<SPACEDIM;dd++) {
            REAL exact=j[d][dd];
            REAL deriv=(gg[dd]-g[dd])/eps;
            REAL error=abs(deriv-exact);
            REAL den=max(abs(exact),abs(deriv));
            if (den>0.) error/=den;
            if (error>eps*32768.) {
              cout << "\tshapeSecondDerivative(" << s << "," << cp 
                   << ")[" << d << "][" << dd << "] = " << exact << " "
                   << deriv << ", error = " << error 
                   << " " << error/eps << endl;
            }
          }
*/
        }
        if (fixed_fixed!=fe_data_data) delete fixed_fixed;
        delete fe_data_data;
      }
      if (fixed!=fe_data) delete fixed;
      delete fe_data;
    }
//
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void debugFE_Q(Tessellation &tess,FiniteElement &finite_element,
UPDATE_FLAGS update_flags) {
//TRACER_CALL(t,"debugFE_Q");
  dofs(finite_element);
//plotCanonicalSupportPoints(finite_element);
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
    const SHAPE_CLASS *e=tess.presentShape();
#if (SPACEDIM>1)
//  for (int f=0;f<e->numberCodimension1Shapes();f++) {
//    boundaryValues(e,f,finite_element);
//  }
#endif
    const LEVEL_CLASS *level=tess.presentLevel();
    prolongation(finite_element,e,level);
//  cout << "\nelement = " << endl;
//  e->printOn(cout);
    int nsf=finite_element.shapeFunctionFamily(e)->getNumber();
    {
//    TRACER_CALL(t,"debugFE_Q canonicalSupportPoint");
      NumPtr<Point<SPACEDIM> > csp;
      e->equallySpacedLatticePoints(finite_element.getDegree(),csp);
#ifdef DEBUG
//    for (int q=0;q<csp.getNumber();q++) {
//      cout << "\tcsp[" << q << "] = " << csp[q] << endl;
//    }
#endif
      for (int q=0;q<nsf;q++) {
        NumPtr<REAL> v;
        finite_element.shapeFunctionFamily(e)->computeValues(csp[q],v);
#ifdef DEBUG
//      cout << "\n\tcsp[" << q << "] = " << csp[q] << endl;
//      for (int s=0;s<v.getNumber();s++) {
//        cout << "\tv[" << s << "] = " << v[s] << endl;
//      }
#endif
        for (int s=0;s<nsf;s++) {
          REAL error=(s==q ? abs(v[s]-1.) : abs(v[s]));
          if (error>DBL_EPSILON*512.) {
            cout << "\tshapeValue(" << s << "," << csp[q] << ") = " 
                 << v[s] << ", error = " << error  
                 << " " << error/DBL_EPSILON << endl;
          }
        }
      }
      NumPtr<int> renumber;
      e->reorderEquallySpacedLatticePoints(finite_element.getDegree(),
        renumber);
#ifdef DEBUG
//    for (int i=0;i<renumber.getNumber();i++) {
//      cout << "\trenumber[" << i << "] = " << renumber[i] << endl;
//    }
#endif
      Quadrature<SPACEDIM> quadrature(csp);
      FiniteElementDataBase *fedb=finite_element.
        getFEData(e->numberVertices(),update_flags,&quadrature);
#ifdef DEBUG
//    for (int q=0;q<nsf;q++) {
//      for (int s=0;s<nsf;s++) {
//        cout << "\tfedb->shapeValue[" << s << "," << q << "] = " 
//             << fedb->shapeValue(s,q) << endl;
//      }
//    }
#endif
      FiniteElementDataBase *fixed=
        finite_element.fixShapeFunctionsFor(e,fedb);
#ifdef DEBUG
//    for (int q=0;q<nsf;q++) {
//      for (int s=0;s<nsf;s++) {
//        cout << "\tfixed->shapeValue[" << s << "," << q << "] = " 
//             << fixed->shapeValue(s,q) << endl;
//      }
//    }
#endif
      for (int q=0;q<nsf;q++) {
        for (int s=0;s<nsf;s++) {
          REAL val=fixed->shapeValue(s,q);
          REAL error=(s==renumber[q] ? abs(val-1.) : abs(val));
          if (error>DBL_EPSILON*512.) {
            cout << "\tshapeValue(" << q << "," << csp[q] << ") = " 
                 << val << ", error = " << error  
                 << " " << error/DBL_EPSILON << endl;
          }
        }
      }
      delete fedb;
      delete fixed;
    }
  }
  shapeGrad(tess,finite_element,update_flags);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void debugFE_H(Tessellation &tess,FE_H &finite_element,
UPDATE_FLAGS update_flags) {
//TRACER_CALL(t,"debugFE_H");
  dofs(finite_element);
#if (SPACEDIM==1)
  GaussianQuadrature<SPACEDIM> gq(finite_element.getDegree());
  const NumPtr<Point<SPACEDIM> > &points=gq.points();
  const NumPtr<REAL> &weights=gq.weights();
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
    const SHAPE_CLASS *e=tess.presentShape();
    ShapeFunction<SPACEDIM> *sf=finite_element.shapeFunctionFamily(e);
    CHECK_POINTER(dynamic_cast<const HierarchicalIntervalPolynomials*>(sf));
#if (SPACEDIM>1)
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      boundaryValues(e,finite_element);
    }
#endif
    const LEVEL_CLASS *level=tess.presentLevel();
    prolongation(finite_element,e,level);
//  cout << "\nelement = " << endl;
//  e->printOn(cout);
    int nsf=sf->getNumber();
    for (int ss=0;ss<e->numberVertices();ss++) {
      NumPtr<REAL> v;
      sf->computeValues(Interval::canonicalVertex(ss),v);
      for (int s=0;s<nsf;s++) {
        REAL error=(s==ss ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*64.) {
          cout << "\tvertex value(" << s << "," << ss << ") = " << v[s]
               << ", error = " << error << endl;
        }
      }
    }
    for (int ss=e->numberVertices();ss<nsf;ss++) {
      NumPtr<REAL> v(nsf);
      v.initialize(0.);
      for (int q=0;q<points.getNumber();q++) {
        NumPtr<Tensor<1,SPACEDIM> > g;
        sf->computeGrads(points[q],g);
        for (int s=0;s<nsf;s++) {
          v[s]+=g[s][0]*g[ss][0]*weights[q];
        }
      }
      for (int s=0;s<nsf;s++) {
        REAL error=(s==ss ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*64.) {
          cout << "\tquadrature(" << s << "," << ss << ") = " << v[s]
               << ", error = " << error << endl;
        }
      }
    }
  }
#else
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
    const SHAPE_CLASS *e=tess.presentShape();
#ifdef DEBUG
//  cout << "\n\te = " << e->elementDataIndex() << endl;
#endif
    const LEVEL_CLASS *level=tess.presentLevel();
    ShapeFunction<SPACEDIM> *sf=finite_element.shapeFunctionFamily(e);
    int nsf=sf->getNumber();
    for (int ss=0;ss<e->numberVertices();ss++) {
      NumPtr<REAL> v;
# if (SPACEDIM==2)
      if (dynamic_cast<const Quadrilateral*>(e)) {
        CHECK_POINTER(dynamic_cast<const HierarchicalQuadrilateralPolynomials*>(sf));
        sf->computeValues(Quadrilateral::canonicalVertex(ss),v);
      } else {
        sf->computeValues(Triangle::canonicalVertex(ss),v);
        CHECK_POINTER(dynamic_cast<const HierarchicalTrianglePolynomials*>(sf));
      }
# else
      if (dynamic_cast<const Hexahedron*>(e)) {
        CHECK_POINTER(dynamic_cast<const HierarchicalHexahedronPolynomials*>(sf));
        sf->computeValuesFor(e,Hexahedron::canonicalVertex(ss),v);
      } else if (dynamic_cast<const Tetrahedron*>(e)) {
#ifdef DEBUG
//      cout << "\tss = " << ss << " "
//           << Tetrahedron::canonicalVertex(ss) << endl;
#endif
        CHECK_POINTER(dynamic_cast<const HierarchicalTetrahedronPolynomials*>(sf));
        if (sf->requiresShapeToComputeValues()) {
          sf->computeValuesFor(e,Tetrahedron::canonicalVertex(ss),v);
        } else {
          sf->computeValues(Tetrahedron::canonicalVertex(ss),v);
        }
      } else {
        CHECK_POINTER(dynamic_cast<const HierarchicalPrismPolynomials*>(sf));
        if (sf->requiresShapeToComputeValues()) {
          sf->computeValuesFor(e,Prism::canonicalVertex(ss),v);
        } else {
          sf->computeValues(Prism::canonicalVertex(ss),v);
        }
      }
#ifdef DEBUG
//    for (int s=0;s<v.getNumber();s++) {
//      cout << "\tv[" << s << "] = " << v[s] << endl;
//    }
#endif
# endif
      for (int s=0;s<nsf;s++) {
        REAL error=(s==ss ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*64.) {
          cout << "\tvertex value(" << s << "," << ss << ") = " << v[s]
               << ", error = " << error << endl;
        }
      }
    }
  }
#endif
  shapeGrad(tess,finite_element,update_flags);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif

    Point<SPACEDIM> low(D_DECL(0.,0.,0.));
    Point<SPACEDIM> edge[SPACEDIM];
      edge[0]=Point<SPACEDIM>(D_DECL(1.,0.,0.));
#if (SPACEDIM>1)
      edge[1]=Point<SPACEDIM>(D_DECL(0.,1.,0.));
#endif
#if (SPACEDIM==3)
      edge[2]=Point<SPACEDIM>(0.,0.,1.);
#endif
    int number_elements[SPACEDIM];
    for (int d=0;d<SPACEDIM;d++) number_elements[d]=2;
    bool set_face_ids=true;

    UPDATE_FLAGS update_flags( UPDATE_VALUES | UPDATE_GRADIENTS
      | UPDATE_SECOND_DERIVATIVES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES 
      | UPDATE_NORMAL_VECTORS | UPDATE_MAPPING_JACOBIAN_INVERSE);
//
    {
//    TRACER_CALL(t,"generateParallelpiped");
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpiped(tess,low,edge,
        number_elements,set_face_ids,boundary,boundary_condition);
/*
      {
        TRACER_CALL(t,"FE_Q");
        for (int degree=1;degree<=D_CHOOSE(6,6,3);degree++) 
//      int degree=2;
        {
          cout << "\tdegree = " << degree << endl;
          FE_Q finite_element(degree);
          debugFE_Q(tess,finite_element,update_flags);
        }
      }
*/
//
      {
        TRACER_CALL(t,"FE_L");
        for (int degree=1;degree<=7;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          debugFE_Q(tess,finite_element,update_flags);
        }
      }
//
//
      {
        TRACER_CALL(t,"FE_H");
        for (int degree=1;degree<=7;degree++) 
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          debugFE_H(tess,finite_element,update_flags);
        }
      }
//    there is not much we can do to test RTFiniteElement,
//      BDFMFiniteElement, Nedelec1FiniteElement or Nedelec2FiniteElement
//    the tests would reproduce the duality tests in
//      testTensorShapeFunctions.C
      {
        TRACER_CALL(t,"RTFiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          RTFiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"BDFMFiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          BDFMFiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"Nedelec1FiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec1FiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"Nedelec2FiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec2FiniteElement finite_element(degree);
        }
      }
//
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
#if (SPACEDIM==2)
//
    {
      TRACER_CALL(t,"generateParallelpipedTriangles");
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpipedTriangles(tess,
        low,edge,number_elements,set_face_ids,boundary,
        boundary_condition);

/*
      {
        TRACER_CALL(t,"FE_Q");
//      for (int degree=1;degree<=4;degree++) 
        int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_Q finite_element(degree);
          debugFE_Q(tess,finite_element,update_flags);
        }
      }
*/
//
      {
        TRACER_CALL(t,"FE_L");
        for (int degree=1;degree<=10;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          debugFE_Q(tess,finite_element,update_flags);
        }
      }
//
//
      {
        TRACER_CALL(t,"FE_H");
        for (int degree=1;degree<=10;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          debugFE_H(tess,finite_element,update_flags);
        }
      }
/*
      {
        TRACER_CALL(t,"RTFiniteElement");
        for (int degree=1;degree<=7;degree++) {
          cout << "\tdegree = " << degree << endl;
          RTFiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"BDFMFiniteElement");
        for (int degree=1;degree<=7;degree++) {
          cout << "\tdegree = " << degree << endl;
          BDFMFiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"Nedelec1FiniteElement");
        for (int degree=1;degree<=7;degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec1FiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"Nedelec2FiniteElement");
        for (int degree=1;degree<=7;degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec2FiniteElement finite_element(degree);
        }
      }
*/

      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
//
    {
//    one quadrilateral and one triangle, with random vertex ordering
      PermutationGenerator pg(3);
      while (pg.hasMore()) {
        const NumPtr<int> &perm=pg.getNext();
        NumPtr<Point<SPACEDIM> > vertices(5);
        vertices[4]=Point<SPACEDIM>(0.,0.);
        vertices[3]=Point<SPACEDIM>(1.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,.5);
        vertices[perm[1]]=Point<SPACEDIM>(1.,.5);
        vertices[perm[0]]=Point<SPACEDIM>(.5,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(4);
        cells[0].vertex(0)=4;
        cells[0].vertex(1)=3;
        cells[0].vertex(2)=perm[2];
        cells[0].vertex(3)=perm[1];
        cells[1].allocate(3);
        cells[1].vertex(0)=perm[2];
        cells[1].vertex(1)=perm[1];
        cells[1].vertex(2)=perm[0];
        Tessellation tess;
        BoundaryData *subcell_data=tess.create(vertices,cells);
        Point<2> centroid((vertices[4]+vertices[3]
          +vertices[perm[2]]+vertices[perm[1]])*0.25);
        NumPtr<BoundaryCondition*> boundary_condition(5);
        NumPtr<Boundary*> boundary(5);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[2]]);
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[1]],
          vertices[3]);
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[1]]);
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[2]],
          vertices[4]);
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[3],
          vertices[4]);
        for (int f=0;f<5;f++) {
          boundary_condition[f]=
            OPERATOR_NEW BoundaryCondition(boundary[f]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<2> center=face_element->center(vertices);
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                break;
              }
            }
          }
        }
        tess.synchronize(subcell_data);
        cout << "\tpermutation = " << perm[0] << " " << perm[1] << " "
             << perm[2] << endl;
/*
        {
          TRACER_CALL(t,"FE_Q");
          for (int degree=1;degree<=4;degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            FE_Q finite_element(degree);
            debugFE_Q(tess,finite_element,update_flags);
          }
        }
*/
        {
          TRACER_CALL(t,"FE_L");
          for (int degree=1;degree<=10;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
            debugFE_Q(tess,finite_element,update_flags);
          }
        }
//
//
        {
          TRACER_CALL(t,"FE_H");
          for (int degree=1;degree<=10;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
            debugFE_H(tess,finite_element,update_flags);
          }
        }
/*
        {
          TRACER_CALL(t,"RTFiniteElement");
          for (int degree=1;degree<=7;degree++) {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"BDFMFiniteElement");
          for (int degree=1;degree<=7;degree++) {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"Nedelec1FiniteElement");
          for (int degree=1;degree<=7;degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"Nedelec2FiniteElement");
          for (int degree=1;degree<=7;degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
          }
        }
*/
        delete subcell_data; subcell_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
#endif
#if (SPACEDIM==3)
//
    {
      TRACER_CALL(t,"generateParallelpipedTetrahedra");
      Tessellation tess;
      int ne[SPACEDIM];
      for (int d=0;d<SPACEDIM;d++) ne[d]=1;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpipedTetrahedra(tess,
        low,edge,ne,set_face_ids,boundary,boundary_condition);

/*
      {
        TRACER_CALL(t,"FE_Q");
        for (int degree=1;degree<=D_CHOOSE(4,4,3);degree++) 
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_Q finite_element(degree);
          debugFE_Q(tess,finite_element,update_flags);
        }
      }
*/
//
      {
        TRACER_CALL(t,"FE_L");
        for (int degree=1;degree<=7;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          debugFE_Q(tess,finite_element,update_flags);
        }
      }
//
//
      {
        TRACER_CALL(t,"FE_H");
        for (int degree=1;degree<=7;degree++) 
//      int degree=2;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          debugFE_H(tess,finite_element,update_flags);
        }
      }
/*
      {
        TRACER_CALL(t,"RTFiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          RTFiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"BDFMFiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          BDFMFiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"Nedelec1FiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec1FiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"Nedelec2FiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec2FiniteElement finite_element(degree);
        }
      }
*/
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
//
    for (int d=0;d<SPACEDIM;d++) 
//  int d=1;
    {
      TRACER_CALL(t,"generateParallelpipedPrisms");
      Tessellation tess;
      int ne[SPACEDIM];
      for (int i=0;i<SPACEDIM;i++) ne[i]=1;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpipedPrisms(tess,
        low,edge,ne,d,set_face_ids,boundary,boundary_condition);

/*
      {
        TRACER_CALL(t,"FE_Q");
        for (int degree=1;degree<=D_CHOOSE(4,4,3);degree++) 
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_Q finite_element(degree);
          debugFE_Q(tess,finite_element,update_flags);
        }
      }
*/
//
      {
        TRACER_CALL(t,"FE_L");
        for (int degree=1;degree<=7;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          debugFE_Q(tess,finite_element,update_flags);
        }
      }
//
//
      {
        TRACER_CALL(t,"FE_H");
        for (int degree=1;degree<=7;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          debugFE_H(tess,finite_element,update_flags);
        }
      }
/*
      {
        TRACER_CALL(t,"RTFiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          RTFiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"BDFMFiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          BDFMFiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"Nedelec1FiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec1FiniteElement finite_element(degree);
        }
      }
      {
        TRACER_CALL(t,"Nedelec2FiniteElement");
        for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec2FiniteElement finite_element(degree);
        }
      }
*/

      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
//
    {
//    one prism and one tetrahedron, with random vertex ordering
      PermutationGenerator pg(4);
      while (pg.hasMore()) {
        const NumPtr<int> &perm=pg.getNext();
        NumPtr<Point<SPACEDIM> > vertices(7);
        vertices[4]=Point<SPACEDIM>(0.,0.,0.);
        vertices[5]=Point<SPACEDIM>(1.,0.,0.);
        vertices[6]=Point<SPACEDIM>(0.,1.,0.);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,.5);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,.5);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,.5);
        vertices[perm[3]]=Point<SPACEDIM>(0.,0.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(6);
        cells[0].vertex(0)=4;
        cells[0].vertex(1)=5;
        cells[0].vertex(2)=6;
        cells[0].vertex(3)=perm[0];
        cells[0].vertex(4)=perm[1];
        cells[0].vertex(5)=perm[2];
        cells[1].allocate(4);
        cells[1].vertex(0)=perm[0];
        cells[1].vertex(1)=perm[1];
        cells[1].vertex(2)=perm[2];
        cells[1].vertex(3)=perm[3];
        Tessellation tess;
        BoundaryData *subcell_data=tess.create(vertices,cells);
        Point<3> centroid((vertices[perm[0]]+vertices[perm[1]]
          +vertices[perm[2]]+vertices[perm[3]])*0.25);
        NumPtr<Boundary*> boundary(5);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
          vertices[6],vertices[perm[3]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
          vertices[5],vertices[perm[3]]); // left
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[4],vertices[5],vertices[6]); // bottom
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[5],vertices[perm[1]],vertices[6]); // right front
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[1]],vertices[perm[2]],vertices[perm[3]]); 
          // right top
        NumPtr<BoundaryCondition*> boundary_condition(5);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
            CHECK_TEST(found);
          }
        }
        tess.synchronize(subcell_data);
        cout << "\tpermutation = " << perm[0] << " " << perm[1] << " "
             << perm[2] << " " << perm[3] << endl;
/*
        {
          TRACER_CALL(t,"FE_Q");
          for (int degree=1;degree<=3;degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            FE_Q finite_element(degree);
            debugFE_Q(tess,finite_element,update_flags);
          }
        }
*/
//
        {
          TRACER_CALL(t,"FE_L");
          for (int degree=1;degree<=7;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
            debugFE_Q(tess,finite_element,update_flags);
          }
        }
//
//
        {
          TRACER_CALL(t,"FE_H");
          for (int degree=1;degree<=7;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
            debugFE_H(tess,finite_element,update_flags);
          }
        }
/*
        {
          TRACER_CALL(t,"RTFiniteElement");
          for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"BDFMFiniteElement");
          for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"Nedelec1FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"Nedelec2FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
          }
        }
*/
        delete subcell_data; subcell_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
//
    {
//    one prism and one hexahedron, with random vertex ordering
//    PermutationGenerator pg(6);
//    while (pg.hasMore())
      for (int count=0;count<50;count++)
      {
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(6);
        RandomPermutation(perm);

        NumPtr<Point<SPACEDIM> > vertices(10);
        vertices[6]=Point<SPACEDIM>(0.,0.,0.);
        vertices[7]=Point<SPACEDIM>(1.,0.,0.);
        vertices[8]=Point<SPACEDIM>(0.,1.,0.);
        vertices[9]=Point<SPACEDIM>(1.,1.,0.);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,.5);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,.5);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,.5);
        vertices[perm[3]]=Point<SPACEDIM>(1.,1.,.5);
        vertices[perm[4]]=Point<SPACEDIM>(0.,0.,1.);
        vertices[perm[5]]=Point<SPACEDIM>(0.,1.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(8);
        cells[0].vertex(0)=6;
        cells[0].vertex(1)=7;
        cells[0].vertex(2)=8;
        cells[0].vertex(3)=9;
        cells[0].vertex(4)=perm[0];
        cells[0].vertex(5)=perm[1];
        cells[0].vertex(6)=perm[2];
        cells[0].vertex(7)=perm[3];
        cells[1].allocate(6);
        cells[1].vertex(0)=perm[0];
        cells[1].vertex(1)=perm[4];
        cells[1].vertex(2)=perm[1];
        cells[1].vertex(3)=perm[2];
        cells[1].vertex(4)=perm[5];
        cells[1].vertex(5)=perm[3];
        Tessellation tess;
        BoundaryData *subcell_data=tess.create(vertices,cells);
        Point<3> centroid(0.5,0.5,0.5);
        NumPtr<Boundary*> boundary(6);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[6],
          vertices[perm[4]],vertices[8]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[7],
          vertices[perm[1]],vertices[9]); // front
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[6],
          vertices[7],vertices[perm[4]]); // left
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[8],vertices[9],vertices[perm[5]]);// right
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[6],vertices[7],vertices[8]); // bottom
        boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[1]],vertices[perm[3]],vertices[perm[4]]); // top
        NumPtr<BoundaryCondition*> boundary_condition(6);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
#ifdef DEBUG
            if (!found) {
              cout << "\tface_element:" << endl;
              face_element->printOn(cout);
              cout << "\tcenter = " << center << endl;
              for (int i=0;i<boundary.getNumber();i++) {
                cout << "\tboundary[" << i << "]: constraint = " 
                     << boundary[i]->constraint(center) << endl;
                boundary[i]->printOn(cout);
              }
            }
#endif
            CHECK_TEST(found);
          }
        }
        tess.synchronize(subcell_data);
        cout << "\tpermutation[" << count << "] = " << perm[0] << " " 
             << perm[1] << " " << perm[2] << " " << perm[3] << " " 
             << perm[4] << " " << perm[5] << endl;
/*
        {
          TRACER_CALL(t,"FE_Q");
          for (int degree=1;degree<=3;degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            FE_Q finite_element(degree);
            debugFE_Q(tess,finite_element,update_flags);
          }
        }
*/
//
        {
          TRACER_CALL(t,"FE_L");
          for (int degree=1;degree<=7;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
            debugFE_Q(tess,finite_element,update_flags);
          }
        }
//
//
        {
          TRACER_CALL(t,"FE_H");
          for (int degree=1;degree<=4;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
            debugFE_H(tess,finite_element,update_flags);
          }
        }
/*
        {
          TRACER_CALL(t,"RTFiniteElement");
          for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"BDFMFiniteElement");
          for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"Nedelec1FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
          }
        }
        {
          TRACER_CALL(t,"Nedelec2FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(7,7,3);degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
          }
        }
*/
        delete subcell_data; subcell_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
#endif
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
