// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Palette.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tessellation.H"
#include "Tracer.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#endif

void richardsonShapeDerivs(int nmax,Mapping &mapping,const SHAPE_CLASS *e,
const Point<SPACEDIM> &x,const Array<2,REAL> &f,
Array<2,Tensor<1,SPACEDIM> > &fp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_VALUES;
  nmax=max(1,nmax);
  for (int i=0;i<SPACEDIM;i++) {
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.size(0));
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[i]+=h;
      Quadrature<SPACEDIM> q(xnew);
      MappingDataBase *mdnew=
        mapping.getMappingData(e->numberVertices(),update_flags,&q);
      Array<2,REAL> fnew(mdnew->nShapeFunctions(),1);
      for (int s=0;s<mdnew->nShapeFunctions();s++) {
        fnew(s,0)=mdnew->shapeValue(s,0); 
        table[nmax-n][s][i]=(fnew(s,0)-f(s,0))/h;
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          table[m][s][i]=table[m-1][s][i]
                        +(table[m-1][s][i]-table[m][s][i])/(p-1.);
        }
        fp(s,0)[i]=table[nmax-1][s][i];
      }
      delete mdnew;
    }
  }
}

void richardsonShapeSecondDerivs(int nmax,Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,
const Array<2,Tensor<1,SPACEDIM> > &fp,
Array<2,Tensor<2,SPACEDIM> > &fpp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_GRADIENTS;
  nmax=max(1,nmax);
  for (int j=0;j<SPACEDIM;j++) {
    for (int i=0;i<SPACEDIM;i++) {
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.size(0));
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
        Quadrature<SPACEDIM> q(xnew);
        MappingDataBase *mdnew=
          mapping.getMappingData(e->numberVertices(),update_flags,&q);
        Array<2,Tensor<1,SPACEDIM> > fpnew(mdnew->nShapeFunctions(),1);
        for (int s=0;s<mdnew->nShapeFunctions();s++) {
          fpnew(s,0)=mdnew->shapeDerivative(s,0); 
          table[nmax-n][s][i][j]=(fpnew(s,0)[i]-fp(s,0)[i])/h;
          REAL p=2.;
          for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
            table[m][s][i][j]=table[m-1][s][i][j]
                          +(table[m-1][s][i][j]-table[m][s][i][j])/(p-1.);
          }
          fpp(s,0)[i][j]=table[nmax-1][s][i][j];
        }
        delete mdnew;
      }
    }
  }
}

void richardsonHigherOrderShapeDerivs(int nmax,Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,const Array<2,REAL> &f,
Array<2,Tensor<1,SPACEDIM> > &fp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_VALUES;
  nmax=max(1,nmax);
  for (int i=0;i<SPACEDIM;i++) {
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.size(0));
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[i]+=h;
      Quadrature<SPACEDIM> q(xnew);
      MappingDataBase *mdnew=
        mapping.getMappingData(e->numberVertices(),update_flags,&q);
      Array<2,REAL> fnew(mdnew->nShapeFunctions(),1);
      for (int s=0;s<mdnew->nShapeFunctions();s++) {
        fnew(s,0)=mdnew->higherOrderShapeValue(s,0); 
        table[nmax-n][s][i]=(fnew(s,0)-f(s,0))/h;
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          table[m][s][i]=table[m-1][s][i]
                        +(table[m-1][s][i]-table[m][s][i])/(p-1.);
        }
        fp(s,0)[i]=table[nmax-1][s][i];
      }
      delete mdnew;
    }
  }
}

void richardsonHigherOrderShapeSecondDerivs(int nmax,Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,
const Array<2,Tensor<1,SPACEDIM> > &fp,
Array<2,Tensor<2,SPACEDIM> > &fpp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_GRADIENTS;
  nmax=max(1,nmax);
  for (int j=0;j<SPACEDIM;j++) {
    for (int i=0;i<SPACEDIM;i++) {
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.size(0));
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
        Quadrature<SPACEDIM> q(xnew);
        MappingDataBase *mdnew=
          mapping.getMappingData(e->numberVertices(),update_flags,&q);
        Array<2,Tensor<1,SPACEDIM> > fpnew(mdnew->nShapeFunctions(),1);
        for (int s=0;s<mdnew->nShapeFunctions();s++) {
          fpnew(s,0)=mdnew->higherOrderShapeDerivative(s,0); 
          table[nmax-n][s][i][j]=(fpnew(s,0)[i]-fp(s,0)[i])/h;
          REAL p=2.;
          for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
            table[m][s][i][j]=table[m-1][s][i][j]
                          +(table[m-1][s][i][j]-table[m][s][i][j])/(p-1.);
          }
          fpp(s,0)[i][j]=table[nmax-1][s][i][j];
        }
        delete mdnew;
      }
    }
  }
}

void debugMapping(Mapping &mapping,Tessellation &tess) {
//TRACER_CALL(t,"debugMapping");
//Palette pal;
//XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
//REAL winsize=0.5;
//XYGraphTool gt("testMapping1","x","y",0.,1.,0.,1.,&cmap,0,winsize);
//REAL pw=gt.getPixelWidth();
  UPDATE_FLAGS update_flags(UPDATE_Q_POINTS | UPDATE_JxW_VALUES
    | UPDATE_NORMAL_VECTORS | UPDATE_MAPPING_JACOBIAN_INVERSE |
    UPDATE_MAPPING_SECOND_DERIVATIVE);
  NewtonCotesQuadrature<SPACEDIM> ncq(2);
  NumPtr<Point<SPACEDIM> > xq(ncq.numberPoints());
  NumPtr<REAL> jxwv(ncq.numberPoints());
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
//  cout << "\n\n\nnew element" << endl;
//  TRACER_CALL(t,"debugMapping");

    const SHAPE_CLASS *e=tess.presentShape();
    const LEVEL_CLASS *level=tess.presentLevel();
    MappingDataBase *mapping_data=
      mapping.getMappingData(e->numberVertices(),update_flags,&ncq);
    bool quadrature_points_ok=false;
    bool JxW_values_ok=false;
    mapping_data->computeFEValues(e,tess,level,xq,jxwv,
      quadrature_points_ok,JxW_values_ok);
    CHECK_TEST(quadrature_points_ok);
    CHECK_TEST(JxW_values_ok);
    for (int q=0;q<ncq.numberPoints();q++) {
      REAL error=abs(vertices[e->vertexIndex(q)][0]-xq[q][0]);
      if (error>DBL_EPSILON) {
        cout << "\tvertex[" << q << "] = " << vertices[e->vertexIndex(q)]
             << "," << xq[q] << " error = " << error << " " 
             << error/DBL_EPSILON << endl;
      }
    }
    REAL error=abs(abs(vertices[e->vertexIndex(1)][0]
                      -vertices[e->vertexIndex(0)][0])-jxwv[0]*2.);
    if (error>DBL_EPSILON) {
      cout << "\tjxwv = " 
           << abs(vertices[e->vertexIndex(1)][0]
                 -vertices[e->vertexIndex(0)][0])
           << " " << jxwv[0]*2.  << " " << jxwv[1]*2. << " error = "
           << error << " " << error/DBL_EPSILON << endl;
    }

    REAL r=drand48();
    r=drand48();
    Point<1> ep=tess.getVertices()[e->vertexIndex(0)]*r
               +tess.getVertices()[e->vertexIndex(1)]*(1-r);
    Point<1> cp=vertices[e->vertexIndex(0)]; // initial guess
    mapping.canonicalPoint(e,tess,mapping_data,ep,cp);
//  cout << "\tep = " << ep << " cp = " << cp << endl;
    error=sqrt((mapping.elementPoint(e,tess,level,cp)-ep).square());
    if (error>DBL_EPSILON) {
      cout << "\telementPoint(canonicalPoint(" << ep << ")) = ("
           << ep << "), error = " 
           << error << " " << error/DBL_EPSILON << endl;
    }
    {
      Point<1> cp(drand48());
      Point<1> cpnp=cp;
      Quadrature<1> q(cp);
      Quadrature<1> qnp(cpnp);
      MappingDataBase *md=
        mapping.getMappingData(e->numberVertices(),update_flags,&q);
      Array<2,REAL> sv(md->nShapeFunctions(),1);
      Array<2,Tensor<1,1> > sg(md->nShapeFunctions(),1);
      Array<2,Tensor<2,1> > sgg(md->nShapeFunctions(),1);
      for (int s=0;s<md->nShapeFunctions();s++) {
        sv(s,0)=md->shapeValue(s,0);
        sg(s,0)=md->shapeDerivative(s,0);
        sgg(s,0)=md->shapeSecondDerivative(s,0);
      }
      Array<2,Tensor<1,SPACEDIM> > sg_dif(md->nShapeFunctions(),1);
      richardsonShapeDerivs(mapping.getDegree(),mapping,e,cp,sv,sg_dif);
      for (int s=0;s<md->nShapeFunctions();s++) {
        REAL exact=sg(s,0)[0];
        REAL deriv=sg_dif(s,0)[0];
        REAL error=abs(exact-deriv);
        REAL den=max(abs(exact),abs(deriv));
        if (den>0.) error/=den;
        if (error>DBL_EPSILON*512.) {
          cout << "\tsg(" << s << ",0)[0] = " << exact
               << " " << deriv << " error = " << error << " "
               << error/DBL_EPSILON << endl;
        }
      }
      Array<2,Tensor<2,SPACEDIM> > sgg_dif(md->nShapeFunctions(),1);
      richardsonShapeSecondDerivs(mapping.getDegree(),mapping,e,cp,sg,
        sgg_dif);
      for (int s=0;s<md->nShapeFunctions();s++) {
        REAL exact=sgg(s,0)[0][0];
        REAL deriv=sgg_dif(s,0)[0][0];
        REAL error=abs(exact-deriv);
        REAL den=max(abs(exact),abs(deriv));
        if (den>0.) error/=den;
        if (error>DBL_EPSILON*512.) {
          cout << "\tsgg(" << s << ",0)[0][0] = " << exact
               << " " << deriv << " error = " << error << " "
               << error/DBL_EPSILON << endl;
        }
      }
      if (md->nHigherOrderShapeFunctions()>0) {
        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
        Array<2,REAL> hosv(md->nHigherOrderShapeFunctions(),1);
        Array<2,Tensor<1,1> > hosg(md->nHigherOrderShapeFunctions(),1);
        Array<2,Tensor<2,1> > hosgg(md->nHigherOrderShapeFunctions(),1);
        for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
          hosv(s,0)=md->higherOrderShapeValue(s,0);
          hosg(s,0)=md->higherOrderShapeDerivative(s,0);
          hosgg(s,0)=md->higherOrderShapeSecondDerivative(s,0);
        }
        richardsonHigherOrderShapeDerivs(mapping.getDegree(),mapping,e,
          cp,hosv,sg_dif);
        for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
          REAL exact=hosg(s,0)[0];
          REAL deriv=sg_dif(s,0)[0];
          REAL error=abs(exact-deriv);
          REAL den=max(abs(exact),abs(deriv));
          if (den>0.) error/=den;
          if (error>DBL_EPSILON*512.) {
            cout << "\thosg(" << s << ",0)[0] = " << exact
                 << " " << deriv << " error = " << error << " "
                 << error/DBL_EPSILON << endl;
          }
        }
        richardsonHigherOrderShapeSecondDerivs(mapping.getDegree(),
          mapping,e,cp,sg,sgg_dif);
        for (int s=0;s<md->nShapeFunctions();s++) {
          REAL exact=hosgg(s,0)[0][0];
          REAL deriv=sgg_dif(s,0)[0][0];
          REAL error=abs(exact-deriv);
          REAL den=max(abs(exact),abs(deriv));
          if (den>0.) error/=den;
          if (error>DBL_EPSILON*512.) {
            cout << "\tsgg(" << s << ",0)[0][0] = " << exact
                 << " " << deriv << " error = " << error << " "
                 << error/DBL_EPSILON << endl;
          }
        }
      }
      delete md;
    }

    {
      MappingDataBase *md=0;
      REAL dia=e->diameter(vertices);
      GaussianQuadrature<1> gq(mapping.getDegree());
      md=mapping.getMappingData(e->numberVertices(),update_flags,&gq);
      NumPtr<Point<1> > xq(gq.numberPoints());
      NumPtr<REAL> jxwv(gq.numberPoints());
      bool quadrature_points_ok=false;
      bool JxW_values_ok=false;
      md->computeFEValues(e,tess,level,xq,jxwv,quadrature_points_ok,
        JxW_values_ok);
      CHECK_TEST(JxW_values_ok);
      REAL quad=0.;
      for (int q=0;q<gq.numberPoints();q++) {
        quad+=jxwv[q];
        Tensor<2,1> dxdxi;
        Tensor<2,1> unit;
        unit[0][0]=1.;
        md->preMultiplyByJacobian(dxdxi,q,unit);
        Tensor<2,1> dxidx_times_dxdxi;
        md->postMultiplyByJacobianInverse(dxidx_times_dxdxi,q,dxdxi);
        REAL error=abs(dxidx_times_dxdxi[0][0]-1.);
        if (error>4.*DBL_EPSILON) {
          cout << "\tdxidx_times_dxdxi = "
               << dxidx_times_dxdxi[0][0] << " " << 1.
               << " " << error << " " << error/DBL_EPSILON << endl;
        }
      }
      REAL area=0.5;
      cout << "\tarea = " << area << " " << quad << " error = "
           << abs(area-quad) << endl;
      delete md;
    }

#if (SPACEDIM>1)
    map<int,MappingDataBase*>::iterator mdm_iterator=
      mapping_data_map.find(e->numberVertices());
    CHECK_TEST(mdm_iterator!=mapping_data_map.end());
    md=mdm_iterator->second;
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      if (level->neighborLevel(e,f)==0) {
        NumPtr<Point<1> > quadrature_points(ncq.numberPoints());
        NumPtr<REAL> JxW_values(1);
        NumPtr<Tensor<1,1> > boundary_forms(1);
        NumPtr<Point<1> > normal_vectors(1);
        md->fillFEFaceValues(f,e,tess,level,f,ncq,
          quadrature_points,JxW_values,boundary_forms,normal_vectors);
      }
    }
#endif

//  gt.newPage();
//  gt.setbgColor("white");
//  gt.setfgColor("black");
//  gt.drawAxes();
//  gt.setfgColor("blue");
//  ep=tess.getVertices()[e->vertexIndex(0)];
//  gt.movePen(ep[0],cp[0]);
//  Point<1> p1=tess.getVertices()[e->vertexIndex(1)];
//  for (;ep[0]<=p1[0];ep[0]+=pw) {
//    cp=mapping.canonicalPoint(e,tess,level,ep);
#ifdef DEBUG
//    cout << "\tep = " << ep << " cp = " << cp << endl;
#endif
//    gt.drawLine(ep[0],cp[0]);
//  }
//  XYGraphTool::WINDOW_TYPE::QuitButton qb;
    delete mapping_data; mapping_data=0;
  }
}

int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif

#if (SPACEDIM==1)
    Point<SPACEDIM> low(0.);
    Point<SPACEDIM> edge[1];
      edge[0]=Point<SPACEDIM>(1.);
    int number_elements[SPACEDIM];
      number_elements[0]=2;
    bool set_face_ids=false;

    Tessellation tess;
    NumPtr<Boundary*> boundary;
    NumPtr<BoundaryCondition*> boundary_condition;
    BoundaryData *sub_cell_data=generateParallelpiped(tess,low,edge,
      number_elements,set_face_ids,boundary,boundary_condition);
//  tess.plotShapes();
//  tess.plotUsedShapes();
//  tess.plotActiveShapes();
    int np=1;
/*
    {
//    TRACER_CALL(t,"MappingQ");
      MappingQ mapping;
      debugMapping(mapping,tess);
    }
*/
    {
//    TRACER_CALL(t,"MappingL");
      MappingL mapping;
      debugMapping(mapping,tess);
    }
/* obsolete:
    {
//    TRACER_CALL(t,"MappingH");
      MappingH mapping;
      debugMapping(mapping,tess);
    }
*/
    delete sub_cell_data;
#endif
  }
  return EXIT_SUCCESS;
}
#include <NumPtr.C>
template class NumPtrBase< NumPtr< Tensor<2,SPACEDIM> >,int>;
template class NumPtr< NumPtr< Tensor<2,SPACEDIM> > >;
