// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "AreaGraphTool.H"
#include "Boundary.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Palette.H"
#include "PermutationGenerator.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tessellation.H"
#include "Tracer.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#endif

Point<SPACEDIM> center(0.,0.);
REAL radius=1.;

void richardsonShapeDerivs(int nmax,const Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,const Array<2,REAL> &f,
Array<2,Tensor<1,SPACEDIM> > &fp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_VALUES;
  nmax=max(1,nmax);
  for (int i=0;i<SPACEDIM;i++) {
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.size(0));
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[i]+=h;
      Quadrature<SPACEDIM> q(xnew);
      MappingDataBase *mdnew=
        mapping.getMappingData(e->numberVertices(),update_flags,&q);
      Array<2,REAL> fnew(mdnew->nShapeFunctions(),1);
      for (int s=0;s<mdnew->nShapeFunctions();s++) {
        fnew(s,0)=mdnew->shapeValue(s,0); 
        table[nmax-n][s][i]=(fnew(s,0)-f(s,0))/h;
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          table[m][s][i]=table[m-1][s][i]
                        +(table[m-1][s][i]-table[m][s][i])/(p-1.);
        }
        fp(s,0)[i]=table[nmax-1][s][i];
      }
      delete mdnew;
    }
  }
}

void richardsonShapeSecondDerivs(int nmax,const Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,
const Array<2,Tensor<1,SPACEDIM> > &fp,
Array<2,Tensor<2,SPACEDIM> > &fpp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_GRADIENTS;
  nmax=max(1,nmax);
  for (int j=0;j<SPACEDIM;j++) {
    for (int i=0;i<SPACEDIM;i++) {
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.size(0));
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
        Quadrature<SPACEDIM> q(xnew);
        MappingDataBase *mdnew=
          mapping.getMappingData(e->numberVertices(),update_flags,&q);
        Array<2,Tensor<1,SPACEDIM> > fpnew(mdnew->nShapeFunctions(),1);
        for (int s=0;s<mdnew->nShapeFunctions();s++) {
          fpnew(s,0)=mdnew->shapeDerivative(s,0); 
          table[nmax-n][s][i][j]=(fpnew(s,0)[i]-fp(s,0)[i])/h;
          REAL p=2.;
          for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
            table[m][s][i][j]=table[m-1][s][i][j]
                          +(table[m-1][s][i][j]-table[m][s][i][j])/(p-1.);
          }
          fpp(s,0)[i][j]=table[nmax-1][s][i][j];
        }
        delete mdnew;
      }
    }
  }
}

void richardsonHigherOrderShapeDerivs(int nmax,const Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,const Array<2,REAL> &f,
Array<2,Tensor<1,SPACEDIM> > &fp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_VALUES;
  nmax=max(1,nmax);
  for (int i=0;i<SPACEDIM;i++) {
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.size(0));
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[i]+=h;
      Quadrature<SPACEDIM> q(xnew);
      MappingDataBase *mdnew=
        mapping.getMappingData(e->numberVertices(),update_flags,&q);
      Array<2,REAL> fnew(mdnew->nHigherOrderShapeFunctions(),1);
      for (int s=0;s<mdnew->nHigherOrderShapeFunctions();s++) {
        fnew(s,0)=mdnew->higherOrderShapeValue(s,0); 
        table[nmax-n][s][i]=(fnew(s,0)-f(s,0))/h;
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          table[m][s][i]=table[m-1][s][i]
                        +(table[m-1][s][i]-table[m][s][i])/(p-1.);
        }
        fp(s,0)[i]=table[nmax-1][s][i];
      }
      delete mdnew;
    }
  }
}

void richardsonHigherOrderShapeSecondDerivs(int nmax,
const Mapping &mapping,const SHAPE_CLASS *e,const Point<SPACEDIM> &x,
const Array<2,Tensor<1,SPACEDIM> > &fp,
Array<2,Tensor<2,SPACEDIM> > &fpp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_GRADIENTS;
  nmax=max(1,nmax);
  for (int j=0;j<SPACEDIM;j++) {
    for (int i=0;i<SPACEDIM;i++) {
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.size(0));
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
        Quadrature<SPACEDIM> q(xnew);
        MappingDataBase *mdnew=
          mapping.getMappingData(e->numberVertices(),update_flags,&q);
        Array<2,Tensor<1,SPACEDIM> >
          fpnew(mdnew->nHigherOrderShapeFunctions(),1);
        for (int s=0;s<mdnew->nHigherOrderShapeFunctions();s++) {
          fpnew(s,0)=mdnew->higherOrderShapeDerivative(s,0); 
          table[nmax-n][s][i][j]=(fpnew(s,0)[i]-fp(s,0)[i])/h;
          REAL p=2.;
          for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
            table[m][s][i][j]=table[m-1][s][i][j]
                          +(table[m-1][s][i][j]-table[m][s][i][j])/(p-1.);
          }
          fpp(s,0)[i][j]=table[nmax-1][s][i][j];
        }
        delete mdnew;
      }
    }
  }
}

void checkMappingContinuity(Tessellation &tess,const Mapping &mapping,
UPDATE_FLAGS update_flags) {
//TRACER_CALL(t,"checkMappingContinuity");
//MappingDataBase::use_full_degree_mapping_on_all_cells=true;
/*
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL winsize=0.5;
  XYGraphTool gt("checkMappingContinuity","x","y",0.,1.,0.,1.,&cmap,0,
    winsize);
*/
  tess.firstShape();
  const SHAPE_CLASS *e0=tess.presentShape();
  const LEVEL_CLASS *level=tess.presentLevel();
  tess.nextShape();
  const SHAPE_CLASS *e1=tess.presentShape();
  int f=e0->sharedFaceElementIndex(e1);
  const CODIMENSION1_SHAPE_CLASS *face=
    level->CODIMENSION1_LEVEL_CLASS::getShape(f);
  CHECK_POINTER(dynamic_cast<const Interval*>(face));

  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  Point<SPACEDIM> face_v0=vertices[face->vertexIndex(0)];
  Point<SPACEDIM> face_v1=vertices[face->vertexIndex(1)];

//int degree=mapping.getDegree();
//int npts=static_cast<int>(1./gt.getPixelWidth());
  int npts=0;
  NumPtr<Point<SPACEDIM> > xq(npts+1);
  REAL r=drand48();
  for (int q=0;q<=npts;q++) {
    r=drand48();
    xq[q]=face_v0*r+face_v1*(1.-r);
  }

  Point<SPACEDIM> point(D_DECL(0.,0.,0.));
  Quadrature<SPACEDIM> point_quadrature(point);
  MappingDataBase *md0=mapping.getMappingData(e0->numberVertices(),
    update_flags,&point_quadrature);
  md0->computeMappingSupportPoints(e0,tess,level);
  if (md0->requiresShapeToInitialize()) md0->computeShapesFor(e0);
  NumPtr<Point<SPACEDIM> > xq0(npts+1);
  for (int q=0;q<=npts;q++) {
    xq0[q]=point;
    md0->canonicalPoint(e0,e0->diameter(tess.getVertices()),xq[q],xq0[q]);
  }

  Quadrature<SPACEDIM> qf0(xq0);
  md0->computeShapes(&qf0);
  if (md0->requiresShapeToInitialize()) md0->computeShapesFor(e0);
  NumPtr<Point<SPACEDIM> > yq0(npts+1);
  NumPtr<REAL> wq0(npts+1);
  wq0.initialize(1.);
  bool quadrature_points_ok=false;
  bool JxW_values_ok=false;
  md0->computeFEValues(e0,tess,level,yq0,wq0,quadrature_points_ok,
    JxW_values_ok);
  CHECK_TEST(quadrature_points_ok);

  MappingDataBase *md1=mapping.getMappingData(e1->numberVertices(),
    update_flags,&point_quadrature);
  md1->computeMappingSupportPoints(e1,tess,level);
  if (md1->requiresShapeToInitialize()) md1->computeShapesFor(e1);
  NumPtr<Point<SPACEDIM> > xq1(npts+1);
  for (int q=0;q<=npts;q++) {
    xq1[q]=point;
    md1->canonicalPoint(e1,e1->diameter(tess.getVertices()),xq[q],xq1[q]);
  }

  Quadrature<SPACEDIM> qf1(xq1);
  md1->computeShapes(&qf1);
  if (md1->requiresShapeToInitialize()) md1->computeShapesFor(e1);
  NumPtr<Point<SPACEDIM> > yq1(npts+1);
  NumPtr<REAL> wq1(npts+1);
  wq1.initialize(1.);
  quadrature_points_ok=false;
  md1->computeFEValues(e1,tess,level,yq1,wq1,quadrature_points_ok,
    JxW_values_ok);
  CHECK_TEST(quadrature_points_ok);

  REAL error=0.;
  for (int q=0;q<=npts;q++) {
    error+=pow(yq0[q].distance(yq1[q]),2);
  }
  error=sqrt(error/static_cast<REAL>(npts+1))/DBL_EPSILON;
  if (error>64.) {
    cout << "\tdifference between coordinate maps on shared face = "
         << error << endl;
  }

/*
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  Point<SPACEDIM> &x0=yq0[0];
  gt.movePen(x0[0],x0[1]);
  for (int q=1;q<=npts;q++) {
    Point<SPACEDIM> &y0=yq0[q];
    //.drawLine(y0[0],y0[1]);
  }
  gt.setfgColor("red");
  Point<SPACEDIM> &x1=yq1[0];
  gt.movePen(x1[0],x1[1]);
  for (int q=1;q<=npts;q++) {
    Point<SPACEDIM> &y1=yq1[q];
    gt.drawLine(y1[0],y1[1]);
  }
  gt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
*/
  delete md0; md0=0;
  delete md1; md1=0;
}

void debugTess(Tessellation &tess,const Mapping &mapping,
UPDATE_FLAGS update_flags) {
//TRACER_CALL(t,"debugTess");
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]; rlo[0]=-radius; rlo[1]=-radius;
  REAL rhi[2]; rhi[0]=radius; rhi[1]=radius;
  AreaGraphTool gt("debugTess","x","y",rlo,rhi,&cmap,0,0.5);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.movePen(center[0]+radius,center[1]);
  REAL angle=0.;
  REAL dangle=2.*M_PI/256.;
  for (int i=0;i<=256;i++,angle+=dangle) {
    gt.drawLine(center[0]+radius*cos(angle),
                center[1]+radius*sin(angle));
  }
  gt.setfgColor("red");

//cf FEFaceValues::FEFaceValues
  map<int,MappingDataBase*> mapping_data_map;
  NewtonCotesQuadrature<1> ncq(mapping.getDegree()+1);

  {
//  TRACER_CALL(t,"debugTess");
    int nfp=3*ncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<3;f++) {
      Triangle::staticProjectToFace(f,ncq.points(),q_points_face);
      weights_face.copyFrom(ncq.weights());
      q_points_face+=ncq.numberPoints();
      weights_face+=ncq.numberPoints();
    }
    Quadrature<SPACEDIM> qf(q_points,weights);
    mapping_data_map[3]=mapping.getMappingData(3,update_flags,&qf);
  }

  {
  //TRACER_CALL(t,"debugTess");
    int nfp=4*ncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<4;f++) {
      Quadrilateral::staticProjectToFace(f,ncq.points(),q_points_face);
      weights_face.copyFrom(ncq.weights());
      q_points_face+=ncq.numberPoints();
      weights_face+=ncq.numberPoints();
    }
    Quadrature<SPACEDIM> qf(q_points,weights);
    mapping_data_map[4]=mapping.getMappingData(4,update_flags,&qf);
  }

  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
#ifdef DEBUG
//for (int i=0;i<vertices.getNumber();i++) {
//  cout << "\tvertices[" << i << "] = " << vertices[i] << endl;
//}
#endif
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
    const SHAPE_CLASS *e=tess.presentShape();
    const LEVEL_CLASS *level=tess.presentLevel();
#ifdef DEBUG
//  cout << "\nelement = " << endl;
//  e->printOn(cout);
#endif

    {
//    TRACER_CALL(t,"debugTess canonicalPoint(elementPoint)");
      Point<SPACEDIM> cp(drand48(),drand48());
      if (e->numberVertices()==3) cp[1]*=1.-cp[0];
      Point<SPACEDIM> ep=
        mapping.elementPoint(e,tess,tess.presentLevel(),cp);
      Quadrature<SPACEDIM> quad(cp);
      MappingDataBase *md=mapping.getMappingData(e->numberVertices(),
        UPDATE_TRANSFORMATION_VALUES | UPDATE_TRANSFORMATION_GRADIENTS,
        &quad);
      md->computeMappingSupportPoints(e,tess,tess.presentLevel());
      if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
      Point<SPACEDIM> cp2=cp;
      md->canonicalPoint(e,e->diameter(tess.getVertices()),ep,cp2);
      REAL error=max(abs(cp2[0]-cp[0]),abs(cp2[1]-cp[1]));
      if (error>DBL_EPSILON*512.) {
        cout << "\tcp = [" << cp << "], ep = [" << ep 
             << "], cp2 = [" << cp2 
             << "], error = " 
             << error << " " << error/DBL_EPSILON << endl;
      }
      delete md; md=0;
    }

    {
//    TRACER_CALL(t,"debugTess");
//    have to change MappingDataBase from private to public
//    in order to debug here
      Point<SPACEDIM> cp(drand48(),drand48());
      Point<SPACEDIM> cpnp=cp;
      Quadrature <SPACEDIM> q(cp);
      Quadrature <SPACEDIM> qnp(cpnp);
      MappingDataBase *md=
        mapping.getMappingData(e->numberVertices(),update_flags,&q);
//    md->printOn(cout);

      Array<2,REAL> sv(md->nShapeFunctions(),1);
      Array<2,Tensor<1,SPACEDIM> > sg(md->nShapeFunctions(),1);
      Array<2,Tensor<2,SPACEDIM> > sgg(md->nShapeFunctions(),1);
      for (int s=0;s<md->nShapeFunctions();s++) {
        sv(s,0)=md->shapeValue(s,0);
        sg(s,0)=md->shapeDerivative(s,0);
        sgg(s,0)=md->shapeSecondDerivative(s,0);
      }
      Array<2,Tensor<1,SPACEDIM> > sg_dif(md->nShapeFunctions(),1);
      richardsonShapeDerivs(mapping.getDegree(),mapping,e,cp,sv,sg_dif);
      for (int d=0;d<SPACEDIM;d++) {
        for (int s=0;s<md->nShapeFunctions();s++) {
          REAL exact=sg(s,0)[d];
          REAL deriv=sg_dif(s,0)[d];
          REAL error=abs(exact-deriv);
          REAL den=max(abs(exact),abs(deriv));
          if (den>0.) error/=den;
          if (error>DBL_EPSILON*8192.) {
            cout << "\tsg(" << s << ",0)[" << d << "] = " << exact 
                 << " " << deriv << " error = " << error << " "
                 << error/DBL_EPSILON << endl;
          }
        }
      }
      Array<2,Tensor<2,SPACEDIM> > sgg_dif(md->nShapeFunctions(),1);
      richardsonShapeSecondDerivs(mapping.getDegree(),mapping,e,cp,sg,
        sgg_dif);
      for (int j=0;j<SPACEDIM;j++) {
        for (int i=0;i<SPACEDIM;i++) {
          for (int s=0;s<md->nShapeFunctions();s++) {
            REAL exact=sgg(s,0)[i][j];
            REAL deriv=sgg_dif(s,0)[i][j];
            REAL error=abs(exact-deriv);
            REAL den=max(abs(exact),abs(deriv));
            if (den>0.) error/=den;
            if (error>DBL_EPSILON*8192.) {
              cout << "\tsgg(" << s << ",0)[" << i << "][" << j
                   << "] = " << exact << " " << deriv << " error = "
                   << error << " " << error/DBL_EPSILON << endl;
            }
          }
        }
      }

      if (md->nHigherOrderShapeFunctions()>0) {
        Array<2,REAL> hosv(md->nHigherOrderShapeFunctions(),1);
        Array<2,Tensor<1,SPACEDIM> >
          hosg(md->nHigherOrderShapeFunctions(),1);
        Array<2,Tensor<2,SPACEDIM> >
          hosgg(md->nHigherOrderShapeFunctions(),1);
        for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
          hosv(s,0)=md->higherOrderShapeValue(s,0);
          hosg(s,0)=md->higherOrderShapeDerivative(s,0);
          hosgg(s,0)=md->higherOrderShapeSecondDerivative(s,0);
        }
        Array<2,Tensor<1,SPACEDIM> >
          hosg_dif(md->nHigherOrderShapeFunctions(),1);
        richardsonHigherOrderShapeDerivs(mapping.getDegree(),mapping,e,
          cp,hosv,hosg_dif);
        for (int d=0;d<SPACEDIM;d++) {
//        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
          for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
            REAL exact=hosg(s,0)[d];
            REAL deriv=hosg_dif(s,0)[d];
            REAL error=abs(exact-deriv);
            REAL den=max(abs(exact),abs(deriv));
            if (den>0.) error/=den;
            if (error>DBL_EPSILON*131072.) {
//            e->printOn(cout);
//            cout << "\tdegree = " << mapping.getDegree() << endl;
//            cout << "\tcp = " << cp << endl;
//            cout << "\thosv(" << s << ") = " << hosv(s,0) << endl;
              cout << "\thosg(" << s << ",0)[" << d << "] = " << exact
                   << " " << deriv << " error = " << error << " "
                   << error/DBL_EPSILON << endl;
              CHECK_TEST(error<=DBL_EPSILON*131072.);
            } 
          } 
        } 
        Array<2,Tensor<2,SPACEDIM> >
          hosgg_dif(md->nHigherOrderShapeFunctions(),1);
        richardsonHigherOrderShapeSecondDerivs(mapping.getDegree(),
          mapping,e,cp,hosg,hosgg_dif);
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            for (int s=0;s<md->nShapeFunctions();s++) {
              REAL exact=hosgg(s,0)[i][j];
              REAL deriv=hosgg_dif(s,0)[i][j];
              REAL error=abs(exact-deriv);
              REAL den=max(abs(exact),abs(deriv));
              if (den>0.) error/=den;
              if (error>DBL_EPSILON*131072.) {
                cout << "\thosgg(" << s << ",0)[" << i << "][" << j
                     << "] = " << exact << " " << deriv << " error = "
                     << error << " " << error/DBL_EPSILON << endl;
                CHECK_TEST(error<=DBL_EPSILON*131072.);
              }
            }
          }
        }
      }
      delete md;
    }

    MappingDataBase *md=0;
    REAL dia=e->diameter(vertices);
    if (e->numberVertices()==3) {
//    TRACER_CALL(t,"debugTess");
      TriangleProductGaussianQuadrature npgq(mapping.getDegree()*SPACEDIM);
      md=mapping.getMappingData(e->numberVertices(),update_flags,&npgq);
      NumPtr<Point<SPACEDIM> > xq(npgq.numberPoints());
      NumPtr<REAL> jxwv(npgq.numberPoints());
      if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
      bool quadrature_points_ok=false;
      bool JxW_values_ok=false;
      md->computeFEValues(e,tess,level,xq,jxwv,quadrature_points_ok,
        JxW_values_ok);
#ifdef DEBUG
//    cout << "\tnpgq:" << endl;
//    npgq.printOn(cout);
//    for (int q=0;q<xq.getNumber();q++) {
//      cout << "\txq[" << q << "] = " << xq[q] << endl;
//    }
//    for (int q=0;q<xq.getNumber();q++) {
//      cout << "\tjxwv[" << q << "] = " << jxwv[q] << endl;
//    }
//    for (int v=0;v<e->numberVertices();v++) {
//      cout << "\tvertices[" << e->vertexIndex(v) << "] = " 
//           << vertices[e->vertexIndex(v)] << endl;
//    }
#endif
      REAL quad=0.;
      for (int q=0;q<npgq.numberPoints();q++) {
        quad+=jxwv[q];
//      Point<SPACEDIM> ep=
//        mapping.elementPoint(e,tess,level,npgq.point(q));
//      REAL error=vertices[e->vertexIndex(q)].distance(ep);
//      if (error>DBL_EPSILON) {
//        cout << "\tvertex[" << q << "] = " 
//             << vertices[e->vertexIndex(q)] << "," << ep 
//             << " error = " << error << " " << error/DBL_EPSILON 
//             << endl;
//      }
        Tensor<2,SPACEDIM> unit;
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            unit[i][j]=(i==j ? 1. : 0.);
          }
        }
        Tensor<2,SPACEDIM> dxdxi;
        md->preMultiplyByJacobian(dxdxi,q,unit);
        Tensor<2,SPACEDIM> dxidx_times_dxdxi;
        md->postMultiplyByJacobianInverse(dxidx_times_dxdxi,q,dxdxi);
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            unit[i][j]=(i==j ? 1. : 0.);
            REAL error=abs(dxidx_times_dxdxi[i][j]-unit[i][j]);
            if (error>4.*DBL_EPSILON) {
              cout << "\tdxidx_times_dxdxi[" << i << "][" << j << "] = "
                   << dxidx_times_dxdxi[i][j] << " " << unit[i][j]
                   << " " << error << " " << error/DBL_EPSILON << endl;
            }
          }
        }
      }
      REAL area=0.25*M_PI*radius*radius;
      cout << "\tarea = " << area << " " << quad << " error = " 
           << abs(area-quad) << endl;
    } else {
//    TRACER_CALL(t,"debugTess");
      GaussianQuadrature<SPACEDIM> gq(mapping.getDegree()*SPACEDIM);
      md=mapping.getMappingData(e->numberVertices(),update_flags,&gq);
      NumPtr<Point<SPACEDIM> > xq(gq.numberPoints());
      NumPtr<REAL> jxwv(gq.numberPoints());
      if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
      bool quadrature_points_ok=false;
      bool JxW_values_ok=false;
      md->computeFEValues(e,tess,level,xq,jxwv,quadrature_points_ok,
        JxW_values_ok);
      REAL quad=0.;
      for (int q=0;q<gq.numberPoints();q++) {
        quad+=jxwv[q];
//      Point<SPACEDIM> ep=mapping.elementPoint(e,tess,level,
//        Quadrilateral::canonicalVertex(q));
//      REAL error=vertices[e->vertexIndex(q)].distance(ep);
//      if (error>256.*DBL_EPSILON) {
//        cout << "\tvertex[" << q << "] = " 
//             << vertices[e->vertexIndex(q)] << "," << ep 
//             << " error = " << error << " " << error/DBL_EPSILON 
//             << endl;
//      }
        Tensor<2,SPACEDIM> unit;
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            unit[i][j]=(i==j ? 1. : 0.);
          }
        }
        Tensor<2,SPACEDIM> dxdxi;
        md->preMultiplyByJacobian(dxdxi,q,unit);
        Tensor<2,SPACEDIM> dxidx_times_dxdxi;
        md->postMultiplyByJacobianInverse(dxidx_times_dxdxi,q,dxdxi);
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            unit[i][j]=(i==j ? 1. : 0.);
            REAL error=abs(dxidx_times_dxdxi[i][j]-unit[i][j]);
            if (error>DBL_EPSILON) {
              cout << "\tdxidx_times_dxdxi[" << i << "][" << j << "] = "
                   << dxidx_times_dxdxi[i][j] << " " << unit[i][j]
                   << " " << error << endl;
            }
          }
        }
      }
      REAL area=0.;
      if (e->elementDataIndex()==2) {
        area=0.4*radius*radius;
      } else {
        area=(M_PI*0.25-0.1)*radius*radius;
      }
      cout << "\tarea = " << area << " " << quad << " error = " 
           << abs(area-quad) << endl;
    }
    delete md;

//  from FEFaceValues::reinit
    map<int,MappingDataBase*>::iterator mdm_iterator=
      mapping_data_map.find(e->numberVertices());
    CHECK_TEST(mdm_iterator!=mapping_data_map.end());
    md=mdm_iterator->second;
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
//    TRACER_CALL(t,"debugTess");
      if (level->neighborLevel(e,f)==0) {
        NumPtr<Point<SPACEDIM> > quadrature_points(ncq.numberPoints());
        NumPtr<REAL> JxW_values(ncq.numberPoints());
        NumPtr<Tensor<1,SPACEDIM> > boundary_forms(ncq.numberPoints());
        NumPtr<Point<SPACEDIM> > normal_vectors(ncq.numberPoints());
//      md->fillFEFaceValues(f*ncq.numberPoints(),e,tess,level,f,ncq,
//        quadrature_points,JxW_values,boundary_forms,normal_vectors);
        bool is_subface=false;
        bool quadrature_points_ok=false;
        bool jxw_values_ok=false;
        bool boundary_forms_ok=false;
        bool normal_vectors_ok=false;
        md->computeFEFaceValues(e,tess,level,f,is_subface,
          ncq.numberPoints(),f*ncq.numberPoints(),ncq.weights(),
          quadrature_points,JxW_values,boundary_forms,normal_vectors,
          quadrature_points_ok,jxw_values_ok,boundary_forms_ok,
          normal_vectors_ok);
        CHECK_TEST(quadrature_points_ok);
#ifdef DEBUG
//      cout << "\te:" << endl;
//      e->printOn(cout);
//      cout << "\tf = " << f << endl;
//      for (int q=0;q<quadrature_points.getNumber();q++) {
//        cout << "\tquadrature_points[" << q << "] = "
//             << quadrature_points[q] << endl;
//      }
//      for (int q=0;q<normal_vectors.getNumber();q++) {
//        cout << "\tnormal_vectors[" << q << "] = "
//             << normal_vectors[q] << endl;
//      }
#endif
        gt.setfgColor("red");
        for (int q=0;q<quadrature_points.getNumber();q++) {
          Point<SPACEDIM> hq=//quadrature_points[q]+
                             normal_vectors[q]*.1;
          COMPLEX base(quadrature_points[q][0],quadrature_points[q][1]);
          COMPLEX head(hq[0],hq[1]);
          COMPLEX arrow(-sqrt(.05),-sqrt(.05));
          gt.drawVector(base,head,arrow);
        }
//      gt.setfgColor("blue");
//      for (int q=0;q<quadrature_points.getNumber();q++) {
//        COMPLEX base(quadrature_points[q][0],quadrature_points[q][1]);
//        COMPLEX head(boundary_forms[q][0],boundary_forms[q][1]);
//        COMPLEX arrow(sqrt(.005),sqrt(.005));
//        gt.drawVector(base,head,arrow);
//      }
      }
    }

    int npts=128;
    gt.setfgColor("green");
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
//    TRACER_CALL(t,"debugTess");
      Point<SPACEDIM> cv,dcv;
      if (e->numberVertices()==4) {
        cv=
          Quadrilateral::canonicalVertex(e->elementVertexFromFace(f,0));
        dcv=(Quadrilateral::canonicalVertex(
          e->elementVertexFromFace(f,1))-cv)/static_cast<REAL>(npts);
      } else {
        cv=Triangle::canonicalVertex(e->elementVertexFromFace(f,0));
        dcv=(Triangle::canonicalVertex(e->elementVertexFromFace(f,1))
            -cv)/static_cast<REAL>(npts);
      }
      Point<SPACEDIM> ep=
        mapping.elementPoint(e,tess,tess.presentLevel(),cv);
      gt.movePen(ep[0],ep[1]);
      cv+=dcv;
      for (int i=0;i<npts;i++,cv+=dcv) {
        ep=mapping.elementPoint(e,tess,tess.presentLevel(),cv);
        gt.drawLine(ep[0],ep[1]);
      }
    }
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;

  for (map<int,MappingDataBase*>::iterator
  mdm_iterator=mapping_data_map.begin();
  mdm_iterator!=mapping_data_map.end();mdm_iterator++) {
    delete mdm_iterator->second; mdm_iterator->second=0;
  }
}

int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
  {
#ifdef MEM_DEBUG
    MemoryDebugger dm(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif


    UPDATE_FLAGS update_flags(UPDATE_VALUES | UPDATE_GRADIENTS |
      UPDATE_SECOND_DERIVATIVES | UPDATE_BOUNDARY_FORMS 
      | UPDATE_Q_POINTS | UPDATE_JxW_VALUES
      | UPDATE_NORMAL_VECTORS | UPDATE_MAPPING_JACOBIAN_INVERSE
      | UPDATE_MAPPING_JACOBIAN | UPDATE_MAPPING_SECOND_DERIVATIVE
      | UPDATE_TRANSFORMATION_VALUES 
      | UPDATE_TRANSFORMATION_GRADIENTS);
    bool set_face_ids=true;
#if (SPACEDIM==2)
//
    { // Triangle mapping vertices
      MappingL mapping(1);
      Tessellation tess;
      Point<SPACEDIM> low(0.,0.);
      Point<SPACEDIM> edge[2];
        edge[0]=Point<SPACEDIM>(2.,0.);
        edge[1]=Point<SPACEDIM>(0.,1.);
      int number_elements[2]={2,2};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpipedTriangles(tess,low,
        edge,number_elements,set_face_ids,boundary,boundary_condition);
      NumPtr<REAL> w(3);
      NumPtr<Point<SPACEDIM> > eslp;
      Triangle::staticEquallySpacedLatticePoints(1,eslp);
      w.initialize(1.);
      Quadrature<SPACEDIM> quad(eslp,w);
      MappingDataBase *md=mapping.getMappingData(
        Triangle::verticesPerShape(),update_flags,&quad);
      md->computeShapes(&quad);
      for (int q=0;q<3;q++) {
        for (int s=0;s<md->nShapeFunctions();s++) {
          REAL sv=md->shapeValue(s,q);
          REAL error=(s==q ? abs(sv-1.) : abs(sv) );
          if (error>DBL_EPSILON) {
            cout << "\teslp[" << q << "] = " << eslp[q] << endl;
            cout << "\tsv[" << s << "," << q << "] = " << sv << endl;
            OBSOLETE("bad shape value");
          }
        }
      }
      for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
        const SHAPE_CLASS *e=tess.presentShape();
        const LEVEL_CLASS *level=tess.presentLevel();
        const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
        NumPtr<Point<SPACEDIM> > qp(3);
        NumPtr<REAL> JxW_values(3);
        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
        bool quadrature_points_ok=false;
        bool JxW_values_ok=false;
        md->computeFEValues(e,tess,level,qp,JxW_values,quadrature_points_ok,
          JxW_values_ok);
        for (int q=0;q<3;q++) {
          REAL error=qp[q].distance(vertices[e->vertexIndex(q)]);
          if (error>DBL_EPSILON) {
            cout << "\tqp[" << q << "] = " << qp[q] << endl;
            cout << "\tvertices[" << e->vertexIndex(q) << "] = " 
                 << vertices[e->vertexIndex(q)] << endl;
            OBSOLETE("bad mapped quadrature point");
          }
        }
      }
      delete md;
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
//
    { // Quadrilateral mapping vertices
      MappingL mapping(1);
      Tessellation tess;
      Point<SPACEDIM> low(0.,0.);
      Point<SPACEDIM> edge[2];
        edge[0]=Point<SPACEDIM>(2.,0.);
        edge[1]=Point<SPACEDIM>(0.,1.);
      int number_elements[2]={2,2};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpiped(tess,low,edge,
        number_elements,set_face_ids,boundary,boundary_condition);
      NumPtr<REAL> w(4);
      NumPtr<Point<SPACEDIM> > eslp;
      Quadrilateral::staticEquallySpacedLatticePoints(1,eslp);
      w.initialize(1.);
      Quadrature<SPACEDIM> quad(eslp,w);
      MappingDataBase *md=mapping.getMappingData(
        Quadrilateral::verticesPerShape(),update_flags,&quad);
      md->computeShapes(&quad);
      for (int q=0;q<4;q++) {
        for (int s=0;s<md->nShapeFunctions();s++) {
          REAL sv=md->shapeValue(s,q);
          REAL error=(s==q ? abs(sv-1.) : abs(sv) );
          if (error>DBL_EPSILON) {
            cout << "\teslp[" << q << "] = " << eslp[q] << endl;
            cout << "\tsv[" << s << "," << q << "] = " << sv << endl;
            OBSOLETE("bad shape value");
          }
        }
      }
      for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
        const SHAPE_CLASS *e=tess.presentShape();
        const LEVEL_CLASS *level=tess.presentLevel();
        const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
        NumPtr<Point<SPACEDIM> > qp(4);
        NumPtr<REAL> JxW_values(4);
        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
        bool quadrature_points_ok=false;
        bool JxW_values_ok=false;
        md->computeFEValues(e,tess,level,qp,JxW_values,
          quadrature_points_ok,JxW_values_ok);
        for (int q=0;q<4;q++) {
          REAL error=qp[q].distance(vertices[e->vertexIndex(q)]);
          if (error>DBL_EPSILON) {
            cout << "\tqp[" << q << "] = " << qp[q] << endl;
            cout << "\tvertices[" << e->vertexIndex(q) << "] = " 
                 << vertices[e->vertexIndex(q)] << endl;
            OBSOLETE("bad mapped quadrature point");
          }
        }
      }
      delete md;
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
    for (int degree=1;degree<=6;degree++) 
//  int degree=6;
    {
      cout << "\n***degree = " << degree << "***" << endl;
//
      { // Triangle mapping equally spaced lattice points
        MappingL mapping(degree);
        int neslp=
          Triangle::staticNumberEquallySpacedLatticePoints(degree);
        NumPtr<REAL> w(neslp);
        NumPtr<Point<SPACEDIM> > eslp;
        Triangle::staticEquallySpacedLatticePoints(degree,eslp);
        w.initialize(1.);
        Quadrature<SPACEDIM> quad(eslp,w);
        MappingDataBase *md=mapping.getMappingData(
          Triangle::verticesPerShape(),update_flags,&quad);
        md->computeShapes(&quad);
        if (!md->requiresShapeToInitialize()) {
#ifdef DEBUG
//        for (int q=0;q<neslp;q++) {
//          cout << "\tshapeValue(" << eslp[q] << ") =";
//          for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
//            cout << " " << md->higherOrderShapeValue(s,q);
//          }
//          cout << endl;
//        }
#endif
          for (int q=0;q<neslp;q++) {
            for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
              REAL sv=md->higherOrderShapeValue(s,q);
              REAL error=(s==q ? abs(sv-1.) : abs(sv) );
              if (error>1024.*DBL_EPSILON) {
                cout << "\teslp[" << q << "] = " << eslp[q] 
                     << " sv(" << s << "," << q << ") = " << sv
                     << " " << error << " " << error/DBL_EPSILON << endl;
                OBSOLETE("bad higher order shape value");
              }
            }
          }
        }
        delete md;
      }
//
//
      { // Quadrilateral mapping equally spaced lattice points
        MappingL mapping(degree);
        int neslp=
          Quadrilateral::staticNumberEquallySpacedLatticePoints(degree);
        NumPtr<REAL> w(neslp);
        NumPtr<Point<SPACEDIM> > eslp;
        Quadrilateral::staticEquallySpacedLatticePoints(degree,eslp);
        w.initialize(1.);
        Quadrature<SPACEDIM> q(eslp,w);
        MappingDataBase *md=mapping.getMappingData(
          Quadrilateral::verticesPerShape(),update_flags,&q);
        md->computeShapes(&q);
        if (!md->requiresShapeToInitialize()) {
#ifdef DEBUG
//        for (int q=0;q<neslp;q++) {
//          cout << "\tshapeValue(" << eslp[q] << ") =";
//          for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
//            cout << " " << md->higherOrderShapeValue(s,q);
//          }
//          cout << endl;
//        }
#endif
          for (int q=0;q<neslp;q++) {
            for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
              REAL sv=md->higherOrderShapeValue(s,q);
              REAL error=(s==q ? abs(sv-1.) : abs(sv) );
              if (error>1024.*DBL_EPSILON) {
                cout << "\teslp[" << q << "] = " << eslp[q] 
                     << " sv(" << s << "," << q << ") = " << sv
                     << " " << error << " " << error/DBL_EPSILON << endl;
                OBSOLETE("bad higher order shape value");
              }
            }
          }
        }
        delete md;
      }
//
//
      {
//      TRACER_CALL(t,"MappingL generateBallTriangles");
        MappingL mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallTriangles(tess,
          center,radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
//
//
      {
//      TRACER_CALL(t,"MappingL generateBall");
        MappingL mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBall(tess,center,radius,
          set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
//
/* obsolete:
      {
//      TRACER_CALL(t,"MappingQ generateBallTriangles");
        MappingQ mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallTriangles(tess,
          center,radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
*/
/* obsolete:
      {
//      TRACER_CALL(t,"MappingQ generateBall");
        MappingQ mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBall(tess,center,radius,
          set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
*/
/* obsolete:
      {
//      TRACER_CALL(t,"MappingH generateBallTriangles");
        MappingH mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallTriangles(tess,center,
          radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
*/
/* obsolete:
      {
//      TRACER_CALL(t,"MappingH generateBall");
        MappingH mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBall(tess,center,radius,
          set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
*/
//
      {
//      TRACER_CALL(t,"checkMappingContinuity");
        PermutationGenerator pg(3);
        while (pg.hasMore()) {
          const NumPtr<int> &perm=pg.getNext();
          NumPtr<Point<SPACEDIM> > vertices(5);
          vertices[4]=Point<SPACEDIM>(0.,0.);
          vertices[3]=Point<SPACEDIM>(1.,0.);
          vertices[perm[2]]=Point<SPACEDIM>(0.,.5);
          vertices[perm[1]]=Point<SPACEDIM>(1.,.5);
          vertices[perm[0]]=Point<SPACEDIM>(.5,1.);
          NumPtr<CellData> cells(2);
          for (int i=0;i<2;i++) {
            cells[i].setMaterialModel(0);
          }
          cells[0].allocate(4);
          cells[0].vertex(0)=4;
          cells[0].vertex(1)=3;
          cells[0].vertex(2)=perm[2];
          cells[0].vertex(3)=perm[1];

          cells[1].allocate(3);
          cells[1].vertex(0)=perm[2];
          cells[1].vertex(1)=perm[1];
          cells[1].vertex(2)=perm[0];
          Tessellation tess;
          BoundaryData *subcell_data=tess.create(vertices,cells);
          Point<2> centroid((vertices[4]+vertices[3]
                            +vertices[perm[2]]+vertices[perm[1]])*0.25);
          NumPtr<Boundary*> boundary(5);
          boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
            vertices[perm[2]]);
          boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[3],
            vertices[perm[1]]);
          boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[3],
            vertices[4]);
          boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
            vertices[perm[2]],vertices[perm[0]]);
          boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
            vertices[perm[1]],vertices[perm[0]]);
          NumPtr<BoundaryCondition*> boundary_condition(5);
          for (int i=0;i<boundary.getNumber();i++) {
            boundary_condition[i]=
              OPERATOR_NEW BoundaryCondition(boundary[i]);
          }
          for (bool doit=tess.firstCodimension1ActiveShape();doit;
          doit=tess.nextCodimension1ActiveShape()) {
            CODIMENSION1_LEVEL_CLASS *codimension1_level=
              tess.presentCodimension1Level();
            if (codimension1_level->getBoundaryCondition()!=0) {
              CODIMENSION1_SHAPE_CLASS *face_element=
                codimension1_level->getShape();
              Point<SPACEDIM> center=face_element->center(vertices);
              bool found=false;
              for (int i=0;i<boundary.getNumber();i++) {
                if (abs(boundary[i]->constraint(center))<1.e-5) {
                  codimension1_level->
                    setBoundaryCondition(boundary_condition[i]);
                  found=true;
                  break;
                }
              }
              CHECK_TEST(found);
            }
          }
          tess.synchronize(subcell_data);

          cout << "\tpermutation = " << perm[0] << " " << perm[1] << " "
               << perm[2] << endl;
          {
            MappingL mapping(degree);
            checkMappingContinuity(tess,mapping,update_flags);
          }
//obsolete{
//          MappingQ mapping(degree);
//          checkMappingContinuity(tess,mapping,update_flags);
//        }
//obsolete{
//          MappingH mapping(degree);
//          checkMappingContinuity(tess,mapping,update_flags);
//        }
          delete subcell_data; subcell_data=0;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            delete boundary_condition[i]; boundary_condition[i]=0;
          }
          for (int i=0;i<boundary.getNumber();i++) {
            delete boundary[i]; boundary[i]=0;
          }
        }
      }
//
    }
#endif
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, MappingDataBase*>, 
  _Select1st<pair<int const, MappingDataBase*> >, 
  less<int>, 
  allocator<pair<int const, MappingDataBase*> > 
>;
#include <NumPtr.C>
template class NumPtrBase< NumPtr< Tensor<2,SPACEDIM> >,int>;
template class NumPtr< NumPtr< Tensor<2,SPACEDIM> > >;
