// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Boundary.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tessellation.H"
#include "Tracer.H"
#include "VolGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#endif

Point<SPACEDIM> center(0.,0.,0.);
REAL radius=1.;

void richardsonShapeDerivs(int nmax,const Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,const Array<2,REAL> &f,
Array<2,Tensor<1,SPACEDIM> > &fp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_VALUES;
  nmax=max(1,nmax);
  for (int i=0;i<SPACEDIM;i++) {
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.size(0));
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[i]+=h;
      Quadrature<SPACEDIM> q(xnew);
      MappingDataBase *mdnew=
        mapping.getMappingData(e->numberVertices(),update_flags,&q);
      Array<2,REAL> fnew(mdnew->nShapeFunctions(),1);
      for (int s=0;s<mdnew->nShapeFunctions();s++) {
        fnew(s,0)=mdnew->shapeValue(s,0); 
        table[nmax-n][s][i]=(fnew(s,0)-f(s,0))/h;
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          table[m][s][i]=table[m-1][s][i]
                        +(table[m-1][s][i]-table[m][s][i])/(p-1.);
        }
        fp(s,0)[i]=table[nmax-1][s][i];
      }
      delete mdnew;
    }
  }
}

void richardsonShapeSecondDerivs(int nmax,const Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,
const Array<2,Tensor<1,SPACEDIM> > &fp,
Array<2,Tensor<2,SPACEDIM> > &fpp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_GRADIENTS;
  nmax=max(1,nmax);
  for (int j=0;j<SPACEDIM;j++) {
    for (int i=0;i<SPACEDIM;i++) {
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.size(0));
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
        Quadrature<SPACEDIM> q(xnew);
        MappingDataBase *mdnew=
          mapping.getMappingData(e->numberVertices(),update_flags,&q);
        Array<2,Tensor<1,SPACEDIM> > fpnew(mdnew->nShapeFunctions(),1);
        for (int s=0;s<mdnew->nShapeFunctions();s++) {
          fpnew(s,0)=mdnew->shapeDerivative(s,0); 
          table[nmax-n][s][i][j]=(fpnew(s,0)[i]-fp(s,0)[i])/h;
          REAL p=2.;
          for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
            table[m][s][i][j]=table[m-1][s][i][j]
                          +(table[m-1][s][i][j]-table[m][s][i][j])/(p-1.);
          }
          fpp(s,0)[i][j]=table[nmax-1][s][i][j];
        }
        delete mdnew;
      }
    }
  }
}

void richardsonHigherOrderShapeDerivs(int nmax,const Mapping &mapping,
const SHAPE_CLASS *e,const Point<SPACEDIM> &x,const Array<2,REAL> &f,
Array<2,Tensor<1,SPACEDIM> > &fp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_VALUES;
  nmax=max(1,nmax);
  for (int i=0;i<SPACEDIM;i++) {
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.size(0));
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[i]+=h;
      Quadrature<SPACEDIM> q(xnew);
      MappingDataBase *mdnew=
        mapping.getMappingData(e->numberVertices(),update_flags,&q);
      Array<2,REAL> fnew(mdnew->nHigherOrderShapeFunctions(),1);
      for (int s=0;s<mdnew->nHigherOrderShapeFunctions();s++) {
        fnew(s,0)=mdnew->higherOrderShapeValue(s,0); 
        table[nmax-n][s][i]=(fnew(s,0)-f(s,0))/h;
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          table[m][s][i]=table[m-1][s][i]
                        +(table[m-1][s][i]-table[m][s][i])/(p-1.);
        }
        fp(s,0)[i]=table[nmax-1][s][i];
      }
      delete mdnew;
    }
  }
}

void richardsonHigherOrderShapeSecondDerivs(int nmax,
const Mapping &mapping,const SHAPE_CLASS *e,const Point<SPACEDIM> &x,
const Array<2,Tensor<1,SPACEDIM> > &fp,
Array<2,Tensor<2,SPACEDIM> > &fpp) {
  UPDATE_FLAGS update_flags=UPDATE_TRANSFORMATION_GRADIENTS;
  nmax=max(1,nmax);
  for (int j=0;j<SPACEDIM;j++) {
    for (int i=0;i<SPACEDIM;i++) {
      NumPtr<NumPtr<Tensor<2,SPACEDIM> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.size(0));
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
        Quadrature<SPACEDIM> q(xnew);
        MappingDataBase *mdnew=
          mapping.getMappingData(e->numberVertices(),update_flags,&q);
        Array<2,Tensor<1,SPACEDIM> >
          fpnew(mdnew->nHigherOrderShapeFunctions(),1);
        for (int s=0;s<mdnew->nHigherOrderShapeFunctions();s++) {
          fpnew(s,0)=mdnew->higherOrderShapeDerivative(s,0); 
          table[nmax-n][s][i][j]=(fpnew(s,0)[i]-fp(s,0)[i])/h;
          REAL p=2.;
          for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
            table[m][s][i][j]=table[m-1][s][i][j]
                          +(table[m-1][s][i][j]-table[m][s][i][j])/(p-1.);
          }
          fpp(s,0)[i][j]=table[nmax-1][s][i][j];
        }
        delete mdnew;
      }
    }
  }
}

void checkMappingContinuity(Tessellation &tess,
const Mapping &mapping,UPDATE_FLAGS update_flags) {
//TRACER_CALL(t,"checkMappingContinuity");
//MappingDataBase::use_full_degree_mapping_on_all_cells=true;
/*
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL winsize=0.5;
  XYGraphTool gt("checkMappingContinuity on edges","x","y",0.,1.,0.,1.,
    &cmap,0,winsize);
*/
  tess.firstShape();
  const SHAPE_CLASS *e0=tess.presentShape();
  const LEVEL_CLASS *level=tess.presentLevel();
  tess.nextShape();
  const SHAPE_CLASS *e1=tess.presentShape();
  int f=e0->sharedFaceElementIndex(e1);
  const CODIMENSION1_SHAPE_CLASS *face=
    level->CODIMENSION1_LEVEL_CLASS::getShape(f);
//int degree=mapping.getDegree();
//int npts_1d=static_cast<int>(1./gt.getPixelWidth());
  int npts_1d=0;

  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (int l=0;l<face->numberCodimension1Shapes();l++) {
    int line_index=face->codimension1ShapeIndex(l);
    const Shape1 *line=level->Level1::getShape(line_index);
    Point<SPACEDIM> line_v0=vertices[line->vertexIndex(0)];
    Point<SPACEDIM> line_v1=vertices[line->vertexIndex(1)];

    NumPtr<Point<SPACEDIM> > xq(npts_1d+1);
    REAL r=drand48();
    for (int q=0;q<=npts_1d;q++) {
      r=drand48();
      xq[q]=line_v0*r+line_v1*(1.-r);
    }

    Point<SPACEDIM> point(D_DECL(0.,0.,0.));
    Quadrature<SPACEDIM> point_quadrature(point);
    MappingDataBase *md0=mapping.getMappingData(e0->numberVertices(),
      update_flags,&point_quadrature);
    md0->computeMappingSupportPoints(e0,tess,level);
    if (md0->requiresShapeToInitialize()) md0->computeShapesFor(e0);
    NumPtr<Point<SPACEDIM> > xq0(npts_1d+1);
    for (int q=0;q<=npts_1d;q++) {
      xq0[q]=point;
      md0->canonicalPoint(e0,e0->diameter(tess.getVertices()),xq[q],
        xq0[q]);
    }

    Quadrature<SPACEDIM> qf0(xq0);
    md0->computeShapes(&qf0);
    if (md0->requiresShapeToInitialize()) md0->computeShapesFor(e0);
    NumPtr<Point<SPACEDIM> > yq0(npts_1d+1);
    NumPtr<REAL> wq0(npts_1d+1);
    wq0.initialize(1.);
    bool quadrature_points_ok=false;
    bool JxW_values_ok=false;
    md0->computeFEValues(e0,tess,level,yq0,wq0,quadrature_points_ok,
      JxW_values_ok);
    CHECK_TEST(quadrature_points_ok);

    MappingDataBase *md1=mapping.getMappingData(e1->numberVertices(),
      update_flags,&point_quadrature);
    md1->computeMappingSupportPoints(e1,tess,level);
    if (md1->requiresShapeToInitialize()) md1->computeShapesFor(e1);
    NumPtr<Point<SPACEDIM> > xq1(npts_1d+1);
    for (int q=0;q<=npts_1d;q++) {
      xq1[q]=point;
      md1->canonicalPoint(e1,e1->diameter(tess.getVertices()),xq[q],
        xq1[q]);
    }

    Quadrature<SPACEDIM> qf1(xq1);
    md1->computeShapes(&qf1);
    if (md1->requiresShapeToInitialize()) md1->computeShapesFor(e1);
    NumPtr<Point<SPACEDIM> > yq1(npts_1d+1);
    NumPtr<REAL> wq1(npts_1d+1);
    wq1.initialize(1.);
    quadrature_points_ok=false;
    JxW_values_ok=false;
    md1->computeFEValues(e1,tess,level,yq1,wq1,quadrature_points_ok,
      JxW_values_ok);
    CHECK_TEST(quadrature_points_ok);

    REAL error=0.;
    for (int q=0;q<=npts_1d;q++) {
      error+=pow(yq0[q].distance(yq1[q]),2);
    }
    error=sqrt(error/static_cast<REAL>(npts_1d+1))/DBL_EPSILON;
    if (error>64.) {
      cout << "\tdifference between coordinate maps on shared face = "
           << error << endl;
    }
    delete md0;
    delete md1;
  }

  int npts=face->numberEquallySpacedLatticePoints(npts_1d);
  NumPtr<Point<SPACEDIM> > xq(npts);
  NumPtr<Point<SPACEDIM> > face_vertex(face->numberVertices());
  for (int i=0;i<face->numberVertices();i++) {
    face_vertex[i]=vertices[face->vertexIndex(i)];
  }
  for (int q=0;q<npts;q++) {
    switch (face->numberVertices()) {
      case 3: {
        REAL r0=drand48();
        REAL r1=drand48()*(1.-r0);
        REAL r2=1.-r0-r1;
        xq[q]=face_vertex[0]*r0+face_vertex[1]*r1+face_vertex[2]*r2;
        break;
      }
      case 4: {
        REAL r0=drand48();
        REAL r1=drand48();
        xq[q]=face_vertex[0]*r0*r1+face_vertex[1]*(1.-r0)*r1
             +face_vertex[2]*r0*(1.-r1)+face_vertex[3]*(1.-r0)*(1.-r1);
        break;
      }
      default:
        OBSOLETE("bad number of face vertices");
    }
  }

  Point<SPACEDIM> point(D_DECL(0.,0.,0.));
  Quadrature<SPACEDIM> point_quadrature(point);
  MappingDataBase *md0=mapping.getMappingData(e0->numberVertices(),
    update_flags,&point_quadrature);
  md0->computeMappingSupportPoints(e0,tess,level);
  if (md0->requiresShapeToInitialize()) md0->computeShapesFor(e0);
  NumPtr<Point<SPACEDIM> > xq0(npts);
  for (int q=0;q<npts;q++) {
    xq0[q]=point;
    md0->canonicalPoint(e0,e0->diameter(tess.getVertices()),xq[q],xq0[q]);
  }

  Quadrature<SPACEDIM> qf0(xq0);
  md0->computeShapes(&qf0);
  if (md0->requiresShapeToInitialize()) md0->computeShapesFor(e0);
  NumPtr<Point<SPACEDIM> > yq0(npts);
  NumPtr<REAL> wq0(npts);
  wq0.initialize(1.);
  bool quadrature_points_ok=false;
  bool JxW_values_ok=false;
  md0->computeFEValues(e0,tess,level,yq0,wq0,quadrature_points_ok,
    JxW_values_ok);
  CHECK_TEST(quadrature_points_ok);

  MappingDataBase *md1=mapping.getMappingData(e1->numberVertices(),
    update_flags,&point_quadrature);
  md1->computeMappingSupportPoints(e1,tess,level);
  if (md1->requiresShapeToInitialize()) md1->computeShapesFor(e1);
  NumPtr<Point<SPACEDIM> > xq1(npts);
  for (int q=0;q<npts;q++) {
    xq1[q]=point;
    md1->canonicalPoint(e1,e1->diameter(tess.getVertices()),xq[q],xq1[q]);
  }

  Quadrature<SPACEDIM> qf1(xq1);
  md1->computeShapes(&qf1);
  if (md1->requiresShapeToInitialize()) md1->computeShapesFor(e1);
  NumPtr<Point<SPACEDIM> > yq1(npts);
  NumPtr<REAL> wq1(npts);
  wq1.initialize(1.);
  quadrature_points_ok=false;
  JxW_values_ok=false;
  md1->computeFEValues(e1,tess,level,yq1,wq1,quadrature_points_ok,
    JxW_values_ok);
  CHECK_TEST(quadrature_points_ok);

  REAL error=0.;
  for (int q=0;q<npts;q++) {
    error+=pow(yq0[q].distance(yq1[q]),2);
  }
  error=sqrt(error/static_cast<REAL>(npts))/DBL_EPSILON;
  if (error>64.) {
    cout << "\tdifference between coordinate maps on shared face = "
         << error << endl;
  }
  delete md0;
  delete md1;
}

void debugTess(Tessellation &tess,const Mapping &mapping,
UPDATE_FLAGS update_flags,VolGraphTool &gt,REAL volume) {
//TRACER_CALL(t,"debugTess");
#ifdef DEBUG
//tess.printVertices();
//tess.printElements();
//tess.printCodimension1Elements();
//tess.printNeighbors();
//tess.checkNeighbors();
//tess.checkFaceVertices();
//tess.checkNeighborLevelAndIndex();
#endif
  volume/=static_cast<REAL>(tess.numberActiveShapes());
  gt.newPage();
  gt.expose();

//cf FEFaceValues::FEFaceValues
  map<int,MappingDataBase*> mapping_data_map;
  map<int,Quadrature<SPACEDIM>*> quadrature_map;
  TriangleFullNewtonCotesQuadrature tncq(mapping.getDegree()+1);
  NewtonCotesQuadrature<2> ncq(mapping.getDegree()+1);
#ifdef DEBUG
//cout << "\ttncq:" << endl;
//tncq.printOn(cout);
//for (int p=0;p<tncq.points().getNumber();p++) {
//  cout << "\ttncq.points[" << p << "] = " << tncq.points()[p] << endl;
//}
//for (int p=0;p<tncq.weights().getNumber();p++) {
//  cout << "\ttncq.weights[" << p << "] = " << tncq.weights()[p] << endl;
//}
//for (int p=0;p<ncq.points().getNumber();p++) {
//  cout << "\tncq.points[" << p << "] = " << ncq.points()[p] << endl;
//}
//for (int p=0;p<ncq.weights().getNumber();p++) {
//  cout << "\tncq.weights[" << p << "] = " << ncq.weights()[p] << endl;
//}
//cout << "\tncq:" << endl;
//ncq.printOn(cout);
#endif

  {
//  TRACER_CALL(t,"debugTess tetrahedron mapping_data");
    int nfp=4*tncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<4;f++) {
      Tetrahedron::staticProjectToFace(f,tncq.points(),q_points_face);
      weights_face.copyFrom(tncq.weights());
      q_points_face+=tncq.numberPoints();
      weights_face+=tncq.numberPoints();
    }   
    Quadrature<SPACEDIM> *qf=
      OPERATOR_NEW Quadrature<SPACEDIM>(q_points,weights);
    quadrature_map[4]=qf;
    mapping_data_map[4]=mapping.getMappingData(4,update_flags,qf);
#ifdef DEBUG
//  {
//    for (int p=0;p<q_points.getNumber();p++) {
//      cout << "\tq_points[" << p << "] = " << q_points[p] << endl;
//    }
//    for (int p=0;p<weights.getNumber();p++) {
//      cout << "\tweights[" << p << "] = " << weights[p] << endl;
//    }
//    cout << "\tqf:" << endl;
//    qf.printOn(cout);
//    cout << "\tmd" << endl;
//    map<int,MappingDataBase*>::iterator mdm_iterator=
//     mapping_data_map.find(4);
//    CHECK_TEST(mdm_iterator!=mapping_data_map.end());
//    MappingDataBase *md=mdm_iterator->second;
//    md->printOn(cout);
//  }
#endif
  }

//
  {
//  TRACER_CALL(t,"debugTess prism mapping_data");
    int nfp=2*tncq.numberPoints()+3*ncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<2;f++) {
      Prism::staticProjectToFace(f,tncq.points(),q_points_face);
      weights_face.copyFrom(tncq.weights());
      q_points_face+=tncq.numberPoints();
      weights_face+=tncq.numberPoints();
    }   
    for (int f=2;f<5;f++) {
      Prism::staticProjectToFace(f,ncq.points(),q_points_face);
      weights_face.copyFrom(ncq.weights());
      q_points_face+=ncq.numberPoints();
      weights_face+=ncq.numberPoints();
    }   
    Quadrature<SPACEDIM> *qf=
      OPERATOR_NEW Quadrature<SPACEDIM>(q_points,weights);
    quadrature_map[6]=qf;
    mapping_data_map[6]=mapping.getMappingData(6,update_flags,qf);
#ifdef DEBUG
//  {
//    for (int p=0;p<q_points.getNumber();p++) {
//      cout << "\tq_points[" << p << "] = " << q_points[p] << endl;
//    }
//    for (int p=0;p<weights.getNumber();p++) {
//      cout << "\tweights[" << p << "] = " << weights[p] << endl;
//    }
//    cout << "\tqf:" << endl;
//    qf.printOn(cout);
//    cout << "\tmd" << endl;
//    map<int,MappingDataBase*>::iterator mdm_iterator=
//     mapping_data_map.find(6);
//    CHECK_TEST(mdm_iterator!=mapping_data_map.end());
//    MappingDataBase *md=mdm_iterator->second;
//    md->printOn(cout);
//  }
#endif
  }
//

//
  {
//  TRACER_CALL(t,"debugTess hexahedron mapping_data");
    int nfp=6*ncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<6;f++) {
      Hexahedron::staticProjectToFace(f,ncq.points(),q_points_face);
      weights_face.copyFrom(ncq.weights());
      q_points_face+=ncq.numberPoints();
      weights_face+=ncq.numberPoints();
    }   
    Quadrature<SPACEDIM> *qf=
      OPERATOR_NEW Quadrature<SPACEDIM>(q_points,weights);
    quadrature_map[8]=qf;
    mapping_data_map[8]=mapping.getMappingData(8,update_flags,qf);
  }
//

  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();

//int npts[2]={2,16};
  int npts[2]={2,mapping.getDegree()};
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) // fix break statement too 
//BOOLEAN do_best=FALSE;
  {
    VolGraphTool::DrawDelimiter dd(do_best,&gt);
    for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
//    TRACER_CALL(t,"debugTess computeMappingSupportPoints");
      const SHAPE_CLASS *e=tess.presentShape();
      const LEVEL_CLASS *level=tess.presentLevel();
#ifdef DEBUG
//    cout << "\n\n\n new element" << endl;
//    e->printOn(cout);
#endif

/*
      {
        TRACER_CALL(t,"debugTess computeMappingSupportPoints");
        NumPtr<Point<SPACEDIM> > eslp;
        e->equallySpacedLatticePoints(mapping.getDegree(),eslp);
        Point<SPACEDIM> cpcp(0.,0.,0.);
        Quadrature<SPACEDIM> point_quadrature(cpcp);
#ifdef DEBUG
//      cout << "\t&point_quadrature = " << &point_quadrature << endl;
#endif
        MappingDataBase *md=mapping.getMappingData(e->numberVertices(),
          UPDATE_TRANSFORMATION_VALUES | UPDATE_TRANSFORMATION_GRADIENTS,
          &point_quadrature);
        md->computeMappingSupportPoints(e,tess,level);
#ifdef DEBUG
//      md->printOn(cout);
#endif
        for (int ii=0;ii<eslp.getNumber();ii++) 
        {
//        Point<SPACEDIM> cp(drand48(),drand48(),drand48());
//        if (e->numberVertices()<8) cp[1]*=1.-cp[0];
//        if (e->numberVertices()==4) cp[2]*=1.-cp[0]-cp[1];
          Point<SPACEDIM> cp=eslp[ii];
          point_quadrature.setPoint(0,cp);
          {
            md->computeShapes(&point_quadrature);
            if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
            Point<SPACEDIM> ep=md->elementPoint();
            Point<SPACEDIM> cp2(drand48(),drand48(),drand48());
            if (e->numberVertices()<8) cp2[1]*=1.-cp2[0];
            if (e->numberVertices()==4) cp2[2]*=1.-cp2[0]-cp2[1];
            cp2=cp*0.99+cp2*0.01; // make initial guess "close"
            md->canonicalPoint(e,e->diameter(tess.getVertices()),ep,cp2);
            REAL error=max(abs(cp2[0]-cp[0]),
                           max(abs(cp2[1]-cp[1]),abs(cp2[2]-cp[2])));
            if (error>134217728.*DBL_EPSILON) {
              cout << "\tcp = [" << cp << "], ep = [" << ep 
                   << "], cp2 = [" << cp2 
                   << "], error = " << error << " " << error/DBL_EPSILON
                   << endl;
              ASSERT(false);
            }
          }
        }
        delete md; md=0;
      }
*/
/* do not use:
      {
        Point<SPACEDIM> cp(drand48(),drand48(),drand48());
        if (e->numberVertices()<8) cp[1]*=1.-cp[0];
        if (e->numberVertices()==4) cp[2]*=1.-cp[0]-cp[1];
        Point<SPACEDIM> ep=
          mapping.elementPoint(e,tess,tess.presentLevel(),cp);
        Point<SPACEDIM> cp2=
          mapping.canonicalPoint(e,tess,tess.presentLevel(),ep);
        REAL error=max(abs(cp2[0]-cp[0]),
                       max(abs(cp2[1]-cp[1]),abs(cp2[2]-cp[2])));
        if (error>32.*DBL_EPSILON) {
          cout << "\tcp = [" << cp << "], ep = [" << ep 
               << "], cp2 = [" << cp2 
               << "], error = " << error << " " << error/DBL_EPSILON
               << endl;
          ASSERT(false);
        }
      }
*/

//
      {
//      TRACER_CALL(t,"debugTess check gradients");
        Point<SPACEDIM> cp(drand48(),drand48(),drand48());
        if (e->numberVertices()<8) cp[1]*=1.-cp[0];
        if (e->numberVertices()==4) cp[2]*=1.-cp[0]-cp[1];
//      have to change MappingDataBase from private to public
//      in order to debug here
        Point<SPACEDIM> cpnp=cp;
        Quadrature <SPACEDIM> q(cp);
        Quadrature <SPACEDIM> qnp(cpnp);
        MappingDataBase *md=
          mapping.getMappingData(e->numberVertices(),update_flags,&q);
        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
#ifdef DEBUG
//      md->printOn(cout);
#endif

        Array<2,REAL> sv(md->nShapeFunctions(),1);
        Array<2,Tensor<1,SPACEDIM> > sg(md->nShapeFunctions(),1);
        Array<2,Tensor<2,SPACEDIM> > sgg(md->nShapeFunctions(),1);
        for (int s=0;s<md->nShapeFunctions();s++) {
          sv(s,0)=md->shapeValue(s,0);
          sg(s,0)=md->shapeDerivative(s,0);
          sgg(s,0)=md->shapeSecondDerivative(s,0);
        }
        Array<2,Tensor<1,SPACEDIM> > sg_dif(md->nShapeFunctions(),1);
        richardsonShapeDerivs(mapping.getDegree(),mapping,e,cp,sv,sg_dif);
        for (int d=0;d<SPACEDIM;d++) {
//        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
          for (int s=0;s<md->nShapeFunctions();s++) {
            REAL exact=sg(s,0)[d];
            REAL deriv=sg_dif(s,0)[d];
            REAL error=abs(exact-deriv);
            if (abs(exact)>0.) error/=abs(exact);
            if (error>8192.*DBL_EPSILON) {
              cout << "\tsg(" << s << ",0)[" << d << "] = " << exact 
                   << " " << deriv << " error = " << error << " " 
                   << error/DBL_EPSILON << endl;
              CHECK_TEST(error<=8192.*DBL_EPSILON);
            }
          }
        }
        Array<2,Tensor<2,SPACEDIM> > sgg_dif(md->nShapeFunctions(),1);
        richardsonShapeSecondDerivs(mapping.getDegree(),mapping,e,cp,sg,
          sgg_dif);
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            for (int s=0;s<md->nShapeFunctions();s++) {
              REAL exact=sgg(s,0)[i][j];
              REAL deriv=sgg_dif(s,0)[i][j];
              REAL error=abs(exact-deriv);
              REAL den=max(abs(exact),abs(deriv));
              if (den>0.) error/=den;
              if (error>DBL_EPSILON*8192.) {
                cout << "\tsgg(" << s << ",0)[" << i << "][" << j
                     << "] = " << exact << " " << deriv << " error = "
                     << error << " " << error/DBL_EPSILON << endl;
              }
            }
          }
        }
        if (md->nHigherOrderShapeFunctions()>0) {
          Array<2,REAL> hosv(md->nHigherOrderShapeFunctions(),1);
          Array<2,Tensor<1,SPACEDIM> > 
            hosg(md->nHigherOrderShapeFunctions(),1);
          Array<2,Tensor<2,SPACEDIM> > 
            hosgg(md->nHigherOrderShapeFunctions(),1);
          for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
            hosv(s,0)=md->higherOrderShapeValue(s,0);
            hosg(s,0)=md->higherOrderShapeDerivative(s,0);
            hosgg(s,0)=md->higherOrderShapeSecondDerivative(s,0);
          }
          Array<2,Tensor<1,SPACEDIM> >
            hosg_dif(md->nHigherOrderShapeFunctions(),1);
          richardsonHigherOrderShapeDerivs(mapping.getDegree(),mapping,e,
            cp,hosv,hosg_dif);
          for (int d=0;d<SPACEDIM;d++) {
            for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
              REAL exact=hosg(s,0)[d];
              REAL deriv=hosg_dif(s,0)[d];
              REAL error=abs(exact-deriv);
              if (abs(exact)>0.) error/=abs(exact);
              if (error>1048576.*DBL_EPSILON) {
                cout << "\thosg(" << s << ",0)[" << d << "] = " << exact 
                     << " " << deriv << " error = " << error << " "
                     << error/DBL_EPSILON << endl;
                CHECK_TEST(error<=1048576.*DBL_EPSILON);
              }
            }
          }
          Array<2,Tensor<2,SPACEDIM> >
            hosgg_dif(md->nHigherOrderShapeFunctions(),1);
          richardsonHigherOrderShapeSecondDerivs(mapping.getDegree(),
            mapping,e,cp,hosg,hosgg_dif);
          for (int j=0;j<SPACEDIM;j++) {
            for (int i=0;i<SPACEDIM;i++) {
              for (int s=0;s<md->nShapeFunctions();s++) {
                REAL exact=hosgg(s,0)[i][j];
                REAL deriv=hosgg_dif(s,0)[i][j];
                REAL error=abs(exact-deriv);
                REAL den=max(abs(exact),abs(deriv));
                if (den>0.) error/=den;
                if (error>DBL_EPSILON*1048576.) {
                  cout << "\thosgg(" << s << ",0)[" << i << "][" << j
                       << "] = " << exact << " " << deriv << " error = "
                       << error << " " << error/DBL_EPSILON << endl;
                  CHECK_TEST(error<=DBL_EPSILON*1048576.);
                }
              }
            }
          }
        }
        delete md; md=0;
      }
//

//
      MappingDataBase *md=0;
      REAL dia=e->diameter(vertices);
      switch (e->numberVertices()) {
        case 4: {
//        TRACER_CALL(t,"debugTess"); 
//        TetrahedronProductGaussianQuadrature 
//          npgq(mapping.getDegree()*SPACEDIM);
          TetrahedronNewtonCotesQuadrature npgq(min(4,mapping.getDegree()));
#ifdef DEBUG
//        cout << "\tnpgq:" << endl;
//        npgq.printOn(cout);
//        const NumPtr<Point<SPACEDIM> > &npgq_points=npgq.points();
//        for (int q=0;q<npgq_points.getNumber();q++) {
//          cout << "\tnpgq_points[" << q << "] = " << npgq_points[q] 
//               << endl;
//        }
#endif
          md=mapping.getMappingData(e->numberVertices(),update_flags,
            &npgq);
          NumPtr<Point<SPACEDIM> > xq(npgq.numberPoints());
          NumPtr<REAL> jxwv(npgq.numberPoints());
          if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
          bool quadrature_points_ok=false;
          bool JxW_values_ok=false;
          md->computeFEValues(e,tess,level,xq,jxwv,quadrature_points_ok,
            JxW_values_ok);
#ifdef DEBUG
//        cout << "\tnpgq:" << endl;
//        npgq.printOn(cout);
//        for (int q=0;q<xq.getNumber();q++) {
//          cout << "\txq[" << q << "] = " << xq[q] << endl;
//        }
//        for (int q=0;q<xq.getNumber();q++) {
//          cout << "\tjxwv[" << q << "] = " << jxwv[q] << endl;
//        }
//        for (int v=0;v<e->numberVertices();v++) {
//          cout << "\tvertices[" << e->vertexIndex(v) << "] = " 
//               << vertices[e->vertexIndex(v)] << endl;
//        }
#endif
          REAL quad=0.;
          for (int q=0;q<npgq.numberPoints();q++) {
            quad+=jxwv[q];
#ifdef DEBUG
//          Point<SPACEDIM> ep=
//            mapping.elementPoint(e,tess,level,npgq.point(q));
//          REAL error=vertices[e->vertexIndex(q)].distance(ep);
//          if (error>DBL_EPSILON) {
//            cout << "\tvertex[" << q << "] = " 
//                 << vertices[e->vertexIndex(q)] << "," << ep 
//                 << " error = " << error << " " << error/DBL_EPSILON 
//                 << endl;
//          }
#endif
            Tensor<2,SPACEDIM> unit;
            for (int j=0;j<SPACEDIM;j++) {
              for (int i=0;i<SPACEDIM;i++) {
                unit[i][j]=(i==j ? 1. : 0.);
              }
            }
            Tensor<2,SPACEDIM> dxdxi;
            md->preMultiplyByJacobian(dxdxi,q,unit);
            Tensor<2,SPACEDIM> dxidx_times_dxdxi;
            md->postMultiplyByJacobianInverse(dxidx_times_dxdxi,
              q,dxdxi);
            for (int j=0;j<SPACEDIM;j++) {
              for (int i=0;i<SPACEDIM;i++) {
                unit[i][j]=(i==j ? 1. : 0.);
                  REAL error=abs(dxidx_times_dxdxi[i][j]-unit[i][j]);
                if (error>128.*DBL_EPSILON) {
                  cout << "\tdxidx_times_dxdxi[" << i << "][" << j 
                       << "] = " << dxidx_times_dxdxi[i][j] << " " 
                       << unit[i][j] << " " << error << " " 
                       << error/DBL_EPSILON << endl;
                  ASSERT(false);
                }
              }
            }
          }
          cout << "\tvolume = " << volume << " " << quad << " error = "
               << abs(volume-quad) << endl;
          break;
        }
        case 6: {
//        TRACER_CALL(t,"debugTess"); 
          TriangleProductGaussianQuadrature 
            tpgq(mapping.getDegree()*SPACEDIM);
          GaussianQuadrature<1> gq(mapping.getDegree()*SPACEDIM);
          ProductQuadrature npgq(tpgq,gq);
          md=mapping.getMappingData(e->numberVertices(),update_flags,
            &npgq);
          NumPtr<Point<SPACEDIM> > xq(npgq.numberPoints());
          NumPtr<REAL> jxwv(npgq.numberPoints());
          if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
          bool quadrature_points_ok=false;
          bool JxW_values_ok=false;
          md->computeFEValues(e,tess,level,xq,jxwv,quadrature_points_ok,
            JxW_values_ok);
#ifdef DEBUG
//        cout << "\tnpgq:" << endl;
//        npgq.printOn(cout);
//        for (int q=0;q<xq.getNumber();q++) {
//          cout << "\txq[" << q << "] = " << xq[q] << endl;
//        }
//        for (int q=0;q<xq.getNumber();q++) {
//          cout << "\tjxwv[" << q << "] = " << jxwv[q] << endl;
//        }
//        for (int v=0;v<e->numberVertices();v++) {
//          cout << "\tvertices[" << e->vertexIndex(v) << "] = " 
//               << vertices[e->vertexIndex(v)] << endl;
//        }
#endif
          REAL quad=0.;
          for (int q=0;q<npgq.numberPoints();q++) {
            quad+=jxwv[q];
//          Point<SPACEDIM> ep=
//            mapping.elementPoint(e,tess,level,npgq.point(q));
//          REAL error=vertices[e->vertexIndex(q)].distance(ep);
//          if (error>DBL_EPSILON) {
//            cout << "\tvertex[" << q << "] = " 
//                 << vertices[e->vertexIndex(q)] << "," << ep 
//                 << " error = " << error << " " << error/DBL_EPSILON 
//                 << endl;
//          }
            Tensor<2,SPACEDIM> unit;
            for (int j=0;j<SPACEDIM;j++) {
              for (int i=0;i<SPACEDIM;i++) {
                unit[i][j]=(i==j ? 1. : 0.);
              }
            }
            Tensor<2,SPACEDIM> dxdxi;
            md->preMultiplyByJacobian(dxdxi,q,unit);
            Tensor<2,SPACEDIM> dxidx_times_dxdxi;
            md->postMultiplyByJacobianInverse(dxidx_times_dxdxi,
              q,dxdxi);
            for (int j=0;j<SPACEDIM;j++) {
              for (int i=0;i<SPACEDIM;i++) {
                unit[i][j]=(i==j ? 1. : 0.);
                  REAL error=abs(dxidx_times_dxdxi[i][j]-unit[i][j]);
                if (error>8192.*DBL_EPSILON) {
                  cout << "\tdxidx_times_dxdxi[" << i << "][" << j 
                       << "] = " << dxidx_times_dxdxi[i][j] << " " 
                       << unit[i][j] << " " << error << " " 
                       << error/DBL_EPSILON << endl;
                  CHECK_TEST(error<=8192.*DBL_EPSILON);
                }
              }
            }
          }
          cout << "\tvolume = " << volume << " " << quad << " error = "
               << abs(volume-quad) << endl;
          break;
        }
        case 8: {
//        TRACER_CALL(t,"debugTess");
          GaussianQuadrature<SPACEDIM> gq(mapping.getDegree()*SPACEDIM);
          md=mapping.getMappingData(e->numberVertices(),update_flags,&gq);
          NumPtr<Point<SPACEDIM> > xq(gq.numberPoints());
          NumPtr<REAL> jxwv(gq.numberPoints());
          if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
          bool quadrature_points_ok=false;
          bool JxW_values_ok=false;
          md->computeFEValues(e,tess,level,xq,jxwv,quadrature_points_ok,
            JxW_values_ok);
          REAL quad=0.;
          for (int q=0;q<gq.numberPoints();q++) {
            quad+=jxwv[q];
//          Point<SPACEDIM> ep=mapping.elementPoint(e,tess,level,
//            Quadrilateral::canonicalVertex(q));
//          REAL error=vertices[e->vertexIndex(q)].distance(ep);
//          if (error>256.*DBL_EPSILON) {
//            cout << "\tvertex[" << q << "] = " 
//                 << vertices[e->vertexIndex(q)] << "," << ep 
//                 << " error = " << error << " " << error/DBL_EPSILON 
//                 << endl;
//          }
            Tensor<2,SPACEDIM> unit;
            for (int j=0;j<SPACEDIM;j++) {
              for (int i=0;i<SPACEDIM;i++) {
                unit[i][j]=(i==j ? 1. : 0.);
              }
            }
            Tensor<2,SPACEDIM> dxdxi;
            md->preMultiplyByJacobian(dxdxi,q,unit);
            Tensor<2,SPACEDIM> dxidx_times_dxdxi;
            md->postMultiplyByJacobianInverse(dxidx_times_dxdxi,
              q,dxdxi);
            for (int j=0;j<SPACEDIM;j++) {
              for (int i=0;i<SPACEDIM;i++) {
                unit[i][j]=(i==j ? 1. : 0.);
                REAL error=abs(dxidx_times_dxdxi[i][j]-unit[i][j]);
                if (error>32.*DBL_EPSILON) {
                  cout << "\tdxidx_times_dxdxi[" << i << "][" << j 
                       << "] = " << dxidx_times_dxdxi[i][j] << " " 
                       << unit[i][j] << " " << error << " "
                       << error/DBL_EPSILON << endl;
                  ASSERT(false);
                }
              }
            }
          }
          REAL area=0.;
          if (e->elementDataIndex()==2) {
            area=0.4*radius*radius;
          } else {
            area=(M_PI*0.25-0.1)*radius*radius;
          }
          cout << "\tarea = " << area << " " << quad << " error = "
               << abs(area-quad) << endl;
          delete md; md=0;
          break;
        }
        default:
          OBSOLETE("unknown number vertices");
      }
      if (md!=0) delete md; md=0;
//

//
//    from FEFaceValues::reinit
      map<int,MappingDataBase*>::iterator mdm_iterator=
        mapping_data_map.find(e->numberVertices());
      CHECK_TEST(mdm_iterator!=mapping_data_map.end());
      md=mdm_iterator->second;
      int offset=0;
      for (int f=0;f<e->numberCodimension1Shapes();f++) {
//      TRACER_CALL(t,"debugTess");
        if (level->neighborLevel(e,f)==0) {
          NumPtr<Point<SPACEDIM> > quadrature_points;
          NumPtr<REAL> JxW_values;
          NumPtr<Tensor<1,SPACEDIM> > boundary_forms;
          NumPtr<Point<SPACEDIM> > normal_vectors;
          bool is_subface=false;
          bool quadrature_points_ok=false;
          bool jxw_values_ok=false;
          bool boundary_forms_ok=false;
          bool normal_vectors_ok=false;
          if (e->numberVerticesOnFace(f)==4) {
            int ncqpts=ncq.numberPoints();
            quadrature_points.allocate(ncqpts);
            JxW_values.allocate(ncqpts);
            boundary_forms.allocate(ncqpts);
            normal_vectors.allocate(ncqpts);
//          md->fillFEFaceValues(f*ncq.numberPoints(),e,tess,level,f,ncq,
//            quadrature_points,JxW_values,boundary_forms,normal_vectors);
            md->computeFEFaceValues(e,tess,level,f,is_subface,ncqpts,
              offset,ncq.weights(),quadrature_points,JxW_values,
              boundary_forms,normal_vectors,quadrature_points_ok,
              jxw_values_ok,boundary_forms_ok,normal_vectors_ok);
          } else {
            int tncqpts=tncq.numberPoints();
            quadrature_points.allocate(tncqpts);
            JxW_values.allocate(tncqpts);
            boundary_forms.allocate(tncqpts);
            normal_vectors.allocate(tncqpts);
//          md->fillFEFaceValues(f*tncq.numberPoints(),e,tess,level,f,tncq,
//            quadrature_points,JxW_values,boundary_forms,normal_vectors);
            md->computeFEFaceValues(e,tess,level,f,is_subface,tncqpts,
              offset,tncq.weights(),quadrature_points,JxW_values,
              boundary_forms,normal_vectors,quadrature_points_ok,
              jxw_values_ok,boundary_forms_ok,normal_vectors_ok);
          }
#ifdef DEBUG
//        cout << "\te:" << endl;
//        e->printOn(cout);
//        cout << "\tf = " << f << endl;
//        for (int q=0;q<quadrature_points.getNumber();q++) {
//          cout << "\tquadrature_points[" << q << "] = "
//               << quadrature_points[q] << endl;
//        }
//        for (int q=0;q<normal_vectors.getNumber();q++) {
//          cout << "\tnormal_vectors[" << q << "] = "
//               << normal_vectors[q] << endl;
//        }
#endif
          for (int q=0;q<quadrature_points.getNumber();q++) {
            Point<SPACEDIM> hq=normal_vectors[q];
            Point<SPACEDIM> qp=quadrature_points[q];
            qp/=sqrt(qp.square());
            REAL err=hq.distance(qp);
            if (err>16.) {
              cout << "\tnormal[f=" << f << ",q=" << q << "] = "
                   << hq << " " << qp << " " << " err = " << err << endl;
              ASSERT(false);
            }
          }
        }
        offset+=(e->numberVerticesOnFace(f)==4 
                ? ncq.numberPoints() : tncq.numberPoints());
      }
//

      int nv=e->numberVertices();
      int degree=mapping.getDegree();
      int edi=e->elementDataIndex();
#ifdef DEBUG
//    bool debug_now=(degree==2 && (edi==5 || edi==7 || edi==11));
//    if (debug_now) {
//      const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
//      for (int v=0;v<nv;v++) {
//        cout << "\tvertex[" << v << "] = " 
//             << vertices[e->vertexIndex(v)] << endl;
//      }
//      NumPtr<Point<3> > cp;
//      e->equallySpacedLatticePoints(mapping.getDegree(),cp);
//      Quadrature<3> quadrature(cp);
//      MappingDataBase *mdata=mapping.getMappingData(nv,
//        UPDATE_TRANSFORMATION_VALUES,&quadrature);
//      mdata->computeMappingSupportPoints(e,tess,level);
//      if (mdata->requiresShapeToInitialize()) {
//        mdata->computeShapesFor(e);
//      }
//      mdata->printOn(cout);
//      NumPtr<Point<3> > ep(cp.getNumber());
//      mdata->elementPoints(ep);
//      for (int i=0;i<ep.getNumber();i++) {
//        cout << "\tep[" << i << "] = " << ep[i] << endl;
//      }
//      delete mdata;
//    }
#endif
      for (int f=0;f<e->numberCodimension1Shapes();f++) {
        NumPtr<Point<SPACEDIM> > cv(e->numberVerticesOnFace(f));
        for (int v=0;v<cv.getNumber();v++) {
          cv[v]=canonicalVertex(nv,e->elementVertexFromFace(f,v));
        }
#ifdef DEBUG
//      if (debug_now) {
//        cout << "\tf = " << f << endl;
//        int face_index=e->codimension1ShapeIndex(f);
//        const BoundaryCondition *bc=
//          level->Level2::getBoundaryCondition(face_index);
//        if (bc!=0) {
//          bc->printOn(cout);
//          for (int v=0;v<e->numberVerticesOnFace(f);v++) {
//            cout << "\tconstraint(" 
//                 << vertices[e->elementVertexFromFace(f,v)] << ") = "
//                 << bc->getBoundary()->
//                    constraint(vertices[e->elementVertexFromFace(f,v)]) 
//                 << endl;
//          }
//        } else cout << "\tface is not on boundary" << endl;
//        for (int v=0;v<cv.getNumber();v++) {
//          cout << "\tcv[" << v << "] = " << cv[v] << endl;
//        }
//      }
#endif
        Point<SPACEDIM> dc1=(cv[1]-cv[0])/static_cast<REAL>(npts[do_best]);
        Point<SPACEDIM> dc2=(cv[2]-cv[0])/static_cast<REAL>(npts[do_best]);
        Point<SPACEDIM> cj=cv[0];
        if (cv.getNumber()==4) {
//
          for (int j=0;j<npts[do_best];j++,cj+=dc2) {
            Point<SPACEDIM> ci=cj;
            for (int i=0;i<npts[do_best];i++,ci+=dc1) {
              Point<SPACEDIM> ep0=
                mapping.elementPoint(e,tess,tess.presentLevel(),ci);
              Point<SPACEDIM> ep1=
                mapping.elementPoint(e,tess,tess.presentLevel(),ci+dc1);
              Point<SPACEDIM> ep2=
                mapping.elementPoint(e,tess,tess.presentLevel(),ci+dc2);
              Point<SPACEDIM> ep3=mapping.elementPoint(e,tess,
                tess.presentLevel(),ci+dc1+dc2);
              Vector3 v0(ep0[0],ep0[1],ep0[2]);
              Vector3 v1(ep1[0],ep1[1],ep1[2]);
              Vector3 v2(ep2[0],ep2[1],ep2[2]);
              Vector3 v3(ep3[0],ep3[1],ep3[2]);
              Vector3 vave=(v0+v1+v2+v3)*0.25;
              gt.colorTriangle(0.5,v0,v1,vave);
              gt.colorTriangle(0.5,vave,v1,v0);
              gt.colorTriangle(0.5,v1,v3,vave);
              gt.colorTriangle(0.5,vave,v3,v1);
              gt.colorTriangle(0.5,v3,v2,vave);
              gt.colorTriangle(0.5,vave,v2,v3);
              gt.colorTriangle(0.5,v2,v0,vave);
              gt.colorTriangle(0.5,vave,v0,v2);
            }
          }
//
        } else {
//
          for (int j=0;j<npts[do_best];j++,cj+=dc2) {
            Point<SPACEDIM> ci=cj;
            for (int i=0;i+j<npts[do_best];i++,ci+=dc1) {
              Point<SPACEDIM> ep0=
                mapping.elementPoint(e,tess,tess.presentLevel(),ci);
              Point<SPACEDIM> ep1=
                mapping.elementPoint(e,tess,tess.presentLevel(),ci+dc1);
              Point<SPACEDIM> ep2=
                mapping.elementPoint(e,tess,tess.presentLevel(),ci+dc2);
              Vector3 v0(ep0[0],ep0[1],ep0[2]);
              Vector3 v1(ep1[0],ep1[1],ep1[2]);
              Vector3 v2(ep2[0],ep2[1],ep2[2]);
              gt.colorTriangle(0.5,v0,v1,v2);
              gt.colorTriangle(0.5,v2,v1,v0);
              if (i+j+1<npts[do_best]) {
                Point<SPACEDIM> ep3=mapping.elementPoint(e,tess,
                  tess.presentLevel(),ci+dc1+dc2);
                Vector3 v3(ep3[0],ep3[1],ep3[2]);
                gt.colorTriangle(0.5,v3,v2,v1);
                gt.colorTriangle(0.5,v1,v2,v3);
              }
            }
          }
//
        }
      }
    }
    if (do_best==TRUE) break;
  }
  gt.eventLoop();

  for (map<int,Quadrature<SPACEDIM>*>::iterator
  q_iterator=quadrature_map.begin();
  q_iterator!=quadrature_map.end();q_iterator++) {
    delete q_iterator->second; q_iterator->second=0;
  }
  for (map<int,MappingDataBase*>::iterator
  mdm_iterator=mapping_data_map.begin();
  mdm_iterator!=mapping_data_map.end();mdm_iterator++) {
    delete mdm_iterator->second; mdm_iterator->second=0;
  }
}


int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
  cout.precision(16);
  {
#ifdef MEM_DEBUG
    MemoryDebugger dm(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif
    UPDATE_FLAGS update_flags(UPDATE_VALUES | UPDATE_GRADIENTS |
      UPDATE_SECOND_DERIVATIVES | UPDATE_BOUNDARY_FORMS 
      | UPDATE_Q_POINTS | UPDATE_JxW_VALUES
      | UPDATE_NORMAL_VECTORS | UPDATE_MAPPING_JACOBIAN_INVERSE
      | UPDATE_MAPPING_JACOBIAN | UPDATE_MAPPING_SECOND_DERIVATIVE
      | UPDATE_TRANSFORMATION_VALUES 
      | UPDATE_TRANSFORMATION_GRADIENTS);
    REAL r=drand48();
    bool set_face_ids=true;
//
    { // Tetrahedron mapping vertices
//    TRACER_CALL(t,"Tetrahedron mapping vertices");
      MappingL mapping(1);
      Tessellation tess;
      Point<SPACEDIM> low(0.,0.,0.);
      Point<SPACEDIM> edge[SPACEDIM];
        edge[0]=Point<SPACEDIM>(3.,0.,0.);
        edge[1]=Point<SPACEDIM>(0.,2.,0.);
        edge[2]=Point<SPACEDIM>(0.,0.,1.);
      int number_elements[3]={2,2,2};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpipedTetrahedra(tess,low,
        edge,number_elements,true,boundary,boundary_condition);
      NumPtr<REAL> w(4);
      NumPtr<Point<SPACEDIM> > eslp;
      Tetrahedron::staticEquallySpacedLatticePoints(1,eslp);
      w.initialize(1.);
      Quadrature<SPACEDIM> quad(eslp,w);
      MappingDataBase *md=mapping.getMappingData(
        Tetrahedron::verticesPerShape(),update_flags,&quad);
      md->computeShapes(&quad);
      for (int q=0;q<4;q++) {
        for (int s=0;s<md->nShapeFunctions();s++) {
          REAL sv=md->shapeValue(s,q);
          REAL error=(s==q ? abs(sv-1.) : abs(sv) );
          if (error>DBL_EPSILON) {
            cout << "\teslp[" << q << "] = " << eslp[q] << endl;
            cout << "\tsv[" << s << "," << q << "] = " << sv << endl;
            OBSOLETE("bad shape value");
          }
        }
      }
      for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
        const SHAPE_CLASS *e=tess.presentShape();
        const LEVEL_CLASS *level=tess.presentLevel();
        const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
        NumPtr<Point<SPACEDIM> > qp(4);
        NumPtr<REAL> JxW_values(4);
        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
        bool quadrature_points_ok=false;
        bool JxW_values_ok=false;
        md->computeFEValues(e,tess,level,qp,JxW_values,
          quadrature_points_ok,JxW_values_ok);
        for (int q=0;q<4;q++) {
          REAL error=qp[q].distance(vertices[e->vertexIndex(q)]);
          if (error>DBL_EPSILON) {
            cout << "\tqp[" << q << "] = " << qp[q] << endl;
            cout << "\tvertices[" << e->vertexIndex(q) << "] = " 
                 << vertices[e->vertexIndex(q)] << endl;
            OBSOLETE("bad mapped quadrature point");
          }
        }
      }
      delete md;
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
//
    { // Prism mapping vertices
//    TRACER_CALL(t,"Prism mapping vertices");
      MappingL mapping(1);
      Tessellation tess;
      Point<SPACEDIM> low(0.,0.,0.);
      Point<SPACEDIM> edge[3];
        edge[0]=Point<SPACEDIM>(3.,0.,0.);
        edge[1]=Point<SPACEDIM>(0.,2.,0.);
        edge[2]=Point<SPACEDIM>(0.,0.,1.);
      int number_elements[3]={2,2,2};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpipedPrisms(tess,low,
        edge,number_elements,0,true,boundary,boundary_condition);
      NumPtr<REAL> w(6);
      NumPtr<Point<SPACEDIM> > eslp;
      Prism::staticEquallySpacedLatticePoints(1,eslp);
      w.initialize(1.);
      Quadrature<SPACEDIM> quad(eslp,w);
      MappingDataBase *md=mapping.getMappingData(Prism::verticesPerShape(),
        update_flags,&quad);
      md->computeShapes(&quad);
      for (int q=0;q<6;q++) {
        for (int s=0;s<md->nShapeFunctions();s++) {
          REAL sv=md->shapeValue(s,q);
          REAL error=(s==q ? abs(sv-1.) : abs(sv) );
          if (error>DBL_EPSILON) {
            cout << "\teslp[" << q << "] = " << eslp[q] << endl;
            cout << "\tsv[" << s << "," << q << "] = " << sv << endl;
            OBSOLETE("bad shape value");
          }
        }
      }
      for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
        const SHAPE_CLASS *e=tess.presentShape();
        const LEVEL_CLASS *level=tess.presentLevel();
        const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
        NumPtr<Point<SPACEDIM> > qp(6);
        NumPtr<REAL> JxW_values(6);
        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
        bool quadrature_points_ok=false;
        bool JxW_values_ok=false;
        md->computeFEValues(e,tess,level,qp,JxW_values,
          quadrature_points_ok,JxW_values_ok);
        for (int q=0;q<6;q++) {
          REAL error=qp[q].distance(vertices[e->vertexIndex(q)]);
          if (error>DBL_EPSILON) {
            cout << "\tqp[" << q << "] = " << qp[q] << endl;
            cout << "\tvertices[" << e->vertexIndex(q) << "] = " 
                 << vertices[e->vertexIndex(q)] << endl;
            OBSOLETE("bad mapped quadrature point");
          }
        }
      }
      delete md;
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
//
    { // Hexahedron mapping vertices
//    TRACER_CALL(t,"Hexahedron mapping vertices");
      MappingL mapping(1);
      Tessellation tess;
      Point<SPACEDIM> low(0.,0.,0.);
      Point<SPACEDIM> edge[3];
        edge[0]=Point<SPACEDIM>(3.,0.,0.);
        edge[1]=Point<SPACEDIM>(0.,2.,0.);
        edge[2]=Point<SPACEDIM>(0.,0.,1.);
      int number_elements[3]={2,2,2};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpiped(tess,low,
        edge,number_elements,true,boundary,boundary_condition);
      NumPtr<REAL> w(8);
      NumPtr<Point<SPACEDIM> > eslp;
      Hexahedron::staticEquallySpacedLatticePoints(1,eslp);
      w.initialize(1.);
      Quadrature<SPACEDIM> quad(eslp,w);
      MappingDataBase *md=mapping.getMappingData(
        Hexahedron::verticesPerShape(),update_flags,&quad);
      md->computeShapes(&quad);
      for (int q=0;q<8;q++) {
        for (int s=0;s<md->nShapeFunctions();s++) {
          REAL sv=md->shapeValue(s,q);
          REAL error=(s==q ? abs(sv-1.) : abs(sv) );
          if (error>DBL_EPSILON) {
            cout << "\teslp[" << q << "] = " << eslp[q] << endl;
            cout << "\tsv[" << s << "," << q << "] = " << sv << endl;
            OBSOLETE("bad shape value");
          }
        }
      }
      for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
        const SHAPE_CLASS *e=tess.presentShape();
        const LEVEL_CLASS *level=tess.presentLevel();
        const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
        NumPtr<Point<SPACEDIM> > qp(8);
        NumPtr<REAL> JxW_values(8);
        if (md->requiresShapeToInitialize()) md->computeShapesFor(e);
        bool quadrature_points_ok=false;
        bool JxW_values_ok=false;
        md->computeFEValues(e,tess,level,qp,JxW_values,
          quadrature_points_ok,JxW_values_ok);
        for (int q=0;q<8;q++) {
          REAL error=qp[q].distance(vertices[e->vertexIndex(q)]);
          if (error>DBL_EPSILON) {
            cout << "\tqp[" << q << "] = " << qp[q] << endl;
            cout << "\tvertices[" << e->vertexIndex(q) << "] = " 
                 << vertices[e->vertexIndex(q)] << endl;
            OBSOLETE("bad mapped quadrature point");
          }
        }
      }
      delete md;
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//

    Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
    Vector3 high(radius*1.25,radius*1.25,radius*1.25);
    Palette pal;
    VolGraphTool gt("debugTess",low,high,pal,TRUE,0.5);
    gt.newPage();
    gt.expose();
    gt.setfgColor("red");

    for (int degree=1;degree<=5;degree++)
//  int degree=2;
    {
      cout << "\n***degree = " << degree << "***" << endl;
//
      { // Tetrahedron mapping equally spaced lattice points
//      TRACER_CALL(t,"Tetrahedron higher-order mapping");
        MappingL mapping(degree);
        int neslp=
          Tetrahedron::staticNumberEquallySpacedLatticePoints(degree);
        NumPtr<REAL> w(neslp);
        NumPtr<Point<SPACEDIM> > eslp;
        Tetrahedron::staticEquallySpacedLatticePoints(degree,eslp);
        w.initialize(1.);
        Quadrature<SPACEDIM> quad(eslp,w);
        MappingDataBase *md=mapping.getMappingData(
          Tetrahedron::verticesPerShape(),update_flags,&quad);
        md->computeShapes(&quad);
        if (!md->requiresShapeToInitialize()) {
#ifdef DEBUG
//        for (int q=0;q<neslp;q++) {
//          cout << "\tshapeValue(" << eslp[q] << ") =";
//          for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
//            cout << " " << md->higherOrderShapeValue(s,q);
//          }
//          cout << endl;
//        }
#endif
          for (int q=0;q<neslp;q++) {
            for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
              REAL sv=md->higherOrderShapeValue(s,q);
              REAL error=(s==q ? abs(sv-1.) : abs(sv) );
              if (error>1024.*DBL_EPSILON) {
                cout << "\teslp[" << q << "] = " << eslp[q] 
                     << " sv(" << s << "," << q << ") = " << sv
                     << " " << error << " " << error/DBL_EPSILON << endl;
                OBSOLETE("bad higher order shape value");
              }
            }
          }
        }
        delete md; md=0;
      }
//
//
      { // Prism mapping equally spaced lattice points
//      TRACER_CALL(t,"Prism higher-order mapping");
        MappingL mapping(degree);
        int neslp=
          Prism::staticNumberEquallySpacedLatticePoints(degree);
        NumPtr<REAL> w(neslp);
        NumPtr<Point<SPACEDIM> > eslp;
        Prism::staticEquallySpacedLatticePoints(degree,eslp);
        w.initialize(1.);
        Quadrature<SPACEDIM> quad(eslp,w);
        MappingDataBase *md=mapping.getMappingData(
          Prism::verticesPerShape(),update_flags,&quad);
        md->computeShapes(&quad);
        if (!md->requiresShapeToInitialize()) {
#ifdef DEBUG
//        if (md->nHigherOrderShapeFunctions()>0) {
//          for (int q=0;q<neslp;q++) {
//            cout << "\thigherOrderShapeValue(" << eslp[q] << ") =";
//            for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
//              cout << " " << md->higherOrderShapeValue(s,q);
//            }
//            cout << endl;
//          }
//        }
#endif
          for (int q=0;q<neslp;q++) {
            for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
              REAL sv=md->higherOrderShapeValue(s,q);
              REAL error=(s==q ? abs(sv-1.) : abs(sv) );
              if (error>1024.*DBL_EPSILON) {
                cout << "\teslp[" << q << "] = " << eslp[q] 
                     << " sv(" << s << "," << q << ") = " << sv
                     << " " << error << " " << error/DBL_EPSILON << endl;
                OBSOLETE("bad higher order shape value");
              }
            }
          }
        }
        delete md; md=0;
      }
//
//
      { // Hexahedron mapping equally spaced lattice points
//      TRACER_CALL(t,"Hexahedron higher-order mapping");
        MappingL mapping(degree);
        int neslp=
          Hexahedron::staticNumberEquallySpacedLatticePoints(degree);
        NumPtr<REAL> w(neslp);
        NumPtr<Point<SPACEDIM> > eslp;
        Hexahedron::staticEquallySpacedLatticePoints(degree,eslp);
        w.initialize(1.);
        Quadrature<SPACEDIM> quad(eslp,w);
        MappingDataBase *md=mapping.getMappingData(
          Hexahedron::verticesPerShape(),update_flags,&quad);
        md->computeShapes(&quad);
        if (!md->requiresShapeToInitialize()) {
#ifdef DEBUG
//        for (int q=0;q<neslp;q++) {
//          cout << "\tshapeValue(" << eslp[q] << ") =";
//          for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
//            cout << " " << md->higherOrderShapeValue(s,q);
//          }
//          cout << endl;
//        }
#endif
          for (int q=0;q<neslp;q++) {
            for (int s=0;s<md->nHigherOrderShapeFunctions();s++) {
              REAL sv=md->higherOrderShapeValue(s,q);
              REAL error=(s==q ? abs(sv-1.) : abs(sv) );
              if (error>1024.*DBL_EPSILON) {
                cout << "\teslp[" << q << "] = " << eslp[q] 
                     << " sv(" << s << "," << q << ") = " << sv
                     << " " << error << " " << error/DBL_EPSILON << endl;
                OBSOLETE("bad higher order shape value");
              }
            }
          }
        }
        delete md; md=0;
      }
//
      REAL volume=4.*M_PI*pow(radius,3)/3.;
//
      {
//      TRACER_CALL(t,"Tetrahedron MappingL");
        cout << "tetrahedra in ball" << endl;
        MappingL mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallTetrahedra(tess,
          center,radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags,gt,volume);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
//
//
      {
//      TRACER_CALL(t,"Prism +Tetrahedron MappingL");
        cout << "prisms + tetrahedra in ball" << endl;
        MappingL mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallPrismsWithTetrahedra(
          tess,center,radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags,gt,volume);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
//
//
      {
//      TRACER_CALL(t,"Prism + Hexahedron MappingL");
        cout << "prisms + hexahedra in ball" << endl;
        MappingL mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallPrismsWithHexahedra(
          tess,center,radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags,gt,volume);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
//
//
      {
//      TRACER_CALL(t,"Hexahedron MappingL");
        cout << "hexahedra in ball" << endl;
        MappingL mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBall(tess,center,radius,
          set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags,gt,volume);
        delete sub_cell_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
//
/* obsolete:
      {
        TRACER_CALL(t,"Tetrahedron MappingH");
        MappingH mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallTetrahedra(tess,center,
          radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags,gt,volume);
        delete sub_cell_data;
      }
*/
/* obsolete:
      {
        TRACER_CALL(t,"Prism MappingH");
        MappingH mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallPrismsWithTetrahedra(
          tess,center,radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags,gt,volume);
        delete sub_cell_data;
      }
*/
/* obsolete:
      {
        TRACER_CALL(t,"Prism MappingH");
        MappingH mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBallPrismsWithHexahedra(
          tess,center,radius,set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags,gt,volume);
        delete sub_cell_data;
      }
*/
/* obsolete:
      {
        TRACER_CALL(t,"Hexahedron MappingH");
        MappingH mapping(degree);
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *sub_cell_data=generateBall(tess,center,radius,
          set_face_ids,boundary,boundary_condition);
//      BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,b);
        debugTess(tess,mapping,update_flags,gt,volume);
        delete sub_cell_data;
      }
*/
//
      if (degree<=1) {//Shapes have multiple faces on boundar; lenoir fails
        PermutationGenerator pg(4);
        while (pg.hasMore()) {
          const NumPtr<int> &perm=pg.getNext();
          NumPtr<Point<SPACEDIM> > vertices(7);
          vertices[4]=Point<SPACEDIM>(0.,0.,0.);
          vertices[5]=Point<SPACEDIM>(1.,0.,0.);
          vertices[6]=Point<SPACEDIM>(0.,1.,0.);
          vertices[perm[0]]=Point<SPACEDIM>(0.,0.,.5);
          vertices[perm[1]]=Point<SPACEDIM>(1.,0.,.5);
          vertices[perm[2]]=Point<SPACEDIM>(0.,1.,.5);
          vertices[perm[3]]=Point<SPACEDIM>(0.,0.,1.);
          NumPtr<CellData> cells(2);
          for (int i=0;i<2;i++) {
            cells[i].setMaterialModel(0);
          }
          cells[0].allocate(6);
          cells[0].vertex(0)=4;
          cells[0].vertex(1)=5;
          cells[0].vertex(2)=6;
          cells[0].vertex(3)=perm[0];
          cells[0].vertex(4)=perm[1];
          cells[0].vertex(5)=perm[2];

          cells[1].allocate(4);
          cells[1].vertex(0)=perm[0];
          cells[1].vertex(1)=perm[1];
          cells[1].vertex(2)=perm[2];
          cells[1].vertex(3)=perm[3];
          Tessellation tess;
          BoundaryData *subcell_data=tess.create(vertices,cells);
          Point<3> centroid((vertices[perm[0]]+vertices[perm[1]]
                            +vertices[perm[2]]+vertices[perm[3]])*0.25);
          NumPtr<Boundary*> boundary(5);
          boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
            vertices[6],vertices[perm[3]]); // back
          boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
            vertices[5],vertices[perm[3]]); // left
          boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,
            vertices[4],vertices[5],vertices[6]); // bottom
          boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
            vertices[5],vertices[perm[1]],vertices[6]); // right front
          boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
            vertices[perm[1]],vertices[perm[2]],vertices[perm[3]]); 
            // right top
          NumPtr<BoundaryCondition*> boundary_condition(5);
          for (int i=0;i<boundary.getNumber();i++) {
            boundary_condition[i]=
              OPERATOR_NEW BoundaryCondition(boundary[i]);
          }
          for (bool doit=tess.firstCodimension1ActiveShape();doit;
          doit=tess.nextCodimension1ActiveShape()) {
            CODIMENSION1_LEVEL_CLASS *codimension1_level=
              tess.presentCodimension1Level();
            if (codimension1_level->getBoundaryCondition()!=0) {
              CODIMENSION1_SHAPE_CLASS *face_element=
                codimension1_level->getShape();
              Point<SPACEDIM> center=face_element->center(vertices);
              bool found=false;
              for (int i=0;i<boundary.getNumber();i++) {
                if (abs(boundary[i]->constraint(center))<1.e-5) {
                  codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                  found=true;
                  break;
                }
              }
              CHECK_TEST(found);
            }
          }
          tess.synchronize(subcell_data);
#ifdef DEBUG
//        tess.plotActiveShapes(gt);
#endif

//        cout << "\tpermutation = " << perm[0] << " " << perm[1] << " "
//             << perm[2] << " " << perm[3] << endl;
          {
            MappingL mapping(degree);
            checkMappingContinuity(tess,mapping,update_flags);
          }
//        {
//          MappingQ mapping(degree);
//          checkMappingContinuityOnTriangle(tess,mapping,update_flags);
//        }
/*
          {
            MappingH mapping(degree);
            checkMappingContinuity(tess,mapping,update_flags);
          }
*/
          delete subcell_data; subcell_data=0;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            delete boundary_condition[i]; boundary_condition[i]=0;
          }
          for (int i=0;i<boundary.getNumber();i++) {
            delete boundary[i]; boundary[i]=0;
          }
        }
      }
//
//
      if (degree<=1) {
        PermutationGenerator pg(6);
        while (pg.hasMore()) {
          const NumPtr<int> &perm=pg.getNext();
          NumPtr<Point<SPACEDIM> > vertices(10);
          vertices[6]=Point<SPACEDIM>(0.,0.,0.);
          vertices[7]=Point<SPACEDIM>(1.,0.,0.);
          vertices[8]=Point<SPACEDIM>(0.,0.,1.);
          vertices[9]=Point<SPACEDIM>(1.,0.,1.);
          vertices[perm[0]]=Point<SPACEDIM>(0.,.5,0.);
          vertices[perm[1]]=Point<SPACEDIM>(1.,.5,0.);
          vertices[perm[2]]=Point<SPACEDIM>(0.,1.,0.);
          vertices[perm[3]]=Point<SPACEDIM>(0.,.5,1.);
          vertices[perm[4]]=Point<SPACEDIM>(1.,.5,1.);
          vertices[perm[5]]=Point<SPACEDIM>(0.,1.,1.);
          NumPtr<CellData> cells(2);
          for (int i=0;i<2;i++) {
            cells[i].setMaterialModel(0);
          }
          cells[0].allocate(8);
          cells[0].vertex(0)=6;
          cells[0].vertex(1)=7;
          cells[0].vertex(2)=perm[0];
          cells[0].vertex(3)=perm[1];
          cells[0].vertex(4)=8;
          cells[0].vertex(5)=9;
          cells[0].vertex(6)=perm[3];
          cells[0].vertex(7)=perm[4];

          cells[1].allocate(6);
          cells[1].vertex(0)=perm[0];
          cells[1].vertex(1)=perm[1];
          cells[1].vertex(2)=perm[2];
          cells[1].vertex(3)=perm[3];
          cells[1].vertex(4)=perm[4];
          cells[1].vertex(5)=perm[5];
          Tessellation tess;
          BoundaryData *subcell_data=tess.create(vertices,cells);
          Point<3> centroid((vertices[perm[0]]+vertices[perm[1]]
                            +vertices[perm[2]]+vertices[perm[3]]
                            +vertices[perm[4]]+vertices[perm[5]])/6.);
          NumPtr<Boundary*> boundary(6);
          boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[6],
            vertices[perm[2]],vertices[8]); // back
          boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[7],
            vertices[perm[1]],vertices[9]); // front
          boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[6],
            vertices[7],vertices[8]); // left
          boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
            vertices[perm[1]],vertices[perm[2]],vertices[perm[4]]);// right
          boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
            vertices[6],vertices[7],vertices[perm[2]]); // bottom
          boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,
            vertices[8],vertices[9],vertices[perm[5]]); // top
          NumPtr<BoundaryCondition*> boundary_condition(6);
          for (int i=0;i<boundary.getNumber();i++) {
            boundary_condition[i]=
              OPERATOR_NEW BoundaryCondition(boundary[i]);
          }
          for (bool doit=tess.firstCodimension1ActiveShape();doit;
          doit=tess.nextCodimension1ActiveShape()) {
            CODIMENSION1_LEVEL_CLASS *codimension1_level=
              tess.presentCodimension1Level();
            if (codimension1_level->getBoundaryCondition()!=0) {
              CODIMENSION1_SHAPE_CLASS *face_element=
                codimension1_level->getShape();
              Point<SPACEDIM> center=face_element->center(vertices);
              bool found=false;
              for (int i=0;i<boundary.getNumber();i++) {
                if (abs(boundary[i]->constraint(center))<1.e-5) {
                  codimension1_level->
                    setBoundaryCondition(boundary_condition[i]);
                  found=true;
                  break;
                }
              }
#ifdef DEBUG
              if (!found) {
                cout << "\tface_element:" << endl;
                face_element->printOn(cout);
                cout << "\tcenter = " << center << endl;
                for (int i=0;i<boundary.getNumber();i++) {
                  cout << "\tboundary[" << i << "]: constraint = " 
                       << boundary[i]->constraint(center) << endl;
                  boundary[i]->printOn(cout);
                }
              }
#endif
              CHECK_TEST(found);
            }
          }
          tess.synchronize(subcell_data);

          cout << "\tpermutation = " << perm[0] << " " << perm[1] << " "
               << perm[2] << " " << perm[3] << " " << perm[4] << " "
               << perm[5] << endl;
          {
            MappingL mapping(degree);
            checkMappingContinuity(tess,mapping,update_flags);
          }
//        {
//          MappingQ mapping(degree);
//          checkMappingContinuityOnQuadrilateral(tess,mapping,
//            update_flags);
//        }
/* obsolete:
          {
            MappingH mapping(degree);
            checkMappingContinuity(tess,mapping,update_flags);
          }
*/
          delete subcell_data; subcell_data=0;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            delete boundary_condition[i]; boundary_condition[i]=0;
          }
          for (int i=0;i<boundary.getNumber();i++) {
            delete boundary[i]; boundary[i]=0;
          }
        }
      }
//
    }
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, Quadrature<3>*>,
  _Select1st<pair<int const, Quadrature<3>*> >, 
  less<int>,
  allocator<pair<int const, Quadrature<3>*> > 
>;
#include <NumPtr.C>
template class NumPtrBase< NumPtr< Tensor<2,SPACEDIM> >,int>;
template class NumPtr< NumPtr< Tensor<2,SPACEDIM> > >;
