//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef MEMORY_DEBUGGER_H
#define MEMORY_DEBUGGER_H "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/MemoryDebugger.H,v 1.1 2009/08/20 17:33:32 johnt Exp $"

#if MEM_DEBUG
#include "malloc.h"
#include <string>
#include "Align.H"
#include "Arch.H"
#include "Errors.H"
#include "IDLListNode.H"
#include "IDLList.H"
#include "Types.H"

//   o Check values are inserted before and after the alloc'ed data
//     to detect pointer underruns and overruns.
//   o Free'd pointers are checked against alloc'ed pointers.
//   o Free'd storage is cleared to smoke out references to free'd data.
//   o The routine mem_checkptr() is provided to check an alloc'ed
//     pointer.

//typedef unsigned long long buffer_type;
//typedef unsigned long long print_type;
typedef unsigned long buffer_type;
typedef unsigned long print_type;
#define FILE_LENGTH 80

class MemoryNode : public IDLListNode {
  private:
    size_t nbytes;  // size of allocation
    void* data;     // allocated data
    static buffer_type beforeval;
    static buffer_type afterval;
    static char mallocval;
    static size_t buffer_bytes;
    static size_t buffer_size;
    char file[FILE_LENGTH]; // must be of fixed length avoid calling new
    int line_number;
    void create(size_t,void*);
  public:
    MemoryNode() : nbytes(0),data(0) {}
    MemoryNode(size_t,void*);
    MemoryNode(size_t,void*,const char*,int);
    static size_t allocationSize(size_t n,size_t buf_sz) {
      if (buffer_size==0U) {
        buffer_bytes=sizeof(buffer_type)*buf_sz; Align(buffer_bytes);
        buffer_size=buffer_bytes/sizeof(buffer_type);
      }
      size_t nb=bytesBeforeData();
      nb+=n; Align(nb);
      nb+=sizeof(buffer_type)*buffer_size; Align(nb); 
      return nb;
    }
    virtual ~MemoryNode() {;}
    static size_t bytesBeforeData() {
      ASSERT(buffer_size>0);
      size_t nb=sizeof(MemoryNode); Align(nb);
      nb+=sizeof(buffer_type)*buffer_size; Align(nb);
      return nb;
    }

    void* ptr() const { return data; }
    size_t size() const { return nbytes; }
    BOOLEAN allocatedByAMR() const { return line_number!=-1; }
    BOOLEAN inBounds(const void *p) const { 
      return p>= data && p< static_cast<char*>(data)+nbytes;
    }
    BOOLEAN checkBuffer() const;
    void printOn(ostream&) const;
    void printOrigin(ostream&) const;
};

class MemoryDebugger {
  private:
    static BOOLEAN inited;     // TRUE if mem package is initialized
    static int count;          // # of unfreed allocs
    static size_t buffer_size; // number of beforeval's and afterval's
    static long numalloc;      // current # bytes allocated
    static long maxalloc;      // max # bytes allocated
    static IDLList<MemoryNode> *alloclist;
    static char badval;
  public:
  MemoryDebugger(size_t=1U);
  ~MemoryDebugger();
    static BOOLEAN active() { return inited; }
    static void check() ;
    static void* malloc(size_t);
    static void* malloc(size_t,const char *,int);
    static void free(void *);
    static void checkPtr(void*);
    static MemoryNode* convert(void *p) {
      ASSERT(inited);
      char *pp=reinterpret_cast<char*>(p)-MemoryNode::bytesBeforeData();
      return reinterpret_cast<MemoryNode*>(pp);
    }
    static const MemoryNode* convert(const void *p) {
      ASSERT(inited);
      const char *pp=reinterpret_cast<const char*>(p)
                    -MemoryNode::bytesBeforeData();
      return reinterpret_cast<const MemoryNode*>(pp);
    }
    static void printOn(ostream&);
};

extern "C" {
void mem_check_ptr(void*);
void mem_check();
void F77_NAME(mem_check)();
}

//operator new with position does not involve MemoryDebugger
void* operator new(size_t,const char*,int);
void* operator new[](size_t,const char*,int);
#define OPERATOR_NEW new(__FILE__,__LINE__)
#define OPERATOR_NEW_BRACKET(T,n) new(__FILE__,__LINE__) T[n]

#else

#define OPERATOR_NEW new
#define OPERATOR_NEW_BRACKET(T,n) new T[n]
#endif

#endif
