//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef NIDLLIST_H
#define NIDLLIST_H "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/NIDLList.H,v 1.1 2009/08/20 17:33:32 johnt Exp $"
#include <iostream>
#include "MemoryDebugger.H"
#include "Types.H"

template<class T> class NIDLListNode {
  private:
    T *data;
    NIDLListNode* bk;
    NIDLListNode* fd;
  public:
    inline NIDLListNode(T *d) : data(d), bk(0), fd(0) {;}
    inline virtual ~NIDLListNode() {;}
    inline T* getData() { return data; }
    inline NIDLListNode* prev() const { return bk; }
    inline NIDLListNode* next() const { return fd; }
    inline NIDLListNode* selfConnect() { return fd=bk=this; }
    inline NIDLListNode* placeBefore(NIDLListNode *n) {
      fd=n; bk=n->bk; n->bk->fd=this; return n->bk=this;
    }
    inline NIDLListNode* placeAfter(NIDLListNode *n) {
      fd=n->fd; bk=n; n->fd->bk=this; return n->fd=this;
    }
    inline NIDLListNode* replace(NIDLListNode *n) {
      n->bk->fd=this; n->fd->bk=this; bk=n->bk; fd=n->fd; return n;
    }
    inline void switchBackConnectionWith(NIDLListNode *n) {
      NIDLListNode<T> *t=n->bk; bk->fd=n; n->bk=bk; bk=t; t->fd=this;
    }
    inline NIDLListNode* disconnect() {
      bk->fd=fd; fd->bk=bk; return this;
    }
    inline void reverseConnections(NIDLListNode *&p,NIDLListNode *&n) {
      p=fd; n=bk; fd=n; bk=p;
    }
    inline void switchWithPrevious() {
      if (bk!=fd) {
        NIDLListNode *bkbk=bk->bk; fd->bk=bk; bkbk->fd=this;
        bk->bk=this; bk->fd=fd; fd=bk; bk=bkbk;
      }
    }
    inline void switchWithNext() {
      if (bk!=fd) {
        NIDLListNode *fdfd=fd->fd; bk->fd=fd; fdfd->bk=this;
        fd->fd=this; fd->bk=bk; bk=fd; fd=fdfd;
      }
    }
    inline void switchWith(NIDLListNode *n) {
      NIDLListNode *b=bk; NIDLListNode *f=fd; fd->bk=bk->fd=n; bk=n->bk;
      fd=n->fd; n->fd->bk=n->bk->fd=this; n->bk=b; n->fd=f;
    }

    virtual void printOn(ostream &os = cerr) const {
      os << "NIDLListNode: this = " << this
         << " bk = " << bk << " fd = " << fd << endl;
    }
};                                                                      
template<class T> inline ostream& operator<<(ostream &strm,
const NIDLListNode<T> &node) {
  strm << "NIDLListNode: &node = " << &node
       << " bk = " << node.prev() << " fd = " << node.next();
  return strm;
}

template <class T> class NIDLList;
template<class T> class NIDLListIterator {
  friend class NIDLList<T>;
  private:
    const NIDLList<T> *list;
    NIDLListNode<T> *node;
  public:
    inline NIDLListIterator() : list(0), node(0) {}
    inline NIDLListIterator(const NIDLList<T> &l): list(&l), 
      node(l.first()) {}
    inline NIDLListIterator(const NIDLListIterator &i): list(i.list), 
      node(i.node) {}
    inline ~NIDLListIterator() {}
    inline T& operator*() { return *node->getData(); }
    inline const T& operator*() const { return *node->getData(); }
    inline operator bool() const { return (node==0 ? FALSE : TRUE); }
    inline BOOLEAN operator!() const { return (node==0?TRUE:FALSE); }
    inline void operator++(int) { 
      if (node!=0) node=node->next(); if (node==list->first()) node=0; 
    }
    inline void operator--(int) { 
      if (node!=0) node=node->prev(); if (node==list->last()) node=0; 
    }
    inline BOOLEAN operator==(const NIDLListIterator &i) const {
      return node==i.node;
    }
    inline BOOLEAN operator!=(const NIDLListIterator &i) const {
      return node!=i.node;
    }
    inline void rewind() { if (list!=0) node=list->first(); }
    inline void fastForward() { if (list!=0) node=list->last(); }
};

//since this list does not allocate the memory for the node_name data,
//  it must return the data when a node is deleted
//the user must take care to allocate space for the data before putting
//  it on the list, and to delete the data when a node is dropped
//
//lists can be cleared easily by
//  while(notEmpty()) delete delAfter(0);

template<class T> class NIDLList {
  protected:
//  can't make this private: see NIDLValueList::prepend
    NIDLListNode<T> *h;

    void setFirst(NIDLListNode<T> *p) { h=p; }
  public:
    typedef NIDLListIterator<T> Iterator;
    inline NIDLList() : h(0) {;}
    inline virtual ~NIDLList() {;}
    inline BOOLEAN empty() const { return h==0; }
    inline BOOLEAN notEmpty() const { return h!=0; }

    inline NIDLListNode<T>* first() const { return h; }
    inline NIDLListNode<T>* next(const NIDLListNode<T> *p) const {
      return (p == 0 || h == 0 || p == h->prev()) ? 
              static_cast<NIDLListNode<T>* >(0) : p->next();
    }
    inline NIDLListNode<T>* prev(const NIDLListNode<T> *p) const {
      return (p==0 || p==h) ? static_cast<NIDLListNode<T>* >(0) 
                            : p->prev();
    }
    inline NIDLListNode<T>* last() const {
      return (h == 0) ? static_cast<NIDLListNode<T>* >(0) : h->prev();
    }

    inline void prepend(T *item) {
      NIDLListNode<T> *t = OPERATOR_NEW NIDLListNode<T>(item);
      if (!h) h=t->selfConnect();
      else h=t->placeBefore(h);
    }
    inline void append(T *item) {
      NIDLListNode<T> *t = OPERATOR_NEW NIDLListNode<T>(item);
      if (!h) h=t->selfConnect();
      else t->placeAfter(h->prev());
    }
    inline void insAfter(NIDLListNode<T> *p,T *item) {
      NIDLListNode<T> *t=OPERATOR_NEW NIDLListNode<T>(item);
      if (!h) h=t->selfConnect();
      else if (!p) h=t->placeBefore(h);
      else t->placeAfter(p);
    }
    inline void insAfter(NIDLListIterator<T> &p,T *item) {
      insAfter(p.node,item);
    }
    inline void insBefore(NIDLListNode<T> *p,T *item) {
      NIDLListNode<T> *t=OPERATOR_NEW NIDLListNode<T>(item);
      if (!h) h=t->selfConnect();
      else if (!p) t->placeAfter(h->prev());
      else { t->placeBefore(p); if (p==h) h=t; }
    }
    inline void insBefore(NIDLListIterator<T> &p,T *item) {
      insBefore(p.node,item);
    }
    T* replace(NIDLListNode<T> *&p,T *item) {
      NIDLListNode<T> *t = OPERATOR_NEW NIDLListNode<T>(item);
      if (p->next() == p) t->selfConnect();
      else t->replace(p);
      if (h == p) h = t;
      T *d=p->getData();
      delete p;
      p = t;
      return d;
    }
    inline T* replace(NIDLListIterator<T> &p,T *item) {
      return replace(p.node,item);
    }
    inline void join(NIDLList &b) {
      NIDLListNode<T> *t = b.h;
      if (!h) h = t;
      else if (t) h->switchBackConnectionWith(t);
      b.h = 0;
    }
    inline void joinAtFront(NIDLList &b) {
      NIDLListNode<T> *t = b.h;
      if (!h) h=t;
      else if (t) { h->switchBackConnectionWith(t); h=t; }
      b.h=0;
    }
    T* delAfter(NIDLListNode<T> *p) {
      CHECK_POINTER(h);
      CHECK_TEST(p != h->prev());
      CHECK_TEST(p == 0 || p->next() != 0);
      if (p==0) p = h->prev(); // ==> delAfter(0) removes first
      NIDLListNode<T> *t = p->next();
      if (t==p) h=0;
      else { if (t == h) h = t->next(); t->disconnect(); }
      T *d=t->getData();
      delete t;
      return d;
    }
    T* delBefore(NIDLListNode<T> *p) {
      CHECK_POINTER(h);
      CHECK_TEST(p != h);
      CHECK_TEST(p == 0 || p->prev() != 0);
      if (p==0) p=h; // ==> delBefore(0) removes last
      NIDLListNode<T> *t = p->prev();
      if (t==p) h=0;
      else { if (t == h) h = t->next(); t->disconnect(); }
      T *d=t->getData();
      delete t;
      return d;
    }
    T* delAndNext(NIDLListNode<T> *&p) {
      CHECK_POINTER(h);
      CHECK_POINTER(p);
      NIDLListNode<T> *q = 0;
      NIDLListNode<T> *t = p->next();
      if (t == p) h = 0;
      else { 
        if (p != h->prev()) q = t; 
        if (p == h) h = t; 
        p->disconnect(); 
      }
      T *d=p->getData();
      delete p;
      p = q;
      return d;
    }
    T* delAndNext(NIDLListIterator<T> &p) { return delAndNext(p.node); }
    T* delAndPrev(NIDLListNode<T> *&p) {
      CHECK_POINTER(h);
      CHECK_POINTER(p);
      NIDLListNode<T> *q = 0;
      NIDLListNode<T> *t = p->prev();
      if (t == p) h = 0;
      else {
        if (p != h) q = t;
        if (p == h) h = p->next();
        p->disconnect();
      }
      T *d=p->getData();
      delete p;
      p = q;
      return d;
    }
    inline T* delAndPrev(NIDLListIterator<T> &p) { 
      return delAndPrev(p.node); 
    }
    T* remove(NIDLListNode<T> *p) {
      if (p==h) {
        h=(h->next()==h ? static_cast<NIDLListNode<T>* >(0) :h->next());
      }
      p->disconnect();
      T *d=p->getData();
      delete p;
      return d;
    }
    inline T* remove(NIDLListIterator<T> &p) { 
      T *d=remove(p.node); p.node=0; return d;
    }

//  if p==h we are swapping first with last
    void swapWithPrev(NIDLListNode<T> *p) {
      CHECK_TEST(!empty());
      CHECK_POINTER(p);
      CHECK_TEST(owns(p));
      NIDLListNode<T> *pbk=p->prev();
      if (pbk==h) h=p;
      else if (p==h) h=pbk;
      p->switchWithPrevious();
    }
//  if p==h->prev() we are swapping first with last
    void swapWithNext(NIDLListNode<T> *p) {
      CHECK_TEST(!empty());
      CHECK_POINTER(p);
      CHECK_TEST(owns(p));
//    CHECK_TEST(p!=h->prev());
      NIDLListNode<T> *pfd=p->next();
      if (p==h) h=pfd;
      else if (pfd==h) h=p;
      p->switchWithNext();
    }

    void reverse();
    void swap(NIDLListNode<T>*,NIDLListNode<T>*);
    int length() const;
    BOOLEAN owns(const NIDLListNode<T>*) const;
    BOOLEAN progress(const NIDLListNode<T>*,const NIDLListNode<T>*) 
      const ;
    BOOLEAN regress(const NIDLListNode<T>*,const NIDLListNode<T>*) 
      const ;
    BOOLEAN ok() const;
    virtual void printOn(ostream& os = cerr) const;
};
template<class T> ostream& operator<<(ostream&,const NIDLList<T>&);

//class T must have defined operators >=,<=
template<class T> class NIDLListC : public NIDLList<T> {
  protected:
  public:
//  typedef NIDLListIterator<T> Iterator;
    inline NIDLListC() {;}
    inline virtual ~NIDLListC() {;}

//  inline void join(NIDLListC &b) { NIDLList<T>::join(b); }
//  inline void joinAtFront(NIDLListC &b) { 
//    NIDLList<T>::joinAtFront(b); 
//  }
    void sort(BOOLEAN=TRUE);
    virtual void printOn(ostream& os = cerr) const;
};
template<class T> ostream& operator<<(ostream&,const NIDLListC<T>&);

template<class T,class V> class NIDLValueListNode : 
public NIDLListNode<T> {
  private:
    V value;
  public:
    inline NIDLValueListNode(T *d,V val): NIDLListNode<T>(d),value(val)
      {;}
    inline V getValue() const { return value; }
    inline void setValue(V new_value) { value=new_value; }
    inline BOOLEAN operator<=(const NIDLValueListNode &p) const {
      return (value <= p.value);
    }
    inline BOOLEAN operator>=(const NIDLValueListNode &p) const {
      return (value >= p.value);
    }
    inline NIDLValueListNode* prev() const { 
      return static_cast<NIDLValueListNode* >(NIDLListNode<T>::prev()); 
    }
    inline NIDLValueListNode* next() const { 
      return static_cast<NIDLValueListNode* >(NIDLListNode<T>::next()); 
    }
};
template<class T,class V> inline ostream& operator<<(ostream &strm,
const NIDLValueListNode<T,V> &node) {
  strm << "NIVDLListNode: &node = " << &node
       << ", value = " << node.getValue();
  return strm;
}

template<class T,class V> class NIDLValueList : public NIDLList<T> {
  private:
    V value;
    inline void setValue(V v) { value=v; }
    inline void setFirst(NIDLListNode<T> *p) { 
      NIDLList<T>::setFirst(p); 
    }
  public:
    inline NIDLValueList() : value(0) {;}
    virtual ~NIDLValueList() {;}

    inline NIDLValueListNode<T,V>* first() const {
      return static_cast<NIDLValueListNode<T,V>* >(
        NIDLList<T>::first());
    }
    inline NIDLValueListNode<T,V>* next(const NIDLValueListNode<T,V> *p)
    const {
      return (p == 0 || NIDLList<T>::first() == 0 || 
              p==static_cast<NIDLValueListNode<T,V>* >(
                NIDLList<T>::first()->prev()))
              ? 0 : static_cast<NIDLValueListNode<T,V>* >(p->next());
    }
    inline NIDLValueListNode<T,V>* prev(const NIDLValueListNode<T,V> *p)
    const {
      return static_cast<NIDLValueListNode<T,V>* >(
        (p==0 || p==first()) ? 0 : p->prev() );
    }
    inline NIDLValueListNode<T,V>* last() const {
      return (first() == 0) ? 0 : static_cast<NIDLValueListNode<T,V>* >(
        NIDLList<T>::first()->prev());
    }
    inline V getValue() const { return value; }

    inline BOOLEAN empty() const { return NIDLList<T>::empty(); }
    inline BOOLEAN ok() const { return NIDLList<T>::ok(); }
    inline void prepend(T *item,V val) {
      value += val;
      NIDLValueListNode<T,V> *t =
        OPERATOR_NEW NIDLValueListNode<T,V>(item,val);
      if (!NIDLList<T>::first()) setFirst(t->selfConnect());
      else setFirst(t->placeBefore(NIDLList<T>::first()));
    }
    inline void append(T *item,V val) {
      value += val;
      NIDLValueListNode<T,V> *t =
        OPERATOR_NEW NIDLValueListNode<T,V>(item,val);
      if (!NIDLList<T>::first()) setFirst(t->selfConnect());
      else t->placeAfter(NIDLList<T>::first()->prev());
    }
    inline void insAfter(NIDLValueListNode<T,V> *p,T *item,V val) {
      value += val;
      NIDLValueListNode<T,V> *t =
        OPERATOR_NEW NIDLValueListNode<T,V>(item,val);
      if (!NIDLList<T>::first()) setFirst(t->selfConnect());
      else if (!p) setFirst(t->placeBefore(NIDLList<T>::first()));
      else t->placeAfter(p);
    }
    inline void insBefore(NIDLValueListNode<T,V> *p,T *item,V val) {
      value += val;
      NIDLValueListNode<T,V> *t =
        OPERATOR_NEW NIDLValueListNode<T,V>(item,val);
      if (!NIDLList<T>::first()) setFirst(t->selfConnect());
      else if (!p) t->placeAfter(NIDLList<T>::first()->prev());
      else { 
        t->placeBefore(p); 
        if (p==NIDLList<T>::first()) setFirst(t); 
      }
    }
    T* replace(NIDLValueListNode<T,V> *&p,T *item,V val) {
      value += val - p->getValue();
      NIDLValueListNode<T,V> *t =
        OPERATOR_NEW NIDLValueListNode<T,V>(item,val);
      if (p->next() == p) t->selfConnect();
      else t->replace(p);
      if (NIDLList<T>::first() == p) setFirst(t);
      T *d=p->getData();
      delete p;
      p = t;
      return d;
    }

    void join(NIDLValueList &b) {
      value += b.value;
      NIDLValueListNode<T,V>* t = b.first();
      if (!NIDLList<T>::first()) setFirst(t);
      else if (t) NIDLList<T>::first()->switchBackConnectionWith(t);
      b.setFirst(0);
      b.setValue(0);
    }
    void joinAtFront(NIDLValueList &b) {
      value += b.value;
      NIDLValueListNode<T,V> *t=b.first();
      if (!NIDLList<T>::first()) setFirst(t);
      else if (t) { 
        NIDLList<T>::first()->switchBackConnectionWith(t); 
        setFirst(t); 
      }
      b.setFirst(0);
      b.setValue(0);
    }
    T* delAfter(NIDLValueListNode<T,V> *p) {
      CHECK_POINTER(NIDLList<T>::first());
      CHECK_TEST(p != NIDLList<T>::first()->prev());
      CHECK_TEST(p == 0 || p->next() != 0);

      if (!p) p = last();
      NIDLValueListNode<T,V> *t = p->next();
      value -= t->getValue();
      if (t==p) setFirst(0);
      else { 
        if (t == NIDLList<T>::first()) setFirst(t->next()); 
        t->disconnect(); 
      }
      T *d=t->getData();
      delete t;
      return d;
    }
    T* delBefore(NIDLValueListNode<T,V> *p) {
      CHECK_POINTER(NIDLList<T>::first());
      CHECK_TEST(p != NIDLList<T>::first());
      CHECK_TEST(p == 0 || p->prev() != 0);

      if (!p) p=first();
      NIDLValueListNode<T,V> *t = p->prev();
      value -= t->getValue();
      if (t==p) setFirst(0);
      else { 
        if (t == NIDLList<T>::first()) setFirst(t->next()); 
        t->disconnect(); 
      }
      T *d=t->getData();
      delete t;
      return d;
    }
    T* delAndNext(NIDLValueListNode<T,V> *&p) {
      CHECK_POINTER(NIDLList<T>::first());
      CHECK_POINTER(p);
      value-=p->getValue();
      NIDLValueListNode<T,V> *q = 0;
      NIDLValueListNode<T,V> *t = p->next();
      if (t == p) setFirst(0);
      else {
        if (p != NIDLList<T>::first()->prev()) q = t;
        if (p == NIDLList<T>::first()) setFirst(t);
        p->disconnect();
      }
      T *d=p->getData();
      delete p;
      p = q;
      return d;
    }
    T* delAndPrev(NIDLValueListNode<T,V> *&p) {
      CHECK_POINTER(NIDLList<T>::first());
      CHECK_POINTER(p);
      value-=p->getValue();
      NIDLValueListNode<T,V> *q = 0;
      NIDLValueListNode<T,V> *t=p->prev();
      if (t == p) setFirst(0);
      else {
        if (p != NIDLList<T>::first()) q = t;
        else setFirst(p->next());
        p->disconnect();
      }
      T *d=p->getData();
      delete p;
      p = q;
      return d;
    }
    T* remove(NIDLValueListNode<T,V> *p) {
      value -= p->getValue();
      return NIDLList<T>::remove(p);
    }

    void selectionSort(BOOLEAN=TRUE);
    void insertionSort(BOOLEAN=TRUE);
    void bubbleSort(BOOLEAN=TRUE);
    void quickSort(NIDLValueListNode<T,V>*,NIDLValueListNode<T,V>*,
      BOOLEAN=TRUE);
    void sort(BOOLEAN largest_first=TRUE) {
      selectionSort(largest_first);
    }
    void balance(INTEGER,NIDLValueList*&) ;
    virtual void printOn(ostream& os = cerr) const;
};
template<class T,class V> ostream& operator<<(ostream&,
  const NIDLValueList<T,V>&);
#endif
