//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/TimedObject.C,v 1.1 2009/08/20 17:33:33 johnt Exp $"
#include "TimedObject.H"
#include <stdio.h>
#ifdef TIMING
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <string.h>
#include "MemoryDebugger.H"
#include "ISLList.C"

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TimedObject::TimedObject(const char *na) : number_calls(0),is_on(FALSE),
total_run_time(ZERO),start(0) {
  int length_name=strlen(na);
  name=OPERATOR_NEW_BRACKET(char,length_name+1);
  strcpy(name,na);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TimedObject::TimedObject(const char *na,TimedObjectList &tol) :
number_calls(0),is_on(FALSE),total_run_time(ZERO),start(0) {
  int length_name=strlen(na);
  name=OPERATOR_NEW_BRACKET(char,length_name+1);
  strcpy(name,na);
  tol.append(*this);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimedObject::turnOn() {
  ASSERT(!is_on);
  is_on=TRUE;
  number_calls++;
  times(&usage);
  start=usage.tms_utime;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL TimedObject::turnOff() {
  if (!is_on) return ZERO;
  times(&usage);
  REAL elapsed=static_cast<REAL>(usage.tms_utime-start)
              /static_cast<REAL>(sysconf(_SC_CLK_TCK));
  total_run_time += elapsed;
  is_on=FALSE;
  return elapsed;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimedObject::printOn(ostream &os) const {
  os << "TimedObject: name = " << name << endl;
  if (number_calls>0) {
    os << "\ttotal_run_time = " << total_run_time
       << " number_calls = " << number_calls
       << "\n\t\ttime/call = " << total_run_time/REAL(number_calls) 
       << endl;
  } else {
    os << "\ttotal_run_time = " << total_run_time
       << " number_calls = " << number_calls << endl;
  }
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
TimedObjectList::~TimedObjectList() {
  while (notEmpty()) delete delAfter(0);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimedObjectList::reset() const {
  for (TimedObject *p=first();p;p=next(p)) p->reset();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimedObjectList::printOn(ostream &os) const {
  for (TimedObject *p=first();p;p=next(p)) {
    if (p->numberCalls()>0) p->printOn(os);
  }
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
IndexedTimedObject::IndexedTimedObject(const char *na,int mi,
IndexedTimedObjectList &tol) : max_indices(mi),object_array(0),
total_length(0) {
  if (max_indices<0) max_indices=0;
  if (max_indices>0) {
    object_array=OPERATOR_NEW_BRACKET( TimedObject*,max_indices);
    total_length=OPERATOR_NEW_BRACKET(long,max_indices);
    for (int i=0;i<max_indices;i++) {
      char title[LENGTH_NAME];
      snprintf(title,LENGTH_NAME,"%s[%d]",na,i);
      object_array[i]=OPERATOR_NEW TimedObject(title);
      total_length[i]=0;
    }
  }
  tol.append(*this);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
IndexedTimedObject::IndexedTimedObject(const char *na,int mi) : 
max_indices(mi),object_array(0),total_length(0) {
  if (max_indices<0) max_indices=0;
  if (max_indices>0) {
    object_array=OPERATOR_NEW_BRACKET(TimedObject*,max_indices);
    total_length=OPERATOR_NEW_BRACKET(long,max_indices);
    for (int i=0;i<max_indices;i++) {
      char title[LENGTH_NAME];
      snprintf(title,LENGTH_NAME,"%s[%d]",na,i);
      object_array[i]=OPERATOR_NEW TimedObject(title);
      total_length[i]=0;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
IndexedTimedObject::~IndexedTimedObject() {
  if (object_array) {
    for (int i=0;i<max_indices;i++) {
      if (object_array[i]) {
        delete object_array[i]; object_array[i]=0;
      }
    }
  }
  delete [] object_array; object_array=0;
  delete [] total_length; total_length=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void IndexedTimedObject::turnOn(INTEGER length,INTEGER index) {
  if (index>=0) {
    CHECK_TEST(index<max_indices);
    CHECK_TEST(!object_array[index]->isOn());
    total_length[index]+=length;
    object_array[index]->turnOn();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void IndexedTimedObject::turnOff(INTEGER index) {
  if (index>=0) {
    CHECK_TEST(index<max_indices);
    object_array[index]->turnOff();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void IndexedTimedObject::reset() {
  for (int i=0;i<max_indices;i++) {
    object_array[i]->reset();
    total_length[i]=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void IndexedTimedObject::printOn(ostream &os) const {
  os << "IndexedTimedObject:" << endl;
  if (object_array) {
    for (INTEGER il=0;il<max_indices;il++) {
      if (object_array[il]->numberCalls()>0) {
        object_array[il]->printOn(os);
        if (object_array[il]->totalRunTime()>ZERO) {
          os << "\ttotal_length = " << total_length[il]
             << " length/time = "
             << REAL(total_length[il])
               /REAL(object_array[il]->totalRunTime()) <<endl;
        }
      }
    }
  }
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
IndexedTimedObjectList::~IndexedTimedObjectList() {
  while (notEmpty()) delete delAfter(0);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void IndexedTimedObjectList::reset() const {
  for (IndexedTimedObject *p=first();p;p=next(p)) p->reset();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void IndexedTimedObjectList::printOn(ostream &os) const {
  os << "IndexedTimedObjectList:" << endl;
  for (IndexedTimedObject *p=first();p;p=next(p)) {
    p->printOn(os);
  }
}

template class ISLList<TimedObject>;
template class ISLList<IndexedTimedObject>;
#endif
