// "$Header: /home/faculty/johnt/cvs/deal_new/mg/Multigrid.C,v 1.1 2009/08/20 17:34:11 johnt Exp $"
//---------------------------  multigrid.all_dimensions.cc  ---------------
//   $Id: Multigrid.C,v 1.1 2009/08/20 17:34:11 johnt Exp $
//   Version: $Name:  $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//---------------------------  multigrid.all_dimensions.cc  ---------------
//
//modified from deal.II/deal.II/source/multigrid/multigrid.all_dimensions.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Multigrid.H>
#include <MGTransfer.H>
#include <MemoryDebugger.H>
#include "Precondition.H"
#include <SolverArrays.H>
#include <Tracer.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Multigrid::Multigrid(const MGDofHandler &mgdh,const SolverMatrix &mc,
Vector &vc,MGTransfer &tr,const SolverPreconditioner &smoother_cloner,
Solver &cs,SolverPreconditioner &csp,
const NumPtr<set<int> > &dirichlet_boundary_dofs,int mnl,int mxl):
  mg_dof_handler(&mgdh),minlevel(mnl),
  maxlevel(min(mgdh.getTessellation().numberLevels()-1,mxl)),
  vector_cloner(&vc),initial_residual(vc,mnl,maxlevel),
  initial_residual_save(vc,mnl,maxlevel),residual(vc,mnl,maxlevel),
  solution(vc,mnl,maxlevel),solution_save(vc,mnl,maxlevel),
  t(vc,mnl,maxlevel),matrices(mc,minlevel,maxlevel),transfer(&tr),
  smoothers(smoother_cloner,minlevel,maxlevel),coarse_solver(&cs),
  coarse_solver_preconditioner(&csp)
{
//build prolongation and restriction on all levels:
  transfer->buildMatrices(dirichlet_boundary_dofs);

  sparsity_patterns.allocate(maxlevel+1);
  sparsity_patterns.initialize(0);
  int max_couplings=mg_dof_handler->maxCouplingsBetweenDofs();
  for (int level=minlevel;level<=maxlevel;level++) {
    int n=mg_dof_handler->numberDofs(level);
    SparsityPattern *sparsity=sparsity_patterns[level]=
      OPERATOR_NEW SparsityPattern(n,n,max_couplings);
    mg_dof_handler->makeSparsityPattern(*sparsity,level);
    sparsity->compress();
    matrices.clone(level);
    SparseMatrix *sparse_matrix_l=
      dynamic_cast<SparseMatrix*>(matrices[level]);
    CHECK_POINTER(sparse_matrix_l);
    sparse_matrix_l->reinit(sparsity);
    initial_residual.clone(level);
    initial_residual[level]->reinit(n);
    initial_residual_save.clone(level);
    initial_residual_save[level]->reinit(n);
    residual.clone(level);
    residual[level]->reinit(n);
    solution.clone(level);
    solution[level]->reinit(n);
    solution_save.clone(level);
    solution_save[level]->reinit(n);
    t.clone(level);
    t[level]->reinit(n);
  }
  coarse_solver->control().setMaxSteps(matrices[minlevel]->size(0));
  coarse_solver->control().setTolerance(DBL_EPSILON*128.);
  mg_dof_handler->copyMatrix(maxlevel,mc,*matrices[maxlevel]);
  if (maxlevel>0) {
    for (int level=maxlevel-1;level>=minlevel;level--) {
      mg_dof_handler->coarsenMatrix(level,*transfer,*matrices[level+1],
        *matrices[level]);
      mg_dof_handler->copyMatrix(level,mc,*matrices[level]);
    }
  }
  for (int level=minlevel+1;level<=maxlevel;level++) {
    smoothers.clone(level);
    smoothers[level]->setMatrix(*matrices[level]);
  }
  if (dynamic_cast<PreconditionUseMatrix*>(coarse_solver_preconditioner)
  !=0) {
    coarse_solver_preconditioner->setMatrix(*matrices[minlevel]);
  } else if (dynamic_cast<PreconditionRelaxation*>(
  coarse_solver_preconditioner)!=0) {
    coarse_solver_preconditioner->setMatrix(*matrices[minlevel]);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Multigrid::~Multigrid() {
  for (int i=0;i<sparsity_patterns.getNumber();i++) {
    if (sparsity_patterns[i]!=0) delete sparsity_patterns[i];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//r_f(0) = b - A x_f(0) (stored in initial_residual) ==>
//                  r_f(0) = A_f ( x_f - x_f(0) )
//d_f(1) = S_f r_f(0) (stored in fine_solution) ==> 
//                x_f(1) = x_f(0) + d_f(1)
//                x_f(1) - x_f = x_f(0) - x_f + S_f A_f ( x_f - x_f(0) )
//                             = [ I - S_f A_f ] ( x_f(0) - x_f )
//r_f(1) = r_f(0) - A_f d_f(1) (stored in residual[level]) ==>
//                  r_f(1) = A_f ( x_f - x_f(0) ) 
//                         - A_f ( x_f(1) - x_f(0) )
//                         = A_f ( x_f - x_f(1) )
//r_c(0) = R r_f(1) ==>
//                  r_c(0) = R A_f ( x_f - x_f(1) )
//d_c = V_c r_c(0) ==>
//                  d_c = V_c R A_f ( x_f - x_f(1) )
//d_f(2) = d_f(1) + P d_c ==>
//            x_f(2) = x_f(1) + P d_c
//                   = x_f(0) + d_f(1) + P d_c
//                   = x_f(0) + d_f(2)
//            x_f(2) - x_f = x_f(1) - x_f + P V_c R A_f ( x_f - x_f(1) )
//                         = [ I - P V_c R A_f ] ( x_f(1) - x_f )
//r_f(2) = r_f(0) - A_f d_f(2) ==>
//            r_f(2) = A_f ( x_f - x_f(0) ) - A_f ( x_f(2) - x_f(0) )
//                   = A_f ( x_f - x_f(2) )
//d_f(3) = d_f(2) + S_f^T r_f(2) ==>
//            x_f(3) = x_f(2) + S_f^T r_f(2)
//                   = s_f(0) + d_f(2) + S_f^T r_f(2) = x_f(0) + d_f(3)
//            x_f(3) - x_f = x_f(2) - x_f + S_f^T A_f ( x_f - x_f(2) )
//                         = [ I - S_f^T A_f ] ( x_f(2) - x_f )
//x_f = x_f(0) + d_f(3) outside of levelMGStep
void Multigrid::levelMGStep(int level,int smoother_iterations,
int coarse_grid_projections) {
  *solution[level]=0.;
  if (level==minlevel) {
    coarse_solver->solve(*matrices[minlevel],*solution[level],
      *initial_residual[level],*coarse_solver_preconditioner);
    return;
  }
  *residual[level] = *initial_residual[level];
  for (int i=0;i<smoother_iterations;i++) {
    smoothers[level]->vmult(*t[level],*residual[level]);
    *solution[level] += *t[level];
    matrices[level]->vmult(*t[level],*solution[level]);
    *residual[level]= *initial_residual[level];
    *residual[level]-= *t[level];
  }
  transfer->restriction(level,*initial_residual[level-1],*residual[level]);
  levelMGStep(level-1,smoother_iterations,coarse_grid_projections);
  if (coarse_grid_projections>1) {
    *initial_residual_save[level-1]= *initial_residual[level-1];
    *solution_save[level-1]= *solution[level-1];
    for (int i=1;i<coarse_grid_projections;i++) {
      matrices[level-1]->vmult(*t[level-1],*solution[level-1]);
      *initial_residual[level-1]= *initial_residual_save[level-1];
      *initial_residual[level-1]-= *t[level-1];
      *solution_save[level-1]= *solution[level-1];
      levelMGStep(level-1,smoother_iterations,coarse_grid_projections);
      *solution[level-1]=*solution_save[level-1];
      *solution[level-1]+= *solution[level-1];
    }
  }
  transfer->prolongationAndAdd(level,*solution[level],*solution[level-1]);
  for (int i=0;i<smoother_iterations;i++) {
    matrices[level]->vmult(*t[level],*solution[level]);
    *residual[level]= *initial_residual[level];
    *residual[level]-= *t[level];
    smoothers[level]->transposeVmult(*t[level],*residual[level]);
    *solution[level] += *t[level];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Multigrid::solve(SolverControl &control,const SolverMatrix &A,
const Vector &b,Vector &x,int smoother_iterations,
int coarse_grid_projections) {
  REAL bnorm=b.linftyNorm();
  RelativeErrorControl *rec=
    dynamic_cast<RelativeErrorControl*>(&coarse_solver->control());
  if (rec!=0) rec->setRhsNorm(bnorm);
  rec=dynamic_cast<RelativeErrorControl*>(&control);
  if (rec!=0) rec->setRhsNorm(bnorm);

  Vector *r=x.clone();
  Vector *dx=x.clone();
  A.vmult(*r,x);
  r->scale(-1.);
  *r += b;
  SOLVER_STATE conv=control.check(0,r->linftyNorm());
  int it=0;
  while (conv==SOLVER_ITERATE) {
    it++;
    transfer->copyToMG(initial_residual,*r);
    levelMGStep(maxlevel,smoother_iterations,coarse_grid_projections);
    transfer->copyFromMG(*dx,solution);
    x+= *dx;
    A.vmult(*r,x);
    r->scale(-1.);
    *r += b;
    conv=control.check(it,r->linftyNorm());
    if (conv!=SOLVER_ITERATE) break;
  }
  delete r;
  delete dx;
  CHECK_TEST(conv==SOLVER_SUCCESS);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// r_c^T ( R r_f ) == ( P r_c )^T r_f
void Multigrid::checkProlongationAndRestrictionRandom() const {
  if (maxlevel<=0) return;
  for (int level=maxlevel-1;level>minlevel;level--) {
    Vector *rf=vector_cloner->cloneEmpty();
    Vector *Rrf=vector_cloner->cloneEmpty();
    Vector *rc=vector_cloner->cloneEmpty();
    Vector *Prc=vector_cloner->cloneEmpty();
    int ps0=transfer->prolongationSize(level,0);
    int ps1=transfer->prolongationSize(level,1);
    rf->reinit(ps0);
    Prc->reinit(ps0);
    rc->reinit(ps1);
    Rrf->reinit(ps1);
    *Prc=0.;
    for (int i=0;i<rf->size();i++) (*rf)[i]=drand48();
    for (int i=0;i<rc->size();i++) (*rc)[i]=drand48();
    transfer->prolongationAndAdd(level+1,*Prc,*rc);
    *Rrf=0.;
    transfer->copySlaveDofs(level,*Rrf,*rc);
    transfer->restriction(level+1,*Rrf,*rf);
    REAL rc_dot_Rrf=(*rc)*(*Rrf);
    REAL Prc_dot_rf=(*Prc)*(*rf);
//  scale so that coarsened matrix is same as matrix on coarse grid
//  Prc_dot_rf*=static_cast<REAL>(childrenPerShape());
    cout << "\tcheckProlongationAndRestrictionRandom[" << level 
         << "] = " << rc_dot_Rrf << " " << Prc_dot_rf << " " 
         << abs(rc_dot_Rrf-Prc_dot_rf)
           /max(abs(rc_dot_Rrf),abs(Prc_dot_rf)) << endl;
    delete rf;
    delete Rrf;
    delete rc;
    delete Prc;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// y^T (A x ) == x^T ( A y )
void Multigrid::checkMatrixSymmetry() const {
  for (int level=maxlevel;level>=minlevel;level--) {
    Vector *x=vector_cloner->cloneEmpty();
    Vector *Ax=vector_cloner->cloneEmpty();
    Vector *y=vector_cloner->cloneEmpty();
    Vector *Ay=vector_cloner->cloneEmpty();
    x->reinit(matrices[level]->size(1));
    Ax->reinit(matrices[level]->size(0));
    y->reinit(matrices[level]->size(1));
    Ay->reinit(matrices[level]->size(0));
    for (int i=0;i<x->size();i++) (*x)[i]=drand48();
    for (int i=0;i<y->size();i++) (*y)[i]=drand48();
    matrices[level]->vmult(*Ax,*x);
    matrices[level]->vmult(*Ay,*y);
    REAL y_dot_Ax=(*y)*(*Ax);
    REAL x_dot_Ay=(*x)*(*Ay);
    cout << "\tcheckMatrixSymmetry[" << level 
         << "] = " << y_dot_Ax << " " << x_dot_Ay << " " 
         << abs(y_dot_Ax-x_dot_Ay)
           /max(abs(y_dot_Ax),abs(x_dot_Ay)) << endl;
    delete x;
    delete Ax;
    delete y;
    delete Ay;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// ( P y_c )^T A_f ( P x_c ) == y_c^T A_c x_c
void Multigrid::checkMatrixCoarsening(
const NumPtr<set<int> > &dirichlet_boundary_dofs) const {
  if (maxlevel<=0) return;
  for (int level=maxlevel-1;level>=minlevel;level--) {
    Vector *xc=vector_cloner->cloneEmpty();
    Vector *sxc=vector_cloner->cloneEmpty();
    Vector *Acxc=vector_cloner->cloneEmpty();
    Vector *Pxc=vector_cloner->cloneEmpty();
    Vector *AfPxc=vector_cloner->cloneEmpty();
    Vector *yc=vector_cloner->cloneEmpty();
    Vector *syc=vector_cloner->cloneEmpty();
    Vector *Pyc=vector_cloner->cloneEmpty();
    int ps0=transfer->prolongationSize(level,0);
    int ps1=transfer->prolongationSize(level,1);
    Pxc->reinit(ps0);
    AfPxc->reinit(ps0);
    Pyc->reinit(ps0);
    xc->reinit(ps1);
    sxc->reinit(ps1);
    Acxc->reinit(ps1);
    yc->reinit(ps1);
    syc->reinit(ps1);
    for (int i=0;i<xc->size();i++) (*xc)[i]=drand48();
    for (int i=0;i<yc->size();i++) (*yc)[i]=drand48();
//  prolongation maps coarse slave dofs to fine dofs
    transfer->copySlaveDofs(level,*sxc,*xc);
    transfer->copySlaveDofs(level,*syc,*yc);
    const set<int> &dbd=dirichlet_boundary_dofs[level];
//  since we set boundary dof diagonal entries of matrix to 1, have to
//  zero boundary dof entries of vectors
    for (set<int>::const_iterator dbd_iterator=dbd.begin();
    dbd_iterator!=dbd.end();dbd_iterator++) {
      (*sxc)[*dbd_iterator]=0.;
      (*syc)[*dbd_iterator]=0.;
    }
    *Pxc=0.;
    *Pyc=0.;
    transfer->prolongationAndAdd(level+1,*Pxc,*sxc);
    transfer->prolongationAndAdd(level+1,*Pyc,*syc);
    matrices[level+1]->vmult(*AfPxc,*Pxc);
    matrices[level]->vmult(*Acxc,*sxc);
    REAL Pyc_dot_AfPxc=(*Pyc)*(*AfPxc);
    REAL yc_dot_Acxc=(*syc)*(*Acxc);
    REAL den=max(abs(Pyc_dot_AfPxc),abs(yc_dot_Acxc));
    if (den<=0.) den=1.; // coarsest level with dirichlet bc's
    cout << "\tcheckMatrixCoarsening[" << level 
         << "] = " << Pyc_dot_AfPxc << " " << yc_dot_Acxc << " " 
         << abs(Pyc_dot_AfPxc-yc_dot_Acxc)/den << endl;
    delete xc;
    delete Acxc;
    delete Pxc;
    delete AfPxc;
    delete yc;
    delete Pyc;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// [ I - P A_c^{-1} R A_f ] P x == 0
void Multigrid::checkCoarseGridProjection() {
  if (maxlevel<=minlevel) return;
  int level=minlevel+1;
  Vector *xc=vector_cloner->cloneEmpty();
  xc->reinit(transfer->prolongationSize(minlevel,1));
  for (int i=0;i<xc->size();i++) (*xc)[i]=drand48();

  Vector *fine_initial_residual=initial_residual[level];
  Vector *fine_residual=residual[level];
  Vector *fine_solution=solution[level];
  Vector *fine_t=t[level];

  *fine_initial_residual=0.;
  *fine_solution=0.;
  transfer->prolongationAndAdd(level,*fine_solution,*xc);
  REAL sn=fine_solution->l2Norm();

  matrices[level]->vmult(*fine_t,*fine_solution);
  *fine_residual = *fine_initial_residual;
  *residual[level] -= *fine_t;
  transfer->restriction(level,*initial_residual[level-1],
    *fine_residual);
  levelMGStep(minlevel,1,1);
  transfer->prolongationAndAdd(level,*fine_solution,*solution[level-1]);

  cout << "\tcheckCoarseGridProjection error = "
       << fine_solution->l2Norm()/sn << endl;
  delete xc;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// y^T ( S x ) == x^T ( S^T y )
void Multigrid::checkSmootherAdjoints() {
  for (int level=maxlevel;level>minlevel;level--) {
    Vector *fine_initial_residual=initial_residual[level];
    Vector *fine_residual=residual[level];
    Vector *fine_solution=solution[level];
    Vector *yf=vector_cloner->cloneEmpty();
    yf->reinit(fine_residual->size());
    *fine_solution=0.;
    *yf=0.;
    for (int i=0;i<fine_initial_residual->size();i++) {
      (*fine_initial_residual)[i]=drand48();
      (*fine_residual)[i]=drand48();
    }

    smoothers[level]->vmult(*fine_solution,*fine_initial_residual);
    smoothers[level]->transposeVmult(*yf,*fine_residual);
    REAL y_dot_S_x=(*fine_residual)*(*fine_solution);
    REAL x_dot_Stranspose_y=(*fine_initial_residual)*(*yf);
    cout << "\tcheckSmootherAdjoints[" << level << "] = "
         << y_dot_S_x << " " << x_dot_Stranspose_y << " error = "
         << abs(y_dot_S_x-x_dot_Stranspose_y)
           /max(abs(y_dot_S_x),abs(x_dot_Stranspose_y)) << endl;
    delete yf;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// || [ I - S^T A ]^{k+1} x || / || [ I - S^T A ]^k x \\ < 1
void Multigrid::checkSmootherConvergence() {
  for (int level=maxlevel;level>minlevel;level--) {
    Vector *fine_residual=residual[level];
    Vector *fine_solution=solution[level];
    Vector *fine_t=t[level];
    for (int i=0;i<fine_solution->size();i++) {
      (*fine_solution)[i]=drand48();
    }

    for (int it=0;it<fine_solution->size();it++) {
      matrices[level]->vmult(*fine_t,*fine_solution);
      *fine_residual=0.;
      *fine_residual -= *fine_t;
      smoothers[level]->transposeVmult(*fine_t,*fine_residual);
      *fine_solution += *fine_t;
    }
    REAL norm_k=fine_solution->l2Norm();
    matrices[level]->vmult(*fine_t,*fine_solution);
    *fine_residual=0.;
    *fine_residual -= *fine_t;
    smoothers[level]->transposeVmult(*fine_t,*fine_residual);
    *fine_solution += *fine_t;
    REAL norm_kp1=fine_solution->l2Norm();

    cout << "\tcheckSmootherConvergence[" << level << "] = "
         << norm_kp1/norm_k << endl;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// y^T ( V x ) == x^T ( V y )
void Multigrid::checkVCycleSymmetry() {
  for (int level=minlevel;level<=maxlevel;level++) {
    Vector *fine_residual=initial_residual[level];
    Vector *fine_solution=solution[level];
    Vector *x=vector_cloner->cloneEmpty();
    Vector *y=vector_cloner->cloneEmpty();
    x->reinit(fine_solution->size());
    y->reinit(fine_solution->size());
    for (int i=0;i<fine_solution->size();i++) {
      (*x)[i]=drand48();
      (*y)[i]=drand48();
    }

    *fine_residual = *x;
    levelMGStep(level,1,1);
    REAL y_dot_V_x= (*y)*(*fine_solution);
    *fine_residual = *y;
    levelMGStep(level,1,1);
    REAL x_dot_V_y= (*x)*(*fine_solution);

    cout << "\tcheckVCycleSymmetry[" << level << "] = "
         << y_dot_V_x << " " << x_dot_V_y << " error = "
         << abs(y_dot_V_x-x_dot_V_y)/max(abs(y_dot_V_x),abs(x_dot_V_y))
         << endl;
    delete x;
    delete y;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// || x - V A x || / || x || < 1
void Multigrid::checkVCycleMagnitude() {
  for (int level=minlevel;level<=maxlevel;level++) {
    Vector *fine_residual=initial_residual[level];
    Vector *fine_solution=solution[level];
    Vector *x=vector_cloner->cloneEmpty();
    x->reinit(fine_solution->size());
    for (int i=0;i<x->size();i++) (*x)[i]=drand48();
    REAL xold_norm= x->l2Norm();

    matrices[level]->vmult(*fine_residual,*x);
    fine_residual->scale(-1.);
    levelMGStep(level,1,1);
    *x += *fine_solution;
    REAL xnew_norm= x->l2Norm();

    cout << "\tcheckVCycleMagnitude[" << level << "] = "
         << xnew_norm/xold_norm << endl;
    delete x;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
extern "C" {
  void F77NAME(dpotrf)(const char &uplo,const int &n,double *a,
    const int &lda,int &info);
}
// A spd; find max eigenvalue
void Multigrid::checkMatricesPositiveDefinite(
const NumPtr<set<int> > &dirichlet_boundary_dofs,
NumPtr<NumPtr<REAL> > &cholesky_factorization,
NumPtr<REAL> &spectral_radius) const {
  for (int level=minlevel;level<=maxlevel;level++) {
    const SolverMatrix *matrix=matrices[level];
    const SparseMatrix *sparse_matrix=
      dynamic_cast<const SparseMatrix*>(matrix);
    int n=sparse_matrix->size(1);
    NumPtr<REAL> &AB=cholesky_factorization[level];
    AB.allocate(n*n);
#ifdef INDEF
    AB.initialize(HUGE_VAL);
#endif
    const set<int> &dbd=dirichlet_boundary_dofs[level];
    REAL gerschgorin=0.;
    for (int j=0;j<n;j++) {
      int jn=j*n;
      REAL offdiagonal_sum=0.;
      for (int i=0;i<=j;i++) {
        REAL aij=AB[i+jn]=sparse_matrix->el(i,j);
        if (i<j) offdiagonal_sum+=abs(aij);
      }
      set<int>::const_iterator dbd_iterator=dbd.find(j);
      if (dbd_iterator==dbd.end()) {
        gerschgorin=max(gerschgorin,AB[j+jn]+2.*offdiagonal_sum);
      }
    }
    spectral_radius[level]=gerschgorin;
    char uplo='U';
    int info=INT_MAX;
    F77NAME(dpotrf)(uplo,n,AB.getData(),n,info);
    CHECK_TEST(info==0);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
extern "C" {
  void F77NAME(dsytrd)(const char &uplo,const int &n,double *a,
    const int &lda,double *d,double *e,double *tau,double *work,
    const int &lwork,int &info);
  void F77NAME(dsterf)(const int &n,double *d,double *e,int &info);
}
// ||| ( I - S A ) v |||_A <= ||| v |||_A
// If A = L * L^T, then for all w we want
//   ||| ( I - S A ) L^{-T} w |||_A^2
//     = w^T [ I - L^T S^T L ] [ I - L^T S L ] w <= w^T w
void Multigrid::checkSpectralRadiusOfSmootherIterativeImprovement(
const NumPtr<NumPtr<REAL> > &cholesky_factorization) const {
  for (int level=minlevel+1;level<=maxlevel;level++) {
    int n=matrices[level]->size(1);
    const NumPtr<REAL> &L=cholesky_factorization[level];
    Matrix error_matrix(n,n,0.);
    for (int j=0;j<n;j++) {
      int jn=j*n;
      Vector x(n,0.);
      for (int i=j;i<n;i++) x[i]=L[j+i*n];
      Vector Sx(n,0.);
      const SolverPreconditioner *smoother=smoothers[level];
      smoother->transposeVmult(Sx,x);
      for (int i=0;i<n;i++) {
        REAL sum=0.;
        for (int k=i;k<n;k++) {
          sum+=L[i+k*n]*Sx[k];
        }
        error_matrix(i,j)=-sum;
      }
      error_matrix(j,j)+=1.;
    }
    Matrix error_matrix2(n,n,0.);
    REAL gerschgorin=0.;
    for (int j=0;j<n;j++) {
      REAL off_diag=0.;
      for (int i=0;i<=j;i++) {
        REAL sum=0.;
        for (int k=0;k<n;k++) {
          sum+=error_matrix(k,i)*error_matrix(k,j);
        }
        if (i<j) off_diag+=abs(sum);
        error_matrix2(i,j)=sum;
      }
      gerschgorin=max(gerschgorin,error_matrix2(j,j)+2.*off_diag);
    }
    char uplo='U';
    double *d=OPERATOR_NEW_BRACKET(double,n);
    double *e=OPERATOR_NEW_BRACKET(double,n-1);
    double *tau=OPERATOR_NEW_BRACKET(double,n-1);
    int lwork=2*n;
    double *work=OPERATOR_NEW_BRACKET(double,lwork);
    int info=INT_MAX;
    F77NAME(dsytrd)(uplo,n,error_matrix2.addr(),n,d,e,tau,
      work,lwork,info);
    CHECK_TEST(info==0);
    F77NAME(dsterf)(n,d,e,info);
    CHECK_TEST(info==0);
    CHECK_TEST(d[0]>=-1. && d[n-1]<=1.);
    delete [] d;
    delete [] e;
    delete [] tau;
    delete [] work;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// ||| ( I - S^T A ) v |||_A <= 
//   [ 1 - || A v ||^2 / rho(A) ||| v |||_A^2 ] ||| v |||_A
// If A = L * L^T, then for all w = L^T v we want
//   || [ I - L^T S^T  L ] w || <= 
//   || w || - || L w ||^2 / [ rho(A) || w || ]
void Multigrid::checkSmootherErrorReduction(
const NumPtr<NumPtr<REAL> > &cholesky_factorization,
const NumPtr<REAL> &spectral_radius) const {
  for (int level=minlevel+1;level<=maxlevel;level++) {
    int n=matrices[level]->size(1);
    const NumPtr<REAL> &L=cholesky_factorization[level];
    Vector w(n);
    for (int it=0;it<1000;it++) {
      for (int i=0;i<n;i++) w[i]=drand48();
      Vector Lw(n,0.);
      for (int i=0;i<n;i++) {
        for (int j=0;j<=i;j++) Lw[i]+=L[j+i*n]*w[j];
      }
      Vector SLw(n,0.);
      const SolverPreconditioner *smoother=smoothers[level];
      smoother->vmult(SLw,Lw);
      Vector LTSLw(n,0.);
      for (int j=0;j<n;j++) {
        for (int i=0;i<=j;i++) {
          LTSLw[i]-=L[i+j*n]*SLw[j];
        }
        LTSLw[j]+=w[j];
      }
      double LTSLw_norm=LTSLw.l2Norm();
      if (LTSLw_norm<=0.) continue;
      double w_norm=w.l2Norm();
      double Lw_norm=Lw.l2Norm();
      if (spectral_radius[level]*w_norm*w_norm<Lw_norm*Lw_norm) {
#ifdef DEBUG
        cout << "\tlevel = " << level << endl;
        cout << "\tLTSLw_norm = " << LTSLw_norm << endl;
        cout << "\tw_norm = " << w_norm << endl;
        cout << "\tLw_norm = " << Lw_norm << endl;
        cout << "\tspectral_radius = " << spectral_radius[level] << endl;
#endif
        OBSOLETE("spectral radius too small");
      }
      if (LTSLw_norm>w_norm-pow(Lw_norm,2)/(w_norm*spectral_radius[level]))
      {
#ifdef DEBUG
        cout << "\tlevel = " << level << endl;
        cout << "\tLTSLw_norm = " << LTSLw_norm << endl;
        cout << "\tw_norm = " << w_norm << endl;
        cout << "\tLw_norm = " << Lw_norm << endl;
        cout << "\tspectral_radius = " << spectral_radius[level] << endl;
#endif
        CHECK_TEST(LTSLw_norm<=
                 w_norm-pow(Lw_norm,2)/(w_norm*spectral_radius[level]));
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
// ||| ( I - P_{cf} A_c^{-1} P_{cf}^T A_f ) v_f |||_{A_f}
//   <= C || A_f v_f || / sqrt(rho(A_f))
// If A_f = L_f * L_f^T, then for all w_f we want
//   sqrt(rho(A_f)) || [ I - L^T P_{cf} A_c^{-1} P_{cf}^T  L_f ] w_f || 
//     / || L_f w_f || <= C 
REAL Multigrid::estimateEllipticProjection(
const NumPtr<NumPtr<REAL> > &cholesky_factorization,
const NumPtr<REAL> &spectral_radius,NumPtr<REAL> &estimated_constant) 
const {
  PreconditionIdentity pi;
  REAL max_estimated_constant=0.;
  for (int level=minlevel+1;level<=maxlevel;level++) {
    int nc=matrices[level-1]->size(1);
    int nf=matrices[level]->size(1);
    SolverControl cg_solver_control(nc,DBL_EPSILON*1.e2,false,false);
    CGSolver cg(cg_solver_control);
    const NumPtr<REAL> &Lf=cholesky_factorization[level];
    Vector wf(nf);
    for (int it=0;it<1000;it++) {
      for (int i=0;i<nf;i++) wf[i]=drand48();
      Vector Lwf(nf,0.);
      for (int i=0;i<nf;i++) {
        for (int j=0;j<=i;j++) Lwf[i]+=Lf[j+i*nf]*wf[j];
      }

      Vector restrict_Lwf(nc,0.);
      transfer->restriction(level,restrict_Lwf,Lwf);
      Vector solve_restrict_Lwf(nc,0.);
      cg.solve(*matrices[level-1],solve_restrict_Lwf,restrict_Lwf,pi);
      Vector prolong_solve_restrict_Lwf(nf,0.);
      transfer->prolongationAndAdd(level,prolong_solve_restrict_Lwf,
        solve_restrict_Lwf);
      Vector LT_prolong_solve_restrict_Lwf(nf,0.);
      for (int j=0;j<nf;j++) {
        for (int i=0;i<=j;i++) {
          LT_prolong_solve_restrict_Lwf[i]-=
            Lf[i+j*nf]*prolong_solve_restrict_Lwf[j];
        }
        LT_prolong_solve_restrict_Lwf[j]+=wf[j];
      }
      estimated_constant[level]=max(estimated_constant[level],
        LT_prolong_solve_restrict_Lwf.l2Norm()/Lwf.l2Norm());
    }
    estimated_constant[level]*=sqrt(spectral_radius[level]);
    max_estimated_constant=
      max(max_estimated_constant,estimated_constant[level]);
  }
  return max_estimated_constant;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Multigrid::printOn(ostream &os) const {
  os << "Multigrid" << endl;
  os << "\tmg_dof_handler = " << mg_dof_handler
     << "\n\tminlevel = " << minlevel
     << "\n\tmaxlevel = " << maxlevel
     << "\n\tvector_cloner = " << vector_cloner
     << "\n\ttransfer = " << transfer 
     << endl;
  os << "\tresidual:" << endl;
  residual.printOn(os);
  os << "\tsolution:" << endl;
  solution.printOn(os);
  os << "\tt:" << endl;
  t.printOn(os);
  os << "\tmatrices:" << endl;
  matrices.printOn(os);
  os << "\tcoarse:" << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void PreconditionMG::printOn(ostream &os) const {
  os << "PreconditionMG:" << endl;
  os << "\tmultigrid = " << multigrid << endl;
}

#include "NumPtr.C"
INSTANTIATE_NUMPTR(SolverMatrix*);
INSTANTIATE_NUMPTR(SolverPreconditioner*);
