#include <float.h>
#include <fstream>
#include <iostream>
#include <math.h>
#include "Arch.H"
#include "Debug.H"
#include "SetTraps.H"
#include "GUIEnumInputParameter.H"
#include "GUIVirtualInput.H"
#include "Palette.H"
#include "TimedObject.H"
#include "Tracer.H"
#include "VGT.H"
#include "XColormap.H"
#include "XYGraphTool.H"
#include <unistd.h>
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

extern "C" {
  void F77NAME(initialize)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,
    double *u,double *x);
  void F77_NAME(explicit_centered)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    const double *unew);
  void F77_NAME(implicit_centered)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(crank_nicolson)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(forward_euler_fem)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77NAME(print_loc)(void *p) {
    cout << p << endl;
    cout << "\tdouble = " << *(double*) p << endl;
    cout << "\tfloat = " << *(float*) p << endl;
    cout << "\tint = " << *(int*) p << endl;
  }
}
struct machine_common {
  double roundoff,small,huge,undefind;
};
extern machine_common F77NAME(machine);
struct constant_common {
  double pi,time;
};
extern constant_common F77NAME(constant);

BOOLEAN skip_gui=FALSE;
int ncells=10;
int nsteps=INT_MAX;
double decay_number=0.45;
double initial_decay_number=0.45;
double diffusion=1.;
double tmax=0.01;
double *u_fine=0;
double *x_fine=0;

enum SCHEME{EXPLICIT_CENTERED,IMPLICIT_CENTERED,CRANK_NICOLSON,
  FORWARD_EULER_FEM};
SCHEME scheme=EXPLICIT_CENTERED;
static const char *scheme_name[4]={
  "explicit_centered","implicit_centered","crank_nicolson",
  "forward_Euler_finite_element"
};
bool time_explicit_centered=true;
bool time_implicit_centered=false;
bool time_crank_nicolson=false;
bool time_forward_euler_fem=false;

enum INITIAL_DATA{ONE_MODE,TWO_MODES,PIECEWISE_CONSTANT};
INITIAL_DATA initial_data=PIECEWISE_CONSTANT;
int iinitial_data=initial_data;

int ratio=1;
double winsize=0.5;
char* display_name=0;
Palette *pal=0;
XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE *cmap=0;
XYGraphTool *gt=0;
XYGraphTool *gte=0;
XYGraphTool *gt1=0;
XYGraphTool *gt2=0;
GUI_INPUT_PARAMETER_LIST_TYPE *param_list=0;
static const char *color_name[11]={
  "red","green","blue","cyan","yellow","magenta","brown","purple",
  "grey","pink","tan"
};

double (*heatSolution)(const double &x,const double &t)=0;
double heatSolutionOneMode(const double &x,const double &t) {
  return exp(-diffusion*t*M_PI_2*M_PI_2)*sin(M_PI_2*x);
}
double heatSolutionTwoModes(const double &x,const double &t) {
  return exp(-diffusion*t*M_PI_2*M_PI_2)*sin(M_PI_2*x)
       - 10.*exp(-diffusion*t*9.*M_PI_2*M_PI_2)*sin(3.*M_PI_2*x);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkMainInput() {
//TRACER_CALL(z,"checkMainInput");
  if (ncells>0) {
    if (time_explicit_centered) scheme=EXPLICIT_CENTERED;
    else if (time_implicit_centered) scheme=IMPLICIT_CENTERED;
    else if (time_crank_nicolson) scheme=CRANK_NICOLSON;
    else if (time_forward_euler_fem) scheme=FORWARD_EULER_FEM;
    if (scheme==EXPLICIT_CENTERED) {
//    param_list->
//      showWarningDialog("checkMainInput:\nunstable timestep");
//    skip_gui=FALSE;
      decay_number=min(0.45,max(0.,decay_number));
    } else if (scheme==FORWARD_EULER_FEM && decay_number>1./6.) {
//    param_list->
//      showWarningDialog("checkMainInput:\nunstable timestep");
//    skip_gui=FALSE;
      decay_number=min(0.16,max(0.,decay_number));
    }
  }
  initial_data=static_cast<INITIAL_DATA>(iinitial_data);
  initial_decay_number=min(initial_decay_number,decay_number);
  switch (initial_data) {
    case ONE_MODE:
      heatSolution=heatSolutionOneMode;
      break;
    case TWO_MODES:
      heatSolution=heatSolutionTwoModes;
      break;
    case PIECEWISE_CONSTANT:
    default:
      heatSolution=0;
  };
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
double errorNorm(int fc,int lc,int ifirst,int ilast,double t,
const double *u,const double *x) {
//TRACER_CALL(z,"errorNorm");
  double sum=0.;
  if (heatSolution) {
//  double elo=HUGE_VAL;
//  double ehi=-HUGE_VAL;
    for (int ic=ifirst;ic<=ilast;ic++) {
      double err=u[ic]-heatSolution(x[ic],t);
      sum+=err*err;
//    elo=min(elo,err);
//    ehi=max(ehi,err);
    }
//  Palette *palen=OPERATOR_NEW Palette();;
//  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE *cmapen=
//    OPERATOR_NEW XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE(palen);
//  XYGraphTool *gten=OPERATOR_NEW XYGraphTool("errorNorm","x","error",
//    x[ifirst],x[ilast],elo,ehi,cmapen,0,winsize);
//  gten->setbgColor("white");
//  gten->setfgColor("black");
//  gten->drawAxes();

//  int ic=ifirst;
//  double dx=x[ifirst+1]-x[ifirst];
//  gten->setfgColor("blue");
//  for (ic=ifirst;ic<=ilast;ic++) {
//    double err=u[ic]-heatSolution(x[ic],t);
//    gten->drawPlus(x[ic],err,0.5*dx);
//  }
//  gten->flush();
//  XYGraphTool::WINDOW_TYPE::QuitButton qb;
//  delete gten;
//  delete cmapen;
//  delete palen;
  } else {
//  TRACER_CALL(z,"errorNorm heatSolution==0");
    for (int ic=ifirst;ic<=ilast;ic++) {
      int icf=ic*ratio;
      if (ic==ifirst) {
        sum+=__cmath_power<double>(u[ic]-u_fine[icf],2);
      } else {
        while (x[ic]>x_fine[icf]) icf++;
        while (x[ic]<x_fine[icf-1]) icf--;
        double dif1=
          (u_fine[icf+1]-u_fine[icf])/(x_fine[icf+1]-x_fine[icf]);
        double dxp=x[ic]-x_fine[icf+1];
        double dx=x[ic]-x_fine[icf];
        double uf=HUGE_VAL;
        if (abs(dx)<abs(dxp)) {
          double dif1m=
            (u_fine[icf]-u_fine[icf-1])/(x_fine[icf]-x_fine[icf-1]);
          double dif2=(dif1-dif1m)/(x_fine[icf+1]-x_fine[icf-1]);
          uf=u_fine[icf+1]+dxp*(dif1+dx*dif2);
        } else {
          double dif1p=
            (u_fine[icf+2]-u_fine[icf+1])/(x_fine[icf+2]-x_fine[icf+1]);
          double dif2=(dif1p-dif1)/(x_fine[icf+2]-x_fine[icf]);
          uf=u_fine[icf]+dx*(dif1+dxp*dif2);
        }
        sum+=__cmath_power<double>(u[ic]-uf,2);
      }
    }
  }
  sum=sqrt(sum/static_cast<double>(ilast-ifirst));
#ifdef DEBUG
//cout << "\terrorNorm = " << sum << endl;
#endif
  return sum;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runOnce(int nc) {
//TRACER_CALL(z,"runOnce");
//array bounds for Fortran calls
  int fi=0;       // for corner-centered arrays, ie u
  int la=nc;
  int ifirst=fi;
  int ilast=la;
  int FI=0;       // for corner-centered arrays, ie U
  int LA=nc*20;
  int IFIRST=FI;
  int ILAST=LA;

//since ncells is determined dynamically, 
//must use dynamic memory allocation
  double *u0=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *u1=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *x=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *matrix=OPERATOR_NEW_BRACKET(double,2*(la-fi+1));

  double *U0=0,*U1=0,*X=0,*MATRIX=0;
  if (heatSolution==0) {
    U0=OPERATOR_NEW_BRACKET(double,LA-FI+1);
    U1=OPERATOR_NEW_BRACKET(double,LA-FI+1);
    X=OPERATOR_NEW_BRACKET(double,LA-FI+1);
    MATRIX=OPERATOR_NEW_BRACKET(double,2*(LA-FI+1));
    ratio=20;
  }

  {
//  TRACER_CALL(tr,"initialization");
#ifdef INDEF
//  initialize arrays to help IEEE exception handling catch unassigned
    for (int i=0;i<=la-fi;i++) {
      u0[i]=HUGE_VAL;
      u1[i]=HUGE_VAL;
      x[i]=HUGE_VAL;
    }
    for (int i=0;i<2*(la-fi+1);i++) {
      matrix[i]=HUGE_VAL;
    }
    if (heatSolution==0) {
      for (int i=0;i<=LA-FI;i++) {
        U0[i]=HUGE_VAL;
        U1[i]=HUGE_VAL;
        X[i]=HUGE_VAL;
      }
      for (int i=0;i<2*(LA-FI+1);i++) {
        MATRIX[i]=HUGE_VAL;
      }
    }
#endif

    if (heatSolution) {
      double dx=1./static_cast<double>(ilast-ifirst);
      for (int ic=ifirst;ic<=ilast;ic++) {
        x[ic]=static_cast<double>(ic-ifirst)*dx;
        u0[ic]=heatSolution(x[ic],0.);
      }
    } else {
      F77NAME(initialize)(fi,la,ifirst,ilast, u0,x);
      F77NAME(initialize)(FI,LA,IFIRST,ILAST, U0,X);
    }
#ifdef DEBUG
//  for (int ic=ifirst;ic<=ilast;ic++) {
//    cout << "\tu,x[" << ic << "] = " << u0[ic] << " " << x[ic] 
//         << endl;
//  }
#endif
  }

//find min,max data values
  double xlo=x[ifirst];
  double xhi=x[ifirst];
  double ulo=u0[ifirst];
  double uhi=u0[ifirst];
  for (int ie=ifirst+1;ie<=ilast;ie++) {
    double xi=x[ie];
    if (xi<xlo) xlo=xi;
    if (xi>xhi) xhi=xi;
    double ui=u0[ie];
    if (ui<ulo) ulo=ui;
    if (ui>uhi) uhi=ui;
  }

//setup interactive graphics
  pal=OPERATOR_NEW Palette();;
  cmap=OPERATOR_NEW XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE(pal);
  gt=OPERATOR_NEW XYGraphTool(scheme_name[scheme],"x","u",xlo,xhi,
    ulo,uhi,cmap,0,winsize);

//initialize graphics display
  gt->setbgColor("white");
  gt->setfgColor("black");
  gt->drawAxes();

//draw numerical solution
  int ic=ifirst;
  double dx=x[ifirst+1]-x[ifirst];
  gt->setfgColor("blue");
  gt->drawPlus(x[ic],u0[ic],0.5*dx);
  for (ic=ifirst+1;ic<=ilast;ic++) {
    gt->drawPlus(x[ic],u0[ic],0.5*dx);
  }

//draw analytical solution
  if (heatSolution) {
    gt->setfgColor("red");
    double gtdx=gt->getPixelWidth();
    double gtx=xlo;
    gt->movePen(gtx,heatSolution(gtx,0.));
    for (;gtx<=xhi;gtx+=gtdx) {
      gt->drawLine(gtx,heatSolution(gtx,0.));
    }
  } else {
    ic=IFIRST;
    gt->movePen(X[ic],U0[ic]);
    for (ic=IFIRST+1;ic<=ILAST;ic++) {
      gt->drawLine(X[ic],U0[ic]);
    }
  }

//force X to perform the requests
  gt->flush();

  double t=0.;
  double *uold=u0;
  double *unew=u1;
  double dt=(scheme==CRANK_NICOLSON ?  initial_decay_number*dx :
    initial_decay_number*dx*dx/diffusion);
  if (tmax<static_cast<double>(nsteps)*dt) {
    nsteps=static_cast<int>(tmax/dt);
  }
  dt=min(dt,tmax/static_cast<double>(nsteps));
  double elo=0.;
  double ehi=0.;
  double *error=OPERATOR_NEW_BRACKET(double,nsteps+1);
  double *time=OPERATOR_NEW_BRACKET(double,nsteps+1);
  time[0]=t;
  error[0]=0.;

  double T=0.;
  double *UOLD=U0;
  double *UNEW=U1;
  double DX=0.,DT=0.;
  int NSTEPS=0;
  if (heatSolution==0) {
    DX=X[IFIRST+1]-X[IFIRST];
    DT=0.5*DX*DX/diffusion;
    NSTEPS=min(1,static_cast<int>(dt/DT));
    DT=min(DT,dt/static_cast<double>(NSTEPS));
  }
#ifdef DEBUG
//cout << "\tdecay_number = " << decay_number << endl;
//cout << "\tdx = " << dx << endl;
//cout << "\tdiffusion = " << diffusion << endl;
//cout << "\ttmax = " << tmax << endl;
//cout << "\tnsteps = " << nsteps << endl;
//cout << "\tdt = " << dt << endl;
//cout << "\tDT = " << DT << endl;
//cout << "\tNSTEPS = " << NSTEPS << endl;
#endif
  int the_step=0;
  for (int step=0;step<nsteps && t<tmax;step++) {
//  TRACER_CALL(tr,"step");
#ifdef DEBUG
//  cout << "t = " << t << endl;
#endif
    F77NAME(constant).time=t;
    switch (scheme) {
      case EXPLICIT_CENTERED:
        F77_NAME(explicit_centered)(fi,la,ifirst,ilast, 
          diffusion,dt,dx, uold, unew);
        break;
      case IMPLICIT_CENTERED:
        F77_NAME(implicit_centered)(fi,la,ifirst,ilast, 
          diffusion,dt,dx, uold, matrix,unew);
        break;
      case CRANK_NICOLSON:
        F77_NAME(crank_nicolson)(fi,la,ifirst,ilast, 
          diffusion,dt,dx, uold, matrix,unew);
        break;
      case FORWARD_EULER_FEM:
        F77_NAME(forward_euler_fem)(fi,la,ifirst,ilast, 
          diffusion,dt,dx, uold, matrix,unew);
        break;
      default:
        cerr << "scheme not recognized" << endl;
        abort();
    }
    t+=dt;
    if (scheme==CRANK_NICOLSON) {
      dt=min(dt*1.1,decay_number*dx);
    } else {
      dt=min(dt*1.1,decay_number*dx*dx/diffusion);
    }
#ifdef DEBUG
//  cout << "dt = " << dt << endl;
#endif
    if (heatSolution==0) {
      for (;T<t;T+=DT) {
#ifdef DEBUG
//      cout << "\tT = " << T << endl;
#endif
        F77_NAME(crank_nicolson)(FI,LA,IFIRST,ILAST, 
          diffusion,DT,DX, UOLD, MATRIX,UNEW);
        double *temp=UOLD;
        UOLD=UNEW;
        UNEW=temp;
      }
      u_fine=UOLD;
      x_fine=X;
    }

//  find min,max data values
    double ulo=uold[0];
    double uhi=uold[0];
    for (int ic=ifirst;ic<=ilast;ic++) {
      double ui=unew[ic];
      if (ui<ulo) ulo=ui;
      if (ui>uhi) uhi=ui;
    }

    the_step=step+1;
    time[the_step]=t;
    error[the_step]=errorNorm(fi,la,ifirst,ilast,t,unew,x);
    elo=min(elo,error[step+1]);
    ehi=max(ehi,error[step+1]);

//  initialize graphics display
    gt->newPage();
    gt->setbgColor("white");
    gt->setfgColor("black");
    gt->drawAxes();

//  draw numerical solution
    int ic=ifirst;
    gt->setfgColor("blue");
    gt->drawPlus(x[ic],unew[ic],0.5*dx);
//  gt->movePen(x[ic],unew[ic]);
    for (ic=ifirst+1;ic<=ilast;ic++) {
      gt->drawPlus(x[ic],unew[ic],0.5*dx);
//    gt->drawLine(x[ic],unew[ic]);
    }

//  draw analytical solution
    gt->setfgColor("red");
    if (heatSolution) {
      double gtdx=gt->getPixelWidth();
      double gtx=xlo;
      gt->movePen(gtx,heatSolution(gtx,t));
      for (;gtx<=xhi;gtx+=gtdx) {
        gt->drawLine(gtx,heatSolution(gtx,t));
      }
    } else {
      ic=IFIRST;
      gt->movePen(X[ic],UNEW[ic]);
      for (ic=IFIRST+1;ic<=ILAST;ic++) {
        gt->drawLine(X[ic],UNEW[ic]);
      }
    }

//  force X to perform the requests
    gt->flush();

    double *temp=uold;
    uold=unew;
    unew=temp;
  } // end of time step loop

  gte=OPERATOR_NEW XYGraphTool("heat equation error vs time","time",
    "error",0.,tmax,elo,ehi,cmap,NULL,winsize);
  gte->newPage();
  gte->setbgColor("white");
  gte->setfgColor("black");
  gte->drawAxes();
  gte->movePen(time[0],error[0]);
  for (int i=1;i<the_step;i++) {
    gte->drawLine(time[i],error[i]);
  }
  gte->flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;

  delete [] u0;
  delete [] u1;
  delete [] x;
  delete [] matrix;
  delete [] error;
  delete [] time;
  if (heatSolution==0) {
    delete [] U0;
    delete [] U1;
    delete [] X;
    delete [] MATRIX;
    u_fine=0;
    x_fine=0;
  }
  nsteps=INT_MAX;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runSimulation(int nc,double &log_error,double &log_dx,
double &log_time,bool finest) {
//TRACER_CALL(tr,"runSimulation");
//switch (scheme) {
//  case EXPLICIT_CENTERED:
//    cout << "\tEXPLICIT_CENTERED" << endl;
//    break;
//  case IMPLICIT_CENTERED:
//    cout << "\tIMPLICIT_CENTERED" << endl;
//    break;
//  case CRANK_NICOLSON:
//    cout << "\tCRANK_NICOLSON" << endl;
//    break;
//  case FORWARD_EULER_FEM:
//    cout << "\tFORWARD_EULER_FEM" << endl;
//    break;
//  default:
//    cerr << "scheme not recognized" << endl;
//    abort();
//}

//if (time_explicit_centered) scheme=EXPLICIT_CENTERED;
//else if (time_implicit_centered) scheme=IMPLICIT_CENTERED;
//else if (time_crank_nicolson) scheme=CRANK_NICOLSON;
//else if (time_forward_euler_fem) scheme=FORWARD_EULER_FEM;

  double log10=log(10.);

//array bounds for Fortran calls
  int fi=0;       // for corner-centered arrays, ie u
  int la=nc;
  int ifirst=fi;
  int ilast=la;

//since ncells is determined dynamically, 
//must use dynamic memory allocation
  double *u0=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *u1=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *x=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *matrix=OPERATOR_NEW_BRACKET(double,2*(la-fi+1));

#ifdef INDEF
//initialize arrays to help IEEE exception handling catch unassigned
  for (int i=0;i<=la-fi;i++) {
    u0[i]=HUGE_VAL;
    u1[i]=HUGE_VAL;
    x[i]=HUGE_VAL;
  }
  for (int i=0;i<2*(la-fi+1);i++) {
    matrix[i]=HUGE_VAL;
  }
#endif

  if (heatSolution) {
    double dx=1./static_cast<double>(ilast-ifirst);
    for (int ic=ifirst;ic<=ilast;ic++) {
      x[ic]=static_cast<double>(ic-ifirst)*dx;
      u0[ic]=heatSolution(x[ic],0.);
    }
  } else {
    F77NAME(initialize)(fi,la,ifirst,ilast, u0,x);
  }

  double t=0.;
  double *uold=u0;
  double *unew=u1;
  double dx=x[ifirst+1]-x[ifirst];
  double dt=initial_decay_number*dx*dx/diffusion;
  nsteps=INT_MAX;
//if (tmax<static_cast<double>(nsteps)*dt) {
//  nsteps=static_cast<int>(tmax/dt);
//}
//dt=min(dt,tmax/static_cast<double>(nsteps));
  TimedObject integrate_timing("integrate");
  {
    Timer timer(&integrate_timing);
    for (int step=0;step<nsteps && t<tmax;step++) {
      switch (scheme) {
        case EXPLICIT_CENTERED:
          F77_NAME(explicit_centered)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, unew);
          break;
        case IMPLICIT_CENTERED:
          F77_NAME(implicit_centered)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, matrix,unew);
          break;
        case CRANK_NICOLSON:
          F77_NAME(crank_nicolson)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, matrix,unew);
          break;
        case FORWARD_EULER_FEM:
          F77_NAME(forward_euler_fem)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, matrix,unew);
          break;
        default:
          cerr << "scheme not recognized" << endl;
          abort();
      }
      t+=dt;
      if (scheme==CRANK_NICOLSON) {
        dt=min(dt*1.1,decay_number*dx);
      } else {
        dt=min(dt*1.1,decay_number*dx*dx/diffusion);
      }
      if (step+1<nsteps && t<tmax && tmax<t+dt) {
        dt=(tmax-t)*(1.+DBL_EPSILON);
        nsteps=step+2;
      }

      double *temp=uold;
      uold=unew;
      unew=temp;
    }
  }


  if (finest && heatSolution==0) {
    for (int ic=ifirst;ic<=ilast;ic++) {
      u_fine[ic]=uold[ic];
      x_fine[ic]=x[ic];
    }
  } else {
    double elapsed=integrate_timing.totalRunTime();
    log_error=log(errorNorm(fi,la,ifirst,ilast,t,uold,x))/log10;
    log_dx=-log(dx)/log10;
    log_time=(elapsed>0. ? log(elapsed)/log10 : -3.);
  }

  delete [] u0;
  delete [] u1;
  delete [] x;
  delete [] matrix;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runMultipleSimulations(unsigned int number_simulations,int nc,
unsigned int runs,double &elo,double &ehi,double &tlo,double &thi,
double &dxlo,double &dxhi,double *log_error,double *log_time,
double *log_dx) {
//TRACER_CALL(z,"runMultipleSimulations");
  unsigned int run_case=1;

  bool run_explicit_centered=runs%(2*run_case);
    runs-=run_explicit_centered*run_case; run_case*=2;
  bool run_implicit_centered=runs%(2*run_case);
    runs-=run_implicit_centered*run_case; run_case*=2;
  bool run_crank_nicolson=runs%(2*run_case);
    runs-=run_crank_nicolson*run_case; run_case*=2;
  bool run_forward_euler_fem=runs%(2*run_case);
    runs-=run_forward_euler_fem*run_case; run_case*=2;

  if (run_explicit_centered) scheme=EXPLICIT_CENTERED;
  else if (run_implicit_centered) scheme=IMPLICIT_CENTERED;
  else if (run_crank_nicolson) scheme=CRANK_NICOLSON;
  else if (run_forward_euler_fem) scheme=FORWARD_EULER_FEM;

  ratio=__cmath_power<int>(2,number_simulations);
  cout << "\n" << scheme_name[scheme] << endl;
  if (heatSolution==0) {
    int nc_fine=nc*ratio;
    int fif=0;
    int laf=nc_fine;
    x_fine=OPERATOR_NEW_BRACKET(double,nc_fine+1);
    u_fine=OPERATOR_NEW_BRACKET(double,nc_fine+1);

    cout << "\trunning fine grid: " << nc_fine << endl;;
    runSimulation(nc_fine,log_error[0],log_dx[0],log_time[0],true);
  }

  for (int nsim=0;nsim<number_simulations;nsim++) {
//  cout << "\trunning coarse grid: " << nc << endl;
    ratio=__cmath_power<int>(2,number_simulations-nsim);
    nsteps=INT_MAX;
    runSimulation(nc,log_error[nsim],log_dx[nsim],log_time[nsim],false);
#ifdef DEBUG
    cout << "\tlog time,error(ncells = " << nc << ") = "
         << log_time[nsim] << " " << log_error[nsim];
    if (nsim>0) {
      cout << " slope = "
           << (log_error[nsim]-log_error[nsim-1])
             /(log_dx[nsim]-log_dx[nsim-1]);
    }
    cout << endl;
#endif
    elo=min(elo,log_error[nsim]);
    ehi=max(ehi,log_error[nsim]);
    tlo=min(tlo,log_time[nsim]);
    thi=max(thi,log_time[nsim]);
    dxlo=min(dxlo,log_dx[nsim]);
    dxhi=max(dxhi,log_dx[nsim]);
    nc*=2;
  }

  if (heatSolution==0) {
    delete [] x_fine; x_fine=0;
    delete [] u_fine; u_fine=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void markPoint(XYGraphTool *g,double x,double y,double dx,int s) {
  switch (s) {
    case 0: 
      g->drawBoxGivenCenter(x,y,dx);
      break;  
    case 1: 
      g->drawDiamondGivenCenter(x,y,dx);
      break;  
    case 2: 
      g->drawBoxGivenCenter(x,y,dx);
      g->drawPlus(x,y,dx);
      break;  
    case 3: 
      g->drawDiamondGivenCenter(x,y,dx);
      g->drawCross(x,y,dx/sqrt(2.));
      break;  
    case 4: 
      g->drawBoxGivenCenter(x,y,dx);
      g->drawCross(x,y,dx);
      break;  
    case 5: 
      g->drawDiamondGivenCenter(x,y,dx);
      g->drawPlus(x,y,dx);
      break;  
    case 6: 
      g->drawBoxGivenCenter(x,y,dx);
      g->drawPlus(x,y,dx);
      g->drawCross(x,y,dx);
      break;  
    case 7: 
      g->drawDiamondGivenCenter(x,y,dx);
      g->drawPlus(x,y,dx);
      g->drawCross(x,y,dx/sqrt(2.));
      break;  
    case 8: 
      g->drawBoxGivenCenter(x,y,dx);
      g->drawDiamondGivenCenter(x,y,dx);
      break;  
    case 9: 
      g->drawBoxGivenCenter(x,y,dx);
      g->drawPlus(x,y,dx);
      g->drawDiamondGivenCenter(x,y,dx);
      break;  
    case 10:
      g->drawBoxGivenCenter(x,y,dx);
      g->drawCross(x,y,dx);
      g->drawDiamondGivenCenter(x,y,dx);
      break;  
    case 11:
      g->drawBoxGivenCenter(x,y,dx);
      g->drawPlus(x,y,dx);
      g->drawCross(x,y,dx);
      g->drawDiamondGivenCenter(x,y,dx);
      break;  
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void plotErrors(XYGraphTool *gt1,XYGraphTool *gt2,SCHEME s,
int number_simulations,double dxhi,double dxlo,double thi,double tlo,
const double *log_error,const double *log_dx,const double *log_time) {
  gt1->setfgColor(color_name[s]);
  gt1->movePen(log_dx[0],log_error[0]);
  for (int i=1;i<number_simulations;i++) {
    gt1->drawLine(log_dx[i],log_error[i]);
  }
  double db=(dxhi-dxlo)/static_cast<double>(number_simulations*10);
  for (int i=0;i<number_simulations;i++) {
    markPoint(gt1,log_dx[i],log_error[i],db,s);
//  gt1->drawBoxGivenCenter(log_dx[i],log_time[i],db);
  }
  gt2->setfgColor(color_name[s]);
  gt2->movePen(log_time[0],log_error[0]);
  for (int i=1;i<number_simulations;i++) {
    gt2->drawLine(log_time[i],log_error[i]);
  }
  db=(thi-tlo)/static_cast<double>(number_simulations*10);
  for (int i=0;i<number_simulations;i++) {
    markPoint(gt2,log_time[i],log_error[i],db,s);
//  gt2->drawBoxGivenCenter(log_time[i],log_error[i],db);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runMain(BOOLEAN /*called_before*/) {
//TRACER_CALL(t,"runMain");
  if (ncells>0) {
    runOnce(ncells);
    return;
  }

  unsigned int number_simulations=5;

  double *log_error_ec=OPERATOR_NEW_BRACKET(double,number_simulations);
  double *log_dx_ec=OPERATOR_NEW_BRACKET(double,number_simulations);
  double *log_time_ec=OPERATOR_NEW_BRACKET(double,number_simulations);

  double *log_error_ic=OPERATOR_NEW_BRACKET(double,number_simulations);
  double *log_dx_ic=OPERATOR_NEW_BRACKET(double,number_simulations);
  double *log_time_ic=OPERATOR_NEW_BRACKET(double,number_simulations);

  double *log_error_cn=OPERATOR_NEW_BRACKET(double,number_simulations);
  double *log_dx_cn=OPERATOR_NEW_BRACKET(double,number_simulations);
  double *log_time_cn=OPERATOR_NEW_BRACKET(double,number_simulations);

  double *log_error_fem=OPERATOR_NEW_BRACKET(double,number_simulations);
  double *log_dx_fem=OPERATOR_NEW_BRACKET(double,number_simulations);
  double *log_time_fem=OPERATOR_NEW_BRACKET(double,number_simulations);

  int nc=100;
  double elo=HUGE_VAL;
  double ehi=-HUGE_VAL;
  double tlo=HUGE_VAL;
  double thi=-HUGE_VAL;
  double dxlo=HUGE_VAL;
  double dxhi=-HUGE_VAL;
  unsigned int runs=1;

  if (time_explicit_centered) {
    decay_number=0.16;
    initial_decay_number=0.16;
    runMultipleSimulations(number_simulations,nc,runs,elo,ehi,tlo,thi,
      dxlo,dxhi,log_error_ec,log_time_ec,log_dx_ec);
  }
  runs*=2;
  if (time_implicit_centered) {
    decay_number=0.5;
    initial_decay_number=0.5;
    runMultipleSimulations(number_simulations,nc,runs,elo,ehi,tlo,thi,
      dxlo,dxhi,log_error_ic,log_time_ic,log_dx_ic);
  }
  runs*=2;
  if (time_crank_nicolson) {
    decay_number=0.03;
    initial_decay_number=0.03;
    runMultipleSimulations(number_simulations,nc*8,runs,elo,ehi,tlo,thi,
      dxlo,dxhi,log_error_cn,log_time_cn,log_dx_cn);
  }
  runs*=2;
  if (time_forward_euler_fem) {
    decay_number=0.16;
    initial_decay_number=0.16;
    runMultipleSimulations(number_simulations,nc,runs,elo,ehi,tlo,thi,
      dxlo,dxhi,log_error_fem,log_time_fem,log_dx_fem);
  }

  pal=OPERATOR_NEW Palette();;
  cmap=OPERATOR_NEW XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE(pal);
  gt1=OPERATOR_NEW XYGraphTool("heat equation error vs mesh width",
    "- log(dx)","log(error)",dxlo,dxhi,elo,ehi,cmap,0,winsize);
  gt2=OPERATOR_NEW XYGraphTool("heat equation error vs time","log(time)",
    "log(error)",tlo,thi,elo,ehi,cmap,0,winsize);

  gt1->newPage();
  gt1->setbgColor("white");
  gt1->setfgColor("black");
  gt1->drawAxes();
  gt2->newPage();
  gt2->setbgColor("white");
  gt2->setfgColor("black");
  gt2->drawAxes();
  if (time_explicit_centered) {
    cout << "\t" << color_name[EXPLICIT_CENTERED] << ": "
         << scheme_name[EXPLICIT_CENTERED] << endl;
    plotErrors(gt1,gt2,EXPLICIT_CENTERED,number_simulations,dxhi,dxlo,
      thi,tlo,log_error_ec,log_dx_ec,log_time_ec);
  }
  if (time_implicit_centered) {
    cout << "\t" << color_name[IMPLICIT_CENTERED] << ": "
         << scheme_name[IMPLICIT_CENTERED] << endl;
    plotErrors(gt1,gt2,IMPLICIT_CENTERED,number_simulations,
      dxhi,dxlo,thi,tlo,log_error_ic,log_dx_ic,log_time_ic);
  }
  if (time_crank_nicolson) {
    cout << "\t" << color_name[CRANK_NICOLSON] << ": "
         << scheme_name[CRANK_NICOLSON] << endl;
    plotErrors(gt1,gt2,CRANK_NICOLSON,number_simulations,
      dxhi,dxlo,thi,tlo,log_error_cn,log_dx_cn,log_time_cn);
  }
  if (time_forward_euler_fem) {
    cout << "\t" << color_name[FORWARD_EULER_FEM] << ": "
         << scheme_name[FORWARD_EULER_FEM] << endl;
    plotErrors(gt1,gt2,FORWARD_EULER_FEM,number_simulations,
      dxhi,dxlo,thi,tlo,log_error_fem,log_dx_fem,log_time_fem);
  }

  gt1->flush();
  gt2->flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;

  delete [] log_error_ec;
  delete [] log_dx_ec;
  delete [] log_time_ec;
  delete [] log_error_ic;
  delete [] log_dx_ic;
  delete [] log_time_ic;
  delete [] log_error_cn;
  delete [] log_dx_cn;
  delete [] log_time_cn;
  delete [] log_error_fem;
  delete [] log_dx_fem;
  delete [] log_time_fem;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void cleanup() {
  if (gt2) delete gt2; gt2=0;
  if (gt1) delete gt1; gt1=0;
  if (gte) delete gte; gte=0;
  if (gt) delete gt; gt=0;
  if (cmap) delete cmap; cmap=0;
  if (pal) delete pal; pal=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void shutdown() {
  while (param_list->notEmpty()) delete param_list->delAfter(0);
  delete param_list; param_list=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
#ifdef MEM_DEBUG
  MemoryDebugger md(128);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  {

//  set machine-dependent constants for Fortran
    F77NAME(machine).roundoff=DBL_EPSILON;
    F77NAME(machine).small=DBL_MIN;
    F77NAME(machine).huge=DBL_MAX;
    F77NAME(machine).undefind=HUGE_VAL;
    F77NAME(constant).pi=M_PI;

//  define input parameters and bounds
    param_list=OPERATOR_NEW GUI_INPUT_PARAMETER_LIST_TYPE("Param List");
    { const char *group="Problem Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        decay_number,"decay number",0.,DBL_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        initial_decay_number,"initial decay number",0.,DBL_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        diffusion,"diffusion",0.,DBL_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(tmax,
        "tmax",0.,DBL_MAX,group) );
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        iinitial_data,"initial data","one mode","two modes",
        "piecewise constant") );
    }
    { const char *group="Method Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(ncells,
        "ncells",0,INT_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(nsteps,
        "nsteps",1,INT_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<bool>(
        time_explicit_centered,"explicit centered",false,true,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<bool>(
        time_implicit_centered,"implicit centered",false,true,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<bool>(
        time_crank_nicolson,"crank nicolson",false,true,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<bool>(
        time_forward_euler_fem,"forward Euler fem",false,true,group) );
    }
    { const char *group="Graphics Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(winsize,
        "winsize",0.,1.,group) );
    }

//  read input file, if it exists
    if (argc>1) {
      ifstream in_file;
      in_file.open(argv[1],ios::in);
      if (in_file) {
        istream &infile(in_file);
        infile.clear(ios::goodbit);
        infile.seekg(0,ios::beg);
        char name[LENGTH_NAME], comment[LENGTH_COMMENT];
        BOOLEAN found_main=FALSE;
        while ( in_file >> setw(LENGTH_NAME) >> name ) {
          if ( strcmp(name,"Main") == 0 ) {
            in_file.getline( comment, LENGTH_COMMENT);
            found_main=TRUE;
            while ( in_file >> setw(LENGTH_NAME) >> name ) {
              if ( strcmp(name,"end") == 0 ) break;
              else if (strcmp(name,"skip_gui") == 0) {
                int iskip_gui; // type bool not read correctly
                in_file >> iskip_gui;
                skip_gui= iskip_gui!=0;
              }
              else param_list->formattedRead(in_file,name);
              in_file.getline( comment, LENGTH_COMMENT);
            }
          } else in_file.getline(comment,LENGTH_COMMENT);
        }
        if ( !found_main ) skip_gui=FALSE;
      }
      if (skip_gui) checkMainInput();
    } else skip_gui=FALSE;

    if (skip_gui) {
      runMain(FALSE);
      cleanup();
      shutdown();
    } else {
#ifdef USE_GTK
      GTKGUI gui(argv[0],display_name,param_list,&runMain,
        &checkMainInput,&cleanup,&shutdown,TRUE);
#else
      GUI gui(argv[0],display_name,param_list,&runMain,&checkMainInput,
        &cleanup,&shutdown,TRUE);
#endif
      gui.createFileMenu();
      gui.createViewMenu();
      gui.createHelpMenu();
      gui.createMainWindow(argc,argv);
      gui.eventLoop();
    }

  } // scope of MemoryDebugger
  return EXIT_SUCCESS;
}

template class InputParameter<int>;
template class InputParameter<double>;
template class InputParameter<bool>;
template double __cmath_power<double>(double,unsigned);
template int __cmath_power<int>(int,unsigned);
