#include <float.h>
#include <fstream>
#include <iostream>
#include <math.h>
#include "Arch.H"
#include "Debug.H"
#include "SetTraps.H"
#include "GUIEnumInputParameter.H"
#include "GUIVirtualInput.H"
#include "Palette.H"
#include "TimedObject.H"
#include "Tracer.H"
#include "VGT.H"
#include "XColormap.H"
#include "XYGraphTool.H"
#include <unistd.h>
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

extern "C" {
  void F77NAME(initialize)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,
    double *u,double *x);
  void F77_NAME(explicit_centered)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    const double *unew);
  void F77_NAME(implicit_centered)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(implicit_centered_periodic)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(crank_nicolson)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(forward_euler_fem)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77NAME(print_loc)(void *p) {
    cout << p << endl;
    cout << "\tdouble = " << *(double*) p << endl;
    cout << "\tfloat = " << *(float*) p << endl;
    cout << "\tint = " << *(int*) p << endl;
  }
}
struct machine_common {
  double roundoff,small,huge,undefind;
};
extern machine_common F77NAME(machine);
struct constant_common {
  double pi;
};
extern constant_common F77NAME(constant);

double heatSolution(const double &diffusion,const double &x,
const double &t) {
  return exp(-diffusion*t*M_PI_2*M_PI_2)*sin(M_PI_2*x);
}

void setField(char* field,INTEGER field_width,INTEGER val) {
  field[field_width]='\0';
  int i=field_width-1;
  int n=val;
  for (;i>=0;i--,n/=10) field[i]='0'+n %10;
}

BOOLEAN skip_gui=FALSE;
int ncells=10;
int nsteps=10;
double decay_number=0.45;
double diffusion=1.;

enum Scheme{explicit_centered,implicit_centered,
  implicit_centered_periodic,crank_nicolson,forward_euler_fem};
Scheme scheme=explicit_centered;
int ischeme=scheme;
static const char *scheme_name[5]={
  "explicit centered","implicit centered","implicit centered periodic",
  "crank nicolson","forward Euler finite element"
};
double tmax=0.01;
double winsize=0.5;
char* display_name=0;
Palette *pal=0;
XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE *cmap=0;
XYGraphTool *gt=0;
GUI_INPUT_PARAMETER_LIST_TYPE *param_list=0;

void checkMainInput() {
  if (ischeme==explicit_centered && decay_number>=0.5) {
    param_list->showWarningDialog("checkMainInput:\nunstable timestep");
    skip_gui=FALSE;
  } else 
  if (ischeme==forward_euler_fem && decay_number>=1./6.) {
    param_list->showWarningDialog("checkMainInput:\nunstable timestep");
    skip_gui=FALSE;
  }
}

void runMain(BOOLEAN /*called_before*/) {
//TRACER_CALL(z,"runMain");
  scheme=static_cast<Scheme>(ischeme);
  char filebase[LENGTH_NAME];
  snprintf(filebase,LENGTH_NAME,scheme_name[ischeme]);
#ifdef DEBUG
  cout << "\t" << filebase << " scheme" << endl;
#endif
//array bounds for Fortran calls
  int fi=0;       // for corner-centered arrays, ie u
  int la=ncells;
  int ifirst=fi;
  int ilast=la;
  int FI=0;       // for corner-centered arrays, ie U
  int LA=ncells*20;
  int IFIRST=FI;
  int ILAST=LA;

//since ncells is determined dynamically, 
//must use dynamic memory allocation
  double *u0=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *u1=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *x=OPERATOR_NEW_BRACKET(double,la-fi+1);
  double *matrix=OPERATOR_NEW_BRACKET(double,2*(la-fi+1));
  double *U0=OPERATOR_NEW_BRACKET(double,LA-FI+1);
  double *U1=OPERATOR_NEW_BRACKET(double,LA-FI+1);
  double *X=OPERATOR_NEW_BRACKET(double,LA-FI+1);
  double *matrix_periodic=OPERATOR_NEW_BRACKET(double,3*(la-fi+1));
  double *MATRIX=OPERATOR_NEW_BRACKET(double,2*(LA-FI+1));

  {
//  TRACER_CALL(tr,"initialization");
#ifdef INDEF
//  initialize arrays to help IEEE exception handling catch unassigned
    for (int i=0;i<=la-fi;i++) {
      u0[i]=HUGE_VAL;
      u1[i]=HUGE_VAL;
      x[i]=HUGE_VAL;
    }
    for (int i=0;i<2*(la-fi+1);i++) {
      matrix[i]=HUGE_VAL;
    }
    for (int i=0;i<3*(la-fi+1);i++) {
      matrix_periodic[i]=HUGE_VAL;
    }
    for (int i=0;i<=LA-FI;i++) {
      U0[i]=HUGE_VAL;
      U1[i]=HUGE_VAL;
      X[i]=HUGE_VAL;
    }
    for (int i=0;i<2*(LA-FI+1);i++) {
      MATRIX[i]=HUGE_VAL;
    }
#endif

    F77NAME(initialize)(fi,la,ifirst,ilast, u0,x);
    F77NAME(initialize)(FI,LA,IFIRST,ILAST, U0,X);
  }

//find min,max data values
  double xlo=x[ifirst];
  double xhi=x[ifirst];
  double ulo=u0[ifirst];
  double uhi=u0[ifirst];
  for (int ie=ifirst+1;ie<=ilast;ie++) {
    double xi=x[ie];
    if (xi<xlo) xlo=xi;
    if (xi>xhi) xhi=xi;
    double ui=u0[ie];
    if (ui<ulo) ulo=ui;
    if (ui>uhi) uhi=ui;
  }

  double dx=HUGE_VAL;
  double DX=HUGE_VAL;
  {
//  TRACER_CALL(tr0,"initialize graphics");
//  setup interactive graphics
    pal=OPERATOR_NEW Palette();;
    cmap=OPERATOR_NEW XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE(pal);
    gt=OPERATOR_NEW XYGraphTool(scheme_name[scheme],"x","u",xlo,xhi,
      ulo,uhi,cmap,0,winsize);

//  initialize graphics display
    gt->setbgColor("white");
    gt->setfgColor("black");
    gt->drawAxes();

//  draw numerical solution
    int ic=ifirst;
    dx=x[ifirst+1]-x[ifirst];
    DX=X[IFIRST+1]-X[IFIRST];
    gt->setfgColor("blue");
    gt->drawPlus(x[ic],u0[ic],0.5*dx);
    for (ic=ifirst+1;ic<=ilast;ic++) {
      gt->drawPlus(x[ic],u0[ic],0.5*dx);
    }

//  force X to perform the requests
    gt->flush();
  }

  int field_width=1;
  int power_of_ten=10;
  while (power_of_ten<nsteps+1) {
    power_of_ten*=10;
    field_width++;
  }
  char *index_field=OPERATOR_NEW_BRACKET(char,field_width+1);
  char filename[LENGTH_NAME];

//TimedObject integrate_timing("integrate");
  double t=0.;
  double T=0.;
  double *uold=u0;
  double *unew=u1;
  double *UOLD=U0;
  double *UNEW=U1;
  double dt=decay_number*dx*dx/diffusion;
  nsteps=nsteps;
  if (tmax<static_cast<double>(nsteps)*dt) {
    nsteps=static_cast<int>(tmax/dt);
  }
  dt=min(dt,tmax/static_cast<double>(nsteps));
  double DT=0.5*DX*DX/diffusion;
  int NSTEPS=min(1,static_cast<int>(dt/DT));
  DT=min(DT,dt/static_cast<double>(NSTEPS));
#ifdef DEBUG
//cout << "\tdecay_number = " << decay_number << endl;
//cout << "\tdx = " << dx << endl;
//cout << "\tdiffusion = " << diffusion << endl;
//cout << "\ttmax = " << tmax << endl;
//cout << "\tnsteps = " << nsteps << endl;
//cout << "\tdt = " << dt << endl;
//cout << "\tDT = " << DT << endl;
//cout << "\tNSTEPS = " << NSTEPS << endl;
#endif
  for (int step=0;step<nsteps && t<tmax;step++) {
//  TRACER_CALL(tr,"step");
//  cout << "t = " << t << endl;
    setField(index_field,field_width,step);
    {
//    TRACER_CALL(tr,"integrate");
//    Timer timer(&integrate_timing);

      switch (scheme) {
        case explicit_centered:
          F77_NAME(explicit_centered)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, unew);
          break;
        case implicit_centered:
          F77_NAME(implicit_centered)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, matrix,unew);
          break;
        case implicit_centered_periodic:
          F77_NAME(implicit_centered_periodic)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, matrix_periodic,unew);
          break;
        case crank_nicolson:
          F77_NAME(crank_nicolson)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, matrix,unew);
          break;
        case forward_euler_fem:
          F77_NAME(forward_euler_fem)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, matrix,unew);
          break;
        default:
          cerr << "scheme not recognized" << endl;
          abort();
      }
      t+=dt;
    }
    for (;T<t;T+=DT) {
#ifdef DEBUG
//    cout << "\tT = " << T << endl;
#endif
      F77_NAME(crank_nicolson)(FI,LA,IFIRST,ILAST, 
        diffusion,DT,DX, UOLD, MATRIX,UNEW);
      double *temp=UOLD;
      UOLD=UNEW;
      UNEW=temp;
    }

//  find min,max data values
    double ulo=uold[0];
    double uhi=uold[0];
    for (int ic=ifirst;ic<=ilast;ic++) {
      double ui=unew[ic];
      if (ui<ulo) ulo=ui;
      if (ui>uhi) uhi=ui;
    }

//  initialize graphics display
    gt->newPage();
    gt->setbgColor("white");
    gt->setfgColor("black");
    gt->drawAxes();

//  draw numerical solution
    int ic=ifirst;
    gt->setfgColor("blue");
    gt->drawPlus(x[ic],unew[ic],0.5*dx);
//  gt->movePen(x[ic],unew[ic]);
    for (ic=ifirst+1;ic<=ilast;ic++) {
      gt->drawPlus(x[ic],unew[ic],0.5*dx);
//    gt->drawLine(x[ic],unew[ic]);
    }

//
//  draw analytical solution
    ic=IFIRST;
    gt->setfgColor("red");
    gt->movePen(X[ic],UNEW[ic]);
    for (ic=IFIRST+1;ic<=ILAST;ic++) {
      gt->drawLine(X[ic],UNEW[ic]);
    }
//

//  force X to perform the requests
    gt->flush();
    snprintf(filename,LENGTH_NAME,"%s_%s",filebase,index_field);
//  gt->writeXPM(filename);

//  if (step%2==0) {
//    XYGraphTool::WINDOW_TYPE::QuitButton qb;
//  }

    double *temp=uold;
    uold=unew;
    unew=temp;
  } // end of time step loop
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
  delete [] index_field;
//integrate_timing.printOn(cout);

  delete [] u0;
  delete [] u1;
  delete [] x;
  delete [] matrix_periodic;
  delete [] matrix;
  delete [] U0;
  delete [] U1;
  delete [] X;
  delete [] MATRIX;
}

void cleanup() {
  if (gt) delete gt; gt=0;
  if (cmap) delete cmap; cmap=0;
  if (pal) delete pal; pal=0;
}

void shutdown() {
  while (param_list->notEmpty()) delete param_list->delAfter(0);
  delete param_list; param_list=0;
}

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  {
//  set machine-dependent constants for Fortran
    F77NAME(machine).roundoff=DBL_EPSILON;
    F77NAME(machine).small=DBL_MIN;
    F77NAME(machine).huge=DBL_MAX;
    F77NAME(machine).undefind=HUGE_VAL;
    F77NAME(constant).pi=M_PI;

//  define input parameters and bounds
    param_list=OPERATOR_NEW GUI_INPUT_PARAMETER_LIST_TYPE("Param List");
    { const char *group="Problem Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        diffusion,"diffusion",0.,DBL_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(tmax,
        "tmax",0.,DBL_MAX,group) );
    }
    { const char *group="Method Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(ncells,
        "ncells",1,INT_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(nsteps,
        "nsteps",1,INT_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        decay_number,"decay",0.,DBL_MAX,group) );
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        ischeme,"scheme",scheme_name[0],scheme_name[1],scheme_name[2],
        scheme_name[3],scheme_name[4]) );
    }
    { const char *group="Graphics Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(winsize,
        "winsize",0.,1.,group) );
    }

//  read input file, if it exists
    if (argc>1) {
      ifstream in_file;
      in_file.open(argv[1],ios::in);
      if (in_file) {
        istream &infile(in_file);
        infile.clear(ios::goodbit);
        infile.seekg(0,ios::beg);
        char name[LENGTH_NAME], comment[LENGTH_COMMENT];
        BOOLEAN found_main=FALSE;
        while ( in_file >> setw(LENGTH_NAME) >> name ) {
          if ( strcmp(name,"Main") == 0 ) {
            in_file.getline( comment, LENGTH_COMMENT);
            found_main=TRUE;
            while ( in_file >> setw(LENGTH_NAME) >> name ) {
              if ( strcmp(name,"end") == 0 ) break;
              else if (strcmp(name,"skip_gui") == 0) {
                int iskip_gui; // type bool not read correctly
                in_file >> iskip_gui;
                skip_gui= iskip_gui!=0;
              }
              else param_list->formattedRead(in_file,name);
              in_file.getline( comment, LENGTH_COMMENT);
            }
          } else in_file.getline(comment,LENGTH_COMMENT);
        }
        if ( !found_main ) skip_gui=FALSE;
      }
      if (skip_gui) checkMainInput();
    } else skip_gui=FALSE;

    if (skip_gui) {
      runMain(FALSE);
      cleanup();
      shutdown();
    } else {
#ifdef USE_GTK
      GTKGUI gui(argv[0],display_name,param_list,&runMain,
        &checkMainInput,&cleanup,&shutdown,TRUE);
#else
      GUI gui(argv[0],display_name,param_list,&runMain,&checkMainInput,
        &cleanup,&shutdown,TRUE);
#endif
      gui.createFileMenu();
      gui.createViewMenu();
      gui.createHelpMenu();
      gui.createMainWindow(argc,argv);
      gui.eventLoop();
    }

  } // scope of MemoryDebugger
  return EXIT_SUCCESS;
}

template class InputParameter<int>;
template class InputParameter<double>;
template class InputParameter<bool>;
