#include <assert.h>
#include <float.h>
#include <stdlib.h>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <string.h>
#include <math.h>
#include "Arch.H"
#include "SetTraps.H"
#include <unistd.h>

#define LENGTH_NAME 80
#define LENGTH_COMMENT 80

extern "C" {
  void F77NAME(initialize)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,
    double *u,double *x);
  void F77_NAME(explicit_centered)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    const double *unew);
  void F77_NAME(implicit_centered)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(forward_euler_fem)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(implicit_centered_periodic)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(crank_nicolson)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77NAME(print_loc)(void *p) {
    cout << p << endl;
    cout << "\tdouble = " << *(double*) p << endl;
    cout << "\tfloat = " << *(float*) p << endl;
    cout << "\tint = " << *(int*) p << endl;
  }
}
struct machine_common {
  double roundoff,small,huge,undefind;
};
extern machine_common F77NAME(machine);

//double heatSolution(const double &diffusion,const double &x,
//const double &t) {
//  return exp(-diffusion*t*M_PI_2*M_PI_2)*sin(M_PI_2*x);
//}

int main(int argc,char* argv[]) {
  assert(argc>1);
  {

//  set machine-dependent constants for Fortran
    F77NAME(machine).roundoff=DBL_EPSILON;
    F77NAME(machine).small=DBL_MIN;
    F77NAME(machine).huge=DBL_MAX;
    F77NAME(machine).undefind=HUGE_VAL;

    int ncells=10;
    int nsteps=10;
    double decay=0.5;
    double diffusion=1.;
    double tmax=1.e10;

//  read input parameters and bounds
    ifstream in_file;
    in_file.open(argv[1],ios::in);
    assert(in_file);
    istream &infile(in_file);
    infile.clear(ios::goodbit);
    infile.seekg(0,ios::beg);
    char name[LENGTH_NAME], comment[LENGTH_COMMENT];
    while ( in_file >> setw(LENGTH_NAME) >> name ) {
      if ( strcmp(name,"ncells") == 0 ) {
        in_file >> ncells;
      } else if ( strcmp(name,"nsteps") == 0 ) {
        in_file >> nsteps;
      } else if ( strcmp(name,"decay") == 0 ) {
        in_file >> decay;
      } else if ( strcmp(name,"diffusion") == 0 ) {
        in_file >> diffusion;
      } else if ( strcmp(name,"tmax") == 0 ) {
        in_file >> tmax;
      }
      in_file.getline( comment, LENGTH_COMMENT);
    }

//  array bounds for Fortran calls
    int fi=0;       // for corner-centered arrays, ie u
    int la=ncells;
    int ifirst=fi;
    int ilast=la;

//  since ncells is determined dynamically, 
//  must use dynamic memory allocation
    double *u0=new double[la-fi+1];
    double *u1=new double[la-fi+1];
    double *x=new double[la-fi+1];
    double *matrix=new double[2*(la-fi+1)];
    double *matrix_periodic=new double[3*(la-fi+1)];

#ifdef INDEF
//  initialize arrays to help IEEE exception handling catch unassigned
    for (int i=0;i<=la-fi;i++) {
      u0[i]=HUGE_VAL;
      u1[i]=HUGE_VAL;
      x[i]=HUGE_VAL;
    }
    for (int i=0;i<2*(la-fi+1);i++) {
      matrix[i]=HUGE_VAL;
    }
    for (int i=0;i<3*(la-fi+1);i++) {
      matrix_periodic[i]=HUGE_VAL;
    }
#endif

    F77NAME(initialize)(fi,la,ifirst,ilast, u0,x);
    double dx=x[1]-x[0];

    double t=0.;
    double *uold=u0;
    double *unew=u1;
    double dt=decay*dx*dx/diffusion;
    nsteps=min(nsteps,static_cast<int>(tmax/dt));
    dt=min(dt,tmax/static_cast<double>(nsteps));
    for (int step=0;step<nsteps && t<tmax;step++) {
      F77_NAME(explicit_centered)(fi,la,ifirst,ilast, 
        diffusion,dt,dx, uold, unew);
//    F77_NAME(implicit_centered)(fi,la,ifirst,ilast, 
//      diffusion,dt,dx, uold, matrix,unew);
//    F77_NAME(forward_euler_fem)(fi,la,ifirst,ilast, 
//      diffusion,dt,dx, uold, matrix,unew);
//    F77_NAME(implicit_centered_periodic)(fi,la,ifirst,ilast, 
//      diffusion,dt,dx, uold, matrix_periodic,unew);
//    F77_NAME(crank_nicolson)(fi,la,ifirst,ilast, 
//      diffusion,dt,dx, uold, matrix,unew);
      t+=dt;
      double *temp=uold;
      uold=unew;
      unew=temp;
    }
    for (int ic=fi;ic<=la;ic++) {
      cout << x[ic] << " " << unew[ic] << endl;
    }
    delete [] u0;
    delete [] u1;
    delete [] x;
    delete [] matrix_periodic;
    delete [] matrix;
  }
  return EXIT_SUCCESS;
}
