#include <float.h>
#include <fstream>
#include <iostream>
#include <math.h>
#include <string.h>
#include "Arch.H"
#include "InputParameter.H"
#include "Palette.H"
#include "SetTraps.H"
#include "TimedObject.H"
#include "Tracer.H"
#include "VGT.H"
#include "XColormap.H"
#include "XYGraphTool.H"
#include <unistd.h>
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif

//#define LENGTH_NAME 80

extern "C" {
  void F77NAME(initialize)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,
    double *u,double *x);
  void F77_NAME(explicit_centered)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    const double *unew);
  void F77_NAME(implicit_centered)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(forward_euler_fem)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(implicit_centered_periodic)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77_NAME(crank_nicolson)(const int &fi,const int &la,
    const int &ifirst,const int &ilast,const double &diffusion,
    const double &dt,const double &dx,
    const double *uold,
    double *matrix,double *unew);
  void F77NAME(print_loc)(void *p) {
    cout << p << endl;
    cout << "\tdouble = " << *(double*) p << endl;
    cout << "\tfloat = " << *(float*) p << endl;
    cout << "\tint = " << *(int*) p << endl;
  }
}
struct machine_common {
  double roundoff,small,huge,undefind;
};
extern machine_common F77NAME(machine);

double heatSolution(const double &diffusion,const double &x,
const double &t) {
  return exp(-diffusion*t*M_PI_2*M_PI_2)*sin(M_PI_2*x);
}

void setField(char* field,INTEGER field_width,INTEGER val) {
  field[field_width]='\0';
  int i=field_width-1;
  int n=val;
  for (;i>=0;i--,n/=10) field[i]='0'+n %10;
}

int main(int argc,char* argv[]) {
  ASSERT(argc>1);
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  {

//  set machine-dependent constants for Fortran
    F77NAME(machine).roundoff=DBL_EPSILON;
    F77NAME(machine).small=DBL_MIN;
    F77NAME(machine).huge=DBL_MAX;
    F77NAME(machine).undefind=HUGE_VAL;

    int ncells=10;
    int nsteps=10;
    double decay=0.5;
    double diffusion=1.;
    double tmax=1.e10;

//  define input parameters and bounds
    InputParameterList param_list("Main List");
    { const char *group="Numerical Method Parameters";
      param_list.append(OPERATOR_NEW InputParameter<int>(ncells,
        "ncells",1,INT_MAX,group) );
      param_list.append(OPERATOR_NEW InputParameter<int>(nsteps,
        "nsteps",1,INT_MAX,group) );
      param_list.append(OPERATOR_NEW InputParameter<double>(decay,
        "decay",0.,DBL_MAX,group) );
      param_list.append(OPERATOR_NEW InputParameter<double>(diffusion,
        "diffusion",0.,DBL_MAX,group) );
      param_list.append(OPERATOR_NEW InputParameter<double>(tmax,"tmax",
        0.,DBL_MAX,group) );
    }

//  read input parameters 
    ifstream in_file;
    in_file.open(argv[1],ios::in);
    ASSERT(in_file);
    istream &infile(in_file);
    infile.clear(ios::goodbit);
    infile.seekg(0,ios::beg);
    char name[LENGTH_NAME], comment[LENGTH_COMMENT];
    while ( in_file >> setw(LENGTH_NAME) >> name ) {
      if ( strcmp(name,"Main") == 0 ) {
        in_file.getline( comment, LENGTH_COMMENT);
        while ( in_file >> setw(LENGTH_NAME) >> name ) {
          if ( strcmp(name,"end") == 0 ) break;
          else param_list.formattedRead(in_file,name);
          in_file.getline( comment, LENGTH_COMMENT);
        }
      } else in_file.getline(comment,LENGTH_COMMENT);
    }

//  array bounds for Fortran calls
    int fi=0;       // for corner-centered arrays, ie u
    int la=ncells;
    int ifirst=fi;
    int ilast=la;
    int FI=0;       // for corner-centered arrays, ie U
    int LA=ncells*20;
    int IFIRST=FI;
    int ILAST=LA;

//  since ncells is determined dynamically, 
//  must use dynamic memory allocation
    double *u0=OPERATOR_NEW_BRACKET(double,la-fi+1);
    double *u1=OPERATOR_NEW_BRACKET(double,la-fi+1);
    double *x=OPERATOR_NEW_BRACKET(double,la-fi+1);
    double *matrix=OPERATOR_NEW_BRACKET(double,2*(la-fi+1));
    double *U0=OPERATOR_NEW_BRACKET(double,LA-FI+1);
    double *U1=OPERATOR_NEW_BRACKET(double,LA-FI+1);
    double *X=OPERATOR_NEW_BRACKET(double,LA-FI+1);
    double *matrix_periodic=OPERATOR_NEW_BRACKET(double,3*(la-fi+1));
    double *MATRIX=OPERATOR_NEW_BRACKET(double,2*(LA-FI+1));

    {
//    TRACER_CALL(tr,"initialization");
#ifdef INDEF
//    initialize arrays to help IEEE exception handling catch unassigned
      for (int i=0;i<=la-fi;i++) {
        u0[i]=HUGE_VAL;
        u1[i]=HUGE_VAL;
        x[i]=HUGE_VAL;
      }
      for (int i=0;i<2*(la-fi+1);i++) {
        matrix[i]=HUGE_VAL;
      }
      for (int i=0;i<3*(la-fi+1);i++) {
        matrix_periodic[i]=HUGE_VAL;
      }
      for (int i=0;i<=LA-FI;i++) {
        U0[i]=HUGE_VAL;
        U1[i]=HUGE_VAL;
        X[i]=HUGE_VAL;
      }
      for (int i=0;i<2*(LA-FI+1);i++) {
        MATRIX[i]=HUGE_VAL;
      }
#endif

      F77NAME(initialize)(fi,la,ifirst,ilast, u0,x);
      F77NAME(initialize)(FI,LA,IFIRST,ILAST, U0,X);
    }

//  find min,max data values
    double xlo=x[ifirst];
    double xhi=x[ifirst];
    double ulo=u0[ifirst];
    double uhi=u0[ifirst];
    for (int ie=ifirst+1;ie<=ilast;ie++) {
      double xi=x[ie];
      if (xi<xlo) xlo=xi;
      if (xi>xhi) xhi=xi;
      double ui=u0[ie];
      if (ui<ulo) ulo=ui;
      if (ui>uhi) uhi=ui;
    }

    {
//    TRACER_CALL(tr0,"initialize graphics");
//    setup interactive graphics
      Palette pal;
      XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
      double winsize=0.5;
      XYGraphTool gt("solution","x","u",xlo,xhi,ulo,uhi,
                     &cmap,NULL,winsize);

//    initialize graphics display
      gt.setbgColor("white");
      gt.setfgColor("black");
      gt.drawAxes();

//    draw numerical solution
      int ic=ifirst;
      double dx=x[ifirst+1]-x[ifirst];
      double DX=X[IFIRST+1]-X[IFIRST];
      gt.setfgColor("blue");
      gt.drawPlus(x[ic],u0[ic],dx);
      for (ic=ifirst+1;ic<=ilast;ic++) {
        gt.drawPlus(x[ic],u0[ic],dx);
      }

//    force X to perform the requests
      gt.flush();

      int field_width=1;
      int power_of_ten=10;
      while (power_of_ten<nsteps+1) {
        power_of_ten*=10;
        field_width++;
      }
      char *index_field=OPERATOR_NEW_BRACKET(char,field_width+1);
      char filename[LENGTH_NAME];

//    TimedObject integrate_timing("integrate");
      double t=0.;
      double T=0.;
      double *uold=u0;
      double *unew=u1;
      double *UOLD=U0;
      double *UNEW=U1;
      double dt=decay*dx*dx/diffusion;
      nsteps=min(nsteps,static_cast<int>(tmax/dt));
      dt=min(dt,tmax/static_cast<double>(nsteps));
      double DT=0.5*DX*DX/diffusion;
      int NSTEPS=min(1,static_cast<int>(dt/DT));
      DT=min(DT,dt/static_cast<double>(NSTEPS));
#ifdef DEBUG
//    cout << "\tdecay = " << decay << endl;
//    cout << "\tdx = " << dx << endl;
//    cout << "\tdiffusion = " << diffusion << endl;
//    cout << "\ttmax = " << tmax << endl;
//    cout << "\tnsteps = " << nsteps << endl;
//    cout << "\tdt = " << dt << endl;
//    cout << "\tDT = " << DT << endl;
//    cout << "\tNSTEPS = " << NSTEPS << endl;
#endif
      for (int step=0;step<nsteps && t<tmax;step++) {
//      TRACER_CALL(tr,"step");
//      cout << "t = " << t << endl;
        setField(index_field,field_width,step);
        {
//        TRACER_CALL(tr,"integrate");
//        Timer timer(&integrate_timing);

          F77_NAME(explicit_centered)(fi,la,ifirst,ilast, 
            diffusion,dt,dx, uold, unew);
//        F77_NAME(implicit_centered)(fi,la,ifirst,ilast, 
//          diffusion,dt,dx, uold, matrix,unew);
//        F77_NAME(forward_euler_fem)(fi,la,ifirst,ilast, 
//          diffusion,dt,dx, uold, matrix,unew);
//        F77_NAME(implicit_centered_periodic)(fi,la,ifirst,ilast, 
//          diffusion,dt,dx, uold, matrix_periodic,unew);
//        F77_NAME(crank_nicolson)(fi,la,ifirst,ilast, 
//          diffusion,dt,dx, uold, matrix,unew);
          t+=dt;
        }
        for (;T<t;T+=DT) {
#ifdef DEBUG
//        cout << "\tT = " << T << endl;
#endif
          F77_NAME(crank_nicolson)(FI,LA,IFIRST,ILAST, 
            diffusion,DT,DX, UOLD, MATRIX,UNEW);
          double *temp=UOLD;
          UOLD=UNEW;
          UNEW=temp;
        }

//      find min,max data values
        double ulo=uold[0];
        double uhi=uold[0];
        for (int ic=ifirst;ic<=ilast;ic++) {
          double ui=unew[ic];
          if (ui<ulo) ulo=ui;
          if (ui>uhi) uhi=ui;
        }

//      initialize graphics display
        gt.newPage();
        gt.setbgColor("white");
        gt.setfgColor("black");
        gt.drawAxes();

//      draw numerical solution
        int ic=ifirst;
        gt.setfgColor("blue");
        gt.drawPlus(x[ic],unew[ic],dx);
//      gt.movePen(x[ic],unew[ic]);
        for (ic=ifirst+1;ic<=ilast;ic++) {
          gt.drawPlus(x[ic],unew[ic],dx);
//        gt.drawLine(x[ic],unew[ic]);
        }

//
//      draw analytical solution
        ic=IFIRST;
        gt.setfgColor("red");
        gt.movePen(X[ic],UNEW[ic]);
        for (ic=IFIRST+1;ic<=ILAST;ic++) {
          gt.drawLine(X[ic],UNEW[ic]);
        }
//

//      force X to perform the requests
        gt.flush();
        snprintf(filename,LENGTH_NAME,"step_%s",index_field);
        gt.writeXPM(filename);

        double *temp=uold;
        uold=unew;
        unew=temp;
      } // end of time step loop
      XYGraphTool::WINDOW_TYPE::QuitButton qb;
      delete [] index_field;
    } // pal,cmap,gt go out of scope here 
//  integrate_timing.printOn(cout);

//  since x, u, flux were created with operator new, must delete
    delete [] u0;
    delete [] u1;
    delete [] x;
    delete [] matrix_periodic;
    delete [] matrix;
    delete [] U0;
    delete [] U1;
    delete [] X;
    delete [] MATRIX;
  } // scope of MemoryDebugger
  return EXIT_SUCCESS;
}

template class InputParameter<int>;
template class InputParameter<double>;
template class InputParameter<bool>;
