// "$Header: /home/faculty/johnt/cvs/deal_new/base/Shape.C,v 1.1 2009/08/20 17:14:10 johnt Exp $"
//----------------------------  geometry_info.cc  ------------------------
// $Id: Shape.C,v 1.1 2009/08/20 17:14:10 johnt Exp $
// Version: $Name:  $
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
// This file is subject to QPL and may not be  distributed
// without copyright and license information. Please refer
// to the file deal.II/doc/license.html for the  text  and
// further information on this license.
//
//----------------------------  geometry_info.cc  ------------------------
//
//modified from deal.II/deal.II/source/grid/geometry_info.cc
// by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Shape.H"
#include <algorithm>
#include <set>
#include <vector>
#include "DDecl.H"
#include "Tensor.H"
#include "Tracer.H"

const Point<1> Interval::canonical_vertices[2]=
  { Point<1>(0.), Point<1>(1.) };
#if (SPACEDIM>1)
const Point<2> Triangle::canonical_vertices[3] =
  { Point<2>(0., 0.), Point<2>(1., 0.), Point<2>(0.,1.) };
const Point<2> Quadrilateral::canonical_vertices[4] =
  { Point<2>(0.,0.),Point<2>(1.,0.),Point<2>(0.,1.),Point<2>(1.,1.) };
#endif
#if (SPACEDIM==3)
const Point<3> Tetrahedron::canonical_vertices[4] =
  { Point<3>(0.,0.,0.), Point<3>(1.,0.,0.), Point<3>(0.,1.,0.),
    Point<3>(0.,0.,1.) };
const Point<3> Prism::canonical_vertices[6] =
  { Point<3>(0.,0.,0.), Point<3>(1.,0.,0.), Point<3>(0.,1.,0.),
    Point<3>(0.,0.,1.), Point<3>(1.,0.,1.), Point<3>(0.,1.,1.) };
const Point<3> Hexahedron::canonical_vertices[8] =
  { Point<3>(0., 0., 0.), Point<3>(1., 0., 0.),
    Point<3>(0., 1., 0.), Point<3>(1., 1., 0.),
    Point<3>(0., 0., 1.), Point<3>(1., 0., 1.),
    Point<3>(0., 1., 1.), Point<3>(1., 1., 1.) };
#endif

const Point<SPACEDIM> Interval::canonical_center(D_DECL(.5,0.,0.));
#if (SPACEDIM>1)
const Point<SPACEDIM> 
  Triangle::canonical_center(D_DECL(1./3.,1./3.,0.));
const Point<SPACEDIM> 
  Quadrilateral::canonical_center(D_DECL(.5,.5,0.));
#endif
#if (SPACEDIM==3)
const Point<SPACEDIM> Tetrahedron::canonical_center(.25,.25,.25);
const Point<SPACEDIM> Prism::canonical_center(1./3.,1./3.,0.5);
const Point<SPACEDIM> Hexahedron::canonical_center(.5,.5,.5);
#endif

#if (SPACEDIM>1)
//unit outer normal * face length / canonical interval length
const Tensor<1,SPACEDIM> Triangle::face_normals[3]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL(-1., 0.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 0.,-1.,0.))),
//Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL(sqrt(.5),sqrt(.5),0.)))
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL(1.,1.,0.)))};
//unit outer normal * face length / canonical interval length
const Tensor<1,SPACEDIM> Quadrilateral::face_normals[4]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL(-1., 0.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 1., 0.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 0.,-1.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 0., 1.,0.)))};
#endif
#if (SPACEDIM==3)
//unit outer normal * face area / canonical triangle area
const Tensor<1,SPACEDIM> Tetrahedron::face_normals[4]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(-1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0.,-1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0.,-1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 1., 1.))};
//unit outer normal * face area / canonical triangle/quadrilateral area
const Tensor<1,SPACEDIM> Prism::face_normals[5]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0.,-1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(-1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0.,-1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 1., 0.))};
//unit outer normal * face area / canonical quadrilateral area
const Tensor<1,SPACEDIM> Hexahedron::face_normals[6]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(-1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0.,-1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0.,-1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.))};
#endif
#if (SPACEDIM>1)
//unit tangent * line length
const Tensor<1,SPACEDIM> Triangle::line_tangents[3]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 0.,-1.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 1., 0.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL(-1., 1.,0.)))};
const Tensor<1,SPACEDIM> Quadrilateral::line_tangents[4]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 0., 1.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 0., 1.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 1., 0.,0.))),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(D_DECL( 1., 0.,0.)))};
#endif
#if (SPACEDIM==3)
const Tensor<1,SPACEDIM> Tetrahedron::line_tangents[6]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(-1., 1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0.,-1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0.,-1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 1.,-1.))};
const Tensor<1,SPACEDIM> Prism::line_tangents[9]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(-1., 1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0.,-1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>(-1., 1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0.,-1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.))};
const Tensor<1,SPACEDIM> Hexahedron::line_tangents[12]={
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 1., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 1., 0., 0.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.)),
  Tensor<1,SPACEDIM>(Point<SPACEDIM>( 0., 0., 1.))};
#endif
#if (SPACEDIM==3)
//faces that share a given line
const int Tetrahedron::faces_from_line[6][2] = { 
  {1,2}, {2,3}, {0,2}, {0,1}, {1,3}, {0,3} 
};
const int Prism::faces_from_line[9][2] = { 
  {0,3}, {0,4}, {0,2}, {1,3}, {1,4}, {1,2}, {2,3}, {3,4}, {2,4} 
};
const int Hexahedron::faces_from_line[12][2] = { 
  {0,4}, {1,4}, {2,4}, {3,4}, {0,5}, {1,5}, 
  {2,5}, {3,5}, {0,2}, {1,2}, {0,3}, {1,3} 
};
#endif

//for each face, { child, child_vertex } interior to that face 
#if (SPACEDIM>1)
const int Triangle::interior_child_vertices[3][2] =
  { {3, 0}, {3, 1}, {3, 2} };
const int Quadrilateral::interior_child_vertices[4][2] =
  { {0, 2}, {3, 2}, {0, 1}, {3, 1} };
#endif
#if (SPACEDIM==3)
const int Prism::interior_child_vertices[5][2] =
  { {INT_MAX, INT_MAX}, {INT_MAX, INT_MAX}, {0, 5}, {0, 4}, {1, 5} };
const int Hexahedron::interior_child_vertices[6][2] =
  { {0, 6}, {1, 7}, {0, 5}, {2, 7}, {0, 3}, {4, 7} };
#endif

//for each line, vertices on that line
//  need to be in order of lattice point traversal
#if (SPACEDIM>1)
const int Triangle::canonical_vertex_indices[3][2] =
  { {0, 2}, {0, 1}, {1, 2} };
const int Quadrilateral::canonical_vertex_indices[4][2] =
  { {0, 2}, {1, 3}, {0, 1}, {2, 3} };
#endif
#if (SPACEDIM==3)
const int Tetrahedron::canonical_vertex_indices[6][2] =
  { {0, 1}, {1, 2}, {0, 2}, {0, 3}, {1, 3}, {2, 3} };
const int Prism::canonical_vertex_indices[9][2] =
  { {0,1}, {1,2}, {0,2}, {3,4}, {4,5}, {3,5}, {0,3}, {1,4}, {2,5} };
const int  Hexahedron::canonical_vertex_indices[12][2] =
  { {0, 2}, {1, 3}, {0, 1}, {2, 3}, 
    {4, 6}, {5, 7}, {4, 5}, {6, 7}, 
    {0, 4}, {1, 5}, {2, 6}, {3, 7}  };
#endif

#if (SPACEDIM>1)
const int Quadrilateral::opposite_vertex[4] = { 3, 2, 1, 0 };
#endif
#if (SPACEDIM==3)
const int Hexahedron::opposite_vertex[8] = { 7, 6, 5, 4, 3, 2, 1, 0 };
#endif

#if (SPACEDIM>1)
const int Triangle::opposite_face[3] = { 1, 2, 0 }; 
const int Quadrilateral::opposite_face[4] = { 1, 0, 3, 2 }; 
#endif
#if (SPACEDIM==3)
const int Tetrahedron::opposite_face[4] = { 1, 2, 3, 0 }; 
const int Prism::opposite_face[5] = { 1, 0, 3, 4, 2 }; 
const int Hexahedron::opposite_face[6] = { 1, 0, 3, 2, 5, 4 }; 
#endif

//see generateParallelpiped:
#if (SPACEDIM>=2)
//faces are ordered left,right,bottom,top
//face_map leads to order left,right,bottom,top
//const int Quadrilateral::face_map[4]={0,1,2,3};
#endif
#if (SPACEDIM==3)
//faces are ordered back,front,left,right,bottom,top
//face_map leads to order back,front,left,right,bottom,top
//const int Hexahedron::face_map[6]={0,1,2,3,4,5};
#endif

//for each face, children having a face on that face
#if (SPACEDIM>1)
const int Triangle::children[3][2] = { {2 , 0}, {0 , 1}, {1 , 2} }; 
const int Quadrilateral::children[4][2] = {{0,2}, {1,3}, {0,1}, {2,3}}; 
#endif
#if (SPACEDIM==3)
const int Tetrahedron::children[4][4] =
  { {0, 2, 3, 7}, {0, 3, 1, 4}, {0, 1, 2, 6}, {1, 3, 2, 5} };
const int Prism::children[5][4] =
  { {0,1,2,3}, {4,5,6,7}, {0,4,2,6}, {0,1,4,5}, {1,2,5,6} };
const int Hexahedron::children[6][4] =
  { {0, 2, 4, 6}, {1, 3, 5, 7}, {0, 1, 4, 5},
    {2, 3, 6, 7}, {0, 1, 2, 3}, {4, 5, 6, 7} };
#endif

#if (SPACEDIM==3)
//for each face and face_vertex, returns element vertex
//vertices in order of lattice traversal
const int Tetrahedron::face_vertex_to_cell_vertex[4][3] =
  { {0,2,3}, {0,1,3}, {0,1,2}, {1,2,3} }; 
const int Prism::face_vertex_to_cell_vertex[5][4] =
  { {0,1,2,6}, {3,4,5,6}, {0,2,3,5}, {0,1,3,4}, {1,2,4,5} };
const int Hexahedron::face_vertex_to_cell_vertex[6][4] = 
  { {0,2,4,6}, {1,3,5,7}, {0,1,4,5}, 
    {2,3,6,7}, {0,1,2,3}, {4,5,6,7} };

//for each face and face_line, returns element line
const int  Tetrahedron::face_line_to_cell_line[4][3] =
  { {3,2,5}, {3,0,4}, {2,0,1}, {4,1,5} };
const int Prism::face_line_to_cell_line[5][4] =
  { {2,0,1,9}, {5,3,4,9}, {6,8,2,5}, {6,7,0,3}, {7,8,1,4} };
const int  Hexahedron::face_line_to_cell_line[6][4] =
  { {8,10,0,4},{9,11,1,5},{8,9,2,6},{10,11,3,7},{0,1,2,3},{4,5,6,7} };
#endif

#if (SPACEDIM==3)
//for each interior child line, {child0,child1} sharing that line
const int Prism::interior_child_line[3][2] =
  { {3,2},{3,1},{3,0} };
const int Hexahedron::interior_child_line[6][2] =
  { {0,6},{1,7},{0,5},{3,7},{0,3},{4,7} };
#endif

//for each interior face, {child0,child1} on that face
const int Interval::child_neighbors[2]={0,1}; 
#if (SPACEDIM>1)
const int Triangle::child_neighbors[3][2]= { {3,2}, {3,0}, {3,1} };
const int Quadrilateral::child_neighbors[4][2]=
  { {0,2}, {1,3}, {0,1}, {2,3} };
#endif
#if (SPACEDIM==3)
const int Tetrahedron::child_neighbors[8][2]={
  {4,5},{4,3},{4,7},{5,1},
  {5,6},{6,2},{6,7},{7,0} };
const int Prism::child_neighbors[10][2]={
  {0,4},{1,5},{2,6},{3,7},{1,3},
  {2,3},{0,3},{5,7},{6,7},{4,7} };
const int Hexahedron::child_neighbors[12][2]={
  {0,1},{4,5},{2,3},{6,7},
  {0,2},{1,3},{4,6},{5,7},
  {0,4},{2,6},{1,5},{3,7} };
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Shape::center(const NumPtr<Point<SPACEDIM> > &vertices)
const {
  Point<SPACEDIM> p(D_DECL(0.,0.,0.));
  int nv=vertex_indices.getNumber();
  for (int v=0;v<nv;v++) {
    p+=vertices[vertex_indices[v]];
  }
  p/=static_cast<REAL>(nv);
  return p;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Shape::vertexEqualToVertexIndex(int vi) const {
  for (int v=0;v<numberVertices();v++) {
    if (vertexIndex(v)==vi) return v;
  }
  return -1;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Shape::sharedVertexIndex(const Shape *other_element) const {
  vector<int> shared_vertex_index(1);
  shared_vertex_index[0]=INT_MAX;
  set<int> vertex_indices,other_vertex_indices;
  for (int v=0;v<numberVertices();v++) {
    vertex_indices.insert(vertexIndex(v));
  }
  for (int v=0;v<other_element->numberVertices();v++) {
    other_vertex_indices.insert(other_element->vertexIndex(v));
  }
  vector<int>::iterator svi_iterator;
  svi_iterator=set_intersection(vertex_indices.begin(),
    vertex_indices.end(),other_vertex_indices.begin(),
    other_vertex_indices.end(),shared_vertex_index.begin());
  CHECK_TEST(shared_vertex_index[0]!=INT_MAX);
  return shared_vertex_index[0];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Shape::sharedFaceElementIndex(const Shape *other_element) const {
  vector<int> shared_face_index(1);
  shared_face_index[0]=INT_MAX;
  set<int> face_indices,other_face_indices;
  for (int f=0;f<numberCodimension1Shapes();f++) {
    face_indices.insert(codimension1ShapeIndex(f));
  }
  for (int f=0;f<other_element->numberCodimension1Shapes();f++)
  {
    other_face_indices.insert(other_element->codimension1ShapeIndex(f));
  }
  vector<int>::iterator sfi_iterator;
  sfi_iterator=set_intersection(face_indices.begin(),face_indices.end(),
    other_face_indices.begin(),other_face_indices.end(),
    shared_face_index.begin());
  CHECK_TEST(shared_face_index[0]!=INT_MAX);
  return shared_face_index[0];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Shape::printOn(ostream &os) const { 
  os << "Shape: element_data_index = " << element_data_index 
     << "\n\tdof_index = " << dof_index << endl;
  os << "\tvertex_indices =";
  for (int i=0;i<vertex_indices.getNumber();i++){
    os << " " << vertex_indices[i];
  }
  os << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Shape1::faceWithIndexEqualTo(int face_index) const {
  for (int f=0;f<numberCodimension1Shapes();f++) {
    if (codimension1ShapeIndex(f)==face_index) return f;
  }
  return -1;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Shape1::printOn(ostream &os) const { 
  os << "Shape1: " << endl;
  Shape::printOn(os);
}
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Shape2::printOn(ostream &os) const { 
  os << "Shape2: codimension1_element_indices =";
  for (int i=0;i<codimension1_element_indices.getNumber();i++){
    os << " " << codimension1_element_indices[i];
  }
  os << endl;
  Shape1::printOn(os);
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Shape3::sharedLineElementIndex(const Shape3 *other_element) const {
  vector<int> shared_line_index(1);
  shared_line_index[0]=INT_MAX;
  set<int> line_indices,other_line_indices;
  for (int f=0;f<numberCodimension2Shapes();f++) {
    line_indices.insert(codimension2ShapeIndex(f));
  }
  for (int f=0;f<other_element->numberCodimension2Shapes();f++)
  {
    other_line_indices.insert(other_element->codimension2ShapeIndex(f));
  }
  vector<int>::iterator sfi_iterator;
  sfi_iterator=set_intersection(line_indices.begin(),line_indices.end(),
    other_line_indices.begin(),other_line_indices.end(),
    shared_line_index.begin());
  CHECK_TEST(shared_line_index[0]!=INT_MAX);
  return shared_line_index[0];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Shape3::lineWithIndexEqualTo(int line_index) const {
  for (int l=0;l<numberCodimension2Shapes();l++) {
    if (codimension2_element_indices[l]==line_index) return l;
  }
  return -1;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Shape3::printOn(ostream &os) const { 
  os << "Shape3: codimension2_element_indices = ";
  for (int i=0;i<codimension2_element_indices.getNumber();i++){
    os << " " << codimension2_element_indices[i];
  }
  os << endl;
  Shape2::printOn(os);
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Interval::Interval(int edi,const int v[2]) : Shape1(edi,2) {
  CHECK_TEST(v[0]!=v[1]);
#if (SPACEDIM==1)
  for (int i=0;i<2;i++) setVertexIndex(i,v[i]);
#else
  if (v[0]<v[1]) {
    setVertexIndex(0,v[0]);
    setVertexIndex(1,v[1]);
  } else {
    setVertexIndex(0,v[1]);
    setVertexIndex(1,v[0]);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Interval::Interval(const int v[2],int di) : Shape1(-1,2) {
  CHECK_TEST(v[0]!=v[1]);
  for (int i=0;i<2;i++) setVertexIndex(i,v[i]);
  dof_index=di;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Interval::contains(const Point<SPACEDIM> &x,
const NumPtr<Point<SPACEDIM> > &vertices) const {
#if (SPACEDIM==1)
  const Point<1> &v0(vertices[vertexIndex(0)]);
  const Point<1> &v1(vertices[vertexIndex(1)]);
  return v0[0]<=x[0] && x[0]<=v1[0];
#else
  return Shape::contains(x,vertices);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// 2 rotations of an interval times 2 children = 4 cases
Point<SPACEDIM> Interval::parentCoordinateForCanonical(
const Point<SPACEDIM> &cp,const Shape *child_shape,int c,
const NumPtr<int>&) const {
#if (SPACEDIM>1)
  OBSOLETE("should not be calling this");
#else
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi>=0);
  
  if (c==0) {
    if (cvi==0) return cp*0.5;
    else {
      OBSOLETE("not checked");
      return Point<1>((1.-cp[0])*0.5);
    }
  } else {
    if (cvi==0) return Point<1>(1.-cp[0]*0.5); 
    else {
      OBSOLETE("not checked");
      return Point<1>((1.+cp[0])*0.5); 
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Interval::childCoordinateOfCanonical(
const Point<SPACEDIM> &pp,const Shape *child_shape,int c,
const NumPtr<int>&) const {
#if (SPACEDIM>1)
  OBSOLETE("should not be calling this");
#else
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi>=0);
  
  if (c==0) {
    if (cvi==0) return pp*2.;
    else {
      OBSOLETE("not checked");
      return Point<1>(1.-pp[0]*2.);
    }
  } else {
    if (cvi==0) return Point<1>((1.-pp[0])*2.); 
    else {
      OBSOLETE("not checked");
      return Point<1>(pp[0]*2.-1.); 
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
const Shape* Interval::reorientFor(const Shape *parent,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  return this;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::staticEquallySpacedLatticePoints(int n,
NumPtr<Point<1> > &points) {
  points.allocate(staticNumberEquallySpacedLatticePoints(n));
  Point<1> p(0.);
  REAL s=(n>0 ? 1./static_cast<REAL>(n) : 0.);
  for (int i=0;i<=n;i++,p[0]+=s) points[i]=p;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::reorderDofs(int n,NumPtr<int> &renumber)
const {
  if (vertexIndex(0)>vertexIndex(1)) {
    for (int fe_s=0;fe_s<=n;fe_s++) renumber[fe_s]=n-fe_s;
  } else {
    for (int fe_s=0;fe_s<=n;fe_s++) renumber[fe_s]=fe_s;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::reorderEquallySpacedLatticePoints(int n,
NumPtr<int> &renumber) const {
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  int m=0;
  renumber[0]=m++;
  renumber[n]=m++;
  for (int i=1;i<n;i++) renumber[i]=m++;
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]==INT_MAX) {
      cout << "\trenumber[" << i << "] = *** " << -1 << " ***" << endl;
      success=false;
    } else if (renumber[i]>=renumber.getNumber()) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::reorderChildEquallySpacedLatticePoints(int n,
const Shape *child_shape,int c,const NumPtr<int>&,NumPtr<int> &renumber) 
const {
  const Interval *child_interval=
    dynamic_cast<const Interval*>(child_shape);
  CHECK_POINTER(child_interval);

  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi>=0);
  int reorder_case=(c==0 ? cvi : 2-cvi);;
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  switch (reorder_case) {
    case 0: {
      for (int i=0;i<=n;i++) renumber[i]=i;
      break;
    }
    case 1: {
      int ni=n;
      for (int i=0;i<=n;i++) renumber[ni--]=i;
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_H::constraints
void Interval::reorderChildDofPoints(int n,const Shape *child_shape,
int c,const NumPtr<int> &licv,NumPtr<int> &renumber) const {
  OBSOLETE("this does not seem right");
  reorderChildEquallySpacedLatticePoints(n,child_shape,c,licv,renumber);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::staticInteriorLatticePointIndices(int n,NumPtr<int> &v) {
  CHECK_TEST(n>0);
  v.allocate(n-1);
  for (int i=0;i+1<n;i++) v[i]=i+1;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::computeInteriorLatticePoints(int degree,int,
NumPtr<Point<SPACEDIM> > &mapping_support_points) const {
  if (degree<=1) return;
  CHECK_SAME(mapping_support_points.getNumber(),degree+1);
  Point<SPACEDIM> p=mapping_support_points[0];
  Point<SPACEDIM> dp=(mapping_support_points[degree]-p)
                    /static_cast<REAL>(degree);
  p+=dp;
  for (int i=1;i<degree;i++,p+=dp) mapping_support_points[i]=p;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Interval::diameter(const NumPtr<Point<SPACEDIM> > &vertices) const{
  return sqrt((vertices[vertexIndex(1)]
              -vertices[vertexIndex(0)]).square());
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Interval::volume(const NumPtr<Point<SPACEDIM> > &vertices) const { 
#if (SPACEDIM==1)
  REAL vol=vertices[vertexIndex(1)][0]-vertices[vertexIndex(0)][0];
  return vol;
#else
  return sqrt((vertices[vertexIndex(1)]
              -vertices[vertexIndex(0)]).square());
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::initializeMappingSupportPoints(int degree,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  NumPtr<Point<1> > eslp;
  staticEquallySpacedLatticePoints(degree,eslp);
  int neslp=eslp.getNumber();
  NumPtr<Point<SPACEDIM> > iv(2);
  for (int v=0;v<2;v++) {
    iv[v]=tessellation_vertices[vertexIndex(v)];
  }
  for (int i=0;i<neslp;i++) {
    REAL eslpi=eslp[i][0];
    points[i]=iv[1]*eslpi+iv[0]*(1.-eslpi);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::getIntermediatePoints(
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  int n=points.getNumber();
  Point<SPACEDIM> p=tessellation_vertices[vertexIndex(0)];
  Point<SPACEDIM> dp=(tessellation_vertices[vertexIndex(1)]-p)
                    /static_cast<REAL>(n+1);
  p+=dp;
  for (int i=0;i<n;i++,p+=dp) points[i]=p;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::getMappedPoints(
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
const NumPtr<Point<1> > &canonical_points,
NumPtr<Point<SPACEDIM> > &points,NumPtr<Point<SPACEDIM> > &dpointsdxi) 
const {
  int n=canonical_points.getNumber();
  points.allocate(n);
  dpointsdxi.allocate(n);
  Point<SPACEDIM> p=tessellation_vertices[vertexIndex(0)];
  Point<SPACEDIM> dp=tessellation_vertices[vertexIndex(1)]-p;
  dpointsdxi.initialize(dp);
  for (int i=0;i<n;i++) {
    points[i]=p+dp*canonical_points[i][0];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Interval::printOn(ostream &os) const {
  os << "Interval:" << endl;
  Shape1::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==2)
void Interval::draw(const NumPtr<Point<2> > &vertices,
AreaGraphTool &agt) const {
  Point<2> p0=vertices[vertexIndex(0)];
  Point<2> p1=vertices[vertexIndex(1)];
  agt.movePen(p0[0],p0[1]);
  agt.drawLine(p1[0],p1[1]);
}
#endif
#if (SPACEDIM==3)
void Interval::draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE&,
const NumPtr<Point<3> > &vertices,VolGraphTool &vgt) const {
  Point<3> p0=vertices[vertexIndex(0)];
  Point<3> p1=vertices[vertexIndex(1)];
  Vector3 v0(p0[0],p0[1],p0[2]);
  Vector3 v1(p1[0],p1[1],p1[2]);
  vgt.drawLine(v0,v1);
}
#endif
#endif

#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool findOtherVertexOnLine(const Interval *interval,int vertex_index,
int &other_vertex_index) {
  if (interval->vertexIndex(0)==vertex_index) {
    other_vertex_index=interval->vertexIndex(1);
    return true;
  } else if (interval->vertexIndex(1)==vertex_index) {
    other_vertex_index=interval->vertexIndex(0);
    return true;
  } 
  return false;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Triangle::Triangle(int edi,const int l[3],const int v[3]) : 
Shape2(edi,3,3) {
  CHECK_TEST(l[0]!=l[1] && l[0]!=l[2] && l[1]!=l[2]);
  CHECK_TEST(v[0]!=v[1] && v[0]!=v[2] && v[1]!=v[2]);
  int vmin=min(min(v[0],v[1]),v[2]);
#if (SPACEDIM==2)
//rotate to get vertex with lowest index first
  if (v[0]==vmin) {
    setCodimension1ShapeIndex(0,l[0]);
    setCodimension1ShapeIndex(1,l[1]);
    setCodimension1ShapeIndex(2,l[2]);
    setVertexIndex(0,v[0]);
    setVertexIndex(1,v[1]);
    setVertexIndex(2,v[2]);
  } else if (v[1]==vmin) {
    setCodimension1ShapeIndex(0,l[1]);
    setCodimension1ShapeIndex(1,l[2]);
    setCodimension1ShapeIndex(2,l[0]);
    setVertexIndex(0,v[1]);
    setVertexIndex(1,v[2]);
    setVertexIndex(2,v[0]);
  } else {
    setCodimension1ShapeIndex(0,l[2]);
    setCodimension1ShapeIndex(1,l[0]);
    setCodimension1ShapeIndex(2,l[1]);
    setVertexIndex(0,v[2]);
    setVertexIndex(1,v[0]);
    setVertexIndex(2,v[1]);
  }
#else
//rotate and/or flip to get vertices with lowest indices first
  if (v[0]==vmin) {
    if (v[1]<v[2]) {
      setCodimension1ShapeIndex(0,l[0]);
      setCodimension1ShapeIndex(1,l[1]);
      setCodimension1ShapeIndex(2,l[2]);
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[2]);
    } else {
      setCodimension1ShapeIndex(0,l[1]);
      setCodimension1ShapeIndex(1,l[0]);
      setCodimension1ShapeIndex(2,l[2]);
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[1]);
    }
  } else if (v[1]==vmin) {
    if (v[2]<v[0]) {
      setCodimension1ShapeIndex(0,l[1]);
      setCodimension1ShapeIndex(1,l[2]);
      setCodimension1ShapeIndex(2,l[0]);
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[0]);
    } else {
      setCodimension1ShapeIndex(0,l[2]);
      setCodimension1ShapeIndex(1,l[1]);
      setCodimension1ShapeIndex(2,l[0]);
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[2]);
    }
  } else {
    if (v[0]<v[1]) {
      setCodimension1ShapeIndex(0,l[2]);
      setCodimension1ShapeIndex(1,l[0]);
      setCodimension1ShapeIndex(2,l[1]);
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[1]);
    } else {
      setCodimension1ShapeIndex(0,l[0]);
      setCodimension1ShapeIndex(1,l[2]);
      setCodimension1ShapeIndex(2,l[1]);
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[0]);
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Triangle::Triangle(const int l[3],const int v[3],int di):Shape2(-1,3,3){
  CHECK_TEST(v[0]!=v[1] && v[0]!=v[2] && v[1]!=v[2]);
  CHECK_TEST(l[0]!=l[1] && l[0]!=l[2] && l[1]!=l[2]);
  for (int i=0;i<3;i++) {
    setVertexIndex(i,v[i]);
    setCodimension1ShapeIndex(i,l[i]);
  }
  dof_index=di;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape1* Triangle::newLineElement(int f) const {
//vertices ordered consistent with lattice indices
  int v[2]={
    vertexIndex(staticElementVertexFromFace(f,0)),
    vertexIndex(staticElementVertexFromFace(f,1))
  };
  return OPERATOR_NEW Interval(v,-1); // second arg is dof_index
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Triangle::contains(const Point<SPACEDIM> &x,
const NumPtr<Point<SPACEDIM> > &vertices) const {
#if (SPACEDIM==2)
  const Point<2> &v0(vertices[vertexIndex(0)]);
  const Point<2> &v1(vertices[vertexIndex(1)]);
  const Point<2> &v2(vertices[vertexIndex(2)]);
  Tensor<1,2> perp;
  crossProduct(perp,Tensor<1,2>(v1-v2));
  if (perp.getValues()*(x-v2)<0.) return false;

  crossProduct(perp,Tensor<1,2>(v2-v0));
  if (perp.getValues()*(x-v0)<0.) return false;

  crossProduct(perp,Tensor<1,2>(v0-v1));
  return perp.getValues()*(x-v1)>=0.;
#else
  return Shape::contains(x,vertices);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_H::prolongationMatrix
// 3 rotations of a square times 4 children = 12 cases
Point<SPACEDIM> Triangle::parentCoordinateForCanonical(
const Point<SPACEDIM> &cp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
#if (SPACEDIM!=2)
  OBSOLETE("should not be calling this");
#else
  if (c==3) {
    int cvi=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[0]);
    CHECK_TEST(cvi>=0);
    switch (cvi) {
      case 0:
        return Point<2>((cp[0]+cp[1])*0.5,(1.-cp[0])*0.5);
      case 1:
        return Point<2>((1.-cp[0])*0.5,(1.-cp[1])*0.5);
      case 2:
        return Point<2>((1.-cp[1])*0.5,(cp[0]+cp[1])*0.5);
    }
  } else {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    CHECK_TEST(vertexIndex(c)==svi);
    int cvi=child_shape->vertexEqualToVertexIndex(svi);
    CHECK_TEST(cvi>=0);

    switch (c) {
      case 0: {
        switch (cvi) {
          case 0:
            return cp*0.5;
          case 1:
            return Point<2>(cp[1]*0.5,(1.-cp[0]-cp[1])*0.5);
          case 2:
            return Point<2>((1.-cp[1])*0.5,cp[0]*0.5);
        }
      }
      case 1: {
        switch (cvi) {
          case 0:
            return Point<2>(1.-(cp[0]+cp[1])*0.5,cp[0]*0.5);
          case 1:
            return Point<2>((1.+cp[0])*0.5,cp[1]*0.5);
          case 2:
            return Point<2>((1.+cp[1])*0.5,(1.-cp[0]-cp[1])*0.5);
        }
        break;
      }
      case 2: {
        switch (cvi) {
          case 0:
            return Point<2>(cp[1]*0.5,1.-(cp[0]+cp[1])*0.5);
          case 1:
            return Point<2>((1.-cp[0]-cp[1])*0.5,(1.+cp[0])*0.5);
          case 2:
            return Point<2>(cp[0]*0.5,(1.+cp[1])*0.5);
        }
        break;
      }
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Triangle::childCoordinateOfCanonical(
const Point<SPACEDIM> &pp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
#if (SPACEDIM!=2)
  OBSOLETE("should not be calling this");
#else
  if (c==3) {
    int cvi=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[0]);
    CHECK_TEST(cvi>=0);
    switch (cvi) {
      case 0:
        return Point<2>(1.-pp[1]*2.,(pp[0]+pp[1])*2.-1.);
      case 1: {
        return Point<2>(1.-pp[0]*2.,1.-pp[1]*2.);
      }
      case 2:
        return Point<2>((pp[0]+pp[1])*2.-1.,1.-pp[0]*2.);
    }
  } else {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    CHECK_TEST(vertexIndex(c)==svi);
    int cvi=child_shape->vertexEqualToVertexIndex(svi);
    CHECK_TEST(cvi>=0);

    switch (c) {
      case 0: {
        switch (cvi) {
          case 0:
            return pp*2.;
          case 1:
            return Point<2>(1.-(pp[0]+pp[1])*2.,pp[0]*2.);
          case 2:
            return Point<2>(pp[1]*2.,1.-pp[0]*2.);
        }
      }
      case 1: {
        switch (cvi) {
          case 0:
            return Point<2>(pp[1]*2.,(1.-pp[0]-pp[1])*2.);
          case 1:
            return Point<2>(pp[0]*2.-1.,pp[1]*2.);
          case 2:
            return Point<2>((1.-pp[0]-pp[1])*2.,pp[0]*2.-1.);
        }
        break;
      }
      case 2: {
        switch (cvi) {
          case 0:
            return Point<2>((1.-pp[0]-pp[1])*2.,pp[0]*2.);
          case 1:
            return Point<2>(2.*pp[1]-1.,(1.-pp[0]-pp[1])*2.);
          case 2:
            return Point<2>(pp[0]*2.,pp[1]*2.-1.);
        }
        break;
      }
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
const Shape* Triangle::reorientFor(const Shape *parent,int c,
const NumPtr<int> &line_interior_child_vertices) const {
#ifdef DEBUG
  if (c<3) CHECK_SAME(vertexIndex(0),parent->vertexIndex(c));
#endif
  if (c<3) return this;
  else {
    int cv0=line_interior_child_vertices[0];
    if (vertexIndex(0)==cv0) return this;
    else if (vertexIndex(1)==cv0) {
      int v[3]={vertexIndex(1),vertexIndex(2),vertexIndex(0)};
      int l[3]={codimension1ShapeIndex(1),codimension1ShapeIndex(2),
                codimension1ShapeIndex(0)};
      return OPERATOR_NEW Triangle(l,v,dof_index);
    } else {
      CHECK_SAME(vertexIndex(2),cv0);
      int v[3]={vertexIndex(2),vertexIndex(0),vertexIndex(1)};
      int l[3]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                codimension1ShapeIndex(1)};
      return OPERATOR_NEW Triangle(l,v,dof_index);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::staticProjectToFace(int f,
const NumPtr<Point<1> > &q_points,NumPtr<Point<2> > &q_points_face) {
  switch (f) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=0.;
        q_points_face[i][1]=q_points[i][0];
      }
      break;
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=0.;
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][1]=q_points[i][0];
        q_points_face[i][0]=1.-q_points[i][0];
      }
      break;
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::projectToSubface(int f,int s,
const NumPtr<Point<1> > &q_points,NumPtr<Point<2> > &q_points_subface) {
  switch (f) {
    case 0: {
      if (s==0) { // child 2
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=0.;
          qps[1]=1.-qp;
        }
      } else { // child 0
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=0.;
          qps[1]=qp;
        }
      }
      break;
    }
    case 1: {
      if (s==0) { // child 0
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=qp;
          qps[1]=0.;
        }
      } else { // child 1
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=1.-qp;
          qps[1]=0.;
        }
      }
      break;
    }
    case 2: {
      if (s==0) { // child 1
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=1.-qp;
          qps[1]=qp;
        }
      } else { // child 2
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=qp;
          qps[1]=1.-qp;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void setTriangleLineIndices(const int v[3],
const map<pair<int,int>,int> &interior_child_line_map,int l[3]) {
  for (int j=0;j<3;j++) {
    int iv[2]={v[(j+2)%3],v[j]};
    pair<int,int> p(min(iv[0],iv[1]),max(iv[0],iv[1]));
    map<pair<int,int>,int>::const_iterator iclm_iterator=
      interior_child_line_map.find(p);
    CHECK_TEST(iclm_iterator!=interior_child_line_map.end());
    l[j]=iclm_iterator->second;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape2* Triangle::newChildShape(int c,int child_element_index,
const NumPtr<int> &interior_child_vertex_indices,
const NumPtr<int> &interior_child_line_vertex_indices,
const map<pair<int,int>,int> &interior_child_line_map) const {
  CHECK_BOUNDS(c,0,4);
  CHECK_SAME(interior_child_vertex_indices.getNumber(),0);
  CHECK_SAME(interior_child_line_vertex_indices.getNumber(),3);
  int v[3];
  switch (c) {
    case 0:
      v[0]=vertex_indices[0];                     // v0
      v[1]=interior_child_line_vertex_indices[1]; // v4
      v[2]=interior_child_line_vertex_indices[0]; // v3
      break;
    case 1:
      v[0]=vertex_indices[1];                     // v1
      v[1]=interior_child_line_vertex_indices[2]; // v5
      v[2]=interior_child_line_vertex_indices[1]; // v4
      break;
    case 2:
      v[0]=vertex_indices[2];                     // v2
      v[1]=interior_child_line_vertex_indices[0]; // v3
      v[2]=interior_child_line_vertex_indices[2]; // v5
      break;
    case 3:
      v[0]=interior_child_line_vertex_indices[0]; // v3
      v[1]=interior_child_line_vertex_indices[1]; // v4
      v[2]=interior_child_line_vertex_indices[2]; // v5
  }
  int l[3];
  setTriangleLineIndices(v,interior_child_line_map,l);
  return OPERATOR_NEW Triangle(child_element_index,l,v);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::staticEquallySpacedLatticePoints(int n,
NumPtr<Point<2> > &points) {
  points.allocate(staticNumberEquallySpacedLatticePoints(n));
  Point<2> p(0.,0.);
  REAL s=(n>0 ? 1./static_cast<REAL>(n) : 0.);
  int k=0;
  for (int j=0;j<=n;j++,p[1]+=s) {
    for (int i=0;i+j<=n;i++,p[0]+=s,k++) points[k]=p;
    p[0]=0.;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Triangle::staticEquallySpacedLatticePointIndexOfVertex(int v,int n) {
  CHECK_TEST(n>0);
  CHECK_TEST(v<3);
  switch (v) {
    case 0:
      return 0;
    case 1: 
      return n;
    case 2:
      return ((n+1)*(n+2))/2-1;
    default:
      OBSOLETE("bad vertex");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::staticEquallySpacedLatticePointIndicesOfFace(int f,int n,
NumPtr<int> &indices) {
  CHECK_TEST(n>0);
  indices.allocate(n+1);
  switch (f) {
    case 0: { // (0,j)
      int ij=0;
      for (int j=0;j<=n;j++) {
        indices[j]=ij;
        ij+=n+1-j;
      }
      break;
    }
    case 1: { // (i,0)
      for (int i=0;i<=n;i++) indices[i]=i;
      break;
    }
    case 2: { // (n-j,j)
      int ij=n;
      for (int j=0;j<=n;j++) {
        indices[j]=ij;
        ij+=n-j;
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::equallySpacedLatticePointIndicesOfFace(int f,int n,
NumPtr<int> &indices) const {
  CHECK_TEST(n>0);
  indices.allocate(n+1);
  switch (f) {
    case 0: { // (0,j)
      if (vertexIndex(0)<vertexIndex(2)) {
        int ij=0;
        for (int j=0;j<=n;j++) {
          indices[j]=ij;
          ij+=n+1-j;
        }
      } else {
        int ij=(n*(n+3))/2;
        for (int j=0;j<=n;j++) {
          indices[j]=ij;
          ij-=n-1+j;
        }
      }
      break;
    }
    case 1: { // (i,0)
      if (vertexIndex(0)<vertexIndex(1)) {
        for (int i=0;i<=n;i++) indices[i]=i;
      } else {
        for (int i=0;i<=n;i++) indices[i]=n-i;
      }
      break;
    }
    case 2: {
      if (vertexIndex(1)<vertexIndex(2)) { // (n-j,j)
        int ij=n;
        for (int j=0;j<=n;j++) {
          indices[j]=ij;
          ij+=n-j;
        }
      } else { // (i,n-i)
        int ij=(n*(n+3))/2;
        for (int i=0;i<=n;i++) {
          indices[i]=ij;
          ij-=i+1;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::staticEquallySpacedLatticePointIndicesOfFaceInterior(
int f,int n,NumPtr<int> &indices) {
  CHECK_TEST(n>0);
  indices.allocate(n-1);
  switch (f) {
    case 0: { // (0,j)
      int ij=n+1;
      for (int j=1;j<n;j++) {
        indices[j-1]=ij;
        ij+=n+1-j;
      }
      break;
    }
    case 1: { // (i,0)
      for (int i=1;i<n;i++) indices[i-1]=i;
      break;
    }
    case 2: { // (n-j,j)
      int ij=2*n;
      for (int j=1;j<n;j++) {
        indices[j-1]=ij;
        ij+=n-j;
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::equallySpacedLatticePointIndicesOfFaceInterior(int f,
int n,NumPtr<int> &indices) const {
  CHECK_TEST(n>0);
  indices.allocate(n-1);
  switch (f) {
    case 0: { // (0,j)
      int ij=n+1;
      for (int j=1;j<n;j++) {
        indices[j-1]=ij;
        ij+=n+1-j;
      }
      break;
    }
    case 1: { // (i,0)
      for (int i=1;i<n;i++) {
        indices[i-1]=i;
      }
      break;
    }
    case 2: {
      if (vertexIndex(1)<vertexIndex(2)) { // (n-j,j)
        int ij=2*n;
        for (int j=1;j<n;j++) {
          indices[j-1]=ij;
          ij+=n-j;
        }
      } else { // (i,n-i)
        int ij=((n-1)*(n+4))/2+1;
        for (int i=1;i<n;i++) {
          indices[i-1]=ij;
          ij-=i+1;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::reorderFaceDofs(int n,NumPtr<int> &renumber) const {
  int nl=n+1;
  for (int i=0;i<3*nl;i++) renumber[i]=i;
  int offset=0;
  NumPtr<int> r=renumber;
  for (int f=0;f<3;f++) {
    int v[2]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1))};
    Interval face_element(v,-1);
    face_element.reorderDofs(n,r);
    for (int i=0;i<nl;i++) r[i]+=offset;
    r+=nl;
    offset+=nl;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<3*nl;i++) {
    if (renumber[i]>=3*nl || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(3*nl);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<3*nl;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<3*nl;i++) CHECK_BOUNDS(renumber_inverse[i],0,3*nl);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::reorderDofs(int n,NumPtr<int> &renumber) const {
  int nt=staticNumberEquallySpacedLatticePoints(n);
  for (int i=0;i<nt;i++) renumber[i]=i;
  int vmin=min(vertexIndex(0),min(vertexIndex(1),vertexIndex(2)));
  if (vertexIndex(0)==vmin) {
    if (vertexIndex(2)<vertexIndex(1)) {
      int m=0;
      int ki=0;
      for (int i=0;i<=n;i++) {
        int k=ki;
        for (int j=0;i+j<=n;j++) {
          renumber[k]=m++;
          k+=n+1-j;
        }
        ki++;
      }
    }
  } else if (vertexIndex(1)==vmin) {
    if (vertexIndex(0)<vertexIndex(2)) {
      int m=0;
      int kj=n;
      for (int j=0;j<=n;j++) {
        int k=kj;
        for (int i=n-j;i>=0;i--) {
          renumber[k]=m++;
          k--;
        }
        kj+=n-j;
      }
    } else {
      int m=0;
      int ki=n;
      for (int ipj=n;ipj>=0;ipj--) {
        int k=ki;
        for (int j=0;j<=ipj;j++) {
          renumber[k]=m++;
          k+=n-j;
        }
        ki--;
      }
    }
  } else {
    if (vertexIndex(0)<vertexIndex(1)) {
      int m=0;
      int ki=staticNumberEquallySpacedLatticePoints(n)-1;
      for (int i=0;i<=n;i++) {
        int k=ki;
        for (int j=n-i;j>=0;j--) {
          renumber[k]=m++;
          k-=n-j+2;;
        }
        ki-=i+1;
      }
    } else {
      int m=0;
      int kj=staticNumberEquallySpacedLatticePoints(n)-1;
      for (int ipj=n;ipj>=0;ipj--) {
        int k=kj;
        for (int j=ipj;j>=0;j--) {
          renumber[k]=m++;
          k-=n-j+1;
        }
        kj-=n+2-ipj;
      }
    }
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<nt;i++) {
    if (renumber[i]>=nt || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] 
           << " ***" << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(nt);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<nt;i++) renumber_inverse[renumber[i]]=i;
  for (int i=0;i<nt;i++) CHECK_BOUNDS(renumber_inverse[i],0,nt);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::reorderEquallySpacedLatticePoints(int n,
NumPtr<int> &renumber) const {
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
//vertices first
  int m=0;
  renumber[0]=m++;
  renumber[n]=m++;
  renumber[renumber.getNumber()-1]=m++;
//face 0
  if (vertexIndex(0)<vertexIndex(2)) {
    int l=n+1;
    for (int j=1;j<n;j++) {
      renumber[l]=m++;
      l+=n-j+1;
    }
  } else {
    int l=renumber.getNumber()-3;
    for (int j=n-1;j>0;j--) {
      renumber[l]=m++;
      l-=n-j+2;
    }
  }
//face 1
  if (vertexIndex(0)<vertexIndex(1)) {
    for (int i=1;i<n;i++) renumber[i]=m++;
  } else {
    int l=n-1;
    for (int i=n-1;i>0;i--) renumber[l--]=m++;
  } 
//face 2
  if (vertexIndex(1)<vertexIndex(2)) {
    int l=2*n;
    for (int j=1;j<n;j++) {
      renumber[l]=m++;
      l+=n-j;
    }
  } else {
    int l=renumber.getNumber()-2;
    for (int i=n-1;i>0;i--) {
      renumber[l]=m++;
      l-=n-i+1;
    }
  }
//interior points last
  int l=n+2;
  for (int j=1;j<n;j++) {
    for (int i=1;i<n-j;i++) renumber[l++]=m++;
    l+=2;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]==INT_MAX) {
      cout << "\trenumber[" << i << "] = *** " << -1 << " ***" << endl;
      success=false;
    } else if (renumber[i]>=renumber.getNumber()) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::reorderChildEquallySpacedLatticePoints(int n,
const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices,NumPtr<int> &renumber) 
const {
  const Triangle *child_triangle=
    dynamic_cast<const Triangle*>(child_shape);
  CHECK_POINTER(child_triangle);

  int reorder_case=-1;
  if (c==3) {
    int cvi=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[0]);
    CHECK_TEST(cvi>=0);
    reorder_case=cvi;
  } else {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    CHECK_TEST(vertexIndex(c)==svi);
    int cvi=child_shape->vertexEqualToVertexIndex(svi);
    CHECK_TEST(cvi>=0);
    reorder_case=cvi;
  }
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  switch (reorder_case) {
    case 0: {
//    renumber[lattice index] = lattice_index:
//    9
//    7 8
//    4 5 6
//    0 1 2 3
      for (int i=0;i<renumber.getNumber();i++) renumber[i]=i;
      break;
    }
    case 1: {
//    renumber[lattice index]:
//    3
//    6 2
//    8 5 1
//    9 7 4 0
      int ij=0;
      int nk=n;
      for (int k=0;k<=n;k++) {
        int nj=nk;
        for (int j=0;j<=n-k;j++) {
//        renumber[ij++]=nj;
          renumber[nj]=ij++;
          nj+=n-j;
        }
        nk--;
      }
      break;
    }
    case 2: {
//    renumber[lattice index]:
//    0
//    1 4
//    2 5 7
//    3 6 8 9
      int ij=0;
      int ni=(n*(n+3))/2;
      for (int i=0;i<=n;i++) {
        int nk=ni;
        for (int k=0;k<=n-i;k++) {
//        renumber[ij++]=nk;
          renumber[nk]=ij++;
          nk-=i+k+2;
        }
        ni-=i+1;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//renumber[child_shape dof]=shape fcn order in
//  Triangle::parentCoordinateOfCanonical(int,Point<2>,int)
//called from FE_H::constraints
void Triangle::reorderChildDofPoints(int n,const Shape *child_shape,
int c,const NumPtr<int> &line_interior_child_vertices,
NumPtr<int> &renumber) const {
  const Triangle *child_triangle=
    dynamic_cast<const Triangle*>(child_shape);
  CHECK_POINTER(child_triangle);

  int reorder_case=-1;
  if (c==3) {
    OBSOLETE("who called this?");
    int cvi=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[0]);
    CHECK_TEST(cvi>=0);
    reorder_case=cvi;
  } else {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    CHECK_TEST(vertexIndex(c)==svi);
    int cvi=child_shape->vertexEqualToVertexIndex(svi);
    CHECK_TEST(cvi>=0);
    reorder_case=cvi;
  }
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  switch (reorder_case) {
    case 0: {
//    dofs on child  parentCoordinateOfCanonical(dof point)
//    2              2
//    4 8            4 8
//    3 9 7          3 9 7
//    0 5 6 1        0 5 6 1
      for (int i=0;i<renumber.getNumber();i++) renumber[i]=i;
      break;
    }
    case 1: {
      OBSOLETE("who called this?");
//    renumber[child_shape dof]
//    1
//    7 6
//    8 9 5
//    2 4 3 0
      int ij=0;
      renumber[1]=ij++;
      renumber[2]=ij++;
      renumber[0]=ij++;
      int d=2*n;
      for (int j=n-1;j>=1;j--) {
        renumber[d--]=ij++;
      }
      d=2*n+1;
      for (int i=0;i<n;i++) {
        renumber[d++]=ij++;
      }
      d=n+1;
      for (int j=n-1;j>=1;j--) {
        renumber[d--]=ij++;
      }
      int di=4*n-3;
      for (int i=n-2;i>=1;i--) {
        int dj=di;
        for (int j=1;j<n-i-1;j++) {
          renumber[dj]=ij++;
          dj+=n-j-1;
        }
        di--;
      }
      break;
    }
    case 2: {
      OBSOLETE("who called this?");
//    renumber[child_shape dof]
//    0
//    5 3
//    6 9 4
//    1 7 8 2
      int ij=0;
      renumber[2]=ij++;
      renumber[0]=ij++;
      renumber[1]=ij++;
      int d=3*n-1;
      for (int j=n-1;j>=1;j--) {
        renumber[d--]=ij++; 
      }
      d=n+1;
      for (int j=n-1;j>=1;j--) {
        renumber[d--]=ij++; 
      }
      d=3;
      for (int i=1;i<n;i++) {
        renumber[d++]=ij++;
      }
      int di=(n*(n+3))/2;
      for (int i=n-2;i>=1;i--) {
        int dj=di;
        for (int j=n-1-i;j>=1;j--) {
          renumber[dj]=ij++;
          dj-=n-j;
        }
        di-=i-1;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::staticInteriorLatticePointIndices(int n,NumPtr<int> &v) {
  v.allocate(staticNumberInteriorEquallySpacedLatticePoints(n));
  int ijk=n+2;
  int iv=0;
  for (int j=1;j<n;j++) {
    for (int i=1;i+j<n;i++) {
      v[iv]=ijk;
      ijk++;
      iv++;
    }
    ijk+=2;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//the following is due to Scott, SIAM J. Num. Anal. 12 (404--427), 1975
void Triangle::computeInteriorLatticePoints(int degree,int f,
NumPtr<Point<SPACEDIM> > &mapping_support_points) const {
  if (degree<=2) return;
  int npf=degree-1;
  CHECK_SAME(mapping_support_points.getNumber(),
             staticNumberEquallySpacedLatticePoints(degree));
  switch (f) {
    case 0: {
      NumPtr<Point<SPACEDIM> > fsp0(npf);
      NumPtr<Point<SPACEDIM> > fsp1(npf);
      int k=degree+1;
      for (int j=1;j<degree;j++)  {
        fsp0[j-1]=mapping_support_points[k];
        k+=degree-j+1;
      }
      for (int i=1;i<degree;i++)  {
        fsp1[i-1]=mapping_support_points[i];
      }
      k=degree+2;
      for (int j=1;j+1<degree;j++) {
        for (int i=1;i+j<degree;i++,k++) {
          mapping_support_points[k]=
            (fsp0[i+j-1]*static_cast<REAL>(j)
            +fsp1[i+j-1]*static_cast<REAL>(i))/static_cast<REAL>(i+j);
        }
        k+=2;
      }
      break;
    }
    case 1: {
      NumPtr<Point<SPACEDIM> > fsp1(npf);
      NumPtr<Point<SPACEDIM> > fsp2(npf);
      for (int i=1;i<degree;i++)  {
        fsp1[i-1]=mapping_support_points[i];
      }
      int k=2*degree;
      for (int j=1;j<degree;j++)  {
        fsp2[j-1]=mapping_support_points[k];
        k+=degree-j;
      }
      k=degree+2;
      for (int j=1;j+1<degree;j++) {
        for (int i=1;i+j<degree;i++,k++) {
          mapping_support_points[k]=
            (fsp1[i-1]*static_cast<REAL>(degree-j-i)
            +fsp2[degree-i-1]*static_cast<REAL>(j))
            /static_cast<REAL>(degree-i);
        }
        k+=2;
      }
      break;
    }
    case 2: {
      NumPtr<Point<SPACEDIM> > fsp0(npf);
      NumPtr<Point<SPACEDIM> > fsp2(npf);
      int k=degree+1;
      for (int j=1;j<degree;j++)  {
        fsp0[j-1]=mapping_support_points[k];
        k+=degree-j+1;
      }
      k=2*degree;
      for (int j=1;j<degree;j++)  {
        fsp2[j-1]=mapping_support_points[k];
        k+=degree-j;
      }
      k=degree+2;
      for (int j=1;j+1<degree;j++) {
        for (int i=1;i+j<degree;i++,k++) {
          mapping_support_points[k]=
            (fsp0[j-1]*static_cast<REAL>(degree-j-i)
            +fsp2[j-1]*static_cast<REAL>(i))
            /static_cast<REAL>(degree-j);
        }
        k+=2;
      }
      break;
    }
    default: { // no face on boundary
      REAL rd=1./static_cast<REAL>(degree);
      Point<SPACEDIM> px=mapping_support_points[
        staticEquallySpacedLatticePointIndexOfVertex(0,degree)];
      Point<SPACEDIM> dpx=(mapping_support_points[
        staticEquallySpacedLatticePointIndexOfVertex(1,degree)]-px)*rd;
      Point<SPACEDIM> dpy=(mapping_support_points[
        staticEquallySpacedLatticePointIndexOfVertex(2,degree)]-px)*rd;
      px+=dpx+dpy;
      int k=degree+2;
      for (int j=1;j+1<degree;j++,k+=2,px+=dpy) {
        Point<SPACEDIM> p=px;
        for (int i=1;i+j<degree;i++,k++,p+=dpx) {
          mapping_support_points[k]=p;
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Triangle::lenoirCanonicalFacePoints(int f,
const NumPtr<Point<2> > &eslp,NumPtr<Point<1> > &canonical_face_points) 
const {
  canonical_face_points.allocate(eslp.getNumber());
  if (f==2) { // cf Tetrahedron::lenoirCanonicalLinePoints case 1
    for (int i=0;i<eslp.getNumber();i++) {
      const Point<2> &eslpi=eslp[i];
      REAL den=eslpi[0]+eslpi[1];
      if (den>0.) canonical_face_points[i][0]=eslpi[1]/den;
      else canonical_face_points[i][0]=0.;
    }
  } else { // cf Tetrahedron::lenoirCanonicalLinePoints cases 2 or 0
    for (int i=0;i<eslp.getNumber();i++) {
      const Point<2> &eslpi=eslp[i];
      REAL den=1.-eslpi[f];
      if (den>0.) canonical_face_points[i][0]=eslpi[1-f]/den;
      else canonical_face_points[i][0]=0.;
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Triangle::lenoirInterpolationForFace(int f,int m,
const NumPtr<Point<2> > &eslp,
const NumPtr<Point<3> > &old_points,
const NumPtr<Point<3> > &new_points,
NumPtr<Point<3> > &higher_order_mapping_support_points) const {
  if (f==2) {
    for (int i=0;i<eslp.getNumber();i++) {
      const Point<2> &eslpi=eslp[i];
      higher_order_mapping_support_points[i]+=
        (new_points[i]-old_points[i])
        *pow(eslpi[0]+eslpi[1],m);
    }
  } else {
    for (int i=0;i<eslp.getNumber();i++) {
      const Point<2> &eslpi=eslp[i];
      higher_order_mapping_support_points[i]+=
        (new_points[i]-old_points[i])*pow(1.-eslpi[f],m);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Triangle::faceCenter(
const NumPtr<Point<SPACEDIM> > &vertices,int f) const {
  switch (f) {
    case 0:
      return (vertices[vertexIndex(2)]+vertices[vertexIndex(0)])*0.5;
    case 1:
      return (vertices[vertexIndex(0)]+vertices[vertexIndex(1)])*0.5;
    case 2:
      return (vertices[vertexIndex(1)]+vertices[vertexIndex(2)])*0.5;
    default:
      OBSOLETE("bad face");
      return Point<SPACEDIM>(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Triangle::diameter(const NumPtr<Point<SPACEDIM> > &vertices) const{
  return sqrt(max((vertices[vertexIndex(1)]
                  -vertices[vertexIndex(0)]).square(),
                  max((vertices[vertexIndex(2)]
                      -vertices[vertexIndex(0)]).square(),
                      (vertices[vertexIndex(2)]
                      -vertices[vertexIndex(1)]).square())));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Triangle::volume(const NumPtr<Point<SPACEDIM> > &vertices) const {
  Tensor<1,SPACEDIM> v10(vertices[vertexIndex(1)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v20(vertices[vertexIndex(2)]
                        -vertices[vertexIndex(0)]);
  REAL vol;
#if (SPACEDIM==2)
  Tensor<1,SPACEDIM> v20_perp;
  crossProduct(v20_perp,v20);
  product(vol,v10,v20_perp);
#else
  Tensor<1,SPACEDIM> cp;
  crossProduct(cp,v10,v20);
  product(vol,cp,cp);
  vol=sqrt(vol);
#endif
  return vol*0.5;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Triangle::lineTangentPointsBackward(int l) const {
  switch (l) {
    case 0:
      return true;
    case 1:
      return false;
    case 2:
      return vertexIndex(2)<vertexIndex(1);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::initializeMappingSupportPoints(int degree,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  NumPtr<Point<2> > eslp;
  staticEquallySpacedLatticePoints(degree,eslp);
  int neslp=eslp.getNumber();
  NumPtr<Point<SPACEDIM> > tv(3);
  for (int v=0;v<3;v++) {
    tv[v]=tessellation_vertices[vertexIndex(v)];
  }
  for (int i=0;i<neslp;i++) {
    const Point<2> &eslpi=eslp[i];
    points[i]=tv[1]*eslpi[0]+tv[2]*eslpi[1]
             +tv[0]*(1.-eslpi[0]-eslpi[1]);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::getIntermediatePoints(
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  int n=points.getNumber();
  int m=
    (static_cast<int>(nearbyint(sqrt(1.+static_cast<REAL>(8*n))))+1)/2;
  CHECK_SAME((m*(m-1))/2,n);
  Point<SPACEDIM> px=tessellation_vertices[vertexIndex(0)];
  REAL rm=1./static_cast<REAL>(m+1);
  Point<SPACEDIM> dpx=(tessellation_vertices[vertexIndex(1)]-px)*rm;
  Point<SPACEDIM> dpy=(tessellation_vertices[vertexIndex(2)]-px)*rm;
  px+=dpx+dpy;
  int k=0;
  for (int j=0;j+1<m;j++,px+=dpy) {
    Point<SPACEDIM> p=px;
    for (int i=0;i+j+1<m;i++,k++,p+=dpx) points[k]=p;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Triangle::getMappedPoints(
const NumPtr<Point<3> > &tessellation_vertices,
const NumPtr<Point<2> > &canonical_points,NumPtr<Point<3> > &points,
NumPtr<Point<3> > &dpointsdxi0,NumPtr<Point<3> > &dpointsdxi1) const {
  int n=canonical_points.getNumber();
  points.allocate(n);
  dpointsdxi0.allocate(n);
  dpointsdxi1.allocate(n);

  Point<3> p=tessellation_vertices[vertexIndex(0)];
  Point<3> dpx=tessellation_vertices[vertexIndex(1)]-p;
  Point<3> dpy=tessellation_vertices[vertexIndex(2)]-p;
  dpointsdxi0.initialize(dpx);
  dpointsdxi1.initialize(dpy);
  for (int i=0;i<n;i++) {
    const Point<2> &cp=canonical_points[i];
    points[i]=p+dpx*cp[0]+dpy*cp[1];
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Triangle::printOn(ostream &os) const {
  os << "Triangle:" << endl;
  Shape2::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Triangle::checkFaceVertices(int f,const Shape1 *face_element) const
{
  CHECK_BOUNDS(f,0,3);
  if (face_element->vertexIndex(0)==
  vertexIndex(canonical_vertex_indices[f][0])) {
    CHECK_SAME(face_element->vertexIndex(1),
               vertexIndex(canonical_vertex_indices[f][1]));
  } else {
    CHECK_SAME(face_element->vertexIndex(1),
               vertexIndex(canonical_vertex_indices[f][0]));
    CHECK_SAME(face_element->vertexIndex(0),
               vertexIndex(canonical_vertex_indices[f][1]));
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==2)
void Triangle::draw(const NumPtr<Point<2> > &vertices,
AreaGraphTool &agt) const {
  Point<2> p0=vertices[vertexIndex(0)];
  Point<2> p1=vertices[vertexIndex(1)];
  Point<2> p2=vertices[vertexIndex(2)];
  agt.movePen(p0[0],p0[1]);
  agt.drawLine(p1[0],p1[1]);
  agt.drawLine(p2[0],p2[1]);
  agt.drawLine(p0[0],p0[1]);
}
#endif
#if (SPACEDIM==3)
void Triangle::draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE &xc,
const NumPtr<Point<3> > &vertices,VolGraphTool &vgt) const {
  Point<3> p0=vertices[vertexIndex(0)];
  Point<3> p1=vertices[vertexIndex(1)];
  Point<3> p2=vertices[vertexIndex(2)];
  Vector3 v0(p0[0],p0[1],p0[2]);
  Vector3 v1(p1[0],p1[1],p1[2]);
  Vector3 v2(p2[0],p2[1],p2[2]);
  vgt.colorTriangle(xc,v2,v1,v0);
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Quadrilateral::Quadrilateral(int edi,const int l[4],const int v[4]) :
Shape2(edi,4,4) {
  CHECK_TEST(l[0]!=l[1] && l[0]!=l[2] && l[0]!=l[3] &&
             l[1]!=l[2] && l[1]!=l[3] && l[2]!=l[3]);
  CHECK_TEST(v[0]!=v[1] && v[0]!=v[2] && v[0]!=v[3] &&
             v[1]!=v[2] && v[1]!=v[3] && v[2]!=v[3]);
  int vmin=min(min(v[0],v[1]),min(v[2],v[3]));
#if (SPACEDIM==2)
//rotate to get vertex with lowest index first
  if (v[0]==vmin) {
    setCodimension1ShapeIndex(0,l[0]);
    setCodimension1ShapeIndex(1,l[1]);
    setCodimension1ShapeIndex(2,l[2]);
    setCodimension1ShapeIndex(3,l[3]);
    setVertexIndex(0,v[0]);
    setVertexIndex(1,v[1]);
    setVertexIndex(2,v[2]);
    setVertexIndex(3,v[3]);
  } else if (v[1]==vmin) {
    setCodimension1ShapeIndex(0,l[2]);
    setCodimension1ShapeIndex(1,l[3]);
    setCodimension1ShapeIndex(2,l[1]);
    setCodimension1ShapeIndex(3,l[0]);
    setVertexIndex(0,v[1]);
    setVertexIndex(1,v[3]);
    setVertexIndex(2,v[0]);
    setVertexIndex(3,v[2]);
  } else if (v[2]==vmin) {
    setCodimension1ShapeIndex(0,l[3]);
    setCodimension1ShapeIndex(1,l[2]);
    setCodimension1ShapeIndex(2,l[0]);
    setCodimension1ShapeIndex(3,l[1]);
    setVertexIndex(0,v[2]);
    setVertexIndex(1,v[0]);
    setVertexIndex(2,v[3]);
    setVertexIndex(3,v[1]);
  } else {
    setCodimension1ShapeIndex(0,l[1]);
    setCodimension1ShapeIndex(1,l[0]);
    setCodimension1ShapeIndex(2,l[3]);
    setCodimension1ShapeIndex(3,l[2]);
    setVertexIndex(0,v[3]);
    setVertexIndex(1,v[2]);
    setVertexIndex(2,v[1]);
    setVertexIndex(3,v[0]);
  }
#else
//rotate and/or flip to get vertices with lowest indices first
  if (v[0]==vmin) {
    if (v[1]<v[2]) {
      setCodimension1ShapeIndex(0,l[0]);
      setCodimension1ShapeIndex(1,l[1]);
      setCodimension1ShapeIndex(2,l[2]);
      setCodimension1ShapeIndex(3,l[3]);
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[2]);
      setVertexIndex(3,v[3]);
    } else {
      setCodimension1ShapeIndex(0,l[2]);
      setCodimension1ShapeIndex(1,l[3]);
      setCodimension1ShapeIndex(2,l[0]);
      setCodimension1ShapeIndex(3,l[1]);
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[1]);
      setVertexIndex(3,v[3]);
    }
  } else if (v[1]==vmin) {
    if (v[3]<v[0]) {
      setCodimension1ShapeIndex(0,l[2]);
      setCodimension1ShapeIndex(1,l[3]);
      setCodimension1ShapeIndex(2,l[1]);
      setCodimension1ShapeIndex(3,l[0]);
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[3]);
      setVertexIndex(2,v[0]);
      setVertexIndex(3,v[2]);
    } else {
      setCodimension1ShapeIndex(0,l[1]);
      setCodimension1ShapeIndex(1,l[0]);
      setCodimension1ShapeIndex(2,l[2]);
      setCodimension1ShapeIndex(3,l[3]);
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[3]);
      setVertexIndex(3,v[2]);
    }
  } else if (v[2]==vmin) {
    if (v[0]<v[3]) {
      setCodimension1ShapeIndex(0,l[3]);
      setCodimension1ShapeIndex(1,l[2]);
      setCodimension1ShapeIndex(2,l[0]);
      setCodimension1ShapeIndex(3,l[1]);
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[3]);
      setVertexIndex(3,v[1]);
    } else {
      setCodimension1ShapeIndex(0,l[0]);
      setCodimension1ShapeIndex(1,l[1]);
      setCodimension1ShapeIndex(2,l[3]);
      setCodimension1ShapeIndex(3,l[2]);
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[3]);
      setVertexIndex(2,v[0]);
      setVertexIndex(3,v[1]);
    }
  } else {
    if (v[2]<v[1]) {
      setCodimension1ShapeIndex(0,l[1]);
      setCodimension1ShapeIndex(1,l[0]);
      setCodimension1ShapeIndex(2,l[3]);
      setCodimension1ShapeIndex(3,l[2]);
      setVertexIndex(0,v[3]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[1]);
      setVertexIndex(3,v[0]);
    } else {
      setCodimension1ShapeIndex(0,l[3]);
      setCodimension1ShapeIndex(1,l[2]);
      setCodimension1ShapeIndex(2,l[1]);
      setCodimension1ShapeIndex(3,l[0]);
      setVertexIndex(0,v[3]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[2]);
      setVertexIndex(3,v[0]);
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Quadrilateral::Quadrilateral(const int l[4],const int v[4],int di) : 
Shape2(-1,4,4) {
  CHECK_TEST(v[0]!=v[1] && v[0]!=v[2] && v[0]!=v[3] &&
             v[1]!=v[2] && v[1]!=v[3] && v[2]!=v[3]);
  CHECK_TEST(l[0]!=l[1] && l[0]!=l[2] && l[0]!=l[3] &&
             l[1]!=l[2] && l[1]!=l[3] && l[2]!=l[3]);
  for (int i=0;i<4;i++) {
    setVertexIndex(i,v[i]);
    setCodimension1ShapeIndex(i,l[i]);
  }
  dof_index=di;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape1* Quadrilateral::newLineElement(int f) const {
  int v[2]={
    vertexIndex(staticElementVertexFromFace(f,0)),
    vertexIndex(staticElementVertexFromFace(f,1))
  };
  return OPERATOR_NEW Interval(v,-1); // second arg is dof_index
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Quadrilateral::contains(const Point<SPACEDIM> &x,
const NumPtr<Point<SPACEDIM> > &vertices) const {
#if (SPACEDIM==2)
  int l[3]={-1,-2,-3};
  int v[3]={vertexIndex(0),vertexIndex(1),vertexIndex(2)};
  Triangle t012(l,v,-1);
  if (t012.contains(x,vertices)) return true;

  {
    int v[3]={vertexIndex(3),vertexIndex(2),vertexIndex(1)};
    Triangle t321(l,v,-1);
    return t321.contains(x,vertices);
  }
#else
  return Shape::contains(x,vertices);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// 4 rotations of a square times 4 children = 16 cases
Point<SPACEDIM> Quadrilateral::parentCoordinateForCanonical(
const Point<SPACEDIM> &cp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
#if (SPACEDIM!=2)
  OBSOLETE("should not be calling this");
#else
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi>=0);

  switch (c) {
    case 0: {
      switch (cvi) {
        case 0:
          return cp*0.5;
        case 1:
          return Point<2>(cp[1]*0.5,(1.-cp[0])*0.5);
        case 2:
          return Point<2>((1.-cp[1])*0.5,cp[0]*0.5);
        case 3:
          OBSOLETE("not checked");
          return Point<2>((1.-cp[0])*0.5,(1.-cp[1])*0.5);
      }
    }
    case 1: {
      switch (cvi) {
        case 0:
          return Point<2>(1.-cp[1]*0.5,cp[0]*0.5);
        case 1:
          return Point<2>((1.+cp[0])*0.5,cp[1]*0.5);
        case 2:
          return Point<2>(1.-cp[0]*0.5,(1.-cp[1])*0.5);
        case 3:
          return Point<2>((1.+cp[1])*0.5,(1.-cp[0])*0.5);
      }
      break;
    }
    case 2: {
      switch (cvi) {
        case 0:
          return Point<2>(cp[1]*0.5,1.-cp[0]*0.5);
        case 1:
          return Point<2>((1.-cp[0])*0.5,1.-cp[1]*0.5);
        case 2:
          return Point<2>(cp[0]*0.5,(1.+cp[1])*0.5);
        case 3:
          return Point<2>((1.-cp[1])*0.5,(1.+cp[0])*0.5);
      }
      break;
    }
    case 3: {
      switch (cvi) {
        case 0:
          return Point<2>(1.-cp[0]*0.5,1.-cp[1]*0.5);
        case 1:
          return Point<2>(1.-cp[1]*0.5,(1.+cp[0])*0.5);
        case 2:
          return Point<2>((1.+cp[1])*0.5,1.-cp[0]*0.5);
        case 3:
          return Point<2>((1.+cp[0])*0.5,(1.+cp[1])*0.5);
      }
      break;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Quadrilateral::childCoordinateOfCanonical(
const Point<SPACEDIM> &pp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
#if (SPACEDIM!=2)
  OBSOLETE("should not be calling this");
#else
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi>=0);

  switch (c) {
    case 0: {
      switch (cvi) {
        case 0:
          return pp*2.;
        case 1:
          return Point<2>(1.-pp[1]*2.,pp[0]*2.);
        case 2:
          return Point<2>(pp[1]*2.,1.-pp[0]*2.);
        case 3:
          return Point<2>(1.-pp[0]*2.,1.-pp[1]*2.);
      }
    }
    case 1: {
      switch (cvi) {
        case 0:
          return Point<2>(pp[1]*2.,(1.-pp[0])*2.);
        case 1:
          return Point<2>(pp[0]*2.-1.,pp[1]*2.);
        case 2:
          return Point<2>((1.-pp[0])*2.,1.-pp[1]*2.);
        case 3:
          return Point<2>(1.-pp[1]*2.,pp[0]*2.-1.);
      }
      break;
    }
    case 2: {
      switch (cvi) {
        case 0:
          return Point<2>((1.-pp[1])*2.,pp[0]*2.);
        case 1:
          return Point<2>(1.-pp[0]*2.,(1.-pp[1])*2.);
        case 2:
          return Point<2>(pp[0]*2.,pp[1]*2.-1.);
        case 3:
          return Point<2>(pp[1]*2.-1.,1.-pp[0]*2.);
      }
      break;
    }
    case 3: {
      switch (cvi) {
        case 0:
          return Point<2>((1.-pp[0])*2.,(1.-pp[1])*2.);
        case 1:
          return Point<2>(pp[1]*2.-1.,(1.-pp[0])*2.);
        case 2:
          return Point<2>((1.-pp[1])*2.,pp[0]*2.-1.);
        case 3:
          return Point<2>(pp[0]*2.-1.,pp[1]*2.-1.);
      }
      break;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//see FE_Q::FE_Q; must be consistent with parentCoordinateOfCanonical
const Shape* Quadrilateral::reorientFor(const Shape *parent,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  CHECK_SAME(vertexIndex(0),parent->vertexIndex(c));
  return this;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::staticProjectToFace(int f,
const NumPtr<Point<1> > &q_points,NumPtr<Point<2> > &q_points_face) {
  switch (f) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=0.;
        q_points_face[i][1]=q_points[i][0];
      }
      break;
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=1.;
        q_points_face[i][1]=q_points[i][0];
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=0.;
      }
      break;
    case 3:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=1.;
      }
      break;
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::projectToFaceForBDFM(int f,
const NumPtr<Point<1> > &q_points,NumPtr<Point<2> > &q_points_face) 
const {
  switch (f) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=0.;
        q_points_face[i][1]=q_points[i][0];
      }
      break;
    case 1:
      if (vertexIndex(1)<vertexIndex(3)) {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][0]=1.;
          q_points_face[i][1]=q_points[i][0];
        }
      } else {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][0]=1.;
          q_points_face[i][1]=1.-q_points[i][0];
        }
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=0.;
      }
      break;
    case 3:
      if (vertexIndex(2)<vertexIndex(3)) {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][0]=q_points[i][0];
          q_points_face[i][1]=1.;
        }
      } else {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][0]=1.-q_points[i][0];
          q_points_face[i][1]=1.;
        }
      }
      break;
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::projectToSubface(int f,int s,
const NumPtr<Point<1> > &q_points,NumPtr<Point<2> > &q_points_subface) {
  switch (f) {
    case 0: {
      if (s==0) { // child 0
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=0.;
          qps[1]=qp;
        }
      } else { // child 2
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=0.;
          qps[1]=1.-qp;
        }
      }
      break;
    }
    case 1: {
      if (s==0) { // child 1
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=1.;
          qps[1]=qp;
        }
      } else { // child 3
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=1.;
          qps[1]=1.-qp;
        }
      }
      break;
    }
    case 2: {
      if (s==0) { // child 0
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=qp;
          qps[1]=0.;
        }
      } else { // child 1
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=1.-qp;
          qps[1]=0.;
        }
      }
      break;
    }
    case 3: {
      if (s==0) { // child 2
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=qp;
          qps[1]=1.;
        }
      } else { // child 3
        for (int q=0;q<q_points.getNumber();q++) {
          Point<2> &qps=q_points_subface[q];
          REAL qp=q_points[q][0]*0.5;
          qps[0]=1.-qp;
          qps[1]=1.;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void setQuadrilateralLineIndices(const int v[4],
const map<pair<int,int>,int> &interior_child_line_map,int l[4]) {
  int k=0;
  for (int j=0;j<2;j++) {
    pair<int,int> p(min(v[j],v[j+2]),max(v[j],v[j+2]));
    map<pair<int,int>,int>::const_iterator iclm_iterator=
      interior_child_line_map.find(p);
    CHECK_TEST(iclm_iterator!=interior_child_line_map.end());
    l[k++]=iclm_iterator->second;
  }
  for (int j=0;j<4;j+=2) {
    pair<int,int> p(min(v[j],v[j+1]),max(v[j],v[j+1]));
    map<pair<int,int>,int>::const_iterator iclm_iterator=
      interior_child_line_map.find(p);
    CHECK_TEST(iclm_iterator!=interior_child_line_map.end());
    l[k++]=iclm_iterator->second;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape2* Quadrilateral::newChildShape(int c,int child_element_index,
const NumPtr<int> &interior_child_vertex_indices,
const NumPtr<int> &interior_child_line_vertex_indices,
const map<pair<int,int>,int> &interior_child_line_map) const {
  CHECK_BOUNDS(c,0,4);
  CHECK_SAME(interior_child_vertex_indices.getNumber(),1);
  CHECK_SAME(interior_child_line_vertex_indices.getNumber(),4);
  int v[4];
  switch (c) {
    case 0:
      v[0]=vertex_indices[0];
      v[1]=interior_child_line_vertex_indices[2];
      v[2]=interior_child_line_vertex_indices[0];
      v[3]=interior_child_vertex_indices[0];
      break;
    case 1:
      v[0]=vertex_indices[1];
      v[1]=interior_child_line_vertex_indices[1];
      v[2]=interior_child_line_vertex_indices[2];
      v[3]=interior_child_vertex_indices[0];
      break;
    case 2:
      v[0]=vertex_indices[2];
      v[1]=interior_child_line_vertex_indices[0];
      v[2]=interior_child_line_vertex_indices[3];
      v[3]=interior_child_vertex_indices[0];
      break;
    case 3:
      v[0]=vertex_indices[3];
      v[1]=interior_child_line_vertex_indices[3];
      v[2]=interior_child_line_vertex_indices[1];
      v[3]=interior_child_vertex_indices[0];
  }
  int l[4];
  setQuadrilateralLineIndices(v,interior_child_line_map,l);
  return OPERATOR_NEW Quadrilateral(child_element_index,l,v);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::staticEquallySpacedLatticePoints(int n0,int n1,
NumPtr<Point<2> > &points) {
  points.allocate(staticNumberEquallySpacedLatticePoints(n0,n1));
  Point<2> p(0.,0.);
  REAL s0=(n0>0 ? 1./static_cast<REAL>(n0) : 0.);
  REAL s1=(n1>0 ? 1./static_cast<REAL>(n1) : 0.);
  int k=0;
  for (int j=0;j<=n1;j++,p[1]+=s1) {
    for (int i=0;i<=n0;i++,p[0]+=s0) points[k++]=p;
    p[0]=0.;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Quadrilateral::staticEquallySpacedLatticePointIndexOfVertex(int v,
int n0,int n1) {
  CHECK_TEST(min(n0,n1)>0);
  CHECK_TEST(v<4);
  switch (v) {
    case 0:
      return 0;
    case 1: 
      return n0;
    case 2:
      return n1*(n0+1);
    case 3:
      return n1*(n0+1)+n0;
    default:
      OBSOLETE("bad vertex");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::staticEquallySpacedLatticePointIndicesOfFace(int f,
int n0,int n1,NumPtr<int> &indices) {
  CHECK_TEST(min(n0,n1)>0);
  int nl=n0+1;
  switch (f) {
    case 0: { // (0,j)
      indices.allocate(n1+1);
      int ij=0;
      for (int j=0;j<=n1;j++) {
        indices[j]=ij;
        ij+=nl;
      }
      break;
    }
    case 1: { // (n,j)
      indices.allocate(n1+1);
      int ij=n0;
      for (int j=0;j<=n1;j++) {
        indices[j]=ij;
        ij+=nl;
      }
      break;
    }
    case 2: { // (i,0)
      indices.allocate(n0+1);
      for (int i=0;i<=n0;i++) indices[i]=i;
      break;
    }
    case 3: {
      indices.allocate(n0+1);
      int ij=n1*nl;
      for (int i=0;i<=n0;i++) indices[i]=ij++;
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::equallySpacedLatticePointIndicesOfFace(int f,int n0,
int n1,NumPtr<int> &indices) const {
  CHECK_TEST(min(n0,n1)>0);
  switch (f) {
    case 0: { // (0,j)
      indices.allocate(n1+1);
      int nl=n0+1;
      int ij=0;
      for (int j=0;j<=n1;j++) {
        indices[j]=ij;
        ij+=nl;
      }
      break;
    }
    case 1: { // (n,j)
      indices.allocate(n1+1);
      int nl=n0+1;
      if (vertexIndex(1)<vertexIndex(3)) {  // (n,j)
        int ij=n0;
        for (int j=0;j<=n1;j++) {
          indices[j]=ij;
          ij+=nl;
        }
      } else { // (n,n-j)
        int ij=(n1+1)*nl-1;
        for (int j=n1;j>=0;j--) {
          indices[j]=ij;
          ij-=nl;
        }
      }
      break;
    }
    case 2: { // (i,0)
      indices.allocate(n0+1);
      for (int i=0;i<=n0;i++) {
        indices[i]=i;
      }
      break;
    }
    case 3:
      indices.allocate(n0+1);
      if (vertexIndex(2)<vertexIndex(3)) { // (i,n)
        int ij=n1*(n0+1);
        for (int i=0;i<=n0;i++) {
          indices[i]=ij++;
        }
      } else { // (n-i,n)
        int ij=(n1+1)*(n0+1)-1;
        int m=0;
        for (int i=n0;i>=0;i--) {
          indices[m++]=ij--;
        }
      }
      break;
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::staticEquallySpacedLatticePointIndicesOfFaceInterior(
int f,int n0,int n1,NumPtr<int> &indices) {
  CHECK_TEST(min(n0,n1)>0);
  int nl=n0+1;
  switch (f) {
    case 0: { // (0,j)
      indices.allocate(n1-1);
      int ij=nl;
      int m=0;
      for (int j=1;j<n1;j++) {
        indices[m++]=ij;
        ij+=nl;
      }
      break;
    }
    case 1: { // (n,j)
      indices.allocate(n1-1);
      int ij=2*nl-1;
      int m=0;
      for (int j=1;j<n1;j++) {
        indices[m++]=ij;
        ij+=nl;
      }
      break;
    }
    case 2: { // (i,0)
      indices.allocate(n0-1);
      int m=0;
      for (int i=1;i<n0;i++) {
        indices[m++]=i;
      }
      break;
    }
    case 3: {
      indices.allocate(n0-1);
      int ij=n1*nl+1;
      int m=0;
      for (int i=1;i<n0;i++) {
        indices[m++]=ij;
        ij++;
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::equallySpacedLatticePointIndicesOfFaceInterior(
int f,int n0,int n1,NumPtr<int> &indices) const {
  CHECK_TEST(min(n0,n1)>0);
  int nl=n0+1;
  switch (f) {
    case 0: { // (0,j)
      indices.allocate(n1-1);
      int ij=nl;
      int m=0;
      for (int j=1;j<n1;j++) {
        indices[m++]=ij;
        ij+=nl;
      }
      break;
    }
    case 1: { // (n,j)
      indices.allocate(n1-1);
      if (vertexIndex(1)<vertexIndex(3)) {  // (n,j)
        int ij=2*nl-1;
        int m=0;
        for (int j=1;j<n1;j++) {
          indices[m++]=ij;
          ij+=nl;
        }
      } else { // (n,n-j)
        int ij=n1*nl-1;
        int m=0;
        for (int j=1;j<n1;j++) {
          indices[m++]=ij;
          ij-=nl;
        }
      }
      break;
    }
    case 2: { // (i,0)
      indices.allocate(n0-1);
      int m=0;
      for (int i=1;i<n0;i++) {
        indices[m++]=i;
      }
      break;
    }
    case 3:
      indices.allocate(n0-1);
      if (vertexIndex(2)<vertexIndex(3)) { // (i,n)
        int ij=n1*nl+1;
        int m=0;
        for (int i=1;i<n0;i++) {
          indices[m++]=ij;
          ij++;
        }
      } else { // (n-i,n)
        int ij=(n1+1)*nl-2;
        int m=0;
        for (int i=n0-1;i>=1;i--) {
          indices[m++]=ij;
          ij--;
        }
      }
      break;
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::reorderFaceDofs(int n,NumPtr<int> &renumber) const {
  int nl=n+1;
  for (int i=0;i<4*nl;i++) renumber[i]=i;
  int offset=0;
  NumPtr<int> r=renumber;
  for (int f=0;f<4;f++) {
    int v[2]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1))};
    Interval face_element(v,-1);
    face_element.reorderDofs(n,r);
    for (int i=0;i<=n;i++) r[i]+=offset;
    r+=nl;
    offset+=nl;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<4*nl;i++) {
    if (renumber[i]>=renumber.getNumber() || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(4*nl);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<4*nl;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<4*nl;i++) CHECK_BOUNDS(renumber_inverse[i],0,4*nl);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::reorderDofs(int n0,int n1,NumPtr<int> &renumber)
const {
  for (int i=0;i<staticNumberEquallySpacedLatticePoints(n0,n1);i++) {
    renumber[i]=i;
  }
  int vmin=min(min(vertexIndex(0),vertexIndex(1)),
               min(vertexIndex(2),vertexIndex(3)));
  int nl=n0+1;
  if (vertexIndex(0)==vmin) {
    if (vertexIndex(2)<vertexIndex(1)) {
      int m=0;
      int ki=0;
      for (int i=0;i<=n0;i++) {
        int k=ki;
        for (int j=0;j<=n1;j++) {
          renumber[k]=m++;
          k+=nl;
        }
        ki++;
      }
    }
  } else if (vertexIndex(1)==vmin) {
    if (vertexIndex(0)<vertexIndex(3)) {
      int m=0;
      int kj=n0;
      for (int j=0;j<=n1;j++) {
        int k=kj;
        for (int i=n0;i>=0;i--) {
          renumber[k]=m++;
          k--;
        }
        kj+=nl;
      }
    } else {
      int m=0;
      int ki=n0;
      for (int i=n0;i>=0;i--) {
        int k=ki;
        for (int j=0;j<=n1;j++) {
          renumber[k]=m++;
          k+=nl;
        }
        ki--;
      }
    }
  } else if (vertexIndex(2)==vmin) {
    if (vertexIndex(0)<vertexIndex(3)) {
      int m=0;
      int ki=n1*nl;
      for (int i=0;i<=n0;i++) {
        int k=ki;
        for (int j=n1;j>=0;j--) {
          renumber[k]=m++;
          k-=nl;
        }
        ki++;
      }
    } else {
      int m=0;
      int kj=n1*nl;
      for (int j=n1;j>=0;j--) {
        int k=kj;
        for (int i=0;i<=n0;i++) {
          renumber[k]=m++;
          k++;
        }
        kj-=nl;
      }
    }
  } else {
    if (vertexIndex(1)<vertexIndex(2)) {
      int m=0;
      int ki=n1*nl+n0;
      for (int i=n0;i>=0;i--) {
        int k=ki;
        for (int j=n1;j>=0;j--) {
          renumber[k]=m++;
          k-=nl;
        }
        ki--;
      }
    } else {
      int m=0;
      int kj=n1*nl+n0;
      for (int j=n1;j>=0;j--) {
        int k=kj;
        for (int i=n0;i>=0;i--) {
          renumber[k]=m++;
          k--;
        }
        kj-=nl;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::reorderEquallySpacedLatticePoints(int n0,int n1,
NumPtr<int> &renumber) const {
  renumber.allocate(staticNumberEquallySpacedLatticePoints(n0,n1));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
//vertices first
  int m=0;
  renumber[0]=m++;
  renumber[n0]=m++;
  renumber[n1*(n0+1)]=m++;
  renumber[renumber.getNumber()-1]=m++;
//face 0
  if (vertexIndex(0)<vertexIndex(2)) {
    int l=n0+1;
    for (int j=1;j<n1;j++) {
      renumber[l]=m++;
      l+=n0+1;
    }
  } else {
    int l=(n1-1)*(n0+1);
    for (int j=n1-1;j>0;j--) {
      renumber[l]=m++;
      l-=n0+1;
    }
  }
//face 1
  if (vertexIndex(1)<vertexIndex(3)) {
    int l=2*n0+1;
    for (int j=1;j<n1;j++) {
      renumber[l]=m++;
      l+=n0+1;
    }
  } else {
    int l=n1*(n0+1)-1;
    for (int j=n1-1;j>0;j--) {
      renumber[l]=m++;
      l-=n0+1;
    }
  }
//face 2
  if (vertexIndex(0)<vertexIndex(1)) {
    for (int i=1;i<n0;i++) renumber[i]=m++;
  } else {
    int l=n0-1;
    for (int i=n0-1;i>0;i--) renumber[l--]=m++;
  } 
//face 3
  if (vertexIndex(2)<vertexIndex(3)) {
    int l=n1*(n0+1)+1;
    for (int i=1;i<n0;i++) renumber[l++]=m++;
  } else {
    int l=renumber.getNumber()-2;
    for (int i=n0-1;i>0;i--) renumber[l--]=m++;
  } 
//interior
  int l=n0+2;
  for (int j=1;j<n1;j++) {
    for (int i=1;i<n0;i++) {
      renumber[l++]=m++;
    }
    l+=2;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]==INT_MAX) {
      cout << "\trenumber[" << i << "] = *** " << -1 << " ***" << endl;
      success=false;
    } else if (renumber[i]>=renumber.getNumber()) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::reorderChildEquallySpacedLatticePoints(int n,
const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices,NumPtr<int> &renumber) 
const {
  const Quadrilateral *child_quadrilateral=
    dynamic_cast<const Quadrilateral*>(child_shape);
  CHECK_POINTER(child_quadrilateral);
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi>=0);
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  int reorder_case=cvi;
  int nl=n+1;
  switch (reorder_case) {
    case 0: {
      for (int i=0;i<renumber.getNumber();i++) renumber[i]=i;
      break;
    }
    case 1: {
      int ij=0;
      int ni=n;
      for (int i=n;i>=0;i--) {
        int nj=ni;
        for (int j=0;j<=n;j++) {
          renumber[nj]=ij++;
          nj+=nl;
        }
        ni--;
      }
      break;
    }
    case 2: {
      int ij=0;
      int ni=n*nl;
      for (int i=0;i<=n;i++) {
        int nj=ni;
        for (int j=n;j>=0;j--) {
          renumber[nj]=ij++;
          nj-=nl;
        }
        ni++;
      }
      break;
    }
    case 3: {
      int ij=0;
      int nj=nl*nl-1;
      for (int j=n;j>=0;j--) {
        int ni=nj;
        for (int i=n;i>=0;i--) {
          renumber[ni]=ij++;
          ni--;
        }
        nj-=nl;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_H::constraints
void Quadrilateral::reorderChildDofPoints(int n,const Shape *child_shape,
int c,const NumPtr<int> &line_interior_child_vertices,
NumPtr<int> &renumber) const {
  const Quadrilateral *child_quadrilateral=
    dynamic_cast<const Quadrilateral*>(child_shape);
  CHECK_POINTER(child_quadrilateral);

  int reorder_case=-1;
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi>=0);
  reorder_case=cvi;

  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  switch (reorder_case) {
    case 0: {
      for (int i=0;i<renumber.getNumber();i++) renumber[i]=i;
      break;
    }
    case 1: {
      int ij=0;
      renumber[ij++]=1;
      renumber[ij++]=3;
      renumber[ij++]=0;
      renumber[ij++]=2;
      int d=3*n;
      for (int i=n-1;i>=1;i--) {
        renumber[ij++]=d--;
      }
      d=4*n-1;
      for (int i=n-1;i>=1;i--) {
        renumber[ij++]=d--;
      }
      d=n+3;
      for (int j=1;j<n;j++) {
        renumber[ij++]=d++;
      }
      d=4;
      for (int j=1;j<n;j++) {
        renumber[ij++]=d++;
      }
      int di=5*n-2;
      for (int i=n-1;i>=1;i--) {
        int dj=di;
        for (int j=1;j<n;j++) {
          renumber[ij++]=dj;
          dj+=n-1;
        }
        di--;
      }
      break;
    }
    case 2: {
      int ij=0;
      renumber[ij++]=2;
      renumber[ij++]=0;
      renumber[ij++]=3;
      renumber[ij++]=1;
      int d=3*n+1;
      for (int i=1;i<n;i++) {
        renumber[ij++]=d++; 
      }
      d=2*n+2;
      for (int i=1;i<n;i++) {
        renumber[ij++]=d++; 
      }
      d=n+2;
      for (int j=n-1;j>=1;j--) {
        renumber[ij++]=d--;
      }
      d=2*n+1;
      for (int j=n-1;j>=1;j--) {
        renumber[ij++]=d--;
      }
      int di=n*(n+1)+2;
      for (int i=1;i<n;i++) {
        int dj=di;
        for (int j=n-1;j>=1;j--) {
          renumber[ij++]=dj;
          dj-=n-1;
        }
        di++;
      }
      break;
    }
    case 3: {
      int ij=0;
      renumber[ij++]=3;
      renumber[ij++]=2;
      renumber[ij++]=1;
      renumber[ij++]=0;
      int d=n+3;
      for (int j=n-1;j>=1;j--) {
        renumber[ij++]=d--; 
      }
      d=n+2;
      for (int j=n-1;j>=1;j--) {
        renumber[ij++]=d--; 
      }
      d=4*n-1;
      for (int i=n-1;i>=1;i--) {
        renumber[ij++]=d--;
      }
      d=3*n;
      for (int i=n-1;i>=1;i--) {
        renumber[ij++]=d--;
      }
      int dj=n*(n+2);
      for (int j=n-1;j>=1;j--) {
        int di=dj;
        for (int i=n-1;i>=1;i--) {
          renumber[ij++]=di;
          di--;
        }
        dj-=n-1;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::staticInteriorLatticePointIndices(int n0,int n1,
NumPtr<int> &v) {
  v.allocate(staticNumberInteriorEquallySpacedLatticePoints(n0,n1));
  if (max(n0,n1)<=1) return;
  int ijk=n0+2;
  int iv=0;
  for (int j=1;j<n1;j++) {
    for (int i=1;i<n0;i++) {
      v[iv]=ijk;
      iv++;
      ijk++;
    }
    ijk+=2;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::computeInteriorLatticePoints(int d0,int d1,int bf,
NumPtr<Point<SPACEDIM> > &mapping_support_points) const {
  if (max(d0,d1)<=1) return;
  CHECK_SAME(mapping_support_points.getNumber(),
             staticNumberEquallySpacedLatticePoints(d0,d1));
  REAL dxi0=1./static_cast<REAL>(d0);
  REAL dxi1=1./static_cast<REAL>(d1);
  const Point<SPACEDIM> &v0=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(0,d0,d1)];
  const Point<SPACEDIM> &v1=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(1,d0,d1)];
  const Point<SPACEDIM> &v2=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(2,d0,d1)];
  const Point<SPACEDIM> &v3=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(3,d0,d1)];
  if (bf<0) { // no face on boundary
    Point<SPACEDIM> dv20=v2-v0;
    Point<SPACEDIM> dv31=v3-v1;
    int k=d0+2;
    REAL xi1=dxi1;
    for (int j=1;j<d1;j++,k+=2,xi1+=dxi1) {
      REAL xi0=dxi0;
      Point<SPACEDIM> v02=v0+dv20*xi1;
      Point<SPACEDIM> v13=v1+dv31*xi1;
      for (int i=1;i<d0;i++,k++,xi0+=dxi0) {
        mapping_support_points[k]=v02*(1.-xi0)+v13*xi0;
      }
    }
  } else { // one or more faces on boundary
    NumPtr<int> indices0;
    NumPtr<int> indices1;
    NumPtr<int> indices2;
    NumPtr<int> indices3;
    staticEquallySpacedLatticePointIndicesOfFace(0,d0,d1,indices0);
    staticEquallySpacedLatticePointIndicesOfFace(1,d0,d1,indices1);
    staticEquallySpacedLatticePointIndicesOfFace(2,d0,d1,indices2);
    staticEquallySpacedLatticePointIndicesOfFace(3,d0,d1,indices3);
    int m=d0+2;
    REAL xi1=dxi1;
    for (int j=1;j<d1;j++,xi1+=dxi1) {
      REAL onemxi1=1.-xi1;
      Point<SPACEDIM> v02=
        mapping_support_points[indices0[j]]-v0-(v2-v0)*xi1;
      Point<SPACEDIM> v13=
        mapping_support_points[indices1[j]]-v1-(v3-v1)*xi1;
      REAL xi0=dxi0;
      for (int i=1;i<d0;i++,m++,xi0+=dxi0) {
        mapping_support_points[m]=v02+(v13-v02)*xi0
          +mapping_support_points[indices2[i]]*onemxi1
          +mapping_support_points[indices3[i]]*xi1;
      }
      m+=2;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Quadrilateral::faceCenter(
const NumPtr<Point<SPACEDIM> > &vertices,int f) const {
  switch (f) {
    case 0:
      return (vertices[vertexIndex(0)]+vertices[vertexIndex(2)])*0.5;
    case 1:
      return (vertices[vertexIndex(1)]+vertices[vertexIndex(3)])*0.5;
    case 2:
      return (vertices[vertexIndex(0)]+vertices[vertexIndex(1)])*0.5;
    case 3:
      return (vertices[vertexIndex(2)]+vertices[vertexIndex(3)])*0.5;
    default:
      OBSOLETE("bad face");
      return Point<SPACEDIM>(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Quadrilateral::diameter(const NumPtr<Point<SPACEDIM> > &vertices) 
const {
  return sqrt(max((vertices[vertexIndex(3)]
                  -vertices[vertexIndex(0)]).square(),
                  (vertices[vertexIndex(2)]
                  -vertices[vertexIndex(1)]).square()));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Quadrilateral::volume(const NumPtr<Point<SPACEDIM> > &vertices) 
const {
  Tensor<1,SPACEDIM> v10(vertices[vertexIndex(1)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v30(vertices[vertexIndex(3)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v20(vertices[vertexIndex(2)]
                        -vertices[vertexIndex(0)]);
  REAL vol13,vol23;
#if (SPACEDIM==2)
  Tensor<1,SPACEDIM> v30_perp;
  crossProduct(v30_perp,v30);
  product(vol13,v10,v30_perp);
  product(vol23,v20,v30_perp);
  return (vol13>0. && vol23<0. ? (vol13-vol23)*0.5 
                               : -(abs(vol13)+abs(vol23))*0.5);
#else
//may not be meaningful: vertices not necessarily co-planar in 3D
  Tensor<1,SPACEDIM> cp;
  crossProduct(cp,v10,v30);
  product(vol13,cp,cp);
  vol13=sqrt(vol13);
  crossProduct(cp,v20,v30);
  product(vol23,cp,cp);
  vol23=sqrt(vol23);
  return (vol13+vol23)*0.5;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Quadrilateral::lineTangentPointsBackward(int l) const {
  switch (l) {
    case 0:
      return false;
    case 1:
      return vertexIndex(3)<vertexIndex(1);
    case 2:
      return false;
    case 3:
      return vertexIndex(3)<vertexIndex(2);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::initializeMappingSupportPoints(int degree,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  NumPtr<Point<2> > eslp;
  staticEquallySpacedLatticePoints(degree,eslp);
  int neslp=eslp.getNumber();
  NumPtr<Point<SPACEDIM> > qv(4);
  for (int v=0;v<4;v++) {
    qv[v]=tessellation_vertices[vertexIndex(v)];
  }
  for (int i=0;i<neslp;i++) {
    const Point<2> &eslpi=eslp[i];
    points[i]=(qv[1]*eslpi[0]+qv[0]*(1.-eslpi[0]))*(1.-eslpi[1])
             +(qv[3]*eslpi[0]+qv[2]*(1.-eslpi[0]))*eslpi[1];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::getIntermediatePoints(
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  int n=points.getNumber();
  int m=static_cast<int>(nearbyint(sqrt(static_cast<REAL>(n))));
  CHECK_SAME(m*m,n);
  Point<SPACEDIM> px=tessellation_vertices[vertexIndex(0)];
  REAL rm=1./static_cast<REAL>(m+1);
  Point<SPACEDIM> dpx=tessellation_vertices[vertexIndex(1)]-px;
  Point<SPACEDIM> dpy=tessellation_vertices[vertexIndex(2)]-px;
  Point<SPACEDIM> dpxy=((tessellation_vertices[vertexIndex(3)]-px)
                       -dpx-dpy)*rm*rm;
  dpx*=rm;
  dpy*=rm;
  dpx+=dpxy;
  px+=dpx+dpy;
  int k=0;
  for (int j=0;j<m;j++,px+=dpy,dpx+=dpxy) {
    Point<SPACEDIM> p=px;
    for (int i=0;i<m;i++,k++,p+=dpx) points[k]=p;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Quadrilateral::getMappedPoints(
const NumPtr<Point<3> > &tessellation_vertices,
const NumPtr<Point<2> > &canonical_points,NumPtr<Point<3> > &points,
NumPtr<Point<3> > &dpointsdxi0,NumPtr<Point<3> > &dpointsdxi1) const {
  int n=canonical_points.getNumber();
  points.allocate(n);
  dpointsdxi0.allocate(n);
  dpointsdxi1.allocate(n);

  Point<3> p=tessellation_vertices[vertexIndex(0)];
  Point<3> dpx=tessellation_vertices[vertexIndex(1)]-p;
  Point<3> dpy=tessellation_vertices[vertexIndex(2)]-p;
  Point<3> dpxy=(tessellation_vertices[vertexIndex(3)]-p)-dpx-dpy;
  for (int i=0;i<n;i++) {
    const Point<2> &cp=canonical_points[i];
    dpointsdxi0[i]=dpx+dpxy*cp[1];
    dpointsdxi1[i]=dpy+dpxy*cp[0];
    points[i]=p+dpointsdxi0[i]*cp[0]+dpy*cp[1];
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Quadrilateral::printOn(ostream &os) const {
  os << "Quadrilateral:" << endl;
  Shape2::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Quadrilateral::checkFaceVertices(int f,const Shape1 *face_element) 
const {
  CHECK_BOUNDS(f,0,4);
  if (face_element->vertexIndex(0)==
  vertexIndex(canonical_vertex_indices[f][0])) {
    CHECK_SAME(face_element->vertexIndex(1),
               vertexIndex(canonical_vertex_indices[f][1]));
  } else {
    CHECK_SAME(face_element->vertexIndex(1),
               vertexIndex(canonical_vertex_indices[f][0]));
    CHECK_SAME(face_element->vertexIndex(0),
               vertexIndex(canonical_vertex_indices[f][1]));
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==2)
void Quadrilateral::draw(const NumPtr<Point<2> > &vertices,
AreaGraphTool &agt) const {
  Point<2> p0=vertices[vertexIndex(0)];
  Point<2> p1=vertices[vertexIndex(1)];
  Point<2> p2=vertices[vertexIndex(2)];
  Point<2> p3=vertices[vertexIndex(3)];
  agt.movePen(p0[0],p0[1]);
  agt.drawLine(p1[0],p1[1]);
  agt.drawLine(p3[0],p3[1]);
  agt.drawLine(p2[0],p2[1]);
  agt.drawLine(p0[0],p0[1]);
}
#endif
#if (SPACEDIM==3)
void Quadrilateral::draw(
const VolGraphTool::WINDOW_TYPE::COLOR_TYPE &xc,
const NumPtr<Point<3> > &vertices,VolGraphTool &vgt) const {
  Point<3> p0=vertices[vertexIndex(0)];
  Point<3> p1=vertices[vertexIndex(1)];
  Point<3> p2=vertices[vertexIndex(2)];
  Point<3> p3=vertices[vertexIndex(3)];
  Vector3 v0(p0[0],p0[1],p0[2]);
  Vector3 v1(p1[0],p1[1],p1[2]);
  Vector3 v2(p2[0],p2[1],p2[2]);
  Vector3 v3(p3[0],p3[1],p3[2]);
  vgt.colorTriangle(xc,v0,v1,v3);
  vgt.colorTriangle(xc,v3,v2,v0);
}
#endif
#endif
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Tetrahedron::Tetrahedron(int edi,const int l[6],const int f[4],
const int v[4]) : Shape3(edi,6,4,4) {
  set<int> vertex_index_set;
  for (int i=0;i<4;i++) vertex_index_set.insert(v[i]);
#ifdef DEBUG
  CHECK_SAME(vertex_index_set.size(),4);
  set<int> face_index_set;
  for (int i=0;i<4;i++) face_index_set.insert(f[i]);
  CHECK_SAME(face_index_set.size(),4);
  set<int> line_index_set;
  for (int i=0;i<6;i++) line_index_set.insert(l[i]);
  CHECK_SAME(line_index_set.size(),6);
#endif
  set<int>::const_iterator vis_iterator=vertex_index_set.begin();
  int vmin=*vis_iterator;
  vis_iterator++;
  int vmin1=*vis_iterator;
  if (vmin==v[0]) {
    if (vmin1==v[1]) {
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[2]);
      setVertexIndex(3,v[3]);
      setCodimension1ShapeIndex(0,f[0]);
      setCodimension1ShapeIndex(1,f[1]);
      setCodimension1ShapeIndex(2,f[2]);
      setCodimension1ShapeIndex(3,f[3]);
      setCodimension2ShapeIndex(0,l[0]);
      setCodimension2ShapeIndex(1,l[1]);
      setCodimension2ShapeIndex(2,l[2]);
      setCodimension2ShapeIndex(3,l[3]);
      setCodimension2ShapeIndex(4,l[4]);
      setCodimension2ShapeIndex(5,l[5]);
    } else if (vmin1==v[2]) {
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[3]);
      setVertexIndex(3,v[1]);
      setCodimension1ShapeIndex(0,f[1]);
      setCodimension1ShapeIndex(1,f[2]);
      setCodimension1ShapeIndex(2,f[0]);
      setCodimension1ShapeIndex(3,f[3]);
      setCodimension2ShapeIndex(0,l[2]);
      setCodimension2ShapeIndex(1,l[5]);
      setCodimension2ShapeIndex(2,l[3]);
      setCodimension2ShapeIndex(3,l[0]);
      setCodimension2ShapeIndex(4,l[1]);
      setCodimension2ShapeIndex(5,l[4]);
    } else {
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[3]);
      setVertexIndex(2,v[1]);
      setVertexIndex(3,v[2]);
      setCodimension1ShapeIndex(0,f[2]);
      setCodimension1ShapeIndex(1,f[0]);
      setCodimension1ShapeIndex(2,f[1]);
      setCodimension1ShapeIndex(3,f[3]);
      setCodimension2ShapeIndex(0,l[3]);
      setCodimension2ShapeIndex(1,l[4]);
      setCodimension2ShapeIndex(2,l[0]);
      setCodimension2ShapeIndex(3,l[2]);
      setCodimension2ShapeIndex(4,l[5]);
      setCodimension2ShapeIndex(5,l[1]);
    }
  } else if (vmin==v[1]) {
    if (vmin1==v[0]) {
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[3]);
      setVertexIndex(3,v[2]);
      setCodimension1ShapeIndex(0,f[3]);
      setCodimension1ShapeIndex(1,f[2]);
      setCodimension1ShapeIndex(2,f[1]);
      setCodimension1ShapeIndex(3,f[0]);
      setCodimension2ShapeIndex(0,l[0]);
      setCodimension2ShapeIndex(1,l[3]);
      setCodimension2ShapeIndex(2,l[4]);
      setCodimension2ShapeIndex(3,l[1]);
      setCodimension2ShapeIndex(4,l[2]);
      setCodimension2ShapeIndex(5,l[5]);
    } else if (vmin1==v[2]) {
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[0]);
      setVertexIndex(3,v[3]);
      setCodimension1ShapeIndex(0,f[1]);
      setCodimension1ShapeIndex(1,f[3]);
      setCodimension1ShapeIndex(2,f[2]);
      setCodimension1ShapeIndex(3,f[0]);
      setCodimension2ShapeIndex(0,l[1]);
      setCodimension2ShapeIndex(1,l[2]);
      setCodimension2ShapeIndex(2,l[0]);
      setCodimension2ShapeIndex(3,l[4]);
      setCodimension2ShapeIndex(4,l[5]);
      setCodimension2ShapeIndex(5,l[3]);
    } else {
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[3]);
      setVertexIndex(2,v[2]);
      setVertexIndex(3,v[0]);
      setCodimension1ShapeIndex(0,f[2]);
      setCodimension1ShapeIndex(1,f[1]);
      setCodimension1ShapeIndex(2,f[3]);
      setCodimension1ShapeIndex(3,f[0]);
      setCodimension2ShapeIndex(0,l[4]);
      setCodimension2ShapeIndex(1,l[5]);
      setCodimension2ShapeIndex(2,l[1]);
      setCodimension2ShapeIndex(3,l[0]);
      setCodimension2ShapeIndex(4,l[3]);
      setCodimension2ShapeIndex(5,l[2]);
    }
  } else if (vmin==v[2]) {
    if (vmin1==v[0]) {
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[1]);
      setVertexIndex(3,v[3]);
      setCodimension1ShapeIndex(0,f[3]);
      setCodimension1ShapeIndex(1,f[0]);
      setCodimension1ShapeIndex(2,f[2]);
      setCodimension1ShapeIndex(3,f[1]);
      setCodimension2ShapeIndex(0,l[2]);
      setCodimension2ShapeIndex(1,l[0]);
      setCodimension2ShapeIndex(2,l[1]);
      setCodimension2ShapeIndex(3,l[5]);
      setCodimension2ShapeIndex(4,l[3]);
      setCodimension2ShapeIndex(5,l[4]);
    } else if (vmin1==v[1]) {
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[3]);
      setVertexIndex(3,v[0]);
      setCodimension1ShapeIndex(0,f[0]);
      setCodimension1ShapeIndex(1,f[2]);
      setCodimension1ShapeIndex(2,f[3]);
      setCodimension1ShapeIndex(3,f[1]);
      setCodimension2ShapeIndex(0,l[1]);
      setCodimension2ShapeIndex(1,l[4]);
      setCodimension2ShapeIndex(2,l[5]);
      setCodimension2ShapeIndex(3,l[2]);
      setCodimension2ShapeIndex(4,l[0]);
      setCodimension2ShapeIndex(5,l[3]);
    } else {
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[3]);
      setVertexIndex(2,v[0]);
      setVertexIndex(3,v[1]);
      setCodimension1ShapeIndex(0,f[2]);
      setCodimension1ShapeIndex(1,f[3]);
      setCodimension1ShapeIndex(2,f[0]);
      setCodimension1ShapeIndex(3,f[1]);
      setCodimension2ShapeIndex(0,l[5]);
      setCodimension2ShapeIndex(1,l[3]);
      setCodimension2ShapeIndex(2,l[2]);
      setCodimension2ShapeIndex(3,l[1]);
      setCodimension2ShapeIndex(4,l[4]);
      setCodimension2ShapeIndex(5,l[0]);
    }
  } else {
    if (vmin1==v[0]) {
      setVertexIndex(0,v[3]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[2]);
      setVertexIndex(3,v[1]);
      setCodimension1ShapeIndex(0,f[3]);
      setCodimension1ShapeIndex(1,f[1]);
      setCodimension1ShapeIndex(2,f[0]);
      setCodimension1ShapeIndex(3,f[2]);
      setCodimension2ShapeIndex(0,l[3]);
      setCodimension2ShapeIndex(1,l[2]);
      setCodimension2ShapeIndex(2,l[5]);
      setCodimension2ShapeIndex(3,l[4]);
      setCodimension2ShapeIndex(4,l[0]);
      setCodimension2ShapeIndex(5,l[1]);
    } else if (vmin1==v[1]) {
      setVertexIndex(0,v[3]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[0]);
      setVertexIndex(3,v[2]);
      setCodimension1ShapeIndex(0,f[0]);
      setCodimension1ShapeIndex(1,f[3]);
      setCodimension1ShapeIndex(2,f[1]);
      setCodimension1ShapeIndex(3,f[2]);
      setCodimension2ShapeIndex(0,l[4]);
      setCodimension2ShapeIndex(1,l[0]);
      setCodimension2ShapeIndex(2,l[3]);
      setCodimension2ShapeIndex(3,l[5]);
      setCodimension2ShapeIndex(4,l[1]);
      setCodimension2ShapeIndex(5,l[2]);
    } else {
      setVertexIndex(0,v[3]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[1]);
      setVertexIndex(3,v[0]);
      setCodimension1ShapeIndex(0,f[1]);
      setCodimension1ShapeIndex(1,f[0]);
      setCodimension1ShapeIndex(2,f[3]);
      setCodimension1ShapeIndex(3,f[2]);
      setCodimension2ShapeIndex(0,l[5]);
      setCodimension2ShapeIndex(1,l[1]);
      setCodimension2ShapeIndex(2,l[4]);
      setCodimension2ShapeIndex(3,l[3]);
      setCodimension2ShapeIndex(4,l[2]);
      setCodimension2ShapeIndex(5,l[0]);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Tetrahedron::Tetrahedron(const int l[6],const int f[4],const int v[4],
int di) : Shape3(-1,6,4,4) {
#ifdef DEBUG
  set<int> vertex_index_set;
  for (int i=0;i<4;i++) vertex_index_set.insert(v[i]);
  CHECK_SAME(vertex_index_set.size(),4);
  set<int> face_index_set;
  for (int i=0;i<4;i++) face_index_set.insert(f[i]);
  CHECK_SAME(face_index_set.size(),4);
  set<int> line_index_set;
  for (int i=0;i<6;i++) line_index_set.insert(l[i]);
  CHECK_SAME(line_index_set.size(),6);
#endif
  for (int i=0;i<4;i++) setVertexIndex(i,v[i]);
  for (int i=0;i<4;i++) setCodimension1ShapeIndex(i,f[i]);
  for (int i=0;i<6;i++) setCodimension2ShapeIndex(i,l[i]);
  dof_index=di;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape2* Tetrahedron::newFaceElement(int f) const {
  int l[3]={
    codimension2ShapeIndex(staticElementLineFromFace(f,0)),
    codimension2ShapeIndex(staticElementLineFromFace(f,1)),
    codimension2ShapeIndex(staticElementLineFromFace(f,2))
  };
  int v[3]={
    vertexIndex(staticElementVertexFromFace(f,0)),
    vertexIndex(staticElementVertexFromFace(f,1)),
    vertexIndex(staticElementVertexFromFace(f,2))
  };
  return OPERATOR_NEW Triangle(l,v,-1); // third arg is dof_index
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape1* Tetrahedron::newLineElement(int l) const {
  int v[2]={
    vertexIndex(staticElementVertexFromLine(l,0)),
    vertexIndex(staticElementVertexFromLine(l,1))
  };
  return OPERATOR_NEW Interval(v,-1); // second arg is dof_index
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Tetrahedron::contains(const Point<3> &x,
const NumPtr<Point<3> > &vertices) const {
  const Point<3> &v0(vertices[vertexIndex(0)]);
  const Point<3> &v1(vertices[vertexIndex(1)]);
  const Point<3> &v2(vertices[vertexIndex(2)]);
  const Point<3> &v3(vertices[vertexIndex(3)]);
  Tensor<1,3> perp;
  crossProduct(perp,Tensor<1,3>(v2-v0),Tensor<1,3>(v3-v0));
  if (perp.getValues()*(x-v0)<0.) return false;

  crossProduct(perp,Tensor<1,3>(v3-v0),Tensor<1,3>(v1-v0));
  if (perp.getValues()*(x-v0)<0.) return false;

  crossProduct(perp,Tensor<1,3>(v1-v0),Tensor<1,3>(v2-v0));
  if (perp.getValues()*(x-v0)<0.) return false;

  crossProduct(perp,Tensor<1,3>(v3-v1),Tensor<1,3>(v2-v1));
  return perp.getValues()*(x-v1)>=0.;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_H::prolongationMatrix
// 12 rotations of a tetrahedron times 8 children = 96 cases
Point<SPACEDIM> Tetrahedron::parentCoordinateForCanonical(
const Point<SPACEDIM> &cp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  const Tetrahedron *child_tetrahedron=
    dynamic_cast<const Tetrahedron*>(child_shape);
  CHECK_POINTER(child_tetrahedron);
  int cvi0=-1;
  int cvi1=-1;
  if (c<4) {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    CHECK_TEST(vertexIndex(c)==svi);
    cvi0=child_shape->vertexEqualToVertexIndex(svi);
    cvi1=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[c]);
  } else {
    cvi0=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[0]);
    switch (c) {
      case 4: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[3]);
        break;
      }
      case 5:
      case 6: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[1]);
        break;
      }
      case 7: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[2]);
        break;
      }
    }
  }
  CHECK_TEST(cvi0>=0);
  CHECK_TEST(cvi1>=0);
  
  switch (c) {
    case 0: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return cp*0.5;
            }
            case 2: { 
              return Point<3>(cp[1]*0.5,cp[2]*0.5,cp[0]*0.5);
            }
            case 3: {
              return Point<3>(cp[2]*0.5,cp[0]*0.5,cp[1]*0.5);
            }
          }
        }
        case 1:
        case 2:
        case 3:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 1: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return 
                Point<3>(1.-(cp[0]+cp[1]+cp[2])*0.5,cp[0]*0.5,cp[2]*0.5);
            }
            case 2: {
              return 
                Point<3>(1.-(cp[0]+cp[1]+cp[2])*0.5,cp[1]*0.5,cp[0]*0.5);
            }
            case 3: {
              return 
                Point<3>(1.-(cp[0]+cp[1]+cp[2])*0.5,cp[2]*0.5,cp[1]*0.5);
            }
          }
        }
        case 1:
        case 2:
        case 3:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 2: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return 
                Point<3>(cp[1]*0.5,1.-(cp[0]+cp[1]+cp[2])*0.5,cp[2]*0.5);
            }
            case 2: { 
              return 
                Point<3>(cp[2]*0.5,1.-(cp[0]+cp[1]+cp[2])*0.5,cp[0]*0.5);
            }
            case 3: {
              return 
                Point<3>(cp[0]*0.5,1.-(cp[0]+cp[1]+cp[2])*0.5,cp[1]*0.5);
            }
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[2]*0.5,(1.+cp[0])*0.5,cp[1]*0.5);
            }
            case 2: { 
              return Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,(1.+cp[0])*0.5,
                              cp[2]*0.5);
            }
            case 3: {
              return Point<3>(cp[1]*0.5,(1.+cp[0])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5);
            }
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[0]*0.5,(1.+cp[1])*0.5,cp[2]*0.5);
            }
            case 1: { 
              return Point<3>(cp[2]*0.5,(1.+cp[1])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5);
            }
            case 3: {
              return Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,(1.+cp[1])*0.5,
                              cp[0]*0.5);
            }
          }
        }
        case 3: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[1]*0.5,(1.+cp[2])*0.5,cp[0]*0.5);
            }
            case 1: { 
              return Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,(1.+cp[2])*0.5,
                              cp[1]*0.5);
            }
            case 2: {
              return Point<3>(cp[0]*0.5,(1.+cp[2])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5);
            }
          }
        }
      }
    }
    case 3: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return 
                Point<3>(cp[2]*0.5,cp[1]*0.5,1.-(cp[0]+cp[1]+cp[2])*0.5);
            }
            case 2: {
              return 
                Point<3>(cp[0]*0.5,cp[2]*0.5,1.-(cp[0]+cp[1]+cp[2])*0.5);
            }
            case 3: {
              return 
                Point<3>(cp[1]*0.5,cp[0]*0.5,1.-(cp[0]+cp[1]+cp[2])*0.5);
            }
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return 
                Point<3>(cp[1]*0.5,cp[2]*0.5,(1.+cp[0])*0.5);
            }
            case 2: {
              return 
                Point<3>(cp[2]*0.5,(1.-cp[0]-cp[1]-cp[2])*0.5,
                         (1.+cp[0])*0.5);
            }
            case 3: {
              return Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,cp[1]*0.5,
                              (1.+cp[0])*0.5);
            }
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[2]*0.5,cp[0]*0.5,(1.+cp[1])*0.5);
            }
            case 1: {
              return 
                Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,cp[2]*0.5,
                         (1.+cp[1])*0.5);
            }
            case 3: {
              return Point<3>(cp[0]*0.5,(1.-cp[0]-cp[1]-cp[2])*0.5,
                              (1.+cp[1])*0.5);
            }
          }
        }
        case 3: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[0]*0.5,cp[1]*0.5,(1.+cp[2])*0.5);
            }
            case 1: {
              return Point<3>(cp[1]*0.5,(1.-cp[0]-cp[1]-cp[2])*0.5,
                              (1.+cp[2])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,cp[0]*0.5,
                              (1.+cp[2])*0.5);
            }
          }
        }
      }
    }
    case 4: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.-cp[0]-cp[2])*0.5,cp[2]*0.5,
                              (cp[0]+cp[1]+cp[2])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[0]-cp[1])*0.5,cp[0]*0.5,
                              (cp[0]+cp[1]+cp[2])*0.5);
            }
            case 3:  {
              return Point<3>((1.-cp[1]-cp[2])*0.5,cp[1]*0.5,
                              (cp[0]+cp[1]+cp[2])*0.5);
            }
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>((cp[0]+cp[2])*0.5,cp[1]*0.5,
                              (1.-cp[0])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[1]-cp[2])*0.5,cp[2]*0.5,
                              (1.-cp[0])*0.5);
            }
            case 3: {
              return Point<3>((cp[0]+cp[1])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5,(1.-cp[0])*0.5);
            }
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>((cp[0]+cp[1])*0.5,cp[2]*0.5,
                              (1.-cp[1])*0.5);
            }
            case 1: {
              return Point<3>((cp[1]+cp[2])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5,(1.-cp[1])*0.5);
            }
            case 3: { 
              return Point<3>((1.-cp[0]-cp[2])*0.5,cp[0]*0.5,
                              (1.-cp[1])*0.5);
            }
          }
        }
        case 3: {
          switch (cvi1) {
            case 0: {
              return Point<3>((cp[1]+cp[2])*0.5,cp[0]*0.5,
                              (1.-cp[2])*0.5);
            }
            case 1: {
              return Point<3>((1.-cp[0]-cp[1])*0.5,cp[1]*0.5,
                              (1.-cp[2])*0.5);
            }
            case 2: { 
              return Point<3>((cp[0]+cp[2])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5,(1.-cp[2])*0.5);
            }
          }
        }
      }
    }
    case 5: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.-cp[1])*0.5,(cp[0]+cp[1])*0.5,
                              (cp[1]+cp[2])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[2])*0.5,(cp[1]+cp[2])*0.5,
                              (cp[0]+cp[2])*0.5);
            }
            case 3: {
              return Point<3>((1.-cp[0])*0.5,(cp[0]+cp[2])*0.5,
                              (cp[0]+cp[1])*0.5);
            }
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[2])*0.5,(1.-cp[0]-cp[1])*0.5,
                              (cp[1]+cp[2])*0.5);
            }
            case 2: {
              return Point<3>((cp[0]+cp[1]+cp[2])*0.5,
                              (1.-cp[0]-cp[2])*0.5,(1.-cp[0]-cp[1])*0.5);
            }
            case 3: { 
              return Point<3>((1.-cp[1])*0.5,(cp[1]+cp[2])*0.5,
                              (1.-cp[0]-cp[2])*0.5);
            }
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[0])*0.5,(1.-cp[1]-cp[2])*0.5,
                              (cp[0]+cp[2])*0.5);
            }
            case 1: {
              return Point<3>((1.-cp[2])*0.5,(cp[0]+cp[2])*0.5,
                              (1.-cp[0]-cp[1])*0.5);
            }
            case 3: {
              return Point<3>((cp[0]+cp[1]+cp[2])*0.5,
                              (1.-cp[0]-cp[1])*0.5,(1.-cp[1]-cp[2])*0.5);
            }
          }
        }
        case 3: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[1])*0.5,(1.-cp[0]-cp[2])*0.5,
                              (cp[0]+cp[1])*0.5);
            }
            case 1: {
              return Point<3>((cp[0]+cp[1]+cp[2])*0.5,
                              (1.-cp[1]-cp[2])*0.5,(1.-cp[0]-cp[2])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[0])*0.5,(cp[0]+cp[1])*0.5,
                              (1.-cp[1]-cp[2])*0.5);
            }
          }
        }
      }
    }
    case 6: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.-cp[1]-cp[2])*0.5,
                              (cp[0]+cp[1]+cp[2])*0.5,cp[2]*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[0]-cp[2])*0.5,
                              (cp[0]+cp[1]+cp[2])*0.5,cp[0]*0.5);
            }
            case 3: {
              return Point<3>((1.-cp[0]-cp[1])*0.5,
                              (cp[0]+cp[1]+cp[2])*0.5,cp[1]*0.5);
            }
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[1]-cp[2])*0.5,
                              (1.-cp[0])*0.5,cp[1]*0.5);
            }
            case 2: {
              return Point<3>((cp[0]+cp[1])*0.5,(1.-cp[0])*0.5,cp[2]*0.5);
            }
            case 3: {
              return Point<3>((cp[0]+cp[2])*0.5,(1.-cp[0])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5);
            }
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[0]-cp[2])*0.5,
                              (1.-cp[1])*0.5,cp[2]*0.5);
            }
            case 1: {
              return Point<3>((cp[0]+cp[1])*0.5,(1.-cp[1])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5);
            }
            case 3: {
              return Point<3>((cp[1]+cp[2])*0.5,(1.-cp[1])*0.5,
                              cp[0]*0.5);
            }
          }
        }
        case 3: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[0]-cp[1])*0.5,
                              (1.-cp[2])*0.5,cp[0]*0.5);
            }
            case 1: {
              return Point<3>((cp[0]+cp[2])*0.5,(1.-cp[2])*0.5,cp[1]*0.5);
            }
            case 2: {
              return Point<3>((cp[1]+cp[2])*0.5,(1.-cp[2])*0.5,
                              (1.-cp[0]-cp[1]-cp[2])*0.5);
            }
          }
        }
      }
    }
    case 7: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,
                              (cp[0]+cp[2])*0.5,(cp[1]+cp[2])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,
                              (cp[0]+cp[1])*0.5,(cp[0]+cp[2])*0.5);
            }
            case 3: {
              return Point<3>((1.-cp[0]-cp[1]-cp[2])*0.5,
                              (cp[1]+cp[2])*0.5,(cp[0]+cp[1])*0.5);
            }
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[0]*0.5,(1.-cp[0]-cp[2])*0.5,
                              (cp[1]+cp[2])*0.5);
            }
            case 2: {
              return Point<3>(cp[0]*0.5,(cp[1]+cp[2])*0.5,
                              (1.-cp[0]-cp[1])*0.5);
            }
            case 3: {
              return Point<3>(cp[0]*0.5,(1.-cp[0]-cp[1])*0.5,
                              (1.-cp[0]-cp[2])*0.5);
            }
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[1]*0.5,(1.-cp[0]-cp[1])*0.5,
                              (cp[0]+cp[2])*0.5);
            }
            case 1: {
              return Point<3>(cp[1]*0.5,(1.-cp[1]-cp[2])*0.5,
                              (1.-cp[0]-cp[1])*0.5);
            }
            case 3: {
              return Point<3>(cp[1]*0.5,(cp[0]+cp[2])*0.5,
                              (1.-cp[1]-cp[2])*0.5);
            }
          }
        }
        case 3: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[2]*0.5,(1.-cp[1]-cp[2])*0.5,
                              (cp[0]+cp[1])*0.5);
            }
            case 1: {
              return Point<3>(cp[2]*0.5,(cp[0]+cp[1])*0.5,
                              (1.-cp[0]-cp[2])*0.5);
            }
            case 2: {
              return Point<3>(cp[2]*0.5,(1.-cp[0]-cp[2])*0.5,
                              (1.-cp[1]-cp[2])*0.5);
            }
          }
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// 12 rotations of a tetrahedron times 8 children = 96 cases
Point<SPACEDIM> Tetrahedron::childCoordinateOfCanonical(
const Point<SPACEDIM> &pp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  const Tetrahedron *child_tetrahedron=
    dynamic_cast<const Tetrahedron*>(child_shape);
  CHECK_POINTER(child_tetrahedron);
  int cvi0=-1;
  int cvi1=-1;
  if (c<4) {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    CHECK_TEST(vertexIndex(c)==svi);
    cvi0=child_shape->vertexEqualToVertexIndex(svi);
    cvi1=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[c]);
  } else {
    cvi0=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[0]);
    switch (c) {
      case 4:
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[3]);
        break;
      case 5:
      case 6:
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[1]);
        break;
      case 7:
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[2]);
        break;
    }
  }
  CHECK_TEST(cvi0>=0);
  CHECK_TEST(cvi1>=0);
  
  switch (c) {
    case 0: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return pp*2.;
            case 2:
              return Point<3>(pp[2]*2.,pp[0]*2.,pp[1]*2.);
            case 3:
              return Point<3>(pp[1]*2.,pp[2]*2.,pp[0]*2.);
          }
        }
        case 1:
        case 2:
        case 3:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 1: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return 
                Point<3>(pp[1]*2.,(1.-pp[0]-pp[1]-pp[2])*2.,pp[2]*2.);
            case 2:
              return 
                Point<3>(pp[2]*2.,pp[1]*2.,(1.-pp[0]-pp[1]-pp[2])*2.);
            case 3:
              return 
                Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,pp[2]*2.,pp[1]*2.);
          }
        }
        case 1:
        case 2:
        case 3:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 2: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return 
                Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,pp[0]*2.,pp[2]*2.);
            case 2:
              return 
                Point<3>(pp[2]*2.,(1.-pp[0]-pp[1]-pp[2])*2.,pp[0]*2.);
            case 3: {
              return 
                Point<3>(pp[0]*2.,pp[2]*2.,(1.-pp[0]-pp[1]-pp[2])*2.);
            }
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[1]*2.-1.,pp[2]*2.,pp[0]*2.);
            case 2:
              return Point<3>(pp[1]*2.-1.,(1.-pp[0]-pp[1]-pp[2])*2.,
                              pp[2]*2.);
            case 3:
              return Point<3>(pp[1]*2.-1.,pp[0]*2.,
                              (1.-pp[0]-pp[1]-pp[2])*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[0]*2.,pp[1]*2.-1.,pp[2]*2.);
            case 1:
              return Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,pp[1]*2.-1.,
                              pp[0]*2.);
            case 3:
              return Point<3>(pp[2]*2.,pp[1]*2.-1.,
                              (1.-pp[0]-pp[1]-pp[2])*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[2]*2.,pp[0]*2.,pp[1]*2.-1.);
            case 1:
              return Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,pp[2]*2.,
                              pp[1]*2.-1.);
            case 2:
              return Point<3>(pp[0]*2.,(1.-pp[0]-pp[1]-pp[2])*2.,
                              pp[1]*2.-1.);
          }
        }
      }
    }
    case 3: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return 
                Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,pp[1]*2.,pp[0]*2.);
            case 2:
              return 
                Point<3>(pp[0]*2.,(1.-pp[0]-pp[1]-pp[2])*2.,pp[1]*2.);
            case 3:
              return 
                Point<3>(pp[1]*2.,pp[0]*2.,(1.-pp[0]-pp[1]-pp[2])*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[2]*2.-1.,pp[0]*2.,pp[1]*2.);
            case 2:
              return 
                Point<3>(pp[2]*2.-1.,(1.-pp[0]-pp[1]-pp[2])*2.,pp[0]*2.);
            case 3:
              return 
                Point<3>(pp[2]*2.-1.,pp[1]*2.,(1.-pp[0]-pp[1]-pp[2])*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: return Point<3>(pp[1]*2.,pp[2]*2.-1.,pp[0]*2.);
            case 1:
              return Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,pp[2]*2.-1.,
                              pp[1]*2.);
            case 3:
              return Point<3>(pp[0]*2.,pp[2]*2.-1.,
                              (1.-pp[0]-pp[1]-pp[2])*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[0]*2.,pp[1]*2.,pp[2]*2.-1.);
            case 1:
              return Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,pp[0]*2.,
                         pp[2]*2.-1.);
            case 2:
              return Point<3>(pp[1]*2.,(1.-pp[0]-pp[1]-pp[2])*2.,
                              pp[2]*2.-1.);
          }
        }
      }
    }
    case 4: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>(1.-(pp[0]+pp[1])*2.,(pp[0]+pp[2])*2.-1.,
                              pp[1]*2.);
            case 2:
              return Point<3>(pp[1]*2.,1.-(pp[0]+pp[1])*2.,
                              (pp[0]+pp[2])*2.-1.);
            case 3:
              return Point<3>((pp[0]+pp[2])*2.-1.,pp[1]*2.,
                              1.-(pp[0]+pp[1])*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(1.-pp[2]*2.,pp[1]*2.,(pp[0]+pp[2])*2.-1.);
            case 2:
              return Point<3>(1.-pp[2]*2.,1.-(pp[0]+pp[1])*2.,pp[1]*2.);
            case 3:
              return Point<3>(1.-pp[2]*2.,(pp[0]+pp[2])*2.-1.,
                              1.-(pp[0]+pp[1])*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>((pp[0]+pp[2])*2.-1.,1.-pp[2]*2.,pp[1]*2.);
            case 1:
              return Point<3>(1.-(pp[0]+pp[1])*2.,1.-pp[2]*2.,
                              (pp[0]+pp[2])*2.-1.);
            case 3:
              return Point<3>(pp[1]*2.,1.-pp[2]*2.,1.-(pp[0]+pp[1])*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[1]*2.,(pp[0]+pp[2])*2.-1.,1.-pp[2]*2.);
            case 1:
              return Point<3>(1.-(pp[0]+pp[1])*2.,pp[1]*2.,
                              1.-pp[2]*2.);
            case 2:
              return Point<3>((pp[0]+pp[2])*2.-1.,1.-(pp[0]+pp[1])*2.,
                              1.-pp[2]*2.);
          }
        }
      }
    }
    case 5: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>((pp[0]+pp[1])*2.-1.,1.-pp[0]*2.,
                              (pp[0]+pp[2])*2.-1.);
            case 2:
              return Point<3>((pp[0]+pp[2])*2.-1.,(pp[0]+pp[1])*2.-1.,
                              1.-pp[0]*2.);
            case 3:
              return Point<3>(1.-pp[0]*2.,(pp[0]+pp[2])*2.-1.,
                              (pp[0]+pp[1])*2.-1.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,
                              (pp[0]+pp[2])*2.-1.,1.-pp[0]*2.);
            case 2:
              return Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,
                              (pp[0]+pp[1])*2.-1.,(pp[0]+pp[2])*2.-1.);
            case 3:
              return Point<3>((1.-pp[0]-pp[1]-pp[2])*2.,1.-pp[0]*2.,
                              (pp[0]+pp[1])*2.-1.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>(1.-pp[0]*2.,(1.-pp[0]-pp[1]-pp[2])*2.,
                              (pp[0]+pp[2])*2.-1.);
            case 1:
              return Point<3>((pp[0]+pp[1])*2.-1.,
                              (1.-pp[0]-pp[1]-pp[2])*2.,1.-pp[0]*2.);
            case 3:
              return Point<3>((pp[0]+pp[2])*2.-1.,
                              (1.-pp[0]-pp[1]-pp[2])*2.,
                              (pp[0]+pp[1])*2.-1.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 0:
              return Point<3>((pp[0]+pp[2])*2.-1.,1.-pp[0]*2.,
                              (1.-pp[0]-pp[1]-pp[2])*2.);
            case 1:
              return Point<3>((pp[0]+pp[1])*2.-1.,(pp[0]+pp[2])*2.-1.,
                              (1.-pp[0]-pp[1]-pp[2])*2.);
            case 2:
              return Point<3>(1.-pp[0]*2.,(pp[0]+pp[1])*2.-1.,
                              (1.-pp[0]-pp[1]-pp[2])*2.);
          }
        }
      }
    }
    case 6: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>((pp[0]+pp[1])*2.-1.,1.-(pp[0]+pp[2])*2.,
                              pp[2]*2.);
            case 2:
              return Point<3>(2.*pp[2],(pp[0]+pp[1])*2.-1.,
                              1.-(pp[0]+pp[2])*2.);
            case 3:
              return Point<3>(1.-(pp[0]+pp[2])*2.,pp[2]*2.,
                              (pp[0]+pp[1])*2.-1.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(1.-pp[1]*2.,pp[2]*2.,1.-(pp[0]+pp[2])*2.);
            case 2:
              return Point<3>(1.-pp[1]*2.,(pp[0]+pp[1])*2.-1.,pp[2]*2.);
            case 3:
              return Point<3>(1.-pp[1]*2.,1.-(pp[0]+pp[2])*2.,
                              (pp[0]+pp[1])*2.-1.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>(1.-(pp[0]+pp[2])*2.,1.-pp[1]*2.,pp[2]*2.);
            case 1:
              return Point<3>((pp[0]+pp[1])*2.-1.,1.-pp[1]*2.,
                              1.-(pp[0]+pp[2])*2.);
            case 3:
              return Point<3>(pp[2]*2.,1.-pp[1]*2.,(pp[0]+pp[1])*2.-1.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[2]*2.,1.-(pp[0]+pp[2])*2.,1.-pp[1]*2.);
            case 1:
              return Point<3>((pp[0]+pp[1])*2.-1.,pp[2]*2.,1.-pp[1]*2.);
            case 2:
              return Point<3>(1.-(pp[0]+pp[2])*2.,(pp[0]+pp[1])*2.-1.,
                              1.-pp[1]*2.);
          }
        }
      }
    }
    case 7: {
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>(1.-(pp[0]+pp[2])*2.,1.-(pp[0]+pp[1])*2.,
                              (pp[0]+pp[1]+pp[2])*2.-1.);
            case 2:
              return Point<3>((pp[0]+pp[1]+pp[2])*2.-1.,
                              1.-(pp[0]+pp[2])*2.,
                              1.-(pp[0]+pp[1])*2.);
            case 3:
              return Point<3>(1.-(pp[0]+pp[1])*2.,
                        (pp[0]+pp[1]+pp[2])*2.-1., 1.-(pp[0]+pp[2])*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[0]*2.,(pp[0]+pp[1]+pp[2])*2.-1.,
                              1.-(pp[0]+pp[1])*2.);
            case 2:
              return Point<3>(pp[0]*2.,1.-(pp[0]+pp[2])*2.,
                              (pp[0]+pp[1]+pp[2])*2.-1.);
            case 3:
              return Point<3>(pp[0]*2.,1.-(pp[0]+pp[1])*2.,
                              1.-(pp[0]+pp[2])*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>(1.-(pp[0]+pp[1])*2.,pp[0]*2.,
                              (pp[0]+pp[1]+pp[2])*2.-1.);
            case 1:
              return Point<3>(1.-(pp[0]+pp[2])*2.,
                              pp[0]*2.,1.-(pp[0]+pp[1])*2.);
            case 3:
              return Point<3>((pp[0]+pp[1]+pp[2])*2.-1.,pp[0]*2.,
                              1.-(pp[0]+pp[2])*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 0:
              return Point<3>((pp[0]+pp[1]+pp[2])*2.-1.,
                              1.-(pp[0]+pp[1])*2.,pp[0]*2.);
            case 1:
              return Point<3>(1.-(pp[0]+pp[2])*2.,
                              (pp[0]+pp[1]+pp[2])*2.-1.,pp[0]*2.);
            case 2:
              return Point<3>(1.-(pp[0]+pp[1])*2.,1.-(pp[0]+pp[2])*2.,
                              pp[0]*2.);
          }
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//see FE_Q::FE_Q; must be consistent with parentCoordinateOfCanonical
const Shape* Tetrahedron::reorientFor(const Shape *parent,int c,
const NumPtr<int> &line_interior_child_vertices) const {
#ifdef DEBUG
  if (c<4) CHECK_SAME(vertexIndex(0),parent->vertexIndex(c));
#endif
  switch (c) {
    case 0: { // v0,v4,v6,v7
//    parentCoordinateOfCanonical(p,0)=p*0.5
      int cv0=line_interior_child_vertices[0];
      if (vertexIndex(1)==cv0) {
        return this;
      } else if (vertexIndex(2)==cv0) {
        int v[4]={vertexIndex(0),vertexIndex(2),
                  vertexIndex(3),vertexIndex(1)};
        int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(2),
                  codimension1ShapeIndex(0),codimension1ShapeIndex(3)};
        int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(5),
                  codimension2ShapeIndex(3),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(1),codimension2ShapeIndex(4)};
        return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
      } else {
        CHECK_SAME(vertexIndex(3),cv0);
        int v[4]={vertexIndex(0),vertexIndex(3),
                  vertexIndex(1),vertexIndex(2)};
        int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(1),codimension1ShapeIndex(3)};
        int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(1)};
        return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
      }
    }
    case 1: { // v1,v4,v8,v5
//    parentCoordinateOfCanonical(p,1)=
//    Point<3>(1.-(p[0]+p[1]+p[2])*0.5,p[2]*0.5,p[1]*0.5);
      int cv1=line_interior_child_vertices[1];
      if (vertexIndex(1)==cv1) {
        int v[4]={vertexIndex(0),vertexIndex(2),
                  vertexIndex(3),vertexIndex(1)};
        int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(2),
                  codimension1ShapeIndex(0),codimension1ShapeIndex(3)};
        int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(5),
                  codimension2ShapeIndex(3),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(1),codimension2ShapeIndex(4)};
        return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
      } else if (vertexIndex(2)==cv1) {
        int v[4]={vertexIndex(0),vertexIndex(3),
                  vertexIndex(1),vertexIndex(2)};
        int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(1),codimension1ShapeIndex(3)};
        int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(1)};
        return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
      } else {
        CHECK_SAME(vertexIndex(3),cv1);
        return this;
      }
    }
    case 2: {
//    parentCoordinateOfCanonical(p,1)=
//    Point<3>(p[0]*0.5,1.-(p[0]+p[1]+p[2])*0.5,p[1]*0.5);
      int cv2=line_interior_child_vertices[2];
      if (vertexIndex(1)==cv2) {
        int v[4]={vertexIndex(0),vertexIndex(2),
                  vertexIndex(3),vertexIndex(1)};
        int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(2),
                  codimension1ShapeIndex(0),codimension1ShapeIndex(3)};
        int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(5),
                  codimension2ShapeIndex(3),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(1),codimension2ShapeIndex(4)};
        return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
      } else if (vertexIndex(2)==cv2) {
        int v[4]={vertexIndex(0),vertexIndex(3),
                  vertexIndex(1),vertexIndex(2)};
        int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(1),codimension1ShapeIndex(3)};
        int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(1)};
        return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
      } else {
        CHECK_SAME(vertexIndex(3),cv2);
        return this;
      }
    }
    case 3: {
//    parentCoordinateOfCanonical(p,1)=
//    Point<3>(p[2]*0.5,p[1]*0.5,1.-(p[0]+p[1]+p[2])*0.5)
      int cv3=line_interior_child_vertices[3];
      if (vertexIndex(1)==cv3) {
        return this;
      } else if (vertexIndex(2)==cv3) {
        int v[4]={vertexIndex(0),vertexIndex(2),
                  vertexIndex(3),vertexIndex(1)};
        int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(2),codimension1ShapeIndex(3)};
        int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(1),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(3),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(4)};
        return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
      } else {
        CHECK_SAME(vertexIndex(3),cv3);
        int v[4]={vertexIndex(0),vertexIndex(3),
                  vertexIndex(1),vertexIndex(2)};
        int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(1),codimension1ShapeIndex(3)};
        int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(1)};
        return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
      }
    }
    case 4: {
//    parentCoordinateOfCanonical(p,1)=
//    Point<3>((1.-p[0]-p[2])*0.5,p[2]*0.5,(p[0]+p[1]+p[2])*0.5);
      int cv4=line_interior_child_vertices[4];
      int cv3=line_interior_child_vertices[3];
      if (vertexIndex(0)==cv4) {
        if (vertexIndex(1)==cv3) {
          int v[4]={vertexIndex(3),vertexIndex(1),
                    vertexIndex(0),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(0),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(4),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(2)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(2)==cv3) {
          int v[4]={vertexIndex(1),vertexIndex(2),
                    vertexIndex(0),vertexIndex(3)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(2),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(1),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(3)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(3),cv3)
          int v[4]={vertexIndex(2),vertexIndex(3),
                    vertexIndex(0),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(5),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(0)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else if (vertexIndex(1)==cv4) {
        if (vertexIndex(2)==cv3) {
          int v[4]={vertexIndex(3),vertexIndex(2),
                    vertexIndex(1),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(5),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(0)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(3)==cv3) {
          int v[4]={vertexIndex(0),vertexIndex(3),
                    vertexIndex(1),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(3)};
          int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(1)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(0),cv3)
          int v[4]={vertexIndex(2),vertexIndex(0),
                    vertexIndex(1),vertexIndex(3)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(2),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(4)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else if (vertexIndex(2)==cv4) {
        if (vertexIndex(3)==cv3) {
          int v[4]={vertexIndex(1),vertexIndex(3),
                    vertexIndex(2),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(1),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(4),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(2)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(0)==cv3) {
          int v[4]={vertexIndex(3),vertexIndex(0),
                    vertexIndex(2),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(1),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(1)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(1),cv3)
          return this;
        }
      } else {
        CHECK_SAME(vertexIndex(3),cv4);
        if (vertexIndex(0)==cv3) {
          int v[4]={vertexIndex(1),vertexIndex(0),
                    vertexIndex(3),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(0),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(5)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(1)==cv3) {
          int v[4]={vertexIndex(2),vertexIndex(1),
                    vertexIndex(3),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(0),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(1),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(3)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(2),cv3)
          int v[4]={vertexIndex(0),vertexIndex(2),
                    vertexIndex(3),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(3)};
          int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(4)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      }
    }
    case 5: {
//    parentCoordinateOfCanonical(p,1)=
//    Point<3>((1.-p[1])*0.5,(p[0]+p[1])*0.5,(p[1]+p[2])*0.5)
      int cv4=line_interior_child_vertices[4];
      int cv5=line_interior_child_vertices[5];
      if (vertexIndex(0)==cv4) {
        if (vertexIndex(1)==cv5) {
          int v[4]={vertexIndex(3),vertexIndex(2),
                    vertexIndex(1),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(5),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(0)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(2)==cv5) {
          int v[4]={vertexIndex(1),vertexIndex(3),
                    vertexIndex(2),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(1),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(4),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(2)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(3),cv5)
          int v[4]={vertexIndex(2),vertexIndex(1),
                    vertexIndex(3),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(0),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(1),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(3)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else if (vertexIndex(1)==cv4) {
        if (vertexIndex(2)==cv5) {
          int v[4]={vertexIndex(3),vertexIndex(0),
                    vertexIndex(2),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(1),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(1)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(3)==cv5) {
          int v[4]={vertexIndex(0),vertexIndex(2),
                    vertexIndex(3),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(3)};
          int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(4)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(0),cv5)
          int v[4]={vertexIndex(2),vertexIndex(3),
                    vertexIndex(0),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(5),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(0)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else if (vertexIndex(2)==cv4) {
        if (vertexIndex(3)==cv5) {
          int v[4]={vertexIndex(1),vertexIndex(0),
                    vertexIndex(3),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(0),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(5)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(0)==cv5) {
          int v[4]={vertexIndex(3),vertexIndex(1),
                    vertexIndex(0),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(0),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(4),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(2)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(1),cv5)
          int v[4]={vertexIndex(0),vertexIndex(3),
                    vertexIndex(1),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(3)};
          int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(1)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else {
        CHECK_SAME(vertexIndex(3),cv4);
        if (vertexIndex(0)==cv5) {
          int v[4]={vertexIndex(1),vertexIndex(2),
                    vertexIndex(0),vertexIndex(3)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(2),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(1),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(3)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(1)==cv5) {
          int v[4]={vertexIndex(2),vertexIndex(0),
                    vertexIndex(1),vertexIndex(3)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(2),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(4)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(2),cv5)
          return this;
        }
      }
    }
    case 6: {
//    parentCoordinateOfCanonical(p,1)=
//    Point<3>((1.-p[1]-p[2])*0.5,(p[0]+p[1]+p[2])*0.5,p[2]*0.5)
      int cv2=line_interior_child_vertices[2];
      int cv5=line_interior_child_vertices[5];
      if (vertexIndex(0)==cv2) {
        if (vertexIndex(1)==cv5) {
          int v[4]={vertexIndex(2),vertexIndex(3),
                    vertexIndex(0),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(5),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(0)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(2)==cv5) {
          int v[4]={vertexIndex(3),vertexIndex(1),
                    vertexIndex(0),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(0),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(4),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(2)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(3),cv5)
          int v[4]={vertexIndex(1),vertexIndex(2),
                    vertexIndex(0),vertexIndex(3)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(2),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(1),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(3)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else if (vertexIndex(1)==cv2) {
        if (vertexIndex(2)==cv5) {
          int v[4]={vertexIndex(0),vertexIndex(3),
                    vertexIndex(1),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(3)};
          int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(1)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(3)==cv5) {
          int v[4]={vertexIndex(2),vertexIndex(0),
                    vertexIndex(1),vertexIndex(3)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(2),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(4)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(0),cv5)
          int v[4]={vertexIndex(3),vertexIndex(2),
                    vertexIndex(1),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(5),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(0)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else if (vertexIndex(2)==cv2) {
        if (vertexIndex(3)==cv5) {
          return this;
        } else if (vertexIndex(0)==cv5) {
          int v[4]={vertexIndex(1),vertexIndex(3),
                    vertexIndex(2),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(1),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(4),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(2)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(1),cv5)
          int v[4]={vertexIndex(3),vertexIndex(0),
                    vertexIndex(2),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(1),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(1)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else {
        CHECK_SAME(vertexIndex(3),cv2);
        if (vertexIndex(0)==cv5) {
          int v[4]={vertexIndex(2),vertexIndex(1),
                    vertexIndex(3),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(0),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(1),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(3)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(1)==cv5) {
          int v[4]={vertexIndex(0),vertexIndex(2),
                    vertexIndex(3),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(3)};
          int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(4)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(2),cv5)
          int v[4]={vertexIndex(1),vertexIndex(0),
                    vertexIndex(3),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(0),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(5)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      }
    }
    case 7: {
//    parentCoordinateOfCanonical(p,1)=
//    Point<3>((1.-p[0]-p[1]-p[2])*0.5,(p[0]+p[2])*0.5,(p[1]+p[2])*0.5)
      int cv2=line_interior_child_vertices[2];
      int cv5=line_interior_child_vertices[5];
      if (vertexIndex(0)==cv5) {
        if (vertexIndex(1)==cv2) {
          int v[4]={vertexIndex(2),vertexIndex(1),
                    vertexIndex(3),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(0),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(1),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(3)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(2)==cv2) {
          int v[4]={vertexIndex(3),vertexIndex(2),
                    vertexIndex(1),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(5),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(0)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(3),cv2)
          int v[4]={vertexIndex(1),vertexIndex(3),
                    vertexIndex(2),vertexIndex(0)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(1),
                    codimension1ShapeIndex(3),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(4),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(2)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else if (vertexIndex(1)==cv5) {
        if (vertexIndex(2)==cv2) {
          int v[4]={vertexIndex(0),vertexIndex(2),
                    vertexIndex(3),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(3)};
          int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(4)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(3)==cv2) {
          int v[4]={vertexIndex(2),vertexIndex(3),
                    vertexIndex(0),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(5),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(0)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(0),cv2)
          int v[4]={vertexIndex(3),vertexIndex(0),
                    vertexIndex(2),vertexIndex(1)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(1),
                    codimension1ShapeIndex(0),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(1)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else if (vertexIndex(2)==cv5) {
        if (vertexIndex(3)==cv2) {
          int v[4]={vertexIndex(0),vertexIndex(3),
                    vertexIndex(1),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(2),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(3)};
          int l[6]={codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(1)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(0)==cv2) {
          int v[4]={vertexIndex(1),vertexIndex(0),
                    vertexIndex(3),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(2),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(0),codimension2ShapeIndex(3),
                    codimension2ShapeIndex(4),codimension2ShapeIndex(1),
                    codimension2ShapeIndex(2),codimension2ShapeIndex(5)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else {
          CHECK_SAME(vertexIndex(1),cv2)
          int v[4]={vertexIndex(3),vertexIndex(1),
                    vertexIndex(0),vertexIndex(2)};
          int f[4]={codimension1ShapeIndex(0),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(1),codimension1ShapeIndex(2)};
          int l[6]={codimension2ShapeIndex(4),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(2)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      } else {
        CHECK_SAME(vertexIndex(3),cv5);
        if (vertexIndex(0)==cv2) {
          int v[4]={vertexIndex(2),vertexIndex(0),
                    vertexIndex(1),vertexIndex(3)};
          int f[4]={codimension1ShapeIndex(3),codimension1ShapeIndex(0),
                    codimension1ShapeIndex(2),codimension1ShapeIndex(1)};
          int l[6]={codimension2ShapeIndex(2),codimension2ShapeIndex(0),
                    codimension2ShapeIndex(1),codimension2ShapeIndex(5),
                    codimension2ShapeIndex(3),codimension2ShapeIndex(4)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        } else if (vertexIndex(1)==cv2) {
          return this;
        } else {
          CHECK_SAME(vertexIndex(2),cv2)
          int v[4]={vertexIndex(1),vertexIndex(2),
                    vertexIndex(0),vertexIndex(3)};
          int f[4]={codimension1ShapeIndex(1),codimension1ShapeIndex(3),
                    codimension1ShapeIndex(2),codimension1ShapeIndex(0)};
          int l[6]={codimension2ShapeIndex(1),codimension2ShapeIndex(2),
                    codimension2ShapeIndex(0),codimension2ShapeIndex(4),
                    codimension2ShapeIndex(5),codimension2ShapeIndex(3)};
          return OPERATOR_NEW Tetrahedron(l,f,v,dof_index);
        }
      }
    }
    default:
      OBSOLETE("bad c");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::staticProjectToLine(int l,
const NumPtr<Point<1> > &q_points,NumPtr<Point<3> > &q_points_line) {
  switch (l) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=q_points[i][0];
        q_points_line[i][1]=0.;
        q_points_line[i][2]=0.;
      }
      break;
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.-q_points[i][0];
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=0.;
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=0.;
      }
      break;
    case 3:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=0.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    case 4:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.-q_points[i][0];
        q_points_line[i][1]=0.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    case 5:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=1.-q_points[i][0];
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::staticProjectToFace(int f,
const NumPtr<Point<2> > &q_points,NumPtr<Point<3> > &q_points_face) {
  switch (f) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=0.;
        q_points_face[i][1]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][1]=0.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][2]=0.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=q_points[i][1];
      }
      break;
    case 3:
      for (int i=0;i<q_points.getNumber();i++) {
//      in order to get RTSimplexDofShapeFunction to be consistent
//        across faces, we have to use the second set of functions
        q_points_face[i][0]=1.-q_points[i][0]-q_points[i][1];
        q_points_face[i][1]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::projectToSubface(int f,int s,
const NumPtr<Point<2> > &q_points,NumPtr<Point<3> > &q_points_subface) {
  switch (f) {
    case 0: {
      switch (s) { // child 0
        case 0: {
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=qp[0]*0.5;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=1.-(qp[0]+qp[1])*0.5;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 2: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=qp[1]*0.5;
            qps[2]=1.-(qp[0]+qp[1])*0.5;
          }
          break;
        }
        case 3: { // child 6
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=(qp[0]+qp[1])*0.5;
            qps[2]=(1.-qp[0])*0.5;
          }
          break;
        }
      }
      break;
    }
    case 1: {
      switch (s) {
        case 0: { // child 0
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=0.;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 3
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=0.;
            qps[2]=1.-(qp[0]+qp[1])*0.5;
          }
          break;
        }
        case 2: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-(qp[0]+qp[1])*0.5;
            qps[1]=0.;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 3: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=(qp[0]+qp[1])*0.5;
            qps[1]=0.;
            qps[2]=(1.-qp[0])*0.5;
          }
          break;
        }
      }
      break;
    }
    case 2: {
      switch (s) {
        case 0: { // child 0
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=qp[1]*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 1: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-(qp[0]+qp[1])*0.5;
            qps[1]=qp[0]*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 2: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=1.-(qp[0]+qp[1])*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 3: { // child 6
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=(qp[0]+qp[1])*0.5;
            qps[1]=(1.-qp[0])*0.5;
            qps[2]=0.;
          }
          break;
        }
      }
      break;
    }
    case 3: {
      switch (s) {
        case 0: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-(qp[0]+qp[1])*0.5;
            qps[1]=qp[0]*0.5;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 3
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=qp[1]*0.5;
            qps[2]=1.-(qp[0]+qp[1])*0.5;
          }
          break;
        }
        case 2: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=1.-(qp[0]+qp[1])*0.5;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 3: { // child 5
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=(1.-qp[1])*0.5;
            qps[1]=(qp[0]+qp[1])*0.5;
            qps[2]=(1.-qp[0])*0.5;
          }
          break;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape2* Tetrahedron::newInteriorCodimension1Shape(int i,
int codimension1_element_index,
const NumPtr<int> &interior_child_vertex_indices,
const NumPtr<int> &interior_child_face_vertex_indices,
const NumPtr<int> &interior_child_line_vertex_indices,
const map<pair<int,int>,int> &interior_child_line_map) const {
  CHECK_BOUNDS(i,0,8);
  CHECK_SAME(interior_child_vertex_indices.getNumber(),0);
  CHECK_SAME(interior_child_face_vertex_indices.getNumber(),4);
  CHECK_SAME(interior_child_line_vertex_indices.getNumber(),6);
  int v[3];
  switch (i) {
    case 0:
      v[0]=interior_child_line_vertex_indices[0]; // v4
      v[1]=interior_child_line_vertex_indices[3]; // v7
      v[2]=interior_child_line_vertex_indices[2]; // v6
      break;
    case 1:
      v[0]=interior_child_line_vertex_indices[0]; // v4
      v[1]=interior_child_line_vertex_indices[1]; // v5
      v[2]=interior_child_line_vertex_indices[4]; // v8
      break;
    case 2:
      v[0]=interior_child_line_vertex_indices[2]; // v6
      v[1]=interior_child_line_vertex_indices[5]; // v9
      v[2]=interior_child_line_vertex_indices[1]; // v5
      break;
    case 3:
      v[0]=interior_child_line_vertex_indices[3]; // v7
      v[1]=interior_child_line_vertex_indices[4]; // v8
      v[2]=interior_child_line_vertex_indices[5]; // v9
      break;
    case 4:
      v[0]=interior_child_line_vertex_indices[3]; // v7
      v[1]=interior_child_line_vertex_indices[0]; // v4
      v[2]=interior_child_line_vertex_indices[5]; // v9
      break;
    case 5:
      v[0]=interior_child_line_vertex_indices[4]; // v8
      v[1]=interior_child_line_vertex_indices[5]; // v9
      v[2]=interior_child_line_vertex_indices[0]; // v4
      break;
    case 6:
      v[0]=interior_child_line_vertex_indices[5]; // v9
      v[1]=interior_child_line_vertex_indices[0]; // v4
      v[2]=interior_child_line_vertex_indices[1]; // v5
      break;
    case 7:
      v[0]=interior_child_line_vertex_indices[2]; // v6
      v[1]=interior_child_line_vertex_indices[5]; // v9
      v[2]=interior_child_line_vertex_indices[0]; // v4
      break;
  }
  int l[3];
  setTriangleLineIndices(v,interior_child_line_map,l);
  return OPERATOR_NEW Triangle(codimension1_element_index,l,v);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape3* Tetrahedron::newChildShape(int c,int child_element_index,
const NumPtr<int> &interior_child_vertex_indices,
const NumPtr<int> &interior_child_face_vertex_indices,
const NumPtr<int> &interior_child_line_vertex_indices,
const map<pair<pair<int,int>,pair<int,int> >,int> 
&interior_child_face_map,
const map<pair<int,int>,int> &interior_child_line_map) const {
  CHECK_BOUNDS(c,0,8);
  CHECK_SAME(interior_child_vertex_indices.getNumber(),0);
  CHECK_SAME(interior_child_face_vertex_indices.getNumber(),4);
  CHECK_SAME(interior_child_line_vertex_indices.getNumber(),6);
  int v[4];
  switch (c) {
    case 0:
      v[0]=vertexIndex(0);                        // v0
      v[1]=interior_child_line_vertex_indices[0]; // v4
      v[2]=interior_child_line_vertex_indices[2]; // v6
      v[3]=interior_child_line_vertex_indices[3]; // v7
      break;
    case 1:
      v[0]=interior_child_line_vertex_indices[0]; // v4
      v[1]=vertexIndex(1);                        // v1
      v[2]=interior_child_line_vertex_indices[1]; // v5
      v[3]=interior_child_line_vertex_indices[4]; // v8
      break;
    case 2:
      v[0]=interior_child_line_vertex_indices[2]; // v6
      v[1]=interior_child_line_vertex_indices[1]; // v5
      v[2]=vertexIndex(2);                        // v2
      v[3]=interior_child_line_vertex_indices[5]; // v9
      break;
    case 3:
      v[0]=interior_child_line_vertex_indices[3]; // v7
      v[1]=interior_child_line_vertex_indices[4]; // v8
      v[2]=interior_child_line_vertex_indices[5]; // v3
      v[3]=vertexIndex(3);                        // v9
      break;
    case 4:
      v[0]=interior_child_line_vertex_indices[4]; // v8
      v[1]=interior_child_line_vertex_indices[3]; // v7
      v[2]=interior_child_line_vertex_indices[5]; // v9
      v[3]=interior_child_line_vertex_indices[0]; // v4
      break;
    case 5:
      v[0]=interior_child_line_vertex_indices[4]; // v8
      v[1]=interior_child_line_vertex_indices[5]; // v9
      v[2]=interior_child_line_vertex_indices[1]; // v5
      v[3]=interior_child_line_vertex_indices[0]; // v4
      break;
    case 6:
      v[0]=interior_child_line_vertex_indices[2]; // v6
      v[1]=interior_child_line_vertex_indices[5]; // v9
      v[2]=interior_child_line_vertex_indices[0]; // v4
      v[3]=interior_child_line_vertex_indices[1]; // v5
      break;
    case 7:
      v[0]=interior_child_line_vertex_indices[5]; // v9
      v[1]=interior_child_line_vertex_indices[2]; // v6
      v[2]=interior_child_line_vertex_indices[0]; // v4
      v[3]=interior_child_line_vertex_indices[3]; // v7
      break;
  }
  int f[4];
  for (int i=0;i<4;i++) {
    int tl[3]={-1,-2,-3};
    int tv[3]={v[face_vertex_to_cell_vertex[i][0]],
               v[face_vertex_to_cell_vertex[i][1]],
               v[face_vertex_to_cell_vertex[i][2]]};
    Triangle bogus_triangle(-1,tl,tv);
    pair<pair<int,int>,pair<int,int> > p(
      make_pair(bogus_triangle.vertexIndex(0),
                bogus_triangle.vertexIndex(1)),
      make_pair(bogus_triangle.vertexIndex(2),-1));
    map<pair<pair<int,int>,pair<int,int> >,int>::const_iterator 
      icfm_iterator=interior_child_face_map.find(p);
    CHECK_TEST(icfm_iterator!=interior_child_face_map.end());
    f[i]=icfm_iterator->second;
  }
  int l[6];
  for (int i=0;i<6;i++) {
    int iv[2]={v[canonical_vertex_indices[i][0]],
               v[canonical_vertex_indices[i][1]]};
    Interval bogus_interval(-1,iv);
    pair<int,int> p(bogus_interval.vertexIndex(0),
                    bogus_interval.vertexIndex(1));
    map<pair<int,int>,int>::const_iterator iclm_iterator=
      interior_child_line_map.find(p);
    CHECK_TEST(iclm_iterator!=interior_child_line_map.end());
    l[i]=iclm_iterator->second;
  }
  return OPERATOR_NEW Tetrahedron(child_element_index,l,f,v);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//point from lowest vertex index to an adjacent vertex
Tensor<1,3> Tetrahedron::faceTangent(int f,int n) const {
  switch (f) {
    case 0: {
      if (vertexIndex(2)<vertexIndex(3)) {
        return (n==0 ? -line_tangents[2] : line_tangents[3]);
      } else {
        return (n==1 ? -line_tangents[2] : line_tangents[3]);
      }
    }
    case 1: {
      if (vertexIndex(1)<vertexIndex(3)) {
        return (n==0 ? line_tangents[0] : line_tangents[3]);
      } else {
        return (n==1 ? line_tangents[0] : line_tangents[3]);
      }
    }
    case 2: {
      if (vertexIndex(1)<vertexIndex(2)) {
        return (n==0 ? line_tangents[0] : -line_tangents[2]);
      } else {
        return (n==1 ? line_tangents[0] : -line_tangents[2]);
      }
      break;
    }
    case 3: {
      int vmin=min(vertexIndex(1),min(vertexIndex(2),vertexIndex(3)));
      if (vertexIndex(1)==vmin) {
        if (vertexIndex(2)<vertexIndex(3)) {
          return (n==0 ? line_tangents[1] : -line_tangents[4]);
        } else {
          return (n==1 ? line_tangents[1] : -line_tangents[4]);
        }
      } else if (vertexIndex(2)==vmin) {
        if (vertexIndex(1)<vertexIndex(3)) {
          return (n==0 ? -line_tangents[1] : -line_tangents[5]);
        } else {
          return (n==1 ? -line_tangents[1] : -line_tangents[5]);
        }
      } else {
        if (vertexIndex(1)<vertexIndex(2)) {
          return (n==0 ? line_tangents[4] : line_tangents[5]);
        } else {
          return (n==1 ? line_tangents[4] : line_tangents[5]);
        }
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Tetrahedron::faceTangentsWindBackward(int f) const {
  switch (f) {
    case 0: // t0 x t1 . n > 0
      return (vertexIndex(2)>=vertexIndex(3));
    case 1:
      return (vertexIndex(1)>=vertexIndex(3));
    case 2:
      return (vertexIndex(1)>=vertexIndex(2));
    case 3: {
      int vmin=min(vertexIndex(1),min(vertexIndex(2),vertexIndex(3)));
      if (vertexIndex(1)==vmin) {
        return (vertexIndex(2)>=vertexIndex(3));
      } else if (vertexIndex(2)==vmin) {
        return(vertexIndex(1)<vertexIndex(3));
      }
      return (vertexIndex(1)>=vertexIndex(2));
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::homogeneousCoordinate(int f,const Point<2> &p,
Point<2> &hp) const {
  switch (f) {
    case 0:
      if (faceTangentsWindBackward(f)) {
        hp[0]=p[1];
        hp[1]=p[0];
      } else hp=p;
      break;
    case 1:
    case 2:
      hp=p;
      break;
    case 3: {
      int vmin=min(vertexIndex(1),min(vertexIndex(2),vertexIndex(3)));
      if (vertexIndex(1)==vmin) {
        if (faceTangentsWindBackward(f)) {
          hp[0]=1.-p[0]-p[1];
          hp[1]=p[1];
        } else {
          hp[0]=p[1];
          hp[1]=1.-p[0]-p[1];
        }
        break;
      } else if (vertexIndex(2)==vmin) {
        if (faceTangentsWindBackward(f)) {
          hp[0]=p[0];
          hp[1]=1.-p[0]-p[1];
        } else {
          hp[0]=1.-p[0]-p[1];
          hp[1]=p[0];
        }
        break;
      }
      if (faceTangentsWindBackward(f)) {
        hp[0]=p[1];
        hp[1]=p[0];
      } else hp=p;
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::homogeneousCoordinateDerivative(int f,const Point<2> &p,
Tensor<2,2> &dhpdp) const {
  switch (f) {
    case 0:
      if (faceTangentsWindBackward(f)) {
        dhpdp[0][0]=0.;
        dhpdp[0][1]=1.;
        dhpdp[1][0]=1.;
        dhpdp[1][1]=0.;
      } else {
        dhpdp[0][0]=1.;
        dhpdp[0][1]=0.;
        dhpdp[1][0]=0.;
        dhpdp[1][1]=1.;
      }
      break;
    case 1:
    case 2:
      dhpdp[0][0]=1.;
      dhpdp[0][1]=0.;
      dhpdp[1][0]=0.;
      dhpdp[1][1]=1.;
      break;
    case 3: {
      int vmin=min(vertexIndex(1),min(vertexIndex(2),vertexIndex(3)));
      if (vertexIndex(1)==vmin) {
        if (faceTangentsWindBackward(f)) {
          dhpdp[0][0]=-1.;
          dhpdp[0][1]=-1.;
          dhpdp[1][0]=0.;
          dhpdp[1][1]=1.;
        } else {
          dhpdp[0][0]=0.;
          dhpdp[0][1]=1.;
          dhpdp[1][0]=-1.;
          dhpdp[1][1]=-1.;
        }
        break;
      } else if (vertexIndex(2)==vmin) {
        if (faceTangentsWindBackward(f)) {
          dhpdp[0][0]=1.;
          dhpdp[0][1]=0.;
          dhpdp[1][0]=-1.;
          dhpdp[1][1]=-1.;
        } else {
          dhpdp[0][0]=-1.;
          dhpdp[0][1]=-1.;
          dhpdp[1][0]=1.;
          dhpdp[1][1]=0.;
        }
        break;
      }
      if (faceTangentsWindBackward(f)) {
        dhpdp[0][0]=0.;
        dhpdp[0][1]=1.;
        dhpdp[1][0]=1.;
        dhpdp[1][1]=0.;
      } else {
        dhpdp[0][0]=1.;
        dhpdp[0][1]=0.;
        dhpdp[1][0]=0.;
        dhpdp[1][1]=1.;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Tetrahedron::lineFromFace(const Shape2 *face,int f,int fl) const {
  CHECK_SAME(face->elementDataIndex(),codimension1ShapeIndex(f));
  switch (f) {
    case 0: {
      int lines[3]={3,2,5};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv2=face->vertexEqualToVertexIndex(vertexIndex(2));
      if (fv2==(fv0+1)%3) return lines[(fl+fv0)%3];
      else return lines[(2+fv2-fl)%3];
    }
    case 1: {
      int lines[3]={3,0,4};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      if (fv1==(fv0+1)%3) return lines[(fl+fv0)%3];
      else return lines[(2+fv1-fl)%3];
    }
    case 2: {
      int lines[3]={2,0,1};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      if (fv1==(fv0+1)%3) return lines[(fl+fv0)%3];
      else return lines[(2+fv1-fl)%3];
    }
    case 3: {
      int lines[3]={4,1,5};
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      int fv2=face->vertexEqualToVertexIndex(vertexIndex(2));
      if (fv2==(fv1+1)%3) return lines[(fl+fv1)%3];
      else return lines[(2+fv2-fl)%3];
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::staticEquallySpacedLatticePoints(int n,
NumPtr<Point<3> > &points) {
  points.allocate(staticNumberEquallySpacedLatticePoints(n));
  Point<3> p(0.,0.,0.);
  REAL s=(n>0 ? 1./static_cast<REAL>(n) : 0.);
  int l=0;
  for (int k=0;k<=n;k++,p[2]+=s) {
    for (int j=0;j<=n-k;j++,p[1]+=s) {
      for (int i=0;i<=n-j-k;i++,p[0]+=s,l++) points[l]=p;
      p[0]=0.;
    }
    p[1]=0.;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Tetrahedron::staticEquallySpacedLatticePointIndexOfVertex(int v,
int n) {
  CHECK_TEST(n>0);
  CHECK_TEST(v<4);
  switch (v) {
    case 0:
      return 0;
    case 1: 
      return n;
    case 2:
      return ((n+1)*(n+2))/2-1;
    case 3:
      return ((n+1)*(n+2)*(n+3))/6-1;
    default:
      OBSOLETE("bad vertex");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::staticEquallySpacedLatticePointIndicesOfFace(int f,
int n,NumPtr<int> &indices) {
  indices.allocate(Triangle::staticNumberEquallySpacedLatticePoints(n));
  switch (f) {
    case 0: { // (0,j,k)
      int ijk=0;
      int jk=0;
      for (int k=0;k<=n;k++) {
        for (int j=0;j+k<=n;j++,jk++) {
          indices[jk]=ijk;
          ijk+=n+1-j-k;
        }
      }
      break;
    }
    case 1: { // (i,0,k)
      int ijk=0;
      int ik=0;
      for (int k=0;k<=n;k++) {
        for (int i=0;i+k<=n;i++,ik++) {
          indices[ik]=ijk;
          ijk++;
        }
        ijk+=(n*(n+1)+k*(k-1))/2-k*n;
      }
      break;
    }
    case 2: { // (i,j,0)
      int ijk=0;
      int ij=0;
      for (int j=0;j<=n;j++) {
        for (int i=0;i+j<=n;i++,ij++) {
          indices[ij]=ijk;
          ijk++;
        }
      }
      break;
    }
    case 3: { // (n-j-k,j,k)
      int ijk=n;
      int jk=0;
      for (int k=0;k<=n;k++) {
        for (int j=0;j+k<=n;j++,jk++) {
          indices[jk]=ijk;
          ijk+=n-j-k;
        }
        ijk+=n-k;
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::equallySpacedLatticePointIndicesOfFace(int f,int n,
NumPtr<int> &indices) const {
  indices.allocate(Triangle::staticNumberEquallySpacedLatticePoints(n));
  switch (f) {
    case 0: { // (0,j,k)
      if (vertexIndex(2)<vertexIndex(3)) {
        int ijk=0;
        int jk=0;
        for (int k=0;k<=n;k++) {
          for (int j=0;j+k<=n;j++,jk++) {
            indices[jk]=ijk;
            if (j+k<n) ijk+=n+1-j-k;
          }
          ijk++;
        }
      } else {
        int ijk=0;
        int jk=0;
        for (int j=0;j<=n;j++) {
          for (int k=0;j+k<=n;k++,jk++) {
            indices[jk]=ijk;
            if (j+k<n) ijk+=(n*(n+3)+k*(k-3))/2+1-j-k*n;
          }
          ijk-=(n*(n+1)*(n+5)-j*(j*j+5))/6-j*n-1;
        }
      }
      break;
    }
    case 1: { // (i,0,k)
      int ijk=0;
      int ik=0;
      for (int k=0;k<=n;k++) {
        for (int i=0;i+k<=n;i++,ik++) {
          indices[ik]=ijk;
          if (i+k<n) ijk++;
        }
        ijk+=(n*(n+1)+k*(k-1))/2+1-k*n;
      }
      break;
    }
    case 2: { // (i,j,0)
      int ijk=0;
      int ij=0;
      for (int j=0;j<=n;j++) {
        for (int i=0;i+j<=n;i++,ij++) {
          indices[ij]=ijk;
          if (i+j<n) ijk++;
        }
        ijk++;
      }
      break;
    }
    case 3: {
      if (vertexIndex(2)<vertexIndex(3)) { // (n-j-k,j,k)
        int ijk=n;
        int jk=0;
        for (int k=0;k<=n;k++) {
          for (int j=0;j+k<=n;j++,jk++) {
            indices[jk]=ijk;
            if (j+k<n) ijk+=n-j-k;
          }
          ijk+=n-k;
        }
      } else { // (n-j-k,j,k)
        int ijk=n;
        int jk=0;
        for (int j=0;j<=n;j++) {
          for (int k=0;j+k<=n;k++,jk++) {
            indices[jk]=ijk;
            if (j+k<n) ijk+=(n*(n+3)+k*(k-3))/2-k*n-j;
          }
          ijk-=(n*(n*(n+6)-1)-j*(j*j-1))/6-j*n;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(
int f,int n,NumPtr<int> &indices) {
  indices.allocate(
    Triangle::staticNumberInteriorEquallySpacedLatticePoints(n));
  switch (f) {
    case 0: { // (0,j,k)
      int ijk=1+(n*(n+5))/2;
      int jk=0;
      for (int k=1;k<n;k++) {
        for (int j=1;j+k<n;j++,jk++) {
          indices[jk]=ijk;
          ijk+=n+1-j-k;
        }
        ijk+=n+1-k;
      }
      break;
    }
    case 1: { // (i,0,k)
      int ijk=2+(n*(n+3))/2;
      int ik=0;
      for (int k=1;k<n;k++) {
        for (int i=1;i+k<n;i++,ik++) {
          indices[ik]=ijk;
          ijk++;
        }
        ijk+=(n*(n+1)+k*(k-1))/2+2-k*n;
      }
      break;
    }
    case 2: { // (i,j,0)
      int ijk=n+2;
      int ij=0;
      for (int j=1;j<n;j++) {
        for (int i=1;i+j<n;i++,ij++) {
          indices[ij]=ijk;
          ijk++;
        }
        ijk+=2;
      }
      break;
    }
    case 3: { // (n-j-k,j,k)
      int ijk=(n*(n+7))/2-1;
      int jk=0;
      for (int k=1;k<n;k++) {
        for (int j=1;j+k<n;j++,jk++) {
          indices[jk]=ijk;
          ijk+=n-j-k;
        }
        ijk+=2*(n-k)-1;
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::staticEquallySpacedLatticePointIndicesOfLine(int l,
int n,NumPtr<int> &indices) {
  indices.allocate(Interval::staticNumberEquallySpacedLatticePoints(n));
  switch (l) {
    case 0: {
      int ijk=0;
      for (int i=0;i<=n;i++) {
        indices[i]=ijk;
        ijk++;
      }
      break;
    }
    case 1: {
      int ijk=n;
      for (int j=0;j<=n;j++) {
        indices[j]=ijk;
        ijk+=n-j;
      }
      break;
    }
    case 2: {
      int ijk=0;
      for (int j=0;j<=n;j++) {
        indices[j]=ijk;
        ijk+=n+1-j;
      }
      break;
    }
    case 3: {
      int ijk=0;
      for (int k=0;k<=n;k++) {
        indices[k]=ijk;
        ijk+=(n*(n+3)+k*(k-3))/2+1-k*n;
      }
      break;
    }
    case 4: {
      int ijk=n;
      for (int k=0;k<=n;k++) {
        indices[k]=ijk;
        ijk+=(n*(n+3)+k*(k-3))/2-k*n;
      }
      break;
    }
    case 5: {
      int ijk=(n*(n+3))/2;
      for (int k=0;k<=n;k++) {
        indices[k]=ijk;
        ijk+=(n*(n+1)+k*(k-1))/2-k*n;
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_L::constraints
//for Mapping, this is superceded by lenoirSupportPointsOnTetrahedron
void Tetrahedron::equallySpacedLatticePointIndicesOfFaceInterior(
int f,int n,NumPtr<int> &indices) const {
  indices.allocate(((n-2)*(n-1))/2);
  if (n<=2) return;
  switch (f) {
    case 0: { // (0,j,k)
      if (vertexIndex(2)<vertexIndex(3)) {
        int ijk=((n+1)*(n+2))/2+n;
        int jk=0;
        for (int k=1;k<n;k++) {
          for (int j=1;j+k<n;j++,jk++) {
            indices[jk]=ijk;
            ijk+=n+1-j-k;
          }
          ijk+=n+1-k;
        }
      } else {
        int ijk=((n+1)*(n+2))/2+n;
        int jk=0;
        for (int j=1;j<n;j++) {
          for (int k=1;j+k<n;k++,jk++) {
            indices[jk]=ijk;
            ijk+=(n*(n+3)+k*(k-3))/2+1-j-k*n;
          }
          ijk-=(n*(n+4)*(n-1)-j*(j+1)*(j-1))/6-j*n-1;
        }
      }
      break;
    }
    case 1: { // (i,0,k)
      int ijk=((n+1)*(n+2))/2+1; 
      int ik=0;
      for (int k=1;k<n;k++) {
        for (int i=1;i+k<n;i++,ik++) {
          indices[ik]=ijk;
          ijk++;
        }
        ijk+=(n*(n+1)+k*(k-1))/2+2-k*n;
      }
      break;
    }
    case 2: { // (i,j,0)
      int ijk=n+2;
      int ij=0;
      for (int j=1;j<n;j++) {
        for (int i=1;i+j<n;i++,ij++) {
          indices[ij]=ijk;
          ijk++;
        }
        ijk+=2;
      }
      break;
    }
    case 3: {
      if (vertexIndex(2)<vertexIndex(3)) { // (n-j-k,j,k)
        int ijk=(n*(n+7))/2-1;
        int jk=0;
        for (int k=1;k<n;k++) {
          for (int j=1;j+k<n;j++,jk++) {
            indices[jk]=ijk;
            ijk+=n-j-k;
          }
          ijk+=2*(n-k)-1;
        }
      } else { // (n-j-k,j,k)
        int ijk=(n*(n+7))/2-1;
        int jk=0;
        for (int j=1;j<n;j++) {
          for (int k=1;j+k<n;k++,jk++) {
            indices[jk]=ijk;
            ijk+=(n*(n+3)+k*(k-3))/2-k*n-j;
          }
          ijk-=(n*(n+5)*(n-2)+j*(7-j*j))/6+1-j*n;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::staticEquallySpacedLatticePointIndicesOfLineInterior(
int l,int n,NumPtr<int> &indices) {
  indices.allocate(n-1);
  if (n<=1) return;
  switch (l) {
    case 0: {
      int ijk=1;
      for (int i=1;i<n;i++) {
        indices[i-1]=ijk;
        ijk++;
      }
      break;
    }
    case 1: {
      int ijk=2*n;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n-j;
      }
      break;
    }
    case 2: {
      int ijk=n+1;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n+1-j;
      }
      break;
    }
    case 3: {
      int ijk=((n+1)*(n+2))/2;
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+3)+k*(k-3))/2+1-k*n;
      }
      break;
    }
    case 4: {
      int ijk=(n*(n+5))/2;
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+3)+k*(k-3))/2-k*n;
      }
      break;
    }
    case 5: {
      int ijk=n*(n+2);
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+1)+k*(k-1))/2-k*n;
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::equallySpacedLatticePointIndicesOfLineInterior(int l,
int n,NumPtr<int> &indices) const {
  indices.allocate(n-1);
  if (n<=1) return;
  switch (l) {
    case 0: {
      int ijk=1;
      for (int i=1;i<n;i++) {
        indices[i-1]=ijk;
        ijk++;
      }
      break;
    }
    case 1: {
      int ijk=2*n;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n-j;
      }
      break;
    }
    case 2: {
      int ijk=n+1;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n+1-j;
      }
      break;
    }
    case 3: {
      int ijk=((n+1)*(n+2))/2;
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+3)+k*(k-3))/2+1-k*n;
      }
      break;
    }
    case 4: {
      int ijk=(n*(n+5))/2;
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+3)+k*(k-3))/2-k*n;
      }
      break;
    }
    case 5: {
      if (vertexIndex(2)<vertexIndex(3)) {
        int ijk=n*(n+2);
        for (int k=1;k<n;k++) {
          indices[k-1]=ijk;
          ijk+=(n*(n+1)+k*(k-1))/2-k*n;
        }
      } else {
        int ijk=(n*(n*(n+6)+11))/6-1;
        for (int j=1;j<n;j++) {
          indices[j-1]=ijk;
          ijk-=(j*(j+3))/2+1;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::reorderLineDofs(int n,NumPtr<int> &renumber) const {
  int nl=n+1;
  for (int i=0;i<6*nl;i++) renumber[i]=i;
  int offset=0;
  NumPtr<int> r=renumber;
  for (int l=0;l<6;l++) {
    int v[2]={vertexIndex(elementVertexFromLine(l,0)),
              vertexIndex(elementVertexFromLine(l,1))};
    Interval line_element(v,-1);
    line_element.reorderDofs(n,r);
    for (int i=0;i<nl;i++) r[i]+=offset;
    r+=nl;
    offset+=nl;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<6*nl;i++) {
    if (renumber[i]>=6*nl || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(6*nl);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<6*nl;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<6*nl;i++) CHECK_BOUNDS(renumber_inverse[i],0,6*nl);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::reorderHdivFaceDofs(int line_offset,int n,
NumPtr<int> &renumber) const {
  int nt=((n+2)*(n+1))/2;
  for (int i=0;i<4*nt;i++) renumber[i]=i;
  NumPtr<int> r=renumber;
  int l[3]={0,1,2};
  int offset=line_offset;
  for (int f=0;f<4;f++) {
    int v[3]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2))};
    Triangle face_element(l,v,-1);
    face_element.reorderDofs(n,r);
    for (int i=0;i<nt;i++) r[i]+=offset;
    r+=nt;
    offset+=nt;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<4*nt;i++) {
    if (renumber[i]-line_offset>=4*nt || renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(4*nt);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<4*nt;i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<4*nt;i++) CHECK_BOUNDS(renumber_inverse[i],0,4*nt);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::reorderHcurlFaceDofs(int line_offset,int n,
NumPtr<int> &renumber) const {
  int nt=((n+2)*(n+1))/2;
  for (int i=0;i<8*nt;i++) renumber[i]=i;
  NumPtr<int> r=renumber;
  int l[3]={0,1,2};
  int offset=line_offset;
  for (int f=0;f<4;f++) {
    int v[3]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2))};
    Triangle face_element(l,v,-1);
    face_element.reorderDofs(n,r);
    if (!faceTangentsWindBackward(f)) {
      for (int i=0;i<nt;i++) {
        r[i]+=offset;
        r[i+nt]=r[i]+nt;
      }
    } else {
      for (int i=0;i<nt;i++) {
        r[i]+=nt+offset;
        r[i+nt]=r[i]-nt;
      }
    }
    r+=2*nt;
    offset+=2*nt;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<8*nt;i++) {
    if (renumber[i]-line_offset>=8*nt || renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(8*nt);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<8*nt;i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<8*nt;i++) CHECK_BOUNDS(renumber_inverse[i],0,8*nt);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::reorderHcurlFaceDofs2(int line_offset,int n,
NumPtr<int> &renumber) const {
  int nt=((n+2)*(n+1))/2;
  int nh=n+1; // homogeneous polynomials;
  for (int i=0;i<4*(2*nt+nh);i++) renumber[i]=i;
  NumPtr<int> r=renumber;
  int l[3]={0,1,2};
  int offset=line_offset;
  for (int f=0;f<4;f++) {
    int v[3]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2))};
    Triangle face_element(l,v,-1);
    face_element.reorderDofs(n,r);
//  if (dot<0.)
    if (!faceTangentsWindBackward(f)) {
      for (int i=0;i<nt;i++) {
        r[i]+=offset;
        r[i+nt]=r[i]+nt;
      }
    } else {
      for (int i=0;i<nt;i++) {
        r[i+nt]=r[i]+offset;
        r[i]+=offset+nt;
      }
    }
    r+=2*nt;
    offset+=2*nt;
//  We required that 
//    Nedelec2SimplexDofShapeFunction::computeFaceValues
//    depends on the Tetrahedron
    for (int i=0;i<nh;i++) r[i]=offset+i;
    r+=nh;
    offset+=nh;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<4*(2*nt+nh);i++) {
    if (renumber[i]-line_offset>=4*(2*nt+nh) || renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(4*(2*nt+nh));
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<4*(2*nt+nh);i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<4*(2*nt+nh);i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,4*(2*nt+nh));
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::reorderEquallySpacedLatticePoints(int n,
NumPtr<int> &renumber) const {
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
//vertices first
  int m=0;
  renumber[0]=m++;
  renumber[n]=m++;
  renumber[(n*(n+3))/2]=m++;
  renumber[renumber.getNumber()-1]=m++;
//line 0
  if (vertexIndex(0)<vertexIndex(1)) {
    for (int i=1;i<n;i++) renumber[i]=m++;
  } else {
    int l=n-1;
    for (int i=n-1;i>0;i--) renumber[l--]=m++;
  } 
//line 1
  if (vertexIndex(1)<vertexIndex(2)) {
    int l=2*n;
    for (int j=1;j<n;j++) {
      renumber[l]=m++;
      l+=n-j;
    }
  } else {
    int l=(n*(n+3))/2-1;
    for (int i=1;i<n;i++) {
      renumber[l]=m++;
      l-=i+1;
    }
  } 
//line 2
  if (vertexIndex(0)<vertexIndex(2)) {
    int l=n+1;
    for (int j=1;j<n;j++) {
      renumber[l]=m++;
      l+=n-j+1;
    }
  } else {
    int l=(n*(n+3))/2-2;
    for (int j=n-1;j>0;j--) {
      renumber[l]=m++;
      l-=n-j+2;
    }
  } 
//line 3
  if (vertexIndex(0)<vertexIndex(3)) {
    int l=((n+1)*(n+2))/2;
    for (int k=1;k<n;k++) {
      renumber[l]=m++;
      l+=((n-k+1)*(n-k+2))/2;
    }
  } else {
    int l=renumber.getNumber()-4;
    for (int k=n-1;k>0;k--) {
      renumber[l]=m++;
      l-=((n-k+2)*(n-k+3))/2;
    }
  } 
//line 4
  if (vertexIndex(1)<vertexIndex(3)) {
    int l=(n*(n+5))/2;
    for (int k=1;k<n;k++) {
      renumber[l]=m++;
      l+=((n+1-k)*(n+2-k))/2-1;
    }
  } else {
    int l=renumber.getNumber()-3;
    for (int i=1;i<n;i++) {
      renumber[l]=m++;
      l-=((i+2)*(i+3))/2-1;
    }
  } 
//line 5
  if (vertexIndex(2)<vertexIndex(3)) {
    int l=n*(n+2);
    for (int k=1;k<n;k++) {
      renumber[l]=m++;
      l+=((n-k)*(n-k+1))/2;
    }
  } else {
    int l=renumber.getNumber()-2;
    for (int k=n-1;k>0;k--) {
      renumber[l]=m++;
      l-=((n-k+1)*(n-k+2))/2;
    }
  } 
//face 0: i=0
//  a(j-1,k+1) = a(j,k)+((n-k)*(n-k+1))/2
//  a(j  ,k+1) = a(j,k)+((n-k+1)*(n-k+2))/2-j
//  a(j-1,k  ) = a(j,k)-(n-j-k+2)
//  a(j+1,k  ) = a(j,k)+(n-j-k+1)
//  a(j  ,k-1) = a(j,k)-((n-k+2)*(n-k+3))/2-j
//  a(j+1,k-1) = a(j,k)+((n-k+1)*(n-k+2))/2
  if (vertexIndex(0)<vertexIndex(2)) {
    if (vertexIndex(2) < vertexIndex(3)) { // 0 < 2 < 3
      int l=(n*(n+5))/2+1;
      for (int k=1;k<n-1;k++) {
        int i=l;
        for (int j=1;j<n-k;j++) {
          renumber[i]=m++;
          i+=n-j-k+1;
        }
        l+=((n-k+1)*(n-k+2))/2-1;
      }
    } else if (vertexIndex(0) < vertexIndex(3)) { // 0 < 3 < 2
      int l=(n*(n+5))/2+1;
      for (int j=1;j<n-1;j++) {
        int i=l;
        for (int k=1;k<n-j;k++) {
          renumber[i]=m++;
          i+=((n-k+1)*(n-k+2))/2-j;
        }
        l+=n-j;
      }
    } else { // 3 < 0 < 2
      int l=renumber.getNumber()-7;
      for (int j=1;j<n-1;j++) {
        int i=l;
        for (int k=n-j-1;k>0;k--) {
          renumber[i]=m++;
          i-=((n-k+2)*(n-k+3))/2-j;
        }
        l-=((j+2)*(j+3))/2;
      }
    }
  } else {
    if (vertexIndex(0)<vertexIndex(3)) { // 2 < 0 < 3
      int l=n*(n+2)-2;
      for (int k=1;k<n-1;k++) {
        int i=l;
        for (int j=n-k-1;j>0;j--) {
          renumber[i]=m++;
          i-=n-j-k+2;
        }
        l+=((n-k)*(n-k+1))/2;
      }
    } else if (vertexIndex(2)<vertexIndex(3)) { // 2 < 3 < 0
      int l=n*(n+2)-2;
      for (int j=n-2;j>0;j--) {
        int i=l;
        for (int k=1;k<=j;k++) {
          renumber[i]=m++;
          i+=((n-k)*(n-k+1))/2;
        }
        l-=n-j+1;
      }
    } else { // 3 < 2 < 0
      int l=renumber.getNumber()-7;
      for (int k=n-2;k>0;k--) {
        int i=l;
        for (int j=1;j<=k;j++) { // move along line j+k'=k+1
          renumber[i]=m++;
          i-=((n-k+j)*(n-k+j+1))/2;
        }
        l-=((n-k+2)*(n-k+3))/2-1;
      }
    }
  }
//face 1: j=0
//  a(i-1,k+1) = a(i,k)+((n-k+1)*(n-k+2))/2-1
//  a(i  ,k+1) = a(i,k)+((n-k+1)*(n-k+2))/2
//  a(i-1,k  ) = a(i,k)-1
//  a(i+1,k  ) = a(i,k)+1
//  a(i  ,k-1) = a(i,k)-((n-k+2)*(n-k+3))/2
//  a(i+1,k-1) = a(i,k)+((n-k+2)*(n-k+3))/2+1
  if (vertexIndex(0)<vertexIndex(1)) {
    if (vertexIndex(1) < vertexIndex(3)) { // 0 < 1 < 3
      int l=(n*(n+3))/2+2;
      for (int k=1;k<n-1;k++) {
        int j=l;
        for (int i=1;i<n-k;i++) {
          renumber[j++]=m++;
        }
        l+=((n-k+1)*(n-k+2))/2;
      }
    } else if (vertexIndex(0) < vertexIndex(3)) { // 0 < 3 < 1
      int l=(n*(n+3))/2+2;
      for (int i=1;i<n-1;i++) {
        int j=l;
        for (int k=1;k<n-i;k++) {
          renumber[j]=m++;
          j+=((n-k+1)*(n-k+2))/2;
        }
        l++;
      }
    } else { // 3 < 0 < 1
      int l=renumber.getNumber()-9;
      for (int i=1;i<n-1;i++) {
        int j=l;
        for (int k=n-i-1;k>0;k--) {
          renumber[j]=m++;
          j-=((n-k+2)*(n-k+3))/2;
        }
        l-=((i+3)*(i+4))/2-1;
      }
    }
  } else {
    if (vertexIndex(0)<vertexIndex(3)) { // 1 < 0 < 3
      int l=(n*(n+5))/2-1;
      for (int k=1;k<n-1;k++) {
        int j=l;
        for (int i=n-k;i>0;i--) {
          renumber[j--]=m++;
        }
        l+=((n-k+1)*(n-k+2))/2-1;
      }
    } else if (vertexIndex(1)<vertexIndex(3)) { // 1 < 3 < 0
      int l=(n*(n+5))/2-1;
      for (int i=n-2;i>0;i--) {
        int j=l;
        for (int k=1;k<=i;k++) {
          renumber[j]=m++;
          j+=((n-k+1)*(n-k+2))/2-1;
        }
        l--;
      }
    } else { // 3 < 1 < 0
      int l=renumber.getNumber()-9;
      for (int k=n-2;k>0;k--) { // l corresponds to i=1
        int j=l;
        for (int i=1;i<=k;i++) { // move along line i+k'=k+1
          renumber[j]=m++;
          j-=((n-k+i+1)*(n-k+i+2))/2-1;
        }
        l-=((n-k+2)*(n-k+3))/2;
      }
    }
  }
//face 2: k=0
//  a(i-1,j+1) = a(i,j)+(n-j)
//  a(i  ,j+1) = a(i,j)+(n-j+1)
//  a(i-1,j  ) = a(i,j)-1
//  a(i+1,j  ) = a(i,j)+1
//  a(i  ,j-1) = a(i,j)-(n-j+2)
//  a(i+1,j-1) = a(i,j)-(n-j+1)
  if (vertexIndex(0)<vertexIndex(1)) {
    if (vertexIndex(1) < vertexIndex(2)) { // 0 < 1 < 2
      int l=n+2;
      for (int j=1;j<n-1;j++) {
        int k=l;
        for (int i=1;i<n-j;i++) {
          renumber[k++]=m++;
        }
        l+=n-j+1;
      }
    } else if (vertexIndex(0) < vertexIndex(2)) { // 0 < 2 < 1
      int l=n+2;
      for (int i=1;i<n-1;i++) {
        int k=l;
        for (int j=1;j<n-i;j++) {
          renumber[k]=m++;
          k+=n-j+1;
        }
        l++;
      }
    } else { // 2 < 0 < 1
      int l=(n*(n+3))/2-4;
      for (int i=1;i<n-1;i++) {
        int k=l;
        for (int j=n-i-1;j>0;j--) {
          renumber[k]=m++;
          k-=n-j+2;
        }
        l-=i+2;
      }
    }
  } else {
    if (vertexIndex(0)<vertexIndex(2)) { // 1 < 0 < 2
      int l=2*n-1;
      for (int j=1;j<n-1;j++) {
        int k=l;
        for (int i=1;i<n-j;i++) {
          renumber[k--]=m++;
        }
        l+=n-j;
      }
    } else if (vertexIndex(1)<vertexIndex(2)) { // 1 < 2 < 0
      int l=2*n-1;
      for (int i=n-2;i>0;i--) {
        int k=l;
        for (int j=1;j<=i;j++) {
          renumber[k]=m++;
          k+=n-j;
        }
        l--;
      }
    } else { // 2 < 1 < 0
      int l=(n*(n+3))/2-4;
      for (int j=n-2;j>0;j--) {
        int k=l;
        for (int i=1;i<=j;i++) {
          renumber[k]=m++;
          k-=n+i-j;
        }
        l-=n-j+2;
      }
    }
  }
//face 3: i+j+k=n
//  a(i,j-1,k+1) = a(i,j,k)+((n-k)*(n-k+1))/2
//  a(i-1,j,k+1) = a(i,j,k)+((n-k)*(n-k+1))/2+i
//  a(i-1,j+1,k) = a(i,j,k)+i
//  a(i+1,j-1,k) = a(i,j,k)-(i+1)
//  a(i,j+1,k-1) = a(i,j,k)-((n-k+1)*(n-k+2))/2
//  a(i+1,j,k-1) = a(i,j,k)-((n-k+1)*(n-k+2))/2-(i+1)
  if (vertexIndex(1)<vertexIndex(2)) {
    if (vertexIndex(2) < vertexIndex(3)) { // 1 < 2 < 3
      int l=(n*(n+7))/2-1;
      for (int k=1;k<n-1;k++) {
        int i=l;
        for (int j=1;j<n-k;j++) {
          renumber[i]=m++;
          i+=n-k-j;
        }
        l+=((n-k+1)*(n-k+2))/2-2;
      }
    } else if (vertexIndex(1) < vertexIndex(3)) { // 1 < 3 < 2
      int l=(n*(n+7))/2-1;
      for (int j=1;j<n-1;j++) {
        int i=l;
        for (int k=1;k<n-j;k++) {
          renumber[i]=m++;
          i+=((n-k+1)*(n-k+2))/2-j-1;
        }
        l+=n-j-1;
      }
    } else { // 3 < 1 < 2
      int l=renumber.getNumber()-6;
      for (int j=1;j<n-1;j++) {
        int k=l;
        for (int i=1;i<n-j;i++) {
          renumber[k]=m++;
          k-=((i+j+1)*(i+j+2))/2+i+1;
        }
        l-=((j+2)*(j+3))/2;
      }
    }
  } else {
    if (vertexIndex(1)<vertexIndex(3)) { // 2 < 1 < 3
      int l=n*(n+2)-1;
      for (int k=1;k<n-1;k++) {
        int j=l;
        for (int i=1;i<n-k;i++) {
          renumber[j]=m++;
          j-=i+1;
        }
        l+=((n-k)*(n-k+1))/2;
      }
    } else if (vertexIndex(2)<vertexIndex(3)) { // 2 < 3 < 1
      int l=n*(n+2)-1;
      for (int i=1;i<n-1;i++) {
        int j=l;
        for (int k=1;k<n-i;k++) {
          renumber[j]=m++;
          j+=((n-k)*(n-k+1))/2;
        }
        l-=i+1;
      }
    } else { // 3 < 2 < 1
      int l=renumber.getNumber()-6;
      for (int i=1;i<n-1;i++) {
        int k=l;
        for (int j=1;j<n-i;j++) {
          renumber[k]=m++;
          k-=((i+j+1)*(i+j+2))/2;
        }
        l-=((i+2)*(i+3))/2+i+1;
      }
    }
  }
//interior points last
  int l2=((n+1)*(n+4))/2;
  for (int k=1;k<n;k++) {
    int l1=l2;
    for (int j=1;j<n-k;j++) {
      int l=l1;
      for (int i=1;i<n-k-j;i++) renumber[l++]=m++;
      l1+=n-j-k+1;
    }
    l2+=((n-k+1)*(n-k+2))/2-1;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]==INT_MAX) {
      cout << "\trenumber[" << i << "] = *** " << -1 << " ***" << endl;
      success=false;
    } else if (renumber[i]>=renumber.getNumber()) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::reorderChildEquallySpacedLatticePoints(int n,
const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices,NumPtr<int> &renumber) 
const {
  const Tetrahedron *child_tetrahedron=
    dynamic_cast<const Tetrahedron*>(child_shape);
  CHECK_POINTER(child_tetrahedron);

  int reorder_case=-1;
  int cvi0=-1;
  int cvi1=-1;
  if (c>=0 && c<4) {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    CHECK_TEST(vertexIndex(c)==svi);
    cvi0=child_shape->vertexEqualToVertexIndex(svi);
    CHECK_TEST(cvi0>=0);
    switch (c) {
      case 0: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[0]);
        break;
      }
      case 1: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[0]);
        break;
      }
      case 2: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[1]);
        break;
      }
      case 3: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[3]);
      }
    }
  } else {
    cvi0=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[0]);
    switch (c) {
      case 4: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[3]);
        break;
      }
      case 5: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[1]);
        break;
      }
      case 6: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[1]);
        break;
      }
      case 7: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[2]);
        break;
      }
    }
  }
  CHECK_TEST(cvi0>=0);
  CHECK_TEST(cvi1>=0);
  switch (cvi0) {
    case 0: {
      switch (cvi1) {
        case 1:
          reorder_case=0;
          break;
        case 2:
          reorder_case=1;
          break;
        case 3:
          reorder_case=2;
          break;
      }
      break;
    }
    case 1: {
      switch (cvi1) {
        case 2:
          reorder_case=3;
          break;
        case 3:
          reorder_case=4;
          break;
        case 0:
          reorder_case=5;
          break;
      }
      break;
    }
    case 2: {
      switch (cvi1) {
        case 3:
          reorder_case=6;
          break;
        case 0:
          reorder_case=7;
          break;
        case 1:
          reorder_case=8;
          break;
      }
      break;
    }
    case 3: {
      switch (cvi1) {
        case 0:
          reorder_case=9;
          break;
        case 1:
          reorder_case=10;
          break;
        case 2:
          reorder_case=11;
          break;
      }
      break;
    }
  }
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  switch (reorder_case) {
    case 0: {
      for (int i=0;i<renumber.getNumber();i++) renumber[i]=i;
      break;
    }
    case 1: {
      int ijk=0;
      for (int i=0;i<=n;i++) {
        int nk=i;
        for (int k=0;i+k<=n;k++) {
          int nj=nk;
          for (int j=0;i+j+k<=n;j++) {
            renumber[nj]=ijk++;
            nj+=n+1-j-k;
          }
          nk+=(n*(n+3)+k*(k-3))/2+1-n*k;
        }
      }
      break;
    }
    case 2: {
      int ijk=0;
      int nj=0;
      for (int j=0;j<=n;j++) {
        int ni=nj;
        for (int i=0;i+j<=n;i++) {
          int nk=ni;
          for (int k=0;i+j+k<=n;k++) {
            renumber[nk]=ijk++;
            nk+=(n*(n+3)+k*(k-3))/2+1-n*k-j;
          }
          ni++;
        }
        nj+=n+1-j;
      }
      break;
    }
    case 3: {
      int ijk=0;
      int nk=n;
      for (int k=0;k<=n;k++) {
        int nl=nk;
        for (int l=0;k+l<=n;l++) {
          int nj=nl;
          for (int j=0;j+k+l<=n;j++) {
            renumber[nj]=ijk++;
            nj+=n-j-k;
          }
          nl--;
        }
        nk+=(n*(n+3)+k*(k-3))/2-n*k;
      }
      break;
    }
    case 4: {
      int ijk=0;
      int nl=n;
      for (int l=0;l<=n;l++) {
        int nj=nl;
        for (int j=0;j+l<=n;j++) {
          int nk=nj;
          for (int k=0;j+k+l<=n;k++) {
            renumber[nk]=ijk++;
            nk+=(n*(n+3)+k*(k-3))/2-j-n*k;
          }
          nj+=n-j;
        }
        nl--;
      }
      break;
    }
    case 5: {
      int ijk=0;
      int nj=n;
      for (int j=0;j<=n;j++) {
        int nk=nj;
        for (int k=0;j+k<=n;k++) {
          int nl=nk;
          for (int l=0;j+k+l<=n;l++) {
            renumber[nl]=ijk++;
            nl--;
          }
          nk+=(n*(n+3)+k*(k-3))/2-j-n*k;
        }
        nj+=n-j;
      }
      break;
    }
    case 6: {
      int ijk=0;
      int ni=(n*(n+3))/2;
      for (int i=0;i<=n;i++) {
        int nl=ni;
        for (int l=0;i+l<=n;l++) {
          int nk=nl;
          for (int k=0;i+k+l<=n;k++) {
            renumber[nk]=ijk++;
            nk+=(n*(n+1)+k*(k-1))/2-n*k;
          }
          nl-=i+l+2;
        }
        ni-=i+1;
      }
      break;
    }
    case 7: {
      int ijk=0;
      int nk=(n*(n+3))/2;
      for (int k=0;k<=n;k++) {
        int ni=nk;
        for (int i=0;i+k<=n;i++) {
          int nl=ni;
          for (int l=0;i+k+l<=n;l++) {
            renumber[nl]=ijk++;
            nl-=i+l+2;
          }
          ni-=i+1;
        }
        nk+=(n*(n+1)+k*(k-1))/2-n*k;
      }
      break;
    }
    case 8: {
      int ijk=0;
      int nl=(n*(n+3))/2;
      for (int l=0;l<=n;l++) {
        int nk=nl;
        for (int k=0;k+l<=n;k++) {
          int ni=nk;
          for (int i=0;i+k+l<=n;i++) {
            renumber[ni]=ijk++;
            ni-=i+l+1;
          }
          nk+=(n*(n+1)+k*(k-1))/2-n*k;
        }
        nl-=l+2;
      }
      break;
    }
    case 9: {
      int ijk=0;
      int ni=n*(11+n*(6+n))/6;
      for (int i=0;i<=n;i++) {
        int nj=ni;
        for (int j=0;i+j<=n;j++) {
          int nl=nj;
          for (int l=0;i+j+l<=n;l++) {
            renumber[nl]=ijk++;
            nl-=(i*(i+5)+j*(j+3)+l*(l+5))/2+3+i*l+i*j+j*l;
          }
          nj-=(i*(i+3)+j*(j+3))/2+1+i*j;
        }
        ni-=(i*(i+5))/2+2;
      }
      break;
    }
    case 10: {
      int ijk=0;
      int nj=n*(11+n*(6+n))/6;
      for (int j=0;j<=n;j++) {
        int nl=nj;
        for (int l=0;j+l<=n;l++) {
          int ni=nl;
          for (int i=0;i+j+l<=n;i++) {
            renumber[ni]=ijk++;
            ni-=(i*(i+5)+j*(j+3)+l*(l+5))/2+2+i*l+i*j+j*l;
          }
          nl-=(j*(j+3)+l*(l+5))/2+3+j*l;
        }
        nj-=(j*(j+3))/2+1;
      }
      break;
    }
    case 11: {
      int ijk=0;
      int nl=n*(11+n*(6+n))/6;
      for (int l=0;l<=n;l++) {
        int ni=nl;
        for (int i=0;i+l<=n;i++) {
          int nj=ni;
          for (int j=0;i+j+l<=n;j++) {
            renumber[nj]=ijk++;
            nj-=(i*(i+3)+j*(j+3)+l*(l+3))/2+1+i*j+i*l+j*l;
          }
          ni-=(i*(i+5)+l*(l+5))/2+2+i*l;
        }
        nl-=(l*(l+5))/2+3;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_H::constraints
void Tetrahedron::reorderChildDofPoints(int n,const Shape *child_shape,
int c,const NumPtr<int> &line_interior_child_vertices,
NumPtr<int> &renumber) const {
  const Tetrahedron *child_tetrahedron=
    dynamic_cast<const Tetrahedron*>(child_shape);
  CHECK_POINTER(child_tetrahedron);

  int reorder_case=-1;
  int cvi0=-1;
  int cvi1=-1;
  if (c>=0 && c<4) {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    CHECK_TEST(vertexIndex(c)==svi);
    cvi0=child_shape->vertexEqualToVertexIndex(svi);
    CHECK_TEST(cvi0>=0);
    switch (c) {
      case 0: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[0]);
        break;
      }
      case 1: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[0]);
        break;
      }
      case 2: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[1]);
        break;
      }
      case 3: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[3]);
      }
    }
  } else {
    cvi0=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[0]);
    switch (c) {
      case 4: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[3]);
        break;
      }
      case 5: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[1]);
        break;
      }
      case 6: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[1]);
        break;
      }
      case 7: {
        cvi1=child_shape->
          vertexEqualToVertexIndex(line_interior_child_vertices[2]);
        break;
      }
    }
  }
  CHECK_TEST(cvi0>=0);
  CHECK_TEST(cvi1>=0);
  switch (cvi0) {
    case 0: {
      switch (cvi1) {
        case 1:
          reorder_case=0;
          break;
        case 2:
          reorder_case=1;
          break;
        case 3:
          reorder_case=2;
          break;
      }
      break;
    }
    case 1: {
      switch (cvi1) {
        case 2:
          reorder_case=3;
          break;
        case 3:
          reorder_case=4;
          break;
        case 0:
          reorder_case=5;
          break;
      }
      break;
    }
    case 2: {
      switch (cvi1) {
        case 3:
          reorder_case=6;
          break;
        case 0:
          reorder_case=7;
          break;
        case 1:
          reorder_case=8;
          break;
      }
      break;
    }
    case 3: {
      switch (cvi1) {
        case 0:
          reorder_case=9;
          break;
        case 1:
          reorder_case=10;
          break;
        case 2:
          reorder_case=11;
          break;
      }
      break;
    }
  }

  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  switch (reorder_case) {
    case 0: {
      for (int i=0;i<renumber.getNumber();i++) renumber[i]=i;
      break;
    }
    case 1: {
      int ijk=0;
      renumber[ijk++]=0;
      renumber[ijk++]=2;
      renumber[ijk++]=3;
      renumber[ijk++]=1;
      int d=2*n+2; // line 2
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++; // (0,j,0)
      }
      d=5*n-1; // line 5
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++; // (0,n-k,k)
      }
      d=3*n+1; // line 3
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++; // (0,0,k)
      }
      d=4; // line 0
      for (int i=1;i<=n-1;i++) {
        renumber[ijk++]=d++; // (i,0,0)
      }
      d=2*n+1; // line 1
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--; // (n-j,j,0)
      }
      d=5*n-2; // line 4
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      {
        int di=(n*(n+9))/2-1; // face 1: 0,3,1
        for (int i=1;i<=n-2;i++) {
          int dk=di;
          for (int k=1;i+k<=n-1;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-1;
          }
          di++;
        }
      }
      {
        int di=n*(n+3); // face 2: 0,2,1
        for (int i=1;i<=n-2;i++) {
          int dj=di;
          for (int j=1;i+j<=n-1;j++) {
            renumber[ijk++]=dj;
            dj+=n-j-1;
          }
          di++;
        }
      }
      {
        int dk=6*n-2; // face 0: 0,2,3
        for (int k=1;k<=n-2;k++) {
          int dj=dk;
          for (int j=1;j+k<=n-1;j++) {
            renumber[ijk++]=dj;
            dj++;
          }
          dk+=n-k-1;
        }
      }
      {
        int dj=(n*(3*n+5))/2-2; // face 3: 2,3,1
        for (int j=n-2;j>=1;j--) {
          int dk=dj;
          for (int k=1;j+k<=n-1;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-2;
          }
          dj--;
        }
      }
      int di=2*n*n+2; // interior: 0,2,3,1
      for (int i=1;i<=n-3;i++) {
        int dk=di;
        for (int k=1;i+k<=n-2;k++) {
          int dj=dk;
          for (int j=1;i+j+k<=n-1;j++) {
            renumber[ijk++]=dj;
            dj+=n-j-k;
          }
          dk+=(n*(n-1)+k*(k+1))/2-n*k;
        }
        di++;
      }
      break;
    }
    case 2: {
      int ijk=0;
      renumber[ijk++]=0;
      renumber[ijk++]=3;
      renumber[ijk++]=1;
      renumber[ijk++]=2;
      int d=3*n+1; // line 3
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=5*n-2; // line 4
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=4; // line 0
      for (int i=1;i<=n-1;i++) {
        renumber[ijk++]=d++;
      }
      d=2*n+2; // line 2
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      d=6*n-3; // line 5
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=n+3; // line 1
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      {
        int dj=n*(n+3); // face 2: 0,1,2
        for (int j=1;j<=n-2;j++) {
          int di=dj;
          for (int i=1;i+j<=n-1;i++) {
            renumber[ijk++]=di;
            di++;
          }
          dj+=n-j-1;
        }
      }
      {
        int dj=6*n-2; // face 0: 0,3,2
        for (int j=1;j<=n-2;j++) {
          int dk=dj;
          for (int k=1;j+k<=n-1;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-1;
          }
          dj++;
        }
      }
      {
        int di=(n*(n+9))/2-1; // face 1: 0,3,1
        for (int i=1;i<=n-2;i++) {
          int dk=di;
          for (int k=1;i+k<=n-1;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-1;
          }
          di++;
        }
      }
      {
        int dj=2*n*n+1; // face 3: 3,1,2
        for (int j=1;j<=n-2;j++) {
          int dk=dj;
          for (int k=n-j-1;k>=1;k--) {
            renumber[ijk++]=dk;
            dk-=n-k;
          }
          dj-=j;
        }
      }
      int dj=2*n*n+2; // interior: 0,3,1,2
      for (int j=1;j<=n-3;j++) {
        int di=dj;
        for (int i=1;i+j<=n-2;i++) {
          int dk=di;
          for (int k=1;i+j+k<=n-1;k++) {
            renumber[ijk++]=dk;
            dk+=(n*(n-1)+k*(k+1))/2-k*n-j+1;
          }
          di++;
        }
        dj+=n-j-1;
      }
      break;
    }
    case 3: { // 1,2,0,3
      int ijk=0;
      renumber[ijk++]=1;
      renumber[ijk++]=2;
      renumber[ijk++]=0;
      renumber[ijk++]=3;
      int d=n+3; // line 1
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      d=3*n; // line 2
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=n+2; // line 0
      for (int i=n-1;i>=1;i--) {
        renumber[ijk++]=d--;
      }
      d=4*n; // line 4
      for (int k=1;k<n;k++) {
        renumber[ijk++]=d++;
      }
      d=5*n-1; // line 5
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=3*n+1; // line 3
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      {
        int dk=(n*(n+11))/2-3; // face 1: 1,0,3
        for (int k=1;k<=n-2;k++) {
          int di=dk;
          for (int i=n-k-1;i>=1;i--) {
            renumber[ijk++]=di;
            di--;
          }
          dk+=n-k-2;
        }
      }
      {
        int dk=(n*(3*n+3))/2+1; // face 3: 1,2,3
        for (int k=1;k<=n-2;k++) {
          int dj=dk;
          for (int j=1;j+k<=n-1;j++) {
            renumber[ijk++]=dj;
            dj++;
          }
          dk+=n-k-1;
        }
      }
      {
        int di=n*(n+4)-2; // face 2: 1,2,0
        for (int i=n-2;i>=1;i--) {
          int dj=di;
          for (int j=1;i+j<=n-1;j++) {
            renumber[ijk++]=dj;
            dj+=n-j-2;
          }
          di--;
        }
      }
      {
        int dk=7*n-4; // face 0: 2,0,3
        for (int k=1;k<=n-2;k++) {
          int dj=dk;
          for (int j=n-k-1;j>=1;j--) {
            renumber[ijk++]=dj;
            dj--;
          }
          dk+=n-k-2;
        }
      }
      int dk=(n*(3*n+7))/2-8; // interior: 1,2,0,3
      for (int k=1;k<=n-3;k++) {
        int di=dk;
        for (int i=n-k-2;i>=1;i--) {
          int dj=di;
          for (int j=1;j<i;j++) {
            renumber[ijk++]=dj;
            dj+=n-j-k-1;
          }
          di--;
        }
        dk+=(n*(n-1)+k*(k+1))/2-k*n-1;
      }
      break;
    }
    case 4: { // 1,3,2,0
      int ijk=0;
      renumber[ijk++]=1;
      renumber[ijk++]=3;
      renumber[ijk++]=2;
      renumber[ijk++]=0;
      int d=4*n; // line 4
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=6*n-3; // line 5
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=n+3; // line 1
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      d=n+2; // line 0
      for (int i=n-1;i>=1;i++) {
        renumber[ijk++]=d--;
      }
      d=4*n-1; // line 3
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=3*n; // line 2
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      {
        int di=n*(n+4)-2; // face 2: 1,2,0
        for (int i=n-2;i>=1;i--) {
          int dj=di;
          for (int j=1;i+j<=n-1;j++) {
            renumber[ijk++]=dj;
            dj+=n-j-2;
          }
          di--;
        }
      }
      {
        int di=(n*(n+11))/2-3; // face 1: 1,3,0
        for (int i=n-1;i>=1;i--) {
          int dk=di;
          for (int k=1;i<=i;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-2;
          }
          di--;
        }
      }
      {
        int dj=(n*(3*n+3))/2+1; // face 3: 1,3,2
        for (int j=1;j<=n-2;j++) {
          int dk=dj;
          for (int k=1;j+k<=n-1;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-1;
          }
          dj++;
        }
      }
      {
        int dk=(n*(n+9))/2-1; // face 0: 3,2,0
        for (int k=n-1;k>=1;k--) {
          int dj=dk;
          for (int j=1;j<=k;j++) {
            renumber[ijk++]=dj;
            dj-=n+j-k-2;
          }
          dk-=n-k;
        }
      }
      int di=n*(2*n+1)-2; // interior: 1,3,2,0
      for (int i=n-3;i>=1;i--) {
        int dj=di;
        for (int j=1;j<=i;j++) {
          int dk=dj;
          for (int k=1;k<=i;k++) {
            renumber[ijk++]=dj;
            dk+=(n*(n-1)+k*(k+1))/2-k*n-j;
          }
          dj+=n-j-2;
        }
        di--;
      }
      break;
    }
    case 5: { // 1,0,3,2
      int ijk=0;
      renumber[ijk++]=1;
      renumber[ijk++]=0;
      renumber[ijk++]=3;
      renumber[ijk++]=2;
      int d=n+2; // line 0
      for (int i=n-1;i>=1;i--) {
        renumber[ijk++]=d--;
      }
      d=3*n+1; // line 3
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=4*n; // line 4
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=n+3; // line 1
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      d=2*n+2; // line 2
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      d=6*n-3; // line 5
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      {
        int dk=(n*(3*n+3))/2+1; // face 3: 1,2,3
        for (int k=1;k<=n-2;k++) {
          int dj=dk;
          for (int j=1;j+k<=n-1;j++) {
            renumber[ijk++]=dj;
            dj++;
          }
          dk+=n-k-1;
        }
      }
      {
        int dj=n*(n+4)-2; // face 2: 1,0,2
        for (int j=1;j<=n-2;j++) {
          int di=dj;
          for (int i=n-j-1;i>=1;i--) {
            renumber[ijk++]=di;
            di--;
          }
          dj+=n-j-2;
        }
      }
      {
        int dk=(n*(n+11))/2-3; // face 1: 1,0,3
        for (int k=1;k<=n-2;k++) {
          int di=dk;
          for (int i=n-k-1;i>=1;i--) {
            renumber[ijk++]=di;
            di--;
          }
          dk+=n-k-2;
        }
      }
      {
        int dj=6*n-2; // face 0: 0,3,2
        for (int j=1;j<=n-2;j++) {
          int dk=dj;
          for (int k=1;j+k<=n-1;k++) {
            renumber[ijk++]=dk;
            dk-=n-k-1;
          }
          dj++;
        }
      }
      int dj=n*(2*n+1)-2; // interior: 1,0,3,2
      for (int j=1;j<=n-3;j++) {
        int dk=dj;
        for (int k=1;j+k<=n-2;k++) {
          int di=dk;
          for (int i=n-j-k-1;i>=1;i--) {
            renumber[ijk++]=di;
            di--;
          }
          dk+=(n*(n-1)+k*(k+1))/2-k*n-j;
        }
        dj+=n-j-2;
      }
      break;
    }
    case 6: { // 2,3,0,1
      int ijk=0;
      renumber[ijk++]=2;
      renumber[ijk++]=3;
      renumber[ijk++]=0;
      renumber[ijk++]=1;
      int d=5*n-1; // line 5
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=4*n-1; // line 3
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=3*n; // line 2
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=2*n+1; // line 1
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=5*n-2; // line 4
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=4; // line 0
      for (int i=1;i<=n-1;i++) {
        renumber[ijk++]=d++;
      }
      {
        int di=(n*(3*n+3))/2+1; // face 2: 2,0,1
        for (int i=1;i<=n-2;i++) {
          int dj=di;
          for (int j=n-i-1;j>=1;j--) {
            renumber[ijk++]=dj;
            dj-=n-j;
          }
          di+=i-1;
        }
      }
      {
        int dj=(n*(3*n+5))/2-2; // face 3: 2,3,1
        for (int j=n-2;j>=1;j--) {
          int dk=dj;
          for (int k=1;k<=j;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-2;
          }
          dj--;
        }
      }
      {
        int dj=7*n-4; // face 0: 2,3,0
        for (int j=n-2;j>=1;j--) {
          int dk=dj;
          for (int k=1;k<=j;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-2;
          }
          dj--;
        }
      }
      {
        int di=n*(n+3); // face 1: 3,0,1
        for (int i=1;i<=n-2;i++) {
          int dk=di;
          for (int k=n-i-1;k>=1;k--) {
            renumber[ijk++]=dk;
            dk-=n-k;
          }
          di+=i-1;
        }
      }
      int di=(n*(5*n-3))/2; // interior: 2,3,0,1
      for (int i=1;i<=n-3;i++) {
        int dj=di;
        for (int j=n-i-2;i>=1;j--) {
          int dk=dj;
          for (int k=1;k<=j;k++) {
            renumber[ijk++]=dk;
            dk+=(n*(n-3)+k*(k+3))/2-k*n+1;
          }
          dj-=n-j;
        }
        di-=i+1;
      }
      break;
    }
    case 7: { // 2,0,1,3
      int ijk=0;
      renumber[ijk++]=2;
      renumber[ijk++]=0;
      renumber[ijk++]=1;
      renumber[ijk++]=3;
      int d=3*n; // line 2
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=4; // line 0
      for (int i=1;i<=n-1;i++) {
        renumber[ijk++]=d++;
      }
      d=2*n+1; // line 1
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=5*n-1; // line 5
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=3*n+1; // line 3
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=4*n; // line 4
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      {
        int dk=(n*(3*n+5))/2-2; // face 3: 2,1,3
        for (int k=1;k<=n-2;k++) {
          int dj=dk;
          for (int j=n-k-1;j>=1;j--) {
            renumber[ijk++]=dj;
            dj--;
          }
          dk+=n-k-2;
        }
      }
      {
        int dk=7*n-4; // face 0: 2,0,3
        for (int k=1;k<=n-2;k++) {
          int dj=dk;
          for (int j=n-k-1;j>=1;j--) {
            renumber[ijk++]=dj;
            dj--;
          }
          dk+=n-k-2;
        }
      }
      {
        int di=(n*(3*n+3))/2+1; // face 2: 2,0,1
        for (int i=1;i<=n-2;i++) {
          int dj=di;
          for (int j=n-i-1;j>=1;j--) {
            renumber[ijk++]=dj;
            dj-=n-j;
          }
          di+=i-1;
        }
      }
      {
        int dk=(n*(n+9))/2-1; // face 1: 0,1,3
        for (int k=1;k<=n-2;k++) {
          int di=dk;
          for (int i=1;i+k<=n-1;i++) {
            renumber[ijk++]=di;
            di++;
          }
          dk+=n-k-1;
        }
      }
      int dk=(n*(5*n-3))/2; // interior: 2,0,1,3
      for (int k=1;k<=n-3;k++) {
        int di=dk;
        for (int i=1;i<=n-k-2;i++) {
          int dj=di;
          for (int j=n-i-k-1;j>=1;j--) {
            renumber[ijk++]=dj;
            dj-=n-k-j+1;
          }
          di-=i+1;
        }
        dk+=(n*(n-3)+k*(k+3))/2-k*n+1;
      }
      break;
    }
    case 8: { // 2,1,3,0
      int ijk=0;
      renumber[ijk++]=2;
      renumber[ijk++]=1;
      renumber[ijk++]=3;
      renumber[ijk++]=0;
      int d=2*n+1; // line 1
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=4*n; // line 4
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=5*n-1; // line 5
      for (int k=1;k<=n-1;k++) {
        renumber[ijk++]=d++;
      }
      d=3*n; // line 2
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=n+2; // line 0
      for (int i=n-1;i>=1;i--) {
        renumber[ijk++]=d--;
      }
      d=4*n-1; // line 3
      for (int k=n-1;k>=11;k--) {
        renumber[ijk++]=d--;
      }
      {
        int dj=7*n-4; // face 0: 2,3,0
        for (int j=n-2;j>=1;j--) {
          int dk=dj;
          for (int k=1;k<=j;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-2;
          }
          dj--;
        }
      }
      {
        int dj=(n*(3*n+3))/2+1; // face 2: 2,1,0
        for (int j=n-2;j>=1;j--) {
          int di=dj;
          for (int i=1;i<=j;i++) {
            renumber[ijk++]=di;
            di-=n+i-j-2;
          }
          dj-=n-j;
        }
      }
      {
        int dk=(n*(3*n+5))/2-2; // face 3: 2,1,3
        for (int k=1;k<=n-2;k++) {
          int dj=dk;
          for (int j=n-k-1;j>=1;j--) {
            renumber[ijk++]=dj;
            dj--;
          }
          dk+=n-k-2;
        }
      }
      {
        int di=(n*(n+11))/2-3; // face 1: 1,3,0
        for (int i=n-2;i>=1;i--) {
          int dk=di;
          for (int k=1;k<=i;k++) {
            renumber[ijk++]=dk;
            dk+=n-k-2;
          }
          di--;
        }
      }
      int dj=(n*(5*n-3))/2; // interior: 2,1,3,0
      for (int j=n-3;j>=1;j--) {
        int dk=dj;
        for (int k=1;k<=j;k++) {
          int di=dk;
          for (int i=1;i<=j;i++) {
            renumber[ijk++]=di;
            di-=n+i-k-j-1;
          }
          dk+=(n*(n-3)+k*(k+3))/2-k*n+j;
        }
        dj-=n-j;
      }
      break;
    }
    case 9: { // 3,0,2,1
      int ijk=0;
      renumber[ijk++]=3;
      renumber[ijk++]=0;
      renumber[ijk++]=2;
      renumber[ijk++]=1;
      int d=4*n-1; // line 3
      for (int k=n-1;k>=11;k--) {
        renumber[ijk++]=d--;
      }
      d=2*n+2; // line 2
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      d=6*n-3; // line 5
      for (int k=n-1;k>=11;k--) {
        renumber[ijk++]=d--;
      }
      d=5*n-2; // line 4
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=4; // line 0
      for (int i=1;i<=n-1;i++) {
        renumber[ijk++]=d++;
      }
      d=2*n+1; // line 1
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      {
        int dk=(n*(3*n+5))/2-2; // face 3: 3,2,1
        for (int k=n-2;k>=1;k--) {
          int dj=dk;
          for (int j=1;j<=k;j++) {
            renumber[ijk++]=dj;
            dj-=n+j-k-2;;
          }
          dk-=n-k;
        }
      }
      {
        int di=n*(n+3); // face 1: 3,0,1
        for (int i=1;i<=n-2;i++) {
          int dk=di;
          for (int k=n-i-1;k>=1;k--) {
            renumber[ijk++]=dk;
            dk-=n-k;
          }
          di-=i;
        }
      }
      {
        int dj=(n*(n+9))/2-1; // face 0: 3,0,2
        for (int j=1;j<=n-2;j++) {
          int dk=dj;
          for (int k=n-j-1;k>=1;k--) {
            renumber[ijk++]=dk;
            dk-=n-k;
          }
          dj-=j;
        }
      }
      {
        int di=n*(n+3); // face 2: 0,2,1
        for (int i=1;i<=n-2;i++) {
          int dj=di;
          for (int j=1;i+j<=n-1;j++) {
            renumber[ijk++]=dj;
            dj+=n-j-1;
          }
          di++;
        }
      }
      int di=(n*(n*(n+9)+2))/6-2; // interior: 3,0,2,1
      for (int i=1;i<=n-3;i++) {
        int dj=di;
        for (int j=1;j<=n-i-2;j++) {
          int dk=dj;
          for (int k=n-i-j-1;k>=1;k--) {
            renumber[ijk++]=dk;
            dk-=(n*(n+1)+k*(k-1))/2-k*n-j+1;
          }
          dj-=(j*(j+1)+i*(i+1))/2+j*i;
        }
        di-=(i*(i+5))/2+2;
      }
      break;
    }
    case 10: { // 3,1,0,2
      int ijk=0;
      renumber[ijk++]=3;
      renumber[ijk++]=1;
      renumber[ijk++]=0;
      renumber[ijk++]=2;
      int d=5*n-2; // line 4
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=n+2; // line 0
      for (int i=n-1;i>=1;i--) {
        renumber[ijk++]=d--;
      }
      d=4*n-1; // line 3
      for (int k=n-1;k>=11;k--) {
        renumber[ijk++]=d--;
      }
      d=6*n-3; // line 5
      for (int k=n-1;k>=11;k--) {
        renumber[ijk++]=d--;
      }
      d=n+3; // line 1
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      d=2*n+2; // line 2
      for (int j=1;j<=n-1;j++) {
        renumber[ijk++]=d++;
      }
      {
        int dk=(n*(n+9))/2-1; // face 0: 3,2,0
        for (int k=n-2;k>=1;k--) {
          int dj=dk;
          for (int j=1;j<=k;j++) {
            renumber[ijk++]=dj;
            dj-=n+j-k-2;
          }
          dk-=n-k;
        }
      }
      {
        int dj=2*n*n+1; // face 3: 3,1,2
        for (int j=1;j<=n-2;j++) {
          int dk=dj;
          for (int k=n-j-1;k>=1;k--) {
            renumber[ijk++]=dk;
            dk-=n-k;;
          }
          dj-=j;
        }
      }
      {
        int dk=n*(n+3); // face 1: 3,1,0
        for (int k=n-2;k>=1;k--) {
          int di=dk;
          for (int i=1;i<=k;i++) {
            renumber[ijk++]=di;
            di-=n+i-k-2;
          }
          dk-=n-k;
        }
      }
      {
        int dj=n*(n+4)-2; // face 2: 1,0,2
        for (int j=1;j<=n-2;j++) {
          int di=dj;
          for (int i=n-j-1;i>=1;i--) {
            renumber[ijk++]=di;
            di--;
          }
          dj+=n-j-2;
        }
      }
      int dj=(n*(n*(n+9)+2))/6-2; // interior: 3,1,0,2
      for (int j=1;j<=n-3;j++) {
        int dk=dj;
        for (int k=n-j-2;k>=1;k--) {
          int di=dk;
          for (int i=1;i<=k;i++) {
            renumber[ijk++]=di;
            di-=(n*(n-1)+k*(k+1)+i*(i-1))/2-k*n-k*i+n*i-j;
          }
          dk-=(n*(n+1)+k*(k-1))/2-k*n-j+1;
        }
        dj-=(j*(j+3))/2+1;
      }
      break;
    }
    case 11: { // 3,2,1,0
      int ijk=0;
      renumber[ijk++]=3;
      renumber[ijk++]=2;
      renumber[ijk++]=1;
      renumber[ijk++]=0;
      int d=6*n-3; // line 5
      for (int k=n-1;k>=11;k--) {
        renumber[ijk++]=d--;
      }
      d=2*n+1; // line 1
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=5*n-2; // line 4
      for (int k=n-1;k>=1;k--) {
        renumber[ijk++]=d--;
      }
      d=4*n-1; // line 3
      for (int k=n-1;k>=11;k--) {
        renumber[ijk++]=d--;
      }
      d=3*n; // line 2
      for (int j=n-1;j>=1;j--) {
        renumber[ijk++]=d--;
      }
      d=n+2; // line 0
      for (int i=n-1;i>=1;i--) {
        renumber[ijk++]=d--;
      }
      {
        int dk=n*(n+3); // face 1: 3,1,0
        for (int k=n-2;k>=1;k--) {
          int di=dk;
          for (int i=1;i<=k;i++) {
            renumber[ijk++]=di;
            di-=n+i-k-2;
          }
          dk-=n-k;
        }
      }
      {
        int dk=(n*(n+9))/2-1; // face 0: 3,2,0
        for (int k=n-2;k>=1;k--) {
          int dj=dk;
          for (int j=1;j<=k;j++) {
            renumber[ijk++]=dj;
            dj-=n+j-k-2;
          }
          dk-=n-k;
        }
      }
      {
        int dk=2*n*n+1; // face 3: 3,2,1
        for (int k=n-2;k>=1;k--) {
          int dj=dk;
          for (int j=1;j<=k;j++) {
            renumber[ijk++]=dj;
            dj-=n+j-k-2;;
          }
          dk-=n-k;
        }
      }
      {
        int dj=(n*(3*n+3))/2+1; // face 2: 2,1,0
        for (int j=n-2;j>=1;j--) {
          int di=dj;
          for (int i=1;i<=j;i++) {
            renumber[ijk++]=di;
            di-=n+i-j-2;;
          }
          dj-=n-j;
        }
      }
      int dk=(n*(n*(n+9)+2))/6-2; // interior: 3,2,1,0
      for (int k=n-3;k>=1;k--) {
        int di=dk;
        for (int i=1;i<=k;i++) {
          int dj=di;
          for (int j=1;j<=k;j++) {
            renumber[ijk++]=dj;
            dj-=(n*(n-3)+k*(k+3)+j*(j-3))/2+n*j-n*k-j*k+i;
          }
          di-=(n*(n-1)+k*(k+1)+i*(i-1))/2+n*i-n*k-k*i-1;
        }
        dk-=(n*(n+1)+k*(k-1))/2-k*n;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::staticInteriorLatticePointIndices(int n,
NumPtr<int> &v) {
  v.allocate(staticNumberInteriorEquallySpacedLatticePoints(n));
  int ijk=((n+1)*(n+4))/2;
  int iv=0;
  for (int k=1;k<n;k++) {
    for (int j=1;j+k<n;j++) {
      for (int i=1;i+j+k<n;i++) {
        v[iv]=ijk;
        iv++;
        ijk++;
      }
      ijk+=2;
    }
    ijk+=n-k+1;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::initializeMappingSupportPoints(int degree,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  NumPtr<Point<3> > eslp;
  staticEquallySpacedLatticePoints(degree,eslp);
  int neslp=eslp.getNumber();
  NumPtr<Point<SPACEDIM> > tv(4);
  for (int v=0;v<4;v++) {
    tv[v]=tessellation_vertices[vertexIndex(v)];
  }
  for (int i=0;i<neslp;i++) {
    const Point<3> &eslpi=eslp[i];
    points[i]=tv[1]*eslpi[0]+tv[2]*eslpi[1]+tv[3]*eslpi[2]
             +tv[0]*(1.-eslpi[0]-eslpi[1]-eslpi[2]);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//see lenoirSupportPointsOnTetrahedron in Tessellation.C:
//  must be consistent with face vertex ordering for consistency on
//  internal faces
void Tetrahedron::lenoirCanonicalFacePoints(int f,
const NumPtr<Point<3> > &eslp,NumPtr<Point<2> > &canonical_face_points) 
const {
  canonical_face_points.allocate(eslp.getNumber());
  switch (f) {
    case 0: { // face vertices v0, v2, v3
      for (int i=0;i<eslp.getNumber();i++) {
        Point<2> &cfp=canonical_face_points[i];
        const Point<3> &eslpi=eslp[i];
        if (eslpi[0]<1.) {
          REAL den=1./(1.-eslpi[0]);
          cfp[0]=eslpi[1]*den;
          cfp[1]=eslpi[2]*den;
        } else cfp=0.;
      }
      break;
    }
    case 1: { // face vertices v0, v1, v3
      for (int i=0;i<eslp.getNumber();i++) {
        Point<2> &cfp=canonical_face_points[i];
        const Point<3> &eslpi=eslp[i];
        if (eslpi[1]<1.) {
          REAL den=1./(1.-eslpi[1]);
          cfp[0]=eslpi[0]*den;
          cfp[1]=eslpi[2]*den;
        } else cfp=0.;
      }
      break;
    }
    case 2: { // face vertices ordered v0, v1, v2
      for (int i=0;i<eslp.getNumber();i++) {
        Point<2> &cfp=canonical_face_points[i];
        const Point<3> &eslpi=eslp[i];
        if (eslpi[2]<1.) {
          REAL den=1./(1.-eslpi[2]);
          cfp[0]=eslpi[0]*den;
          cfp[1]=eslpi[1]*den;
        } else cfp=0.;
      }
      break;
    }
    case 3: { // face vertices ordered v1, v3, v2
      for (int i=0;i<eslp.getNumber();i++) {
        Point<2> &cfp=canonical_face_points[i];
        const Point<3> &eslpi=eslp[i];
        REAL den=eslpi[0]+eslpi[1]+eslpi[2];
        if (den>0.) {
          den=1./den;
          cfp[0]=eslpi[1]*den;
          cfp[1]=eslpi[2]*den;
        } else {
          cfp[0]=1.;
          cfp[1]=0.;
        }
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::lenoirCanonicalLinePoints(int l,
const NumPtr<Point<3> > &eslp,NumPtr<Point<1> > &canonical_line_points) 
const {
  canonical_line_points.allocate(eslp.getNumber());
  switch (l) {
    case 0: {
      for (int i=0;i<eslp.getNumber();i++) {
        Point<1> &clp=canonical_line_points[i];
        const Point<3> &eslpi=eslp[i];
        REAL den=eslpi[1]+eslpi[2];
        if (den<1.) clp=eslpi[0]/(1.-den);
        else clp=0.;
      }
      break;
    }
    case 1: {
      for (int i=0;i<eslp.getNumber();i++) {
        Point<1> &clp=canonical_line_points[i];
        const Point<3> &eslpi=eslp[i];
        REAL den=eslpi[0]+eslpi[1];
        if (den>0.) clp=eslpi[1]/den;
        else clp=0.;
      }
      break;
    }
    case 2: {
      for (int i=0;i<eslp.getNumber();i++) {
        Point<1> &clp=canonical_line_points[i];
        const Point<3> &eslpi=eslp[i];
        REAL den=eslpi[0]+eslpi[2];
        if (den<1.) clp=eslpi[1]/(1.-den);
        else clp=0.;
      }
      break;
    }
    case 3: {
      for (int i=0;i<eslp.getNumber();i++) {
        Point<1> &clp=canonical_line_points[i];
        const Point<3> &eslpi=eslp[i];
        REAL den=eslpi[0]+eslpi[1];
        if (den<1.) clp=eslpi[2]/(1.-den);
        else clp=0.;
      }
      break;
    }
    case 4: {
      for (int i=0;i<eslp.getNumber();i++) {
        Point<1> &clp=canonical_line_points[i];
        const Point<3> &eslpi=eslp[i];
        REAL den=eslpi[0]+eslpi[2];
        if (den>0.) clp=eslpi[2]/den;
        else clp=1.;
      }
      break;
    }
    case 5: {
      for (int i=0;i<eslp.getNumber();i++) {
        Point<1> &clp=canonical_line_points[i];
        const Point<3> &eslpi=eslp[i];
        REAL den=eslpi[1]+eslpi[2];
        if (den>0.) clp=eslpi[2]/den;
        else clp=1.;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::lenoirInterpolationForFace(int f,int m,
const NumPtr<Point<SPACEDIM> > &eslp,
const NumPtr<Point<SPACEDIM> > &old_points,
const NumPtr<Point<SPACEDIM> > &new_points,
NumPtr<Point<SPACEDIM> > &higher_order_mapping_support_points) const {
  if (f==3) {
    for (int i=0;i<eslp.getNumber();i++) {
      const Point<SPACEDIM> &eslpi=eslp[i];
      higher_order_mapping_support_points[i]+=
        (new_points[i]-old_points[i])
        *pow(eslpi[0]+eslpi[1]+eslpi[2],m);
    }
  } else {
    for (int i=0;i<eslp.getNumber();i++) {
      const Point<SPACEDIM> &eslpi=eslp[i];
      higher_order_mapping_support_points[i]+=
        (new_points[i]-old_points[i])*pow(1.-eslpi[f],m);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::lenoirInterpolationForLine(int l,int m,
const NumPtr<Point<SPACEDIM> > &eslp,
const NumPtr<Point<SPACEDIM> > &old_points,
const NumPtr<Point<SPACEDIM> > &new_points,
NumPtr<Point<SPACEDIM> > &higher_order_mapping_support_points) const {
  switch (l) {
    case 0: { // (i,0,0)
      for (int i=0;i<eslp.getNumber();i++) {
        const Point<SPACEDIM> &eslpi=eslp[i];
        higher_order_mapping_support_points[i]+=
          (new_points[i]-old_points[i])*
          pow(1.-eslpi[1]-eslpi[2],m);
      }
      break;
    }
    case 1: { // (n-j,j,0)
      for (int i=0;i<eslp.getNumber();i++) {
        const Point<SPACEDIM> &eslpi=eslp[i];
        higher_order_mapping_support_points[i]+=
          (new_points[i]-old_points[i])*
          pow(eslpi[0]+eslpi[1],m);
      }
      break;
    }
    case 2: { // (0,j,0)
      for (int i=0;i<eslp.getNumber();i++) {
        const Point<SPACEDIM> &eslpi=eslp[i];
        higher_order_mapping_support_points[i]+=
          (new_points[i]-old_points[i])*
          pow(1.-eslpi[0]-eslpi[2],m);
      }
      break;
    }
    case 3: { // (0,0,k)
      for (int i=0;i<eslp.getNumber();i++) {
        const Point<SPACEDIM> &eslpi=eslp[i];
        higher_order_mapping_support_points[i]+=
          (new_points[i]-old_points[i])*
          pow(1.-eslpi[0]-eslpi[1],m);
      }
      break;
    }
    case 4: { // (n-k,0,k)
      for (int i=0;i<eslp.getNumber();i++) {
        const Point<SPACEDIM> &eslpi=eslp[i];
        higher_order_mapping_support_points[i]+=
          (new_points[i]-old_points[i])*
          pow(eslpi[0]+eslpi[2],m);
      }
      break;
    }
    case 5: { // (0,n-k,k)
      for (int i=0;i<eslp.getNumber();i++) {
        const Point<SPACEDIM> &eslpi=eslp[i];
        higher_order_mapping_support_points[i]+=
          (new_points[i]-old_points[i])*
          pow(eslpi[1]+eslpi[2],m);
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Tetrahedron::faceCenter(
const NumPtr<Point<SPACEDIM> > &vertices,int f) const {
  switch (f) {
    case 0:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(2)]
             +vertices[vertexIndex(3)])/3.;
    case 1:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(3)]
             +vertices[vertexIndex(1)])/3.;
    case 2:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(1)]
             +vertices[vertexIndex(2)])/3.;
    case 3:
      return (vertices[vertexIndex(1)]
             +vertices[vertexIndex(3)]
             +vertices[vertexIndex(2)])/3.;
    default:
      OBSOLETE("bad face");
      return Point<SPACEDIM>(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Tetrahedron::diameter(const NumPtr<Point<SPACEDIM> > &vertices) 
const {
  REAL d[6]={
    (vertices[vertexIndex(1)]-vertices[vertexIndex(0)]).square(),
    (vertices[vertexIndex(2)]-vertices[vertexIndex(1)]).square(),
    (vertices[vertexIndex(2)]-vertices[vertexIndex(0)]).square(),
    (vertices[vertexIndex(3)]-vertices[vertexIndex(0)]).square(),
    (vertices[vertexIndex(3)]-vertices[vertexIndex(1)]).square(),
    (vertices[vertexIndex(3)]-vertices[vertexIndex(2)]).square()
  };
  REAL dmax=0.;
  for (int i=0;i<6;i++) dmax=max(d[i],dmax);
  return sqrt(dmax);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Tetrahedron::volume(const NumPtr<Point<SPACEDIM> > &vertices) const
{
  Tensor<1,SPACEDIM> v10(vertices[vertexIndex(1)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v20(vertices[vertexIndex(2)] 
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v30(vertices[vertexIndex(3)]
                        -vertices[vertexIndex(0)]);
  REAL vol;
  Tensor<1,SPACEDIM> cp;
  crossProduct(cp,v20,v30); 
  product(vol,v10,cp);
  return vol/6.;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Tetrahedron::lineTangentPointsBackward(int l) const {
  switch (l) {
    case 0:
      return false;
    case 1:
      return vertexIndex(2)<vertexIndex(1);
    case 2:
      return true;
    case 3:
      return false;
    case 4:
      return vertexIndex(1)<vertexIndex(3);
    case 5:
      return vertexIndex(2)<vertexIndex(3);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::printOn(ostream &os) const {
  os << "Tetrahedron:" << endl;
  Shape3::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG 
void checkTriangleVertices(int v0,int v1,int v2,
const Triangle *triangle) {
  int tl[3]={-1,-2,-3};
  int tv[3]={v0,v1,v2};
  Triangle bogus_triangle(-1,tl,tv);
  CHECK_TEST(bogus_triangle.vertexIndex(0)==triangle->vertexIndex(0) &&
             bogus_triangle.vertexIndex(1)==triangle->vertexIndex(1) &&
             bogus_triangle.vertexIndex(2)==triangle->vertexIndex(2));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tetrahedron::checkFaceVertices(int f,const Shape2 *face_element) 
const {
  CHECK_BOUNDS(f,0,4);
  const Triangle *triangle=dynamic_cast<const Triangle*>(face_element);
  CHECK_POINTER(triangle);
  checkTriangleVertices(vertexIndex(face_vertex_to_cell_vertex[f][0]),
                        vertexIndex(face_vertex_to_cell_vertex[f][1]),
                        vertexIndex(face_vertex_to_cell_vertex[f][2]),
                        triangle);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Tetrahedron::draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE &xc,
const NumPtr<Point<3> > &vertices,VolGraphTool &vgt) const {
  Point<3> p0=vertices[vertexIndex(0)];
  Point<3> p1=vertices[vertexIndex(1)];
  Point<3> p2=vertices[vertexIndex(2)];
  Point<3> p3=vertices[vertexIndex(3)];
  Vector3 v0(p0[0],p0[1],p0[2]);
  Vector3 v1(p1[0],p1[1],p1[2]);
  Vector3 v2(p2[0],p2[1],p2[2]);
  Vector3 v3(p3[0],p3[1],p3[2]);
  vgt.colorTriangle(xc,v2,v1,v0);
  vgt.colorTriangle(xc,v0,v1,v3);
  vgt.colorTriangle(xc,v0,v3,v2);
  vgt.colorTriangle(xc,v1,v2,v3);
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Prism::Prism(int edi,const int l[9],const int f[5],const int v[6]) :
Shape3(edi,9,5,6) {
  set<int> vertex_index_set;
  for (int i=0;i<6;i++) vertex_index_set.insert(v[i]);
#ifdef DEBUG
  CHECK_SAME(vertex_index_set.size(),6);
  set<int> face_index_set;
  for (int i=0;i<5;i++) face_index_set.insert(f[i]);
  CHECK_SAME(face_index_set.size(),5);
  set<int> line_index_set;
  for (int i=0;i<9;i++) line_index_set.insert(l[i]);
  CHECK_SAME(line_index_set.size(),9);
#endif
  set<int>::const_iterator vis_iterator=vertex_index_set.begin();
  int vmin=*vis_iterator;
  if (vmin==v[0]) {
    setVertexIndex(0,v[0]);
    setVertexIndex(1,v[1]);
    setVertexIndex(2,v[2]);
    setVertexIndex(3,v[3]);
    setVertexIndex(4,v[4]);
    setVertexIndex(5,v[5]);
    setCodimension1ShapeIndex(0,f[0]);
    setCodimension1ShapeIndex(1,f[1]);
    setCodimension1ShapeIndex(2,f[2]);
    setCodimension1ShapeIndex(3,f[3]);
    setCodimension1ShapeIndex(4,f[4]);
    setCodimension2ShapeIndex(0,l[0]);
    setCodimension2ShapeIndex(1,l[1]);
    setCodimension2ShapeIndex(2,l[2]);
    setCodimension2ShapeIndex(3,l[3]);
    setCodimension2ShapeIndex(4,l[4]);
    setCodimension2ShapeIndex(5,l[5]);
    setCodimension2ShapeIndex(6,l[6]);
    setCodimension2ShapeIndex(7,l[7]);
    setCodimension2ShapeIndex(8,l[8]);
  } else if (vmin==v[1]) {
    setVertexIndex(0,v[1]);
    setVertexIndex(1,v[2]);
    setVertexIndex(2,v[0]);
    setVertexIndex(3,v[4]);
    setVertexIndex(4,v[5]);
    setVertexIndex(5,v[3]);
    setCodimension1ShapeIndex(0,f[0]);
    setCodimension1ShapeIndex(1,f[1]);
    setCodimension1ShapeIndex(2,f[3]);
    setCodimension1ShapeIndex(3,f[4]);
    setCodimension1ShapeIndex(4,f[2]);
    setCodimension2ShapeIndex(0,l[1]);
    setCodimension2ShapeIndex(1,l[2]);
    setCodimension2ShapeIndex(2,l[0]);
    setCodimension2ShapeIndex(3,l[4]);
    setCodimension2ShapeIndex(4,l[5]);
    setCodimension2ShapeIndex(5,l[3]);
    setCodimension2ShapeIndex(6,l[7]);
    setCodimension2ShapeIndex(7,l[8]);
    setCodimension2ShapeIndex(8,l[6]);
  } else if (vmin==v[2]) {
    setVertexIndex(0,v[2]);
    setVertexIndex(1,v[0]);
    setVertexIndex(2,v[1]);
    setVertexIndex(3,v[5]);
    setVertexIndex(4,v[3]);
    setVertexIndex(5,v[4]);
    setCodimension1ShapeIndex(0,f[0]);
    setCodimension1ShapeIndex(1,f[1]);
    setCodimension1ShapeIndex(2,f[4]);
    setCodimension1ShapeIndex(3,f[2]);
    setCodimension1ShapeIndex(4,f[3]);
    setCodimension2ShapeIndex(0,l[2]);
    setCodimension2ShapeIndex(1,l[0]);
    setCodimension2ShapeIndex(2,l[1]);
    setCodimension2ShapeIndex(3,l[5]);
    setCodimension2ShapeIndex(4,l[3]);
    setCodimension2ShapeIndex(5,l[4]);
    setCodimension2ShapeIndex(6,l[8]);
    setCodimension2ShapeIndex(7,l[6]);
    setCodimension2ShapeIndex(8,l[7]);
  } else if (vmin==v[3]) {
    setVertexIndex(0,v[3]);
    setVertexIndex(1,v[5]);
    setVertexIndex(2,v[4]);
    setVertexIndex(3,v[0]);
    setVertexIndex(4,v[2]);
    setVertexIndex(5,v[1]);
    setCodimension1ShapeIndex(0,f[1]);
    setCodimension1ShapeIndex(1,f[0]);
    setCodimension1ShapeIndex(2,f[3]);
    setCodimension1ShapeIndex(3,f[2]);
    setCodimension1ShapeIndex(4,f[4]);
    setCodimension2ShapeIndex(0,l[5]);
    setCodimension2ShapeIndex(1,l[4]);
    setCodimension2ShapeIndex(2,l[3]);
    setCodimension2ShapeIndex(3,l[2]);
    setCodimension2ShapeIndex(4,l[1]);
    setCodimension2ShapeIndex(5,l[0]);
    setCodimension2ShapeIndex(6,l[6]);
    setCodimension2ShapeIndex(7,l[8]);
    setCodimension2ShapeIndex(8,l[7]);
  } else if (vmin==v[4]) {
    setVertexIndex(0,v[4]);
    setVertexIndex(1,v[3]);
    setVertexIndex(2,v[5]);
    setVertexIndex(3,v[1]);
    setVertexIndex(4,v[0]);
    setVertexIndex(5,v[2]);
    setCodimension1ShapeIndex(0,f[1]);
    setCodimension1ShapeIndex(1,f[0]);
    setCodimension1ShapeIndex(2,f[4]);
    setCodimension1ShapeIndex(3,f[3]);
    setCodimension1ShapeIndex(4,f[2]);
    setCodimension2ShapeIndex(0,l[3]);
    setCodimension2ShapeIndex(1,l[5]);
    setCodimension2ShapeIndex(2,l[4]);
    setCodimension2ShapeIndex(3,l[0]);
    setCodimension2ShapeIndex(4,l[2]);
    setCodimension2ShapeIndex(5,l[1]);
    setCodimension2ShapeIndex(6,l[7]);
    setCodimension2ShapeIndex(7,l[6]);
    setCodimension2ShapeIndex(8,l[8]);
  } else {
    setVertexIndex(0,v[5]);
    setVertexIndex(1,v[4]);
    setVertexIndex(2,v[3]);
    setVertexIndex(3,v[2]);
    setVertexIndex(4,v[1]);
    setVertexIndex(5,v[0]);
    setCodimension1ShapeIndex(0,f[1]);
    setCodimension1ShapeIndex(1,f[0]);
    setCodimension1ShapeIndex(2,f[2]);
    setCodimension1ShapeIndex(3,f[4]);
    setCodimension1ShapeIndex(4,f[3]);
    setCodimension2ShapeIndex(0,l[4]);
    setCodimension2ShapeIndex(1,l[3]);
    setCodimension2ShapeIndex(2,l[5]);
    setCodimension2ShapeIndex(3,l[1]);
    setCodimension2ShapeIndex(4,l[0]);
    setCodimension2ShapeIndex(5,l[2]);
    setCodimension2ShapeIndex(6,l[8]);
    setCodimension2ShapeIndex(7,l[7]);
    setCodimension2ShapeIndex(8,l[6]);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Prism::Prism(const int l[9],const int f[5],const int v[6],int di) : 
Shape3(-1,9,5,6) {
#ifdef DEBUG
  set<int> vertex_index_set;
  for (int i=0;i<6;i++) vertex_index_set.insert(v[i]);
  CHECK_SAME(vertex_index_set.size(),6);
  set<int> face_index_set;
  for (int i=0;i<5;i++) face_index_set.insert(f[i]);
  CHECK_SAME(face_index_set.size(),5);
  set<int> line_index_set;
  for (int i=0;i<9;i++) line_index_set.insert(l[i]);
  CHECK_SAME(line_index_set.size(),9);
#endif
  for (int i=0;i<6;i++) setVertexIndex(i,v[i]);
  for (int i=0;i<5;i++) setCodimension1ShapeIndex(i,f[i]);
  for (int i=0;i<9;i++) setCodimension2ShapeIndex(i,l[i]);
  dof_index=di;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape2* Prism::newFaceElement(int f) const {
  if (f<2) {
    int l[3]={
      codimension2ShapeIndex(staticElementLineFromFace(f,0)),
      codimension2ShapeIndex(staticElementLineFromFace(f,1)),
      codimension2ShapeIndex(staticElementLineFromFace(f,2))
    };
    int v[3]={
      vertexIndex(staticElementVertexFromFace(f,0)),
      vertexIndex(staticElementVertexFromFace(f,1)),
      vertexIndex(staticElementVertexFromFace(f,2))
    };
    return OPERATOR_NEW Triangle(l,v,-1); // third arg is dof_index
  } else {
    int l[4]={
      codimension2ShapeIndex(staticElementLineFromFace(f,0)),
      codimension2ShapeIndex(staticElementLineFromFace(f,1)),
      codimension2ShapeIndex(staticElementLineFromFace(f,2)),
      codimension2ShapeIndex(staticElementLineFromFace(f,3))
    };
    int v[4]={
      vertexIndex(staticElementVertexFromFace(f,0)),
      vertexIndex(staticElementVertexFromFace(f,1)),
      vertexIndex(staticElementVertexFromFace(f,2)),
      vertexIndex(staticElementVertexFromFace(f,3))
    };
    return OPERATOR_NEW Quadrilateral(l,v,-1); // third arg is dof_index
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape1* Prism::newLineElement(int l) const {
  int v[2]={
    vertexIndex(staticElementVertexFromLine(l,0)),
    vertexIndex(staticElementVertexFromLine(l,1))
  };
  return OPERATOR_NEW Interval(v,-1); // second arg is dof_index
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//point from lowest vertex index to an adjacent vertex
Tensor<1,3> Prism::faceTangent(int f,int n) const {
  switch (f) {
    case 0: {
      if (vertexIndex(1)<vertexIndex(2)) {
        return (n==0 ? line_tangents[0] : -line_tangents[2]);
      } else {
        return (n==1 ? line_tangents[0] : -line_tangents[2]);
      }
    }
    case 1: {
      int vmin=min(vertexIndex(3),min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(3)==vmin) {
        if (vertexIndex(4)<vertexIndex(5)) {
          return (n==0 ? line_tangents[3] : -line_tangents[5]);
        } else {
          return (n==1 ? line_tangents[3] : -line_tangents[5]);
        }
      } else if (vertexIndex(4)==vmin) {
        if (vertexIndex(3)<vertexIndex(5)) {
          return (n==0 ? -line_tangents[3] : line_tangents[4]/sqrt(2.));
        } else {
          return (n==1 ? -line_tangents[3] : line_tangents[4]/sqrt(2.));
        }
      } else {
        if (vertexIndex(3)<vertexIndex(4)) {
          return (n==0 ? line_tangents[5] : -line_tangents[4]/sqrt(2.));
        } else {
          return (n==1 ? line_tangents[5] : -line_tangents[4]/sqrt(2.));
        }
      }
    }
    case 2: {
      if (vertexIndex(2)<vertexIndex(3)) {
        return (n==0 ? -line_tangents[2] : line_tangents[6]);
      } else {
        return (n==1 ? -line_tangents[2] : line_tangents[6]);
      }
      break;
    }
    case 3: {
      if (vertexIndex(1)<vertexIndex(3)) {
        return (n==0 ? line_tangents[0] : line_tangents[6]);
      } else {
        return (n==1 ? line_tangents[0] : line_tangents[6]);
      }
      break;
    }
    case 4: {
      int vmin=min(min(vertexIndex(1),vertexIndex(2)),
                   min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(1)==vmin) {
        if (vertexIndex(2)<vertexIndex(4)) {
          return (n==0 ? line_tangents[1] : line_tangents[7]);
        } else {
          return (n==1 ? line_tangents[1] : line_tangents[7]);
        }
      } else if (vertexIndex(2)==vmin) {
        if (vertexIndex(1)<vertexIndex(5)) {
          return (n==0 ? -line_tangents[1] : line_tangents[8]);
        } else {
          return (n==1 ? -line_tangents[1] : line_tangents[8]);
        }
      } else if (vertexIndex(4)==vmin) {
        if (vertexIndex(1)<vertexIndex(5)) {
          return (n==0 ? -line_tangents[7] : line_tangents[4]);
        } else {
          return (n==1 ? -line_tangents[7] : line_tangents[4]);
        }
      } else {
        if (vertexIndex(2)<vertexIndex(4)) {
          return (n==0 ? -line_tangents[8] : -line_tangents[4]);
        } else {
          return (n==1 ? -line_tangents[8] : -line_tangents[4]);
        }
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Prism::faceTangentsWindBackward(int f) const {
  switch (f) {
    case 0:
      return (vertexIndex(1)>=vertexIndex(2));
    case 1: {
      int vmin=min(vertexIndex(3),min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(3)==vmin) return (vertexIndex(4)>=vertexIndex(5));
      else if (vertexIndex(4)==vmin) {
        return (vertexIndex(3)>=vertexIndex(5));
      } else {
        return (vertexIndex(3)>=vertexIndex(4));
      }
    }
    case 2:
      return (vertexIndex(2)>=vertexIndex(3));
    case 3:
      return (vertexIndex(1)>=vertexIndex(3));
    case 4: {
      int vmin=min(min(vertexIndex(1),vertexIndex(2)),
                   min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(1)==vmin) return (vertexIndex(2)>=vertexIndex(4));
      else if (vertexIndex(2)==vmin) {
        return (vertexIndex(1)>=vertexIndex(5));
      } else if (vertexIndex(4)==vmin) {
        return (vertexIndex(1)<vertexIndex(5));
      } 
      return (vertexIndex(2)<vertexIndex(4));
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::homogeneousCoordinate(int f,const Point<2> &p,
Point<2> &hp) const {
  switch (f) {
    case 0:
      if (faceTangentsWindBackward(f)) {
        hp[0]=p[1];
        hp[1]=p[0];
      } else hp=p;
      break;
    case 1: {
      int vmin=min(vertexIndex(3),min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(3)==vmin) {
        if (faceTangentsWindBackward(f)) {
          hp[0]=p[1];
          hp[1]=p[0];
        } else hp=p;
      } else if (vertexIndex(4)==vmin) {
        if (faceTangentsWindBackward(f)) {
          hp[0]=p[1];
          hp[1]=1.-p[0]-p[1];
        } else {
          hp[0]=1.-p[0]-p[1];
          hp[1]=p[1];
        }
      } else {
        if (faceTangentsWindBackward(f)) {
          hp[0]=p[0];
          hp[1]=1.-p[0]-p[1];
        } else {
          hp[0]=1.-p[0]-p[1];
          hp[1]=p[0];
        }
      }
      break;
    }
    case 2:
    case 3:
    case 4: 
      hp=p;
      break;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::homogeneousCoordinateDerivative(int f,const Point<2> &p,
Tensor<2,2> &dhpdp) const {
  switch (f) {
    case 0:
      if (faceTangentsWindBackward(f)) {
        dhpdp[0][0]=0.;
        dhpdp[0][1]=1.;
        dhpdp[1][0]=1.;
        dhpdp[1][1]=0.;
      } else {
        dhpdp[0][0]=1.;
        dhpdp[0][1]=0.;
        dhpdp[1][0]=0.;
        dhpdp[1][1]=1.;
      }
      break;
    case 1: {
      int vmin=min(vertexIndex(3),min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(3)==vmin) {
        if (faceTangentsWindBackward(f)) {
          dhpdp[0][0]=0.;
          dhpdp[0][1]=1.;
          dhpdp[1][0]=1.;
          dhpdp[1][1]=0.;
        } else {
          dhpdp[0][0]=1.;
          dhpdp[0][1]=0.;
          dhpdp[1][0]=0.;
          dhpdp[1][1]=1.;
        }
      } else if (vertexIndex(4)==vmin) {
        if (faceTangentsWindBackward(f)) {
          dhpdp[0][0]=0.;
          dhpdp[0][1]=1.;
          dhpdp[1][0]=-1.;
          dhpdp[1][1]=-1.;
        } else {
          dhpdp[0][0]=-1.;
          dhpdp[0][1]=-1.;
          dhpdp[1][0]=0.;
          dhpdp[1][1]=1.;
        }
      } else {
        if (faceTangentsWindBackward(f)) {
          dhpdp[0][0]=1.;
          dhpdp[0][1]=0.;
          dhpdp[1][0]=-1.;
          dhpdp[1][1]=-1.;
        } else {
          dhpdp[0][0]=-1.;
          dhpdp[0][1]=-1.;
          dhpdp[1][0]=1.;
          dhpdp[1][1]=0.;
        }
      }
      break;
    }
    case 2:
    case 3:
    case 4: 
      dhpdp[0][0]=1.;
      dhpdp[0][1]=0.;
      dhpdp[1][0]=0.;
      dhpdp[1][1]=1.;
      break;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Prism::contains(const Point<3> &x,
const NumPtr<Point<3> > &vertices) const {
  int l[6]={-1,-2,-3,-4,-5,-6};
  int f[4]={-1,-2,-3,-4};
  int v[4]=
    {vertexIndex(0),vertexIndex(1),vertexIndex(2),vertexIndex(3)};
  Tetrahedron t0123(l,f,v,-1);
  if (t0123.contains(x,vertices)) return true;

  {
    int v[4]=
      {vertexIndex(1),vertexIndex(4),vertexIndex(2),vertexIndex(3)};
    Tetrahedron t1423(l,f,v,-1);
    if (t1423.contains(x,vertices)) return true;
  }

  {
    int v[4]=
      {vertexIndex(2),vertexIndex(4),vertexIndex(5),vertexIndex(3)};
    Tetrahedron t2453(l,f,v,-1);
    return t2453.contains(x,vertices);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// 6 rotations of a prism times 8 children = 48 cases
Point<SPACEDIM> Prism::parentCoordinateForCanonical(
const Point<SPACEDIM> &cp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  int cvi=-1;
  if (c!=3 && c!=7) {
    int svi=sharedVertexIndex(child_shape);
#ifdef DEBUG
    if (c<3) CHECK_TEST(vertexIndex(c)==svi);
    if (c>3) CHECK_TEST(vertexIndex(c-1)==svi);
#endif
    CHECK_TEST(svi>=0);
    cvi=child_shape->vertexEqualToVertexIndex(svi);
  } else if (c==3) {
    cvi=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[1]);
  } else {
    cvi=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[5]);
  }
  CHECK_TEST(cvi>=0);
  
  switch (c) {
    case 0: {
      switch (cvi) {
        case 0: {
          return cp*0.5;
        }
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 1: {
      switch (cvi) {
        case 0: {
          return Point<3>(1.-(cp[0]+cp[1])*0.5,cp[0]*0.5,cp[2]*0.5);
        }
        case 1: {
          return Point<3>((1.+cp[0])*0.5,cp[1]*0.5,cp[2]*0.5);
        }
        case 2: {
          return Point<3>((1.+cp[1])*0.5,(1.-cp[0]-cp[1])*0.5,cp[2]*0.5);
        }
        case 3: {
          return Point<3>(1.-(cp[0]+cp[1])*0.5,cp[1]*0.5,(1.-cp[2])*0.5);
        }
        case 4: {
          return Point<3>((1.-cp[1])*0.5,(1.-cp[0])*0.5,(1.-cp[2])*0.5);
        }
        case 5: {
          return Point<3>((1.+cp[1])*0.5,cp[0]*0.5,(1.-cp[2])*0.5);
        }
      }
    }
    case 2: {
      switch (cvi) {
        case 0: {
          return Point<3>(cp[1]*0.5,1.-(cp[0]+cp[1])*0.5,cp[2]*0.5);
        }
        case 1: {
          return Point<3>((1.-cp[0]-cp[1])*0.5,(1.+cp[0])*0.5,cp[2]*0.5);
        }
        case 2: {
          return Point<3>(cp[0]*0.5,(1.+cp[1])*0.5,cp[2]*0.5);
        }
        case 3: {
          return Point<3>(cp[0]*0.5,1.-(cp[0]+cp[1])*0.5,(1.-cp[2])*0.5);
        }
        case 4: {
          return Point<3>(cp[1]*0.5,(1.+cp[0])*0.5,(1.-cp[2])*0.5);
        }
        case 5: {
          return Point<3>((1.-cp[0]-cp[1])*0.5,(1.+cp[1])*0.5,
                          (1.-cp[2])*0.5);
        }
      }
    }
    case 3: {
      switch (cvi) {
        case 0: {
          return Point<3>((1.-cp[0])*0.5,(1.-cp[1])*0.5,cp[2]*0.5);
        }
        case 1: {
          return Point<3>((1.-cp[1])*0.5,(cp[0]+cp[1])*0.5,cp[2]*0.5);
        }
        case 2: {
          return Point<3>((cp[0]+cp[1])*0.5,(1.-cp[0])*0.5,cp[2]*0.5);
        }
        case 3: {
          return Point<3>((1.-cp[1])*0.5,(1.-cp[0])*0.5,(1.-cp[2])*0.5);
        }
        case 4: {
          return Point<3>((cp[0]+cp[1])*0.5,(1.-cp[1])*0.5,
                          (1.-cp[2])*0.5);
        }
        case 5: {
          return Point<3>((1.-cp[0])*0.5,(cp[0]+cp[1])*0.5,
                          (1.-cp[2])*0.5);
        }
      }
    }
    case 4: {
      switch (cvi) {
        case 0: {
          return Point<3>(cp[1]*0.5,cp[0]*0.5,1.-cp[2]*0.5);
        }
        case 1: {
          return Point<3>((1.-cp[0]-cp[1])*0.5,cp[1]*0.5,1.-cp[2]*0.5);
        }
        case 2: {
          return Point<3>((1.-cp[0]-cp[1])*0.5,cp[1]*0.5,1.-cp[2]*0.5);
        }
        case 3: {
          return Point<3>(cp[0]*0.5,cp[1]*0.5,(1.+cp[2])*0.5);
        }
        case 4: {
          return Point<3>(cp[1]*0.5,(1.-cp[0]-cp[1])*0.5,(1.+cp[2])*0.5);
        }
        case 5: {
          return Point<3>((1.-cp[0]-cp[1])*0.5,cp[0]*0.5,(1.+cp[2])*0.5);
        }
      }
    }
    case 5: {
      switch (cvi) {
        case 0: {
          return Point<3>(1.-(cp[0]+cp[1])*0.5,cp[1]*0.5,1.-cp[2]*0.5);
        }
        case 1: {
          return Point<3>((1.+cp[0])*0.5,(1.-cp[0]-cp[1])*0.5,
                          1.-cp[2]*0.5);
        }
        case 2: {
          return Point<3>((1.+cp[1])*0.5,cp[0]*0.5,1.-cp[2]*0.5);
        }
        case 3: {
          return Point<3>(1.-(cp[0]+cp[1])*0.5,cp[0]*0.5,(1.+cp[2])*0.5);
        }
        case 4: {
          return Point<3>((1.+cp[0])*0.5,cp[1]*0.5,(1.+cp[2])*0.5);
        }
        case 5: {
          return Point<3>((1.+cp[1])*0.5,(1.-cp[0]-cp[1])*0.5,
                          (1.+cp[2])*0.5);
        }
      }
    }
    case 6: {
      switch (cvi) {
        case 0: {
          return Point<3>(cp[0]*0.5,1.-(cp[0]+cp[1])*0.5,1.-cp[2]*0.5);
        }
        case 1: {
          return Point<3>(cp[1]*0.5,(1.+cp[0])*0.5,1.-cp[2]*0.5);
        }
        case 2: {
          return Point<3>((1.-cp[0]-cp[1])*0.5,(1.+cp[1])*0.5,
                          1.-cp[2]*0.5);
        }
        case 3: {
          return Point<3>(cp[1]*0.5,1.-(cp[0]+cp[1])*0.5,(1.+cp[2])*0.5);
        }
        case 4: {
          return Point<3>((1.-cp[0]-cp[1])*0.5,(1.+cp[0])*0.5,
                          (1.+cp[2])*0.5);
        }
        case 5: {
          return Point<3>(cp[0]*0.5,(1.+cp[1])*0.5,(1.+cp[2])*0.5);
        }
      }
    }
    case 7: {
      switch (cvi) {
        case 0: {
          return Point<3>((cp[0]+cp[1])*0.5,(1.-cp[1])*0.5,1.-cp[2]*0.5);
        }
        case 1: {
          return Point<3>((1.-cp[0])*0.5,(cp[0]+cp[1])*0.5,1.-cp[2]*0.5);
        }
        case 2: {
          return Point<3>((1.-cp[1])*0.5,(1.-cp[0])*0.5,1.-cp[2]*0.5);
        }
        case 3: {
          return Point<3>((cp[0]+cp[1])*0.5,(1.-cp[0])*0.5,
                          (1.+cp[2])*0.5);
        }
        case 4: {
          return Point<3>((1.-cp[0])*0.5,(1.-cp[1])*0.5,(1.+cp[2])*0.5);
        }
        case 5: {
          return Point<3>((1.-cp[1])*0.5,(cp[0]+cp[1])*0.5,
                          (1.+cp[2])*0.5);
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// 6 rotations of a prism times 8 children = 48 cases
Point<SPACEDIM> Prism::childCoordinateOfCanonical(
const Point<SPACEDIM> &pp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  int cvi=-1;
  if (c!=3 && c!=7) {
    int svi=sharedVertexIndex(child_shape);
#ifdef DEBUG
    if (c<3) CHECK_TEST(vertexIndex(c)==svi);
    if (c>3) CHECK_TEST(vertexIndex(c-1)==svi);
#endif
    CHECK_TEST(svi>=0);
    cvi=child_shape->vertexEqualToVertexIndex(svi);
  } else if (c==3) {
    cvi=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[1]);
  } else {
    cvi=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[5]);
  }
  CHECK_TEST(cvi>=0);
  
  switch (c) {
    case 0: {
      switch (cvi) {
        case 0:
          return pp*2.;
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 1: {
      switch (cvi) {
        case 0:
          return Point<3>(pp[1]*2.,(1.-pp[0]-pp[1])*2.,pp[2]*2.);
        case 1:
          return Point<3>(pp[0]*2.-1.,pp[1]*2.,pp[2]*2.);
        case 2:
          return Point<3>((1.-pp[0]-pp[1])*2.,pp[0]*2.-1.,pp[2]*2.);
        case 3:
          return Point<3>((1.-pp[0]-pp[1])*2.,pp[1]*2.,1.-pp[2]*2.);
        case 4:
          return Point<3>(1.-pp[1]*2.,1.-pp[0]*2.,1.-pp[2]*2.);
        case 5:
          return Point<3>(pp[1]*2.,pp[0]*2.-1.,1.-pp[2]*2.);
      }
    }
    case 2: {
      switch (cvi) {
        case 0:
          return Point<3>((1.-pp[0]-pp[1])*2.,pp[0]*2.,pp[2]*2.);
        case 1:
          return Point<3>(pp[1]*2.-1.,(1.-pp[0]-pp[1])*2.,pp[2]*2.);
        case 2:
          return Point<3>(pp[0]*2.,pp[1]*2.-1.,pp[2]*2.);
        case 3:
          return Point<3>(pp[0]*2.,(1.-pp[0]-pp[1])*2.,1.-pp[2]*2.);
        case 4:
          return Point<3>(pp[1]*2.-1.,pp[0]*2.,1.-pp[2]*2.);
        case 5:
          return Point<3>((1.-pp[0]-pp[1])*2.,pp[1]*2.-1.,1.-pp[2]*2.);
      }
    }
    case 3: {
      switch (cvi) {
        case 0:
          return Point<3>(1.-pp[0]*2.,1.-pp[1]*2.,pp[2]*2.);
        case 1:
          return Point<3>((pp[0]+pp[1])*2.-1.,1.-pp[0]*2.,pp[2]*2.);
        case 2:
          return Point<3>(1.-pp[1]*2.,(pp[0]+pp[1])*2.-1.,pp[2]*2.);
        case 3:
          return Point<3>(1.-pp[1]*2.,1.-pp[0]*2.,1.-pp[2]*2.);
        case 4:
          return Point<3>((pp[0]+pp[1])*2.-1.,1.-pp[1]*2.,1.-pp[2]*2.);
        case 5:
          return Point<3>(1.-pp[0]*2.,(pp[0]+pp[1])*2.-1.,1.-pp[2]*2.);
      }
    }
    case 4: {
      switch (cvi) {
        case 0:
          return Point<3>(pp[1]*2.,pp[0]*2.,(1.-pp[2])*2.);
        case 1:
          return Point<3>(1.-(pp[0]+pp[1])*2.,pp[1]*2.,(1.-pp[2])*2.);
        case 2:
          return Point<3>(1.-(pp[0]+pp[1])*2.,pp[1]*2.,(1.-pp[2])*2.);
        case 3:
          return Point<3>(pp[0]*2.,pp[1]*2.,pp[2]*2.-1.);
        case 4:
          return Point<3>(1.-(pp[0]+pp[1])*2.,pp[0]*2.,pp[2]*2.-1.);
        case 5:
          return Point<3>(pp[1]*2.,1.-(pp[0]+pp[1])*2.,pp[2]*2.-1.);
      }
    }
    case 5: {
      switch (cvi) {
        case 0:
          return Point<3>((1.-pp[0]-pp[1])*2.,pp[1]*2.,(1.-pp[2])*2.);
        case 1:
          return Point<3>(pp[0]*2.-1.,(1.-pp[0]-pp[1])*2.,(1.-pp[2])*2.);
        case 2:
          return Point<3>(pp[1]*2.,pp[0]*2.-1.,(1.-pp[2])*2.);
        case 3:
          return Point<3>(pp[1]*2.,(1.-pp[0]-pp[1])*2.,pp[2]*2.-1.);
        case 4:
          return Point<3>(pp[0]*2.-1.,pp[1]*2.,pp[2]*2.-1.);
        case 5:
          return Point<3>((1.-pp[0]-pp[1])*2.,pp[0]*2.-1.,pp[2]*2.-1.);
      }
    }
    case 6: {
      switch (cvi) {
        case 0:
          return Point<3>(pp[0]*2.,(1.-pp[0]-pp[1])*2.,(1.-pp[2])*2.);
        case 1:
          return Point<3>(pp[1]*2.-1.,pp[0]*2.,(1.-pp[2])*2.);
        case 2:
          return Point<3>((1.-pp[0]-pp[1])*2.,pp[1]*2.-1.,(1.-pp[2])*2.);
        case 3:
          return Point<3>((1.-pp[0]-pp[1])*2.,pp[0]*2.,pp[2]*2.-1.);
        case 4:
          return Point<3>(pp[1]*2.-1.,(1.-pp[0]-pp[1])*2.,pp[2]*2.-1.);
        case 5:
          return Point<3>(pp[0]*2.,pp[1]*2.-1.,pp[2]*2.-1.);
      }
    }
    case 7: {
      switch (cvi) {
        case 0:
          return Point<3>((pp[0]+pp[1])*2.-1.,1.-pp[1]*2.,(1.-pp[2])*2.);
        case 1:
          return Point<3>(1.-pp[0]*2.,(pp[0]+pp[1])*2.-1.,(1.-pp[2])*2.);
        case 2:
          return Point<3>(1.-pp[1]*2.,1.-pp[0]*2.,(1.-pp[2])*2.);
        case 3:
          return Point<3>(1.-pp[1]*2.,(pp[0]+pp[1])*2.-1.,pp[2]*2.-1.);
        case 4:
          return Point<3>(1.-pp[0]*2.,1.-pp[1]*2.,pp[2]*2.-1.);
        case 5:
          return Point<3>((pp[0]+pp[1])*2.-1.,1.-pp[0]*2.,pp[2]*2.-1.);
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//see FE_Q::FE_Q; must be consistent with parentCoordinateOfCanonical
const Shape* Prism::reorientFor(const Shape *parent,int c,
const NumPtr<int> &line_interior_child_vertices) const {
#ifdef DEBUG
  if (c<3) CHECK_SAME(vertexIndex(0),parent->vertexIndex(c));
  if (c>=4 && c<7) CHECK_SAME(vertexIndex(0),parent->vertexIndex(c-1));
#endif
  switch (c) {
    case 0: {
//    parentCoordinateOfCanonical(0,0)=p*0.5
      int v[6]={vertexIndex(0),vertexIndex(1),vertexIndex(2),
                vertexIndex(3),vertexIndex(4),vertexIndex(5)};
      int f[5]={codimension1ShapeIndex(0),codimension1ShapeIndex(1),
               codimension1ShapeIndex(2),codimension1ShapeIndex(3),
               codimension1ShapeIndex(4)};
      int l[9]={codimension2ShapeIndex(0),codimension2ShapeIndex(1),
               codimension2ShapeIndex(2),codimension2ShapeIndex(3),
               codimension2ShapeIndex(4),codimension2ShapeIndex(5),
               codimension2ShapeIndex(6),codimension2ShapeIndex(7),
               codimension2ShapeIndex(8)};
      return OPERATOR_NEW Prism(l,f,v,dof_index);
    }
    case 1: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(1.-(p[0]+p[1])*0.5,p[0]*0.5,p[2]*0.5)
      return this;
    }
    case 2: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(p[1]*0.5,1.-(p[0]+p[1])*0.5,p[2]*0.5)
      return this;
    }
    case 3: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>((1.-p[1])*0.5,(p[0]+p[1])*0.5,p[2]*0.5)
      int cv1=line_interior_child_vertices[1];
      if (vertexIndex(0)==cv1) {
        int v[6]={vertexIndex(2),vertexIndex(0),vertexIndex(1),
                  vertexIndex(5),vertexIndex(3),vertexIndex(4)};
        int f[5]={codimension1ShapeIndex(0),codimension1ShapeIndex(1),
                  codimension1ShapeIndex(4),codimension1ShapeIndex(2),
                  codimension1ShapeIndex(3)};
        int l[9]={codimension2ShapeIndex(2),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(1),codimension2ShapeIndex(5),
                  codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(8),codimension2ShapeIndex(6),
                  codimension2ShapeIndex(7)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      } else if (vertexIndex(1)==cv1) {
        return this;
      } else if (vertexIndex(2)==cv1) {
        int v[6]={vertexIndex(1),vertexIndex(2),vertexIndex(0),
                  vertexIndex(4),vertexIndex(5),vertexIndex(3)};
        int f[5]={codimension1ShapeIndex(0),codimension1ShapeIndex(1),
                  codimension1ShapeIndex(3),codimension1ShapeIndex(4),
                  codimension1ShapeIndex(2)};
        int l[9]={codimension2ShapeIndex(1),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(3),
                  codimension2ShapeIndex(7),codimension2ShapeIndex(8),
                  codimension2ShapeIndex(6)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      } else if (vertexIndex(3)==cv1) {
        int v[6]={vertexIndex(4),vertexIndex(3),vertexIndex(5),
                  vertexIndex(1),vertexIndex(0),vertexIndex(2)};
        int f[5]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(4),codimension1ShapeIndex(3),
                  codimension1ShapeIndex(2)};
        int l[9]={codimension2ShapeIndex(3),codimension2ShapeIndex(5),
                  codimension2ShapeIndex(4),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(2),codimension2ShapeIndex(1),
                  codimension2ShapeIndex(7),codimension2ShapeIndex(6),
                  codimension2ShapeIndex(8)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      } else if (vertexIndex(4)==cv1) {
        int v[6]={vertexIndex(5),vertexIndex(4),vertexIndex(3),
                  vertexIndex(2),vertexIndex(1),vertexIndex(0)};
        int f[5]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(2),codimension1ShapeIndex(4),
                  codimension1ShapeIndex(3)};
        int l[9]={codimension2ShapeIndex(4),codimension2ShapeIndex(3),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(1),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(8),codimension2ShapeIndex(7),
                  codimension2ShapeIndex(6)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      } else {
        CHECK_SAME(vertexIndex(5),cv1);
        int v[6]={vertexIndex(3),vertexIndex(5),vertexIndex(4),
                  vertexIndex(0),vertexIndex(2),vertexIndex(1)};
        int f[5]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(3),codimension1ShapeIndex(2),
                  codimension1ShapeIndex(4)};
        int l[9]={codimension2ShapeIndex(5),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(3),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(1),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(6),codimension2ShapeIndex(8),
                  codimension2ShapeIndex(7)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      }
    }
    case 4: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(p[1]*0.5,p[0]*0.5,1.-p[2]*0.5)
      return this;
    }
    case 5: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(1.-(p[0]+p[1])*0.5,p[1]*0.5,1.-p[2]*0.5)
      return this;
    }
    case 6: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(p[0]*0.5,1.-(p[0]+p[1])*0.5,1.-p[2]*0.5)
      return this;
    }
    case 7: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>((1.-p[1])*0.5,(1.-p[0])*0.5,1.-p[2]*0.5)
      int cv4=line_interior_child_vertices[4];
      if (vertexIndex(0)==cv4) {
        return this;
      } else if (vertexIndex(1)==cv4) {
        int v[6]={vertexIndex(1),vertexIndex(2),vertexIndex(0),
                  vertexIndex(4),vertexIndex(5),vertexIndex(3)};
        int f[5]={codimension1ShapeIndex(0),codimension1ShapeIndex(1),
                  codimension1ShapeIndex(3),codimension1ShapeIndex(4),
                  codimension1ShapeIndex(2)};
        int l[9]={codimension2ShapeIndex(1),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(3),
                  codimension2ShapeIndex(7),codimension2ShapeIndex(8),
                  codimension2ShapeIndex(6)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      } else if (vertexIndex(2)==cv4) {
        int v[6]={vertexIndex(2),vertexIndex(0),vertexIndex(1),
                  vertexIndex(5),vertexIndex(3),vertexIndex(4)};
        int f[5]={codimension1ShapeIndex(0),codimension1ShapeIndex(1),
                  codimension1ShapeIndex(4),codimension1ShapeIndex(2),
                  codimension1ShapeIndex(3)};
        int l[9]={codimension2ShapeIndex(2),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(1),codimension2ShapeIndex(5),
                  codimension2ShapeIndex(3),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(8),codimension2ShapeIndex(6),
                  codimension2ShapeIndex(7)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      } else if (vertexIndex(3)==cv4) {
        int v[6]={vertexIndex(3),vertexIndex(5),vertexIndex(4),
                  vertexIndex(0),vertexIndex(2),vertexIndex(1)};
        int f[5]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(3),codimension1ShapeIndex(2),
                  codimension1ShapeIndex(4)};
        int l[9]={codimension2ShapeIndex(5),codimension2ShapeIndex(4),
                  codimension2ShapeIndex(3),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(1),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(6),codimension2ShapeIndex(8),
                  codimension2ShapeIndex(7)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      } else if (vertexIndex(4)==cv4) {
        int v[6]={vertexIndex(4),vertexIndex(3),vertexIndex(5),
                  vertexIndex(1),vertexIndex(0),vertexIndex(2)};
        int f[5]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(4),codimension1ShapeIndex(3),
                  codimension1ShapeIndex(2)};
        int l[9]={codimension2ShapeIndex(3),codimension2ShapeIndex(5),
                  codimension2ShapeIndex(4),codimension2ShapeIndex(0),
                  codimension2ShapeIndex(2),codimension2ShapeIndex(1),
                  codimension2ShapeIndex(7),codimension2ShapeIndex(6),
                  codimension2ShapeIndex(8)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      } else {
        CHECK_SAME(vertexIndex(5),cv4);
        int v[6]={vertexIndex(5),vertexIndex(4),vertexIndex(3),
                  vertexIndex(2),vertexIndex(1),vertexIndex(0)};
        int f[5]={codimension1ShapeIndex(1),codimension1ShapeIndex(0),
                  codimension1ShapeIndex(2),codimension1ShapeIndex(4),
                  codimension1ShapeIndex(3)};
        int l[9]={codimension2ShapeIndex(4),codimension2ShapeIndex(3),
                  codimension2ShapeIndex(5),codimension2ShapeIndex(1),
                  codimension2ShapeIndex(0),codimension2ShapeIndex(2),
                  codimension2ShapeIndex(8),codimension2ShapeIndex(7),
                  codimension2ShapeIndex(6)};
        return OPERATOR_NEW Prism(l,f,v,dof_index);
      }
    }
    default:
      OBSOLETE("bad c");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::staticProjectToLine(int l,const NumPtr<Point<1> > &q_points,
NumPtr<Point<3> > &q_points_line) {
  switch (l) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=q_points[i][0];
        q_points_line[i][1]=0.;
        q_points_line[i][2]=0.;
      }
      break;
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.-q_points[i][0];
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=0.;
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=0.;
      }
      break;
    case 3:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=q_points[i][0];
        q_points_line[i][1]=0.;
        q_points_line[i][2]=1.;
      }
      break;
    case 4:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.-q_points[i][0];
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=1.;
      }
      break;
    case 5:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=1.;
      }
      break;
    case 6:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=0.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    case 7:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.;
        q_points_line[i][1]=0.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    case 8:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=1.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::staticProjectToFace(int f,const NumPtr<Point<2> > &q_points,
NumPtr<Point<3> > &q_points_face) {
  switch (f) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=q_points[i][1];
        q_points_face[i][2]=0.;
      }
      break;
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=q_points[i][1];
        q_points_face[i][2]=1.;
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=0.;
        q_points_face[i][1]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 3:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][1]=0.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 4:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][1]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
        q_points_face[i][0]=1.-q_points[i][0];
      }
      break;
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::projectToFaceForHcurl2(int f,
const NumPtr<Point<2> > &q_points,NumPtr<Point<3> > &q_points_face)
const {
  switch (f) {
    case 0: {
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=q_points[i][1];
        q_points_face[i][2]=0.;
      }
      break;
    }
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=q_points[i][1];
        q_points_face[i][2]=1.;
      }
      break;
    case 2: {
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=0.;
        q_points_face[i][1]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    }
    case 3: {
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][1]=0.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    }
    case 4: {
      int vmin=min(min(vertexIndex(1),vertexIndex(2)),
                   min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(1)==vmin) {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][1]=q_points[i][0];
          q_points_face[i][2]=q_points[i][1];
          q_points_face[i][0]=1.-q_points[i][0];
        }
      } else if (vertexIndex(2)==vmin) {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][1]=1.-q_points[i][0];
          q_points_face[i][2]=q_points[i][1];
          q_points_face[i][0]=q_points[i][0];
        }
      } else if (vertexIndex(4)==vmin) {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][1]=q_points[i][0];
          q_points_face[i][2]=1.-q_points[i][1];
          q_points_face[i][0]=1.-q_points[i][0];
        }
      } else {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][1]=1.-q_points[i][0];
          q_points_face[i][2]=1.-q_points[i][1];
          q_points_face[i][0]=q_points[i][0];
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::projectToFaceForBDFM(int f,
const NumPtr<Point<2> > &q_points,NumPtr<Point<3> > &q_points_face)
const {
  switch (f) {
    case 0:
      if (vertexIndex(1)<vertexIndex(2)) {
        for (int i=0;i<q_points.getNumber();i++) {
          const Point<2> &qpi=q_points[i];
          Point<3> &qpfi=q_points_face[i];
          qpfi[0]=qpi[0];
          qpfi[1]=qpi[1];
          qpfi[2]=0.;
        }
      } else {
        for (int i=0;i<q_points.getNumber();i++) {
          const Point<2> &qpi=q_points[i];
          Point<3> &qpfi=q_points_face[i];
          qpfi[0]=qpi[1];
          qpfi[1]=qpi[0];
          qpfi[2]=0.;
        }
      }
      break;
    case 1: {
      int vmin=min(min(vertexIndex(3),vertexIndex(4)),vertexIndex(5));
      if (vertexIndex(3)==vmin) {
        if (vertexIndex(4)<vertexIndex(5)) {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[0]=qpi[0];
            qpfi[1]=qpi[1];
            qpfi[2]=1.;
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[0]=qpi[1];
            qpfi[1]=qpi[0];
            qpfi[2]=1.;
          }
        }
      } else if (vertexIndex(4)==vmin) {
        if (vertexIndex(5)<vertexIndex(3)) {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=qpi[0];
            qpfi[0]=1.-qpi[0]-qpi[1];
            qpfi[2]=1.;
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=qpi[1];
            qpfi[0]=1.-qpi[0]-qpi[1];
            qpfi[2]=1.;
          }
        }
      } else {
        if (vertexIndex(3)<vertexIndex(4)) {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=1.-qpi[0]-qpi[1];
            qpfi[0]=qpi[1];
            qpfi[2]=1.;
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=1.-qpi[0]-qpi[1];
            qpfi[0]=qpi[0];
            qpfi[2]=1.;
          }
        }
      }
      break;
    }
    case 2:
      if (vertexIndex(2)<vertexIndex(3)) {
        for (int i=0;i<q_points.getNumber();i++) {
          const Point<2> &qpi=q_points[i];
          Point<3> &qpfi=q_points_face[i];
          qpfi[0]=0.;
          qpfi[1]=qpi[0];
          qpfi[2]=qpi[1];
        }
      } else {
        for (int i=0;i<q_points.getNumber();i++) {
          const Point<2> &qpi=q_points[i];
          Point<3> &qpfi=q_points_face[i];
          qpfi[0]=0.;
          qpfi[1]=qpi[1];
          qpfi[2]=qpi[0];
        }
      }
      break;
    case 3:
      if (vertexIndex(1)<vertexIndex(3)) {
        for (int i=0;i<q_points.getNumber();i++) {
          const Point<2> &qpi=q_points[i];
          Point<3> &qpfi=q_points_face[i];
          qpfi[1]=0.;
          qpfi[0]=qpi[0];
          qpfi[2]=qpi[1];
        }
      } else {
        for (int i=0;i<q_points.getNumber();i++) {
          const Point<2> &qpi=q_points[i];
          Point<3> &qpfi=q_points_face[i];
          qpfi[1]=0.;
          qpfi[0]=qpi[1];
          qpfi[2]=qpi[0];
        }
      }
      break;
    case 4: {
      int vmin=min(min(vertexIndex(1),vertexIndex(2)),
                   min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(1)==vmin) {
        if (vertexIndex(2)<vertexIndex(4)) {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=qpi[0];
            qpfi[2]=qpi[1];
            qpfi[0]=1.-qpi[0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=qpi[1];
            qpfi[2]=qpi[0];
            qpfi[0]=1.-qpi[1];
          }
        }
      } else if (vertexIndex(2)==vmin) {
        if (vertexIndex(1)<vertexIndex(5)) {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=1.-qpi[0];
            qpfi[2]=qpi[1];
            qpfi[0]=qpi[0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=1.-qpi[1];
            qpfi[2]=qpi[0];
            qpfi[0]=qpi[1];
          }
        }
      } else if (vertexIndex(4)==vmin) {
        if (vertexIndex(5)<vertexIndex(1)) {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=qpi[0];
            qpfi[2]=1.-qpi[1];
            qpfi[0]=1.-qpi[0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=qpi[1];
            qpfi[2]=1.-qpi[0];
            qpfi[0]=1.-qpi[1];
          }
        }
      } else {
        if (vertexIndex(4)<vertexIndex(2)) {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=1.-qpi[0];
            qpfi[2]=1.-qpi[1];
            qpfi[0]=qpi[0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            const Point<2> &qpi=q_points[i];
            Point<3> &qpfi=q_points_face[i];
            qpfi[1]=1.-qpi[1];
            qpfi[2]=1.-qpi[0];
            qpfi[0]=qpi[1];
          }
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::projectToSubface(int f,int s,
const NumPtr<Point<2> > &q_points,NumPtr<Point<3> > &q_points_subface) {
  switch (f) {
    case 0: {
      switch (s) {
        case 0: { // child 0
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=qp[1]*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 1: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-(qp[0]+qp[1])*0.5;
            qps[1]=qp[0]*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 2: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=1.-(qp[0]+qp[1])*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 3: { // child 3
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=(qp[0]+qp[1])*0.5;
            qps[1]=(1.-qp[0])*0.5;
            qps[2]=0.;
          }
          break;
        }
      }
      break;
    }
    case 1: {
      switch (s) {
        case 0: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=qp[1]*0.5;
            qps[2]=1.;
          }
          break;
        }
        case 1: { // child 5
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-(qp[0]+qp[1])*0.5;
            qps[1]=qp[0]*0.5;
            qps[2]=1.;
          }
          break;
        }
        case 2: { // child 6
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=1.-(qp[0]+qp[1])*0.5;
            qps[2]=1.;
          }
          break;
        }
        case 3: { // child 7
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=(qp[0]+qp[1])*0.5;
            qps[1]=(1.-qp[0])*0.5;
            qps[2]=1.;
          }
          break;
        }
      }
      break;
    }
    case 2: {
      switch (s) { // child 0
        case 0: {
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=qp[0]*0.5;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=qp[1]*0.5;
            qps[2]=1.-qp[0]*0.5;
          }
          break;
        }
        case 2: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=1.-qp[1]*0.5;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 3: { // child 6
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=1.-qp[0]*0.5;
            qps[2]=1.-qp[1]*0.5;
          }
          break;
        }
      }
      break;
    }
    case 3: {
      switch (s) { // child 0
        case 0: {
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=0.;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[0]*0.5;
            qps[1]=0.;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 2: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=0.;
            qps[2]=1.-qp[1]*0.5;
          }
          break;
        }
        case 3: { // child 5
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[1]*0.5;
            qps[1]=0.;
            qps[2]=1.-qp[0]*0.5;
          }
          break;
        }
      }
      break;
    }
    case 4: {
      switch (s) {
        case 0: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[0]*0.5;
            qps[1]=qp[0]*0.5;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=1.-qp[1]*0.5;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 2: { // child 5
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[1]*0.5;
            qps[1]=qp[1]*0.5;
            qps[2]=1.-qp[0]*0.5;
          }
          break;
        }
        case 3: { // child 6
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=1.-qp[0]*0.5;
            qps[2]=1.-qp[1]*0.5;
          }
          break;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape2* Prism::newInteriorCodimension1Shape(int i,
int codimension1_element_index,
const NumPtr<int> &interior_child_vertex_indices,
const NumPtr<int> &interior_child_face_vertex_indices,
const NumPtr<int> &interior_child_line_vertex_indices,
const map<pair<int,int>,int> &interior_child_line_map) const {
  CHECK_BOUNDS(i,0,10);
  CHECK_SAME(interior_child_vertex_indices.getNumber(),0);
  CHECK_SAME(interior_child_face_vertex_indices.getNumber(),5);
  CHECK_SAME(interior_child_line_vertex_indices.getNumber(),9);
  switch (i) {
    case 0: { // vertices v12,v15,v16
      int v[3]={
        interior_child_line_vertex_indices[6], // v12
        interior_child_face_vertex_indices[2], // v15
        interior_child_face_vertex_indices[3]  // v16
      };
      int l[3];
      setTriangleLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Triangle(codimension1_element_index,l,v);
      break;
    }
    case 1: { // vertices v16,v17,v13
      int v[3]={
        interior_child_face_vertex_indices[3], // v16
        interior_child_face_vertex_indices[4], // v17
        interior_child_line_vertex_indices[7]  // v13
      };
      int l[3];
      setTriangleLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Triangle(codimension1_element_index,l,v);
      break;
    }
    case 2: { // vertices v15,v14,v17
      int v[3]={
        interior_child_face_vertex_indices[2], // v15
        interior_child_line_vertex_indices[8], // v14
        interior_child_face_vertex_indices[4]  // v17
      };
      int l[3];
      setTriangleLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Triangle(codimension1_element_index,l,v);
      break;
    }
    case 3: { // vertices v17,v16,v15
      int v[3]={
        interior_child_face_vertex_indices[4], // v17
        interior_child_face_vertex_indices[3], // v16
        interior_child_face_vertex_indices[2]  // v15
      };
      int l[3];
      setTriangleLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Triangle(codimension1_element_index,l,v);
      break;
    }
    case 4: { // vertices v6,v16,v7,v17
      int v[4]={
        interior_child_line_vertex_indices[0], // v6
        interior_child_face_vertex_indices[3], // v16
        interior_child_line_vertex_indices[1], // v7
        interior_child_face_vertex_indices[4]  // v17
      };
      int l[4];
      setQuadrilateralLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Quadrilateral(codimension1_element_index,l,v);
      break;
    }
    case 5: { // vertices v8,v7,v15,v17
      int v[4]={
        interior_child_line_vertex_indices[2], // v8
        interior_child_line_vertex_indices[1], // v7
        interior_child_face_vertex_indices[2], // v15
        interior_child_face_vertex_indices[4]  // v17
      };
      int l[4];
      setQuadrilateralLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Quadrilateral(codimension1_element_index,l,v);
      break;
    }
    case 6: { // vertices v8,v6,v15,v16
      int v[4]={
        interior_child_line_vertex_indices[2], // v8
        interior_child_line_vertex_indices[0], // v6
        interior_child_face_vertex_indices[2], // v15
        interior_child_face_vertex_indices[3]  // v16
      };
      int l[4];
      setQuadrilateralLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Quadrilateral(codimension1_element_index,l,v);
      break;
    }
    case 7: { // vertices v16,v9,v17,v10
      int v[4]={
        interior_child_face_vertex_indices[3], // v16
        interior_child_line_vertex_indices[3], // v9
        interior_child_face_vertex_indices[4], // v17
        interior_child_line_vertex_indices[4]  // v10
      };
      int l[4];
      setQuadrilateralLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Quadrilateral(codimension1_element_index,l,v);
      break;
    }
    case 8: { // vertices v15,v17,v11,v10
      int v[4]={
        interior_child_face_vertex_indices[2], // v15
        interior_child_face_vertex_indices[4], // v17
        interior_child_line_vertex_indices[5], // v11
        interior_child_line_vertex_indices[4]  // v10
      };
      int l[4];
      setQuadrilateralLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Quadrilateral(codimension1_element_index,l,v);
      break;
    }
    case 9: { // vertices v15,v16,v11,v9
      int v[4]={
        interior_child_face_vertex_indices[2], // v15
        interior_child_face_vertex_indices[3], // v16
        interior_child_line_vertex_indices[5], // v11
        interior_child_line_vertex_indices[3]  // v9
      };
      int l[4];
      setQuadrilateralLineIndices(v,interior_child_line_map,l);
      return OPERATOR_NEW Quadrilateral(codimension1_element_index,l,v);
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape3* Prism::newChildShape(int c,
int child_element_index,
const NumPtr<int> &interior_child_vertex_indices,
const NumPtr<int> &interior_child_face_vertex_indices,
const NumPtr<int> &interior_child_line_vertex_indices,
const map<pair<pair<int,int>,pair<int,int> >,int>
&interior_child_face_map,
const map<pair<int,int>,int> &interior_child_line_map) const {
  CHECK_BOUNDS(c,0,8);
  CHECK_SAME(interior_child_vertex_indices.getNumber(),0);
  CHECK_SAME(interior_child_face_vertex_indices.getNumber(),5);
  CHECK_SAME(interior_child_line_vertex_indices.getNumber(),9);
  int v[6];
  switch (c) {
    case 0: {
      v[0]=vertexIndex(0);                        // v0
      v[1]=interior_child_line_vertex_indices[0]; // v6
      v[2]=interior_child_line_vertex_indices[2]; // v8
      v[3]=interior_child_line_vertex_indices[6]; // v12
      v[4]=interior_child_face_vertex_indices[3]; // v16
      v[5]=interior_child_face_vertex_indices[2]; // v15
      break;
    }
    case 1: {
      v[0]=interior_child_line_vertex_indices[0]; // v6
      v[1]=vertexIndex(1);                        // v1
      v[2]=interior_child_line_vertex_indices[1]; // v7
      v[3]=interior_child_face_vertex_indices[3]; // v16
      v[4]=interior_child_line_vertex_indices[7]; // v13
      v[5]=interior_child_face_vertex_indices[4]; // v17
      break;
    }
    case 2: {
      v[0]=interior_child_line_vertex_indices[2]; // v8
      v[1]=interior_child_line_vertex_indices[1]; // v7
      v[2]=vertexIndex(2);                        // v2
      v[3]=interior_child_face_vertex_indices[2]; // v15
      v[4]=interior_child_face_vertex_indices[4]; // v17
      v[5]=interior_child_line_vertex_indices[8]; // v14
      break;
    }
    case 3: {
      v[0]=interior_child_face_vertex_indices[4]; // v17
      v[1]=interior_child_face_vertex_indices[3]; // v16
      v[2]=interior_child_face_vertex_indices[2]; // v15
      v[3]=interior_child_line_vertex_indices[1]; // v7
      v[4]=interior_child_line_vertex_indices[0]; // v6
      v[5]=interior_child_line_vertex_indices[2]; // v8
      break;
    }
    case 4: {
      v[0]=interior_child_line_vertex_indices[6]; // v12
      v[1]=interior_child_face_vertex_indices[3]; // v16
      v[2]=interior_child_face_vertex_indices[2]; // v15
      v[3]=vertexIndex(3);                        // v3
      v[4]=interior_child_line_vertex_indices[3]; // v9
      v[5]=interior_child_line_vertex_indices[5]; // v11
      break;
    }
    case 5: {
      v[0]=interior_child_face_vertex_indices[3]; // v16
      v[1]=interior_child_line_vertex_indices[7]; // v13
      v[2]=interior_child_face_vertex_indices[4]; // v17
      v[3]=interior_child_line_vertex_indices[3]; // v9
      v[4]=vertexIndex(4);                        // v4
      v[5]=interior_child_line_vertex_indices[4]; // v10
      break;
    }
    case 6: {
      v[0]=interior_child_face_vertex_indices[2]; // v15
      v[1]=interior_child_face_vertex_indices[4]; // v17
      v[2]=interior_child_line_vertex_indices[8]; // v14
      v[3]=interior_child_line_vertex_indices[5]; // v11
      v[4]=interior_child_line_vertex_indices[4]; // v10
      v[5]=vertexIndex(5);                        // v5
      break;
    }
    case 7: {
      v[0]=interior_child_line_vertex_indices[4]; // v10
      v[1]=interior_child_line_vertex_indices[3]; // v9
      v[2]=interior_child_line_vertex_indices[5]; // v11
      v[3]=interior_child_face_vertex_indices[4]; // v17
      v[4]=interior_child_face_vertex_indices[3]; // v16
      v[5]=interior_child_face_vertex_indices[2]; // v15
    }
  }
  int f[5];
  for (int i=0;i<2;i++) {
    int tl[3]={-1,-2,-3};
    int tv[3]={v[face_vertex_to_cell_vertex[i][0]],
               v[face_vertex_to_cell_vertex[i][1]],
               v[face_vertex_to_cell_vertex[i][2]]};
    Triangle bogus_triangle(-1,tl,tv);
    pair<pair<int,int>,pair<int,int> > p(
      make_pair(bogus_triangle.vertexIndex(0),
                bogus_triangle.vertexIndex(1)),
      make_pair(bogus_triangle.vertexIndex(2),-1));
    map<pair<pair<int,int>,pair<int,int> >,int>::const_iterator 
      icfm_iterator=interior_child_face_map.find(p);
    CHECK_TEST(icfm_iterator!=interior_child_face_map.end());
    f[i]=icfm_iterator->second;
  }
  for (int i=2;i<5;i++) {
    int ql[4]={-1,-2,-3,-4};
    int qv[4]={v[face_vertex_to_cell_vertex[i][0]],
               v[face_vertex_to_cell_vertex[i][1]],
               v[face_vertex_to_cell_vertex[i][2]],
               v[face_vertex_to_cell_vertex[i][3]]};
    Quadrilateral bogus_quadrilateral(-1,ql,qv);
    pair<pair<int,int>,pair<int,int> > 
      p(make_pair(bogus_quadrilateral.vertexIndex(0),
                  bogus_quadrilateral.vertexIndex(1)),
        make_pair(bogus_quadrilateral.vertexIndex(2),
                  bogus_quadrilateral.vertexIndex(3)));
    map<pair<pair<int,int>,pair<int,int> >,int>::const_iterator 
      icfm_iterator=interior_child_face_map.find(p);
    CHECK_TEST(icfm_iterator!=interior_child_face_map.end());
    f[i]=icfm_iterator->second;
  }
  int l[9];
  for (int i=0;i<9;i++) {
    int iv[2]={v[canonical_vertex_indices[i][0]],
               v[canonical_vertex_indices[i][1]]};
    Interval bogus_interval(-1,iv);
    pair<int,int> p(bogus_interval.vertexIndex(0),
                    bogus_interval.vertexIndex(1));
    map<pair<int,int>,int>::const_iterator iclm_iterator=
      interior_child_line_map.find(p);
    CHECK_TEST(iclm_iterator!=interior_child_line_map.end());
    l[i]=iclm_iterator->second;
  }
  return OPERATOR_NEW Prism(child_element_index,l,f,v);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Prism::lineFromFace(const Shape2 *face,int f,int fl) const {
  CHECK_SAME(face->elementDataIndex(),codimension1ShapeIndex(f));
  switch (f) {
    case 0: {
      int lines[3]={2,0,1};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      if (fv1==(fv0+1)%3) return lines[(fl+fv0)%3];
      else return lines[(2+fv1-fl)%3];
    }
    case 1: {
      int lines[3]={5,3,4};
      int fv3=face->vertexEqualToVertexIndex(vertexIndex(3));
      int fv4=face->vertexEqualToVertexIndex(vertexIndex(4));
      if (fv4==(fv3+1)%3) return lines[(fl+fv3)%3];
      else return lines[(2+fv4-fl)%3];
    }
    case 2: {
      int lines[4]={6,8,2,5};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv2=face->vertexEqualToVertexIndex(vertexIndex(2));
      switch (fv0) {
        case 0: {
          if (fv2==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv2==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv2==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv2==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
    case 3: {
      int lines[4]={6,7,0,3};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      switch (fv0) {
        case 0: {
          if (fv1==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv1==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv1==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv1==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
    case 4: {
      int lines[4]={7,8,1,4};
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      int fv2=face->vertexEqualToVertexIndex(vertexIndex(2));
      switch (fv1) {
        case 0: {
          if (fv2==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv2==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv2==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv2==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::staticEquallySpacedLatticePoints(int n01,int n2,
NumPtr<Point<3> > &points) {
  points.allocate(staticNumberEquallySpacedLatticePoints(n01,n2));
  Point<3> p(0.,0.,0.);
  REAL s01=(n01>0 ? 1./static_cast<REAL>(n01) : 0.);
  REAL s2=(n2>0 ? 1./static_cast<REAL>(n2) : 0.);
  int l=0;
  for (int k=0;k<=n2;k++,p[2]+=s2) {
    for (int j=0;j<=n01;j++,p[1]+=s01) {
      for (int i=0;i<=n01-j;i++,p[0]+=s01,l++) points[l]=p;
      p[0]=0.;
    }
    p[1]=0.;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Prism::staticEquallySpacedLatticePointIndexOfVertex(int v,int n) {
  CHECK_TEST(n>0);
  CHECK_TEST(v<6);
  switch (v) {
    case 0:
      return 0;
    case 1: 
      return n;
    case 2:
      return ((n+1)*(n+2))/2-1;
    case 3:
      return n*((n+1)*(n+2))/2;
    case 4:
      return n*((n+1)*(n+2))/2+n;
    case 5:
      return (n+1)*((n+1)*(n+2))/2-1;
    default:
      OBSOLETE("bad vertex");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::staticEquallySpacedLatticePointIndicesOfFace(int f,int n,
NumPtr<int> &indices) {
  switch (f) {
    case 0: { // (i,j,0)
      indices.allocate(
        Triangle::staticNumberEquallySpacedLatticePoints(n));
      int ijk=0;
      int ij=0;
      for (int j=0;j<=n;j++) {
        for (int i=0;i+j<=n;i++,ij++) {
          indices[ij]=ijk;
          ijk++;
        }
      }
      break;
    }
    case 1: { // (i,j,n)
      indices.allocate(
        Triangle::staticNumberEquallySpacedLatticePoints(n));
      int ijk=n*((n*(n+3))/2+1);
      int ij=0;
      for (int j=0;j<=n;j++) {
        for (int i=0;i+j<=n;i++,ij++) {
          indices[ij]=ijk;
          ijk++;
        }
      }
      break;
    }
    case 2: { // (0,j,k)
      indices.allocate(
        Quadrilateral::staticNumberEquallySpacedLatticePoints(n));
      int ijk=0;
      int jk=0;
      for (int k=0;k<=n;k++) {
        for (int j=0;j<=n;j++,jk++) {
          indices[jk]=ijk;
          ijk+=n+1-j;
        }
      }
      break;
    }
    case 3: { // (i,0,k)
      indices.allocate(
        Quadrilateral::staticNumberEquallySpacedLatticePoints(n));
      int ijk=0;
      int ik=0;
      for (int k=0;k<=n;k++) {
        for (int i=0;i<=n;i++,ik++) {
          indices[ik]=ijk;
          ijk++;
        }
        ijk+=(n*(n+1))/2;
      }
      break;
    }
    case 4: { // (n-j,j,k)
      indices.allocate(
        Quadrilateral::staticNumberEquallySpacedLatticePoints(n));
      int ijk=n;
      int jk=0;
      for (int k=0;k<=n;k++) {
        for (int j=0;j<=n;j++,jk++) {
          indices[jk]=ijk;
          ijk+=n-j;
        }
        ijk+=n+1;
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::equallySpacedLatticePointIndicesOfFace(int f,int n,
NumPtr<int> &indices) const {
  switch (f) {
    case 0: { // (i,j,0)
      indices.allocate(((n+2)*(n+1))/2);
      if (vertexIndex(1)<vertexIndex(2)) {
        int ijk=0;
        int ij=0;
        for (int j=0;j<=n;j++) {
          for (int i=0;i+j<=n;i++,ij++) {
            indices[ij]=ijk;
            if (i+j<n) ijk++;
          }
          ijk++;
        }
      } else {
        int ijk=0;
        int ij=0;
        for (int i=0;i<=n;i++) {
          for (int j=0;i+j<=n;j++,ij++) {
            indices[ij]=ijk;
            if (i+j<n) ijk+=n+1-j;
          }
          ijk-=(n*(n+3)-i*(i+3))/2-1;
        }
      }
      break;
    }
    case 1: { // (i,j,n)
      indices.allocate(((n+2)*(n+1))/2);
      if (vertexIndex(3)<vertexIndex(4)) {
        if (vertexIndex(4)<vertexIndex(5)) { // v3<v4<v5
          int ijk=(n*(n+1)*(n+2))/2;
          int ij=0;
          for (int j=0;j<=n;j++) {
            for (int i=0;i+j<=n;i++,ij++) {
              indices[ij]=ijk;
              if (i+j<n) ijk++;
            }
            ijk++;
          }
        } else if (vertexIndex(3)<vertexIndex(5)) { // v3<v5<v4
          int ijk=(n*(n+1)*(n+2))/2;
          int ij=0;
          for (int i=0;i<=n;i++) {
            for (int j=0;i+j<=n;j++,ij++) {
              indices[ij]=ijk;
              if (i+j<n) ijk+=n+1-j;
            }
            ijk-=(n*(n+3)-i*(i+3))/2-1;
          }
        } else { // v5<v3<v4
          int ijk=(n*(n*(n+4)+5))/2;
          int ij=0;
          for (int i=0;i<=n;i++) {
            for (int j=n-i;j>=0;j--,ij++) {
              indices[ij]=ijk;
              if (j>0) ijk-=n+2-j;
            }
            ijk+=(n*(n+3)-i*(i+5))/2-1;
          }
        }
      } else {
        if (vertexIndex(3)<vertexIndex(5)) { // v4<v3<v5
          int ijk=(n*(n*(n+3)+4))/2;
          int ij=0;
          for (int j=0;j<=n;j++) {
            for (int i=n-j;i>=0;i--,ij++) {
              indices[ij]=ijk;
              if (i>0) ijk--;
            }
            ijk+=2*(n-j);
          }
        } else if (vertexIndex(4)<vertexIndex(5)) { // v4<v5<v3
          int ijk=(n*(n*(n+3)+4))/2;
          int ij=0;
          for (int i=0;i<=n;i++) {
            for (int j=0;i+j<=n;j++,ij++) {
              indices[ij]=ijk;
              if (i+j<n) ijk+=n-j;
            }
            ijk-=(n*(n+1)-i*(i+1))/2+1;
          }
        } else { // v5<v4<v3
          int ijk=(n*(n*(n+4)+5))/2;
          int ij=0;
          for (int i=0;i<=n;i++) {
            for (int j=n-i;j>=0;j--,ij++) {
              indices[ij]=ijk;
              if (j>0) ijk-=n+1-j;
            }
            ijk+=(n*(n+1)-i*(i+3))/2-2;
          }
        }
      }
      break;
    }
    case 2: { // (0,j,k)
      indices.allocate((n+1)*(n+1));
      if (vertexIndex(2)<vertexIndex(3)) { 
        int ijk=0;
        int jk=0;
        for (int k=0;k<=n;k++) {
          for (int j=0;j<=n;j++,jk++) {
            indices[jk]=ijk;
            if (j<n) ijk+=n+1-j;
          }
          ijk++;
        }
      } else {
        int ijk=0;
        int jk=0;
        for (int j=0;j<=n;j++) {
          for (int k=0;k<=n;k++,jk++) {
            indices[jk]=ijk;
            if (k<n) ijk+=((n+1)*(n+2))/2;
          }
          ijk-=(n*n*(n+3))/2+j-1;
        }
      }
      break;
    }
    case 3: {
      indices.allocate((n+1)*(n+1));
      if (vertexIndex(1)<vertexIndex(3)) { 
        int ijk=0;
        int ik=0;
        for (int k=0;k<=n;k++) {
          for (int i=0;i<=n;i++,ik++) {
            indices[ik]=ijk;
            if (i<n) ijk++;
          }
          ijk+=(n*(n+1))/2+1;
        }
      } else {
        int ijk=0;
        int ik=0;
        for (int i=0;i<=n;i++) {
          for (int k=0;k<=n;k++,ik++) {
            indices[ik]=ijk;
            if (k<n) ijk+=(n*(n+3))/2+1;
          }
          ijk-=(n*(n+1)*(n+2))/2-1;
        }
      }
      break;
    }
    case 4: {
      indices.allocate((n+1)*(n+1));
      int vmin=min(min(vertexIndex(1),vertexIndex(2)),
                   min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(1)==vmin) {
        if (vertexIndex(2)<vertexIndex(4)) {
          int ijk=n;
          int jk=0;
          for (int k=0;k<=n;k++) {
            for (int j=0;j<=n;j++,jk++) {
              indices[jk]=ijk;
              if (j<n) ijk+=n-j;
            }
            ijk+=n+1;
          }
        } else {
          int ijk=n;
          int jk=0;
          for (int j=0;j<=n;j++) {
            for (int k=0;k<=n;k++,jk++) {
              indices[jk]=ijk;
              if (k<n) ijk+=(n*(n+3))/2+1;
            }
            ijk-=(n*n*(n+3))/2+j;
          }
        }
      } else if (vertexIndex(2)==vmin) {
        if (vertexIndex(1)<vertexIndex(5)) {
          int ijk=(n*(n+3))/2;
          int ik=0;
          for (int k=0;k<=n;k++) {
            for (int i=0;i<=n;i++,ik++) {
              indices[ik]=ijk;
              if (i<n) ijk-=i+1;
            }
            ijk+=(n+1)*(n+1);
          }
        } else {
          int ijk=(n*(n+3))/2;
          int ik=0;
          for (int i=0;i<=n;i++) {
            for (int k=0;k<=n;k++,ik++) {
              indices[ik]=ijk;
              if (k<n) ijk+=(n*(n+3))/2+1;
            }
            ijk-=(n*(n+1)*(n+2))/2+i+1;
          }
        }
      } else if (vertexIndex(4)==vmin) {
        if (vertexIndex(1)<vertexIndex(5)) {
          int ijk=(n*(n*(n+3)+4))/2;
          int jk=0;
          for (int j=0;j<=n;j++) {
            for (int k=n;k>=0;k--,jk++) {
              indices[jk]=ijk;
              if (k>0) ijk-=(n*(n+3))/2+1;
            }
            ijk+=(n*(n*(n+3)+4))/2-j;
          }
        } else {
          int ijk=(n*(n*(n+3)+4))/2;
          int jk=0;
          for (int k=n;k>=0;k--) {
            for (int j=0;j<=n;j++,jk++) {
              indices[jk]=ijk;
              if (j<n) ijk+=n-j;
            }
            ijk-=(n+1)*(n+1);
          }
        }
      } else {
        if (vertexIndex(2)<vertexIndex(4)) {
          int ijk=(n*(n*(n+4)+5))/2;
          int ik=0;
          for (int i=0;i<=n;i++) {
            for (int k=n;k>=0;k--,ik++) {
              indices[ik]=ijk;
              if (k>0) ijk-=(n*(n+3))/2+1;
            }
            ijk+=(n*(n+1)*(n+2))/2-i-1;
          }
        } else {
          int ijk=(n*(n*(n+4)+5))/2;
          int ik=0;
          for (int k=n;k>=0;k--) {
            for (int i=0;i<=n;i++,ik++) {
              indices[ik]=ijk;
              if (i<n) ijk-=i+1;
            }
            ijk-=n+1;
          }
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::staticEquallySpacedLatticePointIndicesOfFaceInterior(int f,
int n,NumPtr<int> &indices) {
  switch (f) {
    case 0: { // (i,j,0)
      indices.allocate(
        Triangle::staticNumberInteriorEquallySpacedLatticePoints(n));
      int ijk=n+2;
      int ij=0;
      for (int j=1;j<n;j++) {
        for (int i=1;i+j<n;i++,ij++) {
          indices[ij]=ijk;
          ijk++;
        }
        ijk+=2;
      }
      break;
    }
    case 1: { // (i,j,n)
      indices.allocate(
        Triangle::staticNumberInteriorEquallySpacedLatticePoints(n));
      int ijk=n*((n*(n+3))/2+2)+2;
      int ij=0;
      for (int j=1;j<n;j++) {
        for (int i=1;i+j<n;i++,ij++) {
          indices[ij]=ijk;
          ijk++;
        }
        ijk+=2;
      }
      break;
    }
    case 2: { // (0,j,k)
      indices.allocate(
        Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(n));
      int ijk=(n*(n+5))/2+2;
      int jk=0;
      for (int k=1;k<n;k++) {
        for (int j=1;j<n;j++,jk++) {
          indices[jk]=ijk;
          ijk+=n+1-j;
        }
        ijk+=n+2;
      }
      break;
    }
    case 3: { // (i,0,k)
      indices.allocate(
        Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(n));
      int ijk=(n*(n+3))/2+2;
      int ik=0;
      for (int k=1;k<n;k++) {
        for (int i=1;i<n;i++,ik++) {
          indices[ik]=ijk;
          ijk++;
        }
        ijk+=(n*(n+1))/2+2;
      }
      break;
    }
    case 4: { // (n-j,j,k)
      indices.allocate(
        Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(n));
      int ijk=(n*(n+7))/2+1;
      int jk=0;
      for (int k=1;k<n;k++) {
        for (int j=1;j<n;j++,jk++) {
          indices[jk]=ijk;
          ijk+=n-j;
        }
        ijk+=2*n+1;
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::staticEquallySpacedLatticePointIndicesOfLine(int l,int n,
NumPtr<int> &indices) {
  indices.allocate(Interval::staticNumberEquallySpacedLatticePoints(n));
  switch (l) {
    case 0: {
      int ijk=0;
      for (int i=0;i<=n;i++) {
        indices[i]=ijk;
        ijk++;
      }
      break;
    }
    case 1: {
      int ijk=n;
      for (int j=0;j<=n;j++) {
        indices[j]=ijk;
        ijk+=n-j;
      }
      break;
    }
    case 2: {
      int ijk=0;
      for (int j=0;j<=n;j++) {
        indices[j]=ijk;
        ijk+=n+1-j;
      }
      break;
    }
    case 3: {
      int ijk=n*((n+1)*(n+2))/2;
      for (int i=0;i<=n;i++) {
        indices[i]=ijk;
        ijk++;
      }
      break;
    }
    case 4: {
      int ijk=n*((n+1)*(n+2))/2+n;
      for (int j=0;j<=n;j++) {
        indices[j]=ijk;
        ijk+=n-j;
      }
      break;
    }
    case 5: {
      int ijk=n*((n+1)*(n+2))/2;
      for (int j=0;j<=n;j++) {
        indices[j]=ijk;
        ijk+=n+1-j;
      }
      break;
    }
    case 6: {
      int nt=((n+1)*(n+2))/2;
      int ijk=0;
      for (int k=0;k<=n;k++) {
        indices[k]=ijk;
        ijk+=nt;
      }
      break;
    }
    case 7: {
      int nt=((n+1)*(n+2))/2;
      int ijk=n;
      for (int k=0;k<=n;k++) {
        indices[k]=ijk;
        ijk+=nt;
      }
      break;
    }
    case 8: {
      int nt=((n+1)*(n+2))/2;
      int ijk=(n*(n+3))/2;
      for (int k=0;k<=n;k++) {
        indices[k]=ijk;
        ijk+=nt;
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_L::constraints
void Prism::equallySpacedLatticePointIndicesOfFaceInterior(int f,int n,
NumPtr<int> &indices) const {
  if (n<=1) return;
  switch (f) {
    case 0: { // (i,j,0)
      indices.allocate(((n-2)*(n-1))/2);
      if (vertexIndex(1)<vertexIndex(2)) {
        int ijk=n+2;
        int ij=0;
        for (int j=1;j<n;j++) {
          for (int i=1;i+j<n;i++,ij++) {
            indices[ij]=ijk;
            ijk++;
          }
          ijk+=2;
        }
      } else {
        int ijk=n+2;
        int ij=0;
        for (int i=1;i<n;i++) {
          for (int j=1;i+j<n;j++,ij++) {
            indices[ij]=ijk;
            ijk+=n+1-j;
          }
          ijk-=(n*(n+1)-i*(i+3))/2-2;
        }
      }
      break;
    }
    case 1: { // (i,j,n)
      indices.allocate(((n-2)*(n-1))/2);
      if (vertexIndex(3)<vertexIndex(4)) {
        if (vertexIndex(4)<vertexIndex(5)) { // v3<v4<v5
          int ijk=(n*(n*(n+3)+4))/2+2;
          int ij=0;
          for (int j=1;j<n;j++) {
            for (int i=1;i+j<n;i++,ij++) {
              indices[ij]=ijk;
              ijk++;
            }
            ijk+=2;
          }
        } else if (vertexIndex(3)<vertexIndex(5)) { // v3<v5<v4
          int ijk=(n*(n*(n+3)+4))/2+2;
          int ij=0;
          for (int i=1;i<n;i++) {
            for (int j=1;i+j<n;j++,ij++) {
              indices[ij]=ijk;
              ijk+=n+1-j;
            }
            ijk-=(n*(n+1)-i*(i+3))/2-2;
          }
        } else { // v5<v3<v4
          int ijk=(n*(5+n*(4+n)))/2-4;
          int ij=0;
          for (int i=1;i<n;i++) {
            for (int j=n-i-1;j>=1;j--,ij++) {
              indices[ij]=ijk;
              ijk-=n+2-j;
            }
            ijk+=(n*(n+3)-i*(i+7))/2-4;
          }
        }
      } else {
        if (vertexIndex(3)<vertexIndex(5)) { // v4<v3<v5
          int ijk=(n*(n*(n+3)+6))/2-1;
          int ij=0;
          for (int j=1;j<n;j++) {
            for (int i=n-j-1;i>=1;i--,ij++) {
              indices[ij]=ijk;
              ijk--;
            }
            ijk+=2*(n-j)-1;
          }
        } else if (vertexIndex(4)<vertexIndex(5)) { // v4<v5<v3
          int ijk=(n*(n*(n+3)+6))/2-1;
          int ij=0;
          for (int i=1;i<n;i++) {
            for (int j=1;i+j<n;j++,ij++) {
              indices[ij]=ijk;
              ijk+=n-j;
            }
            ijk-=(n*(n-1)-i*(i+1))/2+1;
          }
        } else { // v5<v4<v3
          int ijk=(n*(5+n*(4+n)))/2-4;
          int ij=0;
          for (int i=1;i<n;i++) {
            for (int j=n-i-1;j>=1;j--,ij++) {
              indices[ij]=ijk;
              ijk-=n+1-j;
            }
            ijk+=(n*(n+1)-i*(i+5))/2-4;
          }
        }
      }
      break;
    }
    case 2: { // (0,j,k)
      indices.allocate((n-1)*(n-1));
      if (vertexIndex(2)<vertexIndex(3)) { 
        int ijk=((n+1)*(n+4))/2;
        int jk=0;
        for (int k=1;k<n;k++) {
          for (int j=1;j<n;j++,jk++) {
            indices[jk]=ijk;
            ijk+=n+1-j;
          }
          ijk+=n+2;
        }
      } else {
        int ijk=((n+1)*(n+4))/2;
        int jk=0;
        for (int j=1;j<n;j++) {
          for (int k=1;k<n;k++,jk++) {
            indices[jk]=ijk;
            ijk+=((n+1)*(n+2))/2;
          }
          ijk-=(n*(n+3)*(n-1))/2+j-2;
        }
      }
      break;
    }
    case 3: {
      indices.allocate((n-1)*(n-1));
      if (vertexIndex(1)<vertexIndex(3)) { 
        int ijk=(n*(n+3))/2+2;
        int ik=0;
        for (int k=1;k<n;k++) {
          for (int i=1;i<n;i++,ik++) {
            indices[ik]=ijk;
            ijk++;
          }
          ijk+=(n*(n+1))/2+2;
        }
      } else {
        int ijk=(n*(n+3))/2+2;
        int ik=0;
        for (int i=1;i<n;i++) {
          for (int k=1;k<n;k++,ik++) {
            indices[ik]=ijk;
            ijk+=(n*(n+3))/2+1;
          }
          ijk-=(n*(-1+n*(2+n)))/2-2;
        }
      }
      break;
    }
    case 4: {
      indices.allocate((n-1)*(n-1));
      int vmin=min(min(vertexIndex(1),vertexIndex(2)),
                   min(vertexIndex(4),vertexIndex(5)));
      if (vertexIndex(1)==vmin) {
        if (vertexIndex(2)<vertexIndex(4)) {
          int ijk=(n*(n+7))/2+1;
          int jk=0;
          for (int k=1;k<n;k++) {
            for (int j=1;j<n;j++,jk++) {
              indices[jk]=ijk;
              ijk+=n-j;
            }
            ijk+=2*n+1;
          }
        } else {
          int ijk=(n*(n+7))/2+1;
          int jk=0;
          for (int j=1;j<n;j++) {
            for (int k=1;k<n;k++,jk++) {
              indices[jk]=ijk;
              ijk+=(n*(n+3))/2+1;
            }
            ijk-=(n*(n*(n+2)-3))/2+j-1;
          }
        }
      } else if (vertexIndex(2)==vmin) {
        if (vertexIndex(1)<vertexIndex(5)) {
          int ijk=n*(n+3);
          int ik=0;
          for (int k=1;k<n;k++) {
            for (int i=1;i<n;i++,ik++) {
              indices[ik]=ijk;
              ijk-=i+1;
            }
            ijk+=n*(n+2);
          }
        } else {
          int ijk=n*(n+3);
          int ik=0;
          for (int i=1;i<n;i++) {
            for (int k=1;k<n;k++,ik++) {
              indices[ik]=ijk;
              ijk+=(n*(n+3))/2+1;
            }
            ijk-=(n*(-1+n*(2+n)))/2+i;
          }
        }
      } else if (vertexIndex(4)==vmin) {
        if (vertexIndex(1)<vertexIndex(5)) {
          int ijk=(n*(3+n*(2+n)))/2-1;
          int jk=0;
          for (int j=1;j<n;j++) {
            for (int k=n-1;k>=1;k--,jk++) {
              indices[jk]=ijk;
              ijk-=(n*(n+3))/2+1;
            }
            ijk+=(n*(1+n*(2+n)))/2-j-1;
          }
        } else {
          int ijk=(n*(3+n*(2+n)))/2-1;
          int jk=0;
          for (int k=n-1;k>=1;k--) {
            for (int j=1;j<n;j++,jk++) {
              indices[jk]=ijk;
              ijk+=n-j;
            }
            ijk-=n*(n+1)+1;
          }
        }
      } else {
        if (vertexIndex(2)<vertexIndex(4)) {
          int ijk=(n*(2+n*(3+n)))/2-2;
          int ik=0;
          for (int i=1;i<n;i++) {
            for (int k=n-1;k>=1;k--,ik++) {
              indices[ik]=ijk;
              ijk-=(n*(n+3))/2+1;
            }
            ijk+=(n*(-1+n*(2+n)))/2-i-2;
          }
        } else {
          int ijk=(n*(2+n*(3+n)))/2-2;
          int ik=0;
          for (int k=n-1;k>=1;k--) {
            for (int i=1;i<n;i++,ik++) {
              indices[ik]=ijk;
              ijk-=i+1;
            }
            ijk-=n+2;
          }
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::staticEquallySpacedLatticePointIndicesOfLineInterior(int l,
int n,NumPtr<int> &indices) {
  indices.allocate(n-1);
  if (n<=1) return;
  switch (l) {
    case 0: {
      int ijk=1;
      for (int i=1;i<n;i++) {
        indices[i-1]=ijk;
        ijk++;
      }
      break;
    }
    case 1: {
      int ijk=2*n;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n-j;
      }
      break;
    }
    case 2: {
      int ijk=n+1;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n+1-j;
      }
      break;
    }
    case 3: {
      int ijk=(n*n*(n+3))/2+n+1;
      for (int i=1;i<n;i++) {
        indices[i-1]=ijk;
        ijk++;
      }
      break;
    }
    case 4: {
      int ijk=(n*n*(n+3))/2+3*n;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n-j;
      }
      break;
    }
    case 5: {
      int ijk=(n*n*(n+3))/2+2*n+1;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n+1-j;
      }
      break;
    }
    case 6: {
      int ijk=(n*(n+3))/2+1;
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+3))/2+1;
      }
      break;
    }
    case 7: {
      int ijk=(n*(n+5))/2+1;
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+3))/2+1;
      }
      break;
    }
    case 8: {
      int ijk=n*(n+3)+1;
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+3))/2+1;
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::equallySpacedLatticePointIndicesOfLineInterior(int l,int n,
NumPtr<int> &indices) const {
  indices.allocate(n-1);
  if (n<=1) return;
  switch (l) {
    case 0: {
      int ijk=1;
      for (int i=1;i<n;i++) {
        indices[i-1]=ijk;
        ijk++;
      }
      break;
    }
    case 1: {
      if (vertexIndex(1)<vertexIndex(2)) {
        int ijk=2*n;
        for (int j=1;j<n;j++) {
          indices[j-1]=ijk;
          ijk+=n-j;
        }
      } else {
        int ijk=(n*(n+3))/2-1;
        for (int i=1;i<n;i++) {
          indices[i-1]=ijk;
          ijk-=i+1;
        }
      }
      break;
    }
    case 2: {
      int ijk=n+1;
      for (int j=1;j<n;j++) {
        indices[j-1]=ijk;
        ijk+=n+1-j;
      }
      break;
    }
    case 3: {
      if (vertexIndex(3)<vertexIndex(4)) {
        int ijk=(n*n*(n+3))/2+n+1;
        for (int i=1;i<n;i++) {
          indices[i-1]=ijk;
          ijk++;
        }
      } else {
        int ijk=(n*n*(n+3))/2+2*n-1;
        int m=0;
        for (int i=n-1;i>=1;i--,m++) {
          indices[m]=ijk;
          ijk--;
        }
      }
      break;
    }
    case 4: {
      if (vertexIndex(4)<vertexIndex(5)) {
        int ijk=(n*n*(n+3))/2+3*n;
        for (int j=1;j<n;j++) {
          indices[j-1]=ijk;
          ijk+=n-j;
        }
      } else {
        int ijk=(n*(n*(n+4)+5))/2-1;
        for (int i=1;i<n;i++) {
          indices[i-1]=ijk;
          ijk-=i+1;
        }
      }
      break;
    }
    case 5: {
      if (vertexIndex(3)<vertexIndex(5)) {
        int ijk=(n*n*(n+3))/2+2*n+1;
        for (int j=1;j<n;j++) {
          indices[j-1]=ijk;
          ijk+=n+1-j;
        }
      } else {
        int ijk=(n*(n*(n+4)+5))/2-2;
        int m=0;
        for (int j=n-1;j>=1;j--,m++) {
          indices[m]=ijk;
          ijk-=n+2-j;
        }
      }
      break;
    }
    case 6: {
      int ijk=(n*(n+3))/2+1;
      for (int k=1;k<n;k++) {
        indices[k-1]=ijk;
        ijk+=(n*(n+3))/2+1;
      }
      break;
    }
    case 7: {
      if (vertexIndex(1)<vertexIndex(4)) {
        int ijk=(n*(n+5))/2+1;
        for (int k=1;k<n;k++) {
          indices[k-1]=ijk;
          ijk+=(n*(n+3))/2+1;
        }
      } else {
        int ijk=(n*(n*(n+2)+1))/2-1;
        int m=0;
        for (int k=n-1;k>=1;k--,m++) {
          indices[m]=ijk;
          ijk-=(n*(n+3))/2+1;
        }
      }
      break;
    }
    case 8: {
      if (vertexIndex(2)<vertexIndex(5)) {
        int ijk=n*(n+3)+1;
        for (int k=1;k<n;k++) {
          indices[k-1]=ijk;
          ijk+=(n*(n+3))/2+1;
        }
      } else {
        int ijk=(n*(n*(n+3)+2))/2-1;
        int m=0;
        for (int k=n-1;k>=1;k--,m++) {
          indices[m]=ijk;
          ijk-=(n*(n+3))/2+1;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::reorderLineDofs(int n,NumPtr<int> &renumber) const {
  int nl=n+1;
  for (int i=0;i<9*nl;i++) renumber[i]=i;
  int offset=0;
  NumPtr<int> r=renumber;
  for (int l=0;l<9;l++) {
    int v[2]={vertexIndex(elementVertexFromLine(l,0)),
              vertexIndex(elementVertexFromLine(l,1))};
    Interval line_element(v,-1);
    line_element.reorderDofs(n,r);
    for (int i=0;i<nl;i++) r[i]+=offset;
    r+=nl;
    offset+=nl;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<9*nl;i++) {
    if (renumber[i]>=9*nl || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(9*nl);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<9*nl;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<9*nl;i++) CHECK_BOUNDS(renumber_inverse[i],0,9*nl);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::reorderRTFaceDofs(int line_offset,int n,
NumPtr<int> &renumber) const {
  int nq=(n+1)*(n+1);
  int nt=((n+2)*(n+1))/2;
  for (int i=0;i<2*nt+3*nq;i++) renumber[i]=i;
  int offset=line_offset;
  NumPtr<int> r=renumber;
  int tl[3]={0,1,2};
  for (int f=0;f<2;f++) {
    int v[3]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2))};
    Triangle face_element(tl,v,-1);
    face_element.reorderDofs(n,r);
    for (int i=0;i<nt;i++) r[i]+=offset;
    r+=nt;
    offset+=nt;
  }
  int ql[4]={0,1,2,3};
  for (int f=2;f<5;f++) {
    int v[4]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2)),
              vertexIndex(elementVertexFromFace(f,3))};
    Quadrilateral face_element(ql,v,-1);
    face_element.reorderDofs(n,n,r);
    for (int i=0;i<nq;i++) r[i]+=offset;
    r+=nq;
    offset+=nq;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<2*nt+3*nq;i++) {
    if (renumber[i]-line_offset>=2*nt+3*nq || renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(2*nt+3*nq);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<2*nt+3*nq;i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<2*nt+3*nq;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,2*nt+3*nq);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::reorderBDFMFaceDofs(int line_offset,int n,
NumPtr<int> &renumber) const {
  int nt=((n+2)*(n+1))/2;
  int nq=nt;
  for (int i=0;i<2*nt+3*nq;i++) renumber[i]=i;
  int offset=line_offset;
  NumPtr<int> r=renumber;
  int tl[3]={0,1,2};
  for (int f=0;f<2;f++) {
    int v[3]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2))};
    Triangle face_element(tl,v,-1);
    face_element.reorderDofs(n,r);
    for (int i=0;i<nt;i++) r[i]+=offset;
    r+=nt;
    offset+=nt;
  }
  for (int f=2;f<5;f++) {
    int v[4]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2)),
              vertexIndex(elementVertexFromFace(f,3))};
    int vmin=min(min(v[0],v[1]),min(v[2],v[3]));
    if (v[0]==vmin) {
      int v3[3]={v[0],v[1],v[2]};
      Triangle face_element(tl,v3,-1);
      face_element.reorderDofs(n,r);
    } else if (v[1]==vmin) {
      int v3[3]={v[1],v[3],v[0]};
      Triangle face_element(tl,v3,-1);
      face_element.reorderDofs(n,r);
    } else if (v[2]==vmin) {
      int v3[3]={v[2],v[0],v[3]};
      Triangle face_element(tl,v3,-1);
      face_element.reorderDofs(n,r);
    } else {
      int v3[3]={v[3],v[2],v[1]};
      Triangle face_element(tl,v3,-1);
      face_element.reorderDofs(n,r);
    }
    for (int i=0;i<nq;i++) r[i]+=offset;
    r+=nq;
    offset+=nq;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<2*nt+3*nq;i++) {
    if (renumber[i]-line_offset>=2*nt+3*nq || renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(2*nt+3*nq);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<2*nt+3*nq;i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<2*nt+3*nq;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,2*nt+3*nq);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::reorderHcurlFaceDofs(int line_offset,int n,
NumPtr<int> &renumber) const {
  int nq=n*(n+1);
  int nt=nq/2;
  for (int i=0;i<4*nt+6*nq;i++) renumber[i]=i;
  int offset=line_offset;
  NumPtr<int> r=renumber;
  int tl[3]={0,1,2};
  for (int f=0;f<2;f++) {
    int v[3]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2))};
    Triangle face_element(tl,v,-1);
    face_element.reorderDofs(n-1,r);
    if (!faceTangentsWindBackward(f)) {
      for (int i=0;i<nt;i++) {
        r[i]+=offset;
        r[i+nt]=r[i]+nt;
      }
    } else {
      for (int i=0;i<nt;i++) {
        r[i]+=nt+offset;
        r[i+nt]=r[i]-nt;
      }
    }
    r+=2*nt;
    offset+=2*nt;
  }
  int ql[4]={0,1,2,3};
  for (int f=2;f<5;f++) {
    int v[4]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2)),
              vertexIndex(elementVertexFromFace(f,3))};
    Quadrilateral face_element(ql,v,-1);
    if (!faceTangentsWindBackward(f)) {
      face_element.reorderDofs(n,n-1,r);
      for (int i=0;i<nq;i++) r[i]+=offset;
      r+=nq;
      offset+=nq;
      face_element.reorderDofs(n-1,n,r);
      for (int i=0;i<nq;i++) r[i]+=offset;
      r+=nq;
    } else {
      face_element.reorderDofs(n-1,n,r);
      for (int i=0;i<nq;i++) r[i]+=offset+nq;
      r+=nq;
      face_element.reorderDofs(n,n-1,r);
      for (int i=0;i<nq;i++) r[i]+=offset;
      offset+=nq;
      r+=nq;
    }
    offset+=nq;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<4*nt+6*nq;i++) {
    if (renumber[i]-line_offset>=4*nt+6*nq || renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(4*nt+6*nq);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<4*nt+6*nq;i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<4*nt+6*nq;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,4*nt+6*nq);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::reorderHcurlFaceDofs2(int line_offset,int n,
NumPtr<int> &renumber) const {
  if (n<=1) return;
  int nq=(n-1)*(n+1);
  int nt=(n*(n-1))/2;
  int nh=n-1;
  for (int i=0;i<4*nt+2*nh+6*nq;i++) renumber[i]=i;
  int offset=line_offset;
  NumPtr<int> r=renumber;
  int tl[3]={0,1,2};
  for (int f=0;f<2;f++) {
    int v[3]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2))};
    Triangle face_element(tl,v,-1);
    face_element.reorderDofs(n-2,r);
    if (!faceTangentsWindBackward(f)) {
      for (int i=0;i<nt;i++) {
        r[i]+=offset;
        r[i+nt]=r[i]+nt;
      }
    } else {
      for (int i=0;i<nt;i++) {
        r[i]+=nt+offset;
        r[i+nt]=r[i]-nt;
      }
    }
    r+=2*nt;
    offset+=2*nt;
    for (int i=0;i<nh;i++) r[i]=offset+i;
    r+=nh;
    offset+=nh;
  }
  int ql[4]={0,1,2,3};
  for (int f=2;f<5;f++) {
    int v[4]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2)),
              vertexIndex(elementVertexFromFace(f,3))};
    Quadrilateral face_element(ql,v,-1);
    face_element.reorderDofs(n,n-2,r);
    for (int i=0;i<nq;i++) r[i]+=offset;
    r+=nq;
    offset+=nq;
    face_element.reorderDofs(n-2,n,r);
    for (int i=0;i<nq;i++) r[i]+=offset;
    r+=nq;
    offset+=nq;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<4*nt+2*nh+6*nq;i++) {
    if (renumber[i]-line_offset>=4*nt+2*nh+6*nq ||
    renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(4*nt+2*nh+6*nq);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<4*nt+2*nh+6*nq;i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<4*nt+2*nh+6*nq;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,4*nt+2*nh+6*nq);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::reorderEquallySpacedLatticePoints(int n,
NumPtr<int> &renumber) const {
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  int nt=((n+1)*(n+2))/2;
//vertices first
  int m=0;
  renumber[0]=m++;
  renumber[n]=m++;
  renumber[nt-1]=m++;
  renumber[n*nt]=m++;
  renumber[n*nt+n]=m++;
  renumber[renumber.getNumber()-1]=m++;
//line 0
  if (vertexIndex(0)<vertexIndex(1)) {
    for (int i=1;i<n;i++) renumber[i]=m++;
  } else {
    for (int i=n-1;i>0;i--) renumber[i]=m++;
  } 
//line 1
  if (vertexIndex(1)<vertexIndex(2)) {
    int l=2*n;
    for (int j=1;j<n;j++) {
      renumber[l]=m++;
      l+=n-j;
    }
  } else {
    int l=(n*(n+3))/2-1;
    for (int i=1;i<n;i++) {
      renumber[l]=m++;
      l-=i+1;
    }
  } 
//line 2
  if (vertexIndex(0)<vertexIndex(2)) {
    int l=n+1;
    for (int j=1;j<n;j++) {
      renumber[l]=m++;
      l+=n+1-j;
    }
  } else {
    int l=(n*(n+3))/2-2;
    for (int j=n-1;j>0;j--) {
      renumber[l]=m++;
      l-=n-j+2;
    }
  } 
//line 3
  if (vertexIndex(3)<vertexIndex(4)) {
    int l=n*nt+1;
    for (int i=1;i<n;i++) renumber[l++]=m++;
  } else {
    int l=n*nt+n-1;
    for (int i=n-1;i>0;i--) {
      renumber[l--]=m++;
    }
  } 
//line 4
  if (vertexIndex(4)<vertexIndex(5)) {
    int l=n*(n*(n+3)+6)/2;
    for (int j=1;j<n;j++) {
      renumber[l]=m++;
      l+=n-j;
    }
  } else {
    int l=renumber.getNumber()-2;
    for (int i=1;i<n;i++) {
      renumber[l]=m++;
      l-=i+1;
    }
  } 
//line 5
  if (vertexIndex(3)<vertexIndex(5)) {
    int l=n*nt+n+1;
    for (int j=1;j<n;j++) {
      renumber[l]=m++;
      l+=n-j+1;
    }
  } else {
    int l=renumber.getNumber()-3;
    for (int j=n-1;j>0;j--) {
      renumber[l]=m++;
      l-=n-j+2;
    }
  } 
//line 6
  if (vertexIndex(0)<vertexIndex(3)) {
    int l=nt;
    for (int k=1;k<n;k++) {
      renumber[l]=m++;
      l+=nt;
    }
  } else {
    int l=(n-1)*nt;
    for (int k=n-1;k>0;k--) {
      renumber[l]=m++;
      l-=nt;
    }
  } 
//line 7
  if (vertexIndex(1)<vertexIndex(4)) {
    int l=(n*(n+5)+2)/2;
    for (int k=1;k<n;k++) {
      renumber[l]=m++;
      l+=nt;
    }
  } else {
    int l=(n*(n*(n+2)+1)-2)/2;
    for (int k=n-1;k>0;k--) {
      renumber[l]=m++;
      l-=nt;
    }
  } 
//line 8
  if (vertexIndex(2)<vertexIndex(5)) {
    int l=2*nt-1;
    for (int k=1;k<n;k++) {
      renumber[l]=m++;
      l+=nt;
    }
  } else {
    int l=n*nt-1;
    for (int k=n-1;k>0;k--) {
      renumber[l]=m++;
      l-=nt;
    }
  } 
//face 0
  if (vertexIndex(0)<vertexIndex(1)) {
    if (vertexIndex(1) < vertexIndex(2)) { // 0 < 1 < 2
      int l=n+2;
      for (int j=1;j<n-1;j++) {
        int k=l;
        for (int i=1;i<n-j;i++) {
          renumber[k++]=m++;
        }
        l+=n-j+1;
      }
    } else if (vertexIndex(0) < vertexIndex(2)) { // 0 < 2 < 1
      int l=n+2;
      for (int i=1;i<n-1;i++) {
        int k=l;
        for (int j=1;j<n-i;j++) {
          renumber[k]=m++;
          k+=n-j+1;
        }
        l++;
      }
    } else { // 2 < 0 < 1
      int l=(n*(n+3))/2-4;
      for (int i=1;i<n-1;i++) {
        int k=l;
        for (int j=n-i-1;j>0;j--) {
          renumber[k]=m++;
          k-=n-j+2;
        }
        l-=i+2;
      }
    }
  } else {
    if (vertexIndex(0)<vertexIndex(2)) { // 1 < 0 < 2
      int l=2*n-1;
      for (int j=1;j<n-1;j++) {
        int k=l;
        for (int i=n-j-1;i>0;i--) {
          renumber[k--]=m++;
        }
        l+=n-j;
      }
    } else if (vertexIndex(1)<vertexIndex(2)) { // 1 < 2 < 0
      int l=2*n-1;
      for (int i=n-2;i>0;i--) {
        int k=l;
        for (int j=1;j<=i;j++) {
          renumber[k]=m++;
          k+=n-j;
        }
        l--;
      }
    } else { // 2 < 1 < 0
      int l=(n*(n+3))/2-4;
      for (int j=n-2;j>0;j--) {
        int k=l;
        for (int i=1;i<=j;i++) { // i+j'=j+1
          renumber[k]=m++;
          k-=n-j+i;
        }
        l-=n-j+2;
      }
    }
  }
//face 1
  if (vertexIndex(3)<vertexIndex(4)) {
    if (vertexIndex(4) < vertexIndex(5)) { // 3 < 4 < 5
      int l=n*nt+n+2;
      for (int j=1;j<n-1;j++) {
        int k=l;
        for (int i=1;i<n-j;i++) {
          renumber[k++]=m++;
        }
        l+=n-j+1;
      }
    } else if (vertexIndex(3) < vertexIndex(5)) { // 3 < 5 < 4
      int l=n*nt+n+2;
      for (int i=1;i<n-1;i++) {
        int k=l;
        for (int j=1;j<n-i;j++) {
          renumber[k]=m++;
          k+=n-j+1;
        }
        l++;
      }
    } else { // 5 < 3 < 4
      int l=renumber.getNumber()-5;
      for (int i=1;i<n-1;i++) {
        int k=l;
        for (int j=n-i-1;j>0;j--) {
          renumber[k]=m++;
          k-=n-j+2;
        }
        l-=i+2;
      }
    }
  } else {
    if (vertexIndex(3)<vertexIndex(5)) { // 4 < 3 < 5
      int l=n*nt+2*n-1;
      for (int j=1;j<n-1;j++) {
        int k=l;
        for (int i=n-j-1;i>0;i--) {
          renumber[k--]=m++;
        }
        l+=n-j;
      }
    } else if (vertexIndex(4)<vertexIndex(5)) { // 4 < 5 < 3
      int l=n*nt+2*n-1;
      for (int i=n-2;i>0;i--) {
        int k=l;
        for (int j=1;j<=i;j++) {
          renumber[k]=m++;
          k+=n-j;
        }
        l--;
      }
    } else { // 5 < 4 < 3
      int l=renumber.getNumber()-5;
      for (int j=n-2;j>0;j--) {
        int k=l;
        for (int i=1;i<=j;i++) { // i+j'=j+1
          renumber[k]=m++;
          k-=n-j+i;
        }
        l-=n-j+2;
      }
    }
  }
//face 2
  int v=min(min(vertexIndex(0),vertexIndex(2)),
            min(vertexIndex(3),vertexIndex(5)));
  if (vertexIndex(0)==v) {
    int l=nt+n+1;
    if (vertexIndex(2) < vertexIndex(3)) { // 0 < 2 < 3
      for (int k=1;k<n;k++) {
        int i=l;
        for (int j=1;j<n;j++) {
          renumber[i]=m++;
          i+=n+1-j;
        }
        l+=nt;
      }
    } else { // 0 < 3 < 2
      for (int j=1;j<n;j++) {
        int i=l;
        for (int k=1;k<n;k++) {
          renumber[i]=m++;
          i+=nt;
        }
        l+=n+1-j;
      }
    }
  } else if (vertexIndex(2)==v) {
    int l=n*(n+3)-1;
    if (vertexIndex(0)<vertexIndex(5)) { // 2 < 0 < 5
      for (int k=1;k<n;k++) {
        int i=l;
        for (int j=n-1;j>0;j--) {
          renumber[i]=m++;
          i-=n-j+2;
        }
        l+=nt;
      }
    } else { // 2 < 5 < 0
      for (int j=n-1;j>0;j--) {
        int i=l;
        for (int k=1;k<n;k++) {
          renumber[i]=m++;
          i+=nt;
        }
        l-=n-j+2;
      }
    }
  } else if (vertexIndex(3)==v) {
    int l=(n-1)*nt+n+1;
    if (vertexIndex(0)<vertexIndex(5)) { // 3 < 0 < 5
      for (int j=1;j<n;j++) {
        int i=l;
        for (int k=n-1;k>0;k--) {
          renumber[i]=m++;
          i-=nt;
        }
        l+=n+1-j;
      }
    } else { // 3 < 5 < 0
      for (int k=n-1;k>0;k--) {
        int i=l;
        for (int j=1;j<n;j++) {
          renumber[i]=m++;
          i+=n+1-j;
        }
        l-=nt;
      }
    }
  } else {
    int l=n*nt-3;
    if (vertexIndex(2)<vertexIndex(3)) { // 5 < 2 < 3
      for (int j=n-1;j>0;j--) {
        int i=l;
        for (int k=n-1;k>0;k--) {
          renumber[i]=m++;
          i-=nt;
        }
        l-=n+2-j;
      }
    } else { // 6 < 3 < 2
      for (int k=n-1;k>0;k--) {
        int i=l;
        for (int j=n-1;j>0;j--) {
          renumber[i]=m++;
          i-=n+2-j;
        }
        l-=nt;
      }
    }
  }
//face 3
  v=min(min(vertexIndex(0),vertexIndex(1)),
        min(vertexIndex(3),vertexIndex(4)));
  if (vertexIndex(0)==v) {
    int l=nt+1;
    if (vertexIndex(1) < vertexIndex(3)) { // 0 < 1 < 3
      for (int k=1;k<n;k++) {
        int j=l;
        for (int k=1;k<n;k++) renumber[j++]=m++;
        l+=nt;
      }
    } else { // 0 < 3 < 1
      for (int i=1;i<n;i++) {
        int j=l;
        for (int k=1;k<n;k++) {
          renumber[j]=m++;
          j+=nt;
        }
        l++;
      }
    }
  } else if (vertexIndex(1)==v) {
    int l=nt+n-1;
    if (vertexIndex(0)<vertexIndex(3)) { // 1 < 0 < 3
      for (int k=1;k<n;k++) {
        int j=l;
        for (int i=n-1;i>0;i--) renumber[j--]=m++;
        l+=nt;
      }
    } else { // 1 < 3 < 0
      for (int i=n-1;i>0;i--) {
        int j=l;
        for (int k=1;k<n;k++) {
          renumber[j]=m++;
          j+=nt;
        }
        l--;
      }
    }
  } else if (vertexIndex(3)==v) {
    int l=(n-1)*nt+1;
    if (vertexIndex(0)<vertexIndex(4)) { // 3 < 0 < 4
      for (int i=1;i<n;i++) {
        int j=l;
        for (int k=n-1;k>0;k--) {
          renumber[j]=m++;
          j-=nt;
        }
        l++;
      }
    } else { // 3 < 4 < 0
      for (int k=n-1;k>0;k--) {
        int j=l;
        for (int i=1;i<n;i++) renumber[j++]=m++;
        l-=nt;
      }
    }
  } else {
    int l=(n-1)*(nt+1);
    if (vertexIndex(1)<vertexIndex(3)) { // 4 < 1 < 3
      for (int i=n-1;i>0;i--) {
        int j=l;
        for (int k=n-1;k>0;k--) {
          renumber[j]=m++;
          j-=nt;
        }
        l--;
      }
    } else { // 4 < 3 < 1
      for (int k=n-1;k>0;k--) {
        int j=l;
        for (int i=n-1;i>0;i--) renumber[j--]=m++;
        l-=nt;
      }
    }
  }
//face 4
  v=min(min(vertexIndex(1),vertexIndex(2)),
        min(vertexIndex(4),vertexIndex(5)));
  if (vertexIndex(1)==v) {
    int l=nt+2*n;
    if (vertexIndex(2) < vertexIndex(4)) { // 1 < 2 < 4
      for (int k=1;k<n;k++) {
        int j=l;
        for (int i=n-1;i>0;i--) {
          renumber[j]=m++;
          j+=i;
        }
        l+=nt;
      }
    } else { // 1 < 4 < 2
      for (int j=1;j<n;j++) {
        int i=l;
        for (int k=1;k<n;k++) {
          renumber[i]=m++;
          i+=nt;
        }
        l+=n-j;
      }
    }
  } else if (vertexIndex(2)==v) {
    int l=n*(n+3);
    if (vertexIndex(1)<vertexIndex(5)) { // 2 < 1 < 5
      for (int k=1;k<n;k++) {
        int j=l;
        for (int i=1;i<n;i++) {
          renumber[j]=m++;
          j-=i+1;
        }
        l+=nt;
      }
    } else { // 2 < 5 < 1
      for (int i=1;i<n;i++) {
        int j=l;
        for (int k=1;k<n;k++) {
          renumber[j]=m++;
          j+=nt;
        }
        l-=i+1;
      }
    }
  } else if (vertexIndex(4)==v) {
    int l=(n-1)*nt+2*n;
    if (vertexIndex(1)<vertexIndex(5)) { // 4 < 1 < 5
      for (int j=1;j<n;j++) {
        int i=l;
        for (int k=n-1;k>0;k--) {
          renumber[i]=m++;
          i-=nt;
        }
        l+=n-j;
      }
    } else { // 4 < 5 < 1
      for (int k=n-1;k>0;k--) {
        int i=l;
        for (int j=1;j<n;j++) {
          renumber[i]=m++;
          i+=n-j;
        }
        l-=nt;
      }
    }
  } else {
    int l=n*nt-2;
    if (vertexIndex(2)<vertexIndex(4)) { // 5 < 2 < 4
      for (int i=1;i<n;i++) {
        int j=l;
        for (int k=n-1;k>0;k--) {
          renumber[j]=m++;
          j-=nt;
        }
        l-=i+1;
      }
    } else { // 5 < 4 < 2
      for (int k=n-1;k>0;k--) {
        int j=l;
        for (int i=1;i<n;i++) {
          renumber[j]=m++;
          j-=i+1;
        }
        l-=nt;
      }
    }
  }
//interior points last
  int l2=nt+n+2;
  for (int k=1;k<n;k++) {
    int l1=l2;
    for (int j=1;j<n;j++) {
      int l=l1;
      for (int i=1;i<n-j;i++) renumber[l++]=m++;
      l1+=n-j+1;
    }
    l2+=nt;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]==INT_MAX) {
      cout << "\trenumber[" << i << "] = *** " << -1 << " ***" << endl;
      success=false;
    } else if (renumber[i]>=renumber.getNumber()) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::reorderChildEquallySpacedLatticePoints(int n,
const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices,NumPtr<int> &renumber) 
const {
  const Prism *child_prism=dynamic_cast<const Prism*>(child_shape);
  CHECK_POINTER(child_prism);

  int reorder_case=-1;
  if (c!=3 && c!=7) {
    int svi=sharedVertexIndex(child_shape);
    CHECK_TEST(svi>=0);
    int pvi=(c<4 ? c : c-1);
    CHECK_TEST(vertexIndex(pvi)==svi);
    reorder_case=child_shape->vertexEqualToVertexIndex(svi);
    CHECK_TEST(reorder_case>=0);
  } else if (c==3) {
    reorder_case=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[1]);
  } else {
    reorder_case=child_shape->
      vertexEqualToVertexIndex(line_interior_child_vertices[4]);
  }
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  switch (reorder_case) {
    case 0: {
      for (int i=0;i<renumber.getNumber();i++) renumber[i]=i;
      break;
    }
    case 1: {
      int ijk=0;
      int nk=n;
      for (int k=0;k<=n;k++) {
        int nl=nk;
        for (int l=0;l<=n;l++) {
          int nj=nl;
          for (int j=0;j+l<=n;j++) {
            renumber[nj]=ijk++;
            nj+=n-j;
          }
          nl-=1;
        }
        nk+=(n*(n+3))/2+1;
      }
      break;
    }
    case 2: {
      int ijk=0;
      int nk=(n*(n+3))/2;
      for (int k=0;k<=n;k++) {
        int ni=nk;
        for (int i=0;i<=n;i++) {
          int nl=ni;
          for (int l=0;i+l<=n;l++) {
            renumber[nl]=ijk++;
            nl-=i+l+2;
          }
          ni-=i+1;
        }
        nk+=(n*(n+3))/2+1;
      }
      break;
    }
    case 3: {
      int ijk=0;
      int nk=(n*(2+n*(3+n)))/2;
      for (int k=n;k>=0;k--) {
        int ni=nk;
        for (int i=0;i<=n;i++) {
          int nj=ni;
          for (int j=0;i+j<=n;j++) {
            renumber[ijk++]=nj;
            nj+=n+1-j;
          }
          ni++;
        }
        nk-=(n*(n+3))/2+1;
      }
      break;
    }
    case 4: {
      int ijk=0;
      int nk=(n*(4+n*(3+n)))/2;
      for (int k=n;k>=0;k--) {
        int nj=nk;
        for (int j=0;j<=n;j++) {
          int nl=nj;
          for (int l=0;j+l<=n;l++) {
            renumber[nl]=ijk++;
            nl--;
          }
          nj+=n-j;
        }
        nk-=(n*(n+3))/2+1;
      }
      break;
    }
    case 5: {
      int ijk=0;
      int nk=(n*(5+n*(4+n)))/2;
      for (int k=n;k>=0;k--) {
        int nl=nk;
        for (int l=0;l<=n;l++) {
          int ni=nl;
          for (int i=0;i+l<=n;i++) {
            renumber[ni]=ijk++;
            ni-=i+l+1;
          }
          nl-=l+2;
        }
        nk-=(n*(n+3))/2+1;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_H::constraints
void Prism::reorderChildDofPoints(int n,const Shape *child_shape,
int c,const NumPtr<int> &line_interior_child_vertices,
NumPtr<int> &renumber) const {
  OBSOLETE("not programmed");
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::staticInteriorLatticePointIndices(int n,NumPtr<int> &v) {
  v.allocate(staticNumberInteriorEquallySpacedLatticePoints(n));
  int ijk=((n+2)*(n+3))/2;
  int iv=0;
  for (int k=1;k<n;k++) {
    for (int j=1;j<n;j++) {
      for (int i=1;i+j<n;i++) {
        v[iv]=ijk;
        iv++;
        ijk++;
      }
      ijk+=2;
    }
    ijk+=n+2;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::computeInteriorLatticePoints(int degree,int bf,
NumPtr<Point<3> > &mapping_support_points) const {
  if (degree<=2) return;
  CHECK_SAME(mapping_support_points.getNumber(),
             staticNumberEquallySpacedLatticePoints(degree));

  switch (bf) {
    case 0:
    case 1: {
      NumPtr<int> indices0;
      NumPtr<int> indices1;
      staticEquallySpacedLatticePointIndicesOfFace(0,degree,indices0);
      staticEquallySpacedLatticePointIndicesOfFace(1,degree,indices1);
      int ijk=((degree+3)*(degree+2))/2;
      for (int k=1;k<degree;k++,ijk+=degree+4) {
        int ij=degree+2;
        for (int j=1;j+1<degree;j++,ijk+=2,ij+=2) {
          for (int i=1;i+j<degree;i++,ijk++,ij++) {
            mapping_support_points[ijk]=
              (mapping_support_points[indices0[ij]]
                *static_cast<REAL>(degree-k)
              +mapping_support_points[indices1[ij]]
                *static_cast<REAL>(k))/static_cast<REAL>(degree);
          }
        }
      }
      break;
    }
    case 2: { // cf Triangle::computeInteriorLatticePoints f=0
      NumPtr<int> indices2;
      NumPtr<int> indices3;
      staticEquallySpacedLatticePointIndicesOfFace(2,degree,indices2);
      staticEquallySpacedLatticePointIndicesOfFace(3,degree,indices3);
      int jk=degree+3;
      int ijk=((degree+3)*(degree+2))/2;
      for (int k=1;k<degree;k++,jk+=2,ijk+=degree+2) {
        for (int j=1;j<degree;j++,jk++,ijk+=2) {
          int ik=jk;
          for (int i=1;i+j<degree;i++,ijk++,ik++) {
            mapping_support_points[ijk]=
              (mapping_support_points[indices2[ik]]*static_cast<REAL>(j)
              +mapping_support_points[indices3[ik]]*static_cast<REAL>(i))
              /static_cast<REAL>(i+j);
          }
        }
      }
      break;
    }
    case 3: { // cf Triangle::computeInteriorLatticePoints f=1
      NumPtr<int> indices3;
      NumPtr<int> indices4;
      staticEquallySpacedLatticePointIndicesOfFace(3,degree,indices3);
      staticEquallySpacedLatticePointIndicesOfFace(4,degree,indices4);
      int ik0=degree+2;
      int ijk=((degree+3)*(degree+2))/2;
      for (int k=1;k<degree;k++,ik0+=degree+1,ijk+=degree+4) {
        for (int j=1;j+1<degree;j++,ijk+=2) {
          int ik=ik0;
          int ki=ik0+degree-2;
          for (int i=1;i+j<degree;i++,ijk++,ik++,ki--) {
            mapping_support_points[ijk]=
              (mapping_support_points[indices3[ik]]
                *static_cast<REAL>(degree-j-i)
              +mapping_support_points[indices4[ki]]*static_cast<REAL>(j))
              /static_cast<REAL>(degree-i);
          }
        }
      }
      break;
    }
    case 4: { // cf Triangle::computeInteriorLatticePoints f=2
      NumPtr<int> indices4;
      NumPtr<int> indices2;
      staticEquallySpacedLatticePointIndicesOfFace(4,degree,indices4);
      staticEquallySpacedLatticePointIndicesOfFace(2,degree,indices2);
      int jk=degree+2;
      int ijk=((degree+3)*(degree+2))/2;
      for (int k=1;k<degree;k++,jk+=2,ijk+=degree+2) {
        for (int j=1;j<degree;j++,jk++,ijk+=2) {
          for (int i=1;i+j<degree;i++,ijk++) {
            mapping_support_points[ijk]=
              (mapping_support_points[indices4[jk]]*static_cast<REAL>(i)
              +mapping_support_points[indices2[jk]]
                *static_cast<REAL>(degree-i-j))
              /static_cast<REAL>(degree-j); 
          }
        }
      }
      break;
    }
    default: {
      OBSOLETE("bad face");
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int line_in_both_faces[5][5] =
  {{-1,-1,2,0,1},{-1,-1,5,3,4},{2,5,-1,6,8},{0,3,6,-1,7},{1,4,8,7,-1}};
int Prism::lineInBothFaces(int f0,int f1) {
  CHECK_TEST(f0!=f1);
  CHECK_BOUNDS(f0,0,5);
  CHECK_BOUNDS(f1,0,5);
  CHECK_TEST(max(f0,f1)>=2);
  return line_in_both_faces[f0][f1];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::initializeMappingSupportPoints(int degree,
const NumPtr<Point<3> > &tessellation_vertices,
NumPtr<Point<3> > &points) const {
  NumPtr<Point<3> > eslp;
  staticEquallySpacedLatticePoints(degree,eslp);
  int neslp=eslp.getNumber();
  NumPtr<Point<3> > wv(6);
  for (int v=0;v<6;v++) {
    wv[v]=tessellation_vertices[vertexIndex(v)];
  }
  for (int i=0;i<neslp;i++) {
    const Point<3> &eslpi=eslp[i];
    REAL bc2=1.-eslpi[0]-eslpi[1];
    points[i]=(wv[1]*eslpi[0]+wv[2]*eslpi[1]+wv[0]*bc2)*(1.-eslpi[2])
             +(wv[4]*eslpi[0]+wv[5]*eslpi[1]+wv[3]*bc2)*eslpi[2];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Prism::faceCenter(
const NumPtr<Point<SPACEDIM> > &vertices,int f) const {
  switch (f) {
    case 0:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(2)]
             +vertices[vertexIndex(1)])/3.;
    case 1:
      return (vertices[vertexIndex(3)]
             +vertices[vertexIndex(4)]
             +vertices[vertexIndex(5)])/3.;
    case 2:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(3)]
             +vertices[vertexIndex(2)]
             +vertices[vertexIndex(5)])*0.25;
    case 3:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(1)]
             +vertices[vertexIndex(3)]
             +vertices[vertexIndex(4)])*0.25;
    case 4:
      return (vertices[vertexIndex(1)]
             +vertices[vertexIndex(2)]
             +vertices[vertexIndex(4)]
             +vertices[vertexIndex(5)])*0.25;
    default:
      OBSOLETE("bad face");
      return Point<SPACEDIM>(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Prism::diameter(const NumPtr<Point<SPACEDIM> > &vertices) const {
  REAL d[6]={
    (vertices[vertexIndex(3)]-vertices[vertexIndex(1)]).square(),
    (vertices[vertexIndex(3)]-vertices[vertexIndex(2)]).square(),
    (vertices[vertexIndex(4)]-vertices[vertexIndex(2)]).square(),
    (vertices[vertexIndex(4)]-vertices[vertexIndex(0)]).square(),
    (vertices[vertexIndex(5)]-vertices[vertexIndex(0)]).square(),
    (vertices[vertexIndex(5)]-vertices[vertexIndex(1)]).square()
  };
  REAL dmax=0.;
  for (int i=0;i<6;i++) dmax=max(d[i],dmax);
  return sqrt(dmax);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Prism::volume(const NumPtr<Point<SPACEDIM> > &vertices) const {
//may not be meaningful: vertices not necessarily co-planar in 3D
  Tensor<1,SPACEDIM> v10(vertices[vertexIndex(1)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v20(vertices[vertexIndex(2)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v30(vertices[vertexIndex(3)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v40(vertices[vertexIndex(4)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v50(vertices[vertexIndex(5)]
                        -vertices[vertexIndex(0)]);
  REAL vola,volb,volc;
  Tensor<1,SPACEDIM> cp;
  crossProduct(cp,v20,v40);
  product(vola,v10,cp); // v0,v1,v2,v4
  crossProduct(cp,v50,v30);
  product(volb,v40,cp); // v0,v4,v5,v3
  crossProduct(cp,v20,v50);
  product(volc,v40,cp); // v0,v4,v2,v3
  return (vola>0. && volb>0. && volc>0. ? (vola+volb+volc)/6.
    : -(abs(vola)+abs(volb)+abs(volc))/6.);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Prism::lineTangentPointsBackward(int l) const {
  switch (l) {
    case 0:
      return false;
    case 1:
      return vertexIndex(2)<vertexIndex(1);
    case 2:
      return true;
    case 3:
      return vertexIndex(4)<vertexIndex(3);
    case 4:
      return vertexIndex(5)<vertexIndex(4);
    case 5:
      return vertexIndex(3)<vertexIndex(5);
    case 6:
      return false;
    case 7:
      return vertexIndex(4)<vertexIndex(1);
    case 8:
      return vertexIndex(5)<vertexIndex(2);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Prism::printOn(ostream &os) const {
  os << "Prism:" << endl;
  Shape3::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkQuadrilateralVertices(int v0,int v1,int v2,int v3,
const Quadrilateral *quadrilateral) {
  int ql[4]={-1,-2,-3,-4};
  int qv[4]={v0,v1,v2,v3};
  Quadrilateral bogus_quadrilateral(-1,ql,qv);
  CHECK_TEST(
    bogus_quadrilateral.vertexIndex(0)==quadrilateral->vertexIndex(0) &&
    bogus_quadrilateral.vertexIndex(1)==quadrilateral->vertexIndex(1) &&
    bogus_quadrilateral.vertexIndex(2)==quadrilateral->vertexIndex(2) &&
    bogus_quadrilateral.vertexIndex(3)==quadrilateral->vertexIndex(3));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Prism::checkFaceVertices(int f,const Shape2 *face_element) const {
  CHECK_BOUNDS(f,0,5);
  if (f<2) {
    const Triangle *triangle=
      dynamic_cast<const Triangle*>(face_element);
    CHECK_POINTER(triangle);
    checkTriangleVertices(vertexIndex(face_vertex_to_cell_vertex[f][0]),
                          vertexIndex(face_vertex_to_cell_vertex[f][1]),
                          vertexIndex(face_vertex_to_cell_vertex[f][2]),
                          triangle);
  } else {
    const Quadrilateral *quadrilateral=
      dynamic_cast<const Quadrilateral*>(face_element);
    CHECK_POINTER(quadrilateral);
    checkQuadrilateralVertices(
      vertexIndex(face_vertex_to_cell_vertex[f][0]),
      vertexIndex(face_vertex_to_cell_vertex[f][1]),
      vertexIndex(face_vertex_to_cell_vertex[f][2]),
      vertexIndex(face_vertex_to_cell_vertex[f][3]),quadrilateral);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Prism::draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE &xc,
const NumPtr<Point<3> > &vertices,VolGraphTool &vgt) const {
  Point<3> p0=vertices[vertexIndex(0)];
  Point<3> p1=vertices[vertexIndex(1)];
  Point<3> p2=vertices[vertexIndex(2)];
  Point<3> p3=vertices[vertexIndex(3)];
  Point<3> p4=vertices[vertexIndex(4)];
  Point<3> p5=vertices[vertexIndex(5)];
  Vector3 v0(p0[0],p0[1],p0[2]);
  Vector3 v1(p1[0],p1[1],p1[2]);
  Vector3 v2(p2[0],p2[1],p2[2]);
  Vector3 v3(p3[0],p3[1],p3[2]);
  Vector3 v4(p4[0],p4[1],p4[2]);
  Vector3 v5(p5[0],p5[1],p5[2]);
  vgt.colorTriangle(xc,v2,v1,v0);
  vgt.colorTriangle(xc,v3,v4,v5);
  vgt.colorTriangle(xc,v3,v2,v0);
  vgt.colorTriangle(xc,v2,v3,v5);
  vgt.colorTriangle(xc,v4,v3,v0);
  vgt.colorTriangle(xc,v0,v1,v4);
  vgt.colorTriangle(xc,v5,v4,v2);
  vgt.colorTriangle(xc,v1,v2,v4);
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Hexahedron::Hexahedron(int edi,const int l[12],const int f[6],
const int v[8]) : Shape3(edi,12,6,8) {
  set<int> vertex_index_set;
  for (int i=0;i<8;i++) vertex_index_set.insert(v[i]);
#ifdef DEBUG
  CHECK_SAME(vertex_index_set.size(),8);
  set<int> face_index_set;
  for (int i=0;i<6;i++) face_index_set.insert(f[i]);
  CHECK_SAME(face_index_set.size(),6);
  set<int> line_index_set;
  for (int i=0;i<12;i++) line_index_set.insert(l[i]);
  CHECK_SAME(line_index_set.size(),12);
#endif
  set<int>::const_iterator vis_iterator=vertex_index_set.begin();
  int vmin=*vis_iterator;
  if (vmin==v[0]) {
    if (v[1]<min(v[2],v[4])) {
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[2]);
      setVertexIndex(3,v[3]);
      setVertexIndex(4,v[4]);
      setVertexIndex(5,v[5]);
      setVertexIndex(6,v[6]);
      setVertexIndex(7,v[7]);
      setCodimension1ShapeIndex(0,f[0]);
      setCodimension1ShapeIndex(1,f[1]);
      setCodimension1ShapeIndex(2,f[2]);
      setCodimension1ShapeIndex(3,f[3]);
      setCodimension1ShapeIndex(4,f[4]);
      setCodimension1ShapeIndex(5,f[5]);
      setCodimension2ShapeIndex(0,l[0]);
      setCodimension2ShapeIndex(1,l[1]);
      setCodimension2ShapeIndex(2,l[2]);
      setCodimension2ShapeIndex(3,l[3]);
      setCodimension2ShapeIndex(4,l[4]);
      setCodimension2ShapeIndex(5,l[5]);
      setCodimension2ShapeIndex(6,l[6]);
      setCodimension2ShapeIndex(7,l[7]);
      setCodimension2ShapeIndex(8,l[8]);
      setCodimension2ShapeIndex(9,l[9]);
      setCodimension2ShapeIndex(10,l[10]);
      setCodimension2ShapeIndex(11,l[11]);
    } else if (v[2]<min(v[1],v[4])) {
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[4]);
      setVertexIndex(3,v[6]);
      setVertexIndex(4,v[1]);
      setVertexIndex(5,v[3]);
      setVertexIndex(6,v[5]);
      setVertexIndex(7,v[7]);
      setCodimension1ShapeIndex(0,f[2]);
      setCodimension1ShapeIndex(1,f[3]);
      setCodimension1ShapeIndex(2,f[4]);
      setCodimension1ShapeIndex(3,f[5]);
      setCodimension1ShapeIndex(4,f[0]);
      setCodimension1ShapeIndex(5,f[1]);
      setCodimension2ShapeIndex(0,l[8]);
      setCodimension2ShapeIndex(1,l[10]);
      setCodimension2ShapeIndex(2,l[0]);
      setCodimension2ShapeIndex(3,l[4]);
      setCodimension2ShapeIndex(4,l[9]);
      setCodimension2ShapeIndex(5,l[11]);
      setCodimension2ShapeIndex(6,l[1]);
      setCodimension2ShapeIndex(7,l[5]);
      setCodimension2ShapeIndex(8,l[2]);
      setCodimension2ShapeIndex(9,l[3]);
      setCodimension2ShapeIndex(10,l[6]);
      setCodimension2ShapeIndex(11,l[7]);
    } else {
      setVertexIndex(0,v[0]);
      setVertexIndex(1,v[4]);
      setVertexIndex(2,v[1]);
      setVertexIndex(3,v[5]);
      setVertexIndex(4,v[2]);
      setVertexIndex(5,v[6]);
      setVertexIndex(6,v[3]);
      setVertexIndex(7,v[7]);
      setCodimension1ShapeIndex(0,f[4]);
      setCodimension1ShapeIndex(1,f[5]);
      setCodimension1ShapeIndex(2,f[0]);
      setCodimension1ShapeIndex(3,f[1]);
      setCodimension1ShapeIndex(4,f[2]);
      setCodimension1ShapeIndex(5,f[3]);
      setCodimension2ShapeIndex(0,l[2]);
      setCodimension2ShapeIndex(1,l[6]);
      setCodimension2ShapeIndex(2,l[8]);
      setCodimension2ShapeIndex(3,l[9]);
      setCodimension2ShapeIndex(4,l[3]);
      setCodimension2ShapeIndex(5,l[7]);
      setCodimension2ShapeIndex(6,l[10]);
      setCodimension2ShapeIndex(7,l[11]);
      setCodimension2ShapeIndex(8,l[0]);
      setCodimension2ShapeIndex(9,l[4]);
      setCodimension2ShapeIndex(10,l[1]);
      setCodimension2ShapeIndex(11,l[5]);
    }
  } else if (vmin==v[1]) {
    if (v[0]<min(v[3],v[5])) {
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[5]);
      setVertexIndex(3,v[4]);
      setVertexIndex(4,v[3]);
      setVertexIndex(5,v[2]);
      setVertexIndex(6,v[7]);
      setVertexIndex(7,v[6]);
      setCodimension1ShapeIndex(0,f[1]);
      setCodimension1ShapeIndex(1,f[0]);
      setCodimension1ShapeIndex(2,f[4]);
      setCodimension1ShapeIndex(3,f[5]);
      setCodimension1ShapeIndex(4,f[2]);
      setCodimension1ShapeIndex(5,f[3]);
      setCodimension2ShapeIndex(0,l[9]);
      setCodimension2ShapeIndex(1,l[8]);
      setCodimension2ShapeIndex(2,l[2]);
      setCodimension2ShapeIndex(3,l[6]);
      setCodimension2ShapeIndex(4,l[11]);
      setCodimension2ShapeIndex(5,l[10]);
      setCodimension2ShapeIndex(6,l[3]);
      setCodimension2ShapeIndex(7,l[7]);
      setCodimension2ShapeIndex(8,l[1]);
      setCodimension2ShapeIndex(9,l[0]);
      setCodimension2ShapeIndex(10,l[5]);
      setCodimension2ShapeIndex(11,l[4]);
    } else if (v[3]<min(v[0],v[5])) {
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[3]);
      setVertexIndex(2,v[0]);
      setVertexIndex(3,v[2]);
      setVertexIndex(4,v[5]);
      setVertexIndex(5,v[7]);
      setVertexIndex(6,v[4]);
      setVertexIndex(7,v[6]);
      setCodimension1ShapeIndex(0,f[2]);
      setCodimension1ShapeIndex(1,f[3]);
      setCodimension1ShapeIndex(2,f[1]);
      setCodimension1ShapeIndex(3,f[0]);
      setCodimension1ShapeIndex(4,f[4]);
      setCodimension1ShapeIndex(5,f[5]);
      setCodimension2ShapeIndex(0,l[2]);
      setCodimension2ShapeIndex(1,l[3]);
      setCodimension2ShapeIndex(2,l[1]);
      setCodimension2ShapeIndex(3,l[0]);
      setCodimension2ShapeIndex(4,l[6]);
      setCodimension2ShapeIndex(5,l[7]);
      setCodimension2ShapeIndex(6,l[5]);
      setCodimension2ShapeIndex(7,l[4]);
      setCodimension2ShapeIndex(8,l[9]);
      setCodimension2ShapeIndex(9,l[11]);
      setCodimension2ShapeIndex(10,l[8]);
      setCodimension2ShapeIndex(11,l[10]);
    } else {
      setVertexIndex(0,v[1]);
      setVertexIndex(1,v[5]);
      setVertexIndex(2,v[3]);
      setVertexIndex(3,v[7]);
      setVertexIndex(4,v[0]);
      setVertexIndex(5,v[4]);
      setVertexIndex(6,v[2]);
      setVertexIndex(7,v[6]);
      setCodimension1ShapeIndex(0,f[4]);
      setCodimension1ShapeIndex(1,f[5]);
      setCodimension1ShapeIndex(2,f[2]);
      setCodimension1ShapeIndex(3,f[3]);
      setCodimension1ShapeIndex(4,f[1]);
      setCodimension1ShapeIndex(5,f[0]);
      setCodimension2ShapeIndex(0,l[1]);
      setCodimension2ShapeIndex(1,l[5]);
      setCodimension2ShapeIndex(2,l[9]);
      setCodimension2ShapeIndex(3,l[11]);
      setCodimension2ShapeIndex(4,l[0]);
      setCodimension2ShapeIndex(5,l[4]);
      setCodimension2ShapeIndex(6,l[8]);
      setCodimension2ShapeIndex(7,l[10]);
      setCodimension2ShapeIndex(8,l[2]);
      setCodimension2ShapeIndex(9,l[6]);
      setCodimension2ShapeIndex(10,l[3]);
      setCodimension2ShapeIndex(11,l[7]);
    }
  } else if (vmin==v[2]) {
    if (v[0]<min(v[3],v[6])) {
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[3]);
      setVertexIndex(3,v[1]);
      setVertexIndex(4,v[6]);
      setVertexIndex(5,v[4]);
      setVertexIndex(6,v[7]);
      setVertexIndex(7,v[5]);
      setCodimension1ShapeIndex(0,f[3]);
      setCodimension1ShapeIndex(1,f[2]);
      setCodimension1ShapeIndex(2,f[0]);
      setCodimension1ShapeIndex(3,f[1]);
      setCodimension1ShapeIndex(4,f[4]);
      setCodimension1ShapeIndex(5,f[5]);
      setCodimension2ShapeIndex(0,l[3]);
      setCodimension2ShapeIndex(1,l[2]);
      setCodimension2ShapeIndex(2,l[0]);
      setCodimension2ShapeIndex(3,l[1]);
      setCodimension2ShapeIndex(4,l[7]);
      setCodimension2ShapeIndex(5,l[6]);
      setCodimension2ShapeIndex(6,l[4]);
      setCodimension2ShapeIndex(7,l[5]);
      setCodimension2ShapeIndex(8,l[10]);
      setCodimension2ShapeIndex(9,l[8]);
      setCodimension2ShapeIndex(10,l[11]);
      setCodimension2ShapeIndex(11,l[9]);
    } else if (v[3]<min(v[0],v[6])) {
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[3]);
      setVertexIndex(2,v[6]);
      setVertexIndex(3,v[7]);
      setVertexIndex(4,v[0]);
      setVertexIndex(5,v[1]);
      setVertexIndex(6,v[4]);
      setVertexIndex(7,v[5]);
      setCodimension1ShapeIndex(0,f[0]);
      setCodimension1ShapeIndex(1,f[1]);
      setCodimension1ShapeIndex(2,f[4]);
      setCodimension1ShapeIndex(3,f[5]);
      setCodimension1ShapeIndex(4,f[3]);
      setCodimension1ShapeIndex(5,f[2]);
      setCodimension2ShapeIndex(0,l[10]);
      setCodimension2ShapeIndex(1,l[11]);
      setCodimension2ShapeIndex(2,l[3]);
      setCodimension2ShapeIndex(3,l[7]);
      setCodimension2ShapeIndex(4,l[8]);
      setCodimension2ShapeIndex(5,l[9]);
      setCodimension2ShapeIndex(6,l[2]);
      setCodimension2ShapeIndex(7,l[6]);
      setCodimension2ShapeIndex(8,l[0]);
      setCodimension2ShapeIndex(9,l[1]);
      setCodimension2ShapeIndex(10,l[4]);
      setCodimension2ShapeIndex(11,l[5]);
    } else {
      setVertexIndex(0,v[2]);
      setVertexIndex(1,v[6]);
      setVertexIndex(2,v[0]);
      setVertexIndex(3,v[4]);
      setVertexIndex(4,v[3]);
      setVertexIndex(5,v[7]);
      setVertexIndex(6,v[1]);
      setVertexIndex(7,v[5]);
      setCodimension1ShapeIndex(0,f[4]);
      setCodimension1ShapeIndex(1,f[5]);
      setCodimension1ShapeIndex(2,f[3]);
      setCodimension1ShapeIndex(3,f[2]);
      setCodimension1ShapeIndex(4,f[0]);
      setCodimension1ShapeIndex(5,f[1]);
      setCodimension2ShapeIndex(0,l[0]);
      setCodimension2ShapeIndex(1,l[4]);
      setCodimension2ShapeIndex(2,l[10]);
      setCodimension2ShapeIndex(3,l[8]);
      setCodimension2ShapeIndex(4,l[1]);
      setCodimension2ShapeIndex(5,l[5]);
      setCodimension2ShapeIndex(6,l[11]);
      setCodimension2ShapeIndex(7,l[9]);
      setCodimension2ShapeIndex(8,l[3]);
      setCodimension2ShapeIndex(9,l[7]);
      setCodimension2ShapeIndex(10,l[2]);
      setCodimension2ShapeIndex(11,l[6]);
    }
  } else if (vmin==v[3]) {
    if (v[1]<min(v[2],v[7])) {
      setVertexIndex(0,v[3]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[7]);
      setVertexIndex(3,v[5]);
      setVertexIndex(4,v[2]);
      setVertexIndex(5,v[0]);
      setVertexIndex(6,v[6]);
      setVertexIndex(7,v[4]);
      setCodimension1ShapeIndex(0,f[3]);
      setCodimension1ShapeIndex(1,f[2]);
      setCodimension1ShapeIndex(2,f[4]);
      setCodimension1ShapeIndex(3,f[5]);
      setCodimension1ShapeIndex(4,f[1]);
      setCodimension1ShapeIndex(5,f[0]);
      setCodimension2ShapeIndex(0,l[11]);
      setCodimension2ShapeIndex(1,l[9]);
      setCodimension2ShapeIndex(2,l[1]);
      setCodimension2ShapeIndex(3,l[5]);
      setCodimension2ShapeIndex(4,l[10]);
      setCodimension2ShapeIndex(5,l[8]);
      setCodimension2ShapeIndex(6,l[0]);
      setCodimension2ShapeIndex(7,l[4]);
      setCodimension2ShapeIndex(8,l[3]);
      setCodimension2ShapeIndex(9,l[2]);
      setCodimension2ShapeIndex(10,l[7]);
      setCodimension2ShapeIndex(11,l[6]);
    } else if (v[2]<min(v[1],v[7])) {
      setVertexIndex(0,v[3]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[1]);
      setVertexIndex(3,v[0]);
      setVertexIndex(4,v[7]);
      setVertexIndex(5,v[6]);
      setVertexIndex(6,v[5]);
      setVertexIndex(7,v[4]);
      setCodimension1ShapeIndex(0,f[1]);
      setCodimension1ShapeIndex(1,f[0]);
      setCodimension1ShapeIndex(2,f[3]);
      setCodimension1ShapeIndex(3,f[2]);
      setCodimension1ShapeIndex(4,f[4]);
      setCodimension1ShapeIndex(5,f[5]);
      setCodimension2ShapeIndex(0,l[1]);
      setCodimension2ShapeIndex(1,l[0]);
      setCodimension2ShapeIndex(2,l[3]);
      setCodimension2ShapeIndex(3,l[2]);
      setCodimension2ShapeIndex(4,l[5]);
      setCodimension2ShapeIndex(5,l[4]);
      setCodimension2ShapeIndex(6,l[7]);
      setCodimension2ShapeIndex(7,l[6]);
      setCodimension2ShapeIndex(8,l[11]);
      setCodimension2ShapeIndex(9,l[10]);
      setCodimension2ShapeIndex(10,l[9]);
      setCodimension2ShapeIndex(11,l[8]);
    } else {
      setVertexIndex(0,v[3]);
      setVertexIndex(1,v[7]);
      setVertexIndex(2,v[2]);
      setVertexIndex(3,v[6]);
      setVertexIndex(4,v[1]);
      setVertexIndex(5,v[5]);
      setVertexIndex(6,v[0]);
      setVertexIndex(7,v[4]);
      setCodimension1ShapeIndex(0,f[4]);
      setCodimension1ShapeIndex(1,f[5]);
      setCodimension1ShapeIndex(2,f[1]);
      setCodimension1ShapeIndex(3,f[0]);
      setCodimension1ShapeIndex(4,f[3]);
      setCodimension1ShapeIndex(5,f[2]);
      setCodimension2ShapeIndex(0,l[3]);
      setCodimension2ShapeIndex(1,l[7]);
      setCodimension2ShapeIndex(2,l[11]);
      setCodimension2ShapeIndex(3,l[10]);
      setCodimension2ShapeIndex(4,l[2]);
      setCodimension2ShapeIndex(5,l[6]);
      setCodimension2ShapeIndex(6,l[9]);
      setCodimension2ShapeIndex(7,l[8]);
      setCodimension2ShapeIndex(8,l[1]);
      setCodimension2ShapeIndex(9,l[5]);
      setCodimension2ShapeIndex(10,l[0]);
      setCodimension2ShapeIndex(11,l[4]);
    }
  } else if (vmin==v[4]) {
    if (v[0]<min(v[5],v[6])) {
      setVertexIndex(0,v[4]);
      setVertexIndex(1,v[0]);
      setVertexIndex(2,v[6]);
      setVertexIndex(3,v[2]);
      setVertexIndex(4,v[5]);
      setVertexIndex(5,v[1]);
      setVertexIndex(6,v[7]);
      setVertexIndex(7,v[3]);
      setCodimension1ShapeIndex(0,f[5]);
      setCodimension1ShapeIndex(1,f[4]);
      setCodimension1ShapeIndex(2,f[2]);
      setCodimension1ShapeIndex(3,f[3]);
      setCodimension1ShapeIndex(4,f[0]);
      setCodimension1ShapeIndex(5,f[1]);
      setCodimension2ShapeIndex(0,l[4]);
      setCodimension2ShapeIndex(1,l[0]);
      setCodimension2ShapeIndex(2,l[8]);
      setCodimension2ShapeIndex(3,l[10]);
      setCodimension2ShapeIndex(4,l[5]);
      setCodimension2ShapeIndex(5,l[1]);
      setCodimension2ShapeIndex(6,l[9]);
      setCodimension2ShapeIndex(7,l[11]);
      setCodimension2ShapeIndex(8,l[6]);
      setCodimension2ShapeIndex(9,l[2]);
      setCodimension2ShapeIndex(10,l[7]);
      setCodimension2ShapeIndex(11,l[3]);
    } else if (v[5]<min(v[0],v[6])) {
      setVertexIndex(0,v[4]);
      setVertexIndex(1,v[5]);
      setVertexIndex(2,v[0]);
      setVertexIndex(3,v[1]);
      setVertexIndex(4,v[6]);
      setVertexIndex(5,v[7]);
      setVertexIndex(6,v[2]);
      setVertexIndex(7,v[3]);
      setCodimension1ShapeIndex(0,f[0]);
      setCodimension1ShapeIndex(1,f[1]);
      setCodimension1ShapeIndex(2,f[5]);
      setCodimension1ShapeIndex(3,f[4]);
      setCodimension1ShapeIndex(4,f[2]);
      setCodimension1ShapeIndex(5,f[3]);
      setCodimension2ShapeIndex(0,l[8]);
      setCodimension2ShapeIndex(1,l[9]);
      setCodimension2ShapeIndex(2,l[6]);
      setCodimension2ShapeIndex(3,l[2]);
      setCodimension2ShapeIndex(4,l[10]);
      setCodimension2ShapeIndex(5,l[11]);
      setCodimension2ShapeIndex(6,l[7]);
      setCodimension2ShapeIndex(7,l[3]);
      setCodimension2ShapeIndex(8,l[4]);
      setCodimension2ShapeIndex(9,l[5]);
      setCodimension2ShapeIndex(10,l[0]);
      setCodimension2ShapeIndex(11,l[1]);
    } else {
      setVertexIndex(0,v[4]);
      setVertexIndex(1,v[6]);
      setVertexIndex(2,v[5]);
      setVertexIndex(3,v[7]);
      setVertexIndex(4,v[0]);
      setVertexIndex(5,v[2]);
      setVertexIndex(6,v[1]);
      setVertexIndex(7,v[3]);
      setCodimension1ShapeIndex(0,f[2]);
      setCodimension1ShapeIndex(1,f[3]);
      setCodimension1ShapeIndex(2,f[0]);
      setCodimension1ShapeIndex(3,f[1]);
      setCodimension1ShapeIndex(4,f[5]);
      setCodimension1ShapeIndex(5,f[4]);
      setCodimension2ShapeIndex(0,l[6]);
      setCodimension2ShapeIndex(1,l[7]);
      setCodimension2ShapeIndex(2,l[4]);
      setCodimension2ShapeIndex(3,l[5]);
      setCodimension2ShapeIndex(4,l[2]);
      setCodimension2ShapeIndex(5,l[3]);
      setCodimension2ShapeIndex(6,l[0]);
      setCodimension2ShapeIndex(7,l[1]);
      setCodimension2ShapeIndex(8,l[8]);
      setCodimension2ShapeIndex(9,l[10]);
      setCodimension2ShapeIndex(10,l[9]);
      setCodimension2ShapeIndex(11,l[11]);
    }
  } else if (vmin==v[5]) {
    if (v[1]<min(v[4],v[7])) {
      setVertexIndex(0,v[5]);
      setVertexIndex(1,v[1]);
      setVertexIndex(2,v[4]);
      setVertexIndex(3,v[0]);
      setVertexIndex(4,v[7]);
      setVertexIndex(5,v[3]);
      setVertexIndex(6,v[6]);
      setVertexIndex(7,v[2]);
      setCodimension1ShapeIndex(0,f[5]);
      setCodimension1ShapeIndex(1,f[4]);
      setCodimension1ShapeIndex(2,f[1]);
      setCodimension1ShapeIndex(3,f[0]);
      setCodimension1ShapeIndex(4,f[2]);
      setCodimension1ShapeIndex(5,f[3]);
      setCodimension2ShapeIndex(0,l[6]);
      setCodimension2ShapeIndex(1,l[2]);
      setCodimension2ShapeIndex(2,l[9]);
      setCodimension2ShapeIndex(3,l[8]);
      setCodimension2ShapeIndex(4,l[7]);
      setCodimension2ShapeIndex(5,l[3]);
      setCodimension2ShapeIndex(6,l[11]);
      setCodimension2ShapeIndex(7,l[10]);
      setCodimension2ShapeIndex(8,l[5]);
      setCodimension2ShapeIndex(9,l[1]);
      setCodimension2ShapeIndex(10,l[4]);
      setCodimension2ShapeIndex(11,l[0]);
    } else if (v[4]<min(v[1],v[7])) {
      setVertexIndex(0,v[5]);
      setVertexIndex(1,v[4]);
      setVertexIndex(2,v[7]);
      setVertexIndex(3,v[6]);
      setVertexIndex(4,v[1]);
      setVertexIndex(5,v[0]);
      setVertexIndex(6,v[3]);
      setVertexIndex(7,v[2]);
      setCodimension1ShapeIndex(0,f[1]);
      setCodimension1ShapeIndex(1,f[0]);
      setCodimension1ShapeIndex(2,f[2]);
      setCodimension1ShapeIndex(3,f[3]);
      setCodimension1ShapeIndex(4,f[5]);
      setCodimension1ShapeIndex(5,f[4]);
      setCodimension2ShapeIndex(0,l[5]);
      setCodimension2ShapeIndex(1,l[4]);
      setCodimension2ShapeIndex(2,l[6]);
      setCodimension2ShapeIndex(3,l[7]);
      setCodimension2ShapeIndex(4,l[1]);
      setCodimension2ShapeIndex(5,l[0]);
      setCodimension2ShapeIndex(6,l[2]);
      setCodimension2ShapeIndex(7,l[3]);
      setCodimension2ShapeIndex(8,l[9]);
      setCodimension2ShapeIndex(9,l[8]);
      setCodimension2ShapeIndex(10,l[11]);
      setCodimension2ShapeIndex(11,l[10]);
    } else {
      setVertexIndex(0,v[5]);
      setVertexIndex(1,v[7]);
      setVertexIndex(2,v[1]);
      setVertexIndex(3,v[3]);
      setVertexIndex(4,v[4]);
      setVertexIndex(5,v[6]);
      setVertexIndex(6,v[0]);
      setVertexIndex(7,v[2]);
      setCodimension1ShapeIndex(0,f[2]);
      setCodimension1ShapeIndex(1,f[3]);
      setCodimension1ShapeIndex(2,f[5]);
      setCodimension1ShapeIndex(3,f[4]);
      setCodimension1ShapeIndex(4,f[1]);
      setCodimension1ShapeIndex(5,f[0]);
      setCodimension2ShapeIndex(0,l[9]);
      setCodimension2ShapeIndex(1,l[11]);
      setCodimension2ShapeIndex(2,l[5]);
      setCodimension2ShapeIndex(3,l[1]);
      setCodimension2ShapeIndex(4,l[8]);
      setCodimension2ShapeIndex(5,l[10]);
      setCodimension2ShapeIndex(6,l[4]);
      setCodimension2ShapeIndex(7,l[0]);
      setCodimension2ShapeIndex(8,l[6]);
      setCodimension2ShapeIndex(9,l[7]);
      setCodimension2ShapeIndex(10,l[2]);
      setCodimension2ShapeIndex(11,l[3]);
    }
  } else if (vmin==v[6]) {
    if (v[2]<min(v[4],v[7])) {
      setVertexIndex(0,v[6]);
      setVertexIndex(1,v[2]);
      setVertexIndex(2,v[7]);
      setVertexIndex(3,v[3]);
      setVertexIndex(4,v[4]);
      setVertexIndex(5,v[0]);
      setVertexIndex(6,v[5]);
      setVertexIndex(7,v[1]);
      setCodimension1ShapeIndex(0,f[5]);
      setCodimension1ShapeIndex(1,f[4]);
      setCodimension1ShapeIndex(2,f[0]);
      setCodimension1ShapeIndex(3,f[1]);
      setCodimension1ShapeIndex(4,f[3]);
      setCodimension1ShapeIndex(5,f[2]);
      setCodimension2ShapeIndex(0,l[7]);
      setCodimension2ShapeIndex(1,l[3]);
      setCodimension2ShapeIndex(2,l[10]);
      setCodimension2ShapeIndex(3,l[11]);
      setCodimension2ShapeIndex(4,l[6]);
      setCodimension2ShapeIndex(5,l[2]);
      setCodimension2ShapeIndex(6,l[8]);
      setCodimension2ShapeIndex(7,l[9]);
      setCodimension2ShapeIndex(8,l[4]);
      setCodimension2ShapeIndex(9,l[0]);
      setCodimension2ShapeIndex(10,l[5]);
      setCodimension2ShapeIndex(11,l[1]);
    } else if (v[4]<min(v[2],v[7])) {
      setVertexIndex(0,v[6]);
      setVertexIndex(1,v[4]);
      setVertexIndex(2,v[2]);
      setVertexIndex(3,v[0]);
      setVertexIndex(4,v[7]);
      setVertexIndex(5,v[5]);
      setVertexIndex(6,v[3]);
      setVertexIndex(7,v[1]);
      setCodimension1ShapeIndex(0,f[3]);
      setCodimension1ShapeIndex(1,f[2]);
      setCodimension1ShapeIndex(2,f[5]);
      setCodimension1ShapeIndex(3,f[4]);
      setCodimension1ShapeIndex(4,f[0]);
      setCodimension1ShapeIndex(5,f[1]);
      setCodimension2ShapeIndex(0,l[10]);
      setCodimension2ShapeIndex(1,l[8]);
      setCodimension2ShapeIndex(2,l[4]);
      setCodimension2ShapeIndex(3,l[0]);
      setCodimension2ShapeIndex(4,l[11]);
      setCodimension2ShapeIndex(5,l[9]);
      setCodimension2ShapeIndex(6,l[5]);
      setCodimension2ShapeIndex(7,l[1]);
      setCodimension2ShapeIndex(8,l[7]);
      setCodimension2ShapeIndex(9,l[6]);
      setCodimension2ShapeIndex(10,l[3]);
      setCodimension2ShapeIndex(11,l[2]);
    } else {
      setVertexIndex(0,v[6]);
      setVertexIndex(1,v[7]);
      setVertexIndex(2,v[4]);
      setVertexIndex(3,v[5]);
      setVertexIndex(4,v[2]);
      setVertexIndex(5,v[3]);
      setVertexIndex(6,v[0]);
      setVertexIndex(7,v[1]);
      setCodimension1ShapeIndex(0,f[0]);
      setCodimension1ShapeIndex(1,f[1]);
      setCodimension1ShapeIndex(2,f[3]);
      setCodimension1ShapeIndex(3,f[2]);
      setCodimension1ShapeIndex(4,f[5]);
      setCodimension1ShapeIndex(5,f[4]);
      setCodimension2ShapeIndex(0,l[4]);
      setCodimension2ShapeIndex(1,l[5]);
      setCodimension2ShapeIndex(2,l[7]);
      setCodimension2ShapeIndex(3,l[6]);
      setCodimension2ShapeIndex(4,l[0]);
      setCodimension2ShapeIndex(5,l[1]);
      setCodimension2ShapeIndex(6,l[3]);
      setCodimension2ShapeIndex(7,l[2]);
      setCodimension2ShapeIndex(8,l[10]);
      setCodimension2ShapeIndex(9,l[11]);
      setCodimension2ShapeIndex(10,l[8]);
      setCodimension2ShapeIndex(11,l[9]);
    }
  } else {
    if (v[3]<min(v[5],v[6])) {
      setVertexIndex(0,v[7]);
      setVertexIndex(1,v[3]);
      setVertexIndex(2,v[5]);
      setVertexIndex(3,v[1]);
      setVertexIndex(4,v[6]);
      setVertexIndex(5,v[2]);
      setVertexIndex(6,v[4]);
      setVertexIndex(7,v[0]);
      setCodimension1ShapeIndex(0,f[5]);
      setCodimension1ShapeIndex(1,f[4]);
      setCodimension1ShapeIndex(2,f[3]);
      setCodimension1ShapeIndex(3,f[2]);
      setCodimension1ShapeIndex(4,f[1]);
      setCodimension1ShapeIndex(5,f[0]);
      setCodimension2ShapeIndex(0,l[5]);
      setCodimension2ShapeIndex(1,l[1]);
      setCodimension2ShapeIndex(2,l[11]);
      setCodimension2ShapeIndex(3,l[9]);
      setCodimension2ShapeIndex(4,l[4]);
      setCodimension2ShapeIndex(5,l[0]);
      setCodimension2ShapeIndex(6,l[10]);
      setCodimension2ShapeIndex(7,l[8]);
      setCodimension2ShapeIndex(8,l[7]);
      setCodimension2ShapeIndex(9,l[3]);
      setCodimension2ShapeIndex(10,l[6]);
      setCodimension2ShapeIndex(11,l[2]);
    } else if (v[5]<min(v[3],v[6])) {
      setVertexIndex(0,v[7]);
      setVertexIndex(1,v[5]);
      setVertexIndex(2,v[6]);
      setVertexIndex(3,v[4]);
      setVertexIndex(4,v[3]);
      setVertexIndex(5,v[1]);
      setVertexIndex(6,v[2]);
      setVertexIndex(7,v[0]);
      setCodimension1ShapeIndex(0,f[3]);
      setCodimension1ShapeIndex(1,f[2]);
      setCodimension1ShapeIndex(2,f[1]);
      setCodimension1ShapeIndex(3,f[0]);
      setCodimension1ShapeIndex(4,f[5]);
      setCodimension1ShapeIndex(5,f[4]);
      setCodimension2ShapeIndex(0,l[7]);
      setCodimension2ShapeIndex(1,l[6]);
      setCodimension2ShapeIndex(2,l[5]);
      setCodimension2ShapeIndex(3,l[4]);
      setCodimension2ShapeIndex(4,l[3]);
      setCodimension2ShapeIndex(5,l[2]);
      setCodimension2ShapeIndex(6,l[1]);
      setCodimension2ShapeIndex(7,l[0]);
      setCodimension2ShapeIndex(8,l[11]);
      setCodimension2ShapeIndex(9,l[9]);
      setCodimension2ShapeIndex(10,l[10]);
      setCodimension2ShapeIndex(11,l[8]);
    } else {
      setVertexIndex(0,v[7]);
      setVertexIndex(1,v[6]);
      setVertexIndex(2,v[3]);
      setVertexIndex(3,v[2]);
      setVertexIndex(4,v[5]);
      setVertexIndex(5,v[4]);
      setVertexIndex(6,v[1]);
      setVertexIndex(7,v[0]);
      setCodimension1ShapeIndex(0,f[1]);
      setCodimension1ShapeIndex(1,f[0]);
      setCodimension1ShapeIndex(2,f[5]);
      setCodimension1ShapeIndex(3,f[4]);
      setCodimension1ShapeIndex(4,f[3]);
      setCodimension1ShapeIndex(5,f[2]);
      setCodimension2ShapeIndex(0,l[11]);
      setCodimension2ShapeIndex(1,l[10]);
      setCodimension2ShapeIndex(2,l[7]);
      setCodimension2ShapeIndex(3,l[3]);
      setCodimension2ShapeIndex(4,l[9]);
      setCodimension2ShapeIndex(5,l[8]);
      setCodimension2ShapeIndex(6,l[6]);
      setCodimension2ShapeIndex(7,l[2]);
      setCodimension2ShapeIndex(8,l[5]);
      setCodimension2ShapeIndex(9,l[4]);
      setCodimension2ShapeIndex(10,l[1]);
      setCodimension2ShapeIndex(11,l[0]);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Hexahedron::Hexahedron(const int l[12],const int f[6],const int v[6],
int di) : Shape3(-1,12,6,8) {
#ifdef DEBUG
  set<int> vertex_index_set;
  for (int i=0;i<8;i++) vertex_index_set.insert(v[i]);
  CHECK_SAME(vertex_index_set.size(),8);
  set<int> face_index_set;
  for (int i=0;i<6;i++) face_index_set.insert(f[i]);
  CHECK_SAME(face_index_set.size(),6);
  set<int> line_index_set;
  for (int i=0;i<12;i++) line_index_set.insert(l[i]);
  CHECK_SAME(line_index_set.size(),12);
#endif
  for (int i=0;i<8;i++) setVertexIndex(i,v[i]);
  for (int i=0;i<6;i++) setCodimension1ShapeIndex(i,f[i]);
  for (int i=0;i<12;i++) setCodimension2ShapeIndex(i,l[i]);
  dof_index=di;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape2* Hexahedron::newFaceElement(int f) const {
  int l[4]={
    codimension2ShapeIndex(staticElementLineFromFace(f,0)),
    codimension2ShapeIndex(staticElementLineFromFace(f,1)),
    codimension2ShapeIndex(staticElementLineFromFace(f,2)),
    codimension2ShapeIndex(staticElementLineFromFace(f,3))
  };
  int v[4]={
    vertexIndex(staticElementVertexFromFace(f,0)),
    vertexIndex(staticElementVertexFromFace(f,1)),
    vertexIndex(staticElementVertexFromFace(f,2)),
    vertexIndex(staticElementVertexFromFace(f,3))
  };
  return OPERATOR_NEW Quadrilateral(l,v,-1); // third arg is dof_index
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape1* Hexahedron::newLineElement(int l) const {
  int v[2]={
    vertexIndex(staticElementVertexFromLine(l,0)),
    vertexIndex(staticElementVertexFromLine(l,1))
  };
  return OPERATOR_NEW Interval(v,-1); // second arg is dof_index
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//point from lowest vertex index to an adjacent vertex
Tensor<1,3> Hexahedron::faceTangent(int f,int n) const {
  switch (f) {
    case 0: {
      if (vertexIndex(2)<vertexIndex(4)) {
        return (n==0 ? line_tangents[0] : line_tangents[8]);
      } else {
        return (n==1 ? line_tangents[0] : line_tangents[8]);
      }
    }
    case 1: {
      int vmin=min(min(vertexIndex(1),vertexIndex(3)),
                   min(vertexIndex(5),vertexIndex(7)));
      if (vertexIndex(1)==vmin) {
        if (vertexIndex(3)<vertexIndex(5)) {
          return (n==0 ? line_tangents[1] : line_tangents[9]);
        } else {
          return (n==1 ? line_tangents[1] : line_tangents[9]);
        }
      } else if (vertexIndex(3)==vmin) {
        if (vertexIndex(1)<vertexIndex(7)) {
          return (n==0 ? -line_tangents[1] : line_tangents[11]);
        } else {
          return (n==1 ? -line_tangents[1] : line_tangents[11]);
        }
      } else if (vertexIndex(5)==vmin) {
        if (vertexIndex(1)<vertexIndex(7)) {
          return (n==0 ? -line_tangents[9] : line_tangents[5]);
        } else {
          return (n==1 ? -line_tangents[9] : line_tangents[5]);
        }
      } else {
        if (vertexIndex(3)<vertexIndex(5)) {
          return (n==0 ? -line_tangents[11] : -line_tangents[5]);
        } else {
          return (n==1 ? -line_tangents[11] : -line_tangents[5]);
        }
      }
    }
    case 2: {
      if (vertexIndex(1)<vertexIndex(4)) {
        return (n==0 ? line_tangents[2] : line_tangents[8]);
      } else {
        return (n==1 ? line_tangents[2] : line_tangents[8]);
      }
      break;
    }
    case 3: {
      int vmin=min(min(vertexIndex(2),vertexIndex(3)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(2)==vmin) {
        if (vertexIndex(3)<vertexIndex(6)) {
          return (n==0 ? line_tangents[3] : line_tangents[10]);
        } else {
          return (n==1 ? line_tangents[3] : line_tangents[10]);
        }
      } else if (vertexIndex(3)==vmin) {
        if (vertexIndex(2)<vertexIndex(7)) {
          return (n==0 ? -line_tangents[3] : line_tangents[11]);
        } else {
          return (n==1 ? -line_tangents[3] : line_tangents[11]);
        }
      } else if (vertexIndex(6)==vmin) {
        if (vertexIndex(2)<vertexIndex(7)) {
          return (n==0 ? -line_tangents[10] : line_tangents[7]);
        } else {
          return (n==1 ? -line_tangents[10] : line_tangents[7]);
        }
      } else {
        if (vertexIndex(3)<vertexIndex(6)) {
          return (n==0 ? -line_tangents[11] : -line_tangents[7]);
        } else {
          return (n==1 ? -line_tangents[11] : -line_tangents[7]);
        }
      }
    }
    case 4: {
      if (vertexIndex(1)<vertexIndex(2)) {
        return (n==0 ? line_tangents[2] : line_tangents[0]);
      } else {
        return (n==1 ? line_tangents[2] : line_tangents[0]);
      }
      break;
    }
    case 5: {
      int vmin=min(min(vertexIndex(4),vertexIndex(5)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(4)==vmin) {
        if (vertexIndex(5)<vertexIndex(6)) {
          return (n==0 ? line_tangents[6] : line_tangents[4]);
        } else {
          return (n==1 ? line_tangents[6] : line_tangents[4]);
        }
      } else if (vertexIndex(5)==vmin) {
        if (vertexIndex(4)<vertexIndex(7)) {
          return (n==0 ? -line_tangents[6] : line_tangents[5]);
        } else {
          return (n==1 ? -line_tangents[6] : line_tangents[5]);
        }
      } else if (vertexIndex(6)==vmin) {
        if (vertexIndex(4)<vertexIndex(7)) {
          return (n==0 ? -line_tangents[4] : line_tangents[7]);
        } else {
          return (n==1 ? -line_tangents[4] : line_tangents[7]);
        }
      } else {
        if (vertexIndex(5)<vertexIndex(6)) {
          return (n==0 ? -line_tangents[5] : -line_tangents[7]);
        } else {
          return (n==1 ? -line_tangents[5] : -line_tangents[7]);
        }
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Hexahedron::faceTangentsWindBackward(int f) const {
  switch (f) {
    case 0:
      return (vertexIndex(2)>=vertexIndex(4));
    case 1: {
      int vmin=min(min(vertexIndex(1),vertexIndex(3)),
                   min(vertexIndex(5),vertexIndex(7)));
      if (vertexIndex(1)==vmin) return (vertexIndex(3)>=vertexIndex(5));
      else if (vertexIndex(3)==vmin) {
        return (vertexIndex(1)>=vertexIndex(7));
      } else if (vertexIndex(5)==vmin) {
        return (vertexIndex(1)<vertexIndex(7));
      }
      return (vertexIndex(3)<vertexIndex(5));
    }
    case 2: {
      return (vertexIndex(1)>=vertexIndex(4));
    }
    case 3: {
      int vmin=min(min(vertexIndex(2),vertexIndex(3)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(2)==vmin) {
        return (vertexIndex(3)>=vertexIndex(6));
      } else if (vertexIndex(3)==vmin) {
        return (vertexIndex(2)>=vertexIndex(7));
      } else if (vertexIndex(6)==vmin) {
        return (vertexIndex(2)<vertexIndex(7));
      }
      return (vertexIndex(3)<vertexIndex(6));
    }
    case 4:
      return (vertexIndex(1)>=vertexIndex(2));
    case 5: {
      int vmin=min(min(vertexIndex(4),vertexIndex(5)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(4)==vmin) {
        return (vertexIndex(5)>=vertexIndex(6));
      } else if (vertexIndex(5)==vmin) {
        return (vertexIndex(4)>=vertexIndex(7));
      } else if (vertexIndex(6)==vmin) {
        return (vertexIndex(4)<vertexIndex(7));
      }
      return (vertexIndex(5)<vertexIndex(6));
      break;
    }
  }
//
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Hexahedron::contains(const Point<3> &x,
const NumPtr<Point<3> > &vertices) const {
  int l[9]={-1,-2,-3,-4,-5,-6,-7,-8,-9};
  int f[5]={-1,-2,-3,-4,-5};
  int v[6]=
    {vertexIndex(0),vertexIndex(1),vertexIndex(2),
     vertexIndex(4),vertexIndex(5),vertexIndex(6)};
  Prism w012456(l,f,v,-1);
  if (w012456.contains(x,vertices)) return true;

  {
    int v[6]=
      {vertexIndex(3),vertexIndex(2),vertexIndex(1),
       vertexIndex(7),vertexIndex(6),vertexIndex(5)};
    Prism w321765(l,f,v,-1);
    return w321765.contains(x,vertices);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// 24 rotations of a cube times 8 children = 192 cases
Point<SPACEDIM> Hexahedron::parentCoordinateForCanonical(
const Point<SPACEDIM> &cp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi0=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi0>=0);

  switch (c) {
    case 0: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[2]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return cp*0.5;
            }
            case 2: {
              return Point<3>(cp[1]*0.5,cp[2]*0.5,cp[0]*0.5);
            }
            case 4: {
              return Point<3>(cp[2]*0.5,cp[0]*0.5,cp[1]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 1: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[1]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>(1.-cp[1]*0.5,cp[0]*0.5,cp[2]*0.5);
            }
            case 2: {
              return Point<3>(1.-cp[2]*0.5,cp[1]*0.5,cp[0]*0.5);
            }
            case 4: {
              return Point<3>(1.-cp[0]*0.5,cp[2]*0.5,cp[1]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 2: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[0]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>(cp[1]*0.5,1.-cp[0]*0.5,cp[2]*0.5);
            }
            case 2: {
              return Point<3>(cp[2]*0.5,1.-cp[1]*0.5,cp[0]*0.5);
            }
            case 4: {
              return Point<3>(cp[0]*0.5,1.-cp[2]*0.5,cp[1]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[2]*0.5,(1.+cp[0])*0.5,cp[1]*0.5);
            }
            case 3: {
              return Point<3>((1.-cp[0])*0.5,(1.-cp[1])*0.5,cp[2]*0.5);
            }
            case 5: {
              return Point<3>(cp[1]*0.5,1.-cp[2]*0.5,(1.-cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[0]*0.5,(1.+cp[1])*0.5,cp[2]*0.5);
            }
            case 3: {
              return Point<3>(cp[2]*0.5,1.-cp[0]*0.5,(1.-cp[1])*0.5);
            }
            case 6: {
              return Point<3>((1.-cp[1])*0.5,(1.-cp[2])*0.5,cp[0]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 3: {
          switch (cvi1) {
            case 1: {
              return Point<3>(cp[2]*0.5,(1.+cp[1])*0.5,(1.-cp[0])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[1])*0.5,(1.+cp[0])*0.5,cp[2]*0.5);
            }
            case 7: {
              return Point<3>((1.-cp[0])*0.5,1.-cp[2]*0.5,(1.-cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 4: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[1]*0.5,(1.+cp[2])*0.5,cp[0]*0.5);
            }
            case 5: {
              return Point<3>((1.-cp[2])*0.5,(1.-cp[0])*0.5,cp[1]*0.5);
            }
            case 6: {
              return Point<3>(cp[0]*0.5,1.-cp[1]*0.5,(1.-cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 5: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.-cp[0])*0.5,(1.+cp[2])*0.5,cp[1]*0.5);
            }
            case 4: {
              return Point<3>(cp[1]*0.5,(1.+cp[0])*0.5,(1.-cp[2])*0.5);
            }
            case 7: {
              return Point<3>((1.-cp[2])*0.5,1.-cp[1]*0.5,(1.-cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 6: {
          switch (cvi1) {
            case 2: {
              return Point<3>(cp[0]*0.5,(1.+cp[2])*0.5,(1.-cp[1])*0.5);
            }
            case 4: {
              return Point<3>((1.-cp[2])*0.5,(1.+cp[1])*0.5,cp[0]*0.5);
            }
            case 7: {
              return Point<3>((1.-cp[1])*0.5,1.-cp[0]*0.5,(1.-cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 7: {
          switch (cvi1) {
            case 3: {
              return Point<3>((1.-cp[1])*0.5,(1.+cp[2])*0.5,
                              (1.-cp[0])*0.5);
            }
            case 5: {
              return Point<3>((1.-cp[1])*0.5,(1.+cp[0])*0.5,
                              (1.-cp[2])*0.5);
            }
            case 6: {
              return Point<3>((1.-cp[2])*0.5,(1.+cp[0])*0.5,
                              (1.-cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
      }
    }
    case 3: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[3]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>(1.-cp[0]*0.5,1.-cp[1]*0.5,cp[2]*0.5);
            }
            case 2: {
              return Point<3>(1.-cp[1]*0.5,1.-cp[2]*0.5,cp[0]*0.5);
            }
            case 4: {
              return Point<3>(1.-cp[2]*0.5,1.-cp[0]*0.5,cp[1]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[0])*0.5,1.-cp[2]*0.5,cp[1]*0.5);
            }
            case 3: {
              return Point<3>(1.-cp[1]*0.5,(1.+cp[0])*0.5,cp[2]*0.5);
            }
            case 5: {
              return Point<3>(1.-cp[2]*0.5,1.-cp[1]*0.5,(1.-cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[1])*0.5,1.-cp[0]*0.5,cp[2]*0.5);
            }
            case 3: {
              return Point<3>(1.-cp[0]*0.5,1.-cp[2]*0.5,(1.-cp[1])*0.5);
            }
            case 6: {
              return Point<3>(1.-cp[2]*0.5,(1.+cp[1])*0.5,cp[0]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 3: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.+cp[1])*0.5,1.-cp[2]*0.5,(1.-cp[0])*0.5);
            }
            case 2: {
              return Point<3>((1.+cp[0])*0.5,(1.+cp[1])*0.5,cp[2]*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[2]*0.5,(1.+cp[0])*0.5,(1.-cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 4: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[2])*0.5,1.-cp[1]*0.5,cp[0]*0.5);
            }
            case 5: {
              return Point<3>(1.-cp[0]*0.5,(1.+cp[2])*0.5,cp[1]*0.5);
            }
            case 6: {
              return Point<3>(1.-cp[1]*0.5,1.-cp[0]*0.5,(1.-cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 5: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.+cp[2])*0.5,(1.+cp[0])*0.5,cp[1]*0.5);
            }
            case 4: {
              return Point<3>((1.+cp[0])*0.5,1.-cp[1]*0.5,(1.-cp[2])*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[1]*0.5,(1.+cp[2])*0.5,(1.-cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 6: {
          switch (cvi1) {
            case 2: {
              return Point<3>((1.+cp[2])*0.5,1.-cp[0]*0.5,(1.-cp[1])*0.5);
            }
            case 4: {
              return Point<3>((1.+cp[1])*0.5,(1.+cp[2])*0.5,cp[0]*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[0]*0.5,(1.+cp[1])*0.5,(1.-cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 7: {
          switch (cvi1) {
            case 3: {
              return Point<3>((1.+cp[1])*0.5,(1.+cp[2])*0.5,
                              (1.-cp[0])*0.5);
            }
            case 5: {
              return Point<3>((1.+cp[1])*0.5,(1.+cp[0])*0.5,
                              (1.-cp[2])*0.5);
            }
            case 6: {
              return Point<3>((1.+cp[0])*0.5,(1.+cp[2])*0.5,
                              (1.-cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
      }
    }
    case 4: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[4]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>(cp[1]*0.5,cp[0]*0.5,1.-cp[2]*0.5);
            }
            case 2: {
              return Point<3>(cp[2]*0.5,cp[1]*0.5,1.-cp[0]*0.5);
            }
            case 4: {
              return Point<3>(cp[0]*0.5,cp[2]*0.5,1.-cp[1]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[2]*0.5,(1.-cp[0])*0.5,1.-cp[1]*0.5);
            }
            case 3: {
              return Point<3>((1.-cp[0])*0.5,cp[1]*0.5,1.-cp[2]*0.5);
            }
            case 5: {
              return Point<3>(cp[1]*0.5,cp[2]*0.5,(1.+cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[0]*0.5,(1.-cp[1])*0.5,1.-cp[2]*0.5);
            }
            case 3: {
              return Point<3>(cp[2]*0.5,cp[0]*0.5,(1.+cp[1])*0.5);
            }
            case 6: {
              return Point<3>((1.-cp[1])*0.5,cp[2]*0.5,1.-cp[0]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 3: {
          switch (cvi1) {
            case 1: {
              return Point<3>(cp[2]*0.5,(1.-cp[1])*0.5,(1.+cp[0])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[1])*0.5,(1.-cp[0])*0.5,1.-cp[2]*0.5);
            }
            case 7: {
              return Point<3>((1.-cp[0])*0.5,cp[2]*0.5,(1.+cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 4: {
          switch (cvi1) {
            case 0: {
              return Point<3>(cp[1]*0.5,(1.-cp[2])*0.5,1.-cp[0]*0.5);
            }
            case 5: {
              return Point<3>((1.-cp[2])*0.5,cp[0]*0.5,1.-cp[1]*0.5);
            }
            case 6: {
              return Point<3>(cp[0]*0.5,cp[1]*0.5,(1.+cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 5: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.-cp[0])*0.5,(1.-cp[2])*0.5,1.-cp[1]*0.5);
            }
            case 4: {
              return Point<3>(cp[1]*0.5,(1.-cp[0])*0.5,(1.+cp[2])*0.5);
            }
            case 7: {
              return Point<3>((1.-cp[2])*0.5,cp[1]*0.5,(1.+cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 6: {
          switch (cvi1) {
            case 2: {
              return Point<3>(cp[0]*0.5,(1.-cp[2])*0.5,(1.+cp[1])*0.5);
            }
            case 4: {
              return Point<3>((1.-cp[2])*0.5,(1.-cp[1])*0.5,1.-cp[0]*0.5);
            }
            case 7: {
              return Point<3>((1.-cp[1])*0.5,cp[0]*0.5,(1.+cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 7: {
          switch (cvi1) {
            case 3: {
              return Point<3>((1.-cp[1])*0.5,(1.-cp[2])*0.5,
                              (1.+cp[0])*0.5);
            }
            case 5: {
              return Point<3>((1.-cp[0])*0.5,(1.-cp[1])*0.5,
                              (1.+cp[2])*0.5);
            }
            case 6: {
              return Point<3>((1.-cp[2])*0.5,(1.-cp[0])*0.5,
                              (1.+cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
      }
    }
    case 5: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[6]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>(1.-cp[0]*0.5,cp[1]*0.5,1.-cp[2]*0.5);
            }
            case 2: {
              return Point<3>(1.-cp[1]*0.5,cp[2]*0.5,1.-cp[0]*0.5);
            }
            case 4: {
              return Point<3>(1.-cp[2]*0.5,cp[0]*0.5,1.-cp[1]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[0])*0.5,cp[2]*0.5,1.-cp[1]*0.5);
            }
            case 3: {
              return Point<3>(1.-cp[1]*0.5,(1.-cp[0])*0.5,1.-cp[2]*0.5);
            }
            case 5: {
              return Point<3>(1.-cp[2]*0.5,cp[1]*0.5,(1.+cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[1])*0.5,cp[0]*0.5,1.-cp[2]*0.5);
            }
            case 3: {
              return Point<3>(1.-cp[0]*0.5,cp[2]*0.5,(1.+cp[1])*0.5);
            }
            case 6: {
              return Point<3>(1.-cp[2]*0.5,(1.-cp[1])*0.5,1.-cp[0]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 3: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.+cp[1])*0.5,cp[2]*0.5,(1.+cp[0])*0.5);
            }
            case 2: {
              return Point<3>((1.+cp[0])*0.5,(1.-cp[1])*0.5,1.-cp[2]*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[2]*0.5,(1.-cp[0])*0.5,(1.+cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 4: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[2])*0.5,cp[1]*0.5,1.-cp[0]*0.5);
            }
            case 5: {
              return Point<3>(1.-cp[0]*0.5,(1.-cp[2])*0.5,1.-cp[1]*0.5);
            }
            case 6: {
              return Point<3>(1.-cp[1]*0.5,cp[0]*0.5,(1.+cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 5: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.+cp[2])*0.5,(1.-cp[0])*0.5,1.-cp[1]*0.5);
            }
            case 4: {
              return Point<3>((1.+cp[0])*0.5,cp[1]*0.5,(1.+cp[2])*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[1]*0.5,(1.-cp[2])*0.5,(1.+cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 6: {
          switch (cvi1) {
            case 2: {
              return Point<3>((1.+cp[2])*0.5,cp[0]*0.5,(1.+cp[1])*0.5);
            }
            case 4: {
              return Point<3>((1.+cp[1])*0.5,(1.-cp[2])*0.5,1.-cp[0]*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[0]*0.5,(1.-cp[1])*0.5,(1.+cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 7: {
          switch (cvi1) {
            case 3: {
              return Point<3>((1.+cp[2])*0.5,(1.-cp[1])*0.5,
                              (1.+cp[0])*0.5);
            }
            case 5: {
              return Point<3>((1.+cp[1])*0.5,(1.-cp[0])*0.5,
                              (1.+cp[2])*0.5);
            }
            case 6: {
              return Point<3>((1.+cp[0])*0.5,(1.-cp[2])*0.5,
                              (1.+cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
      }
    }
    case 6: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[7]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>(cp[0]*0.5,1.-cp[1]*0.5,1.-cp[2]*0.5);
            }
            case 2: {
              return Point<3>(cp[1]*0.5,1.-cp[2]*0.5,1.-cp[0]*0.5);
            }
            case 4: {
              return Point<3>(cp[2]*0.5,1.-cp[0]*0.5,1.-cp[1]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[0])*0.5,1.-cp[2]*0.5,1.-cp[1]*0.5);
            }
            case 3: {
              return Point<3>((1.-cp[0])*0.5,1.-cp[1]*0.5,1.-cp[2]*0.5);
            }
            case 5: {
              return Point<3>(cp[2]*0.5,1.-cp[1]*0.5,(1.+cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[1])*0.5,1.-cp[0]*0.5,1.-cp[2]*0.5);
            }
            case 3: {
              return Point<3>(cp[0]*0.5,1.-cp[2]*0.5,(1.+cp[1])*0.5);
            }
            case 6: {
              return Point<3>(cp[2]*0.5,(1.+cp[1])*0.5,1.-cp[0]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 3: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.-cp[1])*0.5,1.-cp[2]*0.5,(1.+cp[0])*0.5);
            }
            case 2: {
              return Point<3>((1.-cp[0])*0.5,(1.+cp[1])*0.5,1.-cp[2]*0.5);
            }
            case 7: {
              return Point<3>(cp[2]*0.5,(1.+cp[0])*0.5,(1.+cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 4: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.-cp[2])*0.5,1.-cp[1]*0.5,1.-cp[0]*0.5);
            }
            case 5: {
              return Point<3>(cp[0]*0.5,(1.+cp[2])*0.5,1.-cp[1]*0.5);
            }
            case 6: {
              return Point<3>(cp[1]*0.5,1.-cp[0]*0.5,(1.+cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 5: {
          switch (cvi1) {
            case 1: {
              return Point<3>(cp[2]*0.5,(1.+cp[0])*0.5,1.-cp[1]*0.5);
            }
            case 4: {
              return Point<3>((1.-cp[0])*0.5,1.-cp[1]*0.5,(1.+cp[2])*0.5);
            }
            case 7: {
              return Point<3>(cp[1]*0.5,(1.+cp[2])*0.5,(1.+cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 6: {
          switch (cvi1) {
            case 2: {
              return Point<3>((1.-cp[2])*0.5,1.-cp[0]*0.5,(1.+cp[1])*0.5);
            }
            case 4: {
              return Point<3>((1.-cp[1])*0.5,(1.+cp[2])*0.5,1.-cp[0]*0.5);
            }
            case 7: {
              return Point<3>(cp[0]*0.5,(1.+cp[1])*0.5,(1.+cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 7: {
          switch (cvi1) {
            case 3: {
              return Point<3>((1.+cp[2])*0.5,(1.+cp[1])*0.5,
                              (1.+cp[0])*0.5);
            }
            case 5: {
              return Point<3>((1.-cp[1])*0.5,(1.+cp[0])*0.5,
                              (1.+cp[2])*0.5);
            }
            case 6: {
              return Point<3>((1.-cp[0])*0.5,(1.+cp[2])*0.5,
                              (1.+cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
      }
    }
    case 7: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[7]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1: {
              return Point<3>(1.-cp[0]*0.5,1.-cp[2]*0.5,1.-cp[1]*0.5);
            }
            case 2: {
              return Point<3>(1.-cp[1]*0.5,1.-cp[0]*0.5,1.-cp[2]*0.5);
            }
            case 4: {
              return Point<3>(1.-cp[2]*0.5,1.-cp[1]*0.5,1.-cp[0]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 1: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[0])*0.5,1.-cp[1]*0.5,1.-cp[2]*0.5);
            }
            case 3: {
              return Point<3>(1.-cp[1]*0.5,1.-cp[2]*0.5,(1.+cp[0])*0.5);
            }
            case 5: {
              return Point<3>(1.-cp[2]*0.5,(1.+cp[0])*0.5,1.-cp[1]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 2: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[1])*0.5,1.-cp[2]*0.5,1.-cp[0]*0.5);
            }
            case 3: {
              return Point<3>(1.-cp[0]*0.5,(1.+cp[1])*0.5,1.-cp[2]*0.5);
            }
            case 6: {
              return Point<3>(1.-cp[2]*0.5,1.-cp[0]*0.5,(1.+cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 3: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.+cp[1])*0.5,(1.+cp[0])*0.5,1.-cp[2]*0.5);
            }
            case 2: {
              return Point<3>((1.+cp[0])*0.5,1.-cp[2]*0.5,(1.+cp[1])*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[2]*0.5,(1.+cp[1])*0.5,(1.+cp[0])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 4: {
          switch (cvi1) {
            case 0: {
              return Point<3>((1.+cp[2])*0.5,1.-cp[0]*0.5,1.-cp[1]*0.5);
            }
            case 5: {
              return Point<3>(1.-cp[0]*0.5,1.-cp[1]*0.5,(1.+cp[2])*0.5);
            }
            case 6: {
              return Point<3>(1.-cp[1]*0.5,(1.+cp[2])*0.5,1.-cp[0]*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 5: {
          switch (cvi1) {
            case 1: {
              return Point<3>((1.+cp[2])*0.5,1.-cp[1]*0.5,(1.+cp[0])*0.5);
            }
            case 4: {
              return Point<3>((1.+cp[0])*0.5,(1.+cp[2])*0.5,1.-cp[1]*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[1]*0.5,(1.+cp[0])*0.5,(1.+cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 6: {
          switch (cvi1) {
            case 2: {
              return Point<3>((1.+cp[2])*0.5,(1.+cp[1])*0.5,1.-cp[0]*0.5);
            }
            case 4: {
              return Point<3>((1.+cp[1])*0.5,1.-cp[0]*0.5,(1.+cp[2])*0.5);
            }
            case 7: {
              return Point<3>(1.-cp[0]*0.5,(1.+cp[2])*0.5,(1.+cp[1])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
        case 7: {
          switch (cvi1) {
            case 3: {
              return Point<3>((1.+cp[2])*0.5,(1.+cp[0])*0.5,
                              (1.+cp[1])*0.5);
            }
            case 5: {
              return Point<3>((1.+cp[1])*0.5,(1.+cp[2])*0.5,
                              (1.+cp[0])*0.5);
            }
            case 6: {
              return Point<3>((1.+cp[0])*0.5,(1.+cp[1])*0.5,
                              (1.+cp[2])*0.5);
            }
            default:
              OBSOLETE("should not be here");
          }
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// 24 rotations of a cube times 8 children = 192 cases
Point<SPACEDIM> Hexahedron::childCoordinateOfCanonical(
const Point<SPACEDIM> &pp,const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi0=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi0>=0);

  switch (c) {
    case 0: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[2]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return pp*2.;
            case 2:
              return Point<3>(pp[2]*2.,pp[0]*2.,pp[1]*2.);
            case 4:
              return Point<3>(pp[1]*2.,pp[2]*2.,pp[0]*2.);
          }
        }
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 1: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[1]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[1]*2.,(1.-pp[0])*2.,pp[2]*2.);
            case 2:
              return Point<3>(pp[2]*2.,pp[1]*2.,(1.-pp[0])*2.);
            case 4:
              return Point<3>((1.-pp[0])*2.,pp[2]*2.,pp[1]*2.);
          }
        }
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
          OBSOLETE("not programmed");
          Point<SPACEDIM> dummy;
          return dummy;
      }
    }
    case 2: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[0]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>((1.-pp[1])*2.,pp[0]*2.,pp[2]*2.);
            case 2:
              return Point<3>(pp[2]*2.,(1.-pp[1])*2.,pp[0]*2.);
            case 4:
              return Point<3>(pp[0]*2.,pp[2]*2.,(1.-pp[1])*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>((pp[1])*2.-1.,pp[2]*2.,pp[0]*2.);
            case 3:
              return Point<3>(1.-pp[0]*2.,1.-pp[1]*2.,pp[2]*2.);
            case 5:
              return Point<3>(1.-pp[2]*2.,pp[0]*2.,(1.-pp[1])*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[0]*2.,pp[1]*2.-1.,pp[2]*2.);
            case 3:
              return Point<3>((1.-pp[1])*2.,1.-pp[2]*2.,pp[0]*2.);
            case 6:
              return Point<3>(pp[2]*2.,1.-pp[0]*2.,1.-pp[1]*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 1:
              return Point<3>(1.-pp[2]*2.,pp[1]*2.-1.,pp[0]*2.);
            case 2:
              return Point<3>(pp[1]*2.-1.,1.-pp[0]*2.,pp[2]*2.);
            case 7:
              return Point<3>(1.-pp[0]*2.,1.-pp[2]*2.,(1.-pp[1])*2.);
          }
        }
        case 4: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[2]*2.,pp[0]*2.,pp[1]*2.-1.);
            case 5:
              return Point<3>(1.-pp[1]*2.,pp[2]*2.,1.-pp[0]*2.);
            case 6:
              return Point<3>(pp[0]*2.,(1.-pp[1])*2.,1.-pp[2]*2.);
          }
        }
        case 5: {
          switch (cvi1) {
            case 1:
              return Point<3>(1.-pp[0]*2.,pp[2]*2.,pp[1]*2.-1.);
            case 4:
              return Point<3>(pp[1]*2.-1.,pp[0]*2.,1.-pp[2]*2.);
            case 7:
              return Point<3>(1.-pp[2]*2.,(1.-pp[1])*2.,1.-pp[0]*2.);
          }
        }
        case 6: {
          switch (cvi1) {
            case 2:
              return Point<3>(pp[0]*2.,1.-pp[2]*2.,pp[1]*2.-1.);
            case 4:
              return Point<3>(pp[2]*2.,pp[1]*2.-1.,1.-pp[0]*2.);
            case 7:
              return Point<3>((1.-pp[1])*2.,1.-pp[0]*2.,1.-pp[2]*2.);
          }
        }
        case 7: {
          switch (cvi1) {
            case 3:
              return Point<3>(1.-pp[2]*2.,1.-pp[0]*2.,pp[1]*2.-1.);
            case 5:
              return Point<3>(pp[1]*2.-1.,1.-pp[0]*2.,1.-pp[2]*2.);
            case 6:
              return Point<3>(pp[1]*2.-1.,1.-pp[2]*2.,1.-pp[0]*2.);
          }
        }
      }
    }
    case 3: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[3]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>((1.-pp[0])*2.,(1.-pp[1])*2.,pp[2]*2.);
            case 2:
              return Point<3>(pp[2]*2.,(1.-pp[0])*2.,(1.-pp[1])*2.);
            case 4:
              return Point<3>((1.-pp[1])*2.,pp[2]*2.,(1.-pp[0])*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[0]*2.-1.,pp[2]*2.,(1.-pp[1])*2.);
            case 3:
              return Point<3>(pp[1]*2.-1.,(1.-pp[0])*2.,pp[2]*2.);
            case 5:
              return Point<3>(1.-pp[2]*2.,(1.-pp[1])*2.,(1.-pp[0])*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>((1.-pp[1])*2.,pp[0]*2.-1.,pp[2]*2.);
            case 3:
              return Point<3>((1.-pp[0])*2.,1.-pp[2]*2.,(1.-pp[1])*2.);
            case 6:
              return Point<3>(pp[2]*2.,pp[1]*2.-1.,(1.-pp[0])*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 1:
              return Point<3>(1.-pp[2]*2.,pp[0]*2.-1.,(1.-pp[1])*2.);
            case 2:
              return Point<3>(pp[0]*2.-1.,pp[1]*2.-1.,pp[2]*2.);
            case 7:
              return Point<3>(pp[1]*2.-1.,1.-pp[2]*2.,(1.-pp[0])*2.);
          }
        }
        case 4: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[2]*2.,(1.-pp[1])*2.,pp[0]*2.-1.);
            case 5:
              return Point<3>((1.-pp[0])*2.,pp[2]*2.,pp[1]*2.-1.);
            case 6:
              return Point<3>((1.-pp[1])*2.,(1.-pp[0])*2.,1.-pp[2]*2.);
          }
        }
        case 5: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[1]*2.-1.,pp[2]*2.,pp[0]*2.-1.);
            case 4:
              return Point<3>(pp[0]*2.-1.,(1.-pp[1])*2.,1.-pp[2]*2.);
            case 7:
              return Point<3>(1.-pp[2]*2.,(1.-pp[0])*2.,pp[1]*2.-1.);
          }
        }
        case 6: {
          switch (cvi1) {
            case 2:
              return Point<3>((1.-pp[1])*2.,1.-pp[2]*2.,pp[0]*2.-1.);
            case 4:
              return Point<3>(pp[2]*2.,pp[0]*2.-1.,pp[1]*2.-1.);
            case 7:
              return Point<3>((1.-pp[0])*2.,pp[1]*2.-1.,1.-pp[2]*2.);
          }
        }
        case 7: {
          switch (cvi1) {
            case 3:
              return Point<3>(1.-pp[2]*2.,pp[0]*2.-1.,pp[1]*2.-1.);
            case 5:
              return Point<3>(pp[1]*2.-1.,pp[0]*2.-1.,1.-pp[2]*2.);
            case 6:
              return Point<3>(pp[0]*2.-1.,1.-pp[2]*2.,pp[1]*2.-1.);
          }
        }
      }
    }
    case 4: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[4]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[1]*2.,pp[0]*2.,(1.-pp[2])*2.);
            case 2:
              return Point<3>((1.-pp[2])*2.,pp[1]*2.,pp[0]*2.);
            case 4:
              return Point<3>(pp[0]*2.,(1.-pp[2])*2.,pp[1]*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(1.-pp[1]*2.,(1.-pp[2])*2.,pp[0]*2.);
            case 3:
              return Point<3>(1.-pp[0]*2.,pp[1]*2.,(1.-pp[2])*2.);
            case 5:
              return Point<3>(pp[2]*2.-1.,pp[0]*2.,pp[1]*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[0]*2.,1.-pp[1]*2.,(1.-pp[2])*2.);
            case 3:
              return Point<3>(pp[1]*2.,pp[2]*2.-1.,pp[0]*2.);
            case 6:
              return Point<3>((1.-pp[2])*2.,1.-pp[0]*2.,pp[1]*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[2]*2.-1.,1.-pp[1]*2.,pp[0]*2.);
            case 2:
              return Point<3>(1.-pp[1]*2.,1.-pp[0]*2.,(1.-pp[2])*2.);
            case 7:
              return Point<3>(1.-pp[0]*2.,pp[2]*2.-1.,pp[1]*2.);
          }
        }
        case 4: {
          switch (cvi1) {
            case 0:
              return Point<3>((1.-pp[2])*2.,pp[0]*2.,1.-pp[1]*2.);
            case 5:
              return Point<3>(pp[1]*2.,(1.-pp[2])*2.,1.-pp[0]*2.);
            case 6:
              return Point<3>(pp[0]*2.,pp[1]*2.,pp[2]*2.-1.);
          }
        }
        case 5: {
          switch (cvi1) {
            case 1:
              return Point<3>(1.-pp[0]*2.,(1.-pp[2])*2.,1.-pp[1]*2.);
            case 4:
              return Point<3>(1.-pp[1]*2.,pp[0]*2.,pp[2]*2.-1.);
            case 7:
              return Point<3>(pp[2]*2.-1.,pp[1]*2.,1.-pp[0]*2.);
          }
        }
        case 6: {
          switch (cvi1) {
            case 2:
              return Point<3>(pp[0]*2.,pp[2]*2.-1.,1.-pp[1]*2.);
            case 4:
              return Point<3>((1.-pp[2])*2.,1.-pp[1]*2.,1.-pp[0]*2.);
            case 7:
              return Point<3>(pp[1]*2.,1.-pp[0]*2.,pp[2]*2.-1.);
          }
        }
        case 7: {
          switch (cvi1) {
            case 3:
              return Point<3>(pp[2]*2.-1.,1.-pp[0]*2.,1.-pp[1]*2.);
            case 5:
              return Point<3>(1.-pp[0]*2.,1.-pp[1]*2.,pp[2]*2.-1.);
            case 6:
              return Point<3>(1.-pp[1]*2.,pp[2]*2.-1.,1.-pp[0]*2.);
          }
        }
      }
    }
    case 5: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[6]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>((1.-pp[0])*2.,pp[1]*2.,(1.-pp[2])*2.);
            case 2:
              return Point<3>((1.-pp[2])*2.,(1.-pp[0])*2.,pp[1]*2.);
            case 4:
              return Point<3>(pp[1]*2.,(1.-pp[2])*2.,(1.-pp[0])*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[0]*2.-1.,(1.-pp[2])*2.,pp[1]*2.);
            case 3:
              return Point<3>(1.-pp[1]*2.,(1.-pp[0])*2.,(1.-pp[2])*2.);
            case 5:
              return Point<3>(pp[2]*2.-1.,pp[1]*2.,(1.-pp[0])*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[1]*2.,pp[0]*2.-1.,(1.-pp[2])*2.);
            case 3:
              return Point<3>((1.-pp[0])*2.,pp[2]*2.-1.,pp[1]*2.);
            case 6:
              return Point<3>((1.-pp[2])*2.,1.-pp[1]*2.,(1.-pp[0])*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[2]*2.-1.,pp[0]*2.-1.,pp[1]*2.);
            case 2:
              return Point<3>(pp[0]*2.-1.,1.-pp[1]*2.,(1.-pp[2])*2.);
            case 7:
              return Point<3>(1.-pp[1]*2.,pp[2]*2.-1.,(1.-pp[0])*2.);
          }
        }
        case 4: {
          switch (cvi1) {
            case 0:
              return Point<3>((1.-pp[2])*2.,pp[1]*2.,pp[0]*2.-1.);
            case 5:
              return Point<3>((1.-pp[0])*2.,(1.-pp[2])*2.,1.-pp[1]*2.);
            case 6:
              return Point<3>(pp[1]*2.,(1.-pp[0])*2.,pp[2]*2.-1.);
          }
        }
        case 5: {
          switch (cvi1) {
            case 1:
              return Point<3>(1.-pp[1]*2.,(1.-pp[2])*2.,pp[0]*2.-1.);
            case 4:
              return Point<3>(pp[0]*2.-1.,pp[1]*2.,pp[2]*2.-1.);
            case 7:
              return Point<3>(pp[2]*2.-1.,(1.-pp[0])*2.,1.-pp[1]*2.);
          }
        }
        case 6: {
          switch (cvi1) {
            case 2:
              return Point<3>(pp[1]*2.,pp[2]*2.-1.,pp[0]*2.-1.);
            case 4:
              return Point<3>((1.-pp[2])*2.,pp[0]*2.-1.,1.-pp[1]*2.);
            case 7:
              return Point<3>((1.-pp[0])*2.,1.-pp[1]*2.,pp[2]*2.-1.);
          }
        }
        case 7: {
          switch (cvi1) {
            case 3:
              return Point<3>(pp[2]*2.-1.,1.-pp[1]*2.,pp[0]*2.-1.);
            case 5:
              return Point<3>(1.-pp[1]*2.,pp[0]*2.-1.,pp[2]*2.-1.);
            case 6:
              return Point<3>(pp[0]*2.-1.,pp[2]*2.-1.,1.-pp[1]*2.);
          }
        }
      }
    }
    case 6: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[7]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[0]*2.,(1.-pp[1])*2.,(1.-pp[2])*2.);
            case 2:
              return Point<3>((1.-pp[2])*2.,pp[0]*2.,(1.-pp[1])*2.);
            case 4:
              return Point<3>((1.-pp[1])*2.,(1.-pp[2])*2.,pp[0]*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(1.-pp[0]*2.,(1.-pp[2])*2.,(1.-pp[1])*2.);
            case 3:
              return Point<3>(1.-pp[0]*2.,(1.-pp[1])*2.,(1.-pp[2])*2.);
            case 5:
              return Point<3>(pp[2]*2.-1.,(1.-pp[1])*2.,pp[0]*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>((1.-pp[1])*2.,1.-pp[0]*2.,(1.-pp[2])*2.);
            case 3:
              return Point<3>(pp[0]*2.,pp[2]*2.-1.,(1.-pp[1])*2.);
            case 6:
              return Point<3>((1.-pp[2])*2.,pp[1]*2.-1.,pp[0]*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[2]*2.-1.,1.-pp[0]*2.,(1.-pp[1])*2.);
            case 2:
              return Point<3>(1.-pp[0]*2.,pp[1]*2.-1.,(1.-pp[2])*2.);
            case 7:
              return Point<3>(pp[1]*2.-1.,pp[2]*2.-1.,pp[0]*2.);
          }
        }
        case 4: {
          switch (cvi1) {
            case 0:
              return Point<3>((1.-pp[2])*2.,(1.-pp[1])*2.,1.-pp[0]*2.);
            case 5:
              return Point<3>(pp[0]*2.,(1.-pp[2])*2.,pp[1]*2.-1.);
            case 6:
              return Point<3>((1.-pp[1])*2.,pp[0]*2.,pp[2]*2.-1.);
          }
        }
        case 5: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[1]*2.-1.,(1.-pp[2])*2.,pp[0]*2.);
            case 4:
              return Point<3>(1.-pp[0]*2.,(1.-pp[1])*2.,pp[2]*2.-1.);
            case 7:
              return Point<3>(pp[2]*2.-1.,pp[0]*2.,pp[1]*2.-1.);
          }
        }
        case 6: {
          switch (cvi1) {
            case 2:
              return Point<3>((1.-pp[1])*2.,pp[2]*2.-1.,1.-pp[0]*2.);
            case 4:
              return Point<3>((1.-pp[2])*2.,1.-pp[0]*2.,pp[1]*2.-1.);
            case 7:
              return Point<3>(pp[0]*2.,pp[1]*2.-1.,pp[2]*2.-1.);
          }
        }
        case 7: {
          switch (cvi1) {
            case 3:
              return Point<3>(pp[2]*2.-1.,pp[1]*2.-1.,pp[0]*2.-1.);
            case 5:
              return Point<3>(pp[1]*2.-1.,1.-pp[0]*2.,pp[2]*2.-1.);
            case 6:
              return Point<3>(1.-pp[0]*2.,pp[2]*2.-1.,pp[1]*2.-1.);
          }
        }
      }
    }
    case 7: {
      int cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[7]);
      CHECK_TEST(cvi1>=0);
      switch (cvi0) {
        case 0: {
          switch (cvi1) {
            case 1:
              return Point<3>((1.-pp[0])*2.,(1.-pp[2])*2.,(1.-pp[1])*2.);
            case 2:
              return Point<3>((1.-pp[1])*2.,(1.-pp[0])*2.,(1.-pp[2])*2.);
            case 4:
              return Point<3>((1.-pp[2])*2.,(1.-pp[1])*2.,(1.-pp[0])*2.);
          }
        }
        case 1: {
          switch (cvi1) {
            case 0:
              return Point<3>(pp[0]*2.-1.,(1.-pp[1])*2.,(1.-pp[2])*2.);
            case 3:
              return Point<3>(pp[2]*2.-1.,(1.-pp[0])*2.,(1.-pp[1])*2.);
            case 5:
              return Point<3>(pp[1]*2.-1.,(1.-pp[2])*2.,(1.-pp[0])*2.);
          }
        }
        case 2: {
          switch (cvi1) {
            case 0:
              return Point<3>((1.-pp[2])*2.,pp[0]*2.-1.,(1.-pp[1])*2.);
            case 3:
              return Point<3>((1.-pp[0])*2.,pp[1]*2.-1.,(1.-pp[2])*2.);
            case 6:
              return Point<3>((1.-pp[1])*2.,pp[2]*2.-1.,(1.-pp[0])*2.);
          }
        }
        case 3: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[1]*2.-1.,pp[0]*2.-1.,(1.-pp[2])*2.);
            case 2:
              return Point<3>(pp[0]*2.-1.,pp[2]*2.-1.,(1.-pp[1])*2.);
            case 7:
              return Point<3>(pp[2]*2.-1.,pp[1]*2.-1.,(1.-pp[0])*2.);
          }
        }
        case 4: {
          switch (cvi1) {
            case 0:
              return Point<3>((1.-pp[1])*2.,(1.-pp[2])*2.,pp[0]*2.-1.);
            case 5:
              return Point<3>((1.-pp[0])*2.,(1.-pp[1])*2.,pp[2]*2.-1.);
            case 6:
              return Point<3>((1.-pp[2])*2.,(1.-pp[0])*2.,pp[1]*2.-1.);
          }
        }
        case 5: {
          switch (cvi1) {
            case 1:
              return Point<3>(pp[2]*2.-1.,(1.-pp[1])*2.,pp[0]*2.-1.);
            case 4:
              return Point<3>(pp[0]*2.-1.,(1.-pp[2])*2.,pp[1]*2.-1.);
            case 7:
              return Point<3>(pp[1]*2.-1.,(1.-pp[0])*2.,pp[2]*2.-1.);
          }
        }
        case 6: {
          switch (cvi1) {
            case 2:
              return Point<3>((1.-pp[2])*2.,pp[1]*2.-1.,pp[0]*2.-1.);
            case 4:
              return Point<3>((1.-pp[1])*2.,pp[0]*2.-1.,pp[2]*2.-1.);
            case 7:
              return Point<3>((1.-pp[0])*2.,pp[2]*2.-1.,pp[1]*2.-1.);
          }
        }
        case 7: {
          switch (cvi1) {
            case 3:
              return Point<3>(pp[1]*2.-1.,pp[2]*2.-1.,pp[0]*2.-1.);
            case 5:
              return Point<3>(pp[2]*2.-1.,pp[0]*2.-1.,pp[1]*2.-1.);
            case 6:
              return Point<3>(pp[0]*2.-1.,pp[1]*2.-1.,pp[2]*2.-1.);
          }
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//see FE_Q::FE_Q; must be consistent with parentCoordinateOfCanonical
const Shape* Hexahedron::reorientFor(const Shape *parent,int c,
const NumPtr<int> &line_interior_child_vertices) const {
  CHECK_SAME(vertexIndex(0),parent->vertexIndex(c));
  int reorient_case=-1;
  switch (c) {
    case 0: {
//    parentCoordinateOfCanonical(p,0)=p*0.5
      int cv2=line_interior_child_vertices[2];
      if (vertexIndex(1)==cv2) {
        reorient_case=0;
        break;
      } else if (vertexIndex(2)==cv2) {
        reorient_case=1;
        break;
      } else {
        CHECK_SAME(vertexIndex(4),cv2)
        reorient_case=2;
        break;
      }
    }
    case 1: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(1.-p[1]*0.5,p[0]*0.5,p[2]*0.5)
      int cv2=line_interior_child_vertices[2];
      if (vertexIndex(1)==cv2) {
        reorient_case=2;
        break;
      } else if (vertexIndex(2)==cv2) {
        reorient_case=0;
        break;
      } else {
        CHECK_SAME(vertexIndex(4),cv2)
        reorient_case=1;
        break;
      }
    }
    case 2: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(p[1]*0.5,1.-p[0]*0.5,p[2]*0.5)
      int cv3=line_interior_child_vertices[3];
      if (vertexIndex(1)==cv3) {
        reorient_case=2;
        break;
      } else if (vertexIndex(2)==cv3) {
        reorient_case=0;
        break;
      } else {
        CHECK_SAME(vertexIndex(4),cv3)
        reorient_case=1;
        break;
      }
    }
    case 3: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(1.-p[0]*0.5,1.-p[1]*0.5,p[2]*0.5)
      int cv3=line_interior_child_vertices[3];
      if (vertexIndex(1)==cv3) {
        reorient_case=0;
        break;
      } else if (vertexIndex(2)==cv3) {
        reorient_case=1;
        break;
      } else {
        CHECK_SAME(vertexIndex(4),cv3)
        reorient_case=2;
        break;
      }
    }
    case 4: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(p[1]*0.5,p[0]*0.5,1.-p[2]*0.5)
      int cv4=line_interior_child_vertices[4];
      if (vertexIndex(1)==cv4) {
        reorient_case=0;
        break;
      } else if (vertexIndex(2)==cv4) {
        reorient_case=1;
        break;
      } else {
        CHECK_SAME(vertexIndex(4),cv4)
        reorient_case=2;
        break;
      }
    }
    case 5: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(1.-p[0]*0.5,p[1]*0.5,1.-p[2]*0.5)
      int cv5=line_interior_child_vertices[5];
      if (vertexIndex(1)==cv5) {
        reorient_case=2;
        break;
      } else if (vertexIndex(2)==cv5) {
        reorient_case=0;
        break;
      } else {
        CHECK_SAME(vertexIndex(4),cv5)
        reorient_case=1;
        break;
      }
    }
    case 6: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(p[0]*0.5,1.-p[1]*0.5,1.-p[2]*0.5)
      int cv4=line_interior_child_vertices[4];
      if (vertexIndex(1)==cv4) {
        reorient_case=2;
        break;
      } else if (vertexIndex(2)==cv4) {
        reorient_case=0;
        break;
      } else {
        CHECK_SAME(vertexIndex(4),cv4)
        reorient_case=1;
        break;
      }
    }
    case 7: {
//    parentCoordinateOfCanonical(p,0)=
//      Point<3>(1.-p[1]*0.5,1.-p[0]*0.5,1.-p[2]*0.5)
      int cv5=line_interior_child_vertices[5];
      if (vertexIndex(1)==cv5) {
        reorient_case=0;
        break;
      } else if (vertexIndex(2)==cv5) {
        reorient_case=1;
        break;
      } else {
        CHECK_SAME(vertexIndex(4),cv5)
        reorient_case=2;
        break;
      }
    }
    default:
      OBSOLETE("bad c");
  }
  switch (reorient_case) {
    case 0: { // 0 1 2 3 4 5 6 7
      return this;
    }
    case 1: { // 0 2 4 6 1 3 5 7
      int v[8]={vertexIndex(0),vertexIndex(2),
                vertexIndex(4),vertexIndex(6),
                vertexIndex(1),vertexIndex(3),
                vertexIndex(5),vertexIndex(7)};
      int f[6]={codimension1ShapeIndex(2),codimension1ShapeIndex(3),
                codimension1ShapeIndex(4),codimension1ShapeIndex(5),
                codimension1ShapeIndex(0),codimension1ShapeIndex(1)};
      int l[12]={codimension2ShapeIndex(8),codimension2ShapeIndex(10),
                 codimension2ShapeIndex(0),codimension2ShapeIndex(4),
                 codimension2ShapeIndex(9),codimension2ShapeIndex(11),
                 codimension2ShapeIndex(1),codimension2ShapeIndex(5),
                 codimension2ShapeIndex(2),codimension2ShapeIndex(3),
                 codimension2ShapeIndex(6),codimension2ShapeIndex(7)};
      return OPERATOR_NEW Hexahedron(l,f,v,dof_index);
    }
    case 2 : { // 0 4 1 5 2 6 3 7
      int v[8]={vertexIndex(0),vertexIndex(4),
                vertexIndex(1),vertexIndex(5),
                vertexIndex(2),vertexIndex(6),
                vertexIndex(3),vertexIndex(7)};
      int f[6]={codimension1ShapeIndex(4),codimension1ShapeIndex(5),
                codimension1ShapeIndex(0),codimension1ShapeIndex(1),
                codimension1ShapeIndex(2),codimension1ShapeIndex(3)};
      int l[12]={codimension2ShapeIndex(2),codimension2ShapeIndex(6),
                 codimension2ShapeIndex(8),codimension2ShapeIndex(9),
                 codimension2ShapeIndex(3),codimension2ShapeIndex(7),
                 codimension2ShapeIndex(10),codimension2ShapeIndex(11),
                 codimension2ShapeIndex(0),codimension2ShapeIndex(4),
                 codimension2ShapeIndex(1),codimension2ShapeIndex(5)};
      return OPERATOR_NEW Hexahedron(l,f,v,dof_index);
    }
    default:
      OBSOLETE("base reorient_case");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape2* Hexahedron::newInteriorCodimension1Shape(int i,
int codimension1_element_index,
const NumPtr<int> &interior_child_vertex_indices,
const NumPtr<int> &interior_child_face_vertex_indices,
const NumPtr<int> &interior_child_line_vertex_indices,
const map<pair<int,int>,int> &interior_child_line_map) const {
  CHECK_BOUNDS(i,0,12);
  CHECK_SAME(interior_child_vertex_indices.getNumber(),1);
  CHECK_SAME(interior_child_face_vertex_indices.getNumber(),6);
  CHECK_SAME(interior_child_line_vertex_indices.getNumber(),12);
  int v[4];
  switch (i) {
    case 0:
      v[0]=interior_child_line_vertex_indices[2]; // v10
      v[1]=interior_child_face_vertex_indices[2]; // v22
      v[2]=interior_child_face_vertex_indices[4]; // v24
      v[3]=interior_child_vertex_indices[0];      // v26
      break;
    case 1:
      v[0]=interior_child_face_vertex_indices[2]; // v22
      v[1]=interior_child_line_vertex_indices[6]; // v14
      v[2]=interior_child_vertex_indices[0];      // v26
      v[3]=interior_child_face_vertex_indices[5]; // v25
      break;
    case 2:
      v[0]=interior_child_face_vertex_indices[4]; // v24
      v[1]=interior_child_vertex_indices[0];      // v26
      v[2]=interior_child_line_vertex_indices[3]; // v11
      v[3]=interior_child_face_vertex_indices[3]; // v23
      break;
    case 3:
      v[0]=interior_child_vertex_indices[0];      // v26
      v[1]=interior_child_face_vertex_indices[5]; // v25
      v[2]=interior_child_face_vertex_indices[3]; // v23
      v[3]=interior_child_line_vertex_indices[7]; // v15
      break;
    case 4:
      v[0]=interior_child_line_vertex_indices[0]; // v8
      v[1]=interior_child_face_vertex_indices[4]; // v24
      v[2]=interior_child_face_vertex_indices[0]; // v20
      v[3]=interior_child_vertex_indices[0];      // v26
      break;
    case 5:
      v[0]=interior_child_face_vertex_indices[4]; // v24
      v[1]=interior_child_line_vertex_indices[1]; // v9
      v[2]=interior_child_vertex_indices[0];      // v26
      v[3]=interior_child_face_vertex_indices[1]; // v21
      break;
    case 6:
      v[0]=interior_child_face_vertex_indices[0]; // v20
      v[1]=interior_child_vertex_indices[0];      // v26
      v[2]=interior_child_line_vertex_indices[4]; // v12
      v[3]=interior_child_face_vertex_indices[5]; // v25
      break;
    case 7:
      v[0]=interior_child_vertex_indices[0];      // v26
      v[1]=interior_child_face_vertex_indices[1]; // v21
      v[2]=interior_child_face_vertex_indices[5]; // v25
      v[3]=interior_child_line_vertex_indices[5]; // v13
      break;
    case 8:
      v[0]=interior_child_line_vertex_indices[8]; // v16
      v[1]=interior_child_face_vertex_indices[0]; // v20
      v[2]=interior_child_face_vertex_indices[2]; // v22
      v[3]=interior_child_vertex_indices[0];      // v26
      break;
    case 9:
      v[0]=interior_child_face_vertex_indices[0];  // v20
      v[1]=interior_child_line_vertex_indices[10]; // v18
      v[2]=interior_child_vertex_indices[0];       // v26
      v[3]=interior_child_face_vertex_indices[3];  // v23
      break;
    case 10:
      v[0]=interior_child_face_vertex_indices[2]; // v22
      v[1]=interior_child_vertex_indices[0];      // v26
      v[2]=interior_child_line_vertex_indices[9]; // v17
      v[3]=interior_child_face_vertex_indices[1]; // v21
      break;
    case 11:
      v[0]=interior_child_vertex_indices[0];       // v26
      v[1]=interior_child_face_vertex_indices[3];  // v23
      v[2]=interior_child_face_vertex_indices[1];  // v21
      v[3]=interior_child_line_vertex_indices[11]; // v19
      break;
  }
  int l[4];
  setQuadrilateralLineIndices(v,interior_child_line_map,l);
  return OPERATOR_NEW Quadrilateral(codimension1_element_index,l,v);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::staticProjectToFace(int f,
const NumPtr<Point<2> > &q_points,NumPtr<Point<3> > &q_points_face) {
  switch (f) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=0.;
        q_points_face[i][1]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][0]=1.;
        q_points_face[i][1]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][1]=0.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 3:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][1]=1.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 4:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][2]=0.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=q_points[i][1];
      }
      break;
    case 5:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][2]=1.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=q_points[i][1];
      }
      break;
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::projectToFaceForBDFM(int f,
const NumPtr<Point<2> > &q_points,NumPtr<Point<3> > &q_points_face) 
const {
  switch (f) {
    case 0:
      if (vertexIndex(2)<vertexIndex(4)) {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][0]=0.;
          q_points_face[i][1]=q_points[i][0];
          q_points_face[i][2]=q_points[i][1];
        }
      } else {
        for (int i=0;i<q_points.getNumber();i++) {
          q_points_face[i][0]=0.;
          q_points_face[i][1]=q_points[i][1];
          q_points_face[i][2]=q_points[i][0];
        }
      }
      break;
    case 1: {
      int vmin=min(min(vertexIndex(1),vertexIndex(3)),
                   min(vertexIndex(5),vertexIndex(7)));
      if (vertexIndex(1)==vmin) {
        if (vertexIndex(3)<vertexIndex(5)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][0]=1.;
            q_points_face[i][1]=q_points[i][0];
            q_points_face[i][2]=q_points[i][1];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][0]=1.;
            q_points_face[i][1]=q_points[i][1];
            q_points_face[i][2]=q_points[i][0];
          }
        }
      } else if (vertexIndex(3)==vmin) {
        if (vertexIndex(1)<vertexIndex(7)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][0]=1.;
            q_points_face[i][1]=1.-q_points[i][0];
            q_points_face[i][2]=q_points[i][1];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][0]=1.;
            q_points_face[i][1]=1.-q_points[i][1];
            q_points_face[i][2]=q_points[i][0];
          }
        }
      } else if (vertexIndex(5)==vmin) {
        if (vertexIndex(1)<vertexIndex(7)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][0]=1.;
            q_points_face[i][1]=q_points[i][1];
            q_points_face[i][2]=1.-q_points[i][0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][0]=1.;
            q_points_face[i][1]=q_points[i][0];
            q_points_face[i][2]=1.-q_points[i][1];
          }
        }
      } else {
        if (vertexIndex(3)<vertexIndex(5)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][0]=1.;
            q_points_face[i][1]=1.-q_points[i][1];
            q_points_face[i][2]=1.-q_points[i][0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][0]=1.;
            q_points_face[i][1]=1.-q_points[i][0];
            q_points_face[i][2]=1.-q_points[i][1];
          }
        }
      }
      break;
    }
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][1]=0.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][2]=q_points[i][1];
      }
      break;
    case 3: {
      int vmin=min(min(vertexIndex(2),vertexIndex(3)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(2)==vmin) {
        if (vertexIndex(3)<vertexIndex(6)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][1]=1.;
            q_points_face[i][0]=q_points[i][0];
            q_points_face[i][2]=q_points[i][1];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][1]=1.;
            q_points_face[i][0]=q_points[i][1];
            q_points_face[i][2]=q_points[i][0];
          }
        }
      } else if (vertexIndex(3)==vmin) {
        if (vertexIndex(2)<vertexIndex(7)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][1]=1.;
            q_points_face[i][0]=1.-q_points[i][0];
            q_points_face[i][2]=q_points[i][1];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][1]=1.;
            q_points_face[i][0]=1.-q_points[i][1];
            q_points_face[i][2]=q_points[i][0];
          }
        }
      } else if (vertexIndex(6)==vmin) {
        if (vertexIndex(2)<vertexIndex(7)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][1]=1.;
            q_points_face[i][0]=q_points[i][1];
            q_points_face[i][2]=1.-q_points[i][0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][1]=1.;
            q_points_face[i][0]=q_points[i][0];
            q_points_face[i][2]=1.-q_points[i][1];
          }
        }
      } else {
        if (vertexIndex(3)<vertexIndex(6)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][1]=1.;
            q_points_face[i][0]=1.-q_points[i][1];
            q_points_face[i][2]=1.-q_points[i][0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][1]=1.;
            q_points_face[i][0]=1.-q_points[i][0];
            q_points_face[i][2]=1.-q_points[i][1];
          }
        }
      }
      break;
    }
    case 4:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_face[i][2]=0.;
        q_points_face[i][0]=q_points[i][0];
        q_points_face[i][1]=q_points[i][1];
      }
      break;
    case 5: {
      int vmin=min(min(vertexIndex(4),vertexIndex(5)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(4)==vmin) {
        if (vertexIndex(5)<vertexIndex(6)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][2]=1.;
            q_points_face[i][0]=q_points[i][0];
            q_points_face[i][1]=q_points[i][1];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][2]=1.;
            q_points_face[i][0]=q_points[i][1];
            q_points_face[i][1]=q_points[i][0];
          }
        }
      } else if (vertexIndex(5)==vmin) {
        if (vertexIndex(4)<vertexIndex(7)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][2]=1.;
            q_points_face[i][0]=1.-q_points[i][0];
            q_points_face[i][1]=q_points[i][1];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][2]=1.;
            q_points_face[i][0]=1.-q_points[i][1];
            q_points_face[i][1]=q_points[i][0];
          }
        }
      } else if (vertexIndex(6)==vmin) {
        if (vertexIndex(4)<vertexIndex(7)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][2]=1.;
            q_points_face[i][0]=q_points[i][1];
            q_points_face[i][1]=1.-q_points[i][0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][2]=1.;
            q_points_face[i][0]=q_points[i][0];
            q_points_face[i][1]=1.-q_points[i][1];
          }
        }
      } else {
        if (vertexIndex(5)<vertexIndex(6)) {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][2]=1.;
            q_points_face[i][0]=1.-q_points[i][1];
            q_points_face[i][1]=1.-q_points[i][0];
          }
        } else {
          for (int i=0;i<q_points.getNumber();i++) {
            q_points_face[i][2]=1.;
            q_points_face[i][0]=1.-q_points[i][0];
            q_points_face[i][1]=1.-q_points[i][1];
          }
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::staticProjectToLine(int l,
const NumPtr<Point<1> > &q_points,NumPtr<Point<3> > &q_points_line) {
  switch (l) {
    case 0:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=0.;
      }
      break;
    case 1:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.;
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=0.;
      }
      break;
    case 2:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=q_points[i][0];
        q_points_line[i][1]=0.;
        q_points_line[i][2]=0.;
      }
      break;
    case 3:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=q_points[i][0];
        q_points_line[i][1]=1.;
        q_points_line[i][2]=0.;
      }
      break;
    case 4:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=1.;
      }
      break;
    case 5:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.;
        q_points_line[i][1]=q_points[i][0];
        q_points_line[i][2]=1.;
      }
      break;
    case 6:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=q_points[i][0];
        q_points_line[i][1]=0.;
        q_points_line[i][2]=1.;
      }
      break;
    case 7:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=q_points[i][0];
        q_points_line[i][1]=1.;
        q_points_line[i][2]=1.;
      }
      break;
    case 8:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=0.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    case 9:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.;
        q_points_line[i][1]=0.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    case 10:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=0.;
        q_points_line[i][1]=1.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    case 11:
      for (int i=0;i<q_points.getNumber();i++) {
        q_points_line[i][0]=1.;
        q_points_line[i][1]=1.;
        q_points_line[i][2]=q_points[i][0];
      }
      break;
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::projectToSubface(int f,int s,
const NumPtr<Point<2> > &q_points,NumPtr<Point<3> > &q_points_subface) {
  switch (f) {
    case 0: {
      switch (s) {
        case 0: { // child 0
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=qp[0]*0.5;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=1.-qp[1]*0.5;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 2: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=qp[1]*0.5;
            qps[2]=1.-qp[0]*0.5;
          }
          break;
        }
        case 3: { // child 6
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=0.;
            qps[1]=1.-qp[0]*0.5;
            qps[2]=1.-qp[1]*0.5;
          }
          break;
        }
      }
      break;
    }
    case 1: {
      switch (s) {
        case 0: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.;
            qps[1]=qp[0]*0.5;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 3
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.;
            qps[1]=1.-qp[1]*0.5;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 2: { // child 5
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.;
            qps[1]=qp[1]*0.5;
            qps[2]=1.-qp[0]*0.5;
          }
          break;
        }
        case 3: { // child 7
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.;
            qps[1]=1.-qp[0]*0.5;
            qps[2]=1.-qp[1]*0.5;
          }
          break;
        }
      }
      break;
    }
    case 2: {
      switch (s) {
        case 0: { // child 0
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=0.;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[1]*0.5;
            qps[1]=0.;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 2: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=0.;
            qps[2]=1.-qp[0]*0.5;
          }
          break;
        }
        case 3: { // child 5
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[0]*0.5;
            qps[1]=0.;
            qps[2]=1.-qp[1]*0.5;
          }
          break;
        }
      }
      break;
    }
    case 3: {
      switch (s) {
        case 0: { // child 0
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=1.;
            qps[2]=qp[1]*0.5;
          }
          break;
        }
        case 1: { // child 1
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[1]*0.5;
            qps[1]=1.;
            qps[2]=qp[0]*0.5;
          }
          break;
        }
        case 2: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=1.;
            qps[2]=1.-qp[0]*0.5;
          }
          break;
        }
        case 3: { // child 5
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[0]*0.5;
            qps[1]=1.;
            qps[2]=1.-qp[1]*0.5;
          }
          break;
        }
      }
      break;
    }
    case 4: {
      switch (s) {
        case 0: { // child 0
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=qp[1]*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 1: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[1]*0.5;
            qps[1]=qp[0]*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 2: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=1.-qp[0]*0.5;
            qps[2]=0.;
          }
          break;
        }
        case 3: { // child 6
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[0]*0.5;
            qps[1]=1.-qp[1]*0.5;
            qps[2]=0.;
          }
          break;
        }
      }
      break;
    }
    case 5: {
      switch (s) {
        case 0: { // child 0
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[0]*0.5;
            qps[1]=qp[1]*0.5;
            qps[2]=1.;
          }
          break;
        }
        case 1: { // child 2
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[1]*0.5;
            qps[1]=qp[0]*0.5;
            qps[2]=1.;
          }
          break;
        }
        case 2: { // child 4
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=qp[1]*0.5;
            qps[1]=1.-qp[0]*0.5;
            qps[2]=1.;
          }
          break;
        }
        case 3: { // child 6
          for (int q=0;q<q_points.getNumber();q++) {
            Point<3> &qps=q_points_subface[q];
            const Point<2> &qp=q_points[q];
            qps[0]=1.-qp[0]*0.5;
            qps[1]=1.-qp[1]*0.5;
            qps[2]=1.;
          }
          break;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Shape3* Hexahedron::newChildShape(int c,int child_element_index,
const NumPtr<int> &interior_child_vertex_indices,
const NumPtr<int> &interior_child_face_vertex_indices,
const NumPtr<int> &interior_child_line_vertex_indices,
const map<pair<pair<int,int>,pair<int,int> >,int>
&interior_child_face_map,
const map<pair<int,int>,int> &interior_child_line_map) const {
  CHECK_BOUNDS(c,0,8);
  CHECK_SAME(interior_child_vertex_indices.getNumber(),1);
  CHECK_SAME(interior_child_face_vertex_indices.getNumber(),6);
  CHECK_SAME(interior_child_line_vertex_indices.getNumber(),12);
  int v[8];
  switch (c) {
    case 0:
      v[0]=vertexIndex(0);                        // v0
      v[1]=interior_child_line_vertex_indices[2]; // v10
      v[2]=interior_child_line_vertex_indices[0]; // v8
      v[3]=interior_child_face_vertex_indices[4]; // v24
      v[4]=interior_child_line_vertex_indices[8]; // v16
      v[5]=interior_child_face_vertex_indices[2]; // v22
      v[6]=interior_child_face_vertex_indices[0]; // v20
      v[7]=interior_child_vertex_indices[0];      // v26
      break;
    case 1:
      v[0]=interior_child_line_vertex_indices[2]; // v10
      v[1]=vertexIndex(1);                        // v1
      v[2]=interior_child_face_vertex_indices[4]; // v24
      v[3]=interior_child_line_vertex_indices[1]; // v9
      v[4]=interior_child_face_vertex_indices[2]; // v22
      v[5]=interior_child_line_vertex_indices[9]; // v17
      v[6]=interior_child_vertex_indices[0];      // v26
      v[7]=interior_child_face_vertex_indices[1]; // v21
      break;
    case 2:
      v[0]=interior_child_line_vertex_indices[0]; // v8
      v[1]=interior_child_face_vertex_indices[4]; // v24
      v[2]=vertexIndex(2);                        // v2
      v[3]=interior_child_line_vertex_indices[3]; // v11
      v[4]=interior_child_face_vertex_indices[0]; // v20
      v[5]=interior_child_vertex_indices[0];      // v26
      v[6]=interior_child_line_vertex_indices[10];// v18
      v[7]=interior_child_face_vertex_indices[3]; // v23
      break;
    case 3:
      v[0]=interior_child_face_vertex_indices[4]; // v24
      v[1]=interior_child_line_vertex_indices[1]; // v9
      v[2]=interior_child_line_vertex_indices[3]; // v11
      v[3]=vertexIndex(3);                        // v3
      v[4]=interior_child_vertex_indices[0];      // v26
      v[5]=interior_child_face_vertex_indices[1]; // v21
      v[6]=interior_child_face_vertex_indices[3]; // v23
      v[7]=interior_child_line_vertex_indices[11];// v19
      break;
    case 4:
      v[0]=interior_child_line_vertex_indices[8]; // v16
      v[1]=interior_child_face_vertex_indices[2]; // v22
      v[2]=interior_child_face_vertex_indices[0]; // v20
      v[3]=interior_child_vertex_indices[0];      // v26
      v[4]=vertexIndex(4);                        // v4
      v[5]=interior_child_line_vertex_indices[6]; // v14
      v[6]=interior_child_line_vertex_indices[4]; // v12
      v[7]=interior_child_face_vertex_indices[5]; // v25
      break;
    case 5:
      v[0]=interior_child_face_vertex_indices[2]; // v22
      v[1]=interior_child_line_vertex_indices[9]; // v17
      v[2]=interior_child_vertex_indices[0];      // v26
      v[3]=interior_child_face_vertex_indices[1]; // v21
      v[4]=interior_child_line_vertex_indices[6]; // v14
      v[5]=vertexIndex(5);                        // v5
      v[6]=interior_child_face_vertex_indices[5]; // v25
      v[7]=interior_child_line_vertex_indices[5]; // v13
      break;
    case 6:
      v[0]=interior_child_face_vertex_indices[0]; // v20
      v[1]=interior_child_vertex_indices[0];      // v26
      v[2]=interior_child_line_vertex_indices[10];// v18
      v[3]=interior_child_face_vertex_indices[3]; // v23
      v[4]=interior_child_line_vertex_indices[4]; // v12
      v[5]=interior_child_face_vertex_indices[5]; // v25
      v[6]=vertexIndex(6);                        // v6
      v[7]=interior_child_line_vertex_indices[7]; // v15
      break;
    case 7:
      v[0]=interior_child_vertex_indices[0];      // v26
      v[1]=interior_child_face_vertex_indices[1]; // v21
      v[2]=interior_child_face_vertex_indices[3]; // v23
      v[3]=interior_child_line_vertex_indices[11];// v19
      v[4]=interior_child_face_vertex_indices[5]; // v25
      v[5]=interior_child_line_vertex_indices[5]; // v13
      v[6]=interior_child_line_vertex_indices[7]; // v15
      v[7]=vertexIndex(7);                        // v7
      break;
  }
  int f[6];
  for (int i=0;i<6;i++) {
    int ql[4]={-1,-2,-3,-4};
    int qv[4]={v[face_vertex_to_cell_vertex[i][0]],
               v[face_vertex_to_cell_vertex[i][1]],
               v[face_vertex_to_cell_vertex[i][2]],
               v[face_vertex_to_cell_vertex[i][3]]};
    Quadrilateral bogus_quadrilateral(-1,ql,qv);
    pair<pair<int,int>,pair<int,int> >
      p(make_pair(bogus_quadrilateral.vertexIndex(0),
                  bogus_quadrilateral.vertexIndex(1)),
        make_pair(bogus_quadrilateral.vertexIndex(2),
                  bogus_quadrilateral.vertexIndex(3)));
    map<pair<pair<int,int>,pair<int,int> >,int>::const_iterator 
      icfm_iterator=interior_child_face_map.find(p);
    CHECK_TEST(icfm_iterator!=interior_child_face_map.end());
    f[i]=icfm_iterator->second;
  }
  int l[12];
  for (int i=0;i<12;i++) {
    int iv[2]={v[canonical_vertex_indices[i][0]],
               v[canonical_vertex_indices[i][1]]};
    Interval bogus_interval(-1,iv);
    pair<int,int> p(bogus_interval.vertexIndex(0),
                    bogus_interval.vertexIndex(1));
    map<pair<int,int>,int>::const_iterator iclm_iterator=
      interior_child_line_map.find(p);
    CHECK_TEST(iclm_iterator!=interior_child_line_map.end());
    l[i]=iclm_iterator->second;
  }
  return OPERATOR_NEW Hexahedron(child_element_index,l,f,v);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Hexahedron::lineFromFace(const Shape2 *face,int f,int fl) const {
  CHECK_SAME(face->elementDataIndex(),codimension1ShapeIndex(f));
  switch (f) {
    case 0: {
      int lines[4]={8,10,0,4};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv2=face->vertexEqualToVertexIndex(vertexIndex(2));
      switch (fv0) {
        case 0: {
          if (fv2==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv2==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv2==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv2==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
    case 1: {
      int lines[4]={9,11,1,5};
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      int fv3=face->vertexEqualToVertexIndex(vertexIndex(3));
      switch (fv1) {
        case 0: {
          if (fv3==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv3==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv3==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv3==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
    case 2: {
      int lines[4]={8,9,2,6};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      switch (fv0) {
        case 0: {
          if (fv1==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv1==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv1==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv1==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
    case 3: {
      int lines[4]={10,11,3,7};
      int fv2=face->vertexEqualToVertexIndex(vertexIndex(2));
      int fv3=face->vertexEqualToVertexIndex(vertexIndex(3));
      switch (fv2) {
        case 0: {
          if (fv3==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv3==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv3==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv3==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
    case 4: {
      int lines[4]={0,1,2,3};
      int fv0=face->vertexEqualToVertexIndex(vertexIndex(0));
      int fv1=face->vertexEqualToVertexIndex(vertexIndex(1));
      switch (fv0) {
        case 0: {
          if (fv1==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv1==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv1==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv1==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
    case 5: {
      int lines[4]={4,5,6,7};
      int fv4=face->vertexEqualToVertexIndex(vertexIndex(4));
      int fv5=face->vertexEqualToVertexIndex(vertexIndex(5));
      switch (fv4) {
        case 0: {
          if (fv5==1) return lines[fl];
          else return lines[(fl>=2 ? fl-2 : fl+2)];
        }
        case 1: {
          if (fv5==3) return lines[(fl>=2 ? fl-2 : 3-fl)];
          else return lines[(fl>=2 ? fl : 1-fl)];
        }
        case 2: {
          if (fv5==0) return lines[(fl>=2 ? 3-fl : fl+2)];
          else return lines[(fl>=2 ? 5-fl : fl)];
        }
        case 3: {
          if (fv5==2) return lines[(fl>=2 ? 5-fl : 1-fl)];
          else return lines[3-fl];
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::staticEquallySpacedLatticePoints(int n0,int n1,int n2,
NumPtr<Point<3> > &points) {
  points.allocate(staticNumberEquallySpacedLatticePoints(n0,n1,n2));
  Point<3> p(0.,0.,0.);
  REAL s0=(n0>0 ? 1./static_cast<REAL>(n0) : 0.);
  REAL s1=(n1>0 ? 1./static_cast<REAL>(n1) : 0.);
  REAL s2=(n2>0 ? 1./static_cast<REAL>(n2) : 0.);
  int l=0;
  for (int k=0;k<=n2;k++,p[2]+=s2) {
    for (int j=0;j<=n1;j++,p[1]+=s1) {
      for (int i=0;i<=n0;i++,p[0]+=s0,l++) points[l]=p;
      p[0]=0.;
    }
    p[1]=0.;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Hexahedron::staticEquallySpacedLatticePointIndexOfVertex(int v,
int n0,int n1,int n2) {
  CHECK_TEST(min(n0,min(n1,n2))>0);
  CHECK_TEST(v<8);
  switch (v) {
    case 0:
      return 0;
    case 1: 
      return n0;
    case 2:
      return n1*(n0+1);
    case 3:
      return n1*(n0+1)+ n0;
    case 4:
      return n2*(n1+1)*(n0+1);
    case 5:
      return n2*(n1+1)*(n0+1)+n0;
    case 6:
      return (n2*(n1+1)+n1)*(n0+1);
    case 7:
      return (n2+1)*(n1+1)*(n0+1)-1;
    default:
      OBSOLETE("bad vertex");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::staticEquallySpacedLatticePointIndicesOfFace(int f,
int n0,int n1,int n2,NumPtr<int> &indices) {
  int nl=n0+1;
  int np=(n1+1)*nl;
  switch (f) {
    case 0: { // (0,j,k)
      indices.allocate((n1+1)*(n2+1));
      int ijk=0;
      int jk=0;
      for (int k=0;k<=n2;k++) {
        for (int j=0;j<=n1;j++) {
          indices[jk++]=ijk;
          ijk+=nl;
        }
      }
      break;
    }
    case 1: { // (n,j,k)
      indices.allocate((n1+1)*(n2+1));
      int ijk=n0;
      int jk=0;
      for (int k=0;k<=n2;k++) {
        for (int j=0;j<=n1;j++) {
          indices[jk++]=ijk;
          ijk+=nl;
        }
      }
      break;
    }
    case 2: { // (i,0,k)
      indices.allocate((n2+1)*(n0+1));
      int ijk=0;
      int ik=0;
      for (int k=0;k<=n2;k++) {
        for (int i=0;i<=n0;i++) {
          indices[ik++]=ijk;
          ijk++;
        }
        ijk+=np-nl;
      }
      break;
    }
    case 3: { // (i,n,k)
      indices.allocate((n2+1)*(n0+1));
      int ijk=np-nl;
      int ik=0;
      for (int k=0;k<=n2;k++) {
        for (int i=0;i<=n0;i++) {
          indices[ik++]=ijk;
          ijk++;
        }
        ijk+=np-nl;
      }
      break;
    }
    case 4: { // (i,j,0)
      indices.allocate(np);
      int ijk=0;
      int ij=0;
      for (int j=0;j<=n1;j++) {
        for (int i=0;i<=n0;i++) {
          indices[ij++]=ijk;
          ijk++;
        }
      }
      break;
    }
    case 5: { // (i,j,n)
      indices.allocate(np);
      int ijk=n2*np;
      int ij=0;
      for (int j=0;j<=n1;j++) {
        for (int i=0;i<=n0;i++) {
          indices[ij++]=ijk;
          ijk++;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::equallySpacedLatticePointIndicesOfFace(int f,int n0,
int n1,int n2,NumPtr<int> &indices) const {
  int nl=n0+1;
  int np=(n1+1)*nl;
  switch (f) {
    case 0: {
      indices.allocate((n1+1)*(n2+1));
      if (vertexIndex(2)<vertexIndex(4)) { // 0 < 2 < 4
        int m=0;
        int jk=0;
        for (int k=0;k<=n2;k++) {
          int ijk=m;
          for (int j=0;j<=n1;j++,jk++) {
            indices[jk]=ijk;
            ijk+=nl;
          }
          m+=np;
        }
      } else { // 0 < 4 < 2
        int m=0;
        int jk=0;
        for (int j=0;j<=n1;j++) {
          int ijk=m;
          for (int k=0;k<=n2;k++,jk++) {
            indices[jk]=ijk;
            ijk+=np;
          }
          m+=nl;
        }
      }
      break;
    }
    case 1: {
      indices.allocate((n1+1)*(n2+1));
      if (vertexIndex(3)<vertexIndex(5)) { // 1 < 3 < 5
        int m=n0;
        int jk=0;
        for (int k=0;k<=n2;k++) {
          int ijk=m;
          for (int j=0;j<=n1;j++,jk++) {
            indices[jk]=ijk;
            ijk+=nl;
          }
          m+=np;
        }
      } else { // 1 < 5 < 3
        int m=n0;
        int jk=0;
        for (int j=0;j<=n1;j++) {
          int ijk=m;
          for (int k=0;k<=n2;k++,jk++) {
            indices[jk]=ijk;
            ijk+=np;
          }
          m+=nl;
        }
      }
      break;
    }
    case 2: { // 0 < 1 < 4
      indices.allocate((n2+1)*(n0+1));
      int m=0;
      int ik=0;
      for (int k=0;k<=n2;k++) {
        int ijk=m;
        for (int i=0;i<=n0;i++,ik++) {
          indices[ik]=ijk;
          ijk++;
        }
        m+=np;
      }
      break;
    }
    case 3: {
      indices.allocate((n2+1)*(n0+1));
      int vmin=min(min(vertexIndex(2),vertexIndex(3)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(2)==vmin) {
        if (vertexIndex(3)<vertexIndex(6)) {  // 2 < 3 < 6
          int m=n1*nl;
          int ik=0;
          for (int k=0;k<=n2;k++) {
            int ijk=m;
            for (int i=0;i<=n0;i++,ik++) {
              indices[ik]=ijk;
              ijk++;
            }
            m+=np;
          }
        } else {  // 2 < 6 < 3
          int m=n1*nl;
          int ik=0;
          for (int i=0;i<=n0;i++) {
            int ijk=m;
            for (int k=0;k<=n2;k++,ik++) {
              indices[ik]=ijk;
              ijk+=np;
            }
            m++;
          }
        }
      } else if (vertexIndex(3)==vmin) {
        if (vertexIndex(2)<vertexIndex(7)) {  // 3 < 2 < 7
          int m=np-1;
          int ik=0;
          for (int k=0;k<=n2;k++) {
            int ijk=m;
            for (int i=n0;i>=0;i--,ik++) {
              indices[ik]=ijk;
              ijk--;
            }
            m+=np;
          }
        } else {  // 3 < 7 < 2
          int m=np-1;
          int ik=0;
          for (int i=n0;i>=0;i--) {
            int ijk=m;
            for (int k=0;k<=n2;k++,ik++) {
              indices[ik]=ijk;
              ijk+=np;
            }
            m--;
          }
        }
      } else if (vertexIndex(6)==vmin) {
        if (vertexIndex(2)<vertexIndex(7)) { // 6 < 2 < 7
          int m=n2*np+n1*nl;
          int ik=0;
          for (int i=0;i<=n0;i++) {
            int ijk=m;
            for (int k=n2;k>=0;k--,ik++) {
              indices[ik]=ijk;
              ijk-=np;
            }
            m++;
          }
        } else { // 6 < 7 < 2
          int m=n2*np+n1*nl;
          int ik=0;
          for (int k=n2;k>=0;k--) {
            int ijk=m;
            for (int i=0;i<=n0;i++,ik++) {
              indices[ik]=ijk;
              ijk++;
            }
            m-=np;
          }
        }
      } else {
        if (vertexIndex(3)<vertexIndex(6)) { // 7 < 3 < 6
          int m=(n2+1)*np-1;
          int ik=0;
          for (int i=n0;i>=0;i--) {
            int ijk=m;
            for (int k=n2;k>=0;k--,ik++) {
              indices[ik]=ijk;
              ijk-=np;
            }
            m--;
          }
        } else { // 7 < 6 < 3
          int m=(n2+1)*np-1;
          int ik=0;
          for (int k=n2;k>=0;k--) {
            int ijk=m;
            for (int i=n0;i>=0;i--,ik++) {
              indices[ik]=ijk;
              ijk--;
            }
            m-=np;
          }
        }
      }
      break;
    }
    case 4: { // 0 < 1 < 2
      indices.allocate((n0+1)*(n1+1));
      int m=0;
      int ij=0;
      for (int j=0;j<=n1;j++) {
        int ijk=m;
        for (int i=0;i<=n0;i++,ij++) {
          indices[ij]=ijk;
          ijk++;
        }
        m+=nl;;
      }
      break;
    }
    case 5: {
      indices.allocate((n0+1)*(n1+1));
      int vmin=min(min(vertexIndex(4),vertexIndex(5)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(4)==vmin) {
        if (vertexIndex(5)<vertexIndex(6)) { // 4 < 5 < 6
          int m=n2*np;
          int ij=0;
          for (int j=0;j<=n1;j++) {
            int ijk=m;
            for (int i=0;i<=n0;i++,ij++) {
              indices[ij]=ijk;
              ijk++;
            }
            m+=nl;
          }
        } else { // 4 < 6 < 5
          int m=n2*np;
          int ij=0;
          for (int i=0;i<=n0;i++) {
            int ijk=m;
            for (int j=0;j<=n1;j++,ij++) {
              indices[ij]=ijk;
              ijk+=nl;
            }
            m++;
          }
        }
      } else if (vertexIndex(5)==vmin) {
        if (vertexIndex(4)<vertexIndex(7)) { // 5 < 4 < 7
          int m=n2*np+n0;
          int ij=0;
          for (int j=0;j<=n1;j++) {
            int ijk=m;
            for (int i=n0;i>=0;i--,ij++) {
              indices[ij]=ijk;
              ijk--;
            }
            m+=nl;
          }
        } else { // 5 < 7 < 4
          int m=n2*np+n0;
          int ij=0;
          for (int i=n0;i>=0;i--) {
            int ijk=m;
            for (int j=0;j<=n1;j++,ij++) {
              indices[ij]=ijk;
              ijk+=nl;
            }
            m--;
          }
        }
      } else if (vertexIndex(6)==vmin) {
        if (vertexIndex(4)<vertexIndex(7)) { // 6 < 4 < 7
          int m=n2*np+n1*nl;
          int ij=0;
          for (int i=0;i<=n0;i++) {
            int ijk=m;
            for (int j=n1;j>=0;j--,ij++) {
              indices[ij]=ijk;
              ijk-=nl;
            }
            m++;
          }
        } else { // 6 < 7 < 4
          int m=n2*np+n1*nl;
          int ij=0;
          for (int j=n1;j>=0;j--) {
            int ijk=m;
            for (int i=0;i<=n0;i++,ij++) {
              indices[ij]=ijk;
              ijk++;
            }
            m-=nl;
          }
        }
      } else { // 7 < 5 < 6
        if (vertexIndex(5)<vertexIndex(6)) {
          int m=(n2+1)*np-1;
          int ij=0;
          for (int j=n1;j>=0;j--) {
            int ijk=m;
            for (int i=n0;i>=0;i--,ij++) {
              indices[ij]=ijk;
              ijk-=nl;;
            }
            m--;
          }
        } else { // 7 < 6 < 5
          int m=(n2+1)*np-1;
          int ij=0;
          for (int j=n1;j>=0;j--) {
            int ijk=m;
            for (int i=n0;i>=0;i--,ij++) {
              indices[ij]=ijk;
              ijk--;
            }
            m-=nl;
          }
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::staticEquallySpacedLatticePointIndicesOfLine(int l,
int n0,int n1,int n2,NumPtr<int> &indices) {
  int nl=n0+1;
  int np=(n1+1)*nl;
  switch (l) {
    case 0: {
      indices.allocate(n1+1);
      int ijk=0;
      for (int j=0;j<=n1;j++) {
        indices[j]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 1: {
      indices.allocate(n1+1);
      int ijk=n0;
      for (int j=0;j<=n1;j++) {
        indices[j]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 2: {
      indices.allocate(nl);
      for (int i=0;i<=n0;i++) indices[i]=i;
      break;
    }
    case 3: {
      indices.allocate(nl);
      int ijk=n1*nl;
      for (int i=0;i<=n0;i++) {
        indices[i]=ijk++;
      }
      break;
    }
    case 4: {
      indices.allocate(n1+1);
      int ijk=n2*np;
      for (int j=0;j<=n1;j++) {
        indices[j]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 5: {
      indices.allocate(n1+1);
      int ijk=n2*np+n0;
      for (int j=0;j<=n1;j++) {
        indices[j]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 6: {
      indices.allocate(nl);
      int ijk=n2*np;
      for (int i=0;i<=n0;i++) {
        indices[i]=ijk++;
      }
      break;
    }
    case 7: {
      indices.allocate(nl);
      int ijk=n2*np+n1*nl;
      for (int i=0;i<=n0;i++) {
        indices[i]=ijk;
        ijk++;
      }
      break;
    }
    case 8: {
      indices.allocate(n2+1);
      int ijk=0;
      for (int k=0;k<=n2;k++) {
        indices[k]=ijk;
        ijk+=np;
      }
      break;
    }
    case 9: {
      indices.allocate(n2+1);
      int ijk=n0;
      for (int k=0;k<=n2;k++) {
        indices[k]=ijk;
        ijk+=np;
      }
      break;
    }
    case 10: {
      indices.allocate(n2+1);
      int ijk=n1*nl;
      for (int k=0;k<=n2;k++) {
        indices[k]=ijk;
        ijk+=np;
      }
      break;
    }
    case 11: {
      indices.allocate(n2+1);
      int ijk=np-1;
      for (int k=0;k<=n2;k++) {
        indices[k]=ijk;
        ijk+=np;
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(
int f,int n0,int n1,int n2,NumPtr<int> &indices) {
  if (max(n0,max(n1,n2))<=1) return;
  int nl=n0+1;
  int np=(n1+1)*nl;
  switch (f) {
    case 0: {
      indices.allocate((n1-1)*(n2-1));
      int m=np+nl;
      int jk=0;
      for (int k=1;k<n2;k++) {
        int ijk=m;
        for (int j=1;j<n1;j++) {
          indices[jk++]=ijk;
          ijk+=nl;
        }
        m+=np;
      }
      break;
    }
    case 1: {
      indices.allocate((n1-1)*(n2-1));
      int m=np+2*nl-1;
      int jk=0;
      for (int k=1;k<n2;k++) {
        int ijk=m;
        for (int j=1;j<n1;j++) {
          indices[jk++]=ijk;
          ijk+=nl;
        }
        m+=np;
      }
      break;
    }
    case 2: { // 0 < 1 < 4
      indices.allocate((n2-1)*(n0-1));
      int m=np+1;
      int ik=0;
      for (int k=1;k<n2;k++) {
        int ijk=m;
        for (int i=1;i<n1;i++) {
          indices[ik++]=ijk;
          ijk++;
        }
        m+=np;
      }
      break;
    }
    case 3: {
      indices.allocate((n2-1)*(n0-1));
      int m=np+n1*nl+1;
      int ik=0;
      for (int k=1;k<n2;k++) {
        int ijk=m;
        for (int i=1;i<n0;i++) {
          indices[ik++]=ijk;
          ijk++;
        }
        m+=np;
      }
      break;
    }
    case 4: {
      indices.allocate((n0-1)*(n1-1));
      int m=nl+1;
      int ij=0;
      for (int j=1;j<n1;j++) {
        int ijk=m;
        for (int i=1;i<n0;i++) {
          indices[ij++]=ijk;
          ijk++;
        }
        m+=nl;
      }
      break;
    }
    case 5: {
      indices.allocate((n0-1)*(n1-1));
      int m=n2*np+nl+1;
      int ij=0;
      for (int j=1;j<n1;j++) {
        int ijk=m;
        for (int i=1;i<n0;i++) {
          indices[ij++]=ijk;
          ijk++;
        }
        m+=nl;
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::equallySpacedLatticePointIndicesOfFaceInterior(
int f,int n0,int n1,int n2,NumPtr<int> &indices) const {
  if (max(n0,max(n1,n2))<=1) return;
  int nl=n0+1;
  int np=(n1+1)*nl;
  switch (f) {
    case 0: {
      indices.allocate((n1-1)*(n2-1));
      if (vertexIndex(2)<vertexIndex(4)) { // 0 < 2 < 4
        int m=np+nl;
        int jk=0;
        for (int k=1;k<n2;k++) {
          int ijk=m;
          for (int j=1;j<n1;j++) {
            indices[jk++]=ijk;
            ijk+=nl;
          }
          m+=np;
        }
      } else { // 0 < 4 < 2
        int m=np+nl;
        int jk=0;
        for (int j=1;j<n1;j++) {
          int ijk=m;
          for (int k=1;k<n2;k++) {
            indices[jk++]=ijk;
            ijk+=np;
          }
          m+=nl;
        }
      }
      break;
    }
    case 1: {
      indices.allocate((n1-1)*(n2-1));
      if (vertexIndex(3)<vertexIndex(5)) { // 1 < 3 < 5
        int m=np+2*nl-1;
        int jk=0;
        for (int k=1;k<n2;k++) {
          int ijk=m;
          for (int j=1;j<n1;j++) {
            indices[jk++]=ijk;
            ijk+=nl;
          }
          m+=np;
        }
      } else {
        int m=np+2*nl-1;
        int jk=0;
        for (int j=1;j<n1;j++) {
          int ijk=m;
          for (int k=1;k<n2;k++) {
            indices[jk++]=ijk;
            ijk+=np;
          }
          m+=nl;
        }
      }
      break;
    }
    case 2: { // 0 < 1 < 4
      indices.allocate((n2-1)*(n0-1));
      int m=np+1;
      int ik=0;
      for (int k=1;k<n2;k++) {
        int ijk=m;
        for (int i=1;i<n1;i++) {
          indices[ik++]=ijk;
          ijk++;
        }
        m+=np;
      }
      break;
    }
    case 3: {
      indices.allocate((n2-1)*(n0-1));
      int vmin=min(min(vertexIndex(2),vertexIndex(3)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(2)==vmin) {
        if (vertexIndex(3)<vertexIndex(6)) { // 2 < 3 < 6
          int m=np+n1*nl+1;
          int ik=0;
          for (int k=1;k<n2;k++) {
            int ijk=m;
            for (int i=1;i<n0;i++) {
              indices[ik++]=ijk;
              ijk++;
            }
            m+=np;
          }
        } else { // 2 < 6 < 3
          int m=np+n1*nl+1;
          int ik=0;
          for (int i=1;i<n0;i++) {
            int ijk=m;
            for (int k=1;k<n2;k++) {
              indices[ik++]=ijk;
              ijk+=np;
            }
            m++;
          }
        }
      } else if (vertexIndex(3)==vmin) {
        if (vertexIndex(2)<vertexIndex(7)) { // 3 < 2 < 7
          int m=2*np-2;
          int ik=0;
          for (int k=1;k<n2;k++) {
            int ijk=m;
            for (int i=n0-1;i>=1;i--) {
              indices[ik++]=ijk;
              ijk--;
            }
            m+=np;
          }
        } else { // 3 < 7 < 2
          int m=2*np-2;
          int ik=0;
          for (int i=n0-1;i>=1;i--) {
            int ijk=m;
            for (int k=1;k<n2;k++) {
              indices[ik++]=ijk;
              ijk+=np;
            }
            m--;
          }
        }
      } else if (vertexIndex(6)==vmin) {
        if (vertexIndex(2)<vertexIndex(7)) { // 6 < 2 < 7
          int m=n2*np-n0;
          int ik=0;
          for (int i=1;i<n0;i++) {
            int ijk=m;
            for (int k=n2-1;k>=1;k--) {
              indices[ik++]=ijk;
              ijk-=np;
            }
            m++;
          }
        } else { // 6 < 7 < 2
          int m=n2*np-n0;
          int ik=0;
          for (int k=n2-1;k>=1;k--) {
            int ijk=m;
            for (int i=1;i<n0;i++) {
              indices[ik++]=ijk;
              ijk++;
            }
            m-=np;
          }
        }
      } else {
        if (vertexIndex(3)<vertexIndex(6)) { // 7 < 3 < 6
          int m=n2*np-2;
          int ik=0;
          for (int i=n0-1;i>=1;i--) {
            int ijk=m;
            for (int k=n2-1;k>=1;k--) {
              indices[ik++]=ijk;
              ijk-=np;
            }
            m--;
          }
        } else { // 7 < 6 < 3
          int m=n2*np-2;
          int ik=0;
          for (int k=n2-1;k>=1;k--) {
            int ijk=m;
            for (int i=n0-1;i>=1;i--) {
              indices[ik++]=ijk;
              ijk--;
            }
            m-=np;
          }
        }
      }
      break;
    }
    case 4: {
      indices.allocate((n0-1)*(n1-1));
      int m=nl+1;
      int ij=0;
      for (int j=1;j<n1;j++) {
        int ijk=m;
        for (int i=1;i<n0;i++) {
          indices[ij++]=ijk;
          ijk++;
        }
        m+=nl;
      }
      break;
    }
    case 5: {
      indices.allocate((n0-1)*(n1-1));
      int vmin=min(min(vertexIndex(4),vertexIndex(5)),
                   min(vertexIndex(6),vertexIndex(7)));
      if (vertexIndex(4)==vmin) {
        if (vertexIndex(5)<vertexIndex(6)) { // 4 < 5 < 6
          int m=n2*np+nl+1;
          int ij=0;
          for (int j=1;j<n1;j++) {
            int ijk=m;
            for (int i=1;i<n0;i++) {
              indices[ij++]=ijk;
              ijk++;
            }
            m+=nl;
          }
        } else { // 4 < 6 < 5
          int m=n2*np+nl+1;
          int ij=0;
          for (int i=1;i<n0;i++) {
            int ijk=m;
            for (int j=1;j<n1;j++) {
              indices[ij++]=ijk;
              ijk+=nl;
            }
            m++;
          }
        }
      } else if (vertexIndex(5)==vmin) {
        if (vertexIndex(4)<vertexIndex(7)) { // 5 < 4 < 7
          int m=n2*np+2*n0;
          int ij=0;
          for (int j=1;j<n1;j++) {
            int ijk=m;
            for (int i=n0-1;i>=1;i--) {
              indices[ij++]=ijk;
              ijk--;
            }
            m+=nl;
          }
        } else { // 5 < 7 < 4
          int m=n2*np+2*n0;
          int ij=0;
          for (int i=n0-1;i>=1;i--) {
            int ijk=m;
            for (int j=1;j<n1;j++) {
              indices[ij++]=ijk;
              ijk+=nl;
            }
            m--;
          }
        }
      } else if (vertexIndex(6)==vmin) {
        if (vertexIndex(4)<vertexIndex(7)) { // 6 < 4 < 7
          int m=n2*np+(n1-1)*nl+1;
          int ij=0;
          for (int i=1;i<n0;i++) {
            int ijk=m;
            for (int j=n1-1;j>=1;j--) {
              indices[ij++]=ijk;
              ijk-=nl;
            }
            m++;
          }
        } else { // 6 < 7 < 4
          int m=n2*np+(n1-1)*nl+1;
          int ij=0;
          for (int j=n1-1;j>=1;j--) {
            int ijk=m;
            for (int i=1;i<n0;i++) {
              indices[ij++]=ijk;
              ijk++;
            }
            m-=nl;
          }
        }
      } else {
        if (vertexIndex(5)<vertexIndex(6)) { // 7 < 5 < 6
          int m=n2*np+n1*nl-2;
          int ij=0;
          for (int i=n0-1;i>=1;i--) {
            int ijk=m;
            for (int j=n1-1;j>=1;j--) {
              indices[ij++]=ijk;
              ijk-=nl;;
            }
            m--;
          }
        } else { // 7 < 6 < 5
          int m=n2*np+n1*nl-2;
          int ij=0;
          for (int j=n1-1;j>=1;j--) {
            int ijk=m;
            for (int i=n0-1;i>=1;i--) {
              indices[ij++]=ijk;
              ijk--;
            }
            m-=nl;
          }
        }
      }
      break;
    }
    default:
      OBSOLETE("bad face");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::staticEquallySpacedLatticePointIndicesOfLineInterior(
int l,int n0,int n1,int n2,NumPtr<int> &indices) {
  if (max(n1,max(n1,n2))<=1) return;
  int nl=n0+1;
  int np=(n1+1)*nl;
  switch (l) {
    case 0: {
      indices.allocate(n1-1);
      int ijk=nl;
      for (int j=1;j<n1;j++) {
        indices[j-1]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 1: {
      indices.allocate(n1-1);
      int ijk=nl+n0;
      for (int j=1;j<n1;j++) {
        indices[j-1]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 2: {
      indices.allocate(n0-1);
      int ijk=1;
      for (int i=1;i<n0;i++) {
        indices[i-1]=ijk++;
      }
      break;
    }
    case 3: {
      indices.allocate(n0-1);
      int ijk=n1*nl+1;
      for (int i=1;i<n0;i++) {
        indices[i-1]=ijk++;
      }
      break;
    }
    case 4: {
      indices.allocate(n1-1);
      int ijk=n2*np+nl;
      for (int j=1;j<n1;j++) {
        indices[j-1]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 5: {
      indices.allocate(n1-1);
      int ijk=n2*np+nl+n0;
      for (int j=1;j<n1;j++) {
        indices[j-1]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 6: {
      indices.allocate(n0-1);
      int ijk=n2*np+1;
      for (int i=1;i<n0;i++) {
        indices[i-1]=ijk++;
      }
      break;
    }
    case 7: {
      indices.allocate(n0-1);
      int ijk=n2*np+n1*nl+1;
      for (int i=1;i<n0;i++) {
        indices[i-1]=ijk++;
      }
      break;
    }
    case 8: {
      indices.allocate(n2-1);
      int ijk=np;
      for (int k=1;k<n2;k++) {
        indices[k-1]=ijk;
        ijk+=np;
      }
      break;
    }
    case 9: {
      indices.allocate(n2-1);
      int ijk=np+n0;
      for (int k=1;k<n2;k++) {
        indices[k-1]=ijk;
        ijk+=np;
      }
      break;
    }
    case 10: {
      indices.allocate(n2-1);
      int ijk=np+n1*nl;
      for (int k=1;k<n2;k++) {
        indices[k-1]=ijk;
        ijk+=np;
      }
      break;
    }
    case 11: {
      indices.allocate(n2-1);
      int ijk=2*np-1;
      for (int k=1;k<n2;k++) {
        indices[k-1]=ijk;
        ijk+=np;
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::equallySpacedLatticePointIndicesOfLineInterior(int l,
int n0,int n1,int n2,NumPtr<int> &indices) const {
  if (max(n0,max(n1,n2))<=1) return;
  int nl=n0+1;
  int np=(n1+1)*nl;
  switch (l) {
    case 0: {
      indices.allocate(n1-1);
      int ijk=nl;
      for (int j=1;j<n1;j++) {
        indices[j-1]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 1: {
      indices.allocate(n1-1);
      int ijk=nl+n0;
      for (int j=1;j<n1;j++) {
        indices[j-1]=ijk;
        ijk+=nl;
      }
      break;
    }
    case 2: {
      indices.allocate(n0-1);
      int ijk=1;
      for (int i=1;i<n0;i++) {
        indices[i-1]=ijk;
        ijk++;
      }
      break;
    }
    case 3: {
      indices.allocate(n0-1);
      if (vertexIndex(2)<vertexIndex(3)) {
        int ijk=n1*nl+1;
        for (int i=1;i<n0;i++) {
          indices[i-1]=ijk;
          ijk++;
        }
      } else {
        int ijk=np-2;
        int m=0;
        for (int i=n0-1;i>=1;i--) {
          indices[m++]=ijk;
          ijk--;
        }
      }
      break;
    }
    case 4: {
      indices.allocate(n1-1);
      if (vertexIndex(4)<vertexIndex(6)) {
        int ijk=n2*np+nl;
        for (int j=1;j<n1;j++) {
          indices[j-1]=ijk;
          ijk+=nl;
        }
      } else {
        int ijk=n2*np+(n1-1)*nl;
        int m=0;
        for (int j=n1-1;j>=1;j--) {
          indices[m++]=ijk;
          ijk-=nl;
        }
      }
      break;
    }
    case 5: {
      indices.allocate(n1-1);
      if (vertexIndex(5)<vertexIndex(7)) {
        int ijk=n2*np+nl+n0;
        for (int j=1;j<n1;j++) {
          indices[j-1]=ijk;
          ijk+=nl;
        }
      } else {
        int ijk=n2*np+n1*nl-1;
        int m=0;
        for (int j=n1-1;j>=1;j--) {
          indices[m++]=ijk;
          ijk-=nl;
        }
      }
      break;
    }
    case 6: {
      indices.allocate(n0-1);
      if (vertexIndex(4)<vertexIndex(5)) {
        int ijk=n2*np+1;
        for (int i=1;i<n0;i++) {
          indices[i-1]=ijk;
          ijk++;
        }
      } else {
        int ijk=n2*np+n0-1;
        int m=0;
        for (int i=n0-1;i>=1;i--) {
          indices[m++]=ijk;
          ijk--;
        }
      }
      break;
    }
    case 7: {
      indices.allocate(n0-1);
      if (vertexIndex(6)<vertexIndex(7)) {
        int ijk=n2*np+n1*nl+1;
        for (int i=1;i<n0;i++) {
          indices[i-1]=ijk;
          ijk++;
        }
      } else {
        int ijk=n2*np+n1*nl+n0-1;
        int m=0;
        for (int i=n0-1;i>=1;i--) {
          indices[m++]=ijk;
          ijk--;
        }
      }
      break;
    }
    case 8: {
      indices.allocate(n2-1);
      int ijk=np;
      for (int k=1;k<n2;k++) {
        indices[k-1]=ijk;
        ijk+=np;
      }
      break;
    }
    case 9: {
      indices.allocate(n2-1);
      int ijk=np+n0;
      for (int k=1;k<n2;k++) {
        indices[k-1]=ijk;
        ijk+=np;
      }
      break;
    }
    case 10: {
      indices.allocate(n2-1);
      if (vertexIndex(2)<vertexIndex(6)) {
        int ijk=np+n1*nl;
        for (int k=1;k<n2;k++) {
          indices[k-1]=ijk;
          ijk+=np;
        }
      } else {
        int ijk=n2*np-nl;
        int m=0;
        for (int k=n2-1;k>=1;k--) {
          indices[m++]=ijk;
          ijk-=np;
        }
      }
      break;
    }
    case 11: {
      indices.allocate(n2-1);
      if (vertexIndex(3)<vertexIndex(7)) {
        int ijk=2*np-1;
        for (int k=1;k<n2;k++) {
          indices[k-1]=ijk;
          ijk+=np;
        }
      } else {
        int ijk=n2*np-1;
        int m=0;
        for (int k=n2-1;k>=1;k--) {
          indices[m++]=ijk;
          ijk-=np;
        }
      }
      break;
    }
    default:
      OBSOLETE("bad line");
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::reorderLineDofs(int n,NumPtr<int> &renumber) const {
  int nl=n+1;
  for (int i=0;i<12*nl;i++) renumber[i]=i;
  int offset=0;
  NumPtr<int> r=renumber;
  for (int l=0;l<12;l++) {
    int v[2]={vertexIndex(elementVertexFromLine(l,0)),
              vertexIndex(elementVertexFromLine(l,1))};
    Interval line_element(v,-1);
    line_element.reorderDofs(n,r);
    for (int i=0;i<nl;i++) r[i]+=offset;
    r+=nl;
    offset+=nl;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<12*nl;i++) {
    if (renumber[i]>=12*nl || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(12*nl);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<12*nl;i++) renumber_inverse[renumber[i]]=i;
  for (int i=0;i<12*nl;i++) CHECK_BOUNDS(renumber_inverse[i],0,12*nl);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::reorderHdivFaceDofs(int line_offset,int n,
NumPtr<int> &renumber) const {
  int nq=(n+1)*(n+1);
  for (int i=0;i<6*nq;i++) renumber[i]=i;
  int offset=line_offset;
  NumPtr<int> r=renumber;
  int l[4]={0,1,2,3};
  for (int f=0;f<6;f++) {
    int v[4]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2)),
              vertexIndex(elementVertexFromFace(f,3))};
    Quadrilateral face_element(l,v,-1);
    face_element.reorderDofs(n,n,r);
    for (int i=0;i<nq;i++) r[i]+=offset;
    r+=nq;
    offset+=nq;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<6*nq;i++) {
    if (renumber[i]-line_offset>=6*nq || renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(6*nq);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<6*nq;i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<6*nq;i++) CHECK_BOUNDS(renumber_inverse[i],0,6*nq);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::reorderHcurlFaceDofs(int line_offset,int n0,int n1,
NumPtr<int> &renumber) const {
  int nq=(n0+1)*(n1+1);
  for (int i=0;i<12*nq;i++) renumber[i]=i;
  int offset=line_offset;
  NumPtr<int> r=renumber;
  int l[4]={0,1,2,3};
  for (int f=0;f<6;f++) {
    int v[4]={vertexIndex(elementVertexFromFace(f,0)),
              vertexIndex(elementVertexFromFace(f,1)),
              vertexIndex(elementVertexFromFace(f,2)),
              vertexIndex(elementVertexFromFace(f,3))};
    Quadrilateral face_element(l,v,-1);
    if (!faceTangentsWindBackward(f)) {
      face_element.reorderDofs(n0,n1,r);
      for (int i=0;i<nq;i++) r[i]+=offset;
      r+=nq;
      offset+=nq;
      face_element.reorderDofs(n1,n0,r);
      for (int i=0;i<nq;i++) r[i]+=offset;
      r+=nq;
    } else {
      face_element.reorderDofs(n1,n0,r);
      for (int i=0;i<nq;i++) r[i]+=offset;
      r+=nq;
      offset+=nq;
      face_element.reorderDofs(n0,n1,r);
      for (int i=0;i<nq;i++) r[i]+=offset;
      r+=nq;
    }
    offset+=nq;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<12*nq;i++) {
    if (renumber[i]-line_offset>=12*nq || renumber[i]<line_offset) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(12*nq);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<12*nq;i++) {
    renumber_inverse[renumber[i]-line_offset]=i;
  }
  for (int i=0;i<12*nq;i++) CHECK_BOUNDS(renumber_inverse[i],0,12*nq);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::reorderEquallySpacedLatticePoints(int n0,int n1,int n2,
NumPtr<int> &renumber) const {
  renumber.allocate(staticNumberEquallySpacedLatticePoints(n0,n1,n2));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  int nl=n0+1;
  int np=nl*(n1+1);
//vertices first
  int m=0;
  renumber[0]=m++;
  renumber[n0]=m++;
  renumber[n1*nl]=m++;
  renumber[np-1]=m++;
  renumber[n2*np]=m++;
  renumber[n2*np+n0]=m++;
  renumber[renumber.getNumber()-nl]=m++;
  renumber[renumber.getNumber()-1]=m++;
//line 0
  if (vertexIndex(0)<vertexIndex(2)) {
    int l=nl;
    for (int j=1;j<n1;j++) {
      renumber[l]=m++;
      l+=nl;
    }
  } else {
    int l=(n1-1)*nl;
    for (int j=n1-1;j>0;j--) {
      renumber[l]=m++;
      l-=nl;
    }
  } 
//line 1
  if (vertexIndex(1)<vertexIndex(3)) {
    int l=2*nl-1;
    for (int j=1;j<n1;j++) {
      renumber[l]=m++;
      l+=nl;
    }
  } else {
    int l=n1*nl-1;
    for (int j=n1-1;j>0;j--) {
      renumber[l]=m++;
      l-=nl;
    }
  } 
//line 2
  if (vertexIndex(0)<vertexIndex(1)) {
    for (int i=1;i<n0;i++) renumber[i]=m++;
  } else {
    for (int i=n0-1;i>0;i--) renumber[i]=m++;
  } 
//line 3
  if (vertexIndex(2)<vertexIndex(3)) {
    int l=n1*nl+1;
    for (int i=1;i<n0;i++) renumber[l++]=m++;
  } else {
    int l=np-2;
    for (int i=n0-1;i>0;i--) renumber[l--]=m++;
  } 
//line 4
  if (vertexIndex(4)<vertexIndex(6)) {
    int l=n2*np+nl;
    for (int j=1;j<n1;j++) {
      renumber[l]=m++;
      l+=nl;
    }
  } else {
    int l=n2*np+(n1-1)*nl;
    for (int j=n1-1;j>0;j--) {
      renumber[l]=m++;
      l-=nl;
    }
  } 
//line 5
  if (vertexIndex(5)<vertexIndex(7)) {
    int l=n2*np+2*nl-1;
    for (int j=1;j<n1;j++) {
      renumber[l]=m++;
      l+=nl;
    }
  } else {
    int l=renumber.getNumber()-nl-1;
    for (int j=n1-1;j>0;j--) {
      renumber[l]=m++;
      l-=nl;
    }
  } 
//line 6
  if (vertexIndex(4)<vertexIndex(5)) {
    int l=n2*np+1;
    for (int i=1;i<n0;i++) renumber[l++]=m++;
  } else {
    int l=n2*np+n0-1;
    for (int i=n0-1;i>0;i--) renumber[l--]=m++;
  } 
//line 7
  if (vertexIndex(6)<vertexIndex(7)) {
    int l=renumber.getNumber()-n0;
    for (int i=1;i<n0;i++) renumber[l++]=m++;
  } else {
    int l=renumber.getNumber()-2;
    for (int i=n0-1;i>0;i--) renumber[l--]=m++;
  } 
//line 8
  if (vertexIndex(0)<vertexIndex(4)) {
    int l=np;
    for (int k=1;k<n2;k++) {
      renumber[l]=m++;
      l+=np;
    }
  } else {
    int l=(n2-1)*np;
    for (int k=n2-1;k>0;k--) {
      renumber[l]=m++;
      l-=np;
    }
  } 
//line 9
  if (vertexIndex(1)<vertexIndex(5)) {
    int l=np+n0;
    for (int k=1;k<n2;k++) {
      renumber[l]=m++;
      l+=np;
    }
  } else {
    int l=(n2-1)*np+n0;
    for (int k=n2-1;k>0;k--) {
      renumber[l]=m++;
      l-=np;
    }
  } 
//line 10
  if (vertexIndex(2)<vertexIndex(6)) {
    int l=np+n1*nl;
    for (int k=1;k<n2;k++) {
      renumber[l]=m++;
      l+=np;
    }
  } else {
    int l=(n2-1)*np+n1*nl;
    for (int k=n2-1;k>0;k--) {
      renumber[l]=m++;
      l-=np;
    }
  } 
//line 11
  if (vertexIndex(3)<vertexIndex(7)) {
    int l=2*np-1;
    for (int k=1;k<n2;k++) {
      renumber[l]=m++;
      l+=np;
    }
  } else {
    int l=n2*np-1;
    for (int k=n2-1;k>0;k--) {
      renumber[l]=m++;
      l-=np;
    }
  } 
//face 0
  int v=min(min(vertexIndex(0),vertexIndex(2)),
            min(vertexIndex(4),vertexIndex(6)));
  if (vertexIndex(0)==v) {
    int l=np+nl;
    if (vertexIndex(2) < vertexIndex(4)) { // 0 < 2 < 4
      for (int k=1;k<n2;k++) {
        int i=l;
        for (int j=1;j<n1;j++) {
          renumber[i]=m++;
          i+=nl;
        }
        l+=np;
      }
    } else { // 0 < 4 < 2
      for (int j=1;j<n1;j++) {
        int i=l;
        for (int k=1;k<n2;k++) {
          renumber[i]=m++;
          i+=np;
        }
        l+=nl;
      }
    }
  } else if (vertexIndex(2)==v) {
    int l=np+(n1-1)*nl;
    if (vertexIndex(0)<vertexIndex(6)) { // 2 < 0 < 6
      for (int k=1;k<n2;k++) {
        int i=l;
        for (int j=n1-1;j>0;j--) {
          renumber[i]=m++;
          i-=nl;
        }
        l+=np;
      }
    } else { // 2 < 6 < 0
      for (int j=n1-1;j>0;j--) {
        int i=l;
        for (int k=1;k<n2;k++) {
          renumber[i]=m++;
          i+=np;
        }
        l-=nl;
      }
    }
  } else if (vertexIndex(4)==v) {
    int l=(n2-1)*np+nl;
    if (vertexIndex(0)<vertexIndex(6)) { // 4 < 0 < 6
      for (int j=1;j<n1;j++) {
        int i=l;
        for (int k=n2-1;k>0;k--) {
          renumber[i]=m++;
          i-=np;
        }
        l+=nl;
      }
    } else { // 4 < 6 < 0
      for (int k=n2-1;k>0;k--) {
        int i=l;
        for (int j=1;j<n1;j++) {
          renumber[i]=m++;
          i+=nl;
        }
        l-=np;
      }
    }
  } else {
    int l=(n2-1)*np+(n1-1)*nl;
    if (vertexIndex(0)<vertexIndex(6)) { // 6 < 2 < 4
      for (int j=n1-1;j>0;j--) {
        int i=l;
        for (int k=n2-1;k>0;k--) {
          renumber[i]=m++;
          i-=np;
        }
        l-=nl;
      }
    } else { // 6 < 4 < 0
      for (int k=n2-1;k>0;k--) {
        int i=l;
        for (int j=n1-1;j>0;j--) {
          renumber[i]=m++;
          i-=nl;
        }
        l-=np;
      }
    }
  }
//face 1
  v=min(min(vertexIndex(1),vertexIndex(3)),
        min(vertexIndex(5),vertexIndex(7)));
  if (vertexIndex(1)==v) {
    int l=np+2*nl-1;
    if (vertexIndex(3) < vertexIndex(5)) { // 1 < 3 < 5
      for (int k=1;k<n2;k++) {
        int i=l;
        for (int j=1;j<n1;j++) {
          renumber[i]=m++;
          i+=nl;
        }
        l+=np;
      }
    } else { // 1 < 5 < 3
      for (int j=1;j<n1;j++) {
        int i=l;
        for (int k=1;k<n2;k++) {
          renumber[i]=m++;
          i+=np;
        }
        l+=nl;
      }
    }
  } else if (vertexIndex(3)==v) {
    int l=np+n1*nl-1;
    if (vertexIndex(1)<vertexIndex(7)) { // 3 < 1 < 7
      for (int k=1;k<n2;k++) {
        int i=l;
        for (int j=n1-1;j>0;j--) {
          renumber[i]=m++;
          i-=nl;
        }
        l+=np;
      }
    } else { // 3 < 7 < 1
      for (int j=n1-1;j>0;j--) {
        int i=l;
        for (int k=1;k<n2;k++) {
          renumber[i]=m++;
          i+=np;
        }
        l-=nl;
      }
    }
  } else if (vertexIndex(5)==v) {
    int l=(n2-1)*np+2*nl-1;
    if (vertexIndex(1)<vertexIndex(7)) { // 5 < 1 < 7
      for (int j=1;j<n1;j++) {
        int i=l;
        for (int k=n2-1;k>0;k--) {
          renumber[i]=m++;
          i-=np;
        }
        l+=nl;
      }
    } else { // 5 < 7 < 1
      for (int k=n2-1;k>0;k--) {
        int i=l;
        for (int j=1;j<n1;j++) {
          renumber[i]=m++;
          i+=nl;
        }
        l-=np;
      }
    }
  } else {
    int l=(n2-1)*np+n1*nl-1;
    if (vertexIndex(3)<vertexIndex(5)) { // 7 < 3 < 5
      for (int j=n1-1;j>0;j--) {
        int i=l;
        for (int k=n2-1;k>0;k--) {
          renumber[i]=m++;
          i-=np;
        }
        l-=nl;
      }
    } else { // 7 < 5 < 3
      for (int k=n2-1;k>0;k--) {
        int i=l;
        for (int j=n1-1;j>0;j--) {
          renumber[i]=m++;
          i-=nl;
        }
        l-=np;
      }
    }
  }
//face 2
  v=min(min(vertexIndex(0),vertexIndex(1)),
        min(vertexIndex(4),vertexIndex(5)));
  if (vertexIndex(0)==v) {
    int l=np+1;
    if (vertexIndex(1) < vertexIndex(4)) { // 0 < 1 < 4
      for (int k=1;k<n2;k++) {
        int j=l;
        for (int i=1;i<n0;i++) {
          renumber[j++]=m++;
        }
        l+=np;
      }
    } else { // 0 < 4 < 1
      for (int i=1;i<n0;i++) {
        int j=l;
        for (int k=1;k<n2;k++) {
          renumber[j]=m++;
          j+=np;
        }
        l++;
      }
    }
  } else if (vertexIndex(1)==v) {
    int l=np+n0-1;
    if (vertexIndex(0)<vertexIndex(5)) { // 1 < 0 < 5
      for (int k=1;k<n2;k++) {
        int j=l;
        for (int i=n0-1;i>0;i--) {
          renumber[j--]=m++;
        }
        l+=np;
      }
    } else { // 1 < 5 < 0
      for (int i=n0-1;i>0;i--) {
        int j=l;
        for (int k=1;k<n2;k++) {
          renumber[j]=m++;
          j+=np;
        }
        l--;
      }
    }
  } else if (vertexIndex(4)==v) {
    int l=(n2-1)*np+1;
    if (vertexIndex(0)<vertexIndex(5)) { // 4 < 0 < 5
      for (int i=1;i<n0;i++) {
        int j=l;
        for (int k=n2-1;k>0;k--) {
          renumber[j]=m++;
          j-=np;
        }
        l++;
      }
    } else { // 4 < 5 < 0
      for (int k=n2-1;k>0;k--) {
        int j=l;
        for (int i=1;i<n0;i++) {
          renumber[j++]=m++;
        }
        l-=np;
      }
    }
  } else {
    int l=(n2-1)*np+n0-1;
    if (vertexIndex(1)<vertexIndex(4)) { // 5 < 1 < 4
      for (int i=n0-1;i>0;i--) {
        int j=l;
        for (int k=n2-1;k>0;k--) {
          renumber[j]=m++;
          j-=np;
        }
        l--;
      }
    } else { // 5 < 4 < 1
      for (int k=n2-1;k>0;k--) {
        int j=l;
        for (int i=n0-1;i>0;i--) {
          renumber[j--]=m++;
        }
        l-=np;
      }
    }
  }
//face 3
  v=min(min(vertexIndex(2),vertexIndex(3)),
        min(vertexIndex(6),vertexIndex(7)));
  if (vertexIndex(2)==v) {
    int l=2*np-n0;
    if (vertexIndex(3) < vertexIndex(6)) { // 2 < 3 < 6
      for (int k=1;k<n2;k++) {
        int j=l;
        for (int i=1;i<n0;i++) {
          renumber[j++]=m++;
        }
        l+=np;
      }
    } else { // 2 < 6 < 3
      for (int i=1;i<n0;i++) {
        int j=l;
        for (int k=1;k<n2;k++) {
          renumber[j]=m++;
          j+=np;
        }
        l++;
      }
    }
  } else if (vertexIndex(3)==v) {
    int l=2*np-2;
    if (vertexIndex(2)<vertexIndex(7)) { // 3 < 2 < 7
      for (int k=1;k<n2;k++) {
        int j=l;
        for (int i=n0-1;i>0;i--) {
          renumber[j--]=m++;
        }
        l+=np;
      }
    } else { // 3 < 7 < 2
      for (int i=n0-1;i>0;i--) {
        int j=l;
        for (int k=1;k<n2;k++) {
          renumber[j]=m++;
          j+=np;
        }
        l--;
      }
    }
  } else if (vertexIndex(6)==v) {
    int l=n2*np-n0;
    if (vertexIndex(2)<vertexIndex(7)) { // 6 < 2 < 7
      for (int i=1;i<n0;i++) {
        int j=l;
        for (int k=n2-1;k>0;k--) {
          renumber[j]=m++;
          j-=np;
        }
        l++;
      }
    } else { // 6 < 7 < 2
      for (int k=n2-1;k>0;k--) {
        int j=l;
        for (int i=1;i<n0;i++) {
          renumber[j++]=m++;
        }
        l-=np;
      }
    }
  } else {
    int l=n2*np-2;
    if (vertexIndex(3)<vertexIndex(6)) { // 7 < 3 < 6
      for (int i=n0-1;i>0;i--) {
        int j=l;
        for (int k=n2-1;k>0;k--) {
          renumber[j]=m++;
          j-=np;
        }
        l--;
      }
    } else { // 7 < 6 < 3
      for (int k=n2-1;k>0;k--) {
        int j=l;
        for (int i=n0-1;i>0;i--) {
          renumber[j--]=m++;
        }
        l-=np;
      }
    }
  }
//face 4
  v=min(min(vertexIndex(0),vertexIndex(1)),
        min(vertexIndex(2),vertexIndex(3)));
  if (vertexIndex(0)==v) {
    int l=nl+1;
    if (vertexIndex(1) < vertexIndex(2)) { // 0 < 1 < 2
      for (int j=1;j<n1;j++) {
        int k=l;
        for (int i=1;i<n0;i++) {
          renumber[k++]=m++;
        }
        l+=nl;
      }
    } else { // 0 < 2 < 1
      for (int i=1;i<n0;i++) {
        int k=l;
        for (int j=1;j<n1;j++) {
          renumber[k]=m++;
          k+=nl;
        }
        l++;
      }
    }
  } else if (vertexIndex(1)==v) {
    int l=2*n0;
    if (vertexIndex(0)<vertexIndex(3)) { // 1 < 0 < 3
      for (int j=1;j<n1;j++) {
        int k=l;
        for (int i=n0-1;i>0;i--) {
          renumber[k--]=m++;
        }
        l+=nl;
      }
    } else { // 1 < 3 < 0
      for (int i=n0-1;i>0;i--) {
        int k=l;
        for (int j=1;j<n1;j++) {
          renumber[k]=m++;
          k+=nl;
        }
        l--;
      }
    }
  } else if (vertexIndex(2)==v) {
    int l=n1*nl+1;
    if (vertexIndex(0)<vertexIndex(3)) { // 2 < 0 < 3
      for (int i=1;i<n0;i++) {
        int k=l;
        for (int j=n1-1;j>0;j--) {
          renumber[k]=m++;
          k-=nl;
        }
        l++;
      }
    } else { // 2 < 3 < 0
      for (int j=n1-1;j>0;j--) {
        int k=l;
        for (int i=1;i<n0;i++) {
          renumber[k++]=m++;
        }
        l-=nl;
      }
    }
  } else {
    int l=n1*nl-2;
    if (vertexIndex(1)<vertexIndex(2)) { // 3 < 1 < 2
      for (int i=n0-1;i>0;i--) {
        int k=l;
        for (int j=n1-1;j>0;j--) {
          renumber[k]=m++;
          k-=nl;
        }
        l--;
      }
    } else { // 3 < 2 < 1
      for (int j=n1-1;j>0;j--) {
        int k=l;
        for (int i=n0-1;i>0;i--) {
          renumber[k--]=m++;
        }
        l-=nl;
      }
    }
  }
//face 5
  v=min(min(vertexIndex(4),vertexIndex(5)),
        min(vertexIndex(6),vertexIndex(7)));
  if (vertexIndex(4)==v) {
    int l=n2*np+nl+1;
    if (vertexIndex(5) < vertexIndex(6)) { // 4 < 5 < 6
      for (int j=1;j<n1;j++) {
        int k=l;
        for (int i=1;i<n0;i++) {
          renumber[k++]=m++;
        }
        l+=nl;
      }
    } else { // 4 < 6 < 5
      for (int i=1;i<n0;i++) {
        int k=l;
        for (int j=1;j<n1;j++) {
          renumber[k]=m++;
          k+=nl;
        }
        l++;
      }
    }
  } else if (vertexIndex(5)==v) {
    int l=n2*np+2*n0;
    if (vertexIndex(4)<vertexIndex(7)) { // 5 < 4 < 7
      for (int j=1;j<n1;j++) {
        int k=l;
        for (int i=n0-1;i>0;i--) {
          renumber[k--]=m++;
        }
        l+=nl;
      }
    } else { // 5 < 7 < 4
      for (int i=n0-1;i>0;i--) {
        int k=l;
        for (int j=1;j<n1;j++) {
          renumber[k]=m++;
          k+=nl;
        }
        l--;
      }
    }
  } else if (vertexIndex(6)==v) {
    int l=n2*np+(n1-1)*nl+1;
    if (vertexIndex(4)<vertexIndex(7)) { // 6 < 4 < 7
      for (int i=1;i<n0;i++) {
        int k=l;
        for (int j=n1-1;j>0;j--) {
          renumber[k]=m++;
          k-=nl;
        }
        l++;
      }
    } else { // 6 < 7 < 4
      for (int j=n1-1;j>0;j--) {
        int k=l;
        for (int i=1;i<n0;i++) {
          renumber[k++]=m++;
        }
        l-=nl;
      }
    }
  } else {
    int l=n2*np+n1*nl-2;
    if (vertexIndex(5)<vertexIndex(6)) { // 7 < 5 < 6
      for (int i=n0-1;i>0;i--) {
        int k=l;
        for (int j=n1-1;j>0;j--) {
          renumber[k]=m++;
          k-=nl;
        }
        l--;
      }
    } else { // 7 < 6 < 5
      for (int j=n1-1;j>0;j--) {
        int k=l;
        for (int i=n0-1;i>0;i--) {
          renumber[k--]=m++;
        }
        l-=nl;
      }
    }
  }
//interior points last
  int l2=np+nl+1;
  for (int k=1;k<n2;k++) {
    int l1=l2;
    for (int j=1;j<n1;j++) {
      int l=l1;
      for (int i=1;i<n0;i++) renumber[l++]=m++;
      l1+=nl;
    }
    l2+=np;
  }
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]==INT_MAX) {
      cout << "\trenumber[" << i << "] = *** " << -1 << " ***" << endl;
      success=false;
    } else if (renumber[i]>=renumber.getNumber()) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::reorderChildEquallySpacedLatticePoints(int n,
const Shape *child_shape,int c,
const NumPtr<int> &line_interior_child_vertices,NumPtr<int> &renumber) 
const {
  const Hexahedron *child_hexahedron=
    dynamic_cast<const Hexahedron*>(child_shape);
  CHECK_POINTER(child_hexahedron);

  int reorder_case=-1;
  int svi=sharedVertexIndex(child_shape);
  CHECK_TEST(svi>=0);
  CHECK_TEST(vertexIndex(c)==svi);
  int cvi0=child_shape->vertexEqualToVertexIndex(svi);
  CHECK_TEST(cvi0>=0);
  int cvi1=-1;
  switch (c) {
    case 0:
      cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[2]);
      break;
    case 1:
      cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[1]);
      break;
    case 2:
      cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[0]);
      break;
    case 3:
      cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[3]);
      break;
    case 4:
      cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[4]);
      break;
    case 5:
      cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[6]);
      break;
    case 6:
      cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[7]);
      break;
    case 7:
      cvi1=child_shape->
        vertexEqualToVertexIndex(line_interior_child_vertices[5]);
      break;
  }
  switch (cvi0) {
    case 0: {
      switch (cvi1) {
        case 1:
          reorder_case=0;
          break;
        case 2:
          reorder_case=1;
          break;
        case 4:
          reorder_case=2;
          break;
      }
      break;
    }
    case 1: {
      switch (cvi1) {
        case 0:
          reorder_case=3;
          break;
        case 3:
          reorder_case=4;
          break;
        case 5:
          reorder_case=5;
          break;
      }
      break;
    }
    case 2: {
      switch (cvi1) {
        case 0:
          reorder_case=6;
          break;
        case 3:
          reorder_case=7;
          break;
        case 6:
          reorder_case=8;
          break;
      }
      break;
    }
    case 3: {
      switch (cvi1) {
        case 1:
          reorder_case=9;
          break;
        case 2:
          reorder_case=10;
          break;
        case 7:
          reorder_case=11;
          break;
      }
      break;
    }
    case 4: {
      switch (cvi1) {
        case 0:
          reorder_case=12;
          break;
        case 5:
          reorder_case=13;
          break;
        case 6:
          reorder_case=14;
          break;
      }
      break;
    }
    case 5: {
      switch (cvi1) {
        case 1:
          reorder_case=15;
          break;
        case 4:
          reorder_case=16;
          break;
        case 7:
          reorder_case=17;
          break;
      }
      break;
    }
    case 6: {
      switch (cvi1) {
        case 2:
          reorder_case=18;
          break;
        case 4:
          reorder_case=19;
          break;
        case 7:
          reorder_case=20;
          break;
      }
      break;
    }
    case 7: {
      switch (cvi1) {
        case 3:
          reorder_case=21;
          break;
        case 5:
          reorder_case=22;
          break;
        case 6:
          reorder_case=23;
          break;
      }
      break;
    }
  }
  renumber.allocate(numberEquallySpacedLatticePoints(n));
#ifdef INDEF
  renumber.initialize(INT_MAX);
#endif
  int nl=n+1;
  int np=nl*nl;
  switch (reorder_case) {
    case 0: {
      for (int i=0;i<renumber.getNumber();i++) renumber[i]=i;
      break;
    }
    case 1: {
      int ijk=0;
      int ni=0;
      for (int i=0;i<=n;i++) {
        int nk=ni;
        for (int k=0;k<=n;k++) {
          int nj=nk;
          for (int j=0;j<=n;j++) {
            renumber[nj]=ijk++;
            nj+=nl;
          }
          nk+=np;
        }
        ni++;
      }
      break;
    }
    case 2: {
      int ijk=0;
      int nj=0;
      for (int j=0;j<=n;j++) {
        int ni=nj;
        for (int i=0;i<=n;i++) {
          int nk=ni;
          for (int k=0;k<=n;k++) {
            renumber[nk]=ijk++;
            nk+=np;
          }
          ni++;
        }
        nj+=nl;
      }
      break;
    }
    case 3: {
      int ijk=0;
      int nj=n;
      for (int j=0;j<=n;j++) {
        int nk=nj;
        for (int k=0;k<=n;k++) {
          int ni=nk;
          for (int i=n;i>=0;i--) {
            renumber[ni]=ijk++;
            ni--;
          }
          nk+=np;
        }
        nj+=nl;
      }
      break;
    }
    case 4: {
      int ijk=0;
      int nk=n;
      for (int k=0;k<=n;k++) {
        int ni=nk;
        for (int i=n;i>=0;i--) {
          int nj=ni;
          for (int j=0;j<=n;j++) {
            renumber[nj]=ijk++;
            nj+=nl;
          }
          ni--;
        }
        nk+=np;
      }
      break;
    }
    case 5: {
      int ijk=0;
      int ni=n;
      for (int i=n;i>=0;i--) {
        int nj=ni;
        for (int j=0;j<=n;j++) {
          int nk=nj;
          for (int k=0;k<=n;k++) {
            renumber[nk]=ijk++;
            nk+=np;
          }
          nj+=nl;
        }
        ni--;
      }
      break;
    }
    case 6: {
      int ijk=0;
      int nk=n*nl;
      for (int k=0;k<=n;k++) {
        int ni=nk;
        for (int i=0;i<=n;i++) {
          int nj=ni;
          for (int j=n;j>=0;j--) {
            renumber[nj]=ijk++;
            nj-=nl;
          }
          ni++;
        }
        nk+=np;;
      }
      break;
    }
    case 7: {
      int ijk=0;
      int nj=n*nl;
      for (int j=n;j>=0;j--) {
        int nk=nj;
        for (int k=0;k<=n;k++) {
          int ni=nk;
          for (int i=0;i<=n;i++) {
            renumber[ni]=ijk++;
            ni++;
          }
          nk+=np;
        }
        nj-=nl;
      }
      break;
    }
    case 8: {
      int ijk=0;
      int ni=n*nl;
      for (int i=0;i<=n;i++) {
        int nj=ni;
        for (int j=n;j>=0;j--) {
          int nk=nj;
          for (int k=0;k<=n;k++) {
            renumber[nk]=ijk++;
            nk+=np;
          }
          nj-=nl;
        }
        ni++;
      }
      break;
    }
    case 9: {
      int ijk=0;
      int ni=np-1;
      for (int i=n;i>=0;i--) {
        int nk=ni;
        for (int k=0;k<=n;k++) {
          int nj=nk;
          for (int j=n;j>=0;j--) {
            renumber[nj]=ijk++;
            nj-=nl;
          }
          nk+=np;
        }
        ni--;
      }
      break;
    }
    case 10: {
      int ijk=0;
      int nk=np-1;
      for (int k=0;k<=n;k++) {
        int nj=nk;
        for (int j=n;j>=0;j--) {
          int ni=nj;
          for (int i=n;i>=0;i--) {
            renumber[ni]=ijk++;
            ni--;
          }
          nj-=nl;
        }
        nk+=np;
      }
      break;
    }
    case 11: {
      int ijk=0;
      int nj=np-1;
      for (int j=n;j>=0;j--) {
        int ni=nj;
        for (int i=n;i>=0;i--) {
          int nk=ni;
          for (int k=0;k<=n;k++) {
            renumber[nk]=ijk++;
            nk+=np;
          }
          ni--;
        }
        nj-=nl;
      }
      break;
    }
    case 12: {
      int ijk=0;
      int ni=n*np;
      for (int i=0;i<=n;i++) {
        int nj=ni;
        for (int j=0;j<=n;j++) {
          int nk=nj;
          for (int k=n;k>=0;k--) {
            renumber[nk]=ijk++;
            nk-=np;
          }
          nj+=nl;
        }
        ni++;
      }
      break;
    }
    case 13: {
      int ijk=0;
      int nj=n*np;
      for (int j=0;j<=n;j++) {
        int nk=nj;
        for (int k=n;k>=0;k--) {
          int ni=nk;
          for (int i=0;i<=n;i++) {
            renumber[ni]=ijk++;
            ni++;
          }
          nk-=np;
        }
        nj+=nl;
      }
      break;
    }
    case 14: {
      int ijk=0;
      int nk=n*np;
      for (int k=n;k>=0;k--) {
        int ni=nk;
        for (int i=0;i<=n;i++) {
          int nj=ni;
          for (int j=0;j<=n;j++) {
            renumber[nj]=ijk++;
            nj+=nl;
          }
          ni++;
        }
        nk-=np;
      }
      break;
    }
    case 15: {
      int ijk=0;
      int nj=n*(np+1);
      for (int j=0;j<=n;j++) {
        int ni=nj;
        for (int i=n;i>=0;i--) {
          int nk=ni;
          for (int k=n;k>=0;k--) {
            renumber[nk]=ijk++;
            nk-=np;
          }
          ni--;
        }
        nj+=nl;
      }
      break;
    }
    case 16: {
      int ijk=0;
      int nk=n*(np+1);
      for (int k=n;k>=0;k--) {
        int nj=nk;
        for (int j=0;j<=n;j++) {
          int ni=nj;
          for (int i=n;i>=0;i--) {
            renumber[ni]=ijk++;
            ni--;
          }
          nj+=nl;
        }
        nk-=np;
      }
      break;
    }
    case 17: {
      int ijk=0;
      int ni=n*(np+1);
      for (int i=n;i>=0;i--) {
        int nk=ni;
        for (int k=n;k>=0;k--) {
          int nj=nk;
          for (int j=0;j<=n;j++) {
            renumber[nj]=ijk++;
            nj+=nl;
          }
          nk-=np;
        }
        ni--;
      }
      break;
    }
    case 18: {
      int ijk=0;
      int nj=n*(np+nl);
      for (int j=n;j>=0;j--) {
        int ni=nj;
        for (int i=0;i<=n;i++) {
          int nk=ni;
          for (int k=n;k>=0;k--) {
            renumber[nk]=ijk++;
            nk-=np;
          }
          ni++;
        }
        nj-=nl;
      }
      break;
    }
    case 19: {
      int ijk=0;
      int ni=n*(np+nl);
      for (int i=0;i<=n;i++) {
        int nk=ni;
        for (int k=n;k>=0;k--) {
          int nj=nk;
          for (int j=n;j>=0;j--) {
            renumber[nj]=ijk++;
            nj-=nl;
          }
          nk-=np;
        }
        ni++;
      }
      break;
    }
    case 20: {
      int ijk=0;
      int nk=n*(np+nl);
      for (int k=n;k>=0;k--) {
        int nj=nk;
        for (int j=n;j>=0;j--) {
          int ni=nj;
          for (int i=0;i<=n;i++) {
            renumber[ni]=ijk++;
            ni++;
          }
          nj-=nl;
        }
        nk-=np;
      }
      break;
    }
    case 21: {
      int ijk=0;
      int ni=nl*np-1;
      for (int i=n;i>=0;i--) {
        int nj=ni;
        for (int j=n;j>=0;j--) {
          int nk=nj;
          for (int k=n;k>=0;k--) {
            renumber[nk]=ijk++;
            nk-=np;
          }
          nj-=nl;
        }
        ni--;
      }
      break;
    }
    case 22: {
      int ijk=0;
      int nk=nl*np-1;
      for (int k=n;k>=0;k--) {
        int ni=nk;
        for (int i=n;i>=0;i--) {
          int nj=ni;
          for (int j=n;j>=0;j--) {
            renumber[nj]=ijk++;
            nj-=nl;
          }
          ni--;
        }
        nk-=np;
      }
      break;
    }
    case 23: {
      int ijk=0;
      int nj=nl*np-1;
      for (int j=n;j>=0;j--) {
        int nk=nj;
        for (int k=n;k>=0;k--) {
          int ni=nk;
          for (int i=n;i>=0;i--) {
            renumber[ni]=ijk++;
            ni--;
          }
          nk-=np;
        }
        nj-=nl;
      }
      break;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called from FE_H::constraints
void Hexahedron::reorderChildDofPoints(int n,const Shape *child_shape,
int c,const NumPtr<int> &line_interior_child_vertices,
NumPtr<int> &renumber) const {
  OBSOLETE("not programmed");
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::staticInteriorLatticePointIndices(int n0,int n1,int n2,
NumPtr<int> &v) {
  v.allocate(staticNumberInteriorEquallySpacedLatticePoints(n0,n1,n2));
  int nl=n0+1;
  int np=(n1+1)*nl;
  int ijk=np+nl+1;
  int iv=0;
  for (int k=1;k<n2;k++) {
    for (int j=1;j<n1;j++) {
      for (int i=1;i<n0;i++) {
        v[iv]=ijk;
        iv++;
        ijk++;
      }
      ijk+=2;
    }
    ijk+=2*nl;;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::computeInteriorLatticePoints(int d0,int d1,int d2,
int bf,NumPtr<Point<SPACEDIM> > &mapping_support_points) const {
  if (max(d0,max(d1,d2))<=1) return;
  CHECK_SAME(mapping_support_points.getNumber(),
             staticNumberEquallySpacedLatticePoints(d0,d1,d2));
  REAL dxi0=1./static_cast<REAL>(d0);
  REAL dxi1=1./static_cast<REAL>(d1);
  REAL dxi2=1./static_cast<REAL>(d2);
  const Point<SPACEDIM> &v0=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(0,d0,d1,d2)];
  const Point<SPACEDIM> &v1=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(1,d0,d1,d2)];
  const Point<SPACEDIM> &v2=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(2,d0,d1,d2)];
  const Point<SPACEDIM> &v3=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(3,d0,d1,d2)];
  const Point<SPACEDIM> &v4=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(4,d0,d1,d2)];
  const Point<SPACEDIM> &v5=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(5,d0,d1,d2)];
  const Point<SPACEDIM> &v6=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(6,d0,d1,d2)];
  const Point<SPACEDIM> &v7=mapping_support_points[
    staticEquallySpacedLatticePointIndexOfVertex(7,d0,d1,d2)];
  int nl=d0+1;
  int np=(d1+1)*nl;
  if (bf<0) { // no face on boundary, or all boundary faces are plane
    Point<SPACEDIM> dv40=v4-v0;
    Point<SPACEDIM> dv51=v5-v1;
    Point<SPACEDIM> dv62=v6-v2;
    Point<SPACEDIM> dv73=v7-v3;
    int ijk=np+nl+1;
    REAL xi2=dxi2;
    for (int k=1;k<d2;k++,ijk+=2*nl,xi2+=dxi2) {
      REAL xi1=dxi1;
      Point<SPACEDIM> v04=v0+dv40*xi2;
      Point<SPACEDIM> v15=v1+dv51*xi2;
      Point<SPACEDIM> v26=v2+dv62*xi2;
      Point<SPACEDIM> v37=v3+dv73*xi2;
      for (int j=1;j<d1;j++,ijk+=2,xi1+=dxi1) {
        REAL xi0=dxi0;
        Point<SPACEDIM> v0246=v04+(v26-v04)*xi1;
        Point<SPACEDIM> v1357=v15+(v37-v15)*xi1;
        for (int i=1;i<d0;i++,ijk++,xi0+=dxi0) {
          mapping_support_points[ijk]=v0246+(v1357-v0246)*xi0;
        }
      }
    }
  } else { // one or more faces on a non-plane boundary
    NumPtr<int> indices0;
    NumPtr<int> indices1;
    NumPtr<int> indices2;
    NumPtr<int> indices3;
    NumPtr<int> indices4;
    NumPtr<int> indices5;
    staticEquallySpacedLatticePointIndicesOfFace(0,d0,d1,d2,indices0);
    staticEquallySpacedLatticePointIndicesOfFace(1,d0,d1,d2,indices1);
    staticEquallySpacedLatticePointIndicesOfFace(2,d0,d1,d2,indices2);
    staticEquallySpacedLatticePointIndicesOfFace(3,d0,d1,d2,indices3);
    staticEquallySpacedLatticePointIndicesOfFace(4,d0,d1,d2,indices4);
    staticEquallySpacedLatticePointIndicesOfFace(5,d0,d1,d2,indices5);
    int ijk=np+nl+1;
    int jk=d1+2;
    int ik_save=d0+2;
    for (int k=1;k<d2;k++,jk+=2,ik_save+=d0+1,ijk+=2*nl) {
      REAL xi2=static_cast<REAL>(k)*dxi2;
      REAL onemxi2=1.-xi2;
      int ij=nl+1;
      for (int j=1;j<d1;j++,jk++,ij+=2,ijk+=2) {
        REAL xi1=static_cast<REAL>(j)*dxi1;
        REAL onemxi1=1.-xi1;
        int ik=ik_save;
        for (int i=1;i<d0;i++,ij++,ik++,ijk++) {
          REAL xi0=static_cast<REAL>(i)*dxi0;
          REAL onemxi0=1.-xi0;
          mapping_support_points[ijk]=
            (mapping_support_points[indices0[jk]]
            -((v0*onemxi1+v2*xi1)*onemxi2+(v4*onemxi1+v6*xi1)*xi2)*2.)
            *onemxi0
           +(mapping_support_points[indices1[jk]]
            -((v1*onemxi1+v3*xi1)*onemxi2+(v5*onemxi1+v7*xi1)*xi2)*2.)*xi0
           +mapping_support_points[indices2[ik]]*onemxi1
           +mapping_support_points[indices3[ik]]*xi1
           +mapping_support_points[indices4[ij]]*onemxi2
           +mapping_support_points[indices5[ij]]*xi2;
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::initializeMappingSupportPoints(int degree,
const NumPtr<Point<SPACEDIM> > &tessellation_vertices,
NumPtr<Point<SPACEDIM> > &points) const {
  NumPtr<Point<3> > eslp;
  staticEquallySpacedLatticePoints(degree,eslp);
  int neslp=eslp.getNumber();
  NumPtr<Point<SPACEDIM> > wv(8);
  for (int v=0;v<8;v++) {
    wv[v]=tessellation_vertices[vertexIndex(v)];
  }
  for (int i=0;i<neslp;i++) {
    const Point<3> &eslpi=eslp[i];
    points[i]=
      ((wv[1]*eslpi[0]+wv[0]*(1.-eslpi[0]))*(1.-eslpi[1])
      +(wv[3]*eslpi[0]+wv[2]*(1.-eslpi[0]))*eslpi[1])*(1.-eslpi[2])
     +((wv[5]*eslpi[0]+wv[4]*(1.-eslpi[0]))*(1.-eslpi[1])
      +(wv[7]*eslpi[0]+wv[6]*(1.-eslpi[0]))*eslpi[1])*eslpi[2];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Hexahedron::faceCenter(
const NumPtr<Point<SPACEDIM> > &vertices,int f) const {
  switch (f) {
    case 0:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(4)]
             +vertices[vertexIndex(2)]
             +vertices[vertexIndex(6)])*0.25;
    case 1:
      return (vertices[vertexIndex(1)]
             +vertices[vertexIndex(5)]
             +vertices[vertexIndex(3)]
             +vertices[vertexIndex(7)])*0.25;
    case 2:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(1)]
             +vertices[vertexIndex(4)]
             +vertices[vertexIndex(5)])*0.25;
    case 3:
      return (vertices[vertexIndex(2)]
             +vertices[vertexIndex(3)]
             +vertices[vertexIndex(6)]
             +vertices[vertexIndex(7)])*0.25;
    case 4:
      return (vertices[vertexIndex(0)]
             +vertices[vertexIndex(2)]
             +vertices[vertexIndex(1)]
             +vertices[vertexIndex(3)])*0.25;
    case 5:
      return (vertices[vertexIndex(4)]
             +vertices[vertexIndex(6)]
             +vertices[vertexIndex(5)]
             +vertices[vertexIndex(7)])*0.25;
    default:
      OBSOLETE("bad face");
      return Point<SPACEDIM>(D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL));
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Hexahedron::diameter(const NumPtr<Point<SPACEDIM> > &vertices) 
const {
  REAL d[4]={
    (vertices[vertexIndex(4)]-vertices[vertexIndex(3)]).square(),
    (vertices[vertexIndex(5)]-vertices[vertexIndex(2)]).square(),
    (vertices[vertexIndex(6)]-vertices[vertexIndex(1)]).square(),
    (vertices[vertexIndex(7)]-vertices[vertexIndex(0)]).square()
  };
  REAL dmax=0.;
  for (int i=0;i<4;i++) dmax=max(d[i],dmax);
  return sqrt(dmax);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL Hexahedron::volume(const NumPtr<Point<SPACEDIM> > &vertices) const
{
//may not be meaningful: vertices not necessarily co-planar in 3D
  Tensor<1,SPACEDIM> v10(vertices[vertexIndex(1)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v20(vertices[vertexIndex(2)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v40(vertices[vertexIndex(4)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v50(vertices[vertexIndex(5)]
                        -vertices[vertexIndex(0)]);
  Tensor<1,SPACEDIM> v60(vertices[vertexIndex(6)]
                        -vertices[vertexIndex(0)]);
//first prism:
  REAL vola,volb,volc,vold,vole,volf;
  Tensor<1,SPACEDIM> cp;
  crossProduct(cp,v20,v60);
  product(vola,v10,cp); // v0,v1,v2,v6
  crossProduct(cp,v60,v50);
  product(volb,v10,cp); // v0,v1,v6,v5
  crossProduct(cp,v60,v40);
  product(volc,v50,cp); // v0,v5,v6,v4
  Tensor<1,SPACEDIM> v17(vertices[vertexIndex(1)]
                        -vertices[vertexIndex(7)]);
  Tensor<1,SPACEDIM> v27(vertices[vertexIndex(2)]
                        -vertices[vertexIndex(7)]);
  Tensor<1,SPACEDIM> v37(vertices[vertexIndex(3)]
                        -vertices[vertexIndex(7)]);
  Tensor<1,SPACEDIM> v57(vertices[vertexIndex(5)]
                        -vertices[vertexIndex(7)]);
  Tensor<1,SPACEDIM> v67(vertices[vertexIndex(6)]
                        -vertices[vertexIndex(7)]);
//second prism:
  crossProduct(cp,v67,v17);
  product(vold,v57,cp); // v7,v5,v6,v1
  crossProduct(cp,v67,v27);
  product(vole,v17,cp); // v7,v1,v6,v2
  crossProduct(cp,v27,v37);
  product(volf,v17,cp); // v7,v1,v2,v3
  return (vola>0. && volb>0. && volc>0. && vold>0. && vole>0. && volf>0.
    ? (vola+volb+volc+vold+vole+volf)/6.
    : -(abs(vola)+abs(volb)+abs(volc)+abs(vold)+abs(vole)+abs(volf))/6.);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Hexahedron::lineTangentPointsBackward(int l) const {
  switch (l) {
    case 0:
      return false;
    case 1:
      return vertexIndex(3)<vertexIndex(1);
    case 2:
      return false;
    case 3:
      return vertexIndex(3)<vertexIndex(2);
    case 4:
      return vertexIndex(6)<vertexIndex(4);
    case 5:
      return vertexIndex(7)<vertexIndex(5);
    case 6:
      return vertexIndex(5)<vertexIndex(4);
    case 7:
      return vertexIndex(7)<vertexIndex(6);
    case 8:
      return false;
    case 9:
      return vertexIndex(5)<vertexIndex(1);
    case 10:
      return vertexIndex(6)<vertexIndex(2);
    case 11:
      return vertexIndex(7)<vertexIndex(3);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Hexahedron::printOn(ostream &os) const {
  os << "Hexahedron:" << endl;
  Shape3::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Hexahedron::checkFaceVertices(int f,const Shape2 *face_element) 
const {
  CHECK_BOUNDS(f,0,6);
  const Quadrilateral *quadrilateral=
    dynamic_cast<const Quadrilateral*>(face_element);
  CHECK_POINTER(quadrilateral);
  checkQuadrilateralVertices(
    vertexIndex(face_vertex_to_cell_vertex[f][0]),
    vertexIndex(face_vertex_to_cell_vertex[f][1]),
    vertexIndex(face_vertex_to_cell_vertex[f][2]),
    vertexIndex(face_vertex_to_cell_vertex[f][3]),
    quadrilateral);
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Hexahedron::draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE &xc,
const NumPtr<Point<3> > &vertices,VolGraphTool &vgt) const {
  Point<3> p0=vertices[vertexIndex(0)];
  Point<3> p1=vertices[vertexIndex(1)];
  Point<3> p2=vertices[vertexIndex(2)];
  Point<3> p3=vertices[vertexIndex(3)];
  Point<3> p4=vertices[vertexIndex(4)];
  Point<3> p5=vertices[vertexIndex(5)];
  Point<3> p6=vertices[vertexIndex(6)];
  Point<3> p7=vertices[vertexIndex(7)];
  Vector3 v0(p0[0],p0[1],p0[2]);
  Vector3 v1(p1[0],p1[1],p1[2]);
  Vector3 v2(p2[0],p2[1],p2[2]);
  Vector3 v3(p3[0],p3[1],p3[2]);
  Vector3 v4(p4[0],p4[1],p4[2]);
  Vector3 v5(p5[0],p5[1],p5[2]);
  Vector3 v6(p6[0],p6[1],p6[2]);
  Vector3 v7(p7[0],p7[1],p7[2]);
  vgt.colorTriangle(xc,v0,v4,v2);
  vgt.colorTriangle(xc,v6,v2,v4);
  vgt.colorTriangle(xc,v1,v3,v5);
  vgt.colorTriangle(xc,v7,v5,v3);
  vgt.colorTriangle(xc,v0,v1,v4);
  vgt.colorTriangle(xc,v5,v4,v1);
  vgt.colorTriangle(xc,v2,v6,v3);
  vgt.colorTriangle(xc,v7,v3,v6);
  vgt.colorTriangle(xc,v0,v2,v1);
  vgt.colorTriangle(xc,v3,v1,v2);
  vgt.colorTriangle(xc,v4,v5,v6);
  vgt.colorTriangle(xc,v7,v6,v5);
}
#endif
#endif

#include <NumPtr.C>
INSTANTIATE_NUMPTR(int)
#if (SPACEDIM>1)
INSTANTIATE_NUMPTR(Shape*);
INSTANTIATE_NUMPTR(const Shape*);
INSTANTIATE_NUMPTR(NumPtr<const Shape*>);
#endif

template __gnu_cxx::__normal_iterator<
  int*, 
  vector<int, allocator<int> > 
> set_intersection<
  _Rb_tree_const_iterator<int>, 
  _Rb_tree_const_iterator<int>, 
  __gnu_cxx::__normal_iterator<
    int*, 
    vector<int, allocator<int> > > 
>(
  _Rb_tree_const_iterator<int>, 
  _Rb_tree_const_iterator<int>, 
  _Rb_tree_const_iterator<int>, 
  _Rb_tree_const_iterator<int>, 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >
);

template class _Rb_tree<
  int, 
  int, 
  _Identity<int>, 
  less<int>, 
  allocator<int> 
>;

#if (SPACEDIM>1)
template class _Rb_tree<
  pair<int, int>, 
  pair<pair<int, int> const, int>,
  _Select1st<pair<pair<int, int> const, int> >,
  less<pair<int, int> >,
  allocator<pair<pair<int, int> const, int> >
>;
#endif
#if (SPACEDIM==3)
template class _Rb_tree<
  pair<pair<int, int>, pair<int, int> >,
  pair<pair<pair<int, int>, pair<int, int> > const, int>,
  _Select1st<pair<pair<pair<int, int>, pair<int, int> > const, int> >,
  less<pair<pair<int, int>, pair<int, int> > >,
  allocator<pair<pair<pair<int, int>, pair<int, int> > const, int> >
>;
#endif

template int* fill_n<int*, unsigned int, int>(
  int*, unsigned int, int const&);
