// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <TensorShapeFunction.H>
#include <Quadrature.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<int dim> void TensorShapeFunction<dim>::printOn(ostream &os)
const {
  os << "TensorShapeFunction: dim = " << dim << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofTensorShapeFunction::printOn(
ostream &os) const {
  os << "HdivDofTensorShapeFunction:" << endl;
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDualTensorShapeFunction::printOn(
ostream &os) const {
  os << "HdivDualTensorShapeFunction:" << endl;
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivPrismDofShapeFunction::printOn(ostream &os) const {
  os << "HdivPrismDofShapeFunction = " << endl;
  TensorShapeFunction<3>::printOn(os);
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTSimplexShapeFunction::RTSimplexShapeFunction(int degree) :
order(degree),n_polys(0) {
#if (SPACEDIM==1)
  simplex_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree+1);
#elif (SPACEDIM==2)
  simplex_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
#else
  simplex_polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
#endif
  n_polys=simplex_polys->getNumber()*SPACEDIM;
#if (SPACEDIM>=2)
# if (SPACEDIM==2)
  homo_polys=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
# elif (SPACEDIM==3)
  homo_polys=OPERATOR_NEW HomogeneousLagrangePolynomials3(degree);
# endif
  n_polys+=homo_polys->getNumber();
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void RTSimplexShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  simplex_polys->computeValues(p,pv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) v[iv++][d]=pv[s];
  }
#if (SPACEDIM>=2)
  NumPtr<REAL> hv;
  homo_polys->computeValues(p,hv);
  for (int s=0;s<homo_polys->getNumber();s++) {
    for (int d=0;d<SPACEDIM;d++) v[iv][d]=p[d]*hv[s];
    iv++;
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void RTSimplexShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > gv;
  simplex_polys->computeGrads(p,gv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) g[iv++][d]=gv[s];
  }
#if (SPACEDIM>=2)
  NumPtr<REAL> hv;
  NumPtr<Tensor<1,SPACEDIM> > hg;
  homo_polys->computeValues(p,hv);
  homo_polys->computeGrads(p,hg);
  for (int s=0;s<homo_polys->getNumber();s++) {
    Tensor<1,SPACEDIM> &hgs=hg[s];
    REAL &hvs=hv[s];
    for (int d=0;d<SPACEDIM;d++) {
      Tensor<1,SPACEDIM> &gvd=g[iv][d];
      for (int i=0;i<SPACEDIM;i++) gvd[i]=p[d]*hgs[i];
      gvd[d]+=hvs;
    }
    iv++;
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void RTSimplexShapeFunction::computeGradGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > ggv;
  simplex_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) gg[iv++][d]=ggv[s];
  }
#if (SPACEDIM>=2)
  NumPtr<Tensor<1,SPACEDIM> > hg;
  NumPtr<Tensor<2,SPACEDIM> > hgg;
  homo_polys->computeGrads(p,hg);
  homo_polys->computeGradGrads(p,hgg);
  for (int s=0;s<homo_polys->getNumber();s++) {
    Tensor<2,SPACEDIM> &hggs=hgg[s];
    Tensor<1,SPACEDIM> &hgs=hg[s];
    for (int d=0;d<SPACEDIM;d++) {
      Tensor<2,SPACEDIM> &ggvd=gg[iv][d];
      for (int j=0;j<SPACEDIM;j++) {
        for (int i=0;i<SPACEDIM;i++) {
          ggvd[i][j]=p[d]*hggs[i][j];
          if (i==d) ggvd[d][j]+=hgs[j];
          if (j==d) ggvd[i][d]+=hgs[i];
        }
      }
    }
    iv++;
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTSimplexShapeFunction::printOn(ostream &os) const {
  os << "RTSimplexShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tsimplex_polys:" << endl;
  simplex_polys->printOn(cout);
#if (SPACEDIM>=2)
  cout << "\thomo_polys:" << endl;
  homo_polys->printOn(cout);
#endif
  TensorShapeFunction<SPACEDIM>::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTSimplexDofShapeFunction::RTSimplexDofShapeFunction(int degree) :
order(degree),n_polys(0) {
#if (SPACEDIM==1)
  if (degree<1) {
    interior_polys=0;
    n_polys=0;
  } else {
    interior_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree-1);
    n_polys=interior_polys->getNumber();
  }
#elif (SPACEDIM==2)
  interior_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree-1);
  face_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_polys=SPACEDIM*interior_polys->getNumber();
#else
  face_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
  interior_polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree-1);
  n_polys=SPACEDIM*interior_polys->getNumber();
#endif
#if (SPACEDIM>=2)
  n_face_polys=face_polys->getNumber();
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void RTSimplexDofShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> pv;
    interior_polys->computeValues(p,pv);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<interior_polys->getNumber();s++) v[iv++][d]=pv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void RTSimplexDofShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,SPACEDIM> > gv;
    interior_polys->computeGrads(p,gv);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<interior_polys->getNumber();s++) g[iv++][d]=gv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void RTSimplexDofShapeFunction::computeGradGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,SPACEDIM> > ggv;
    interior_polys->computeGradGrads(p,ggv);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<interior_polys->getNumber();s++) gg[iv++][d]=ggv[s];
    }
  }
}
#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTSimplexDofShapeFunction::computeFaceValues(
const Point<SPACEDIM-1> &p,NumPtr<REAL > &v) const
{
  v.allocate(n_face_polys);
  v.initialize(0.);
  NumPtr<REAL> pv;
  face_polys->computeValues(p,pv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) v[iv++]=pv[s];
}
#endif
#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void RTSimplexDofShapeFunction::computeFaceGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<1,SPACEDIM-1> > &g) const {
  g.allocate(n_face_polys);
  Tensor<1,SPACEDIM-1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM-1> > gv;
  face_polys->computeGrads(p,gv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) g[iv++]=gv[s];
}
#endif
#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void RTSimplexDofShapeFunction::computeFaceGradGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<2,SPACEDIM-1> > &gg) const {
  gg.allocate(n_face_polys);
  Tensor<2,SPACEDIM-1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM-1> > ggv;
  face_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) gg[iv++]=ggv[s];
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//renumber[fe_s] = dof_s:
void RTSimplexDofShapeFunction::shapeFunctionsOrder(const Shape *e,
NumPtr<int> &renumber) const {
#if (SPACEDIM==1)
  renumber.allocate(n_polys+2);
#else
  renumber.allocate(n_polys+(SPACEDIM+1)*n_face_polys);
#endif
  for (int fe_s=0;fe_s<renumber.getNumber();fe_s++) renumber[fe_s]=fe_s;
#if (SPACEDIM==2)
  const Triangle *element=dynamic_cast<const Triangle*>(e);
  CHECK_POINTER(element);
  element->reorderFaceDofs(order,renumber);
#elif (SPACEDIM==3)
  const Tetrahedron *element=dynamic_cast<const Tetrahedron*>(e);
  CHECK_POINTER(element);
  element->reorderHdivFaceDofs(0,order,renumber);
#endif
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]>=renumber.getNumber() || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(n_polys);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<n_polys;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<n_polys;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,n_polys);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTSimplexDofShapeFunction::printOn(ostream &os) const {
  os << "RTSimplexDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
#if (SPACEDIM>=2)
     << "\n\tn_face_polys = " << n_face_polys 
#endif
     << endl;
#if (SPACEDIM>=2)
  cout << "\tface_polys:" << endl;
  face_polys->printOn(cout);
#endif
  cout << "\tinterior_polys:" << endl;
  interior_polys->printOn(cout);
  HdivDofTensorShapeFunction::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTSimplexDualShapeFunction::RTSimplexDualShapeFunction(int degree) :
dof_polys(degree),order(degree),n_polys(0) {
#if (SPACEDIM==1)
  simplex_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree+1);
#elif (SPACEDIM==2)
  simplex_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
#else
  simplex_polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
#endif
  int spn=simplex_polys->getNumber();
  n_polys=spn*SPACEDIM;
#if (SPACEDIM>=2)
# if (SPACEDIM==2)
  homo_polys=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
# elif (SPACEDIM==3)
  homo_polys=OPERATOR_NEW HomogeneousLagrangePolynomials3(degree);
# endif
  int hpn=homo_polys->getNumber();
  n_polys+=hpn;
#endif

  int dpn=dof_polys.getNumber();
#if (SPACEDIM>=2)
  int dpfn=dof_polys.getFaceNumber();
#endif
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

#if (SPACEDIM==1)
  CHECK_SAME(n_polys,2+dpn);
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+2));
#elif (SPACEDIM==2)
  TriangleGaussianQuadrature 
    gq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  GaussianQuadrature<1>
    fgq(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
#else
  TetrahedronGaussianQuadrature 
    gq(TetrahedronGaussianQuadrature::argGivenOrder(2*degree+2));
  TriangleGaussianQuadrature 
    fgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
#endif
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
#if (SPACEDIM>=2)
  CHECK_SAME(n_polys,dpn+(SPACEDIM+1)*dpfn);
  const NumPtr<Point<SPACEDIM-1> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
#endif

  int ioffset=0;
#if (SPACEDIM==1)
  Point<1> p; p[0]=0.;
  NumPtr<REAL> pv;
  simplex_polys->computeValues(p,pv);
  for (int j=0;j<n_polys;j++) matrix(0,j)=-pv[j]; // normal points left
  p[0]=1.;
  pv.cleanup();
  simplex_polys->computeValues(p,pv);
  for (int j=0;j<n_polys;j++) matrix(1,j)=pv[j];
  ioffset+=2;
#else
  NumPtr<Point<SPACEDIM> > q_points_face[SPACEDIM+1];
  for (int f=0;f<SPACEDIM+1;f++) {
    q_points_face[f].allocate(fgq_points.getNumber());
# if (SPACEDIM==2)
    Triangle::staticProjectToFace(f,fgq_points,q_points_face[f]);
# else
    Tetrahedron::staticProjectToFace(f,fgq_points,q_points_face[f]);
# endif
  }
  NumPtr<NumPtr<REAL> > dpfv(fgq_points.getNumber());
  for (int q=0;q<fgq_points.getNumber();q++) {
    dof_polys.computeFaceValues(fgq_points[q],dpfv[q]);
  }
  int hoffset=SPACEDIM*spn;;
  for (int f=0;f<SPACEDIM;f++) {
    int joffset=f*spn;
    for (int q=0;q<fgq_points.getNumber();q++) {
      REAL wq=fgq_weights[q];
      NumPtr<REAL> spv;
      simplex_polys->computeValues(q_points_face[f][q],spv);
      for (int j=0;j<spn;j++) {
        REAL spvj=spv[j];
        for (int i=0;i<dpfn;i++) {
//        normal is -e_f; simplex poly is e_f * spv
          matrix(i+ioffset,j+joffset)-=dpfv[q][i]*wq*spvj;
        }
      }
      NumPtr<REAL> hpv;
      homo_polys->computeValues(q_points_face[f][q],hpv);
      for (int j=0;j<hpn;j++) {
//      normal component is -e_f . qp * hpv
        REAL hpvj=-q_points_face[f][q][f]*hpv[j];
        for (int i=0;i<dpfn;i++) {
          matrix(i+ioffset,j+hoffset)+=dpfv[q][i]*wq*hpvj;
        }
      }
    }
    ioffset+=dpfn;
  }
  int f=SPACEDIM;
  for (int q=0;q<fgq_points.getNumber();q++) {
    REAL wq=fgq_weights[q];
    NumPtr<REAL> spv;
    simplex_polys->computeValues(q_points_face[f][q],spv);
    for (int j=0;j<spn;j++) {
      REAL spvj=spv[j]; // normal ~ sum e_j; simplex_poly = e_j * spv
      for (int i=0;i<dpfn;i++) {
        REAL term=dpfv[q][i]*wq*spvj;
        for (int d=0;d<SPACEDIM;d++) {
          matrix(i+ioffset,j+d*spn)+=term;
        }
      }
    }
    NumPtr<REAL> hpv;
    homo_polys->computeValues(q_points_face[f][q],hpv);
    for (int j=0;j<hpn;j++) {
      REAL hpvj=hpv[j]; // normal . xi = 1 / sqrt(SPACEDIM)
      for (int i=0;i<dpfn;i++) {
        matrix(i+ioffset,j+hoffset)+=dpfv[q][i]*wq*hpvj;
      }
    }
  }
  ioffset+=dpfn;
#endif
  NumPtr<NumPtr<Tensor<1,SPACEDIM> > > dpv(gq_points.getNumber());
  for (int q=0;q<gq_points.getNumber();q++) {
    dof_polys.computeValues(gq_points[q],dpv[q]);
  }
  for (int q=0;q<gq_points.getNumber();q++) {
    REAL wq=gq_weights[q];
    NumPtr<REAL> spv;
    const Point<SPACEDIM> &gqpq=gq_points[q];
    simplex_polys->computeValues(gqpq,spv);
    int joffset=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int j=0;j<spn;j++) {
        REAL spvj=spv[j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[q][i][d]*wq*spvj;
        }
      }
      joffset+=spn;
    }
#if (SPACEDIM>=2)
    NumPtr<REAL> hpv;
    homo_polys->computeValues(gqpq,hpv);
    for (int j=0;j<hpn;j++) {
      REAL hpvj=hpv[j];
      for (int i=0;i<dpn;i++) {
        REAL dot=0.;
        for (int d=0;d<SPACEDIM;d++) dot+=dpv[q][i][d]*gqpq[d];
        matrix(i+ioffset,j+joffset)+=dot*wq*hpvj;
      }
    }
#endif
  }
  soln=factorNSolve(matrix,rhs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void RTSimplexDualShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> spv;
  simplex_polys->computeValues(p,spv);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      REAL spvs=spv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=spvs*(*soln)(s+offset,iv);
      }
    }
    offset+=spn;
  }
#if (SPACEDIM>=2)
  NumPtr<REAL> hv;
  homo_polys->computeValues(p,hv);
  int hpn=homo_polys->getNumber();
  for (int s=0;s<hpn;s++) {
    REAL hvs=hv[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      for (int d=0;d<SPACEDIM;d++) {
        v[iv][d]+=p[d]*hvs*ssiv;
      }
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void RTSimplexDualShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > spg;
  simplex_polys->computeGrads(p,spg);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      Tensor<1,SPACEDIM> &spgs=spg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int dd=0;dd<SPACEDIM;dd++) {
          givd[dd]+=spgs[dd]*ssiv;
        }
      }
    }
    offset+=spn;
  }
#if (SPACEDIM>=2)
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  homo_polys->computeValues(p,hpv);
  homo_polys->computeGrads(p,hpg);
  int hpn=homo_polys->getNumber();
  for (int s=0;s<hpn;s++) {
    Tensor<1,SPACEDIM> &hpgs=hpg[s];
    REAL &hpvs=hpv[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      for (int d=0;d<SPACEDIM;d++) {
        Tensor<1,SPACEDIM> &gvd=g[iv][d];
        for (int i=0;i<SPACEDIM;i++) gvd[i]+=p[d]*hpgs[i]*ssiv;
        gvd[d]+=hpvs*ssiv;
      }
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void RTSimplexDualShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > spgg;
  simplex_polys->computeGradGrads(p,spgg);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      Tensor<2,SPACEDIM> &spggs=spgg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            ggivd[i][j]+=spggs[i][j]*ssiv;;
          }
        }
      }
    }
    offset+=spn;
  }
#if (SPACEDIM>=2)
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  NumPtr<Tensor<2,SPACEDIM> > hpgg;
  homo_polys->computeGrads(p,hpg);
  homo_polys->computeGradGrads(p,hpgg);
  int hpn=homo_polys->getNumber();
  for (int s=0;s<hpn;s++) {
    Tensor<2,SPACEDIM> &hpggs=hpgg[s];
    Tensor<1,SPACEDIM> &hpgs=hpg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      for (int d=0;d<SPACEDIM;d++) {
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            ggivd[i][j]+=p[d]*hpggs[i][j]*ssiv;
            if (i==d) ggivd[d][j]+=hpgs[j]*ssiv;
            if (j==d) ggivd[i][d]+=hpgs[i]*ssiv;
          }
        }
      }
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTSimplexDualShapeFunction::nonzeroShapesOnFace(const Shape1 *e,
int f,NumPtr<int> &indices) const {
#if (SPACEDIM==1)
  indices.allocate(1);
  indices[0]=f;
#elif (SPACEDIM==2)
  int nl=order+1;
  indices.allocate(nl);
  int offset=f*nl;
  for (int i=0;i<nl;i++) indices[i]=i+offset;
#else
  int nt=((order+1)*(order+2))/2;
  indices.allocate(nt);
  int offset=f*nt;
  for (int i=0;i<nt;i++) indices[i]=i+offset;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTSimplexDualShapeFunction::printOn(ostream &os) const {
  os << "RTSimplexDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tsimplex_polys:" << endl;
  simplex_polys->printOn(cout);
#if (SPACEDIM>=2)
  cout << "\thomo_polys:" << endl;
  homo_polys->printOn(cout);
#endif
  HdivDualTensorShapeFunction::printOn(os);
}

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTParallelpipedShapeFunction::RTParallelpipedShapeFunction(int degree) :
order(degree),n_polys(0) {
# if (SPACEDIM==2)
  polys[0]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree+1,degree);
  polys[1]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree+1);
# else
  polys[0]=
    OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree+1,degree,degree);
  polys[1]=
    OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree,degree+1,degree);
  polys[2]=
    OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree,degree,degree+1);
# endif
  for (int d=0;d<SPACEDIM;d++) n_polys+=polys[d]->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void RTParallelpipedShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<REAL> pv;
    polys[d]->computeValues(p,pv);
    for (int s=0;s<polys[d]->getNumber();s++) v[iv++][d]=pv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void RTParallelpipedShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<1,SPACEDIM> > gv;
    polys[d]->computeGrads(p,gv);
    for (int s=0;s<polys[d]->getNumber();s++) g[iv++][d]=gv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void RTParallelpipedShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<2,SPACEDIM> > ggv;
    polys[d]->computeGradGrads(p,ggv);
    for (int s=0;s<polys[d]->getNumber();s++) gg[iv++][d]=ggv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTParallelpipedShapeFunction::printOn(ostream &os) const {
  os << "RTParallelpipedShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  for (int d=0;d<SPACEDIM;d++) {
    cout << "\tpolys[" << d << "]:" << endl;
    polys[d]->printOn(cout);
  }
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTParallelpipedDofShapeFunction::RTParallelpipedDofShapeFunction(
int degree) : order(degree),n_polys(0),n_face_polys(0) {
# if (SPACEDIM==2)
  face_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_face_polys=face_polys->getNumber();
  if (degree<1) {
    interior_polys[0]=0;
    interior_polys[1]=0;
    n_polys=0;
  } else {
    interior_polys[0]=
      OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree-1,degree);
    interior_polys[1]=
      OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree-1);
    for (int d=0;d<SPACEDIM;d++) n_polys+=interior_polys[d]->getNumber();
  }
# else
  face_polys=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree);
  n_face_polys=face_polys->getNumber();
  if (degree<1) {
    interior_polys[0]=0;
    interior_polys[1]=0;
    interior_polys[2]=0;
    n_polys=0;
  } else {
    interior_polys[0]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree-1,degree,degree);
    interior_polys[1]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree,degree-1,degree);
    interior_polys[2]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree,degree,degree-1);
    for (int d=0;d<SPACEDIM;d++) n_polys+=interior_polys[d]->getNumber();
  }
# endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void RTParallelpipedDofShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<REAL> pv;
      interior_polys[d]->computeValues(p,pv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        v[iv++][d]=pv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void RTParallelpipedDofShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<Tensor<1,SPACEDIM> > gv;
      interior_polys[d]->computeGrads(p,gv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        g[iv++][d]=gv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void RTParallelpipedDofShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<Tensor<2,SPACEDIM> > ggv;
      interior_polys[d]->computeGradGrads(p,ggv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        gg[iv++][d]=ggv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTParallelpipedDofShapeFunction::computeFaceValues(
const Point<SPACEDIM-1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_face_polys);
  v.initialize(0.);
  NumPtr<REAL> pv;
  face_polys->computeValues(p,pv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) v[iv++]=pv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void RTParallelpipedDofShapeFunction::computeFaceGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<1,SPACEDIM-1> > &g) const {
  g.allocate(n_face_polys);
  Tensor<1,SPACEDIM-1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM-1> > gv;
  face_polys->computeGrads(p,gv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) g[iv++]=gv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void RTParallelpipedDofShapeFunction::computeFaceGradGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<2,SPACEDIM-1> > &gg) const {
  gg.allocate(n_face_polys);
  Tensor<2,SPACEDIM-1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM-1> > ggv;
  face_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) gg[iv++]=ggv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//renumber[fe_s] = dof_s:
void RTParallelpipedDofShapeFunction::shapeFunctionsOrder(
const Shape *e,NumPtr<int> &renumber) const {
  renumber.allocate(n_polys+2*SPACEDIM*n_face_polys);
  for (int fe_s=0;fe_s<renumber.getNumber();fe_s++) renumber[fe_s]=fe_s;
#if (SPACEDIM==2)
  const Quadrilateral *element=dynamic_cast<const Quadrilateral*>(e);
  CHECK_POINTER(element);
  element->reorderFaceDofs(order,renumber);
#elif (SPACEDIM==3)
  const Hexahedron *element=dynamic_cast<const Hexahedron*>(e);
  CHECK_POINTER(element);
  element->reorderHdivFaceDofs(0,order,renumber);
#endif
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]>=renumber.getNumber() || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(n_polys);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<n_polys;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<n_polys;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,n_polys);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTParallelpipedDofShapeFunction::printOn(ostream &os) const {
  os << "RTParallelpipedDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tn_face_polys = " << n_face_polys << endl;
  cout << "\tface_polys:" << endl;
  face_polys->printOn(cout);
  for (int d=0;d<SPACEDIM;d++) {
    cout << "\tinterior_polys[" << d << "]:" << endl;
    interior_polys[d]->printOn(cout);
  }
  HdivDofTensorShapeFunction::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTParallelpipedDualShapeFunction::RTParallelpipedDualShapeFunction(
int degree) : dof_polys(degree),order(degree),n_polys(0) {
# if (SPACEDIM==2)
  polys[0]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree+1,degree);
  polys[1]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree+1);
# else
  polys[0]=
    OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree+1,degree,degree);
  polys[1]=
    OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree,degree+1,degree);
  polys[2]=
    OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree,degree,degree+1);
# endif
  for (int d=0;d<SPACEDIM;d++) n_polys+=polys[d]->getNumber();

  int dpn=dof_polys.getNumber();
  int dpfn=dof_polys.getFaceNumber();
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();

  int ioffset=0;
  CHECK_SAME(n_polys,dpn+2*SPACEDIM*dpfn);
  GaussianQuadrature<SPACEDIM-1>
    fgq(GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM-1> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  NumPtr<Point<SPACEDIM> > q_points_face[2*SPACEDIM]; 
  for (int f=0;f<2*SPACEDIM;f++) {
    q_points_face[f].allocate(fgq_points.getNumber());
# if (SPACEDIM==2)
    Quadrilateral::staticProjectToFace(f,fgq_points,q_points_face[f]);
# else
    Hexahedron::staticProjectToFace(f,fgq_points,q_points_face[f]);
# endif
  }
  NumPtr<NumPtr<REAL> > dpfv(fgq_points.getNumber());
  for (int q=0;q<fgq_points.getNumber();q++) {
    dof_polys.computeFaceValues(fgq_points[q],dpfv[q]);
  }
  int joffset=0;
  for (int d=0;d<SPACEDIM;d++) {
    int pdn=polys[d]->getNumber();
    for (int f=2*d;f<2*d+2;f++) {
      for (int q=0;q<fgq_points.getNumber();q++) {
        REAL wq=fgq_weights[q];
        NumPtr<REAL> pv;
        polys[d]->computeValues(q_points_face[f][q],pv);
        for (int j=0;j<pdn;j++) {
          REAL pvj=pv[j];
          if (f%2==0) pvj=-pvj; // normal = - e_{f/2}
          for (int i=0;i<dpfn;i++) {
            matrix(i+ioffset,j+joffset)+=dpfv[q][i]*wq*pvj;
          }
        }
      }
      ioffset+=dpfn;
    }
    joffset+=pdn;
  }

  NumPtr<NumPtr<Tensor<1,SPACEDIM> > > dpv(gq_points.getNumber());
  for (int q=0;q<gq_points.getNumber();q++) {
    dof_polys.computeValues(gq_points[q],dpv[q]);
  }
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    int pdn=polys[d]->getNumber();
    for (int q=0;q<gq_points.getNumber();q++) {
      REAL wq=gq_weights[q];
      NumPtr<REAL> pv;
      polys[d]->computeValues(gq_points[q],pv);
      for (int j=0;j<pdn;j++) {
        REAL pvj=pv[j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+offset)+=dpv[q][i][d]*wq*pvj;
        }
      }
    }
    offset+=pdn;
  }
  soln=factorNSolve(matrix,rhs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void RTParallelpipedDualShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<REAL> pv;
    polys[d]->computeValues(p,pv);
    int pdn=polys[d]->getNumber();
    for (int s=0;s<pdn;s++) {
      REAL pvs=pv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=pvs*(*soln)(s+offset,iv);
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void RTParallelpipedDualShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<1,SPACEDIM> > gv;
    polys[d]->computeGrads(p,gv);
    int pdn=polys[d]->getNumber();
    for (int s=0;s<pdn;s++) {
      Tensor<1,SPACEDIM> &gvs=gv[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int i=0;i<SPACEDIM;i++) {
          givd[i]+=gvs[i]*ssiv;
        }
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void RTParallelpipedDualShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<2,SPACEDIM> > ggv;
    polys[d]->computeGradGrads(p,ggv);
    int pdn=polys[d]->getNumber();
    for (int s=0;s<pdn;s++) {
      Tensor<2,SPACEDIM> &ggvs=ggv[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int i=0;i<SPACEDIM;i++) {
          for (int j=0;j<SPACEDIM;j++) {
            ggivd[i][j]+=ggvs[i][j]*ssiv;
          }
        }
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTParallelpipedDualShapeFunction::nonzeroShapesOnFace(
const Shape1 *e,int f,NumPtr<int> &indices) const {
# if (SPACEDIM==2)
  int nl=order+1;
  indices.allocate(nl);
  int offset=f*nl;
  for (int i=0;i<nl;i++) indices[i]=i+offset;
# else
  int nq=(order+1)*(order+1);
  indices.allocate(nq);
  int offset=f*nq;
  for (int i=0;i<nq;i++) indices[i]=i+offset;
# endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTParallelpipedDualShapeFunction::printOn(ostream &os) const {
  os << "RTParallelpipedDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  for (int d=0;d<SPACEDIM;d++) {
    cout << "\tpolys[" << d << "]:" << endl;
    polys[d]->printOn(cout);
  }
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTPrismShapeFunction::RTPrismShapeFunction(int degree) :
order(degree),n_polys(0) {
  prism_poly_01=OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree);
  homo_poly=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
  mono_poly=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  prism_poly_2=OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree+1);
  n_polys=2*prism_poly_01->getNumber()+prism_poly_2->getNumber()
         +homo_poly->getNumber()*mono_poly->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void RTPrismShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> wp01v;
  prism_poly_01->computeValues(p,wp01v);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<prism_poly_01->getNumber();s++) {
      v[iv++][d]=wp01v[s];
    }
  }

  NumPtr<REAL> wp2v;
  prism_poly_2->computeValues(p,wp2v);
  for (int s=0;s<prism_poly_2->getNumber();s++) {
    v[iv++][2]=wp2v[s];
  }

  Point<SPACEDIM-1> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  homo_poly->computeValues(p01,hpv);
  NumPtr<REAL> mv;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mv);
  for (int i=0;i<=order;i++) {
    REAL &mvi=mv[i];
    for (int s=0;s<homo_poly->getNumber();s++) {
      v[iv][0]=p[0]*hpv[s]*mvi;
      v[iv][1]=p[1]*hpv[s]*mvi;
      iv++;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void RTPrismShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > wp01g;
  prism_poly_01->computeGrads(p,wp01g);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<prism_poly_01->getNumber();s++) g[iv++][d]=wp01g[s];
  }

  NumPtr<Tensor<1,SPACEDIM> > wp2g;
  prism_poly_2->computeGrads(p,wp2g);
  for (int s=0;s<prism_poly_2->getNumber();s++) {
    g[iv++][2]=wp2g[s];
  }

  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,2> > hpg;
  homo_poly->computeValues(p01,hpv);
  homo_poly->computeGrads(p01,hpg);
  NumPtr<REAL> mv;
  NumPtr<Tensor<1,1> > mg;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mv);
  mono_poly->computeGrads(p2,mg);
  for (int i=0;i<=order;i++) {
    REAL &mvi=mv[i];
    REAL &mgi=mg[i][0];
    for (int s=0;s<homo_poly->getNumber();s++) {
      g[iv][0][0]=(p[0]*hpg[s][0]+hpv[s])*mvi;
      g[iv][0][1]=p[0]*hpg[s][1]*mvi;
      g[iv][0][2]=p[0]*hpv[s]*mgi;
      g[iv][1][0]=p[1]*hpg[s][0]*mvi;
      g[iv][1][1]=(p[1]*hpg[s][1]+hpv[s])*mvi;
      g[iv][1][2]=p[1]*hpv[s]*mgi;
      iv++;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void RTPrismShapeFunction::computeGradGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > wp01gg;
  prism_poly_01->computeGradGrads(p,wp01gg);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<prism_poly_01->getNumber();s++) gg[iv++][d]=wp01gg[s];
  }

  NumPtr<Tensor<2,SPACEDIM> > wp2gg;
  prism_poly_2->computeGradGrads(p,wp2gg);
  for (int s=0;s<prism_poly_2->getNumber();s++) {
    gg[iv++][2]=wp2gg[s];
  }

  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,2> > hpg;
  NumPtr<Tensor<2,2> > hpgg;
  homo_poly->computeValues(p01,hpv);
  homo_poly->computeGrads(p01,hpg);
  homo_poly->computeGradGrads(p01,hpgg);
  NumPtr<REAL> mv;
  NumPtr<Tensor<1,1> > mg;
  NumPtr<Tensor<2,1> > mgg;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mv);
  mono_poly->computeGrads(p2,mg);
  mono_poly->computeGradGrads(p2,mgg);
  for (int i=0;i<=order;i++) {
    REAL &mvi=mv[i];
    REAL &mgi=mg[i][0];
    REAL &mggi=mgg[i][0][0];
    for (int s=0;s<homo_poly->getNumber();s++) {
      gg[iv][0][0][0]=(p[0]*hpgg[s][0][0]+2.*hpg[s][0])*mvi;
      gg[iv][0][0][1]=(p[0]*hpgg[s][0][1]+hpg[s][1])*mvi;
      gg[iv][0][0][2]=(p[0]*hpg[s][0]+hpv[s])*mgi;
      gg[iv][0][1][1]=p[0]*hpgg[s][1][1]*mvi;
      gg[iv][0][1][2]=p[0]*hpg[s][1]*mgi;
      gg[iv][0][2][2]=p[0]*hpv[s]*mggi;
      gg[iv][0][1][0]=gg[iv][0][0][1];
      gg[iv][0][2][0]=gg[iv][0][0][2];
      gg[iv][0][2][1]=gg[iv][0][1][2];

      gg[iv][1][0][0]=p[1]*hpgg[s][0][0]*mvi;
      gg[iv][1][0][1]=(p[1]*hpgg[s][0][1]+hpg[s][0])*mvi;
      gg[iv][1][0][2]=p[1]*hpg[s][0]*mgi;
      gg[iv][1][1][1]=(p[1]*hpgg[s][1][1]+2.*hpg[s][1])*mvi;
      gg[iv][1][1][2]=(p[1]*hpg[s][1]+hpv[s])*mgi;
      gg[iv][1][2][2]=p[1]*hpv[s]*mggi;
      gg[iv][1][1][0]=gg[iv][1][0][1];
      gg[iv][1][2][0]=gg[iv][1][0][2];
      gg[iv][1][2][1]=gg[iv][1][1][2];
      iv++;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTPrismShapeFunction::printOn(ostream &os) const {
  os << "RTPrismShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tprism_poly_01:" << endl;
  prism_poly_01->printOn(cout);
  cout << "\tprism_poly_2:" << endl;
  prism_poly_2->printOn(cout);
  cout << "\thomo_polys:" << endl;
  homo_poly->printOn(cout);
  cout << "\tmono_polys:" << endl;
  mono_poly->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTPrismDofShapeFunction::RTPrismDofShapeFunction(int degree) :
order(degree),n_polys(0) {
  if (degree<1) {
    interior_polys01=0;
    interior_polys2=0;
    n_polys=0;
  } else {
    interior_polys01=
      OPERATOR_NEW C0LagrangePrismPolynomials(degree-1,degree);
    interior_polys2=
      OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree-1);
    n_polys=2*interior_polys01->getNumber()
           +interior_polys2->getNumber();
  }
  triangle_face_polys=
    OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
  quadrilateral_face_polys=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree);
  n_triangle_face_polys=triangle_face_polys->getNumber();
  n_quadrilateral_face_polys=quadrilateral_face_polys->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void RTPrismDofShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> wp01v;
    interior_polys01->computeValues(p,wp01v);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<interior_polys01->getNumber();s++) {
        v[iv++][d]=wp01v[s];
      }
    }

    NumPtr<REAL> wp2v;
    interior_polys2->computeValues(p,wp2v);
    for (int s=0;s<interior_polys2->getNumber();s++) {
      v[iv++][2]=wp2v[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void RTPrismDofShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,SPACEDIM> > wp01g;
    interior_polys01->computeGrads(p,wp01g);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<interior_polys01->getNumber();s++) {
        g[iv++][d]=wp01g[s];
      }
    }

    NumPtr<Tensor<1,SPACEDIM> > wp2g;
    interior_polys2->computeGrads(p,wp2g);
    for (int s=0;s<interior_polys2->getNumber();s++) {
      g[iv++][2]=wp2g[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void RTPrismDofShapeFunction::computeGradGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,SPACEDIM> > wp01gg;
    interior_polys01->computeGradGrads(p,wp01gg);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<interior_polys01->getNumber();s++) {
        gg[iv++][d]=wp01gg[s];
      }
    }

    NumPtr<Tensor<2,SPACEDIM> > wp2gg;
    interior_polys2->computeGradGrads(p,wp2gg);
    for (int s=0;s<interior_polys2->getNumber();s++) {
      gg[iv++][2]=wp2gg[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void RTPrismDofShapeFunction::computeTriangleFaceValues(
const Point<SPACEDIM-1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_triangle_face_polys);
  v.initialize(0.);
  NumPtr<REAL> tfpv;
  triangle_face_polys->computeValues(p,tfpv);
  int iv=0;
  for (int s=0;s<triangle_face_polys->getNumber();s++) v[iv++]=tfpv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void RTPrismDofShapeFunction::computeTriangleFaceGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<1,SPACEDIM-1> > &g) const {
  g.allocate(n_triangle_face_polys);
  Tensor<1,SPACEDIM-1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM-1> > tfpg;
  triangle_face_polys->computeGrads(p,tfpg);
  int iv=0;
  for (int s=0;s<triangle_face_polys->getNumber();s++) g[iv++]=tfpg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void RTPrismDofShapeFunction::computeTriangleFaceGradGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<2,SPACEDIM-1> > &gg) const {
  gg.allocate(n_triangle_face_polys);
  Tensor<2,SPACEDIM-1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM-1> > tfpgg;
  triangle_face_polys->computeGradGrads(p,tfpgg);
  int iv=0;
  for (int s=0;s<triangle_face_polys->getNumber();s++) {
    gg[iv++]=tfpgg[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void RTPrismDofShapeFunction::computeQuadrilateralFaceValues(
const Point<SPACEDIM-1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_quadrilateral_face_polys);
  v.initialize(0.);
  NumPtr<REAL> tfpv;
  quadrilateral_face_polys->computeValues(p,tfpv);
  int iv=0;
  for (int s=0;s<quadrilateral_face_polys->getNumber();s++) v[iv++]=tfpv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void RTPrismDofShapeFunction::computeQuadrilateralFaceGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<1,SPACEDIM-1> > &g) const {
  g.allocate(n_quadrilateral_face_polys);
  Tensor<1,SPACEDIM-1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM-1> > tfpg;
  quadrilateral_face_polys->computeGrads(p,tfpg);
  int iv=0;
  for (int s=0;s<quadrilateral_face_polys->getNumber();s++) g[iv++]=tfpg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void RTPrismDofShapeFunction::computeQuadrilateralFaceGradGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<2,SPACEDIM-1> > &gg) const {
  gg.allocate(n_quadrilateral_face_polys);
  Tensor<2,SPACEDIM-1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM-1> > tfpgg;
  quadrilateral_face_polys->computeGradGrads(p,tfpgg);
  int iv=0;
  for (int s=0;s<quadrilateral_face_polys->getNumber();s++) {
    gg[iv++]=tfpgg[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//renumber[fe_s] = dof_s:
void RTPrismDofShapeFunction::shapeFunctionsOrder(const Shape *e,
NumPtr<int> &renumber) const {
  renumber.allocate(n_polys+2*n_triangle_face_polys
                           +3*n_quadrilateral_face_polys);
  int rn=renumber.getNumber();
  for (int fe_s=0;fe_s<rn;fe_s++) renumber[fe_s]=fe_s;
  const Prism *element=dynamic_cast<const Prism*>(e);
  CHECK_POINTER(element);
  element->reorderRTFaceDofs(0,order,renumber);
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<rn;i++) {
    if (renumber[i]>=rn || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(rn);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<rn;i++) renumber_inverse[renumber[i]]=i;
  for (int i=0;i<rn;i++) CHECK_BOUNDS(renumber_inverse[i],0,rn);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTPrismDofShapeFunction::printOn(ostream &os) const {
  os << "RTPrismDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tn_triangle_face_polys = " << n_triangle_face_polys
     << "\n\tn_quadrilateral_face_polys = " << n_quadrilateral_face_polys
     << endl;
  cout << "\ttriangle_face_polys:" << endl;
  triangle_face_polys->printOn(cout);
  cout << "\tquadrilateral_face_polys:" << endl;
  quadrilateral_face_polys->printOn(cout);
  cout << "\tinterior_polys01:" << endl;
  interior_polys01->printOn(cout);
  cout << "\tinterior_polys2:" << endl;
  interior_polys2->printOn(cout);
  HdivPrismDofShapeFunction::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
RTPrismDualShapeFunction::RTPrismDualShapeFunction(int degree) :
dof_polys(degree),order(degree),n_polys(0) {
  prism_poly_01=OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree);
  homo_poly=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
  mono_poly=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  prism_poly_2=OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree+1);
  int wp01n=prism_poly_01->getNumber();
  int wp2n=prism_poly_2->getNumber();
  int hpn=homo_poly->getNumber();
  int mpn=mono_poly->getNumber();
  n_polys=2*wp01n+wp2n+hpn*mpn;

  int dpn=dof_polys.getNumber();
  int dptfn=dof_polys.getTriangleFaceNumber();
  int dpqfn=dof_polys.getQuadrilateralFaceNumber();
  CHECK_SAME(n_polys,dptfn*2+dpqfn*3+dpn);

  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

  TriangleGaussianQuadrature
    tgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  GaussianQuadrature<1>
    gq1(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
  ProductQuadrature pq(tgq,gq1);

  GaussianQuadrature<2>
    qgq(GaussianQuadrature<2>::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM> > &pq_points=pq.points();
  const NumPtr<REAL> &pq_weights=pq.weights();
  const NumPtr<Point<SPACEDIM-1> > &tgq_points=tgq.points();
  const NumPtr<REAL> &tgq_weights=tgq.weights();
  const NumPtr<Point<SPACEDIM-1> > &qgq_points=qgq.points();
  const NumPtr<REAL> &qgq_weights=qgq.weights();

//triangle faces:
  NumPtr<Point<SPACEDIM> > tq_points_face[2];
  for (int f=0;f<2;f++) {
    tq_points_face[f].allocate(tgq_points.getNumber());
    Prism::staticProjectToFace(f,tgq_points,tq_points_face[f]);
  }
  NumPtr<NumPtr<REAL> > dptfv(tgq_points.getNumber());
  for (int q=0;q<tgq_points.getNumber();q++) {
    dof_polys.computeTriangleFaceValues(tgq_points[q],dptfv[q]);
  }
  int ioffset=0;
  for (int f=0;f<2;f++) {
    for (int q=0;q<tgq_points.getNumber();q++) {
      REAL wq=tgq_weights[q];
//    prism_poly_01 has zero normal component in triangle faces
      int joffset=2*wp01n;
      NumPtr<REAL> wp2v;
      prism_poly_2->computeValues(tq_points_face[f][q],wp2v);
      for (int j=0;j<wp2n;j++) {
        REAL wp2vj=wp2v[j];
        if (f==0) wp2vj=-wp2vj; // normal = - e_2
        for (int i=0;i<dptfn;i++) {
          matrix(i+ioffset,j+joffset)+=dptfv[q][i]*wq*wp2vj;
        }
      }
//    homo_poly has zero normal component in triangle faces
    }
    ioffset+=dptfn;
  }

//quadrilateral faces:
  NumPtr<Point<SPACEDIM> > qq_points_face[3];
  for (int f=2;f<5;f++) {
    qq_points_face[f-2].allocate(qgq_points.getNumber());
    Prism::staticProjectToFace(f,qgq_points,qq_points_face[f-2]);
  }
  NumPtr<NumPtr<REAL> > dpqfv(qgq_points.getNumber());
  for (int q=0;q<qgq_points.getNumber();q++) {
    dof_polys.computeQuadrilateralFaceValues(qgq_points[q],dpqfv[q]);
  }
  for (int f=2;f<5;f++) {
    Tensor<1,SPACEDIM> face_normal=Prism::faceNormal(f);
    for (int q=0;q<qgq_points.getNumber();q++) {
      REAL wq=qgq_weights[q];
      NumPtr<REAL> wp01v;
      const Point<SPACEDIM> &qp=qq_points_face[f-2][q];
      prism_poly_01->computeValues(qp,wp01v);
      int joffset=0;
      for (int d=0;d<2;d++) {
        for (int j=0;j<wp01n;j++) {
          REAL wp01vj=wp01v[j];
          for (int i=0;i<dpqfn;i++) {
            matrix(i+ioffset,j+joffset)+=
              dpqfv[q][i]*wq*face_normal[d]*wp01vj;
          }
        }
        joffset+=wp01n;
      }
//    prism_poly_2 has zero normal component in quadrilateral faces
      joffset+=wp2n;
      Point<SPACEDIM-1> qp01; qp01[0]=qp[0]; qp01[1]=qp[1];
      NumPtr<REAL> hpv;
      homo_poly->computeValues(qp01,hpv);
      NumPtr<REAL> mpv;
      Point<1> qp2(qp[2]);
      mono_poly->computeValues(qp2,mpv);
      REAL dot=face_normal[0]*qp[0]+face_normal[1]*qp[1];
      for (int jm=0;jm<mpn;jm++) {
        REAL mpvj=mpv[jm];
        for (int jh=0;jh<hpn;jh++) {
          REAL hpvj=hpv[jh];
          for (int i=0;i<dpqfn;i++) {
            matrix(i+ioffset,jh+jm*hpn+joffset)+=
              dpqfv[q][i]*wq*dot*hpvj*mpvj;
          }
        }
      }
    }
    ioffset+=dpqfn;
  }

//interior:
  for (int q=0;q<pq_points.getNumber();q++) {
    REAL wq=pq_weights[q];
    NumPtr<REAL> wp01v;
    const Point<SPACEDIM> &qp=pq_points[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    prism_poly_01->computeValues(qp,wp01v);
    int joffset=0;
    for (int d=0;d<2;d++) {
      for (int j=0;j<wp01n;j++) {
        REAL wp01vj=wp01v[j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[i][d]*wq*wp01vj;
        }
      }
      joffset+=wp01n;
    }
    NumPtr<REAL> wp2v;
    prism_poly_2->computeValues(qp,wp2v);
    for (int j=0;j<wp2n;j++) {
      REAL wp2vj=wp2v[j];
      for (int i=0;i<dpn;i++) {
        matrix(i+ioffset,j+joffset)+=dpv[i][2]*wq*wp2vj;
      }
    }
    joffset+=wp2n;
    Point<SPACEDIM-1> qp01; qp01[0]=qp[0]; qp01[1]=qp[1];
    NumPtr<REAL> hpv;
    homo_poly->computeValues(qp01,hpv);
    NumPtr<REAL> mpv;
    Point<1> qp2(qp[2]);
    mono_poly->computeValues(qp2,mpv);
    for (int jm=0;jm<mpn;jm++) {
      for (int jh=0;jh<hpn;jh++) {
        for (int i=0;i<dpn;i++) {
          REAL dot=0.;
          for (int d=0;d<2;d++) dot+=dpv[i][d]*qp[d];
          matrix(i+ioffset,jh+jm*hpn+joffset)+=dot*wq*hpv[jh]*mpv[jm];
        }
      }
    }
  }
  soln=factorNSolve(matrix,rhs); 
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void RTPrismDualShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> wp01v;
  prism_poly_01->computeValues(p,wp01v);
  int wp01n=prism_poly_01->getNumber();
  int offset=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<wp01n;s++) {
      REAL wp01vs=wp01v[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=wp01vs*(*soln)(s+offset,iv);
      }
    }
    offset+=wp01n;
  }

  NumPtr<REAL> wp2v;
  prism_poly_2->computeValues(p,wp2v);
  int wp2n=prism_poly_2->getNumber();
  for (int s=0;s<wp2n;s++) {
    REAL wp2vs=wp2v[s];
    for (int iv=0;iv<n_polys;iv++) {
      v[iv][2]+=wp2v[s]*(*soln)(s+offset,iv);
    }
  }
  offset+=wp2n;

  Point<SPACEDIM-1> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  homo_poly->computeValues(p01,hpv);
  int hpn=homo_poly->getNumber();
  NumPtr<REAL> mv;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mv);
  int mpn=mono_poly->getNumber();
  for (int i=0;i<mpn;i++) {
    REAL mvi=mv[i];
    for (int s=0;s<hpn;s++) {
      REAL hpvs=hpv[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+i*hpn+offset,iv);
        v[iv][0]+=p[0]*hpvs*mvi*ssiv;
        v[iv][1]+=p[1]*hpvs*mvi*ssiv;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void RTPrismDualShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > wp01g;
  prism_poly_01->computeGrads(p,wp01g);
  int wp01n=prism_poly_01->getNumber();
  int offset=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<wp01n;s++) {
      Tensor<1,SPACEDIM> &wp01gs=wp01g[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int j=0;j<SPACEDIM;j++) {
          givd[j]+=wp01gs[j]*ssiv;
        }
      }
    }
    offset+=wp01n;
  }

  NumPtr<Tensor<1,SPACEDIM> > wp2g;
  prism_poly_2->computeGrads(p,wp2g);
  int wp2n=prism_poly_2->getNumber();
  for (int s=0;s<wp2n;s++) {
    Tensor<1,SPACEDIM> &wp2gs=wp2g[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      Tensor<1,SPACEDIM> &giv2=g[iv][2];
      for (int j=0;j<SPACEDIM;j++) {
        giv2[j]+=wp2gs[j]*ssiv;
      }
    }
  }
  offset+=wp2n;

  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,2> > hpg;
  homo_poly->computeValues(p01,hpv);
  homo_poly->computeGrads(p01,hpg);
  int hpn=homo_poly->getNumber();
  NumPtr<REAL> mv;
  NumPtr<Tensor<1,1> > mg;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mv);
  mono_poly->computeGrads(p2,mg);
  int mpn=mono_poly->getNumber();
  for (int i=0;i<mpn;i++) {
    REAL &mvi=mv[i];
    REAL &mgi=mg[i][0];
    for (int s=0;s<hpn;s++) {
      REAL hpvs=hpv[s];
      Tensor<1,2> &hpgs=hpg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+i*hpn+offset,iv);
        Tensor<2,SPACEDIM> &giv=g[iv];
        giv[0][0]+=(p[0]*hpgs[0]+hpvs)*mvi*ssiv;
        giv[0][1]+=p[0]*hpgs[1]*mvi*ssiv;
        giv[0][2]+=p[0]*hpvs*mgi*ssiv;
        giv[1][0]+=p[1]*hpgs[0]*mvi*ssiv;
        giv[1][1]+=(p[1]*hpgs[1]+hpvs)*mvi*ssiv;
        giv[1][2]+=p[1]*hpvs*mgi*ssiv;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void RTPrismDualShapeFunction::computeGradGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > wp01gg;
  prism_poly_01->computeGradGrads(p,wp01gg);
  int wp01n=prism_poly_01->getNumber();
  int offset=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<wp01n;s++) {
      Tensor<2,SPACEDIM> &wp01ggs=wp01gg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            ggivd[i][j]+=wp01ggs[i][j]*ssiv;
          }
        }
      }
    }
    offset+=wp01n;
  }

  NumPtr<Tensor<2,SPACEDIM> > wp2gg;
  prism_poly_2->computeGradGrads(p,wp2gg);
  int wp2n=prism_poly_2->getNumber();
  for (int s=0;s<wp2n;s++) {
    Tensor<2,SPACEDIM> &wp2ggs=wp2gg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      Tensor<2,SPACEDIM> &ggiv2=gg[iv][2];
      for (int j=0;j<SPACEDIM;j++) {
        for (int i=0;i<SPACEDIM;i++) {
          ggiv2[i][j]+=wp2ggs[i][j]*ssiv;
        }
      }
    }
  }
  offset+=wp2n;

  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,2> > hpg;
  NumPtr<Tensor<2,2> > hpgg;
  homo_poly->computeValues(p01,hpv);
  homo_poly->computeGrads(p01,hpg);
  homo_poly->computeGradGrads(p01,hpgg);
  int hpn=homo_poly->getNumber();
  NumPtr<REAL> mv;
  NumPtr<Tensor<1,1> > mg;
  NumPtr<Tensor<2,1> > mgg;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mv);
  mono_poly->computeGrads(p2,mg);
  mono_poly->computeGradGrads(p2,mgg);
  int mpn=mono_poly->getNumber();
  for (int i=0;i<mpn;i++) {
    REAL &mvi=mv[i];
    REAL &mgi=mg[i][0];
    REAL &mggi=mgg[i][0][0];
    for (int s=0;s<hpn;s++) {
      Tensor<2,SPACEDIM-1> &hpggs=hpgg[s];
      Tensor<1,SPACEDIM-1> &hpgs=hpg[s];
      REAL hpvs=hpv[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+i*hpn+offset,iv);
        Tensor<3,SPACEDIM> &ggiv=gg[iv];
        ggiv[0][0][0]+=(p[0]*hpggs[0][0]+2.*hpgs[0])*mvi*ssiv;
        ggiv[0][0][1]+=(p[0]*hpggs[0][1]+hpgs[1])*mvi*ssiv;
        ggiv[0][0][2]+=(p[0]*hpgs[0]+hpvs)*mgi*ssiv;
        ggiv[0][1][1]+=p[0]*hpggs[1][1]*mvi*ssiv;
        ggiv[0][1][2]+=p[0]*hpgs[1]*mgi*ssiv;
        ggiv[0][2][2]+=p[0]*hpvs*mggi*ssiv;
        ggiv[0][1][0]=gg[iv][0][0][1];
        ggiv[0][2][0]=gg[iv][0][0][2];
        ggiv[0][2][1]=gg[iv][0][1][2];

        ggiv[1][0][0]+=p[1]*hpggs[0][0]*mvi*ssiv;
        ggiv[1][0][1]+=(p[1]*hpggs[0][1]+hpgs[0])*mvi*ssiv;
        ggiv[1][0][2]+=p[1]*hpgs[0]*mgi*ssiv;
        ggiv[1][1][1]+=(p[1]*hpggs[1][1]+2.*hpgs[1])*mvi*ssiv;
        ggiv[1][1][2]+=(p[1]*hpgs[1]+hpvs)*mgi*ssiv;
        ggiv[1][2][2]+=p[1]*hpvs*mggi*ssiv;
        ggiv[1][1][0]=gg[iv][1][0][1];
        ggiv[1][2][0]=gg[iv][1][0][2];
        ggiv[1][2][1]=gg[iv][1][1][2];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTPrismDualShapeFunction::nonzeroShapesOnFace(const Shape1 *e,
int f,NumPtr<int> &indices) const {
  int nt=((order+1)*(order+2))/2;
  if (f<2) {
    indices.allocate(nt);
    int offset=f*nt;
    for (int i=0;i<nt;i++) indices[i]=i+offset;
  } else {
    int nq=(order+1)*(order+1);
    indices.allocate(nq);
    int offset=2*nt+(f-2)*nq;
    for (int i=0;i<nq;i++) indices[i]=i+offset;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void RTPrismDualShapeFunction::printOn(ostream &os) const {
  os << "RTPrismDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tprism_poly_01:" << endl;
  prism_poly_01->printOn(cout);
  cout << "\tprism_poly_2:" << endl;
  prism_poly_2->printOn(cout);
  cout << "\thomo_polys:" << endl;
  homo_poly->printOn(cout);
  cout << "\tmono_polys:" << endl;
  mono_poly->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMSimplexShapeFunction::printOn(ostream &os) const {
  os << "BDFMSimplexShapeFunction" << endl;
  RTSimplexShapeFunction::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMSimplexDofShapeFunction::printOn(ostream &os) const {
  os << "BDFMSimplexDofShapeFunction" << endl;
  RTSimplexDofShapeFunction::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMSimplexDualShapeFunction::printOn(ostream &os) const {
  os << "BDFMSimplexDualShapeFunction" << endl;
  RTSimplexDualShapeFunction::printOn(os);
}

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BDFMParallelpipedShapeFunction::BDFMParallelpipedShapeFunction(
int degree) : order(degree),n_polys(0) {
# if (SPACEDIM==2)
  polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
  homo_polys=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
# else
  polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
  homo_polys=OPERATOR_NEW HomogeneousLagrangePolynomials3(degree);
# endif
  n_polys=(polys->getNumber()+homo_polys->getNumber())*SPACEDIM;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void BDFMParallelpipedShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  polys->computeValues(p,pv);
  int pn=polys->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) v[iv++][d]=pv[s];
  }
  NumPtr<REAL> hpv;
  homo_polys->computeValues(p,hpv);
  int hpn=homo_polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<hpn;s++) v[iv++][d]=p[d]*hpv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void BDFMParallelpipedShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > gv;
  polys->computeGrads(p,gv);
  int pn=polys->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) g[iv++][d]=gv[s];
  }
  NumPtr<REAL> hpv;
  homo_polys->computeValues(p,hpv);
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  homo_polys->computeGrads(p,hpg);
  int hpn=homo_polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<hpn;s++) {
      for (int j=0;j<SPACEDIM;j++) g[iv][d][j]=p[d]*hpg[s][j];
      g[iv++][d][d]+=hpv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void BDFMParallelpipedShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > ggv;
  polys->computeGradGrads(p,ggv);
  int pn=polys->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) gg[iv++][d]=ggv[s];
  }
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  homo_polys->computeGrads(p,hpg);
  NumPtr<Tensor<2,SPACEDIM> > hpgg;
  homo_polys->computeGradGrads(p,hpgg);
  int hpn=homo_polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<hpn;s++) {
      Tensor<2,SPACEDIM> &ggivd=gg[iv++][d];
      Tensor<2,SPACEDIM> &hpggs=hpgg[s];
      Tensor<1,SPACEDIM> &hpgs=hpg[s];
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          ggivd[i][j]=p[d]*hpggs[i][j];
          if (i==d) ggivd[d][j]+=hpgs[j];
          if (j==d) ggivd[i][d]+=hpgs[i];
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMParallelpipedShapeFunction::printOn(ostream &os) const {
  os << "BDFMParallelpipedShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tpolys:" << endl;
  polys->printOn(cout);
  cout << "\thomo_polys:" << endl;
  homo_polys->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BDFMParallelpipedDofShapeFunction::BDFMParallelpipedDofShapeFunction(
int degree) : order(degree),n_polys(0),n_face_polys(0) {
# if (SPACEDIM==2)
  face_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_face_polys=face_polys->getNumber();
  if (degree<1) {
    interior_polys=0;
    n_polys=0;
  } else {
    interior_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree-1);
    n_polys=interior_polys->getNumber()*SPACEDIM;
  }
# else
  face_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
  n_face_polys=face_polys->getNumber();
  if (degree<1) {
    interior_polys=0;
    n_polys=0;
  } else {
    interior_polys=OPERATOR_NEW
      C0LagrangeTetrahedronPolynomials(degree-1);
    n_polys=interior_polys->getNumber()*SPACEDIM;
  }
# endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void BDFMParallelpipedDofShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> pv;
    interior_polys->computeValues(p,pv);
    int pn=interior_polys->getNumber();
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<pn;s++) v[iv++][d]=pv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void BDFMParallelpipedDofShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,SPACEDIM> > gv;
    interior_polys->computeGrads(p,gv);
    int pn=interior_polys->getNumber();
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<pn;s++) g[iv++][d]=gv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void BDFMParallelpipedDofShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,SPACEDIM> > ggv;
    interior_polys->computeGradGrads(p,ggv);
    int pn=interior_polys->getNumber();
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<pn;s++) gg[iv++][d]=ggv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMParallelpipedDofShapeFunction::computeFaceValues(
const Point<SPACEDIM-1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_face_polys);
  v.initialize(0.);
  NumPtr<REAL> pv;
  face_polys->computeValues(p,pv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) v[iv++]=pv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void BDFMParallelpipedDofShapeFunction::computeFaceGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<1,SPACEDIM-1> > &g) const {
  g.allocate(n_face_polys);
  Tensor<1,SPACEDIM-1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM-1> > gv;
  face_polys->computeGrads(p,gv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) g[iv++]=gv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void BDFMParallelpipedDofShapeFunction::computeFaceGradGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<2,SPACEDIM-1> > &gg) const {
  gg.allocate(n_face_polys);
  Tensor<2,SPACEDIM-1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM-1> > ggv;
  face_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) gg[iv++]=ggv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//renumber[fe_s] = dof_s:
void BDFMParallelpipedDofShapeFunction::shapeFunctionsOrder(
const Shape *e,NumPtr<int> &renumber) const {
  renumber.allocate(n_polys+2*SPACEDIM*n_face_polys);
  int rn=renumber.getNumber();
  for (int fe_s=0;fe_s<rn;fe_s++) renumber[fe_s]=fe_s;
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<rn;i++) {
    if (renumber[i]>=rn || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(n_polys);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<n_polys;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<n_polys;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,n_polys);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMParallelpipedDofShapeFunction::printOn(ostream &os) const {
  os << "BDFMParallelpipedDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tn_face_polys = " << n_face_polys << endl;
  cout << "\tface_polys:" << endl;
  face_polys->printOn(cout);
  cout << "\tinterior_polys:" << endl;
  interior_polys->printOn(cout);
  HdivDofTensorShapeFunction::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BDFMParallelpipedDualShapeFunction::BDFMParallelpipedDualShapeFunction(
int degree) : dof_polys(degree),order(degree),n_polys(0),soln(0) {
# if (SPACEDIM==2)
  polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
  homo_polys=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
# else
  polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
  homo_polys=OPERATOR_NEW HomogeneousLagrangePolynomials3(degree);
# endif
  n_polys=(polys->getNumber()+homo_polys->getNumber())*SPACEDIM;

  int dpn=dof_polys.getNumber();
  int dpfn=dof_polys.getFaceNumber();
  CHECK_SAME(n_polys,dpn+2*SPACEDIM*dpfn);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMParallelpipedDualShapeFunction::computeCoefficientsFor(
const Shape *element) {
  const D_CHOOSE(Interval,Quadrilateral,Hexahedron) *parallelpiped=
    dynamic_cast<const D_CHOOSE(Interval,Quadrilateral,Hexahedron)*>(
      element);
  CHECK_POINTER(parallelpiped);

  int dpn=dof_polys.getNumber();
  int dpfn=dof_polys.getFaceNumber();
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*order+1));
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();

  GaussianQuadrature<SPACEDIM-1>
    fgq(GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*order+2));
  const NumPtr<Point<SPACEDIM-1> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  NumPtr<NumPtr<REAL> > dpfv(fgq_points.getNumber());
  for (int q=0;q<fgq_points.getNumber();q++) {
    dof_polys.computeFaceValues(fgq_points[q],dpfv[q]);
  }
  int pn=polys->getNumber();
  int ioffset=0;
  int joffset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int f=2*d;f<2*d+2;f++) {
      NumPtr<Point<SPACEDIM> > q_points_face(fgq_points.getNumber());
      parallelpiped->projectToFaceForBDFM(f,fgq_points,q_points_face);
      for (int q=0;q<fgq_points.getNumber();q++) {
        REAL wq=fgq_weights[q];
        NumPtr<REAL> pv;
        polys->computeValues(q_points_face[q],pv);
        for (int j=0;j<pn;j++) {
          REAL pvj=pv[j];
          if (f%2==0) pvj=-pvj; // normal = - e_{f/2}
          for (int i=0;i<dpfn;i++) {
            matrix(i+ioffset,j+joffset)+=dpfv[q][i]*wq*pvj;
          }
        }
      }
      ioffset+=dpfn;
    }
    joffset+=pn;
  }
  ioffset=0;
  int hpn=homo_polys->getNumber();
  int hoffset=SPACEDIM*pn;
  for (int d=0;d<SPACEDIM;d++) {
    for (int f=2*d;f<2*d+2;f++) {
      NumPtr<Point<SPACEDIM> > q_points_face(fgq_points.getNumber());
      parallelpiped->projectToFaceForBDFM(f,fgq_points,q_points_face);
      for (int q=0;q<fgq_points.getNumber();q++) {
        const Point<SPACEDIM> &qp=q_points_face[q];
        REAL wq=fgq_weights[q];
        NumPtr<REAL> hpv;
        homo_polys->computeValues(q_points_face[q],hpv);
        for (int j=0;j<hpn;j++) {
          REAL hpvj=qp[d]*hpv[j];
          for (int i=0;i<dpfn;i++) {
            matrix(i+ioffset,j+hoffset)+=dpfv[q][i]*wq*hpvj;
          }
        }
      }
      ioffset+=dpfn;
    }
    hoffset+=hpn;
  }

  NumPtr<NumPtr<Tensor<1,SPACEDIM> > > dpv(gq_points.getNumber());
  NumPtr<NumPtr<REAL> > pv(gq_points.getNumber());
  NumPtr<NumPtr<REAL> > hpv(gq_points.getNumber());
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    dof_polys.computeValues(qp,dpv[q]);
    polys->computeValues(qp,pv[q]);
    homo_polys->computeValues(qp,hpv[q]);
  }
  ioffset=2*SPACEDIM*dpfn;
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int q=0;q<gq_points.getNumber();q++) {
      REAL wq=gq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL pvj=pv[q][j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+offset)+=dpv[q][i][d]*wq*pvj;
        }
      }
    }
    offset+=pn;
  }
  for (int d=0;d<SPACEDIM;d++) {
    for (int q=0;q<gq_points.getNumber();q++) {
      REAL wq=gq_weights[q];
      const Point<SPACEDIM> &qp=gq_points[q];
      for (int j=0;j<hpn;j++) {
        REAL hpvj=hpv[q][j];
        for (int i=0;i<dpn;i++) {
          REAL dot=dpv[q][i][d]*qp[d];
          matrix(i+ioffset,j+offset)+=dot*wq*hpvj;
        }
      }
    }
    offset+=hpn;
  }
  if (soln) delete soln;
  soln=factorNSolve(matrix,rhs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void BDFMParallelpipedDualShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  polys->computeValues(p,pv);
  int pn=polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) {
      REAL pvs=pv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=pvs*(*soln)(s+offset,iv);
      }
    }
    offset+=pn;
  }
  NumPtr<REAL> hv;
  homo_polys->computeValues(p,hv);
  int hpn=homo_polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<hpn;s++) {
      REAL hvs=hv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=p[d]*hvs*(*soln)(s+offset,iv);
      }
    }
    offset+=hpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void BDFMParallelpipedDualShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > gv;
  polys->computeGrads(p,gv);
  int pn=polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) {
      Tensor<1,SPACEDIM> &gvs=gv[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int i=0;i<SPACEDIM;i++) {
          givd[i]+=gvs[i]*ssiv;
        }
      }
    }
    offset+=pn;
  }
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  homo_polys->computeValues(p,hpv);
  homo_polys->computeGrads(p,hpg);
  int hpn=homo_polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<hpn;s++) {
      Tensor<1,SPACEDIM> &hpgs=hpg[s];
      REAL &hpvs=hpv[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &gvd=g[iv][d];
        for (int i=0;i<SPACEDIM;i++) gvd[i]+=p[d]*hpgs[i]*ssiv;
        gvd[d]+=hpvs*ssiv;
      }
    }
    offset+=hpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void BDFMParallelpipedDualShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > ggv;
  polys->computeGradGrads(p,ggv);
  int pn=polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) {
      Tensor<2,SPACEDIM> &ggvs=ggv[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int i=0;i<SPACEDIM;i++) {
          for (int j=0;j<SPACEDIM;j++) {
            ggivd[i][j]+=ggvs[i][j]*ssiv;
          }
        }
      }
    }
    offset+=pn;
  }
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  NumPtr<Tensor<2,SPACEDIM> > hpgg;
  homo_polys->computeGrads(p,hpg);
  homo_polys->computeGradGrads(p,hpgg);
  int hpn=homo_polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<hpn;s++) {
      Tensor<2,SPACEDIM> &hpggs=hpgg[s];
      Tensor<1,SPACEDIM> &hpgs=hpg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            ggivd[i][j]+=p[d]*hpggs[i][j]*ssiv;
            if (i==d) ggivd[d][j]+=hpgs[j]*ssiv;
            if (j==d) ggivd[i][d]+=hpgs[i]*ssiv;
          }
        }
      }
    }
    offset+=hpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMParallelpipedDualShapeFunction::nonzeroShapesOnFace(
const Shape1 *e,int f,NumPtr<int> &indices) const {
#if (SPACEDIM==2)
  int nl=order+1;
  indices.allocate(nl);
  int offset=f*nl;
  for (int i=0;i<nl;i++) indices[i]=i+offset;
#else
  int nt=((order+1)*(order+2))/2;
  indices.allocate(nt);
  int offset=f*nt;
  for (int i=0;i<nt;i++) indices[i]=i+offset;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMParallelpipedDualShapeFunction::printOn(ostream &os) const {
  os << "BDFMParallelpipedDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tpolys:" << endl;
  polys->printOn(cout);
  cout << "\thomo_polys:" << endl;
  homo_polys->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BDFMPrismShapeFunction::BDFMPrismShapeFunction(int degree) :
order(degree),n_polys(0) {
  poly=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
  homo_poly=OPERATOR_NEW HomogeneousLagrangePolynomials3(degree);
  n_polys=SPACEDIM*poly->getNumber()+2*homo_poly->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void BDFMPrismShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  poly->computeValues(p,pv);
  int pn=poly->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) v[iv++][d]=pv[s];
  }

  NumPtr<REAL> hpv;
  homo_poly->computeValues(p,hpv);
  int hpn=homo_poly->getNumber();
  for (int s=0;s<hpn;s++) {
    v[iv][0]=p[0]*hpv[s];
    v[iv][1]=p[1]*hpv[s];
    iv++;
  }
  for (int s=0;s<hpn;s++) v[iv++][2]=p[2]*hpv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void BDFMPrismShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > pg;
  poly->computeGrads(p,pg);
  int pn=poly->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) g[iv++][d]=pg[s];
  }

  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  homo_poly->computeValues(p,hpv);
  homo_poly->computeGrads(p,hpg);
  int hpn=homo_poly->getNumber();
  for (int s=0;s<hpn;s++) {
    g[iv][0][0]=p[0]*hpg[s][0]+hpv[s];
    g[iv][0][1]=p[0]*hpg[s][1];
    g[iv][0][2]=p[0]*hpg[s][2];
    g[iv][1][0]=p[1]*hpg[s][0];
    g[iv][1][1]=p[1]*hpg[s][1]+hpv[s];
    g[iv][1][2]=p[1]*hpg[s][2];
    iv++;
  }

  for (int s=0;s<hpn;s++) {
    g[iv][2][0]=p[2]*hpg[s][0];
    g[iv][2][1]=p[2]*hpg[s][1];
    g[iv][2][2]=p[2]*hpg[s][2]+hpv[s];
    iv++;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void BDFMPrismShapeFunction::computeGradGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > pgg;
  poly->computeGradGrads(p,pgg);
  int pn=poly->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) gg[iv++][d]=pgg[s];
  }

  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  NumPtr<Tensor<2,SPACEDIM> > hpgg;
  homo_poly->computeValues(p,hpv);
  homo_poly->computeGrads(p,hpg);
  homo_poly->computeGradGrads(p,hpgg);
  int hpn=homo_poly->getNumber();
  for (int s=0;s<homo_poly->getNumber();s++) {
    gg[iv][0][0][0]=p[0]*hpgg[s][0][0]+2.*hpg[s][0];
    gg[iv][0][0][1]=p[0]*hpgg[s][0][1]+hpg[s][1];
    gg[iv][0][0][2]=p[0]*hpgg[s][0][2]+hpg[s][2];
    gg[iv][0][1][1]=p[0]*hpgg[s][1][1];
    gg[iv][0][1][2]=p[0]*hpgg[s][1][2];
    gg[iv][0][2][2]=p[0]*hpgg[s][2][2];
    gg[iv][0][1][0]=gg[iv][0][0][1];
    gg[iv][0][2][0]=gg[iv][0][0][2];
    gg[iv][0][2][1]=gg[iv][0][1][2];

    gg[iv][1][0][0]=p[1]*hpgg[s][0][0];
    gg[iv][1][0][1]=p[1]*hpgg[s][0][1]+hpg[s][0];
    gg[iv][1][0][2]=p[1]*hpgg[s][0][2];
    gg[iv][1][1][1]=p[1]*hpgg[s][1][1]+2.*hpg[s][1];
    gg[iv][1][1][2]=p[1]*hpgg[s][1][2]+hpg[s][2];
    gg[iv][1][2][2]=p[1]*hpgg[s][2][2];
    gg[iv][1][1][0]=gg[iv][1][0][1];
    gg[iv][1][2][0]=gg[iv][1][0][2];
    gg[iv][1][2][1]=gg[iv][1][1][2];
    iv++;
  }

  for (int s=0;s<homo_poly->getNumber();s++) {
    gg[iv][2][0][0]=p[2]*hpgg[s][0][0];
    gg[iv][2][0][1]=p[2]*hpgg[s][0][1];
    gg[iv][2][0][2]=p[2]*hpgg[s][0][2]+hpg[s][0];
    gg[iv][2][1][1]=p[2]*hpgg[s][1][1];
    gg[iv][2][1][2]=p[2]*hpgg[s][1][2]+hpg[s][1];
    gg[iv][2][2][2]=p[2]*hpgg[s][2][2]+2.*hpg[s][2];
    gg[iv][2][1][0]=gg[iv][2][0][1];
    gg[iv][2][2][0]=gg[iv][2][0][2];
    gg[iv][2][2][1]=gg[iv][2][1][2];
    iv++;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMPrismShapeFunction::printOn(ostream &os) const {
  os << "BDFMPrismShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tpoly:" << endl;
  poly->printOn(cout);
  cout << "\thomo_poly:" << endl;
  homo_poly->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BDFMPrismDofShapeFunction::BDFMPrismDofShapeFunction(int degree) :
order(degree),n_polys(0) {
  if (degree<1) {
    interior_polys=0;
    n_polys=0;
  } else {
    interior_polys=
      OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree-1);
    n_polys=SPACEDIM*interior_polys->getNumber();
  }
  face_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
  n_face_polys=face_polys->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void BDFMPrismDofShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> pv;
    interior_polys->computeValues(p,pv);
    int pn=interior_polys->getNumber();
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<pn;s++) v[iv++][d]=pv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void BDFMPrismDofShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,SPACEDIM> > pg;
    interior_polys->computeGrads(p,pg);
    int pn=interior_polys->getNumber();
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<pn;s++) g[iv++][d]=pg[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void BDFMPrismDofShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,SPACEDIM> > pgg;
    interior_polys->computeGradGrads(p,pgg);
    int pn=interior_polys->getNumber();
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<pn;s++) gg[iv++][d]=pgg[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void BDFMPrismDofShapeFunction::computeTriangleFaceValues(
const Point<SPACEDIM-1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_face_polys);
  v.initialize(0.);
  NumPtr<REAL> tfpv;
  face_polys->computeValues(p,tfpv);
  int fpn=face_polys->getNumber();
  int iv=0;
  for (int s=0;s<fpn;s++) v[iv++]=tfpv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void BDFMPrismDofShapeFunction::computeTriangleFaceGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<1,SPACEDIM-1> > &g) const {
  g.allocate(n_face_polys);
  Tensor<1,SPACEDIM-1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM-1> > tfpg;
  face_polys->computeGrads(p,tfpg);
  int fpn=face_polys->getNumber();
  int iv=0;
  for (int s=0;s<fpn;s++) g[iv++]=tfpg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void BDFMPrismDofShapeFunction::computeTriangleFaceGradGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<2,SPACEDIM-1> > &gg) const {
  gg.allocate(n_face_polys);
  Tensor<2,SPACEDIM-1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM-1> > tfpgg;
  face_polys->computeGradGrads(p,tfpgg);
  int fpn=face_polys->getNumber();
  int iv=0;
  for (int s=0;s<fpn;s++) gg[iv++]=tfpgg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void BDFMPrismDofShapeFunction::computeQuadrilateralFaceValues(
const Point<SPACEDIM-1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_face_polys);
  v.initialize(0.);
  NumPtr<REAL> tfpv;
  face_polys->computeValues(p,tfpv);
  int fpn=face_polys->getNumber();
  int iv=0;
  for (int s=0;s<fpn;s++) v[iv++]=tfpv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void BDFMPrismDofShapeFunction::computeQuadrilateralFaceGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<1,SPACEDIM-1> > &g) const {
  g.allocate(n_face_polys);
  Tensor<1,SPACEDIM-1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM-1> > tfpg;
  face_polys->computeGrads(p,tfpg);
  int fpn=face_polys->getNumber();
  int iv=0;
  for (int s=0;s<fpn;s++) g[iv++]=tfpg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void BDFMPrismDofShapeFunction::computeQuadrilateralFaceGradGrads(
const Point<SPACEDIM-1> &p,NumPtr<Tensor<2,SPACEDIM-1> > &gg) const {
  gg.allocate(n_face_polys);
  Tensor<2,SPACEDIM-1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM-1> > tfpgg;
  face_polys->computeGradGrads(p,tfpgg);
  int fpn=face_polys->getNumber();
  int iv=0;
  for (int s=0;s<fpn;s++) gg[iv++]=tfpgg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//renumber[fe_s] = dof_s:
void BDFMPrismDofShapeFunction::shapeFunctionsOrder(const Shape *e,
NumPtr<int> &renumber) const {
  renumber.allocate(n_polys+5*n_face_polys);
  int rn=renumber.getNumber();
  for (int fe_s=0;fe_s<rn;fe_s++) renumber[fe_s]=fe_s;
  const Prism *element=dynamic_cast<const Prism*>(e);
  CHECK_POINTER(element);
  element->reorderBDFMFaceDofs(0,order,renumber);
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<rn;i++) {
    if (renumber[i]>=rn || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(rn);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<rn;i++) renumber_inverse[renumber[i]]=i;
  for (int i=0;i<rn;i++) CHECK_BOUNDS(renumber_inverse[i],0,rn);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMPrismDofShapeFunction::printOn(ostream &os) const {
  os << "BDFMPrismDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tn_face_polys = " << n_face_polys << endl;
  cout << "\tface_polys:" << endl;
  face_polys->printOn(cout);
  cout << "\tinterior_polys:" << endl;
  interior_polys->printOn(cout);
  HdivPrismDofShapeFunction::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BDFMPrismDualShapeFunction::BDFMPrismDualShapeFunction(int degree) :
dof_polys(degree),order(degree),n_polys(0),soln(0) {
  poly=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
  homo_poly=OPERATOR_NEW HomogeneousLagrangePolynomials3(degree);
  int pn=poly->getNumber();
  int hpn=homo_poly->getNumber();
  n_polys=SPACEDIM*pn+2*hpn;

  int dpn=dof_polys.getNumber();
  int dptfn=dof_polys.getTriangleFaceNumber();
  int dpqfn=dof_polys.getQuadrilateralFaceNumber();
  CHECK_SAME(n_polys,dptfn*2+dpqfn*3+dpn);
};
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMPrismDualShapeFunction::computeCoefficientsFor(
const Shape *element) {
  const Prism *prism=dynamic_cast<const Prism*>(element);
  CHECK_POINTER(prism);

  NumPtr<int> renumber; // renumber[fe_s]=dof_s
  dof_polys.shapeFunctionsOrder(prism,renumber);
  int pn=poly->getNumber();
  int hpn=homo_poly->getNumber();
  int dpn=dof_polys.getNumber();
  int dptfn=dof_polys.getTriangleFaceNumber();
  int dpqfn=dof_polys.getQuadrilateralFaceNumber();

  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

  TriangleGaussianQuadrature
    tgq(TriangleGaussianQuadrature::argGivenOrder(2*order+2));
  GaussianQuadrature<1>
    gq1(GaussianQuadrature<1>::argGivenOrder(2*order+1));
  ProductQuadrature pq(tgq,gq1);

  GaussianQuadrature<2>
    qgq(GaussianQuadrature<2>::argGivenOrder(2*order+2));
  const NumPtr<Point<SPACEDIM> > &pq_points=pq.points();
  const NumPtr<REAL> &pq_weights=pq.weights();
  const NumPtr<Point<SPACEDIM-1> > &tgq_points=tgq.points();
  const NumPtr<REAL> &tgq_weights=tgq.weights();
  const NumPtr<Point<SPACEDIM-1> > &qgq_points=qgq.points();
  const NumPtr<REAL> &qgq_weights=qgq.weights();

//triangle faces:
  NumPtr<NumPtr<REAL> > dptfv(tgq_points.getNumber());
  for (int q=0;q<tgq_points.getNumber();q++) {
    dof_polys.computeTriangleFaceValues(tgq_points[q],dptfv[q]);
  }
  int ioffset=0;
  for (int f=0;f<2;f++) {
    NumPtr<Point<SPACEDIM> > tq_points_face(tgq_points.getNumber());
    prism->projectToFaceForBDFM(f,tgq_points,tq_points_face);
    for (int q=0;q<tgq_points.getNumber();q++) {
      const Point<SPACEDIM> &qp=tq_points_face[q];
      REAL wq=tgq_weights[q];
      int joffset=2*pn;
      NumPtr<REAL> pv;
      poly->computeValues(qp,pv);
      for (int j=0;j<pn;j++) {
        REAL pvj=pv[j];
        if (f==0) pvj=-pvj; // normal = - e_2
        for (int i=0;i<dptfn;i++) {
          matrix(i+ioffset,j+joffset)+=dptfv[q][i]*wq*pvj;
        }
      }
      joffset+=pn+hpn;
      NumPtr<REAL> hpv;
      homo_poly->computeValues(qp,hpv);
      for (int j=0;j<hpn;j++) {
        REAL hpvj=qp[2]*hpv[j];
        if (f==0) hpvj=-hpvj;
        for (int i=0;i<dptfn;i++) {
          matrix(i+ioffset,j+joffset)+=dptfv[q][i]*wq*hpvj;
        }
      }
    }
    ioffset+=dptfn;
  }

//quadrilateral faces:
  NumPtr<NumPtr<REAL> > dpqfv(qgq_points.getNumber());
  for (int q=0;q<qgq_points.getNumber();q++) {
    dof_polys.computeQuadrilateralFaceValues(qgq_points[q],dpqfv[q]);
  }
  for (int f=2;f<5;f++) {
    NumPtr<Point<SPACEDIM> > qq_points_face(qgq_points.getNumber());
    prism->projectToFaceForBDFM(f,qgq_points,qq_points_face);
    Tensor<1,SPACEDIM> face_normal=Prism::faceNormal(f);
    for (int q=0;q<qgq_points.getNumber();q++) {
      REAL wq=qgq_weights[q];
      const Point<SPACEDIM> &qp=qq_points_face[q];
      NumPtr<REAL> pv;
      poly->computeValues(qp,pv);
      int joffset=0;
      for (int d=0;d<2;d++) {
        for (int j=0;j<pn;j++) {
          REAL pvj=pv[j];
          for (int i=0;i<dpqfn;i++) {
            matrix(i+ioffset,j+joffset)+=
              dpqfv[q][i]*wq*face_normal[d]*pvj;
          }
        }
        joffset+=pn;
      }
      joffset+=pn;
      NumPtr<REAL> hpv;
      homo_poly->computeValues(qp,hpv);
      REAL dot=face_normal[0]*qp[0]+face_normal[1]*qp[1];
      for (int j=0;j<hpn;j++) {
        REAL hpvj=hpv[j];
        for (int i=0;i<dpqfn;i++) {
          matrix(i+ioffset,j+joffset)+=dpqfv[q][i]*wq*dot*hpvj;
        }
      }
    }
    ioffset+=dpqfn;
  }

//interior:
  for (int q=0;q<pq_points.getNumber();q++) {
    REAL wq=pq_weights[q];
    const Point<SPACEDIM> &qp=pq_points[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<REAL> pv;
    poly->computeValues(qp,pv);
    int joffset=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int j=0;j<pn;j++) {
        REAL pvj=pv[j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[i][d]*wq*pvj;
        }
      }
      joffset+=pn;
    }
    NumPtr<REAL> hpv;
    homo_poly->computeValues(qp,hpv);
    for (int j=0;j<hpn;j++) {
      REAL hpvj=hpv[j];
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i][0]*qp[0]+dpv[i][1]*qp[1];
        matrix(i+ioffset,j+joffset)+=dot*wq*hpvj;
      }
    }
    joffset+=hpn;
    for (int j=0;j<hpn;j++) {
      REAL hpvj=hpv[j];
      for (int i=0;i<dpn;i++) {
        matrix(i+ioffset,j+joffset)+=dpv[i][2]*qp[2]*wq*hpvj;
      }
    }
  }
  if (soln) delete soln;
  soln=factorNSolve(matrix,rhs); 
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void BDFMPrismDualShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  poly->computeValues(p,pv);
  int pn=poly->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) {
      REAL pvs=pv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=pvs*(*soln)(s+offset,iv);
      }
    }
    offset+=pn;
  }

  NumPtr<REAL> hpv;
  homo_poly->computeValues(p,hpv);
  int hpn=homo_poly->getNumber();
  for (int s=0;s<hpn;s++) {
    REAL hpvs=hpv[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      v[iv][0]+=p[0]*hpvs*ssiv;
      v[iv][1]+=p[1]*hpvs*ssiv;
    }
  }
  offset+=hpn;
  for (int s=0;s<hpn;s++) {
    REAL hpvs=hpv[s];
    for (int iv=0;iv<n_polys;iv++) {
      v[iv][2]+=p[2]*hpvs*(*soln)(s+offset,iv);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void BDFMPrismDualShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > pg;
  poly->computeGrads(p,pg);
  int pn=poly->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) {
      Tensor<1,SPACEDIM> &pgs=pg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int j=0;j<SPACEDIM;j++) givd[j]+=pgs[j]*ssiv;
      }
    }
    offset+=pn;
  }

  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  homo_poly->computeValues(p,hpv);
  homo_poly->computeGrads(p,hpg);
  int hpn=homo_poly->getNumber();
  for (int s=0;s<hpn;s++) {
    REAL hpvs=hpv[s];
    Tensor<1,SPACEDIM> &hpgs=hpg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      Tensor<2,SPACEDIM> &giv=g[iv];
      giv[0][0]+=(p[0]*hpgs[0]+hpvs)*ssiv;
      giv[0][1]+=p[0]*hpgs[1]*ssiv;
      giv[0][2]+=p[0]*hpgs[2]*ssiv;
      giv[1][0]+=p[1]*hpgs[0]*ssiv;
      giv[1][1]+=(p[1]*hpgs[1]+hpvs)*ssiv;
      giv[1][2]+=p[1]*hpgs[2]*ssiv;
    }
  }
  offset+=hpn;
  for (int s=0;s<hpn;s++) {
    REAL hpvs=hpv[s];
    Tensor<1,SPACEDIM> &hpgs=hpg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      Tensor<2,SPACEDIM> &giv=g[iv];
      giv[2][0]+=p[2]*hpgs[0]*ssiv;
      giv[2][1]+=p[2]*hpgs[1]*ssiv;
      giv[2][2]+=(p[2]*hpgs[2]+hpvs)*ssiv;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void BDFMPrismDualShapeFunction::computeGradGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > pgg;
  poly->computeGradGrads(p,pgg);
  int pn=poly->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) {
      Tensor<2,SPACEDIM> &pggs=pgg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            ggivd[i][j]+=pggs[i][j]*ssiv;
          }
        }
      }
    }
    offset+=pn;
  }

  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,SPACEDIM> > hpg;
  NumPtr<Tensor<2,SPACEDIM> > hpgg;
  homo_poly->computeValues(p,hpv);
  homo_poly->computeGrads(p,hpg);
  homo_poly->computeGradGrads(p,hpgg);
  int hpn=homo_poly->getNumber();
  for (int s=0;s<hpn;s++) {
    Tensor<2,SPACEDIM> &hpggs=hpgg[s];
    Tensor<1,SPACEDIM> &hpgs=hpg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      Tensor<3,SPACEDIM> &ggiv=gg[iv];
      ggiv[0][0][0]+=(p[0]*hpggs[0][0]+2.*hpgs[0])*ssiv;
      ggiv[0][0][1]+=(p[0]*hpggs[0][1]+hpgs[1])*ssiv;
      ggiv[0][0][2]+=(p[0]*hpggs[0][2]+hpgs[2])*ssiv;
      ggiv[0][1][1]+=p[0]*hpggs[1][1]*ssiv;
      ggiv[0][1][2]+=p[0]*hpggs[1][2]*ssiv;
      ggiv[0][2][2]+=p[0]*hpggs[2][2]*ssiv;
      ggiv[0][1][0]=gg[iv][0][0][1];
      ggiv[0][2][0]=gg[iv][0][0][2];
      ggiv[0][2][1]=gg[iv][0][1][2];

      ggiv[1][0][0]+=p[1]*hpggs[0][0]*ssiv;
      ggiv[1][0][1]+=(p[1]*hpggs[0][1]+hpgs[0])*ssiv;
      ggiv[1][0][2]+=p[1]*hpggs[0][2]*ssiv;
      ggiv[1][1][1]+=(p[1]*hpggs[1][1]+2.*hpgs[1])*ssiv;
      ggiv[1][1][2]+=(p[1]*hpggs[1][2]+hpgs[2])*ssiv;
      ggiv[1][2][2]+=p[1]*hpggs[2][2]*ssiv;
      ggiv[1][1][0]=gg[iv][1][0][1];
      ggiv[1][2][0]=gg[iv][1][0][2];
      ggiv[1][2][1]=gg[iv][1][1][2];
    }
  }
  offset+=hpn;

  for (int s=0;s<hpn;s++) {
    Tensor<2,SPACEDIM> &hpggs=hpgg[s];
    Tensor<1,SPACEDIM> &hpgs=hpg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      Tensor<3,SPACEDIM> &ggiv=gg[iv];
      ggiv[2][0][0]+=p[2]*hpggs[0][0]*ssiv;
      ggiv[2][0][1]+=p[2]*hpggs[0][1]*ssiv;
      ggiv[2][0][2]+=(p[2]*hpggs[0][2]+hpgs[0])*ssiv;
      ggiv[2][1][1]+=p[2]*hpggs[1][1]*ssiv;
      ggiv[2][1][2]+=(p[2]*hpggs[1][2]+hpgs[1])*ssiv;
      ggiv[2][2][2]+=(p[2]*hpggs[2][2]+2.*hpgs[2])*ssiv;
      ggiv[2][1][0]=gg[iv][2][0][1];
      ggiv[2][2][0]=gg[iv][2][0][2];
      ggiv[2][2][1]=gg[iv][2][1][2];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMPrismDualShapeFunction::nonzeroShapesOnFace(const Shape1 *e,
int f,NumPtr<int> &indices) const {
  int nt=((order+1)*(order+2))/2;
  indices.allocate(nt);
  int offset=f*nt;
  for (int i=0;i<nt;i++) indices[i]=i+offset;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BDFMPrismDualShapeFunction::printOn(ostream &os) const {
  os << "BDFMPrismDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tpoly:" << endl;
  poly->printOn(cout);
  cout << "\thomo_poly:" << endl;
  homo_poly->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
NedelecXShapeFunction2::NedelecXShapeFunction2(int degree) :
order(degree),n_polys(0) {
  homo_poly=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
  n_polys=homo_poly->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NedelecXShapeFunction2::computeValues(const Point<2> &p,
NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_polys);
  Tensor<1,2> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> hv;
  homo_poly->computeValues(p,hv);
  int hn=homo_poly->getNumber();
  int iv=0;
  for (int s=0;s<hn;s++) {
    Tensor<1,2> &vv=v[iv++];
    REAL hvs=hv[s];
    vv[0]=-p[1]*hvs;
    vv[1]= p[0]*hvs;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NedelecXShapeFunction2::computeGrads(const Point<2> &p,
NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_polys);
  Tensor<2,2> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<REAL> hv;
  NumPtr<Tensor<1,2> > hg;
  homo_poly->computeValues(p,hv);
  homo_poly->computeGrads(p,hg);
  int hn=homo_poly->getNumber();
  int iv=0;
  for (int s=0;s<hn;s++) {
    Tensor<2,2> &gv=g[iv++];
    REAL hvs=hv[s];
    Tensor<1,2> &hgs=hg[s];
    gv[0][0]=-p[1]*hgs[0];
    gv[0][1]=-p[1]*hgs[1]-hvs;
    gv[1][0]= p[0]*hgs[0]+hvs;
    gv[1][1]= p[0]*hgs[1];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NedelecXShapeFunction2::computeGradGrads(const Point<2> &p,
NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,2> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<1,2> > hg;
  NumPtr<Tensor<2,2> > hgg;
  homo_poly->computeGrads(p,hg);
  homo_poly->computeGradGrads(p,hgg);
  int hn=homo_poly->getNumber();
  int iv=0;
  for (int s=0;s<hn;s++) {
    Tensor<3,2> &ggv=gg[iv++];
    Tensor<1,2> &hgs=hg[s];
    Tensor<2,2> &hggs=hgg[s];
    ggv[0][0][0]=-p[1]*hggs[0][0];
    ggv[0][0][1]=-p[1]*hggs[0][1]-hgs[0];
    ggv[0][1][1]=-p[1]*hggs[1][1]-2.*hgs[1];
    ggv[0][1][0]=ggv[0][0][1];
    ggv[1][0][0]= p[0]*hggs[0][0]+2.*hgs[0];
    ggv[1][0][1]= p[0]*hggs[0][1]+hgs[1];
    ggv[1][1][1]= p[0]*hggs[1][1];
    ggv[1][1][0]=ggv[1][0][1];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NedelecXShapeFunction2::printOn(ostream &os) const {
  os << "NedelecXShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  os << "\thomo_poly:" << endl;
  homo_poly->printOn(os);
  TensorShapeFunction<2>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
NedelecXShapeFunction3::NedelecXShapeFunction3(int degree) :
order(degree),n_polys(0) {
  homo_poly=OPERATOR_NEW HomogeneousLagrangePolynomials3(degree);
  homo_poly_01=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
  n_polys=2*homo_poly->getNumber()+homo_poly_01->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NedelecXShapeFunction3::computeValues(const Point<3> &p,
NumPtr<Tensor<1,3> > &v) const {
  v.allocate(n_polys);
  Tensor<1,3> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> hv;
  homo_poly->computeValues(p,hv);
  int hn=homo_poly->getNumber();
  int iv=0;
  for (int s=0;s<hn;s++) {
    Tensor<1,3> &vv=v[iv++];
    REAL hvs=hv[s];
    vv[1]=-p[2]*hvs;
    vv[2]= p[1]*hvs;
  }
  for (int s=0;s<hn;s++) {
    Tensor<1,3> &vv=v[iv++];
    REAL hvs=hv[s];
    vv[0]= p[2]*hvs;
    vv[2]=-p[0]*hvs;
  }
  NumPtr<REAL> hv01;
  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  homo_poly_01->computeValues(p01,hv01);
  int hn01=homo_poly_01->getNumber();
  for (int s=0;s<hn01;s++) {
    Tensor<1,3> &vv=v[iv++];
    REAL hvs01=hv01[s];
    vv[0]=-p01[1]*hvs01;
    vv[1]= p01[0]*hvs01;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NedelecXShapeFunction3::computeGrads(const Point<3> &p,
NumPtr<Tensor<2,3> > &g) const {
  g.allocate(n_polys);
  Tensor<2,3> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<REAL> hv;
  NumPtr<Tensor<1,3> > hg;
  homo_poly->computeValues(p,hv);
  homo_poly->computeGrads(p,hg);
  int hn=homo_poly->getNumber();
  int iv=0;
  for (int s=0;s<hn;s++) {
    Tensor<2,3> &gv=g[iv++];
    REAL hvs=hv[s];
    Tensor<1,3> &hgs=hg[s];
    gv[1][0]=-p[2]*hgs[0];
    gv[1][1]=-p[2]*hgs[1];
    gv[1][2]=-p[2]*hgs[2]-hvs;
    gv[2][0]= p[1]*hgs[0];
    gv[2][1]= p[1]*hgs[1]+hvs;
    gv[2][2]= p[1]*hgs[2];
  }
  for (int s=0;s<hn;s++) {
    Tensor<2,3> &gv=g[iv++];
    REAL hvs=hv[s];
    Tensor<1,3> &hgs=hg[s];
    gv[0][0]= p[2]*hgs[0];
    gv[0][1]= p[2]*hgs[1];
    gv[0][2]= p[2]*hgs[2]+hvs;
    gv[2][0]=-p[0]*hgs[0]-hvs;
    gv[2][1]=-p[0]*hgs[1];
    gv[2][2]=-p[0]*hgs[2];
  }
  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hv01;
  NumPtr<Tensor<1,2> > hg01;
  homo_poly_01->computeValues(p01,hv01);
  homo_poly_01->computeGrads(p01,hg01);
  int hn01=homo_poly_01->getNumber();
  for (int s=0;s<hn01;s++) {
    Tensor<2,3> &gv=g[iv++];
    REAL hvs01=hv01[s];
    Tensor<1,2> &hgs01=hg01[s];
    gv[0][0]=-p01[1]*hgs01[0];
    gv[0][1]=-p01[1]*hgs01[1]-hvs01;
    gv[1][0]= p01[0]*hgs01[0]+hvs01;
    gv[1][1]= p01[0]*hgs01[1];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NedelecXShapeFunction3::computeGradGrads(const Point<3> &p,
NumPtr<Tensor<3,3> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,3> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<1,3> > hg;
  NumPtr<Tensor<2,3> > hgg;
  homo_poly->computeGrads(p,hg);
  homo_poly->computeGradGrads(p,hgg);
  int hn=homo_poly->getNumber();
  int iv=0;
  for (int s=0;s<hn;s++) {
    Tensor<3,3> &ggv=gg[iv++];
    Tensor<1,3> &hgs=hg[s];
    Tensor<2,3> &hggs=hgg[s];
    ggv[1][0][0]=-p[2]*hggs[0][0];
    ggv[1][0][1]=-p[2]*hggs[0][1];
    ggv[1][0][2]=-p[2]*hggs[0][2]-hgs[0];
    ggv[1][1][1]=-p[2]*hggs[1][1];
    ggv[1][1][2]=-p[2]*hggs[1][2]-hgs[1];
    ggv[1][2][2]=-p[2]*hggs[2][2]-2.*hgs[2];
    ggv[1][1][0]=ggv[1][0][1];
    ggv[1][2][0]=ggv[1][0][2];
    ggv[1][2][1]=ggv[1][1][2];
    ggv[2][0][0]= p[1]*hggs[0][0];
    ggv[2][0][1]= p[1]*hggs[0][1]+hgs[0];
    ggv[2][0][2]= p[1]*hggs[0][2];
    ggv[2][1][1]= p[1]*hggs[1][1]+2.*hgs[1];
    ggv[2][1][2]= p[1]*hggs[1][2]+hgs[2];
    ggv[2][2][2]= p[1]*hggs[2][2];
    ggv[2][1][0]=ggv[2][0][1];
    ggv[2][2][0]=ggv[2][0][2];
    ggv[2][2][1]=ggv[2][1][2];
  }
  for (int s=0;s<hn;s++) {
    Tensor<3,3> &ggv=gg[iv++];
    Tensor<1,3> &hgs=hg[s];
    Tensor<2,3> &hggs=hgg[s];
    ggv[0][0][0]= p[2]*hggs[0][0];
    ggv[0][0][1]= p[2]*hggs[0][1];
    ggv[0][0][2]= p[2]*hggs[0][2]+hgs[0];
    ggv[0][1][1]= p[2]*hggs[1][1];
    ggv[0][1][2]= p[2]*hggs[1][2]+hgs[1];
    ggv[0][2][2]= p[2]*hggs[2][2]+2.*hgs[2];
    ggv[0][1][0]=ggv[0][0][1];
    ggv[0][2][0]=ggv[0][0][2];
    ggv[0][2][1]=ggv[0][1][2];
    ggv[2][0][0]=-p[0]*hggs[0][0]-2.*hgs[0];
    ggv[2][0][1]=-p[0]*hggs[0][1]-hgs[1];
    ggv[2][0][2]=-p[0]*hggs[0][2]-hgs[2];
    ggv[2][1][1]=-p[0]*hggs[1][1];
    ggv[2][1][2]=-p[0]*hggs[1][2];
    ggv[2][2][2]=-p[0]*hggs[2][2];
    ggv[2][1][0]=ggv[2][0][1];
    ggv[2][2][0]=ggv[2][0][2];
    ggv[2][2][1]=ggv[2][1][2];
  }
  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<Tensor<1,2> > hg01;
  NumPtr<Tensor<2,2> > hgg01;
  homo_poly_01->computeGrads(p01,hg01);
  homo_poly_01->computeGradGrads(p01,hgg01);
  int hn01=homo_poly_01->getNumber();
  for (int s=0;s<hn01;s++) {
    Tensor<3,3> &ggv=gg[iv++];
    Tensor<1,2> &hgs01=hg01[s];
    Tensor<2,2> &hggs01=hgg01[s];
    ggv[0][0][0]=-p01[1]*hggs01[0][0];
    ggv[0][0][1]=-p01[1]*hggs01[0][1]-hgs01[0];
    ggv[0][1][1]=-p01[1]*hggs01[1][1]-2.*hgs01[1];
    ggv[0][1][0]=ggv[0][0][1];
    ggv[1][0][0]= p01[0]*hggs01[0][0]+2.*hgs01[0];
    ggv[1][0][1]= p01[0]*hggs01[0][1]+hgs01[1];
    ggv[1][1][1]= p01[0]*hggs01[1][1];
    ggv[1][1][0]=ggv[1][0][1];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void NedelecXShapeFunction3::printOn(ostream &os) const {
  os << "NedelecXShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  os << "\thomo_poly:" << endl;
  homo_poly->printOn(os);
  os << "\thomo_poly_01:" << endl;
  homo_poly_01->printOn(os);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDofTensorShapeFunction::printOn(ostream &os) const {
  os << "HcurlDofTensorShapeFunction:" << endl;
  TensorShapeFunction<SPACEDIM>::printOn(os);
}

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlPrismDofShapeFunction::printOn(ostream &os) const {
  os << "HcurlPrismDofShapeFunction:" << endl;
  TensorShapeFunction<3>::printOn(os);
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDualTensorShapeFunction::shapeFunctionsOrder(const Shape *e,
NumPtr<int> &renumber) const {
//since requiresShapeToComputeValues=true, no reordering is needed
#if (SPACEDIM==1)
  int n=getNumberInteriorDofs();
#elif (SPACEDIM==2)
  int n=getNumberInteriorDofs()
       +getNumberEdgeDofs()*e->numberCodimension1Shapes();
#else
  int n=getNumberInteriorDofs()
       +getNumberFaceDofs()*e->numberCodimension1Shapes();
       +getNumberEdgeDofs()*
        dynamic_cast<const Shape3*>(e)->numberCodimension2Shapes();
#endif
  renumber.allocate(n);
  for (int i=0;i<n;i++) renumber[i]=i;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlDualTensorShapeFunction::printOn(ostream &os) const {
  os << "HcurlDualTensorShapeFunction:" << endl;
  TensorShapeFunction<SPACEDIM>::printOn(os);
}

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierTetrahedronShapeFunction::computeValues(
const Point<3> &p,NumPtr<Tensor<1,3> > &v) const {
  v.allocate(n_polys);
  Tensor<1,3> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  polys->computeValues(p,pv);
  int iv=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<pv.getNumber();s++) v[iv++][d]=pv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierTetrahedronShapeFunction::computeGrads(
const Point<3> &p,NumPtr<Tensor<2,3> > &g) const {
  g.allocate(n_polys);
  Tensor<2,3> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,3> > gv;
  polys->computeGrads(p,gv);
  int iv=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<gv.getNumber();s++) g[iv++][d]=gv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierTetrahedronShapeFunction::computeGradGrads(
const Point<3> &p,NumPtr<Tensor<3,3> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,3> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,3> > ggv;
  polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<ggv.getNumber();s++) gg[iv++][d]=ggv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierTetrahedronShapeFunction::printOn(ostream &os) const
{
  os << "HcurlMultiplierTetrahedronShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tpolys = " << polys
     << endl;
  TensorShapeFunction<3>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierPrismShapeFunction::computeValues(
const Point<3> &p,NumPtr<Tensor<1,3> > &v) const {
  v.allocate(n_polys);
  Tensor<1,3> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv01;
  polys01->computeValues(p,pv01);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<pv01.getNumber();s++) v[iv++][d]=pv01[s];
  }
  NumPtr<REAL> pv2;
  polys2->computeValues(p,pv2);
  for (int s=0;s<pv2.getNumber();s++) v[iv++][2]=pv2[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierPrismShapeFunction::computeGrads(
const Point<3> &p,NumPtr<Tensor<2,3> > &g) const {
  g.allocate(n_polys);
  Tensor<2,3> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,3> > gv01;
  polys01->computeGrads(p,gv01);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<gv01.getNumber();s++) g[iv++][d]=gv01[s];
  }
  NumPtr<Tensor<1,3> > gv2;
  polys2->computeGrads(p,gv2);
  for (int s=0;s<gv2.getNumber();s++) g[iv++][2]=gv2[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierPrismShapeFunction::computeGradGrads(
const Point<3> &p,NumPtr<Tensor<3,3> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,3> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,3> > ggv01;
  polys01->computeGradGrads(p,ggv01);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<ggv01.getNumber();s++) gg[iv++][d]=ggv01[s];
  }
  NumPtr<Tensor<2,3> > ggv2;
  polys2->computeGradGrads(p,ggv2);
  for (int s=0;s<ggv2.getNumber();s++) gg[iv++][2]=ggv2[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierPrismShapeFunction::printOn(ostream &os) const
{
  os << "HcurlMultiplierPrismShapeFunction: order01,order2 = "
     << order01 << " " << order2 
     << "\n\tn_polys = " << n_polys
     << "\n\tpolys01,polys2 = " << polys01 << " " << polys2
     << endl;
  TensorShapeFunction<3>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierHexahedronShapeFunction::computeValues(
const Point<3> &p,NumPtr<Tensor<1,3> > &v) const {
  v.allocate(n_polys);
  Tensor<1,3> zero_tensor;
  v.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<3;d++) {
    NumPtr<REAL> pv;
    polys[d]->computeValues(p,pv);
    for (int s=0;s<pv.getNumber();s++) v[iv++][d]=pv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierHexahedronShapeFunction::computeGrads(
const Point<3> &p,NumPtr<Tensor<2,3> > &g) const {
  g.allocate(n_polys);
  Tensor<2,3> zero_tensor;
  g.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<3;d++) {
    NumPtr<Tensor<1,3> > gv;
    polys[d]->computeGrads(p,gv);
    for (int s=0;s<gv.getNumber();s++) g[iv++][d]=gv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierHexahedronShapeFunction::computeGradGrads(
const Point<3> &p,NumPtr<Tensor<3,3> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,3> zero_tensor;
  gg.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<3;d++) {
    NumPtr<Tensor<2,3> > ggv;
    polys[d]->computeGradGrads(p,ggv);
    for (int s=0;s<ggv.getNumber();s++) gg[iv++][d]=ggv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HcurlMultiplierHexahedronShapeFunction::printOn(ostream &os) const
{
  os << "HcurlMultiplierHexahedronShapeFunction: ordera,orderb = "
     << ordera << " " << orderb
     << "\n\tn_polys = " << n_polys
     << "\tpolys = " << polys[0] << " " << polys[1] << " " << polys[2]
     << endl;
  TensorShapeFunction<3>::printOn(os);
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1SimplexShapeFunction::Nedelec1SimplexShapeFunction(int degree) :
order(degree),n_polys(0) {
#if (SPACEDIM==1)
  simplex_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree+1);
  n_polys=0;
#elif (SPACEDIM==2)
  simplex_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
  x_polys=OPERATOR_NEW NedelecXShapeFunction2(degree);
  n_polys=x_polys->getNumber();
# else
  simplex_polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
  x_polys=OPERATOR_NEW NedelecXShapeFunction3(degree);
  n_polys=x_polys->getNumber();
# endif
  n_polys+=simplex_polys->getNumber()*SPACEDIM;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1SimplexShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  simplex_polys->computeValues(p,pv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) v[iv++][d]=pv[s];
  }
#if (SPACEDIM>1)
  NumPtr<Tensor<1,SPACEDIM> > xv;
  x_polys->computeValues(p,xv);
  for (int s=0;s<x_polys->getNumber();s++) v[iv++]=xv[s];
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1SimplexShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > gv;
  simplex_polys->computeGrads(p,gv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) g[iv++][d]=gv[s];
  }
#if (SPACEDIM>1)
  NumPtr<Tensor<2,SPACEDIM> > xg;
  x_polys->computeGrads(p,xg);
  for (int s=0;s<x_polys->getNumber();s++) g[iv++]=xg[s];
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1SimplexShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > ggv;
  simplex_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) gg[iv++][d]=ggv[s];
  }
#if (SPACEDIM>1)
  NumPtr<Tensor<3,SPACEDIM> > xgg;
  x_polys->computeGradGrads(p,xgg);
  for (int s=0;s<x_polys->getNumber();s++) gg[iv++]=xgg[s];
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1SimplexShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1SimplexShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tsimplex_polys:" << endl;
  simplex_polys->printOn(cout);
#if (SPACEDIM>1)
  cout << "\tx_polys:" << endl;
  x_polys->printOn(cout);
#endif
  TensorShapeFunction<SPACEDIM>::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1SimplexDofShapeFunction::Nedelec1SimplexDofShapeFunction(
int degree) : order(degree),n_polys(0),interior_polys(0) {
#if (SPACEDIM==1)
  if (degree<1) {
    interior_polys=0;
    n_polys=0;
  } else {
    interior_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree-1);
    n_polys=interior_polys->getNumber();
  }
#else
  edge_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_edge_polys=edge_polys->getNumber();
# if (SPACEDIM==2)
  interior_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree-1);
  n_polys=2*interior_polys->getNumber();
# else
  face_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree-1);
  n_face_polys=2*face_polys->getNumber();
  if (degree>=2) {
    interior_polys=
      OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree-2);
    n_polys=3*interior_polys->getNumber();
  }
# endif
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1SimplexDofShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> pv;
    interior_polys->computeValues(p,pv);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<interior_polys->getNumber();s++) v[iv++][d]=pv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1SimplexDofShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,SPACEDIM> > gv;
    interior_polys->computeGrads(p,gv);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<interior_polys->getNumber();s++) g[iv++][d]=gv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1SimplexDofShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,SPACEDIM> > ggv;
    interior_polys->computeGradGrads(p,ggv);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<interior_polys->getNumber();s++) gg[iv++][d]=ggv[s];
    }
  }
}
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1SimplexDofShapeFunction::computeEdgeValues(
const Point<1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_edge_polys);
  v.initialize(0.);
  NumPtr<REAL> pv;
  edge_polys->computeValues(p,pv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) v[iv++]=pv[s];
}
#endif
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec1SimplexDofShapeFunction::computeEdgeGrads(
const Point<1> &p,NumPtr<Tensor<1,1> > &g) const {
  g.allocate(n_edge_polys);
  Tensor<1,1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,1> > gv;
  edge_polys->computeGrads(p,gv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) g[iv++]=gv[s];
}
#endif
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec1SimplexDofShapeFunction::computeEdgeGradGrads(
const Point<1> &p,NumPtr<Tensor<2,1> > &gg) const {
  gg.allocate(n_edge_polys);
  Tensor<2,1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,1> > ggv;
  edge_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) gg[iv++]=ggv[s];
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1SimplexDofShapeFunction::computeFaceValues(
const Point<2> &p,NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_face_polys);
  Tensor<1,2> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> fpv;
  face_polys->computeValues(p,fpv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) v[iv++][0]=fpv[s];
  for (int s=0;s<face_polys->getNumber();s++) v[iv++][1]=fpv[s];
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec1SimplexDofShapeFunction::computeFaceGrads(
const Point<2> &p,NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_face_polys);
  Tensor<2,2> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,2> > fgv;
  face_polys->computeGrads(p,fgv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) g[iv++][0]=fgv[s];
  for (int s=0;s<face_polys->getNumber();s++) g[iv++][1]=fgv[s];
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec1SimplexDofShapeFunction::computeFaceGradGrads(
const Point<2> &p,NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_face_polys);
  Tensor<3,2> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,2> > fggv;
  face_polys->computeGradGrads(p,fggv);
  int iv=0;
  for (int s=0;s<face_polys->getNumber();s++) gg[iv++][0]=fggv[s];
  for (int s=0;s<face_polys->getNumber();s++) gg[iv++][1]=fggv[s];
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1SimplexDofShapeFunction::shapeFunctionsOrder(
const Shape *e,NumPtr<int> &renumber) const {
#if (SPACEDIM==1)
  renumber.allocate(n_polys+2);
#elif (SPACEDIM==2)
  renumber.allocate(n_polys+3*n_edge_polys);
#else
  renumber.allocate(n_polys+6*n_edge_polys+4*n_face_polys);
#endif
  int rn=renumber.getNumber();
  for (int fe_s=0;fe_s<rn;fe_s++) renumber[fe_s]=fe_s;
#if (SPACEDIM==2)
  const Triangle *element=dynamic_cast<const Triangle*>(e);
  CHECK_POINTER(element);
  element->reorderFaceDofs(order,renumber);
#elif (SPACEDIM==3)
  const Tetrahedron *element=dynamic_cast<const Tetrahedron*>(e);
  CHECK_POINTER(element);
  NumPtr<int> r=renumber;
  element->reorderLineDofs(order,r);
  r+=6*n_edge_polys;
  if (order>0) element->reorderHcurlFaceDofs(6*n_edge_polys,order-1,r);
#endif
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<rn;i++) {
    if (renumber[i]>=rn || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(rn);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<rn;i++) renumber_inverse[renumber[i]]=i;
  for (int i=0;i<rn;i++) CHECK_BOUNDS(renumber_inverse[i],0,rn);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1SimplexDofShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1SimplexDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
#if (SPACEDIM>1)
     << "\n\tn_edge_polys = " << n_edge_polys 
#endif
#if (SPACEDIM==3)
     << "\n\tn_face_polys = " << n_face_polys 
#endif
     << endl;
#if (SPACEDIM>1)
  cout << "\tedge_polys:" << endl;
  edge_polys->printOn(cout);
#endif
#if (SPACEDIM==3)
  cout << "\tface_polys:" << endl;
  face_polys->printOn(cout);
#endif
  if (interior_polys!=0) {
    cout << "\tinterior_polys:" << endl;
    interior_polys->printOn(cout);
  }
  HcurlDofTensorShapeFunction::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1SimplexDualShapeFunction::Nedelec1SimplexDualShapeFunction(
int degree) : dof_polys(degree),order(degree),n_polys(0),soln_ok(false),
soln(0) {
#if (SPACEDIM==1)
  simplex_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree+1);
  n_polys=0;
#elif (SPACEDIM==2)
  simplex_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
  x_polys=OPERATOR_NEW NedelecXShapeFunction2(degree);
  n_polys=x_polys->getNumber();
#else
  simplex_polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
  x_polys=OPERATOR_NEW NedelecXShapeFunction3(degree);
  n_polys=x_polys->getNumber();
#endif
  n_polys+=simplex_polys->getNumber()*SPACEDIM;
#ifdef DEBUG
//cout << "\tn_polys = " << n_polys << endl;
#endif
  int dpn=dof_polys.getNumber();
#if (SPACEDIM==1)
  CHECK_SAME(n_polys,2+dpn);
#elif (SPACEDIM==2)
  CHECK_SAME(n_polys,dpn+3*dof_polys.getEdgeNumber());
#else
  CHECK_SAME(n_polys,dpn+6*dof_polys.getEdgeNumber()
                        +4*dof_polys.getFaceNumber());
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1SimplexDualShapeFunction::computeCoefficientsFor(
const Shape *element) {
#if (SPACEDIM==1)
  const Interval *simplex=dynamic_cast<const Interval*>(element);
#elif (SPACEDIM==2)
  const Triangle *simplex=dynamic_cast<const Triangle*>(element);
#else
  const Tetrahedron *simplex=dynamic_cast<const Tetrahedron*>(element);
#endif
  CHECK_POINTER(simplex);
  NumPtr<int> renumber; // renumber[fe_s]=dof_s
  dof_polys.shapeFunctionsOrder(simplex,renumber);
  int dpn=dof_polys.getNumber();
  int spn=simplex_polys->getNumber();
#if (SPACEDIM>1)
  int dpen=dof_polys.getEdgeNumber();
  int xpn=x_polys->getNumber();
#endif
#if (SPACEDIM==3)
  int dpfn=dof_polys.getFaceNumber();
#endif
  if (soln!=0) delete soln; soln=0;
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

#if (SPACEDIM==1)
  GaussianQuadrature<1>
    gq(GaussianQuadrature<1>::argGivenOrder(2*order+2));
#else
  GaussianQuadrature<1> 
    egq(GaussianQuadrature<1>::argGivenOrder(2*order+2));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
# if (SPACEDIM==2)
  TriangleGaussianQuadrature 
    gq(TriangleGaussianQuadrature::argGivenOrder(2*order+1));
# else
  TetrahedronGaussianQuadrature 
    gq(TetrahedronGaussianQuadrature::argGivenOrder(2*order));
  TriangleGaussianQuadrature 
    fgq(TriangleGaussianQuadrature::argGivenOrder(2*order+1));
  const NumPtr<Point<2> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
# endif
#endif
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();

  int ioffset=0;
#if (SPACEDIM==1)
  Point<1> p; p[0]=0.;
  NumPtr<REAL> pv;
  simplex_polys->computeValues(p,pv);
  for (int j=0;j<n_polys;j++) matrix(0,j)=pv[j];
  p[0]=1.;
  pv.cleanup();
  simplex_polys->computeValues(p,pv);
  for (int j=0;j<n_polys;j++) matrix(1,j)=pv[j];
  ioffset+=2;
#else
  NumPtr<NumPtr<REAL> > dpev(egq_points.getNumber());
  for (int q=0;q<egq_points.getNumber();q++) {
    dof_polys.computeEdgeValues(egq_points[q],dpev[q]);
  }
  for (int e=0;e<D_CHOOSE(0,3,6);e++) {
    NumPtr<Point<SPACEDIM> > q_points_edge(egq_points.getNumber());
# if (SPACEDIM==2)
    Triangle::staticProjectToFace(e,egq_points,q_points_edge);
    Tensor<1,2> tangent=Triangle::lineTangent(e);
# else
    Tetrahedron::staticProjectToLine(e,egq_points,q_points_edge);
    Tensor<1,3> tangent=Tetrahedron::lineTangent(e);
# endif
    if (simplex->lineTangentPointsBackward(e)) tangent=-tangent;
    for (int q=0;q<egq_points.getNumber();q++) {
      NumPtr<REAL> spev;
      simplex_polys->computeValues(q_points_edge[q],spev);
      REAL wq=egq_weights[q];
      int joffset=0;
      for (int d=0;d<SPACEDIM;d++) {
        for (int j=0;j<spn;j++) {
          REAL spevj=tangent[d]*spev[j]*wq;
          for (int i=0;i<dpen;i++) {
            matrix(renumber[i+ioffset],j+joffset)+=dpev[q][i]*spevj;
          }
        }
        joffset+=spn;
      }
      NumPtr<Tensor<1,SPACEDIM> > xpv;
      x_polys->computeValues(q_points_edge[q],xpv);
      for (int j=0;j<xpn;j++) {
        REAL dot=tangent*xpv[j]*wq;
        for (int i=0;i<dpen;i++) {
          matrix(renumber[i+ioffset],j+joffset)+=dpev[q][i]*dot;
        }
      }
    }
    ioffset+=dpen;
  }
#endif
#if (SPACEDIM==3)
  NumPtr<NumPtr<Tensor<1,2> > > dpfv(fgq_points.getNumber());
  for (int q=0;q<fgq_points.getNumber();q++) {
    dof_polys.computeFaceValues(fgq_points[q],dpfv[q]);
  }
  for (int f=0;f<4;f++) {
    Tensor<1,3> ft0=simplex->faceTangent(f,0);
    Tensor<1,3> ft1=simplex->faceTangent(f,1);
    bool ftwb=simplex->faceTangentsWindBackward(f);
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(fgq_points.getNumber());
    Tetrahedron::staticProjectToFace(f,fgq_points,q_points_face);
    for (int q=0;q<fgq_points.getNumber();q++) {
      REAL wq=fgq_weights[q];
      NumPtr<REAL> spv;
      simplex_polys->computeValues(q_points_face[q],spv);
      int joffset=0;
      for (int d=0;d<3;d++) {
        for (int j=0;j<spn;j++) {
          REAL spvj=spv[j]*wq;
          for (int i=0;i<dpfn;i++) {
            Tensor<1,2> &dpfvqi=dpfv[q][i];
            if (ftwb) {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dpfvqi[1]*ft0[d]+dpfvqi[0]*ft1[d])*spvj;
            } else {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dpfvqi[0]*ft0[d]+dpfvqi[1]*ft1[d])*spvj;
            }
          }
        }
        joffset+=spn;
      }

      NumPtr<Tensor<1,3> > xpv;
      x_polys->computeValues(q_points_face[q],xpv);
      for (int j=0;j<xpn;j++) {
        Tensor<1,3> &xpvj=xpv[j];
        REAL dot0=(xpvj*ft0)*wq;
        REAL dot1=(xpvj*ft1)*wq;
        for (int i=0;i<dpfn;i++) {
          Tensor<1,2> &dpfvqi=dpfv[q][i];
          if (ftwb) {
            matrix(renumber[i+ioffset],j+joffset)+=
              dpfvqi[1]*dot0+dpfvqi[0]*dot1;
          } else {
            matrix(renumber[i+ioffset],j+joffset)+=
              dpfvqi[0]*dot0+dpfvqi[1]*dot1;
          }
        }
      }
    }
    ioffset+=dpfn;
  }
#endif
  NumPtr<NumPtr<Tensor<1,SPACEDIM> > > dpv(gq_points.getNumber());
  for (int q=0;q<gq_points.getNumber();q++) {
    dof_polys.computeValues(gq_points[q],dpv[q]);
  }
  for (int q=0;q<gq_points.getNumber();q++) {
    REAL wq=gq_weights[q];
    NumPtr<REAL> spv;
    const Point<SPACEDIM> &gqpq=gq_points[q];
    simplex_polys->computeValues(gqpq,spv);
    int joffset=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int j=0;j<spn;j++) {
        REAL spvj=spv[j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[q][i][d]*wq*spvj;
        }
      }
      joffset+=spn;
    }
#if (SPACEDIM>1)
    NumPtr<Tensor<1,SPACEDIM> > xpv;
    x_polys->computeValues(gqpq,xpv);
    for (int j=0;j<xpn;j++) {
      Tensor<1,SPACEDIM> &xpvj=xpv[j];
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[q][i]*xpvj;
        matrix(i+ioffset,j+joffset)+=dot*wq;
      }
    }
#endif
  }
  soln=factorNSolve(matrix,rhs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1SimplexDualShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> spv;
  simplex_polys->computeValues(p,spv);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      REAL spvs=spv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=spvs*(*soln)(s+offset,iv);
      }
    }
    offset+=spn;
  }
#if (SPACEDIM>1)
  NumPtr<Tensor<1,SPACEDIM> > xv;
  x_polys->computeValues(p,xv);
  int xpn=x_polys->getNumber();
  for (int s=0;s<xpn;s++) {
    Tensor<1,SPACEDIM> &xvs=xv[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      for (int d=0;d<SPACEDIM;d++) v[iv][d]+=xvs[d]*ssiv;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1SimplexDualShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > spg;
  simplex_polys->computeGrads(p,spg);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      Tensor<1,SPACEDIM> &spgs=spg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int dd=0;dd<SPACEDIM;dd++) givd[dd]+=spgs[dd]*ssiv;
      }
    }
    offset+=spn;
  }
#if (SPACEDIM>1)
  NumPtr<Tensor<2,SPACEDIM> > xpg;
  x_polys->computeGrads(p,xpg);
  int xpn=x_polys->getNumber();
  for (int s=0;s<xpn;s++) {
    Tensor<2,SPACEDIM> &xpgs=xpg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      for (int d=0;d<SPACEDIM;d++) {
        Tensor<1,SPACEDIM> &gvd=g[iv][d];
        Tensor<1,SPACEDIM> &xpgsd=xpgs[d];
        for (int i=0;i<SPACEDIM;i++) gvd[i]+=xpgsd[i]*ssiv;
      }
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1SimplexDualShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > spgg;
  simplex_polys->computeGradGrads(p,spgg);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      Tensor<2,SPACEDIM> &spggs=spgg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            ggivd[i][j]+=spggs[i][j]*ssiv;;
          }
        }
      }
    }
    offset+=spn;
  }
#if (SPACEDIM>1)
  NumPtr<Tensor<3,SPACEDIM> > xpgg;
  x_polys->computeGradGrads(p,xpgg);
  int xpn=x_polys->getNumber();
  for (int s=0;s<xpn;s++) {
    Tensor<3,SPACEDIM> &xpggs=xpgg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      for (int d=0;d<SPACEDIM;d++) {
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        Tensor<2,SPACEDIM> &xpggsd=xpggs[d];
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) ggivd[i][j]+=xpggsd[i][j]*ssiv;
        }
      }
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1SimplexDualShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1SimplexDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tsimplex_polys:" << endl;
  simplex_polys->printOn(cout);
#if (SPACEDIM>1)
  cout << "\tx_polys:" << endl;
  x_polys->printOn(cout);
#endif
  TensorShapeFunction<SPACEDIM>::printOn(os);
}

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1ParallelpipedShapeFunction::Nedelec1ParallelpipedShapeFunction(
int degree) : order(degree),n_polys(0) {
# if (SPACEDIM==2)
  polys[0]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree+1);
  polys[1]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree+1,degree);
# else
  polys[0]=OPERATOR_NEW
    C0LagrangeHexahedronPolynomials(degree,degree+1,degree+1);
  polys[1]=OPERATOR_NEW
    C0LagrangeHexahedronPolynomials(degree+1,degree,degree+1);
  polys[2]=OPERATOR_NEW
    C0LagrangeHexahedronPolynomials(degree+1,degree+1,degree);
# endif
  for (int d=0;d<SPACEDIM;d++) n_polys+=polys[d]->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void Nedelec1ParallelpipedShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<REAL> pv;
    polys[d]->computeValues(p,pv);
    for (int s=0;s<polys[d]->getNumber();s++) v[iv++][d]=pv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec1ParallelpipedShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<1,SPACEDIM> > gv;
    polys[d]->computeGrads(p,gv);
    for (int s=0;s<polys[d]->getNumber();s++) g[iv++][d]=gv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec1ParallelpipedShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<2,SPACEDIM> > ggv;
    polys[d]->computeGradGrads(p,ggv);
    for (int s=0;s<polys[d]->getNumber();s++) gg[iv++][d]=ggv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1ParallelpipedShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1ParallelpipedShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  for (int d=0;d<SPACEDIM;d++) {
    cout << "\tpolys[" << d << "]:" << endl;
    polys[d]->printOn(cout);
  }
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1ParallelpipedDofShapeFunction::
Nedelec1ParallelpipedDofShapeFunction(int degree) : order(degree),
n_polys(0) {
  edge_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_edge_polys=edge_polys->getNumber();
# if (SPACEDIM==3)
  if (degree<1) {
    face_polys[0]=0;
    face_polys[1]=0;
    n_face_polys=0;
  } else {
    face_polys[0]=OPERATOR_NEW
      C0LagrangeQuadrilateralPolynomials(degree,degree-1);
    face_polys[1]=OPERATOR_NEW
      C0LagrangeQuadrilateralPolynomials(degree-1,degree);
    n_face_polys=face_polys[0]->getNumber()+face_polys[1]->getNumber();
  }
# endif
  if (degree<1) {
    for (int d=0;d<SPACEDIM;d++) interior_polys[d]=0;
  } else {
# if (SPACEDIM==2)
    interior_polys[0]=OPERATOR_NEW
      C0LagrangeQuadrilateralPolynomials(degree,degree-1);
    interior_polys[1]=OPERATOR_NEW
      C0LagrangeQuadrilateralPolynomials(degree-1,degree);
# else
    interior_polys[0]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree,degree-1,degree-1);
    interior_polys[1]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree-1,degree,degree-1);
    interior_polys[2]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree-1,degree-1,degree);
# endif
    for (int d=0;d<SPACEDIM;d++) n_polys+=interior_polys[d]->getNumber();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void Nedelec1ParallelpipedDofShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<REAL> pv;
      interior_polys[d]->computeValues(p,pv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        v[iv++][d]=pv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec1ParallelpipedDofShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<Tensor<1,SPACEDIM> > gv;
      interior_polys[d]->computeGrads(p,gv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        g[iv++][d]=gv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec1ParallelpipedDofShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<Tensor<2,SPACEDIM> > ggv;
      interior_polys[d]->computeGradGrads(p,ggv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        gg[iv++][d]=ggv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1ParallelpipedDofShapeFunction::computeEdgeValues(
const Point<1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_edge_polys);
  v.initialize(0.);
  NumPtr<REAL> pv;
  edge_polys->computeValues(p,pv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) v[iv++]=pv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec1ParallelpipedDofShapeFunction::computeEdgeGrads(
const Point<1> &p,NumPtr<Tensor<1,1> > &g) const {
  g.allocate(n_edge_polys);
  Tensor<1,1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,1> > gv;
  edge_polys->computeGrads(p,gv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) g[iv++]=gv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec1ParallelpipedDofShapeFunction::computeEdgeGradGrads(
const Point<1> &p,NumPtr<Tensor<2,1> > &gg) const {
  gg.allocate(n_edge_polys);
  Tensor<2,1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,1> > ggv;
  edge_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) gg[iv++]=ggv[s];
}
# if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1ParallelpipedDofShapeFunction::computeFaceValues(
const Point<2> &p,NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<1,2> zero_tensor;
    v.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<REAL> pv;
      face_polys[d]->computeValues(p,pv);
      for (int s=0;s<face_polys[d]->getNumber();s++) v[iv++][d]=pv[s];
    }
  }
}
# endif
# if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec1ParallelpipedDofShapeFunction::computeFaceGrads(
const Point<2> &p,NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<2,2> zero_tensor;
    g.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<Tensor<1,2> > gv;
      face_polys[d]->computeGrads(p,gv);
      for (int s=0;s<face_polys[d]->getNumber();s++) g[iv++][d]=gv[s];
    }
  }
}
# endif
# if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec1ParallelpipedDofShapeFunction::computeFaceGradGrads(
const Point<2> &p,NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<3,2> zero_tensor;
    gg.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<Tensor<2,2> > ggv;
      face_polys[d]->computeGradGrads(p,ggv);
      for (int s=0;s<face_polys[d]->getNumber();s++) gg[iv++][d]=ggv[s];
    }
  }
}
# endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1ParallelpipedDofShapeFunction::shapeFunctionsOrder(
const Shape *e,NumPtr<int> &renumber) const {
# if (SPACEDIM==2)
  renumber.allocate(n_polys+4*n_edge_polys);
# else
  renumber.allocate(n_polys+12*n_edge_polys+6*n_face_polys);
# endif
  int rn=renumber.getNumber();
  for (int fe_s=0;fe_s<rn;fe_s++) renumber[fe_s]=fe_s;
# if (SPACEDIM==2)
  const Quadrilateral *element=dynamic_cast<const Quadrilateral*>(e);
  CHECK_POINTER(element);
  element->reorderFaceDofs(order,renumber);
# else
  const Hexahedron *element=dynamic_cast<const Hexahedron*>(e);
  CHECK_POINTER(element);
  NumPtr<int> r=renumber;
  element->reorderLineDofs(order,r);
  r+=12*n_edge_polys;
  if (order>0) {
    element->reorderHcurlFaceDofs(12*n_edge_polys,order,order-1,r);
  }
# endif
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<rn;i++) {
    if (renumber[i]>=rn || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(rn);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<rn;i++) renumber_inverse[renumber[i]]=i;
  for (int i=0;i<rn;i++) CHECK_BOUNDS(renumber_inverse[i],0,rn);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1ParallelpipedDofShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1ParallelpipedDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tn_edge_polys = " << n_edge_polys
# if (SPACEDIM==3)
     << "\n\tn_face_polys = " << n_face_polys
# endif
     << endl;
  cout << "\tedge_polys:" << endl;
  edge_polys->printOn(cout);
# if (SPACEDIM==3)
  for (int d=0;d<2;d++) {
    cout << "\tface_polys[" << d << "]:" << endl;
    if (face_polys[d]) face_polys[d]->printOn(cout);
  }
# endif
  for (int d=0;d<SPACEDIM;d++) {
    cout << "\tinterior_polys[" << d << "]:" << endl;
    if (interior_polys[d]) interior_polys[d]->printOn(cout);
  }
  HcurlDofTensorShapeFunction::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1ParallelpipedDualShapeFunction::
Nedelec1ParallelpipedDualShapeFunction(int degree) : dof_polys(degree),
order(degree),n_polys(0),soln_ok(false),soln(0) {
# if (SPACEDIM==2)
  polys[0]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree+1);
  polys[1]=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree+1,degree);
# else
  polys[0]=OPERATOR_NEW
    C0LagrangeHexahedronPolynomials(degree,degree+1,degree+1);
  polys[1]=OPERATOR_NEW
    C0LagrangeHexahedronPolynomials(degree+1,degree,degree+1);
  polys[2]=OPERATOR_NEW
    C0LagrangeHexahedronPolynomials(degree+1,degree+1,degree);
# endif
  for (int d=0;d<SPACEDIM;d++) n_polys+=polys[d]->getNumber();
# if (SPACEDIM==2)
  CHECK_SAME(n_polys,dof_polys.getNumber()+4*dof_polys.getEdgeNumber());
# else
  CHECK_SAME(n_polys,dof_polys.getNumber()+12*dof_polys.getEdgeNumber()
                                          +6*dof_polys.getFaceNumber());
# endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1ParallelpipedDualShapeFunction::computeCoefficientsFor(
const Shape *element) {
# if (SPACEDIM==2)
  const Quadrilateral *parallelpiped=
    dynamic_cast<const Quadrilateral*>(element);
# else
  const Hexahedron *parallelpiped=
    dynamic_cast<const Hexahedron*>(element);
# endif
  CHECK_POINTER(parallelpiped);
  NumPtr<int> renumber; // renumber[fe_s]=dof_s
  dof_polys.shapeFunctionsOrder(parallelpiped,renumber);
  int dpn=dof_polys.getNumber();
  int dpen=dof_polys.getEdgeNumber();
# if (SPACEDIM==3)
  int dpfn=dof_polys.getFaceNumber();
# endif
  if (soln!=0) delete soln; soln=0;
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*order+2));
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();

  GaussianQuadrature<1>
    egq(GaussianQuadrature<1>::argGivenOrder(2*order+2));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
  NumPtr<NumPtr<REAL> > dpev(egq_points.getNumber());
  for (int q=0;q<egq_points.getNumber();q++) {
    dof_polys.computeEdgeValues(egq_points[q],dpev[q]);
  }
  int ioffset=0;
  for (int e=0;e<D_CHOOSE(0,4,12);e++) {
    NumPtr<Point<SPACEDIM> > q_points_edge(egq_points.getNumber());
# if (SPACEDIM==2)
    Quadrilateral::staticProjectToFace(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> tangent=Quadrilateral::lineTangent(e);
# else
    Hexahedron::staticProjectToLine(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> tangent=Hexahedron::lineTangent(e);
# endif
    if (parallelpiped->lineTangentPointsBackward(e)) tangent=-tangent;
    tangent.normalize();
    int joffset=0;
    for (int d=0;d<SPACEDIM;d++) {
      int pdn=polys[d]->getNumber();
      REAL dot=tangent[d];
      for (int q=0;q<egq_points.getNumber();q++) {
        REAL wq=egq_weights[q];
        NumPtr<REAL> pv;
        polys[d]->computeValues(q_points_edge[q],pv);
        for (int j=0;j<pdn;j++) {
          REAL pvj=dot*pv[j]*wq;
          for (int i=0;i<dpen;i++) {
            matrix(renumber[i+ioffset],j+joffset)+=dpev[q][i]*pvj;
          }
        }
      }
      joffset+=pdn;
    }
    ioffset+=dpen;
  }

# if (SPACEDIM==3)
  GaussianQuadrature<2>
    fgq(GaussianQuadrature<2>::argGivenOrder(2*order+2));
  const NumPtr<Point<2> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  NumPtr<Point<3> > q_points_face[6]; 
  for (int f=0;f<6;f++) {
    q_points_face[f].allocate(fgq_points.getNumber());
    Hexahedron::staticProjectToFace(f,fgq_points,q_points_face[f]);
  }
  NumPtr<NumPtr<Tensor<1,2> > > dpfv(fgq_points.getNumber());
  for (int q=0;q<fgq_points.getNumber();q++) {
    dof_polys.computeFaceValues(fgq_points[q],dpfv[q]);
  }
  int dpfn2=dpfn/2;
  for (int f=0;f<6;f++) {
    Tensor<1,3> ft0=parallelpiped->faceTangent(f,0);
    Tensor<1,3> ft1=parallelpiped->faceTangent(f,1);
    ft0.normalize();
    ft1.normalize();
    bool ftwb=parallelpiped->faceTangentsWindBackward(f);
    int joffset=0;
    for (int d=0;d<3;d++) {
      int pdn=polys[d]->getNumber();
      if (abs(ft0[d])>0.) {
        for (int q=0;q<fgq_points.getNumber();q++) {
          REAL wq=fgq_weights[q];
          NumPtr<REAL> pv;
          polys[d]->computeValues(q_points_face[f][q],pv);
          for (int j=0;j<pdn;j++) {
            REAL pvj=ft0[d]*pv[j]*wq;
            for (int i=0;i<dpfn2;i++) {
              REAL dpfvqi=(ftwb ? dpfv[q][i+dpfn2][1] : dpfv[q][i][0]);
              matrix(renumber[i+ioffset],j+joffset)+=dpfvqi*pvj;
            }
          }
        }
      }
      joffset+=pdn;
    }
    ioffset+=dpfn2;
    joffset=0;
    for (int d=0;d<3;d++) {
      int pdn=polys[d]->getNumber();
      if (abs(ft1[d])>0.) {
        for (int q=0;q<fgq_points.getNumber();q++) {
          REAL wq=fgq_weights[q];
          NumPtr<REAL> pv;
          polys[d]->computeValues(q_points_face[f][q],pv);
          for (int j=0;j<pdn;j++) {
            REAL pvj=ft1[d]*pv[j]*wq;
            for (int i=0;i<dpfn2;i++) {
              REAL dpfvqi=(ftwb ? dpfv[q][i][0] : dpfv[q][i+dpfn2][1]);
              matrix(renumber[i+ioffset],j+joffset)+=dpfvqi*pvj;
            }
          }
        }
      }
      joffset+=pdn;
    }
    ioffset+=dpfn2;
  }
# endif

  NumPtr<NumPtr<Tensor<1,SPACEDIM> > > dpv(gq_points.getNumber());
  for (int q=0;q<gq_points.getNumber();q++) {
    dof_polys.computeValues(gq_points[q],dpv[q]);
  }
  int joffset=0;
  int doffset=0;
  int dpns=dpn/SPACEDIM;
  for (int d=0;d<SPACEDIM;d++) {
    int pdn=polys[d]->getNumber();
    for (int q=0;q<gq_points.getNumber();q++) {
      REAL wq=gq_weights[q];
      NumPtr<REAL> pv;
      polys[d]->computeValues(gq_points[q],pv);
      for (int j=0;j<pdn;j++) {
        REAL pvj=pv[j];
        for (int i=0;i<dpns;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[q][i+doffset][d]*wq*pvj;
        }
      }
    }
    joffset+=pdn;
    ioffset+=dpns;
    doffset+=dpns;
  }
  soln=factorNSolve(matrix,rhs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void Nedelec1ParallelpipedDualShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<REAL> pv;
    polys[d]->computeValues(p,pv);
    int pdn=polys[d]->getNumber();
    for (int s=0;s<pdn;s++) {
      REAL pvs=pv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=pvs*(*soln)(s+offset,iv);
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec1ParallelpipedDualShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<1,SPACEDIM> > pg;
    polys[d]->computeGrads(p,pg);
    int pdn=polys[d]->getNumber();
    for (int s=0;s<pdn;s++) {
      Tensor<1,SPACEDIM> &pgs=pg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int i=0;i<SPACEDIM;i++) givd[i]+=pgs[i]*ssiv;
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec1ParallelpipedDualShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<2,SPACEDIM> > pgg;
    polys[d]->computeGradGrads(p,pgg);
    int pdn=polys[d]->getNumber();
    for (int s=0;s<pdn;s++) {
      Tensor<2,SPACEDIM> &pggs=pgg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int i=0;i<SPACEDIM;i++) {
          for (int j=0;j<SPACEDIM;j++) {
            ggivd[i][j]+=pggs[i][j]*ssiv;
          }
        }
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1ParallelpipedDualShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1ParallelpipedDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  for (int d=0;d<SPACEDIM;d++) {
    cout << "\tpolys[" << d << "]:" << endl;
    polys[d]->printOn(cout);
  }
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1PrismShapeFunction::Nedelec1PrismShapeFunction(int degree) :
order(degree),n_polys(0) {
  prism_poly_01=
    OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree+1);
  prism_poly_2=OPERATOR_NEW C0LagrangePrismPolynomials(degree+1,degree);
  homo_poly=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
  mono_poly=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree+1);
  n_polys=2*prism_poly_01->getNumber()+prism_poly_2->getNumber()
         +homo_poly->getNumber()*mono_poly->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1PrismShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> wp01v;
  prism_poly_01->computeValues(p,wp01v);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<prism_poly_01->getNumber();s++) {
      v[iv++][d]=wp01v[s];
    }
  }

  NumPtr<REAL> wp2v;
  prism_poly_2->computeValues(p,wp2v);
  for (int s=0;s<prism_poly_2->getNumber();s++) {
    v[iv++][2]=wp2v[s];
  }

  Point<SPACEDIM-1> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  homo_poly->computeValues(p01,hpv);
  NumPtr<REAL> mpv;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mpv);
  for (int s2=0;s2<mono_poly->getNumber();s2++) {
    REAL mpvs=mpv[s2];
    for (int s01=0;s01<homo_poly->getNumber();s01++) {
      REAL hpvs=hpv[s01];
      Tensor<1,3> &vv=v[iv++];
      vv[0]=-p[1]*hpvs*mpvs;
      vv[1]= p[0]*hpvs*mpvs;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1PrismShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > wp01g;
  prism_poly_01->computeGrads(p,wp01g);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<prism_poly_01->getNumber();s++) g[iv++][d]=wp01g[s];
  }

  NumPtr<Tensor<1,SPACEDIM> > wp2g;
  prism_poly_2->computeGrads(p,wp2g);
  for (int s=0;s<prism_poly_2->getNumber();s++) {
    g[iv++][2]=wp2g[s];
  }

  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,2> > hpg;
  homo_poly->computeValues(p01,hpv);
  homo_poly->computeGrads(p01,hpg);
  NumPtr<REAL> mpv;
  NumPtr<Tensor<1,1> > mpg;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mpv);
  mono_poly->computeGrads(p2,mpg);
  for (int s2=0;s2<mono_poly->getNumber();s2++) {
    REAL mpvs=mpv[s2];
    Tensor<1,1> mpgs=mpg[s2];
    for (int s01=0;s01<homo_poly->getNumber();s01++) {
      REAL &hpvs=hpv[s01];
      Tensor<1,2> &hpgs=hpg[s01];
      Tensor<2,3> &gv=g[iv++];
      gv[0][0]= -p[1]*hpgs[0]*mpvs;
      gv[0][1]=-(p[1]*hpgs[1]+hpvs)*mpvs;
      gv[0][2]= -p[1]*hpvs*mpgs[0];
      gv[1][0]=(p[0]*hpgs[0]+hpvs)*mpvs;
      gv[1][1]= p[0]*hpgs[1]*mpvs;
      gv[1][2]= p[0]*hpvs*mpgs[0];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1PrismShapeFunction::computeGradGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > wp01gg;
  prism_poly_01->computeGradGrads(p,wp01gg);
  int iv=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<prism_poly_01->getNumber();s++) gg[iv++][d]=wp01gg[s];
  }

  NumPtr<Tensor<2,SPACEDIM> > wp2gg;
  prism_poly_2->computeGradGrads(p,wp2gg);
  for (int s=0;s<prism_poly_2->getNumber();s++) {
    gg[iv++][2]=wp2gg[s];
  }
  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,2> > hpg;
  NumPtr<Tensor<2,2> > hpgg;
  homo_poly->computeValues(p01,hpv);
  homo_poly->computeGrads(p01,hpg);
  homo_poly->computeGradGrads(p01,hpgg);
  NumPtr<REAL> mpv;
  NumPtr<Tensor<1,1> > mpg;
  NumPtr<Tensor<2,1> > mpgg;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mpv);
  mono_poly->computeGrads(p2,mpg);
  mono_poly->computeGradGrads(p2,mpgg);
  for (int s2=0;s2<mono_poly->getNumber();s2++) {
    REAL mpvs=mpv[s2];
    Tensor<1,1> mpgs=mpg[s2];
    Tensor<2,1> mpggs=mpgg[s2];
    for (int s01=0;s01<homo_poly->getNumber();s01++) {
      REAL hpvs=hpv[s01];
      Tensor<1,2> &hpgs=hpg[s01];
      Tensor<2,2> &hpggs=hpgg[s01];
      Tensor<3,3> &ggv=gg[iv++];
      ggv[0][0][0]= -p[1]*hpggs[0][0]*mpvs;
      ggv[0][0][1]=-(p[1]*hpggs[0][1]+hpgs[0])*mpvs;
      ggv[0][0][2]= -p[1]*hpgs[0]*mpgs[0];
      ggv[0][1][1]=-(p[1]*hpggs[1][1]+2.*hpgs[1])*mpvs;
      ggv[0][1][2]=-(p[1]*hpgs[1]+hpvs)*mpgs[0];
      ggv[0][2][2]= -p[1]*hpvs*mpggs[0][0];
      ggv[0][1][0]=ggv[0][0][1];
      ggv[0][2][0]=ggv[0][0][2];
      ggv[0][2][1]=ggv[0][1][2];
      ggv[1][0][0]=(p[0]*hpggs[0][0]+2.*hpgs[0])*mpvs;
      ggv[1][0][1]=(p[0]*hpggs[0][1]+hpgs[1])*mpvs;
      ggv[1][0][2]=(p[0]*hpgs[0]+hpvs)*mpgs[0];
      ggv[1][1][1]= p[0]*hpggs[1][1]*mpvs;
      ggv[1][1][2]= p[0]*hpgs[1]*mpgs[0];
      ggv[1][2][2]= p[0]*hpvs*mpggs[0][0];
      ggv[1][1][0]=ggv[1][0][1];
      ggv[1][2][0]=ggv[1][0][2];
      ggv[1][2][1]=ggv[1][1][2];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1PrismShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1PrismShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tprism_poly_01:" << endl;
  prism_poly_01->printOn(cout);
  cout << "\tprism_poly_2:" << endl;
  prism_poly_2->printOn(cout);
  cout << "\thomo_polys:" << endl;
  homo_poly->printOn(cout);
  cout << "\tmono_polys:" << endl;
  mono_poly->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1PrismDofShapeFunction::Nedelec1PrismDofShapeFunction(int degree) :
order(degree),n_polys(0),n_triangle_face_polys(0),
n_quadrilateral_face_polys(0) {
  edge_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_edge_polys=edge_polys->getNumber();
  interior_polys2=0;
  if (degree<1) {
    interior_polys01=0;
    triangle_face_polys=0;
    quadrilateral_face_polys[0]=0;
    quadrilateral_face_polys[1]=0;
    n_polys=0;
  } else {
    interior_polys01=
      OPERATOR_NEW C0LagrangePrismPolynomials(degree-1,degree-1);
    n_polys=2*interior_polys01->getNumber();
    if (degree>=2) {
      interior_polys2=
        OPERATOR_NEW C0LagrangePrismPolynomials(degree-2,degree);
      n_polys+=interior_polys2->getNumber();
    }
    triangle_face_polys=
      OPERATOR_NEW C0LagrangeTrianglePolynomials(degree-1);
    quadrilateral_face_polys[0]=
      OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree-1);
    quadrilateral_face_polys[1]=
      OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree-1,degree);
    n_triangle_face_polys=2*triangle_face_polys->getNumber();
    n_quadrilateral_face_polys=
      2*quadrilateral_face_polys[0]->getNumber();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1PrismDofShapeFunction::computeValues(const Point<3> &p,
NumPtr<Tensor<1,3> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,3> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> wp01v;
    interior_polys01->computeValues(p,wp01v);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<interior_polys01->getNumber();s++) {
        v[iv++][d]=wp01v[s];
      }
    }

    if (interior_polys2) {
      NumPtr<REAL> wp2v;
      interior_polys2->computeValues(p,wp2v);
      for (int s=0;s<interior_polys2->getNumber();s++) {
        v[iv++][2]=wp2v[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1PrismDofShapeFunction::computeGrads(const Point<3> &p,
NumPtr<Tensor<2,3> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,3> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,3> > wp01g;
    interior_polys01->computeGrads(p,wp01g);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<interior_polys01->getNumber();s++) {
        g[iv++][d]=wp01g[s];
      }
    }

    if (interior_polys2) {
      NumPtr<Tensor<1,3> > wp2g;
      interior_polys2->computeGrads(p,wp2g);
      for (int s=0;s<interior_polys2->getNumber();s++) {
        g[iv++][2]=wp2g[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1PrismDofShapeFunction::computeGradGrads(const Point<3> &p,
NumPtr<Tensor<3,3> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,3> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,3> > wp01gg;
    interior_polys01->computeGradGrads(p,wp01gg);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<interior_polys01->getNumber();s++) {
        gg[iv++][d]=wp01gg[s];
      }
    }

    if (interior_polys2) {
      NumPtr<Tensor<2,3> > wp2gg;
      interior_polys2->computeGradGrads(p,wp2gg);
      for (int s=0;s<interior_polys2->getNumber();s++) {
        gg[iv++][2]=wp2gg[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1PrismDofShapeFunction::computeEdgeValues(const Point<1> &p,
NumPtr<REAL> &v) const {
  v.allocate(n_edge_polys);
  v.initialize(0.);
  NumPtr<REAL> epv;
  edge_polys->computeValues(p,epv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) v[iv++]=epv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1PrismDofShapeFunction::computeEdgeGrads(const Point<1> &p,
NumPtr<Tensor<1,1> > &g) const {
  g.allocate(n_edge_polys);
  Tensor<1,1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,1> > epg;
  edge_polys->computeGrads(p,epg);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) g[iv++]=epg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1PrismDofShapeFunction::computeEdgeGradGrads(
const Point<1> &p,NumPtr<Tensor<2,1> > &gg) const {
  gg.allocate(n_edge_polys);
  Tensor<2,1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,1> > epgg;
  edge_polys->computeGradGrads(p,epgg);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) gg[iv++]=epgg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1PrismDofShapeFunction::computeTriangleFaceValues(
const Point<2> &p,NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_triangle_face_polys);
  if (n_triangle_face_polys>0) {
    Tensor<1,2> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> tfpv;
    triangle_face_polys->computeValues(p,tfpv);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<triangle_face_polys->getNumber();s++) {
        v[iv++][d]=tfpv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1PrismDofShapeFunction::computeTriangleFaceGrads(
const Point<2> &p,NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_triangle_face_polys);
  if (n_triangle_face_polys>0) {
    Tensor<2,2> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,2> > tfpg;
    triangle_face_polys->computeGrads(p,tfpg);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<triangle_face_polys->getNumber();s++) {
        g[iv++][d]=tfpg[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1PrismDofShapeFunction::computeTriangleFaceGradGrads(
const Point<2> &p,NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_triangle_face_polys);
  if (n_triangle_face_polys>0) {
    Tensor<3,2> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,2> > tfpgg;
    triangle_face_polys->computeGradGrads(p,tfpgg);
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<triangle_face_polys->getNumber();s++) {
        gg[iv++][d]=tfpgg[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1PrismDofShapeFunction::computeQuadrilateralFaceValues(
const Point<2> &p,NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_quadrilateral_face_polys);
  if (n_quadrilateral_face_polys>0) {
    Tensor<1,2> zero_tensor;
    v.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<REAL> qfpv;
      quadrilateral_face_polys[d]->computeValues(p,qfpv);
      for (int s=0;s<quadrilateral_face_polys[d]->getNumber();s++) {
        v[iv++][d]=qfpv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1PrismDofShapeFunction::computeQuadrilateralFaceGrads(
const Point<2> &p,NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_quadrilateral_face_polys);
  if (n_quadrilateral_face_polys>0) {
    Tensor<2,2> zero_tensor;
    g.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<Tensor<1,2> > qfpg;
      quadrilateral_face_polys[d]->computeGrads(p,qfpg);
      for (int s=0;s<quadrilateral_face_polys[d]->getNumber();s++) {
        g[iv++][d]=qfpg[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1PrismDofShapeFunction::computeQuadrilateralFaceGradGrads(
const Point<2> &p,NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_quadrilateral_face_polys);
  if (n_quadrilateral_face_polys>0) {
    Tensor<3,2> zero_tensor;
    gg.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<Tensor<2,2> > qfpgg;
      quadrilateral_face_polys[d]->computeGradGrads(p,qfpgg);
      for (int s=0;s<quadrilateral_face_polys[d]->getNumber();s++) {
        gg[iv++][d]=qfpgg[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1PrismDofShapeFunction::shapeFunctionsOrder(
const Shape *e,NumPtr<int> &renumber) const {
  renumber.allocate(n_polys+9*n_edge_polys+2*n_triangle_face_polys
                   +3*n_quadrilateral_face_polys);
  int rn=renumber.getNumber();
  for (int fe_s=0;fe_s<rn;fe_s++) renumber[fe_s]=fe_s;
  const Prism *element=dynamic_cast<const Prism*>(e);
  CHECK_POINTER(element);
  NumPtr<int> r=renumber;
  element->reorderLineDofs(order,r);
  r+=9*n_edge_polys;
  element->reorderHcurlFaceDofs(9*n_edge_polys,order,r);
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<rn;i++) {
    if (renumber[i]>=rn || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(rn);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<rn;i++) renumber_inverse[renumber[i]]=i;
  for (int i=0;i<rn;i++) CHECK_BOUNDS(renumber_inverse[i],0,rn);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1PrismDofShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1PrismDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tn_triangle_face_polys = " << n_triangle_face_polys
     << "\n\tn_quadrilateral_face_polys = " << n_quadrilateral_face_polys
     << endl;
  cout << "\tedge_polys:" << endl;
  edge_polys->printOn(cout);
  cout << "\ttriangle_face_polys:" << endl;
  if (triangle_face_polys) triangle_face_polys->printOn(cout);
  for (int d=0;d<2;d++) {
    cout << "\tquadrilateral_face_polys[" << d << "]:" << endl;
    if (quadrilateral_face_polys[d]) {
      quadrilateral_face_polys[d]->printOn(cout);
    }
  }
  cout << "\tinterior_polys01:" << endl;
  if (interior_polys01) interior_polys01->printOn(cout);
  cout << "\tinterior_polys2:" << endl;
  if (interior_polys2) interior_polys2->printOn(cout);
  HcurlPrismDofShapeFunction::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec1PrismDualShapeFunction::Nedelec1PrismDualShapeFunction(
int degree) : dof_polys(degree),order(degree),n_polys(0),soln(0) {
  prism_poly_01=
    OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree+1);
  prism_poly_2=OPERATOR_NEW
    C0LagrangePrismPolynomials(degree+1,degree);
  homo_poly=OPERATOR_NEW HomogeneousLagrangePolynomials2(degree);
  mono_poly=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree+1);
  int wp01n=prism_poly_01->getNumber();
  int wp2n=prism_poly_2->getNumber();
  int hpn=homo_poly->getNumber();
  int mpn=mono_poly->getNumber();
  n_polys=2*wp01n+wp2n+hpn*mpn;
  CHECK_SAME(n_polys,dof_polys.getNumber()
                    +9*dof_polys.getEdgeNumber()
                    +2*dof_polys.getTriangleFaceNumber()
                    +3*dof_polys.getQuadrilateralFaceNumber());
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1PrismDualShapeFunction::computeCoefficientsFor(
const Shape *element) {
  const Prism *prism=dynamic_cast<const Prism*>(element);
  CHECK_POINTER(prism);
  NumPtr<int> renumber; // renumber[fe_s]=dof_s
  dof_polys.shapeFunctionsOrder(prism,renumber);
  int dpn=dof_polys.getNumber();
  int dpen=dof_polys.getEdgeNumber();
  int dptfn=dof_polys.getTriangleFaceNumber();
  int dpqfn=dof_polys.getQuadrilateralFaceNumber();
  int wp01n=prism_poly_01->getNumber();
  int wp2n=prism_poly_2->getNumber();
  int hpn=homo_poly->getNumber();
  int mpn=mono_poly->getNumber();

  if (soln!=0) delete soln; soln=0;
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

  TriangleGaussianQuadrature
    tgq(TriangleGaussianQuadrature::argGivenOrder(2*order+2));
  GaussianQuadrature<1>
    gq1(GaussianQuadrature<1>::argGivenOrder(2*order+2));
  ProductQuadrature pq(tgq,gq1);

  GaussianQuadrature<2>
    qgq(GaussianQuadrature<2>::argGivenOrder(2*order+2));
  const NumPtr<Point<SPACEDIM> > &pq_points=pq.points();
  const NumPtr<REAL> &pq_weights=pq.weights();
  const NumPtr<Point<SPACEDIM-1> > &tgq_points=tgq.points();
  const NumPtr<REAL> &tgq_weights=tgq.weights();
  const NumPtr<Point<SPACEDIM-1> > &qgq_points=qgq.points();
  const NumPtr<REAL> &qgq_weights=qgq.weights();

//edges:
  const NumPtr<Point<1> > &egq_points=gq1.points();
  const NumPtr<REAL> &egq_weights=gq1.weights();
  NumPtr<NumPtr<REAL> > dpev(egq_points.getNumber());
  for (int q=0;q<egq_points.getNumber();q++) {
    dof_polys.computeEdgeValues(egq_points[q],dpev[q]);
  }
  int ioffset=0;
  for (int e=0;e<9;e++) {
    Tensor<1,3> tangent=Prism::lineTangent(e);
    if (prism->lineTangentPointsBackward(e)) tangent=-tangent;
    NumPtr<Point<3> > q_points_edge(egq_points.getNumber());
    Prism::staticProjectToLine(e,egq_points,q_points_edge);
    NumPtr<NumPtr<REAL> >  wp01v;
    wp01v.allocate(egq_points.getNumber());
    for (int q=0;q<egq_points.getNumber();q++) {
      prism_poly_01->computeValues(q_points_edge[q],wp01v[q]);
    }
    int joffset=0;
    for (int d=0;d<2;d++) {
      REAL dot=tangent[d];
      for (int q=0;q<egq_points.getNumber();q++) {
        REAL wq=egq_weights[q];
        NumPtr<REAL> &wp01veq=wp01v[q];
        for (int j=0;j<wp01n;j++) {
          REAL wp01veqj=dot*wp01veq[j]*wq;
          for (int i=0;i<dpen;i++) {
            matrix(renumber[i+ioffset],j+joffset)+=dpev[q][i]*wp01veqj;
          }
        }
      }
      joffset+=wp01n;
    }
    REAL dot=tangent[2];
    for (int q=0;q<egq_points.getNumber();q++) {
      REAL wq=egq_weights[q];
      NumPtr<REAL> wp2v;
      prism_poly_2->computeValues(q_points_edge[q],wp2v);
      for (int j=0;j<wp2n;j++) {
        REAL wp2vj=dot*wp2v[j]*wq;
        for (int i=0;i<dpen;i++) {
          matrix(renumber[i+ioffset],j+joffset)+=dpev[q][i]*wp2vj;
        }
      }
    }
    joffset+=wp2n;
    for (int q=0;q<egq_points.getNumber();q++) {
      const Point<3> &qpeq=q_points_edge[q];
      REAL dot=-tangent[0]*qpeq[1]+tangent[1]*qpeq[0];
      REAL wq=egq_weights[q];
      NumPtr<REAL> hpv;
      Point<2> p01; 
      for (int d=0;d<2;d++) p01[d]=qpeq[d];
      homo_poly->computeValues(p01,hpv);
      NumPtr<REAL> mpv;
      Point<1> qpeq2(qpeq[2]);
      mono_poly->computeValues(qpeq2,mpv);
      for (int j2=0;j2<mpn;j2++) {
        for (int j01=0;j01<hpn;j01++) {
          REAL hpvmpvj=dot*hpv[j01]*mpv[j2]*wq;
          for (int i=0;i<dpen;i++) {
            matrix(renumber[i+ioffset],j01+j2*hpn+joffset)+=
              dpev[q][i]*hpvmpvj;
          }
        }
      }
    }
    ioffset+=dpen;
  }

//triangle faces:
  NumPtr<NumPtr<Tensor<1,2> > > dptfv(tgq_points.getNumber());
  for (int q=0;q<tgq_points.getNumber();q++) {
    dof_polys.computeTriangleFaceValues(tgq_points[q],dptfv[q]);
  }
  for (int f=0;f<2;f++) {
    Tensor<1,3> ft0=prism->faceTangent(f,0);
    Tensor<1,3> ft1=prism->faceTangent(f,1);
    bool ftwb=prism->faceTangentsWindBackward(f);
    NumPtr<Point<SPACEDIM> > tq_points_face(tgq_points.getNumber());
    Prism::staticProjectToFace(f,tgq_points,tq_points_face);
    NumPtr<NumPtr<REAL> > wp01v;
    wp01v.allocate(tgq_points.getNumber());
    for (int q=0;q<tgq_points.getNumber();q++) {
      prism_poly_01->computeValues(tq_points_face[q],wp01v[q]);
    }
    int joffset=0;
    for (int d=0;d<2;d++) {
      for (int q=0;q<tgq_points.getNumber();q++) {
        REAL wq=tgq_weights[q];
        NumPtr<REAL > &wp01vq=wp01v[q];
        for (int j=0;j<wp01n;j++) {
          REAL wp01vqj=wp01vq[j]*wq;
          for (int i=0;i<dptfn;i++) {
            Tensor<1,2> &dptfvqi=dptfv[q][i];
            if (ftwb) {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dptfvqi[1]*ft0[d]+dptfvqi[0]*ft1[d])*wp01vqj;
            } else {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dptfvqi[0]*ft0[d]+dptfvqi[1]*ft1[d])*wp01vqj;
            }
          }
        }
      }
      joffset+=wp01n;
    }
    joffset+=wp2n;
    for (int q=0;q<tgq_points.getNumber();q++) {
      const Point<3> &qp=tq_points_face[q];
      REAL wq=tgq_weights[q];
      NumPtr<REAL> hpv;
      Point<2> qp01; 
      for (int d=0;d<2;d++) qp01[d]=tq_points_face[q][d];
      homo_poly->computeValues(qp01,hpv);
      NumPtr<REAL> mpv;
      Point<1> qp2(qp[2]);
      mono_poly->computeValues(qp2,mpv);
      REAL dot0=-ft0[0]*qp[1]+ft0[1]*qp[0];
      REAL dot1=-ft1[0]*qp[1]+ft1[1]*qp[0];
      for (int j2=0;j2<mpn;j2++) {
        for (int j01=0;j01<hpn;j01++) {
          REAL hpvmpvj=hpv[j01]*mpv[j2]*wq;
          for (int i=0;i<dptfn;i++) {
            Tensor<1,2> &dptfvqi=dptfv[q][i];
            if (ftwb) {
              matrix(renumber[i+ioffset],j01+j2*hpn+joffset)+=
                (dptfvqi[1]*dot0+dptfvqi[0]*dot1)*hpvmpvj;
            } else {
              matrix(renumber[i+ioffset],j01+j2*hpn+joffset)+=
                (dptfvqi[0]*dot0+dptfvqi[1]*dot1)*hpvmpvj;
            }
          }
        }
      }
    }
    ioffset+=dptfn;
  }

//quadrilateral faces:
  NumPtr<NumPtr<Tensor<1,2> > > dpqfv(qgq_points.getNumber());
  for (int q=0;q<qgq_points.getNumber();q++) {
    dof_polys.computeQuadrilateralFaceValues(qgq_points[q],dpqfv[q]);
  }
  for (int f=2;f<5;f++) {
    Tensor<1,3> ft0=prism->faceTangent(f,0);
    Tensor<1,3> ft1=prism->faceTangent(f,1);
    bool ftwb=prism->faceTangentsWindBackward(f);
    NumPtr<Point<SPACEDIM> > qq_points_face(qgq_points.getNumber());
    Prism::staticProjectToFace(f,qgq_points,qq_points_face);
    NumPtr<NumPtr<REAL> > wp01v;
    wp01v.allocate(qgq_points.getNumber());
    for (int q=0;q<qgq_points.getNumber();q++) {
      prism_poly_01->computeValues(qq_points_face[q],wp01v[q]);
    }
    int joffset=0;
    for (int d=0;d<2;d++) {
      for (int q=0;q<qgq_points.getNumber();q++) {
        REAL wq=qgq_weights[q];
        NumPtr<REAL> &wp01vq=wp01v[q];
        for (int j=0;j<wp01n;j++) {
          REAL wp01vqj=wp01vq[j]*wq;
          for (int i=0;i<dpqfn;i++) {
            Tensor<1,2> &dpqfvqi=dpqfv[q][i];
            if (ftwb) {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dpqfvqi[1]*ft0[d]+dpqfvqi[0]*ft1[d])*wp01vqj;
            } else {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dpqfvqi[0]*ft0[d]+dpqfvqi[1]*ft1[d])*wp01vqj;
            }
          }
        }
      }
      joffset+=wp01n;
    }
    for (int q=0;q<qgq_points.getNumber();q++) {
      REAL wq=qgq_weights[q];
      NumPtr<REAL> wp2v;
      prism_poly_2->computeValues(qq_points_face[q],wp2v);
      for (int j=0;j<wp2n;j++) {
        REAL wp2vqj=wp2v[j]*wq;
        for (int i=0;i<dpqfn;i++) {
          Tensor<1,2> &dpqfvqi=dpqfv[q][i];
          if (ftwb) {
            matrix(renumber[i+ioffset],j+joffset)+=
              (dpqfvqi[1]*ft0[2]+dpqfvqi[0]*ft1[2])*wp2vqj;
          } else {
            matrix(renumber[i+ioffset],j+joffset)+=
              (dpqfvqi[0]*ft0[2]+dpqfvqi[1]*ft1[2])*wp2vqj;
          }
        }
      }
    }
    joffset+=wp2n;
    for (int q=0;q<qgq_points.getNumber();q++) {
      const Point<3> &qp=qq_points_face[q];
      REAL wq=qgq_weights[q];
      NumPtr<REAL> hpv;
      Point<2> qp01; 
      for (int d=0;d<2;d++) qp01[d]=qq_points_face[q][d];
      homo_poly->computeValues(qp01,hpv);
      NumPtr<REAL> mpv;
      Point<1> qp2(qp[2]);
      mono_poly->computeValues(qp2,mpv);
      REAL dot0=-ft0[0]*qp[1]+ft0[1]*qp[0];
      REAL dot1=-ft1[0]*qp[1]+ft1[1]*qp[0];
      for (int j2=0;j2<mpn;j2++) {
        for (int j01=0;j01<hpn;j01++) {
          REAL hpvmpvj=hpv[j01]*mpv[j2]*wq;
          for (int i=0;i<dpqfn;i++) {
            Tensor<1,2> &dpqfvqi=dpqfv[q][i];
            if (ftwb) {
              matrix(renumber[i+ioffset],j01+j2*hpn+joffset)+=
                (dpqfvqi[1]*dot0+dpqfvqi[0]*dot1)*hpvmpvj;
            } else {
              matrix(renumber[i+ioffset],j01+j2*hpn+joffset)+=
                (dpqfvqi[0]*dot0+dpqfvqi[1]*dot1)*hpvmpvj;
            }
          }
        }
      }
    }
    ioffset+=dpqfn;
  }

//interior:
  for (int q=0;q<pq_points.getNumber();q++) {
    REAL wq=pq_weights[q];
    NumPtr<REAL> wp01v;
    const Point<SPACEDIM> &qp=pq_points[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    prism_poly_01->computeValues(qp,wp01v);
    int joffset=0;
    for (int d=0;d<2;d++) {
      for (int j=0;j<wp01n;j++) {
        REAL wp01vj=wp01v[j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[i][d]*wq*wp01vj;
        }
      }
      joffset+=wp01n;
    }
    NumPtr<REAL> wp2v;
    prism_poly_2->computeValues(qp,wp2v);
    for (int j=0;j<wp2n;j++) {
      REAL wp2vj=wp2v[j];
      for (int i=0;i<dpn;i++) {
        matrix(i+ioffset,j+joffset)+=dpv[i][2]*wq*wp2vj;
      }
    }
    joffset+=wp2n;
    Point<SPACEDIM-1> qp01; qp01[0]=qp[0]; qp01[1]=qp[1];
    NumPtr<REAL> hpv;
    homo_poly->computeValues(qp01,hpv);
    NumPtr<REAL> mpv;
    Point<1> qp2(qp[2]);
    mono_poly->computeValues(qp2,mpv);
    for (int j2=0;j2<mpn;j2++) {
      for (int j01=0;j01<hpn;j01++) {
        for (int i=0;i<dpn;i++) {
          REAL dot=-dpv[i][0]*qp[1]+dpv[i][1]*qp[0];
          matrix(i+ioffset,j01+joffset)+=dot*wq*hpv[j01]*mpv[j2];
        }
      }
      joffset+=hpn;
    }
  }
  soln=factorNSolve(matrix,rhs); 
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec1PrismDualShapeFunction::computeValues(const Point<3> &p,
NumPtr<Tensor<1,3> > &v) const {
  v.allocate(n_polys);
  Tensor<1,3> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> wp01v;
  prism_poly_01->computeValues(p,wp01v);
  int wp01n=prism_poly_01->getNumber();
  int offset=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<wp01n;s++) {
      REAL wp01vs=wp01v[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=wp01vs*(*soln)(s+offset,iv);
      }
    }
    offset+=wp01n;
  }

  NumPtr<REAL> wp2v;
  prism_poly_2->computeValues(p,wp2v);
  int wp2n=prism_poly_2->getNumber();
  for (int s=0;s<wp2n;s++) {
    REAL wp2vs=wp2v[s];
    for (int iv=0;iv<n_polys;iv++) {
      v[iv][2]+=wp2v[s]*(*soln)(s+offset,iv);
    }
  }
  offset+=wp2n;

  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  homo_poly->computeValues(p01,hpv);
  int hpn=homo_poly->getNumber();
  NumPtr<REAL> mpv;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mpv);
  int mpn=mono_poly->getNumber();
  for (int s2=0;s2<mpn;s2++) {
    for (int s01=0;s01<hpn;s01++) {
      REAL hpvmpv=hpv[s01]*mpv[s2];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s01+offset,iv);
        v[iv][0]+=-p[1]*hpvmpv*ssiv;
        v[iv][1]+= p[0]*hpvmpv*ssiv;
      }
    }
    offset+=hpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec1PrismDualShapeFunction::computeGrads(const Point<3> &p,
NumPtr<Tensor<2,3> > &g) const {
  g.allocate(n_polys);
  Tensor<2,3> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,3> > wp01g;
  prism_poly_01->computeGrads(p,wp01g);
  int wp01n=prism_poly_01->getNumber();
  int offset=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<wp01n;s++) {
      Tensor<1,3> &wp01gs=wp01g[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,3> &givd=g[iv][d];
        for (int j=0;j<3;j++) givd[j]+=wp01gs[j]*ssiv;
      }
    }
    offset+=wp01n;
  }

  NumPtr<Tensor<1,3> > wp2g;
  prism_poly_2->computeGrads(p,wp2g);
  int wp2n=prism_poly_2->getNumber();
  for (int s=0;s<wp2n;s++) {
    Tensor<1,3> &wp2gs=wp2g[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      Tensor<1,3> &giv2=g[iv][2];
      for (int j=0;j<3;j++) giv2[j]+=wp2gs[j]*ssiv;
    }
  }
  offset+=wp2n;

  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,2> > hpg;
  homo_poly->computeValues(p01,hpv);
  homo_poly->computeGrads(p01,hpg);
  int hpn=homo_poly->getNumber();
  NumPtr<REAL> mpv;
  NumPtr<Tensor<1,1> > mpg;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mpv);
  mono_poly->computeGrads(p2,mpg);
  int mpn=mono_poly->getNumber();
  for (int s2=0;s2<mpn;s2++) {
    for (int s01=0;s01<hpn;s01++) {
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s01+offset,iv);
        Tensor<2,3> &gv=g[iv];
          gv[0][0]+= -p[1]*hpg[s01][0]*mpv[s2]*ssiv;
          gv[0][1]+=-(p[1]*hpg[s01][1]+hpv[s01])*mpv[s2]*ssiv;
          gv[0][2]+= -p[1]*hpv[s01]*mpg[s2][0]*ssiv;
          gv[1][0]+=(p[0]*hpg[s01][0]+hpv[s01])*mpv[s2]*ssiv;
          gv[1][1]+= p[0]*hpg[s01][1]*mpv[s2]*ssiv;
          gv[1][2]+= p[0]*hpv[s01]*mpg[s2][0]*ssiv;
      }
    }
    offset+=hpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec1PrismDualShapeFunction::computeGradGrads(const Point<3> &p,
NumPtr<Tensor<3,3> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,3> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,3> > wp01gg;
  prism_poly_01->computeGradGrads(p,wp01gg);
  int wp01n=prism_poly_01->getNumber();
  int offset=0;
  for (int d=0;d<2;d++) {
    for (int s=0;s<wp01n;s++) {
      Tensor<2,3> &wp01ggs=wp01gg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,3> &ggivd=gg[iv][d];
        for (int j=0;j<3;j++) {
          for (int i=0;i<3;i++) ggivd[i][j]+=wp01ggs[i][j]*ssiv;
        }
      }
    }
    offset+=wp01n;
  }

  NumPtr<Tensor<2,3> > wp2gg;
  prism_poly_2->computeGradGrads(p,wp2gg);
  int wp2n=prism_poly_2->getNumber();
  for (int s=0;s<wp2n;s++) {
    Tensor<2,3> &wp2ggs=wp2gg[s];
    for (int iv=0;iv<n_polys;iv++) {
      REAL ssiv=(*soln)(s+offset,iv);
      Tensor<2,3> &ggiv2=gg[iv][2];
      for (int j=0;j<3;j++) {
        for (int i=0;i<3;i++) ggiv2[i][j]+=wp2ggs[i][j]*ssiv;
      }
    }
  }
  offset+=wp2n;

  Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
  NumPtr<REAL> hpv;
  NumPtr<Tensor<1,2> > hpg;
  NumPtr<Tensor<2,2> > hpgg;
  homo_poly->computeValues(p01,hpv);
  homo_poly->computeGrads(p01,hpg);
  homo_poly->computeGradGrads(p01,hpgg);
  int hpn=homo_poly->getNumber();
  NumPtr<REAL> mpv;
  NumPtr<Tensor<1,1> > mpg;
  NumPtr<Tensor<2,1> > mpgg;
  Point<1> p2(p[2]);
  mono_poly->computeValues(p2,mpv);
  mono_poly->computeGrads(p2,mpg);
  mono_poly->computeGradGrads(p2,mpgg);
  int mpn=mono_poly->getNumber();
  for (int s2=0;s2<mpn;s2++) {
    for (int s01=0;s01<hpn;s01++) {
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s01+offset,iv);
        Tensor<3,3> &ggv=gg[iv];
        ggv[0][0][0]+= -p[1]*hpgg[s01][0][0]*mpv[s2]*ssiv;
        ggv[0][0][1]+=-(p[1]*hpgg[s01][0][1]+hpg[s01][0])*mpv[s2]*ssiv;
        ggv[0][0][2]+= -p[1]*hpg[s01][0]*mpg[s2][0]*ssiv;
        ggv[0][1][1]+=-(p[1]*hpgg[s01][1][1]+2.*hpg[s01][1])*mpv[s2]*ssiv;
        ggv[0][1][2]+=-(p[1]*hpg[s01][1]+hpv[s01])*mpg[s2][0]*ssiv;
        ggv[0][2][2]+= -p[1]*hpv[s01]*mpgg[s2][0][0]*ssiv;
        ggv[0][1][0]=ggv[0][0][1];
        ggv[0][2][0]=ggv[0][0][2];
        ggv[0][2][1]=ggv[0][1][2];
        ggv[1][0][0]+=(p[0]*hpgg[s01][0][0]+2.*hpg[s01][0])*mpv[s2]*ssiv;
        ggv[1][0][1]+=(p[0]*hpgg[s01][0][1]+hpg[s01][1])*mpv[s2]*ssiv;
        ggv[1][0][2]+=(p[0]*hpg[s01][0]+hpv[s01])*mpg[s2][0]*ssiv;
        ggv[1][1][1]+= p[0]*hpgg[s01][1][1]*mpv[s2]*ssiv;
        ggv[1][1][2]+= p[0]*hpg[s01][1]*mpg[s2][0]*ssiv;
        ggv[1][2][2]+= p[0]*hpv[s01]*mpgg[s2][0][0]*ssiv;
        ggv[1][1][0]=ggv[1][0][1];
        ggv[1][2][0]=ggv[1][0][2];
        ggv[1][2][1]=ggv[1][1][2];
      }
    }
    offset+=hpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1PrismDualShapeFunction::computeFaceValues(int,
const Point<SPACEDIM-1>&,NumPtr<REAL>&) const {
  OBSOLETE("not programmed");
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec1PrismDualShapeFunction::printOn(ostream &os) const {
  os << "Nedelec1PrismDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tprism_poly_01:" << endl;
  prism_poly_01->printOn(cout);
  cout << "\tprism_poly_2:" << endl;
  prism_poly_2->printOn(cout);
  cout << "\thomo_polys:" << endl;
  homo_poly->printOn(cout);
  cout << "\tmono_polys:" << endl;
  mono_poly->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2SimplexShapeFunction::Nedelec2SimplexShapeFunction(int degree) :
order(degree),n_polys(0) {
  simplex_polys=OPERATOR_NEW 
    D_CHOOSE(C0LagrangeIntervalPolynomials,C0LagrangeTrianglePolynomials,
             C0LagrangeTetrahedronPolynomials)(degree);
  n_polys=simplex_polys->getNumber()*SPACEDIM;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2SimplexShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  simplex_polys->computeValues(p,pv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) v[iv++][d]=pv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2SimplexShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > gv;
  simplex_polys->computeGrads(p,gv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) g[iv++][d]=gv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2SimplexShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > ggv;
  simplex_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<simplex_polys->getNumber();s++) gg[iv++][d]=ggv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2SimplexShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2SimplexShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tsimplex_polys:" << endl;
  simplex_polys->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2SimplexDofShapeFunction::Nedelec2SimplexDofShapeFunction(
int degree) : order(degree),n_polys(0),interior_simplex_polys(0) {
  CHECK_TEST(degree>=1);
#if (SPACEDIM==1)
  if (degree>=2) {
    interior_simplex_polys=
      OPERATOR_NEW C0LagrangeIntervalPolynomials(degree-2);
    n_polys=interior_simplex_polys->getNumber();
  } else {
    interior_simplex_polys=0;
    n_polys=0;
  }
#else
  interior_homo_polys=0;
  edge_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_edge_polys=edge_polys->getNumber();
# if (SPACEDIM==2)
  if (degree>=2) {
    interior_simplex_polys=OPERATOR_NEW
      C0LagrangeTrianglePolynomials(degree-2);
    interior_homo_polys=OPERATOR_NEW
      HomogeneousLagrangePolynomials2(degree-2);
    n_polys=2*interior_simplex_polys->getNumber()
           +interior_homo_polys->getNumber();
  }
# else
  face_simplex_polys=0;
  face_homo_polys=0;
  n_face_polys=0;
  if (degree>=2) {
    face_simplex_polys=OPERATOR_NEW 
      C0LagrangeTrianglePolynomials(degree-2);
    face_homo_polys=OPERATOR_NEW
      HomogeneousLagrangePolynomials2(degree-2);
    n_face_polys=2*face_simplex_polys->getNumber()
                +face_homo_polys->getNumber();
  }
  if (degree>=3) {
    interior_simplex_polys=
      OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree-3);
    interior_homo_polys=OPERATOR_NEW
      HomogeneousLagrangePolynomials3(degree-3);
    n_polys=3*interior_simplex_polys->getNumber()
             +interior_homo_polys->getNumber();
  }
# endif
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2SimplexDofShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> spv;
    interior_simplex_polys->computeValues(p,spv);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<interior_simplex_polys->getNumber();s++) {
        v[iv++][d]=spv[s];
      }
    }
#if (SPACEDIM>1)
    NumPtr<REAL> hpv;
    interior_homo_polys->computeValues(p,hpv);
    for (int s=0;s<interior_homo_polys->getNumber();s++) {
      REAL hpvs=hpv[s];
      Tensor<1,SPACEDIM> &vv=v[iv++];
      for (int d=0;d<SPACEDIM;d++) vv[d]=p[d]*hpvs;
    }
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2SimplexDofShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,SPACEDIM> > spg;
    interior_simplex_polys->computeGrads(p,spg);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<interior_simplex_polys->getNumber();s++) {
        g[iv++][d]=spg[s];
      }
    }
#if (SPACEDIM>1)
    NumPtr<REAL> hpv;
    NumPtr<Tensor<1,SPACEDIM> > hpg;
    interior_homo_polys->computeValues(p,hpv);
    interior_homo_polys->computeGrads(p,hpg);
    for (int s=0;s<interior_homo_polys->getNumber();s++) {
      REAL hpvs=hpv[s];
      Tensor<1,SPACEDIM> &hpgs=hpg[s];
      Tensor<2,SPACEDIM> &gv=g[iv++];
      for (int d=0;d<SPACEDIM;d++) {
        for (int i=0;i<SPACEDIM;i++) gv[d][i]=p[d]*hpgs[i];
        gv[d][d]+=hpvs;
      }
    }
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2SimplexDofShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,SPACEDIM> > spgg;
    interior_simplex_polys->computeGradGrads(p,spgg);
    int nisp=interior_simplex_polys->getNumber();
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int s=0;s<nisp;s++) gg[iv++][d]=spgg[s];
    }
#if (SPACEDIM>1)
    NumPtr<Tensor<1,SPACEDIM> > hpg;
    NumPtr<Tensor<2,SPACEDIM> > hpgg;
    interior_homo_polys->computeGrads(p,hpg);
    interior_homo_polys->computeGradGrads(p,hpgg);
    for (int s=0;s<interior_homo_polys->getNumber();s++) {
      Tensor<1,SPACEDIM> &hpgs=hpg[s];
      Tensor<2,SPACEDIM> &hpggs=hpgg[s];
      Tensor<3,SPACEDIM> &ggv=gg[iv++];
      for (int d=0;d<SPACEDIM;d++) {
        Tensor<2,SPACEDIM> &ggvd=ggv[d];
        for (int i=0;i<SPACEDIM;i++) {
          for (int j=0;j<SPACEDIM;j++) {
            ggvd[i][j]=p[d]*hpggs[i][j];
            if (i==d) ggvd[d][j]+=hpgs[j]; 
            if (j==d) ggvd[i][d]+=hpgs[i]; 
          }
        }
      }
    }
#endif
  }
}
#if (SPACEDIM>1)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2SimplexDofShapeFunction::computeEdgeValues(
const Point<1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_edge_polys);
  v.initialize(0.);
  NumPtr<REAL> pv;
  edge_polys->computeValues(p,pv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) v[iv++]=pv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec2SimplexDofShapeFunction::computeEdgeGrads(
const Point<1> &p,NumPtr<Tensor<1,1> > &g) const {
  g.allocate(n_edge_polys);
  Tensor<1,1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,1> > gv;
  edge_polys->computeGrads(p,gv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) g[iv++]=gv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec2SimplexDofShapeFunction::computeEdgeGradGrads(
const Point<1> &p,NumPtr<Tensor<2,1> > &gg) const {
  gg.allocate(n_edge_polys);
  Tensor<2,1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,1> > ggv;
  edge_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) gg[iv++]=ggv[s];
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2SimplexDofShapeFunction::computeFaceValues(
const Point<2> &p,NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<1,2> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> fspv;
    face_simplex_polys->computeValues(p,fspv);
    int fspn=face_simplex_polys->getNumber();
    int iv=0;
    for (int s=0;s<fspn;s++) v[iv++][0]=fspv[s];
    for (int s=0;s<fspn;s++) v[iv++][1]=fspv[s];
    NumPtr<REAL> fhpv;
    face_homo_polys->computeValues(p,fhpv);
    int fhpn=face_homo_polys->getNumber();
    for (int s=0;s<fhpn;s++) {
      Tensor<1,2> &vv=v[iv++];
      vv[0]=p[0]*fhpv[s];
      vv[1]=p[1]*fhpv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec2SimplexDofShapeFunction::computeFaceGrads(
const Point<2> &p,NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<2,2> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,2> > fspg;
    face_simplex_polys->computeGrads(p,fspg);
    int fspn=face_simplex_polys->getNumber();
    int iv=0;
    for (int s=0;s<fspn;s++) g[iv++][0]=fspg[s];
    for (int s=0;s<fspn;s++) g[iv++][1]=fspg[s];
    NumPtr<REAL> fhpv;
    NumPtr<Tensor<1,2> > fhpg;
    face_homo_polys->computeValues(p,fhpv);
    face_homo_polys->computeGrads(p,fhpg);
    int fhpn=face_homo_polys->getNumber();
    for (int s=0;s<fhpn;s++) {
      Tensor<2,2> &gv=g[iv++];
      const Tensor<1,2> &fhpgs=fhpg[s];
      REAL fhpvs=fhpv[s];
      gv[0][0]=fhpvs+p[0]*fhpgs[0];
      gv[1][0]=p[1]*fhpgs[0];
      gv[0][1]=p[0]*fhpgs[1];
      gv[1][1]=fhpvs+p[1]*fhpgs[1];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec2SimplexDofShapeFunction::computeFaceGradGrads(
const Point<2> &p,NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<3,2> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,2> > fspgg;
    face_simplex_polys->computeGradGrads(p,fspgg);
    int fspn=face_simplex_polys->getNumber();
    int iv=0;
    for (int s=0;s<fspn;s++) gg[iv++][0]=fspgg[s];
    for (int s=0;s<fspn;s++) gg[iv++][1]=fspgg[s];
    NumPtr<Tensor<1,2> > fhpg;
    NumPtr<Tensor<2,2> > fhpgg;
    face_homo_polys->computeGrads(p,fhpg);
    face_homo_polys->computeGradGrads(p,fhpgg);
    int fhpn=face_homo_polys->getNumber();
    for (int s=0;s<fhpn;s++) {
      Tensor<3,2> &ggv=gg[iv++];
      Tensor<1,2> &fhpgs=fhpg[s];
      Tensor<2,2> &fhpggs=fhpgg[s];
      for (int d=0;d<2;d++) {
        ggv[d][0][0]=p[d]*fhpggs[0][0];
        ggv[d][0][1]=p[d]*fhpggs[0][1];
        ggv[d][1][1]=p[d]*fhpggs[1][1];
      }
      ggv[0][0][0]+=2.*fhpgs[0];
      ggv[0][0][1]+=fhpgs[1];
      ggv[0][1][0]=ggv[0][0][1];
      ggv[1][1][1]+=2.*fhpgs[1];
      ggv[1][0][1]+=fhpgs[0];
      ggv[1][1][0]=ggv[1][0][1];
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2SimplexDofShapeFunction::shapeFunctionsOrder(
const Shape *e,NumPtr<int> &renumber) const {
#if (SPACEDIM==1)
  renumber.allocate(n_polys+2);
#elif (SPACEDIM==2)
  renumber.allocate(n_polys+3*n_edge_polys);
#else
  renumber.allocate(n_polys+6*n_edge_polys+4*n_face_polys);
#endif
  int rn=renumber.getNumber();
  for (int fe_s=0;fe_s<rn;fe_s++) renumber[fe_s]=fe_s;
#if (SPACEDIM==2)
  const Triangle *element=dynamic_cast<const Triangle*>(e);
  CHECK_POINTER(element);
  element->reorderFaceDofs(order,renumber);
#elif (SPACEDIM==3)
  const Tetrahedron *element=dynamic_cast<const Tetrahedron*>(e);
  CHECK_POINTER(element);
  NumPtr<int> r=renumber;
  element->reorderLineDofs(order,r);
  r+=6*n_edge_polys;
  if (order>1) {
    dynamic_cast<const Tetrahedron*>(element)->
      reorderHcurlFaceDofs2(6*n_edge_polys,order-2,r);
  }
#endif
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]>=renumber.getNumber() || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(n_polys);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<n_polys;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<n_polys;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,n_polys);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2SimplexDofShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2SimplexDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
#if (SPACEDIM>1)
     << "\n\tn_edge_polys = " << n_edge_polys 
#endif
#if (SPACEDIM==3)
     << "\n\tn_face_polys = " << n_face_polys 
#endif
     << endl;
#if (SPACEDIM>1)
  cout << "\tedge_polys:" << endl;
  edge_polys->printOn(cout);
#endif
#if (SPACEDIM==3)
  cout << "\tface_simplex_polys:" << endl;
  face_simplex_polys->printOn(cout);
  cout << "\tface_homo_polys:" << endl;
  face_homo_polys->printOn(cout);
#endif
  cout << "\tinterior_simplex_polys:" << endl;
  interior_simplex_polys->printOn(cout);
#if (SPACEDIM>1)
  cout << "\tinterior_homo_polys:" << endl;
  interior_homo_polys->printOn(cout);
#endif
  HcurlDofTensorShapeFunction::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2SimplexDualShapeFunction::Nedelec2SimplexDualShapeFunction(
int degree) : dof_polys(degree),order(degree),n_polys(0),soln_ok(false),
soln(0) {
#if (SPACEDIM==1)
  simplex_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
#elif (SPACEDIM==2)
  simplex_polys=OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
#else
  simplex_polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
#endif
  int spn=simplex_polys->getNumber();
  n_polys=spn*SPACEDIM;
  int dpn=dof_polys.getNumber();
#if (SPACEDIM==1)
  CHECK_SAME(n_polys,2+dpn);
#elif (SPACEDIM==2)
  CHECK_SAME(n_polys,dpn+3*dof_polys.getEdgeNumber());
#else
  CHECK_SAME(n_polys,dpn+6*dof_polys.getEdgeNumber()
                        +4*dof_polys.getFaceNumber());
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2SimplexDualShapeFunction::computeCoefficientsFor(
const Shape *element) {
#if (SPACEDIM==1)
  const Interval *simplex=dynamic_cast<const Interval*>(element);
#elif (SPACEDIM==2)
  const Triangle *simplex=dynamic_cast<const Triangle*>(element);
#else
  const Tetrahedron *simplex=dynamic_cast<const Tetrahedron*>(element);
#endif
  CHECK_POINTER(simplex);
  NumPtr<int> renumber; // renumber[fe_s]=dof_s
  dof_polys.shapeFunctionsOrder(simplex,renumber);
  int dpn=dof_polys.getNumber();
  int spn=simplex_polys->getNumber();
#if (SPACEDIM>1)
  int dpen=dof_polys.getEdgeNumber();
#endif
#if (SPACEDIM==3)
  int dpfn=dof_polys.getFaceNumber();
#endif
  if (soln!=0) delete soln; soln=0;
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

#if (SPACEDIM==1)
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*order));
#else
  GaussianQuadrature<1> 
    egq(GaussianQuadrature<1>::argGivenOrder(2*order+1));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
# if (SPACEDIM==2)
  TriangleGaussianQuadrature 
    gq(TriangleGaussianQuadrature::argGivenOrder(2*order));
# else
  TetrahedronGaussianQuadrature 
    gq(TetrahedronGaussianQuadrature::argGivenOrder(2*order-1));
  TriangleGaussianQuadrature 
    fgq(TriangleGaussianQuadrature::argGivenOrder(2*order));
  const NumPtr<Point<2> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
# endif
#endif
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();

  int ioffset=0;
#if (SPACEDIM==1)
  Point<1> p; p[0]=0.;
  NumPtr<REAL> pv;
  simplex_polys->computeValues(p,pv);
  for (int j=0;j<n_polys;j++) matrix(0,j)=pv[j];
  p[0]=1.;
  pv.cleanup();
  simplex_polys->computeValues(p,pv);
  for (int j=0;j<n_polys;j++) matrix(1,j)=pv[j];
  ioffset+=2;
#else
  NumPtr<NumPtr<REAL> > dpev(egq_points.getNumber());
  for (int q=0;q<egq_points.getNumber();q++) {
    dof_polys.computeEdgeValues(egq_points[q],dpev[q]);
  }
  for (int e=0;e<D_CHOOSE(0,3,6);e++) {
    NumPtr<Point<SPACEDIM> > q_points_edge(egq_points.getNumber());
# if (SPACEDIM==2)
    Tensor<1,2> tangent=Triangle::lineTangent(e);
    Triangle::staticProjectToFace(e,egq_points,q_points_edge);
# else
    Tensor<1,3> tangent=Tetrahedron::lineTangent(e);
    Tetrahedron::staticProjectToLine(e,egq_points,q_points_edge);
# endif
    bool ltpb=simplex->lineTangentPointsBackward(e);
    if (ltpb) tangent=-tangent;
    for (int q=0;q<egq_points.getNumber();q++) {
      NumPtr<REAL> spev;
      simplex_polys->computeValues(q_points_edge[q],spev);
      NumPtr<REAL> &dpevq=dpev[q];
      REAL wq=egq_weights[q];
      int joffset=0;
      for (int d=0;d<SPACEDIM;d++) {
        for (int j=0;j<spn;j++) {
          REAL spevj=tangent[d]*spev[j]*wq;
          for (int i=0;i<dpen;i++) {
            matrix(renumber[i+ioffset],j+joffset)+=dpevq[i]*spevj;
          }
        }
        joffset+=spn;
      }
    }
    ioffset+=dpen;
  }
#endif

#if (SPACEDIM==3)
  NumPtr<NumPtr<Tensor<1,2> > > dpfv(fgq_points.getNumber());
  for (int q=0;q<fgq_points.getNumber();q++) {
    const Point<2> &fgq_pointsq=fgq_points[q];
    dof_polys.computeFaceValues(fgq_pointsq,dpfv[q]);
  }
  for (int f=0;f<4;f++) {
    Tensor<1,3> ft0=simplex->faceTangent(f,0);
    Tensor<1,3> ft1=simplex->faceTangent(f,1);
    bool ftwb=simplex->faceTangentsWindBackward(f);
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(fgq_points.getNumber());
    Tetrahedron::staticProjectToFace(f,fgq_points,q_points_face);
    for (int q=0;q<fgq_points.getNumber();q++) {
      REAL wq=fgq_weights[q];
      NumPtr<REAL> spv;
      simplex_polys->computeValues(q_points_face[q],spv);
      NumPtr<Tensor<1,2> > &dpfvq=dpfv[q];
      int joffset=0;
      for (int d=0;d<3;d++) {
        for (int j=0;j<spn;j++) {
          REAL spvj=spv[j]*wq;
          for (int i=0;i<dpfn;i++) {
            Tensor<1,2> &dpfvqi=dpfvq[i];
            if (ftwb) {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dpfvqi[1]*ft0[d]+dpfvqi[0]*ft1[d])*spvj;
            } else {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dpfvqi[0]*ft0[d]+dpfvqi[1]*ft1[d])*spvj;
            }
          }
        }
        joffset+=spn;
      }
    }
    ioffset+=dpfn;
  }
#endif

  NumPtr<NumPtr<Tensor<1,SPACEDIM> > > dpv(gq_points.getNumber());
  for (int q=0;q<gq_points.getNumber();q++) {
    dof_polys.computeValues(gq_points[q],dpv[q]);
  }
  for (int q=0;q<gq_points.getNumber();q++) {
    REAL wq=gq_weights[q];
    NumPtr<REAL> spv;
    const Point<SPACEDIM> &gqpq=gq_points[q];
    simplex_polys->computeValues(gqpq,spv);
    int joffset=0;
    for (int d=0;d<SPACEDIM;d++) {
      for (int j=0;j<spn;j++) {
        REAL spvj=spv[j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[q][i][d]*wq*spvj;
        }
      }
      joffset+=spn;
    }
  }
  soln=factorNSolve(matrix,rhs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2SimplexDualShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> spv;
  simplex_polys->computeValues(p,spv);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      REAL spvs=spv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=spvs*(*soln)(s+offset,iv);
      }
    }
    offset+=spn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2SimplexDualShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > spg;
  simplex_polys->computeGrads(p,spg);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      Tensor<1,SPACEDIM> &spgs=spg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int dd=0;dd<SPACEDIM;dd++) givd[dd]+=spgs[dd]*ssiv;
      }
    }
    offset+=spn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2SimplexDualShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > spgg;
  simplex_polys->computeGradGrads(p,spgg);
  int spn=simplex_polys->getNumber();
  int offset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<spn;s++) {
      Tensor<2,SPACEDIM> &spggs=spgg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            ggivd[i][j]+=spggs[i][j]*ssiv;;
          }
        }
      }
    }
    offset+=spn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2SimplexDualShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2SimplexDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tsimplex_polys:" << endl;
  simplex_polys->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2ParallelpipedShapeFunction::Nedelec2ParallelpipedShapeFunction(
int degree) : order(degree),n_polys(0) {
# if (SPACEDIM==2)
  polys=
    OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree);
# else
  polys=OPERATOR_NEW
    C0LagrangeHexahedronPolynomials(degree,degree,degree);
# endif
  n_polys=SPACEDIM*polys->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void Nedelec2ParallelpipedShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> pv;
  polys->computeValues(p,pv);
  int pn=polys->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) v[iv++][d]=pv[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec2ParallelpipedShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > pg;
  polys->computeGrads(p,pg);
  int pn=polys->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) g[iv++][d]=pg[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec2ParallelpipedShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > pgg;
  polys->computeGradGrads(p,pgg);
  int pn=polys->getNumber();
  int iv=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int s=0;s<pn;s++) gg[iv++][d]=pgg[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2ParallelpipedShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2ParallelpipedShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tpolys:" << endl;
  polys->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2ParallelpipedDofShapeFunction::
Nedelec2ParallelpipedDofShapeFunction(int degree) : order(degree),
n_polys(0) {
  edge_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_edge_polys=edge_polys->getNumber();
# if (SPACEDIM==3)
  if (degree<2) {
    face_polys[0]=0;
    face_polys[1]=0;
    n_face_polys=0;
  } else {
    face_polys[0]=OPERATOR_NEW
      C0LagrangeQuadrilateralPolynomials(degree,degree-2);
    face_polys[1]=OPERATOR_NEW
      C0LagrangeQuadrilateralPolynomials(degree-2,degree);
    n_face_polys=face_polys[0]->getNumber()+face_polys[1]->getNumber();
  }
# endif
  if (degree<2) {
    for (int d=0;d<SPACEDIM;d++) interior_polys[d]=0;
  } else {
# if (SPACEDIM==2)
    interior_polys[0]=OPERATOR_NEW
      C0LagrangeQuadrilateralPolynomials(degree,degree-2);
    interior_polys[1]=OPERATOR_NEW
      C0LagrangeQuadrilateralPolynomials(degree-2,degree);
# else
    interior_polys[0]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree,degree-2,degree-2);
    interior_polys[1]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree-2,degree,degree-2);
    interior_polys[2]=OPERATOR_NEW
      C0LagrangeHexahedronPolynomials(degree-2,degree-2,degree);
# endif
    for (int d=0;d<SPACEDIM;d++) n_polys+=interior_polys[d]->getNumber();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void Nedelec2ParallelpipedDofShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,SPACEDIM> zero_tensor;
    v.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<REAL> pv;
      interior_polys[d]->computeValues(p,pv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        v[iv++][d]=pv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec2ParallelpipedDofShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,SPACEDIM> zero_tensor;
    g.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<Tensor<1,SPACEDIM> > gv;
      interior_polys[d]->computeGrads(p,gv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        g[iv++][d]=gv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec2ParallelpipedDofShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,SPACEDIM> zero_tensor;
    gg.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<SPACEDIM;d++) {
      NumPtr<Tensor<2,SPACEDIM> > ggv;
      interior_polys[d]->computeGradGrads(p,ggv);
      for (int s=0;s<interior_polys[d]->getNumber();s++) {
        gg[iv++][d]=ggv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2ParallelpipedDofShapeFunction::computeEdgeValues(
const Point<1> &p,NumPtr<REAL> &v) const {
  v.allocate(n_edge_polys);
  v.initialize(0.);
  NumPtr<REAL> pv;
  edge_polys->computeValues(p,pv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) v[iv++]=pv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec2ParallelpipedDofShapeFunction::computeEdgeGrads(
const Point<1> &p,NumPtr<Tensor<1,1> > &g) const {
  g.allocate(n_edge_polys);
  Tensor<1,1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,1> > gv;
  edge_polys->computeGrads(p,gv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) g[iv++]=gv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec2ParallelpipedDofShapeFunction::computeEdgeGradGrads(
const Point<1> &p,NumPtr<Tensor<2,1> > &gg) const {
  gg.allocate(n_edge_polys);
  Tensor<2,1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,1> > ggv;
  edge_polys->computeGradGrads(p,ggv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) gg[iv++]=ggv[s];
}
# if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2ParallelpipedDofShapeFunction::computeFaceValues(
const Point<2> &p,NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<1,2> zero_tensor;
    v.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<REAL> pv;
      face_polys[d]->computeValues(p,pv);
      for (int s=0;s<face_polys[d]->getNumber();s++) v[iv++][d]=pv[s];
    }
  }
}
# endif
# if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec2ParallelpipedDofShapeFunction::computeFaceGrads(
const Point<2> &p,NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<2,2> zero_tensor;
    g.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<Tensor<1,2> > gv;
      face_polys[d]->computeGrads(p,gv);
      for (int s=0;s<face_polys[d]->getNumber();s++) g[iv++][d]=gv[s];
    }
  }
}
# endif
# if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec2ParallelpipedDofShapeFunction::computeFaceGradGrads(
const Point<2> &p,NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_face_polys);
  if (n_face_polys>0) {
    Tensor<3,2> zero_tensor;
    gg.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<Tensor<2,2> > ggv;
      face_polys[d]->computeGradGrads(p,ggv);
      for (int s=0;s<face_polys[d]->getNumber();s++) gg[iv++][d]=ggv[s];
    }
  }
}
# endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2ParallelpipedDofShapeFunction::shapeFunctionsOrder(
const Shape *e,NumPtr<int> &renumber) const {
# if (SPACEDIM==2)
  renumber.allocate(n_polys+4*n_edge_polys);
# else
  renumber.allocate(n_polys+12*n_edge_polys+6*n_face_polys);
# endif
  for (int fe_s=0;fe_s<renumber.getNumber();fe_s++) renumber[fe_s]=fe_s;
# if (SPACEDIM==2)
  const Quadrilateral *element=dynamic_cast<const Quadrilateral*>(e);
  CHECK_POINTER(element);
  element->reorderFaceDofs(order,renumber);
# else
  const Hexahedron *element=dynamic_cast<const Hexahedron*>(e);
  CHECK_POINTER(element);
  NumPtr<int> r=renumber;
  element->reorderLineDofs(order,r);
  r+=12*n_edge_polys;
  if (order>=2) {
    element->reorderHcurlFaceDofs(12*n_edge_polys,order,order-2,r);
  }
# endif
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<renumber.getNumber();i++) {
    if (renumber[i]>=renumber.getNumber() || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(n_polys);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<n_polys;i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<n_polys;i++) {
    CHECK_BOUNDS(renumber_inverse[i],0,n_polys);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2ParallelpipedDofShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2ParallelpipedDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tn_edge_polys = " << n_edge_polys
# if (SPACEDIM==3)
     << "\n\tn_face_polys = " << n_face_polys
# endif
     << endl;
  cout << "\tedge_polys:" << endl;
  edge_polys->printOn(cout);
# if (SPACEDIM==3)
  for (int d=0;d<2;d++) {
    cout << "\tface_polys[" << d << "]:" << endl;
    if (face_polys[d]) face_polys[d]->printOn(cout);
  }
# endif
  for (int d=0;d<SPACEDIM;d++) {
    cout << "\tinterior_polys[" << d << "]:" << endl;
    if (interior_polys[d]) interior_polys[d]->printOn(cout);
  }
  HcurlDofTensorShapeFunction::printOn(os);
}
#endif

#if (SPACEDIM>=2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2ParallelpipedDualShapeFunction::
Nedelec2ParallelpipedDualShapeFunction(int degree) : dof_polys(degree),
order(degree),n_polys(0),soln(0) {
  CHECK_TEST(degree>=1);
# if (SPACEDIM==2)
  polys=OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree);
# else
  polys=OPERATOR_NEW
    C0LagrangeHexahedronPolynomials(degree,degree,degree);
# endif
  int pn=polys->getNumber();
  n_polys=SPACEDIM*pn;
# if (SPACEDIM==2)
  CHECK_SAME(n_polys,dof_polys.getNumber()+4*dof_polys.getEdgeNumber());
# else
  CHECK_SAME(n_polys,dof_polys.getNumber()+12*dof_polys.getEdgeNumber()
                                          +6*dof_polys.getFaceNumber());
# endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2ParallelpipedDualShapeFunction::computeCoefficientsFor(
const Shape *element) {
# if (SPACEDIM==2)
  const Quadrilateral *parallelpiped=
    dynamic_cast<const Quadrilateral*>(element);
# else
  const Hexahedron *parallelpiped=
    dynamic_cast<const Hexahedron*>(element);
# endif
  CHECK_POINTER(parallelpiped);
  NumPtr<int> renumber; // renumber[fe_s]=dof_s
  dof_polys.shapeFunctionsOrder(parallelpiped,renumber);
  int pn=polys->getNumber();
  int dpn=dof_polys.getNumber();
  int dpen=dof_polys.getEdgeNumber();
# if (SPACEDIM==3)
  int dpfn=dof_polys.getFaceNumber();
# endif
  if (soln!=0) delete soln; soln=0;
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*order+1));
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();

  int ioffset=0;
  GaussianQuadrature<1>
    egq(GaussianQuadrature<1>::argGivenOrder(2*order+1));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
  NumPtr<Point<SPACEDIM> > q_points_edge[D_CHOOSE(0,4,12)]; 
  for (int e=0;e<D_CHOOSE(0,4,12);e++) {
    q_points_edge[e].allocate(egq_points.getNumber());
# if (SPACEDIM==2)
    Quadrilateral::staticProjectToFace(e,egq_points,q_points_edge[e]);
# else
    Hexahedron::staticProjectToLine(e,egq_points,q_points_edge[e]);
# endif
  }
  NumPtr<NumPtr<REAL> > dpev(egq_points.getNumber());
  for (int q=0;q<egq_points.getNumber();q++) {
    dof_polys.computeEdgeValues(egq_points[q],dpev[q]);
  }
  for (int e=0;e<D_CHOOSE(0,4,12);e++) {
    Tensor<1,SPACEDIM> tangent=
      D_CHOOSE(Interval,Quadrilateral,Hexahedron)::lineTangent(e);
    if (parallelpiped->lineTangentPointsBackward(e)) tangent=-tangent;
    int joffset=0;
    for (int d=0;d<SPACEDIM;d++) {
      REAL dot=tangent[d];
      for (int q=0;q<egq_points.getNumber();q++) {
        REAL wq=egq_weights[q];
        NumPtr<REAL> pv;
        polys->computeValues(q_points_edge[e][q],pv);
        for (int j=0;j<pn;j++) {
          REAL pvj=dot*pv[j]*wq;
          for (int i=0;i<dpen;i++) {
            matrix(renumber[i+ioffset],j+joffset)+=dpev[q][i]*pvj;
          }
        }
      }
      joffset+=pn;
    }
    ioffset+=dpen;
  }

# if (SPACEDIM==3)
  GaussianQuadrature<2>
    fgq(GaussianQuadrature<2>::argGivenOrder(2*order+1));
  const NumPtr<Point<2> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  NumPtr<Point<3> > q_points_face[6]; 
  for (int f=0;f<6;f++) {
    q_points_face[f].allocate(fgq_points.getNumber());
    Hexahedron::staticProjectToFace(f,fgq_points,q_points_face[f]);
  }
  NumPtr<NumPtr<Tensor<1,2> > > dpfv(fgq_points.getNumber());
  for (int q=0;q<fgq_points.getNumber();q++) {
    dof_polys.computeFaceValues(fgq_points[q],dpfv[q]);
  }
  int dpfn2=dpfn/2;
  for (int f=0;f<6;f++) {
    Tensor<1,3> ft0=parallelpiped->faceTangent(f,0);
    Tensor<1,3> ft1=parallelpiped->faceTangent(f,1);
    ft0.normalize();
    ft1.normalize();
    bool ftwb=parallelpiped->faceTangentsWindBackward(f);
    for (int q=0;q<fgq_points.getNumber();q++) {
      REAL wq=fgq_weights[q];
      NumPtr<REAL> pv;
      polys->computeValues(q_points_face[f][q],pv);
      int joffset=0;
      for (int d=0;d<3;d++) {
        if (abs(ft0[d])>0.) {
          for (int j=0;j<pn;j++) {
            REAL pvj=ft0[d]*pv[j]*wq;
            for (int i=0;i<dpfn2;i++) {
              REAL dpfvqi=(ftwb ? dpfv[q][i+dpfn2][1] : dpfv[q][i][0]);
              matrix(renumber[i+ioffset],j+joffset)+=dpfvqi*pvj;
            }
          }
        }
        joffset+=pn;
      }
    }
    ioffset+=dpfn2;
    for (int q=0;q<fgq_points.getNumber();q++) {
      REAL wq=fgq_weights[q];
      NumPtr<REAL> pv;
      polys->computeValues(q_points_face[f][q],pv);
      int joffset=0;
      for (int d=0;d<3;d++) {
        if (abs(ft1[d])>0.) {
          for (int j=0;j<pn;j++) {
            REAL pvj=ft1[d]*pv[j]*wq;
            for (int i=0;i<dpfn2;i++) {
              REAL dpfvqi=(ftwb ? dpfv[q][i][0] : dpfv[q][i+dpfn2][1]);
              matrix(renumber[i+ioffset],j+joffset)+=dpfvqi*pvj;
            }
          }
        }
        joffset+=pn;
      }
    }
    ioffset+=dpfn2;
  }
# endif

  NumPtr<NumPtr<Tensor<1,SPACEDIM> > > dpv(gq_points.getNumber());
  for (int q=0;q<gq_points.getNumber();q++) {
    dof_polys.computeValues(gq_points[q],dpv[q]);
  }
  int joffset=0;
  int doffset=0;
  int dpns=dpn/SPACEDIM;
  for (int d=0;d<SPACEDIM;d++) {
    for (int q=0;q<gq_points.getNumber();q++) {
      REAL wq=gq_weights[q];
      NumPtr<REAL> pv;
      polys->computeValues(gq_points[q],pv);
      for (int j=0;j<pn;j++) {
        REAL pvj=pv[j];
        for (int i=0;i<dpns;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[q][i+doffset][d]*wq*pvj;
        }
      }
    }
    joffset+=pn;
    ioffset+=dpns;
    doffset+=dpns;
  }
  soln=factorNSolve(matrix,rhs);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s] is isomorphic to a Point<SPACEDIM>
void Nedelec2ParallelpipedDualShapeFunction::computeValues(
const Point<SPACEDIM> &p,NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  int offset=0;
  int pdn=polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<REAL> pv;
    polys->computeValues(p,pv);
    for (int s=0;s<pdn;s++) {
      REAL pvs=pv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=pvs*(*soln)(s+offset,iv);
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d] contains all first partial derivatives of v[s][d]
void Nedelec2ParallelpipedDualShapeFunction::computeGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  int offset=0;
  int pdn=polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<1,SPACEDIM> > pg;
    polys->computeGrads(p,pg);
    for (int s=0;s<pdn;s++) {
      Tensor<1,SPACEDIM> &pgs=pg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,SPACEDIM> &givd=g[iv][d];
        for (int i=0;i<SPACEDIM;i++) givd[i]+=pgs[i]*ssiv;
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d] contains all second partial derivatives of v[s][d]
void Nedelec2ParallelpipedDualShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  int offset=0;
  int pdn=polys->getNumber();
  for (int d=0;d<SPACEDIM;d++) {
    NumPtr<Tensor<2,SPACEDIM> > pgg;
    polys->computeGradGrads(p,pgg);
    for (int s=0;s<pdn;s++) {
      Tensor<2,SPACEDIM> &pggs=pgg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,SPACEDIM> &ggivd=gg[iv][d];
        for (int i=0;i<SPACEDIM;i++) {
          for (int j=0;j<SPACEDIM;j++) {
            ggivd[i][j]+=pggs[i][j]*ssiv;
          }
        }
      }
    }
    offset+=pdn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2ParallelpipedDualShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2ParallelpipedDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tpolys:" << endl;
  polys->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2PrismShapeFunction::Nedelec2PrismShapeFunction(int degree) :
order(degree),n_polys(0) {
  prism_poly=OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree);
  n_polys=3*prism_poly->getNumber();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2PrismShapeFunction::computeValues(const Point<SPACEDIM> &p,
NumPtr<Tensor<1,SPACEDIM> > &v) const {
  v.allocate(n_polys);
  Tensor<1,SPACEDIM> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> wpv;
  prism_poly->computeValues(p,wpv);
  int iv=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<prism_poly->getNumber();s++) {
      v[iv++][d]=wpv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2PrismShapeFunction::computeGrads(const Point<SPACEDIM> &p,
NumPtr<Tensor<2,SPACEDIM> > &g) const {
  g.allocate(n_polys);
  Tensor<2,SPACEDIM> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,SPACEDIM> > wpg;
  prism_poly->computeGrads(p,wpg);
  int iv=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<prism_poly->getNumber();s++) g[iv++][d]=wpg[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2PrismShapeFunction::computeGradGrads(
const Point<SPACEDIM> &p,NumPtr<Tensor<3,SPACEDIM> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,SPACEDIM> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,SPACEDIM> > wpgg;
  prism_poly->computeGradGrads(p,wpgg);
  int iv=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<prism_poly->getNumber();s++) gg[iv++][d]=wpgg[s];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2PrismShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2PrismShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tprism_poly:" << endl;
  prism_poly->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2PrismDofShapeFunction::Nedelec2PrismDofShapeFunction(int degree) :
order(degree),n_polys(0),n_triangle_face_polys(0),
n_quadrilateral_face_polys(0) {
  edge_polys=OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  n_edge_polys=edge_polys->getNumber();

  triangle_face_polys=0;
  triangle_face_homo_polys=0;
  quadrilateral_face_polys[0]=0;
  quadrilateral_face_polys[1]=0;
  interior_polys01=0;
  interior_polys01_2=0;
  interior_homo_polys_01=0;
  interior_polys2=0;
  if (degree>=2) {
    triangle_face_polys=
      OPERATOR_NEW C0LagrangeTrianglePolynomials(degree-2);
    triangle_face_homo_polys=
      OPERATOR_NEW HomogeneousLagrangePolynomials2(degree-2);
    n_triangle_face_polys=2*triangle_face_polys->getNumber()
                         +triangle_face_homo_polys->getNumber();
    quadrilateral_face_polys[0]=
      OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree,degree-2);
    quadrilateral_face_polys[1]=
      OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree-2,degree);
    n_quadrilateral_face_polys=quadrilateral_face_polys[0]->getNumber()
                              +quadrilateral_face_polys[1]->getNumber();
    interior_polys01=
      OPERATOR_NEW C0LagrangeTrianglePolynomials(degree-2);
    interior_polys01_2=
      OPERATOR_NEW C0LagrangeIntervalPolynomials(degree-2);
    interior_homo_polys_01=
      OPERATOR_NEW HomogeneousLagrangePolynomials2(degree-2);
    n_polys=(2*interior_polys01->getNumber()
            +interior_homo_polys_01->getNumber())
            *interior_polys01_2->getNumber();
  }
  if (degree>=3) {
    interior_polys2=
      OPERATOR_NEW C0LagrangePrismPolynomials(degree-3,degree);
    n_polys+=interior_polys2->getNumber();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2PrismDofShapeFunction::computeValues(const Point<3> &p,
NumPtr<Tensor<1,3> > &v) const {
  v.allocate(n_polys);
  if (n_polys>0) {
    Tensor<1,3> zero_tensor;
    v.initialize(zero_tensor);
    Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
    NumPtr<REAL> ip01v;
    interior_polys01->computeValues(p01,ip01v);
    int nip01=interior_polys01->getNumber();
    NumPtr<REAL> ip012v;
    Point<1> p2(p[2]);
    interior_polys01_2->computeValues(p2,ip012v);
    int nip012=interior_polys01_2->getNumber();
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s2=0;s2<nip012;s2++) {
        REAL ip012vs=ip012v[s2];
        for (int s01=0;s01<nip01;s01++) {
          v[iv++][d]=ip01v[s01]*ip012vs;
        }
      }
    }
    NumPtr<REAL> ihp01v;
    interior_homo_polys_01->computeValues(p01,ihp01v);
    int nihp01=interior_homo_polys_01->getNumber();
    for (int s2=0;s2<nip012;s2++) {
      REAL ip012vs=ip012v[s2];
      for (int s01=0;s01<nihp01;s01++) {
        Tensor<1,3> &vv=v[iv++];
        REAL ihp01vs=ihp01v[s01];
        vv[0]=p[0]*ihp01vs*ip012vs;
        vv[1]=p[1]*ihp01vs*ip012vs;
      }
    }
    if (interior_polys2) {
      NumPtr<REAL> ip2v;
      interior_polys2->computeValues(p,ip2v);
      int nip2=interior_polys2->getNumber();
      for (int s=0;s<nip2;s++) v[iv++][2]=ip2v[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2PrismDofShapeFunction::computeGrads(const Point<3> &p,
NumPtr<Tensor<2,3> > &g) const {
  g.allocate(n_polys);
  if (n_polys>0) {
    Tensor<2,3> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,3> > wp01g;
    Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
    NumPtr<REAL> ip01v;
    NumPtr<Tensor<1,2> > ip01g;
    interior_polys01->computeValues(p01,ip01v);
    interior_polys01->computeGrads(p01,ip01g);
    int nip01=interior_polys01->getNumber();
    NumPtr<REAL> ip012v;
    NumPtr<Tensor<1,1> > ip012g;
    Point<1> p2(p[2]);
    interior_polys01_2->computeValues(p2,ip012v);
    interior_polys01_2->computeGrads(p2,ip012g);
    int nip012=interior_polys01_2->getNumber();
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s2=0;s2<nip012;s2++) {
        REAL ip012vs=ip012v[s2];
        Tensor<1,1> &ip012gs=ip012g[s2];
        for (int s01=0;s01<nip01;s01++) {
          Tensor<1,3> &gvd=g[iv++][d];
          REAL ip01vs=ip01v[s01];
          Tensor<1,2> &ip01gs=ip01g[s01];
          gvd[0]=ip01gs[0]*ip012vs;
          gvd[1]=ip01gs[1]*ip012vs;
          gvd[2]=ip01vs*ip012gs[0];
        }
      }
    }
    NumPtr<REAL> ihp01v;
    NumPtr<Tensor<1,2> > ihp01g;
    interior_homo_polys_01->computeValues(p01,ihp01v);
    interior_homo_polys_01->computeGrads(p01,ihp01g);
    int nihp01=interior_homo_polys_01->getNumber();
    for (int s2=0;s2<nip012;s2++) {
      REAL ip012vs=ip012v[s2];
      Tensor<1,1> &ip012gs=ip012g[s2];
      for (int s01=0;s01<nihp01;s01++) {
        Tensor<2,3> &gv=g[iv++];
        REAL ihp01vs=ihp01v[s01];
        Tensor<1,2> &ihp01gs=ihp01g[s01];
        gv[0][0]=(p[0]*ihp01gs[0]+ihp01vs)*ip012vs;
        gv[0][1]=p[0]*ihp01gs[1]*ip012vs;
        gv[0][2]=p[0]*ihp01vs*ip012gs[0];
        gv[1][0]=p[1]*ihp01gs[0]*ip012vs;
        gv[1][1]=(p[1]*ihp01gs[1]+ihp01vs)*ip012vs;
        gv[1][2]=p[1]*ihp01vs*ip012gs[0];
      }
    }
    if (interior_polys2) {
      NumPtr<Tensor<1,3> > ip2g;
      interior_polys2->computeGrads(p,ip2g);
      for (int s=0;s<interior_polys2->getNumber();s++) {
        g[iv++][2]=ip2g[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2PrismDofShapeFunction::computeGradGrads(const Point<3> &p,
NumPtr<Tensor<3,3> > &gg) const {
  gg.allocate(n_polys);
  if (n_polys>0) {
    Tensor<3,3> zero_tensor;
    gg.initialize(zero_tensor);
    Point<2> p01; p01[0]=p[0]; p01[1]=p[1];
    NumPtr<REAL> ip01v;
    NumPtr<Tensor<1,2> > ip01g;
    NumPtr<Tensor<2,2> > ip01gg;
    interior_polys01->computeValues(p01,ip01v);
    interior_polys01->computeGrads(p01,ip01g);
    interior_polys01->computeGradGrads(p01,ip01gg);
    int nip01=interior_polys01->getNumber();
    NumPtr<REAL> ip012v;
    NumPtr<Tensor<1,1> > ip012g;
    NumPtr<Tensor<2,1> > ip012gg;
    Point<1> p2(p[2]);
    interior_polys01_2->computeValues(p2,ip012v);
    interior_polys01_2->computeGrads(p2,ip012g);
    interior_polys01_2->computeGradGrads(p2,ip012gg);
    int nip012=interior_polys01_2->getNumber();
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s2=0;s2<nip012;s2++) {
        REAL ip012vs=ip012v[s2];
        Tensor<1,1> &ip012gs=ip012g[s2];
        Tensor<2,1> &ip012ggs=ip012gg[s2];
        for (int s01=0;s01<nip01;s01++) {
          Tensor<2,3> &ggvd=gg[iv++][d];
          REAL ip01vs=ip01v[s01];
          Tensor<1,2> &ip01gs=ip01g[s01];
          Tensor<2,2> &ip01ggs=ip01gg[s01];
          ggvd[0][0]=ip01ggs[0][0]*ip012vs;
          ggvd[0][1]=ip01ggs[0][1]*ip012vs;
          ggvd[0][2]=ip01gs[0]*ip012gs[0];
          ggvd[1][1]=ip01ggs[1][1]*ip012vs;
          ggvd[1][2]=ip01gs[1]*ip012gs[0];
          ggvd[2][2]=ip01vs*ip012ggs[0][0];
          ggvd[1][0]=ggvd[0][1];
          ggvd[2][0]=ggvd[0][2];
          ggvd[2][1]=ggvd[1][2];
        }
      }
    }
    NumPtr<REAL> ihp01v;
    NumPtr<Tensor<1,2> > ihp01g;
    NumPtr<Tensor<2,2> > ihp01gg;
    interior_homo_polys_01->computeValues(p01,ihp01v);
    interior_homo_polys_01->computeGrads(p01,ihp01g);
    interior_homo_polys_01->computeGradGrads(p01,ihp01gg);
    int nihp01=interior_homo_polys_01->getNumber();
    for (int s2=0;s2<nip012;s2++) {
      REAL ip012vs=ip012v[s2];
      Tensor<1,1> &ip012gs=ip012g[s2];
      Tensor<2,1> &ip012ggs=ip012gg[s2];
      for (int s01=0;s01<nihp01;s01++) {
        Tensor<3,3> &ggv=gg[iv++];
        REAL ihp01vs=ihp01v[s01];
        Tensor<1,2> &ihp01gs=ihp01g[s01];
        Tensor<2,2> &ihp01ggs=ihp01gg[s01];
        ggv[0][0][0]=(p[0]*ihp01ggs[0][0]+2.*ihp01gs[0])*ip012vs;
        ggv[0][0][1]=(p[0]*ihp01ggs[0][1]+ihp01gs[1])*ip012vs;
        ggv[0][0][2]=(p[0]*ihp01gs[0]+ihp01vs)*ip012gs[0];
        ggv[0][1][1]=p[0]*ihp01ggs[1][1]*ip012vs;
        ggv[0][1][2]=p[0]*ihp01gs[1]*ip012gs[0];
        ggv[0][2][2]=p[0]*ihp01vs*ip012ggs[0][0];
        ggv[0][1][0]=ggv[0][0][1];
        ggv[0][2][0]=ggv[0][0][2];
        ggv[0][2][1]=ggv[0][1][2];
        ggv[1][0][0]=p[1]*ihp01ggs[0][0]*ip012vs;
        ggv[1][0][1]=(p[1]*ihp01ggs[0][1]+ihp01gs[0])*ip012vs;
        ggv[1][0][2]=p[1]*ihp01gs[0]*ip012gs[0];
        ggv[1][1][1]=(p[1]*ihp01ggs[1][1]+2.*ihp01gs[1])*ip012vs;
        ggv[1][1][2]=(p[1]*ihp01gs[1]+ihp01vs)*ip012gs[0];
        ggv[1][2][2]=p[1]*ihp01vs*ip012ggs[0][0];
        ggv[1][1][0]=ggv[1][0][1];
        ggv[1][2][0]=ggv[1][0][2];
        ggv[1][2][1]=ggv[1][1][2];
      }
    }
    if (interior_polys2) {
      NumPtr<Tensor<2,3> > ip2gg;
      interior_polys2->computeGradGrads(p,ip2gg);
      int nihp2=interior_polys2->getNumber();
      for (int s=0;s<nihp2;s++) {
        gg[iv++][2]=ip2gg[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2PrismDofShapeFunction::computeEdgeValues(const Point<1> &p,
NumPtr<REAL> &v) const {
  v.allocate(n_edge_polys);
  v.initialize(0.);
  NumPtr<REAL> epv;
  edge_polys->computeValues(p,epv);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) v[iv++]=epv[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2PrismDofShapeFunction::computeEdgeGrads(const Point<1> &p,
NumPtr<Tensor<1,1> > &g) const {
  g.allocate(n_edge_polys);
  Tensor<1,1> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,1> > epg;
  edge_polys->computeGrads(p,epg);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) g[iv++]=epg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2PrismDofShapeFunction::computeEdgeGradGrads(
const Point<1> &p,NumPtr<Tensor<2,1> > &gg) const {
  gg.allocate(n_edge_polys);
  Tensor<2,1> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,1> > epgg;
  edge_polys->computeGradGrads(p,epgg);
  int iv=0;
  for (int s=0;s<edge_polys->getNumber();s++) gg[iv++]=epgg[s];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2PrismDofShapeFunction::computeTriangleFaceValues(
const Point<2> &p,NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_triangle_face_polys);
  if (n_triangle_face_polys>0) {
    Tensor<1,2> zero_tensor;
    v.initialize(zero_tensor);
    NumPtr<REAL> tfpv;
    triangle_face_polys->computeValues(p,tfpv);
    int ntfp=triangle_face_polys->getNumber();
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<ntfp;s++) v[iv++][d]=tfpv[s];
    }
    NumPtr<REAL> tfhpv;
    triangle_face_homo_polys->computeValues(p,tfhpv);
    int ntfhp=triangle_face_homo_polys->getNumber();
    for (int s=0;s<ntfhp;s++) {
      Tensor<1,2> &vv=v[iv++];
      vv[0]=p[0]*tfhpv[s];
      vv[1]=p[1]*tfhpv[s];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2PrismDofShapeFunction::computeTriangleFaceGrads(
const Point<2> &p,NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_triangle_face_polys);
  if (n_triangle_face_polys>0) {
    Tensor<2,2> zero_tensor;
    g.initialize(zero_tensor);
    NumPtr<Tensor<1,2> > tfpg;
    triangle_face_polys->computeGrads(p,tfpg);
    int ntfp=triangle_face_polys->getNumber();
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<ntfp;s++) g[iv++][d]=tfpg[s];
    }
    NumPtr<REAL> tfhpv;
    NumPtr<Tensor<1,2> > tfhpg;
    triangle_face_homo_polys->computeValues(p,tfhpv);
    triangle_face_homo_polys->computeGrads(p,tfhpg);
    int ntfhp=triangle_face_homo_polys->getNumber();
    for (int s=0;s<ntfhp;s++) {
      Tensor<2,2> &gv=g[iv++];
      const Tensor<1,2> &tfhpgs=tfhpg[s];
      REAL tfhpvs=tfhpv[s];

      gv[0][0]=tfhpvs+p[0]*tfhpgs[0];
      gv[1][0]=p[1]*tfhpgs[0];
      gv[0][1]=p[0]*tfhpgs[1];
      gv[1][1]=tfhpvs+p[1]*tfhpgs[1];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2PrismDofShapeFunction::computeTriangleFaceGradGrads(
const Point<2> &p,NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_triangle_face_polys);
  if (n_triangle_face_polys>0) {
    Tensor<3,2> zero_tensor;
    gg.initialize(zero_tensor);
    NumPtr<Tensor<2,2> > tfpgg;
    triangle_face_polys->computeGradGrads(p,tfpgg);
    int ntfp=triangle_face_polys->getNumber();
    int iv=0;
    for (int d=0;d<2;d++) {
      for (int s=0;s<ntfp;s++) gg[iv++][d]=tfpgg[s];
    }
    NumPtr<Tensor<1,2> > tfhpg;
    NumPtr<Tensor<2,2> > tfhpgg;
    triangle_face_homo_polys->computeGrads(p,tfhpg);
    triangle_face_homo_polys->computeGradGrads(p,tfhpgg);
    int ntfhp=triangle_face_homo_polys->getNumber();
    for (int s=0;s<ntfhp;s++) {
      Tensor<3,2> &ggv=gg[iv++];
      Tensor<1,2> &tfhpgs=tfhpg[s];
      Tensor<2,2> &tfhpggs=tfhpgg[s];
      for (int d=0;d<2;d++) {
        ggv[d][0][0]=p[d]*tfhpggs[0][0];
        ggv[d][0][1]=p[d]*tfhpggs[0][1];
        ggv[d][1][1]=p[d]*tfhpggs[1][1];
      }
      ggv[0][0][0]+=2.*tfhpgs[0];
      ggv[0][0][1]+=tfhpgs[1];
      ggv[0][1][0]=ggv[0][0][1];
      ggv[1][1][1]+=2.*tfhpgs[1];
      ggv[1][0][1]+=tfhpgs[0];
      ggv[1][1][0]=ggv[1][0][1];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2PrismDofShapeFunction::computeQuadrilateralFaceValues(
const Point<2> &p,NumPtr<Tensor<1,2> > &v) const {
  v.allocate(n_quadrilateral_face_polys);
  if (n_quadrilateral_face_polys>0) {
    Tensor<1,2> zero_tensor;
    v.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<REAL> qfpv;
      quadrilateral_face_polys[d]->computeValues(p,qfpv);
      for (int s=0;s<quadrilateral_face_polys[d]->getNumber();s++) {
        v[iv++][d]=qfpv[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2PrismDofShapeFunction::computeQuadrilateralFaceGrads(
const Point<2> &p,NumPtr<Tensor<2,2> > &g) const {
  g.allocate(n_quadrilateral_face_polys);
  if (n_quadrilateral_face_polys>0) {
    Tensor<2,2> zero_tensor;
    g.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<Tensor<1,2> > qfpg;
      quadrilateral_face_polys[d]->computeGrads(p,qfpg);
      for (int s=0;s<quadrilateral_face_polys[d]->getNumber();s++) {
        g[iv++][d]=qfpg[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2PrismDofShapeFunction::computeQuadrilateralFaceGradGrads(
const Point<2> &p,NumPtr<Tensor<3,2> > &gg) const {
  gg.allocate(n_quadrilateral_face_polys);
  if (n_quadrilateral_face_polys>0) {
    Tensor<3,2> zero_tensor;
    gg.initialize(zero_tensor);
    int iv=0;
    for (int d=0;d<2;d++) {
      NumPtr<Tensor<2,2> > qfpgg;
      quadrilateral_face_polys[d]->computeGradGrads(p,qfpgg);
      for (int s=0;s<quadrilateral_face_polys[d]->getNumber();s++) {
        gg[iv++][d]=qfpgg[s];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2PrismDofShapeFunction::shapeFunctionsOrder(
const Shape *e,NumPtr<int> &renumber) const {
  renumber.allocate(n_polys+9*n_edge_polys+2*n_triangle_face_polys
                   +3*n_quadrilateral_face_polys);
  int rn=renumber.getNumber();
  for (int fe_s=0;fe_s<rn;fe_s++) renumber[fe_s]=fe_s;
  const Prism *prism=dynamic_cast<const Prism*>(e);
  CHECK_POINTER(prism);
  NumPtr<int> r=renumber;
  prism->reorderLineDofs(order,r);
  r+=9*n_edge_polys;
  prism->reorderHcurlFaceDofs2(9*n_edge_polys,order,r);
#ifdef DEBUG
  bool success=true;
  for (int i=0;i<rn;i++) {
    if (renumber[i]>=rn || renumber[i]<0) {
      cout << "\trenumber[" << i << "] = *** " << renumber[i] << " ***"
           << endl;
      success=false;
    }
  }
  CHECK_TEST(success);
  NumPtr<int> renumber_inverse(rn);
  renumber_inverse.initialize(INT_MAX);
  for (int i=0;i<rn;i++) renumber_inverse[renumber[i]]=i;
  for (int i=0;i<rn;i++) CHECK_BOUNDS(renumber_inverse[i],0,rn);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2PrismDofShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2PrismDofShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys
     << "\n\tn_triangle_face_polys = " << n_triangle_face_polys
     << "\n\tn_quadrilateral_face_polys = " << n_quadrilateral_face_polys
     << endl;
  cout << "\tedge_polys:" << endl;
  edge_polys->printOn(cout);
  cout << "\ttriangle_face_polys:" << endl;
  if (triangle_face_polys) triangle_face_polys->printOn(cout);
  for (int d=0;d<2;d++) {
    cout << "\tquadrilateral_face_polys[" << d << "]:" << endl;
    if (quadrilateral_face_polys[d]) {
      quadrilateral_face_polys[d]->printOn(cout);
    }
  }
  cout << "\tinterior_polys01:" << endl;
  if (interior_polys01) interior_polys01->printOn(cout);
  cout << "\tinterior_polys2:" << endl;
  if (interior_polys2) interior_polys2->printOn(cout);
  HcurlPrismDofShapeFunction::printOn(os);
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Nedelec2PrismDualShapeFunction::Nedelec2PrismDualShapeFunction(
int degree) : dof_polys(degree),order(degree),n_polys(0),soln(0) {
  prism_poly=
    OPERATOR_NEW C0LagrangePrismPolynomials(degree,degree);
  int wpn=prism_poly->getNumber();
  n_polys=3*wpn;
  CHECK_SAME(n_polys,dof_polys.getTriangleFaceNumber()*2
                    +dof_polys.getQuadrilateralFaceNumber()*3
                    +9*dof_polys.getEdgeNumber()+dof_polys.getNumber());
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2PrismDualShapeFunction::computeCoefficientsFor(
const Shape *element) {
  const Prism *prism=dynamic_cast<const Prism*>(element);
  CHECK_POINTER(prism);
  NumPtr<int> renumber; // renumber[fe_s]=dof_s
  dof_polys.shapeFunctionsOrder(prism,renumber);

  int wpn=prism_poly->getNumber();
  int dpn=dof_polys.getNumber();
  int dpen=dof_polys.getEdgeNumber();
  int dptfn=dof_polys.getTriangleFaceNumber();
  int dpqfn=dof_polys.getQuadrilateralFaceNumber();
  if (soln!=0) delete soln; soln=0;
  LAMatrix matrix(n_polys,n_polys,0.);
  LAMatrix rhs(n_polys,n_polys,0.);
  for (int i=0;i<n_polys;i++) rhs(i,i)=1.;

  TriangleGaussianQuadrature
    tgq(TriangleGaussianQuadrature::argGivenOrder(2*order+1));
  GaussianQuadrature<1>
    gq1(GaussianQuadrature<1>::argGivenOrder(2*order+1));
  ProductQuadrature pq(tgq,gq1);

  GaussianQuadrature<2>
    qgq(GaussianQuadrature<2>::argGivenOrder(2*order+1));
  const NumPtr<Point<SPACEDIM> > &pq_points=pq.points();
  const NumPtr<REAL> &pq_weights=pq.weights();
  const NumPtr<Point<SPACEDIM-1> > &tgq_points=tgq.points();
  const NumPtr<REAL> &tgq_weights=tgq.weights();
  const NumPtr<Point<SPACEDIM-1> > &qgq_points=qgq.points();
  const NumPtr<REAL> &qgq_weights=qgq.weights();

//edges:
  const NumPtr<Point<1> > &egq_points=gq1.points();
  const NumPtr<REAL> &egq_weights=gq1.weights();
  NumPtr<NumPtr<Point<3> > > q_points_edge(9);
  for (int e=0;e<9;e++) {
    q_points_edge[e].allocate(egq_points.getNumber());
    Prism::staticProjectToLine(e,egq_points,q_points_edge[e]);
  }
  NumPtr<NumPtr<REAL> > dpev(egq_points.getNumber());
  for (int q=0;q<egq_points.getNumber();q++) {
    dof_polys.computeEdgeValues(egq_points[q],dpev[q]);
  }
  int ioffset=0;
  for (int e=0;e<9;e++) {
    Tensor<1,3> tangent=Prism::lineTangent(e);
    if (prism->lineTangentPointsBackward(e)) tangent=-tangent;
//  tangent.normalize();
    NumPtr<NumPtr<REAL> >  wpv;
    wpv.allocate(egq_points.getNumber());
    for (int q=0;q<egq_points.getNumber();q++) {
      prism_poly->computeValues(q_points_edge[e][q],wpv[q]);
    }
    int joffset=0;
    for (int d=0;d<3;d++) {
      REAL dot=tangent[d];
      for (int q=0;q<egq_points.getNumber();q++) {
        REAL wq=egq_weights[q];
        NumPtr<REAL> &wpveq=wpv[q];
        for (int j=0;j<wpn;j++) {
          REAL wpveqj=dot*wpveq[j]*wq;
          for (int i=0;i<dpen;i++) {
            matrix(renumber[i+ioffset],j+joffset)+=dpev[q][i]*wpveqj;
          }
        }
      }
      joffset+=wpn;
    }
    ioffset+=dpen;
  }

//triangle faces:
  NumPtr<NumPtr<Tensor<1,2> > > dptfv(tgq_points.getNumber());
  for (int q=0;q<tgq_points.getNumber();q++) {
    dof_polys.computeTriangleFaceValues(tgq_points[q],dptfv[q]);
  }
  for (int f=0;f<2;f++) {
    Tensor<1,3> ft0=prism->faceTangent(f,0);
    Tensor<1,3> ft1=prism->faceTangent(f,1);
    bool ftwb=prism->faceTangentsWindBackward(f);
    NumPtr<Point<SPACEDIM> > tq_points_face(tgq_points.getNumber());
    Prism::staticProjectToFace(f,tgq_points,tq_points_face);
    NumPtr<NumPtr<REAL> > wpv;
    wpv.allocate(tgq_points.getNumber());
    for (int q=0;q<tgq_points.getNumber();q++) {
      prism_poly->computeValues(tq_points_face[q],wpv[q]);
    }
    int joffset=0;
    for (int d=0;d<3;d++) {
      for (int q=0;q<tgq_points.getNumber();q++) {
        REAL wq=tgq_weights[q];
        NumPtr<REAL > &wpvq=wpv[q];
        for (int j=0;j<wpn;j++) {
          REAL wpvqj=wpvq[j]*wq;
          for (int i=0;i<dptfn;i++) {
            Tensor<1,2> &dptfvqi=dptfv[q][i];
            if (ftwb) {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dptfvqi[1]*ft0[d]+dptfvqi[0]*ft1[d])*wpvqj;
            } else {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dptfvqi[0]*ft0[d]+dptfvqi[1]*ft1[d])*wpvqj;
            }
          }
        }
      }
      joffset+=wpn;
    }
    ioffset+=dptfn;
  }

//quadrilateral faces:
  NumPtr<NumPtr<Tensor<1,2> > > dpqfv(qgq_points.getNumber());
  for (int q=0;q<qgq_points.getNumber();q++) {
    dof_polys.computeQuadrilateralFaceValues(qgq_points[q],dpqfv[q]);
  }
  for (int f=2;f<5;f++) {
    Tensor<1,3> ft0=prism->faceTangent(f,0);
    Tensor<1,3> ft1=prism->faceTangent(f,1);
    bool ftwb=prism->faceTangentsWindBackward(f);
    NumPtr<Point<SPACEDIM> > qq_points_face(qgq_points.getNumber());
    prism->staticProjectToFace(f,qgq_points,qq_points_face);
    NumPtr<NumPtr<REAL> > wpv;
    wpv.allocate(qgq_points.getNumber());
    for (int q=0;q<qgq_points.getNumber();q++) {
      prism_poly->computeValues(qq_points_face[q],wpv[q]);
    }
    int joffset=0;
    for (int d=0;d<3;d++) {
      for (int q=0;q<qgq_points.getNumber();q++) {
        REAL wq=qgq_weights[q];
        NumPtr<REAL> &wpvq=wpv[q];
        for (int j=0;j<wpn;j++) {
          REAL wpvqj=wpvq[j]*wq;
          for (int i=0;i<dpqfn;i++) {
            Tensor<1,2> &dpqfvqi=dpqfv[q][i];
            if (ftwb) {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dpqfvqi[1]*ft0[d]+dpqfvqi[0]*ft1[d])*wpvqj;
            } else {
              matrix(renumber[i+ioffset],j+joffset)+=
                (dpqfvqi[0]*ft0[d]+dpqfvqi[1]*ft1[d])*wpvqj;
            }
          }
        }
      }
      joffset+=wpn;
    }
    ioffset+=dpqfn;
  }

//interior:
  for (int q=0;q<pq_points.getNumber();q++) {
    REAL wq=pq_weights[q];
    NumPtr<REAL> wpv;
    const Point<SPACEDIM> &qp=pq_points[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    prism_poly->computeValues(qp,wpv);
    int joffset=0;
    for (int d=0;d<3;d++) {
      for (int j=0;j<wpn;j++) {
        REAL wpvj=wpv[j];
        for (int i=0;i<dpn;i++) {
          matrix(i+ioffset,j+joffset)+=dpv[i][d]*wq*wpvj;
        }
      }
      joffset+=wpn;
    }
  }
  soln=factorNSolve(matrix,rhs); 
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//v[s][d] = axis_d . v_s
void Nedelec2PrismDualShapeFunction::computeValues(const Point<3> &p,
NumPtr<Tensor<1,3> > &v) const {
  v.allocate(n_polys);
  Tensor<1,3> zero_tensor;
  v.initialize(zero_tensor);
  NumPtr<REAL> wpv;
  prism_poly->computeValues(p,wpv);
  int wpn=prism_poly->getNumber();
  int offset=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<wpn;s++) {
      REAL wpvs=wpv[s];
      for (int iv=0;iv<n_polys;iv++) {
        v[iv][d]+=wpvs*(*soln)(s+offset,iv);
      }
    }
    offset+=wpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//g[s][d][i]=partial ( axis_d . v_s ) / partial xi_i
void Nedelec2PrismDualShapeFunction::computeGrads(const Point<3> &p,
NumPtr<Tensor<2,3> > &g) const {
  g.allocate(n_polys);
  Tensor<2,3> zero_tensor;
  g.initialize(zero_tensor);
  NumPtr<Tensor<1,3> > wpg;
  prism_poly->computeGrads(p,wpg);
  int wpn=prism_poly->getNumber();
  int offset=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<wpn;s++) {
      Tensor<1,3> &wpgs=wpg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<1,3> &givd=g[iv][d];
        for (int j=0;j<3;j++) givd[j]+=wpgs[j]*ssiv;
      }
    }
    offset+=wpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//gg[s][d][i][j]=partial^2 ( axis_d . v_s )
//              / partial xi_i partial xi_j
void Nedelec2PrismDualShapeFunction::computeGradGrads(const Point<3> &p,
NumPtr<Tensor<3,3> > &gg) const {
  gg.allocate(n_polys);
  Tensor<3,3> zero_tensor;
  gg.initialize(zero_tensor);
  NumPtr<Tensor<2,3> > wpgg;
  prism_poly->computeGradGrads(p,wpgg);
  int wpn=prism_poly->getNumber();
  int offset=0;
  for (int d=0;d<3;d++) {
    for (int s=0;s<wpn;s++) {
      Tensor<2,3> &wpggs=wpgg[s];
      for (int iv=0;iv<n_polys;iv++) {
        REAL ssiv=(*soln)(s+offset,iv);
        Tensor<2,3> &ggivd=gg[iv][d];
        for (int j=0;j<3;j++) {
          for (int i=0;i<3;i++) ggivd[i][j]+=wpggs[i][j]*ssiv;
        }
      }
    }
    offset+=wpn;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Nedelec2PrismDualShapeFunction::printOn(ostream &os) const {
  os << "Nedelec2PrismDualShapeFunction: order = " << order
     << "\n\tn_polys = " << n_polys << endl;
  cout << "\tprism_poly:" << endl;
  prism_poly->printOn(cout);
  TensorShapeFunction<SPACEDIM>::printOn(os);
}
#endif

template class TensorShapeFunction<SPACEDIM>;
#if (SPACEDIM==3)
template class TensorShapeFunction<SPACEDIM-1>;
#endif

#include "NumPtr.C"
template class NumPtrBase<Tensor<1,SPACEDIM>,int>;
template class NumPtr<Tensor<1,SPACEDIM> >;
template class NumPtrBase<Tensor<2,SPACEDIM>,int>;
template class NumPtr<Tensor<2,SPACEDIM> >;
template class NumPtrBase<Tensor<3,SPACEDIM>,int>;
template class NumPtr<Tensor<3,SPACEDIM> >;
template class NumPtrBase<NumPtr<Tensor<1,SPACEDIM> >,int>;
template class NumPtr<NumPtr<Tensor<1,SPACEDIM> > >;
#if (SPACEDIM>=2)
template class NumPtrBase<Tensor<1,SPACEDIM-1>,int>;
template class NumPtr<Tensor<1,SPACEDIM-1> >;
template class NumPtrBase<Tensor<2,SPACEDIM-1>,int>;
template class NumPtr<Tensor<2,SPACEDIM-1> >;
#endif
#if (SPACEDIM==3)
template class NumPtrBase<NumPtr<Point<3> >,int>;
template class NumPtr<NumPtr<Point<3> > >;
template class NumPtrBase<Tensor<1,1>,int>;
template class NumPtr<Tensor<1,1> >;
template class NumPtrBase<NumPtr<Tensor<1,2> >,int>;
template class NumPtr<NumPtr<Tensor<1,2> > >;
template class NumPtrBase<Tensor<2,1>,int>;
template class NumPtr<Tensor<2,1> >;
template class NumPtrBase<Tensor<3,2>,int>;
template class NumPtr<Tensor<3,2> >;
#endif
