// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <fstream>
#include <iomanip>
#include <iostream>
#include <stdlib.h>
#include "Shape.H"
#include "Matrix.H"
#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "TensorShapeFunction.H"
#include "Surface.H"
#include "SurfaceDataSet.H"
#include "Tracer.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif
REAL winsize=0.5;

extern "C" {
void F77NAME(dpotrf)(char *uplo,int *n,double *a,int *lda,int *info);
void F77NAME(dpotri)(char *uplo,int *n,double *a,int *lda,int *info);
}

void debugPoly(const TensorShapeFunction<SPACEDIM> *pols,int degree) {
//TRACER_CALL(t,"debugPoly");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int n=pols->getNumber();
  const RTSimplexShapeFunction *rtssf=
    dynamic_cast<const RTSimplexShapeFunction*>(pols);
  if (rtssf) {
//  TRACER_CALL(t,"debugPoly RTSimplexShapeFunction");
//  lagrange polynomials:
    NumPtr<Point<SPACEDIM> > points;
#if (SPACEDIM==1)
    Interval::staticEquallySpacedLatticePoints(degree+1,points);
#elif (SPACEDIM==2)
    Triangle::staticEquallySpacedLatticePoints(degree,points);
#else
    Tetrahedron::staticEquallySpacedLatticePoints(degree,points);
#endif
    int pn=points.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtssf->computeValues(points[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s][0]-1.) : abs(v[s][0]));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error[0] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
      }
#if (SPACEDIM>=2)
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+pn][1]-1.) : abs(v[s+pn][1]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s+pn] << ", error[1] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
#endif
#if (SPACEDIM==3)
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+2*pn][2]-1.) : abs(v[s+2*pn][2]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s+2*pn] << ", error[2] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
#endif
    }
#if (SPACEDIM>=2)
//  homogeneous polynomials:
    Point<SPACEDIM> point;
    for (int d=0;d<SPACEDIM;d++) point[d]=drand48();
    NumPtr<Tensor<1,SPACEDIM> > v;
    rtssf->computeValues(point,v);
    for (int s=SPACEDIM*pn;s<n;s++) {
      REAL error=v[s][1]/v[s][0]-point[1]/point[0];
      if (error>DBL_EPSILON*512.*abs(point[1]/point[0])) {
        cout << "\tcomputeValue(" << s << "," << point
             << ") = " << v[s] << ", error = " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
# if (SPACEDIM==3)
      error=v[s][2]/v[s][0]-point[2]/point[0];
      if (error>DBL_EPSILON*512.*abs(point[2]/point[0])) {
        cout << "\tcomputeValue(" << s << "," << point
             << ") = " << v[s] << ", error = " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
# endif
//    if (s<n-1) {
//      error=v[s+1][0]/v[s][0]-point[0]/point[1];
//      if (error>DBL_EPSILON*512.) {
//        cout << "\tcomputeValue(" << s << "," << point
//             << ") = " << v[s] << " " << v[s+1] << ", error = "
//             << error << " " << error/DBL_EPSILON << endl;
//        CHECK_TEST(error<=DBL_EPSILON*512.);
//      }
//    }
    }
#endif
  }
  const RTSimplexDofShapeFunction *rtsdsf=
    dynamic_cast<const RTSimplexDofShapeFunction*>(pols);
  if (rtsdsf) {
//  TRACER_CALL(t,"debugPoly RTSimplexDofShapeFunction");
//  lagrange polynomials:
    NumPtr<Point<SPACEDIM> > points;
#if (SPACEDIM==1)
    Interval::staticEquallySpacedLatticePoints(degree-1,points);
#elif (SPACEDIM==2)
    Triangle::staticEquallySpacedLatticePoints(degree-1,points);
#else
    Tetrahedron::staticEquallySpacedLatticePoints(degree-1,points);
#endif
    int pn=points.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtsdsf->computeValues(points[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s][0]-1.) : abs(v[s][0]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error[0] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
#if (SPACEDIM>=2)
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+pn][1]-1.) : abs(v[s+pn][1]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s+pn] << ", error[1] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
#endif
#if (SPACEDIM==3)
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+2*pn][2]-1.) : abs(v[s+2*pn][2]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s+2*pn] << ", error[2] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
#endif
    }
#if (SPACEDIM>=2)
    NumPtr<Point<SPACEDIM-1> > points1;
# if (SPACEDIM==2)
    Interval::staticEquallySpacedLatticePoints(degree,points1);
# else
    Triangle::staticEquallySpacedLatticePoints(degree,points1);
# endif
    for (int pi=0;pi<points1.getNumber();pi++) {
      NumPtr<REAL> v;
      rtsdsf->computeFaceValues(points1[pi],v);
      for (int s=0;s<points1.getNumber();s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
#endif
  }

#if (SPACEDIM>=2)
  const RTParallelpipedShapeFunction *rtpsf=
    dynamic_cast<const RTParallelpipedShapeFunction*>(pols);
  if (rtpsf) {
//  TRACER_CALL(t,"debugPoly RTParallelpipedShapeFunction");
    NumPtr<Point<SPACEDIM> > points;
    D_CHOOSE(Interval,Quadrilateral,Hexahedron)::
      staticEquallySpacedLatticePoints(D_DECL(degree+1,degree,degree),
      points);
    int pn=points.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtpsf->computeValues(points[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s][0]-1.) : abs(v[s][0]));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error[0] = "
               << error << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
        for (int d=1;d<SPACEDIM;d++) {
          REAL error=abs(v[s][d]);
          if (error>DBL_EPSILON*1024.) {
            cout << "\tcomputeValue(" << s << "," << points[pi] 
                 << ") = " << v[s] << ", error[" << d << "] = "
                 << error << " " << error/DBL_EPSILON << endl; 
            CHECK_TEST(error<=DBL_EPSILON*1024.);
          }
        }
      }
//    for (int s=pn;s<n;s++) {
//      for (int d=0;d<SPACEDIM;d++) {
//        REAL error=abs(v[s][d]);
//        if (error>DBL_EPSILON*1024.) {
//          cout << "\tcomputeValue(" << s << "," << points[pi] 
//               << ") = " << v[s] << ", error[" << d << "] = "
//               << error << " " << error/DBL_EPSILON << endl; 
//          CHECK_TEST(error<=DBL_EPSILON*1024.);
//        }
//      }
//    }
    }
    NumPtr<Point<SPACEDIM> > points1;
    D_CHOOSE(Interval,Quadrilateral,Hexahedron)::
      staticEquallySpacedLatticePoints(D_DECL(degree,degree+1,degree),
      points1);
    pn=points1.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtpsf->computeValues(points1[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+pn][1]-1.) : abs(v[s+pn][1]));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s+pn << "," << points1[pi] 
               << ") = " << v[s+pn] << ", error[1] = "
               << error << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
        for (int d=0;d<SPACEDIM;d++) {
          if (d != 1) {
            REAL error=abs(v[s+pn][d]);
            if (error>DBL_EPSILON*1024.) {
              cout << "\tcomputeValue(" << s+pn << "," << points1[pi] 
                   << ") = " << v[s+pn] << ", error[" << d << "] = "
                   << error << " " << error/DBL_EPSILON << endl; 
              CHECK_TEST(error<=DBL_EPSILON*1024.);
            }
          }
        }
      }
//    for (int s=0;s<pn;s++) {
//      for (int d=1;d<SPACEDIM;d++) {
//        REAL error=abs(v[s][0]);
//        if (error>DBL_EPSILON*1024.) {
//          cout << "\tcomputeValue(" << s << "," << points1[pi] 
//               << ") = " << v[s] << ", error[" << d << "] = "
//               << error << " " << error/DBL_EPSILON << endl; 
//          CHECK_TEST(error<=DBL_EPSILON*1024.);
//        }
//      }
//    }
#if (SPACEDIM==3)
//    for (int s=2*pn;s<n;s++) {
//      for (int d=1;d<SPACEDIM;d++) {
//        REAL error=abs(v[s][0]);
//        if (error>DBL_EPSILON*1024.) {
//          cout << "\tcomputeValue(" << s << "," << points1[pi] 
//               << ") = " << v[s] << ", error[d] = "
//               << error << " " << error/DBL_EPSILON << endl; 
//          CHECK_TEST(error<=DBL_EPSILON*1024.);
//        }
//      }
//    }
#endif
    }
# if (SPACEDIM==3)
    NumPtr<Point<SPACEDIM> > points2;
    Hexahedron::staticEquallySpacedLatticePoints(degree,degree,degree+1,
      points2);
    pn=points2.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtpsf->computeValues(points2[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+2*pn][2]-1.) : abs(v[s+2*pn][2]));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tcomputeValue(" << s+2*pn << "," << points2[pi] 
               << ") = " << v[s+2*pn] << ", error[2] = "
               << error << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
        for (int d=0;d<2;d++) {
          REAL error=abs(v[s+2*pn][d]);
          if (error>DBL_EPSILON*1024.) {
            cout << "\tcomputeValue(" << s+2*pn << "," << points2[pi] 
                 << ") = " << v[s+2*pn] << ", error[" << d << "] = "
                 << error << " " << error/DBL_EPSILON << endl; 
            CHECK_TEST(error<=DBL_EPSILON*1024.);
          }
        }
      }
//    for (int s=0;s<2*pn;s++) {
//      for (int d=1;d<SPACEDIM;d++) {
//        REAL error=abs(v[s][0]);
//        if (error>DBL_EPSILON*1024.) {
//          cout << "\tcomputeValue(" << s << "," << points2[pi] 
//               << ") = " << v[s] << ", error[" << d << "] = "
//               << error << " " << error/DBL_EPSILON << endl; 
//          CHECK_TEST(error<=DBL_EPSILON*1024.);
//        }
//      }
//    }
    }
# endif
  }
#endif
#if (SPACEDIM>=2)
  const RTParallelpipedDofShapeFunction *rtpdsf=
    dynamic_cast<const RTParallelpipedDofShapeFunction*>(pols);
  if (rtpdsf) {
//  TRACER_CALL(t,"debugPoly RTParallelpipedDofShapeFunction");
    NumPtr<Point<SPACEDIM> > points;
    D_CHOOSE(Interval,Quadrilateral,Hexahedron)::
      staticEquallySpacedLatticePoints(D_DECL(degree-1,degree,degree),
      points);
    int pn=points.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtpdsf->computeValues(points[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s][0]-1.) : abs(v[s][0]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error[0] = " << error 
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=1;d<SPACEDIM;d++) {
          REAL error=abs(v[s][d]);
          if (error>DBL_EPSILON*1024.) {
            cout << "\tcomputeValue(" << s << "," << points[pi] 
                 << ") = " << v[s] << ", error[" << d << "] = "
                 << error << " " << error/DBL_EPSILON << endl; 
            CHECK_TEST(error<=DBL_EPSILON*1024.);
          }
        }
      }
//    for (int s=pn;s<n;s++) {
//      for (int d=0;d<SPACEDIM;d++) {
//        REAL error=abs(v[s][0]);
//        if (error>DBL_EPSILON*1024.) {
//          cout << "\tcomputeValue(" << s << "," << points[pi] 
//               << ") = " << v[s] << ", error[" << d << "] = "
//               << error << " " << error/DBL_EPSILON << endl; 
//          CHECK_TEST(error<=DBL_EPSILON*1024.);
//        }
//      }
//    }
    }
    NumPtr<Point<SPACEDIM> > points1;
    D_CHOOSE(Interval,Quadrilateral,Hexahedron)::
      staticEquallySpacedLatticePoints(D_DECL(degree,degree-1,degree),
      points1);
    pn=points1.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtpdsf->computeValues(points1[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+pn][1]-1.) : abs(v[s+pn][1]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s+pn << "," << points1[pi] 
               << ") = " << v[s+pn] << ", error[1] = " << error 
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=0;d<SPACEDIM;d++) {
          if (d != 1) {
            REAL error=abs(v[s+pn][d]);
            if (error>DBL_EPSILON*1024.) {
              cout << "\tcomputeValue(" << s+pn << "," << points1[pi] 
                   << ") = " << v[s+pn] << ", error[" << d << "] = "
                   << error << " " << error/DBL_EPSILON << endl; 
              CHECK_TEST(error<=DBL_EPSILON*1024.);
            }
          }
        }
      }
    }
# if (SPACEDIM==3)
    NumPtr<Point<SPACEDIM> > points2;
    Hexahedron::staticEquallySpacedLatticePoints(degree,degree,degree-1,
      points2);
    pn=points2.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtpdsf->computeValues(points2[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+2*pn][2]-1.) : abs(v[s+2*pn][2]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s+2*pn << "," << points2[pi] 
               << ") = " << v[s+2*pn] << ", error[2] = " << error 
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=0;d<2;d++) {
          REAL error=abs(v[s+2*pn][d]);
          if (error>DBL_EPSILON*1024.) {
            cout << "\tcomputeValue(" << s+2*pn << "," << points2[pi] 
                 << ") = " << v[s+2*pn] << ", error[" << d << "] = "
                 << error << " " << error/DBL_EPSILON << endl; 
            CHECK_TEST(error<=DBL_EPSILON*1024.);
          }
        }
      }
    }
# endif
    NumPtr<Point<SPACEDIM-1> > points3;
# if (SPACEDIM==2)
    Interval::staticEquallySpacedLatticePoints(degree,points3);
# else
    Quadrilateral::staticEquallySpacedLatticePoints(degree,points3);
# endif
    for (int pi=0;pi<points3.getNumber();pi++) {
      NumPtr<REAL > v;
      rtpdsf->computeFaceValues(points3[pi],v);
      for (int s=0;s<points3.getNumber();s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points3[pi] 
               << ") = " << v[s] << ", error = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }
#endif

#if (SPACEDIM==3)
  const RTPrismShapeFunction *rtwsf=
    dynamic_cast<const RTPrismShapeFunction*>(pols);
  if (rtwsf) {
//  TRACER_CALL(t,"debugPoly RTPrismShapeFunction");
//  lagrange polynomials:
    NumPtr<Point<SPACEDIM> > points;
    Prism::staticEquallySpacedLatticePoints(degree,degree,points);
    int pn=points.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtwsf->computeValues(points[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s][0]-1.) : abs(v[s][0]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error[0] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=1;d<SPACEDIM;d++) {
          REAL error=abs(v[s][d]);
          if (error>DBL_EPSILON*1024.) {
            cout << "\tcomputeValue(" << s << "," << points[pi] 
                 << ") = " << v[s] << ", error[" << d << "] = "
                 << error << " " << error/DBL_EPSILON << endl; 
            CHECK_TEST(error<=DBL_EPSILON*1024.);
          }
        }
      }
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+pn][1]-1.) : abs(v[s+pn][1]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s+pn << "," << points[pi] 
               << ") = " << v[s+pn] << ", error[1] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=0;d<SPACEDIM;d++) {
          if (d!=1) {
            REAL error=abs(v[s+pn][d]);
            if (error>DBL_EPSILON*1024.) {
              cout << "\tcomputeValue(" << s+pn << "," << points[pi] 
                   << ") = " << v[s+pn] << ", error[" << d << "] = "
                   << error << " " << error/DBL_EPSILON << endl; 
              CHECK_TEST(error<=DBL_EPSILON*1024.);
            }
          }
        }
      }
    }
    NumPtr<Point<SPACEDIM> > points2;
    Prism::staticEquallySpacedLatticePoints(degree,degree+1,points2);
    int pn2=points2.getNumber();
    for (int pi=0;pi<pn2;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtwsf->computeValues(points2[pi],v);
      for (int s=0;s<pn2;s++) {
        REAL error=(s==pi ? abs(v[s+2*pn][2]-1.) : abs(v[s+2*pn][2]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s+2*pn << "," << points2[pi] 
               << ") = " << v[s+2*pn] << ", error[2] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=0;d<SPACEDIM-1;d++) {
          REAL error=abs(v[s+2*pn][d]);
          if (error>DBL_EPSILON*1024.) {
            cout << "\tcomputeValue(" << s+2*pn << "," << points2[pi] 
                 << ") = " << v[s+2*pn] << ", error[" << d << "] = "
                 << error << " " << error/DBL_EPSILON << endl; 
            CHECK_TEST(error<=DBL_EPSILON*1024.);
          }
        }
      }
    }
//  homogeneous polynomials:
    Point<SPACEDIM> point;
    for (int d=0;d<SPACEDIM;d++) point[d]=drand48();
    NumPtr<Tensor<1,SPACEDIM> > v;
    rtwsf->computeValues(point,v);
    for (int s=2*pn+pn2;s<n;s++) {
      REAL error=v[s][1]/v[s][0]-point[1]/point[0];
      if (error>DBL_EPSILON*512.) {
        cout << "\tcomputeValue(" << s << "," << point
             << ") = " << v[s] << ", error = " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*512.);
      }
//    if (s+1<n) {
//      error=v[s+1][0]/v[s][0]-point[0]/point[1];
//      if (error>DBL_EPSILON*512.) {
//        cout << "\tcomputeValue(" << s << "," << point
//             << ") = " << v[s] << " " << v[s+1] << ", error = "
//             << error << " " << error/DBL_EPSILON << endl;
//        CHECK_TEST(error<=DBL_EPSILON*512.);
//      }
//    }
    }
  }
#endif
#if (SPACEDIM==3)
  const RTPrismDofShapeFunction *rtwdsf=
    dynamic_cast<const RTPrismDofShapeFunction*>(pols);
  if (rtwdsf) {
//  TRACER_CALL(t,"debugPoly RTPrismDofShapeFunction");
//  lagrange polynomials:
    NumPtr<Point<SPACEDIM> > points;
    Prism::staticEquallySpacedLatticePoints(degree-1,degree,points);
    int pn=points.getNumber();
    for (int pi=0;pi<pn;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtwdsf->computeValues(points[pi],v);
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s][0]-1.) : abs(v[s][0]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points[pi] 
               << ") = " << v[s] << ", error[0] = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=1;d<SPACEDIM;d++) {
          REAL error=abs(v[s][d]);
          if (error>DBL_EPSILON*512.) {
            cout << "\tcomputeValue(" << s << "," << points[pi] 
                 << ") = " << v[s] << ", error[" << d << "] = " << error
                 << " " << error/DBL_EPSILON << endl; 
            CHECK_TEST(error<=DBL_EPSILON*512.);
          }
        }
      }
      for (int s=0;s<pn;s++) {
        REAL error=(s==pi ? abs(v[s+pn][1]-1.) : abs(v[s+pn][1]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s+pn << "," << points[pi]
               << ") = " << v[s+pn] << ", error[1] = " << error  << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=0;d<SPACEDIM;d++) {
          if (d!=1) {
            REAL error=abs(v[s+pn][d]);
            if (error>DBL_EPSILON*1024.) {
              cout << "\tcomputeValue(" << s+pn << "," << points[pi]
                   << ") = " << v[s+pn] << ", error[" << d << "] = "
                   << error << " " << error/DBL_EPSILON << endl;
              CHECK_TEST(error<=DBL_EPSILON*1024.);
            }
          }
        }
      }
    }
    NumPtr<Point<SPACEDIM> > points2;
    Prism::staticEquallySpacedLatticePoints(degree,degree-1,points2);
    int pn2=points2.getNumber();
    for (int pi=0;pi<pn2;pi++) {
      NumPtr<Tensor<1,SPACEDIM> > v;
      rtwdsf->computeValues(points2[pi],v);
      for (int s=0;s<pn2;s++) {
        REAL error=(s==pi ? abs(v[s+2*pn][2]-1.) : abs(v[s+2*pn][2]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points2[pi] 
               << ") = " << v[s+2*pn] << ", error = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
        for (int d=0;d<SPACEDIM-1;d++) {
          REAL error=abs(v[s+2*pn][d]);
          if (error>DBL_EPSILON*1024.) {
            cout << "\tcomputeValue(" << s+2*pn << "," << points2[pi]
                 << ") = " << v[s+2*pn] << ", error[" << d << "] = "
                 << error << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=DBL_EPSILON*1024.);
          }
        }
      }
    }
    NumPtr<Point<SPACEDIM-1> > points3;
    Triangle::staticEquallySpacedLatticePoints(degree,points3);
    int pn3=points3.getNumber();
    for (int pi=0;pi<pn3;pi++) {
      NumPtr<REAL > v;
      rtwdsf->computeTriangleFaceValues(points3[pi],v);
      for (int s=0;s<pn3;s++) {
        REAL error=(s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points3[pi] 
               << ") = " << v[s+2*pn+pn2] << ", error = " << error  << " "
               << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
    NumPtr<Point<SPACEDIM-1> > points4;
    Quadrilateral::staticEquallySpacedLatticePoints(degree,points4);
    int pn4=points4.getNumber();
    for (int pi=0;pi<pn4;pi++) {
      NumPtr<REAL > v;
      rtwdsf->computeQuadrilateralFaceValues(points4[pi],v);
      for (int s=0;s<pn4;s++) {
        REAL error= (s==pi ? abs(v[s]-1.) : abs(v[s]));
        if (error>DBL_EPSILON*512.) {
          cout << "\tcomputeValue(" << s << "," << points4[pi] 
               << ") = " << v[s+2*pn+pn2+pn3] << ", error = " << error
               << " " << error/DBL_EPSILON << endl; 
          CHECK_TEST(error<=DBL_EPSILON*512.);
        }
      }
    }
  }
#endif
}

void richardsonDerivs(int nmax,const TensorShapeFunction<SPACEDIM> *pols,
const Point<SPACEDIM> &x,const NumPtr<Tensor<1,SPACEDIM> > &f,
NumPtr<Tensor<2,SPACEDIM> > &fp) {
//TRACER_CALL(t,"richardsonDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<SPACEDIM;i++) {
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<Tensor<1,SPACEDIM> > fnew;
      Point<SPACEDIM> xnew; xnew=x; xnew[i]+=h;
      pols->computeValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<SPACEDIM;d++) {
          table[nmax-n][s][d]=(fnew[s][d]-f[s][d])/h;
        }
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          for (int d=0;d<SPACEDIM;d++) {
            table[m][s][d]=table[m-1][s][d]
                          +(table[m-1][s][d]-table[m][s][d])/(p-1.);
          }
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<SPACEDIM;d++) {
          fp[s][d][i]=table[nmax-1][s][d];
        }
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardson2ndDerivs(int nmax,
const TensorShapeFunction<SPACEDIM> *pols,const Point<SPACEDIM> &x,
const NumPtr<Tensor<2,SPACEDIM> > &fp,NumPtr<Tensor<3,SPACEDIM> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<SPACEDIM;i++) {
    for (int j=0;j<SPACEDIM;j++) {
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<2,SPACEDIM> > fpnew;
        Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
        pols->computeGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          for (int d=0;d<SPACEDIM;d++) {
            table[nmax-n][s][d]=(fpnew[s][d][i]-fp[s][d][i])/h;
          }
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            for (int d=0;d<SPACEDIM;d++) {
              table[m][s][d]=table[m-1][s][d]
                            +(table[m-1][s][d]-table[m][s][d])/(p-1.);
            }
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        for (int d=0;d<SPACEDIM;d++) {
          fpp[s][d][i][j]=table[nmax-1][s][d];
        }
      }
    }
  }
}

#if (SPACEDIM>=2)
void richardsonFaceDerivs(int nmax,const HdivDofTensorShapeFunction *pols,
const Point<SPACEDIM-1> &x,const NumPtr<REAL> &f,
NumPtr<Tensor<1,SPACEDIM-1> > &fp) {
//TRACER_CALL(t,"richardsonFaceDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<SPACEDIM-1;i++) {
    NumPtr<NumPtr<REAL> > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<REAL> fnew;
      Point<SPACEDIM-1> xnew; xnew=x; xnew[i]+=h;
      pols->computeFaceValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        table[nmax-n][s]=(fnew[s]-f[s])/h;
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        fp[s][i]=table[nmax-1][s];
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardsonFace2ndDerivs(int nmax,
const HdivDofTensorShapeFunction *pols,const Point<SPACEDIM-1> &x,
const NumPtr<Tensor<1,SPACEDIM-1> > &fp,
NumPtr<Tensor<2,SPACEDIM-1> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<SPACEDIM-1;i++) {
    for (int j=0;j<SPACEDIM-1;j++) {
      NumPtr<NumPtr<REAL> > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<1,SPACEDIM-1> > fpnew;
        Point<SPACEDIM-1> xnew; xnew=x; xnew[j]+=h;
        pols->computeFaceGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          table[nmax-n][s]=(fpnew[s][i]-fp[s][i])/h;
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        fpp[s][i][j]=table[nmax-1][s];
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void nedelec2RichardsonFaceDerivs(int nmax,
const Nedelec2SimplexDofShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<1,2> > &f,NumPtr<Tensor<2,2> > &fp) {
//TRACER_CALL(t,"nedelec2RichardsonFaceDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<2;i++) {
    NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<Tensor<1,2> > fnew;
      Point<2> xnew; xnew=x; xnew[i]+=h;
      pols->computeFaceValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        table[nmax-n][s]=(fnew[s]-f[s])/h;
#ifdef DEBUG
//      cout << "\ts = " << s << endl;
//      cout << "\tf = " << f[s] << endl;
//      cout << "\tfnew = " << fnew[s] << endl;
//      cout << "\ttable[" << nmax-n << "][" << s << "] = "
//           << table[nmax-n][s] << endl;
#endif
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fp[s][d][i]=table[nmax-1][s][d];
        }
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void nedelec2RichardsonFace2ndDerivs(int nmax,
const Nedelec2SimplexDofShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<2,2> > &fp,NumPtr<Tensor<3,2> > &fpp) {
//TRACER_CALL(t,"nedelec2RichardsonFace2ndDerivs");
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<2;i++) {
    for (int j=0;j<2;j++) {
      NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
#ifdef DEBUG
//    cout << "\ti,j = " << i << " " << j << endl;
#endif
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<2,2> > fpnew;
        Point<2> xnew; xnew=x; xnew[j]+=h;
        pols->computeFaceGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          for (int d=0;d<2;d++) {
            table[nmax-n][s][d]=(fpnew[s][d][i]-fp[s][d][i])/h;
          }
#ifdef DEBUG
//        cout << "\ttable[" << nmax-n << "][" << s << "] = "
//             << table[nmax-n][s] << endl;
#endif
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fpp[s][d][i][j]=table[nmax-1][s][d];
#ifdef DEBUG
//        cout << "\tfpp[" << s << "][" << d << "][" << i << "]["
//             << j << "] = " << fpp[s][d][i][j] << endl;
#endif
        }
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void richardsonPrismTriangleFaceDerivs(int nmax,
const HdivPrismDofShapeFunction *pols,const Point<2> &x,
const NumPtr<REAL> &f,NumPtr<Tensor<1,2> > &fp) {
//TRACER_CALL(t,"richardsonPrismTriangleFaceDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<2;i++) {
    NumPtr<NumPtr<REAL> > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<REAL> fnew;
      Point<2> xnew; xnew=x; xnew[i]+=h;
      pols->computeTriangleFaceValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        table[nmax-n][s]=(fnew[s]-f[s])/h;
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        fp[s][i]=table[nmax-1][s];
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardsonPrismTriangleFace2ndDerivs(int nmax,
const HdivPrismDofShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<1,2> > &fp,NumPtr<Tensor<2,2> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<2;i++) {
    for (int j=0;j<2;j++) {
      NumPtr<NumPtr<REAL> > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<1,2> > fpnew;
        Point<2> xnew; xnew=x; xnew[j]+=h;
        pols->computeTriangleFaceGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          table[nmax-n][s]=(fpnew[s][i]-fp[s][i])/h;
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        fpp[s][i][j]=table[nmax-1][s];
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void nedelec2RichardsonPrismTriangleFaceDerivs(int nmax,
const Nedelec2PrismDofShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<1,2> > &f,NumPtr<Tensor<2,2> > &fp) {
//TRACER_CALL(t,"nedelec2RichardsonPrismTriangleFaceDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<2;i++) {
    NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<Tensor<1,2> > fnew;
      Point<2> xnew; xnew=x; xnew[i]+=h;
      pols->computeTriangleFaceValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        table[nmax-n][s]=(fnew[s]-f[s])/h;
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fp[s][d][i]=table[nmax-1][s][d];
        }
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void nedelec2RichardsonPrismTriangleFace2ndDerivs(int nmax,
const Nedelec2PrismDofShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<2,2> > &fp,NumPtr<Tensor<3,2> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<2;i++) {
    for (int j=0;j<2;j++) {
      NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<2,2> > fpnew;
        Point<2> xnew; xnew=x; xnew[j]+=h;
        pols->computeTriangleFaceGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          for (int d=0;d<2;d++) {
            table[nmax-n][s][d]=(fpnew[s][d][i]-fp[s][d][i])/h;
          }
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fpp[s][d][i][j]=table[nmax-1][s][d];
        }
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void richardsonPrismQuadrilateralFaceDerivs(int nmax,
const HdivPrismDofShapeFunction *pols,const Point<2> &x,
const NumPtr<REAL> &f,NumPtr<Tensor<1,2> > &fp) {
//TRACER_CALL(t,"richardsonPrismQuadrilateralFaceDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<2;i++) {
    NumPtr<NumPtr<REAL> > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<REAL> fnew;
      Point<2> xnew; xnew=x; xnew[i]+=h;
      pols->computeQuadrilateralFaceValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        table[nmax-n][s]=(fnew[s]-f[s])/h;
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        fp[s][i]=table[nmax-1][s];
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardsonPrismQuadrilateralFace2ndDerivs(int nmax,
const HdivPrismDofShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<1,2> > &fp,NumPtr<Tensor<2,2> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<2;i++) {
    for (int j=0;j<2;j++) {
      NumPtr<NumPtr<REAL> > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<1,2> > fpnew;
        Point<2> xnew; xnew=x; xnew[j]+=h;
        pols->computeQuadrilateralFaceGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          table[nmax-n][s]=(fpnew[s][i]-fp[s][i])/h;
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        fpp[s][i][j]=table[nmax-1][s];
      }
    }
  }
}
#endif

#if (SPACEDIM>=2)
void richardsonEdgeDerivs(int nmax,const HcurlDofTensorShapeFunction *pols,
const Point<1> &x,const NumPtr<REAL> &f,
NumPtr<Tensor<1,1> > &fp) {
//TRACER_CALL(t,"richardsonEdgeDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<1;i++) {
    NumPtr<NumPtr<REAL> > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<REAL> fnew;
      Point<1> xnew; xnew=x; xnew[i]+=h;
      pols->computeEdgeValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        table[nmax-n][s]=(fnew[s]-f[s])/h;
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        fp[s][i]=table[nmax-1][s];
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardsonEdge2ndDerivs(int nmax,
const HcurlDofTensorShapeFunction *pols,const Point<1> &x,
const NumPtr<Tensor<1,1> > &fp,NumPtr<Tensor<2,1> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<1;i++) {
    for (int j=0;j<1;j++) {
      NumPtr<NumPtr<REAL> > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<1,1> > fpnew;
        Point<1> xnew; xnew=x; xnew[j]+=h;
        pols->computeEdgeGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          table[nmax-n][s]=(fpnew[s][i]-fp[s][i])/h;
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        fpp[s][i][j]=table[nmax-1][s];
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void richardsonPrismEdgeDerivs(int nmax,
const HcurlPrismDofShapeFunction *pols,
const Point<1> &x,const NumPtr<REAL> &f,
NumPtr<Tensor<1,1> > &fp) {
//TRACER_CALL(t,"richardsonEdgeDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<1;i++) {
    NumPtr<NumPtr<REAL> > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<REAL> fnew;
      Point<1> xnew; xnew=x; xnew[i]+=h;
      pols->computeEdgeValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        table[nmax-n][s]=(fnew[s]-f[s])/h;
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        fp[s][i]=table[nmax-1][s];
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardsonPrismEdge2ndDerivs(int nmax,
const HcurlPrismDofShapeFunction *pols,const Point<1> &x,
const NumPtr<Tensor<1,1> > &fp,
NumPtr<Tensor<2,1> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<1;i++) {
    for (int j=0;j<1;j++) {
      NumPtr<NumPtr<REAL> > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<1,1> > fpnew;
        Point<1> xnew; xnew=x; xnew[j]+=h;
        pols->computeEdgeGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          table[nmax-n][s]=(fpnew[s][i]-fp[s][i])/h;
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            table[m][s]=table[m-1][s]+(table[m-1][s]-table[m][s])/(p-1.);
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        fpp[s][i][j]=table[nmax-1][s];
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void richardsonFaceDerivs(int nmax,
const HcurlDofTensorShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<1,2> > &f,NumPtr<Tensor<2,2> > &fp) {
//TRACER_CALL(t,"richardsonFaceDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<2;i++) {
    NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<Tensor<1,2> > fnew;
      Point<2> xnew; xnew=x; xnew[i]+=h;
      pols->computeFaceValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<2;d++) {
          table[nmax-n][s][d]=(fnew[s][d]-f[s][d])/h;
        }
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          for (int d=0;d<2;d++) {
            table[m][s][d]=table[m-1][s][d]
                          +(table[m-1][s][d]-table[m][s][d])/(p-1.);
          }
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fp[s][d][i]=table[nmax-1][s][d];
        }
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardsonFace2ndDerivs(int nmax,
const HcurlDofTensorShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<2,2> > &fp,
NumPtr<Tensor<3,2> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<2;i++) {
    for (int j=0;j<2;j++) {
      NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<2,2> > fpnew;
        Point<2> xnew; xnew=x; xnew[j]+=h;
        pols->computeFaceGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          for (int d=0;d<2;d++) {
            table[nmax-n][s][d]=(fpnew[s][d][i]-fp[s][d][i])/h;
          }
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            for (int d=0;d<2;d++) {
              table[m][s][d]=table[m-1][s][d]
                            +(table[m-1][s][d]-table[m][s][d])/(p-1.);
            }
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fpp[s][d][i][j]=table[nmax-1][s][d];
        }
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void richardsonPrismTriangleFaceDerivs(int nmax,
const HcurlPrismDofShapeFunction *pols,
const Point<2> &x,const NumPtr<Tensor<1,2> > &f,
NumPtr<Tensor<2,2> > &fp) {
//TRACER_CALL(t,"richardsonPrismTriangleFaceDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<2;i++) {
    NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<Tensor<1,2> > fnew;
      Point<2> xnew; xnew=x; xnew[i]+=h;
      pols->computeTriangleFaceValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<2;d++) {
          table[nmax-n][s][d]=(fnew[s][d]-f[s][d])/h;
        }
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          for (int d=0;d<2;d++) {
            table[m][s][d]=table[m-1][s][d]
                          +(table[m-1][s][d]-table[m][s][d])/(p-1.);
          }
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fp[s][d][i]=table[nmax-1][s][d];
        }
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardsonPrismTriangleFace2ndDerivs(int nmax,
const HcurlPrismDofShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<2,2> > &fp,
NumPtr<Tensor<3,2> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<2;i++) {
    for (int j=0;j<2;j++) {
      NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<2,2> > fpnew;
        Point<2> xnew; xnew=x; xnew[j]+=h;
        pols->computeTriangleFaceGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          for (int d=0;d<2;d++) {
            table[nmax-n][s][d]=(fpnew[s][d][i]-fp[s][d][i])/h;
          }
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            for (int d=0;d<2;d++) {
              table[m][s][d]=table[m-1][s][d]
                            +(table[m-1][s][d]-table[m][s][d])/(p-1.);
            }
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fpp[s][d][i][j]=table[nmax-1][s][d];
        }
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void richardsonPrismQuadrilateralFaceDerivs(int nmax,
const HcurlPrismDofShapeFunction *pols,
const Point<2> &x,const NumPtr<Tensor<1,2> > &f,
NumPtr<Tensor<2,2> > &fp) {
//TRACER_CALL(t,"richardsonPrismQuadrilateralFaceDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//cout << "\tf.getNumber = " << f.getNumber() << endl;
#endif
  nmax=max(1,nmax);
  fp.allocate(f.getNumber());
  for (int i=0;i<2;i++) {
    NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
    for (int n=0;n<nmax;n++) table[n].allocate(f.getNumber());
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      NumPtr<Tensor<1,2> > fnew;
      Point<2> xnew; xnew=x; xnew[i]+=h;
      pols->computeQuadrilateralFaceValues(xnew,fnew);
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<2;d++) {
          table[nmax-n][s][d]=(fnew[s][d]-f[s][d])/h;
        }
      }
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        for (int s=0;s<f.getNumber();s++) {
          for (int d=0;d<2;d++) {
            table[m][s][d]=table[m-1][s][d]
                          +(table[m-1][s][d]-table[m][s][d])/(p-1.);
          }
        }
      }
      for (int s=0;s<f.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fp[s][d][i]=table[nmax-1][s][d];
        }
      }
    }
  }
#ifdef DEBUG
//for (int s=0;s<fp.getNumber();s++) {
//  cout << "\tfp[" << s << "] = " << fp[s] << endl;
//}
#endif
}

void richardsonPrismQuadrilateralFace2ndDerivs(int nmax,
const HcurlPrismDofShapeFunction *pols,const Point<2> &x,
const NumPtr<Tensor<2,2> > &fp,
NumPtr<Tensor<3,2> > &fpp) {
  nmax=max(1,nmax);
  fpp.allocate(fp.getNumber());
  for (int i=0;i<2;i++) {
    for (int j=0;j<2;j++) {
      NumPtr<NumPtr<Tensor<1,2> > > table(nmax);
      for (int n=0;n<nmax;n++) table[n].allocate(fp.getNumber());
      REAL h=0.1;
      for (int n=1;n<=nmax;n++,h*=0.5) {
        NumPtr<Tensor<2,2> > fpnew;
        Point<2> xnew; xnew=x; xnew[j]+=h;
        pols->computeQuadrilateralFaceGrads(xnew,fpnew);
        for (int s=0;s<fp.getNumber();s++) {
          for (int d=0;d<2;d++) {
            table[nmax-n][s][d]=(fpnew[s][d][i]-fp[s][d][i])/h;
          }
        }
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          for (int s=0;s<fp.getNumber();s++) {
            for (int d=0;d<2;d++) {
              table[m][s][d]=table[m-1][s][d]
                            +(table[m-1][s][d]-table[m][s][d])/(p-1.);
            }
          }
        }
      }
      for (int s=0;s<fp.getNumber();s++) {
        for (int d=0;d<2;d++) {
          fpp[s][d][i][j]=table[nmax-1][s][d];
        }
      }
    }
  }
}
#endif

void debugDerivs(int nmax,const TensorShapeFunction<SPACEDIM> *pols) {
//TRACER_CALL(t,"debugDerivs");
#ifdef DEBUG
//cout << "\tnmax = " << nmax << endl;
//pols->printOn(cout);
#endif
  int n=pols->getNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  Point<SPACEDIM> p(D_DECL(drand48(),drand48(),drand48()));
  NumPtr<Tensor<1,SPACEDIM> > v;
  NumPtr<Tensor<2,SPACEDIM> > g;
  NumPtr<Tensor<3,SPACEDIM> > j;
  pols->computeValues(p,v);
  pols->computeGrads(p,g);
  pols->computeGradGrads(p,j);
#ifdef DEBUG
//for (int s=0;s<v.getNumber();s++) {
//  cout << "\tv[" << s << "] = " << v[s] << endl;
//}
//for (int s=0;s<g.getNumber();s++) {
//  cout << "\tg[" << s << "] = " << g[s] << endl;
//}
//for (int s=0;s<j.getNumber();s++) {
//  cout << "\tj[" << s << "] = " << j[s] << endl;
//}
#endif
  NumPtr<Tensor<2,SPACEDIM> > gg;
  NumPtr<Tensor<3,SPACEDIM> > jj;
  richardsonDerivs(nmax,pols,p,v,gg);
  richardson2ndDerivs(nmax,pols,p,g,jj);
  for (int d=0;d<SPACEDIM;d++) {
//  Point<SPACEDIM> pp(p);
//  pp[d]+=eps;
//  NumPtr<Tensor<1,SPACEDIM> > vv;
//  NumPtr<Tensor<2,SPACEDIM> > gg;
//  pols->computeValues(pp,vv);
//  pols->computeGrads(pp,gg);
    for (int pi=0;pi<n;pi++) {
      for (int c=0;c<SPACEDIM;c++) {
        REAL exact=g[pi][c][d];
//      REAL deriv=(vv[pi][c]-v[pi][c])/eps;
        REAL deriv=gg[pi][c][d];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>DBL_EPSILON*268435456.) {
          if (abs(exact)>268435456.*DBL_EPSILON) {
            if (abs(exact)>268435456.*DBL_EPSILON*abs(v[pi][c])) {
              cout << "\tcomputeGrad(" << pi << "," << p << ")[" << c
                   << "][" << d << "] = " << exact << " " << deriv
                   << ", error = " << error << " " << error/DBL_EPSILON
                   << endl;
              cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                   << endl;
              cout << "\tv = " << v[pi][c] << endl;
//            if (abs(v[pi][c])>0.) {
//              cout << "\tderiv/(eps*v[pi][c]) = " 
//                   << deriv/(eps*v[pi][c]) << endl;
//            }
//            CHECK_TEST(error<=DBL_EPSILON*268435456.);
            }
          }
        }
        for (int dd=0;dd<SPACEDIM;dd++) {
          REAL exact=j[pi][c][d][dd];
//        REAL deriv=(gg[pi][c][dd]-g[pi][c][dd])/eps;
          REAL deriv=jj[pi][c][d][dd];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>DBL_EPSILON*268435456.) {
            if (abs(exact)>268435456.*DBL_EPSILON) {
              if (abs(exact)>268435456.*DBL_EPSILON*abs(g[pi][c][dd])) {
                cout << "\tcomputeGradGrad(" << pi << "," << p << ")[" 
                     << c << "][" << d << "][" << dd << "] = " << exact
                     << " " << deriv << ", error = " << error << " "
                     << error/DBL_EPSILON << endl;
                cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                     << endl;
                cout << "\tg = " << g[pi][c][dd] << endl;
//              if (abs(g[pi][c][dd])>0.) {
//                cout << "\tderiv/(eps*g[pi][c][dd]) = "
//                     << deriv/(eps*g[pi][c][dd]) << endl;
//              }
//              CHECK_TEST(error<=DBL_EPSILON*268435456.);
              }
            }
          }
        }
      }
    }
  }
}

#if (SPACEDIM>=2)
void debugFaceDerivs(int nmax,const HdivDofTensorShapeFunction *pols) {
//TRACER_CALL(t,"debugFaceDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int fn=pols->getFaceNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
# if (SPACEDIM==2)
  Point<1> p(drand48());
# else
  Point<2> p(drand48(),drand48());
# endif
  NumPtr<REAL> v;
  NumPtr<Tensor<1,SPACEDIM-1> > g;
  NumPtr<Tensor<2,SPACEDIM-1> > j;
  pols->computeFaceValues(p,v);
  pols->computeFaceGrads(p,g);
  pols->computeFaceGradGrads(p,j);
  NumPtr<Tensor<1,SPACEDIM-1> > gg;
  NumPtr<Tensor<2,SPACEDIM-1> > jj;
  richardsonFaceDerivs(nmax,pols,p,v,gg);
  richardsonFace2ndDerivs(nmax,pols,p,g,jj);
  for (int d=0;d<SPACEDIM-1;d++) {
//  Point<SPACEDIM-1> pp(p);
//  pp[d]+=eps;
//  NumPtr<REAL> vv;
//  NumPtr<Tensor<1,SPACEDIM-1> > gg;
//  pols->computeFaceValues(pp,vv);
//  pols->computeFaceGrads(pp,gg);
    for (int pi=0;pi<fn;pi++) {
      REAL exact=g[pi][d];
//    REAL deriv=(vv[pi]-v[pi])/eps;
      REAL deriv=gg[pi][d];
      REAL error=abs(deriv-exact);
      REAL den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*4096.) {
        if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi])) {
          cout << "\tcomputeFaceGrad(" << pi << "," << p << ")[" << d 
               << "] = " << exact << " " << deriv << ", error = "
               << error << " " << error/eps << endl;
          cout << "\tv = " << v[pi] << endl;
          if (abs(v[pi])>0.) {
            cout << "\tderiv/(eps*v[pi]) = " 
                 << deriv/(eps*v[pi]) << endl;
          }
          CHECK_TEST(error<=eps*4096.);
        }
      }
      for (int dd=0;dd<SPACEDIM-1;dd++) {
        REAL exact=j[pi][d][dd];
//      REAL deriv=(gg[pi][dd]-g[pi][dd])/eps;
        REAL deriv=jj[pi][d][dd];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>eps*4096.) {
          if (abs(exact)>67108864.*DBL_EPSILON) {
            if (max(abs(exact),abs(deriv))>
            131072.*eps*abs(g[pi][dd])) {
              cout << "\tcomputeFaceGradGrad(" << pi << "," << p << ")[" 
                   << d << "][" << dd << "] = " << exact
                   << " " << deriv << ", error = " << error << " "
                   << error/eps << endl;
              cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                   << " " << deriv/DBL_EPSILON << endl;
              cout << "\tg = " << g[pi][dd] << endl;
              if (abs(g[pi][dd])>0.) {
                cout << "\tderiv/(eps*g[pi][dd]) = "
                     << deriv/(eps*g[pi][dd]) << endl;
              }
              CHECK_TEST(error<=eps*4096.);
            }
          }
        }
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void debugPrismFaceDerivs(int nmax,
const HdivPrismDofShapeFunction *pols) {
//TRACER_CALL(t,"debugPrismFaceDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int tfn=pols->getTriangleFaceNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  Point<2> p(drand48(),drand48());
  NumPtr<REAL> v;
  NumPtr<Tensor<1,2> > g;
  NumPtr<Tensor<2,2> > j;
  NumPtr<Tensor<1,2> > gg;
  NumPtr<Tensor<2,2> > jj;
  {
    pols->computeTriangleFaceValues(p,v);
    pols->computeTriangleFaceGrads(p,g);
    pols->computeTriangleFaceGradGrads(p,j);
    richardsonPrismTriangleFaceDerivs(nmax,pols,p,v,gg);
    richardsonPrismTriangleFace2ndDerivs(nmax,pols,p,g,jj);
    for (int d=0;d<2;d++) {
//    Point<2> pp(p);
//    pp[d]+=eps;
//    NumPtr<REAL> vv;
//    NumPtr<Tensor<1,2> > gg;
//    pols->computeTriangleFaceValues(pp,vv);
//    pols->computeTriangleFaceGrads(pp,gg);
      for (int pi=0;pi<tfn;pi++) {
        REAL exact=g[pi][d];
//      REAL deriv=(vv[pi]-v[pi])/eps;
        REAL deriv=gg[pi][d];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>eps*4096.) {
          if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi])) {
            cout << "\tcomputeTriangleFaceGrad(" << pi << "," << p
                 << ")[" << d << "] = " << exact << " " << deriv
                 << ", error = " << error << " " << error/eps << endl;
            cout << "\tv = " << v[pi] << endl;
            if (abs(v[pi])>0.) {
              cout << "\tderiv/(eps*v[pi]) = " 
                   << deriv/(eps*v[pi]) << endl;
            }
            CHECK_TEST(error<=eps*4096.);
          }
        }
        for (int dd=0;dd<2;dd++) {
          REAL exact=j[pi][d][dd];
//        REAL deriv=(gg[pi][dd]-g[pi][dd])/eps;
          REAL deriv=jj[pi][d][dd];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>eps*1024.) {
            if (abs(exact)>67108864.*DBL_EPSILON) {
              if (max(abs(exact),abs(deriv))>
              131072.*eps*abs(g[pi][dd])) {
                cout << "\tcomputeTriangleFaceGradGrad(" << pi << ","
                     << p << ")[" << d << "][" << dd << "] = " << exact
                     << " " << deriv << ", error = " << error << " "
                     << error/eps << endl;
                cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                     << " " << deriv/DBL_EPSILON << endl;
                cout << "\tg = " << g[pi][dd] << endl;
                if (abs(g[pi][dd])>0.) {
                  cout << "\tderiv/(eps*g[pi][dd]) = "
                       << deriv/(eps*g[pi][dd]) << endl;
                }
                CHECK_TEST(error<=eps*1024.);
              }
            }
          }
        }
      }
    }
  }
  v.cleanup();
  g.cleanup();
  j.cleanup();
  gg.cleanup();
  jj.cleanup();
  {
    int qfn=pols->getQuadrilateralFaceNumber();
    pols->computeQuadrilateralFaceValues(p,v);
    pols->computeQuadrilateralFaceGrads(p,g);
    pols->computeQuadrilateralFaceGradGrads(p,j);
    richardsonPrismQuadrilateralFaceDerivs(nmax,pols,p,v,gg);
    richardsonPrismQuadrilateralFace2ndDerivs(nmax,pols,p,g,jj);
    for (int d=0;d<2;d++) {
//    Point<2> pp(p);
//    pp[d]+=eps;
//    NumPtr<REAL> vv;
//    NumPtr<Tensor<1,2> > gg;
//    pols->computeQuadrilateralFaceValues(pp,vv);
//    pols->computeQuadrilateralFaceGrads(pp,gg);
      for (int pi=0;pi<qfn;pi++) {
        REAL exact=g[pi][d];
//      REAL deriv=(vv[pi]-v[pi])/eps;
        REAL deriv=gg[pi][d];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>eps*4096.) {
          if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi])) {
            cout << "\tcomputeQuadrilateralFaceGrad(" << pi << "," << p
                 << ")[" << d << "] = " << exact << " " << deriv
                 << ", error = " << error << " " << error/eps << endl;
            cout << "\tv = " << v[pi] << endl;
            if (abs(v[pi])>0.) {
              cout << "\tderiv/(eps*v[pi]) = " 
                   << deriv/(eps*v[pi]) << endl;
            }
            CHECK_TEST(error<=eps*4096.);
          }
        }
        for (int dd=0;dd<2;dd++) {
          REAL exact=j[pi][d][dd];
//        REAL deriv=(gg[pi][dd]-g[pi][dd])/eps;
          REAL deriv=jj[pi][d][dd];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>eps*1024.) {
            if (abs(exact)>67108864.*DBL_EPSILON) {
              if (max(abs(exact),abs(deriv))>
              131072.*eps*abs(g[pi][dd])) {
                cout << "\tcomputeQuadrilateralFaceGradGrad(" << pi << ","
                     << p << ")[" << d << "][" << dd << "] = " << exact
                     << " " << deriv << ", error = " << error << " "
                     << error/eps << endl;
                cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                     << " " << deriv/DBL_EPSILON << endl;
                cout << "\tg = " << g[pi][dd] << endl;
                if (abs(g[pi][dd])>0.) {
                  cout << "\tderiv/(eps*g[pi][dd]) = "
                       << deriv/(eps*g[pi][dd]) << endl;
                }
                CHECK_TEST(error<=eps*1024.);
              }
            }
          }
        }
      }
    }
  }
}
#endif

/*
#if (SPACEDIM==3)
void debugPrismFaceDerivs(int nmax,
const HcurlPrismDofShapeFunction *pols) {
//TRACER_CALL(t,"debugPrismFaceDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int tfn=pols->getTriangleFaceNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  Point<2> p(drand48(),drand48());
  NumPtr<Tensor<1,2> > v;
  NumPtr<Tensor<2,2> > g;
  NumPtr<Tensor<3,2> > j;
  NumPtr<Tensor<2,2> > gg;
  NumPtr<Tensor<3,2> > jj;
  {
    pols->computeTriangleFaceValues(p,v);
    pols->computeTriangleFaceGrads(p,g);
    pols->computeTriangleFaceGradGrads(p,j);
    richardsonPrismTriangleFaceDerivs(nmax,pols,p,v,gg);
    richardsonPrismTriangleFace2ndDerivs(nmax,pols,p,g,jj);
    for (int d=0;d<2;d++) {
//    Point<2> pp(p);
//    pp[d]+=eps;
//    NumPtr<REAL> vv;
//    NumPtr<Tensor<1,2> > gg;
//    pols->computeTriangleFaceValues(pp,vv);
//    pols->computeTriangleFaceGrads(pp,gg);
      for (int pi=0;pi<tfn;pi++) {
        REAL exact=g[pi][d];
//      REAL deriv=(vv[pi]-v[pi])/eps;
        REAL deriv=gg[pi][d];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>eps*4096.) {
          if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi])) {
            cout << "\tcomputeTriangleFaceGrad(" << pi << "," << p
                 << ")[" << d << "] = " << exact << " " << deriv
                 << ", error = " << error << " " << error/eps << endl;
            cout << "\tv = " << v[pi] << endl;
            if (abs(v[pi])>0.) {
              cout << "\tderiv/(eps*v[pi]) = " 
                   << deriv/(eps*v[pi]) << endl;
            }
            CHECK_TEST(error<=eps*4096.);
          }
        }
        for (int dd=0;dd<2;dd++) {
          REAL exact=j[pi][d][dd];
//        REAL deriv=(gg[pi][dd]-g[pi][dd])/eps;
          REAL deriv=jj[pi][d][dd];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>eps*1024.) {
            if (abs(exact)>67108864.*DBL_EPSILON) {
              if (max(abs(exact),abs(deriv))>
              131072.*eps*abs(g[pi][dd])) {
                cout << "\tcomputeTriangleFaceGradGrad(" << pi << ","
                     << p << ")[" << d << "][" << dd << "] = " << exact
                     << " " << deriv << ", error = " << error << " "
                     << error/eps << endl;
                cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                     << " " << deriv/DBL_EPSILON << endl;
                cout << "\tg = " << g[pi][dd] << endl;
                if (abs(g[pi][dd])>0.) {
                  cout << "\tderiv/(eps*g[pi][dd]) = "
                       << deriv/(eps*g[pi][dd]) << endl;
                }
                CHECK_TEST(error<=eps*1024.);
              }
            }
          }
        }
      }
    }
  }
  v.cleanup();
  g.cleanup();
  j.cleanup();
  gg.cleanup();
  jj.cleanup();
  {
    int qfn=pols->getQuadrilateralFaceNumber();
    pols->computeQuadrilateralFaceValues(p,v);
    pols->computeQuadrilateralFaceGrads(p,g);
    pols->computeQuadrilateralFaceGradGrads(p,j);
    richardsonPrismQuadrilateralFaceDerivs(nmax,pols,p,v,gg);
    richardsonPrismQuadrilateralFace2ndDerivs(nmax,pols,p,g,jj);
    for (int d=0;d<2;d++) {
//    Point<2> pp(p);
//    pp[d]+=eps;
//    NumPtr<REAL> vv;
//    NumPtr<Tensor<1,2> > gg;
//    pols->computeQuadrilateralFaceValues(pp,vv);
//    pols->computeQuadrilateralFaceGrads(pp,gg);
      for (int pi=0;pi<qfn;pi++) {
        REAL exact=g[pi][d];
//      REAL deriv=(vv[pi]-v[pi])/eps;
        REAL deriv=gg[pi][d];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>eps*4096.) {
          if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi])) {
            cout << "\tcomputeQuadrilateralFaceGrad(" << pi << "," << p
                 << ")[" << d << "] = " << exact << " " << deriv
                 << ", error = " << error << " " << error/eps << endl;
            cout << "\tv = " << v[pi] << endl;
            if (abs(v[pi])>0.) {
              cout << "\tderiv/(eps*v[pi]) = " 
                   << deriv/(eps*v[pi]) << endl;
            }
            CHECK_TEST(error<=eps*4096.);
          }
        }
        for (int dd=0;dd<2;dd++) {
          REAL exact=j[pi][d][dd];
//        REAL deriv=(gg[pi][dd]-g[pi][dd])/eps;
          REAL deriv=jj[pi][d][dd];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>eps*1024.) {
            if (abs(exact)>67108864.*DBL_EPSILON) {
              if (max(abs(exact),abs(deriv))>
              131072.*eps*abs(g[pi][dd])) {
                cout << "\tcomputeQuadrilateralFaceGradGrad(" << pi << ","
                     << p << ")[" << d << "][" << dd << "] = " << exact
                     << " " << deriv << ", error = " << error << " "
                     << error/eps << endl;
                cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                     << " " << deriv/DBL_EPSILON << endl;
                cout << "\tg = " << g[pi][dd] << endl;
                if (abs(g[pi][dd])>0.) {
                  cout << "\tderiv/(eps*g[pi][dd]) = "
                       << deriv/(eps*g[pi][dd]) << endl;
                }
                CHECK_TEST(error<=eps*1024.);
              }
            }
          }
        }
      }
    }
  }
}
#endif
*/

#if (SPACEDIM>=2)
void debugEdgeDerivs(int nmax,const HcurlDofTensorShapeFunction *pols) {
//TRACER_CALL(t,"debugEdgeDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int en=pols->getEdgeNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  {
    Point<1> p(drand48());
    NumPtr<REAL> ev;
    NumPtr<Tensor<1,1> > eg;
    NumPtr<Tensor<2,1> > ej;
    pols->computeEdgeValues(p,ev);
    pols->computeEdgeGrads(p,eg);
    pols->computeEdgeGradGrads(p,ej);
    NumPtr<Tensor<1,1> > egg;
    NumPtr<Tensor<2,1> > ejj;
    richardsonEdgeDerivs(nmax,pols,p,ev,egg);
    richardsonEdge2ndDerivs(nmax,pols,p,eg,ejj);
//  Point<1> pp(p);
//  pp[0]+=eps;
//  NumPtr<REAL> evv;
//  NumPtr<Tensor<1,1> > egg;
//  pols->computeEdgeValues(pp,evv);
//  pols->computeEdgeGrads(pp,egg);
    for (int pi=0;pi<en;pi++) {
      REAL exact=eg[pi][0];
//    REAL deriv=(evv[pi]-ev[pi])/eps;
      REAL deriv=egg[pi][0];
      REAL error=abs(deriv-exact);
      REAL den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*4096.) {
        if (max(abs(exact),abs(deriv))>16384.*eps*abs(ev[pi])) {
          cout << "\tcomputeEdgeGrad(" << pi << "," << p << ") = "
               << exact << " " << deriv << ", error = "
               << error << " " << error/eps << endl;
          cout << "\tv = " << ev[pi] << endl;
          if (abs(ev[pi])>0.) {
            cout << "\tderiv/(eps*ev[pi]) = " 
                 << deriv/(eps*ev[pi]) << endl;
          }
          CHECK_TEST(error<=eps*4096.);
        }
      }
      exact=ej[pi][0][0];
//    deriv=(egg[pi][0]-eg[pi][0])/eps;
      deriv=ej[pi][0][0];
      error=abs(deriv-exact);
      den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*1024.) {
        if (abs(exact)>67108864.*DBL_EPSILON) {
          if (max(abs(exact),abs(deriv))>
          131072.*eps*abs(eg[pi][0])) {
            cout << "\tcomputeEdgeGradGrad(" << pi << "," << p << ")[" 
                 << 0 << "][" << 0 << "][" << 0 << "] = " << exact
                 << " " << deriv << ", error = " << error << " "
                 << error/eps << endl;
            cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                 << " " << deriv/DBL_EPSILON << endl;
            cout << "\tg = " << eg[pi][0] << endl;
            if (abs(eg[pi][0])>0.) {
              cout << "\tderiv/(eps*eg[pi][0]) = "
                   << deriv/(eps*eg[pi][0]) << endl;
            }
            CHECK_TEST(error<=eps*1024.);
          }
        }
      }
    }
  }
# if (SPACEDIM==3)
  int n=pols->getFaceNumber();
  Point<2> p(drand48(),drand48());
  NumPtr<Tensor<1,2> > v;
  NumPtr<Tensor<2,2> > g;
  NumPtr<Tensor<3,2> > j;
  pols->computeFaceValues(p,v);
  pols->computeFaceGrads(p,g);
  pols->computeFaceGradGrads(p,j);
  NumPtr<Tensor<2,2> > gg;
  NumPtr<Tensor<3,2> > jj;
  richardsonFaceDerivs(nmax,pols,p,v,gg);
  richardsonFace2ndDerivs(nmax,pols,p,g,jj);
  for (int d=0;d<2;d++) {
//  Point<2> pp(p);
//  pp[d]+=eps;
//  NumPtr<Tensor<1,2> > vv;
//  NumPtr<Tensor<2,2> > gg;
//  pols->computeFaceValues(pp,vv);
//  pols->computeFaceGrads(pp,gg);
    for (int pi=0;pi<n;pi++) {
      for (int c=0;c<2;c++) {
        REAL exact=g[pi][c][d];
//      REAL deriv=(vv[pi][c]-v[pi][c])/eps;
        REAL deriv=gg[pi][c][d];
        REAL error=abs(deriv-exact);
        REAL den=max(abs(deriv),abs(exact));
        if (den>0.) error/=den;
        if (error>eps*4096.) {
          if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi][c])) {
            cout << "\tcomputeFaceGrad(" << pi << "," << p << ")[" << d 
                 << "] = " << exact << " " << deriv << ", error = "
                 << error << " " << error/eps << endl;
            cout << "\tv = " << v[pi][c] << endl;
            if (abs(v[pi][c])>0.) {
              cout << "\tderiv/(eps*v[pi][c]) = " 
                   << deriv/(eps*v[pi][c]) << endl;
            }
            CHECK_TEST(error<=eps*4096.);
          }
        }
        for (int dd=0;dd<2;dd++) {
          REAL exact=j[pi][c][d][dd];
//        REAL deriv=(gg[pi][c][dd]-g[pi][c][dd])/eps;
          REAL deriv=jj[pi][c][d][dd];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>eps*1024.) {
            if (abs(exact)>67108864.*DBL_EPSILON) {
              if (max(abs(exact),abs(deriv))>
              131072.*eps*abs(g[pi][c][dd])) {
                cout << "\tcomputeFaceGradGrad(" << pi << "," << p << ")[" 
                     << c << "][" << d << "][" << dd << "] = " << exact
                     << " " << deriv << ", error = " << error << " "
                     << error/eps << endl;
                cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                     << " " << deriv/DBL_EPSILON << endl;
                cout << "\tg = " << g[pi][c][dd] << endl;
                if (abs(g[pi][c][dd])>0.) {
                  cout << "\tderiv/(eps*g[pi][c][dd]) = "
                       << deriv/(eps*g[pi][c][dd]) << endl;
                }
                CHECK_TEST(error<=eps*1024.);
              }
            }
          }
        }
      }
    }
  }
# endif
}
#endif

#if (SPACEDIM>=2)
void debugNedelec2EdgeDerivs(int nmax,
const Nedelec2SimplexDofShapeFunction *pols) {
//TRACER_CALL(t,"debugNedelec2EdgeDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int en=pols->getEdgeNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  {
    Point<1> p(drand48());
    NumPtr<REAL> ev;
    NumPtr<Tensor<1,1> > eg;
    NumPtr<Tensor<2,1> > ej;
    pols->computeEdgeValues(p,ev);
    pols->computeEdgeGrads(p,eg);
    pols->computeEdgeGradGrads(p,ej);
    NumPtr<Tensor<1,1> > egg;
    NumPtr<Tensor<2,1> > ejj;
    richardsonEdgeDerivs(nmax,pols,p,ev,egg);
    richardsonEdge2ndDerivs(nmax,pols,p,eg,ejj);
//  Point<1> pp(p);
//  pp[0]+=eps;
//  NumPtr<REAL> evv;
//  NumPtr<Tensor<1,1> > egg;
//  pols->computeEdgeValues(pp,evv);
//  pols->computeEdgeGrads(pp,egg);
    for (int pi=0;pi<en;pi++) {
      REAL exact=eg[pi][0];
//    REAL deriv=(evv[pi]-ev[pi])/eps;
      REAL deriv=egg[pi][0];
      REAL error=abs(deriv-exact);
      REAL den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*4096.) {
        if (max(abs(exact),abs(deriv))>16384.*eps*abs(ev[pi])) {
          cout << "\tcomputeEdgeGrad(" << pi << "," << p << ") = "
               << exact << " " << deriv << ", error = "
               << error << " " << error/eps << endl;
          cout << "\tv = " << ev[pi] << endl;
          if (abs(ev[pi])>0.) {
            cout << "\tderiv/(eps*ev[pi]) = " 
                 << deriv/(eps*ev[pi]) << endl;
          }
          CHECK_TEST(error<=eps*4096.);
        }
      }
      exact=ej[pi][0][0];
//    deriv=(egg[pi][0]-eg[pi][0])/eps;
      deriv=ej[pi][0][0];
      error=abs(deriv-exact);
      den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*1024.) {
        if (abs(exact)>67108864.*DBL_EPSILON) {
          if (max(abs(exact),abs(deriv))>
          131072.*eps*abs(eg[pi][0])) {
            cout << "\tcomputeEdgeGradGrad(" << pi << "," << p << ")[" 
                 << 0 << "][" << 0 << "][" << 0 << "] = " << exact
                 << " " << deriv << ", error = " << error << " "
                 << error/eps << endl;
            cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                 << " " << deriv/DBL_EPSILON << endl;
            cout << "\tg = " << eg[pi][0] << endl;
            if (abs(eg[pi][0])>0.) {
              cout << "\tderiv/(eps*eg[pi][0]) = "
                   << deriv/(eps*eg[pi][0]) << endl;
            }
            CHECK_TEST(error<=eps*1024.);
          }
        }
      }
    }
  }
# if (SPACEDIM==3)
  int n=pols->getFaceNumber();
  Point<2> p(drand48(),drand48());
//const Nedelec2SimplexDofShapeFunction *n2sdsf=
//  dynamic_cast<const Nedelec2SimplexDofShapeFunction*>(pols);
//CHECK_POINTER(n2sdsf);
  PermutationGenerator pg(3);
  while (pg.hasMore())
  {
//  TRACER_CALL(t,"debugNedelec2EdgeDerivs");
    const NumPtr<int> &permutation=pg.getNext();
    int vertices[4]; vertices[0]=0;
    for (int i=0;i<3;i++) vertices[i+1]=permutation[i]+1;
    int lines[6]={0,1,2,3,4,5};
    Tetrahedron e(-1,lines,vertices,vertices);
    for (int f=0;f<4;f++)
//  int f=3;
    {
#   ifdef DEBUG
//    e.printOn(cout);
//    cout << "\tf = " << f << endl;
#   endif
      NumPtr<Tensor<1,2> > v;
      NumPtr<Tensor<2,2> > g;
      NumPtr<Tensor<3,2> > j;
      pols->computeFaceValues(p,v);
      pols->computeFaceGrads(p,g);
      pols->computeFaceGradGrads(p,j);
      NumPtr<Tensor<2,2> > gg;
      NumPtr<Tensor<3,2> > jj;
      nedelec2RichardsonFaceDerivs(nmax,pols,p,v,gg);
      nedelec2RichardsonFace2ndDerivs(nmax,pols,p,g,jj);
      for (int d=0;d<2;d++) {
        for (int pi=0;pi<n;pi++) {
          for (int c=0;c<2;c++) {
            REAL exact=g[pi][c][d];
            REAL deriv=gg[pi][c][d];
            REAL error=abs(deriv-exact);
            REAL den=max(abs(deriv),abs(exact));
            if (den>0.) error/=den;
            if (error>DBL_EPSILON*131072.) {
              if (max(abs(exact),abs(deriv))
              >16384.*DBL_EPSILON*abs(v[pi][c])) {
                cout << "\tcomputeFaceGrad(" << pi << "," << p
                     << ")[" << d << "] = " << exact << " " << deriv
                     << ", error = " << error << " " << error/DBL_EPSILON
                     << endl;
                cout << "\tv = " << v[pi][c] << endl;
                if (abs(v[pi][c])>0.) {
                  cout << "\tderiv/(DBL_EPSILON*v[pi][c]) = " 
                       << deriv/(DBL_EPSILON*v[pi][c]) << endl;
                }
                CHECK_TEST(error<=eps*131072.);
              }
            }
            for (int dd=0;dd<2;dd++) {
              REAL exact=j[pi][c][d][dd];
//            REAL deriv=(gg[pi][c][dd]-g[pi][c][dd])/eps;
              REAL deriv=jj[pi][c][d][dd];
              REAL error=abs(deriv-exact);
              REAL den=max(abs(deriv),abs(exact));
              if (den>0.) error/=den;
              if (error>eps*131072.) {
                if (abs(exact)>67108864.*DBL_EPSILON) {
                  if (max(abs(exact),abs(deriv))>
                  131072.*eps*abs(g[pi][c][dd])) {
                    cout << "\tcomputeFaceGradGrad(" << pi << "," << p
                         << ")[" << c << "][" << d << "][" << dd
                         << "] = " << exact << " " << deriv
                         << ", error = " << error << " " << error/eps
                         << endl;
                    cout << "\texact/DBL_EPSILON = "
                         << exact/DBL_EPSILON
                         << " " << deriv/DBL_EPSILON << endl;
                    cout << "\tg = " << g[pi][c][dd] << endl;
                    if (abs(g[pi][c][dd])>0.) {
                      cout << "\tderiv/(eps*g[pi][c][dd]) = "
                           << deriv/(eps*g[pi][c][dd]) << endl;
                    }
                    CHECK_TEST(error<=eps*131072.);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
# endif
}
#endif

#if (SPACEDIM==3)
void debugPrismEdgeDerivs(int nmax,
const HcurlPrismDofShapeFunction *pols) {
//TRACER_CALL(t,"debugPrismEdgeDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int en=pols->getEdgeNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  {
    Point<1> p(drand48());
    NumPtr<REAL> ev;
    NumPtr<Tensor<1,1> > eg;
    NumPtr<Tensor<2,1> > ej;
    pols->computeEdgeValues(p,ev);
    pols->computeEdgeGrads(p,eg);
    pols->computeEdgeGradGrads(p,ej);
    NumPtr<Tensor<1,1> > egg;
    NumPtr<Tensor<2,1> > ejj;
    richardsonPrismEdgeDerivs(nmax,pols,p,ev,egg);
    richardsonPrismEdge2ndDerivs(nmax,pols,p,eg,ejj);
//  Point<1> pp(p);
//  pp[0]+=eps;
//  NumPtr<REAL> evv;
//  NumPtr<Tensor<1,1> > egg;
//  pols->computeEdgeValues(pp,evv);
//  pols->computeEdgeGrads(pp,egg);
    for (int pi=0;pi<en;pi++) {
      REAL exact=eg[pi][0];
//    REAL deriv=(evv[pi]-ev[pi])/eps;
      REAL deriv=egg[pi][0];
      REAL error=abs(deriv-exact);
      REAL den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*4096.) {
        if (max(abs(exact),abs(deriv))>16384.*eps*abs(ev[pi])) {
          cout << "\tcomputeEdgeGrad(" << pi << "," << p << ") = "
               << exact << " " << deriv << ", error = "
               << error << " " << error/eps << endl;
          cout << "\tv = " << ev[pi] << endl;
          if (abs(ev[pi])>0.) {
            cout << "\tderiv/(eps*ev[pi]) = " 
                 << deriv/(eps*ev[pi]) << endl;
          }
          CHECK_TEST(error<=eps*4096.);
        }
      }
      exact=ej[pi][0][0];
//    deriv=(egg[pi][0]-eg[pi][0])/eps;
      deriv=ejj[pi][0][0];
      error=abs(deriv-exact);
      den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*1024.) {
        if (abs(exact)>67108864.*DBL_EPSILON) {
          if (max(abs(exact),abs(deriv))>
          131072.*eps*abs(eg[pi][0])) {
            cout << "\tcomputeEdgeGradGrad(" << pi << "," << p << ")[" 
                 << 0 << "][" << 0 << "][" << 0 << "] = " << exact
                 << " " << deriv << ", error = " << error << " "
                 << error/eps << endl;
            cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                 << " " << deriv/DBL_EPSILON << endl;
            cout << "\tg = " << eg[pi][0] << endl;
            if (abs(eg[pi][0])>0.) {
              cout << "\tderiv/(eps*eg[pi][0]) = "
                   << deriv/(eps*eg[pi][0]) << endl;
            }
            CHECK_TEST(error<=eps*1024.);
          }
        }
      }
    }
  }
  {
    int n=pols->getTriangleFaceNumber();
    Point<2> p(drand48(),drand48());
    NumPtr<Tensor<1,2> > v;
    NumPtr<Tensor<2,2> > g;
    NumPtr<Tensor<3,2> > j;
    pols->computeTriangleFaceValues(p,v);
    pols->computeTriangleFaceGrads(p,g);
    pols->computeTriangleFaceGradGrads(p,j);
    NumPtr<Tensor<2,2> > gg;
    NumPtr<Tensor<3,2> > jj;
    richardsonPrismTriangleFaceDerivs(nmax,pols,p,v,gg);
    richardsonPrismTriangleFace2ndDerivs(nmax,pols,p,g,jj);
    for (int d=0;d<2;d++) {
//    Point<2> pp(p);
//    pp[d]+=eps;
//    NumPtr<Tensor<1,2> > vv;
//    NumPtr<Tensor<2,2> > gg;
//    pols->computeTriangleFaceValues(pp,vv);
//    pols->computeTriangleFaceGrads(pp,gg);
      for (int pi=0;pi<n;pi++) {
        for (int c=0;c<2;c++) {
          REAL exact=g[pi][c][d];
//        REAL deriv=(vv[pi][c]-v[pi][c])/eps;
          REAL deriv=gg[pi][c][d];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>eps*4096.) {
            if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi][c])) {
              cout << "\tcomputeTriangleFaceGrad(" << pi << "," << p
                   << ")[" << d << "] = " << exact << " " << deriv
                   << ", error = " << error << " " << error/eps << endl;
              cout << "\tv = " << v[pi][c] << endl;
              if (abs(v[pi][c])>0.) {
                cout << "\tderiv/(eps*v[pi][c]) = " 
                     << deriv/(eps*v[pi][c]) << endl;
              }
              CHECK_TEST(error<=eps*4096.);
            }
          }
          for (int dd=0;dd<2;dd++) {
            REAL exact=j[pi][c][d][dd];
//          REAL deriv=(gg[pi][c][dd]-g[pi][c][dd])/eps;
            REAL deriv=jj[pi][c][d][dd];
            REAL error=abs(deriv-exact);
            REAL den=max(abs(deriv),abs(exact));
            if (den>0.) error/=den;
            if (error>eps*1024.) {
              if (abs(exact)>67108864.*DBL_EPSILON) {
                if (max(abs(exact),abs(deriv))>
                131072.*eps*abs(g[pi][c][dd])) {
                  cout << "\tcomputeTriangleFaceGradGrad(" << pi << ","
                       << p << ")[" << c << "][" << d << "][" << dd
                       << "] = " << exact << " " << deriv
                       << ", error = " << error << " " << error/eps
                       << endl;
                  cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                       << " " << deriv/DBL_EPSILON << endl;
                  cout << "\tg = " << g[pi][c][dd] << endl;
                  if (abs(g[pi][c][dd])>0.) {
                    cout << "\tderiv/(eps*g[pi][c][dd]) = "
                         << deriv/(eps*g[pi][c][dd]) << endl;
                  }
                  CHECK_TEST(error<=eps*1024.);
                }
              }
            }
          }
        }
      }
    }
  }
  {
    int n=pols->getQuadrilateralFaceNumber();
    Point<2> p(drand48(),drand48());
    NumPtr<Tensor<1,2> > v;
    NumPtr<Tensor<2,2> > g;
    NumPtr<Tensor<3,2> > j;
    pols->computeQuadrilateralFaceValues(p,v);
    pols->computeQuadrilateralFaceGrads(p,g);
    pols->computeQuadrilateralFaceGradGrads(p,j);
    NumPtr<Tensor<2,2> > gg;
    NumPtr<Tensor<3,2> > jj;
    richardsonPrismQuadrilateralFaceDerivs(nmax,pols,p,v,gg);
    richardsonPrismQuadrilateralFace2ndDerivs(nmax,pols,p,g,jj);
    for (int d=0;d<2;d++) {
//    Point<2> pp(p);
//    pp[d]+=eps;
//    NumPtr<Tensor<1,2> > vv;
//    NumPtr<Tensor<2,2> > gg;
//    pols->computeQuadrilateralFaceValues(pp,vv);
//    pols->computeQuadrilateralFaceGrads(pp,gg);
      for (int pi=0;pi<n;pi++) {
        for (int c=0;c<2;c++) {
          REAL exact=g[pi][c][d];
//        REAL deriv=(vv[pi][c]-v[pi][c])/eps;
          REAL deriv=gg[pi][c][d];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>eps*4096.) {
            if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi][c])) {
              cout << "\tcomputeQuadrilateralFaceGrad(" << pi << "," << p
                   << ")[" << d << "] = " << exact << " " << deriv
                   << ", error = " << error << " " << error/eps << endl;
              cout << "\tv = " << v[pi][c] << endl;
              if (abs(v[pi][c])>0.) {
                cout << "\tderiv/(eps*v[pi][c]) = " 
                     << deriv/(eps*v[pi][c]) << endl;
              }
              CHECK_TEST(error<=eps*4096.);
            }
          }
          for (int dd=0;dd<2;dd++) {
            REAL exact=j[pi][c][d][dd];
//          REAL deriv=(gg[pi][c][dd]-g[pi][c][dd])/eps;
            REAL deriv=jj[pi][c][d][dd];
            REAL error=abs(deriv-exact);
            REAL den=max(abs(deriv),abs(exact));
            if (den>0.) error/=den;
            if (error>eps*1024.) {
              if (abs(exact)>67108864.*DBL_EPSILON) {
                if (max(abs(exact),abs(deriv))>
                131072.*eps*abs(g[pi][c][dd])) {
                  cout << "\tcomputeQuadrilateralFaceGradGrad(" << pi
                       << "," << p << ")[" << c << "][" << d << "][" << dd
                       << "] = " << exact << " " << deriv
                       << ", error = " << error << " " << error/eps
                       << endl;
                  cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                       << " " << deriv/DBL_EPSILON << endl;
                  cout << "\tg = " << g[pi][c][dd] << endl;
                  if (abs(g[pi][c][dd])>0.) {
                    cout << "\tderiv/(eps*g[pi][c][dd]) = "
                         << deriv/(eps*g[pi][c][dd]) << endl;
                  }
                  CHECK_TEST(error<=eps*1024.);
                }
              }
            }
          }
        }
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void debugNedelec2PrismEdgeDerivs(int nmax,
const Nedelec2PrismDofShapeFunction *pols) {
//TRACER_CALL(t,"debugNedelec2PrismEdgeDerivs");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int en=pols->getEdgeNumber();
  REAL eps=sqrt(DBL_EPSILON);
  drand48();
  {
    Point<1> p(drand48());
    NumPtr<REAL> ev;
    NumPtr<Tensor<1,1> > eg;
    NumPtr<Tensor<2,1> > ej;
    pols->computeEdgeValues(p,ev);
    pols->computeEdgeGrads(p,eg);
    pols->computeEdgeGradGrads(p,ej);
    NumPtr<Tensor<1,1> > egg;
    NumPtr<Tensor<2,1> > ejj;
    richardsonPrismEdgeDerivs(nmax,pols,p,ev,egg);
    richardsonPrismEdge2ndDerivs(nmax,pols,p,eg,ejj);
//  Point<1> pp(p);
//  pp[0]+=eps;
//  NumPtr<REAL> evv;
//  NumPtr<Tensor<1,1> > egg;
//  pols->computeEdgeValues(pp,evv);
//  pols->computeEdgeGrads(pp,egg);
    for (int pi=0;pi<en;pi++) {
      REAL exact=eg[pi][0];
//    REAL deriv=(evv[pi]-ev[pi])/eps;
      REAL deriv=egg[pi][0];
      REAL error=abs(deriv-exact);
      REAL den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*4096.) {
        if (max(abs(exact),abs(deriv))>16384.*eps*abs(ev[pi])) {
          cout << "\tcomputeEdgeGrad(" << pi << "," << p << ") = "
               << exact << " " << deriv << ", error = "
               << error << " " << error/eps << endl;
          cout << "\tv = " << ev[pi] << endl;
          if (abs(ev[pi])>0.) {
            cout << "\tderiv/(eps*ev[pi]) = " 
                 << deriv/(eps*ev[pi]) << endl;
          }
          CHECK_TEST(error<=eps*4096.);
        }
      }
      exact=ej[pi][0][0];
//    deriv=(egg[pi][0]-eg[pi][0])/eps;
      deriv=ejj[pi][0][0];
      error=abs(deriv-exact);
      den=max(abs(deriv),abs(exact));
      if (den>0.) error/=den;
      if (error>eps*1024.) {
        if (abs(exact)>67108864.*DBL_EPSILON) {
          if (max(abs(exact),abs(deriv))>
          131072.*eps*abs(eg[pi][0])) {
            cout << "\tcomputeEdgeGradGrad(" << pi << "," << p << ")[" 
                 << 0 << "][" << 0 << "][" << 0 << "] = " << exact
                 << " " << deriv << ", error = " << error << " "
                 << error/eps << endl;
            cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                 << " " << deriv/DBL_EPSILON << endl;
            cout << "\tg = " << eg[pi][0] << endl;
            if (abs(eg[pi][0])>0.) {
              cout << "\tderiv/(eps*eg[pi][0]) = "
                   << deriv/(eps*eg[pi][0]) << endl;
            }
            CHECK_TEST(error<=eps*1024.);
          }
        }
      }
    }
  }
  {
    int n=pols->getTriangleFaceNumber();
    Point<2> p(drand48(),drand48());
    PermutationGenerator pg(5);
    while (pg.hasMore())
    {
//    TRACER_CALL(t,"debugNedelec2PrismEdgeDerivs");
      const NumPtr<int> &permutation=pg.getNext();
      int vertices[6]; vertices[0]=0;
      for (int i=0;i<5;i++) vertices[i+1]=permutation[i]+1;
      int lines[9]={0,1,2,3,4,5,6,7,8};
      int faces[5]={0,1,2,3,4};
      Prism e(-1,lines,faces,vertices);
      for (int f=0;f<2;f++) 
//    int f=0;
      {
#     ifdef DEBUG
//      e.printOn(cout);
//      cout << "\tf = " << f << endl;
#     endif
        NumPtr<Tensor<1,2> > v;
        NumPtr<Tensor<2,2> > g;
        NumPtr<Tensor<3,2> > j;
        pols->computeTriangleFaceValues(p,v);
        pols->computeTriangleFaceGrads(p,g);
        pols->computeTriangleFaceGradGrads(p,j);
        NumPtr<Tensor<2,2> > gg;
        NumPtr<Tensor<3,2> > jj;
        nedelec2RichardsonPrismTriangleFaceDerivs(nmax,pols,p,v,gg);
        nedelec2RichardsonPrismTriangleFace2ndDerivs(nmax,pols,p,g,jj);
        for (int d=0;d<2;d++) {
          for (int pi=0;pi<n;pi++) {
            for (int c=0;c<2;c++) {
              REAL exact=g[pi][c][d];
              REAL deriv=gg[pi][c][d];
              REAL error=abs(deriv-exact);
              REAL den=max(abs(deriv),abs(exact));
              if (den>0.) error/=den;
              if (error>eps*4096.) {
                if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi][c])) {
                  cout << "\tcomputeTriangleFaceGrad(" << pi << "," << p
                       << ")[" << d << "] = " << exact << " " << deriv
                       << ", error = " << error << " " << error/eps
                       << endl;
                  cout << "\tv = " << v[pi][c] << endl;
                  if (abs(v[pi][c])>0.) {
                    cout << "\tderiv/(eps*v[pi][c]) = " 
                         << deriv/(eps*v[pi][c]) << endl;
                  }
                  CHECK_TEST(error<=eps*4096.);
                }
              }
              for (int dd=0;dd<2;dd++) {
                REAL exact=j[pi][c][d][dd];
                REAL deriv=jj[pi][c][d][dd];
                REAL error=abs(deriv-exact);
                REAL den=max(abs(deriv),abs(exact));
                if (den>0.) error/=den;
                if (error>eps*1024.) {
                  if (abs(exact)>67108864.*DBL_EPSILON) {
                    if (max(abs(exact),abs(deriv))>
                    131072.*eps*abs(g[pi][c][dd])) {
                      cout << "\tcomputeTriangleFaceGradGrad(" << pi << ","
                           << p << ")[" << c << "][" << d << "][" << dd
                           << "] = " << exact << " " << deriv
                           << ", error = " << error << " " << error/eps
                           << endl;
                      cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                           << " " << deriv/DBL_EPSILON << endl;
                      cout << "\tg = " << g[pi][c][dd] << endl;
                      if (abs(g[pi][c][dd])>0.) {
                        cout << "\tderiv/(eps*g[pi][c][dd]) = "
                             << deriv/(eps*g[pi][c][dd]) << endl;
                      }
                      CHECK_TEST(error<=eps*1024.);
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  {
    int n=pols->getQuadrilateralFaceNumber();
    Point<2> p(drand48(),drand48());
    NumPtr<Tensor<1,2> > v;
    NumPtr<Tensor<2,2> > g;
    NumPtr<Tensor<3,2> > j;
    pols->computeQuadrilateralFaceValues(p,v);
    pols->computeQuadrilateralFaceGrads(p,g);
    pols->computeQuadrilateralFaceGradGrads(p,j);
    NumPtr<Tensor<2,2> > gg;
    NumPtr<Tensor<3,2> > jj;
    richardsonPrismQuadrilateralFaceDerivs(nmax,pols,p,v,gg);
    richardsonPrismQuadrilateralFace2ndDerivs(nmax,pols,p,g,jj);
    for (int d=0;d<2;d++) {
//    Point<2> pp(p);
//    pp[d]+=eps;
//    NumPtr<Tensor<1,2> > vv;
//    NumPtr<Tensor<2,2> > gg;
//    pols->computeQuadrilateralFaceValues(pp,vv);
//    pols->computeQuadrilateralFaceGrads(pp,gg);
      for (int pi=0;pi<n;pi++) {
        for (int c=0;c<2;c++) {
          REAL exact=g[pi][c][d];
//        REAL deriv=(vv[pi][c]-v[pi][c])/eps;
          REAL deriv=gg[pi][c][d];
          REAL error=abs(deriv-exact);
          REAL den=max(abs(deriv),abs(exact));
          if (den>0.) error/=den;
          if (error>eps*4096.) {
            if (max(abs(exact),abs(deriv))>16384.*eps*abs(v[pi][c])) {
              cout << "\tcomputeQuadrilateralFaceGrad(" << pi << "," << p
                   << ")[" << d << "] = " << exact << " " << deriv
                   << ", error = " << error << " " << error/eps << endl;
              cout << "\tv = " << v[pi][c] << endl;
              if (abs(v[pi][c])>0.) {
                cout << "\tderiv/(eps*v[pi][c]) = " 
                     << deriv/(eps*v[pi][c]) << endl;
              }
              CHECK_TEST(error<=eps*4096.);
            }
          }
          for (int dd=0;dd<2;dd++) {
            REAL exact=j[pi][c][d][dd];
//          REAL deriv=(gg[pi][c][dd]-g[pi][c][dd])/eps;
            REAL deriv=jj[pi][c][d][dd];
            REAL error=abs(deriv-exact);
            REAL den=max(abs(deriv),abs(exact));
            if (den>0.) error/=den;
            if (error>eps*1024.) {
              if (abs(exact)>67108864.*DBL_EPSILON) {
                if (max(abs(exact),abs(deriv))>
                131072.*eps*abs(g[pi][c][dd])) {
                  cout << "\tcomputeQuadrilateralFaceGradGrad(" << pi
                       << "," << p << ")[" << c << "][" << d << "][" << dd
                       << "] = " << exact << " " << deriv
                       << ", error = " << error << " " << error/eps
                       << endl;
                  cout << "\texact/DBL_EPSILON = " << exact/DBL_EPSILON
                       << " " << deriv/DBL_EPSILON << endl;
                  cout << "\tg = " << g[pi][c][dd] << endl;
                  if (abs(g[pi][c][dd])>0.) {
                    cout << "\tderiv/(eps*g[pi][c][dd]) = "
                         << deriv/(eps*g[pi][c][dd]) << endl;
                  }
                  CHECK_TEST(error<=eps*1024.);
                }
              }
            }
          }
        }
      }
    }
  }
}
#endif

void debugSimplexDuality(const RTSimplexDualShapeFunction *pols,
int degree) {
//TRACER_CALL(t,"debugSimplexDuality");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int pn=pols->getNumber();
  RTSimplexDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
//face dofs:
#if (SPACEDIM==1)
  for (int f=0;f<2;f++) {
    Point<1> p; p[0]=static_cast<REAL>(f);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(p,pv);
    for (int s=0;s<pols->getNumber();s++) {
      REAL error=(s==f ? abs((f==0 ? pv[s][0]+1. : pv[s][0]-1.)) 
                       : abs(pv[s][0]));
      if (error>DBL_EPSILON*1024) {
        cout << "\tcomputeValue(" << s << "," << p << ") = " << pv[s]
             << ", error[0] = " << error << " " << error/DBL_EPSILON
             << endl;
        CHECK_TEST(error<=DBL_EPSILON*1024);
      }
    }
  }
  int ioffset=2;
#else
# if (SPACEDIM==2)
  GaussianQuadrature<SPACEDIM-1>
    fgq(GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
# else
  TriangleGaussianQuadrature
    fgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
# endif
  const NumPtr<Point<SPACEDIM-1> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  int dpfn=dof_polys.getFaceNumber();
  int ioffset=0;
  for (int f=0;f<SPACEDIM+1;f++) {
    NumPtr<Point<SPACEDIM> > q_points_face;
    q_points_face.allocate(fgq_points.getNumber());
# if (SPACEDIM==2)
    Triangle::staticProjectToFace(f,fgq_points,q_points_face);
    Tensor<1,SPACEDIM> face_normal=Triangle::faceNormal(f);
# else
    Tetrahedron::staticProjectToFace(f,fgq_points,q_points_face);
    Tensor<1,SPACEDIM> face_normal=Tetrahedron::faceNormal(f);
# endif
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<SPACEDIM-1> &fqp=fgq_points[q];
      const Point<SPACEDIM> &qp=q_points_face[q];
      NumPtr<REAL> dpv;
      dof_polys.computeFaceValues(fqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=fgq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=face_normal*pv[j];
//      if (f==SPACEDIM) dotj*=sqrt(static_cast<REAL>(SPACEDIM));
        for (int i=0;i<dpfn;i++) {
          matrix(i,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*32768.) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*32768.);
        }
      }
    }
    ioffset+=dpfn;
  }
#endif
//interior
#if (SPACEDIM==1)
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+2));
#elif (SPACEDIM==2)
  TriangleGaussianQuadrature
    gq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
#else
  TetrahedronGaussianQuadrature
    gq(TetrahedronGaussianQuadrature::argGivenOrder(2*degree+2));
#endif
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i]*pv[j];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*4096.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*4096.);
      }
    }
  }
}

void debugSimplexDuality(const Nedelec1SimplexDualShapeFunction *pols,
int degree,const SHAPE_CLASS &element) {
//TRACER_CALL(t,"debugSimplexDuality");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int pn=pols->getNumber();
  Nedelec1SimplexDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
  NumPtr<int> renumber;
  dof_polys.shapeFunctionsOrder(&element,renumber);
#if (SPACEDIM==1)
  for (int f=0;f<2;f++) {
    Point<1> p; p[0]=static_cast<REAL>(f);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(p,pv);
    for (int s=0;s<pols->getNumber();s++) {
      REAL error=(s==f ? abs((f==0 ? pv[s][0]-1. : pv[s][0]-1.)) 
                       : abs(pv[s][0]));
      if (error>DBL_EPSILON*1024) {
        cout << "\tcomputeValue(" << s << "," << p << ") = " << pv[s]
             << ", error[0] = " << error << " " << error/DBL_EPSILON
             << endl;
        CHECK_TEST(error<=DBL_EPSILON*1024);
      }
    }
  }
  int ioffset=2;
#else
//edge dofs:
  GaussianQuadrature<1>
    egq(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
  int dpen=dof_polys.getEdgeNumber();
  int ioffset=0;
  for (int e=0;e<D_CHOOSE(0,3,6);e++) {
    NumPtr<Point<SPACEDIM> > q_points_edge;
    q_points_edge.allocate(egq_points.getNumber());
# if (SPACEDIM==2)
    Triangle::staticProjectToFace(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> edge_tangent=Triangle::lineTangent(e);
# else
    Tetrahedron::staticProjectToLine(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> edge_tangent=Tetrahedron::lineTangent(e);
# endif
    if (element.lineTangentPointsBackward(e)) edge_tangent=-edge_tangent;
//  edge_tangent.normalize();
    LAMatrix matrix(dpen,pn,0.);
    for (int q=0;q<q_points_edge.getNumber();q++) {
      const Point<1> &eqp=egq_points[q];
      const Point<SPACEDIM> &qp=q_points_edge[q];
      NumPtr<REAL> dpv;
      dof_polys.computeEdgeValues(eqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=egq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=edge_tangent*pv[j];
        for (int i=0;i<dpen;i++) {
          matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpen;i++) {
        REAL error=(j==i+ioffset ? abs(matrix(i,j)-1.) 
                                 : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tedge dof(" << e << "," << i+ioffset << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
    ioffset+=dpen;
  }
#endif
//face dofs:
# if (SPACEDIM==3)
  TriangleGaussianQuadrature
    fgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<2> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  int dpfn=dof_polys.getFaceNumber();
  for (int f=0;f<4;f++) {
    bool ftwb=element.faceTangentsWindBackward(f);
    Tensor<1,3> ft0=element.faceTangent(f,0);
    Tensor<1,3> ft1=element.faceTangent(f,1);
//  ft0.normalize();
//  ft1.normalize();
//  if (f==3) {
//    ft0=-ft0;
//    ft1=-ft1;
//  }
    Tensor<1,3> fn=Tetrahedron::faceNormal(f);
    CHECK_TEST(abs(fn*ft0)<DBL_EPSILON*16.);
    CHECK_TEST(abs(fn*ft1)<DBL_EPSILON*16.);
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(fgq_points.getNumber());
    Tetrahedron::staticProjectToFace(f,fgq_points,q_points_face);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<2> &fqp=fgq_points[q];
      const Point<3> &qp=q_points_face[q];
      NumPtr<Tensor<1,2> > dpv;
      dof_polys.computeFaceValues(fqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=fgq_weights[q];
      for (int j=0;j<pn;j++) {
        Tensor<1,3> &pvj=pv[j];
        REAL dot0=ft0*pvj;
        REAL dot1=ft1*pvj;
        for (int i=0;i<dpfn;i++) {
          Tensor<1,2> &dpvi=dpv[i];
          if (ftwb) {
            matrix(renumber[i+ioffset]-ioffset,j)
              +=(dpvi[1]*dot0+dpvi[0]*dot1)*wq;
          } else {
            matrix(renumber[i+ioffset]-ioffset,j)
              +=(dpvi[0]*dot0+dpvi[1]*dot1)*wq;
          }
        }
      }
    }
# ifdef DEBUG
//  cout << "\tf = " << f << endl;
//  for (int i=0;i<dpfn;i++) {
//    cout << "\tmatrix(" << i+6*dpen+f*dpfn << ",*) = ";
//    for (int j=6*dpen+f*dpfn;j<6*dpen+(f+1)*dpfn;j++) {
//      REAL mat=
//        ( abs(matrix(i,j))<DBL_EPSILON*1048576. ? 0. : matrix(i,j));
//      cout << mat << " ";
//    }
//    cout << endl;
//  }
# endif
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.)
                                  : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tface dof(f = " << f << ", i j = " << i+ioffset
               << " " << j << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
    ioffset+=dpfn;
  }
# endif
//interior
#if (SPACEDIM==1)
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+2));
#elif (SPACEDIM==2)
  TriangleGaussianQuadrature
    gq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
#else
  TetrahedronGaussianQuadrature
    gq(TetrahedronGaussianQuadrature::argGivenOrder(2*degree+2));
#endif
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i]*pv[j];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*1048576.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*1048576.);
      }
    }
  }
}

void debugSimplexDuality(const Nedelec2SimplexDualShapeFunction *pols,
int degree,const SHAPE_CLASS &element) {
//TRACER_CALL(t,"debugSimplexDuality");
#ifdef DEBUG
//element.printOn(cout);
//pols->printOn(cout);
#endif
  int pn=pols->getNumber();
  Nedelec2SimplexDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
  NumPtr<int> renumber;
  dof_polys.shapeFunctionsOrder(&element,renumber);
#ifdef DEBUG
//cout << "\trenumber = ";
//for (int i=0;i<renumber.getNumber();i++) cout << renumber[i] << " ";
//cout << endl;
#endif
#if (SPACEDIM==1)
  for (int f=0;f<2;f++) {
    Point<1> p; p[0]=static_cast<REAL>(f);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(p,pv);
    for (int s=0;s<pols->getNumber();s++) {
      REAL error=(s==f ? abs((f==0 ? pv[s][0]-1. : pv[s][0]-1.)) 
                       : abs(pv[s][0]));
      if (error>DBL_EPSILON*1024) {
        cout << "\tcomputeValue(" << s << "," << p << ") = " << pv[s]
             << ", error[0] = " << error << " " << error/DBL_EPSILON
             << endl;
        CHECK_TEST(error<=DBL_EPSILON*1024);
      }
    }
  }
  int ioffset=2;
#else
//edge dofs:
  GaussianQuadrature<1>
    egq(GaussianQuadrature<1>::argGivenOrder(2*degree+1));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
  int dpen=dof_polys.getEdgeNumber();
  int ioffset=0;
  for (int e=0;e<D_CHOOSE(0,3,6);e++) {
    NumPtr<Point<SPACEDIM> > q_points_edge;
    q_points_edge.allocate(egq_points.getNumber());
# if (SPACEDIM==2)
    Triangle::staticProjectToFace(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> edge_tangent=Triangle::lineTangent(e);
# else
    Tetrahedron::staticProjectToLine(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> edge_tangent=Tetrahedron::lineTangent(e);
# endif
    bool ltpb=element.lineTangentPointsBackward(e);
    if (ltpb) edge_tangent=-edge_tangent;
//  edge_tangent.normalize();
    LAMatrix matrix(dpen,pn,0.);
    for (int q=0;q<q_points_edge.getNumber();q++) {
      const Point<1> &eqp=egq_points[q];
      const Point<SPACEDIM> &qp=q_points_edge[q];
      NumPtr<REAL> dpv;
//    if (ltpb) {
//      dof_polys.computeEdgeValues(
//        egq_points[egq_points.getNumber()-1-q],dpv);
//    } else {
        dof_polys.computeEdgeValues(eqp,dpv);
//    }
# ifdef DEBUG
//    cout << "\tdpv[" << q << "] = ";
//    for (int s=0;s<dpv.getNumber();s++) {
//      cout << dpv[s] << " ";
//    }
//    cout << endl;
# endif
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=egq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=edge_tangent*pv[j];
        for (int i=0;i<dpen;i++) {
          matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i]*wq*dotj;
        }
      }
    }
# ifdef DEBUG
//  for (int i=0;i<dpen;i++) {
//    cout << "\tmatrix(" << i+ioffset << ",*) = ";
//    for (int j=0;j<pn;j++) {
//      REAL mat=(abs(matrix(i,j))<1.e-5 ? 0. : matrix(i,j));
//      cout << mat << " ";
//    }
//    cout << endl;
//  }
# endif
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpen;i++) {
        REAL error=(j==i+ioffset ? abs(matrix(i,j)-1.) 
                                 : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tedge dof(e = " << e << ", i j = " << i+ioffset
               << " " << j << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
    ioffset+=dpen;
  }
#endif
//face dofs:
#if (SPACEDIM==3)
  TriangleGaussianQuadrature
    fgq(TriangleGaussianQuadrature::argGivenOrder(2*degree));
  const NumPtr<Point<2> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  int dpfn=dof_polys.getFaceNumber();
  for (int f=0;f<4;f++) {
# ifdef DEBUG
//  cout << "\tf,ioffset = " << f << " " << ioffset << endl;
# endif
    bool ftwb=element.faceTangentsWindBackward(f);
    Tensor<1,3> ft0=element.faceTangent(f,0);
    Tensor<1,3> ft1=element.faceTangent(f,1);
//  ft0.normalize();
//  ft1.normalize();
//  if (f==3) {
//    ft0=-ft0;
//    ft1=-ft1;
//  }
//  if (ftwb) {
//    ft0=-ft0;
//    ft1=-ft1;
//  }
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(fgq_points.getNumber());
    Tetrahedron::staticProjectToFace(f,fgq_points,q_points_face);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<2> &fqp=fgq_points[q];
      const Point<3> &qp=q_points_face[q];
      NumPtr<Tensor<1,2> > dpv;
//    if (ftwb) {
//      Point<2> fqpr(fqp[1],fqp[0]);
//      dof_polys.computeFaceValues(fqpr,dpv);
//    } else {
        dof_polys.computeFaceValues(fqp,dpv);
//    }
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=fgq_weights[q];
      for (int j=0;j<pn;j++) {
        Tensor<1,3> &pvj=pv[j];
        for (int i=0;i<dpfn;i++) {
          Tensor<1,2> &dpvi=dpv[i];
          if (ftwb) {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[1]*(pvj*ft0)+dpvi[0]*(pvj*ft1))*wq;
          } else {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[0]*(pvj*ft0)+dpvi[1]*(pvj*ft1))*wq;
          }
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tface dof(f = " << f << ", i j = " << i+ioffset
               << " " << j << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
    ioffset+=dpfn;
  }
#endif
//interior
#if (SPACEDIM==1)
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+2));
#elif (SPACEDIM==2)
  TriangleGaussianQuadrature
    gq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
#else
  TetrahedronGaussianQuadrature
    gq(TetrahedronGaussianQuadrature::argGivenOrder(2*degree+2));
#endif
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i]*pv[j];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*1048576.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*1048576.);
      }
    }
  }
}

#if (SPACEDIM>=2)
void debugParallelpipedDuality(
const RTParallelpipedDualShapeFunction *pols,int degree) {
//TRACER_CALL(t,"debugParallelpipedDuality");
# ifdef DEBUG
//pols->printOn(cout);
# endif
  int pn=pols->getNumber();
  RTParallelpipedDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
//face dofs:
  GaussianQuadrature<SPACEDIM-1>
    fgq(GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM-1> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  int dpfn=dof_polys.getFaceNumber();
  int ioffset=0;
  for (int f=0;f<2*SPACEDIM;f++) {
    NumPtr<Point<SPACEDIM> > q_points_face;
    q_points_face.allocate(fgq_points.getNumber());
# if (SPACEDIM==2)
    Quadrilateral::staticProjectToFace(f,fgq_points,q_points_face);
    Tensor<1,SPACEDIM> face_normal=Quadrilateral::faceNormal(f);
# else
    Hexahedron::staticProjectToFace(f,fgq_points,q_points_face);
    Tensor<1,SPACEDIM> face_normal=Hexahedron::faceNormal(f);
# endif
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<SPACEDIM-1> &fqp=fgq_points[q];
      const Point<SPACEDIM> &qp=q_points_face[q];
      NumPtr<REAL> dpv;
      dof_polys.computeFaceValues(fqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=fgq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=face_normal*pv[j];
        for (int i=0;i<dpfn;i++) {
          matrix(i,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1024) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1024);
        }
      }
    }
    ioffset+=dpfn;
  }
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=0.;
        for (int d=0;d<SPACEDIM;d++) dot+=dpv[i][d]*pv[j][d];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*1024) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*1024);
      }
    }
  }
}
#endif

#if (SPACEDIM>=2)
void debugParallelpipedDuality(const Shape *element,
const BDFMParallelpipedDualShapeFunction *pols,int degree) {
//TRACER_CALL(t,"debugParallelpipedDuality");
# ifdef DEBUG
//pols->printOn(cout);
# endif
  const D_CHOOSE(Interval,Quadrilateral,Hexahedron) *parallelpiped=
    dynamic_cast<const D_CHOOSE(Interval,Quadrilateral,Hexahedron)*>(
      element);
  CHECK_POINTER(parallelpiped);

  int pn=pols->getNumber();
  BDFMParallelpipedDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
  NumPtr<int> renumber;
  dof_polys.shapeFunctionsOrder(element,renumber);
//face dofs:
  GaussianQuadrature<SPACEDIM-1>
    fgq(GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM-1> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  int dpfn=dof_polys.getFaceNumber();
  int ioffset=0;
  for (int d=0;d<SPACEDIM;d++) {
    for (int f=2*d;f<2*d+2;f++) {
      NumPtr<Point<SPACEDIM> > q_points_face(fgq_points.getNumber());;
      parallelpiped->projectToFaceForBDFM(f,fgq_points,q_points_face);
      LAMatrix matrix(dpfn,pn,0.);
      for (int q=0;q<q_points_face.getNumber();q++) {
        REAL wq=fgq_weights[q];
        const Point<SPACEDIM-1> &fqp=fgq_points[q];
        const Point<SPACEDIM> &qp=q_points_face[q];
        NumPtr<REAL> dpv;
        dof_polys.computeFaceValues(fqp,dpv);
        NumPtr<Tensor<1,SPACEDIM> > pv;
        pols->computeValues(qp,pv);
        for (int j=0;j<pn;j++) {
          REAL dotj=pv[j][d];
          if (f%2==0) dotj=-dotj;
          for (int i=0;i<dpfn;i++) {
            matrix(renumber[i],j)+=dpv[i]*wq*dotj;
          }
        }
      }
# ifdef DEBUG
//    cout << "\n\tf = " << f << endl;
//    for (int i=0;i<dpfn;i++) {
//      cout << "\tmatrix(" << i << ",*) = ";
//      for (int j=0;j<pn;j++) {
//        cout << (abs(matrix(i,j))<sqrt(DBL_EPSILON) ? 0. : matrix(i,j) )
//             << " ";
//      }
//      cout << endl;
//    }
# endif
      for (int j=0;j<pn;j++) {
        for (int i=0;i<dpfn;i++) {
          REAL error=
            (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
          if (error>DBL_EPSILON*2097152.) {
            cout << "\tface dof(" << f << "," << i << " " << j
                 << ") = " << matrix(i,j)
                 << ", error = " << error << " " << error/DBL_EPSILON
                 << endl;
            CHECK_TEST(error<=DBL_EPSILON*2097152.);
          }
        }
      }
      ioffset+=dpfn;
    }
  }
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i]*pv[j];
        matrix(renumber[i],j)+=dot*wq;
      }
    }
  }
# ifdef DEBUG
//for (int i=0;i<dpn;i++) {
//  cout << "\tmatrix(" << i << ",*) = ";
//  for (int j=0;j<pn;j++) {
//    cout << matrix(i,j) << " ";
//  }
//  cout << endl;
//}
# endif
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*67108864) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*67108864);
      }
    }
  }
}
#endif

#if (SPACEDIM>=2)
void debugParallelpipedDuality(
const Nedelec1ParallelpipedDualShapeFunction *pols,int degree,
const SHAPE_CLASS &element) {
//TRACER_CALL(t,"debugParallelpipedDuality");
# ifdef DEBUG
//pols->printOn(cout);
//element.printOn(cout);
# endif
  int pn=pols->getNumber();
  Nedelec1ParallelpipedDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
  NumPtr<int> renumber;
  dof_polys.shapeFunctionsOrder(&element,renumber);
//edge dofs:
  GaussianQuadrature<1>
    egq(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
  int dpen=dof_polys.getEdgeNumber();
  int ioffset=0;
  for (int e=0;e<D_CHOOSE(0,4,12);e++) {
    NumPtr<Point<SPACEDIM> > q_points_edge;
    q_points_edge.allocate(egq_points.getNumber());
# if (SPACEDIM==2)
    Quadrilateral::staticProjectToFace(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> edge_tangent=Quadrilateral::lineTangent(e);
# else
    Hexahedron::staticProjectToLine(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> edge_tangent=Hexahedron::lineTangent(e);
# endif
    if (element.lineTangentPointsBackward(e)) edge_tangent=-edge_tangent;
    edge_tangent.normalize();
    LAMatrix matrix(dpen,pn,0.);
    for (int q=0;q<q_points_edge.getNumber();q++) {
      const Point<1> &fqp=egq_points[q];
      const Point<SPACEDIM> &qp=q_points_edge[q];
      NumPtr<REAL> dpv;
      dof_polys.computeEdgeValues(fqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=egq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=edge_tangent*pv[j];
        for (int i=0;i<dpen;i++) {
          matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpen;i++) {
        REAL error=(j==i+ioffset ? abs(matrix(i,j)-1.) 
                                 : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tedge dof(" << e << "," << i+ioffset << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
      }
    }
    ioffset+=dpen;
  }
# if (SPACEDIM==3)
//face dofs:
#ifdef DEBUG
//cout << "\trenumber = ";
//for (int i=0;i<renumber.getNumber();i++) cout << renumber[i] << " ";
//cout << endl;
#endif
  GaussianQuadrature<2>
    fgq(GaussianQuadrature<2>::argGivenOrder(2*degree+2));
  const NumPtr<Point<2> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  int dpfn=dof_polys.getFaceNumber();
  int dpfn2=dpfn/2;
  for (int f=0;f<6;f++) {
    bool ftwb=element.faceTangentsWindBackward(f);
    Tensor<1,3> ft0=element.faceTangent(f,0);
    Tensor<1,3> ft1=element.faceTangent(f,1);
    ft0.normalize();
    ft1.normalize();
#ifdef DEBUG
//  cout << "\tftwb = " << ftwb << endl;
//  cout << "\tft0 = " << ft0 << endl;
//  cout << "\tft1 = " << ft1 << endl;
#endif
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(fgq_points.getNumber());
    Hexahedron::staticProjectToFace(f,fgq_points,q_points_face);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<2> &fqp=fgq_points[q];
      const Point<3> &qp=q_points_face[q];
      NumPtr<Tensor<1,2> > dpv;
      dof_polys.computeFaceValues(fqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=fgq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL pvj0=(ft0*pv[j])*wq;
        REAL pvj1=(ft1*pv[j])*wq;
        for (int i=0;i<dpfn2;i++) {
          if (ftwb) {
            matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i+dpfn2][1]*pvj0;
            matrix(renumber[i+dpfn2+ioffset]-ioffset,j)+=dpv[i][0]*pvj1;
          } else {
            matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i][0]*pvj0;
            matrix(renumber[i+dpfn2+ioffset]-ioffset,j)
              +=dpv[i+dpfn2][1]*pvj1;
          }
        }
      }
    }
#ifdef DEBUG
//  for (int i=0;i<dpfn;i++) {
//    cout << "\tmatrix(" << i+ioffset << ",*) = ";
//    for (int j=0;j<pn;j++) {
//      REAL mat=(abs(matrix(i,j))>1.e-5 ? matrix(i,j) : 0.);
//      cout << mat << " ";
//    }
//    cout << endl;
//  }
#endif
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=(j==i+ioffset ? abs(matrix(i,j)-1.) 
                                  : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1024) {
//        cout << "\ti,j,renumber = " << i+ioffset << " " << j << " "
//             << renumber[j] << endl;
          cout << "\tface dof(f = " << f << ",i j = " << i+ioffset << " "
               << j << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1024);
        }
      }
    }
    ioffset+=dpfn;
  }
# endif
//interior dofs:
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=0.;
        for (int d=0;d<SPACEDIM;d++) dot+=dpv[i][d]*pv[j][d];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*65536.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*65536.);
      }
    }
  }
}
#endif

#if (SPACEDIM>=2)
void debugParallelpipedDuality(
const Nedelec2ParallelpipedDualShapeFunction *pols,int degree,
const SHAPE_CLASS &element) {
//TRACER_CALL(t,"debugParallelpipedDuality");
# ifdef DEBUG
//pols->printOn(cout);
# endif
  int pn=pols->getNumber();
  Nedelec2ParallelpipedDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
  NumPtr<int> renumber;
  dof_polys.shapeFunctionsOrder(&element,renumber);
//edge dofs:
  GaussianQuadrature<1>
    egq(GaussianQuadrature<1>::argGivenOrder(2*degree+1));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
  int dpen=dof_polys.getEdgeNumber();
  int ioffset=0;
  for (int e=0;e<D_CHOOSE(0,4,12);e++) {
    NumPtr<Point<SPACEDIM> > q_points_edge;
    q_points_edge.allocate(egq_points.getNumber());
# if (SPACEDIM==2)
    Quadrilateral::staticProjectToFace(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> edge_tangent=Quadrilateral::lineTangent(e);
# else
    Hexahedron::staticProjectToLine(e,egq_points,q_points_edge);
    Tensor<1,SPACEDIM> edge_tangent=Hexahedron::lineTangent(e);
# endif
    if (element.lineTangentPointsBackward(e)) edge_tangent=-edge_tangent;
    edge_tangent.normalize();
    LAMatrix matrix(dpen,pn,0.);
    for (int q=0;q<q_points_edge.getNumber();q++) {
      const Point<1> &fqp=egq_points[q];
      const Point<SPACEDIM> &qp=q_points_edge[q];
      NumPtr<REAL> dpv;
      dof_polys.computeEdgeValues(fqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=egq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=edge_tangent*pv[j];
        for (int i=0;i<dpen;i++) {
          matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpen;i++) {
        REAL error=(j==i+ioffset ? abs(matrix(i,j)-1.) 
                                 : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1024.) {
          cout << "\tedge dof(" << e << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1024.);
        }
      }
    }
    ioffset+=dpen;
  }
# if (SPACEDIM==3)
//face dofs:
  GaussianQuadrature<2>
    fgq(GaussianQuadrature<2>::argGivenOrder(2*degree+1));
  const NumPtr<Point<2> > &fgq_points=fgq.points();
  const NumPtr<REAL> &fgq_weights=fgq.weights();
  int dpfn=dof_polys.getFaceNumber();
  int dpfn2=dpfn/2;
  for (int f=0;f<6;f++) {
    bool ftwb=element.faceTangentsWindBackward(f);
    Tensor<1,3> ft0=element.faceTangent(f,0);
    Tensor<1,3> ft1=element.faceTangent(f,1);
    ft0.normalize();
    ft1.normalize();
#ifdef DEBUG
//  cout << "\tft0,ft1 = " << ft0 << " " << ft1 << endl;
#endif
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(fgq_points.getNumber());
    Hexahedron::staticProjectToFace(f,fgq_points,q_points_face);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<2> &fqp=fgq_points[q];
      const Point<3> &qp=q_points_face[q];
      NumPtr<Tensor<1,2> > dpv;
      dof_polys.computeFaceValues(fqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=fgq_weights[q];
      for (int j=0;j<pn;j++) {
        Tensor<1,3> &pvj=pv[j];
        REAL pvj0=(ft0*pvj)*wq;
        REAL pvj1=(ft1*pvj)*wq;
        for (int i=0;i<dpfn2;i++) {
          if (ftwb) {
            matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i+dpfn2][1]*pvj0;
            matrix(renumber[i+dpfn2+ioffset]-ioffset,j)+=dpv[i][0]*pvj1;
          } else {
            matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i][0]*pvj0;
            matrix(renumber[i+dpfn2+ioffset]-ioffset,j)+=
              dpv[i+dpfn2][1]*pvj1;
          }
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1024) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1024);
        }
      }
    }
    ioffset+=dpfn;
  }
# endif
//interior dofs:
  GaussianQuadrature<SPACEDIM>
    gq(GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+1));
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=0.;
        for (int d=0;d<SPACEDIM;d++) dot+=dpv[i][d]*pv[j][d];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*65536.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*65536.);
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void debugPrismDuality(const RTPrismDualShapeFunction *pols,int degree) {
//TRACER_CALL(t,"debugPrismDuality");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int pn=pols->getNumber();
  RTPrismDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
//triangle face dofs:
  TriangleGaussianQuadrature
    tgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM-1> > &tgq_points=tgq.points();
  const NumPtr<REAL> &tgq_weights=tgq.weights();
  int dpfn=dof_polys.getTriangleFaceNumber();
  int ioffset=0;
  for (int f=0;f<2;f++) {
    NumPtr<Point<SPACEDIM> > q_points_face;
    q_points_face.allocate(tgq_points.getNumber());
    Prism::staticProjectToFace(f,tgq_points,q_points_face);
    Tensor<1,SPACEDIM> face_normal=Prism::faceNormal(f);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<SPACEDIM-1> &fqp=tgq_points[q];
      const Point<SPACEDIM> &qp=q_points_face[q];
      NumPtr<REAL> dpv;
      dof_polys.computeTriangleFaceValues(fqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=tgq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=face_normal*pv[j];
        for (int i=0;i<dpfn;i++) {
          matrix(i,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*32768.) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*32768.);
        }
      }
    }
    ioffset+=dpfn;
  }
//quadrilateral face dofs:
  GaussianQuadrature<2>
    qgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM-1> > &qgq_points=qgq.points();
  const NumPtr<REAL> &qgq_weights=qgq.weights();
  dpfn=dof_polys.getQuadrilateralFaceNumber();
  for (int f=2;f<5;f++) {
    NumPtr<Point<SPACEDIM> > q_points_face;
    q_points_face.allocate(qgq_points.getNumber());
    Prism::staticProjectToFace(f,qgq_points,q_points_face);
    Tensor<1,SPACEDIM> face_normal=Prism::faceNormal(f);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<SPACEDIM-1> &fqp=qgq_points[q];
      const Point<SPACEDIM> &qp=q_points_face[q];
      NumPtr<REAL> dpv;
      dof_polys.computeQuadrilateralFaceValues(fqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=qgq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=face_normal*pv[j];
        for (int i=0;i<dpfn;i++) {
          matrix(i,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
    ioffset+=dpfn;
  }
//interior
  GaussianQuadrature<1>
    gq1(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
    ProductQuadrature gq(tgq,gq1);
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i]*pv[j];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*524288.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*524288.);
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void debugPrismDuality(const Shape *element,
const BDFMPrismDualShapeFunction *pols,int degree) {
//TRACER_CALL(t,"debugPrismDuality");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  const Prism *prism=dynamic_cast<const Prism*>(element);
  CHECK_POINTER(prism);

  int pn=pols->getNumber();
  BDFMPrismDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
//triangle face dofs:
  TriangleGaussianQuadrature
    tgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM-1> > &tgq_points=tgq.points();
  const NumPtr<REAL> &tgq_weights=tgq.weights();
  int dpfn=dof_polys.getTriangleFaceNumber();
  int ioffset=0;
  for (int f=0;f<2;f++) {
    NumPtr<Point<SPACEDIM> > q_points_face;
    q_points_face.allocate(tgq_points.getNumber());
    prism->projectToFaceForBDFM(f,tgq_points,q_points_face);
    Tensor<1,SPACEDIM> face_normal=Prism::faceNormal(f);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<SPACEDIM-1> &fqp=tgq_points[q];
      const Point<SPACEDIM> &qp=q_points_face[q];
      NumPtr<REAL> dpv;
      dof_polys.computeTriangleFaceValues(fqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=tgq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=face_normal*pv[j];
        for (int i=0;i<dpfn;i++) {
          matrix(i,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*32768.) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*32768.);
        }
      }
    }
    ioffset+=dpfn;
  }
//quadrilateral face dofs:
  GaussianQuadrature<2>
    qgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<SPACEDIM-1> > &qgq_points=qgq.points();
  const NumPtr<REAL> &qgq_weights=qgq.weights();
  dpfn=dof_polys.getQuadrilateralFaceNumber();
  for (int f=2;f<5;f++) {
    NumPtr<Point<SPACEDIM> > q_points_face;
    q_points_face.allocate(qgq_points.getNumber());
    prism->projectToFaceForBDFM(f,qgq_points,q_points_face);
    Tensor<1,SPACEDIM> face_normal=Prism::faceNormal(f);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<SPACEDIM-1> &fqp=qgq_points[q];
      const Point<SPACEDIM> &qp=q_points_face[q];
      NumPtr<REAL> dpv;
      dof_polys.computeQuadrilateralFaceValues(fqp,dpv);
      NumPtr<Tensor<1,SPACEDIM> > pv;
      pols->computeValues(qp,pv);
      REAL wq=qgq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=face_normal*pv[j];
        for (int i=0;i<dpfn;i++) {
          matrix(i,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
    ioffset+=dpfn;
  }
//interior
  GaussianQuadrature<1>
    gq1(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
    ProductQuadrature gq(tgq,gq1);
  const NumPtr<Point<SPACEDIM> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<SPACEDIM> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,SPACEDIM> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,SPACEDIM> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i]*pv[j];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*524288.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*524288.);
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void debugPrismDuality(const Nedelec1PrismDualShapeFunction *pols,
int degree,const Shape3 &element) {
//TRACER_CALL(t,"debugPrismDuality");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int pn=pols->getNumber();
  Nedelec1PrismDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
  NumPtr<int> renumber;
  dof_polys.shapeFunctionsOrder(&element,renumber);
//edge dofs:
  GaussianQuadrature<1>
    egq(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
  int dpen=dof_polys.getEdgeNumber();
  int ioffset=0;
  for (int e=0;e<9;e++) {
    NumPtr<Point<3> > q_points_edge;
    q_points_edge.allocate(egq_points.getNumber());
    Prism::staticProjectToLine(e,egq_points,q_points_edge);
    Tensor<1,3> edge_tangent=Prism::lineTangent(e);
    if (element.lineTangentPointsBackward(e)) edge_tangent=-edge_tangent;
//  edge_tangent.normalize();
    LAMatrix matrix(dpen,pn,0.);
    for (int q=0;q<q_points_edge.getNumber();q++) {
      const Point<1> &eqp=egq_points[q];
      const Point<3> &qp=q_points_edge[q];
      NumPtr<REAL> dpv;
      dof_polys.computeEdgeValues(eqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=egq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=edge_tangent*pv[j];
        for (int i=0;i<dpen;i++) {
          matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i]*wq*dotj;
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpen;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*32768.) {
          cout << "\tedge dof(" << e << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*32768.);
        }
      }
    }
    ioffset+=dpen;
  }
//triangle face dofs:
  TriangleGaussianQuadrature
    tgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<2> > &tgq_points=tgq.points();
  const NumPtr<REAL> &tgq_weights=tgq.weights();
  int dpfn=dof_polys.getTriangleFaceNumber();
  for (int f=0;f<2;f++) {
    bool ftwb=element.faceTangentsWindBackward(f);
    Tensor<1,3> ft0=element.faceTangent(f,0);
    Tensor<1,3> ft1=element.faceTangent(f,1);
//  ft0.normalize();
//  ft1.normalize();
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(tgq_points.getNumber());
    Prism::staticProjectToFace(f,tgq_points,q_points_face);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<2> &fqp=tgq_points[q];
      const Point<3> &qp=q_points_face[q];
      NumPtr<Tensor<1,2> > dpv;
      dof_polys.computeTriangleFaceValues(fqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=tgq_weights[q];
      for (int j=0;j<pn;j++) {
        Tensor<1,3> &pvj=pv[j];
        REAL dot0=ft0*pvj;
        REAL dot1=ft1*pvj;
        for (int i=0;i<dpfn;i++) {
          Tensor<1,2> &dpvi=dpv[i];
          if (ftwb) {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[1]*dot0+dpvi[0]*dot1)*wq;
          } else {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[0]*dot0+dpvi[1]*dot1)*wq;
          }
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*32768.) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*32768.);
        }
      }
    }
    ioffset+=dpfn;
  }
//quadrilateral face dofs:
  GaussianQuadrature<2>
    qgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<2> > &qgq_points=qgq.points();
  const NumPtr<REAL> &qgq_weights=qgq.weights();
  dpfn=dof_polys.getQuadrilateralFaceNumber();
  for (int f=2;f<5;f++) {
    bool ftwb=element.faceTangentsWindBackward(f);
    Tensor<1,3> ft0=element.faceTangent(f,0);
    Tensor<1,3> ft1=element.faceTangent(f,1);
//  ft0.normalize();
//  ft1.normalize();
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(qgq_points.getNumber());
    Prism::staticProjectToFace(f,qgq_points,q_points_face);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<2> &fqp=qgq_points[q];
      const Point<3> &qp=q_points_face[q];
      NumPtr<Tensor<1,2> > dpv;
      dof_polys.computeQuadrilateralFaceValues(fqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=qgq_weights[q];
      for (int j=0;j<pn;j++) {
        Tensor<1,3> &pvj=pv[j];
        REAL dot0=ft0*pvj;
        REAL dot1=ft1*pvj;
        for (int i=0;i<dpfn;i++) {
          Tensor<1,2> &dpvi=dpv[i];
          if (ftwb) {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[1]*dot0+dpvi[0]*dot1)*wq;
          } else {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[0]*dot0+dpvi[1]*dot1)*wq;
          }
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tface dof(f = " << f << ", i j = " << i+ioffset
               << " " << j << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
    ioffset+=dpfn;
  }
//interior
  GaussianQuadrature<1>
    gq1(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
    ProductQuadrature gq(tgq,gq1);
  const NumPtr<Point<3> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<3> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,3> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,3> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i]*pv[j];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*524288.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*524288.);
      }
    }
  }
}
#endif

#if (SPACEDIM==3)
void debugPrismDuality(const Nedelec2PrismDualShapeFunction *pols,
int degree,const Shape3 &element) {
//TRACER_CALL(t,"debugPrismDuality");
#ifdef DEBUG
//pols->printOn(cout);
#endif
  int pn=pols->getNumber();
  Nedelec2PrismDofShapeFunction dof_polys(degree);
  int dpn=dof_polys.getNumber();
  NumPtr<int> renumber;
  dof_polys.shapeFunctionsOrder(&element,renumber);
//edge dofs:
  GaussianQuadrature<1>
    egq(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
  const NumPtr<Point<1> > &egq_points=egq.points();
  const NumPtr<REAL> &egq_weights=egq.weights();
  int dpen=dof_polys.getEdgeNumber();
  int ioffset=0;
  for (int e=0;e<9;e++) {
    NumPtr<Point<3> > q_points_edge;
    q_points_edge.allocate(egq_points.getNumber());
    Prism::staticProjectToLine(e,egq_points,q_points_edge);
    Tensor<1,3> edge_tangent=Prism::lineTangent(e);
    if (element.lineTangentPointsBackward(e)) edge_tangent=-edge_tangent;
//  edge_tangent.normalize();
    LAMatrix matrix(dpen,pn,0.);
    for (int q=0;q<q_points_edge.getNumber();q++) {
      const Point<1> &eqp=egq_points[q];
      const Point<3> &qp=q_points_edge[q];
      NumPtr<REAL> dpv;
      dof_polys.computeEdgeValues(eqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=egq_weights[q];
      for (int j=0;j<pn;j++) {
        REAL dotj=edge_tangent*pv[j];
        for (int i=0;i<dpen;i++) {
          matrix(renumber[i+ioffset]-ioffset,j)+=dpv[i]*wq*dotj;
        }
      }
    }
#ifdef DEBUG
//  for (int i=0;i<dpen;i++) {
//    cout << "\tmatrix(" << i << ",*) = ";
//    for (int j=0;j<pn;j++) {
//      cout << matrix(i,j) << " ";
//    }
//    cout << endl;
//  }
#endif
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpen;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*32768.) {
          cout << "\tedge dof(" << e << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*32768.);
        }
      }
    }
    ioffset+=dpen;
  }
//triangle face dofs:
  TriangleGaussianQuadrature
    tgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<2> > &tgq_points=tgq.points();
  const NumPtr<REAL> &tgq_weights=tgq.weights();
  int dpfn=dof_polys.getTriangleFaceNumber();
  for (int f=0;f<2;f++) {
    bool ftwb=element.faceTangentsWindBackward(f);
    Tensor<1,3> ft0=element.faceTangent(f,0);
    Tensor<1,3> ft1=element.faceTangent(f,1);
//  ft0.normalize();
//  ft1.normalize();
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(tgq_points.getNumber());
    Prism::staticProjectToFace(f,tgq_points,q_points_face);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<2> &fqp=tgq_points[q];
      const Point<3> &qp=q_points_face[q];
      NumPtr<Tensor<1,2> > dpv;
      dof_polys.computeTriangleFaceValues(fqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=tgq_weights[q];
      for (int j=0;j<pn;j++) {
        Tensor<1,3> &pvj=pv[j];
        for (int i=0;i<dpfn;i++) {
          Tensor<1,2> &dpvi=dpv[i];
          if (ftwb) {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[1]*(pvj*ft0)+dpvi[0]*(pvj*ft1))*wq;
          } else {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[0]*(pvj*ft0)+dpvi[1]*(pvj*ft1))*wq;
          }
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*131072.) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*131072.);
        }
      }
    }
    ioffset+=dpfn;
  }
//quadrilateral face dofs:
  GaussianQuadrature<2>
    qgq(TriangleGaussianQuadrature::argGivenOrder(2*degree+2));
  const NumPtr<Point<2> > &qgq_points=qgq.points();
  const NumPtr<REAL> &qgq_weights=qgq.weights();
  dpfn=dof_polys.getQuadrilateralFaceNumber();
  for (int f=2;f<5;f++) {
    bool ftwb=element.faceTangentsWindBackward(f);
    Tensor<1,3> ft0=element.faceTangent(f,0);
    Tensor<1,3> ft1=element.faceTangent(f,1);
//  ft0.normalize();
//  ft1.normalize();
    NumPtr<Point<3> > q_points_face;
    q_points_face.allocate(qgq_points.getNumber());
    Prism::staticProjectToFace(f,qgq_points,q_points_face);
//  dynamic_cast<const Prism&>(element).projectToFaceForHcurl2(f,
//    qgq_points,q_points_face);
    LAMatrix matrix(dpfn,pn,0.);
    for (int q=0;q<q_points_face.getNumber();q++) {
      const Point<2> &fqp=qgq_points[q];
      const Point<3> &qp=q_points_face[q];
      NumPtr<Tensor<1,2> > dpv;
      dof_polys.computeQuadrilateralFaceValues(fqp,dpv);
      NumPtr<Tensor<1,3> > pv;
      pols->computeValues(qp,pv);
      REAL wq=qgq_weights[q];
      for (int j=0;j<pn;j++) {
        Tensor<1,3> &pvj=pv[j];
        for (int i=0;i<dpfn;i++) {
          Tensor<1,2> &dpvi=dpv[i];
          if (ftwb) {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[1]*(ft0*pvj)+dpvi[0]*(ft1*pvj))*wq;
          } else {
            matrix(renumber[i+ioffset]-ioffset,j)+=
              (dpvi[0]*(ft0*pvj)+dpvi[1]*(ft1*pvj))*wq;
          }
        }
      }
    }
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpfn;i++) {
        REAL error=
          (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
        if (error>DBL_EPSILON*1048576.) {
          cout << "\tface dof(" << f << "," << i << " " << j
               << ") = " << matrix(i,j)
               << ", error = " << error << " " << error/DBL_EPSILON
               << endl;
          CHECK_TEST(error<=DBL_EPSILON*1048576.);
        }
      }
    }
    ioffset+=dpfn;
  }
//interior
  GaussianQuadrature<1>
    gq1(GaussianQuadrature<1>::argGivenOrder(2*degree+2));
    ProductQuadrature gq(tgq,gq1);
  const NumPtr<Point<3> > &gq_points=gq.points();
  const NumPtr<REAL> &gq_weights=gq.weights();
  LAMatrix matrix(dpn,pn,0.);
  for (int q=0;q<gq_points.getNumber();q++) {
    const Point<3> &qp=gq_points[q];
    REAL wq=gq_weights[q];
    NumPtr<Tensor<1,3> > dpv;
    dof_polys.computeValues(qp,dpv);
    NumPtr<Tensor<1,3> > pv;
    pols->computeValues(qp,pv);
    for (int j=0;j<pn;j++) {
      for (int i=0;i<dpn;i++) {
        REAL dot=dpv[i]*pv[j];
        matrix(i,j)+=dot*wq;
      }
    }
  }
  for (int j=0;j<pn;j++) {
    for (int i=0;i<dpn;i++) {
      REAL error=
        (j==i+ioffset ? abs(matrix(i,j)-1.) : abs(matrix(i,j)));
      if (error>DBL_EPSILON*524288.) {
        cout << "\tinterior dof(" << i << " " << j << ") = "
             << matrix(i,j) << ", error= " << error << " "
             << error/DBL_EPSILON << endl;
        CHECK_TEST(error<=DBL_EPSILON*524288.);
      }
    }
  }
}
#endif

int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//
    {
      TRACER_CALL(t,"RTSimplexShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        RTSimplexShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tRTSimplexShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"RTSimplexDofShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        RTSimplexDofShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree,&pols);
#if (SPACEDIM>=2)
        debugFaceDerivs(degree,&pols);
#endif
      }
      cout << "\tRTSimplexDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"RTSimplexDualShapeFunction");
      for (int degree=0;degree<=D_CHOOSE(10,4,3);degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        RTSimplexDualShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
        debugSimplexDuality(&pols,degree);
      }
      cout << "\tRTSimplexDualShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMSimplexShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        BDFMSimplexShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tBDFMSimplexShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMSimplexDofShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        BDFMSimplexDofShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree,&pols);
#if (SPACEDIM>=2)
        debugFaceDerivs(degree,&pols);
#endif
      }
      cout << "\tBDFMSimplexDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMSimplexDualShapeFunction");
      for (int degree=0;degree<=D_CHOOSE(10,4,4);degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        BDFMSimplexDualShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
        debugSimplexDuality(&pols,degree);
      }
      cout << "\tBDFMSimplexDualShapeFunction OK" << endl;
    }
//
//
#if (SPACEDIM==3)
    {
      TRACER_CALL(t,"HcurlMultiplierTetrahedronShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        HcurlMultiplierTetrahedronShapeFunction pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tHcurlMultiplierTetrahedronShapeFunction OK" << endl;
    }
#endif
//
//
#if (SPACEDIM==3)
    {
      TRACER_CALL(t,"HcurlMultiplierPrismShapeFunction");
      for (int degreea=0;degreea<=10;degreea++)
//    int degreea=1;
      {
        cout << "\tdegreea = " << degreea << endl;
        for (int degreeb=degreea;degreeb<=degreea+1;degreeb++) {
          cout << "\tdegreeb = " << degreeb << endl;
          HcurlMultiplierPrismShapeFunction pols(degreea,degreeb);
          debugDerivs(degreeb+1,&pols);
        }
      }
      cout << "\tHcurlMultiplierPrismShapeFunction OK" << endl;
    }
#endif
//
//
#if (SPACEDIM==3)
    {
      TRACER_CALL(t,"HcurlMultiplierHexahedronShapeFunction");
      for (int degreea=0;degreea<=10;degreea++)
//    int degreea=1;
      {
        cout << "\tdegreea = " << degreea << endl;
        for (int degreeb=degreea;degreeb<=degreea+1;degreeb++) {
          cout << "\tdegreeb = " << degreeb << endl;
          HcurlMultiplierHexahedronShapeFunction pols(degreea,degreeb);
          debugDerivs(degreeb+1,&pols);
        }
      }
      cout << "\tHcurlMultiplierHexahedronShapeFunction OK" << endl;
    }
#endif
//
//
    {
      TRACER_CALL(t,"Nedelec1SimplexShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec1SimplexShapeFunction pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tNedelec1SimplexShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec1SimplexDofShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec1SimplexDofShapeFunction pols(degree);
        debugDerivs(degree,&pols);
#if (SPACEDIM>1)
        debugEdgeDerivs(degree,&pols);
#endif
      }
      cout << "\tNedelec1SimplexDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec1SimplexDualShapeFunction");
      for (int degree=0;degree<=D_CHOOSE(10,5,4);degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        PermutationGenerator pg(SPACEDIM);
        while (pg.hasMore())
        {
          int v[SPACEDIM+1]; v[0]=0;
          const NumPtr<int> &p=pg.getNext();
          for (int i=0;i<SPACEDIM;i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=3; v[3]=2;
#ifdef DEBUG
//        cout << "\tv = ";
//        for (int i=0;i<=SPACEDIM;i++) cout << v[i] << " ";
//        cout << endl;
#endif
#if (SPACEDIM==1)
          Interval e(-1,v);
#elif (SPACEDIM==2)
          Triangle e(-1,v,v);
#else
          int l[6]={0,1,2,3,4,5};
          Tetrahedron e(-1,l,v,v);
#endif
          Nedelec1SimplexDualShapeFunction pols(degree);
          pols.computeCoefficientsFor(&e);
          debugSimplexDuality(&pols,degree,e);
          debugDerivs(degree+1,&pols);
        }
      }
      cout << "\tNedelec1SimplexDualShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2SimplexShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec2SimplexShapeFunction pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tNedelec2SimplexShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2SimplexDofShapeFunction");
      for (int degree=1;degree<=10;degree++)
//    int degree=4;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec2SimplexDofShapeFunction pols(degree);
        debugDerivs(degree,&pols);
#if (SPACEDIM>1)
        debugNedelec2EdgeDerivs(degree,&pols);
#endif
      }
      cout << "\tNedelec2SimplexDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2SimplexDualShapeFunction");
      for (int degree=1;degree<=D_CHOOSE(10,5,2);degree++)
//    int degree=3;
      {
        cout << "\tdegree = " << degree << endl;
        PermutationGenerator pg(SPACEDIM);
        while (pg.hasMore())
        {
          const NumPtr<int> &p=pg.getNext();
          int v[SPACEDIM+1]; v[0]=0;
          for (int i=0;i<SPACEDIM;i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=2; v[3]=3;
#if (SPACEDIM==1)
          Interval e(-1,v);
#elif (SPACEDIM==2)
          Triangle e(-1,v,v);
//        cout << "\tv = ";
//        for (int i=0;i<=SPACEDIM;i++) cout << v[i] << " ";
//        cout << endl;
#else
          int l[6]={0,1,2,3,4,5};
          Tetrahedron e(-1,l,v,v);
#endif
          Nedelec2SimplexDualShapeFunction pols(degree);
          pols.computeCoefficientsFor(&e);
          debugSimplexDuality(&pols,degree,e);
          debugDerivs(degree+1,&pols);
        }
      }
      cout << "\tNedelec2SimplexDualShapeFunction OK" << endl;
    }
//
#if (SPACEDIM==2)
//
    {
      TRACER_CALL(t,"NedelecXShapeFunction2");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        NedelecXShapeFunction2 pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tNedelecXShapeFunction2 OK" << endl;
    }
//
#endif
#if (SPACEDIM>=2)
//
    {
      TRACER_CALL(t,"RTParallelpipedShapeFunction");
      for (int degree=0;degree<=10;degree++) {
        cout << "\tdegree = " << degree << endl;
        RTParallelpipedShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tRTParallelpipedShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"RTParallelpipedDofShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        RTParallelpipedDofShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree,&pols);
        debugFaceDerivs(degree,&pols);
      }
      cout << "\tRTParallelpipedDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"RTParallelpipedDualShapeFunction");
      for (int degree=0;degree<=D_CHOOSE(0,8,4);degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        RTParallelpipedDualShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
        debugParallelpipedDuality(&pols,degree);
      }
      cout << "\tRTParallelpipedDualShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMParallelpipedShapeFunction");
      for (int degree=0;degree<=D_CHOOSE(0,10,6);degree++) {
        cout << "\tdegree = " << degree << endl;
        BDFMParallelpipedShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tBDFMParallelpipedShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMParallelpipedDofShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        BDFMParallelpipedDofShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree,&pols);
        debugFaceDerivs(degree,&pols);
      }
      cout << "\tBDFMParallelpipedDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMParallelpipedDualShapeFunction");
      for (int degree=0;degree<=D_CHOOSE(0,5,3);degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        BDFMParallelpipedDualShapeFunction pols(degree);
# if (SPACEDIM==2)
        PermutationGenerator pg(3);
        while (pg.hasMore())
        {
          const NumPtr<int> &p=pg.getNext();
//        cout << "\tp = ";
//        for (int i=0;i<3;i++) cout << p[i] << " ";
//        cout << endl;
          int v[D_CHOOSE(2,4,8)]; v[0]=0;
          for (int i=0;i<p.getNumber();i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=3; v[3]=2;
          Quadrilateral e(-1,v,v);
          pols.computeCoefficientsFor(&e);
          debugPoly(&pols,degree);
          debugDerivs(degree+1,&pols);
          debugParallelpipedDuality(&e,&pols,degree);
        }
# else
        for (int count=0;count<200;count++)
        {
          NumPtr<int> p(7);
          RandomPermutation(p);
          int v[D_CHOOSE(2,4,8)]; v[0]=0;
          for (int i=0;i<p.getNumber();i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=2; v[3]=3; v[4]=5; v[5]=6; v[6]=4; v[7]=7;
          int l[12]={0,1,2,3,4,5,6,7,8,9,10,11};
          int f[6]={0,1,2,3,4,5};
          Hexahedron e(-1,l,f,v);
          pols.computeCoefficientsFor(&e);
          debugPoly(&pols,degree);
          debugDerivs(degree+1,&pols);
          debugParallelpipedDuality(&e,&pols,degree);
        }
# endif
      }
      cout << "\tBDFMParallelpipedDualShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec1ParallelpipedShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec1ParallelpipedShapeFunction pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tNedelec1ParallelpipedShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec1ParallelpipedDofShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec1ParallelpipedDofShapeFunction pols(degree);
        debugDerivs(degree,&pols);
        debugEdgeDerivs(degree,&pols);
      }
      cout << "\tNedelec1ParallelpipedDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec1ParallelpipedDualShapeFunction");
      for (int degree=0;degree<=D_CHOOSE(0,6,2);degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
//      PermutationGenerator pg(D_CHOOSE(1,3,7));
//      while (pg.hasMore())
        for (int count=0;count<200;count++)
        {
//        const NumPtr<int> &p=pg.getNext();
          NumPtr<int> p(D_CHOOSE(1,3,7));
          RandomPermutation(p);
          int v[D_CHOOSE(2,4,8)]; v[0]=0;
          for (int i=0;i<p.getNumber();i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=2; v[3]=3; v[4]=5; v[5]=6; v[6]=4; v[7]=7;
# if (SPACEDIM==2)
          Quadrilateral e(-1,v,v);
# else
          int l[12]={0,1,2,3,4,5,6,7,8,9,10,11};
          int f[6]={0,1,2,3,4,5};
          Hexahedron e(-1,l,f,v);
# endif
          Nedelec1ParallelpipedDualShapeFunction pols(degree);
          pols.computeCoefficientsFor(&e);
          debugParallelpipedDuality(&pols,degree,e);
          debugDerivs(degree+1,&pols);
        }
      }
      cout << "\tNedelec1ParallelpipedDualShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2ParallelpipedShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec2ParallelpipedShapeFunction pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tNedelec2ParallelpipedShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2ParallelpipedDofShapeFunction");
      for (int degree=1;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec2ParallelpipedDofShapeFunction pols(degree);
        debugDerivs(degree,&pols);
        debugEdgeDerivs(degree,&pols);
      }
      cout << "\tNedelec2ParallelpipedDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2ParallelpipedDualShapeFunction");
      for (int degree=1;degree<=D_CHOOSE(0,5,3);degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
//      PermutationGenerator pg(D_CHOOSE(1,3,7));
//      while (pg.hasMore())
        for (int count=0;count<200;count++)
        {
//        const NumPtr<int> &p=pg.getNext();
          NumPtr<int> p(D_CHOOSE(1,3,7));
          RandomPermutation(p);
          int v[D_CHOOSE(2,4,8)]; v[0]=0;
          for (int i=0;i<p.getNumber();i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=2; v[3]=3; v[4]=5; v[5]=6; v[6]=4; v[7]=7;
# if (SPACEDIM==2)
          Quadrilateral e(-1,v,v);
# else
          int l[12]={0,1,2,3,4,5,6,7,8,9,10,11};
          int f[6]={0,1,2,3,4,5};
          Hexahedron e(-1,l,f,v);
# endif
          Nedelec2ParallelpipedDualShapeFunction pols(degree);
          pols.computeCoefficientsFor(&e);
          debugParallelpipedDuality(&pols,degree,e);
          debugDerivs(degree+1,&pols);
        }
      }
      cout << "\tNedelec2ParallelpipedDualShapeFunction OK" << endl;
    }
//
#endif
#if (SPACEDIM==3)
//
    {
      TRACER_CALL(t,"RTPrismShapeFunction");
      for (int degree=0;degree<=9;degree++) 
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        RTPrismShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tRTPrismShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"RTPrismDofShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        RTPrismDofShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree,&pols);
        debugPrismFaceDerivs(degree,&pols);
      }
      cout << "\tRTPrismDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"RTPrismDualShapeFunction");
      for (int degree=0;degree<=4;degree++)//higher order bad rounding 
//    int degree=0;
      {
        cout << "\tdegree = " << degree << endl;
        RTPrismDualShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugPrismDuality(&pols,degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tRTPrismDualShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMPrismShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        BDFMPrismShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tBDFMPrismShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMPrismDofShapeFunction");
      for (int degree=0;degree<=10;degree++) 
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        BDFMPrismDofShapeFunction pols(degree);
        debugPoly(&pols,degree);
        debugDerivs(degree,&pols);
        debugPrismFaceDerivs(degree,&pols);
      }
      cout << "\tBDFMPrismDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"BDFMPrismDualShapeFunction");
      for (int degree=0;degree<=3;degree++)//higher order bad rounding 
//    int degree=0;
      {
        cout << "\tdegree = " << degree << endl;
        BDFMPrismDualShapeFunction pols(degree);
        PermutationGenerator pg(5);
        while (pg.hasMore())
        {
          const NumPtr<int> &p=pg.getNext();
          int v[6]; v[0]=0;
          for (int i=0;i<p.getNumber();i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=2; v[3]=5; v[4]=3; v[5]=4;
          int l[9]={0,1,2,3,4,5,6,7,8};
          int f[5]={0,1,2,3,4};
          Prism e(-1,l,f,v);
          pols.computeCoefficientsFor(&e);
          debugPoly(&pols,degree);
          debugPrismDuality(&e,&pols,degree);
          debugDerivs(degree+1,&pols);
        }
      }
      cout << "\tBDFMPrismDualShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"NedelecXShapeFunction3");
      for (int degree=0;degree<=8;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        NedelecXShapeFunction3 pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tNedelecXShapeFunction3 OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec1PrismShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec1PrismShapeFunction pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tNedelec1PrismShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec1PrismDofShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec1PrismDofShapeFunction pols(degree);
        debugDerivs(degree,&pols);
        debugPrismEdgeDerivs(degree,&pols);
//      debugPrismFaceDerivs(degree,&pols);
      }
      cout << "\tNedelec1PrismDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec1PrismDualShapeFunction");
      for (int degree=0;degree<=3;degree++)
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        PermutationGenerator pg(5);
        while (pg.hasMore())
        {
          const NumPtr<int> &p=pg.getNext();
          int v[6]; v[0]=0;
          for (int i=0;i<p.getNumber();i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=2; v[3]=5; v[4]=3; v[5]=4;
          int l[9]={0,1,2,3,4,5,6,7,8};
          int f[5]={0,1,2,3,4};
          Prism e(-1,l,f,v);
          Nedelec1PrismDualShapeFunction pols(degree);
          pols.computeCoefficientsFor(&e);
          debugPrismDuality(&pols,degree,e);
          debugDerivs(degree+1,&pols);
        }
      }
      cout << "\tNedelec1PrismDualShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2PrismShapeFunction");
      for (int degree=0;degree<=10;degree++)
//    int degree=1;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec2PrismShapeFunction pols(degree);
        debugDerivs(degree+1,&pols);
      }
      cout << "\tNedelec2PrismShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2PrismDofShapeFunction");
      for (int degree=1;degree<=10;degree++)
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        Nedelec2PrismDofShapeFunction pols(degree);
        debugDerivs(degree,&pols);
        debugNedelec2PrismEdgeDerivs(degree,&pols);
//      debugPrismFaceDerivs(degree,&pols);
      }
      cout << "\tNedelec2PrismDofShapeFunction OK" << endl;
    }
//
//
    {
      TRACER_CALL(t,"Nedelec2PrismDualShapeFunction");
      for (int degree=1;degree<=2;degree++)
//    int degree=2;
      {
        cout << "\tdegree = " << degree << endl;
        PermutationGenerator pg(5);
        while (pg.hasMore())
        {
          const NumPtr<int> &p=pg.getNext();
          int v[6]; v[0]=0;
          for (int i=0;i<p.getNumber();i++) v[i+1]=p[i]+1;
//        v[1]=1; v[2]=2; v[3]=5; v[4]=3; v[5]=4;
          int l[9]={0,1,2,3,4,5,6,7,8};
          int f[5]={0,1,2,3,4};
          Prism e(-1,l,f,v);
          Nedelec2PrismDualShapeFunction pols(degree);
          pols.computeCoefficientsFor(&e);
          debugPrismDuality(&pols,degree,e);
          debugDerivs(degree+1,&pols);
        }
      }
      cout << "\tNedelec2PrismDualShapeFunction OK" << endl;
    }
//
#endif
  }
  return EXIT_SUCCESS;
}

#include "TensorShapeFunction.C"
//template class TensorShapeFunction<SPACEDIM>;
#if (SPACEDIM==3)
//template class TensorShapeFunction<2>;
//template class HomogeneousPolynomials<3>;
#endif
