// "$Header:$"
//----------------------------  tria.cc  ---------------------------
//    $Id: tria.cc,v 1.151 2003/05/01 14:47:12 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  tria.cc  ---------------------------
//
//modified from deal.II/deal.II/source/grid/tria.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <algorithm>
#include <cmath>
#include <iostream>
#include <map>
#include <numeric>
#include <set>
#include <Tessellation.H>
#include <Boundary.H>
#include <MemoryDebugger.H>
#include <NIDLList.H>
#include <NISLList.H>
#include <ShapeFunction.H>
#include <Tracer.H>

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void BoundaryData::printOn(ostream &os) const {
  for (int i=0;i<boundary_faces.getNumber();i++) {
    cout << "\tboundary_faces[" << i << "] = " << endl;
    boundary_faces[i].printOn(cout);
  }
#if (SPACEDIM==3)
  for (int i=0;i<boundary_lines.getNumber();i++) {
    cout << "\tboundary_lines[" << i << "] = " << endl;
    boundary_lines[i].printOn(cout);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
ostream& operator<<(ostream &os,const Tessellation::MeshSmoothing &ms) {
  switch (ms) {
    case Tessellation::mesh_smoothing_none:
      os << "none";
      break;
    case 
    Tessellation::mesh_smoothing_limit_level_difference_at_vertices:
      os << "limit level difference at vertices";
      break;
    case Tessellation::mesh_smoothing_eliminate_unrefined_islands:
      os << "eliminate unrefined islands";
      break;
    case Tessellation::mesh_smoothing_patch_level_1:
      os << "patch level 1";
      break;
    case Tessellation::mesh_smoothing_eliminate_refined_inner_islands:
      os << "eliminate refined inner islands";
      break;
    case 
    Tessellation::mesh_smoothing_eliminate_refined_boundary_islands:
      os << "eliminate refined boundary islands";
      break;
    case Tessellation::mesh_smoothing_do_not_produce_unrefined_islands:
      os << "do not produce unrefined islands";
      break;
//  case Tessellation::mesh_smoothing_on_refinement:
//    os << "on refinement";
//    break;
//  case Tessellation::mesh_smoothing_on_coarsening:
//    os << "on coarsening";
//    break;
  }
  return os;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Tessellation::~Tessellation() {
  if (levels_own_element_data_shapes) {
    for (LEVEL_CLASS *level=coarsest_level;level;
    level=level->finerLevel()) {
      level->deleteElementDataShapes();
    }
  }
  delete coarsest_level;
  vertices.cleanup();
  vertices_used.cleanup();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
BoundaryData* Tessellation::create(const NumPtr<Point<SPACEDIM> > &v,
const NumPtr<CellData> &cells) {
  CHECK_SAME(vertices.getNumber(),0);
  CHECK_TEST(coarsest_level==0);
  vertices.allocate(v.getNumber());
  vertices.copyFrom(v);
  vertices_used.allocate(vertices.getNumber());
  vertices_used.initialize(true);
  
  NumPtr<NISLList<int> > lines_at_vertex(vertices.getNumber());

  coarsest_level=OPERATOR_NEW Level1(0);
  coarsest_level->Level0::reserveSpace(cells.getNumber());
  coarsest_level->Level1::reserveSpace(cells.getNumber());

  firstShape(coarsest_level);
  for (int cell=0;cell<cells.getNumber();cell++,nextShape()) {
    int l[2]={cells[cell].vertex(0),cells[cell].vertex(1)};
    present_level->setShape(OPERATOR_NEW 
      Interval(present_level->presentIndex(),l));
    present_level->setUsedFlag();
    present_level->setMaterialModel(cells[cell].getMaterialModel());
    present_level->clearUserFlag();
    present_level->clearUserPointer();
    present_level->setSubdomainId(0);

    lines_at_vertex[cells[cell].vertex(0)].
      append(OPERATOR_NEW int(cell));
    lines_at_vertex[cells[cell].vertex(1)].
      append(OPERATOR_NEW int(cell));
  }
  int boundary_nodes=0;
  for (int i=0;i<lines_at_vertex.getNumber();++i) {
    int lav=lines_at_vertex[i].length();
    CHECK_BOUNDS(lav,1,3);
    if (lav==1) {
      boundary_nodes++;
    }
  }
  CHECK_TEST(boundary_nodes==2);

  BoundaryCondition bogus_boundary_condition; 
  BoundaryData *boundary_data=OPERATOR_NEW BoundaryData();
  boundary_data->boundary_faces.allocate(2);
  for (bool doit=firstActiveShape(coarsest_level);doit;
  doit=nextActiveShape()) {
    int element_index=present_level->presentIndex();
    const SHAPE_CLASS *element=present_level->getShape();
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const NISLList<int> &elements_at_face=lines_at_vertex[face_index];
      NISLListNode<int> *node=elements_at_face.first();
      int ei=*(node->getData());
      if (ei==element_index) {
        NISLListNode<int> *other_node=elements_at_face.next(node);
        if (other_node) {
          present_level->
            setNeighbor(f,present_level,*(other_node->getData()));
        } else {
          present_level->setNeighbor(f,0,-1);
          present_level->
            Level0::setBoundaryCondition(f,&bogus_boundary_condition);
          FaceData &boundary_vertex=
            boundary_data->boundary_faces[f];
          boundary_vertex.allocate(1);
          boundary_vertex.vertex(0)=element->vertexIndex(f);
          boundary_vertex.
            setBoundaryCondition(&bogus_boundary_condition);
        }
      } else {
        present_level->setNeighbor(f,coarsest_level,ei);
      }
    }
  }
  for (int i=0;i<lines_at_vertex.getNumber();i++) {
    while (lines_at_vertex[i].notEmpty()) {
      delete lines_at_vertex[i].delAfter(0);
    }
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
BoundaryData* Tessellation::create(const NumPtr<Point<SPACEDIM> > &v,
const NumPtr<CellData> &cells) {
  CHECK_SAME(vertices.getNumber(),0);
  CHECK_TEST(coarsest_level==0);
  vertices.allocate(v.getNumber());
  vertices.copyFrom(v);
  vertices_used.allocate(vertices.getNumber());
  vertices_used.initialize(true);

//the following leaks 1280 bytes
  map<pair<int,int>,int> needed_lines;

//store the cell vertex indices as ordered pairs in needed_lines
//  since neighboring triangles will traverse the lines in reverse order
//  store the pairs in order of increasing vertex index
  for (int cell=0;cell<cells.getNumber();cell++) {
    const CellData &cd=cells[cell];
    int nv=cd.numberVertices();
    for (int vertex=0;vertex<nv;++vertex) {
      int cv=cd.vertex(vertex);
      CHECK_BOUNDS(cv,0,vertices.getNumber());
    }
    if (nv==4) {
      int cv[4]={cd.vertex(0),cd.vertex(1),cd.vertex(2),cd.vertex(3)};
      for (int l=0;l<4;l++) {
        needed_lines[MAKE_LINE_PAIR(
          cv[Quadrilateral::staticElementVertexFromFace(l,0)],
          cv[Quadrilateral::staticElementVertexFromFace(l,1)])]=-1;
      }
    } else {
      int cv[3]={cd.vertex(0),cd.vertex(1),cd.vertex(2)};
      for (int l=0;l<3;l++) {
        needed_lines[MAKE_LINE_PAIR(
          cv[Triangle::staticElementVertexFromFace(l,0)],
          cv[Triangle::staticElementVertexFromFace(l,1)])]=-1;
      }
    }
  }

//count the number of times each vertex appears 
//  in one of the ordered pairs in needed_lines
  {
    NumPtr<int> vertex_touch_count(v.getNumber());
    vertex_touch_count.initialize(0);
    map<pair<int,int>,int>::iterator i;
    for (i=needed_lines.begin(); i!=needed_lines.end(); i++) {
      ++vertex_touch_count[i->first.first];
      ++vertex_touch_count[i->first.second];
    }
  }

//construct new Level with no coarser Level
//store needed_lines in codimension1 Level
  coarsest_level=OPERATOR_NEW Level2(0);
  coarsest_level->Level0::reserveSpace(cells.getNumber());
  coarsest_level->Level1::reserveSpace(needed_lines.size());
  coarsest_level->Level2::reserveSpace(cells.getNumber());
  {
    map<pair<int,int>,int>::iterator i=needed_lines.begin();
    for (bool doit=firstCodimension1Shape();doit;
    doit=nextCodimension1Shape(),i++) {
      int cv[2]={i->first.first,i->first.second};
      present_codimension1_level->setShape(OPERATOR_NEW Interval(
        present_codimension1_level->presentIndex(),cv));
      present_codimension1_level->setUsedFlag();
      present_codimension1_level->clearUserFlag();
      present_codimension1_level->clearUserPointer();
      i->second=present_codimension1_level->presentIndex();
    }
  }

//construct Shapes and store in Level2
//for each element side, append cell index to list of adjacent cells
  map<int,NISLList<int> > adjacent_cells;
  {
    firstShape();
    for (int c=0;c<cells.getNumber();c++,nextShape()) {
      const CellData &cd=cells[c];
      if (cd.numberVertices()==4) {
        int cv[4]={cd.vertex(0),cd.vertex(1),cd.vertex(2),cd.vertex(3)};
        int lines[4];
        for (int l=0;l<4;l++) {
          lines[l]=needed_lines[MAKE_LINE_PAIR(
            cv[Quadrilateral::staticElementVertexFromFace(l,0)],
            cv[Quadrilateral::staticElementVertexFromFace(l,1)])];
        }
        present_level->setShape(OPERATOR_NEW Quadrilateral(
          present_level->presentIndex(),lines,cv));
        present_level->setUsedFlag();
        present_level->setMaterialModel(cd.getMaterialModel());
        present_level->clearUserFlag();
        present_level->clearUserPointer();
        present_level->setSubdomainId(0);

        for (int line=0; line<4; ++line) {
          adjacent_cells[lines[line]].
            append(OPERATOR_NEW int(present_level->presentIndex()));
        }
#ifdef DEBUG
        present_codimension1_level=static_cast<Level1*>(present_level);
        NumPtr<int> vtc(4);
        vtc.initialize(0);
        for (int i=0;i<4;i++) {
          for (int j=0;j<2;j++) {
            int vi=present_codimension1_level->vertexIndex(lines[i],j);
            for (int k=0;k<4;k++) {
              if (vi==cv[k]) { vtc[k]++; break; }
            }
          }
        }
        for (int i=0;i<4;i++) CHECK_SAME(vtc[i],2);

        CHECK_TEST(present_level->getShape()->volume(vertices)>0.);
        for (int f=0;f<4;f++) {
          int face_element_index=
            present_level->getShape()->codimension1ShapeIndex(f);
          const Shape1 *face_element=
            present_level->Level1::getShape(face_element_index);
          present_level->getShape()->checkFaceVertices(f,face_element);
        }
#endif
      } else {
        int cv[3]={cd.vertex(0),cd.vertex(1),cd.vertex(2)};
        int lines[3];
        for (int l=0;l<3;l++) {
          lines[l]=needed_lines[MAKE_LINE_PAIR(
            cv[Triangle::staticElementVertexFromFace(l,0)],
            cv[Triangle::staticElementVertexFromFace(l,1)])];
        }
        present_level->setShape(OPERATOR_NEW Triangle(
          present_level->presentIndex(),lines,cv));
        present_level->setUsedFlag();
        present_level->setMaterialModel(cd.getMaterialModel());
        present_level->clearUserFlag();
        present_level->clearUserPointer();
        present_level->setSubdomainId(0);

        for (int line=0; line<3; ++line) {
          adjacent_cells[lines[line]].
            append(OPERATOR_NEW int(present_level->presentIndex()));
        }
#ifdef DEBUG
        present_codimension1_level=static_cast<Level1*>(present_level);
        NumPtr<int> vtc(3);
        vtc.initialize(0);
        for (int i=0;i<3;i++) {
          for (int j=0;j<2;j++) {
            int vi=present_codimension1_level->vertexIndex(lines[i],j);
            for (int k=0;k<3;k++) {
              if (vi==cv[k]) { vtc[k]++; break; }
            }
          }
        }
        for (int i=0;i<3;i++) CHECK_SAME(vtc[i],2);

        CHECK_TEST(present_level->getShape()->volume(vertices)>0.);
        for (int f=0;f<3;f++) {
          int face_element_index=
            present_level->getShape()->codimension1ShapeIndex(f);
          const Shape1 *face_element=
            present_level->Level1::getShape(face_element_index);
          present_level->getShape()->checkFaceVertices(f,face_element);
        }
#endif
      }
    }
  }

//for element sides with one adjacent cell (boundary line), set id=0
//  else (interior line) set id=MAX_MATERIAL_ID
//count number of boundary lines

//the following will go out of scope
//it must be replaced without being deleted
  BoundaryCondition bogus_boundary_condition; 
  int n_boundary_lines=0;
  for (bool doit=firstCodimension1Shape();doit;
  doit=nextCodimension1Shape()) {
    int n_adj_cells=adjacent_cells[
      present_codimension1_level->presentIndex()].length();
    CHECK_BOUNDS(n_adj_cells,1,3);
    if (n_adj_cells==1) {
      present_codimension1_level->
        setBoundaryCondition(&bogus_boundary_condition);
      n_boundary_lines++;
    } else {
      present_codimension1_level->setBoundaryCondition(0);
    }
  }

  BoundaryData *boundary_data=OPERATOR_NEW BoundaryData();
  boundary_data->boundary_faces.allocate(n_boundary_lines);
  n_boundary_lines=0;
  for (bool doit=firstCodimension1Shape();doit;
  doit=nextCodimension1Shape()) {
    int line_index=present_codimension1_level->presentIndex();
    if (adjacent_cells[line_index].length()==1) {
      FaceData &boundary_line=
        boundary_data->boundary_faces[n_boundary_lines];
      boundary_line.allocate(2);
      boundary_line.vertex(0)=
        present_codimension1_level->vertexIndex(0);
      boundary_line.vertex(1)=
        present_codimension1_level->vertexIndex(1);
      boundary_line.setBoundaryCondition(&bogus_boundary_condition);
      n_boundary_lines++;
    }
  }
//set neighbor of each side of each element
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    const Shape2 *element=present_level->getShape();
    int element_index=present_level->presentIndex();
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      NISLList<int> &adjacent_cell_list=
        adjacent_cells[element->codimension1ShapeIndex(f)];
      NISLListNode<int> *first_adjacent_cell_node=
        adjacent_cell_list.first();
      int adjacent_cell=*first_adjacent_cell_node->getData();
      if (adjacent_cell==element_index) {
        NISLListNode<int> *other_adjacent_cell_node=
          adjacent_cell_list.next(first_adjacent_cell_node);
        if (other_adjacent_cell_node) {
          present_level->setNeighbor(f,present_level,
            *other_adjacent_cell_node->getData());
        } else present_level->setNeighbor(f,0,-1);
      } else {
        present_level->setNeighbor(f,present_level,adjacent_cell);
      }
    }
  }

//delete lists of adjacent cells
  map<int,NISLList<int> >::iterator adjacent_cell;
  for (adjacent_cell=adjacent_cells.begin();
  adjacent_cell!=adjacent_cells.end();adjacent_cell++) {
    NISLList<int> &adjacent_cell_list=adjacent_cell->second;
    while (adjacent_cell_list.notEmpty()) {
      delete adjacent_cell_list.delAfter(0);
    }
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
BoundaryData* Tessellation::create(const NumPtr<Point<3> > &v,
const NumPtr<CellData> &cells) {
  CHECK_SAME(vertices.getNumber(),0);
  CHECK_TEST(coarsest_level==0);
  vertices.allocate(v.getNumber());
  vertices.copyFrom(v);
  vertices_used.allocate(vertices.getNumber());
  vertices_used.initialize(true);

//store the cell vertex indices as ordered pairs in needed_lines
  map<pair<int,int>,int> needed_lines;
  for (int cell=0;cell<cells.getNumber();cell++) {
    const CellData &cd=cells[cell];
    int nv=cd.numberVertices();
    for (int vertex=0; vertex<nv; ++vertex) {
      int cv=cd.vertex(vertex);
      CHECK_BOUNDS(cv,0,vertices.getNumber());
    }
    switch (nv) {
      case 4: {
        int cv[4]={cd.vertex(0),cd.vertex(1),cd.vertex(2),cd.vertex(3)};
        for (int l=0;l<Tetrahedron::linesPerShape();l++) {
          needed_lines[MAKE_LINE_PAIR(
            cv[Tetrahedron::staticElementVertexFromLine(l,0)],
            cv[Tetrahedron::staticElementVertexFromLine(l,1)])]=-1;
        }
        break;
      }
      case 6: {
        int cv[6]={cd.vertex(0),cd.vertex(1),cd.vertex(2),
                   cd.vertex(3),cd.vertex(4),cd.vertex(5)};
        for (int l=0;l<Prism::linesPerShape();l++) {
          needed_lines[MAKE_LINE_PAIR(
            cv[Prism::staticElementVertexFromLine(l,0)],
            cv[Prism::staticElementVertexFromLine(l,1)])]=-1;
        }
        break;
      }
      case 8: {
        int cv[8]={cd.vertex(0),cd.vertex(1),cd.vertex(2),cd.vertex(3),
                   cd.vertex(4),cd.vertex(5),cd.vertex(6),cd.vertex(7)};
        for (int l=0;l<Hexahedron::linesPerShape();l++) {
          needed_lines[MAKE_LINE_PAIR(
            cv[Hexahedron::staticElementVertexFromLine(l,0)],
            cv[Hexahedron::staticElementVertexFromLine(l,1)])]=-1;
        }
        break;
      }
      default:
        OBSOLETE("invalid numberVertices");
    }
  }

#ifdef DEBUG
//count the number of times each vertex appears 
//  in one of the ordered pairs in needed_lines
  {
    NumPtr<int> vertex_touch_count(v.getNumber());
    vertex_touch_count.initialize(0);
    map<pair<int,int>,int>::iterator i;
    for (i=needed_lines.begin(); i!=needed_lines.end(); i++) {
      ++vertex_touch_count[i->first.first];
      ++vertex_touch_count[i->first.second];
    };
    for (int i=0;i<vertex_touch_count.getNumber();i++) {
      if (vertex_touch_count[i]<3) {
        cout << "\tv[" << i << "] = " << v[i] << endl;
        cout << "\tvertex_touch_count = " << vertex_touch_count[i] 
             << endl;
        CHECK_TEST(vertex_touch_count[i]>=3);
      }
    }
  }
#endif

//construct new Level with no coarser Level
//store needed_lines in codimension2 Level
  coarsest_level=OPERATOR_NEW Level3(0);
  coarsest_level->Level0::reserveSpace(cells.getNumber());
  coarsest_level->Level1::reserveSpace(needed_lines.size());
  {
    map<pair<int,int>,int>::iterator i=needed_lines.begin();
    for (bool doit=firstCodimension2Shape();doit;
    doit=nextCodimension2Shape(),i++) {
      int cv[2]={i->first.first,i->first.second};
      present_codimension2_level->setShape(OPERATOR_NEW Interval(
        present_codimension2_level->presentIndex(),cv));
      present_codimension2_level->setUsedFlag();
      present_codimension2_level->clearUserFlag();
      present_codimension2_level->clearUserPointer();
      i->second=present_codimension2_level->presentIndex();
    }
  }

//construct faces and store in needed_faces
  map<pair<pair<int,int>,pair<int,int> >,int> needed_faces;
  for (int cell=0;cell<cells.getNumber();cell++) {
    const CellData &cd=cells[cell];
    if (cd.numberVertices()==Tetrahedron::verticesPerShape()) {
      for (int f=0;f<Tetrahedron::facesPerShape();f++) {
        int tl[3]={-1,-2,-3};
        int tv[3]={
          cd.vertex(Tetrahedron::staticElementVertexFromFace(f,0)),
          cd.vertex(Tetrahedron::staticElementVertexFromFace(f,1)),
          cd.vertex(Tetrahedron::staticElementVertexFromFace(f,2))};
        Triangle bogus_triangle(-1,tl,tv);
        pair<pair<int,int>,pair<int,int> > 
          p(make_pair(bogus_triangle.vertexIndex(0),
                      bogus_triangle.vertexIndex(1)),
            make_pair(bogus_triangle.vertexIndex(2),-1));
        needed_faces[p]=-1;
      }
    } else if (cd.numberVertices()==Prism::verticesPerShape()) {
      for (int f=0;f<Prism::facesPerShape();f++) {
        if (Prism::staticNumberVerticesOnFace(f)==3) {
          int tl[3]={-1,-2,-3};
          int tv[3]={
            cd.vertex(Prism::staticElementVertexFromFace(f,0)),
            cd.vertex(Prism::staticElementVertexFromFace(f,1)),
            cd.vertex(Prism::staticElementVertexFromFace(f,2))};
          Triangle bogus_triangle(-1,tl,tv);
          pair<pair<int,int>,pair<int,int> > 
            p(make_pair(bogus_triangle.vertexIndex(0),
                        bogus_triangle.vertexIndex(1)),
              make_pair(bogus_triangle.vertexIndex(2),-1));
          needed_faces[p]=-1;
        } else {
          int ql[4]={-1,-2,-3,-4};
          int qv[4]={
            cd.vertex(Prism::staticElementVertexFromFace(f,0)),
            cd.vertex(Prism::staticElementVertexFromFace(f,1)),
            cd.vertex(Prism::staticElementVertexFromFace(f,2)),
            cd.vertex(Prism::staticElementVertexFromFace(f,3))};
          Quadrilateral bogus_quadrilateral(-1,ql,qv);
          pair<pair<int,int>,pair<int,int> > 
            p(make_pair(bogus_quadrilateral.vertexIndex(0),
                        bogus_quadrilateral.vertexIndex(1)),
              make_pair(bogus_quadrilateral.vertexIndex(2),
                        bogus_quadrilateral.vertexIndex(3)));
          needed_faces[p]=-1;
        }
      }
    } else {
      for (int f=0;f<Hexahedron::facesPerShape();f++) {
        int ql[4]={-1,-2,-3,-4};
        int qv[4]={
          cd.vertex(Hexahedron::staticElementVertexFromFace(f,0)),
          cd.vertex(Hexahedron::staticElementVertexFromFace(f,1)),
          cd.vertex(Hexahedron::staticElementVertexFromFace(f,2)),
          cd.vertex(Hexahedron::staticElementVertexFromFace(f,3))};
        Quadrilateral bogus_quadrilateral(-1,ql,qv);
        pair<pair<int,int>,pair<int,int> > 
          p(make_pair(bogus_quadrilateral.vertexIndex(0),
                      bogus_quadrilateral.vertexIndex(1)),
            make_pair(bogus_quadrilateral.vertexIndex(2),
                      bogus_quadrilateral.vertexIndex(3)));
        needed_faces[p]=-1;
      }
    }
  }

//construct faces and store in Level2
  coarsest_level->Level2::reserveSpace(needed_faces.size());
  {
    map<pair<pair<int,int>,pair<int,int> >,int>::iterator 
      q=needed_faces.begin();
    for (bool doit=firstCodimension1Shape();doit;
    doit=nextCodimension1Shape(),q++) {
      if (q->first.second.second<0) {
        int cv[3]={q->first.first.first,q->first.first.second,
                   q->first.second.first};
        int cl[3];
        for (int f=0;f<3;f++) {
          pair<int,int> p=MAKE_LINE_PAIR(
            cv[Triangle::staticElementVertexFromFace(f,0)],
            cv[Triangle::staticElementVertexFromFace(f,1)]);
          CHECK_TEST(needed_lines.find(p)!=needed_lines.end());
          cl[f]=needed_lines[p];
        }
        present_codimension1_level->setShape(OPERATOR_NEW Triangle(
          present_codimension1_level->presentIndex(),cl,cv));
#ifdef DEBUG
        for (int f=0;f<3;f++) {
          int face_element_index=present_codimension1_level->
            getShape()->codimension1ShapeIndex(f);
          const Shape1 *face_element=present_codimension1_level->
              Level1::getShape(face_element_index);
          present_codimension1_level->getShape()->
            checkFaceVertices(f,face_element);
        }
#endif
      } else {
        int cv[4]={q->first.first.first,q->first.first.second,
                   q->first.second.first,q->first.second.second};
        int cl[4];
        for (int f=0;f<4;f++) {
          pair<int,int> p=MAKE_LINE_PAIR(
            cv[Quadrilateral::staticElementVertexFromFace(f,0)],
            cv[Quadrilateral::staticElementVertexFromFace(f,1)]);
          CHECK_TEST(needed_lines.find(p)!=needed_lines.end());
          cl[f]=needed_lines[p];
        }
        present_codimension1_level->setShape(OPERATOR_NEW 
          Quadrilateral(present_codimension1_level->presentIndex(),
                        cl,cv));
#ifdef DEBUG
        for (int f=0;f<4;f++) {
          int face_element_index=present_codimension1_level->
            getShape()->codimension1ShapeIndex(f);
          const Shape1 *face_element=present_codimension1_level->
              Level1::getShape(face_element_index);
          present_codimension1_level->getShape()->
            checkFaceVertices(f,face_element);
        }
#endif
      }
      present_codimension1_level->setUsedFlag();
      present_codimension1_level->clearUserFlag();
      present_codimension1_level->clearUserPointer();
      q->second=present_codimension1_level->presentIndex();
    }
  }

//construct Shapes and store in Level3
//for each element face, append cell index to list of adjacent cells
  coarsest_level->Level3::reserveSpace(cells.getNumber());  
  map<int,NISLList<int> > adjacent_cells;
  {
    firstShape();
    for (int c=0;c<cells.getNumber();c++,nextShape()) {
      const CellData &cd=cells[c];
      switch (cd.numberVertices()) {
        case 4: {
          int cv[4]=
            {cd.vertex(0),cd.vertex(1),cd.vertex(2),cd.vertex(3)};
          int faces[4];
          for (int f=0;f<4;f++) {
            int tl[3]={-1,-2,-3};
            int tv[3]={
              cv[Tetrahedron::staticElementVertexFromFace(f,0)],
              cv[Tetrahedron::staticElementVertexFromFace(f,1)],
              cv[Tetrahedron::staticElementVertexFromFace(f,2)]};
            Triangle bogus_triangle(-1,tl,tv);
            pair<pair<int,int>,pair<int,int> > 
              p(make_pair(bogus_triangle.vertexIndex(0),
                          bogus_triangle.vertexIndex(1)),
                make_pair(bogus_triangle.vertexIndex(2),-1));
            faces[f]=needed_faces[p];
          }
          int lines[6];
          for (int l=0;l<6;l++) {
            int iv[2]={cv[Tetrahedron::staticElementVertexFromLine(l,0)],
                       cv[Tetrahedron::staticElementVertexFromLine(l,1)]};
            Interval bogus_interval(-1,iv);
            pair<int,int> p(bogus_interval.vertexIndex(0),
                            bogus_interval.vertexIndex(1));
            lines[l]=needed_lines[p];
          }
          present_level->setShape(OPERATOR_NEW Tetrahedron(
            present_level->presentIndex(),lines,faces,cv));
          for (int f=0;f<4;f++) {
            adjacent_cells[faces[f]].
              append(OPERATOR_NEW int(present_level->presentIndex()));
          }
#ifdef DEBUG
          NumPtr<int> vtc(4);
          vtc.initialize(0);
          for (int f=0;f<Tetrahedron::facesPerShape();f++) {
            for (int fv=0;
            fv<Tetrahedron::staticNumberVerticesOnFace(fv);fv++) {
              int vi=present_level->Level2::vertexIndex(faces[f],fv);
              for (int v=0;v<4;v++) {
                if (vi==cv[v]) { vtc[v]++; break; }
              }
            }
          }
          for (int v=0;v<4;v++) CHECK_SAME(vtc[v],3);

          if (present_level->getShape()->volume(vertices)<=0.) {
            present_level->getShape()->printOn(cout);
            CHECK_TEST(present_level->getShape()->volume(vertices)>0.);
          }
          for (int f=0;f<4;f++) {
            int face_element_index=present_level->getShape()->
              codimension1ShapeIndex(f);
            const Shape2 *face_element=present_level->
                Level2::getShape(face_element_index);
            present_level->getShape()->
              checkFaceVertices(f,face_element);
          }
#endif
          break;
        }
        case 6: {
          int cv[6]={cd.vertex(0),cd.vertex(1),cd.vertex(2),
                     cd.vertex(3),cd.vertex(4),cd.vertex(5)};
          int faces[5];
          for (int f=0;f<5;f++) {
            if (Prism::staticNumberVerticesOnFace(f)==3) {
              int tl[3]={-1,-2,-3};
              int tv[3]={cv[Prism::staticElementVertexFromFace(f,0)],
                         cv[Prism::staticElementVertexFromFace(f,1)],
                         cv[Prism::staticElementVertexFromFace(f,2)]};
              Triangle bogus_triangle(-1,tl,tv);
              pair<pair<int,int>,pair<int,int> > 
                p(make_pair(bogus_triangle.vertexIndex(0),
                            bogus_triangle.vertexIndex(1)),
                  make_pair(bogus_triangle.vertexIndex(2),-1));
              faces[f]=needed_faces[p];
            } else {
              int ql[4]={-1,-2,-3,-4};
              int qv[4]={cv[Prism::staticElementVertexFromFace(f,0)],
                         cv[Prism::staticElementVertexFromFace(f,1)],
                         cv[Prism::staticElementVertexFromFace(f,2)],
                         cv[Prism::staticElementVertexFromFace(f,3)]};
              Quadrilateral bogus_quadrilateral(-1,ql,qv);
              pair<pair<int,int>,pair<int,int> > 
                p(make_pair(bogus_quadrilateral.vertexIndex(0),
                            bogus_quadrilateral.vertexIndex(1)),
                  make_pair(bogus_quadrilateral.vertexIndex(2),
                            bogus_quadrilateral.vertexIndex(3)));
              faces[f]=needed_faces[p];
            }
          }
          int lines[9];
          for (int l=0;l<9;l++) {
            int iv[2]={cv[Prism::staticElementVertexFromLine(l,0)],
                       cv[Prism::staticElementVertexFromLine(l,1)]};
            Interval bogus_interval(-1,iv);
            pair<int,int> p(bogus_interval.vertexIndex(0),
                            bogus_interval.vertexIndex(1));
            lines[l]=needed_lines[p];
          }
          present_level->setShape(OPERATOR_NEW Prism(
            present_level->presentIndex(),lines,faces,cv));
          for (int f=0;f<5;f++) {
            adjacent_cells[faces[f]].
              append(OPERATOR_NEW int(present_level->presentIndex()));
          }
#ifdef DEBUG
          NumPtr<int> vtc(6);
          vtc.initialize(0);
          for (int f=0;f<Prism::facesPerShape();f++) {
            for (int fv=0;
            fv<Prism::staticNumberVerticesOnFace(f);fv++) {
              int vi=present_level->Level2::vertexIndex(faces[f],fv);
              for (int v=0;v<6;v++) {
                if (vi==cv[v]) { vtc[v]++; break; }
              }
            }
          }
          for (int v=0;v<6;v++) CHECK_SAME(vtc[v],3);

          if (present_level->getShape()->volume(vertices)<=0.) {
            present_level->getShape()->printOn(cout);
            CHECK_TEST(present_level->getShape()->volume(vertices)>0.);
          }
          for (int f=0;f<5;f++) {
            int face_element_index=present_level->getShape()->
              codimension1ShapeIndex(f);
            const Shape2 *face_element=present_level->
                Level2::getShape(face_element_index);
            present_level->getShape()->
              checkFaceVertices(f,face_element);
          }
#endif
          break;
        }
        case 8: {
          int cv[8]={
            cd.vertex(0),cd.vertex(1),cd.vertex(2),cd.vertex(3),
            cd.vertex(4),cd.vertex(5),cd.vertex(6),cd.vertex(7)};
          int faces[6];
          for (int f=0;f<6;f++) {
            int ql[4]={-1,-2,-3,-4};
            int qv[4]={
              cv[Hexahedron::staticElementVertexFromFace(f,0)],
              cv[Hexahedron::staticElementVertexFromFace(f,1)],
              cv[Hexahedron::staticElementVertexFromFace(f,2)],
              cv[Hexahedron::staticElementVertexFromFace(f,3)]};
            Quadrilateral bogus_quadrilateral(-1,ql,qv);
            pair<pair<int,int>,pair<int,int> > 
              p(make_pair(bogus_quadrilateral.vertexIndex(0),
                          bogus_quadrilateral.vertexIndex(1)),
                make_pair(bogus_quadrilateral.vertexIndex(2),
                          bogus_quadrilateral.vertexIndex(3)));
            faces[f]=needed_faces[p];
          }
          int lines[12];
          for (int l=0;l<12;l++) {
            int iv[2]={cv[Hexahedron::staticElementVertexFromLine(l,0)],
                       cv[Hexahedron::staticElementVertexFromLine(l,1)]};
            Interval bogus_interval(-1,iv);
            pair<int,int> p(bogus_interval.vertexIndex(0),
                            bogus_interval.vertexIndex(1));
            lines[l]=needed_lines[p];
          }
          present_level->setShape(OPERATOR_NEW Hexahedron(
            present_level->presentIndex(),lines,faces,cv));
          for (int f=0;f<6;f++) {
            adjacent_cells[faces[f]].
              append(OPERATOR_NEW int(present_level->presentIndex()));
          }
#ifdef DEBUG
          NumPtr<int> vtc(8);
          vtc.initialize(0);
          for (int f=0;f<Hexahedron::facesPerShape();f++) {
            for (int fv=0;
            fv<Hexahedron::staticNumberVerticesOnFace(f);fv++) {
              int vi=present_level->Level2::vertexIndex(faces[f],fv);
              for (int k=0;k<8;k++) {
                if (vi==cv[k]) { vtc[k]++; break; }
              }
            }
          }
          for (int v=0;v<8;v++) CHECK_SAME(vtc[v],3);

          if (present_level->getShape()->volume(vertices)<=0.) {
            present_level->getShape()->printOn(cout);
            CHECK_TEST(present_level->getShape()->volume(vertices)>0.);
          }
          for (int f=0;f<6;f++) {
            int face_element_index=present_level->getShape()->
              codimension1ShapeIndex(f);
            const Shape2 *face_element=present_level->
                Level2::getShape(face_element_index);
            present_level->getShape()->
              checkFaceVertices(f,face_element);
          }
#endif
          break;
        }
        default:
          OBSOLETE("invalid numberVertices");
      }
      present_level->setUsedFlag();
      present_level->setMaterialModel(cd.getMaterialModel());
      present_level->clearUserFlag();
      present_level->clearUserPointer();
      present_level->setSubdomainId(0);
    }
  }

//for element faces with one adjacent cell (boundary face), set id=0
//  else (interior line) set id=MAX_MATERIAL_ID
//count number of boundary faces

//the following will go out of scope
//it must be replaced without being deleted
  BoundaryCondition bogus_boundary_condition;
  int n_boundary_faces=0;
  for (bool doit=firstCodimension1Shape();doit;
  doit=nextCodimension1Shape()) {
    int n_adj_cells=adjacent_cells[
      present_codimension1_level->presentIndex()].length();
    CHECK_BOUNDS(n_adj_cells,1,3);
    if (n_adj_cells==1) {
      present_codimension1_level->
        setBoundaryCondition(&bogus_boundary_condition);
      n_boundary_faces++;
    } else {
      present_codimension1_level->setBoundaryCondition(0);
    }
  }

  BoundaryData *boundary_data=OPERATOR_NEW BoundaryData();
  boundary_data->boundary_faces.allocate(n_boundary_faces);
  n_boundary_faces=0;
  for (bool doit=firstCodimension1Shape();doit;
  doit=nextCodimension1Shape()) {
    int face_index=present_codimension1_level->presentIndex();
    if (adjacent_cells[face_index].length()==1) {
      FaceData &boundary_face=
        boundary_data->boundary_faces[n_boundary_faces];
      int nv=
        present_codimension1_level->getShape()->numberVertices();
      boundary_face.allocate(nv);
      for (int v=0;v<nv;v++) {
        boundary_face.vertex(v)=
          present_codimension1_level->vertexIndex(v);
      }
      boundary_face.setBoundaryCondition(&bogus_boundary_condition);
      n_boundary_faces++;
    }
  }

//default all lines to have no boundary conditions
  for (bool doit=firstCodimension2Shape();doit;
  doit=nextCodimension2Shape()) {
    present_codimension2_level->setBoundaryCondition(0);
  }

//set boundary conditions for lines in boundary faces
//since face boundary conditions are bogus, so will be the line bc's
  for (bool doit=firstCodimension1Shape();doit;
  doit=nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      present_codimension1_level->getBoundaryCondition();
    if (bc!=0) {
      const Shape2 *element=present_codimension1_level->getShape();
      for (int l=0;l<element->numberCodimension1Shapes();l++) {
        int line_index=element->codimension1ShapeIndex(l);
        present_codimension1_level->
          Level1::setBoundaryCondition(line_index,bc);
      }
    }
  }

  int n_boundary_lines=0;
  for (bool doit=firstCodimension2Shape();doit;
  doit=nextCodimension2Shape()) {
    const BoundaryCondition* bc=
      present_codimension2_level->getBoundaryConditions().first;
    if (bc!=0) n_boundary_lines++;
  }

//set boundary_line vertices
  boundary_data->boundary_lines.allocate(n_boundary_lines);
  n_boundary_lines=0;
  for (bool doit=firstCodimension2Shape();doit;
  doit=nextCodimension2Shape()) {
    int line_index=present_codimension2_level->presentIndex();
    const BoundaryCondition* bc=
      present_codimension2_level->getBoundaryConditions().first;
    if (bc!=0) {
      for (int v=0;v<2;v++) {
        boundary_data->boundary_lines[n_boundary_lines].vertex(v)=
          present_codimension2_level->vertexIndex(v);
      }
      boundary_data->boundary_lines[n_boundary_lines].
        setBoundaryCondition(bc);
      n_boundary_lines++;
    }
  }

//set neighbor of each face of each element
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    const Shape3 *element=present_level->getShape();
    int element_index=present_level->presentIndex();
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      NISLList<int> &adjacent_cell_list=
        adjacent_cells[element->codimension1ShapeIndex(f)];
      NISLListNode<int> *first_adjacent_cell_node=
        adjacent_cell_list.first();
      int adjacent_cell=*first_adjacent_cell_node->getData();
      if (adjacent_cell==element_index) {
        NISLListNode<int> *other_adjacent_cell_node=
          adjacent_cell_list.next(first_adjacent_cell_node);
        if (other_adjacent_cell_node) {
          present_level->setNeighbor(f,present_level,
            *other_adjacent_cell_node->getData());
        } else {
          present_level->setNeighbor(f,0,-1);
        }
      } else {
        present_level->setNeighbor(f,present_level,adjacent_cell);
      }
    }
  }
#ifdef DEBUG
  checkNeighbors();
#endif

//delete lists of adjacent cells
  map<int,NISLList<int> >::iterator adjacent_cell;
  for (adjacent_cell=adjacent_cells.begin();
  adjacent_cell!=adjacent_cells.end();adjacent_cell++) {
    NISLList<int> &adjacent_cell_list=adjacent_cell->second;
    while (adjacent_cell_list.notEmpty()) {
      delete adjacent_cell_list.delAfter(0);
    }
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::synchronize(BoundaryData *boundary_data) {
#if (SPACEDIM==1)
  Level0 *codimension1_level=coarsestLevel();
  for (int f=0;f<2;f++) {
    const BoundaryCondition *bc=
      codimension1_level->getBoundaryCondition(f);
    boundary_data->boundary_faces[f].setBoundaryCondition(bc);
  }
#else
# if (SPACEDIM==3)
  for (bool doit=firstCodimension2Shape();doit;
  doit=nextCodimension2Shape()) { 
    presentCodimension2Level()->setBoundaryCondition(0);
  }
# endif
  int bf=0;
  for (bool doit=firstCodimension1Shape();doit;
  doit=nextCodimension1Shape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      presentCodimension1Level();
    const BoundaryCondition *bc=codimension1_level->getBoundaryCondition();
    if (bc!=0) {
      boundary_data->boundary_faces[bf].setBoundaryCondition(bc);
# if (SPACEDIM==3)
      const CODIMENSION1_SHAPE_CLASS *face=presentCodimension1Shape();
      for (int f=0;f<face->numberCodimension1Shapes();f++) {
        int li=face->codimension1ShapeIndex(f);
        codimension1_level->Level1::setBoundaryCondition(li,bc);
      }
# endif
      bf++;
    } 
  }
# if (SPACEDIM==3)
  int bl=0;
  for (bool doit=firstCodimension2Shape();doit;
  doit=nextCodimension2Shape()) { 
    Level1 *codimension2_level=presentCodimension2Level();
    boundary_data->boundary_lines[bl].setBoundaryConditions(
      codimension2_level->getBoundaryConditions());
  }
# endif
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::setMaterialModel(MaterialModel *mm) {
  for (bool doit=firstActiveShape(coarsest_level);doit;
  doit=nextActiveShapeInLevel()) {
    present_level->setMaterialModel(mm);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Tessellation::numberActiveShapes(const LEVEL_CLASS *level) {
  int n=0;
  for (bool doit=firstActiveShape(level);doit;
  doit=nextActiveShapeInLevel()) {
    n++;
  }
  return n;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Tessellation::numberActiveShapes() {
  int n=0;
  for (bool doit=firstActiveShape();doit;doit=nextActiveShape()) {
    n++;
  }
  return n;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
int Tessellation::numberActiveCodimension1Shapes(
const CODIMENSION1_LEVEL_CLASS *level) {
  int n=0;
  for (bool doit=firstCodimension1ActiveShape(level);doit;
  doit=nextCodimension1ActiveShapeInLevel()) {
    n++;
  }
  return n;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
int Tessellation::numberActiveCodimension1Shapes() {
  int n=0;
  for (bool doit=firstCodimension1ActiveShape();doit;
  doit=nextCodimension1ActiveShape()) {
    n++;
  }
  return n;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::saveCoarsenFlags(NumPtr<bool> &v) {
  v.allocate(numberActiveShapes());
  v.initialize(false);
  int i=0;
  for (bool doit=firstActiveShape();doit;doit=nextActiveShape(),i++)
  {
    v[i]=presentLevel()->coarsenFlagSet();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::saveRefineFlags(NumPtr<bool> &v) {
  v.allocate(numberActiveShapes());
  v.initialize(false);
  int i=0;
  for (bool doit=firstActiveShape();doit;doit=nextActiveShape(),i++)
  {
    v[i]=presentLevel()->refineFlagSet();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::clearUserFlags() {
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    present_level->clearUserFlag();
  }
#if (SPACEDIM>=2)
  for (bool doit=firstCodimension1UsedShape();doit;
  doit=nextCodimension1UsedShape()) {
    present_codimension1_level->clearUserFlag();
  }
#endif
#if (SPACEDIM==3)
  for (bool doit=firstCodimension2UsedShape();doit;
  doit=nextCodimension2UsedShape()) {
    present_codimension2_level->clearUserFlag();
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::fixCoarsenFlags() {
  clearUserFlags();
  int cpc=childrenPerShape();
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    if (!present_level->hasChildren()) { // active
      if (present_level==coarsest_level) {
        present_level->clearCoarsenFlag();
      }
    } else {
      int flagged_children=0;
      LEVEL_CLASS *present_finer_level=present_level->finerLevel();
      for (int c=0; c<cpc; ++c) {
        int child_index=present_level->childIndex(c);
        if (!present_finer_level->hasChildren(child_index) &&
        present_finer_level->coarsenFlagSet(child_index)) {
          flagged_children++;
          present_finer_level->clearCoarsenFlag(child_index);
        }
      }
      if (flagged_children==cpc) present_level->setUserFlag();
    }
  }
#if DEBUG
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    CHECK_TEST(!present_level->coarsenFlagSet());
  }
#endif
  for (bool doit=lastUsedShape();doit;doit=prevUsedShape()) {
    if (present_level->userFlagSet()) {
      present_level->clearUserFlag();
      bool coarsening_allowed=true;
      int nf=present_level->getShape()->numberCodimension1Shapes();
      LEVEL_CLASS *present_finer_level=present_level->finerLevel();
      for (int c=0;c<cpc;c++) {
        int child_index=present_level->childIndex(c);
        for (int f=0;f<nf;f++) {
          int child_neighbor_index=
            present_finer_level->neighborIndex(child_index,f);
          if (child_neighbor_index>=0) {
            const LEVEL_CLASS *child_neighbor_level=
              present_finer_level->neighborLevel(child_index,f);
            if (child_neighbor_level!=present_finer_level) break;
            if (child_neighbor_level->hasChildren(child_neighbor_index) 
            || child_neighbor_level->refineFlagSet(child_neighbor_index))
            {
              coarsening_allowed=false;
            }
          }
        }
      }
      if (coarsening_allowed) {
        for (int c=0;c<cpc;c++) {
          int child_index=present_level->childIndex(c);
          CHECK_TEST(!present_finer_level->refineFlagSet(child_index));
          present_finer_level->setCoarsenFlag(child_index);
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Tessellation::prepareRefinement() {
  for (bool doit=firstCodimension2UsedShape();doit;
  doit=nextCodimension2UsedShape()) {
    present_codimension2_level->clearUserFlag();
  }
  bool mesh_changed=false;
  do {
    mesh_changed=false;
    for (bool doit=firstActiveShape();doit;doit=nextActiveShape()) {
      if (present_level->refineFlagSet()) {
        const Shape3 *element=present_level->getShape();
        int nl=element->numberCodimension2Shapes();
        Level1 *present_level1=static_cast<Level1*>(present_level);
        for (int l=0;l<nl;l++) {
          int line_index=element->codimension2ShapeIndex(l);
          if (!present_level1->hasChildren(line_index)) {
            present_level1->setUserFlag(line_index);
          }
        }
      }
    }
    for (bool doit=lastActiveShape();doit;doit=prevActiveShape()) {
      const Shape3 *element=present_level->getShape();
      Level1 *present_level1=static_cast<Level1*>(present_level);
      int nl=element->numberCodimension2Shapes();
      for (int l=0;l<nl;l++) {
        int line_index=element->codimension2ShapeIndex(l);
        if (present_level1->hasChildren(line_index)) {
          bool offending_line_found=false;
          Level1 *present_finer_level=present_level1->finerLevel();
          for (int c=0;c<2;c++) {
            int line_child_index=
              present_level1->childIndex(line_index,c);
//          if line is refined, its children must not have children
            CHECK_TEST(!present_finer_level->
                        hasChildren(line_child_index));
//          if line child will be refined, then this cell must be refined
            if (present_finer_level->userFlagSet(line_child_index) &&
            !present_level->refineFlagSet()) {
              present_level->clearCoarsenFlag();
              present_level->setRefineFlag();
              offending_line_found=true;
              for (int ll=0;ll<nl;ll++) {
                int second_line_index=
                  element->codimension2ShapeIndex(ll);
                if (!present_level1->hasChildren(second_line_index)) {
                  present_level->setUserFlag(second_line_index);
                }
              }
              break;
            }
          }
          if (offending_line_found) {
            mesh_changed=true;
            break;
          }
        }
        if (present_level1->hasChildren(line_index) ||
        present_level1->userFlagSet(line_index)) {
          if (present_level->coarsenFlagSet()) {
            present_level->clearCoarsenFlag();
            mesh_changed=true;
            break;
          }
        }
      }
    }
  } while (mesh_changed);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool Tessellation::prepareCoarseningAndRefinement() {
  NumPtr<bool> coarsen_flags_before,refine_flags_before;
  saveCoarsenFlags(coarsen_flags_before);
  saveRefineFlags(refine_flags_before);
#if (SPACEDIM==1)
  fixCoarsenFlags();
  NumPtr<bool> coarsen_flags_after,refine_flags_after;
  saveCoarsenFlags(coarsen_flags_after);
  saveRefineFlags(refine_flags_after);
  for (int i=0;i<coarsen_flags_before.getNumber();i++) {
    if (coarsen_flags_after[i]!=coarsen_flags_before[i]) return true;
  }
  for (int i=0;i<refine_flags_before.getNumber();i++) {
    if (refine_flags_after[i]!=refine_flags_before[i]) return true;
  }
  return false;
#else
  int cpc=childrenPerShape();
  NumPtr<bool> 
    coarsen_flags_before_loop(coarsen_flags_before.getNumber()),
    refine_flags_before_loop(refine_flags_before.getNumber());
  coarsen_flags_before_loop.copyFrom(coarsen_flags_before);
  refine_flags_before_loop.copyFrom(refine_flags_before);
  bool mesh_changed_in_this_loop=false;
  do {
    if (mesh_smoothing & mesh_smoothing_do_not_produce_unrefined_islands
    && !(mesh_smoothing & mesh_smoothing_patch_level_1)) {
      for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
        if (present_level->hasChildren()) {
          LEVEL_CLASS *present_finer_level=present_level->finerLevel();
          int n_childrens_coarsen_flags_set=0;
          for (int c=0;c<cpc;c++) {
            int child_index=present_level->childIndex(c);
            if (!present_finer_level->hasChildren(child_index) &&
            present_finer_level->coarsenFlagSet(child_index)) {
              ++n_childrens_coarsen_flags_set;
            }
          }
          if (n_childrens_coarsen_flags_set==cpc) {
            int n_neighbors=0;
            int count=0;
            int nf=
              present_level->getShape()->numberCodimension1Shapes();
            for (int f=0;f<nf;f++) {
              int neighbor_index=present_level->neighborIndex(f);
              if (neighbor_index>=0) {
                n_neighbors++;
                bool not_active_neighbor_will_be_coarsened=false;
                int n_neighbors_childrens_coarsen_flags_set=0;
                if (present_level->hasChildren(neighbor_index)) {
                  for (int c=0;c<cpc;c++) {
                    int neighbor_child_index=
                      present_level->childIndex(neighbor_index,c);
                    if (!present_finer_level->hasChildren(
                    neighbor_child_index) && 
                    present_finer_level->coarsenFlagSet(
                    neighbor_child_index)) {
                      n_neighbors_childrens_coarsen_flags_set++;
                    }
                  }
                  if (n_neighbors_childrens_coarsen_flags_set ==cpc){
                    not_active_neighbor_will_be_coarsened=true;
                  }
                }
                if ((!present_level->hasChildren(neighbor_index) &&
                present_level->refineFlagSet(neighbor_index)) ||
                !not_active_neighbor_will_be_coarsened) {
                  ++count;
                }
              }
            }
            if (count==n_neighbors || 
            (count>=n_neighbors-1 && n_neighbors==nf)) 
            {
              for (int c=0;c<cpc;c++) {
                present_finer_level->clearCoarsenFlag(
                  present_level->childIndex(c));
              }
            }
          }
        }
      }
    }
    if (mesh_smoothing & 
    (mesh_smoothing_eliminate_refined_inner_islands |
    mesh_smoothing_eliminate_refined_boundary_islands) &&
    !(mesh_smoothing & mesh_smoothing_patch_level_1)) {
      for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
        if (present_level->hasChildren() ||
        (!present_level->hasChildren()&&present_level->refineFlagSet()))
        {
          bool all_children_active=true;
          if (present_level->hasChildren()) {
            for (int c=0;c<cpc;c++) {
              int child_index=present_level->childIndex(c);
              LEVEL_CLASS *present_finer_level=
                present_level->finerLevel();
              if (present_finer_level->hasChildren(child_index)) {
                all_children_active=false;
                break;
              }
            }
          }
          if (all_children_active) {
            int unrefined_neighbors=0,total_neighbors=0;
            int fpc=
              present_level->getShape()->numberCodimension1Shapes();
            for (int f=0;f<fpc;f++) {
              int neighbor_index=present_level->neighborIndex(f);
              LEVEL_CLASS *neighbor_level=
                present_level->neighborLevel(f);
              if (neighbor_index>=0) {
                total_neighbors++;
                if ((!neighbor_level->hasChildren(neighbor_index) &&
                !neighbor_level->refineFlagSet(neighbor_index)) ||
                neighbor_level==present_level->coarserLevel()) {
                  unrefined_neighbors++;
                } else {
                  if (neighbor_level->hasChildren(neighbor_index)) {
                    int tagged_children=0;
                    for (int c=0;c<cpc;c++) {
                      int neighbor_child_index=
                        neighbor_level->childIndex(neighbor_index,c);
                      if (neighbor_level->finerLevel()->
                      coarsenFlagSet(neighbor_child_index)) {
                        tagged_children++;
                      }
                    }
                    if (tagged_children==cpc) unrefined_neighbors++;
                  }
                }
              }
            }
            if ((unrefined_neighbors==total_neighbors) &&
            (((unrefined_neighbors==fpc)
            && (mesh_smoothing & 
                mesh_smoothing_eliminate_refined_inner_islands)) ||
            ((unrefined_neighbors<fpc) &&
             (mesh_smoothing & 
              mesh_smoothing_eliminate_refined_boundary_islands)) )
            && (total_neighbors!=0)) {
              if (present_level->hasChildren()) {
                LEVEL_CLASS *present_finer_level=
                  present_level->finerLevel();
                for (int c=0;c<cpc;c++) {
                  int child_index=present_level->childIndex(c);
                  present_finer_level->clearRefineFlag(child_index);
                  present_finer_level->setCoarsenFlag(child_index);
                }
              } else {
                present_level->clearRefineFlag();
              }
            }
          }
        }
      }
    }
    if (mesh_smoothing & 
    mesh_smoothing_limit_level_difference_at_vertices) {
      NumPtr<LEVEL_CLASS*> vertex_level(vertices.getNumber());
      vertex_level.initialize(coarsest_level);
      for (bool doit=firstActiveShape();doit;doit=nextActiveShape())
      {
        const SHAPE_CLASS *element=present_level->getShape();
        int nv=element->numberVertices();
        for (int v=0;v<nv;v++) {
          int element_vertex_index=element->vertexIndex(v);
          LEVEL_CLASS *element_vertex_level=
            vertex_level[element_vertex_index];
          if (present_level->refineFlagSet()) {
            LEVEL_CLASS *present_finer_level=
              present_level->finerLevel();
            vertex_level[present_level->presentIndex()]=
              (present_finer_level->isFinerThan(element_vertex_level) ?
              present_finer_level : element_vertex_level);
          } else {
            vertex_level[element_vertex_index]=
              (present_level->isFinerThan(element_vertex_level) ?
               present_level : element_vertex_level);
          }
        }
      }
      for (bool doit=lastActiveShape();doit;doit=prevActiveShape()){
        if (!present_level->refineFlagSet()) {
          const SHAPE_CLASS *element=present_level->getShape();
          int nv=element->numberVertices();
          LEVEL_CLASS *present_finer_level=present_level->finerLevel();
          for (int v=0;v<nv;v++) {
            if (vertex_level[element->vertexIndex(v)]->
            isFinerThan(present_finer_level)) {
              present_level->clearCoarsenFlag();
              present_level->setRefineFlag();
              for (int vv=0;vv<nv;++vv) {
                LEVEL_CLASS *element_vertex_level=
                  vertex_level[element->vertexIndex(vv)];
                vertex_level[element->vertexIndex(vv)]=
                  (element_vertex_level->
                  isFinerThan(present_finer_level) ? 
                  element_vertex_level : present_finer_level);
              }
              break;
            }
          }
        }
      }
    }
    if (mesh_smoothing & mesh_smoothing_eliminate_unrefined_islands) {
      for (bool doit=firstActiveShape();doit;doit=nextActiveShape())
      {
        if (present_level->refineFlagSet()) continue;
        int refined_neighbors=0,unrefined_neighbors=0;
        const SHAPE_CLASS *element=present_level->getShape();
        int nf=element->numberCodimension1Shapes();
        for (int f=0;f<nf;f++) {
          int neighbor_index=present_level->neighborIndex(f);
          if (neighbor_index!=-1) {
            LEVEL_CLASS *neighbor_level=present_level->neighborLevel(f);
            CHECK_TEST(neighbor_level==present_level ||
              neighbor_level==present_level->coarserLevel());
            if (neighbor_level==present_level && 
            (neighbor_level->refineFlagSet(neighbor_index) ||
            neighbor_level->hasChildren(neighbor_index))) {
              refined_neighbors++;
            } else unrefined_neighbors++;
          }
        }
        if (unrefined_neighbors<refined_neighbors) {
          if (present_level->coarsenFlagSet()) {
            present_level->clearCoarsenFlag();
          }
          present_level->setRefineFlag();
        }
      }
    }
    if (mesh_smoothing & mesh_smoothing_patch_level_1) {
      for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
        if (present_level->hasChildren()) {
          int child_index=present_level->childIndex(0);
          LEVEL_CLASS *present_finer_level=present_level->finerLevel();
          if (!present_finer_level->hasChildren(child_index)) {
            bool any_refine_flag_set=false;
            for (int c=0;c<cpc;c++) {
              child_index=present_level->childIndex(c);
              CHECK_TEST(!present_finer_level->
                hasChildren(child_index));
              if (present_finer_level->refineFlagSet(child_index)) {
                any_refine_flag_set=true;
                break;
              }
            }
            if (any_refine_flag_set) {
              for (int c=0;c<cpc;c++) {
                child_index=present_level->childIndex(c);
                present_finer_level->clearCoarsenFlag(child_index);
                present_finer_level->setRefineFlag(child_index);
              }
            }
          }
        }
      }
      for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
        int present_child_index=present_level->childIndex(0);
        if (present_level->hasChildren()) {
          LEVEL_CLASS *present_finer_level=present_level->finerLevel();
          if (present_finer_level->hasChildren(present_child_index)) {
            int present_grand_child_index=
              present_finer_level->childIndex(0);
            LEVEL_CLASS *present_twice_finer_level=
              present_finer_level->finerLevel();
            if (!present_twice_finer_level->
            hasChildren(present_grand_child_index)) {
              int n_coarsen_flags=0;
              for (int c=0;c<cpc;c++) {
                present_child_index=present_level->childIndex(c);
                CHECK_TEST(present_finer_level->
                  hasChildren(present_child_index));
                present_grand_child_index=
                  present_finer_level->childIndex(0);
                if (!present_twice_finer_level->
                hasChildren(present_grand_child_index)) {
                  for (int cc=0;cc<cpc;cc++) {
                    present_grand_child_index=
                      present_finer_level->childIndex(cc);
                    if (present_twice_finer_level->
                    coarsenFlagSet(present_grand_child_index)) {
                      n_coarsen_flags++;
                    }
                  }
                }
              }
              if (n_coarsen_flags!=cpc*cpc) {
                for (int c=0;c<cpc;c++) {
                  present_child_index=present_level->childIndex(c);
                  CHECK_TEST(present_finer_level->
                    hasChildren(present_child_index));
                  present_grand_child_index=
                    present_finer_level->childIndex(0);
                  if (!present_twice_finer_level->
                  hasChildren(present_grand_child_index)) {
                    for (int cc=0;cc<cpc;cc++) {
                      present_grand_child_index=
                        present_finer_level->childIndex(cc);
                      CHECK_TEST(!present_twice_finer_level->
                        hasChildren(present_grand_child_index));
                      present_twice_finer_level->
                        clearCoarsenFlag(present_grand_child_index);
                    }
                  }
                }
              }
            } 
          }
        }
      }
    }
//  if refine current element, 
//    if neighbor is coarser: don't coarsen neighbor, refine it 
//    if neighbor is same level: don't coarsen it
    for (bool doit=lastActiveShape();doit;doit=prevActiveShape()) {
      if (present_level->refineFlagSet()) {
        int nf=present_level->getShape()->numberCodimension1Shapes();
        for (int f=0;f<nf;f++) {
          int neighbor_index=present_level->neighborIndex(f);
          if (neighbor_index>=0) {
            LEVEL_CLASS *neighbor_level=present_level->neighborLevel(f);
            if (neighbor_level==present_level) {
              neighbor_level->clearCoarsenFlag(neighbor_index);
            } else if (neighbor_level==present_level->coarserLevel()) {
              neighbor_level->clearCoarsenFlag(neighbor_index);
              neighbor_level->setRefineFlag(neighbor_index);
            }
          }
        }
      }
    }
#if (SPACEDIM==3)
    prepareRefinement();
#endif
//  if element has no children,
//    if element is on coarsest level, clear coarsen flag
//  if element has children,
//    if all children are marked for coarsening
//      if child neighbor is not same level as child (ie coarser) or 
//      (has children and is not marked for refinement) then coarsen 
    fixCoarsenFlags();

#if (SPACEDIM>=2)
//cf DoFHandler::maxCouplingsBetweenDofs
    int nuv=numberUsedVertices();
    NumPtr<int> min_level(nuv); min_level.initialize(INT_MAX);
    NumPtr<int> max_level(nuv); max_level.initialize(0);
    for (bool doit=firstActiveShape();doit;doit=nextActiveShape()) {
      const SHAPE_CLASS *element=present_level->getShape();
      int present_level_number=present_level->numberCoarserLevels();
      int element_level=
        (present_level->coarsenFlagSet() ? present_level_number-1 
          : (present_level->refineFlagSet() ? present_level_number+1 
                                            : present_level_number));
      for (int v=0;v<element->numberVertices();v++) {
        int iv=element->vertexIndex(v);
        min_level[iv]=min(min_level[iv],element_level);
        max_level[iv]=max(max_level[iv],element_level);
      }
    }
    for (bool doit=firstActiveShape();doit;doit=nextActiveShape()) {
      const SHAPE_CLASS *element=present_level->getShape();
      bool needs_refinement=false;
      for (int v=0;v<element->numberVertices();v++) {
        int iv=element->vertexIndex(v);
        if (max_level[iv]>min_level[iv]+1) {
          needs_refinement=true;
          break;
        }
      }
      if (needs_refinement) present_level->setRefineFlag();
    }
#endif
    NumPtr<bool> coarsen_flags_after_loop,refine_flags_after_loop;
    saveCoarsenFlags(coarsen_flags_after_loop);
    saveRefineFlags(refine_flags_after_loop);
    mesh_changed_in_this_loop=false;
    for (int i=0;i<coarsen_flags_before_loop.getNumber();i++) {
      if (coarsen_flags_after_loop[i]!=coarsen_flags_before_loop[i]) {
        mesh_changed_in_this_loop=true;
        break;
      }
    }
    if (!mesh_changed_in_this_loop) {
      for (int i=0;i<refine_flags_before_loop.getNumber();i++){
        if (refine_flags_after_loop[i]!=refine_flags_before_loop[i]) {
          mesh_changed_in_this_loop=true;
          break;
        }
      }
    }
    coarsen_flags_before_loop.copy(coarsen_flags_after_loop);
    refine_flags_before_loop.copy(refine_flags_after_loop);
  } while (mesh_changed_in_this_loop);

  for (int i=0;i<coarsen_flags_before.getNumber();i++) {
    if (coarsen_flags_before_loop[i]!=coarsen_flags_before[i]) {
      return true;
    }
  }
  for (int i=0;i<refine_flags_before.getNumber();i++) {
    if (refine_flags_before_loop[i]!=refine_flags_before[i]) {
      return true;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void Tessellation::deleteChildren() {
  LEVEL_CLASS *present_finer_level=present_level->finerLevel();
  const SHAPE_CLASS *element=present_level->getShape();
  int cpc=childrenPerShape();
#ifdef DEBUG
  for (int c=0;c<cpc;c++) {
    CHECK_TEST(!present_finer_level->
      hasChildren(present_level->childIndex(c)));
  }
#endif
  int nf=element->numberCodimension1Shapes();
  int present_index=present_level->presentIndex();
  for (int c=0;c<cpc;c++) {
    int child_index=present_level->childIndex(c);
    for (int f=0;f<nf;f++) {
      LEVEL_CLASS *neighbor_level=
        present_finer_level->neighborLevel(child_index,f);
      int neighbor_index=
        present_finer_level->neighborIndex(child_index,f);
      if (neighbor_index<0) continue;
      CHECK_TEST((neighbor_level==present_level) ||
                 (neighbor_level==present_finer_level));
      if (neighbor_level==present_finer_level) {
        const SHAPE_CLASS *neighbor_element=
          neighbor_level->getShape(neighbor_index);
        int nnf=neighbor_element->numberCodimension1Shapes();
        for (int ff=0;ff<nnf;ff++) {
          int neighbor_neighbor_index=
            neighbor_level->neighborIndex(neighbor_index,ff);
          const LEVEL_CLASS *neighbor_neighbor_level=
            neighbor_level->neighborLevel(neighbor_index,ff);
          if ((neighbor_neighbor_level==present_finer_level) &&
          (neighbor_neighbor_index==child_index)) {
            neighbor_level->
              setNeighbor(neighbor_index,ff,present_level,present_index);
          }
        }
      }
    }
  }
  int present_child_index=present_level->childIndex(0);
  vertices_used[present_finer_level->vertexIndex(present_child_index,1)]
    =false;
  for (int c=0;c<cpc;c++) {
    present_child_index=present_level->childIndex(c);
    present_finer_level->clearUserPointer(present_child_index);
    present_finer_level->clearUserFlag(present_child_index);
    present_finer_level->clearUsedFlag(present_child_index);
    delete present_finer_level->getShape(present_child_index);
    present_finer_level->setShape(present_child_index,0);
    present_finer_level->setParent(present_child_index,-1);
  };
  present_level->setChildren(-1);
  present_level->clearUserFlag();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void Tessellation::deleteChildren() {
  const SHAPE_CLASS *element=present_level->getShape();
  int present_index=present_level->presentIndex();
  int nf=element->numberCodimension1Shapes();
  int cpc=childrenPerShape();
  LEVEL_CLASS *present_finer_level=present_level->finerLevel();
  for (int c=0;c<cpc;c++) {
    int child_index=present_level->childIndex(c);
    for (int f=0;f<nf;f++) {
      LEVEL_CLASS *neighbor_level=
        present_finer_level->neighborLevel(child_index,f);
      int neighbor_index=
        present_finer_level->neighborIndex(child_index,f);
      if (neighbor_index<0) continue;
      CHECK_TEST((neighbor_level==present_level) ||
                 (neighbor_level==present_finer_level));
      if (neighbor_level==present_finer_level) {
        const SHAPE_CLASS *neighbor_element=
          neighbor_level->getShape(neighbor_index);
        for (int ff=0;
        ff<neighbor_element->numberCodimension1Shapes();ff++) {
          int neighbor_neighbor_index=
            neighbor_level->neighborIndex(neighbor_index,ff);
          LEVEL_CLASS *neighbor_neighbor_level=
            neighbor_level->neighborLevel(neighbor_index,ff);
          if ((neighbor_neighbor_level==present_finer_level) &&
          (neighbor_neighbor_index==child_index)) {
            neighbor_level->setNeighbor(neighbor_index,ff,present_level,
              present_index);
          }
        }
      }
    }
  }
  for (int i=0;
  i<element->numberInteriorChildVertices();i++) {
    int c(INT_MAX),v(INT_MAX);
    present_level->interiorChildVertex(present_index,c,v);
    int child_index=present_level->childIndex(c);
    vertices_used[present_finer_level->vertexIndex(child_index,v)]=
      false;
  }
  for (int i=0;
  i<element->numberInteriorChildCodimension1Shapes();i++) {
    int c0,c1;
    element->interiorChildCodimension1ShapeNeighbors(i,c0,c1);
    int child_index0=present_level->childIndex(c0);
    const SHAPE_CLASS *child_element0=
      present_finer_level->getShape(child_index0);
    int child_index1=present_level->childIndex(c1);
    const SHAPE_CLASS *child_element1=
      present_finer_level->getShape(child_index1);
    int shared_face_index=
      child_element0->sharedFaceElementIndex(child_element1);
    for (int f=0;
    f<child_element0->numberCodimension1Shapes();f++) {
      if (child_element0->codimension1ShapeIndex(f)==shared_face_index){
        int child_face_index=
          child_element0->codimension1ShapeIndex(f);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUserPointer(child_face_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUserFlag(child_face_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUsedFlag(child_face_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::setParent(child_face_index,-1);
        break;
      }
    }
  }

  for (int f=0; f<nf; ++f) { 
    LEVEL_CLASS *neighbor_level=present_level->neighborLevel(f);
    int neighbor_index=present_level->neighborIndex(f);
    if (neighbor_index<0 || 
    neighbor_level==present_level->coarserLevel() ||
    (neighbor_level==present_level && 
    !neighbor_level->hasChildren(neighbor_index))) {
      int face_index=element->codimension1ShapeIndex(f);
      int c(INT_MAX),v(INT_MAX);
      int face_interior_vertex_index=present_level->
        CODIMENSION1_LEVEL_CLASS::interiorChildVertex(face_index,c,v);
      vertices_used[face_interior_vertex_index]=false;
      for (int subface=0;
      subface<childrenPerCodimension1Shape();subface++) {
        int present_face_child_index=present_level->
          CODIMENSION1_LEVEL_CLASS::childIndex(face_index,subface);
        present_finer_level->CODIMENSION1_LEVEL_CLASS::clearUserPointer(
          present_face_child_index);
        present_finer_level->CODIMENSION1_LEVEL_CLASS::clearUserFlag(
          present_face_child_index);
        present_finer_level->CODIMENSION1_LEVEL_CLASS::clearUsedFlag(
          present_face_child_index);
        present_finer_level->CODIMENSION1_LEVEL_CLASS::setParent(
          present_face_child_index,-1);
      }
      present_level->CODIMENSION1_LEVEL_CLASS::setChildren(face_index,-1);
    }
  }
  for (int c=0;c<cpc;++c) {
    int child_index=present_level->childIndex(c);
    present_finer_level->clearUserPointer(child_index);
    present_finer_level->clearUserFlag(child_index);
    present_finer_level->clearUsedFlag(child_index);
    present_finer_level->setParent(child_index,-1);
  };
  present_level->setChildren(-1);
  present_level->clearUserFlag();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Tessellation::deleteChildren() {
  int cpc=childrenPerShape();
  const SHAPE_CLASS *element=present_level->getShape();
  int present_index=present_level->presentIndex();
  int nf=element->numberCodimension1Shapes();
  LEVEL_CLASS *present_finer_level=present_level->finerLevel();
  for (int c=0;c<cpc;c++) {
    int child_index=present_level->childIndex(c);
    for (int f=0;f<nf;f++) {
      LEVEL_CLASS *neighbor_level=
        present_finer_level->neighborLevel(child_index,f);
      int neighbor_index=
        present_finer_level->neighborIndex(child_index,f);
      if (neighbor_index<0) continue;
      CHECK_TEST((neighbor_level==present_level) ||
                 (neighbor_level==present_finer_level));
      if (neighbor_level==present_finer_level) {
        const SHAPE_CLASS *neighbor_element=
          neighbor_level->getShape(neighbor_index);
        int nnf=neighbor_element->numberCodimension1Shapes();
        for (int ff=0;ff<nnf;ff++) {
          int neighbor_neighbor_index=
            neighbor_level->neighborIndex(neighbor_index,ff);
          LEVEL_CLASS *neighbor_neighbor_level=
            neighbor_level->neighborLevel(neighbor_index,ff);
          if ((neighbor_neighbor_level==present_finer_level) &&
          (neighbor_neighbor_index==child_index)) {
            neighbor_level->setNeighbor(neighbor_index,ff,present_level,
              present_index);
          }
        }
      }
    }
  }
  for (int i=0;i<element->numberInteriorChildVertices();i++) {
    int c(INT_MAX),v(INT_MAX);
    int icv=present_level->interiorChildVertex(present_index,c,v);
    vertices_used[icv]=false;
  }
  for (int i=0;
  i<element->numberInteriorChildCodimension1Shapes();i++) {
    int c0,c1;
    element->interiorChildCodimension1ShapeNeighbors(i,c0,c1);
    int child_index0=present_level->childIndex(c0);
    const SHAPE_CLASS *child_element0=
      present_finer_level->getShape(child_index0);
    int child_index1=present_level->childIndex(c1);
    const SHAPE_CLASS *child_element1=
      present_finer_level->getShape(child_index1);
    int shared_face_index=
      child_element0->sharedFaceElementIndex(child_element1);
    for (int f=0;
    f<child_element0->numberCodimension1Shapes();f++) {
      if (child_element0->codimension1ShapeIndex(f)==shared_face_index){
        int child_face_index=
          child_element0->codimension1ShapeIndex(f);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUserPointer(child_face_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUserFlag(child_face_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUsedFlag(child_face_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::setParent(child_face_index,-1);
        break;
      }
    }
  }
  for (int i=0;
  i<element->numberInteriorChildCodimension2Shapes();i++) {
    int c0,c1;
    element->interiorChildCodimension2ShapeNeighbors(i,c0,c1);
    int child_index0=present_level->childIndex(c0);
    const SHAPE_CLASS *child_element0=
      present_finer_level->getShape(child_index0);
    int child_index1=present_level->childIndex(c1);
    const SHAPE_CLASS *child_element1=
      present_finer_level->getShape(child_index1);
    int child_line_index=
      child_element0->sharedLineElementIndex(child_element1);
    present_finer_level->Level1::clearUserPointer(child_line_index);
    present_finer_level->Level1::clearUserFlag(child_line_index);
    present_finer_level->Level1::clearUsedFlag(child_line_index);
    present_finer_level->Level1::setParent(child_line_index,-1);
  }
  for (int f=0;f<nf;f++) {
    CODIMENSION1_LEVEL_CLASS *neighbor_level=
      present_level->neighborLevel(f);
    int neighbor_index=present_level->neighborIndex(f);
    if (neighbor_index<0 ||
    neighbor_level==present_level->coarserLevel() ||
    (neighbor_level==present_level &&
    !neighbor_level->hasChildren(neighbor_index))) {
      int face_index=element->codimension1ShapeIndex(f);
      const Shape2 *face_element=
        present_level->CODIMENSION1_LEVEL_CLASS::getShape(face_index);
      for (int i=0;
      i<face_element->numberInteriorChildVertices();i++) {
        int c,v;
        face_element->interiorChildVertex(c,v);
        int face_child_index=
          present_level->CODIMENSION1_LEVEL_CLASS::childIndex(c);
        int face_interior_vertex_index=present_finer_level->
          CODIMENSION1_LEVEL_CLASS::vertexIndex(face_child_index,v);
        vertices_used[face_interior_vertex_index]=false;
      }
      for (int i=0;
      i<face_element->numberInteriorChildCodimension1Shapes();i++) {
        int c0,c1;
        face_element->interiorChildCodimension1ShapeNeighbors(i,c0,c1);
        int face_child_index0=
          present_level->CODIMENSION1_LEVEL_CLASS::childIndex(c0);
        const Shape2 *face_child_element0=present_finer_level->
          CODIMENSION1_LEVEL_CLASS::getShape(face_child_index0);
        int face_child_index1=
          present_level->CODIMENSION1_LEVEL_CLASS::childIndex(c1);
        const Shape2 *face_child_element1=present_finer_level->
          CODIMENSION1_LEVEL_CLASS::getShape(face_child_index1);
        int face_interior_line_index=face_child_element0->
          sharedFaceElementIndex(face_child_element1);
        present_finer_level->
          Level1::clearUserPointer(face_interior_line_index);
        present_finer_level->
          Level1::clearUserFlag(face_interior_line_index);
        present_finer_level->
          Level1::clearUsedFlag(face_interior_line_index);
        present_finer_level->
          Level1::setParent(face_interior_line_index,-1);
      }
      for (int ff=0;ff<childrenPerCodimension1Shape();ff++) {
        int face_child_index=present_level->
          CODIMENSION1_LEVEL_CLASS::childIndex(face_index,ff);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUserPointer(face_child_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUserFlag(face_child_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::clearUsedFlag(face_child_index);
        present_finer_level->
          CODIMENSION1_LEVEL_CLASS::setParent(face_child_index,-1);
      }
      present_level->CODIMENSION1_LEVEL_CLASS::setChildren(face_index,-1);
    }
  }
  for (int c=0;c<cpc;c++) {
    int child_index=present_level->childIndex(c);
    present_finer_level->clearUserPointer(child_index);
    present_finer_level->clearUserFlag(child_index);
    present_finer_level->clearUsedFlag(child_index);
    present_finer_level->setParent(child_index,-1);
  }
  present_level->setChildren(-1);
  present_level->clearUserFlag();
  NumPtrWithValue<bool> 
    line_is_needed(element->numberCodimension2Shapes());
  for (int i=0;i<line_is_needed.getNumber();i++) {
    int line_index=element->codimension2ShapeIndex(i);
    line_is_needed.setValue(i,false,line_index);
#ifdef DEBUG
    for (int c=0;c<2;c++) {
      int line_child_index=
        present_level->Level1::childIndex(line_index,c);
      CHECK_TEST(!present_finer_level->
        Level1::hasChildren(line_child_index));
    }
#endif
  }
  line_is_needed.heapSort();
  for (int f=0;f<element->numberCodimension1Shapes();f++) {
    LEVEL_CLASS *neighbor_level=present_level->neighborLevel(f);
    int neighbor_index=present_level->neighborIndex(f);
    if (neighbor_index<0) continue;
    CHECK_SAME(neighbor_level,present_level);
    if (neighbor_level->hasChildren(neighbor_index)) {
      int face_index=element->codimension1ShapeIndex(f);
      const Shape2 *face=
        present_level->CODIMENSION1_LEVEL_CLASS::getShape(face_index);
      for (int l=0;l<face->numberCodimension1Shapes();l++) {
        int face_line_index=face->codimension1ShapeIndex(l);
        int i=line_is_needed.binarySearch(face_line_index);
        CHECK_TEST(i>=0);
        bool is_needed;
        int line_index;
        line_is_needed.getValue(i,is_needed,line_index);
        line_is_needed.setValue(i,true,line_index);;
      }
      continue;
    }
    for (int ff=0;ff<element->numberCodimension1Shapes();ff++) {
      LEVEL_CLASS *neighbor_neighbor_level=
        neighbor_level->neighborLevel(neighbor_index,ff);
      int neighbor_neighbor_index=
        neighbor_level->neighborIndex(neighbor_index,ff);
      if (neighbor_neighbor_index<0) continue;
      CHECK_TEST(neighbor_neighbor_level==present_level ||
        neighbor_neighbor_level==present_level->coarserLevel());
      if (neighbor_neighbor_level==present_level->coarserLevel()) {
        CHECK_TEST(!neighbor_neighbor_level->
          hasChildren(neighbor_neighbor_index));
        continue;
      }
      if (neighbor_neighbor_level->hasChildren(neighbor_neighbor_index))
      {
        const SHAPE_CLASS *neighbor_neighbor_element=
          neighbor_neighbor_level->getShape(neighbor_neighbor_index);
        for (int l=0;
        l<neighbor_neighbor_element->numberCodimension2Shapes();l++) {
          int neighbor_neighbor_line_index=neighbor_neighbor_element->
            codimension2ShapeIndex(l);
          int i=
            line_is_needed.binarySearch(neighbor_neighbor_line_index);
          if (i>=0) {
            bool is_needed;
            int line_index;
            line_is_needed.getValue(i,is_needed,line_index);
            line_is_needed.setValue(i,true,line_index);
          }
        }
      }
    }
  }
  for (int i=0;i<line_is_needed.getNumber();i++) {
    bool is_needed;
    int line_index;
    line_is_needed.getValue(i,is_needed,line_index);
    if (!is_needed) {
      int c(INT_MAX),v(INT_MAX);
      int interior_line_vertex=
        present_level->Level1::interiorChildVertex(line_index,c,v);
      vertices_used[interior_line_vertex]=false;
      const Shape1 *unneeded_line=
        present_level->Level1::getShape(line_index);
      for (int c=0;c<childrenPerCodimension2Shape();c++) {
        int unneeded_line_child_index=
          present_level->Level1::childIndex(line_index,c);
        present_finer_level->
          Level1::clearUserPointer(unneeded_line_child_index);
        present_finer_level->
          Level1::clearUserFlag(unneeded_line_child_index);
        present_finer_level->
          Level1::clearUsedFlag(unneeded_line_child_index);
        present_finer_level->
          Level1::setParent(unneeded_line_child_index,-1);
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//should have at least one boundary line
void Tessellation::lenoirSupportPointsOnTetrahedron(
const Shape3 *element,int degree,const Level3 *level,
NumPtr<Point<3> > &higher_order_mapping_support_points) const {
  CHECK_TEST(degree>1);
  const Tetrahedron *tetrahedron=
    dynamic_cast<const Tetrahedron*>(element);
  CHECK_POINTER(tetrahedron);
  int nbf=0;
  int bf=-1;
  for (int f=0;f<element->numberCodimension1Shapes();f++) {
    int face_element_index=element->codimension1ShapeIndex(f);
    const Shape2 *face_element=
      level->Level2::getShape(face_element_index);
    const BoundaryCondition *bc=
      level->Level2::getBoundaryCondition(face_element);
    if (bc!=0) {
      if (!bc->getBoundary()->isPlanar()) {
        nbf++;
        bf=f;
      }
    }
  }
  int nbl=0;
  int bl=-1;
  for (int l=0;l<element->numberCodimension2Shapes();l++) {
    int line_element_index=element->codimension2ShapeIndex(l);
    const Shape1 *line_element=
      level->Level1::getShape(line_element_index);
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
      level->Level1::getBoundaryConditions(line_element);
    if (bcs.first!=0) {
      if (!(bcs.first->getBoundary()->isPlanar())) {
        nbl++;
        bl=l;
      } else if (bcs.second!=0) {
        if (!bcs.second->getBoundary()->isPlanar()) {
          nbl++;
          bl=l;
        }
      }
    }
  }
//if use_full_degree_mapping_on_all_cells, default to first face
  if (nbl==0 && nbf==0) {
    nbf=1;
    bf=0;
  }
//Lenoir assumes that at most one face is on the boundary, or
//  at most one line is on the boundary
  CHECK_TEST((nbf==0 && nbl==1) || (nbf==1 && nbl==3));

  element->initializeMappingSupportPoints(degree,vertices,
    higher_order_mapping_support_points);

  NumPtr<Point<3> > *old_points=OPERATOR_NEW NumPtr<Point<3> >;
  old_points->
    allocate(higher_order_mapping_support_points.getNumber());
  NumPtr<Point<3> > *new_points=OPERATOR_NEW NumPtr<Point<3> >;
  new_points->allocate(old_points->getNumber());
  if (nbf==1) {
    NumPtr<Point<3> > eslp;
    Tetrahedron::staticEquallySpacedLatticePoints(degree,eslp);
    NumPtr<Point<2> > canonical_points;
    tetrahedron->lenoirCanonicalFacePoints(bf,eslp,canonical_points);
    int face_element_index=element->codimension1ShapeIndex(bf);
//  since Tetrahedron::lenoirCanonicalFacePoints uses Tetrahedron vertex
//    ordering, must not reorder vertices in face_element
    Shape2 *new_face_element=element->newFaceElement(bf);
    const Boundary *bo=level->Level2::getBoundaryCondition(
      face_element_index)->getBoundary();

    C0LagrangeTrianglePolynomials c0ltp(1);
    NumPtr<Point<3> > fv(new_face_element->numberVertices());
    for (int i=0;i<new_face_element->numberVertices();i++) {
      fv[i]=vertices[new_face_element->vertexIndex(i)];
    }
    for (int ijk=0;ijk<old_points->getNumber();ijk++) {
      NumPtr<REAL> values;
      c0ltp.computeValues(canonical_points[ijk],values);
      (*old_points)[ijk]=fv[0]*values[0]+fv[1]*values[1]+fv[2]*values[2];
    }
    for (int m=2;m<=degree;m++) {
      int neslp=new_face_element->numberEquallySpacedLatticePoints(m);
      NumPtr<Point<3> > face_mapping_support_points(neslp);
      for (int v=0;v<new_face_element->numberVertices();v++) {
        int iv=Triangle::staticEquallySpacedLatticePointIndexOfVertex(v,m);
        face_mapping_support_points[iv]=fv[v];
      }
      for (int l=0;l<new_face_element->numberCodimension1Shapes();l++) {
        int line_element_index=new_face_element->codimension1ShapeIndex(l);
        Shape1 *new_line_element=new_face_element->newLineElement(l);
        int nieslp=
          new_line_element->numberInteriorEquallySpacedLatticePoints(m);
        NumPtr<Point<SPACEDIM> > fsp(nieslp);
        pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
          level->Level1::getBoundaryConditions(line_element_index);
        if (bcs.first==0) {
          new_line_element->getIntermediatePoints(vertices,fsp);
        } else if (bcs.second==0) {
          bcs.first->getBoundary()->
            getIntermediatePointsOn(*new_line_element,vertices,fsp);
        } else {
          bcs.first->getBoundary()->
            getIntermediatePointsOnIntersectionWith(
            bcs.second->getBoundary(),*new_line_element,vertices,fsp);
        }
        NumPtr<int> indices;
        Triangle::staticEquallySpacedLatticePointIndicesOfFaceInterior(
          l,m,indices);
        for (int i=0;i<fsp.getNumber();i++) {
          int iv=indices[i];
          face_mapping_support_points[iv]=fsp[i];
        }
        delete new_line_element; new_line_element=0;
      }
      bo->computeInteriorLatticePoints(*new_face_element,m,bf,
        face_mapping_support_points);
      C0LagrangeTrianglePolynomials c0ltp(m);

      for (int ijk=0;ijk<new_points->getNumber();ijk++) {
        NumPtr<REAL> values;
//      values will be in dof order
        c0ltp.computeValues(canonical_points[ijk],values);
        (*new_points)[ijk]=0.;
        for (int v=0;v<values.getNumber();v++) {
          (*new_points)[ijk]+=face_mapping_support_points[v]*values[v];
        }
      }
      tetrahedron->lenoirInterpolationForFace(bf,m,eslp,
        *old_points,*new_points,higher_order_mapping_support_points);
//    in bits/stl_algobase.h:
      swap(old_points,new_points);
    }
    delete new_face_element; new_face_element=0;
  } else if (nbl==1) {
    NumPtr<Point<3> > eslp;
    Tetrahedron::staticEquallySpacedLatticePoints(degree,eslp);
    NumPtr<Point<1> > canonical_points;
    tetrahedron->lenoirCanonicalLinePoints(bl,eslp,canonical_points);
    int line_element_index=element->codimension2ShapeIndex(bl);
//  since Tetrahedron::lenoirCanonicalLinePoints uses Tetrahedron vertex
//    ordering, must not reorder vertices in face_element
    Shape1 *new_line_element=element->newLineElement(bl);
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
      level->Level1::getBoundaryConditions(line_element_index);

    C0LagrangeIntervalPolynomials c0lip(1);

    NumPtr<Point<3> > fv(new_line_element->numberVertices());
    for (int i=0;i<new_line_element->numberVertices();i++) {
      fv[i]=vertices[new_line_element->vertexIndex(i)];
    }
    for (int ijk=0;ijk<old_points->getNumber();ijk++) {
      NumPtr<REAL> values;
      c0lip.computeValues(canonical_points[ijk],values);
      (*old_points)[ijk]=fv[0]*values[0]+fv[1]*values[1];
    }
    for (int m=2;m<=degree;m++) {
      int neslp=new_line_element->numberEquallySpacedLatticePoints(m);
      NumPtr<Point<3> > line_mapping_support_points(neslp);
      for (int v=0;v<new_line_element->numberVertices();v++) {
        int iv=
          Interval::staticEquallySpacedLatticePointIndexOfVertex(v,m);
        line_mapping_support_points[iv]=fv[v];
      }
      NumPtr<Point<3> > lsp(line_mapping_support_points,1,m-1);
      if (bcs.first==0) {
        new_line_element->getIntermediatePoints(vertices,lsp);
      } else if (bcs.second==0) {
        bcs.first->getBoundary()->
          getIntermediatePointsOn(*new_line_element,vertices,lsp);
      } else {
        bcs.first->getBoundary()->
          getIntermediatePointsOnIntersectionWith(
          bcs.second->getBoundary(),*new_line_element,vertices,lsp);
      }
      C0LagrangeIntervalPolynomials c0lip(m);
      for (int ijk=0;ijk<new_points->getNumber();ijk++) {
        NumPtr<REAL> values;
        c0lip.computeValues(canonical_points[ijk],values);
        (*new_points)[ijk]=0.;
        for (int v=0;v<values.getNumber();v++) {
          (*new_points)[ijk]+=line_mapping_support_points[v]*values[v];
        }
      }
      tetrahedron->lenoirInterpolationForLine(bl,m,eslp,
        *old_points,*new_points,higher_order_mapping_support_points);
//    in bits/stl_algobase.h:
      swap(old_points,new_points);
    }
    delete new_line_element; new_line_element=0;
  }
  delete old_points;
  delete new_points;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//should have at least one boundary line
void Tessellation::lenoirSupportPointsOnPrism(const Shape3 *element,
int degree,const Level3 *level,
NumPtr<Point<3> > &higher_order_mapping_support_points) const {
  CHECK_TEST(degree>1);
  CHECK_POINTER(dynamic_cast<const Prism*>(element));

  int nbf=0;
  int nqbf=0;
  int bf=-1;
  for (int f=0;f<element->numberCodimension1Shapes();f++) {
    int face_element_index=element->codimension1ShapeIndex(f);
    const Shape2 *face_element=
      level->Level2::getShape(face_element_index);
    const BoundaryCondition *bc=
      level->Level2::getBoundaryCondition(face_element_index);
    if (bc!=0) {
      if (!bc->getBoundary()->isPlanar()) {
        nbf++;
        if (dynamic_cast<const Quadrilateral*>(face_element)!=0) {
          nqbf++;
          bf=f;
        }
      }
    }
  }
  int nll=0;
  int bl=-1;
  for (int l=0;l<element->numberCodimension2Shapes();l++) {
    int line_element_index=element->codimension2ShapeIndex(l);
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
      level->Level1::getBoundaryConditions(line_element_index);
    if (bcs.first!=0) {
      if (!bcs.first->getBoundary()->isPlanar()) {
        nll++;
        if (l<6) bl=l;
      }
    }
  }
  CHECK_TEST((nbf>0 && nqbf==0) // one or more Triangle faces on boundary
          || (nbf==1 && nqbf==1) // one Quadrilateral face on boundary
          || (nll==1 && bl<6) // one Quadrilateral-Triangle line on bdry
          || (nll>0 && bl==-1)); // one or more Quad-Quad lines on bdry
  if ((nbf>0 && nqbf==0) || (nbf==0 && nll>0 && bl==-1)) {
    for (int v=0;v<element->numberVertices();v++) {
      int iv=
        Prism::staticEquallySpacedLatticePointIndexOfVertex(v,degree);
      higher_order_mapping_support_points[iv]=
        vertices[element->vertexIndex(v)];
    }
//  either Quad-Quad lines on bdry, one or more Triangle faces on boundary
    for (int l=0;l<Prism::linesPerShape();l++) {
      if (Prism::isTriangleLine(l)) {
        Shape1 *new_line_element=element->newLineElement(l);
        int line_element_index=element->codimension2ShapeIndex(l);
//      const Shape1 *line_element=
//        level->Level1::getShape(line_element_index);
        NumPtr<Point<SPACEDIM> > lsp(degree-1);
        pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
          level->Level1::getBoundaryConditions(line_element_index);
        if (bcs.first==0) {
          new_line_element->getIntermediatePoints(vertices,lsp);
        } else if (bcs.second==0) {
          bcs.first->getBoundary()->
          getIntermediatePointsOn(*new_line_element,vertices,lsp);
        } else {
          bcs.first->getBoundary()->
          getIntermediatePointsOnIntersectionWith(
            bcs.second->getBoundary(),*new_line_element,vertices,lsp);
        }
        NumPtr<int> indices;
        Prism::staticEquallySpacedLatticePointIndicesOfLineInterior(l,
          degree,indices);
        for (int i=0;i<lsp.getNumber();i++) {
          int iv=indices[i];
          higher_order_mapping_support_points[iv]=lsp[i];
        }
        delete new_line_element; new_line_element=0;
      }
    }
    for (int l=0;l<Prism::linesPerShape();l++) {
      if (!Prism::isTriangleLine(l)) {
        Shape1 *new_line_element=element->newLineElement(l);
        int line_element_index=element->codimension2ShapeIndex(l);
        NumPtr<Point<SPACEDIM> > lsp(degree-1);
        pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
          level->Level1::getBoundaryConditions(line_element_index);
        if (bcs.first==0) {
          new_line_element->getIntermediatePoints(vertices,lsp);
        } else if (bcs.second==0) {
          bcs.first->getBoundary()->
          getIntermediatePointsOn(*new_line_element,vertices,lsp);
        } else {
          bcs.first->getBoundary()->
          getIntermediatePointsOnIntersectionWith(
            bcs.second->getBoundary(),*new_line_element,vertices,lsp);
        }
        NumPtr<int> indices;
        Prism::staticEquallySpacedLatticePointIndicesOfLineInterior(l,
          degree,indices);
        for (int i=0;i<lsp.getNumber();i++) {
          int iv=indices[i];
          higher_order_mapping_support_points[iv]=lsp[i];
        }
        delete new_line_element; new_line_element=0;
      }
    }
    for (int f=0;f<Prism::facesPerShape();f++) {
      if (Prism::isTriangleFace(f)) {
        const Shape2 *new_face_element=
          element->newFaceElement(f);
        int face_element_index=element->codimension1ShapeIndex(f);
        int neslp=
          new_face_element->numberEquallySpacedLatticePoints(degree);
        NumPtr<Point<SPACEDIM> > fsp(neslp);
        const BoundaryCondition *bc=level->CODIMENSION1_LEVEL_CLASS::
          getBoundaryCondition(face_element_index);
        if (bc!=0) {
          if (!bc->getBoundary()->isPlanar()) bf=f;
        }
        for (int v=0;v<new_face_element->numberVertices();v++) {
          int face_vertex_index=Triangle::
            staticEquallySpacedLatticePointIndexOfVertex(v,degree);
          fsp[face_vertex_index]=
            vertices[new_face_element->vertexIndex(v)];
        }
        int bl=-1;
        for (int l=0;l<new_face_element->numberCodimension1Shapes();l++) {
          int line_index=new_face_element->codimension1ShapeIndex(l);
          pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
            level->Level1::getBoundaryConditions(line_index);
          if (bcs.first!=0) {
            if (!bcs.first->getBoundary()->isPlanar()) bl=l;
            else if (bcs.second!=0) {
              if (!bcs.second->getBoundary()->isPlanar()) bl=l;
            }
          }
          NumPtr<int> element_line_indices;
          Prism::staticEquallySpacedLatticePointIndicesOfLineInterior(
            element->elementLineFromFace(f,l),degree,element_line_indices);
          NumPtr<int> face_line_indices;
          Triangle::staticEquallySpacedLatticePointIndicesOfFaceInterior(
            l,degree,face_line_indices);
          for (int i=0;i<face_line_indices.getNumber();i++) {
            fsp[face_line_indices[i]]=
              higher_order_mapping_support_points[element_line_indices[i]];
          }
        }
        if (bc==0 || bl<0) {
          new_face_element->getIntermediatePoints(vertices,fsp);
        } else {
          bc->getBoundary()->
            computeInteriorLatticePoints(*new_face_element,degree,bl,fsp);
        }
        NumPtr<int> indices;
        Prism::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
          degree,indices);
        NumPtr<int> face_indices;
        Triangle::staticInteriorLatticePointIndices(degree,face_indices);
        for (int i=0;i<face_indices.getNumber();i++) {
          int iv=indices[i];
          int iw=face_indices[i];
          higher_order_mapping_support_points[iv]=fsp[iw];
        }
        delete new_face_element; new_face_element=0;
      }
    }
    for (int f=0;f<Prism::facesPerShape();f++) {
      if (!Prism::isTriangleFace(f)) {
        Shape2 *new_face_element=element->newFaceElement(f);
        int face_element_index=element->codimension1ShapeIndex(f);
        int neslp=
          new_face_element->numberEquallySpacedLatticePoints(degree);
        NumPtr<Point<SPACEDIM> > fsp(neslp);
        const BoundaryCondition *bc=level->CODIMENSION1_LEVEL_CLASS::
          getBoundaryCondition(face_element_index);
        for (int v=0;v<new_face_element->numberVertices();v++) {
          int face_vertex_index=Quadrilateral::
            staticEquallySpacedLatticePointIndexOfVertex(v,degree);
          fsp[face_vertex_index]=
            vertices[new_face_element->vertexIndex(v)];
        }
        int bl=-1;
        for (int l=0;l<new_face_element->numberCodimension1Shapes();l++) {
          int line_index=new_face_element->codimension1ShapeIndex(l);
          pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
            level->Level1::getBoundaryConditions(line_index);
          if (bcs.first!=0) {
            if (!bcs.first->getBoundary()->isPlanar()) bl=l;
            else if (bcs.second!=0) {
              if (!bcs.second->getBoundary()->isPlanar()) bl=l;
            }
          }
          if (bf==-1 && bl>=0) bf=f; //no face & 1 quad line on bdry
          NumPtr<int> element_line_indices;
          int el=element->elementLineFromFace(f,l);
          element->equallySpacedLatticePointIndicesOfLineInterior(
            el,degree,element_line_indices);
          NumPtr<int> face_line_indices;
          Quadrilateral::
            staticEquallySpacedLatticePointIndicesOfFaceInterior(
            l,degree,face_line_indices);
          for (int i=0;i<face_line_indices.getNumber();i++) {
            int iv=face_line_indices[i];
            fsp[iv]=
              higher_order_mapping_support_points[element_line_indices[i]];
          }
        }
        if (bc==0 || bl<0) {
          new_face_element->computeInteriorLatticePoints(degree,bl,fsp);
        } else {
          bc->getBoundary()->
            computeInteriorLatticePoints(*new_face_element,degree,bl,fsp);
        }
        NumPtr<int> indices;
        element->equallySpacedLatticePointIndicesOfFaceInterior(f,degree,
          indices);
        NumPtr<int> face_indices;
        Quadrilateral::staticInteriorLatticePointIndices(degree,
          face_indices);
        for (int i=0;i<face_indices.getNumber();i++) {
          int iv=indices[i];
          int iw=face_indices[i];
          higher_order_mapping_support_points[iv]=fsp[iw];
        }
        delete new_face_element; new_face_element=0;
      }
    }
    element->computeInteriorLatticePoints(degree,bf,
      higher_order_mapping_support_points);
  } else {
    int tneslp=Triangle::staticNumberEquallySpacedLatticePoints(degree);
    NumPtr<Point<3> > *old_points=OPERATOR_NEW NumPtr<Point<3> >;
    old_points->allocate(tneslp);
    NumPtr<Point<3> > *new_points=OPERATOR_NEW NumPtr<Point<3> >;
    new_points->allocate(tneslp);
    if (nbf==1) { // one Quadrilateral face on boundary
      Shape2 *new_quad_face_element=element->newFaceElement(bf);
      NumPtr<Point<2> > eslp;
      Triangle::staticEquallySpacedLatticePoints(degree,eslp);
      for (int f=0;f<Prism::facesPerShape();f++) { 
        if (Prism::isTriangleFace(f)) {
//        may coincide with Tetrahedron ==> use Lenoir
          Shape2 *new_face_element=element->newFaceElement(f);
          int face_element_index=element->codimension1ShapeIndex(f);
          int line_element_index=new_face_element->
            sharedFaceElementIndex(new_quad_face_element);
          int bl=
            new_face_element->faceWithIndexEqualTo(line_element_index);
          int fneslp=
            new_face_element->numberEquallySpacedLatticePoints(degree);
          NumPtr<Point<3> > face_mapping_support_points(fneslp);
          new_face_element->initializeMappingSupportPoints(degree,vertices,
            face_mapping_support_points);
          NumPtr<Point<1> > canonical_points;
          const Triangle *new_face_triangle=
            dynamic_cast<const Triangle*>(new_face_element);
          CHECK_POINTER(new_face_triangle);
          new_face_triangle->
            lenoirCanonicalFacePoints(bl,eslp,canonical_points);
          Shape1 *new_line_element=new_face_element->newLineElement(bl);
          C0LagrangeIntervalPolynomials c0lip(1);
          NumPtr<Point<3> > fv(new_line_element->numberVertices());
          for (int i=0;i<new_line_element->numberVertices();i++) {
            fv[i]=vertices[new_line_element->vertexIndex(i)];
          }
          for (int ijk=0;ijk<old_points->getNumber();ijk++) {
            NumPtr<REAL> values;
            c0lip.computeValues(canonical_points[ijk],values);
            (*old_points)[ijk]=fv[0]*values[0]+fv[1]*values[1];
          }
          pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
            level->Level1::getBoundaryConditions(line_element_index);
          for (int m=2;m<=degree;m++) {
            int neslp=
              new_line_element->numberEquallySpacedLatticePoints(m);
            NumPtr<Point<3> > line_mapping_support_points(neslp);
            for (int v=0;v<new_line_element->numberVertices();v++) {
              int iv=Interval::
                staticEquallySpacedLatticePointIndexOfVertex(v,m);
              line_mapping_support_points[iv]=
                vertices[new_line_element->vertexIndex(v)];
            }
            NumPtr<Point<3> > lsp(line_mapping_support_points,1,m-1);
            if (bcs.second==0) {
              bcs.first->getBoundary()->getIntermediatePointsOn(
                *new_line_element,vertices,lsp);
            } else {
              bcs.first->getBoundary()->
                getIntermediatePointsOnIntersectionWith(
                  bcs.second->getBoundary(),*new_line_element,vertices,
                  lsp);
            }
            C0LagrangeIntervalPolynomials c0lip(m);
            for (int ijk=0;ijk<new_points->getNumber();ijk++) {
              NumPtr<REAL> values;
              c0lip.computeValues(canonical_points[ijk],values);
              (*new_points)[ijk]=0.;
              for (int v=0;v<values.getNumber();v++) {
                (*new_points)[ijk]+=
                  line_mapping_support_points[v]*values[v];
              }
            }
            new_face_triangle->lenoirInterpolationForFace(bl,m,eslp,
              *old_points,*new_points,face_mapping_support_points);
            swap(old_points,new_points);
          }
          NumPtr<int> indices;
          Prism::staticEquallySpacedLatticePointIndicesOfFace(f,degree,
            indices);
          for (int i=0;i<indices.getNumber();i++) {
            int iv=indices[i];
            higher_order_mapping_support_points[iv]=
              face_mapping_support_points[i];
          }
          delete new_line_element; new_line_element=0;
          delete new_face_element; new_face_element=0;
        }
      }
      for (int l=0;l<Prism::linesPerShape();l++) {
        if (!Prism::isTriangleLine(l)) {
          Shape1 *new_line_element=element->newLineElement(l);
          int line_element_index=element->codimension2ShapeIndex(l);
          pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
            level->Level1::getBoundaryConditions(line_element_index);
          int neslp=new_line_element->
            numberInteriorEquallySpacedLatticePoints(degree);
          NumPtr<Point<3> > lsp(neslp);
          if (bcs.first==0) {
            new_line_element->getIntermediatePoints(vertices,lsp);
          } else if (bcs.second==0) {
            bcs.first->getBoundary()->getIntermediatePointsOn(
              *new_line_element,vertices,lsp);
          } else {
            bcs.first->getBoundary()->
              getIntermediatePointsOnIntersectionWith(
                bcs.second->getBoundary(),*new_line_element,vertices,
                lsp);
          }
          NumPtr<int> element_indices;
          Prism::staticEquallySpacedLatticePointIndicesOfLineInterior(
            l,degree,element_indices);
          for (int i=0;i<element_indices.getNumber();i++) {
            int iw=element_indices[i];
            higher_order_mapping_support_points[iw]=lsp[i];
          }
          delete new_line_element; new_line_element=0;
        }
      }
      for (int f=0;f<Prism::facesPerShape();f++) {
        if (!Prism::isTriangleFace(f)) {
          Shape2 *new_face_element=element->newFaceElement(f);
          int face_element_index=element->codimension1ShapeIndex(f);
          int fneslp=
            new_face_element->numberEquallySpacedLatticePoints(degree);
          NumPtr<Point<3> > face_mapping_support_points(fneslp);
          for (int v=0;v<new_face_element->numberVertices();v++) {
            int iv=Quadrilateral::
              staticEquallySpacedLatticePointIndexOfVertex(v,degree);
            face_mapping_support_points[iv]=
              vertices[new_face_element->vertexIndex(v)];
          }
          int bl=-1;
          for (int l=0;l<Quadrilateral::facesPerShape();l++) {
            int line_element_index=
              new_face_element->codimension1ShapeIndex(l);
            int wl=element->lineWithIndexEqualTo(line_element_index);
            NumPtr<int> indices;
            Quadrilateral::staticEquallySpacedLatticePointIndicesOfFace(
              l,degree,indices);
            NumPtr<int> element_indices;
            Prism::staticEquallySpacedLatticePointIndicesOfLine(wl,degree,
              element_indices);
            for (int i=0;i<indices.getNumber();i++) {
              int iv=indices[i];
              int iw=element_indices[i];
              face_mapping_support_points[iv]=
                higher_order_mapping_support_points[iw];
            }
            pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
              level->Level1::getBoundaryConditions(line_element_index);
            if (bcs.first!=0) {
              if (!bcs.first->getBoundary()->isPlanar()) bl=l;
              else if (bcs.second!=0) {
                if (!bcs.second->getBoundary()->isPlanar()) bl=l;
              }
            }
          }
          const BoundaryCondition *bc=
            level->Level2::getBoundaryCondition(face_element_index);
          if (bc==0) {
            new_face_element->computeInteriorLatticePoints(degree,bl,
              face_mapping_support_points);
          } else {
            bc->getBoundary()->computeInteriorLatticePoints(
              *new_face_element,degree,bl,face_mapping_support_points);
          }
          NumPtr<int> face_indices;
          Quadrilateral::staticInteriorLatticePointIndices(degree,
            face_indices);
          NumPtr<int> element_indices;
          Prism::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
            degree,element_indices);
          for (int i=0;i<face_indices.getNumber();i++) {
            int iv=element_indices[i];
            higher_order_mapping_support_points[iv]=
              face_mapping_support_points[face_indices[i]];
          }
          delete new_face_element; new_face_element=0;
        }
      }
      element->computeInteriorLatticePoints(degree,bf,
        higher_order_mapping_support_points);
      delete new_quad_face_element; new_quad_face_element=0;
    } else { // one Quadrilateral-Triangle line on bdry
      int bf=-1;
      for (int f=0;f<Prism::facesPerShape();f++) {
        if (Prism::isTriangleFace(f)) {
//        may coincide with Tetrahedron ==> use Lenoir
          Shape2 *new_face_element=element->newFaceElement(f);
          int face_element_index=element->codimension1ShapeIndex(f);
          int fneslp=
            new_face_element->numberEquallySpacedLatticePoints(degree);
          NumPtr<Point<3> > face_mapping_support_points(fneslp);
          new_face_element->initializeMappingSupportPoints(degree,vertices,
            face_mapping_support_points);
          int bl=-1;
          for (int l=0;l<new_face_element->numberCodimension1Shapes();
          l++) {
            int line_index=new_face_element->codimension1ShapeIndex(l);
            pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
              level->Level1::getBoundaryConditions(line_index);
            if (bcs.first!=0) {
              if (!bcs.first->getBoundary()->isPlanar()) {
                bl=l;
                break;
              } else if (bcs.second!=0) {
                if (!bcs.second->getBoundary()->isPlanar()) {
                  bl=l;
                  break;
                }
              }
            }
          }
          if (bl>=0) {
            bf=f;
            NumPtr<Point<2> > eslp;
            Triangle::staticEquallySpacedLatticePoints(degree,eslp);
            const Triangle *new_face_triangle=
              dynamic_cast<const Triangle*>(new_face_element);
            CHECK_POINTER(new_face_triangle);

            NumPtr<Point<1> > canonical_points;
            new_face_triangle->
              lenoirCanonicalFacePoints(bl,eslp,canonical_points);
            Shape1 *new_line_element=new_face_element->newLineElement(bl);
            int line_element_index=
              new_face_element->codimension1ShapeIndex(bl);
            NumPtr<Point<3> > fv(new_line_element->numberVertices());
            for (int i=0;i<new_line_element->numberVertices();i++) {
              fv[i]=vertices[new_line_element->vertexIndex(i)];
            }
            C0LagrangeIntervalPolynomials c0lip(1);
            for (int ijk=0;ijk<old_points->getNumber();ijk++) {
              NumPtr<REAL> values;
              c0lip.computeValues(canonical_points[ijk],values);
              (*old_points)[ijk]=fv[0]*values[0]+fv[1]*values[1];
            }
            pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
              level->Level1::getBoundaryConditions(line_element_index);
            for (int m=2;m<=degree;m++) {
              int neslp=
                new_line_element->numberEquallySpacedLatticePoints(m);
              NumPtr<Point<3> > line_mapping_support_points(neslp);
              for (int v=0;v<new_line_element->numberVertices();v++) {
                int iv=Interval::
                  staticEquallySpacedLatticePointIndexOfVertex(v,m);
                line_mapping_support_points[iv]=fv[v];
              }
              NumPtr<Point<3> > 
                lsp(line_mapping_support_points,1,neslp-2);
              if (bcs.second==0) {
                bcs.first->getBoundary()->getIntermediatePointsOn(
                  *new_line_element,vertices,lsp);
              } else {
                bcs.first->getBoundary()->
                  getIntermediatePointsOnIntersectionWith(
                    bcs.second->getBoundary(),*new_line_element,vertices,
                    lsp);
              }
              C0LagrangeIntervalPolynomials c0lip(m);
              for (int ijk=0;ijk<new_points->getNumber();ijk++) {
                NumPtr<REAL> values;
                c0lip.computeValues(canonical_points[ijk],values);
                (*new_points)[ijk]=0.;
                for (int v=0;v<values.getNumber();v++) {
                  (*new_points)[ijk]+=
                    line_mapping_support_points[v]*values[v];
                }
              }
              new_face_triangle->lenoirInterpolationForFace(bl,m,eslp,
                *old_points,*new_points,face_mapping_support_points);
              swap(old_points,new_points);
            }
            delete new_line_element; new_line_element=0;
          }
          NumPtr<int> indices;
          Prism::staticEquallySpacedLatticePointIndicesOfFace(f,degree,
            indices);
          for (int i=0;i<indices.getNumber();i++) {
            int iv=indices[i];
            higher_order_mapping_support_points[iv]=
              face_mapping_support_points[i];
          }
          delete new_face_element; new_face_element=0;
        }
      }
      for (int l=0;l<Prism::linesPerShape();l++) {
        if (!Prism::isTriangleLine(l)) {
          Shape1 *new_line_element=element->newLineElement(l);
          int line_element_index=element->codimension2ShapeIndex(l);
          NumPtr<Point<SPACEDIM> > lsp(degree-1);
          new_line_element->getIntermediatePoints(vertices,lsp);
          NumPtr<int> indices;
          Prism::staticEquallySpacedLatticePointIndicesOfLineInterior(l,
            degree,indices);
          for (int i=0;i<lsp.getNumber();i++) {
            int iv=indices[i];
            higher_order_mapping_support_points[iv]=lsp[i];
          }
          delete new_line_element; new_line_element=0;
        }
      }
      for (int f=0;f<Prism::facesPerShape();f++) {
        if (!Prism::isTriangleFace(f)) {
          const Shape2 *new_face_element=element->newFaceElement(f);
          int face_element_index=element->codimension1ShapeIndex(f);
          int neslp=new_face_element->
            numberEquallySpacedLatticePoints(degree);
          NumPtr<Point<SPACEDIM> > fsp(neslp);
          for (int v=0;v<new_face_element->numberVertices();v++) {
            int face_vertex_index=Quadrilateral::
              staticEquallySpacedLatticePointIndexOfVertex(v,degree);
            fsp[face_vertex_index]=
              vertices[new_face_element->vertexIndex(v)];
          }
          int bl=-1;
          for (int l=0;l<new_face_element->numberCodimension1Shapes();
          l++) {
            int line_index=new_face_element->codimension1ShapeIndex(l);
            int wl=element->lineWithIndexEqualTo(line_index);

            NumPtr<int> face_line_indices;
            Quadrilateral::
              staticEquallySpacedLatticePointIndicesOfFaceInterior(l,
              degree,face_line_indices);
            NumPtr<int> element_line_indices;
            Prism::staticEquallySpacedLatticePointIndicesOfLineInterior(
              wl,degree,element_line_indices);
            for (int i=0;i<face_line_indices.getNumber();i++) {
              int iv=face_line_indices[i];
              fsp[iv]=higher_order_mapping_support_points[
                element_line_indices[i]];
            }
            if (Prism::isTriangleLine(wl)) {
              pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
                level->Level1::getBoundaryConditions(line_index);
              if (bcs.first!=0) {
                if (!bcs.first->getBoundary()->isPlanar()) bl=l;
                else if (bcs.second!=0) {
                  if (!bcs.second->getBoundary()->isPlanar()) bl=l;
                }
              }
            }
          }
          new_face_element->computeInteriorLatticePoints(degree,bl,fsp);
          NumPtr<int> element_indices;
          Prism::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
            degree,element_indices);
          NumPtr<int> face_indices;
          Quadrilateral::staticInteriorLatticePointIndices(degree,
            face_indices);
          for (int i=0;i<face_indices.getNumber();i++) {
            int iv=element_indices[i];
            higher_order_mapping_support_points[iv]=fsp[face_indices[i]];
          }
          delete new_face_element; new_face_element=0;
        }
      }
      element->computeInteriorLatticePoints(degree,bf,
        higher_order_mapping_support_points);
    }
    delete old_points; old_points=0;
    delete new_points; new_points=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void Tessellation::mappingSupportPointsOnBoundaryOf(
const SHAPE_CLASS *element,int degree,const LEVEL_CLASS *level,
NumPtr<Point<SPACEDIM> > &higher_order_mapping_support_points) const {
# ifdef INDEF
  for (int i=0;i<higher_order_mapping_support_points.getNumber();i++) {
    higher_order_mapping_support_points[i]=HUGE_VAL;
  }
# endif
# ifdef DEBUG
  bool debug_now=(element->elementDataIndex()==0);
# endif
  if (degree<=1) {
    int nv=element->numberVertices();
    for (int v=0;v<nv;v++) {
      int iv=element->equallySpacedLatticePointIndexOfVertex(v,degree);
      higher_order_mapping_support_points[iv]=
        vertices[element->vertexIndex(v)];
    }
  } else {
# if (SPACEDIM==2)
    int bf=-1;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_element_index=element->codimension1ShapeIndex(f);
      const BoundaryCondition *bc=level->CODIMENSION1_LEVEL_CLASS::
        getBoundaryCondition(face_element_index);
      if (bc!=0) {
        if (!bc->getBoundary()->isPlanar()) {
          bf=f;
          break;
        }
      }
    }
    if (bf==-1) { // no face on a non-plane boundary
      element->initializeMappingSupportPoints(degree,vertices,
        higher_order_mapping_support_points);
    } else {
      int nv=element->numberVertices();
      for (int v=0;v<nv;v++) {
        int iv=element->equallySpacedLatticePointIndexOfVertex(v,degree);
        higher_order_mapping_support_points[iv]=
          vertices[element->vertexIndex(v)];
      }
      for (int f=0;f<element->numberCodimension1Shapes();f++) {
        int face_element_index=element->codimension1ShapeIndex(f);
        const CODIMENSION1_SHAPE_CLASS *face_element=
          level->CODIMENSION1_LEVEL_CLASS::getShape(face_element_index);
        int nieslp=
          face_element->numberInteriorEquallySpacedLatticePoints(degree);
        if (nieslp>0) {
          NumPtr<Point<SPACEDIM> > fsp(nieslp);
          const BoundaryCondition *bc=level->
            CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_element);
          if (bc==0) {
            face_element->getIntermediatePoints(vertices,fsp);
          } else {
            bf=f;
            bc->getBoundary()->
              getIntermediatePointsOn(*face_element,vertices,fsp);
          }
          NumPtr<int> indices;
          element->equallySpacedLatticePointIndicesOfFaceInterior(f,degree,
            indices);
          for (int i=0;i<fsp.getNumber();i++) {
            int iv=indices[i];
            higher_order_mapping_support_points[iv]=fsp[i];
          }
        }
      }
      element->computeInteriorLatticePoints(degree,bf,
        higher_order_mapping_support_points);
    }
# else // 3D
    int bl=-1;
    for (int l=0;l<element->numberCodimension2Shapes();l++) {
      int line_index=element->codimension2ShapeIndex(l);
      pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
        level->Level1::getBoundaryConditions(line_index);
      if (bcs.first!=0) {
        if (!bcs.first->getBoundary()->isPlanar()) {
          bl=l;
          break;
        } else if (bcs.second!=0) {
          if (!bcs.second->getBoundary()->isPlanar()) {
            bl=l;
            break;
          }
        }
      }
    }
    if (bl==-1) { // no line on a non-plane boundary
      element->initializeMappingSupportPoints(degree,vertices,
        higher_order_mapping_support_points);
    } else {
      if (dynamic_cast<const Tetrahedron*>(element)!=0) {
        lenoirSupportPointsOnTetrahedron(element,degree,level,
          higher_order_mapping_support_points);
        return;
      } else if (dynamic_cast<const Prism*>(element)!=0) {
        lenoirSupportPointsOnPrism(element,degree,level,
          higher_order_mapping_support_points);
        return;
      } 
//    must be a hexahedron...
      int nv=element->numberVertices();
      for (int v=0;v<nv;v++) {
        int iv=element->equallySpacedLatticePointIndexOfVertex(v,degree);
        higher_order_mapping_support_points[iv]=
          vertices[element->vertexIndex(v)];
      }
      for (int l=0;l<element->numberCodimension2Shapes();l++) {
        int line_index=element->codimension2ShapeIndex(l);
        Shape1 *new_line_element=element->newLineElement(l);
        NumPtr<Point<SPACEDIM> > fsp(degree-1);
        pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
          level->Level1::getBoundaryConditions(line_index);
        if (bcs.first==0) {
          new_line_element->getIntermediatePoints(vertices,fsp);
        } else if (bcs.second==0) {
          bcs.first->getBoundary()->
            getIntermediatePointsOn(*new_line_element,vertices,fsp);
        } else {
          bcs.first->getBoundary()->
            getIntermediatePointsOnIntersectionWith(
            bcs.second->getBoundary(),*new_line_element,vertices,fsp);
        }
        NumPtr<int> indices;
        Hexahedron::staticEquallySpacedLatticePointIndicesOfLineInterior(
          l,degree,indices);
        for (int i=0;i<fsp.getNumber();i++) {
          int iv=indices[i];
          higher_order_mapping_support_points[iv]=fsp[i];
        }
        delete new_line_element; new_line_element=0;
      }
      int bf=-1;
      for (int f=0;f<element->numberCodimension1Shapes();f++) {
        int face_element_index=element->codimension1ShapeIndex(f);
        Shape2 *new_face_element=element->newFaceElement(f);
        int neslp=
          new_face_element->numberEquallySpacedLatticePoints(degree);
        NumPtr<Point<SPACEDIM> > fsp(neslp);
        const BoundaryCondition *bc=level->CODIMENSION1_LEVEL_CLASS::
          getBoundaryCondition(face_element_index);
        if (bc!=0) {
          if (!bc->getBoundary()->isPlanar()) {
            bf=f;
          }
        }
        for (int v=0;v<new_face_element->numberVertices();v++) {
          int face_vertex_index=Quadrilateral::
            staticEquallySpacedLatticePointIndexOfVertex(v,degree);
          int element_vertex_index=
            Hexahedron::staticEquallySpacedLatticePointIndexOfVertex(
            Hexahedron::staticElementVertexFromFace(f,v),degree);
          fsp[face_vertex_index]=
            higher_order_mapping_support_points[element_vertex_index];
        }
        int bl=-1;
        for (int l=0;l<new_face_element->numberCodimension1Shapes();l++) {
          int line_index=new_face_element->codimension1ShapeIndex(l);
          pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
            level->Level1::getBoundaryConditions(line_index);
          if (bcs.first!=0) {
            if (!bcs.first->getBoundary()->isPlanar()) bl=l;
            else if (bcs.second!=0) {
              if (!bcs.second->getBoundary()->isPlanar()) bl=l;
            }
          }

          NumPtr<int> element_line_indices;
          int el=Hexahedron::staticElementLineFromFace(f,l);
          Hexahedron::
            staticEquallySpacedLatticePointIndicesOfLineInterior(el,
            degree,element_line_indices);
          NumPtr<int> face_line_indices;
          Quadrilateral::
            staticEquallySpacedLatticePointIndicesOfFaceInterior(l,
            degree,face_line_indices);
          for (int i=0;i<face_line_indices.getNumber();i++) {
            fsp[face_line_indices[i]]=
              higher_order_mapping_support_points[element_line_indices[i]];
          }
        }
        if (bc==0 || bl<0) {
          new_face_element->getIntermediatePoints(vertices,fsp);
        } else {
          bc->getBoundary()->
            computeInteriorLatticePoints(*new_face_element,degree,bl,fsp);
        }
        NumPtr<int> indices;
        Hexahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(
          f,degree,indices);
        NumPtr<int> face_indices;
        Quadrilateral::staticInteriorLatticePointIndices(degree,
          face_indices);
        for (int i=0;i<face_indices.getNumber();i++) {
          int iv=indices[i];
          int iw=face_indices[i];
          higher_order_mapping_support_points[iv]=fsp[iw];
        }
        delete new_face_element; new_face_element=0;
      }
      element->computeInteriorLatticePoints(degree,bf,
        higher_order_mapping_support_points);
    }
# endif
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void Tessellation::getMappedMeasureOn(const Shape1 &face_element,
const Level1 *level,const NumPtr<Point<SPACEDIM-1> > &canonical_points,
NumPtr<Point<SPACEDIM> > &points,NumPtr<Point<SPACEDIM> > &normals,
NumPtr<REAL> &surface_measure) const {
  const BoundaryCondition *bc=
    level->getBoundaryCondition(&face_element);
  CHECK_POINTER(bc);
  bc->getBoundary()->getMappedMeasureOn(face_element,vertices,
    canonical_points,points,normals,surface_measure);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Tessellation::getMappedPointsOnLine(int line_index,
const Level1 *level,const NumPtr<Point<1> > &line_canonical_points,
NumPtr<Point<3> > &boundary_points,NumPtr<Point<SPACEDIM> > &dpointsdxi)
const {
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Tessellation::getMappedMeasureOn(const Shape2 &face_element,
const Level2 *level,const NumPtr<Point<2> > &canonical_points,
NumPtr<Point<3> > &points,NumPtr<Point<3> > &normals,
NumPtr<REAL> &surface_measure) const {
  const BoundaryCondition *bc=level->getBoundaryCondition(&face_element);
  CHECK_POINTER(bc);
  bc->getBoundary()->getMappedMeasureOn(face_element,vertices,
    canonical_points,points,normals,surface_measure);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::executeCoarsening () {
  clearUserFlags();
  if (coarsest_level->numberFinerLevels()<1) return;
  int cpc=childrenPerShape();
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    if (present_level->hasChildren()) {
      int present_child_index=present_level->childIndex(0);
      LEVEL_CLASS *present_finer_level=present_level->finerLevel();
      if (present_finer_level->coarsenFlagSet(present_child_index)) {
        present_level->setUserFlag();
        for (int c=0;c<cpc;c++) {
          present_child_index=present_level->childIndex(c);
          CHECK_TEST(present_finer_level->coarsenFlagSet(present_child_index));
          present_finer_level->clearCoarsenFlag(present_child_index);
        }
      }
    }
  }
  LEVEL_CLASS *next_to_finest_level=
    coarsest_level->finestLevel()->coarserLevel();
  for (bool doit=lastUsedShape(next_to_finest_level);doit;
  doit=prevUsedShape()) {
    if (present_level->userFlagSet()) deleteChildren();
  }
#ifdef DEBUG
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    CHECK_TEST(!present_level->userFlagSet());
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//we modified this to be more similar to 3D
#if (SPACEDIM==1)
void Tessellation::executeRefinement() {
//if we want to refine an element on the finest level, make new level
  LEVEL_CLASS *finest_level=coarsest_level->finestLevel();
  for (bool doit=firstShape(finest_level);doit;
  doit=nextShapeInLevel()) {
    if (present_level->used() && present_level->refineFlagSet()) {
      LEVEL_CLASS *new_level=OPERATOR_NEW LEVEL_CLASS(finest_level);
      break;
    }
  }
  finest_level=coarsest_level->finestLevel();
  LEVEL_CLASS *next_to_finest_level=finest_level->coarserLevel();
  int needed_vertices=0;
  int cpc=childrenPerShape();
  int cpcd1e=childrenPerCodimension1Shape();
  for (LEVEL_CLASS *level=next_to_finest_level;level;
  level=level->coarserLevel()) {
    int flagged_cells=0;
    for (bool doit=firstActiveShape(level);present_level==level && doit;
    doit=nextActiveShapeInLevel()) {
      if (present_level->refineFlagSet()) {
        flagged_cells++;
        needed_vertices+=
          present_level->getShape()->numberInteriorChildVertices();
      }
    }
    LEVEL_CLASS *finer_level=level->finerLevel();
    int used_elements=finer_level->numberUsedShapes();
    int nc=flagged_cells*cpc;
    finer_level->Level0::reserveSpace(used_elements+nc);
    finer_level->LEVEL_CLASS::reserveSpace(nc);
  }
  for (int i=0;i<vertices_used.getNumber();i++) {
    needed_vertices+=vertices_used[i];
  }
  if (needed_vertices>vertices.getNumber()) {
    NumPtr<Point<SPACEDIM> > vertices_extension(needed_vertices);
    vertices_extension.copyFrom(vertices);
    vertices.copy(vertices_extension);
    NumPtr<bool> vertices_used_extension(needed_vertices);
    vertices_used_extension.initialize(false);
    vertices_used_extension.copyFrom(vertices_used);
    vertices_used.copy(vertices_used_extension);
  }

//original deal code looped fine to coarse in 1D, coarse to fine in 2D, 3D
//  we loop coarse to fine in all dimensions
  int next_unused_vertex=0;
  for (LEVEL_CLASS *level=coarsest_level;level!=finest_level;
  level=level->finerLevel()) {
    LEVEL_CLASS *finer_level=level->finerLevel();
    int next_unused_element=0;
    for (bool doit=firstActiveShape(level);present_level==level && doit;
    doit=nextActiveShapeInLevel()) {
      if (present_level->refineFlagSet()) {
        present_level->clearRefineFlag();
        const SHAPE_CLASS *element=present_level->getShape();
        int present_index=present_level->presentIndex();
#ifdef DEBUG
        for (int f=0;f<element->numberCodimension1Shapes();
        f++) {
          int neighbor_index=
            present_level->neighborIndex(present_index,f);
          if (neighbor_index>=0) {
            const LEVEL_CLASS *neighbor_level=
              present_level->neighborLevel(present_index,f);
            CHECK_TEST( (neighbor_level==present_level && 
              !neighbor_level->coarsenFlagSet(neighbor_index)) ||
              (neighbor_level==present_level->coarserLevel() &&
              neighbor_level->refineFlagSet(neighbor_index)));
          }
        }
#endif
        NumPtr<int> interior_child_vertex_indices(
          element->numberInteriorChildVertices());
//      compute new child vertex interior to this element
        if (element->numberInteriorChildVertices()>0) {
          while (vertices_used[next_unused_vertex]) {
            next_unused_vertex++;
          }
          interior_child_vertex_indices[0]=next_unused_vertex;
          vertices_used[next_unused_vertex]=true;
          vertices[next_unused_vertex]=
            element->interiorChildPoint(vertices);
        }
        while (finer_level->used(next_unused_element)) {
          next_unused_element++;
        }
        CHECK_TEST(next_unused_element<finer_level->numberShapes());
        NumPtr<int> new_element_indices(cpc);
        for (int c=0;c<cpc;c++) {
          new_element_indices[c]=next_unused_element;
          CHECK_TEST(!finer_level->used(next_unused_element));
          if (c==0) present_level->setChildren(next_unused_element);
          Shape1 *child_element=element->newChildShape(c,
            next_unused_element,interior_child_vertex_indices);
          finer_level->setShape(next_unused_element,child_element);
          finer_level->setUsedFlag(next_unused_element);
          finer_level->clearUserFlag(next_unused_element);
          finer_level->clearUserPointer(next_unused_element);
          finer_level->setChildren(next_unused_element,-1);
          finer_level->setParent(next_unused_element,present_index);
          finer_level->setMaterialModel(next_unused_element,
                           present_level->getMaterialModel(element));
          finer_level->setSubdomainId(next_unused_element,
                          present_level->subdomainId(element));
          next_unused_element++;
        }
//      set child neighbors at child faces interior to parent element
        level->setInteriorChildNeighbors(new_element_indices);
        for (int f=0;f<element->numberCodimension1Shapes();f++) {
          int neighbor_index=
            present_level->neighborIndex(element,f);
          LEVEL_CLASS *neighbor_level=
            present_level->neighborLevel(element,f);
          if (neighbor_index<0) {
//          parent face has no neighbor:
//            set child neighbors to parent neighbors
            for (int s=0;s<cpcd1e;s++) {
              int c,cf;
              present_level->childFaceOnFaceOf(element,f,s,c,cf);
              finer_level->setNeighbor(new_element_indices[c],cf,0,-1);
            }
          } else {
            CHECK_SAME(neighbor_level,present_level);
            if (!neighbor_level->hasChildren(neighbor_index)) {
//            parent face has neighbor and neighbor has no children:
//              set child neighbors to parent neighbor            
              for (int s=0;s<cpcd1e;s++) {
                int c,cf;
                present_level->childFaceOnFaceOf(element,f,s,c,cf);
                finer_level->setNeighbor(new_element_indices[c],cf,
                  neighbor_level,neighbor_index);
              }
            } else {
//            parent face has neighbor and neighbor has children:
//              set child neighbors to children of neighbor of parent 
              int face_element_index=element->codimension1ShapeIndex(f);
              const Shape1 *neighbor_element=
                neighbor_level->getShape(neighbor_index);
              int neighbor_face=
                neighbor_element->faceWithIndexEqualTo(face_element_index);
              LEVEL_CLASS *neighbor_finer_level=
                neighbor_level->finerLevel();
              CHECK_BOUNDS(neighbor_face,0,
                neighbor_element->numberCodimension1Shapes());
//            set neighbors of child faces to neighbor children
              for (int s=0;s<cpcd1e;s++) {
                int c,cf;
                present_level->childFaceOnFaceOf(element,f,s,c,cf);
                int child_element_index=
                  present_level->childIndex(element,c);
                const Shape1 *child_element=
                  finer_level->getShape(child_element_index);
                int child_face_element_index=
                  child_element->codimension1ShapeIndex(cf);
                bool found=false;
                for (int ns=0;ns<cpcd1e;ns++) {
                  int nc=neighbor_element->childOnFace(neighbor_face,ns);
                  int nci=neighbor_level->childIndex(neighbor_index,nc);
                  const Shape1 *neighbor_child_element=
                    neighbor_finer_level->getShape(nci);
                  for (int ncf=0;
                  ncf<neighbor_child_element->numberCodimension1Shapes();
                  ncf++) {
                    if (neighbor_child_element->
                    codimension1ShapeIndex(ncf)==child_face_element_index){
                      finer_level->setNeighbor(new_element_indices[c],cf,
                        neighbor_finer_level,nci);
                      neighbor_finer_level->setNeighbor(nci,ncf,
                        finer_level,new_element_indices[c]);
                      found=true;
                      break;
                    }
                  }
                  if (found) break;
                }

              }
            }
          }
        }
      }
    }
  }
# ifdef DEBUG
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    CHECK_TEST(!present_level->refineFlagSet());
  }
  checkNeighbors();
# endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void Tessellation::executeRefinement() {
//if we want to refine an element on the finest level, make new level
  LEVEL_CLASS *finest_level=coarsest_level->finestLevel();
  for (bool doit=firstShape(finest_level);doit;
  doit=nextShapeInLevel()) {
    CHECK_SAME(present_level,finest_level)
    if (finest_level->used() && finest_level->refineFlagSet()) {
      LEVEL_CLASS *new_level=OPERATOR_NEW LEVEL_CLASS(finest_level);
      break;  
    }
  }
  for (bool doit=firstCodimension1UsedShape();doit;
  doit=nextCodimension1UsedShape()) {
    present_codimension1_level->clearUserFlag();
  }
//determine space needed for new elements and new vertices
  int needed_vertices=0;
  int cpc=childrenPerShape();
  int cpcd1e=childrenPerCodimension1Shape();
  LEVEL_CLASS *next_to_coarsest_level=
    coarsest_level->finestLevel()->coarserLevel();
  for (LEVEL_CLASS *level=next_to_coarsest_level;level;
  level=level->coarserLevel()) {
    int flagged_cells=0,needed_lines=0;
//  count elements to be refined (flagged_cells),
//    child vertices interior to those elements (needed_vertices),
//    child faces interior to those elements (needed_lines)
    for (bool doit=firstActiveShape(level);present_level==level && doit;
    doit=nextActiveShapeInLevel()) {
      if (present_level->refineFlagSet()) {
        const SHAPE_CLASS *element=level->getShape();
        flagged_cells++;
        needed_vertices+=element->numberInteriorChildVertices();
        needed_lines+=
          element->numberInteriorChildCodimension1Shapes();
        for (int f=0;f<element->numberCodimension1Shapes();
        f++) {
          int face_index=element->codimension1ShapeIndex(f);
          if (!level->Level1::hasChildren(face_index)) {
            level->Level1::setUserFlag(face_index);
          }
        }
      }
    }
//  add faces interior to element faces (needed_lines),
//    vertices interior to element faces
    for (bool doit=firstCodimension1UsedShape(level);
    present_codimension1_level==
    static_cast<Level1*>(level) && doit;
    doit=nextCodimension1UsedShapeInLevel()) {
      if (present_codimension1_level->userFlagSet()) {
        CHECK_TEST(!present_codimension1_level->hasChildren());
        const Shape1 *face_element=
          present_codimension1_level->getShape();
        needed_lines+=cpcd1e;
        needed_vertices+=face_element->numberInteriorChildVertices();
      }
    }
    LEVEL_CLASS *finer_level=level->finerLevel();
    int used_elements=finer_level->numberUsedShapes();
    int nc=flagged_cells*childrenPerShape();
    finer_level->Level0::reserveSpace(used_elements+nc);
    finer_level->Level1::reserveSpace(needed_lines);
    finer_level->LEVEL_CLASS::reserveSpace(nc);
  }
  for (int i=0;i<vertices_used.getNumber();i++) {
    needed_vertices+=vertices_used[i];
  }
//make space for new vertices, and copy old vertices
  if (needed_vertices>vertices.getNumber()) {
    NumPtr<Point<SPACEDIM> > vertices_extension(needed_vertices);
    vertices_extension.copyFrom(vertices);
    vertices.copy(vertices_extension);
    NumPtr<bool> vertices_used_extension(needed_vertices);
    vertices_used_extension.initialize(false);
    vertices_used_extension.copyFrom(vertices_used);
    vertices_used.copy(vertices_used_extension);
  }
#ifdef DEBUG
  for (bool doit=firstActiveShape();doit;doit=nextActiveShape()) {
    if (!present_level->refineFlagSet()) {
      const Shape2 *element=present_level->getShape();
      LEVEL_CLASS *finer_level=present_level->finerLevel();
      for (int f=0;f<element->numberCodimension1Shapes();f++)
      {
        int face_index=element->codimension1ShapeIndex(f);
        if (present_level->
        Level1::hasChildren(face_index)) {
          for (int c=0;c<cpcd1e;c++){
            int face_child_index=present_level->
              Level1::childIndex(face_index,c);
            CHECK_TEST(!finer_level->
              Level1::userFlagSet(face_child_index));
          }
        }
      }
    }
  }
#endif
  int next_unused_vertex=0;
//refine flagged element faces
  for (LEVEL_CLASS *level=coarsest_level;level->finerLevel();
  level=level->finerLevel()) {
    int next_unused_line=0;
    for (bool doit=firstCodimension1ActiveShape(level);
    present_codimension1_level==static_cast<Level1*>(level) && doit;
    doit=nextCodimension1ActiveShapeInLevel()) {
      if (present_codimension1_level->userFlagSet()) {
        const BoundaryCondition *bc=
          present_codimension1_level->getBoundaryCondition();
        const Shape1 *face_element=
          present_codimension1_level->getShape();
        int present_index=present_codimension1_level->presentIndex();
        NumPtr<int> interior_child_vertex_indices(
          face_element->numberInteriorChildVertices());
//      compute new child vertex interior to face element
        if (face_element->numberInteriorChildVertices()>0) {
          while (vertices_used[next_unused_vertex]) {
            next_unused_vertex++;
          }
          vertices_used[next_unused_vertex]=true;
          interior_child_vertex_indices[0]=next_unused_vertex;
          if (bc!=0) {
//          if the following fails, then you forgot to call
//          tess.setBoundary(mid,Boundary*)
            const Boundary *b=bc->getBoundary();
            CHECK_POINTER(b);
            vertices[next_unused_vertex]=
              b->getNewPointOn(*face_element,vertices);
          } else {
            vertices[next_unused_vertex]=
              face_element->interiorChildPoint(vertices);
          }
        }
        Level1 *present_finer_codimension1_level=
          present_codimension1_level->finerLevel();
//      construct child elements for present codimension 1 element
        for (int c=0;c<cpcd1e;c++) {
          while (present_finer_codimension1_level->
          used(next_unused_line)) {
            next_unused_line++;
          }
          if (c==0) {
            present_codimension1_level->setChildren(next_unused_line);
          }
          Shape1 *new_child_element=face_element->newChildShape(
            c,next_unused_line,interior_child_vertex_indices);
          present_finer_codimension1_level->
            setShape(next_unused_line,new_child_element);
          present_finer_codimension1_level->
            setChildren(next_unused_line,-1);
          present_finer_codimension1_level->
            setParent(next_unused_line,present_index);
          present_finer_codimension1_level->
            setUsedFlag(next_unused_line);
          present_finer_codimension1_level->
            clearUserPointer(next_unused_line);
          present_finer_codimension1_level->
            clearUserFlag(next_unused_line);
          present_finer_codimension1_level->
            setBoundaryCondition(next_unused_line,bc);
        }
        present_codimension1_level->clearUserFlag();
      }
    }
  }
//refine flagged elements
  for (LEVEL_CLASS *level=coarsest_level;level->finerLevel();
  level=level->finerLevel()) {
    int next_unused_element=0,next_unused_face=0;
    for (bool doit=firstActiveShape(level);present_level==level && doit;
    doit=nextActiveShapeInLevel()) {
      if (present_level->refineFlagSet()) {
        const SHAPE_CLASS *element=level->getShape();
        int present_index=level->presentIndex();
        LEVEL_CLASS *finer_level=level->finerLevel();
#ifdef DEBUG
        for (int f=0;f<element->numberCodimension1Shapes();
        f++) {
          int neighbor_index=level->neighborIndex(f);
          if (neighbor_index>=0) {
            LEVEL_CLASS *neighbor_level=level->neighborLevel(f);
            CHECK_TEST((neighbor_level==level &&
              !neighbor_level->coarsenFlagSet(neighbor_index)) ||
              (neighbor_level==level->coarserLevel() &&
              neighbor_level->refineFlagSet(neighbor_index)));
          }
        }
#endif
        level->clearRefineFlag();
        NumPtr<int> interior_child_vertex_indices(
          element->numberInteriorChildVertices());
//      compute child vertex interior to element
        if (element->numberInteriorChildVertices()>0) {
          while (vertices_used[next_unused_vertex]) {
            next_unused_vertex++;
          }
          interior_child_vertex_indices[0]=next_unused_vertex;
          vertices_used[next_unused_vertex]=true;
          NumPtr<Point<SPACEDIM> > 
            element_vertices(element->numberVertices());
          for (int i=0;i<element->numberVertices();i++) {
            element_vertices[i]=vertices[level->vertexIndex(i)];
          }
          NumPtr<int> interior_points_of_faces(
            element->numberCodimension1Shapes());
          for (int f=0;f<element->numberCodimension1Shapes();
          f++) {
            int face_index=element->codimension1ShapeIndex(f);
            int c(INT_MAX),v(INT_MAX);
            interior_points_of_faces[f]=level->
              Level1::interiorChildVertex(face_index,c,v);
          }
          const Quadrilateral *quadrilateral=
            dynamic_cast<const Quadrilateral*>(element);
          CHECK_POINTER(quadrilateral);
          vertices[next_unused_vertex]=quadrilateral->
            interiorChildPoint(vertices,interior_points_of_faces);
        }
//      store child face elements
        map<pair<int,int>,int> interior_child_line_map;
        for (int f=0;
        f<element->numberCodimension1Shapes();f++) {
          int face_index=element->codimension1ShapeIndex(f);
          for (int c=0;c<cpcd1e;c++){
            int child_face_index=
              level->Level1::childIndex(face_index,c);
            Shape1 *child_face_element=
              finer_level->Level1::getShape(child_face_index); 
            pair<int,int> p(child_face_element->vertexIndex(0),
                            child_face_element->vertexIndex(1));
            interior_child_line_map[p]=child_face_index;
          }
        }
        NumPtr<int> interior_child_line_vertex_indices(
          element->numberCodimension1Shapes());
        for (int l=0;l<element->numberCodimension1Shapes();
        l++) {
          int line_index=element->codimension1ShapeIndex(l);
          int c(INT_MAX),v(INT_MAX);
          interior_child_line_vertex_indices[l]=
            level->Level1::interiorChildVertex(line_index,c,v);
        }
//      construct codimension 1 elements interior to element
        for (int i=0;
        i<element->numberInteriorChildCodimension1Shapes();i++) {
          while (finer_level->
          Level1::used(next_unused_face)) {
            next_unused_face++;
          }
          Shape1 *interior_child_face_element=element->
            newInteriorCodimension1Shape(i,next_unused_face,
              interior_child_vertex_indices,
              interior_child_line_vertex_indices);
          finer_level->Level1::setShape(next_unused_face,
            interior_child_face_element);
          finer_level->Level1::setUsedFlag(next_unused_face);
          finer_level->Level1::clearUserFlag(next_unused_face);
          finer_level->Level1::clearUserPointer(next_unused_face);
          finer_level->Level1::setChildren(next_unused_face,-1);
          finer_level->Level1::setParent(next_unused_face,-1);
          finer_level->
            Level1::setBoundaryCondition(next_unused_face,0);
          pair<int,int> p(interior_child_face_element->vertexIndex(0),
                          interior_child_face_element->vertexIndex(1));
          interior_child_line_map[p]=
            interior_child_face_element->elementDataIndex();
        }
//      construct child elements
        NumPtr<int> new_element_indices(childrenPerShape());
        MaterialModel *mm=present_level->getMaterialModel();
        int sid=present_level->subdomainId();
        for (int c=0;c<childrenPerShape();c++) {
          while (finer_level->used(next_unused_element)) {
            next_unused_element++;
          }
          if (c==0) level->setChildren(next_unused_element);
          new_element_indices[c]=next_unused_element;
          Shape2 *child_element=element->newChildShape(c,
            next_unused_element,interior_child_vertex_indices,
            interior_child_line_vertex_indices,interior_child_line_map);
#ifdef DEBUG
          CHECK_TEST(child_element->volume(vertices)>0.);
          for (int f=0;
          f<child_element->numberCodimension1Shapes();f++) {
            const Shape1 *child_face_element=finer_level->Level1::
              getShape(child_element->codimension1ShapeIndex(f));
            child_element->checkFaceVertices(f,child_face_element);
          }
#endif
          finer_level->
            setShape(next_unused_element,child_element);
          finer_level->setUsedFlag(next_unused_element);
          finer_level->clearUserFlag(next_unused_element);
          finer_level->clearUserPointer(next_unused_element);
          finer_level->setChildren(next_unused_element,-1);
          finer_level->setParent(next_unused_element,present_index);
          finer_level->setMaterialModel(next_unused_element,mm);
          finer_level->setSubdomainId(next_unused_element,sid);
        }
//      set neighbors of child faces interior to element
        level->setInteriorChildNeighbors(new_element_indices);
//      set neighbors of child faces on element faces
        for (int f=0;f<element->numberCodimension1Shapes();f++) {
          int neighbor_index=level->neighborIndex(f);
          LEVEL_CLASS *neighbor_level=level->neighborLevel(f);
          if (neighbor_index<0) {
//          parent face has no neighbor:
//            set child neighbors to parent neighbors
            for (int s=0;s<cpcd1e;s++) {
              int c,cf;
              present_level->childFaceOnFaceOf(element,f,s,c,cf);
              finer_level->setNeighbor(new_element_indices[c],cf,0,-1);
            }
          } else {
            CHECK_SAME(neighbor_level,present_level);
            if (!neighbor_level->hasChildren(neighbor_index)) {
//            parent face has neighbor and neighbor has no children:
//              set child neighbors to parent neighbor            
              for (int s=0;s<cpcd1e;s++) {
                int c,cf;
                present_level->childFaceOnFaceOf(element,f,s,c,cf);
                finer_level->setNeighbor(new_element_indices[c],cf,
                  neighbor_level,neighbor_index);
              }
            } else {
//            parent face has neighbor and neighbor has children:
//
              int face_element_index=element->codimension1ShapeIndex(f);
              const Shape2 *neighbor_element=
                neighbor_level->getShape(neighbor_index);
              int neighbor_face=
                neighbor_element->faceWithIndexEqualTo(face_element_index);
              LEVEL_CLASS *neighbor_finer_level=
                neighbor_level->finerLevel();
              CHECK_BOUNDS(neighbor_face,0,
                neighbor_element->numberCodimension1Shapes());
//            set neighbors of child faces to neighbor children
              for (int s=0;s<cpcd1e;s++) {
                int c,cf;
                present_level->childFaceOnFaceOf(element,f,s,c,cf);
                int child_element_index=present_level->childIndex(c);
                const Shape2 *child_element=
                  finer_level->getShape(child_element_index);
                int child_face_element_index=
                  child_element->codimension1ShapeIndex(cf);
                bool found=false;
                for (int ns=0;ns<cpcd1e;ns++) {
                  int nc=neighbor_element->childOnFace(neighbor_face,ns);
                  int nci=neighbor_level->childIndex(neighbor_index,nc);
                  const Shape2 *neighbor_child_element=
                    neighbor_finer_level->getShape(nci);
                  for (int ncf=0;
                  ncf<neighbor_child_element->numberCodimension1Shapes();
                  ncf++) {
                    if (neighbor_child_element->
                    codimension1ShapeIndex(ncf)==child_face_element_index){
                      finer_level->setNeighbor(new_element_indices[c],cf,
                        neighbor_finer_level,nci);
                      neighbor_finer_level->setNeighbor(nci,ncf,
                        finer_level,new_element_indices[c]);
                      found=true;
                      break;
                    }
                  }
                  if (found) break;
                }
              }
            }
          }
        }
#ifdef DEBUG
        present_level->checkNeighbors(present_level->presentIndex());
#endif
      }
    }
  }
#ifdef DEBUG
  for (bool doit=firstCodimension1UsedShape();doit;
  doit=nextCodimension1UsedShape()) {
    CHECK_TEST(!present_codimension1_level->userFlagSet());
  }
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    CHECK_TEST(!present_level->refineFlagSet());
  }
  checkNeighbors();
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Tessellation::executeRefinement() {
  Level3 *finest_level=coarsest_level->finestLevel();
  for (bool doit=firstShape(finest_level);doit;
  doit=nextShapeInLevel()) {
    if (present_level->used() && present_level->refineFlagSet()) {
      Level3 *new_level=OPERATOR_NEW Level3(finest_level);
      break;  
    }
  }

  for (bool doit=firstCodimension1UsedShape();doit;
  doit=nextCodimension1UsedShape()) {
    present_codimension1_level->clearUserFlag();
  }
  for (bool doit=firstCodimension2UsedShape();doit;
  doit=nextCodimension2UsedShape()) {
    present_codimension2_level->clearUserFlag();
  }

  int needed_vertices=0;
  int cpc=childrenPerShape();
  int cpcd1e=childrenPerCodimension1Shape();
  for (Level3 *level=coarsest_level->finestLevel()->coarserLevel();
  level; level=level->coarserLevel()) {
    int flagged_cells=0,needed_lines=0,needed_faces=0;
    for (bool doit=firstActiveShape(level);present_level==level && doit;
    doit=nextActiveShapeInLevel()) {
      if (level->refineFlagSet()) {
        const Shape3 *element=level->getShape();
        flagged_cells++;
        needed_vertices+=element->numberInteriorChildVertices();
        needed_faces+=
          element->numberInteriorChildCodimension1Shapes();
        needed_lines+=
          element->numberInteriorChildCodimension2Shapes();;
        for (int f=0;f<element->numberCodimension1Shapes();f++)
        {
          int face_index=element->codimension1ShapeIndex(f);
          const Shape2 *face_element=
            level->Level2::getShape(face_index);
          if (!level->Level2::hasChildren(face_index)) {
            level->Level2::setUserFlag(face_index);
          }
        }
        for (int l=0;l<element->numberCodimension2Shapes();l++)
        {
          int line_index=element->codimension2ShapeIndex(l);
          if (!level->Level1::hasChildren(line_index)) {
            level->Level1::setUserFlag(line_index);
          }
        }
      }
    }
//  add faces interior to element faces (needed_faces),
//    lines interior to element faces (needed_lines),
//    vertices interior to element faces
    for (bool doit=firstCodimension1UsedShape(level);
    present_codimension1_level==static_cast<Level2*>(level) && doit;
    doit=nextCodimension1UsedShapeInLevel()) {
      if (present_codimension1_level->userFlagSet()) {
        CHECK_TEST(!present_codimension1_level->hasChildren());
        const Shape2 *face_element=
          present_codimension1_level->getShape();
        needed_faces+=cpcd1e;
        needed_lines+=
          face_element->numberInteriorChildCodimension1Shapes();
        needed_vertices+=face_element->numberInteriorChildVertices();
      }
    }
//  add lines interior to element lines (needed_lines),
//    vertices interior to element lines
    for (bool doit=firstCodimension2UsedShape(level);
    present_codimension2_level==static_cast<Level1*>(level) && doit;
    doit=nextCodimension2UsedShapeInLevel()) {
      if (present_codimension2_level->userFlagSet()) {
        CHECK_TEST(!present_codimension2_level->hasChildren());
        const Shape1 *line_element=
          present_codimension2_level->getShape();
        needed_lines+=childrenPerCodimension2Shape();
        needed_vertices+=line_element->numberInteriorChildVertices();
      }
    }
    Level3 *finer_level=level->finerLevel();
    int used_elements=finer_level->numberUsedShapes();
    int nc=flagged_cells*childrenPerShape();
    finer_level->Level0::reserveSpace(used_elements+nc);
    finer_level->Level1::reserveSpace(needed_lines);
    finer_level->Level2::reserveSpace(needed_faces);
    finer_level->reserveSpace(nc);
  }

  for (int i=0;i<vertices_used.getNumber();i++) {
    needed_vertices+=vertices_used[i];
  }

  if (needed_vertices>vertices.getNumber()) {
    NumPtr<Point<SPACEDIM> > vertices_extension(needed_vertices);
    vertices_extension.copyFrom(vertices);
    vertices.copy(vertices_extension);
    NumPtr<bool> vertices_used_extension(needed_vertices);
    vertices_used_extension.initialize(false);
    vertices_used_extension.copyFrom(vertices_used);
    vertices_used.copy(vertices_used_extension);
  }

#ifdef DEBUG
  for (bool doit=firstActiveShape();doit;doit=nextActiveShape()) {
    if (!present_level->refineFlagSet()) {
      const Shape3 *element=present_level->getShape();
      Level3 *finer_level=present_level->finerLevel();
      for (int l=0;l<element->numberCodimension2Shapes();l++) {
        int line_index=element->codimension2ShapeIndex(l);
        if (present_level->Level1::hasChildren(line_index)) {
          for (int c=0;c<childrenPerCodimension2Shape();c++){
            int line_child_index=present_level->
              Level1::childIndex(line_index,c);
            CHECK_TEST(!finer_level->
                Level1::userFlagSet(line_child_index));
          }
        }
      }
    }
  }
#endif

  int next_unused_vertex=0;
  for (Level3 *level=coarsest_level;level->finerLevel();
  level=level->finerLevel()) {
    Level3 *finer_level=level->finerLevel();
    int next_unused_line=0;
    for (bool doit=firstCodimension2ActiveShape(level);
    present_codimension2_level==static_cast<Level1*>(level) && doit;
    doit=nextCodimension2ActiveShapeInLevel()) {
      if (present_codimension2_level->userFlagSet()) {
        while (vertices_used[next_unused_vertex]) next_unused_vertex++;
        vertices_used[next_unused_vertex]=true;
        pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
          present_codimension2_level->getBoundaryConditions();
        const Shape1 *line_element=
          present_codimension2_level->getShape();
        int present_index=present_codimension2_level->presentIndex();
        if (bcs.first!=0) {
          if (bcs.second==0) {
            vertices[next_unused_vertex]=bcs.first->getBoundary()->
              getNewPointOn(*line_element,vertices);
          } else {
            vertices[next_unused_vertex]=bcs.first->getBoundary()->
              getNewPointOnIntersectionWith(bcs.second->getBoundary(),
                *line_element,vertices);
          }
        } else {
          vertices[next_unused_vertex]=
            line_element->interiorChildPoint(vertices);
        }
        NumPtr<int> interior_child_vertex_indices(
          line_element->numberInteriorChildVertices());
        for (int v=0;
        v<interior_child_vertex_indices.getNumber();v++) {
          interior_child_vertex_indices[v]=next_unused_vertex;
        }
        for (int c=0;c<childrenPerCodimension2Shape();c++) {
          while (finer_level->Level1::used(next_unused_line)) {
            next_unused_line++;
          }
          if (c==0) {
            present_codimension2_level->setChildren(next_unused_line);
          }
          Shape1 *new_child_element=line_element->newChildShape(
            c,next_unused_line,interior_child_vertex_indices);
          finer_level->
            Level1::setShape(next_unused_line,new_child_element);
          finer_level->Level1::setChildren(next_unused_line,-1);
          finer_level->Level1::setParent(next_unused_line,present_index);
          finer_level->Level1::setUsedFlag(next_unused_line);
          finer_level->Level1::clearUserPointer(next_unused_line);
          finer_level->Level1::clearUserFlag(next_unused_line);
          finer_level->
            Level1::setBoundaryConditions(next_unused_line,bcs);
        }
        present_codimension2_level->clearUserFlag();
      }
    }
  }
  for (Level3 *level=coarsest_level;level->finerLevel();
  level=level->finerLevel()) {
    Level3 *finer_level=level->finerLevel();
    int next_unused_line=0,next_unused_face=0;
    for (bool doit=firstCodimension1ActiveShape(level);
    present_codimension1_level==static_cast<Level2*>(level) && doit;
    doit=nextCodimension1ActiveShapeInLevel()) {
      if (present_codimension1_level->userFlagSet()) {
        const BoundaryCondition *bc=
          present_codimension1_level->getBoundaryCondition();
        const Shape2 *face_element=
          present_codimension1_level->getShape();
        int present_index=present_codimension1_level->presentIndex();
        NumPtr<int> interior_child_vertex_indices(
          face_element->numberInteriorChildVertices());
        const Triangle *face_triangle=
          dynamic_cast<const Triangle*>(face_element);
        const Quadrilateral *face_quadrilateral=
          dynamic_cast<const Quadrilateral*>(face_element);
//      compute new child vertex interior to face element
        if (face_element->numberInteriorChildVertices()>0) {
          while (vertices_used[next_unused_vertex]) {
            next_unused_vertex++;
          }
          vertices_used[next_unused_vertex]=true;
          interior_child_vertex_indices[0]=next_unused_vertex;
          CHECK_POINTER(face_quadrilateral);
          if (bc!=0) {
            vertices[next_unused_vertex]=bc->getBoundary()->
              getNewPointOn(*face_element,vertices);
          } else {
            NumPtr<int> interior_points_of_face_lines(
              face_element->numberCodimension1Shapes());
            for (int f=0;f<face_element->numberCodimension1Shapes();f++) {
              int face_line_index=
                face_element->codimension1ShapeIndex(f);
              int c(INT_MAX),v(INT_MAX);
              interior_points_of_face_lines[f]=
                level->Level1::interiorChildVertex(face_line_index,c,v);
            }
            vertices[next_unused_vertex]=
              face_quadrilateral->interiorChildPoint(vertices,
                interior_points_of_face_lines);
          }
        }
//      store child line elements
        map<pair<int,int>,int> interior_child_line_map;
        for (int f=0;
        f<face_element->numberCodimension1Shapes();f++) {
          int line_index=face_element->codimension1ShapeIndex(f);
          for (int c=0;c<childrenPerCodimension2Shape();
          c++){
            int line_child_index=present_codimension1_level->
              Level1::childIndex(line_index,c);
            const Shape1 *child_line_element=
              finer_level->Level1::getShape(line_child_index);
            pair<int,int> p(child_line_element->vertexIndex(0),
                            child_line_element->vertexIndex(1));
            interior_child_line_map[p]=line_child_index;
          }
        }
        NumPtr<int> interior_child_line_vertex_indices(
          face_element->numberCodimension1Shapes());
        for (int l=0;l<face_element->numberCodimension1Shapes();
        l++) {
          int line_index=face_element->codimension1ShapeIndex(l);
          int c(INT_MAX),v(INT_MAX);
          interior_child_line_vertex_indices[l]=
            level->Level1::interiorChildVertex(line_index,c,v);
        }
//      construct interior child line elements for present face element
        for (int i=0;
        i<face_element->numberInteriorChildCodimension1Shapes();i++) {
          while (finer_level->Level1::used(next_unused_line)) {
            next_unused_line++;
          }
          Shape1 *new_line=face_element->newInteriorCodimension1Shape(i,
              next_unused_line,interior_child_vertex_indices,
              interior_child_line_vertex_indices);
          finer_level->
            Level1::setShape(next_unused_line,new_line);
          finer_level->Level1::setChildren(next_unused_line,-1);
          finer_level->Level1::setParent(next_unused_line,-1);
          finer_level->Level1::setUsedFlag(next_unused_line);
          finer_level->Level1::clearUserPointer(next_unused_line);
          finer_level->Level1::clearUserFlag(next_unused_line);
          finer_level->Level1::setBoundaryCondition(next_unused_line,bc);
          pair<int,int> p(new_line->vertexIndex(0),
                          new_line->vertexIndex(1));
          interior_child_line_map[p]=new_line->elementDataIndex();

        }
//      construct child elements for present codimension 1 element
        for (int c=0;c<cpcd1e;c++) {
          while (finer_level->Level2::used(next_unused_face)) {
            next_unused_face++;
          }
          if (c==0) {
            present_codimension1_level->setChildren(next_unused_face);
          }
          Shape2 *new_child_element=face_element->newChildShape(
            c,next_unused_face,interior_child_vertex_indices,
            interior_child_line_vertex_indices,interior_child_line_map);
          finer_level->
            Level2::setShape(next_unused_face,new_child_element);
          finer_level->Level2::setChildren(next_unused_face,-1);
          finer_level->Level2::setParent(next_unused_face,present_index);
          finer_level->Level2::setUsedFlag(next_unused_face);
          finer_level->Level2::clearUserPointer(next_unused_face);
          finer_level->Level2::clearUserFlag(next_unused_face);
          finer_level->Level2::setBoundaryCondition(next_unused_face,bc);
#ifdef DEBUG
          for (int f=0;
          f<new_child_element->numberCodimension1Shapes();f++) {
            int face_element_index=
              new_child_element->codimension1ShapeIndex(f);
            const Shape1 *face_element=finer_level->
              Level1::getShape(face_element_index);
            new_child_element->checkFaceVertices(f,face_element);
          }
#endif
        }
        present_codimension1_level->clearUserFlag();
      }
    }
  }

  for (Level3 *level=coarsest_level;level->finerLevel();
  level=level->finerLevel()) {
    int next_unused_element=0,next_unused_face=0,next_unused_line=0;
    for (bool doit=firstActiveShape(level);present_level==level && doit;
    doit=nextActiveShapeInLevel()) {
      if (present_level->refineFlagSet()) {
        const Shape3 *element=level->getShape();
        int present_index=present_level->presentIndex();
        Level3 *finer_level=level->finerLevel();
#ifdef DEBUG
        for (int f=0;f<element->numberCodimension1Shapes();
        f++) {
          int neighbor_index=level->neighborIndex(f);
          if (neighbor_index>=0) {
            Level3 *neighbor_level=level->neighborLevel(f);
            CHECK_TEST((neighbor_level==level &&
              !neighbor_level->coarsenFlagSet(neighbor_index)) ||
              (neighbor_level==level->coarserLevel() &&
              neighbor_level->refineFlagSet(neighbor_index)));
          }
        }
#endif
        level->clearRefineFlag();
        NumPtr<int> interior_child_vertex_indices(
          element->numberInteriorChildVertices());
        const Hexahedron *hexahedron=
          dynamic_cast<const Hexahedron*>(element);
        if (element->numberInteriorChildVertices()>0) {
          CHECK_POINTER(hexahedron);
          while (vertices_used[next_unused_vertex]) {
            next_unused_vertex++;
          }
          vertices_used[next_unused_vertex]=true;
          NumPtr<int> interior_points_of_faces(
            element->numberCodimension1Shapes());
          for (int f=0;f<element->numberCodimension1Shapes();
          f++) {
            int c(INT_MAX),v(INT_MAX);
            int interior_child_vertex=
              level->Level2::interiorChildVertex(
                element->codimension1ShapeIndex(f),c,v);
            interior_points_of_faces[f]=interior_child_vertex;
          }
          NumPtr<int> interior_points_of_element_lines(
            element->numberCodimension2Shapes());
          for (int l=0;l<element->numberCodimension2Shapes();
          l++) {
            int line_index=element->codimension2ShapeIndex(l);
            int c(INT_MAX),v(INT_MAX);
            interior_points_of_element_lines[l]=
              level->Level1::interiorChildVertex(line_index,c,v);
          }
          interior_child_vertex_indices[0]=next_unused_vertex;
          vertices[next_unused_vertex]=hexahedron->interiorChildPoint(
            vertices,interior_points_of_faces,
            interior_points_of_element_lines);
        }
//      store child line elements
        map<pair<int,int>,int> interior_child_line_map;
        for (int l=0;l<element->numberCodimension2Shapes();l++)
        {
          int line_index=element->codimension2ShapeIndex(l);
          for (int c=0;c<childrenPerCodimension2Shape();c++) {
            int line_child_index=
              level->Level1::childIndex(line_index,c);
            const Shape1 *line_element=
              finer_level->Level1::getShape(line_child_index);
            pair<int,int> p(line_element->vertexIndex(0),
                            line_element->vertexIndex(1));
            interior_child_line_map[p]=line_child_index;
          }
        }
//      store child face elements
        NumPtr<int> interior_child_face_vertex_indices(
          element->numberCodimension1Shapes());
        map<pair<pair<int,int>,pair<int,int> >,int> 
          interior_child_face_map;
        for (int f=0;f<element->numberCodimension1Shapes();f++)
        {
          int face_index=element->codimension1ShapeIndex(f);
          const Shape2 *face_element=
            level->Level2::getShape(face_index);
          if (face_element->numberInteriorChildVertices()>0) {
            int c(INT_MAX),v(INT_MAX);
            interior_child_face_vertex_indices[f]=
              level->Level2::interiorChildVertex(face_index,c,v);
          } else {
            interior_child_face_vertex_indices[f]=INT_MAX;
          }
          for (int c=0;c<cpcd1e;c++){
            int child_face_index=
              level->Level2::childIndex(face_index,c);
            const Shape2 *child_face_element=finer_level->
              Level2::getShape(child_face_index);
            if (child_face_element->numberVertices()==3) {
              pair<pair<int,int>,pair<int,int> >
                p(make_pair(child_face_element->vertexIndex(0),
                            child_face_element->vertexIndex(1)),
                  make_pair(child_face_element->vertexIndex(2),-1));
              interior_child_face_map[p]=child_face_index;
            } else {
              pair<pair<int,int>,pair<int,int> >
                p(make_pair(child_face_element->vertexIndex(0),
                            child_face_element->vertexIndex(1)),
                  make_pair(child_face_element->vertexIndex(2),
                            child_face_element->vertexIndex(3)));
              interior_child_face_map[p]=child_face_index;
            }
          }
          for (int i=0;
          i<face_element->numberInteriorChildCodimension1Shapes();i++) {
            int c0,c1;
            face_element->
              interiorChildCodimension1ShapeNeighbors(i,c0,c1);
            int child_face_index0=level->
              Level2::childIndex(face_index,c0);
            const Shape2 *child_face_element0=finer_level->
              Level2::getShape(child_face_index0);
            int child_face_index1=level->
              Level2::childIndex(face_index,c1);
            const Shape2 *child_face_element1=finer_level->
              Level2::getShape(child_face_index1);
            int child_face_interior_line_index=child_face_element0->
              sharedFaceElementIndex(child_face_element1);
            const Shape1 *child_face_interior_line=finer_level->
              Level1::getShape(child_face_interior_line_index);
            pair<int,int> p(child_face_interior_line->vertexIndex(0),
                            child_face_interior_line->vertexIndex(1));
            interior_child_line_map[p]=child_face_interior_line_index;
          }
        }
        NumPtr<int> interior_child_line_vertex_indices(
          element->numberCodimension2Shapes());
        for (int ln=0;ln<element->numberCodimension2Shapes();
        ln++) {
          int line_index=element->codimension2ShapeIndex(ln);
          int c(INT_MAX),v(INT_MAX);
          interior_child_line_vertex_indices[ln]=
            level->Level1::interiorChildVertex(line_index,c,v);
        }
//      construct lines interior to element
        for (int l=0;
        l<element->numberInteriorChildCodimension2Shapes();l++) {
          while (finer_level->Level1::used(next_unused_line)) {
            next_unused_line++;
          }
          Shape1 *interior_child_line_element=element->
            newInteriorCodimension2Shape(l,next_unused_line,
              interior_child_vertex_indices,
              interior_child_face_vertex_indices,
              interior_child_line_vertex_indices);
          finer_level->Level1::setShape(next_unused_line,
            interior_child_line_element);
          finer_level->Level1::setUsedFlag(next_unused_line);
          finer_level->Level1::clearUserFlag(next_unused_line);
          finer_level->Level1::clearUserPointer(next_unused_line);
          finer_level->Level1::setChildren(next_unused_line,-1);
          finer_level->Level1::setParent(next_unused_line,-1);
          finer_level->Level1::setBoundaryCondition(next_unused_line,0);
          interior_child_line_map[
            make_pair(interior_child_line_element->vertexIndex(0),
                      interior_child_line_element->vertexIndex(1))]=
            interior_child_line_element->elementDataIndex();
        }
//      construct codimension 1 elements interior to element
        for (int i=0;
        i<element->numberInteriorChildCodimension1Shapes();i++) {
          while (finer_level->Level2::used(next_unused_face)) {
            next_unused_face++;
          }
          Shape2 *interior_child_face_element=element->
            newInteriorCodimension1Shape(i,next_unused_face,
              interior_child_vertex_indices,
              interior_child_face_vertex_indices,
              interior_child_line_vertex_indices,
              interior_child_line_map);
          finer_level->Level2::setShape(
            next_unused_face,interior_child_face_element);
          finer_level->Level2::setUsedFlag(next_unused_face);
          finer_level->Level2::clearUserFlag(next_unused_face);
          finer_level->Level2::clearUserPointer(next_unused_face);
          finer_level->Level2::setChildren(next_unused_face,-1);
          finer_level->Level2::setParent(next_unused_face,-1);
          finer_level->Level2::setBoundaryCondition(next_unused_face,0);
          if (interior_child_face_element->numberVertices()==3) {
            pair<pair<int,int>,pair<int,int> > 
              p(make_pair(interior_child_face_element->vertexIndex(0),
                          interior_child_face_element->vertexIndex(1)),
                make_pair(interior_child_face_element->vertexIndex(2),
                          -1));
            interior_child_face_map[p]=
              interior_child_face_element->elementDataIndex();
          } else {
            pair<pair<int,int>,pair<int,int> > 
              p(make_pair(interior_child_face_element->vertexIndex(0),
                          interior_child_face_element->vertexIndex(1)),
                make_pair(interior_child_face_element->vertexIndex(2),
                          interior_child_face_element->vertexIndex(3)));
            interior_child_face_map[p]=
              interior_child_face_element->elementDataIndex();
          }
#ifdef DEBUG
          for (int f=0;f<interior_child_face_element->
          numberCodimension1Shapes();f++) {
            int face_element_index=interior_child_face_element->
              codimension1ShapeIndex(f);
            const Shape1 *face_element=finer_level->
              Level1::getShape(face_element_index);
            interior_child_face_element->
              checkFaceVertices(f,face_element);
          }
#endif
        }
//      construct child elements
        NumPtr<int> new_element_indices(childrenPerShape());
        MaterialModel *mm=present_level->getMaterialModel();
        int sid=present_level->subdomainId();
        for (int c=0;c<childrenPerShape();c++) {
          while (finer_level->used(next_unused_element)) {
            next_unused_element++;
          }
          if (c==0) level->setChildren(next_unused_element);
          new_element_indices[c]=next_unused_element;
          Shape3 *new_child_element=element->newChildShape(c,
            next_unused_element,interior_child_vertex_indices,
            interior_child_face_vertex_indices,
            interior_child_line_vertex_indices,interior_child_face_map,
            interior_child_line_map);
          finer_level->
            setShape(next_unused_element,new_child_element);
          finer_level->setUsedFlag(next_unused_element);
          finer_level->clearUserFlag(next_unused_element);
          finer_level->clearUserPointer(next_unused_element);
          finer_level->setChildren(next_unused_element,-1);
          finer_level->setParent(next_unused_element,present_index);
          finer_level->setMaterialModel(next_unused_element,mm);
          finer_level->setSubdomainId(next_unused_element,sid);
#ifdef DEBUG
          if (new_child_element->volume(vertices)<=0.) {
            cout << "\telement:" << endl;
            element->printOn(cout);
            for (int i=0;i<element->numberVertices();i++) {
              cout << "\tvertices[" << i << "] = "
                   << vertices[element->vertexIndex(i)] << endl;
            }
            cout << "\tvolume = " << element->volume(vertices) << endl;
            cout << "\tc = " << c << endl;
            cout << "\tchild_element:" << endl;
            new_child_element->printOn(cout);
            for (int i=0;i<new_child_element->numberVertices();
            i++) {
              cout << "\tchild element vertices[" << i << "] = "
                   << vertices[new_child_element->vertexIndex(i)] << endl;
            }
            cout << "\tchild volume = " 
                 << new_child_element->volume(vertices) << endl;
            CHECK_TEST(new_child_element->volume(vertices)>0.);
          }
          for (int f=0;
          f<new_child_element->numberCodimension1Shapes();f++) {
            int face_element_index=
              new_child_element->codimension1ShapeIndex(f);
            const Shape2 *face_element=finer_level->
              Level2::getShape(face_element_index);
            new_child_element->checkFaceVertices(f,face_element);
          }
#endif
        }
//      set neighbors of child faces interior to element
        level->setInteriorChildNeighbors(new_element_indices);
//      set neighbors of child faces on element faces
        for (int f=0;f<element->numberCodimension1Shapes();f++) {
          int neighbor_index=level->neighborIndex(f);
          LEVEL_CLASS *neighbor_level=level->neighborLevel(f);
          if (neighbor_index<0) {
//          parent face has no neighbor:
//            set child neighbors to parent neighbors
            for (int s=0;s<cpcd1e;s++) {
              int c,cf;
              present_level->childFaceOnFaceOf(element,f,s,c,cf);
              finer_level->setNeighbor(new_element_indices[c],cf,0,-1);
            }
          } else {
            CHECK_SAME(neighbor_level,level);
            if (!neighbor_level->hasChildren(neighbor_index)) {
//            parent face has neighbor and neighbor has no children:
//              set child neighbors to parent neighbor            
              for (int s=0;s<cpcd1e;s++) {
                int c,cf;
                present_level->childFaceOnFaceOf(element,f,s,c,cf);
                finer_level->setNeighbor(new_element_indices[c],cf,
                  neighbor_level,neighbor_index);
              }
            } else {
//            parent face has neighbor and neighbor has children:
              int face_element_index=element->codimension1ShapeIndex(f);
              const Shape3 *neighbor_element=
                neighbor_level->getShape(neighbor_index);
              int neighbor_face=
                neighbor_element->faceWithIndexEqualTo(face_element_index);
              LEVEL_CLASS *neighbor_finer_level=
                neighbor_level->finerLevel();
              CHECK_BOUNDS(neighbor_face,0,
                neighbor_element->numberCodimension1Shapes());
//            set neighbors of child faces to neighbor children
              for (int s=0;s<cpcd1e;s++) {
                int c,cf;
                present_level->childFaceOnFaceOf(element,f,s,c,cf);
                int child_element_index=present_level->childIndex(c);
                const Shape3 *child_element=
                  finer_level->getShape(child_element_index);
                int child_face_element_index=
                  child_element->codimension1ShapeIndex(cf);
                bool found=false;
                for (int ns=0;ns<cpcd1e;ns++) {
                  int nc=neighbor_element->childOnFace(neighbor_face,ns);
                  int nci=neighbor_level->childIndex(neighbor_index,nc);
                  const Shape3 *neighbor_child_element=
                    neighbor_finer_level->getShape(nci);
                  for (int ncf=0;
                  ncf<neighbor_child_element->numberCodimension1Shapes();
                  ncf++) {
                    if (neighbor_child_element->
                    codimension1ShapeIndex(ncf)==child_face_element_index){
                      finer_level->setNeighbor(new_element_indices[c],cf,
                        neighbor_finer_level,nci);
                      neighbor_finer_level->setNeighbor(nci,ncf,
                        finer_level,new_element_indices[c]);
                      found=true;
                      break;
                    }
                  }
                  if (found) break;
                }
              }
            }
          }
        }
      }
    }
  }
#ifdef DEBUG
  for (bool doit=firstCodimension2UsedShape();doit;
  doit=nextCodimension2UsedShape()) {
    CHECK_TEST(!present_codimension2_level->userFlagSet());
  }
  for (bool doit=firstCodimension1UsedShape();doit;
  doit=nextCodimension1UsedShape()) {
    CHECK_TEST(!present_codimension1_level->userFlagSet());
  }
  for (bool doit=firstUsedShape();doit;doit=nextUsedShape()) {
    CHECK_TEST(!present_level->refineFlagSet());
  }
  checkNeighbors();
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::executeCoarseningAndRefinement () {
  bool flags_changed=prepareCoarseningAndRefinement();
  executeCoarsening();
  executeRefinement();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::refineGlobal(int times) {
  for (int i=0;i<times;i++) {
    for (bool doit=firstActiveShape();doit;doit=nextActiveShape()) {
      present_level->clearCoarsenFlag();
      present_level->setRefineFlag();
    }
    executeCoarseningAndRefinement();
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::refine(const Vector &criteria,REAL threshold) {
  int n_cells=criteria.size();
#ifdef DEBUG
  CHECK_SAME(n_cells,numberActiveShapes());
  for (int i=0;i<n_cells;i++) ASSERT(criteria[i]>=0.);
#endif
  if (criteria.allZero()) return;

  REAL new_threshold=threshold;
  if (abs(new_threshold)==0.) {
    for (int i=0;i<n_cells;i++) {
      if (criteria[i]>0. && 
      (criteria[i]<new_threshold || abs(new_threshold)==0.)) {
        new_threshold=criteria[i];
      }
    }
  }
  int i=0;
  for (bool doit=firstActiveShape();doit;doit=nextActiveShape(),i++) {
    if (abs(criteria[i])>=new_threshold) {
      presentLevel()->setRefineFlag();
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::coarsen(const Vector &criteria,REAL threshold) {
  int n_cells=criteria.size();
#ifdef DEBUG
  CHECK_SAME(n_cells,numberActiveShapes());
  for (int i=0;i<criteria.size();i++) {
    ASSERT(criteria[i]>=0.);
  }
#endif
  int i=0;
  for (bool doit=firstActiveShape();doit;doit=nextActiveShape(),i++) {
    if (abs(criteria[i])<=threshold) {
      if (!presentLevel()->refineFlagSet()) {
        presentLevel()->setCoarsenFlag();
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::refineAndCoarsenFixedNumber(const Vector &criteria,
REAL top_fraction,REAL bottom_fraction) {
  int n_cells=criteria.size();
#ifdef DEBUG
  CHECK_SAME(n_cells,numberActiveShapes());
  CHECK_TEST(top_fraction>=0. && top_fraction<=1.);
  CHECK_TEST(bottom_fraction>=0. && bottom_fraction<=1.);
  CHECK_TEST(top_fraction+bottom_fraction<=1.);
  for (int i=0;i<criteria.size();i++) {
    ASSERT(criteria[i]>=0.);
  }
#endif
  REAL rn_cells=static_cast<REAL>(n_cells);
  int refine_cells=static_cast<int>(top_fraction*rn_cells);
  int coarsen_cells=static_cast<int>(bottom_fraction*rn_cells);
  if (refine_cells>0 || coarsen_cells>0) {
    NumPtrC<REAL> tmp(n_cells);
    for (int i=0;i<n_cells;i++) tmp[i]=criteria[i];
    tmp.heapSort(); // ordered smallest first
    if (refine_cells) {
      refine(criteria,tmp[n_cells-refine_cells]);
    }
    if (coarsen_cells) {
      coarsen(criteria,tmp[coarsen_cells-1]);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::refineAndCoarsenFixedFraction(const Vector &criteria,
REAL top_fraction,REAL bottom_fraction) {
  int n_cells=criteria.size();
#ifdef DEBUG
  CHECK_SAME(n_cells,numberActiveShapes());
  CHECK_TEST(top_fraction>=0. && top_fraction<=1.);
  CHECK_TEST(bottom_fraction>=0. && bottom_fraction<=1.);
  CHECK_TEST(top_fraction+bottom_fraction<=1.);
  for (int i=0;i<criteria.size();i++) {
    ASSERT(criteria[i]>=0.);
  }
#endif
  REAL total_error=0.;
  for (int i=0;i<n_cells;i++) total_error+=criteria[i];
  top_fraction*=total_error;
  bottom_fraction*=total_error;
  
  NumPtrC<REAL> tmp(n_cells);
  for (int i=0;i<n_cells;i++) tmp[i]=criteria[i];
  tmp.heapSort();
  
  int i=0;
  for (REAL sum=0.;sum<bottom_fraction && i<n_cells;sum+=tmp[i++]);
  REAL bottom_threshold=(i>0 && i<n_cells-1 ? 0.5*(tmp[i]+tmp[i-1]) : 
    (i>0 ? tmp[i-1] : tmp[i]));
  i=n_cells-1;
  for (REAL sum=0.;sum<top_fraction && i>0;sum+=tmp[i--]);
  REAL top_threshold=(i<n_cells-1 && i>0 ? 0.5*(tmp[i]+tmp[i-1]) : tmp[i]);
  if (top_threshold==tmp[n_cells-1] && top_fraction!=total_error) {
    top_threshold*=0.999;
  }
  if (bottom_threshold>=top_threshold) {
    bottom_threshold=0.999*top_threshold;
  }
  if (top_threshold<tmp[n_cells-1]) refine(criteria,top_threshold);
  if (bottom_threshold>tmp[0]) coarsen(criteria,bottom_threshold);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::refineOptimize(const Vector &criteria) {
  int n_cells=criteria.size();
#ifdef DEBUG
  CHECK_SAME(n_cells,numberActiveShapes());
  for (int i=0;i<criteria.size();i++) {
    ASSERT(criteria[i]>=0.);
  }
#endif
  NIDLValueList<int,REAL> criteria_list;
  for (int i=0;i<n_cells;i++) {
    criteria_list.append(OPERATOR_NEW int(i),criteria[i]);
  }
  criteria_list.sort(FALSE);
//The first M cells are refined to minimize the expected error
//multiplied with the expected number of cells. We assume that the 
//error is decreased by 3/4 a_K if the cell K with error indicator a_K
//is refined. The expected number of cells is N+3*M (N is the current 
//number of cells)
  NIDLValueListNode<int,REAL> *node=criteria_list.first();
  REAL s0=0.75*criteria[*(node->getData())];
  REAL E=0.;
  for (int i=0;i<n_cells;i++) E+=criteria[i];
  int N=n_cells;
  int M=0;
  REAL the_min=static_cast<REAL>(3*(1+M)+N)*(E-s0);
  NIDLValueListNode<int,REAL> *min_arg=criteria_list.last();
  for (M++,criteria_list.next(node);M<n_cells;
  M++,criteria_list.next(node)) {
    s0+=0.75*criteria[*(node->getData())];
    REAL err=static_cast<REAL>(3*(1+M)+N)*(E-s0);
    if (err<=the_min) {
      the_min=err;
      min_arg=node;
    }
  }
  refine(criteria,criteria[*(node->getData())]);
  while (!criteria_list.empty()) delete criteria_list.delAfter(0);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::saveUserFlagsForShapes(NumPtr<bool> &v) {
  int i=0;
  for (bool doit=firstShape();doit;doit=nextShape(),i++) {
    v[i]=present_level->userFlagSet();
  }
}

#if (SPACEDIM>1)
void Tessellation::saveUserFlagsForCodimension1Shapes(NumPtr<bool> &v)
{
  int i=0;
  for (bool doit=firstCodimension1Shape();doit;
  doit=nextCodimension1Shape(),i++) {
    v[i]=present_codimension1_level->userFlagSet();
  }
}
#endif

#if (SPACEDIM==3)
void Tessellation::saveUserFlagsForCodimension2Shapes(NumPtr<bool> &v)
{
  int i=0;
  for (bool doit=firstCodimension2Shape();doit;
  doit=nextCodimension2Shape(),i++) {
    v[i]=present_codimension2_level->userFlagSet();
  }
}
#endif

void Tessellation::saveUserFlags(NumPtr<bool> &v) {
  v.cleanup();
  int n=numberShapes();
#if (SPACEDIM==1)
  v.allocate(n);
#endif
#if (SPACEDIM==2)
  int n1=numberCodimension1Shapes();
  v.allocate(n+n1);
#endif
#if (SPACEDIM==3)
  int n1=numberCodimension1Shapes();
  int n2=numberCodimension2Shapes();
  v.allocate(n+n1+n2);
#endif
  NumPtr<bool> p=v;
#if (SPACEDIM==3)
  saveUserFlagsForCodimension2Shapes(p);
  p+=n2;
#endif
#if (SPACEDIM>1)
  saveUserFlagsForCodimension1Shapes(p);
  p+=n1;
#endif
  saveUserFlagsForShapes(p);
}

void Tessellation::loadUserFlagsForShapes(const NumPtr<bool> &v) {
  CHECK_TEST(v.getNumber()>=numberShapes());
  int i=0;
  for (bool doit=firstShape();doit;doit=nextShape(),i++) {
    if (v[i]) present_level->setUserFlag();
    else present_level->clearUserFlag();
  }
}

#if (SPACEDIM>1)
void Tessellation::loadUserFlagsForCodimension1Shapes(
const NumPtr<bool> &v) {
  CHECK_TEST(v.getNumber()>=numberCodimension1Shapes());
  int i=0;
  for (bool doit=firstCodimension1Shape();doit;
  doit=nextCodimension1Shape(),i++) {
    if (v[i]) present_codimension1_level->setUserFlag();
    else present_codimension1_level->clearUserFlag();
  }
}
#endif

#if (SPACEDIM==3)
void Tessellation::loadUserFlagsForCodimension2Shapes(
const NumPtr<bool> &v) {
  CHECK_TEST(v.getNumber()>=numberCodimension2Shapes());
  int i=0;
  for (bool doit=firstCodimension2Shape();doit;
  doit=nextCodimension2Shape(),i++) {
    if (v[i]) present_codimension2_level->setUserFlag();
    else present_codimension2_level->clearUserFlag();
  }
}
#endif

void Tessellation::loadUserFlags(const NumPtr<bool> &v) {
  NumPtr<bool> p=v;
#if (SPACEDIM==3)
  int n2=numberCodimension2Shapes();
  loadUserFlagsForCodimension2Shapes(p);
  p+=n2;
#endif
#if (SPACEDIM>1)
  int n1=numberCodimension1Shapes();
  loadUserFlagsForCodimension1Shapes(p);
  p+=n1;
#endif
  loadUserFlagsForShapes(p);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Tessellation::numberShapes() const {
  int n=0;
  for (LEVEL_CLASS *level=coarsest_level;level;
  level=level->finerLevel()) {
    n+=level->numberShapes();
  }
  return n;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
int Tessellation::numberCodimension1Shapes() const {
  int n=0;
  for (CODIMENSION1_LEVEL_CLASS *level=
  static_cast<CODIMENSION1_LEVEL_CLASS*>(coarsest_level);level;
  level=level->finerLevel()) {
    n+=level->numberShapes();
  }
  return n;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
int Tessellation::numberCodimension2Shapes() const {
  int n=0;
  for (Level1 *level=static_cast<Level1*>(coarsest_level);level;
  level=level->finerLevel()) {
    n+=level->numberShapes();
  }
  return n;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Tessellation::numberLevels() {
  if (coarsest_level==0) return 0;
  LEVEL_CLASS *finest_level=coarsest_level->finestLevel();
  int number_levels=coarsest_level->numberFinerLevels()+1;
  for (bool doit=firstShape(finest_level);doit;
  doit=nextShapeInLevel()) {
    if (present_level->used()) return number_levels;
  }
  OBSOLETE("finest level has no used elements");
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Tessellation::numberUsedVertices() const {
  int n=0;
  for (int i=0;i<vertices_used.getNumber();i++) {
    n+=vertices_used[i];
  }
  return n;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Tessellation::printOn(ostream &os) const {
  os << "Tessellation: mesh_smoothing = " << mesh_smoothing << endl;
  os << "\tboundary:" << endl;
  os << "\tvertices:" << endl;
  for (int i=0;i<vertices.getNumber();i++) {
    os << "\tvertices[" << i << "] = " << vertices[i] << endl;
  }
  for (int i=0;i<vertices_used.getNumber();i++) {
    os << "\tvertices_used[" << i << "] = " << vertices_used[i] << endl;
  }
  os << "\tcoarsest_level = " << coarsest_level << endl;
  if (coarsest_level) coarsest_level->printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Tessellation::printVertices() const {
  for (int i=0;i<vertices.getNumber();i++) {
    cout << "\tvertices[" << i << "] = " << vertices[i];
    cout << (vertices_used[i] ? " is used" : " is not used") << endl;
  }
}
void Tessellation::printUsedVertices() const {
  for (int i=0;i<vertices.getNumber();i++) {
    if (vertices_used[i]) {
      cout << "\tvertices[" << i << "] = " << vertices[i] << endl;
    }
  }
}
void Tessellation::printUnusedVertices() const {
  for (int i=0;i<vertices.getNumber();i++) {
    if (!vertices_used[i]) {
      cout << "\tvertices[" << i << "] = " << vertices[i] << endl;
    }
  }
}
void Tessellation::printElements(LEVEL_CLASS *level) {
  cout << "\n\tlevel = " << level->numberCoarserLevels() << endl;
  for (bool doit=firstShape(level);doit && present_level==level;
  doit=nextShapeInLevel()) {
    const SHAPE_CLASS *element=level->getShape();
    if (element!=0) {
      cout << "\telement[" << element->elementDataIndex() 
           << "], vertices =";
      for (int v=0;v<element->numberVertices();v++) {
        cout << " " << element->vertexIndex(v);
      }
      cout << endl;
    }
  }
}
void Tessellation::printElements() {
  for (LEVEL_CLASS *level=coarsest_level;level;level=level->finerLevel())
  {
    printElements(level);
  }
}
void Tessellation::printActiveElements(LEVEL_CLASS *level) {
  cout << "\n\tlevel = " << level->numberCoarserLevels() << endl;
  for (bool doit=firstActiveShape(level);doit && present_level==level;
  doit=nextActiveShapeInLevel()) {
    const SHAPE_CLASS *element=level->getShape();
    cout << "\telement[" << element->elementDataIndex() 
         << "], vertices =";
    for (int v=0;v<element->numberVertices();v++) {
      cout << " " << element->vertexIndex(v);
    }
    cout << endl;
  }
}
void Tessellation::printActiveElements() {
  for (LEVEL_CLASS *level=coarsest_level;level;level=level->finerLevel())
  {
    printActiveElements(level);
  }
}
#if (SPACEDIM>1)
void Tessellation::printCodimension1Elements(
CODIMENSION1_LEVEL_CLASS *level) {
  cout << "\n\tlevel = " << level->numberCoarserLevels() << endl;
  for (bool doit=firstCodimension1Shape(level);
  doit && presentCodimension1Level()==level;
  doit=nextCodimension1ShapeInLevel()) {
    const CODIMENSION1_SHAPE_CLASS *e=
      presentCodimension1Level()->getShape();
    if (e!=0) {
      cout << "\telement[" << e->elementDataIndex() << "], vertices =";
      for (int v=0;v<e->numberVertices();v++) {
        cout << " " << e->vertexIndex(v);
      }
      cout << " bc = " << level->getBoundaryCondition() << endl;
    }
  }
}
void Tessellation::printCodimension1Elements() {
  for (CODIMENSION1_LEVEL_CLASS *level=
  static_cast<CODIMENSION1_LEVEL_CLASS*>(coarsestLevel());
  level;level=level->finerLevel()) {
    printCodimension1Elements(level);
  }
}
void Tessellation::printActiveCodimension1Elements(
CODIMENSION1_LEVEL_CLASS *level) {
  cout << "\n\tlevel = " << level->numberCoarserLevels() << endl;
  for (bool doit=firstCodimension1ActiveShape(level);
  doit && presentCodimension1Level()==level;
  doit=nextCodimension1ActiveShapeInLevel()) {
    const CODIMENSION1_SHAPE_CLASS *e=
      presentCodimension1Level()->getShape();
    cout << "\telement[" << e->elementDataIndex() << "], vertices =";
    for (int v=0;v<e->numberVertices();v++) {
      cout << " " << e->vertexIndex(v);
    }
    cout << " bc = " << level->getBoundaryCondition() << endl;
  }
}
void Tessellation::printActiveCodimension1Elements() {
  for (CODIMENSION1_LEVEL_CLASS *level=
  static_cast<CODIMENSION1_LEVEL_CLASS*>(coarsestLevel());
  level;level=level->finerLevel()) {
    printActiveCodimension1Elements(level);
  }
}
#endif
#if (SPACEDIM==3)
void Tessellation::printCodimension2Elements(Level1 *level) {
  cout << "\n\tlevel = " << level->numberCoarserLevels() << endl;
  for (bool doit=firstCodimension2Shape(level);
  doit && presentCodimension2Level()==level;
  doit=nextCodimension2ShapeInLevel()) {
    const Shape1 *e=presentCodimension2Level()->getShape();
    if (e!=0) {
      cout << "\telement[" << e->elementDataIndex() << "], vertices =";
      for (int v=0;v<e->numberVertices();v++) {
        cout << " " << e->vertexIndex(v);
      }
      cout << " bc's = " << level->getBoundaryConditions().first
           << " " << level->getBoundaryConditions().second << endl;
    }
  }
}
void Tessellation::printCodimension2Elements() {
  for (Level1 *level=static_cast<Level1*>(coarsestLevel());
  level;level=level->finerLevel()) {
    printCodimension2Elements(level);
  }
}
void Tessellation::printActiveCodimension2Elements(Level1 *level) {
  cout << "\n\tlevel = " << level->numberCoarserLevels() << endl;
  for (bool doit=firstCodimension2ActiveShape(level);
  doit && presentCodimension2Level()==level;
  doit=nextCodimension2ActiveShapeInLevel()) {
    const Shape1 *e=presentCodimension2Level()->getShape();
    cout << "\telement[" << e->elementDataIndex() << "], vertices =";
    for (int v=0;v<e->numberVertices();v++) {
      cout << " " << e->vertexIndex(v);
    }
    cout << " bc's = " << level->getBoundaryConditions().first
         << " " << level->getBoundaryConditions().second << endl;
  }
}
void Tessellation::printActiveCodimension2Elements() {
  for (Level1 *level=static_cast<Level1*>(coarsestLevel());
  level;level=level->finerLevel()) {
    printActiveCodimension2Elements(level);
  }
}
#endif
void Tessellation::printNeighbors(LEVEL_CLASS *level) {
  for (bool doit=firstActiveShape(level);doit && present_level==level;
  doit=nextActiveShapeInLevel()) {
    level->printNeighbors();
  }
}
void Tessellation::printNeighbors() {
  for (bool doit=firstActiveShape();doit;doit=nextActiveShape()) {
    present_level->printNeighbors();
  }
}
void Tessellation::checkNeighbors(LEVEL_CLASS *level) {
  for (bool doit=firstActiveShape(level);doit && present_level==level;
  doit=nextActiveShapeInLevel()) {
    level->checkNeighbors();
  }
}
void Tessellation::checkNeighbors() {
  for (LEVEL_CLASS *level=coarsestLevel();level;
  level=level->finerLevel()) {
    checkNeighbors(level);
  }
}
#if (SPACEDIM>1)
void Tessellation::checkFaceVertices(LEVEL_CLASS *level) {
  level->checkFaceVertices();
}
void Tessellation::checkFaceVertices() {
  for (LEVEL_CLASS *level=coarsestLevel();level;
  level=level->finerLevel()) {
    level->checkFaceVertices();
  }
}
void Tessellation::checkFaceBoundaryCondition(
NumPtr<BoundaryCondition*> &boundary_condition) {
  for (bool doit=firstCodimension1ActiveShape();doit;
  doit=nextCodimension1ActiveShape()) {
    const BoundaryCondition *bc=
      presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      bool found=false;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        if (bc==boundary_condition[i]) {
          const CODIMENSION1_SHAPE_CLASS *e=
            presentCodimension1Level()->getShape();
          const Boundary *b=bc->getBoundary();
          for (int v=0;v<e->numberVertices();v++) {
            int vi=e->vertexIndex(v);
            const Point<SPACEDIM> &ver=vertices[vi];
            REAL con=b->constraint(ver);
            if (abs(con)>=1.e-5) {
              cout << "\tboundary:" << endl;
              b->printOn(cout);
              cout << "\telement:" << endl;
              e->printOn(cout);
              cout << "\tvertex[" << vi << "] = " << ver
                   << " is not on boundary: constraint = " << con << endl;
              CHECK_TEST(abs(con)<1.e-5);
            }
          }
          found=true;
          break;
        }
      }
      if (!found) {
        cout << "\tbogus boundary condition for face "
             << presentCodimension1Level()->presentIndex() << " "
             << bc << endl;
        presentCodimension1Level()->getShape()->printOn(cout);
        CHECK_TEST(found);
      }
    }
  }
}
#if (SPACEDIM==3)
void Tessellation::checkLineBoundaryCondition(
NumPtr<BoundaryCondition*> &boundary_condition) {
  for (bool doit=firstCodimension2ActiveShape();doit;
  doit=nextCodimension2ActiveShape()) {
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
      presentCodimension2Level()->getBoundaryConditions();
    const BoundaryCondition *bc=bcs.first;
    if (bc!=0) {
      const Boundary *b=bc->getBoundary();
      bool found=false;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        if (bc==boundary_condition[i]) {
          const CODIMENSION2_SHAPE_CLASS *e=
            presentCodimension2Level()->getShape();
          for (int v=0;v<e->numberVertices();v++) {
            int vi=e->vertexIndex(v);
            const Point<SPACEDIM> &ver=vertices[vi];
            REAL con=b->constraint(ver);
            if (abs(con)>=1.e-5) {
              cout << "\tbc = " << bc << endl;
              cout << "\tboundary:" << endl;
              b->printOn(cout);
              cout << "\telement:" << endl;
              e->printOn(cout);
              cout << "\tvertex[" << vi << "] = " << ver
                   << " is not on boundary: constraint = " << con << endl;
              CHECK_TEST(abs(con)<1.e-5);
            }
          }
          found=true;
          break;
        }
      }
      if (!found) {
        cout << "\tbogus boundary condition for line "
             << presentCodimension2Level()->presentIndex() << " "
             << bc << endl;
        presentCodimension2Level()->getShape()->printOn(cout);
        CHECK_TEST(found);
      }
      bc=bcs.second;
      if (bc!=0) {
        const Boundary *b=bc->getBoundary();
        bool found=false;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          if (bc==boundary_condition[i]) {
            const CODIMENSION2_SHAPE_CLASS *e=
              presentCodimension2Level()->getShape();
            for (int v=0;v<e->numberVertices();v++) {
              int vi=e->vertexIndex(v);
              const Point<SPACEDIM> &ver=vertices[vi];
              REAL con=b->constraint(ver);
              if (abs(con)>=1.e-5) {
                cout << "\tbc = " << bc << endl;
                cout << "\tboundary:" << endl;
                b->printOn(cout);
                cout << "\telement:" << endl;
                e->printOn(cout);
                cout << "\tvertex[" << vi << "] = " << ver
                     << " is not on boundary: constraint = " << con 
                     << endl;
                CHECK_TEST(abs(con)<1.e-5);
              }
            }
            found=true;
            break;
          }
        }
        if (!found) {
          cout << "\tbogus boundary condition for line "
               << presentCodimension2Level()->presentIndex() << " "
               << bc << endl;
          presentCodimension2Level()->getShape()->printOn(cout);
          CHECK_TEST(found);
        }
      }
    }
  }
}
#endif
void Tessellation::checkNeighborLevelAndIndex(LEVEL_CLASS *level) {
  for (bool doit=firstActiveShape(level);doit && present_level==level;
  doit=nextActiveShapeInLevel()) {
    level->checkNeighborLevelAndIndex(level->presentIndex());
  }
}
void Tessellation::checkNeighborLevelAndIndex() {
  for (LEVEL_CLASS *level=coarsestLevel();level;
  level=level->finerLevel()) {
    checkNeighbors(level);
  }
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include <NISLList.C>
template class NISLListNode<int>;
template ostream& operator<<(ostream&,const NISLListNode<int> &);
template class NISLList<int>;
template ostream& operator<<(ostream&,const NISLList<int> &);

#include <NIDLList.C>
template class NIDLListNode<int>;
template class NIDLList<int>;
template class  NIDLValueListNode<int,REAL>;
template class NIDLValueList<int,REAL>;

#include <NumPtr.C>
template class NumPtrC<REAL>;
INSTANTIATE_NUMPTR(bool)
template class NumPtrWithValue<bool>;
INSTANTIATE_NUMPTR(NumPtr<int>)
INSTANTIATE_NUMPTR(LEVEL_CLASS*)
template class NumPtrBase<pair<Level0*,int>,int>;
template class NumPtr<pair<Level0*,int> >;
INSTANTIATE_NUMPTR( NISLList<int> );
INSTANTIATE_NUMPTR( CellData )
INSTANTIATE_NUMPTR( FaceData )
#if (SPACEDIM==3)
INSTANTIATE_NUMPTR( LineData )
#endif
#if (SPACEDIM>1)
template class _Rb_tree<
    int, 
    pair<int const, NISLList<int> >,
    _Select1st<pair<int const, NISLList<int> > >,
    less<int>, 
    allocator<pair<int const, NISLList<int> > > 
  >;
#endif
