// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Boundary.H"
#include "BoundaryCondition.H"
#include "DataOut.H"
#include "DofHandler.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include "HcurlDataOut.H"
#include "HcurlDofHandler.H"
#include "HcurlFiniteElement.H"
#include "HdivDataOut.H"
#include "HdivDofHandler.H"
#include "HdivFiniteElement.H"
//#include "Mapping.H"
#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tessellation.H"
#include "Tracer.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
class ExactFunctionSystem : public FunctionSystem {
  public: ExactFunctionSystem(int nc=1) : FunctionSystem(nc) { }
  ~ExactFunctionSystem() {}
  REAL value(const Point<SPACEDIM> &p,int c) const { 
    return D_CHOOSE(p[0],p[0]+p[1],p[0]+p[1]+p[2]);
  }
  void value(const Point<SPACEDIM> &p,Vector &v) const {
    if (v.size()!=n_components) v.reinit(n_components,true);
    REAL psum=D_CHOOSE(p[0],p[0]+p[1],p[0]+p[1]+p[2]);
    for (int c=0;c<n_components;c++) v[c]=psum;
  }
  Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int c) const {
    return Tensor<1,SPACEDIM>();
  }
  void gradient(NumPtr<Tensor<1,SPACEDIM> > &g,const Point<SPACEDIM> &p)
  const {
    if (g.getNumber()!=n_components) {
      g.cleanup();
      g.allocate(n_components);
    }
    for (int c=0;c<n_components;c++) g[c].initialize(0.);
  }
//REAL divGrad(const Point<SPACEDIM>&,int) const { return 0.; }
//Vector divGrad(const Point<SPACEDIM>&) const {
//  Vector v(n_components,0.);
//  return v;
//}
};
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void constraintMatrix(Tessellation &tess,const Mapping &mapping,
FiniteElement &fe) {
//TRACER_CALL(t,"constraintMatrix");
  DofHandler dof_handler(tess,fe);
  dof_handler.distributeDofs();
#ifdef DEBUG
//tess.printVertices();
//tess.printElements();
//dof_handler.printDofs();
#endif
  LEVEL_CLASS *level=0;
  for (bool doit=tess.firstActiveShape(level);doit;
  doit=tess.nextActiveShapeInLevel()) {
    if (level==0) level=tess.presentLevel();
    if (tess.presentLevel()!=level) break;
    const SHAPE_CLASS *e=level->getShape();
#ifdef DEBUG
//  cout << "\n\n\n\nnew shape:" << endl;
//  e->printOn(cout);
#endif
    CODIMENSION1_LEVEL_CLASS *face_level=
      dynamic_cast<CODIMENSION1_LEVEL_CLASS*>(level);
    LEVEL_CLASS *finer_level=level->finerLevel();
    for (int f=0;f<e->numberCodimension1Shapes();f++) 
//  int f=1;
    {
      int face_index=e->codimension1ShapeIndex(f);
      if (!face_level->hasChildren(face_index)) continue;
      CODIMENSION1_SHAPE_CLASS *face_shape=
        face_level->getShape(face_index);

      int neighbor_index=level->neighborIndex(e,f);
      const SHAPE_CLASS *neighbor_shape=level->getShape(neighbor_index);
      int nf=neighbor_shape->faceWithIndexEqualTo(face_index);
      CHECK_TEST(level->hasChildren(neighbor_index));
#ifdef DEBUG
//    cout << "\n\n\n\tf = " << f << endl;
//    cout <<"\tface_shape:" << endl;
//    face_shape->printOn(cout);
//    cout << "\tneighbor_index = " << neighbor_index << endl;
//    cout << "\tnf = " << nf << endl;
//    cout <<"\tneighbor_shape:" << endl;
//    neighbor_shape->printOn(cout);
#endif
      NumPtr<int> dofs_on_mother;
      dof_handler.getDofIndices(face_shape,face_level,dofs_on_mother);
      NumPtr<int> dofs_on_children;
      dof_handler.dofsOnChildren(face_shape,face_level,dofs_on_children);
#ifdef DEBUG
//    for (int d=0;d<dofs_on_mother.getNumber();d++) {
//      cout << "\tdofs_on_mother[" << d << "] = " << dofs_on_mother[d]
//           << endl;
//    }
//    for (int d=0;d<dofs_on_children.getNumber();d++) {
//      cout << "\tdofs_on_children[" << d << "] = " 
//           << dofs_on_children[d] << endl;
//    }
#endif

      Matrix m;
      fe.constraints(level,e,face_shape,f,m);
#ifdef DEBUG
//    for (int i=0;i<m.size(0);i++) {
//      cout << "\tm[" << i << ",*] =";
//      for (int j=0;j<m.size(1);j++) cout << " " << m(i,j);
//      cout << endl;
//    }
#endif

      Vector vf(m.size(0));
      Vector vc(m.size(1));
      for (int k=0;k<vc.size();k++) 
      {
//      TRACER_CALL(t,"constraintMatrix loop over vc");
        vc=0.;
        vc[k]=1.;
//      for (int i=0;i<vc.size();i++) vc[i]=drand48();
        m.vmult(vf,vc);
#ifdef DEBUG
//      cout << "\n\n***********\n\n" << endl;
//      for (int i=0;i<vc.size();i++) {
//        cout << "\tvc[" << i << "] = " << vc[i] << endl;
//      }
//      for (int i=0;i<vf.size();i++) {
//        cout << "\tvf[" << i << "] = " << vf[i] << endl;
//      }
#endif
        NumPtr<int> parent_dofs;
        dof_handler.getDofIndices(neighbor_shape,level,parent_dofs);
        Vector parent_dof_values(fe.dofsPerCell(neighbor_shape));
        parent_dof_values=0.;
        for (int d=0;d<parent_dof_values.size();d++) {
          for (int dd=0;dd<dofs_on_mother.getNumber();dd++) {
            if (parent_dofs[d]==dofs_on_mother[dd]) {
              parent_dof_values[d]=vc[dd];
            }
          }
        }
#ifdef DEBUG
//      for (int d=0;d<parent_dofs.getNumber();d++) {
//        cout << "\tparent_dofs[" << d << "] = " << parent_dofs[d] 
//             << endl;
//      }
//      for (int d=0;d<parent_dof_values.size();d++) {
//        cout << "\tparent_dof_values[" << d << "] = " 
//             << parent_dof_values[d] << endl;
//      }
#endif
        LEVEL_CLASS *finer_level=level->finerLevel();
        CODIMENSION1_LEVEL_CLASS *face_child_level=
          face_level->finerLevel();
        for (int s=0;s<childrenPerCodimension1Shape();s++) 
//      int s=1;
        {
//        TRACER_CALL(t,"constraintMatrix loop over s");
          int nc,ncf;
          level->childFaceOnFaceOf(neighbor_shape,nf,s,nc,ncf);
          int neighbor_child_index=level->childIndex(neighbor_index,nc);
          const SHAPE_CLASS *neighbor_child_shape=
            finer_level->getShape(neighbor_child_index);
          int neighbor_child_face_index=
            neighbor_child_shape->codimension1ShapeIndex(ncf);
          const CODIMENSION1_SHAPE_CLASS *neighbor_child_face_shape=
            face_child_level->getShape(neighbor_child_face_index);
          NumPtr<int> child_dofs;
          dof_handler.getDofIndices(neighbor_child_shape,finer_level,
            child_dofs);
#ifdef DEBUG
//        cout << "\n\n" << endl;
//        cout << "\tnc = " << nc << endl;
//        cout << "\tncf = " << ncf << endl;
//        cout << "\tneighbor_child_index = " << neighbor_child_index 
//             << endl;
//        for (int d=0;d<child_dofs.getNumber();d++) {
//          cout << "\tchild_dofs[" << d << "] = " << child_dofs[d] 
//               << endl;
//        }
#endif
          Vector child_dof_values(child_dofs.getNumber());
          child_dof_values=0.;
          for (int d=0;d<child_dofs.getNumber();d++) {
            for (int dd=0;dd<dofs_on_children.getNumber();dd++) {
              if (child_dofs[d]==dofs_on_children[dd]) {
                child_dof_values[d]=vf[dd];
#ifdef DEBUG
//              cout << "\tchild_dof_values[" << d << "] = vf[" << dd 
//                   << "] = " << vf[dd] << endl;
#endif
                break;
              }
            }
            for (int dd=0;dd<dofs_on_mother.getNumber();dd++) {
              if (child_dofs[d]==dofs_on_mother[dd]) {
                child_dof_values[d]=vc[dd];
                break;
              }
            }
          }
#ifdef DEBUG
//        for (int d=0;d<child_dof_values.size();d++) {
//          cout << "\tchild_dof_values[" << d << "] = " 
//               << child_dof_values[d] << endl;
//        }
#endif

//        NumPtr<Point<SPACEDIM-1> > feslp;
//        equallySpacedLatticePointsForFace(neighbor_child_face_shape->
//          numberVertices(),fe.getDegree(),feslp);
//        for (int q=0;q<feslp.getNumber();q++) 
          {
//          TRACER_CALL(t,"constraintMatrix loop over vc loop over feslp");
            NumPtr<Point<SPACEDIM-1> > neighbor_child_face_points(1);
//          for debugging 3D hexahedron, try
//            neighbor_child_face_points[0]=Point<2>(0.,0.);
//            neighbor_child_face_points[0]=Point<2>(1.,0.);
//            neighbor_child_face_points[0]=Point<2>(0.,1.);
//            neighbor_child_face_points[0]=Point<2>(1.,1.);
//            neighbor_child_face_points[0]=Point<2>(0.5,0.);
//            neighbor_child_face_points[0]=Point<2>(0.5,1.);
//            neighbor_child_face_points[0]=Point<2>(0.,0.5);
//            neighbor_child_face_points[0]=Point<2>(1.,0.5);
//            neighbor_child_face_points[0]=Point<2>(0.5,0.5);

            neighbor_child_face_points[0][0]=drand48();
//          neighbor_child_face_points[0][0]=0.5;
#if (SPACEDIM==3)
            neighbor_child_face_points[0][1]=drand48();
            if (neighbor_child_face_shape->numberVertices()==3) {
              neighbor_child_face_points[0][1]*=
                1.-neighbor_child_face_points[0][0];
            }
//          neighbor_child_face_points[0][1]=0.5;
#endif
//          neighbor_child_face_points[0]=feslp[q];
#ifdef DEBUG
//          cout << "\tneighbor_child_face_points[0] = "
//               << neighbor_child_face_points[0] << endl;
#endif
            NumPtr<Point<SPACEDIM> > neighbor_child_points(1);
            neighbor_child_shape->projectToFace(ncf,
              neighbor_child_face_points,neighbor_child_points);
            Quadrature<SPACEDIM> 
              neighbor_child_quad(neighbor_child_points);
            map<int,Quadrature<SPACEDIM> > neighbor_child_qmap;
            neighbor_child_qmap[neighbor_child_shape->numberVertices()]=
              neighbor_child_quad;
            FEValues neighbor_child_fev(mapping,fe,neighbor_child_qmap,
              UPDATE_VALUES | UPDATE_Q_POINTS);
            neighbor_child_fev.mapShapes(tess,neighbor_child_shape,
              finer_level);

            NumPtr<REAL> neighbor_child_values;
            neighbor_child_fev.getFunctionValues(child_dof_values,
              neighbor_child_shape,neighbor_child_values);
#ifdef DEBUG
//          neighbor_child_points ==> neighbor_child_quad
//            ==> neighbor_child_qmap
//            ==> neighbor_child_fev
//            ==> neighbor_child_values, mapped_child_points
//                                   ==> neighbor_points  
//                                   ==> neighbor_quad
//                                   ==> neighbor_qmap
//                                   ==> neighbor_fev
//                                   ==> neighbor_values
//
//          cout << "\tneighbor_child_face_point = " 
//               << neighbor_child_face_points[0] << endl;
//          cout << "\tneighbor_child_point = " 
//               << neighbor_child_points[0] << endl;
//          for (int q=0;q<neighbor_child_values.getNumber();q++) {
//            cout << "\tneighbor_child_values[" << q << "] = " 
//                 << neighbor_child_values[q] << endl;
//          }
#endif

            const NumPtr<Point<SPACEDIM> > &mapped_child_points=
              neighbor_child_fev.getQuadraturePoints(neighbor_child_shape);
            NumPtr<Point<SPACEDIM> > neighbor_lattice_points;
            equallySpacedLatticePoints(neighbor_shape->numberVertices(),
              mapping.getDegree(),neighbor_lattice_points);
            Quadrature<SPACEDIM> 
              neighbor_quadrature(neighbor_lattice_points);
            MappingDataBase *neighbor_md=mapping.getMappingData(
              neighbor_shape->numberVertices(),
              UPDATE_TRANSFORMATION_VALUES | 
              UPDATE_TRANSFORMATION_GRADIENTS,&neighbor_quadrature);
            neighbor_md->computeMappingSupportPoints(neighbor_shape,
              tess,level);
            NumPtr<Point<SPACEDIM> > neighbor_mapped_lattice_points(
              neighbor_lattice_points.getNumber());
            neighbor_md->elementPoints(neighbor_mapped_lattice_points);
            NumPtr<int> neighbor_face_lattice_indices;
            equallySpacedLatticePointIndicesOfFace(
              neighbor_shape->numberVertices(),nf,mapping.getDegree(),
              neighbor_face_lattice_indices);
            NumPtr<Point<SPACEDIM> > 
              neighbor_points(mapped_child_points.getNumber());
            for (int i=0;i<mapped_child_points.getNumber();i++) {
              const Point<SPACEDIM> &mcp=mapped_child_points[i];
              REAL min_dist=HUGE_VAL;
              int min_j=-1;
              for (int j=0;j<neighbor_face_lattice_indices.getNumber();
              j++) {
                REAL dist=mcp.distance(neighbor_mapped_lattice_points[
                  neighbor_face_lattice_indices[j]]);
                if (dist<min_dist) {
                  min_dist=dist;
                  min_j=j;
                }
              }
              neighbor_points[i]=neighbor_lattice_points[min_j];
              mapping.canonicalPoint(neighbor_shape,tess,neighbor_md,
                mapped_child_points[i],neighbor_points[i]);
            }
#ifdef DEBUG
//          for (int q=0;q<mapped_child_points.getNumber();q++) {
//            cout << "\tmapped_child_points[" << q << "] = " 
//                 << mapped_child_points[q] << endl;
//          }
//          for (int q=0;q<neighbor_points.getNumber();q++) {
//            cout << "\tneighbor_points[" << q << "] = " 
//                 << neighbor_points[q] << endl;
//          }
#endif
            Quadrature<SPACEDIM> neighbor_quad(neighbor_points);
            map<int,Quadrature<SPACEDIM> > neighbor_qmap;
            neighbor_qmap[neighbor_shape->numberVertices()]=neighbor_quad;
            FEValues neighbor_fev(mapping,fe,neighbor_qmap,UPDATE_VALUES);
            neighbor_fev.mapShapes(tess,neighbor_shape,level);

            NumPtr<REAL> neighbor_values;
            neighbor_fev.getFunctionValues(parent_dof_values,
              neighbor_shape,neighbor_values);
#ifdef DEBUG
//          for (int q=0;q<neighbor_values.getNumber();q++) {
//            cout << "\tneighbor_values[" << q << "] = " 
//                 << neighbor_values[q] << endl;
//          }
#endif
            for (int q=0;q<neighbor_values.getNumber();q++) {
              REAL error=abs(neighbor_values[q]-neighbor_child_values[q]);
              if (error>1048576.*DBL_EPSILON) {
                cout << "\tf = " << f << endl;
                cout << "\ts = " << s << endl;
                cout << "\tconstraints error[" << q << "] = "
                     << error << " " << error/DBL_EPSILON << endl;
                for (int q2=0;q2<neighbor_values.getNumber();q2++) {
                  cout << "\tneighbor_values,neighbor_child_values["
                       << q2 << "] = " << neighbor_values[q2] << " "
                       << neighbor_child_values[q2] << endl;
                }
                OBSOLETE("fix bugs");
              }
            }
            delete neighbor_md; neighbor_md=0;
          }
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void debugDofHandler(Tessellation &tess,FiniteElement &finite_element) {
  DofHandler dof_handler(tess,finite_element);
  dof_handler.distributeDofs();
  int n_dofs=dof_handler.numberDofs();
//cout << "\tnumberDofs = " << n_dofs << endl;
  int mcbd=dof_handler.maxCouplingsBetweenDofs();
  cout << "\tmaxCouplingsBetweenDofs = " << mcbd << endl;
  Vector solution(n_dofs,0.);

  MappingL mapping(finite_element.getDegree());
  for (int i=0;i<n_dofs;i++) 
//int i=13;
  {
#ifdef DEBUG
//  cout << "\t\ti = " << i << endl;
#endif
    solution=0.;
    solution[i]=1.;
    DataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
    data_out.buildPatches(D_CHOOSE(640,50,10));
//  data_out.buildPatches(D_CHOOSE(640,2,20));
#if (SPACEDIM>=2)
//  data_out.drawGrid();
#endif
//  data_out.plot("solution");
#if (SPACEDIM==2)
//  data_out.contourPlot();
#endif
#if (SPACEDIM>=2)
//  data_out.plotSurface("solution");
#endif
  }

  SparsityPattern sparsity_pattern(n_dofs,n_dofs,mcbd);
  dof_handler.makeSparsityPattern(sparsity_pattern);
  sparsity_pattern.compress();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void debugDofHandler(Tessellation &tess,
HdivFiniteElement &finite_element) {
  HdivDofHandler dof_handler(tess,finite_element);
  dof_handler.distributeDofs();
  int n_dofs=dof_handler.numberDofs();
//cout << "\tnumberDofs = " << n_dofs << endl;
  int mcbd=dof_handler.maxCouplingsBetweenDofs();
  cout << "\tmaxCouplingsBetweenDofs = " << mcbd << endl;
  Vector solution(n_dofs,0.);

  MappingL mapping(finite_element.getDegree());
  for (int i=0;i<n_dofs;i++) 
//int i=13;
  {
#ifdef DEBUG
//  cout << "\t\ti = " << i << endl;
#endif
    solution=0.;
    solution[i]=1.;
    HdivDataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
    data_out.buildPatches(D_CHOOSE(640,50,10));
//  data_out.buildPatches(D_CHOOSE(640,2,20));
#if (SPACEDIM>=2)
//  data_out.drawGrid();
#endif
//  data_out.plot("solution");
#if (SPACEDIM==2)
//  data_out.contourPlot();
#endif
#if (SPACEDIM>=2)
//  data_out.plotSurface("solution");
#endif
  }

  int tsd=dof_handler.totalSolutionDofs();
  SparsityPattern sparsity_pattern_A(tsd,tsd,mcbd);
  SparsityPattern sparsity_pattern_B(dof_handler.numberDofs()-tsd,tsd,
    mcbd);
  dof_handler.makeSparsityPattern(sparsity_pattern_A,
    sparsity_pattern_B);
  sparsity_pattern_A.compress();
  sparsity_pattern_B.compress();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void debugDofHandler(Tessellation &tess,
HcurlFiniteElement &finite_element) {
  HcurlDofHandler dof_handler(tess,finite_element);
  dof_handler.distributeDofs();
  int n_dofs=dof_handler.numberDofs();
//cout << "\tnumberDofs = " << n_dofs << endl;
  int mcbd=dof_handler.maxCouplingsBetweenDofs();
  cout << "\tmaxCouplingsBetweenDofs = " << mcbd << endl;
  Vector solution(n_dofs,0.);

  MappingL mapping(finite_element.getDegree());
  for (int i=0;i<n_dofs;i++) 
//int i=13;
  {
#ifdef DEBUG
//  cout << "\t\ti = " << i << endl;
#endif
    solution=0.;
    solution[i]=1.;
    HcurlDataOut data_out(dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
    data_out.buildPatches(D_CHOOSE(640,50,10));
//  data_out.buildPatches(D_CHOOSE(640,2,20));
#if (SPACEDIM>=2)
//  data_out.drawGrid();
#endif
//  data_out.plot("solution");
#if (SPACEDIM==2)
//  data_out.contourPlot();
#endif
#if (SPACEDIM>=2)
//  data_out.plotSurface("solution");
#endif
  }

  SparsityPattern sparsity_pattern(n_dofs,n_dofs,mcbd);
  dof_handler.makeSparsityPattern(sparsity_pattern);
  sparsity_pattern.compress();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkContinuity(Tessellation &tess,FiniteElement &finite_element) {
//TRACER_CALL(t,"checkContinuity");
#ifdef DEBUG
//tess.printVertices();
//tess.printElements();
#if (SPACEDIM>1)
//tess.printCodimension1Elements();
#endif
//tess.printNeighbors();
#endif
  DofHandler dof_handler(tess,finite_element);
  dof_handler.distributeDofs();
  int n_dofs=dof_handler.numberDofs();
  Vector solution(n_dofs,0.);
  REAL r=drand48();
  for (int i=0;i<n_dofs;i++) solution[i]=drand48();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *level=tess.presentLevel();
    const SHAPE_CLASS *element=tess.presentShape();
#ifdef DEBUG
//  cout << "\n\n***new element" << endl;
//  element->printOn(cout);
#endif
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
//    interior face
      if (face_index>=0) {
        int neighbor_index=level->neighborIndex(f);
//      did not already process this face with an earlier element
        if (neighbor_index>element->elementDataIndex()) {
          SHAPE_CLASS *neighbor_element=level->getShape(neighbor_index);
#ifdef DEBUG
//        cout << "\tf = " << f << endl;
//        cout << "\n***new neighbor" << endl;
//        neighbor_element->printOn(cout);
#endif
//        MappingL mapping(finite_element.getDegree());
          MappingL mapping;
          NumPtr<Point<SPACEDIM> > element_random_canonical_points(1);
          NumPtr<Point<SPACEDIM> > neighbor_random_canonical_points(1);
#if (SPACEDIM==1)
          element_random_canonical_points[0]=(f==0 ? 0. : 1.);
          int shared_vertex=element->sharedVertexIndex(neighbor_element);
          int neighbor_f=
            neighbor_element->faceWithIndexEqualTo(shared_vertex);
          neighbor_random_canonical_points[0]=(neighbor_f==0 ? 0. : 1.);
          NumPtr<Point<SPACEDIM> > vertex(1);
          vertex[0]=tess.getVertices()[shared_vertex];
          int element_v=element->vertexEqualToVertexIndex(shared_vertex);
          int neighbor_element_v=
            neighbor_element->vertexEqualToVertexIndex(shared_vertex);
          NumPtr<Point<SPACEDIM> > element_canonical_points(1);
          element_canonical_points[0]=
            Interval::canonicalVertex(element_v);
          NumPtr<Point<SPACEDIM> > neighbor_canonical_points(1);
          neighbor_canonical_points[0]=
            Interval::canonicalVertex(neighbor_element_v);
# ifdef DEBUG
//        cout << "\tshared_vertex = " << shared_vertex
//             << vertex[0] << endl;
//        cout << "\telement_v = " << element_v
//             << " neighbor_element_v = " << neighbor_element_v << endl;
# endif
#else
          NumPtr<Point<SPACEDIM-1> > random_face_point(1);
          random_face_point[0][0]=drand48();
#  if (SPACEDIM==3)
          random_face_point[0][1]=drand48();
#  endif
          element->projectToFace(f,random_face_point,
            element_random_canonical_points);
          int shared_face=
            element->sharedFaceElementIndex(neighbor_element);
          int neighbor_f=
            neighbor_element->faceWithIndexEqualTo(shared_face);
          NumPtr<Point<SPACEDIM> > element_lattice_points;
          equallySpacedLatticePoints(element->numberVertices(),
            finite_element.getDegree(),element_lattice_points);
          NumPtr<Point<SPACEDIM> > element_points(
            element_lattice_points.getNumber());
          mapping.elementPoints(element,tess,level,
            element_lattice_points,element_points);
# ifdef DEBUG
//        for (int i=0;i<element_lattice_points.getNumber();i++) {
//          cout << "\tlattice_point,element_point[" << i << "] = "
//               << element_lattice_points[i] << " " 
//               << element_points[i] << endl;
//        }
# endif

          NumPtr<Point<SPACEDIM> > neighbor_element_lattice_points;
          equallySpacedLatticePoints(neighbor_element->numberVertices(),
            finite_element.getDegree(),neighbor_element_lattice_points);
          NumPtr<Point<SPACEDIM> > neighbor_element_points(
            neighbor_element_lattice_points.getNumber());
          mapping.elementPoints(neighbor_element,tess,level,
            neighbor_element_lattice_points,neighbor_element_points);
# ifdef DEBUG
//        for (int i=0;i<neighbor_element_lattice_points.getNumber();i++)
//        {
//          cout << "\tneighbor lattice_point,element_point[" << i 
//               << "] = " << neighbor_element_lattice_points[i] << " " 
//               << neighbor_element_points[i] << endl;
//        }
# endif
          NumPtr<int> element_face_lattice_point_indices;
          equallySpacedLatticePointIndicesOfFace(
            element->numberVertices(),f,finite_element.getDegree(),
            element_face_lattice_point_indices);

          NumPtr<Point<SPACEDIM> > element_canonical_points(
            element_face_lattice_point_indices.getNumber());
          NumPtr<Point<SPACEDIM> > neighbor_canonical_points(
            element_face_lattice_point_indices.getNumber());
          Point<SPACEDIM> point(D_DECL(0.,0.,0.));
          Quadrature<SPACEDIM> point_quadrature(point);
          MappingDataBase *md=
            mapping.getMappingData(neighbor_element->numberVertices(),
            UPDATE_Q_POINTS | UPDATE_TRANSFORMATION_VALUES,
            &point_quadrature);
          md->computeMappingSupportPoints(neighbor_element,tess,level);
          Point<SPACEDIM> element_random_mapped_point=
            mapping.elementPoint(element,tess,level,
            element_random_canonical_points[0]);
          REAL min_dist=HUGE_VAL;
          int min_j=-1;
          for (int j=0;j<neighbor_element_points.getNumber();j++) {
            REAL dist=element_random_mapped_point.distance(
              neighbor_element_points[j]);
            if (dist<min_dist) {
              min_j=j;
              min_dist=dist;
            }
          }
          neighbor_random_canonical_points[0]=
            neighbor_element_lattice_points[min_j];
# ifdef DEBUG
//        cout << "\n\telement_random_canonical_point = "
//             << element_random_canonical_points[0] << endl;
//        cout << "\telement_random_mapped_point = "
//             << element_random_mapped_point << endl;
//        cout << "\tneighbor_random_canonical_point = "
//             << neighbor_random_canonical_points[0] << endl;
# endif
          mapping.canonicalPoint(neighbor_element,tess,md,
            element_random_mapped_point,
            neighbor_random_canonical_points[0]);
# ifdef DEBUG
//        cout << "\tneighbor_random_canonical_point = "
//             << neighbor_random_canonical_points[0] << endl;
# endif
          for (int i=0;i<element_face_lattice_point_indices.getNumber();
          i++) {
            int eflpi=element_face_lattice_point_indices[i];
            element_canonical_points[i]=element_lattice_points[eflpi];
            Point<SPACEDIM> face_point=element_points[eflpi];
            REAL min_dist=HUGE_VAL;
            int min_j=-1;
            for (int j=0;j<neighbor_element_points.getNumber();j++) {
              REAL dist=face_point.distance(neighbor_element_points[j]);
              if (dist<min_dist) {
                min_j=j;
                min_dist=dist;
              }
            }
            neighbor_canonical_points[i]=
              neighbor_element_lattice_points[min_j];
            mapping.canonicalPoint(neighbor_element,tess,md,face_point,
              neighbor_canonical_points[i]);
# ifdef DEBUG
//          cout << "\n\telement_canonical_points[" << i << "] = "
//               << element_canonical_points[i] << endl;
//          cout << "\tface_point = " << face_point << endl;
//          cout << "\tneighbor_canonical_points = "
//               << neighbor_canonical_points[i] << endl;
# endif
          }
#endif
#ifdef DEBUG
# if (SPACEDIM>1)
//        cout << "\tshared_face = " << shared_face << endl;
# endif
//        cout << "\tneighbor_f = " << neighbor_f << endl;
#endif
          UPDATE_FLAGS flags=
            UPDATE_VALUES | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;

          Quadrature<SPACEDIM> 
            element_random_quad(element_random_canonical_points);
          Quadrature<SPACEDIM> 
            neighbor_random_quad(neighbor_random_canonical_points);
          map<int,Quadrature<SPACEDIM> > element_random_qmap;
          element_random_qmap[element->numberVertices()]=
            element_random_quad;
          map<int,Quadrature<SPACEDIM> > neighbor_random_qmap;
          neighbor_random_qmap[neighbor_element->numberVertices()]=
            neighbor_random_quad;
          FEValues element_random_fe_values(mapping,finite_element,
            element_random_qmap,flags);
          element_random_fe_values.mapShapes(tess,element,level);
          FEValues neighbor_random_fe_values(mapping,finite_element,
            neighbor_random_qmap,flags);
          neighbor_random_fe_values.mapShapes(tess,neighbor_element,level);

          Quadrature<SPACEDIM> element_quad(element_canonical_points);
          Quadrature<SPACEDIM> neighbor_quad(neighbor_canonical_points);

          map<int,Quadrature<SPACEDIM> > element_qmap;
          element_qmap[element->numberVertices()]=element_quad;
          map<int,Quadrature<SPACEDIM> > neighbor_qmap;
          neighbor_qmap[neighbor_element->numberVertices()]=
            neighbor_quad;

          FEValues element_fe_values(mapping,finite_element,element_qmap,
            flags);
          element_fe_values.mapShapes(tess,element,level);
          FEValues neighbor_fe_values(mapping,finite_element,
            neighbor_qmap,flags);
          neighbor_fe_values.mapShapes(tess,neighbor_element,level);

          NumPtr<int> face_dofs;
#if (SPACEDIM>1)
          CODIMENSION1_SHAPE_CLASS *face_element=
            level->CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          dof_handler.getDofIndices(face_element,level,face_dofs);
#else
          NumPtr<int> element_dofs_1d;
          dof_handler.getDofIndices(element,level,element_dofs_1d);
          face_dofs.allocate(1);
          face_dofs[0]=element_dofs_1d[element_v];
#endif
          NumPtr<int> element_dofs;
          dof_handler.getDofIndices(element,level,element_dofs);
          NumPtr<int> neighbor_dofs;
          dof_handler.getDofIndices(neighbor_element,level,
            neighbor_dofs);
#ifdef DEBUG
//        for (int d=0;d<face_dofs.getNumber();d++) {
//          cout << "\tface_dofs[" << d << "] = " << face_dofs[d] << endl;
//        }
//        for (int d=0;d<element_dofs.getNumber();d++) {
//          cout << "\telement_dofs[" << d << "] = "
//               << element_dofs[d] << endl;
//        }
//        for (int d=0;d<neighbor_dofs.getNumber();d++) {
//          cout << "\tneighbor_dofs[" << d << "] = "
//               << neighbor_dofs[d] << endl;
//        }
#endif
          for (int d=0;d<face_dofs.getNumber();d++) 
//        int d=11;
          {
#ifdef DEBUG
//          cout << "\n\n\tface_dofs[" << d << "] = " << face_dofs[d]
//               << endl;
#endif
            solution=0.;
            solution[face_dofs[d]]=1.;
            NumPtr<REAL> element_random_values;
            NumPtr<REAL> element_values;
            {
//            TRACER_CALL(t,"checkContinuity element_values");
              dof_handler.getFunctionValues(element_random_fe_values,
                element,level,solution,element_random_values);
              dof_handler.getFunctionValues(element_fe_values,element,
                level,solution,element_values);
            }

            NumPtr<REAL> neighbor_random_values;
            NumPtr<REAL> neighbor_values;
            {
//            TRACER_CALL(t,"checkContinuity neighbor_values");
              dof_handler.getFunctionValues(neighbor_random_fe_values,
                neighbor_element,level,solution,neighbor_random_values);
              dof_handler.getFunctionValues(neighbor_fe_values,
                neighbor_element,level,solution,neighbor_values);
            }
#ifdef DEBUG
//          for (int q=0;q<element_canonical_points.getNumber();q++) {
//            cout << "\telement_values,neighbor_values[" << q << "] = "
//                 << element_values[q] << " "
//                 << neighbor_values[q] << endl;
//          }
#endif

            REAL random_error=abs(element_random_values[0]
              -neighbor_random_values[0]);
            REAL error=0;
            REAL size=0.;
            for (int q=0;q<element_canonical_points.getNumber();q++) {
              error+=abs(element_values[q]-neighbor_values[q]);
              size=max(size,max(abs(element_values[q]),
                                abs(neighbor_values[q])));
            }
            if (error>1024.*DBL_EPSILON*size) {
              cout << "\telement = " << element->elementDataIndex()
                << " face = " << f << " neighbor_f = " << neighbor_f
                << " error = " << error/DBL_EPSILON << endl;
              const Array<2,REAL> *element_random_shape_values=
                element_random_fe_values.getShapeValues(element);  
              const Array<2,REAL> *neighbor_random_shape_values=
                neighbor_random_fe_values.getShapeValues(neighbor_element);
              for (int d=0;d<face_dofs.getNumber();d++) {
                int fd=face_dofs[d];
                for (int se=0;se<element_dofs.getNumber();se++) {
                  if (element_dofs[se]==fd) {
                    for (int sn=0;sn<neighbor_dofs.getNumber();sn++) {
                      if (neighbor_dofs[sn]==fd) {
                        cout << "\tshape_value[" << fd << "] = "
                             << (*element_random_shape_values)(se,0) << " "
                             << (*neighbor_random_shape_values)(sn,0)
                             << endl;
                        break; 
                      }
                    }
                    break; 
                  }
                }
              }
              for (int q=0;q<element_canonical_points.getNumber();q++) {
                cout << "\telement_values,neighbor_values[" << q << "] = "
                     << element_values[q] << " "
                     << neighbor_values[q] << endl;
              }
              CHECK_TEST(error<=1024.*DBL_EPSILON*size)
              OBSOLETE(false);
            }
          }
#if (SPACEDIM>1)
          delete md; md=0;
#endif
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkContinuity(Tessellation &tess,
HdivFiniteElement &finite_element) {
//TRACER_CALL(t,"checkContinuity");
#ifdef DEBUG
//tess.printVertices();
//tess.printElements();
# if (SPACEDIM>1)
//tess.printCodimension1Elements();
# endif
//tess.printNeighbors();
#endif
  HdivDofHandler dof_handler(tess,finite_element);
  dof_handler.distributeDofs();
#ifdef DEBUG
//dof_handler.printDofs();
#endif
  int n_dofs=dof_handler.numberDofs();
  Vector solution(n_dofs,0.);
  REAL r=drand48();
  for (int i=0;i<n_dofs;i++) solution[i]=drand48();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *level=tess.presentLevel();
    const SHAPE_CLASS *element=tess.presentShape();
#ifdef DEBUG
//  cout << "\n\n***new element" << endl;
//  element->printOn(cout);
#endif
    for (int f=0;f<element->numberCodimension1Shapes();f++) 
//  int f=4;
    {
      int face_index=element->codimension1ShapeIndex(f);
//    interior face
      if (face_index>=0) {
        int neighbor_index=level->neighborIndex(f);
//      did not already process this face with an earlier element
        if (neighbor_index>element->elementDataIndex()) {
          SHAPE_CLASS *neighbor_element=level->getShape(neighbor_index);
#ifdef DEBUG
//        cout << "\n***new neighbor" << endl;
//        cout << "\tf = " << f << endl;
//        neighbor_element->printOn(cout);
#endif
//        MappingL mapping(finite_element.getDegree());
          MappingL mapping;
          NumPtr<Point<SPACEDIM> > element_random_canonical_points(1);
          NumPtr<Point<SPACEDIM> > neighbor_random_canonical_points(1);
#if (SPACEDIM==1)
          element_random_canonical_points[0]=(f==0 ? 0. : 1.);
          int shared_vertex=element->sharedVertexIndex(neighbor_element);
          int shared_face=
            element->sharedFaceElementIndex(neighbor_element);
          int neighbor_f=
            neighbor_element->faceWithIndexEqualTo(shared_face);
          neighbor_random_canonical_points[0]=(neighbor_f==0 ? 0. : 1.);
          NumPtr<Point<SPACEDIM> > vertex(1);
          vertex[0]=tess.getVertices()[shared_vertex];
          int element_v=element->vertexEqualToVertexIndex(shared_vertex);
          int neighbor_element_v=
            neighbor_element->vertexEqualToVertexIndex(shared_vertex);
          NumPtr<Point<SPACEDIM> > element_canonical_points(1);
          element_canonical_points[0]=
            Interval::canonicalVertex(element_v);
          NumPtr<Point<SPACEDIM> > neighbor_canonical_points(1);
          neighbor_canonical_points[0]=
            Interval::canonicalVertex(neighbor_element_v);
# ifdef DEBUG
//        cout << "\tshared_vertex = " << shared_vertex
//             << vertex[0] << endl;
# endif
#else
          NumPtr<Point<SPACEDIM-1> > random_face_point(1);
          random_face_point[0][0]=drand48();
#  if (SPACEDIM==3)
          random_face_point[0][1]=drand48();
#  endif
          element->projectToFace(f,random_face_point,
            element_random_canonical_points);
          int shared_face=
            element->sharedFaceElementIndex(neighbor_element);
          int neighbor_f=
            neighbor_element->faceWithIndexEqualTo(shared_face);
          Tensor<1,SPACEDIM> element_face_normal=
            faceNormal(element->numberVertices(),f);
          Tensor<1,SPACEDIM> neighbor_face_normal=
            faceNormal(neighbor_element->numberVertices(),neighbor_f);
          NumPtr<Point<SPACEDIM> > element_lattice_points;
          equallySpacedLatticePoints(element->numberVertices(),
            finite_element.getDegree()+1,element_lattice_points);
          NumPtr<Point<SPACEDIM> > element_points(
            element_lattice_points.getNumber());
          mapping.elementPoints(element,tess,level,
            element_lattice_points,element_points);
# ifdef DEBUG
//        cout << "\tneighbor_f = " << neighbor_f << endl;
//        cout << "\telement_face_normal = " << element_face_normal
//             << endl;
//        cout << "\tneighbor_face_normal = " << neighbor_face_normal
//             << endl;
//        for (int i=0;i<element_lattice_points.getNumber();i++) {
//          cout << "\ti = " << i << " element_lattice_points = "
//               << element_lattice_points[i] << " element_points = "
//               << element_points[i] << endl;
//        }
# endif
          NumPtr<Point<SPACEDIM> > neighbor_element_lattice_points;
          equallySpacedLatticePoints(neighbor_element->numberVertices(),
            finite_element.getDegree()+1,neighbor_element_lattice_points);
          NumPtr<Point<SPACEDIM> > neighbor_element_points(
            neighbor_element_lattice_points.getNumber());
          mapping.elementPoints(neighbor_element,tess,level,
            neighbor_element_lattice_points,neighbor_element_points);
# ifdef DEBUG
//        for (int i=0;i<neighbor_element_lattice_points.getNumber();i++)
//        {
//          cout << "\ti = " << i 
//               << " neighbor_element_lattice_points = "
//               << neighbor_element_lattice_points[i] 
//               << " neighbor_element_points = "
//               << neighbor_element_points[i] << endl;
//        }
# endif
          NumPtr<int> element_face_lattice_point_indices;
          equallySpacedLatticePointIndicesOfFace(
            element->numberVertices(),f,finite_element.getDegree()+1,
            element_face_lattice_point_indices);

          NumPtr<Point<SPACEDIM> > element_canonical_points(
            element_face_lattice_point_indices.getNumber());
          NumPtr<Point<SPACEDIM> > neighbor_canonical_points(
            element_face_lattice_point_indices.getNumber());
          Point<SPACEDIM> point(D_DECL(0.,0.,0.));
          Quadrature<SPACEDIM> point_quadrature(point);
          MappingDataBase *md=
            mapping.getMappingData(neighbor_element->numberVertices(),
            UPDATE_Q_POINTS | UPDATE_TRANSFORMATION_VALUES,
            &point_quadrature);
          md->computeMappingSupportPoints(neighbor_element,tess,level);
          Point<SPACEDIM> element_random_mapped_point=
            mapping.elementPoint(element,tess,level,
            element_random_canonical_points[0]);
          REAL min_dist=HUGE_VAL;
          int min_j=-1;
          for (int j=0;j<neighbor_element_points.getNumber();j++) {
            REAL dist=element_random_mapped_point.distance(
              neighbor_element_points[j]);
            if (dist<min_dist) {
              min_j=j;
              min_dist=dist;
            }
          }
          neighbor_random_canonical_points[0]=
            neighbor_element_lattice_points[min_j];
          mapping.canonicalPoint(neighbor_element,tess,md,
            element_random_mapped_point,
            neighbor_random_canonical_points[0]);
          for (int i=0;i<element_face_lattice_point_indices.getNumber();
          i++) {
            int eflpi=element_face_lattice_point_indices[i];
            element_canonical_points[i]=element_lattice_points[eflpi];
            Point<SPACEDIM> face_point=element_points[eflpi];
            REAL min_dist=HUGE_VAL;
            int min_j=-1;
            for (int j=0;j<neighbor_element_points.getNumber();j++) {
              REAL dist=face_point.distance(neighbor_element_points[j]);
              if (dist<min_dist) {
                min_j=j;
                min_dist=dist;
              }
            }
            neighbor_canonical_points[i]=
              neighbor_element_lattice_points[min_j];
# ifdef DEBUG
//          cout << "\tneighbor_canonical_points = "
//               << neighbor_canonical_points[i] << endl;
# endif
//          the following is overkill
            mapping.canonicalPoint(neighbor_element,tess,md,face_point,
              neighbor_canonical_points[i]);
# ifdef DEBUG
//          cout << "\n\telement_canonical_points[" << i << "] = "
//               << element_canonical_points[i] << endl;
//          cout << "\tface_point = " << face_point << endl;
//          cout << "\tneighbor_canonical_points = "
//               << neighbor_canonical_points[i] << endl;
# endif
          }
#endif
          UPDATE_FLAGS flags=
            UPDATE_VALUES | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
          Quadrature<SPACEDIM>
            element_random_quad(element_random_canonical_points);
          Quadrature<SPACEDIM>
            neighbor_random_quad(neighbor_random_canonical_points);
          map<int,Quadrature<SPACEDIM> > element_random_qmap;
          element_random_qmap[element->numberVertices()]=
            element_random_quad;
          map<int,Quadrature<SPACEDIM> > neighbor_random_qmap;
          neighbor_random_qmap[neighbor_element->numberVertices()]=
            neighbor_random_quad;
          HdivFEValues element_random_fe_values(mapping,finite_element,
            element_random_qmap,flags);
          element_random_fe_values.mapShapes(tess,element,level);
          HdivFEValues neighbor_random_fe_values(mapping,finite_element,
            neighbor_random_qmap,flags);
          neighbor_random_fe_values.mapShapes(tess,neighbor_element,level);

          Quadrature<SPACEDIM> element_quad(element_canonical_points);
          Quadrature<SPACEDIM> neighbor_quad(neighbor_canonical_points);

          map<int,Quadrature<SPACEDIM> > element_qmap;
          element_qmap[element->numberVertices()]=element_quad;
          map<int,Quadrature<SPACEDIM> > neighbor_qmap;
          neighbor_qmap[neighbor_element->numberVertices()]=
            neighbor_quad;

          HdivFEValues element_fe_values(mapping,finite_element,
            element_qmap, flags);
          element_fe_values.mapShapes(tess,element,level);

          HdivFEValues neighbor_fe_values(mapping,finite_element,
            neighbor_qmap,flags);
          neighbor_fe_values.mapShapes(tess,neighbor_element,level);

          NumPtr<int> face_dofs;
#if (SPACEDIM>1)
          CODIMENSION1_SHAPE_CLASS *face_element=
            level->CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          dof_handler.getFaceDofIndices(face_element,level,face_dofs);
#else
          NumPtr<int> element_dofs_1d;
          dof_handler.getDofIndices(element,level,element_dofs_1d);
          face_dofs.allocate(1);
          face_dofs[0]=element_dofs_1d[shared_vertex];
#endif
          NumPtr<int> element_dofs;
          dof_handler.getDofIndices(element,level,element_dofs);
          NumPtr<int> neighbor_dofs;
          dof_handler.getDofIndices(neighbor_element,level,
            neighbor_dofs);
#ifdef DEBUG
//        for (int d=0;d<face_dofs.getNumber();d++) {
//          cout << "\tface_dofs[" << d << "] = " << face_dofs[d] << endl;
//        }
# if (SPACEDIM>=2)
/*
//        {
//          const MappingDataBase *element_md=
//            element_fe_values.mappingData(element);
//          for (int q=0;q<element_canonical_points.getNumber();q++) {
//            const Tensor<2,SPACEDIM> &mj=element_md->mappingJacobian(q);
//            cout << "\telement mapping jacobian[" << q << "] = " << endl;
//            for (int i=0;i<SPACEDIM;i++) {
//              cout << "\t\t";
//              for (int j=0;j<SPACEDIM;j++) cout << mj[i][j] << " ";
//              cout << endl;
//            }
//          }
//        }
//        {
//          const MappingDataBase *neighbor_md=
//            neighbor_fe_values.mappingData(neighbor_element);
//          for (int q=0;q<element_canonical_points.getNumber();q++) {
//            const Tensor<2,SPACEDIM> &mj=neighbor_md->mappingJacobian(q);
//            cout << "\tneighbor mapping jacobian[" << q << "] = "
//                 << endl;
//            for (int i=0;i<SPACEDIM;i++) {
//              cout << "\t\t";
//              for (int j=0;j<SPACEDIM;j++) cout << mj[i][j] << " ";
//              cout << endl;
//            }
//          }
//        }
          for (int d=0;d<face_dofs.getNumber();d++) {
            solution=0.;
            solution[face_dofs[d]]=1.;
            NumPtr<Tensor<1,SPACEDIM> > element_values;
            dof_handler.getFunctionValues(element_fe_values,element,
              level,solution,element_values);
            NumPtr<Tensor<1,SPACEDIM> > neighbor_values;
            dof_handler.getFunctionValues(neighbor_fe_values,
              neighbor_element,level,solution,neighbor_values);
            cout << "\n\td = " << d << endl;
            for (int q=0;q<element_canonical_points.getNumber();q++) {
              const MappingDataBase *element_md=
                element_fe_values.mappingData(element);
              Tensor<1,SPACEDIM> element_mapped_normal;
              adjointProduct(element_mapped_normal,
                element_md->mappingJacobian(q),element_face_normal);
              REAL element_mapped_normal_length;
              product(element_mapped_normal_length,element_mapped_normal,
                element_mapped_normal);
              element_mapped_normal/=sqrt(element_mapped_normal_length);
              const MappingDataBase *neighbor_md=
                neighbor_fe_values.mappingData(neighbor_element);
              Tensor<1,SPACEDIM> neighbor_mapped_normal;
              adjointProduct(neighbor_mapped_normal,
                neighbor_md->mappingJacobian(q),neighbor_face_normal);
              REAL neighbor_mapped_normal_length;
              product(neighbor_mapped_normal_length,
                neighbor_mapped_normal,neighbor_mapped_normal);
              neighbor_mapped_normal/=sqrt(neighbor_mapped_normal_length);
              REAL element_dot=element_mapped_normal*element_values[q];
              REAL neighbor_dot=neighbor_mapped_normal*neighbor_values[q];
              cout << "\tdots[" << q << "] = " << element_dot << " "
                   << neighbor_dot << endl;
            }
          }
*/
# endif
#endif
          for (int d=0;d<face_dofs.getNumber();d++) 
//        int d=11;
          {
#ifdef DEBUG
//          cout << "\tface_dofs[" << d << "] = " << face_dofs[d]
//               << endl;
#endif
            solution=0.;
            solution[face_dofs[d]]=1.;
            NumPtr<Tensor<1,SPACEDIM> > element_random_values;
            NumPtr<Tensor<1,SPACEDIM> > element_values;
            {
//            TRACER_CALL(t,"checkContinuity element_values");
              dof_handler.getFunctionValues(element_random_fe_values,
                element,level,solution,element_random_values);
              dof_handler.getFunctionValues(element_fe_values,element,
                level,solution,element_values);
            }
            NumPtr<Tensor<1,SPACEDIM> > neighbor_random_values;
            NumPtr<Tensor<1,SPACEDIM> > neighbor_values;
            {
//            TRACER_CALL(t,"checkContinuity neighbor_values");
              dof_handler.getFunctionValues(neighbor_random_fe_values,
                neighbor_element,level,solution,neighbor_random_values);
              dof_handler.getFunctionValues(neighbor_fe_values,
                neighbor_element,level,solution,neighbor_values);
            }
#ifdef DEBUG
//          for (int q=0;q<element_canonical_points.getNumber();q++) {
//            cout << "\telement_values,neighbor_values[" << q << "] = "
//                 << element_values[q] << " "
//                 << neighbor_values[q] << endl;
//          }
#endif

#if (SPACEDIM==1)
            REAL random_error=
              abs(element_random_values[0][0]
                 +neighbor_random_values[0][0]); ;
            for (int q=0;q<element_canonical_points.getNumber();q++) {
              REAL error=abs(element_values[q][0]-neighbor_values[q][0]);
              REAL size=max(size,max(abs(element_values[q][0]),
                                abs(neighbor_values[q][0])));
              if (error>65536.*DBL_EPSILON*size) {
                cout << "\telement = " << element->elementDataIndex()
                  << " face = " << f << " error = " << error/DBL_EPSILON
                  << endl;
                for (int q=0;q<element_canonical_points.getNumber();q++) {
                  cout << "\telement_values,neighbor_values[" << q
                       << "] = " << element_values[q] << " "
                       << neighbor_values[q] << endl;
                }
                CHECK_TEST(error<=65536.*DBL_EPSILON*size)
                OBSOLETE(false);
              }
            }
#else
            REAL random_error=
              abs(element_face_normal*element_random_values[0]
                 -neighbor_face_normal*neighbor_random_values[0]); ;
            for (int q=0;q<element_canonical_points.getNumber();q++) {
              const MappingDataBase *element_md=
                element_fe_values.mappingData(element);
              Tensor<1,SPACEDIM> element_mapped_normal;
              adjointProduct(element_mapped_normal,
                element_md->mappingJacobian(q),element_face_normal);
              REAL element_mapped_normal_length;
              product(element_mapped_normal_length,element_mapped_normal,
                element_mapped_normal);
              element_mapped_normal/=sqrt(element_mapped_normal_length);
              const MappingDataBase *neighbor_md=
                neighbor_fe_values.mappingData(neighbor_element);
              Tensor<1,SPACEDIM> neighbor_mapped_normal;
              adjointProduct(neighbor_mapped_normal,
                neighbor_md->mappingJacobian(q),neighbor_face_normal);
              REAL neighbor_mapped_normal_length;
              product(neighbor_mapped_normal_length,
                neighbor_mapped_normal,neighbor_mapped_normal);
              neighbor_mapped_normal/=sqrt(neighbor_mapped_normal_length);
              REAL element_dot=element_mapped_normal*element_values[q];
              REAL neighbor_dot=neighbor_mapped_normal*neighbor_values[q];
              REAL error=abs(element_dot+neighbor_dot);
              REAL size=max(size,max(abs(element_dot),abs(neighbor_dot)));
              if (error>65536.*DBL_EPSILON*size) {
                cout << "\telement = " << element->elementDataIndex()
                  << " face = " << f << " error = " << error/DBL_EPSILON
                  << endl;
                const Array<2,Tensor<1,SPACEDIM> > 
                  *element_random_shape_values=
                  element_random_fe_values.getShapeValues(element);
                const Array<2,Tensor<1,SPACEDIM> > 
                  *neighbor_random_shape_values=
                  neighbor_random_fe_values.getShapeValues(
                    neighbor_element);
                for (int d=0;d<face_dofs.getNumber();d++) {
                  int fd=face_dofs[d];
                  for (int se=0;se<element_dofs.getNumber();se++) {
                    if (element_dofs[se]==fd) {
                      for (int sn=0;sn<neighbor_dofs.getNumber();sn++) {
                        if (neighbor_dofs[sn]==fd) {
                          cout << "\tshape_value[" << fd << "] = "
                               << (*element_random_shape_values)(se,0)
                               << " "
                               << (*neighbor_random_shape_values)(sn,0)
                               << endl;
                          break;
                        }
                      }
                      break;
                    }
                  }
                }
                for (int q=0;q<element_canonical_points.getNumber();q++) {
                  cout << "\telement_values,neighbor_values[" << q
                       << "] = " << element_values[q] << " "
                       << neighbor_values[q] << endl;
                }
                cout << "\telement_mapped_normal = "
                     << element_mapped_normal << endl;
                cout << "\tneighbor_mapped_normal = "
                     << neighbor_mapped_normal << endl;
                cout << "\telement_dot, neighbor_dot = "
                     << element_dot << " " << neighbor_dot << endl;
                CHECK_TEST(error<=65536.*DBL_EPSILON*size)
                OBSOLETE(false);
              }
            }
#endif
          }
#if (SPACEDIM>1)
          delete md; md=0;
#endif
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkContinuity(Tessellation &tess,
HcurlFiniteElement &finite_element) {
//TRACER_CALL(t,"checkContinuity");
#ifdef DEBUG
//tess.printVertices();
//tess.printElements();
# if (SPACEDIM>1)
//tess.printCodimension1Elements();
# endif
# if (SPACEDIM==3)
//tess.printCodimension2Elements();
# endif
//tess.printNeighbors();
#endif
  HcurlDofHandler dof_handler(tess,finite_element);
  dof_handler.distributeDofs();
#ifdef DEBUG
//dof_handler.printDofs();
#endif
  int n_dofs=dof_handler.numberDofs();
  Vector solution(n_dofs,0.);
  REAL r=drand48();
  for (int i=0;i<n_dofs;i++) solution[i]=drand48();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *level=tess.presentLevel();
    const SHAPE_CLASS *element=tess.presentShape();
#ifdef DEBUG
//  cout << "\n\n***new element" << endl;
//  element->printOn(cout);
#endif
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      int neighbor_index=level->neighborIndex(f);
//    did not already process this face with an earlier element
      if (neighbor_index>element->elementDataIndex()) {
        SHAPE_CLASS *neighbor_element=level->getShape(neighbor_index);
#ifdef DEBUG
//      cout << "\n***new neighbor for f = " << f << endl;
//      neighbor_element->printOn(cout);
#endif
//      MappingL mapping(finite_element.getDegree());
        MappingL mapping;
        NumPtr<Point<SPACEDIM> > element_random_canonical_points(1);
        NumPtr<Point<SPACEDIM> > neighbor_random_canonical_points(1);
#if (SPACEDIM==1)
        element_random_canonical_points[0]=(f==0 ? 0. : 1.);
        int shared_vertex=element->sharedVertexIndex(neighbor_element);
        int neighbor_f=
          neighbor_element->faceWithIndexEqualTo(shared_vertex);
        neighbor_random_canonical_points[0]=(neighbor_f==0 ? 0. : 1.);
        NumPtr<Point<SPACEDIM> > vertex(1);
        vertex[0]=tess.getVertices()[shared_vertex];
        int element_v=element->vertexEqualToVertexIndex(shared_vertex);
        int neighbor_element_v=
          neighbor_element->vertexEqualToVertexIndex(shared_vertex);
        NumPtr<Point<SPACEDIM> > element_canonical_points(1);
        element_canonical_points[0]=
          Interval::canonicalVertex(element_v);
        NumPtr<Point<SPACEDIM> > neighbor_canonical_points(1);
        neighbor_canonical_points[0]=
          Interval::canonicalVertex(neighbor_element_v);
#else
        NumPtr<Point<SPACEDIM-1> > random_face_point(1);
        random_face_point[0][0]=drand48();
# if (SPACEDIM==3)
        random_face_point[0][1]=drand48();
# endif
        element->projectToFace(f,random_face_point,
          element_random_canonical_points);
        int shared_face=
          element->sharedFaceElementIndex(neighbor_element);
        int neighbor_f=
          neighbor_element->faceWithIndexEqualTo(shared_face);
        Tensor<1,SPACEDIM> element_face_normal=
          faceNormal(element->numberVertices(),f);
//      Tensor<1,SPACEDIM> neighbor_face_normal=
//        faceNormal(neighbor_element->numberVertices(),neighbor_f);
        NumPtr<Point<SPACEDIM> > element_lattice_points;
        equallySpacedLatticePoints(element->numberVertices(),
          max(1,finite_element.getDegree()),element_lattice_points);
        NumPtr<Point<SPACEDIM> > element_points(
          element_lattice_points.getNumber());
        mapping.elementPoints(element,tess,level,
          element_lattice_points,element_points);
# ifdef DEBUG
//      cout << "\tneighbor_f = " << neighbor_f << endl;
//      cout << "\telement_face_normal = " << element_face_normal
//           << endl;
//      for (int i=0;i<element_lattice_points.getNumber();i++) {
//        cout << "\tlattice_point,element_point[" << i << "] = "
//             << element_lattice_points[i] << " " 
//             << element_points[i] << endl;
//      }
# endif
        NumPtr<Point<SPACEDIM> > neighbor_element_lattice_points;
        equallySpacedLatticePoints(neighbor_element->numberVertices(),
          max(1,finite_element.getDegree()),
          neighbor_element_lattice_points);
        NumPtr<Point<SPACEDIM> > neighbor_element_points(
          neighbor_element_lattice_points.getNumber());
        mapping.elementPoints(neighbor_element,tess,level,
          neighbor_element_lattice_points,neighbor_element_points);
# ifdef DEBUG
//      for (int i=0;i<neighbor_element_lattice_points.getNumber();i++)
//      {
//        cout << "\tneighbor lattice_point,element_point[" << i 
//             << "] = " << neighbor_element_lattice_points[i] << " " 
//             << neighbor_element_points[i] << endl;
//      }
# endif
        NumPtr<int> element_face_lattice_point_indices;
        equallySpacedLatticePointIndicesOfFace(
          element->numberVertices(),f,max(1,finite_element.getDegree()),
          element_face_lattice_point_indices);
# ifdef DEBUG
//      for (int i=0;i<element_face_lattice_point_indices.getNumber();i++)
//      {
//        cout << "\telement_face_lattice_point_indices[" << i 
//             << "] = " << element_face_lattice_point_indices[i] << endl;
//      }
# endif
        NumPtr<Point<SPACEDIM> > element_canonical_points(
          element_face_lattice_point_indices.getNumber());
        NumPtr<Point<SPACEDIM> > neighbor_canonical_points(
          element_face_lattice_point_indices.getNumber());
        Point<SPACEDIM> point(D_DECL(0.,0.,0.));
        Quadrature<SPACEDIM> point_quadrature(point);
        MappingDataBase *md=
          mapping.getMappingData(neighbor_element->numberVertices(),
          UPDATE_Q_POINTS | UPDATE_TRANSFORMATION_VALUES,
          &point_quadrature);
        md->computeMappingSupportPoints(neighbor_element,tess,level);
        Point<SPACEDIM> element_random_mapped_point=
          mapping.elementPoint(element,tess,level,
          element_random_canonical_points[0]);
        REAL min_dist=HUGE_VAL;
        int min_j=-1;
        for (int j=0;j<neighbor_element_points.getNumber();j++) {
          REAL dist=element_random_mapped_point.distance(
            neighbor_element_points[j]);
          if (dist<min_dist) {
            min_j=j;
            min_dist=dist;
          }
        }
        neighbor_random_canonical_points[0]=
          neighbor_element_lattice_points[min_j];
        mapping.canonicalPoint(neighbor_element,tess,md,
          element_random_mapped_point,
          neighbor_random_canonical_points[0]);
        for (int i=0;i<element_face_lattice_point_indices.getNumber();
        i++) {
          int eflpi=element_face_lattice_point_indices[i];
          element_canonical_points[i]=element_lattice_points[eflpi];
          Point<SPACEDIM> face_point=element_points[eflpi];
          REAL min_dist=HUGE_VAL;
          int min_j=-1;
          for (int j=0;j<neighbor_element_points.getNumber();j++) {
            REAL dist=face_point.distance(neighbor_element_points[j]);
            if (dist<min_dist) {
              min_j=j;
              min_dist=dist;
            }
          }
          neighbor_canonical_points[i]=
            neighbor_element_lattice_points[min_j];
          mapping.canonicalPoint(neighbor_element,tess,md,face_point,
            neighbor_canonical_points[i]);
# ifdef DEBUG
//        cout << "\n\telement_canonical_points[" << i << "] = "
//             << element_canonical_points[i] << endl;
//        cout << "\tface_point = " << face_point << endl;
//        cout << "\tneighbor_canonical_points = "
//             << neighbor_canonical_points[i] << endl;
# endif
        }
#endif
        UPDATE_FLAGS flags=
          UPDATE_VALUES | UPDATE_JxW_VALUES | UPDATE_Q_POINTS |
          UPDATE_MAPPING_JACOBIAN_INVERSE;
        Quadrature<SPACEDIM>
          element_random_quad(element_random_canonical_points);
        Quadrature<SPACEDIM>
          neighbor_random_quad(neighbor_random_canonical_points);
        map<int,Quadrature<SPACEDIM> > element_random_qmap;
        element_random_qmap[element->numberVertices()]=
          element_random_quad;
        map<int,Quadrature<SPACEDIM> > neighbor_random_qmap;
        neighbor_random_qmap[neighbor_element->numberVertices()]=
          neighbor_random_quad;
        HcurlFEValues element_random_fe_values(mapping,finite_element,
          element_random_qmap,flags);
        element_random_fe_values.mapShapes(tess,element,level);
        HcurlFEValues neighbor_random_fe_values(mapping,finite_element,
          neighbor_random_qmap,flags);
        neighbor_random_fe_values.mapShapes(tess,neighbor_element,level);

        Quadrature<SPACEDIM> element_quad(element_canonical_points);
        Quadrature<SPACEDIM> neighbor_quad(neighbor_canonical_points);

        map<int,Quadrature<SPACEDIM> > element_qmap;
        element_qmap[element->numberVertices()]=element_quad;
        map<int,Quadrature<SPACEDIM> > neighbor_qmap;
        neighbor_qmap[neighbor_element->numberVertices()]=
          neighbor_quad;

        HcurlFEValues element_fe_values(mapping,finite_element,
          element_qmap, flags);
        element_fe_values.mapShapes(tess,element,level);

        HcurlFEValues neighbor_fe_values(mapping,finite_element,
          neighbor_qmap,flags);
        neighbor_fe_values.mapShapes(tess,neighbor_element,level);

        NumPtr<int> face_dofs;
#if (SPACEDIM>1)
        CODIMENSION1_SHAPE_CLASS *face_element=
          level->CODIMENSION1_LEVEL_CLASS::getShape(face_index);
        dof_handler.getFaceDofIndices(face_element,level,face_dofs);
#else
        NumPtr<int> element_dofs_1d;
        dof_handler.getDofIndices(element,level,element_dofs_1d);
        face_dofs.allocate(1);
        face_dofs[0]=element_dofs_1d[element_v];
#endif
        NumPtr<int> element_dofs;
        dof_handler.getDofIndices(element,level,element_dofs);
        NumPtr<int> neighbor_dofs;
        dof_handler.getDofIndices(neighbor_element,level,
          neighbor_dofs);
#ifdef DEBUG
//      element_fe_values.printOn(cout);
//      cout << "\tface_element:" << endl;
//      face_element->printOn(cout);
//      for (int d=0;d<face_dofs.getNumber();d++) {
//        cout << "\tface_dofs[" << d << "] = " << face_dofs[d] << endl;
//      }
#endif
        const MappingDataBase *element_md=
          element_fe_values.mappingData(element);
        const MappingDataBase *neighbor_md=
          neighbor_fe_values.mappingData(neighbor_element);
#ifdef DEBUG
        {
          const HcurlFiniteElementDataBase *element_fedata=
            const_cast<const HcurlFEValues&>(element_fe_values).
            getFEData(element);
          const HcurlFiniteElementDataBase *neighbor_fedata=
            const_cast<const HcurlFEValues&>(neighbor_fe_values).
            getFEData(neighbor_element);
# if (SPACEDIM==3)
//        element_fedata->printOn(cout);
//        int nq=element_fedata->getQuadrature()->numberPoints();
/*
          for (int fl=0;fl<face_element->numberCodimension1Shapes();fl++) {
            int line_index=face_element->codimension1ShapeIndex(fl);
            int element_line=element->lineWithIndexEqualTo(line_index);
            int neighbor_line=
              neighbor_element->lineWithIndexEqualTo(line_index);
            int element_offset=element_line*finite_element.dofsPerLine();
            int neighbor_offset=neighbor_line*finite_element.dofsPerLine();
            Tensor<1,3> element_line_tangent=
              lineTangent(element->numberVertices(),element_line);
            if (element->lineTangentPointsBackward(element_line)) {
              element_line_tangent=-element_line_tangent;
            }
            Tensor<1,3> neighbor_line_tangent=lineTangent(
              neighbor_element->numberVertices(),neighbor_line);
            if (neighbor_element->lineTangentPointsBackward(neighbor_line))
            {
              neighbor_line_tangent=-neighbor_line_tangent;
            }
            cout << "\tfl = " << fl 
                 << " element_line = " << element_line
                 << " neighbor_line = " << neighbor_line
                 << endl;
            for (int i=0;i<finite_element.dofsPerLine();i++) {
              for (int q=0;q<nq;q++) {
                REAL element_dot=element_line_tangent*
                  element_fedata->shapeValue(i+element_offset,q);
                REAL neighbor_dot=neighbor_line_tangent*
                  neighbor_fedata->shapeValue(i+neighbor_offset,q);
                REAL error=abs(element_dot-neighbor_dot);
                if (error>sqrt(DBL_EPSILON)) {
                  cout << "\t\tdots[" << i << "][" << q << "] = "
                       << element_dot << " " << neighbor_dot << endl;
                }
              }
            }
          }
*/
/*
          Tensor<1,3> ft0=element->faceTangent(f,0);
          Tensor<1,3> ft1=element->faceTangent(f,1);
          bool ftwb=element->faceTangentsWindBackward(f);
          Tensor<1,3> nft0=neighbor_element->faceTangent(neighbor_f,0);
          Tensor<1,3> nft1=neighbor_element->faceTangent(neighbor_f,1);
          bool nftwb=
            neighbor_element->faceTangentsWindBackward(neighbor_f);
          int element_offset=0;
          for (int i=0;i<f;i++) {
            element_offset+=
              finite_element.dofsPerFace(element->numberVerticesOnFace(i));
          }
          element_offset+=element->numberCodimension2Shapes()
                         *finite_element.dofsPerLine();
          int neighbor_offset=0;
          for (int i=0;i<neighbor_f;i++) {
            neighbor_offset+=finite_element.
              dofsPerFace(neighbor_element->numberVerticesOnFace(i));
          }
          neighbor_offset+=neighbor_element->numberCodimension2Shapes()
                          *finite_element.dofsPerLine();
          for (int i=0;
          i<finite_element.dofsPerFace(face_element->numberVertices());i++)
          {
            for (int q=0;q<nq;q++) {
              REAL element_dot0=
                ft0*element_fedata->shapeValue(i+element_offset,q);
              REAL element_dot1=
                ft1*element_fedata->shapeValue(i+element_offset,q);
              REAL neighbor_dot0=
                nft0*neighbor_fedata->shapeValue(i+neighbor_offset,q);
              REAL neighbor_dot1=
                nft1*neighbor_fedata->shapeValue(i+neighbor_offset,q);
              REAL error0=abs(element_dot0-neighbor_dot0);
              REAL error1=abs(element_dot1-neighbor_dot1);
              if (max(error0,error1)>sqrt(DBL_EPSILON)) {
                cout << "\telement_dots[" << i << "][" << q << "] = "
                     << element_dot0 << " " << element_dot1 << endl;
                cout << "\tneighbor_dots[" << i << "][" << q << "] = "
                     << neighbor_dot0 << " " << neighbor_dot1 << endl;
              }
            }
          }
*/
# endif
/*
          for (int d=0;d<face_dofs.getNumber();d++) {
            solution=0.;
            solution[face_dofs[d]]=1.;
            NumPtr<Tensor<1,SPACEDIM> > element_values;
            dof_handler.getFunctionValues(element_fe_values,element,
              level,solution,element_values);
            NumPtr<Tensor<1,SPACEDIM> > neighbor_values;
            dof_handler.getFunctionValues(neighbor_fe_values,
              neighbor_element,level,solution,neighbor_values);
            cout << "\n\td = " << d << endl;
            for (int q=0;q<element_canonical_points.getNumber();q++) {
              Tensor<1,SPACEDIM> element_mapped_normal;
              adjointProduct(element_mapped_normal,
                element_md->mappingJacobian(q),element_face_normal);
              element_mapped_normal.normalize();
              const Tensor<1,SPACEDIM> &evq=element_values[q];
              const Tensor<1,SPACEDIM> &nvq=neighbor_values[q];
# if (SPACEDIM==2)
              REAL element_dot=element_mapped_normal[1]*evq[0]
                              -element_mapped_normal[0]*evq[1];
              REAL neighbor_dot=element_mapped_normal[1]*nvq[0]
                               -element_mapped_normal[0]*nvq[1];
              cout << "\tdots[" << q << "] = " 
                   << (abs(element_dot)>sqrt(DBL_EPSILON)?element_dot:0.)
                   << " "
                   << (abs(neighbor_dot)>sqrt(DBL_EPSILON)?neighbor_dot:0.)
                   << endl;
# else
              Tensor<1,SPACEDIM> element_cp;
              crossProduct(element_cp,element_mapped_normal,evq);
              Tensor<1,SPACEDIM> neighbor_cp;
              crossProduct(neighbor_cp,element_mapped_normal,nvq);
              cout << "\tcps[" << q << "] = <";
              for (int i=0;i<3;i++) {
                REAL cpi=element_cp[i];
                cout << (abs(cpi)>sqrt(DBL_EPSILON) ? cpi : 0.) << " ";
              }
              cout << "> <";
              for (int i=0;i<3;i++) {
                REAL cpi=neighbor_cp[i];
                cout << (abs(cpi)>sqrt(DBL_EPSILON) ? cpi : 0.) << " ";
              }
              cout << ">" << endl;
# endif
            }
          }
*/
        }
#endif
        for (int d=0;d<face_dofs.getNumber();d++) 
//      int d=11;
        {
#ifdef DEBUG
//        cout << "\tface_dofs[" << d << "] = " << face_dofs[d]
//             << endl;
#endif
          solution=0.;
          solution[face_dofs[d]]=1.;
          NumPtr<Tensor<1,SPACEDIM> > element_random_values;
          NumPtr<Tensor<1,SPACEDIM> > element_values;
          {
//          TRACER_CALL(t,"checkContinuity element_values");
            dof_handler.getFunctionValues(element_random_fe_values,
              element,level,solution,element_random_values);
            dof_handler.getFunctionValues(element_fe_values,element,
              level,solution,element_values);
          }
          NumPtr<Tensor<1,SPACEDIM> > neighbor_random_values;
          NumPtr<Tensor<1,SPACEDIM> > neighbor_values;
          {
//          TRACER_CALL(t,"checkContinuity neighbor_values");
            dof_handler.getFunctionValues(neighbor_random_fe_values,
              neighbor_element,level,solution,neighbor_random_values);
            dof_handler.getFunctionValues(neighbor_fe_values,
              neighbor_element,level,solution,neighbor_values);
          }
#ifdef DEBUG
//        for (int q=0;q<element_canonical_points.getNumber();q++) {
//          cout << "\telement_values,neighbor_values[" << q << "] = "
//               << element_values[q] << " "
//               << neighbor_values[q] << endl;
//        }
#endif
#if (SPACEDIM==1)
          REAL random_error=abs(element_random_values[0][0]
            -neighbor_random_values[0][0]);
          for (int q=0;q<element_canonical_points.getNumber();q++) {
            REAL error=abs(element_values[q][0]-neighbor_values[q][0]);
            REAL size=max(size,max(abs(element_values[q][0]),
                              abs(neighbor_values[q][0])));
            if (error>65536.*DBL_EPSILON*size) {
              cout << "\telement = " << element->elementDataIndex()
                << " face = " << f << " error = " << error/DBL_EPSILON
                << endl;
              for (int q=0;q<element_canonical_points.getNumber();q++) {
                cout << "\telement_values,neighbor_values[" << q
                     << "] = " << element_values[q] << " "
                     << neighbor_values[q] << endl;
              }
              CHECK_TEST(error<=65536.*DBL_EPSILON*size)
            }
          }
#else
          const MappingDataBase *element_random_md=
            element_random_fe_values.mappingData(element);
          Tensor<1,SPACEDIM> element_mapped_normal;
          adjointProduct(element_mapped_normal,
            element_random_md->mappingJacobian(0),element_face_normal);
          element_mapped_normal.normalize();
# if (SPACEDIM==2)
          const Tensor<1,SPACEDIM> &evq=element_random_values[0];
          const Tensor<1,SPACEDIM> &nvq=neighbor_random_values[0];
          REAL element_dot=element_mapped_normal[1]*evq[0]
                          -element_mapped_normal[0]*evq[1];
          REAL neighbor_dot=element_mapped_normal[1]*nvq[0]
                           -element_mapped_normal[0]*nvq[1];
          REAL random_error=abs(element_dot-neighbor_dot);
# else
          const Tensor<1,SPACEDIM> &evq=element_random_values[0];
          const Tensor<1,SPACEDIM> &nvq=neighbor_random_values[0];
          Tensor<1,SPACEDIM> element_cp;
          crossProduct(element_cp,element_mapped_normal,evq);
          Tensor<1,SPACEDIM> neighbor_cp;
          crossProduct(neighbor_cp,element_mapped_normal,nvq);
          REAL random_error=0.;
          for (int i=0;i<SPACEDIM;i++) {
            random_error+=abs(element_cp[i]-neighbor_cp[i]);
          }
# endif
          for (int q=0;q<element_canonical_points.getNumber();q++) {
            Tensor<1,SPACEDIM> element_mapped_normal;
            adjointProduct(element_mapped_normal,
              element_md->mappingJacobian(q),element_face_normal);
            element_mapped_normal.normalize();
//          Tensor<1,SPACEDIM> neighbor_mapped_normal;
//          adjointProduct(neighbor_mapped_normal,
//            neighbor_md->mappingJacobian(q),neighbor_face_normal);
//          neighbor_mapped_normal.normalize();
# ifdef DEBUG
//          cout << "\telement_mapped_normal = " << element_mapped_normal
//               << endl;
# endif
            const Tensor<1,SPACEDIM> &evq=element_values[q];
            const Tensor<1,SPACEDIM> &nvq=neighbor_values[q];
# if (SPACEDIM==2)
            REAL element_dot=element_mapped_normal[1]*evq[0]
                            -element_mapped_normal[0]*evq[1];
//          if (element->lineTangentPointsBackward(f)) {
//            element_dot=-element_dot;
//          }
            REAL neighbor_dot=element_mapped_normal[1]*nvq[0]
                             -element_mapped_normal[0]*nvq[1];
//          if (neighbor_element->lineTangentPointsBackward(neighbor_f)) {
//            neighbor_dot=-element_dot;
//          }
            REAL error=abs(element_dot-neighbor_dot);
            REAL size=max(size,max(abs(element_dot),abs(neighbor_dot)));
# else
            Tensor<1,SPACEDIM> element_cp;
            crossProduct(element_cp,element_mapped_normal,evq);
            Tensor<1,SPACEDIM> neighbor_cp;
            crossProduct(neighbor_cp,element_mapped_normal,nvq);
            REAL error=0.,size=0.;
            for (int i=0;i<SPACEDIM;i++) {
              error+=abs(element_cp[i]-neighbor_cp[i]);
              size=max(size,max(abs(evq[i]),abs(nvq[i])));
            }
# endif
            if (error>65536.*DBL_EPSILON /* *size */ ) {
              cout << "\telement = " << element->elementDataIndex()
                << " f = " << f << " d = " << d << " q = " << q
                << " error = " << error/DBL_EPSILON
                << endl;
              cout << "\tneighbor = "
                   << neighbor_element->elementDataIndex() << endl;
              cout << "\tface_dofs[" << d << "] = " << face_dofs[d]
                   << endl;
              cout << "\telement_values,neighbor_values[" << q << "] = "
                   << element_values[q] << " "
                   << neighbor_values[q] << endl;
              cout << "\telement_mapped_normal = "
                   << element_mapped_normal << endl;
# if (SPACEDIM==2)
              cout << "\telement_dot, neighbor_dot = "
                   << element_dot << " " << neighbor_dot << endl;
# else
              cout << "\telement_cp, neighbor_cp = "
                   << element_cp << " " << neighbor_cp << endl;
# endif
              const Array<2,Tensor<1,SPACEDIM> >
                *element_random_shape_values=
                element_random_fe_values.getShapeValues(element);
              const Array<2,Tensor<1,SPACEDIM> >
                *neighbor_random_shape_values=
                neighbor_random_fe_values.getShapeValues(neighbor_element);
              for (int d=0;d<face_dofs.getNumber();d++) {
                int fd=face_dofs[d];
                for (int se=0;se<element_dofs.getNumber();se++) {
                  if (element_dofs[se]==fd) {
                    for (int sn=0;sn<neighbor_dofs.getNumber();sn++) {
                      if (neighbor_dofs[sn]==fd) {
                        cout << "\tshape_value[" << fd << "] = "
                             << (*element_random_shape_values)(se,0) << " "
                             << (*neighbor_random_shape_values)(sn,0)
                             << endl;
                        break; 
                      }
                    }
                    break; 
                  }
                }
              }
              for (int q=0;q<element_canonical_points.getNumber();q++) {
                cout << "\telement_values,neighbor_values[" << q << "] = "
                     << element_values[q] << " "
                     << neighbor_values[q] << endl;
              }
              CHECK_TEST(error<=65536.*DBL_EPSILON /* *size */ )
            }
          }
#endif
        }
#if (SPACEDIM==3)
        for (int fl=0;fl<face_element->numberCodimension1Shapes();fl++) {
          int element_line=element->lineFromFace(face_element,f,fl);
          int element_line_index=
            element->codimension2ShapeIndex(element_line);
          Shape1 *edge_element=
            level->Level1::getShape(element_line_index);
          int neighbor_line=
            neighbor_element->lineWithIndexEqualTo(element_line_index);
          NumPtr<int> edge_dofs;
          dof_handler.getLineDofIndices(edge_element,level,edge_dofs);
          Tensor<1,3> element_line_tangent=
            lineTangent(element->numberVertices(),element_line);
//        Tensor<1,3> neighbor_line_tangent=
//          lineTangent(neighbor_element->numberVertices(),neighbor_line);
# ifdef DEBUG
//        cout << "\tfl = " << fl << endl;
//        cout << "\telement_line = " << element_line << endl;
//        cout << "\tneighbor_line = " << neighbor_line << endl;
//        cout << "\telement_line_index = " << element_line_index << endl;
//        cout << "\tedge_element:" << endl;
//        edge_element->printOn(cout);
//        for (int d=0;d<edge_dofs.getNumber();d++) {
//          cout << "\tedge_dofs[" << d << "] = " << edge_dofs[d]
//               << endl;
//        }
/*
          for (int d=0;d<edge_dofs.getNumber();d++) {
            solution=0.;
            solution[edge_dofs[d]]=1.;
            NumPtr<Tensor<1,SPACEDIM> > element_values;
            dof_handler.getFunctionValues(element_fe_values,element,
              level,solution,element_values);
            NumPtr<Tensor<1,SPACEDIM> > neighbor_values;
            dof_handler.getFunctionValues(neighbor_fe_values,
              neighbor_element,level,solution,neighbor_values);
            cout << "\n\td = " << d << endl;
            for (int q=0;q<element_canonical_points.getNumber();q++) {
              Tensor<1,SPACEDIM> element_mapped_tangent;
              product(element_mapped_tangent,
                element_md->mappingJacobian(q),element_line_tangent);
              element_mapped_tangent.normalize();
              const Tensor<1,SPACEDIM> &evq=element_values[q];
              const Tensor<1,SPACEDIM> &nvq=neighbor_values[q];
              REAL element_dot=element_mapped_tangent*evq;
              REAL neighbor_dot=element_mapped_tangent*nvq;
              cout << "\tdots[" << q << "] = " 
                   << (abs(element_dot)>sqrt(DBL_EPSILON)?element_dot:0.)
                   << " "
                   << (abs(neighbor_dot)>sqrt(DBL_EPSILON)?neighbor_dot:0.)
                   << endl;
            }
          }
*/
# endif
          for (int d=0;d<edge_dofs.getNumber();d++)
          {
            solution=0.;
            solution[edge_dofs[d]]=1.;
            NumPtr<Tensor<1,SPACEDIM> > element_values;
            dof_handler.getFunctionValues(element_fe_values,element,
              level,solution,element_values);
            NumPtr<Tensor<1,SPACEDIM> > neighbor_values;
            dof_handler.getFunctionValues(neighbor_fe_values,
              neighbor_element,level,solution,neighbor_values);
# ifdef DEBUG
//          for (int q=0;q<element_canonical_points.getNumber();q++) {
//            cout << "\telement_values,neighbor_values[" << q << "] = "
//                 << element_values[q] << " "
//                 << neighbor_values[q] << endl;
//          }
/*
            cout << "\n\td = " << d << endl;
            for (int q=0;q<element_canonical_points.getNumber();q++) {
              Tensor<1,3> element_mapped_tangent;
              product(element_mapped_tangent,
                element_md->mappingJacobian(q),element_line_tangent);
              element_mapped_tangent.normalize();
              const Tensor<1,SPACEDIM> &evq=element_values[q];
              const Tensor<1,SPACEDIM> &nvq=neighbor_values[q];
              REAL element_dot=element_mapped_tangent*evq;
              REAL neighbor_dot=element_mapped_tangent*nvq;
              cout << "\tdots[" << q << "] = "
                   << (abs(element_dot)>sqrt(DBL_EPSILON)?element_dot:0.)
                   << " "
                   << (abs(neighbor_dot)>sqrt(DBL_EPSILON)?neighbor_dot:0.)
                   << endl;
            }
*/
# endif
            for (int q=0;q<element_canonical_points.getNumber();q++) {
              Tensor<1,3> element_mapped_tangent;
              product(element_mapped_tangent,
                element_md->mappingJacobian(q),element_line_tangent);
              element_mapped_tangent.normalize();
//            Tensor<1,3> neighbor_mapped_tangent;
//            product(neighbor_mapped_tangent,
//              neighbor_md->mappingJacobian(q),neighbor_line_tangent);
//            REAL neighbor_mapped_tangent_length;
//            product(neighbor_mapped_tangent_length,
//              neighbor_mapped_tangent,neighbor_mapped_tangent);
//            neighbor_mapped_tangent/=
//              sqrt(neighbor_mapped_tangent_length);
              const Tensor<1,SPACEDIM> &evq=element_values[q];
              const Tensor<1,SPACEDIM> &nvq=neighbor_values[q];
              REAL element_dot=element_mapped_tangent*evq;
              REAL neighbor_dot=element_mapped_tangent*nvq;
              REAL error=abs(element_dot-neighbor_dot);
//            REAL size=max(abs(element_dot),abs(neighbor_dot));
              if (error>1024.*DBL_EPSILON /* *size */ ) {
                cout << "\telement = " << element->elementDataIndex()
                     << " line = " << element_line << " error = "
                     << error << " " << error/DBL_EPSILON << endl;
                for (int q=0;q<element_canonical_points.getNumber();q++) {
                  cout << "\telement_values,neighbor_values[" << q
                       << "] = " << element_values[q] << " "
                       << neighbor_values[q] << endl;
                }
                cout << "\telement_mapped_tangent = "
                     << element_mapped_tangent << endl;
                cout << "\telement_dot, neighbor_dot = "
                     << element_dot << " " << neighbor_dot << endl;
                CHECK_TEST(error<=65536.*DBL_EPSILON /* *size */ )
              }
            }
          }
        }
#endif
#if (SPACEDIM>1)
        delete md; md=0;
#endif
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif

    Point<SPACEDIM> low(D_DECL(0.,0.,0.));
    Point<SPACEDIM> edge[SPACEDIM];
      edge[0]=Point<SPACEDIM>(D_DECL(1.,0.,0.));
#if (SPACEDIM>1)
      edge[1]=Point<SPACEDIM>(D_DECL(0.,1.,0.));
#endif
#if (SPACEDIM==3)
      edge[2]=Point<SPACEDIM>(0.,0.,1.);
#endif
    bool set_face_ids=true;

    UPDATE_FLAGS update_flags( UPDATE_VALUES | UPDATE_GRADIENTS
      | UPDATE_SECOND_DERIVATIVES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES 
      | UPDATE_NORMAL_VECTORS | UPDATE_MAPPING_JACOBIAN_INVERSE);
#if (SPACEDIM==1)
//
    {
//    two intervals, with random vertex ordering
      TRACER_CALL(t,"testDofHandler checkContinuity two intervals");
      PermutationGenerator pg(3);
      while (pg.hasMore()) 
      {
        const NumPtr<int> &perm=pg.getNext();
//      NumPtr<int> perm(3);
//      perm[0]=0;
//      perm[1]=1;
//      perm[2]=2;
        NumPtr<Point<SPACEDIM> > vertices(3);
        vertices[perm[0]]=Point<SPACEDIM>(0.);
        vertices[perm[1]]=Point<SPACEDIM>(.5);
        vertices[perm[2]]=Point<SPACEDIM>(1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(2);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[1].allocate(2);
        cells[1].vertex(0)=perm[1];
        cells[1].vertex(1)=perm[2];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<1> centroid(.5);
        NumPtr<BoundaryCondition*> boundary_condition(2);
        NumPtr<Boundary*> boundary(2);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]]); // left
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[2]); // right
        for (int f=0;f<2;f++) {
          boundary_condition[f]=
            OPERATOR_NEW BoundaryCondition(boundary[f]);
          tess.coarsestLevel()->
            setBoundaryCondition(f,boundary_condition[f]);
        }
        tess.synchronize(boundary_data);
        cout << "\tpermutation = " << perm[0] << " " << perm[1] << " "
             << perm[2] << endl;
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=10;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
//        TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=10;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=8;degree++) {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=8;degree++) {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=10;degree++) 
//        int degree=0;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=10;degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
#endif
#if (SPACEDIM==2)
//
    {
//    two quadrilaterals, with random vertex ordering
      TRACER_CALL(t,"testDofHandler checkContinuity two quadrilaterals");
      int count=-1;
//    PermutationGenerator pg(6);
//    while (pg.hasMore()) 
      for (int count=0;count<=50;count++)
      {
        count++;
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(6);
        RandomPermutation(perm);
//      perm[0]=5;
//      perm[1]=3;
//      perm[2]=0;
//      perm[3]=1;
//      perm[4]=2;
//      perm[5]=4;
        NumPtr<Point<SPACEDIM> > vertices(6);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(.5,0.);
        vertices[perm[2]]=Point<SPACEDIM>(1.,0.);
        vertices[perm[3]]=Point<SPACEDIM>(0.,1.);
        vertices[perm[4]]=Point<SPACEDIM>(.5,1.);
        vertices[perm[5]]=Point<SPACEDIM>(1.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(4);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[3];
        cells[0].vertex(3)=perm[4];
        cells[1].allocate(4);
        cells[1].vertex(0)=perm[1];
        cells[1].vertex(1)=perm[2];
        cells[1].vertex(2)=perm[4];
        cells[1].vertex(3)=perm[5];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<2> centroid(.5,.5);
        NumPtr<BoundaryCondition*> boundary_condition(4);
        NumPtr<Boundary*> boundary(4);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[3]]); // left
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[2]],
          vertices[perm[5]]); // right
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[2]]); // bottom
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[3]],
          vertices[perm[5]]); // top
        for (int f=0;f<4;f++) {
          boundary_condition[f]=
            OPERATOR_NEW BoundaryCondition(boundary[f]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<2> center=face_element->center(vertices);
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                break;
              }
            }
          }
        }
        tess.synchronize(boundary_data);
        cout << "\tpermutation[" << count << "] = " << perm[0] << " "
             << perm[1] << " " << perm[2] << " " << perm[3] << " "
             << perm[4] << " " << perm[5] << endl;
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
         for (int degree=1;degree<=4;degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=4;degree++) 
//        int degree=0;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
//
    {
//    two triangles, with random vertex ordering
      TRACER_CALL(t,"testDofHandler checkContinuity two triangles");
      int count=-1;
      PermutationGenerator pg(4);
      while (pg.hasMore()) 
      {
        count++;
        const NumPtr<int> &perm=pg.getNext();
//      NumPtr<int> perm(6);
//      perm[0]=0;
//      perm[1]=1;
//      perm[2]=2;
//      perm[3]=3;
        NumPtr<Point<SPACEDIM> > vertices(4);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.);
        vertices[perm[3]]=Point<SPACEDIM>(1.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(3);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[2];
        cells[1].allocate(3);
        cells[1].vertex(0)=perm[1];
        cells[1].vertex(1)=perm[3];
        cells[1].vertex(2)=perm[2];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<2> centroid(.5,.5);
        NumPtr<BoundaryCondition*> boundary_condition(4);
        NumPtr<Boundary*> boundary(4);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[2]]); // left
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[1]],
          vertices[perm[3]]); // right
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[1]]); // bottom
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[2]],
          vertices[perm[3]]); // top
        for (int f=0;f<4;f++) {
          boundary_condition[f]=
            OPERATOR_NEW BoundaryCondition(boundary[f]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<2> center=face_element->center(vertices);
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                break;
              }
            }
          }
        }
        tess.synchronize(boundary_data);
        cout << "\tpermutation[" << count << "] = " << perm[0] << " "
             << perm[1] << " " << perm[2] << " " << perm[3] << endl;
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
//        TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=4;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=4;degree++) 
//        int degree=0;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
//
    {
//    one quadrilateral and one triangle, with random vertex ordering
      TRACER_CALL(t,"testDofHandler checkContinuity one quadrilateral and one triangle");
      int count=-1;
      PermutationGenerator pg(5);
      while (pg.hasMore()) 
      {
        count++;
        const NumPtr<int> &perm=pg.getNext();
//      NumPtr<int> perm(5);
//      RandomPermutation(perm);
//      NumPtr<int> perm(3);
//      perm[0]=0; perm[1]=2; perm[2]=1;
        NumPtr<Point<SPACEDIM> > vertices(5);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,.5);
        vertices[perm[3]]=Point<SPACEDIM>(1.,.5);
        vertices[perm[4]]=Point<SPACEDIM>(.5,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(4);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[2];
        cells[0].vertex(3)=perm[3];
        cells[1].allocate(3);
        cells[1].vertex(0)=perm[2];
        cells[1].vertex(1)=perm[3];
        cells[1].vertex(2)=perm[4];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<2> centroid(.5,.5);
        NumPtr<BoundaryCondition*> boundary_condition(5);
        NumPtr<Boundary*> boundary(5);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[2]]); // left
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[1]],
          vertices[perm[3]]); // right
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[1]]); // bottom
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[2]],
          vertices[perm[4]]); // top left
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[3]],
          vertices[perm[4]]); // top right
        for (int f=0;f<5;f++) {
          boundary_condition[f]=
            OPERATOR_NEW BoundaryCondition(boundary[f]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<2> center=face_element->center(vertices);
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                break;
              }
            }
          }
        }
        tess.synchronize(boundary_data);
        cout << "\tpermutation[" << count << "] = " << perm[0] << " "
             << perm[1] << " " << perm[2] << endl;
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
//        TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=4;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=4;degree++) 
//        int degree=0;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=4;degree++) {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
#endif
#if (SPACEDIM==3)
/*
    {
      TRACER_CALL(t,"testDofHandler checkContinuity two hexahedra");
//    two hexahedra, with random vertex ordering
      int count=-1;
//    PermutationGenerator pg(12);
//    while (pg.hasMore()) 
      for (int count=0;count<500;count++)
      {
//      count++;
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(12);
        RandomPermutation(perm);
        NumPtr<Point<SPACEDIM> > vertices(12);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,0.);
        vertices[perm[3]]=Point<SPACEDIM>(1.,1.,0.);
        vertices[perm[4]]=Point<SPACEDIM>(0.,0.,.5);
        vertices[perm[5]]=Point<SPACEDIM>(1.,0.,.5);
        vertices[perm[6]]=Point<SPACEDIM>(0.,1.,.5);
        vertices[perm[7]]=Point<SPACEDIM>(1.,1.,.5);
        vertices[perm[8]]=Point<SPACEDIM>(0.,0.,1.);
        vertices[perm[9]]=Point<SPACEDIM>(1.,0.,1.);
        vertices[perm[10]]=Point<SPACEDIM>(0.,1.,1.);
        vertices[perm[11]]=Point<SPACEDIM>(1.,1.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(8);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[2];
        cells[0].vertex(3)=perm[3];
        cells[0].vertex(4)=perm[4];
        cells[0].vertex(5)=perm[5];
        cells[0].vertex(6)=perm[6];
        cells[0].vertex(7)=perm[7];
        cells[1].allocate(8);
        cells[1].vertex(0)=perm[4];
        cells[1].vertex(1)=perm[5];
        cells[1].vertex(2)=perm[6];
        cells[1].vertex(3)=perm[7];
        cells[1].vertex(4)=perm[8];
        cells[1].vertex(5)=perm[9];
        cells[1].vertex(6)=perm[10];
        cells[1].vertex(7)=perm[11];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<3> centroid(0.5,0.5,0.5);
        NumPtr<Boundary*> boundary(6);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[2]],vertices[perm[8]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[1]],vertices[perm[3]],vertices[perm[9]]); // front
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[8]]); // left
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[2]],vertices[perm[3]],vertices[perm[10]]);// right
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[2]]); // bottom
        boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[8]],vertices[perm[9]],vertices[perm[10]]); // top
        NumPtr<BoundaryCondition*> boundary_condition(6);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
#ifdef DEBUG
            if (!found) {
              cout << "\tface_element:" << endl;
              face_element->printOn(cout);
              cout << "\tcenter = " << center << endl;
              for (int i=0;i<boundary.getNumber();i++) {
                cout << "\tboundary[" << i << "]: constraint = " 
                     << boundary[i]->constraint(center) << endl;
                boundary[i]->printOn(cout);
              }
            }
#endif
            CHECK_TEST(found);
          }
        }
        tess.synchronize(boundary_data);
#ifdef DEBUG
        cout << "\tpermutation[" << count << "] = " << perm[0] << " " 
             << perm[1] << " " << perm[2] << " " << perm[3] << " " 
             << perm[4] << " " << perm[5] << " " << perm[6] << " " 
             << perm[7] << " " << perm[8] << " " << perm[9] << " " 
             << perm[10] << " " << perm[11] << endl;
        for (bool doit=tess.firstActiveShape();doit;
        doit=tess.nextActiveShape()) {
          LEVEL_CLASS *level=tess.presentLevel();
          const SHAPE_CLASS *element=tess.presentShape();
          for (int f=0;f<element->numberCodimension1Shapes();f++)
          {
            int face_index=element->codimension1ShapeIndex(f);
            if (face_index>=0) {
              CODIMENSION1_SHAPE_CLASS *face_element=
                level->CODIMENSION1_LEVEL_CLASS::getShape(face_index);
              int minv=INT_MAX;
              for (int v=0;v<face_element->numberVertices();v++) {
                minv=min(minv,face_element->vertexIndex(v));
              }
              int neighbor_index=level->neighborIndex(f);
              if (neighbor_index>element->elementDataIndex()) {
                SHAPE_CLASS *neighbor_element=
                  level->getShape(neighbor_index);
                int neighbor_f=neighbor_element->faceWithIndexEqualTo(
                  element->sharedFaceElementIndex(neighbor_element));
                cout << "\t\tfaces = " << f << " " << neighbor_f 
                     << " min vertex on face = " 
                     << element->vertexEqualToVertexIndex(minv) << " "
                     << neighbor_element->vertexEqualToVertexIndex(minv)
                     << endl;
                cout << "\t\tvertices = [ ";
                for (int v=0;v<element->numberVertices();v++) {
                  cout << element->vertexIndex(v) << " ";
                }
                cout << "] [ ";
                for (int v=0;v<neighbor_element->numberVertices();v++) {
                  cout << neighbor_element->vertexIndex(v) << " ";
                }
                cout << " ]" << endl;
                cout << "\t\tvertices on face = [ ";
                for (int v=0;v<face_element->numberVertices();v++) {
                  cout << element->vertexIndex(
                          element->elementVertexFromFace(f,v)) << " ";
                }
                cout << "] [ ";
                for (int v=0;v<face_element->numberVertices();v++) {
                  cout << neighbor_element->vertexIndex(neighbor_element->
                          elementVertexFromFace(neighbor_f,v)) << " ";
                }
                cout << " ]" << endl;
              }
            }
          }
        }
#endif
      }
    }
*/
//
    {
      TRACER_CALL(t,"testDofHandler checkContinuity two hexahedra");
//    two hexahedra, with random vertex ordering
      int count=-1;
//    PermutationGenerator pg(12);
//    while (pg.hasMore()) 
      for (int count=0;count<100;count++)
      {
//      count++;
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(12);
        RandomPermutation(perm);
//      for (int i=0;i<perm.getNumber();i++) perm[i]=i;
//      perm[0]=5;
//      perm[1]=1;
//      perm[2]=9;
//      perm[3]=0;
//      perm[4]=3;
//      perm[5]=6;
//      perm[6]=11;
//      perm[7]=4;
//      perm[8]=10;
//      perm[9]=8;
//      perm[10]=7;
//      perm[11]=2;
        NumPtr<Point<SPACEDIM> > vertices(12);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,0.);
        vertices[perm[3]]=Point<SPACEDIM>(1.,1.,0.);
        vertices[perm[4]]=Point<SPACEDIM>(0.,0.,.5);
        vertices[perm[5]]=Point<SPACEDIM>(1.,0.,.5);
        vertices[perm[6]]=Point<SPACEDIM>(0.,1.,.5);
        vertices[perm[7]]=Point<SPACEDIM>(1.,1.,.5);
        vertices[perm[8]]=Point<SPACEDIM>(0.,0.,1.);
        vertices[perm[9]]=Point<SPACEDIM>(1.,0.,1.);
        vertices[perm[10]]=Point<SPACEDIM>(0.,1.,1.);
        vertices[perm[11]]=Point<SPACEDIM>(1.,1.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(8);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[2];
        cells[0].vertex(3)=perm[3];
        cells[0].vertex(4)=perm[4];
        cells[0].vertex(5)=perm[5];
        cells[0].vertex(6)=perm[6];
        cells[0].vertex(7)=perm[7];
        cells[1].allocate(8);
        cells[1].vertex(0)=perm[4];
        cells[1].vertex(1)=perm[5];
        cells[1].vertex(2)=perm[6];
        cells[1].vertex(3)=perm[7];
        cells[1].vertex(4)=perm[8];
        cells[1].vertex(5)=perm[9];
        cells[1].vertex(6)=perm[10];
        cells[1].vertex(7)=perm[11];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<3> centroid(0.5,0.5,0.5);
        NumPtr<Boundary*> boundary(6);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[2]],vertices[perm[8]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[1]],vertices[perm[3]],vertices[perm[9]]); // front
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[8]]); // left
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[2]],vertices[perm[3]],vertices[perm[10]]);// right
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[2]]); // bottom
        boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[8]],vertices[perm[9]],vertices[perm[10]]); // top
        NumPtr<BoundaryCondition*> boundary_condition(6);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
#ifdef DEBUG
            if (!found) {
              cout << "\tface_element:" << endl;
              face_element->printOn(cout);
              cout << "\tcenter = " << center << endl;
              for (int i=0;i<boundary.getNumber();i++) {
                cout << "\tboundary[" << i << "]: constraint = " 
                     << boundary[i]->constraint(center) << endl;
                boundary[i]->printOn(cout);
              }
            }
#endif
            CHECK_TEST(found);
          }
        }
        tess.synchronize(boundary_data);
#ifdef DEBUG
        cout << "\tpermutation[" << count << "] = " << perm[0] << " " 
             << perm[1] << " " << perm[2] << " " << perm[3] << " " 
             << perm[4] << " " << perm[5] << " " << perm[6] << " " 
             << perm[7] << " " << perm[8] << " " << perm[9] << " " 
             << perm[10] << " " << perm[11] << endl;
#endif
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=4;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=4;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=3;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=3;degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"testDofHandler checkContinuity two prisms sharing triangle face");
//    two prisms, with random vertex ordering
      int count=-1;
//    PermutationGenerator pg(9);
//    while (pg.hasMore()) 
      for (int count=0;count<50;count++)
      {
//      count++;
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(9);
        RandomPermutation(perm);
//      for (int i=0;i<perm.getNumber();i++) perm[i]=i;
//      perm[0]=2;
//      perm[1]=1;
//      perm[2]=0;
//      perm[3]=3;
//      perm[4]=4;
//      perm[5]=8;
//      perm[6]=6;
//      perm[7]=7;
//      perm[8]=5;
        NumPtr<Point<SPACEDIM> > vertices(9);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,0.);
        vertices[perm[3]]=Point<SPACEDIM>(0.,0.,.5);
        vertices[perm[4]]=Point<SPACEDIM>(1.,0.,.5);
        vertices[perm[5]]=Point<SPACEDIM>(0.,1.,.5);
        vertices[perm[6]]=Point<SPACEDIM>(0.,0.,1.);
        vertices[perm[7]]=Point<SPACEDIM>(1.,0.,1.);
        vertices[perm[8]]=Point<SPACEDIM>(0.,1.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(6);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[2];
        cells[0].vertex(3)=perm[3];
        cells[0].vertex(4)=perm[4];
        cells[0].vertex(5)=perm[5];
        cells[1].allocate(6);
        cells[1].vertex(0)=perm[3];
        cells[1].vertex(1)=perm[4];
        cells[1].vertex(2)=perm[5];
        cells[1].vertex(3)=perm[6];
        cells[1].vertex(4)=perm[7];
        cells[1].vertex(5)=perm[8];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<3> centroid(1./3.,1./3.,0.5);
        NumPtr<Boundary*> boundary(5);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[2]],vertices[perm[6]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[1]],vertices[perm[2]],vertices[perm[7]]); // front
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[6]]); // left
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[2]]); // bottom
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[6]],vertices[perm[7]],vertices[perm[8]]); // top
        NumPtr<BoundaryCondition*> boundary_condition(5);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
#ifdef DEBUG
            if (!found) {
              cout << "\tface_element:" << endl;
              face_element->printOn(cout);
              cout << "\tcenter = " << center << endl;
              for (int i=0;i<boundary.getNumber();i++) {
                cout << "\tboundary[" << i << "]: constraint = " 
                     << boundary[i]->constraint(center) << endl;
                boundary[i]->printOn(cout);
              }
            }
#endif
            CHECK_TEST(found);
          }
        }
        tess.synchronize(boundary_data);
#ifdef DEBUG
        cout << "\tpermutation[" << count << "] = " << perm[0] << " " 
             << perm[1] << " " << perm[2] << " " << perm[3] << " " 
             << perm[4] << " " << perm[5] << " " << perm[6] << " " 
             << perm[7] << " " << perm[8] << endl;
#endif
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=3;degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"testDofHandler checkContinuity two prisms quadrilateral face");
//    two prisms, with random vertex ordering
      int count=-1;
//    PermutationGenerator pg(8);
//    while (pg.hasMore()) 
      for (int count=0;count<100;count++)
      {
//      count++;
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(8);
        RandomPermutation(perm);
//      for (int i=0;i<perm.getNumber();i++) perm[i]=i;
//      perm[0]=2;
//      perm[1]=1;
//      perm[2]=0;
//      perm[3]=3;
//      perm[4]=4;
//      perm[5]=8;
//      perm[6]=6;
//      perm[7]=7;
        NumPtr<Point<SPACEDIM> > vertices(8);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,0.);
        vertices[perm[3]]=Point<SPACEDIM>(1.,1.,0.);
        vertices[perm[4]]=Point<SPACEDIM>(0.,0.,1.);
        vertices[perm[5]]=Point<SPACEDIM>(1.,0.,1.);
        vertices[perm[6]]=Point<SPACEDIM>(0.,1.,1.);
        vertices[perm[7]]=Point<SPACEDIM>(1.,1.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(6);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[2];
        cells[0].vertex(3)=perm[4];
        cells[0].vertex(4)=perm[5];
        cells[0].vertex(5)=perm[6];
        cells[1].allocate(6);
        cells[1].vertex(0)=perm[1];
        cells[1].vertex(1)=perm[3];
        cells[1].vertex(2)=perm[2];
        cells[1].vertex(3)=perm[5];
        cells[1].vertex(4)=perm[7];
        cells[1].vertex(5)=perm[6];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<3> centroid(0.5,0.5,0.5);
        NumPtr<Boundary*> boundary(6);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[2]],vertices[perm[4]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[1]],vertices[perm[3]],vertices[perm[5]]); // front
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[4]]); // left
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[2]],vertices[perm[3]],vertices[perm[6]]); // right
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[2]]); // bottom
        boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[4]],vertices[perm[5]],vertices[perm[6]]); // top
        NumPtr<BoundaryCondition*> boundary_condition(6);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
#ifdef DEBUG
            if (!found) {
              cout << "\tface_element:" << endl;
              face_element->printOn(cout);
              cout << "\tcenter = " << center << endl;
              for (int i=0;i<boundary.getNumber();i++) {
                cout << "\tboundary[" << i << "]: constraint = " 
                     << boundary[i]->constraint(center) << endl;
                boundary[i]->printOn(cout);
              }
            }
#endif
            CHECK_TEST(found);
          }
        }
        tess.synchronize(boundary_data);
#ifdef DEBUG
        cout << "\tpermutation[" << count << "] = " << perm[0] << " " 
             << perm[1] << " " << perm[2] << " " << perm[3] << " " 
             << perm[4] << " " << perm[5] << " " << perm[6] << " " 
             << perm[7] << endl;
#endif
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,0);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"testDofHandler checkContinuity two tetrahedra");
//    two tetrahedra, with random vertex ordering
      int count=-1;
//    PermutationGenerator pg(5);
//    while (pg.hasMore()) 
      for (int count=0;count<50;count++)
      {
//      count++;
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(5);
        RandomPermutation(perm);
//      for (int i=0;i<perm.getNumber();i++) perm[i]=i;
//      perm[0]=0;
//      perm[1]=1;
//      perm[2]=3;
//      perm[3]=2;
//      perm[4]=4;
        NumPtr<Point<SPACEDIM> > vertices(5);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(.5,0.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(1.,0.,0.);
        vertices[perm[3]]=Point<SPACEDIM>(.5,1.,0.);
        vertices[perm[4]]=Point<SPACEDIM>(0.,0.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(4);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[3];
        cells[0].vertex(3)=perm[4];
        cells[1].allocate(4);
        cells[1].vertex(0)=perm[1];
        cells[1].vertex(1)=perm[2];
        cells[1].vertex(2)=perm[3];
        cells[1].vertex(3)=perm[4];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<3> centroid(0.5,1./3.,1./3.);
        NumPtr<Boundary*> boundary(4);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[3]],vertices[perm[4]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[2]],vertices[perm[3]],vertices[perm[4]]); // front
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[2]],vertices[perm[4]]); // left
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[2]],vertices[perm[3]]); // bottom
        NumPtr<BoundaryCondition*> boundary_condition(4);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
#ifdef DEBUG
            if (!found) {
              cout << "\tface_element:" << endl;
              face_element->printOn(cout);
              cout << "\tcenter = " << center << endl;
              for (int i=0;i<boundary.getNumber();i++) {
                cout << "\tboundary[" << i << "]: constraint = " 
                     << boundary[i]->constraint(center) << endl;
                boundary[i]->printOn(cout);
              }
            }
#endif
            CHECK_TEST(found);
          }
        }
        tess.synchronize(boundary_data);
#ifdef DEBUG
        cout << "\tpermutation[" << count << "] = " << perm[0] << " " 
             << perm[1] << " " << perm[2] << " " << perm[3] << " " 
             << perm[4] << endl;
#endif
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=2;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"testDofHandler checkContinuity one prism one hexahedron");
//    one prism and one hexahedron, with random vertex ordering
      int count=-1;
//    PermutationGenerator pg(10);
//    while (pg.hasMore()) 
      for (int count=0;count<50;count++)
      {
//      count++;
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(10);
        RandomPermutation(perm);
//      perm[0]=0;
//      perm[1]=2;
//      perm[2]=1;
//      perm[3]=3;
//      perm[4]=4;
//      perm[5]=5;
//      perm[6]=6;
//      perm[7]=7;
//      perm[8]=8;
//      perm[9]=9;
        NumPtr<Point<SPACEDIM> > vertices(10);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,0.);
        vertices[perm[3]]=Point<SPACEDIM>(1.,1.,0.);
        vertices[perm[4]]=Point<SPACEDIM>(0.,0.,.5);
        vertices[perm[5]]=Point<SPACEDIM>(1.,0.,.5);
        vertices[perm[6]]=Point<SPACEDIM>(0.,1.,.5);
        vertices[perm[7]]=Point<SPACEDIM>(1.,1.,.5);
        vertices[perm[8]]=Point<SPACEDIM>(0.,0.,1.);
        vertices[perm[9]]=Point<SPACEDIM>(0.,1.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(8);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[2];
        cells[0].vertex(3)=perm[3];
        cells[0].vertex(4)=perm[4];
        cells[0].vertex(5)=perm[5];
        cells[0].vertex(6)=perm[6];
        cells[0].vertex(7)=perm[7];
        cells[1].allocate(6);
        cells[1].vertex(0)=perm[4];
        cells[1].vertex(1)=perm[8];
        cells[1].vertex(2)=perm[5];
        cells[1].vertex(3)=perm[6];
        cells[1].vertex(4)=perm[9];
        cells[1].vertex(5)=perm[7];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<3> centroid(0.5,0.5,0.5);
        NumPtr<Boundary*> boundary(6);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[2]],vertices[perm[8]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[1]],
          vertices[perm[3]],vertices[perm[5]]); // front
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[1]],vertices[perm[8]]); // left
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[2]],vertices[perm[3]],vertices[perm[9]]);// right
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[0]],vertices[perm[1]],vertices[perm[2]]); // bottom
        boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[8]],vertices[perm[5]],vertices[perm[9]]); // top
        NumPtr<BoundaryCondition*> boundary_condition(6);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
#ifdef DEBUG
            if (!found) {
              cout << "\tface_element:" << endl;
              face_element->printOn(cout);
              cout << "\tcenter = " << center << endl;
              for (int i=0;i<boundary.getNumber();i++) {
                cout << "\tboundary[" << i << "]: constraint = " 
                     << boundary[i]->constraint(center) << endl;
                boundary[i]->printOn(cout);
              }
            }
#endif
            CHECK_TEST(found);
          }
        }
        tess.synchronize(boundary_data);
#ifdef DEBUG
        cout << "\tpermutation[" << count << "] = " << perm[0] << " " 
             << perm[1] << " " << perm[2] << " " << perm[3] << " " 
             << perm[4] << " " << perm[5] << endl;
#endif
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=0;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
//
    {
      TRACER_CALL(t,"testDofHandler checkContinuity one prism one tetrahedron");
//    one prism and one tetrahedron, with random vertex ordering
      int count=-1;
//    PermutationGenerator pg(7);
//    while (pg.hasMore()) 
      for (int count=0;count<50;count++)
      {
        count++;
//      const NumPtr<int> &perm=pg.getNext();
        NumPtr<int> perm(7);
        RandomPermutation(perm);
//      perm[0]=0;
//      perm[1]=2;
//      perm[2]=1;
//      perm[3]=3;
//      perm[4]=4;
//      perm[5]=5;
//      perm[6]=6;
        NumPtr<Point<SPACEDIM> > vertices(7);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,0.);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,0.);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,0.);
        vertices[perm[3]]=Point<SPACEDIM>(0.,0.,.5);
        vertices[perm[4]]=Point<SPACEDIM>(1.,0.,.5);
        vertices[perm[5]]=Point<SPACEDIM>(0.,1.,.5);
        vertices[perm[6]]=Point<SPACEDIM>(0.,0.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(6);
        cells[0].vertex(0)=perm[0];
        cells[0].vertex(1)=perm[1];
        cells[0].vertex(2)=perm[2];
        cells[0].vertex(3)=perm[3];
        cells[0].vertex(4)=perm[4];
        cells[0].vertex(5)=perm[5];
        cells[1].allocate(4);
        cells[1].vertex(0)=perm[3];
        cells[1].vertex(1)=perm[4];
        cells[1].vertex(2)=perm[5];
        cells[1].vertex(3)=perm[6];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<3> centroid(1./3.,1./3.,0.5);
        NumPtr<Boundary*> boundary(5);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[2]],vertices[perm[6]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[1]],
          vertices[perm[2]],vertices[perm[4]]); // front
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[1]],vertices[perm[6]]); // left
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[perm[0]],
          vertices[perm[1]],vertices[perm[2]]); // bottom
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[4]],vertices[perm[5]],vertices[perm[6]]); // top
        NumPtr<BoundaryCondition*> boundary_condition(5);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                  setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
#ifdef DEBUG
            if (!found) {
              cout << "\tface_element:" << endl;
              face_element->printOn(cout);
              cout << "\tcenter = " << center << endl;
              for (int i=0;i<boundary.getNumber();i++) {
                cout << "\tboundary[" << i << "]: constraint = " 
                     << boundary[i]->constraint(center) << endl;
                boundary[i]->printOn(cout);
              }
            }
#endif
            CHECK_TEST(found);
          }
        }
        tess.synchronize(boundary_data);
#ifdef DEBUG
        cout << "\tpermutation[" << count << "] = " << perm[0] << " " 
             << perm[1] << " " << perm[2] << " " << perm[3] << " " 
             << perm[4] << " " << perm[5] << " " << perm[6] << endl;
#endif
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=0;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=D_CHOOSE(6,6,1);degree++) 
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
//
#endif

/*
    {
      TRACER_CALL(t,"testDofHandler generateParallelpiped");
      Tessellation tess;
      int number_elements[SPACEDIM];
      for (int d=0;d<SPACEDIM;d++) number_elements[d]=2;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
        number_elements,set_face_ids,boundary,boundary_condition);
//
      {
        TRACER_CALL(t,"testDofHandler FE_L");
        for (int degree=1;degree<=6;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler FE_H");
        for (int degree=1;degree<=6;degree++) 
//      int degree=2;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler RTFiniteElement");
        for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          RTFiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
        for (int degree=0;degree<=D_CHOOSE(5,5,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          BDFMFiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
#if (SPACEDIM>=2)
//
      {
        TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
        for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec1FiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
        for (int degree=1;degree<=D_CHOOSE(6,6,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec2FiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
#endif
      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
#if (SPACEDIM==2)
/*
    {
      TRACER_CALL(t,"testDofHandler generateParallelpipedTriangles");
      Tessellation tess;
      int number_elements[SPACEDIM];
      for (int d=0;d<SPACEDIM;d++) number_elements[d]=1;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *boundary_data=generateParallelpipedTriangles(tess,
        low,edge,number_elements,set_face_ids,boundary,
        boundary_condition);
//    Point<SPACEDIM> center(D_DECL(0.5,0.5,0.5));
//    PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(center);
//    tess.setBoundary(0,sb);

//
      {
        TRACER_CALL(t,"testDofHandler FE_L");
        for (int degree=1;degree<=10;degree++) 
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler FE_H");
        for (int degree=1;degree<=10;degree++) 
//      int degree=3;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler RTFiniteElement");
        for (int degree=0;degree<=8;degree++) 
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          RTFiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
//      for (int degree=0;degree<=8;degree++) 
        int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          BDFMFiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
        for (int degree=0;degree<=8;degree++) 
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          Nedelec1FiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
        for (int degree=1;degree<=8;degree++) 
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          Nedelec2FiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//

      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
/*
    {
      TRACER_CALL(t,"testDofHandler constraintMatrix one triangle");
      NumPtr<Point<SPACEDIM> > vertices(3);
      vertices[0]=Point<SPACEDIM>(0.,0.);
      vertices[1]=Point<SPACEDIM>(1.,0.);
      vertices[2]=Point<SPACEDIM>(0.,1.);
      NumPtr<CellData> cells(1);
      cells[0].setMaterialModel(0);
      cells[0].allocate(3);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;

      Tessellation tess;
      BoundaryData *boundary_data=tess.create(vertices,cells);
      NumPtr<BoundaryCondition*> boundary_condition(3);
      NumPtr<Boundary*> boundary(3);
      Point<SPACEDIM> centroid((vertices[0]+vertices[1]+vertices[2])/3.);
      boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
        vertices[0]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
        vertices[2]);
      for (int i=0;i<boundary_condition.getNumber();i++) {
        boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
          Point<2> face_center=
            tess.presentCodimension1Shape()->center(vertices);
          for (int f=0;f<boundary.getNumber();f++) {
            if (abs(boundary[f]->constraint(face_center))<1.e-5) {
              codimension1_level->setBoundaryCondition(
                boundary_condition[f]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
#ifdef DEBUG
      tess.checkFaceBoundaryCondition(boundary_condition);
#endif
      tess.refineGlobal(1);

      Vector criteria(tess.numberActiveShapes());
      criteria=2.;
      criteria[3]=0.;
      tess.refine(criteria,1.);
      tess.executeCoarseningAndRefinement();
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif

      MappingL mapping;
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_L");
        for (int degree=1;degree<=10;degree++)
 //     int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_H");
        for (int degree=1;degree<=10;degree++)
//      int degree=3;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
#endif
/*
    {
//    TRACER_CALL(t,"testDofHandler constraintMatrix generateParallelpiped");
      Tessellation tess;
      Point<SPACEDIM> low(D_DECL(0.,0.,0.));
      Point<SPACEDIM> edge[SPACEDIM];
        edge[0]=Point<SPACEDIM>(D_DECL(1.,0.,0.));
        edge[1]=Point<SPACEDIM>(D_DECL(0.,1.,0.));
//      edge[2]=Point<SPACEDIM>(D_DECL(0.,0.,1.));
      int nc[SPACEDIM]={D_DECL(3,3,3)};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
        nc,set_face_ids,boundary,boundary_condition);

      Vector criteria(tess.numberActiveShapes());
      criteria=0.;
      criteria[1]=2.;
      criteria[3]=2.;
      criteria[5]=2.;
      criteria[7]=2.;
      tess.refine(criteria,1.);
      tess.executeCoarseningAndRefinement();
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif

      MappingL mapping;
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_L");
        for (int degree=1;degree<=10;degree++)
 //     int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_H");
        for (int degree=1;degree<=10;degree++)
//      int degree=3;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
#if (SPACEDIM==3)
/*
    {
      TRACER_CALL(t,"testDofHandler generateParallelpipedTetrahedra");
      Tessellation tess;
      int number_elements[SPACEDIM];
      for (int d=0;d<SPACEDIM;d++) number_elements[d]=1;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *boundary_data=generateParallelpipedTetrahedra(tess,
        low,edge,number_elements,set_face_ids,boundary,
        boundary_condition);

//
      {
        TRACER_CALL(t,"testDofHandler FE_L");
        for (int degree=1;degree<=6;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler FE_H");
        for (int degree=1;degree<=6;degree++) 
//      int degree=3;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler RTFiniteElement");
        for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) 
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          RTFiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
        for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++) {
          cout << "\tdegree = " << degree << endl;
          BDFMFiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
        for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++)
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          Nedelec1FiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
        for (int degree=1;degree<=D_CHOOSE(6,6,2);degree++)
//      int degree=2;
        {
          cout << "\tdegree = " << degree << endl;
          Nedelec2FiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//

      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
/*
    {
      TRACER_CALL(t,"testDofHandler constraintMatrix 5 tetrahedra");
      NumPtr<Point<SPACEDIM> > vertices(8);
      vertices[0]=Point<SPACEDIM>(0.,0.,0.);
      vertices[1]=Point<SPACEDIM>(1.,0.,0.);
      vertices[2]=Point<SPACEDIM>(0.,1.,0.);
      vertices[3]=Point<SPACEDIM>(0.,0.,1.);
      vertices[4]=Point<SPACEDIM>(-1.,0.,0.);
      vertices[5]=Point<SPACEDIM>(0.,-1.,0.);
      vertices[6]=Point<SPACEDIM>(0.,0.,-1.);
      vertices[7]=Point<SPACEDIM>(1.,1.,1.);
      NumPtr<CellData> cells(5);
      cells[0].setMaterialModel(0);
      cells[0].allocate(4);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=3;

      cells[1].setMaterialModel(0);
      cells[1].allocate(4);
      cells[1].vertex(0)=0;
      cells[1].vertex(1)=2;
      cells[1].vertex(2)=4;
      cells[1].vertex(3)=3;

      cells[2].setMaterialModel(0);
      cells[2].allocate(4);
      cells[2].vertex(0)=0;
      cells[2].vertex(1)=5;
      cells[2].vertex(2)=1;
      cells[2].vertex(3)=3;

      cells[3].setMaterialModel(0);
      cells[3].allocate(4);
      cells[3].vertex(0)=0;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=1;
      cells[3].vertex(3)=6;

      cells[4].setMaterialModel(0);
      cells[4].allocate(4);
      cells[4].vertex(0)=1;
      cells[4].vertex(1)=2;
      cells[4].vertex(2)=3;
      cells[4].vertex(3)=7;

      Tessellation tess;
      BoundaryData *boundary_data=tess.create(vertices,cells);
      Point<SPACEDIM> centroid((vertices[0]+vertices[1]+vertices[2]
        +vertices[3])*0.25);

      NumPtr<Boundary*> boundary(6);
      NumPtr<BoundaryCondition*> boundary_condition(6);

      boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1],vertices[5]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1],vertices[6]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[2],vertices[6]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
        vertices[2],vertices[6]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
        vertices[3],vertices[5]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
        vertices[3],vertices[4]);
      for (int i=0;i<boundary.getNumber();i++) {
        boundary_condition[i]=
          OPERATOR_NEW BoundaryCondition(boundary[i]);
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<3> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              found=true;
              break;
            }
          }
          if (!found) {
            face_element->printOn(cout);
            CHECK_TEST(found);
          }
        }
      }
      tess.synchronize(boundary_data);

      Vector criteria(tess.numberActiveShapes());
      criteria=2.;
      criteria[0]=0.;
      tess.refine(criteria,1.);
      tess.executeCoarseningAndRefinement();
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif

      MappingL mapping;
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_L");
        for (int degree=1;degree<=7;degree++)
//      int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_H");
        for (int degree=1;degree<=6;degree++)
//      int degree=3;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
/*
    {
      TRACER_CALL(t,"testDofHandler generateParallelpipedPrisms");
      int number_elements[SPACEDIM];
      for (int d=0;d<SPACEDIM;d++) number_elements[d]=1;
      for (int d=0;d<SPACEDIM;d++) {
        Tessellation tess;
        NumPtr<Boundary*> boundary;
        NumPtr<BoundaryCondition*> boundary_condition;
        BoundaryData *boundary_data=generateParallelpipedPrisms(tess,
          low,edge,number_elements,d,set_face_ids,boundary,
          boundary_condition);

//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=6;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler RTFiniteElement");
          for (int degree=0;degree<=D_CHOOSE(6,6,3);degree++)
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            RTFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
          for (int degree=0;degree<=3;degree++)
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            BDFMFiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
          for (int degree=0;degree<=1;degree++)
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec1FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
          for (int degree=1;degree<=2;degree++)
//        int degree=1;
          {
            cout << "\tdegree = " << degree << endl;
            Nedelec2FiniteElement finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//

        delete boundary_data;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
*/
/*
    {
//    TRACER_CALL(t,"testDofHandler constraintMatrix one prism");
      NumPtr<Point<SPACEDIM> > vertices(6);
      vertices[0]=Point<SPACEDIM>(0.,0.,0.);
      vertices[1]=Point<SPACEDIM>(1.,0.,0.);
      vertices[2]=Point<SPACEDIM>(0.,1.,0.);
      vertices[3]=Point<SPACEDIM>(0.,0.,1.);
      vertices[4]=Point<SPACEDIM>(1.,0.,1.);
      vertices[5]=Point<SPACEDIM>(0.,1.,1.);
      NumPtr<CellData> cells(1);
      cells[0].setMaterialModel(0);
      cells[0].allocate(6);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=3;
      cells[0].vertex(4)=4;
      cells[0].vertex(5)=5;

      Tessellation tess;
      BoundaryData *boundary_data=tess.create(vertices,cells);
      NumPtr<Boundary*> boundary(5);
      Point<SPACEDIM> centroid(1./3.,1./3.,0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1],vertices[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[3],
        vertices[4],vertices[5]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[2],vertices[3]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1],vertices[3]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
        vertices[2],vertices[4]);
      NumPtr<BoundaryCondition*> boundary_condition(5);
      for (int i=0;i<boundary.getNumber();i++) {
        boundary_condition[i]=
          OPERATOR_NEW BoundaryCondition(boundary[i]);
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              found=true;
              break;
            }
          }
          CHECK_TEST(found);
        }
      }
      tess.synchronize(boundary_data);

      tess.refineGlobal(1);

      Vector criteria(tess.numberActiveShapes());
      criteria=2.;
      criteria[3]=0.;
      criteria[4]=0.;
      criteria[5]=0.;
      criteria[6]=0.;
      tess.refine(criteria,1.);
      tess.executeCoarseningAndRefinement();
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif

      MappingL mapping;
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_L");
        for (int degree=1;degree<=6;degree++)
 //     int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_H");
        for (int degree=1;degree<=6;degree++)
//      int degree=2;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
/*
    {
      TRACER_CALL(t,"testDofHandler constraintMatrix four hexahedra");
      NumPtr<Point<SPACEDIM> > vertices(20);
      vertices[0]=Point<SPACEDIM>(0.,0.,0.);
      vertices[1]=Point<SPACEDIM>(1.,0.,0.);
      vertices[2]=Point<SPACEDIM>(0.,1.,0.);
      vertices[3]=Point<SPACEDIM>(1.,1.,0.);
      vertices[4]=Point<SPACEDIM>(0.,0.,1.);
      vertices[5]=Point<SPACEDIM>(1.,0.,1.);
      vertices[6]=Point<SPACEDIM>(0.,1.,1.);
      vertices[7]=Point<SPACEDIM>(1.,1.,1.);
      vertices[8]=Point<SPACEDIM>(0.,-1.,0.);
      vertices[9]=Point<SPACEDIM>(1.,-1.,0.);
      vertices[10]=Point<SPACEDIM>(0.,-1.,1.);
      vertices[11]=Point<SPACEDIM>(1.,-1.,1.);
      vertices[12]=Point<SPACEDIM>(-1.,0.,0.);
      vertices[13]=Point<SPACEDIM>(-1.,1.,0.);
      vertices[14]=Point<SPACEDIM>(-1.,0.,1.);
      vertices[15]=Point<SPACEDIM>(-1.,1.,1.);
      vertices[16]=Point<SPACEDIM>(0.,0.,-1.);
      vertices[17]=Point<SPACEDIM>(1.,0.,-1.);
      vertices[18]=Point<SPACEDIM>(0.,1.,-1.);
      vertices[19]=Point<SPACEDIM>(1.,1.,-1.);
      NumPtr<CellData> cells(4);

      for (int i=0;i<cells.getNumber();i++) {
        cells[i].setMaterialModel(0);
        cells[i].allocate(8);
      }

      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=3;
      cells[0].vertex(4)=4;
      cells[0].vertex(5)=5;
      cells[0].vertex(6)=6;
      cells[0].vertex(7)=7;

      cells[1].vertex(0)=8;
      cells[1].vertex(1)=9;
      cells[1].vertex(2)=0;
      cells[1].vertex(3)=1;
      cells[1].vertex(4)=10;
      cells[1].vertex(5)=11;
      cells[1].vertex(6)=4;
      cells[1].vertex(7)=5;

      cells[2].vertex(0)=12;
      cells[2].vertex(1)=0;
      cells[2].vertex(2)=13;
      cells[2].vertex(3)=2;
      cells[2].vertex(4)=14;
      cells[2].vertex(5)=4;
      cells[2].vertex(6)=15;
      cells[2].vertex(7)=6;

      cells[3].vertex(0)=16;
      cells[3].vertex(1)=17;
      cells[3].vertex(2)=18;
      cells[3].vertex(3)=19;
      cells[3].vertex(4)=0;
      cells[3].vertex(5)=1;
      cells[3].vertex(6)=2;
      cells[3].vertex(7)=3;

      Tessellation tess;
      BoundaryData *boundary_data=tess.create(vertices,cells);
      Point<SPACEDIM> centroid(0.5,0.5,0.5);

      NumPtr<Boundary*> boundary(9);
      NumPtr<BoundaryCondition*> boundary_condition(9);

      boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1],vertices[8]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1],vertices[16]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[2],vertices[16]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
        vertices[3],vertices[5]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
        vertices[3],vertices[6]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
        vertices[5],vertices[6]);
      boundary[6]=OPERATOR_NEW PlaneBoundary(centroid,vertices[8],
        vertices[9],vertices[10]);
      boundary[7]=OPERATOR_NEW PlaneBoundary(centroid,vertices[12],
        vertices[13],vertices[14]);
      boundary[8]=OPERATOR_NEW PlaneBoundary(centroid,vertices[16],
        vertices[17],vertices[18]);
      for (int i=0;i<boundary.getNumber();i++) {
        boundary_condition[i]=
          OPERATOR_NEW BoundaryCondition(boundary[i]);
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<3> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              found=true;
              break;
            }
          }
          if (!found) {
            face_element->printOn(cout);
            CHECK_TEST(found);
          }
        }
      }
      tess.synchronize(boundary_data);

      tess.refineGlobal(1);

      Vector criteria(tess.numberActiveShapes());
      criteria=2.;
      criteria[0]=0.;
      tess.refine(criteria,1.);
      tess.executeCoarseningAndRefinement();
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif

      MappingL mapping;
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_L");
        for (int degree=1;degree<=6;degree++)
 //     int degree=1;
        {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler constraintMatrix FE_H");
        for (int degree=1;degree<=D_CHOOSE(6,6,4);degree++)
//      int degree=3;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
          constraintMatrix(tess,mapping,finite_element);
        }
      }
//
      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
/*
    {
//    TRACER_CALL(t,"testDofHandler constraintMatrix one prism one hexahedron");
//    one prism and one tetrahedron, with random vertex ordering
      PermutationGenerator pg(4);
      while (pg.hasMore()) {
        const NumPtr<int> &perm=pg.getNext();
        NumPtr<Point<SPACEDIM> > vertices(7);
        vertices[4]=Point<SPACEDIM>(0.,0.,0.);
        vertices[5]=Point<SPACEDIM>(1.,0.,0.);
        vertices[6]=Point<SPACEDIM>(0.,1.,0.);
        vertices[perm[0]]=Point<SPACEDIM>(0.,0.,.5);
        vertices[perm[1]]=Point<SPACEDIM>(1.,0.,.5);
        vertices[perm[2]]=Point<SPACEDIM>(0.,1.,.5);
        vertices[perm[3]]=Point<SPACEDIM>(0.,0.,1.);
        NumPtr<CellData> cells(2);
        for (int i=0;i<2;i++) {
          cells[i].setMaterialModel(0);
        }
        cells[0].allocate(6);
        cells[0].vertex(0)=4;
        cells[0].vertex(1)=5;
        cells[0].vertex(2)=6;
        cells[0].vertex(3)=perm[0];
        cells[0].vertex(4)=perm[1];
        cells[0].vertex(5)=perm[2];
        cells[1].allocate(4);
        cells[1].vertex(0)=perm[0];
        cells[1].vertex(1)=perm[1];
        cells[1].vertex(2)=perm[2];
        cells[1].vertex(3)=perm[3];
        Tessellation tess;
        BoundaryData *boundary_data=tess.create(vertices,cells);
        Point<3> centroid((vertices[perm[0]]+vertices[perm[1]]
          +vertices[perm[2]]+vertices[perm[3]])*0.25);
        NumPtr<Boundary*> boundary(5);
        boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
          vertices[6],vertices[perm[3]]); // back
        boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
          vertices[5],vertices[perm[3]]); // left
        boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[4],vertices[5],vertices[6]); // bottom
        boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[5],vertices[perm[1]],vertices[6]); // right front
        boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,
          vertices[perm[1]],vertices[perm[2]],vertices[perm[3]]); 
          // right top
        NumPtr<BoundaryCondition*> boundary_condition(5);
        for (int i=0;i<boundary.getNumber();i++) {
          boundary_condition[i]=
            OPERATOR_NEW BoundaryCondition(boundary[i]);
        }
        for (bool doit=tess.firstCodimension1ActiveShape();doit;
        doit=tess.nextCodimension1ActiveShape()) {
          CODIMENSION1_LEVEL_CLASS *codimension1_level=
            tess.presentCodimension1Level();
          if (codimension1_level->getBoundaryCondition()!=0) {
            CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape();
            Point<SPACEDIM> center=face_element->center(vertices);
            bool found=false;
            for (int i=0;i<boundary.getNumber();i++) {
              if (abs(boundary[i]->constraint(center))<1.e-5) {
                codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
                found=true;
                break;
              }
            }
            CHECK_TEST(found);
          }
        }
        tess.synchronize(boundary_data);
        cout << "\tpermutation = " << perm[0] << " " << perm[1] << " "
             << perm[2] << " " << perm[3] << endl;
//
        {
          TRACER_CALL(t,"testDofHandler FE_L");
          for (int degree=1;degree<=6;degree++) {
            cout << "\tdegree = " << degree << endl;
            FE_L finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
//
        {
          TRACER_CALL(t,"testDofHandler FE_H");
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\tdegree = " << degree << endl;
            FE_H finite_element(degree);
//          debugDofHandler(tess,finite_element);
            checkContinuity(tess,finite_element);
          }
        }
//
        delete boundary_data; boundary_data=0;
        for (int i=0;i<boundary_condition.getNumber();i++) {
          delete boundary_condition[i]; boundary_condition[i]=0;
        }
        for (int i=0;i<boundary.getNumber();i++) {
          delete boundary[i]; boundary[i]=0;
        }
      }
    }
*/
/*
    {
      TRACER_CALL(t,"testDofHandler twoTetrahedraOnePrismOneHexahedron");
      NumPtr<Point<3> > vertices(12);
      vertices[ 0]=Point<3>(0.,0.,0.);
      vertices[ 1]=Point<3>(1.,0.,0.);
      vertices[ 2]=Point<3>(0.,1.,0.);
      vertices[ 3]=Point<3>(1.,1.,0.);
      vertices[ 4]=Point<3>(0.,0.,.5);
      vertices[ 5]=Point<3>(1.,0.,.5);
      vertices[ 6]=Point<3>(0.,1.,.5);
      vertices[ 7]=Point<3>(1.,1.,.5);
      vertices[ 8]=Point<3>(0.,.5,1.);
      vertices[ 9]=Point<3>(.3,.5,1.);
      vertices[10]=Point<3>(.6,.5,1.);
      vertices[11]=Point<3>(1.,.5,1.);

      NumPtr<CellData> cells(4);
      cells[0].allocate(8);
      for (int j=0;j<8;j++) cells[0].vertex(j)=j;
      cells[0].setMaterialModel(0);

      cells[1].allocate(6);
      cells[1].vertex(0)=4;
      cells[1].vertex(1)=6;
      cells[1].vertex(2)=9;
      cells[1].vertex(3)=5;
      cells[1].vertex(4)=7;
      cells[1].vertex(5)=10;
      cells[1].setMaterialModel(0);

      cells[2].allocate(4);
      cells[2].vertex(0)=4;
      cells[2].vertex(1)=6;
      cells[2].vertex(2)=8;
      cells[2].vertex(3)=9;
      cells[2].setMaterialModel(0);

      cells[3].allocate(4);
      cells[3].vertex(0)=5;
      cells[3].vertex(1)=7;
      cells[3].vertex(2)=10;
      cells[3].vertex(3)=11;
      cells[3].setMaterialModel(0);

      Tessellation tess;
      BoundaryData *boundary_data=tess.create(vertices,cells);
      NumPtr<Boundary*> boundary(7);
      NumPtr<BoundaryCondition*> boundary_condition(7);
      Point<3> centroid((vertices[0]+vertices[1]
                        +vertices[2]+vertices[3])*0.25);
      boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[2],vertices[4]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
        vertices[3],vertices[5]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1],vertices[4]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
        vertices[3],vertices[6]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
        vertices[1],vertices[2]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
        vertices[5],vertices[8]);
      boundary[6]=OPERATOR_NEW PlaneBoundary(centroid,vertices[6],
        vertices[7],vertices[8]);
      for (int f=0;f<7;f++) {
        boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
#ifdef DEBUG
//      cout << "\tboundary_condition[" << f << "] = "
//           << boundary_condition[f] << " boundary = " << boundary[f] 
//           << endl;
#endif
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<3> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
#ifdef DEBUG
      tess.checkFaceBoundaryCondition(boundary_condition);
#endif
      tess.synchronize(boundary_data);
//
      {
        TRACER_CALL(t,"testDofHandler FE_L");
        for (int degree=1;degree<=6;degree++) {
          cout << "\tdegree = " << degree << endl;
          FE_L finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler FE_H");
        for (int degree=1;degree<=6;degree++) 
//      int degree=3;
        {
          cout << "\tdegree = " << degree << endl;
          FE_H finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler RTFiniteElement");
        for (int degree=1;degree<=6;degree++) {
          cout << "\tdegree = " << degree << endl;
          RTFiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler BDFMFiniteElement");
        for (int degree=1;degree<=6;degree++) {
          cout << "\tdegree = " << degree << endl;
          BDFMFiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      {
        TRACER_CALL(t,"testDofHandler Nedelec1FiniteElement");
        for (int degree=1;degree<=6;degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec1FiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
      {
        TRACER_CALL(t,"testDofHandler Nedelec2FiniteElement");
        for (int degree=1;degree<=6;degree++) {
          cout << "\tdegree = " << degree << endl;
          Nedelec2FiniteElement finite_element(degree);
//        debugDofHandler(tess,finite_element);
          checkContinuity(tess,finite_element);
        }
      }
//
//
      delete boundary_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
*/
#endif
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
